/**
 * Copyright 2025 Google LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
import{a as rA,b as Ye,c as J7,d as Je,e as jQ,f as Ao}from"./chunk-EQDQRRRY.js";var PP=Je((X0e,OP)=>{"use strict";OP.exports=[{value:"#B0171F",name:"indian red"},{value:"#DC143C",css:!0,name:"crimson"},{value:"#FFB6C1",css:!0,name:"lightpink"},{value:"#FFAEB9",name:"lightpink 1"},{value:"#EEA2AD",name:"lightpink 2"},{value:"#CD8C95",name:"lightpink 3"},{value:"#8B5F65",name:"lightpink 4"},{value:"#FFC0CB",css:!0,name:"pink"},{value:"#FFB5C5",name:"pink 1"},{value:"#EEA9B8",name:"pink 2"},{value:"#CD919E",name:"pink 3"},{value:"#8B636C",name:"pink 4"},{value:"#DB7093",css:!0,name:"palevioletred"},{value:"#FF82AB",name:"palevioletred 1"},{value:"#EE799F",name:"palevioletred 2"},{value:"#CD6889",name:"palevioletred 3"},{value:"#8B475D",name:"palevioletred 4"},{value:"#FFF0F5",name:"lavenderblush 1"},{value:"#FFF0F5",css:!0,name:"lavenderblush"},{value:"#EEE0E5",name:"lavenderblush 2"},{value:"#CDC1C5",name:"lavenderblush 3"},{value:"#8B8386",name:"lavenderblush 4"},{value:"#FF3E96",name:"violetred 1"},{value:"#EE3A8C",name:"violetred 2"},{value:"#CD3278",name:"violetred 3"},{value:"#8B2252",name:"violetred 4"},{value:"#FF69B4",css:!0,name:"hotpink"},{value:"#FF6EB4",name:"hotpink 1"},{value:"#EE6AA7",name:"hotpink 2"},{value:"#CD6090",name:"hotpink 3"},{value:"#8B3A62",name:"hotpink 4"},{value:"#872657",name:"raspberry"},{value:"#FF1493",name:"deeppink 1"},{value:"#FF1493",css:!0,name:"deeppink"},{value:"#EE1289",name:"deeppink 2"},{value:"#CD1076",name:"deeppink 3"},{value:"#8B0A50",name:"deeppink 4"},{value:"#FF34B3",name:"maroon 1"},{value:"#EE30A7",name:"maroon 2"},{value:"#CD2990",name:"maroon 3"},{value:"#8B1C62",name:"maroon 4"},{value:"#C71585",css:!0,name:"mediumvioletred"},{value:"#D02090",name:"violetred"},{value:"#DA70D6",css:!0,name:"orchid"},{value:"#FF83FA",name:"orchid 1"},{value:"#EE7AE9",name:"orchid 2"},{value:"#CD69C9",name:"orchid 3"},{value:"#8B4789",name:"orchid 4"},{value:"#D8BFD8",css:!0,name:"thistle"},{value:"#FFE1FF",name:"thistle 1"},{value:"#EED2EE",name:"thistle 2"},{value:"#CDB5CD",name:"thistle 3"},{value:"#8B7B8B",name:"thistle 4"},{value:"#FFBBFF",name:"plum 1"},{value:"#EEAEEE",name:"plum 2"},{value:"#CD96CD",name:"plum 3"},{value:"#8B668B",name:"plum 4"},{value:"#DDA0DD",css:!0,name:"plum"},{value:"#EE82EE",css:!0,name:"violet"},{value:"#FF00FF",vga:!0,name:"magenta"},{value:"#FF00FF",vga:!0,css:!0,name:"fuchsia"},{value:"#EE00EE",name:"magenta 2"},{value:"#CD00CD",name:"magenta 3"},{value:"#8B008B",name:"magenta 4"},{value:"#8B008B",css:!0,name:"darkmagenta"},{value:"#800080",vga:!0,css:!0,name:"purple"},{value:"#BA55D3",css:!0,name:"mediumorchid"},{value:"#E066FF",name:"mediumorchid 1"},{value:"#D15FEE",name:"mediumorchid 2"},{value:"#B452CD",name:"mediumorchid 3"},{value:"#7A378B",name:"mediumorchid 4"},{value:"#9400D3",css:!0,name:"darkviolet"},{value:"#9932CC",css:!0,name:"darkorchid"},{value:"#BF3EFF",name:"darkorchid 1"},{value:"#B23AEE",name:"darkorchid 2"},{value:"#9A32CD",name:"darkorchid 3"},{value:"#68228B",name:"darkorchid 4"},{value:"#4B0082",css:!0,name:"indigo"},{value:"#8A2BE2",css:!0,name:"blueviolet"},{value:"#9B30FF",name:"purple 1"},{value:"#912CEE",name:"purple 2"},{value:"#7D26CD",name:"purple 3"},{value:"#551A8B",name:"purple 4"},{value:"#9370DB",css:!0,name:"mediumpurple"},{value:"#AB82FF",name:"mediumpurple 1"},{value:"#9F79EE",name:"mediumpurple 2"},{value:"#8968CD",name:"mediumpurple 3"},{value:"#5D478B",name:"mediumpurple 4"},{value:"#483D8B",css:!0,name:"darkslateblue"},{value:"#8470FF",name:"lightslateblue"},{value:"#7B68EE",css:!0,name:"mediumslateblue"},{value:"#6A5ACD",css:!0,name:"slateblue"},{value:"#836FFF",name:"slateblue 1"},{value:"#7A67EE",name:"slateblue 2"},{value:"#6959CD",name:"slateblue 3"},{value:"#473C8B",name:"slateblue 4"},{value:"#F8F8FF",css:!0,name:"ghostwhite"},{value:"#E6E6FA",css:!0,name:"lavender"},{value:"#0000FF",vga:!0,css:!0,name:"blue"},{value:"#0000EE",name:"blue 2"},{value:"#0000CD",name:"blue 3"},{value:"#0000CD",css:!0,name:"mediumblue"},{value:"#00008B",name:"blue 4"},{value:"#00008B",css:!0,name:"darkblue"},{value:"#000080",vga:!0,css:!0,name:"navy"},{value:"#191970",css:!0,name:"midnightblue"},{value:"#3D59AB",name:"cobalt"},{value:"#4169E1",css:!0,name:"royalblue"},{value:"#4876FF",name:"royalblue 1"},{value:"#436EEE",name:"royalblue 2"},{value:"#3A5FCD",name:"royalblue 3"},{value:"#27408B",name:"royalblue 4"},{value:"#6495ED",css:!0,name:"cornflowerblue"},{value:"#B0C4DE",css:!0,name:"lightsteelblue"},{value:"#CAE1FF",name:"lightsteelblue 1"},{value:"#BCD2EE",name:"lightsteelblue 2"},{value:"#A2B5CD",name:"lightsteelblue 3"},{value:"#6E7B8B",name:"lightsteelblue 4"},{value:"#778899",css:!0,name:"lightslategray"},{value:"#708090",css:!0,name:"slategray"},{value:"#C6E2FF",name:"slategray 1"},{value:"#B9D3EE",name:"slategray 2"},{value:"#9FB6CD",name:"slategray 3"},{value:"#6C7B8B",name:"slategray 4"},{value:"#1E90FF",name:"dodgerblue 1"},{value:"#1E90FF",css:!0,name:"dodgerblue"},{value:"#1C86EE",name:"dodgerblue 2"},{value:"#1874CD",name:"dodgerblue 3"},{value:"#104E8B",name:"dodgerblue 4"},{value:"#F0F8FF",css:!0,name:"aliceblue"},{value:"#4682B4",css:!0,name:"steelblue"},{value:"#63B8FF",name:"steelblue 1"},{value:"#5CACEE",name:"steelblue 2"},{value:"#4F94CD",name:"steelblue 3"},{value:"#36648B",name:"steelblue 4"},{value:"#87CEFA",css:!0,name:"lightskyblue"},{value:"#B0E2FF",name:"lightskyblue 1"},{value:"#A4D3EE",name:"lightskyblue 2"},{value:"#8DB6CD",name:"lightskyblue 3"},{value:"#607B8B",name:"lightskyblue 4"},{value:"#87CEFF",name:"skyblue 1"},{value:"#7EC0EE",name:"skyblue 2"},{value:"#6CA6CD",name:"skyblue 3"},{value:"#4A708B",name:"skyblue 4"},{value:"#87CEEB",css:!0,name:"skyblue"},{value:"#00BFFF",name:"deepskyblue 1"},{value:"#00BFFF",css:!0,name:"deepskyblue"},{value:"#00B2EE",name:"deepskyblue 2"},{value:"#009ACD",name:"deepskyblue 3"},{value:"#00688B",name:"deepskyblue 4"},{value:"#33A1C9",name:"peacock"},{value:"#ADD8E6",css:!0,name:"lightblue"},{value:"#BFEFFF",name:"lightblue 1"},{value:"#B2DFEE",name:"lightblue 2"},{value:"#9AC0CD",name:"lightblue 3"},{value:"#68838B",name:"lightblue 4"},{value:"#B0E0E6",css:!0,name:"powderblue"},{value:"#98F5FF",name:"cadetblue 1"},{value:"#8EE5EE",name:"cadetblue 2"},{value:"#7AC5CD",name:"cadetblue 3"},{value:"#53868B",name:"cadetblue 4"},{value:"#00F5FF",name:"turquoise 1"},{value:"#00E5EE",name:"turquoise 2"},{value:"#00C5CD",name:"turquoise 3"},{value:"#00868B",name:"turquoise 4"},{value:"#5F9EA0",css:!0,name:"cadetblue"},{value:"#00CED1",css:!0,name:"darkturquoise"},{value:"#F0FFFF",name:"azure 1"},{value:"#F0FFFF",css:!0,name:"azure"},{value:"#E0EEEE",name:"azure 2"},{value:"#C1CDCD",name:"azure 3"},{value:"#838B8B",name:"azure 4"},{value:"#E0FFFF",name:"lightcyan 1"},{value:"#E0FFFF",css:!0,name:"lightcyan"},{value:"#D1EEEE",name:"lightcyan 2"},{value:"#B4CDCD",name:"lightcyan 3"},{value:"#7A8B8B",name:"lightcyan 4"},{value:"#BBFFFF",name:"paleturquoise 1"},{value:"#AEEEEE",name:"paleturquoise 2"},{value:"#AEEEEE",css:!0,name:"paleturquoise"},{value:"#96CDCD",name:"paleturquoise 3"},{value:"#668B8B",name:"paleturquoise 4"},{value:"#2F4F4F",css:!0,name:"darkslategray"},{value:"#97FFFF",name:"darkslategray 1"},{value:"#8DEEEE",name:"darkslategray 2"},{value:"#79CDCD",name:"darkslategray 3"},{value:"#528B8B",name:"darkslategray 4"},{value:"#00FFFF",name:"cyan"},{value:"#00FFFF",css:!0,name:"aqua"},{value:"#00EEEE",name:"cyan 2"},{value:"#00CDCD",name:"cyan 3"},{value:"#008B8B",name:"cyan 4"},{value:"#008B8B",css:!0,name:"darkcyan"},{value:"#008080",vga:!0,css:!0,name:"teal"},{value:"#48D1CC",css:!0,name:"mediumturquoise"},{value:"#20B2AA",css:!0,name:"lightseagreen"},{value:"#03A89E",name:"manganeseblue"},{value:"#40E0D0",css:!0,name:"turquoise"},{value:"#808A87",name:"coldgrey"},{value:"#00C78C",name:"turquoiseblue"},{value:"#7FFFD4",name:"aquamarine 1"},{value:"#7FFFD4",css:!0,name:"aquamarine"},{value:"#76EEC6",name:"aquamarine 2"},{value:"#66CDAA",name:"aquamarine 3"},{value:"#66CDAA",css:!0,name:"mediumaquamarine"},{value:"#458B74",name:"aquamarine 4"},{value:"#00FA9A",css:!0,name:"mediumspringgreen"},{value:"#F5FFFA",css:!0,name:"mintcream"},{value:"#00FF7F",css:!0,name:"springgreen"},{value:"#00EE76",name:"springgreen 1"},{value:"#00CD66",name:"springgreen 2"},{value:"#008B45",name:"springgreen 3"},{value:"#3CB371",css:!0,name:"mediumseagreen"},{value:"#54FF9F",name:"seagreen 1"},{value:"#4EEE94",name:"seagreen 2"},{value:"#43CD80",name:"seagreen 3"},{value:"#2E8B57",name:"seagreen 4"},{value:"#2E8B57",css:!0,name:"seagreen"},{value:"#00C957",name:"emeraldgreen"},{value:"#BDFCC9",name:"mint"},{value:"#3D9140",name:"cobaltgreen"},{value:"#F0FFF0",name:"honeydew 1"},{value:"#F0FFF0",css:!0,name:"honeydew"},{value:"#E0EEE0",name:"honeydew 2"},{value:"#C1CDC1",name:"honeydew 3"},{value:"#838B83",name:"honeydew 4"},{value:"#8FBC8F",css:!0,name:"darkseagreen"},{value:"#C1FFC1",name:"darkseagreen 1"},{value:"#B4EEB4",name:"darkseagreen 2"},{value:"#9BCD9B",name:"darkseagreen 3"},{value:"#698B69",name:"darkseagreen 4"},{value:"#98FB98",css:!0,name:"palegreen"},{value:"#9AFF9A",name:"palegreen 1"},{value:"#90EE90",name:"palegreen 2"},{value:"#90EE90",css:!0,name:"lightgreen"},{value:"#7CCD7C",name:"palegreen 3"},{value:"#548B54",name:"palegreen 4"},{value:"#32CD32",css:!0,name:"limegreen"},{value:"#228B22",css:!0,name:"forestgreen"},{value:"#00FF00",vga:!0,name:"green 1"},{value:"#00FF00",vga:!0,css:!0,name:"lime"},{value:"#00EE00",name:"green 2"},{value:"#00CD00",name:"green 3"},{value:"#008B00",name:"green 4"},{value:"#008000",vga:!0,css:!0,name:"green"},{value:"#006400",css:!0,name:"darkgreen"},{value:"#308014",name:"sapgreen"},{value:"#7CFC00",css:!0,name:"lawngreen"},{value:"#7FFF00",name:"chartreuse 1"},{value:"#7FFF00",css:!0,name:"chartreuse"},{value:"#76EE00",name:"chartreuse 2"},{value:"#66CD00",name:"chartreuse 3"},{value:"#458B00",name:"chartreuse 4"},{value:"#ADFF2F",css:!0,name:"greenyellow"},{value:"#CAFF70",name:"darkolivegreen 1"},{value:"#BCEE68",name:"darkolivegreen 2"},{value:"#A2CD5A",name:"darkolivegreen 3"},{value:"#6E8B3D",name:"darkolivegreen 4"},{value:"#556B2F",css:!0,name:"darkolivegreen"},{value:"#6B8E23",css:!0,name:"olivedrab"},{value:"#C0FF3E",name:"olivedrab 1"},{value:"#B3EE3A",name:"olivedrab 2"},{value:"#9ACD32",name:"olivedrab 3"},{value:"#9ACD32",css:!0,name:"yellowgreen"},{value:"#698B22",name:"olivedrab 4"},{value:"#FFFFF0",name:"ivory 1"},{value:"#FFFFF0",css:!0,name:"ivory"},{value:"#EEEEE0",name:"ivory 2"},{value:"#CDCDC1",name:"ivory 3"},{value:"#8B8B83",name:"ivory 4"},{value:"#F5F5DC",css:!0,name:"beige"},{value:"#FFFFE0",name:"lightyellow 1"},{value:"#FFFFE0",css:!0,name:"lightyellow"},{value:"#EEEED1",name:"lightyellow 2"},{value:"#CDCDB4",name:"lightyellow 3"},{value:"#8B8B7A",name:"lightyellow 4"},{value:"#FAFAD2",css:!0,name:"lightgoldenrodyellow"},{value:"#FFFF00",vga:!0,name:"yellow 1"},{value:"#FFFF00",vga:!0,css:!0,name:"yellow"},{value:"#EEEE00",name:"yellow 2"},{value:"#CDCD00",name:"yellow 3"},{value:"#8B8B00",name:"yellow 4"},{value:"#808069",name:"warmgrey"},{value:"#808000",vga:!0,css:!0,name:"olive"},{value:"#BDB76B",css:!0,name:"darkkhaki"},{value:"#FFF68F",name:"khaki 1"},{value:"#EEE685",name:"khaki 2"},{value:"#CDC673",name:"khaki 3"},{value:"#8B864E",name:"khaki 4"},{value:"#F0E68C",css:!0,name:"khaki"},{value:"#EEE8AA",css:!0,name:"palegoldenrod"},{value:"#FFFACD",name:"lemonchiffon 1"},{value:"#FFFACD",css:!0,name:"lemonchiffon"},{value:"#EEE9BF",name:"lemonchiffon 2"},{value:"#CDC9A5",name:"lemonchiffon 3"},{value:"#8B8970",name:"lemonchiffon 4"},{value:"#FFEC8B",name:"lightgoldenrod 1"},{value:"#EEDC82",name:"lightgoldenrod 2"},{value:"#CDBE70",name:"lightgoldenrod 3"},{value:"#8B814C",name:"lightgoldenrod 4"},{value:"#E3CF57",name:"banana"},{value:"#FFD700",name:"gold 1"},{value:"#FFD700",css:!0,name:"gold"},{value:"#EEC900",name:"gold 2"},{value:"#CDAD00",name:"gold 3"},{value:"#8B7500",name:"gold 4"},{value:"#FFF8DC",name:"cornsilk 1"},{value:"#FFF8DC",css:!0,name:"cornsilk"},{value:"#EEE8CD",name:"cornsilk 2"},{value:"#CDC8B1",name:"cornsilk 3"},{value:"#8B8878",name:"cornsilk 4"},{value:"#DAA520",css:!0,name:"goldenrod"},{value:"#FFC125",name:"goldenrod 1"},{value:"#EEB422",name:"goldenrod 2"},{value:"#CD9B1D",name:"goldenrod 3"},{value:"#8B6914",name:"goldenrod 4"},{value:"#B8860B",css:!0,name:"darkgoldenrod"},{value:"#FFB90F",name:"darkgoldenrod 1"},{value:"#EEAD0E",name:"darkgoldenrod 2"},{value:"#CD950C",name:"darkgoldenrod 3"},{value:"#8B6508",name:"darkgoldenrod 4"},{value:"#FFA500",name:"orange 1"},{value:"#FF8000",css:!0,name:"orange"},{value:"#EE9A00",name:"orange 2"},{value:"#CD8500",name:"orange 3"},{value:"#8B5A00",name:"orange 4"},{value:"#FFFAF0",css:!0,name:"floralwhite"},{value:"#FDF5E6",css:!0,name:"oldlace"},{value:"#F5DEB3",css:!0,name:"wheat"},{value:"#FFE7BA",name:"wheat 1"},{value:"#EED8AE",name:"wheat 2"},{value:"#CDBA96",name:"wheat 3"},{value:"#8B7E66",name:"wheat 4"},{value:"#FFE4B5",css:!0,name:"moccasin"},{value:"#FFEFD5",css:!0,name:"papayawhip"},{value:"#FFEBCD",css:!0,name:"blanchedalmond"},{value:"#FFDEAD",name:"navajowhite 1"},{value:"#FFDEAD",css:!0,name:"navajowhite"},{value:"#EECFA1",name:"navajowhite 2"},{value:"#CDB38B",name:"navajowhite 3"},{value:"#8B795E",name:"navajowhite 4"},{value:"#FCE6C9",name:"eggshell"},{value:"#D2B48C",css:!0,name:"tan"},{value:"#9C661F",name:"brick"},{value:"#FF9912",name:"cadmiumyellow"},{value:"#FAEBD7",css:!0,name:"antiquewhite"},{value:"#FFEFDB",name:"antiquewhite 1"},{value:"#EEDFCC",name:"antiquewhite 2"},{value:"#CDC0B0",name:"antiquewhite 3"},{value:"#8B8378",name:"antiquewhite 4"},{value:"#DEB887",css:!0,name:"burlywood"},{value:"#FFD39B",name:"burlywood 1"},{value:"#EEC591",name:"burlywood 2"},{value:"#CDAA7D",name:"burlywood 3"},{value:"#8B7355",name:"burlywood 4"},{value:"#FFE4C4",name:"bisque 1"},{value:"#FFE4C4",css:!0,name:"bisque"},{value:"#EED5B7",name:"bisque 2"},{value:"#CDB79E",name:"bisque 3"},{value:"#8B7D6B",name:"bisque 4"},{value:"#E3A869",name:"melon"},{value:"#ED9121",name:"carrot"},{value:"#FF8C00",css:!0,name:"darkorange"},{value:"#FF7F00",name:"darkorange 1"},{value:"#EE7600",name:"darkorange 2"},{value:"#CD6600",name:"darkorange 3"},{value:"#8B4500",name:"darkorange 4"},{value:"#FFA54F",name:"tan 1"},{value:"#EE9A49",name:"tan 2"},{value:"#CD853F",name:"tan 3"},{value:"#CD853F",css:!0,name:"peru"},{value:"#8B5A2B",name:"tan 4"},{value:"#FAF0E6",css:!0,name:"linen"},{value:"#FFDAB9",name:"peachpuff 1"},{value:"#FFDAB9",css:!0,name:"peachpuff"},{value:"#EECBAD",name:"peachpuff 2"},{value:"#CDAF95",name:"peachpuff 3"},{value:"#8B7765",name:"peachpuff 4"},{value:"#FFF5EE",name:"seashell 1"},{value:"#FFF5EE",css:!0,name:"seashell"},{value:"#EEE5DE",name:"seashell 2"},{value:"#CDC5BF",name:"seashell 3"},{value:"#8B8682",name:"seashell 4"},{value:"#F4A460",css:!0,name:"sandybrown"},{value:"#C76114",name:"rawsienna"},{value:"#D2691E",css:!0,name:"chocolate"},{value:"#FF7F24",name:"chocolate 1"},{value:"#EE7621",name:"chocolate 2"},{value:"#CD661D",name:"chocolate 3"},{value:"#8B4513",name:"chocolate 4"},{value:"#8B4513",css:!0,name:"saddlebrown"},{value:"#292421",name:"ivoryblack"},{value:"#FF7D40",name:"flesh"},{value:"#FF6103",name:"cadmiumorange"},{value:"#8A360F",name:"burntsienna"},{value:"#A0522D",css:!0,name:"sienna"},{value:"#FF8247",name:"sienna 1"},{value:"#EE7942",name:"sienna 2"},{value:"#CD6839",name:"sienna 3"},{value:"#8B4726",name:"sienna 4"},{value:"#FFA07A",name:"lightsalmon 1"},{value:"#FFA07A",css:!0,name:"lightsalmon"},{value:"#EE9572",name:"lightsalmon 2"},{value:"#CD8162",name:"lightsalmon 3"},{value:"#8B5742",name:"lightsalmon 4"},{value:"#FF7F50",css:!0,name:"coral"},{value:"#FF4500",name:"orangered 1"},{value:"#FF4500",css:!0,name:"orangered"},{value:"#EE4000",name:"orangered 2"},{value:"#CD3700",name:"orangered 3"},{value:"#8B2500",name:"orangered 4"},{value:"#5E2612",name:"sepia"},{value:"#E9967A",css:!0,name:"darksalmon"},{value:"#FF8C69",name:"salmon 1"},{value:"#EE8262",name:"salmon 2"},{value:"#CD7054",name:"salmon 3"},{value:"#8B4C39",name:"salmon 4"},{value:"#FF7256",name:"coral 1"},{value:"#EE6A50",name:"coral 2"},{value:"#CD5B45",name:"coral 3"},{value:"#8B3E2F",name:"coral 4"},{value:"#8A3324",name:"burntumber"},{value:"#FF6347",name:"tomato 1"},{value:"#FF6347",css:!0,name:"tomato"},{value:"#EE5C42",name:"tomato 2"},{value:"#CD4F39",name:"tomato 3"},{value:"#8B3626",name:"tomato 4"},{value:"#FA8072",css:!0,name:"salmon"},{value:"#FFE4E1",name:"mistyrose 1"},{value:"#FFE4E1",css:!0,name:"mistyrose"},{value:"#EED5D2",name:"mistyrose 2"},{value:"#CDB7B5",name:"mistyrose 3"},{value:"#8B7D7B",name:"mistyrose 4"},{value:"#FFFAFA",name:"snow 1"},{value:"#FFFAFA",css:!0,name:"snow"},{value:"#EEE9E9",name:"snow 2"},{value:"#CDC9C9",name:"snow 3"},{value:"#8B8989",name:"snow 4"},{value:"#BC8F8F",css:!0,name:"rosybrown"},{value:"#FFC1C1",name:"rosybrown 1"},{value:"#EEB4B4",name:"rosybrown 2"},{value:"#CD9B9B",name:"rosybrown 3"},{value:"#8B6969",name:"rosybrown 4"},{value:"#F08080",css:!0,name:"lightcoral"},{value:"#CD5C5C",css:!0,name:"indianred"},{value:"#FF6A6A",name:"indianred 1"},{value:"#EE6363",name:"indianred 2"},{value:"#8B3A3A",name:"indianred 4"},{value:"#CD5555",name:"indianred 3"},{value:"#A52A2A",css:!0,name:"brown"},{value:"#FF4040",name:"brown 1"},{value:"#EE3B3B",name:"brown 2"},{value:"#CD3333",name:"brown 3"},{value:"#8B2323",name:"brown 4"},{value:"#B22222",css:!0,name:"firebrick"},{value:"#FF3030",name:"firebrick 1"},{value:"#EE2C2C",name:"firebrick 2"},{value:"#CD2626",name:"firebrick 3"},{value:"#8B1A1A",name:"firebrick 4"},{value:"#FF0000",vga:!0,name:"red 1"},{value:"#FF0000",vga:!0,css:!0,name:"red"},{value:"#EE0000",name:"red 2"},{value:"#CD0000",name:"red 3"},{value:"#8B0000",name:"red 4"},{value:"#8B0000",css:!0,name:"darkred"},{value:"#800000",vga:!0,css:!0,name:"maroon"},{value:"#8E388E",name:"sgi beet"},{value:"#7171C6",name:"sgi slateblue"},{value:"#7D9EC0",name:"sgi lightblue"},{value:"#388E8E",name:"sgi teal"},{value:"#71C671",name:"sgi chartreuse"},{value:"#8E8E38",name:"sgi olivedrab"},{value:"#C5C1AA",name:"sgi brightgray"},{value:"#C67171",name:"sgi salmon"},{value:"#555555",name:"sgi darkgray"},{value:"#1E1E1E",name:"sgi gray 12"},{value:"#282828",name:"sgi gray 16"},{value:"#515151",name:"sgi gray 32"},{value:"#5B5B5B",name:"sgi gray 36"},{value:"#848484",name:"sgi gray 52"},{value:"#8E8E8E",name:"sgi gray 56"},{value:"#AAAAAA",name:"sgi lightgray"},{value:"#B7B7B7",name:"sgi gray 72"},{value:"#C1C1C1",name:"sgi gray 76"},{value:"#EAEAEA",name:"sgi gray 92"},{value:"#F4F4F4",name:"sgi gray 96"},{value:"#FFFFFF",vga:!0,css:!0,name:"white"},{value:"#F5F5F5",name:"white smoke"},{value:"#F5F5F5",name:"gray 96"},{value:"#DCDCDC",css:!0,name:"gainsboro"},{value:"#D3D3D3",css:!0,name:"lightgrey"},{value:"#C0C0C0",vga:!0,css:!0,name:"silver"},{value:"#A9A9A9",css:!0,name:"darkgray"},{value:"#808080",vga:!0,css:!0,name:"gray"},{value:"#696969",css:!0,name:"dimgray"},{value:"#696969",name:"gray 42"},{value:"#000000",vga:!0,css:!0,name:"black"},{value:"#FCFCFC",name:"gray 99"},{value:"#FAFAFA",name:"gray 98"},{value:"#F7F7F7",name:"gray 97"},{value:"#F2F2F2",name:"gray 95"},{value:"#F0F0F0",name:"gray 94"},{value:"#EDEDED",name:"gray 93"},{value:"#EBEBEB",name:"gray 92"},{value:"#E8E8E8",name:"gray 91"},{value:"#E5E5E5",name:"gray 90"},{value:"#E3E3E3",name:"gray 89"},{value:"#E0E0E0",name:"gray 88"},{value:"#DEDEDE",name:"gray 87"},{value:"#DBDBDB",name:"gray 86"},{value:"#D9D9D9",name:"gray 85"},{value:"#D6D6D6",name:"gray 84"},{value:"#D4D4D4",name:"gray 83"},{value:"#D1D1D1",name:"gray 82"},{value:"#CFCFCF",name:"gray 81"},{value:"#CCCCCC",name:"gray 80"},{value:"#C9C9C9",name:"gray 79"},{value:"#C7C7C7",name:"gray 78"},{value:"#C4C4C4",name:"gray 77"},{value:"#C2C2C2",name:"gray 76"},{value:"#BFBFBF",name:"gray 75"},{value:"#BDBDBD",name:"gray 74"},{value:"#BABABA",name:"gray 73"},{value:"#B8B8B8",name:"gray 72"},{value:"#B5B5B5",name:"gray 71"},{value:"#B3B3B3",name:"gray 70"},{value:"#B0B0B0",name:"gray 69"},{value:"#ADADAD",name:"gray 68"},{value:"#ABABAB",name:"gray 67"},{value:"#A8A8A8",name:"gray 66"},{value:"#A6A6A6",name:"gray 65"},{value:"#A3A3A3",name:"gray 64"},{value:"#A1A1A1",name:"gray 63"},{value:"#9E9E9E",name:"gray 62"},{value:"#9C9C9C",name:"gray 61"},{value:"#999999",name:"gray 60"},{value:"#969696",name:"gray 59"},{value:"#949494",name:"gray 58"},{value:"#919191",name:"gray 57"},{value:"#8F8F8F",name:"gray 56"},{value:"#8C8C8C",name:"gray 55"},{value:"#8A8A8A",name:"gray 54"},{value:"#878787",name:"gray 53"},{value:"#858585",name:"gray 52"},{value:"#828282",name:"gray 51"},{value:"#7F7F7F",name:"gray 50"},{value:"#7D7D7D",name:"gray 49"},{value:"#7A7A7A",name:"gray 48"},{value:"#787878",name:"gray 47"},{value:"#757575",name:"gray 46"},{value:"#737373",name:"gray 45"},{value:"#707070",name:"gray 44"},{value:"#6E6E6E",name:"gray 43"},{value:"#666666",name:"gray 40"},{value:"#636363",name:"gray 39"},{value:"#616161",name:"gray 38"},{value:"#5E5E5E",name:"gray 37"},{value:"#5C5C5C",name:"gray 36"},{value:"#595959",name:"gray 35"},{value:"#575757",name:"gray 34"},{value:"#545454",name:"gray 33"},{value:"#525252",name:"gray 32"},{value:"#4F4F4F",name:"gray 31"},{value:"#4D4D4D",name:"gray 30"},{value:"#4A4A4A",name:"gray 29"},{value:"#474747",name:"gray 28"},{value:"#454545",name:"gray 27"},{value:"#424242",name:"gray 26"},{value:"#404040",name:"gray 25"},{value:"#3D3D3D",name:"gray 24"},{value:"#3B3B3B",name:"gray 23"},{value:"#383838",name:"gray 22"},{value:"#363636",name:"gray 21"},{value:"#333333",name:"gray 20"},{value:"#303030",name:"gray 19"},{value:"#2E2E2E",name:"gray 18"},{value:"#2B2B2B",name:"gray 17"},{value:"#292929",name:"gray 16"},{value:"#262626",name:"gray 15"},{value:"#242424",name:"gray 14"},{value:"#212121",name:"gray 13"},{value:"#1F1F1F",name:"gray 12"},{value:"#1C1C1C",name:"gray 11"},{value:"#1A1A1A",name:"gray 10"},{value:"#171717",name:"gray 9"},{value:"#141414",name:"gray 8"},{value:"#121212",name:"gray 7"},{value:"#0F0F0F",name:"gray 6"},{value:"#0D0D0D",name:"gray 5"},{value:"#0A0A0A",name:"gray 4"},{value:"#080808",name:"gray 3"},{value:"#050505",name:"gray 2"},{value:"#030303",name:"gray 1"},{value:"#F5F5F5",css:!0,name:"whitesmoke"}]});var VP=Je(($0e,T2)=>{"use strict";var p8=PP(),jP=p8.filter(function(t){return!!t.css}),qP=p8.filter(function(t){return!!t.vga});T2.exports=function(t){var e=T2.exports.get(t);return e&&e.value};T2.exports.get=function(t){return t=t||"",t=t.trim().toLowerCase(),p8.filter(function(e){return e.name.toLowerCase()===t}).pop()};T2.exports.all=T2.exports.get.all=function(){return p8};T2.exports.get.css=function(t){return t?(t=t||"",t=t.trim().toLowerCase(),jP.filter(function(e){return e.name.toLowerCase()===t}).pop()):jP};T2.exports.get.vga=function(t){return t?(t=t||"",t=t.trim().toLowerCase(),qP.filter(function(e){return e.name.toLowerCase()===t}).pop()):qP}});var Qj=Je((A2e,Ej)=>{"use strict";var N3A=1/0,_3A="[object Symbol]",G3A=/[^\x00-\x2f\x3a-\x40\x5b-\x60\x7b-\x7f]+/g,ij="\\ud800-\\udfff",U3A="\\u0300-\\u036f\\ufe20-\\ufe23",K3A="\\u20d0-\\u20f0",nj="\\u2700-\\u27bf",oj="a-z\\xdf-\\xf6\\xf8-\\xff",Y3A="\\xac\\xb1\\xd7\\xf7",J3A="\\x00-\\x2f\\x3a-\\x40\\x5b-\\x60\\x7b-\\xbf",T3A="\\u2000-\\u206f",H3A=" \\t\\x0b\\f\\xa0\\ufeff\\n\\r\\u2028\\u2029\\u1680\\u180e\\u2000\\u2001\\u2002\\u2003\\u2004\\u2005\\u2006\\u2007\\u2008\\u2009\\u200a\\u202f\\u205f\\u3000",rj="A-Z\\xc0-\\xd6\\xd8-\\xde",z3A="\\ufe0e\\ufe0f",sj=Y3A+J3A+T3A+H3A,aj="['\u2019]",ZP="["+sj+"]",O3A="["+U3A+K3A+"]",cj="\\d+",P3A="["+nj+"]",lj="["+oj+"]",gj="[^"+ij+sj+cj+nj+oj+rj+"]",j3A="\\ud83c[\\udffb-\\udfff]",q3A="(?:"+O3A+"|"+j3A+")",V3A="[^"+ij+"]",Ij="(?:\\ud83c[\\udde6-\\uddff]){2}",Cj="[\\ud800-\\udbff][\\udc00-\\udfff]",UB="["+rj+"]",Z3A="\\u200d",WP="(?:"+lj+"|"+gj+")",W3A="(?:"+UB+"|"+gj+")",XP="(?:"+aj+"(?:d|ll|m|re|s|t|ve))?",$P="(?:"+aj+"(?:D|LL|M|RE|S|T|VE))?",dj=q3A+"?",Bj="["+z3A+"]?",X3A="(?:"+Z3A+"(?:"+[V3A,Ij,Cj].join("|")+")"+Bj+dj+")*",$3A=Bj+dj+X3A,AfA="(?:"+[P3A,Ij,Cj].join("|")+")"+$3A,efA=RegExp([UB+"?"+lj+"+"+XP+"(?="+[ZP,UB,"$"].join("|")+")",W3A+"+"+$P+"(?="+[ZP,UB+WP,"$"].join("|")+")",UB+"?"+WP+"+"+XP,UB+"+"+$P,cj,AfA].join("|"),"g"),tfA=/[a-z][A-Z]|[A-Z]{2,}[a-z]|[0-9][a-zA-Z]|[a-zA-Z][0-9]|[^a-zA-Z0-9 ]/,ifA=typeof global=="object"&&global&&global.Object===Object&&global,nfA=typeof self=="object"&&self&&self.Object===Object&&self,ofA=ifA||nfA||Function("return this")();function rfA(t){return t.match(G3A)||[]}function sfA(t){return tfA.test(t)}function afA(t){return t.match(efA)||[]}var cfA=Object.prototype,lfA=cfA.toString,Aj=ofA.Symbol,ej=Aj?Aj.prototype:void 0,tj=ej?ej.toString:void 0;function gfA(t){if(typeof t=="string")return t;if(CfA(t))return tj?tj.call(t):"";var e=t+"";return e=="0"&&1/t==-N3A?"-0":e}function IfA(t){return!!t&&typeof t=="object"}function CfA(t){return typeof t=="symbol"||IfA(t)&&lfA.call(t)==_3A}function dfA(t){return t==null?"":gfA(t)}function BfA(t,e,A){return t=dfA(t),e=A?void 0:e,e===void 0?sfA(t)?afA(t):rfA(t):t.match(e)||[]}Ej.exports=BfA});var Lj=Je((e2e,xj)=>{"use strict";var EfA=1/0,QfA="[object Symbol]",hfA=/^\s+/,Vk="\\ud800-\\udfff",pj="\\u0300-\\u036f\\ufe20-\\ufe23",wj="\\u20d0-\\u20f0",Dj="\\ufe0e\\ufe0f",ufA="["+Vk+"]",jk="["+pj+wj+"]",qk="\\ud83c[\\udffb-\\udfff]",ffA="(?:"+jk+"|"+qk+")",yj="[^"+Vk+"]",vj="(?:\\ud83c[\\udde6-\\uddff]){2}",bj="[\\ud800-\\udbff][\\udc00-\\udfff]",Mj="\\u200d",kj=ffA+"?",Sj="["+Dj+"]?",mfA="(?:"+Mj+"(?:"+[yj,vj,bj].join("|")+")"+Sj+kj+")*",pfA=Sj+kj+mfA,wfA="(?:"+[yj+jk+"?",jk,vj,bj,ufA].join("|")+")",DfA=RegExp(qk+"(?="+qk+")|"+wfA+pfA,"g"),yfA=RegExp("["+Mj+Vk+pj+wj+Dj+"]"),vfA=typeof global=="object"&&global&&global.Object===Object&&global,bfA=typeof self=="object"&&self&&self.Object===Object&&self,MfA=vfA||bfA||Function("return this")();function kfA(t){return t.split("")}function SfA(t,e,A,i){for(var n=t.length,o=A+(i?1:-1);i?o--:++o<n;)if(e(t[o],o,t))return o;return-1}function RfA(t,e,A){if(e!==e)return SfA(t,xfA,A);for(var i=A-1,n=t.length;++i<n;)if(t[i]===e)return i;return-1}function xfA(t){return t!==t}function LfA(t,e){for(var A=-1,i=t.length;++A<i&&RfA(e,t[A],0)>-1;);return A}function FfA(t){return yfA.test(t)}function hj(t){return FfA(t)?NfA(t):kfA(t)}function NfA(t){return t.match(DfA)||[]}var _fA=Object.prototype,GfA=_fA.toString,uj=MfA.Symbol,fj=uj?uj.prototype:void 0,mj=fj?fj.toString:void 0;function UfA(t,e,A){var i=-1,n=t.length;e<0&&(e=-e>n?0:n+e),A=A>n?n:A,A<0&&(A+=n),n=e>A?0:A-e>>>0,e>>>=0;for(var o=Array(n);++i<n;)o[i]=t[i+e];return o}function Rj(t){if(typeof t=="string")return t;if(JfA(t))return mj?mj.call(t):"";var e=t+"";return e=="0"&&1/t==-EfA?"-0":e}function KfA(t,e,A){var i=t.length;return A=A===void 0?i:A,!e&&A>=i?t:UfA(t,e,A)}function YfA(t){return!!t&&typeof t=="object"}function JfA(t){return typeof t=="symbol"||YfA(t)&&GfA.call(t)==QfA}function TfA(t){return t==null?"":Rj(t)}function HfA(t,e,A){if(t=TfA(t),t&&(A||e===void 0))return t.replace(hfA,"");if(!t||!(e=Rj(e)))return t;var i=hj(t),n=LfA(i,hj(e));return KfA(i,n).join("")}xj.exports=HfA});var $j=Je((t2e,Xj)=>{"use strict";var Zk=1/0,zfA=9007199254740991,OfA=17976931348623157e292,Fj=NaN,PfA="[object Symbol]",jfA=/^\s+|\s+$/g,qfA=/^[-+]0x[0-9a-f]+$/i,VfA=/^0b[01]+$/i,ZfA=/^0o[0-7]+$/i,AS="\\ud800-\\udfff",Yj="\\u0300-\\u036f\\ufe20-\\ufe23",Jj="\\u20d0-\\u20f0",Tj="\\ufe0e\\ufe0f",WfA="["+AS+"]",Wk="["+Yj+Jj+"]",Xk="\\ud83c[\\udffb-\\udfff]",XfA="(?:"+Wk+"|"+Xk+")",Hj="[^"+AS+"]",zj="(?:\\ud83c[\\udde6-\\uddff]){2}",Oj="[\\ud800-\\udbff][\\udc00-\\udfff]",Pj="\\u200d",jj=XfA+"?",qj="["+Tj+"]?",$fA="(?:"+Pj+"(?:"+[Hj,zj,Oj].join("|")+")"+qj+jj+")*",AmA=qj+jj+$fA,emA="(?:"+[Hj+Wk+"?",Wk,zj,Oj,WfA].join("|")+")",$k=RegExp(Xk+"(?="+Xk+")|"+emA+AmA,"g"),tmA=RegExp("["+Pj+AS+Yj+Jj+Tj+"]"),imA=parseInt,nmA=typeof global=="object"&&global&&global.Object===Object&&global,omA=typeof self=="object"&&self&&self.Object===Object&&self,rmA=nmA||omA||Function("return this")(),smA=cmA("length");function amA(t){return t.split("")}function cmA(t){return function(e){return e?.[t]}}function eS(t){return tmA.test(t)}function Vj(t){return eS(t)?gmA(t):smA(t)}function lmA(t){return eS(t)?ImA(t):amA(t)}function gmA(t){for(var e=$k.lastIndex=0;$k.test(t);)e++;return e}function ImA(t){return t.match($k)||[]}var CmA=Object.prototype,dmA=CmA.toString,Nj=rmA.Symbol,BmA=Math.ceil,EmA=Math.floor,_j=Nj?Nj.prototype:void 0,Gj=_j?_j.toString:void 0;function Uj(t,e){var A="";if(!t||e<1||e>zfA)return A;do e%2&&(A+=t),e=EmA(e/2),e&&(t+=t);while(e);return A}function QmA(t,e,A){var i=-1,n=t.length;e<0&&(e=-e>n?0:n+e),A=A>n?n:A,A<0&&(A+=n),n=e>A?0:A-e>>>0,e>>>=0;for(var o=Array(n);++i<n;)o[i]=t[i+e];return o}function Zj(t){if(typeof t=="string")return t;if(Wj(t))return Gj?Gj.call(t):"";var e=t+"";return e=="0"&&1/t==-Zk?"-0":e}function hmA(t,e,A){var i=t.length;return A=A===void 0?i:A,!e&&A>=i?t:QmA(t,e,A)}function umA(t,e){e=e===void 0?" ":Zj(e);var A=e.length;if(A<2)return A?Uj(e,t):e;var i=Uj(e,BmA(t/Vj(e)));return eS(e)?hmA(lmA(i),0,t).join(""):i.slice(0,t)}function Kj(t){var e=typeof t;return!!t&&(e=="object"||e=="function")}function fmA(t){return!!t&&typeof t=="object"}function Wj(t){return typeof t=="symbol"||fmA(t)&&dmA.call(t)==PfA}function mmA(t){if(!t)return t===0?t:0;if(t=wmA(t),t===Zk||t===-Zk){var e=t<0?-1:1;return e*OfA}return t===t?t:0}function pmA(t){var e=mmA(t),A=e%1;return e===e?A?e-A:e:0}function wmA(t){if(typeof t=="number")return t;if(Wj(t))return Fj;if(Kj(t)){var e=typeof t.valueOf=="function"?t.valueOf():t;t=Kj(e)?e+"":e}if(typeof t!="string")return t===0?t:+t;t=t.replace(jfA,"");var A=VfA.test(t);return A||ZfA.test(t)?imA(t.slice(2),A?2:8):qfA.test(t)?Fj:+t}function DmA(t){return t==null?"":Zj(t)}function ymA(t,e,A){t=DmA(t),e=pmA(e);var i=e?Vj(t):0;return e&&i<e?t+umA(e-i,A):t}Xj.exports=ymA});var eq=Je((i2e,Aq)=>{"use strict";Aq.exports=(t,e,A,i)=>{let n=(t+(i||"")).toString().includes("%");if(typeof t=="string"?[t,e,A,i]=t.match(/(0?\.?\d{1,3})%?\b/g).map(Number):i!==void 0&&(i=parseFloat(i)),typeof t!="number"||typeof e!="number"||typeof A!="number"||t>255||e>255||A>255)throw new TypeError("Expected three numbers below 256");if(typeof i=="number"){if(!n&&i>=0&&i<=1)i=Math.round(255*i);else if(n&&i>=0&&i<=100)i=Math.round(255*i/100);else throw new TypeError(`Expected alpha value (${i}) as a fraction or percentage`);i=(i|256).toString(16).slice(1)}else i="";return(A|e<<8|t<<16|1<<24).toString(16).slice(1)+i}});var iq=Je((n2e,tq)=>{"use strict";var Ju="a-f\\d",vmA=`#?[${Ju}]{3}[${Ju}]?`,bmA=`#?[${Ju}]{6}([${Ju}]{2})?`,MmA=new RegExp(`[^#${Ju}]`,"gi"),kmA=new RegExp(`^${vmA}$|^${bmA}$`,"i");tq.exports=(t,e={})=>{if(typeof t!="string"||MmA.test(t)||!kmA.test(t))throw new TypeError("Expected a valid hex string");t=t.replace(/^#/,"");let A=1;t.length===8&&(A=Number.parseInt(t.slice(6,8),16)/255,t=t.slice(0,6)),t.length===4&&(A=Number.parseInt(t.slice(3,4).repeat(2),16)/255,t=t.slice(0,3)),t.length===3&&(t=t[0]+t[0]+t[1]+t[1]+t[2]+t[2]);let i=Number.parseInt(t,16),n=i>>16,o=i>>8&255,r=i&255,s=typeof e.alpha=="number"?e.alpha:A;if(e.format==="array")return[n,o,r,s];if(e.format==="css"){let a=s===1?"":` / ${Number((s*100).toFixed(2))}%`;return`rgb(${n} ${o} ${r}${a})`}return{red:n,green:o,blue:r,alpha:s}}});var rq=Je((o2e,oq)=>{"use strict";var SmA=VP(),RmA=Qj(),xmA=Lj(),LmA=$j(),FmA=eq(),nq=iq(),tS=.75,iS=.25,nS=16777215,NmA=49979693;oq.exports=function(t){return"#"+UmA(String(JSON.stringify(t)))};function _mA(t){var e=RmA(t),A=[];return e.forEach(function(i){var n=SmA(i);n&&A.push(nq(xmA(n,"#"),{format:"array"}))}),A}function GmA(t){var e=[0,0,0];return t.forEach(function(A){for(var i=0;i<3;i++)e[i]+=A[i]}),[e[0]/t.length,e[1]/t.length,e[2]/t.length]}function UmA(t){var e,A=_mA(t);A.length>0&&(e=GmA(A));var i=1,n=0,o=1;if(t.length>0)for(var r=0;r<t.length;r++)t[r].charCodeAt(0)>n&&(n=t[r].charCodeAt(0)),o=parseInt(nS/n),i=(i+t[r].charCodeAt(0)*o*NmA)%nS;var s=(i*t.length%nS).toString(16);s=LmA(s,6,s);var a=nq(s,{format:"array"});return e?FmA(iS*a[0]+tS*e[0],iS*a[1]+tS*e[1],iS*a[2]+tS*e[2]):s}});var xq=Je(QS=>{"use strict";var Rq={b:"\b",f:"\f",n:`
`,r:"\r",t:"	",'"':'"',"/":"/","\\":"\\"},NpA=97;QS.parse=function(t,e,A){var i={},n=0,o=0,r=0,s=A&&A.bigint&&typeof BigInt<"u";return{data:a("",!0),pointers:i};function a(U,H){c();var q;L(U,"value");var j=E();switch(j){case"t":B("rue"),q=!0;break;case"f":B("alse"),q=!1;break;case"n":B("ull"),q=null;break;case'"':q=l();break;case"[":q=C(U);break;case"{":q=d(U);break;default:h(),"-0123456789".indexOf(j)>=0?q=I():_()}return L(U,"valueEnd"),c(),H&&r<t.length&&_(),q}function c(){A:for(;r<t.length;){switch(t[r]){case" ":o++;break;case"	":o+=4;break;case"\r":o=0;break;case`
`:o=0,n++;break;default:break A}r++}}function l(){for(var U="",H;H=E(),H!='"';)H=="\\"?(H=E(),H in Rq?U+=Rq[H]:H=="u"?U+=u():K()):U+=H;return U}function I(){var U="",H=!0;t[r]=="-"&&(U+=E()),U+=t[r]=="0"?E():D(),t[r]=="."&&(U+=E()+D(),H=!1),(t[r]=="e"||t[r]=="E")&&(U+=E(),(t[r]=="+"||t[r]=="-")&&(U+=E()),U+=D(),H=!1);var q=+U;return s&&H&&(q>Number.MAX_SAFE_INTEGER||q<Number.MIN_SAFE_INTEGER)?BigInt(U):q}function C(U){c();var H=[],q=0;if(E()=="]")return H;for(h();;){var j=U+"/"+q;H.push(a(j)),c();var gA=E();if(gA=="]")break;gA!=","&&K(),c(),q++}return H}function d(U){c();var H={};if(E()=="}")return H;for(h();;){var q=w();E()!='"'&&K();var j=l(),gA=U+"/"+ES(j);R(gA,"key",q),L(gA,"keyEnd"),c(),E()!=":"&&K(),c(),H[j]=a(gA),c();var QA=E();if(QA=="}")break;QA!=","&&K(),c()}return H}function B(U){for(var H=0;H<U.length;H++)E()!==U[H]&&K()}function E(){z();var U=t[r];return r++,o++,U}function h(){r--,o--}function u(){for(var U=4,H=0;U--;){H<<=4;var q=E().toLowerCase();q>="a"&&q<="f"?H+=q.charCodeAt()-NpA+10:q>="0"&&q<="9"?H+=+q:K()}return String.fromCharCode(H)}function D(){for(var U="";t[r]>="0"&&t[r]<="9";)U+=E();if(U.length)return U;z(),_()}function L(U,H){R(U,H,w())}function R(U,H,q){i[U]=i[U]||{},i[U][H]=q}function w(){return{line:n,column:o,pos:r}}function _(){throw new SyntaxError("Unexpected token "+t[r]+" in JSON at position "+r)}function K(){h(),_()}function z(){if(r>=t.length)throw new SyntaxError("Unexpected end of JSON input")}};QS.stringify=function(t,e,A){if(!x8(t))return;var i=0,n,o,r=typeof A=="object"?A.space:A;switch(typeof r){case"number":var s=r>10?10:r<0?0:Math.floor(r);r=s&&R(s," "),n=s,o=s;break;case"string":r=r.slice(0,10),n=0,o=0;for(var a=0;a<r.length;a++){var c=r[a];switch(c){case" ":o++;break;case"	":o+=4;break;case"\r":o=0;break;case`
`:o=0,i++;break;default:throw new Error("whitespace characters not allowed in JSON")}n++}break;default:r=void 0}var l="",I={},C=0,d=0,B=0,E=A&&A.es6&&typeof Map=="function";return h(t,0,""),{json:l,pointers:I};function h(w,_,K){switch(L(K,"value"),typeof w){case"number":case"bigint":case"boolean":u(""+w);break;case"string":u(L8(w));break;case"object":w===null?u("null"):typeof w.toJSON=="function"?u(L8(w.toJSON())):Array.isArray(w)?z():E?w.constructor.BYTES_PER_ELEMENT?z():w instanceof Map?H():w instanceof Set?H(!0):U():U()}L(K,"valueEnd");function z(){if(w.length){u("[");for(var q=_+1,j=0;j<w.length;j++){j&&u(","),D(q);var gA=x8(w[j])?w[j]:null,QA=K+"/"+j;h(gA,q,QA)}D(_),u("]")}else u("[]")}function U(){var q=Object.keys(w);if(q.length){u("{");for(var j=_+1,gA=0;gA<q.length;gA++){var QA=q[gA],BA=w[QA];if(x8(BA)){gA&&u(",");var lA=K+"/"+ES(QA);D(j),L(lA,"key"),u(L8(QA)),L(lA,"keyEnd"),u(":"),r&&u(" "),h(BA,j,lA)}}D(_),u("}")}else u("{}")}function H(q){if(w.size){u("{");for(var j=_+1,gA=!0,QA=w.entries(),BA=QA.next();!BA.done;){var lA=BA.value,vA=lA[0],tA=q?!0:lA[1];if(x8(tA)){gA||u(","),gA=!1;var cA=K+"/"+ES(vA);D(j),L(cA,"key"),u(L8(vA)),L(cA,"keyEnd"),u(":"),r&&u(" "),h(tA,j,cA)}BA=QA.next()}D(_),u("}")}else u("{}")}}function u(w){d+=w.length,B+=w.length,l+=w}function D(w){if(r){for(l+=`
`+R(w,r),C++,d=0;w--;)i?(C+=i,d=o):d+=o,B+=n;B+=1}}function L(w,_){I[w]=I[w]||{},I[w][_]={line:C,column:d,pos:B}}function R(w,_){return Array(w+1).join(_)}};var _pA=["number","bigint","boolean","string","object"];function x8(t){return _pA.indexOf(typeof t)>=0}var GpA=/"|\\/g,UpA=/[\b]/g,KpA=/\f/g,YpA=/\n/g,JpA=/\r/g,TpA=/\t/g;function L8(t){return t=t.replace(GpA,"\\$&").replace(KpA,"\\f").replace(UpA,"\\b").replace(YpA,"\\n").replace(JpA,"\\r").replace(TpA,"\\t"),'"'+t+'"'}var HpA=/~/g,zpA=/\//g;function ES(t){return t.replace(HpA,"~0").replace(zpA,"~1")}});var XW=Je((zfe,WW)=>{"use strict";var ZW=function(t,e){var A,i,n=1,o=0,r=0,s=String.alphabet;function a(c,l,I){if(I){for(A=l;I=a(c,A),I<76&&I>65;)++A;return+c.slice(l-1,A)}return I=s&&s.indexOf(c.charAt(l)),I>-1?I+76:(I=c.charCodeAt(l)||0,I<45||I>127?I:I<46?65:I<48?I-1:I<58?I+18:I<65?I-11:I<91?I+11:I<97?I-37:I<123?I+5:I-63)}if((t+="")!=(e+="")){for(;n;)if(i=a(t,o++),n=a(e,r++),i<76&&n<76&&i>66&&n>66&&(i=a(t,o,o),n=a(e,r,o=A),r=A),i!=n)return i<n?-1:1}return 0};try{WW.exports=ZW}catch{String.naturalCompare=ZW}});var B4=Je(vn=>{"use strict";Object.defineProperty(vn,"__esModule",{value:!0});vn.regexpCode=vn.getEsmExportName=vn.getProperty=vn.safeStringify=vn.stringify=vn.strConcat=vn.addCodeArg=vn.str=vn._=vn.nil=vn._Code=vn.Name=vn.IDENTIFIER=vn._CodeOrName=void 0;var C4=class{};vn._CodeOrName=C4;vn.IDENTIFIER=/^[a-z$_][a-z$_0-9]*$/i;var IC=class extends C4{constructor(e){if(super(),!vn.IDENTIFIER.test(e))throw new Error("CodeGen: name must be a valid identifier");this.str=e}toString(){return this.str}emptyStr(){return!1}get names(){return{[this.str]:1}}};vn.Name=IC;var Nc=class extends C4{constructor(e){super(),this._items=typeof e=="string"?[e]:e}toString(){return this.str}emptyStr(){if(this._items.length>1)return!1;let e=this._items[0];return e===""||e==='""'}get str(){var e;return(e=this._str)!==null&&e!==void 0?e:this._str=this._items.reduce((A,i)=>`${A}${i}`,"")}get names(){var e;return(e=this._names)!==null&&e!==void 0?e:this._names=this._items.reduce((A,i)=>(i instanceof IC&&(A[i.str]=(A[i.str]||0)+1),A),{})}};vn._Code=Nc;vn.nil=new Nc("");function AX(t,...e){let A=[t[0]],i=0;for(;i<e.length;)eR(A,e[i]),A.push(t[++i]);return new Nc(A)}vn._=AX;var AR=new Nc("+");function eX(t,...e){let A=[d4(t[0])],i=0;for(;i<e.length;)A.push(AR),eR(A,e[i]),A.push(AR,d4(t[++i]));return FvA(A),new Nc(A)}vn.str=eX;function eR(t,e){e instanceof Nc?t.push(...e._items):e instanceof IC?t.push(e):t.push(GvA(e))}vn.addCodeArg=eR;function FvA(t){let e=1;for(;e<t.length-1;){if(t[e]===AR){let A=NvA(t[e-1],t[e+1]);if(A!==void 0){t.splice(e-1,3,A);continue}t[e++]="+"}e++}}function NvA(t,e){if(e==='""')return t;if(t==='""')return e;if(typeof t=="string")return e instanceof IC||t[t.length-1]!=='"'?void 0:typeof e!="string"?`${t.slice(0,-1)}${e}"`:e[0]==='"'?t.slice(0,-1)+e.slice(1):void 0;if(typeof e=="string"&&e[0]==='"'&&!(t instanceof IC))return`"${t}${e.slice(1)}`}function _vA(t,e){return e.emptyStr()?t:t.emptyStr()?e:eX`${t}${e}`}vn.strConcat=_vA;function GvA(t){return typeof t=="number"||typeof t=="boolean"||t===null?t:d4(Array.isArray(t)?t.join(","):t)}function UvA(t){return new Nc(d4(t))}vn.stringify=UvA;function d4(t){return JSON.stringify(t).replace(/\u2028/g,"\\u2028").replace(/\u2029/g,"\\u2029")}vn.safeStringify=d4;function KvA(t){return typeof t=="string"&&vn.IDENTIFIER.test(t)?new Nc(`.${t}`):AX`[${t}]`}vn.getProperty=KvA;function YvA(t){if(typeof t=="string"&&vn.IDENTIFIER.test(t))return new Nc(`${t}`);throw new Error(`CodeGen: invalid export name: ${t}, use explicit $id name mapping`)}vn.getEsmExportName=YvA;function JvA(t){return new Nc(t.toString())}vn.regexpCode=JvA});var nR=Je(Ma=>{"use strict";Object.defineProperty(Ma,"__esModule",{value:!0});Ma.ValueScope=Ma.ValueScopeName=Ma.Scope=Ma.varKinds=Ma.UsedValueState=void 0;var ba=B4(),tR=class extends Error{constructor(e){super(`CodeGen: "code" for ${e} not defined`),this.value=e.value}},f5=function(t){return t[t.Started=0]="Started",t[t.Completed=1]="Completed",t}(f5||(Ma.UsedValueState=f5={}));Ma.varKinds={const:new ba.Name("const"),let:new ba.Name("let"),var:new ba.Name("var")};var m5=class{constructor({prefixes:e,parent:A}={}){this._names={},this._prefixes=e,this._parent=A}toName(e){return e instanceof ba.Name?e:this.name(e)}name(e){return new ba.Name(this._newName(e))}_newName(e){let A=this._names[e]||this._nameGroup(e);return`${e}${A.index++}`}_nameGroup(e){var A,i;if(!((i=(A=this._parent)===null||A===void 0?void 0:A._prefixes)===null||i===void 0)&&i.has(e)||this._prefixes&&!this._prefixes.has(e))throw new Error(`CodeGen: prefix "${e}" is not allowed in this scope`);return this._names[e]={prefix:e,index:0}}};Ma.Scope=m5;var p5=class extends ba.Name{constructor(e,A){super(A),this.prefix=e}setValue(e,{property:A,itemIndex:i}){this.value=e,this.scopePath=(0,ba._)`.${new ba.Name(A)}[${i}]`}};Ma.ValueScopeName=p5;var TvA=(0,ba._)`\n`,iR=class extends m5{constructor(e){super(e),this._values={},this._scope=e.scope,this.opts=Ye(rA({},e),{_n:e.lines?TvA:ba.nil})}get(){return this._scope}name(e){return new p5(e,this._newName(e))}value(e,A){var i;if(A.ref===void 0)throw new Error("CodeGen: ref must be passed in value");let n=this.toName(e),{prefix:o}=n,r=(i=A.key)!==null&&i!==void 0?i:A.ref,s=this._values[o];if(s){let l=s.get(r);if(l)return l}else s=this._values[o]=new Map;s.set(r,n);let a=this._scope[o]||(this._scope[o]=[]),c=a.length;return a[c]=A.ref,n.setValue(A,{property:o,itemIndex:c}),n}getValue(e,A){let i=this._values[e];if(i)return i.get(A)}scopeRefs(e,A=this._values){return this._reduceValues(A,i=>{if(i.scopePath===void 0)throw new Error(`CodeGen: name "${i}" has no value`);return(0,ba._)`${e}${i.scopePath}`})}scopeCode(e=this._values,A,i){return this._reduceValues(e,n=>{if(n.value===void 0)throw new Error(`CodeGen: name "${n}" has no value`);return n.value.code},A,i)}_reduceValues(e,A,i={},n){let o=ba.nil;for(let r in e){let s=e[r];if(!s)continue;let a=i[r]=i[r]||new Map;s.forEach(c=>{if(a.has(c))return;a.set(c,f5.Started);let l=A(c);if(l){let I=this.opts.es5?Ma.varKinds.var:Ma.varKinds.const;o=(0,ba._)`${o}${I} ${c} = ${l};${this.opts._n}`}else if(l=n?.(c))o=(0,ba._)`${o}${l}${this.opts._n}`;else throw new tR(c);a.set(c,f5.Completed)})}return o}};Ma.ValueScope=iR});var An=Je(Xi=>{"use strict";Object.defineProperty(Xi,"__esModule",{value:!0});Xi.or=Xi.and=Xi.not=Xi.CodeGen=Xi.operators=Xi.varKinds=Xi.ValueScopeName=Xi.ValueScope=Xi.Scope=Xi.Name=Xi.regexpCode=Xi.stringify=Xi.getProperty=Xi.nil=Xi.strConcat=Xi.str=Xi._=void 0;var Bn=B4(),ul=nR(),I1=B4();Object.defineProperty(Xi,"_",{enumerable:!0,get:function(){return I1._}});Object.defineProperty(Xi,"str",{enumerable:!0,get:function(){return I1.str}});Object.defineProperty(Xi,"strConcat",{enumerable:!0,get:function(){return I1.strConcat}});Object.defineProperty(Xi,"nil",{enumerable:!0,get:function(){return I1.nil}});Object.defineProperty(Xi,"getProperty",{enumerable:!0,get:function(){return I1.getProperty}});Object.defineProperty(Xi,"stringify",{enumerable:!0,get:function(){return I1.stringify}});Object.defineProperty(Xi,"regexpCode",{enumerable:!0,get:function(){return I1.regexpCode}});Object.defineProperty(Xi,"Name",{enumerable:!0,get:function(){return I1.Name}});var b5=nR();Object.defineProperty(Xi,"Scope",{enumerable:!0,get:function(){return b5.Scope}});Object.defineProperty(Xi,"ValueScope",{enumerable:!0,get:function(){return b5.ValueScope}});Object.defineProperty(Xi,"ValueScopeName",{enumerable:!0,get:function(){return b5.ValueScopeName}});Object.defineProperty(Xi,"varKinds",{enumerable:!0,get:function(){return b5.varKinds}});Xi.operators={GT:new Bn._Code(">"),GTE:new Bn._Code(">="),LT:new Bn._Code("<"),LTE:new Bn._Code("<="),EQ:new Bn._Code("==="),NEQ:new Bn._Code("!=="),NOT:new Bn._Code("!"),OR:new Bn._Code("||"),AND:new Bn._Code("&&"),ADD:new Bn._Code("+")};var N0=class{optimizeNodes(){return this}optimizeNames(e,A){return this}},oR=class extends N0{constructor(e,A,i){super(),this.varKind=e,this.name=A,this.rhs=i}render({es5:e,_n:A}){let i=e?ul.varKinds.var:this.varKind,n=this.rhs===void 0?"":` = ${this.rhs}`;return`${i} ${this.name}${n};`+A}optimizeNames(e,A){if(e[this.name.str])return this.rhs&&(this.rhs=cE(this.rhs,e,A)),this}get names(){return this.rhs instanceof Bn._CodeOrName?this.rhs.names:{}}},D5=class extends N0{constructor(e,A,i){super(),this.lhs=e,this.rhs=A,this.sideEffects=i}render({_n:e}){return`${this.lhs} = ${this.rhs};`+e}optimizeNames(e,A){if(!(this.lhs instanceof Bn.Name&&!e[this.lhs.str]&&!this.sideEffects))return this.rhs=cE(this.rhs,e,A),this}get names(){let e=this.lhs instanceof Bn.Name?{}:rA({},this.lhs.names);return v5(e,this.rhs)}},rR=class extends D5{constructor(e,A,i,n){super(e,i,n),this.op=A}render({_n:e}){return`${this.lhs} ${this.op}= ${this.rhs};`+e}},sR=class extends N0{constructor(e){super(),this.label=e,this.names={}}render({_n:e}){return`${this.label}:`+e}},aR=class extends N0{constructor(e){super(),this.label=e,this.names={}}render({_n:e}){return`break${this.label?` ${this.label}`:""};`+e}},cR=class extends N0{constructor(e){super(),this.error=e}render({_n:e}){return`throw ${this.error};`+e}get names(){return this.error.names}},lR=class extends N0{constructor(e){super(),this.code=e}render({_n:e}){return`${this.code};`+e}optimizeNodes(){return`${this.code}`?this:void 0}optimizeNames(e,A){return this.code=cE(this.code,e,A),this}get names(){return this.code instanceof Bn._CodeOrName?this.code.names:{}}},E4=class extends N0{constructor(e=[]){super(),this.nodes=e}render(e){return this.nodes.reduce((A,i)=>A+i.render(e),"")}optimizeNodes(){let{nodes:e}=this,A=e.length;for(;A--;){let i=e[A].optimizeNodes();Array.isArray(i)?e.splice(A,1,...i):i?e[A]=i:e.splice(A,1)}return e.length>0?this:void 0}optimizeNames(e,A){let{nodes:i}=this,n=i.length;for(;n--;){let o=i[n];o.optimizeNames(e,A)||(HvA(e,o.names),i.splice(n,1))}return i.length>0?this:void 0}get names(){return this.nodes.reduce((e,A)=>CC(e,A.names),{})}},_0=class extends E4{render(e){return"{"+e._n+super.render(e)+"}"+e._n}},gR=class extends E4{},IR=(()=>{class t extends _0{}return t.kind="else",t})(),w5=(()=>{class t extends _0{constructor(A,i){super(i),this.condition=A}render(A){let i=`if(${this.condition})`+super.render(A);return this.else&&(i+="else "+this.else.render(A)),i}optimizeNodes(){super.optimizeNodes();let A=this.condition;if(A===!0)return this.nodes;let i=this.else;if(i){let n=i.optimizeNodes();i=this.else=Array.isArray(n)?new IR(n):n}if(i)return A===!1?i instanceof t?i:i.nodes:this.nodes.length?this:new t(rX(A),i instanceof t?[i]:i.nodes);if(!(A===!1||!this.nodes.length))return this}optimizeNames(A,i){var n;if(this.else=(n=this.else)===null||n===void 0?void 0:n.optimizeNames(A,i),!!(super.optimizeNames(A,i)||this.else))return this.condition=cE(this.condition,A,i),this}get names(){let A=super.names;return v5(A,this.condition),this.else&&CC(A,this.else.names),A}}return t.kind="if",t})(),M5=(()=>{class t extends _0{}return t.kind="for",t})(),CR=class extends M5{constructor(e){super(),this.iteration=e}render(e){return`for(${this.iteration})`+super.render(e)}optimizeNames(e,A){if(super.optimizeNames(e,A))return this.iteration=cE(this.iteration,e,A),this}get names(){return CC(super.names,this.iteration.names)}},dR=class extends M5{constructor(e,A,i,n){super(),this.varKind=e,this.name=A,this.from=i,this.to=n}render(e){let A=e.es5?ul.varKinds.var:this.varKind,{name:i,from:n,to:o}=this;return`for(${A} ${i}=${n}; ${i}<${o}; ${i}++)`+super.render(e)}get names(){let e=v5(super.names,this.from);return v5(e,this.to)}},y5=class extends M5{constructor(e,A,i,n){super(),this.loop=e,this.varKind=A,this.name=i,this.iterable=n}render(e){return`for(${this.varKind} ${this.name} ${this.loop} ${this.iterable})`+super.render(e)}optimizeNames(e,A){if(super.optimizeNames(e,A))return this.iterable=cE(this.iterable,e,A),this}get names(){return CC(super.names,this.iterable.names)}},tX=(()=>{class t extends _0{constructor(A,i,n){super(),this.name=A,this.args=i,this.async=n}render(A){return`${this.async?"async ":""}function ${this.name}(${this.args})`+super.render(A)}}return t.kind="func",t})(),iX=(()=>{class t extends E4{render(A){return"return "+super.render(A)}}return t.kind="return",t})(),BR=class extends _0{render(e){let A="try"+super.render(e);return this.catch&&(A+=this.catch.render(e)),this.finally&&(A+=this.finally.render(e)),A}optimizeNodes(){var e,A;return super.optimizeNodes(),(e=this.catch)===null||e===void 0||e.optimizeNodes(),(A=this.finally)===null||A===void 0||A.optimizeNodes(),this}optimizeNames(e,A){var i,n;return super.optimizeNames(e,A),(i=this.catch)===null||i===void 0||i.optimizeNames(e,A),(n=this.finally)===null||n===void 0||n.optimizeNames(e,A),this}get names(){let e=super.names;return this.catch&&CC(e,this.catch.names),this.finally&&CC(e,this.finally.names),e}},nX=(()=>{class t extends _0{constructor(A){super(),this.error=A}render(A){return`catch(${this.error})`+super.render(A)}}return t.kind="catch",t})(),oX=(()=>{class t extends _0{render(A){return"finally"+super.render(A)}}return t.kind="finally",t})(),ER=class{constructor(e,A={}){this._values={},this._blockStarts=[],this._constants={},this.opts=Ye(rA({},A),{_n:A.lines?`
`:""}),this._extScope=e,this._scope=new ul.Scope({parent:e}),this._nodes=[new gR]}toString(){return this._root.render(this.opts)}name(e){return this._scope.name(e)}scopeName(e){return this._extScope.name(e)}scopeValue(e,A){let i=this._extScope.value(e,A);return(this._values[i.prefix]||(this._values[i.prefix]=new Set)).add(i),i}getScopeValue(e,A){return this._extScope.getValue(e,A)}scopeRefs(e){return this._extScope.scopeRefs(e,this._values)}scopeCode(){return this._extScope.scopeCode(this._values)}_def(e,A,i,n){let o=this._scope.toName(A);return i!==void 0&&n&&(this._constants[o.str]=i),this._leafNode(new oR(e,o,i)),o}const(e,A,i){return this._def(ul.varKinds.const,e,A,i)}let(e,A,i){return this._def(ul.varKinds.let,e,A,i)}var(e,A,i){return this._def(ul.varKinds.var,e,A,i)}assign(e,A,i){return this._leafNode(new D5(e,A,i))}add(e,A){return this._leafNode(new rR(e,Xi.operators.ADD,A))}code(e){return typeof e=="function"?e():e!==Bn.nil&&this._leafNode(new lR(e)),this}object(...e){let A=["{"];for(let[i,n]of e)A.length>1&&A.push(","),A.push(i),(i!==n||this.opts.es5)&&(A.push(":"),(0,Bn.addCodeArg)(A,n));return A.push("}"),new Bn._Code(A)}if(e,A,i){if(this._blockNode(new w5(e)),A&&i)this.code(A).else().code(i).endIf();else if(A)this.code(A).endIf();else if(i)throw new Error('CodeGen: "else" body without "then" body');return this}elseIf(e){return this._elseNode(new w5(e))}else(){return this._elseNode(new IR)}endIf(){return this._endBlockNode(w5,IR)}_for(e,A){return this._blockNode(e),A&&this.code(A).endFor(),this}for(e,A){return this._for(new CR(e),A)}forRange(e,A,i,n,o=this.opts.es5?ul.varKinds.var:ul.varKinds.let){let r=this._scope.toName(e);return this._for(new dR(o,r,A,i),()=>n(r))}forOf(e,A,i,n=ul.varKinds.const){let o=this._scope.toName(e);if(this.opts.es5){let r=A instanceof Bn.Name?A:this.var("_arr",A);return this.forRange("_i",0,(0,Bn._)`${r}.length`,s=>{this.var(o,(0,Bn._)`${r}[${s}]`),i(o)})}return this._for(new y5("of",n,o,A),()=>i(o))}forIn(e,A,i,n=this.opts.es5?ul.varKinds.var:ul.varKinds.const){if(this.opts.ownProperties)return this.forOf(e,(0,Bn._)`Object.keys(${A})`,i);let o=this._scope.toName(e);return this._for(new y5("in",n,o,A),()=>i(o))}endFor(){return this._endBlockNode(M5)}label(e){return this._leafNode(new sR(e))}break(e){return this._leafNode(new aR(e))}return(e){let A=new iX;if(this._blockNode(A),this.code(e),A.nodes.length!==1)throw new Error('CodeGen: "return" should have one node');return this._endBlockNode(iX)}try(e,A,i){if(!A&&!i)throw new Error('CodeGen: "try" without "catch" and "finally"');let n=new BR;if(this._blockNode(n),this.code(e),A){let o=this.name("e");this._currNode=n.catch=new nX(o),A(o)}return i&&(this._currNode=n.finally=new oX,this.code(i)),this._endBlockNode(nX,oX)}throw(e){return this._leafNode(new cR(e))}block(e,A){return this._blockStarts.push(this._nodes.length),e&&this.code(e).endBlock(A),this}endBlock(e){let A=this._blockStarts.pop();if(A===void 0)throw new Error("CodeGen: not in self-balancing block");let i=this._nodes.length-A;if(i<0||e!==void 0&&i!==e)throw new Error(`CodeGen: wrong number of nodes: ${i} vs ${e} expected`);return this._nodes.length=A,this}func(e,A=Bn.nil,i,n){return this._blockNode(new tX(e,A,i)),n&&this.code(n).endFunc(),this}endFunc(){return this._endBlockNode(tX)}optimize(e=1){for(;e-- >0;)this._root.optimizeNodes(),this._root.optimizeNames(this._root.names,this._constants)}_leafNode(e){return this._currNode.nodes.push(e),this}_blockNode(e){this._currNode.nodes.push(e),this._nodes.push(e)}_endBlockNode(e,A){let i=this._currNode;if(i instanceof e||A&&i instanceof A)return this._nodes.pop(),this;throw new Error(`CodeGen: not in block "${A?`${e.kind}/${A.kind}`:e.kind}"`)}_elseNode(e){let A=this._currNode;if(!(A instanceof w5))throw new Error('CodeGen: "else" without "if"');return this._currNode=A.else=e,this}get _root(){return this._nodes[0]}get _currNode(){let e=this._nodes;return e[e.length-1]}set _currNode(e){let A=this._nodes;A[A.length-1]=e}};Xi.CodeGen=ER;function CC(t,e){for(let A in e)t[A]=(t[A]||0)+(e[A]||0);return t}function v5(t,e){return e instanceof Bn._CodeOrName?CC(t,e.names):t}function cE(t,e,A){if(t instanceof Bn.Name)return i(t);if(!n(t))return t;return new Bn._Code(t._items.reduce((o,r)=>(r instanceof Bn.Name&&(r=i(r)),r instanceof Bn._Code?o.push(...r._items):o.push(r),o),[]));function i(o){let r=A[o.str];return r===void 0||e[o.str]!==1?o:(delete e[o.str],r)}function n(o){return o instanceof Bn._Code&&o._items.some(r=>r instanceof Bn.Name&&e[r.str]===1&&A[r.str]!==void 0)}}function HvA(t,e){for(let A in e)t[A]=(t[A]||0)-(e[A]||0)}function rX(t){return typeof t=="boolean"||typeof t=="number"||t===null?!t:(0,Bn._)`!${QR(t)}`}Xi.not=rX;var zvA=sX(Xi.operators.AND);function OvA(...t){return t.reduce(zvA)}Xi.and=OvA;var PvA=sX(Xi.operators.OR);function jvA(...t){return t.reduce(PvA)}Xi.or=jvA;function sX(t){return(e,A)=>e===Bn.nil?A:A===Bn.nil?e:(0,Bn._)`${QR(e)} ${t} ${QR(A)}`}function QR(t){return t instanceof Bn.Name?t:(0,Bn._)`(${t})`}});var bn=Je(en=>{"use strict";Object.defineProperty(en,"__esModule",{value:!0});en.checkStrictMode=en.getErrorPath=en.Type=en.useFunc=en.setEvaluated=en.evaluatedPropsToName=en.mergeEvaluated=en.eachItem=en.unescapeJsonPointer=en.escapeJsonPointer=en.escapeFragment=en.unescapeFragment=en.schemaRefOrVal=en.schemaHasRulesButRef=en.schemaHasRules=en.checkUnknownRules=en.alwaysValidSchema=en.toHash=void 0;var Do=An(),qvA=B4();function VvA(t){let e={};for(let A of t)e[A]=!0;return e}en.toHash=VvA;function ZvA(t,e){return typeof e=="boolean"?e:Object.keys(e).length===0?!0:(lX(t,e),!gX(e,t.self.RULES.all))}en.alwaysValidSchema=ZvA;function lX(t,e=t.schema){let{opts:A,self:i}=t;if(!A.strictSchema||typeof e=="boolean")return;let n=i.RULES.keywords;for(let o in e)n[o]||dX(t,`unknown keyword: "${o}"`)}en.checkUnknownRules=lX;function gX(t,e){if(typeof t=="boolean")return!t;for(let A in t)if(e[A])return!0;return!1}en.schemaHasRules=gX;function WvA(t,e){if(typeof t=="boolean")return!t;for(let A in t)if(A!=="$ref"&&e.all[A])return!0;return!1}en.schemaHasRulesButRef=WvA;function XvA({topSchemaRef:t,schemaPath:e},A,i,n){if(!n){if(typeof A=="number"||typeof A=="boolean")return A;if(typeof A=="string")return(0,Do._)`${A}`}return(0,Do._)`${t}${e}${(0,Do.getProperty)(i)}`}en.schemaRefOrVal=XvA;function $vA(t){return IX(decodeURIComponent(t))}en.unescapeFragment=$vA;function A9A(t){return encodeURIComponent(uR(t))}en.escapeFragment=A9A;function uR(t){return typeof t=="number"?`${t}`:t.replace(/~/g,"~0").replace(/\//g,"~1")}en.escapeJsonPointer=uR;function IX(t){return t.replace(/~1/g,"/").replace(/~0/g,"~")}en.unescapeJsonPointer=IX;function e9A(t,e){if(Array.isArray(t))for(let A of t)e(A);else e(t)}en.eachItem=e9A;function aX({mergeNames:t,mergeToName:e,mergeValues:A,resultToName:i}){return(n,o,r,s)=>{let a=r===void 0?o:r instanceof Do.Name?(o instanceof Do.Name?t(n,o,r):e(n,o,r),r):o instanceof Do.Name?(e(n,r,o),o):A(o,r);return s===Do.Name&&!(a instanceof Do.Name)?i(n,a):a}}en.mergeEvaluated={props:aX({mergeNames:(t,e,A)=>t.if((0,Do._)`${A} !== true && ${e} !== undefined`,()=>{t.if((0,Do._)`${e} === true`,()=>t.assign(A,!0),()=>t.assign(A,(0,Do._)`${A} || {}`).code((0,Do._)`Object.assign(${A}, ${e})`))}),mergeToName:(t,e,A)=>t.if((0,Do._)`${A} !== true`,()=>{e===!0?t.assign(A,!0):(t.assign(A,(0,Do._)`${A} || {}`),fR(t,A,e))}),mergeValues:(t,e)=>t===!0?!0:rA(rA({},t),e),resultToName:CX}),items:aX({mergeNames:(t,e,A)=>t.if((0,Do._)`${A} !== true && ${e} !== undefined`,()=>t.assign(A,(0,Do._)`${e} === true ? true : ${A} > ${e} ? ${A} : ${e}`)),mergeToName:(t,e,A)=>t.if((0,Do._)`${A} !== true`,()=>t.assign(A,e===!0?!0:(0,Do._)`${A} > ${e} ? ${A} : ${e}`)),mergeValues:(t,e)=>t===!0?!0:Math.max(t,e),resultToName:(t,e)=>t.var("items",e)})};function CX(t,e){if(e===!0)return t.var("props",!0);let A=t.var("props",(0,Do._)`{}`);return e!==void 0&&fR(t,A,e),A}en.evaluatedPropsToName=CX;function fR(t,e,A){Object.keys(A).forEach(i=>t.assign((0,Do._)`${e}${(0,Do.getProperty)(i)}`,!0))}en.setEvaluated=fR;var cX={};function t9A(t,e){return t.scopeValue("func",{ref:e,code:cX[e.code]||(cX[e.code]=new qvA._Code(e.code))})}en.useFunc=t9A;var hR=function(t){return t[t.Num=0]="Num",t[t.Str=1]="Str",t}(hR||(en.Type=hR={}));function i9A(t,e,A){if(t instanceof Do.Name){let i=e===hR.Num;return A?i?(0,Do._)`"[" + ${t} + "]"`:(0,Do._)`"['" + ${t} + "']"`:i?(0,Do._)`"/" + ${t}`:(0,Do._)`"/" + ${t}.replace(/~/g, "~0").replace(/\\//g, "~1")`}return A?(0,Do.getProperty)(t).toString():"/"+uR(t)}en.getErrorPath=i9A;function dX(t,e,A=t.opts.strictSchema){if(A){if(e=`strict mode: ${e}`,A===!0)throw new Error(e);t.self.logger.warn(e)}}en.checkStrictMode=dX});var G0=Je(mR=>{"use strict";Object.defineProperty(mR,"__esModule",{value:!0});var Ss=An(),n9A={data:new Ss.Name("data"),valCxt:new Ss.Name("valCxt"),instancePath:new Ss.Name("instancePath"),parentData:new Ss.Name("parentData"),parentDataProperty:new Ss.Name("parentDataProperty"),rootData:new Ss.Name("rootData"),dynamicAnchors:new Ss.Name("dynamicAnchors"),vErrors:new Ss.Name("vErrors"),errors:new Ss.Name("errors"),this:new Ss.Name("this"),self:new Ss.Name("self"),scope:new Ss.Name("scope"),json:new Ss.Name("json"),jsonPos:new Ss.Name("jsonPos"),jsonLen:new Ss.Name("jsonLen"),jsonPart:new Ss.Name("jsonPart")};mR.default=n9A});var Q4=Je(Rs=>{"use strict";Object.defineProperty(Rs,"__esModule",{value:!0});Rs.extendErrors=Rs.resetErrorsCount=Rs.reportExtraError=Rs.reportError=Rs.keyword$DataError=Rs.keywordError=void 0;var fn=An(),k5=bn(),Zs=G0();Rs.keywordError={message:({keyword:t})=>(0,fn.str)`must pass "${t}" keyword validation`};Rs.keyword$DataError={message:({keyword:t,schemaType:e})=>e?(0,fn.str)`"${t}" keyword must be ${e} ($data)`:(0,fn.str)`"${t}" keyword is invalid ($data)`};function o9A(t,e=Rs.keywordError,A,i){let{it:n}=t,{gen:o,compositeRule:r,allErrors:s}=n,a=QX(t,e,A);i??(r||s)?BX(o,a):EX(n,(0,fn._)`[${a}]`)}Rs.reportError=o9A;function r9A(t,e=Rs.keywordError,A){let{it:i}=t,{gen:n,compositeRule:o,allErrors:r}=i,s=QX(t,e,A);BX(n,s),o||r||EX(i,Zs.default.vErrors)}Rs.reportExtraError=r9A;function s9A(t,e){t.assign(Zs.default.errors,e),t.if((0,fn._)`${Zs.default.vErrors} !== null`,()=>t.if(e,()=>t.assign((0,fn._)`${Zs.default.vErrors}.length`,e),()=>t.assign(Zs.default.vErrors,null)))}Rs.resetErrorsCount=s9A;function a9A({gen:t,keyword:e,schemaValue:A,data:i,errsCount:n,it:o}){if(n===void 0)throw new Error("ajv implementation error");let r=t.name("err");t.forRange("i",n,Zs.default.errors,s=>{t.const(r,(0,fn._)`${Zs.default.vErrors}[${s}]`),t.if((0,fn._)`${r}.instancePath === undefined`,()=>t.assign((0,fn._)`${r}.instancePath`,(0,fn.strConcat)(Zs.default.instancePath,o.errorPath))),t.assign((0,fn._)`${r}.schemaPath`,(0,fn.str)`${o.errSchemaPath}/${e}`),o.opts.verbose&&(t.assign((0,fn._)`${r}.schema`,A),t.assign((0,fn._)`${r}.data`,i))})}Rs.extendErrors=a9A;function BX(t,e){let A=t.const("err",e);t.if((0,fn._)`${Zs.default.vErrors} === null`,()=>t.assign(Zs.default.vErrors,(0,fn._)`[${A}]`),(0,fn._)`${Zs.default.vErrors}.push(${A})`),t.code((0,fn._)`${Zs.default.errors}++`)}function EX(t,e){let{gen:A,validateName:i,schemaEnv:n}=t;n.$async?A.throw((0,fn._)`new ${t.ValidationError}(${e})`):(A.assign((0,fn._)`${i}.errors`,e),A.return(!1))}var dC={keyword:new fn.Name("keyword"),schemaPath:new fn.Name("schemaPath"),params:new fn.Name("params"),propertyName:new fn.Name("propertyName"),message:new fn.Name("message"),schema:new fn.Name("schema"),parentSchema:new fn.Name("parentSchema")};function QX(t,e,A){let{createErrors:i}=t.it;return i===!1?(0,fn._)`{}`:c9A(t,e,A)}function c9A(t,e,A={}){let{gen:i,it:n}=t,o=[l9A(n,A),g9A(t,A)];return I9A(t,e,o),i.object(...o)}function l9A({errorPath:t},{instancePath:e}){let A=e?(0,fn.str)`${t}${(0,k5.getErrorPath)(e,k5.Type.Str)}`:t;return[Zs.default.instancePath,(0,fn.strConcat)(Zs.default.instancePath,A)]}function g9A({keyword:t,it:{errSchemaPath:e}},{schemaPath:A,parentSchema:i}){let n=i?e:(0,fn.str)`${e}/${t}`;return A&&(n=(0,fn.str)`${n}${(0,k5.getErrorPath)(A,k5.Type.Str)}`),[dC.schemaPath,n]}function I9A(t,{params:e,message:A},i){let{keyword:n,data:o,schemaValue:r,it:s}=t,{opts:a,propertyName:c,topSchemaRef:l,schemaPath:I}=s;i.push([dC.keyword,n],[dC.params,typeof e=="function"?e(t):e||(0,fn._)`{}`]),a.messages&&i.push([dC.message,typeof A=="function"?A(t):A]),a.verbose&&i.push([dC.schema,r],[dC.parentSchema,(0,fn._)`${l}${I}`],[Zs.default.data,o]),c&&i.push([dC.propertyName,c])}});var uX=Je(lE=>{"use strict";Object.defineProperty(lE,"__esModule",{value:!0});lE.boolOrEmptySchema=lE.topBoolOrEmptySchema=void 0;var C9A=Q4(),d9A=An(),B9A=G0(),E9A={message:"boolean schema is false"};function Q9A(t){let{gen:e,schema:A,validateName:i}=t;A===!1?hX(t,!1):typeof A=="object"&&A.$async===!0?e.return(B9A.default.data):(e.assign((0,d9A._)`${i}.errors`,null),e.return(!0))}lE.topBoolOrEmptySchema=Q9A;function h9A(t,e){let{gen:A,schema:i}=t;i===!1?(A.var(e,!1),hX(t)):A.var(e,!0)}lE.boolOrEmptySchema=h9A;function hX(t,e){let{gen:A,data:i}=t,n={gen:A,keyword:"false schema",data:i,schema:!1,schemaCode:!1,schemaValue:!1,params:{},it:t};(0,C9A.reportError)(n,E9A,void 0,e)}});var pR=Je(gE=>{"use strict";Object.defineProperty(gE,"__esModule",{value:!0});gE.getRules=gE.isJSONType=void 0;var u9A=["string","number","integer","boolean","null","object","array"],f9A=new Set(u9A);function m9A(t){return typeof t=="string"&&f9A.has(t)}gE.isJSONType=m9A;function p9A(){let t={number:{type:"number",rules:[]},string:{type:"string",rules:[]},array:{type:"array",rules:[]},object:{type:"object",rules:[]}};return{types:Ye(rA({},t),{integer:!0,boolean:!0,null:!0}),rules:[{rules:[]},t.number,t.string,t.array,t.object],post:{rules:[]},all:{},keywords:{}}}gE.getRules=p9A});var wR=Je(C1=>{"use strict";Object.defineProperty(C1,"__esModule",{value:!0});C1.shouldUseRule=C1.shouldUseGroup=C1.schemaHasRulesForType=void 0;function w9A({schema:t,self:e},A){let i=e.RULES.types[A];return i&&i!==!0&&fX(t,i)}C1.schemaHasRulesForType=w9A;function fX(t,e){return e.rules.some(A=>mX(t,A))}C1.shouldUseGroup=fX;function mX(t,e){var A;return t[e.keyword]!==void 0||((A=e.definition.implements)===null||A===void 0?void 0:A.some(i=>t[i]!==void 0))}C1.shouldUseRule=mX});var h4=Je(xs=>{"use strict";Object.defineProperty(xs,"__esModule",{value:!0});xs.reportTypeError=xs.checkDataTypes=xs.checkDataType=xs.coerceAndCheckDataType=xs.getJSONTypes=xs.getSchemaTypes=xs.DataType=void 0;var D9A=pR(),y9A=wR(),v9A=Q4(),Ui=An(),pX=bn(),IE=function(t){return t[t.Correct=0]="Correct",t[t.Wrong=1]="Wrong",t}(IE||(xs.DataType=IE={}));function b9A(t){let e=wX(t.type);if(e.includes("null")){if(t.nullable===!1)throw new Error("type: null contradicts nullable: false")}else{if(!e.length&&t.nullable!==void 0)throw new Error('"nullable" cannot be used without "type"');t.nullable===!0&&e.push("null")}return e}xs.getSchemaTypes=b9A;function wX(t){let e=Array.isArray(t)?t:t?[t]:[];if(e.every(D9A.isJSONType))return e;throw new Error("type must be JSONType or JSONType[]: "+e.join(","))}xs.getJSONTypes=wX;function M9A(t,e){let{gen:A,data:i,opts:n}=t,o=k9A(e,n.coerceTypes),r=e.length>0&&!(o.length===0&&e.length===1&&(0,y9A.schemaHasRulesForType)(t,e[0]));if(r){let s=yR(e,i,n.strictNumbers,IE.Wrong);A.if(s,()=>{o.length?S9A(t,e,o):vR(t)})}return r}xs.coerceAndCheckDataType=M9A;var DX=new Set(["string","number","integer","boolean","null"]);function k9A(t,e){return e?t.filter(A=>DX.has(A)||e==="array"&&A==="array"):[]}function S9A(t,e,A){let{gen:i,data:n,opts:o}=t,r=i.let("dataType",(0,Ui._)`typeof ${n}`),s=i.let("coerced",(0,Ui._)`undefined`);o.coerceTypes==="array"&&i.if((0,Ui._)`${r} == 'object' && Array.isArray(${n}) && ${n}.length == 1`,()=>i.assign(n,(0,Ui._)`${n}[0]`).assign(r,(0,Ui._)`typeof ${n}`).if(yR(e,n,o.strictNumbers),()=>i.assign(s,n))),i.if((0,Ui._)`${s} !== undefined`);for(let c of A)(DX.has(c)||c==="array"&&o.coerceTypes==="array")&&a(c);i.else(),vR(t),i.endIf(),i.if((0,Ui._)`${s} !== undefined`,()=>{i.assign(n,s),R9A(t,s)});function a(c){switch(c){case"string":i.elseIf((0,Ui._)`${r} == "number" || ${r} == "boolean"`).assign(s,(0,Ui._)`"" + ${n}`).elseIf((0,Ui._)`${n} === null`).assign(s,(0,Ui._)`""`);return;case"number":i.elseIf((0,Ui._)`${r} == "boolean" || ${n} === null
              || (${r} == "string" && ${n} && ${n} == +${n})`).assign(s,(0,Ui._)`+${n}`);return;case"integer":i.elseIf((0,Ui._)`${r} === "boolean" || ${n} === null
              || (${r} === "string" && ${n} && ${n} == +${n} && !(${n} % 1))`).assign(s,(0,Ui._)`+${n}`);return;case"boolean":i.elseIf((0,Ui._)`${n} === "false" || ${n} === 0 || ${n} === null`).assign(s,!1).elseIf((0,Ui._)`${n} === "true" || ${n} === 1`).assign(s,!0);return;case"null":i.elseIf((0,Ui._)`${n} === "" || ${n} === 0 || ${n} === false`),i.assign(s,null);return;case"array":i.elseIf((0,Ui._)`${r} === "string" || ${r} === "number"
              || ${r} === "boolean" || ${n} === null`).assign(s,(0,Ui._)`[${n}]`)}}}function R9A({gen:t,parentData:e,parentDataProperty:A},i){t.if((0,Ui._)`${e} !== undefined`,()=>t.assign((0,Ui._)`${e}[${A}]`,i))}function DR(t,e,A,i=IE.Correct){let n=i===IE.Correct?Ui.operators.EQ:Ui.operators.NEQ,o;switch(t){case"null":return(0,Ui._)`${e} ${n} null`;case"array":o=(0,Ui._)`Array.isArray(${e})`;break;case"object":o=(0,Ui._)`${e} && typeof ${e} == "object" && !Array.isArray(${e})`;break;case"integer":o=r((0,Ui._)`!(${e} % 1) && !isNaN(${e})`);break;case"number":o=r();break;default:return(0,Ui._)`typeof ${e} ${n} ${t}`}return i===IE.Correct?o:(0,Ui.not)(o);function r(s=Ui.nil){return(0,Ui.and)((0,Ui._)`typeof ${e} == "number"`,s,A?(0,Ui._)`isFinite(${e})`:Ui.nil)}}xs.checkDataType=DR;function yR(t,e,A,i){if(t.length===1)return DR(t[0],e,A,i);let n,o=(0,pX.toHash)(t);if(o.array&&o.object){let r=(0,Ui._)`typeof ${e} != "object"`;n=o.null?r:(0,Ui._)`!${e} || ${r}`,delete o.null,delete o.array,delete o.object}else n=Ui.nil;o.number&&delete o.integer;for(let r in o)n=(0,Ui.and)(n,DR(r,e,A,i));return n}xs.checkDataTypes=yR;var x9A={message:({schema:t})=>`must be ${t}`,params:({schema:t,schemaValue:e})=>typeof t=="string"?(0,Ui._)`{type: ${t}}`:(0,Ui._)`{type: ${e}}`};function vR(t){let e=L9A(t);(0,v9A.reportError)(e,x9A)}xs.reportTypeError=vR;function L9A(t){let{gen:e,data:A,schema:i}=t,n=(0,pX.schemaRefOrVal)(t,i,"type");return{gen:e,keyword:"type",data:A,schema:i.type,schemaCode:n,schemaValue:n,parentSchema:i,params:{},it:t}}});var vX=Je(S5=>{"use strict";Object.defineProperty(S5,"__esModule",{value:!0});S5.assignDefaults=void 0;var CE=An(),F9A=bn();function N9A(t,e){let{properties:A,items:i}=t.schema;if(e==="object"&&A)for(let n in A)yX(t,n,A[n].default);else e==="array"&&Array.isArray(i)&&i.forEach((n,o)=>yX(t,o,n.default))}S5.assignDefaults=N9A;function yX(t,e,A){let{gen:i,compositeRule:n,data:o,opts:r}=t;if(A===void 0)return;let s=(0,CE._)`${o}${(0,CE.getProperty)(e)}`;if(n){(0,F9A.checkStrictMode)(t,`default is ignored for: ${s}`);return}let a=(0,CE._)`${s} === undefined`;r.useDefaults==="empty"&&(a=(0,CE._)`${a} || ${s} === null || ${s} === ""`),i.if(a,(0,CE._)`${s} = ${(0,CE.stringify)(A)}`)}});var _c=Je(ro=>{"use strict";Object.defineProperty(ro,"__esModule",{value:!0});ro.validateUnion=ro.validateArray=ro.usePattern=ro.callValidateCode=ro.schemaProperties=ro.allSchemaProperties=ro.noPropertyInData=ro.propertyInData=ro.isOwnProperty=ro.hasPropFunc=ro.reportMissingProp=ro.checkMissingProp=ro.checkReportMissingProp=void 0;var zo=An(),bR=bn(),d1=G0(),_9A=bn();function G9A(t,e){let{gen:A,data:i,it:n}=t;A.if(kR(A,i,e,n.opts.ownProperties),()=>{t.setParams({missingProperty:(0,zo._)`${e}`},!0),t.error()})}ro.checkReportMissingProp=G9A;function U9A({gen:t,data:e,it:{opts:A}},i,n){return(0,zo.or)(...i.map(o=>(0,zo.and)(kR(t,e,o,A.ownProperties),(0,zo._)`${n} = ${o}`)))}ro.checkMissingProp=U9A;function K9A(t,e){t.setParams({missingProperty:e},!0),t.error()}ro.reportMissingProp=K9A;function bX(t){return t.scopeValue("func",{ref:Object.prototype.hasOwnProperty,code:(0,zo._)`Object.prototype.hasOwnProperty`})}ro.hasPropFunc=bX;function MR(t,e,A){return(0,zo._)`${bX(t)}.call(${e}, ${A})`}ro.isOwnProperty=MR;function Y9A(t,e,A,i){let n=(0,zo._)`${e}${(0,zo.getProperty)(A)} !== undefined`;return i?(0,zo._)`${n} && ${MR(t,e,A)}`:n}ro.propertyInData=Y9A;function kR(t,e,A,i){let n=(0,zo._)`${e}${(0,zo.getProperty)(A)} === undefined`;return i?(0,zo.or)(n,(0,zo.not)(MR(t,e,A))):n}ro.noPropertyInData=kR;function MX(t){return t?Object.keys(t).filter(e=>e!=="__proto__"):[]}ro.allSchemaProperties=MX;function J9A(t,e){return MX(e).filter(A=>!(0,bR.alwaysValidSchema)(t,e[A]))}ro.schemaProperties=J9A;function T9A({schemaCode:t,data:e,it:{gen:A,topSchemaRef:i,schemaPath:n,errorPath:o},it:r},s,a,c){let l=c?(0,zo._)`${t}, ${e}, ${i}${n}`:e,I=[[d1.default.instancePath,(0,zo.strConcat)(d1.default.instancePath,o)],[d1.default.parentData,r.parentData],[d1.default.parentDataProperty,r.parentDataProperty],[d1.default.rootData,d1.default.rootData]];r.opts.dynamicRef&&I.push([d1.default.dynamicAnchors,d1.default.dynamicAnchors]);let C=(0,zo._)`${l}, ${A.object(...I)}`;return a!==zo.nil?(0,zo._)`${s}.call(${a}, ${C})`:(0,zo._)`${s}(${C})`}ro.callValidateCode=T9A;var H9A=(0,zo._)`new RegExp`;function z9A({gen:t,it:{opts:e}},A){let i=e.unicodeRegExp?"u":"",{regExp:n}=e.code,o=n(A,i);return t.scopeValue("pattern",{key:o.toString(),ref:o,code:(0,zo._)`${n.code==="new RegExp"?H9A:(0,_9A.useFunc)(t,n)}(${A}, ${i})`})}ro.usePattern=z9A;function O9A(t){let{gen:e,data:A,keyword:i,it:n}=t,o=e.name("valid");if(n.allErrors){let s=e.let("valid",!0);return r(()=>e.assign(s,!1)),s}return e.var(o,!0),r(()=>e.break()),o;function r(s){let a=e.const("len",(0,zo._)`${A}.length`);e.forRange("i",0,a,c=>{t.subschema({keyword:i,dataProp:c,dataPropType:bR.Type.Num},o),e.if((0,zo.not)(o),s)})}}ro.validateArray=O9A;function P9A(t){let{gen:e,schema:A,keyword:i,it:n}=t;if(!Array.isArray(A))throw new Error("ajv implementation error");if(A.some(a=>(0,bR.alwaysValidSchema)(n,a))&&!n.opts.unevaluated)return;let r=e.let("valid",!1),s=e.name("_valid");e.block(()=>A.forEach((a,c)=>{let l=t.subschema({keyword:i,schemaProp:c,compositeRule:!0},s);e.assign(r,(0,zo._)`${r} || ${s}`),t.mergeValidEvaluated(l,s)||e.if((0,zo.not)(r))})),t.result(r,()=>t.reset(),()=>t.error(!0))}ro.validateUnion=P9A});var RX=Je(pg=>{"use strict";Object.defineProperty(pg,"__esModule",{value:!0});pg.validateKeywordUsage=pg.validSchemaType=pg.funcKeywordCode=pg.macroKeywordCode=void 0;var Ws=An(),BC=G0(),j9A=_c(),q9A=Q4();function V9A(t,e){let{gen:A,keyword:i,schema:n,parentSchema:o,it:r}=t,s=e.macro.call(r.self,n,o,r),a=SX(A,i,s);r.opts.validateSchema!==!1&&r.self.validateSchema(s,!0);let c=A.name("valid");t.subschema({schema:s,schemaPath:Ws.nil,errSchemaPath:`${r.errSchemaPath}/${i}`,topSchemaRef:a,compositeRule:!0},c),t.pass(c,()=>t.error(!0))}pg.macroKeywordCode=V9A;function Z9A(t,e){var A;let{gen:i,keyword:n,schema:o,parentSchema:r,$data:s,it:a}=t;X9A(a,e);let c=!s&&e.compile?e.compile.call(a.self,o,r,a):e.validate,l=SX(i,n,c),I=i.let("valid");t.block$data(I,C),t.ok((A=e.valid)!==null&&A!==void 0?A:I);function C(){if(e.errors===!1)E(),e.modifying&&kX(t),h(()=>t.error());else{let u=e.async?d():B();e.modifying&&kX(t),h(()=>W9A(t,u))}}function d(){let u=i.let("ruleErrs",null);return i.try(()=>E((0,Ws._)`await `),D=>i.assign(I,!1).if((0,Ws._)`${D} instanceof ${a.ValidationError}`,()=>i.assign(u,(0,Ws._)`${D}.errors`),()=>i.throw(D))),u}function B(){let u=(0,Ws._)`${l}.errors`;return i.assign(u,null),E(Ws.nil),u}function E(u=e.async?(0,Ws._)`await `:Ws.nil){let D=a.opts.passContext?BC.default.this:BC.default.self,L=!("compile"in e&&!s||e.schema===!1);i.assign(I,(0,Ws._)`${u}${(0,j9A.callValidateCode)(t,l,D,L)}`,e.modifying)}function h(u){var D;i.if((0,Ws.not)((D=e.valid)!==null&&D!==void 0?D:I),u)}}pg.funcKeywordCode=Z9A;function kX(t){let{gen:e,data:A,it:i}=t;e.if(i.parentData,()=>e.assign(A,(0,Ws._)`${i.parentData}[${i.parentDataProperty}]`))}function W9A(t,e){let{gen:A}=t;A.if((0,Ws._)`Array.isArray(${e})`,()=>{A.assign(BC.default.vErrors,(0,Ws._)`${BC.default.vErrors} === null ? ${e} : ${BC.default.vErrors}.concat(${e})`).assign(BC.default.errors,(0,Ws._)`${BC.default.vErrors}.length`),(0,q9A.extendErrors)(t)},()=>t.error())}function X9A({schemaEnv:t},e){if(e.async&&!t.$async)throw new Error("async keyword in sync schema")}function SX(t,e,A){if(A===void 0)throw new Error(`keyword "${e}" failed to compile`);return t.scopeValue("keyword",typeof A=="function"?{ref:A}:{ref:A,code:(0,Ws.stringify)(A)})}function $9A(t,e,A=!1){return!e.length||e.some(i=>i==="array"?Array.isArray(t):i==="object"?t&&typeof t=="object"&&!Array.isArray(t):typeof t==i||A&&typeof t>"u")}pg.validSchemaType=$9A;function AbA({schema:t,opts:e,self:A,errSchemaPath:i},n,o){if(Array.isArray(n.keyword)?!n.keyword.includes(o):n.keyword!==o)throw new Error("ajv implementation error");let r=n.dependencies;if(r?.some(s=>!Object.prototype.hasOwnProperty.call(t,s)))throw new Error(`parent schema must have dependencies of ${o}: ${r.join(",")}`);if(n.validateSchema&&!n.validateSchema(t[o])){let a=`keyword "${o}" value is invalid at path "${i}": `+A.errorsText(n.validateSchema.errors);if(e.validateSchema==="log")A.logger.error(a);else throw new Error(a)}}pg.validateKeywordUsage=AbA});var LX=Je(B1=>{"use strict";Object.defineProperty(B1,"__esModule",{value:!0});B1.extendSubschemaMode=B1.extendSubschemaData=B1.getSubschema=void 0;var wg=An(),xX=bn();function ebA(t,{keyword:e,schemaProp:A,schema:i,schemaPath:n,errSchemaPath:o,topSchemaRef:r}){if(e!==void 0&&i!==void 0)throw new Error('both "keyword" and "schema" passed, only one allowed');if(e!==void 0){let s=t.schema[e];return A===void 0?{schema:s,schemaPath:(0,wg._)`${t.schemaPath}${(0,wg.getProperty)(e)}`,errSchemaPath:`${t.errSchemaPath}/${e}`}:{schema:s[A],schemaPath:(0,wg._)`${t.schemaPath}${(0,wg.getProperty)(e)}${(0,wg.getProperty)(A)}`,errSchemaPath:`${t.errSchemaPath}/${e}/${(0,xX.escapeFragment)(A)}`}}if(i!==void 0){if(n===void 0||o===void 0||r===void 0)throw new Error('"schemaPath", "errSchemaPath" and "topSchemaRef" are required with "schema"');return{schema:i,schemaPath:n,topSchemaRef:r,errSchemaPath:o}}throw new Error('either "keyword" or "schema" must be passed')}B1.getSubschema=ebA;function tbA(t,e,{dataProp:A,dataPropType:i,data:n,dataTypes:o,propertyName:r}){if(n!==void 0&&A!==void 0)throw new Error('both "data" and "dataProp" passed, only one allowed');let{gen:s}=e;if(A!==void 0){let{errorPath:c,dataPathArr:l,opts:I}=e,C=s.let("data",(0,wg._)`${e.data}${(0,wg.getProperty)(A)}`,!0);a(C),t.errorPath=(0,wg.str)`${c}${(0,xX.getErrorPath)(A,i,I.jsPropertySyntax)}`,t.parentDataProperty=(0,wg._)`${A}`,t.dataPathArr=[...l,t.parentDataProperty]}if(n!==void 0){let c=n instanceof wg.Name?n:s.let("data",n,!0);a(c),r!==void 0&&(t.propertyName=r)}o&&(t.dataTypes=o);function a(c){t.data=c,t.dataLevel=e.dataLevel+1,t.dataTypes=[],e.definedProperties=new Set,t.parentData=e.data,t.dataNames=[...e.dataNames,c]}}B1.extendSubschemaData=tbA;function ibA(t,{jtdDiscriminator:e,jtdMetadata:A,compositeRule:i,createErrors:n,allErrors:o}){i!==void 0&&(t.compositeRule=i),n!==void 0&&(t.createErrors=n),o!==void 0&&(t.allErrors=o),t.jtdDiscriminator=e,t.jtdMetadata=A}B1.extendSubschemaMode=ibA});var SR=Je((lme,FX)=>{"use strict";FX.exports=function t(e,A){if(e===A)return!0;if(e&&A&&typeof e=="object"&&typeof A=="object"){if(e.constructor!==A.constructor)return!1;var i,n,o;if(Array.isArray(e)){if(i=e.length,i!=A.length)return!1;for(n=i;n--!==0;)if(!t(e[n],A[n]))return!1;return!0}if(e.constructor===RegExp)return e.source===A.source&&e.flags===A.flags;if(e.valueOf!==Object.prototype.valueOf)return e.valueOf()===A.valueOf();if(e.toString!==Object.prototype.toString)return e.toString()===A.toString();if(o=Object.keys(e),i=o.length,i!==Object.keys(A).length)return!1;for(n=i;n--!==0;)if(!Object.prototype.hasOwnProperty.call(A,o[n]))return!1;for(n=i;n--!==0;){var r=o[n];if(!t(e[r],A[r]))return!1}return!0}return e!==e&&A!==A}});var _X=Je((gme,NX)=>{"use strict";var E1=NX.exports=function(t,e,A){typeof e=="function"&&(A=e,e={}),A=e.cb||A;var i=typeof A=="function"?A:A.pre||function(){},n=A.post||function(){};R5(e,i,n,t,"",t)};E1.keywords={additionalItems:!0,items:!0,contains:!0,additionalProperties:!0,propertyNames:!0,not:!0,if:!0,then:!0,else:!0};E1.arrayKeywords={items:!0,allOf:!0,anyOf:!0,oneOf:!0};E1.propsKeywords={$defs:!0,definitions:!0,properties:!0,patternProperties:!0,dependencies:!0};E1.skipKeywords={default:!0,enum:!0,const:!0,required:!0,maximum:!0,minimum:!0,exclusiveMaximum:!0,exclusiveMinimum:!0,multipleOf:!0,maxLength:!0,minLength:!0,pattern:!0,format:!0,maxItems:!0,minItems:!0,uniqueItems:!0,maxProperties:!0,minProperties:!0};function R5(t,e,A,i,n,o,r,s,a,c){if(i&&typeof i=="object"&&!Array.isArray(i)){e(i,n,o,r,s,a,c);for(var l in i){var I=i[l];if(Array.isArray(I)){if(l in E1.arrayKeywords)for(var C=0;C<I.length;C++)R5(t,e,A,I[C],n+"/"+l+"/"+C,o,n,l,i,C)}else if(l in E1.propsKeywords){if(I&&typeof I=="object")for(var d in I)R5(t,e,A,I[d],n+"/"+l+"/"+nbA(d),o,n,l,i,d)}else(l in E1.keywords||t.allKeys&&!(l in E1.skipKeywords))&&R5(t,e,A,I,n+"/"+l,o,n,l,i)}A(i,n,o,r,s,a,c)}}function nbA(t){return t.replace(/~/g,"~0").replace(/\//g,"~1")}});var u4=Je(ka=>{"use strict";Object.defineProperty(ka,"__esModule",{value:!0});ka.getSchemaRefs=ka.resolveUrl=ka.normalizeId=ka._getFullPath=ka.getFullPath=ka.inlineRef=void 0;var obA=bn(),rbA=SR(),sbA=_X(),abA=new Set(["type","format","pattern","maxLength","minLength","maxProperties","minProperties","maxItems","minItems","maximum","minimum","uniqueItems","multipleOf","required","enum","const"]);function cbA(t,e=!0){return typeof t=="boolean"?!0:e===!0?!RR(t):e?GX(t)<=e:!1}ka.inlineRef=cbA;var lbA=new Set(["$ref","$recursiveRef","$recursiveAnchor","$dynamicRef","$dynamicAnchor"]);function RR(t){for(let e in t){if(lbA.has(e))return!0;let A=t[e];if(Array.isArray(A)&&A.some(RR)||typeof A=="object"&&RR(A))return!0}return!1}function GX(t){let e=0;for(let A in t){if(A==="$ref")return 1/0;if(e++,!abA.has(A)&&(typeof t[A]=="object"&&(0,obA.eachItem)(t[A],i=>e+=GX(i)),e===1/0))return 1/0}return e}function UX(t,e="",A){A!==!1&&(e=dE(e));let i=t.parse(e);return KX(t,i)}ka.getFullPath=UX;function KX(t,e){return t.serialize(e).split("#")[0]+"#"}ka._getFullPath=KX;var gbA=/#\/?$/;function dE(t){return t?t.replace(gbA,""):""}ka.normalizeId=dE;function IbA(t,e,A){return A=dE(A),t.resolve(e,A)}ka.resolveUrl=IbA;var CbA=/^[a-z_][-a-z0-9._]*$/i;function dbA(t,e){if(typeof t=="boolean")return{};let{schemaId:A,uriResolver:i}=this.opts,n=dE(t[A]||e),o={"":n},r=UX(i,n,!1),s={},a=new Set;return sbA(t,{allKeys:!0},(I,C,d,B)=>{if(B===void 0)return;let E=r+C,h=o[B];typeof I[A]=="string"&&(h=u.call(this,I[A])),D.call(this,I.$anchor),D.call(this,I.$dynamicAnchor),o[C]=h;function u(L){let R=this.opts.uriResolver.resolve;if(L=dE(h?R(h,L):L),a.has(L))throw l(L);a.add(L);let w=this.refs[L];return typeof w=="string"&&(w=this.refs[w]),typeof w=="object"?c(I,w.schema,L):L!==dE(E)&&(L[0]==="#"?(c(I,s[L],L),s[L]=I):this.refs[L]=E),L}function D(L){if(typeof L=="string"){if(!CbA.test(L))throw new Error(`invalid anchor "${L}"`);u.call(this,`#${L}`)}}}),s;function c(I,C,d){if(C!==void 0&&!rbA(I,C))throw l(d)}function l(I){return new Error(`reference "${I}" resolves to more than one schema`)}}ka.getSchemaRefs=dbA});var p4=Je(Q1=>{"use strict";Object.defineProperty(Q1,"__esModule",{value:!0});Q1.getData=Q1.KeywordCxt=Q1.validateFunctionCode=void 0;var zX=uX(),YX=h4(),LR=wR(),x5=h4(),BbA=vX(),m4=RX(),xR=LX(),yt=An(),Bi=G0(),EbA=u4(),U0=bn(),f4=Q4();function QbA(t){if(jX(t)&&(qX(t),PX(t))){fbA(t);return}OX(t,()=>(0,zX.topBoolOrEmptySchema)(t))}Q1.validateFunctionCode=QbA;function OX({gen:t,validateName:e,schema:A,schemaEnv:i,opts:n},o){n.code.es5?t.func(e,(0,yt._)`${Bi.default.data}, ${Bi.default.valCxt}`,i.$async,()=>{t.code((0,yt._)`"use strict"; ${JX(A,n)}`),ubA(t,n),t.code(o)}):t.func(e,(0,yt._)`${Bi.default.data}, ${hbA(n)}`,i.$async,()=>t.code(JX(A,n)).code(o))}function hbA(t){return(0,yt._)`{${Bi.default.instancePath}="", ${Bi.default.parentData}, ${Bi.default.parentDataProperty}, ${Bi.default.rootData}=${Bi.default.data}${t.dynamicRef?(0,yt._)`, ${Bi.default.dynamicAnchors}={}`:yt.nil}}={}`}function ubA(t,e){t.if(Bi.default.valCxt,()=>{t.var(Bi.default.instancePath,(0,yt._)`${Bi.default.valCxt}.${Bi.default.instancePath}`),t.var(Bi.default.parentData,(0,yt._)`${Bi.default.valCxt}.${Bi.default.parentData}`),t.var(Bi.default.parentDataProperty,(0,yt._)`${Bi.default.valCxt}.${Bi.default.parentDataProperty}`),t.var(Bi.default.rootData,(0,yt._)`${Bi.default.valCxt}.${Bi.default.rootData}`),e.dynamicRef&&t.var(Bi.default.dynamicAnchors,(0,yt._)`${Bi.default.valCxt}.${Bi.default.dynamicAnchors}`)},()=>{t.var(Bi.default.instancePath,(0,yt._)`""`),t.var(Bi.default.parentData,(0,yt._)`undefined`),t.var(Bi.default.parentDataProperty,(0,yt._)`undefined`),t.var(Bi.default.rootData,Bi.default.data),e.dynamicRef&&t.var(Bi.default.dynamicAnchors,(0,yt._)`{}`)})}function fbA(t){let{schema:e,opts:A,gen:i}=t;OX(t,()=>{A.$comment&&e.$comment&&ZX(t),ybA(t),i.let(Bi.default.vErrors,null),i.let(Bi.default.errors,0),A.unevaluated&&mbA(t),VX(t),MbA(t)})}function mbA(t){let{gen:e,validateName:A}=t;t.evaluated=e.const("evaluated",(0,yt._)`${A}.evaluated`),e.if((0,yt._)`${t.evaluated}.dynamicProps`,()=>e.assign((0,yt._)`${t.evaluated}.props`,(0,yt._)`undefined`)),e.if((0,yt._)`${t.evaluated}.dynamicItems`,()=>e.assign((0,yt._)`${t.evaluated}.items`,(0,yt._)`undefined`))}function JX(t,e){let A=typeof t=="object"&&t[e.schemaId];return A&&(e.code.source||e.code.process)?(0,yt._)`/*# sourceURL=${A} */`:yt.nil}function pbA(t,e){if(jX(t)&&(qX(t),PX(t))){wbA(t,e);return}(0,zX.boolOrEmptySchema)(t,e)}function PX({schema:t,self:e}){if(typeof t=="boolean")return!t;for(let A in t)if(e.RULES.all[A])return!0;return!1}function jX(t){return typeof t.schema!="boolean"}function wbA(t,e){let{schema:A,gen:i,opts:n}=t;n.$comment&&A.$comment&&ZX(t),vbA(t),bbA(t);let o=i.const("_errs",Bi.default.errors);VX(t,o),i.var(e,(0,yt._)`${o} === ${Bi.default.errors}`)}function qX(t){(0,U0.checkUnknownRules)(t),DbA(t)}function VX(t,e){if(t.opts.jtd)return TX(t,[],!1,e);let A=(0,YX.getSchemaTypes)(t.schema),i=(0,YX.coerceAndCheckDataType)(t,A);TX(t,A,!i,e)}function DbA(t){let{schema:e,errSchemaPath:A,opts:i,self:n}=t;e.$ref&&i.ignoreKeywordsWithRef&&(0,U0.schemaHasRulesButRef)(e,n.RULES)&&n.logger.warn(`$ref: keywords ignored in schema at path "${A}"`)}function ybA(t){let{schema:e,opts:A}=t;e.default!==void 0&&A.useDefaults&&A.strictSchema&&(0,U0.checkStrictMode)(t,"default is ignored in the schema root")}function vbA(t){let e=t.schema[t.opts.schemaId];e&&(t.baseId=(0,EbA.resolveUrl)(t.opts.uriResolver,t.baseId,e))}function bbA(t){if(t.schema.$async&&!t.schemaEnv.$async)throw new Error("async schema in sync schema")}function ZX({gen:t,schemaEnv:e,schema:A,errSchemaPath:i,opts:n}){let o=A.$comment;if(n.$comment===!0)t.code((0,yt._)`${Bi.default.self}.logger.log(${o})`);else if(typeof n.$comment=="function"){let r=(0,yt.str)`${i}/$comment`,s=t.scopeValue("root",{ref:e.root});t.code((0,yt._)`${Bi.default.self}.opts.$comment(${o}, ${r}, ${s}.schema)`)}}function MbA(t){let{gen:e,schemaEnv:A,validateName:i,ValidationError:n,opts:o}=t;A.$async?e.if((0,yt._)`${Bi.default.errors} === 0`,()=>e.return(Bi.default.data),()=>e.throw((0,yt._)`new ${n}(${Bi.default.vErrors})`)):(e.assign((0,yt._)`${i}.errors`,Bi.default.vErrors),o.unevaluated&&kbA(t),e.return((0,yt._)`${Bi.default.errors} === 0`))}function kbA({gen:t,evaluated:e,props:A,items:i}){A instanceof yt.Name&&t.assign((0,yt._)`${e}.props`,A),i instanceof yt.Name&&t.assign((0,yt._)`${e}.items`,i)}function TX(t,e,A,i){let{gen:n,schema:o,data:r,allErrors:s,opts:a,self:c}=t,{RULES:l}=c;if(o.$ref&&(a.ignoreKeywordsWithRef||!(0,U0.schemaHasRulesButRef)(o,l))){n.block(()=>XX(t,"$ref",l.all.$ref.definition));return}a.jtd||SbA(t,e),n.block(()=>{for(let C of l.rules)I(C);I(l.post)});function I(C){(0,LR.shouldUseGroup)(o,C)&&(C.type?(n.if((0,x5.checkDataType)(C.type,r,a.strictNumbers)),HX(t,C),e.length===1&&e[0]===C.type&&A&&(n.else(),(0,x5.reportTypeError)(t)),n.endIf()):HX(t,C),s||n.if((0,yt._)`${Bi.default.errors} === ${i||0}`))}}function HX(t,e){let{gen:A,schema:i,opts:{useDefaults:n}}=t;n&&(0,BbA.assignDefaults)(t,e.type),A.block(()=>{for(let o of e.rules)(0,LR.shouldUseRule)(i,o)&&XX(t,o.keyword,o.definition,e.type)})}function SbA(t,e){t.schemaEnv.meta||!t.opts.strictTypes||(RbA(t,e),t.opts.allowUnionTypes||xbA(t,e),LbA(t,t.dataTypes))}function RbA(t,e){if(e.length){if(!t.dataTypes.length){t.dataTypes=e;return}e.forEach(A=>{WX(t.dataTypes,A)||FR(t,`type "${A}" not allowed by context "${t.dataTypes.join(",")}"`)}),NbA(t,e)}}function xbA(t,e){e.length>1&&!(e.length===2&&e.includes("null"))&&FR(t,"use allowUnionTypes to allow union type keyword")}function LbA(t,e){let A=t.self.RULES.all;for(let i in A){let n=A[i];if(typeof n=="object"&&(0,LR.shouldUseRule)(t.schema,n)){let{type:o}=n.definition;o.length&&!o.some(r=>FbA(e,r))&&FR(t,`missing type "${o.join(",")}" for keyword "${i}"`)}}}function FbA(t,e){return t.includes(e)||e==="number"&&t.includes("integer")}function WX(t,e){return t.includes(e)||e==="integer"&&t.includes("number")}function NbA(t,e){let A=[];for(let i of t.dataTypes)WX(e,i)?A.push(i):e.includes("integer")&&i==="number"&&A.push("integer");t.dataTypes=A}function FR(t,e){let A=t.schemaEnv.baseId+t.errSchemaPath;e+=` at "${A}" (strictTypes)`,(0,U0.checkStrictMode)(t,e,t.opts.strictTypes)}var L5=class{constructor(e,A,i){if((0,m4.validateKeywordUsage)(e,A,i),this.gen=e.gen,this.allErrors=e.allErrors,this.keyword=i,this.data=e.data,this.schema=e.schema[i],this.$data=A.$data&&e.opts.$data&&this.schema&&this.schema.$data,this.schemaValue=(0,U0.schemaRefOrVal)(e,this.schema,i,this.$data),this.schemaType=A.schemaType,this.parentSchema=e.schema,this.params={},this.it=e,this.def=A,this.$data)this.schemaCode=e.gen.const("vSchema",$X(this.$data,e));else if(this.schemaCode=this.schemaValue,!(0,m4.validSchemaType)(this.schema,A.schemaType,A.allowUndefined))throw new Error(`${i} value must be ${JSON.stringify(A.schemaType)}`);("code"in A?A.trackErrors:A.errors!==!1)&&(this.errsCount=e.gen.const("_errs",Bi.default.errors))}result(e,A,i){this.failResult((0,yt.not)(e),A,i)}failResult(e,A,i){this.gen.if(e),i?i():this.error(),A?(this.gen.else(),A(),this.allErrors&&this.gen.endIf()):this.allErrors?this.gen.endIf():this.gen.else()}pass(e,A){this.failResult((0,yt.not)(e),void 0,A)}fail(e){if(e===void 0){this.error(),this.allErrors||this.gen.if(!1);return}this.gen.if(e),this.error(),this.allErrors?this.gen.endIf():this.gen.else()}fail$data(e){if(!this.$data)return this.fail(e);let{schemaCode:A}=this;this.fail((0,yt._)`${A} !== undefined && (${(0,yt.or)(this.invalid$data(),e)})`)}error(e,A,i){if(A){this.setParams(A),this._error(e,i),this.setParams({});return}this._error(e,i)}_error(e,A){(e?f4.reportExtraError:f4.reportError)(this,this.def.error,A)}$dataError(){(0,f4.reportError)(this,this.def.$dataError||f4.keyword$DataError)}reset(){if(this.errsCount===void 0)throw new Error('add "trackErrors" to keyword definition');(0,f4.resetErrorsCount)(this.gen,this.errsCount)}ok(e){this.allErrors||this.gen.if(e)}setParams(e,A){A?Object.assign(this.params,e):this.params=e}block$data(e,A,i=yt.nil){this.gen.block(()=>{this.check$data(e,i),A()})}check$data(e=yt.nil,A=yt.nil){if(!this.$data)return;let{gen:i,schemaCode:n,schemaType:o,def:r}=this;i.if((0,yt.or)((0,yt._)`${n} === undefined`,A)),e!==yt.nil&&i.assign(e,!0),(o.length||r.validateSchema)&&(i.elseIf(this.invalid$data()),this.$dataError(),e!==yt.nil&&i.assign(e,!1)),i.else()}invalid$data(){let{gen:e,schemaCode:A,schemaType:i,def:n,it:o}=this;return(0,yt.or)(r(),s());function r(){if(i.length){if(!(A instanceof yt.Name))throw new Error("ajv implementation error");let a=Array.isArray(i)?i:[i];return(0,yt._)`${(0,x5.checkDataTypes)(a,A,o.opts.strictNumbers,x5.DataType.Wrong)}`}return yt.nil}function s(){if(n.validateSchema){let a=e.scopeValue("validate$data",{ref:n.validateSchema});return(0,yt._)`!${a}(${A})`}return yt.nil}}subschema(e,A){let i=(0,xR.getSubschema)(this.it,e);(0,xR.extendSubschemaData)(i,this.it,e),(0,xR.extendSubschemaMode)(i,e);let n=Ye(rA(rA({},this.it),i),{items:void 0,props:void 0});return pbA(n,A),n}mergeEvaluated(e,A){let{it:i,gen:n}=this;i.opts.unevaluated&&(i.props!==!0&&e.props!==void 0&&(i.props=U0.mergeEvaluated.props(n,e.props,i.props,A)),i.items!==!0&&e.items!==void 0&&(i.items=U0.mergeEvaluated.items(n,e.items,i.items,A)))}mergeValidEvaluated(e,A){let{it:i,gen:n}=this;if(i.opts.unevaluated&&(i.props!==!0||i.items!==!0))return n.if(A,()=>this.mergeEvaluated(e,yt.Name)),!0}};Q1.KeywordCxt=L5;function XX(t,e,A,i){let n=new L5(t,A,e);"code"in A?A.code(n,i):n.$data&&A.validate?(0,m4.funcKeywordCode)(n,A):"macro"in A?(0,m4.macroKeywordCode)(n,A):(A.compile||A.validate)&&(0,m4.funcKeywordCode)(n,A)}var _bA=/^\/(?:[^~]|~0|~1)*$/,GbA=/^([0-9]+)(#|\/(?:[^~]|~0|~1)*)?$/;function $X(t,{dataLevel:e,dataNames:A,dataPathArr:i}){let n,o;if(t==="")return Bi.default.rootData;if(t[0]==="/"){if(!_bA.test(t))throw new Error(`Invalid JSON-pointer: ${t}`);n=t,o=Bi.default.rootData}else{let c=GbA.exec(t);if(!c)throw new Error(`Invalid JSON-pointer: ${t}`);let l=+c[1];if(n=c[2],n==="#"){if(l>=e)throw new Error(a("property/index",l));return i[e-l]}if(l>e)throw new Error(a("data",l));if(o=A[e-l],!n)return o}let r=o,s=n.split("/");for(let c of s)c&&(o=(0,yt._)`${o}${(0,yt.getProperty)((0,U0.unescapeJsonPointer)(c))}`,r=(0,yt._)`${r} && ${o}`);return r;function a(c,l){return`Cannot access ${c} ${l} levels up, current level is ${e}`}}Q1.getData=$X});var F5=Je(_R=>{"use strict";Object.defineProperty(_R,"__esModule",{value:!0});var NR=class extends Error{constructor(e){super("validation failed"),this.errors=e,this.ajv=this.validation=!0}};_R.default=NR});var w4=Je(KR=>{"use strict";Object.defineProperty(KR,"__esModule",{value:!0});var GR=u4(),UR=class extends Error{constructor(e,A,i,n){super(n||`can't resolve reference ${i} from id ${A}`),this.missingRef=(0,GR.resolveUrl)(e,A,i),this.missingSchema=(0,GR.normalizeId)((0,GR.getFullPath)(e,this.missingRef))}};KR.default=UR});var _5=Je(Gc=>{"use strict";Object.defineProperty(Gc,"__esModule",{value:!0});Gc.resolveSchema=Gc.getCompilingSchema=Gc.resolveRef=Gc.compileSchema=Gc.SchemaEnv=void 0;var fl=An(),UbA=F5(),EC=G0(),ml=u4(),A$=bn(),KbA=p4(),BE=class{constructor(e){var A;this.refs={},this.dynamicAnchors={};let i;typeof e.schema=="object"&&(i=e.schema),this.schema=e.schema,this.schemaId=e.schemaId,this.root=e.root||this,this.baseId=(A=e.baseId)!==null&&A!==void 0?A:(0,ml.normalizeId)(i?.[e.schemaId||"$id"]),this.schemaPath=e.schemaPath,this.localRefs=e.localRefs,this.meta=e.meta,this.$async=i?.$async,this.refs={}}};Gc.SchemaEnv=BE;function JR(t){let e=e$.call(this,t);if(e)return e;let A=(0,ml.getFullPath)(this.opts.uriResolver,t.root.baseId),{es5:i,lines:n}=this.opts.code,{ownProperties:o}=this.opts,r=new fl.CodeGen(this.scope,{es5:i,lines:n,ownProperties:o}),s;t.$async&&(s=r.scopeValue("Error",{ref:UbA.default,code:(0,fl._)`require("ajv/dist/runtime/validation_error").default`}));let a=r.scopeName("validate");t.validateName=a;let c={gen:r,allErrors:this.opts.allErrors,data:EC.default.data,parentData:EC.default.parentData,parentDataProperty:EC.default.parentDataProperty,dataNames:[EC.default.data],dataPathArr:[fl.nil],dataLevel:0,dataTypes:[],definedProperties:new Set,topSchemaRef:r.scopeValue("schema",this.opts.code.source===!0?{ref:t.schema,code:(0,fl.stringify)(t.schema)}:{ref:t.schema}),validateName:a,ValidationError:s,schema:t.schema,schemaEnv:t,rootId:A,baseId:t.baseId||A,schemaPath:fl.nil,errSchemaPath:t.schemaPath||(this.opts.jtd?"":"#"),errorPath:(0,fl._)`""`,opts:this.opts,self:this},l;try{this._compilations.add(t),(0,KbA.validateFunctionCode)(c),r.optimize(this.opts.code.optimize);let I=r.toString();l=`${r.scopeRefs(EC.default.scope)}return ${I}`,this.opts.code.process&&(l=this.opts.code.process(l,t));let d=new Function(`${EC.default.self}`,`${EC.default.scope}`,l)(this,this.scope.get());if(this.scope.value(a,{ref:d}),d.errors=null,d.schema=t.schema,d.schemaEnv=t,t.$async&&(d.$async=!0),this.opts.code.source===!0&&(d.source={validateName:a,validateCode:I,scopeValues:r._values}),this.opts.unevaluated){let{props:B,items:E}=c;d.evaluated={props:B instanceof fl.Name?void 0:B,items:E instanceof fl.Name?void 0:E,dynamicProps:B instanceof fl.Name,dynamicItems:E instanceof fl.Name},d.source&&(d.source.evaluated=(0,fl.stringify)(d.evaluated))}return t.validate=d,t}catch(I){throw delete t.validate,delete t.validateName,l&&this.logger.error("Error compiling schema, function code:",l),I}finally{this._compilations.delete(t)}}Gc.compileSchema=JR;function YbA(t,e,A){var i;A=(0,ml.resolveUrl)(this.opts.uriResolver,e,A);let n=t.refs[A];if(n)return n;let o=HbA.call(this,t,A);if(o===void 0){let r=(i=t.localRefs)===null||i===void 0?void 0:i[A],{schemaId:s}=this.opts;r&&(o=new BE({schema:r,schemaId:s,root:t,baseId:e}))}if(o!==void 0)return t.refs[A]=JbA.call(this,o)}Gc.resolveRef=YbA;function JbA(t){return(0,ml.inlineRef)(t.schema,this.opts.inlineRefs)?t.schema:t.validate?t:JR.call(this,t)}function e$(t){for(let e of this._compilations)if(TbA(e,t))return e}Gc.getCompilingSchema=e$;function TbA(t,e){return t.schema===e.schema&&t.root===e.root&&t.baseId===e.baseId}function HbA(t,e){let A;for(;typeof(A=this.refs[e])=="string";)e=A;return A||this.schemas[e]||N5.call(this,t,e)}function N5(t,e){let A=this.opts.uriResolver.parse(e),i=(0,ml._getFullPath)(this.opts.uriResolver,A),n=(0,ml.getFullPath)(this.opts.uriResolver,t.baseId,void 0);if(Object.keys(t.schema).length>0&&i===n)return YR.call(this,A,t);let o=(0,ml.normalizeId)(i),r=this.refs[o]||this.schemas[o];if(typeof r=="string"){let s=N5.call(this,t,r);return typeof s?.schema!="object"?void 0:YR.call(this,A,s)}if(typeof r?.schema=="object"){if(r.validate||JR.call(this,r),o===(0,ml.normalizeId)(e)){let{schema:s}=r,{schemaId:a}=this.opts,c=s[a];return c&&(n=(0,ml.resolveUrl)(this.opts.uriResolver,n,c)),new BE({schema:s,schemaId:a,root:t,baseId:n})}return YR.call(this,A,r)}}Gc.resolveSchema=N5;var zbA=new Set(["properties","patternProperties","enum","dependencies","definitions"]);function YR(t,{baseId:e,schema:A,root:i}){var n;if(((n=t.fragment)===null||n===void 0?void 0:n[0])!=="/")return;for(let s of t.fragment.slice(1).split("/")){if(typeof A=="boolean")return;let a=A[(0,A$.unescapeFragment)(s)];if(a===void 0)return;A=a;let c=typeof A=="object"&&A[this.opts.schemaId];!zbA.has(s)&&c&&(e=(0,ml.resolveUrl)(this.opts.uriResolver,e,c))}let o;if(typeof A!="boolean"&&A.$ref&&!(0,A$.schemaHasRulesButRef)(A,this.RULES)){let s=(0,ml.resolveUrl)(this.opts.uriResolver,e,A.$ref);o=N5.call(this,i,s)}let{schemaId:r}=this.opts;if(o=o||new BE({schema:A,schemaId:r,root:i,baseId:e}),o.schema!==o.root.schema)return o}});var t$=Je((hme,ObA)=>{ObA.exports={$id:"https://raw.githubusercontent.com/ajv-validator/ajv/master/lib/refs/data.json#",description:"Meta-schema for $data reference (JSON AnySchema extension proposal)",type:"object",required:["$data"],properties:{$data:{type:"string",anyOf:[{format:"relative-json-pointer"},{format:"json-pointer"}]}},additionalProperties:!1}});var n$=Je((ume,i$)=>{"use strict";var PbA={0:0,1:1,2:2,3:3,4:4,5:5,6:6,7:7,8:8,9:9,a:10,A:10,b:11,B:11,c:12,C:12,d:13,D:13,e:14,E:14,f:15,F:15};i$.exports={HEX:PbA}});var I$=Je((fme,g$)=>{"use strict";var{HEX:jbA}=n$(),qbA=/^(?:(?:25[0-5]|2[0-4]\d|1\d{2}|[1-9]\d|\d)\.){3}(?:25[0-5]|2[0-4]\d|1\d{2}|[1-9]\d|\d)$/u;function a$(t){if(l$(t,".")<3)return{host:t,isIPV4:!1};let e=t.match(qbA)||[],[A]=e;return A?{host:ZbA(A,"."),isIPV4:!0}:{host:t,isIPV4:!1}}function TR(t,e=!1){let A="",i=!0;for(let n of t){if(jbA[n]===void 0)return;n!=="0"&&i===!0&&(i=!1),i||(A+=n)}return e&&A.length===0&&(A="0"),A}function VbA(t){let e=0,A={error:!1,address:"",zone:""},i=[],n=[],o=!1,r=!1,s=!1;function a(){if(n.length){if(o===!1){let c=TR(n);if(c!==void 0)i.push(c);else return A.error=!0,!1}n.length=0}return!0}for(let c=0;c<t.length;c++){let l=t[c];if(!(l==="["||l==="]"))if(l===":"){if(r===!0&&(s=!0),!a())break;if(e++,i.push(":"),e>7){A.error=!0;break}c-1>=0&&t[c-1]===":"&&(r=!0);continue}else if(l==="%"){if(!a())break;o=!0}else{n.push(l);continue}}return n.length&&(o?A.zone=n.join(""):s?i.push(n.join("")):i.push(TR(n))),A.address=i.join(""),A}function c$(t){if(l$(t,":")<2)return{host:t,isIPV6:!1};let e=VbA(t);if(e.error)return{host:t,isIPV6:!1};{let A=e.address,i=e.address;return e.zone&&(A+="%"+e.zone,i+="%25"+e.zone),{host:A,escapedHost:i,isIPV6:!0}}}function ZbA(t,e){let A="",i=!0,n=t.length;for(let o=0;o<n;o++){let r=t[o];r==="0"&&i?(o+1<=n&&t[o+1]===e||o+1===n)&&(A+=r,i=!1):(r===e?i=!0:i=!1,A+=r)}return A}function l$(t,e){let A=0;for(let i=0;i<t.length;i++)t[i]===e&&A++;return A}var o$=/^\.\.?\//u,r$=/^\/\.(?:\/|$)/u,s$=/^\/\.\.(?:\/|$)/u,WbA=/^\/?(?:.|\n)*?(?=\/|$)/u;function XbA(t){let e=[];for(;t.length;)if(t.match(o$))t=t.replace(o$,"");else if(t.match(r$))t=t.replace(r$,"/");else if(t.match(s$))t=t.replace(s$,"/"),e.pop();else if(t==="."||t==="..")t="";else{let A=t.match(WbA);if(A){let i=A[0];t=t.slice(i.length),e.push(i)}else throw new Error("Unexpected dot segment condition")}return e.join("")}function $bA(t,e){let A=e!==!0?escape:unescape;return t.scheme!==void 0&&(t.scheme=A(t.scheme)),t.userinfo!==void 0&&(t.userinfo=A(t.userinfo)),t.host!==void 0&&(t.host=A(t.host)),t.path!==void 0&&(t.path=A(t.path)),t.query!==void 0&&(t.query=A(t.query)),t.fragment!==void 0&&(t.fragment=A(t.fragment)),t}function AMA(t){let e=[];if(t.userinfo!==void 0&&(e.push(t.userinfo),e.push("@")),t.host!==void 0){let A=unescape(t.host),i=a$(A);if(i.isIPV4)A=i.host;else{let n=c$(i.host);n.isIPV6===!0?A=`[${n.escapedHost}]`:A=t.host}e.push(A)}return(typeof t.port=="number"||typeof t.port=="string")&&(e.push(":"),e.push(String(t.port))),e.length?e.join(""):void 0}g$.exports={recomposeAuthority:AMA,normalizeComponentEncoding:$bA,removeDotSegments:XbA,normalizeIPv4:a$,normalizeIPv6:c$,stringArrayToHexStripped:TR}});var h$=Je((mme,Q$)=>{"use strict";var eMA=/^[\da-f]{8}-[\da-f]{4}-[\da-f]{4}-[\da-f]{4}-[\da-f]{12}$/iu,tMA=/([\da-z][\d\-a-z]{0,31}):((?:[\w!$'()*+,\-.:;=@]|%[\da-f]{2})+)/iu;function C$(t){return typeof t.secure=="boolean"?t.secure:String(t.scheme).toLowerCase()==="wss"}function d$(t){return t.host||(t.error=t.error||"HTTP URIs must have a host."),t}function B$(t){let e=String(t.scheme).toLowerCase()==="https";return(t.port===(e?443:80)||t.port==="")&&(t.port=void 0),t.path||(t.path="/"),t}function iMA(t){return t.secure=C$(t),t.resourceName=(t.path||"/")+(t.query?"?"+t.query:""),t.path=void 0,t.query=void 0,t}function nMA(t){if((t.port===(C$(t)?443:80)||t.port==="")&&(t.port=void 0),typeof t.secure=="boolean"&&(t.scheme=t.secure?"wss":"ws",t.secure=void 0),t.resourceName){let[e,A]=t.resourceName.split("?");t.path=e&&e!=="/"?e:void 0,t.query=A,t.resourceName=void 0}return t.fragment=void 0,t}function oMA(t,e){if(!t.path)return t.error="URN can not be parsed",t;let A=t.path.match(tMA);if(A){let i=e.scheme||t.scheme||"urn";t.nid=A[1].toLowerCase(),t.nss=A[2];let n=`${i}:${e.nid||t.nid}`,o=HR[n];t.path=void 0,o&&(t=o.parse(t,e))}else t.error=t.error||"URN can not be parsed.";return t}function rMA(t,e){let A=e.scheme||t.scheme||"urn",i=t.nid.toLowerCase(),n=`${A}:${e.nid||i}`,o=HR[n];o&&(t=o.serialize(t,e));let r=t,s=t.nss;return r.path=`${i||e.nid}:${s}`,e.skipEscape=!0,r}function sMA(t,e){let A=t;return A.uuid=A.nss,A.nss=void 0,!e.tolerant&&(!A.uuid||!eMA.test(A.uuid))&&(A.error=A.error||"UUID is not valid."),A}function aMA(t){let e=t;return e.nss=(t.uuid||"").toLowerCase(),e}var E$={scheme:"http",domainHost:!0,parse:d$,serialize:B$},cMA={scheme:"https",domainHost:E$.domainHost,parse:d$,serialize:B$},G5={scheme:"ws",domainHost:!0,parse:iMA,serialize:nMA},lMA={scheme:"wss",domainHost:G5.domainHost,parse:G5.parse,serialize:G5.serialize},gMA={scheme:"urn",parse:oMA,serialize:rMA,skipNormalize:!0},IMA={scheme:"urn:uuid",parse:sMA,serialize:aMA,skipNormalize:!0},HR={http:E$,https:cMA,ws:G5,wss:lMA,urn:gMA,"urn:uuid":IMA};Q$.exports=HR});var f$=Je((pme,K5)=>{"use strict";var{normalizeIPv6:CMA,normalizeIPv4:dMA,removeDotSegments:D4,recomposeAuthority:BMA,normalizeComponentEncoding:U5}=I$(),zR=h$();function EMA(t,e){return typeof t=="string"?t=Dg(K0(t,e),e):typeof t=="object"&&(t=K0(Dg(t,e),e)),t}function QMA(t,e,A){let i=Object.assign({scheme:"null"},A),n=u$(K0(t,i),K0(e,i),i,!0);return Dg(n,Ye(rA({},i),{skipEscape:!0}))}function u$(t,e,A,i){let n={};return i||(t=K0(Dg(t,A),A),e=K0(Dg(e,A),A)),A=A||{},!A.tolerant&&e.scheme?(n.scheme=e.scheme,n.userinfo=e.userinfo,n.host=e.host,n.port=e.port,n.path=D4(e.path||""),n.query=e.query):(e.userinfo!==void 0||e.host!==void 0||e.port!==void 0?(n.userinfo=e.userinfo,n.host=e.host,n.port=e.port,n.path=D4(e.path||""),n.query=e.query):(e.path?(e.path.charAt(0)==="/"?n.path=D4(e.path):((t.userinfo!==void 0||t.host!==void 0||t.port!==void 0)&&!t.path?n.path="/"+e.path:t.path?n.path=t.path.slice(0,t.path.lastIndexOf("/")+1)+e.path:n.path=e.path,n.path=D4(n.path)),n.query=e.query):(n.path=t.path,e.query!==void 0?n.query=e.query:n.query=t.query),n.userinfo=t.userinfo,n.host=t.host,n.port=t.port),n.scheme=t.scheme),n.fragment=e.fragment,n}function hMA(t,e,A){return typeof t=="string"?(t=unescape(t),t=Dg(U5(K0(t,A),!0),Ye(rA({},A),{skipEscape:!0}))):typeof t=="object"&&(t=Dg(U5(t,!0),Ye(rA({},A),{skipEscape:!0}))),typeof e=="string"?(e=unescape(e),e=Dg(U5(K0(e,A),!0),Ye(rA({},A),{skipEscape:!0}))):typeof e=="object"&&(e=Dg(U5(e,!0),Ye(rA({},A),{skipEscape:!0}))),t.toLowerCase()===e.toLowerCase()}function Dg(t,e){let A={host:t.host,scheme:t.scheme,userinfo:t.userinfo,port:t.port,path:t.path,query:t.query,nid:t.nid,nss:t.nss,uuid:t.uuid,fragment:t.fragment,reference:t.reference,resourceName:t.resourceName,secure:t.secure,error:""},i=Object.assign({},e),n=[],o=zR[(i.scheme||A.scheme||"").toLowerCase()];o&&o.serialize&&o.serialize(A,i),A.path!==void 0&&(i.skipEscape?A.path=unescape(A.path):(A.path=escape(A.path),A.scheme!==void 0&&(A.path=A.path.split("%3A").join(":")))),i.reference!=="suffix"&&A.scheme&&n.push(A.scheme,":");let r=BMA(A);if(r!==void 0&&(i.reference!=="suffix"&&n.push("//"),n.push(r),A.path&&A.path.charAt(0)!=="/"&&n.push("/")),A.path!==void 0){let s=A.path;!i.absolutePath&&(!o||!o.absolutePath)&&(s=D4(s)),r===void 0&&(s=s.replace(/^\/\//u,"/%2F")),n.push(s)}return A.query!==void 0&&n.push("?",A.query),A.fragment!==void 0&&n.push("#",A.fragment),n.join("")}var uMA=Array.from({length:127},(t,e)=>/[^!"$&'()*+,\-.;=_`a-z{}~]/u.test(String.fromCharCode(e)));function fMA(t){let e=0;for(let A=0,i=t.length;A<i;++A)if(e=t.charCodeAt(A),e>126||uMA[e])return!0;return!1}var mMA=/^(?:([^#/:?]+):)?(?:\/\/((?:([^#/?@]*)@)?(\[[^#/?\]]+\]|[^#/:?]*)(?::(\d*))?))?([^#?]*)(?:\?([^#]*))?(?:#((?:.|[\n\r])*))?/u;function K0(t,e){let A=Object.assign({},e),i={scheme:void 0,userinfo:void 0,host:"",port:void 0,path:"",query:void 0,fragment:void 0},n=t.indexOf("%")!==-1,o=!1;A.reference==="suffix"&&(t=(A.scheme?A.scheme+":":"")+"//"+t);let r=t.match(mMA);if(r){if(i.scheme=r[1],i.userinfo=r[3],i.host=r[4],i.port=parseInt(r[5],10),i.path=r[6]||"",i.query=r[7],i.fragment=r[8],isNaN(i.port)&&(i.port=r[5]),i.host){let a=dMA(i.host);if(a.isIPV4===!1){let c=CMA(a.host);i.host=c.host.toLowerCase(),o=c.isIPV6}else i.host=a.host,o=!0}i.scheme===void 0&&i.userinfo===void 0&&i.host===void 0&&i.port===void 0&&i.query===void 0&&!i.path?i.reference="same-document":i.scheme===void 0?i.reference="relative":i.fragment===void 0?i.reference="absolute":i.reference="uri",A.reference&&A.reference!=="suffix"&&A.reference!==i.reference&&(i.error=i.error||"URI is not a "+A.reference+" reference.");let s=zR[(A.scheme||i.scheme||"").toLowerCase()];if(!A.unicodeSupport&&(!s||!s.unicodeSupport)&&i.host&&(A.domainHost||s&&s.domainHost)&&o===!1&&fMA(i.host))try{i.host=URL.domainToASCII(i.host.toLowerCase())}catch(a){i.error=i.error||"Host's domain name can not be converted to ASCII: "+a}(!s||s&&!s.skipNormalize)&&(n&&i.scheme!==void 0&&(i.scheme=unescape(i.scheme)),n&&i.host!==void 0&&(i.host=unescape(i.host)),i.path&&(i.path=escape(unescape(i.path))),i.fragment&&(i.fragment=encodeURI(decodeURIComponent(i.fragment)))),s&&s.parse&&s.parse(i,A)}else i.error=i.error||"URI can not be parsed.";return i}var OR={SCHEMES:zR,normalize:EMA,resolve:QMA,resolveComponents:u$,equal:hMA,serialize:Dg,parse:K0};K5.exports=OR;K5.exports.default=OR;K5.exports.fastUri=OR});var p$=Je(PR=>{"use strict";Object.defineProperty(PR,"__esModule",{value:!0});var m$=f$();m$.code='require("ajv/dist/runtime/uri").default';PR.default=m$});var S$=Je(ls=>{"use strict";Object.defineProperty(ls,"__esModule",{value:!0});ls.CodeGen=ls.Name=ls.nil=ls.stringify=ls.str=ls._=ls.KeywordCxt=void 0;var pMA=p4();Object.defineProperty(ls,"KeywordCxt",{enumerable:!0,get:function(){return pMA.KeywordCxt}});var EE=An();Object.defineProperty(ls,"_",{enumerable:!0,get:function(){return EE._}});Object.defineProperty(ls,"str",{enumerable:!0,get:function(){return EE.str}});Object.defineProperty(ls,"stringify",{enumerable:!0,get:function(){return EE.stringify}});Object.defineProperty(ls,"nil",{enumerable:!0,get:function(){return EE.nil}});Object.defineProperty(ls,"Name",{enumerable:!0,get:function(){return EE.Name}});Object.defineProperty(ls,"CodeGen",{enumerable:!0,get:function(){return EE.CodeGen}});var wMA=F5(),b$=w4(),DMA=pR(),y4=_5(),yMA=An(),v4=u4(),Y5=h4(),qR=bn(),w$=t$(),vMA=p$(),M$=(t,e)=>new RegExp(t,e);M$.code="new RegExp";var bMA=["removeAdditional","useDefaults","coerceTypes"],MMA=new Set(["validate","serialize","parse","wrapper","root","schema","keyword","pattern","formats","validate$data","func","obj","Error"]),kMA={errorDataPath:"",format:"`validateFormats: false` can be used instead.",nullable:'"nullable" keyword is supported by default.',jsonPointers:"Deprecated jsPropertySyntax can be used instead.",extendRefs:"Deprecated ignoreKeywordsWithRef can be used instead.",missingRefs:"Pass empty schema with $id that should be ignored to ajv.addSchema.",processCode:"Use option `code: {process: (code, schemaEnv: object) => string}`",sourceCode:"Use option `code: {source: true}`",strictDefaults:"It is default now, see option `strict`.",strictKeywords:"It is default now, see option `strict`.",uniqueItems:'"uniqueItems" keyword is always validated.',unknownFormats:"Disable strict mode or pass `true` to `ajv.addFormat` (or `formats` option).",cache:"Map is used as cache, schema object as key.",serialize:"Map is used as cache, schema object as key.",ajvErrors:"It is default now."},SMA={ignoreKeywordsWithRef:"",jsPropertySyntax:"",unicode:'"minLength"/"maxLength" account for unicode characters by default.'},D$=200;function RMA(t){var e,A,i,n,o,r,s,a,c,l,I,C,d,B,E,h,u,D,L,R,w,_,K,z,U;let H=t.strict,q=(e=t.code)===null||e===void 0?void 0:e.optimize,j=q===!0||q===void 0?1:q||0,gA=(i=(A=t.code)===null||A===void 0?void 0:A.regExp)!==null&&i!==void 0?i:M$,QA=(n=t.uriResolver)!==null&&n!==void 0?n:vMA.default;return{strictSchema:(r=(o=t.strictSchema)!==null&&o!==void 0?o:H)!==null&&r!==void 0?r:!0,strictNumbers:(a=(s=t.strictNumbers)!==null&&s!==void 0?s:H)!==null&&a!==void 0?a:!0,strictTypes:(l=(c=t.strictTypes)!==null&&c!==void 0?c:H)!==null&&l!==void 0?l:"log",strictTuples:(C=(I=t.strictTuples)!==null&&I!==void 0?I:H)!==null&&C!==void 0?C:"log",strictRequired:(B=(d=t.strictRequired)!==null&&d!==void 0?d:H)!==null&&B!==void 0?B:!1,code:t.code?Ye(rA({},t.code),{optimize:j,regExp:gA}):{optimize:j,regExp:gA},loopRequired:(E=t.loopRequired)!==null&&E!==void 0?E:D$,loopEnum:(h=t.loopEnum)!==null&&h!==void 0?h:D$,meta:(u=t.meta)!==null&&u!==void 0?u:!0,messages:(D=t.messages)!==null&&D!==void 0?D:!0,inlineRefs:(L=t.inlineRefs)!==null&&L!==void 0?L:!0,schemaId:(R=t.schemaId)!==null&&R!==void 0?R:"$id",addUsedSchema:(w=t.addUsedSchema)!==null&&w!==void 0?w:!0,validateSchema:(_=t.validateSchema)!==null&&_!==void 0?_:!0,validateFormats:(K=t.validateFormats)!==null&&K!==void 0?K:!0,unicodeRegExp:(z=t.unicodeRegExp)!==null&&z!==void 0?z:!0,int32range:(U=t.int32range)!==null&&U!==void 0?U:!0,uriResolver:QA}}var b4=class{constructor(e={}){this.schemas={},this.refs={},this.formats={},this._compilations=new Set,this._loading={},this._cache=new Map,e=this.opts=rA(rA({},e),RMA(e));let{es5:A,lines:i}=this.opts.code;this.scope=new yMA.ValueScope({scope:{},prefixes:MMA,es5:A,lines:i}),this.logger=GMA(e.logger);let n=e.validateFormats;e.validateFormats=!1,this.RULES=(0,DMA.getRules)(),y$.call(this,kMA,e,"NOT SUPPORTED"),y$.call(this,SMA,e,"DEPRECATED","warn"),this._metaOpts=NMA.call(this),e.formats&&LMA.call(this),this._addVocabularies(),this._addDefaultMetaSchema(),e.keywords&&FMA.call(this,e.keywords),typeof e.meta=="object"&&this.addMetaSchema(e.meta),xMA.call(this),e.validateFormats=n}_addVocabularies(){this.addKeyword("$async")}_addDefaultMetaSchema(){let{$data:e,meta:A,schemaId:i}=this.opts,n=w$;i==="id"&&(n=rA({},w$),n.id=n.$id,delete n.$id),A&&e&&this.addMetaSchema(n,n[i],!1)}defaultMeta(){let{meta:e,schemaId:A}=this.opts;return this.opts.defaultMeta=typeof e=="object"?e[A]||e:void 0}validate(e,A){let i;if(typeof e=="string"){if(i=this.getSchema(e),!i)throw new Error(`no schema with key or ref "${e}"`)}else i=this.compile(e);let n=i(A);return"$async"in i||(this.errors=i.errors),n}compile(e,A){let i=this._addSchema(e,A);return i.validate||this._compileSchemaEnv(i)}compileAsync(e,A){if(typeof this.opts.loadSchema!="function")throw new Error("options.loadSchema should be a function");let{loadSchema:i}=this.opts;return n.call(this,e,A);function n(l,I){return Ao(this,null,function*(){yield o.call(this,l.$schema);let C=this._addSchema(l,I);return C.validate||r.call(this,C)})}function o(l){return Ao(this,null,function*(){l&&!this.getSchema(l)&&(yield n.call(this,{$ref:l},!0))})}function r(l){return Ao(this,null,function*(){try{return this._compileSchemaEnv(l)}catch(I){if(!(I instanceof b$.default))throw I;return s.call(this,I),yield a.call(this,I.missingSchema),r.call(this,l)}})}function s({missingSchema:l,missingRef:I}){if(this.refs[l])throw new Error(`AnySchema ${l} is loaded but ${I} cannot be resolved`)}function a(l){return Ao(this,null,function*(){let I=yield c.call(this,l);this.refs[l]||(yield o.call(this,I.$schema)),this.refs[l]||this.addSchema(I,l,A)})}function c(l){return Ao(this,null,function*(){let I=this._loading[l];if(I)return I;try{return yield this._loading[l]=i(l)}finally{delete this._loading[l]}})}}addSchema(e,A,i,n=this.opts.validateSchema){if(Array.isArray(e)){for(let r of e)this.addSchema(r,void 0,i,n);return this}let o;if(typeof e=="object"){let{schemaId:r}=this.opts;if(o=e[r],o!==void 0&&typeof o!="string")throw new Error(`schema ${r} must be string`)}return A=(0,v4.normalizeId)(A||o),this._checkUnique(A),this.schemas[A]=this._addSchema(e,i,A,n,!0),this}addMetaSchema(e,A,i=this.opts.validateSchema){return this.addSchema(e,A,!0,i),this}validateSchema(e,A){if(typeof e=="boolean")return!0;let i;if(i=e.$schema,i!==void 0&&typeof i!="string")throw new Error("$schema must be a string");if(i=i||this.opts.defaultMeta||this.defaultMeta(),!i)return this.logger.warn("meta-schema not available"),this.errors=null,!0;let n=this.validate(i,e);if(!n&&A){let o="schema is invalid: "+this.errorsText();if(this.opts.validateSchema==="log")this.logger.error(o);else throw new Error(o)}return n}getSchema(e){let A;for(;typeof(A=v$.call(this,e))=="string";)e=A;if(A===void 0){let{schemaId:i}=this.opts,n=new y4.SchemaEnv({schema:{},schemaId:i});if(A=y4.resolveSchema.call(this,n,e),!A)return;this.refs[e]=A}return A.validate||this._compileSchemaEnv(A)}removeSchema(e){if(e instanceof RegExp)return this._removeAllSchemas(this.schemas,e),this._removeAllSchemas(this.refs,e),this;switch(typeof e){case"undefined":return this._removeAllSchemas(this.schemas),this._removeAllSchemas(this.refs),this._cache.clear(),this;case"string":{let A=v$.call(this,e);return typeof A=="object"&&this._cache.delete(A.schema),delete this.schemas[e],delete this.refs[e],this}case"object":{let A=e;this._cache.delete(A);let i=e[this.opts.schemaId];return i&&(i=(0,v4.normalizeId)(i),delete this.schemas[i],delete this.refs[i]),this}default:throw new Error("ajv.removeSchema: invalid parameter")}}addVocabulary(e){for(let A of e)this.addKeyword(A);return this}addKeyword(e,A){let i;if(typeof e=="string")i=e,typeof A=="object"&&(this.logger.warn("these parameters are deprecated, see docs for addKeyword"),A.keyword=i);else if(typeof e=="object"&&A===void 0){if(A=e,i=A.keyword,Array.isArray(i)&&!i.length)throw new Error("addKeywords: keyword must be string or non-empty array")}else throw new Error("invalid addKeywords parameters");if(KMA.call(this,i,A),!A)return(0,qR.eachItem)(i,o=>jR.call(this,o)),this;JMA.call(this,A);let n=Ye(rA({},A),{type:(0,Y5.getJSONTypes)(A.type),schemaType:(0,Y5.getJSONTypes)(A.schemaType)});return(0,qR.eachItem)(i,n.type.length===0?o=>jR.call(this,o,n):o=>n.type.forEach(r=>jR.call(this,o,n,r))),this}getKeyword(e){let A=this.RULES.all[e];return typeof A=="object"?A.definition:!!A}removeKeyword(e){let{RULES:A}=this;delete A.keywords[e],delete A.all[e];for(let i of A.rules){let n=i.rules.findIndex(o=>o.keyword===e);n>=0&&i.rules.splice(n,1)}return this}addFormat(e,A){return typeof A=="string"&&(A=new RegExp(A)),this.formats[e]=A,this}errorsText(e=this.errors,{separator:A=", ",dataVar:i="data"}={}){return!e||e.length===0?"No errors":e.map(n=>`${i}${n.instancePath} ${n.message}`).reduce((n,o)=>n+A+o)}$dataMetaSchema(e,A){let i=this.RULES.all;e=JSON.parse(JSON.stringify(e));for(let n of A){let o=n.split("/").slice(1),r=e;for(let s of o)r=r[s];for(let s in i){let a=i[s];if(typeof a!="object")continue;let{$data:c}=a.definition,l=r[s];c&&l&&(r[s]=k$(l))}}return e}_removeAllSchemas(e,A){for(let i in e){let n=e[i];(!A||A.test(i))&&(typeof n=="string"?delete e[i]:n&&!n.meta&&(this._cache.delete(n.schema),delete e[i]))}}_addSchema(e,A,i,n=this.opts.validateSchema,o=this.opts.addUsedSchema){let r,{schemaId:s}=this.opts;if(typeof e=="object")r=e[s];else{if(this.opts.jtd)throw new Error("schema must be object");if(typeof e!="boolean")throw new Error("schema must be object or boolean")}let a=this._cache.get(e);if(a!==void 0)return a;i=(0,v4.normalizeId)(r||i);let c=v4.getSchemaRefs.call(this,e,i);return a=new y4.SchemaEnv({schema:e,schemaId:s,meta:A,baseId:i,localRefs:c}),this._cache.set(a.schema,a),o&&!i.startsWith("#")&&(i&&this._checkUnique(i),this.refs[i]=a),n&&this.validateSchema(e,!0),a}_checkUnique(e){if(this.schemas[e]||this.refs[e])throw new Error(`schema with key or id "${e}" already exists`)}_compileSchemaEnv(e){if(e.meta?this._compileMetaSchema(e):y4.compileSchema.call(this,e),!e.validate)throw new Error("ajv implementation error");return e.validate}_compileMetaSchema(e){let A=this.opts;this.opts=this._metaOpts;try{y4.compileSchema.call(this,e)}finally{this.opts=A}}};b4.ValidationError=wMA.default;b4.MissingRefError=b$.default;ls.default=b4;function y$(t,e,A,i="error"){for(let n in t){let o=n;o in e&&this.logger[i](`${A}: option ${n}. ${t[o]}`)}}function v$(t){return t=(0,v4.normalizeId)(t),this.schemas[t]||this.refs[t]}function xMA(){let t=this.opts.schemas;if(t)if(Array.isArray(t))this.addSchema(t);else for(let e in t)this.addSchema(t[e],e)}function LMA(){for(let t in this.opts.formats){let e=this.opts.formats[t];e&&this.addFormat(t,e)}}function FMA(t){if(Array.isArray(t)){this.addVocabulary(t);return}this.logger.warn("keywords option as map is deprecated, pass array");for(let e in t){let A=t[e];A.keyword||(A.keyword=e),this.addKeyword(A)}}function NMA(){let t=rA({},this.opts);for(let e of bMA)delete t[e];return t}var _MA={log(){},warn(){},error(){}};function GMA(t){if(t===!1)return _MA;if(t===void 0)return console;if(t.log&&t.warn&&t.error)return t;throw new Error("logger must implement log, warn and error methods")}var UMA=/^[a-z_$][a-z0-9_$:-]*$/i;function KMA(t,e){let{RULES:A}=this;if((0,qR.eachItem)(t,i=>{if(A.keywords[i])throw new Error(`Keyword ${i} is already defined`);if(!UMA.test(i))throw new Error(`Keyword ${i} has invalid name`)}),!!e&&e.$data&&!("code"in e||"validate"in e))throw new Error('$data keyword must have "code" or "validate" function')}function jR(t,e,A){var i;let n=e?.post;if(A&&n)throw new Error('keyword with "post" flag cannot have "type"');let{RULES:o}=this,r=n?o.post:o.rules.find(({type:a})=>a===A);if(r||(r={type:A,rules:[]},o.rules.push(r)),o.keywords[t]=!0,!e)return;let s={keyword:t,definition:Ye(rA({},e),{type:(0,Y5.getJSONTypes)(e.type),schemaType:(0,Y5.getJSONTypes)(e.schemaType)})};e.before?YMA.call(this,r,s,e.before):r.rules.push(s),o.all[t]=s,(i=e.implements)===null||i===void 0||i.forEach(a=>this.addKeyword(a))}function YMA(t,e,A){let i=t.rules.findIndex(n=>n.keyword===A);i>=0?t.rules.splice(i,0,e):(t.rules.push(e),this.logger.warn(`rule ${A} is not defined`))}function JMA(t){let{metaSchema:e}=t;e!==void 0&&(t.$data&&this.opts.$data&&(e=k$(e)),t.validateSchema=this.compile(e,!0))}var TMA={$ref:"https://raw.githubusercontent.com/ajv-validator/ajv/master/lib/refs/data.json#"};function k$(t){return{anyOf:[t,TMA]}}});var R$=Je(VR=>{"use strict";Object.defineProperty(VR,"__esModule",{value:!0});var HMA={keyword:"id",code(){throw new Error('NOT SUPPORTED: keyword "id", use "$id" for schema ID')}};VR.default=HMA});var N$=Je(QC=>{"use strict";Object.defineProperty(QC,"__esModule",{value:!0});QC.callRef=QC.getValidate=void 0;var zMA=w4(),x$=_c(),Sa=An(),QE=G0(),L$=_5(),J5=bn(),OMA={keyword:"$ref",schemaType:"string",code(t){let{gen:e,schema:A,it:i}=t,{baseId:n,schemaEnv:o,validateName:r,opts:s,self:a}=i,{root:c}=o;if((A==="#"||A==="#/")&&n===c.baseId)return I();let l=L$.resolveRef.call(a,c,n,A);if(l===void 0)throw new zMA.default(i.opts.uriResolver,n,A);if(l instanceof L$.SchemaEnv)return C(l);return d(l);function I(){if(o===c)return T5(t,r,o,o.$async);let B=e.scopeValue("root",{ref:c});return T5(t,(0,Sa._)`${B}.validate`,c,c.$async)}function C(B){let E=F$(t,B);T5(t,E,B,B.$async)}function d(B){let E=e.scopeValue("schema",s.code.source===!0?{ref:B,code:(0,Sa.stringify)(B)}:{ref:B}),h=e.name("valid"),u=t.subschema({schema:B,dataTypes:[],schemaPath:Sa.nil,topSchemaRef:E,errSchemaPath:A},h);t.mergeEvaluated(u),t.ok(h)}}};function F$(t,e){let{gen:A}=t;return e.validate?A.scopeValue("validate",{ref:e.validate}):(0,Sa._)`${A.scopeValue("wrapper",{ref:e})}.validate`}QC.getValidate=F$;function T5(t,e,A,i){let{gen:n,it:o}=t,{allErrors:r,schemaEnv:s,opts:a}=o,c=a.passContext?QE.default.this:Sa.nil;i?l():I();function l(){if(!s.$async)throw new Error("async schema referenced by sync schema");let B=n.let("valid");n.try(()=>{n.code((0,Sa._)`await ${(0,x$.callValidateCode)(t,e,c)}`),d(e),r||n.assign(B,!0)},E=>{n.if((0,Sa._)`!(${E} instanceof ${o.ValidationError})`,()=>n.throw(E)),C(E),r||n.assign(B,!1)}),t.ok(B)}function I(){t.result((0,x$.callValidateCode)(t,e,c),()=>d(e),()=>C(e))}function C(B){let E=(0,Sa._)`${B}.errors`;n.assign(QE.default.vErrors,(0,Sa._)`${QE.default.vErrors} === null ? ${E} : ${QE.default.vErrors}.concat(${E})`),n.assign(QE.default.errors,(0,Sa._)`${QE.default.vErrors}.length`)}function d(B){var E;if(!o.opts.unevaluated)return;let h=(E=A?.validate)===null||E===void 0?void 0:E.evaluated;if(o.props!==!0)if(h&&!h.dynamicProps)h.props!==void 0&&(o.props=J5.mergeEvaluated.props(n,h.props,o.props));else{let u=n.var("props",(0,Sa._)`${B}.evaluated.props`);o.props=J5.mergeEvaluated.props(n,u,o.props,Sa.Name)}if(o.items!==!0)if(h&&!h.dynamicItems)h.items!==void 0&&(o.items=J5.mergeEvaluated.items(n,h.items,o.items));else{let u=n.var("items",(0,Sa._)`${B}.evaluated.items`);o.items=J5.mergeEvaluated.items(n,u,o.items,Sa.Name)}}}QC.callRef=T5;QC.default=OMA});var _$=Je(ZR=>{"use strict";Object.defineProperty(ZR,"__esModule",{value:!0});var PMA=R$(),jMA=N$(),qMA=["$schema","$id","$defs","$vocabulary",{keyword:"$comment"},"definitions",PMA.default,jMA.default];ZR.default=qMA});var G$=Je(WR=>{"use strict";Object.defineProperty(WR,"__esModule",{value:!0});var H5=An(),h1=H5.operators,z5={maximum:{okStr:"<=",ok:h1.LTE,fail:h1.GT},minimum:{okStr:">=",ok:h1.GTE,fail:h1.LT},exclusiveMaximum:{okStr:"<",ok:h1.LT,fail:h1.GTE},exclusiveMinimum:{okStr:">",ok:h1.GT,fail:h1.LTE}},VMA={message:({keyword:t,schemaCode:e})=>(0,H5.str)`must be ${z5[t].okStr} ${e}`,params:({keyword:t,schemaCode:e})=>(0,H5._)`{comparison: ${z5[t].okStr}, limit: ${e}}`},ZMA={keyword:Object.keys(z5),type:"number",schemaType:"number",$data:!0,error:VMA,code(t){let{keyword:e,data:A,schemaCode:i}=t;t.fail$data((0,H5._)`${A} ${z5[e].fail} ${i} || isNaN(${A})`)}};WR.default=ZMA});var U$=Je(XR=>{"use strict";Object.defineProperty(XR,"__esModule",{value:!0});var M4=An(),WMA={message:({schemaCode:t})=>(0,M4.str)`must be multiple of ${t}`,params:({schemaCode:t})=>(0,M4._)`{multipleOf: ${t}}`},XMA={keyword:"multipleOf",type:"number",schemaType:"number",$data:!0,error:WMA,code(t){let{gen:e,data:A,schemaCode:i,it:n}=t,o=n.opts.multipleOfPrecision,r=e.let("res"),s=o?(0,M4._)`Math.abs(Math.round(${r}) - ${r}) > 1e-${o}`:(0,M4._)`${r} !== parseInt(${r})`;t.fail$data((0,M4._)`(${i} === 0 || (${r} = ${A}/${i}, ${s}))`)}};XR.default=XMA});var Y$=Je($R=>{"use strict";Object.defineProperty($R,"__esModule",{value:!0});function K$(t){let e=t.length,A=0,i=0,n;for(;i<e;)A++,n=t.charCodeAt(i++),n>=55296&&n<=56319&&i<e&&(n=t.charCodeAt(i),(n&64512)===56320&&i++);return A}$R.default=K$;K$.code='require("ajv/dist/runtime/ucs2length").default'});var J$=Je(Ax=>{"use strict";Object.defineProperty(Ax,"__esModule",{value:!0});var hC=An(),$MA=bn(),AkA=Y$(),ekA={message({keyword:t,schemaCode:e}){let A=t==="maxLength"?"more":"fewer";return(0,hC.str)`must NOT have ${A} than ${e} characters`},params:({schemaCode:t})=>(0,hC._)`{limit: ${t}}`},tkA={keyword:["maxLength","minLength"],type:"string",schemaType:"number",$data:!0,error:ekA,code(t){let{keyword:e,data:A,schemaCode:i,it:n}=t,o=e==="maxLength"?hC.operators.GT:hC.operators.LT,r=n.opts.unicode===!1?(0,hC._)`${A}.length`:(0,hC._)`${(0,$MA.useFunc)(t.gen,AkA.default)}(${A})`;t.fail$data((0,hC._)`${r} ${o} ${i}`)}};Ax.default=tkA});var T$=Je(ex=>{"use strict";Object.defineProperty(ex,"__esModule",{value:!0});var ikA=_c(),O5=An(),nkA={message:({schemaCode:t})=>(0,O5.str)`must match pattern "${t}"`,params:({schemaCode:t})=>(0,O5._)`{pattern: ${t}}`},okA={keyword:"pattern",type:"string",schemaType:"string",$data:!0,error:nkA,code(t){let{data:e,$data:A,schema:i,schemaCode:n,it:o}=t,r=o.opts.unicodeRegExp?"u":"",s=A?(0,O5._)`(new RegExp(${n}, ${r}))`:(0,ikA.usePattern)(t,i);t.fail$data((0,O5._)`!${s}.test(${e})`)}};ex.default=okA});var H$=Je(tx=>{"use strict";Object.defineProperty(tx,"__esModule",{value:!0});var k4=An(),rkA={message({keyword:t,schemaCode:e}){let A=t==="maxProperties"?"more":"fewer";return(0,k4.str)`must NOT have ${A} than ${e} properties`},params:({schemaCode:t})=>(0,k4._)`{limit: ${t}}`},skA={keyword:["maxProperties","minProperties"],type:"object",schemaType:"number",$data:!0,error:rkA,code(t){let{keyword:e,data:A,schemaCode:i}=t,n=e==="maxProperties"?k4.operators.GT:k4.operators.LT;t.fail$data((0,k4._)`Object.keys(${A}).length ${n} ${i}`)}};tx.default=skA});var z$=Je(ix=>{"use strict";Object.defineProperty(ix,"__esModule",{value:!0});var S4=_c(),R4=An(),akA=bn(),ckA={message:({params:{missingProperty:t}})=>(0,R4.str)`must have required property '${t}'`,params:({params:{missingProperty:t}})=>(0,R4._)`{missingProperty: ${t}}`},lkA={keyword:"required",type:"object",schemaType:"array",$data:!0,error:ckA,code(t){let{gen:e,schema:A,schemaCode:i,data:n,$data:o,it:r}=t,{opts:s}=r;if(!o&&A.length===0)return;let a=A.length>=s.loopRequired;if(r.allErrors?c():l(),s.strictRequired){let d=t.parentSchema.properties,{definedProperties:B}=t.it;for(let E of A)if(d?.[E]===void 0&&!B.has(E)){let h=r.schemaEnv.baseId+r.errSchemaPath,u=`required property "${E}" is not defined at "${h}" (strictRequired)`;(0,akA.checkStrictMode)(r,u,r.opts.strictRequired)}}function c(){if(a||o)t.block$data(R4.nil,I);else for(let d of A)(0,S4.checkReportMissingProp)(t,d)}function l(){let d=e.let("missing");if(a||o){let B=e.let("valid",!0);t.block$data(B,()=>C(d,B)),t.ok(B)}else e.if((0,S4.checkMissingProp)(t,A,d)),(0,S4.reportMissingProp)(t,d),e.else()}function I(){e.forOf("prop",i,d=>{t.setParams({missingProperty:d}),e.if((0,S4.noPropertyInData)(e,n,d,s.ownProperties),()=>t.error())})}function C(d,B){t.setParams({missingProperty:d}),e.forOf(d,i,()=>{e.assign(B,(0,S4.propertyInData)(e,n,d,s.ownProperties)),e.if((0,R4.not)(B),()=>{t.error(),e.break()})},R4.nil)}}};ix.default=lkA});var O$=Je(nx=>{"use strict";Object.defineProperty(nx,"__esModule",{value:!0});var x4=An(),gkA={message({keyword:t,schemaCode:e}){let A=t==="maxItems"?"more":"fewer";return(0,x4.str)`must NOT have ${A} than ${e} items`},params:({schemaCode:t})=>(0,x4._)`{limit: ${t}}`},IkA={keyword:["maxItems","minItems"],type:"array",schemaType:"number",$data:!0,error:gkA,code(t){let{keyword:e,data:A,schemaCode:i}=t,n=e==="maxItems"?x4.operators.GT:x4.operators.LT;t.fail$data((0,x4._)`${A}.length ${n} ${i}`)}};nx.default=IkA});var P5=Je(ox=>{"use strict";Object.defineProperty(ox,"__esModule",{value:!0});var P$=SR();P$.code='require("ajv/dist/runtime/equal").default';ox.default=P$});var j$=Je(sx=>{"use strict";Object.defineProperty(sx,"__esModule",{value:!0});var rx=h4(),gs=An(),CkA=bn(),dkA=P5(),BkA={message:({params:{i:t,j:e}})=>(0,gs.str)`must NOT have duplicate items (items ## ${e} and ${t} are identical)`,params:({params:{i:t,j:e}})=>(0,gs._)`{i: ${t}, j: ${e}}`},EkA={keyword:"uniqueItems",type:"array",schemaType:"boolean",$data:!0,error:BkA,code(t){let{gen:e,data:A,$data:i,schema:n,parentSchema:o,schemaCode:r,it:s}=t;if(!i&&!n)return;let a=e.let("valid"),c=o.items?(0,rx.getSchemaTypes)(o.items):[];t.block$data(a,l,(0,gs._)`${r} === false`),t.ok(a);function l(){let B=e.let("i",(0,gs._)`${A}.length`),E=e.let("j");t.setParams({i:B,j:E}),e.assign(a,!0),e.if((0,gs._)`${B} > 1`,()=>(I()?C:d)(B,E))}function I(){return c.length>0&&!c.some(B=>B==="object"||B==="array")}function C(B,E){let h=e.name("item"),u=(0,rx.checkDataTypes)(c,h,s.opts.strictNumbers,rx.DataType.Wrong),D=e.const("indices",(0,gs._)`{}`);e.for((0,gs._)`;${B}--;`,()=>{e.let(h,(0,gs._)`${A}[${B}]`),e.if(u,(0,gs._)`continue`),c.length>1&&e.if((0,gs._)`typeof ${h} == "string"`,(0,gs._)`${h} += "_"`),e.if((0,gs._)`typeof ${D}[${h}] == "number"`,()=>{e.assign(E,(0,gs._)`${D}[${h}]`),t.error(),e.assign(a,!1).break()}).code((0,gs._)`${D}[${h}] = ${B}`)})}function d(B,E){let h=(0,CkA.useFunc)(e,dkA.default),u=e.name("outer");e.label(u).for((0,gs._)`;${B}--;`,()=>e.for((0,gs._)`${E} = ${B}; ${E}--;`,()=>e.if((0,gs._)`${h}(${A}[${B}], ${A}[${E}])`,()=>{t.error(),e.assign(a,!1).break(u)})))}}};sx.default=EkA});var q$=Je(cx=>{"use strict";Object.defineProperty(cx,"__esModule",{value:!0});var ax=An(),QkA=bn(),hkA=P5(),ukA={message:"must be equal to constant",params:({schemaCode:t})=>(0,ax._)`{allowedValue: ${t}}`},fkA={keyword:"const",$data:!0,error:ukA,code(t){let{gen:e,data:A,$data:i,schemaCode:n,schema:o}=t;i||o&&typeof o=="object"?t.fail$data((0,ax._)`!${(0,QkA.useFunc)(e,hkA.default)}(${A}, ${n})`):t.fail((0,ax._)`${o} !== ${A}`)}};cx.default=fkA});var V$=Je(lx=>{"use strict";Object.defineProperty(lx,"__esModule",{value:!0});var L4=An(),mkA=bn(),pkA=P5(),wkA={message:"must be equal to one of the allowed values",params:({schemaCode:t})=>(0,L4._)`{allowedValues: ${t}}`},DkA={keyword:"enum",schemaType:"array",$data:!0,error:wkA,code(t){let{gen:e,data:A,$data:i,schema:n,schemaCode:o,it:r}=t;if(!i&&n.length===0)throw new Error("enum must have non-empty array");let s=n.length>=r.opts.loopEnum,a,c=()=>a??(a=(0,mkA.useFunc)(e,pkA.default)),l;if(s||i)l=e.let("valid"),t.block$data(l,I);else{if(!Array.isArray(n))throw new Error("ajv implementation error");let d=e.const("vSchema",o);l=(0,L4.or)(...n.map((B,E)=>C(d,E)))}t.pass(l);function I(){e.assign(l,!1),e.forOf("v",o,d=>e.if((0,L4._)`${c()}(${A}, ${d})`,()=>e.assign(l,!0).break()))}function C(d,B){let E=n[B];return typeof E=="object"&&E!==null?(0,L4._)`${c()}(${A}, ${d}[${B}])`:(0,L4._)`${A} === ${E}`}}};lx.default=DkA});var Z$=Je(gx=>{"use strict";Object.defineProperty(gx,"__esModule",{value:!0});var ykA=G$(),vkA=U$(),bkA=J$(),MkA=T$(),kkA=H$(),SkA=z$(),RkA=O$(),xkA=j$(),LkA=q$(),FkA=V$(),NkA=[ykA.default,vkA.default,bkA.default,MkA.default,kkA.default,SkA.default,RkA.default,xkA.default,{keyword:"type",schemaType:["string","array"]},{keyword:"nullable",schemaType:"boolean"},LkA.default,FkA.default];gx.default=NkA});var Cx=Je(F4=>{"use strict";Object.defineProperty(F4,"__esModule",{value:!0});F4.validateAdditionalItems=void 0;var uC=An(),Ix=bn(),_kA={message:({params:{len:t}})=>(0,uC.str)`must NOT have more than ${t} items`,params:({params:{len:t}})=>(0,uC._)`{limit: ${t}}`},GkA={keyword:"additionalItems",type:"array",schemaType:["boolean","object"],before:"uniqueItems",error:_kA,code(t){let{parentSchema:e,it:A}=t,{items:i}=e;if(!Array.isArray(i)){(0,Ix.checkStrictMode)(A,'"additionalItems" is ignored when "items" is not an array of schemas');return}W$(t,i)}};function W$(t,e){let{gen:A,schema:i,data:n,keyword:o,it:r}=t;r.items=!0;let s=A.const("len",(0,uC._)`${n}.length`);if(i===!1)t.setParams({len:e.length}),t.pass((0,uC._)`${s} <= ${e.length}`);else if(typeof i=="object"&&!(0,Ix.alwaysValidSchema)(r,i)){let c=A.var("valid",(0,uC._)`${s} <= ${e.length}`);A.if((0,uC.not)(c),()=>a(c)),t.ok(c)}function a(c){A.forRange("i",e.length,s,l=>{t.subschema({keyword:o,dataProp:l,dataPropType:Ix.Type.Num},c),r.allErrors||A.if((0,uC.not)(c),()=>A.break())})}}F4.validateAdditionalItems=W$;F4.default=GkA});var dx=Je(N4=>{"use strict";Object.defineProperty(N4,"__esModule",{value:!0});N4.validateTuple=void 0;var X$=An(),j5=bn(),UkA=_c(),KkA={keyword:"items",type:"array",schemaType:["object","array","boolean"],before:"uniqueItems",code(t){let{schema:e,it:A}=t;if(Array.isArray(e))return $$(t,"additionalItems",e);A.items=!0,!(0,j5.alwaysValidSchema)(A,e)&&t.ok((0,UkA.validateArray)(t))}};function $$(t,e,A=t.schema){let{gen:i,parentSchema:n,data:o,keyword:r,it:s}=t;l(n),s.opts.unevaluated&&A.length&&s.items!==!0&&(s.items=j5.mergeEvaluated.items(i,A.length,s.items));let a=i.name("valid"),c=i.const("len",(0,X$._)`${o}.length`);A.forEach((I,C)=>{(0,j5.alwaysValidSchema)(s,I)||(i.if((0,X$._)`${c} > ${C}`,()=>t.subschema({keyword:r,schemaProp:C,dataProp:C},a)),t.ok(a))});function l(I){let{opts:C,errSchemaPath:d}=s,B=A.length,E=B===I.minItems&&(B===I.maxItems||I[e]===!1);if(C.strictTuples&&!E){let h=`"${r}" is ${B}-tuple, but minItems or maxItems/${e} are not specified or different at path "${d}"`;(0,j5.checkStrictMode)(s,h,C.strictTuples)}}}N4.validateTuple=$$;N4.default=KkA});var AAA=Je(Bx=>{"use strict";Object.defineProperty(Bx,"__esModule",{value:!0});var YkA=dx(),JkA={keyword:"prefixItems",type:"array",schemaType:["array"],before:"uniqueItems",code:t=>(0,YkA.validateTuple)(t,"items")};Bx.default=JkA});var tAA=Je(Ex=>{"use strict";Object.defineProperty(Ex,"__esModule",{value:!0});var eAA=An(),TkA=bn(),HkA=_c(),zkA=Cx(),OkA={message:({params:{len:t}})=>(0,eAA.str)`must NOT have more than ${t} items`,params:({params:{len:t}})=>(0,eAA._)`{limit: ${t}}`},PkA={keyword:"items",type:"array",schemaType:["object","boolean"],before:"uniqueItems",error:OkA,code(t){let{schema:e,parentSchema:A,it:i}=t,{prefixItems:n}=A;i.items=!0,!(0,TkA.alwaysValidSchema)(i,e)&&(n?(0,zkA.validateAdditionalItems)(t,n):t.ok((0,HkA.validateArray)(t)))}};Ex.default=PkA});var iAA=Je(Qx=>{"use strict";Object.defineProperty(Qx,"__esModule",{value:!0});var Uc=An(),q5=bn(),jkA={message:({params:{min:t,max:e}})=>e===void 0?(0,Uc.str)`must contain at least ${t} valid item(s)`:(0,Uc.str)`must contain at least ${t} and no more than ${e} valid item(s)`,params:({params:{min:t,max:e}})=>e===void 0?(0,Uc._)`{minContains: ${t}}`:(0,Uc._)`{minContains: ${t}, maxContains: ${e}}`},qkA={keyword:"contains",type:"array",schemaType:["object","boolean"],before:"uniqueItems",trackErrors:!0,error:jkA,code(t){let{gen:e,schema:A,parentSchema:i,data:n,it:o}=t,r,s,{minContains:a,maxContains:c}=i;o.opts.next?(r=a===void 0?1:a,s=c):r=1;let l=e.const("len",(0,Uc._)`${n}.length`);if(t.setParams({min:r,max:s}),s===void 0&&r===0){(0,q5.checkStrictMode)(o,'"minContains" == 0 without "maxContains": "contains" keyword ignored');return}if(s!==void 0&&r>s){(0,q5.checkStrictMode)(o,'"minContains" > "maxContains" is always invalid'),t.fail();return}if((0,q5.alwaysValidSchema)(o,A)){let E=(0,Uc._)`${l} >= ${r}`;s!==void 0&&(E=(0,Uc._)`${E} && ${l} <= ${s}`),t.pass(E);return}o.items=!0;let I=e.name("valid");s===void 0&&r===1?d(I,()=>e.if(I,()=>e.break())):r===0?(e.let(I,!0),s!==void 0&&e.if((0,Uc._)`${n}.length > 0`,C)):(e.let(I,!1),C()),t.result(I,()=>t.reset());function C(){let E=e.name("_valid"),h=e.let("count",0);d(E,()=>e.if(E,()=>B(h)))}function d(E,h){e.forRange("i",0,l,u=>{t.subschema({keyword:"contains",dataProp:u,dataPropType:q5.Type.Num,compositeRule:!0},E),h()})}function B(E){e.code((0,Uc._)`${E}++`),s===void 0?e.if((0,Uc._)`${E} >= ${r}`,()=>e.assign(I,!0).break()):(e.if((0,Uc._)`${E} > ${s}`,()=>e.assign(I,!1).break()),r===1?e.assign(I,!0):e.if((0,Uc._)`${E} >= ${r}`,()=>e.assign(I,!0)))}}};Qx.default=qkA});var rAA=Je(yg=>{"use strict";Object.defineProperty(yg,"__esModule",{value:!0});yg.validateSchemaDeps=yg.validatePropertyDeps=yg.error=void 0;var hx=An(),VkA=bn(),_4=_c();yg.error={message:({params:{property:t,depsCount:e,deps:A}})=>{let i=e===1?"property":"properties";return(0,hx.str)`must have ${i} ${A} when property ${t} is present`},params:({params:{property:t,depsCount:e,deps:A,missingProperty:i}})=>(0,hx._)`{property: ${t},
    missingProperty: ${i},
    depsCount: ${e},
    deps: ${A}}`};var ZkA={keyword:"dependencies",type:"object",schemaType:"object",error:yg.error,code(t){let[e,A]=WkA(t);nAA(t,e),oAA(t,A)}};function WkA({schema:t}){let e={},A={};for(let i in t){if(i==="__proto__")continue;let n=Array.isArray(t[i])?e:A;n[i]=t[i]}return[e,A]}function nAA(t,e=t.schema){let{gen:A,data:i,it:n}=t;if(Object.keys(e).length===0)return;let o=A.let("missing");for(let r in e){let s=e[r];if(s.length===0)continue;let a=(0,_4.propertyInData)(A,i,r,n.opts.ownProperties);t.setParams({property:r,depsCount:s.length,deps:s.join(", ")}),n.allErrors?A.if(a,()=>{for(let c of s)(0,_4.checkReportMissingProp)(t,c)}):(A.if((0,hx._)`${a} && (${(0,_4.checkMissingProp)(t,s,o)})`),(0,_4.reportMissingProp)(t,o),A.else())}}yg.validatePropertyDeps=nAA;function oAA(t,e=t.schema){let{gen:A,data:i,keyword:n,it:o}=t,r=A.name("valid");for(let s in e)(0,VkA.alwaysValidSchema)(o,e[s])||(A.if((0,_4.propertyInData)(A,i,s,o.opts.ownProperties),()=>{let a=t.subschema({keyword:n,schemaProp:s},r);t.mergeValidEvaluated(a,r)},()=>A.var(r,!0)),t.ok(r))}yg.validateSchemaDeps=oAA;yg.default=ZkA});var aAA=Je(ux=>{"use strict";Object.defineProperty(ux,"__esModule",{value:!0});var sAA=An(),XkA=bn(),$kA={message:"property name must be valid",params:({params:t})=>(0,sAA._)`{propertyName: ${t.propertyName}}`},ASA={keyword:"propertyNames",type:"object",schemaType:["object","boolean"],error:$kA,code(t){let{gen:e,schema:A,data:i,it:n}=t;if((0,XkA.alwaysValidSchema)(n,A))return;let o=e.name("valid");e.forIn("key",i,r=>{t.setParams({propertyName:r}),t.subschema({keyword:"propertyNames",data:r,dataTypes:["string"],propertyName:r,compositeRule:!0},o),e.if((0,sAA.not)(o),()=>{t.error(!0),n.allErrors||e.break()})}),t.ok(o)}};ux.default=ASA});var mx=Je(fx=>{"use strict";Object.defineProperty(fx,"__esModule",{value:!0});var V5=_c(),pl=An(),eSA=G0(),Z5=bn(),tSA={message:"must NOT have additional properties",params:({params:t})=>(0,pl._)`{additionalProperty: ${t.additionalProperty}}`},iSA={keyword:"additionalProperties",type:["object"],schemaType:["boolean","object"],allowUndefined:!0,trackErrors:!0,error:tSA,code(t){let{gen:e,schema:A,parentSchema:i,data:n,errsCount:o,it:r}=t;if(!o)throw new Error("ajv implementation error");let{allErrors:s,opts:a}=r;if(r.props=!0,a.removeAdditional!=="all"&&(0,Z5.alwaysValidSchema)(r,A))return;let c=(0,V5.allSchemaProperties)(i.properties),l=(0,V5.allSchemaProperties)(i.patternProperties);I(),t.ok((0,pl._)`${o} === ${eSA.default.errors}`);function I(){e.forIn("key",n,h=>{!c.length&&!l.length?B(h):e.if(C(h),()=>B(h))})}function C(h){let u;if(c.length>8){let D=(0,Z5.schemaRefOrVal)(r,i.properties,"properties");u=(0,V5.isOwnProperty)(e,D,h)}else c.length?u=(0,pl.or)(...c.map(D=>(0,pl._)`${h} === ${D}`)):u=pl.nil;return l.length&&(u=(0,pl.or)(u,...l.map(D=>(0,pl._)`${(0,V5.usePattern)(t,D)}.test(${h})`))),(0,pl.not)(u)}function d(h){e.code((0,pl._)`delete ${n}[${h}]`)}function B(h){if(a.removeAdditional==="all"||a.removeAdditional&&A===!1){d(h);return}if(A===!1){t.setParams({additionalProperty:h}),t.error(),s||e.break();return}if(typeof A=="object"&&!(0,Z5.alwaysValidSchema)(r,A)){let u=e.name("valid");a.removeAdditional==="failing"?(E(h,u,!1),e.if((0,pl.not)(u),()=>{t.reset(),d(h)})):(E(h,u),s||e.if((0,pl.not)(u),()=>e.break()))}}function E(h,u,D){let L={keyword:"additionalProperties",dataProp:h,dataPropType:Z5.Type.Str};D===!1&&Object.assign(L,{compositeRule:!0,createErrors:!1,allErrors:!1}),t.subschema(L,u)}}};fx.default=iSA});var gAA=Je(wx=>{"use strict";Object.defineProperty(wx,"__esModule",{value:!0});var nSA=p4(),cAA=_c(),px=bn(),lAA=mx(),oSA={keyword:"properties",type:"object",schemaType:"object",code(t){let{gen:e,schema:A,parentSchema:i,data:n,it:o}=t;o.opts.removeAdditional==="all"&&i.additionalProperties===void 0&&lAA.default.code(new nSA.KeywordCxt(o,lAA.default,"additionalProperties"));let r=(0,cAA.allSchemaProperties)(A);for(let I of r)o.definedProperties.add(I);o.opts.unevaluated&&r.length&&o.props!==!0&&(o.props=px.mergeEvaluated.props(e,(0,px.toHash)(r),o.props));let s=r.filter(I=>!(0,px.alwaysValidSchema)(o,A[I]));if(s.length===0)return;let a=e.name("valid");for(let I of s)c(I)?l(I):(e.if((0,cAA.propertyInData)(e,n,I,o.opts.ownProperties)),l(I),o.allErrors||e.else().var(a,!0),e.endIf()),t.it.definedProperties.add(I),t.ok(a);function c(I){return o.opts.useDefaults&&!o.compositeRule&&A[I].default!==void 0}function l(I){t.subschema({keyword:"properties",schemaProp:I,dataProp:I},a)}}};wx.default=oSA});var BAA=Je(Dx=>{"use strict";Object.defineProperty(Dx,"__esModule",{value:!0});var IAA=_c(),W5=An(),CAA=bn(),dAA=bn(),rSA={keyword:"patternProperties",type:"object",schemaType:"object",code(t){let{gen:e,schema:A,data:i,parentSchema:n,it:o}=t,{opts:r}=o,s=(0,IAA.allSchemaProperties)(A),a=s.filter(E=>(0,CAA.alwaysValidSchema)(o,A[E]));if(s.length===0||a.length===s.length&&(!o.opts.unevaluated||o.props===!0))return;let c=r.strictSchema&&!r.allowMatchingProperties&&n.properties,l=e.name("valid");o.props!==!0&&!(o.props instanceof W5.Name)&&(o.props=(0,dAA.evaluatedPropsToName)(e,o.props));let{props:I}=o;C();function C(){for(let E of s)c&&d(E),o.allErrors?B(E):(e.var(l,!0),B(E),e.if(l))}function d(E){for(let h in c)new RegExp(E).test(h)&&(0,CAA.checkStrictMode)(o,`property ${h} matches pattern ${E} (use allowMatchingProperties)`)}function B(E){e.forIn("key",i,h=>{e.if((0,W5._)`${(0,IAA.usePattern)(t,E)}.test(${h})`,()=>{let u=a.includes(E);u||t.subschema({keyword:"patternProperties",schemaProp:E,dataProp:h,dataPropType:dAA.Type.Str},l),o.opts.unevaluated&&I!==!0?e.assign((0,W5._)`${I}[${h}]`,!0):!u&&!o.allErrors&&e.if((0,W5.not)(l),()=>e.break())})})}}};Dx.default=rSA});var EAA=Je(yx=>{"use strict";Object.defineProperty(yx,"__esModule",{value:!0});var sSA=bn(),aSA={keyword:"not",schemaType:["object","boolean"],trackErrors:!0,code(t){let{gen:e,schema:A,it:i}=t;if((0,sSA.alwaysValidSchema)(i,A)){t.fail();return}let n=e.name("valid");t.subschema({keyword:"not",compositeRule:!0,createErrors:!1,allErrors:!1},n),t.failResult(n,()=>t.reset(),()=>t.error())},error:{message:"must NOT be valid"}};yx.default=aSA});var QAA=Je(vx=>{"use strict";Object.defineProperty(vx,"__esModule",{value:!0});var cSA=_c(),lSA={keyword:"anyOf",schemaType:"array",trackErrors:!0,code:cSA.validateUnion,error:{message:"must match a schema in anyOf"}};vx.default=lSA});var hAA=Je(bx=>{"use strict";Object.defineProperty(bx,"__esModule",{value:!0});var X5=An(),gSA=bn(),ISA={message:"must match exactly one schema in oneOf",params:({params:t})=>(0,X5._)`{passingSchemas: ${t.passing}}`},CSA={keyword:"oneOf",schemaType:"array",trackErrors:!0,error:ISA,code(t){let{gen:e,schema:A,parentSchema:i,it:n}=t;if(!Array.isArray(A))throw new Error("ajv implementation error");if(n.opts.discriminator&&i.discriminator)return;let o=A,r=e.let("valid",!1),s=e.let("passing",null),a=e.name("_valid");t.setParams({passing:s}),e.block(c),t.result(r,()=>t.reset(),()=>t.error(!0));function c(){o.forEach((l,I)=>{let C;(0,gSA.alwaysValidSchema)(n,l)?e.var(a,!0):C=t.subschema({keyword:"oneOf",schemaProp:I,compositeRule:!0},a),I>0&&e.if((0,X5._)`${a} && ${r}`).assign(r,!1).assign(s,(0,X5._)`[${s}, ${I}]`).else(),e.if(a,()=>{e.assign(r,!0),e.assign(s,I),C&&t.mergeEvaluated(C,X5.Name)})})}}};bx.default=CSA});var uAA=Je(Mx=>{"use strict";Object.defineProperty(Mx,"__esModule",{value:!0});var dSA=bn(),BSA={keyword:"allOf",schemaType:"array",code(t){let{gen:e,schema:A,it:i}=t;if(!Array.isArray(A))throw new Error("ajv implementation error");let n=e.name("valid");A.forEach((o,r)=>{if((0,dSA.alwaysValidSchema)(i,o))return;let s=t.subschema({keyword:"allOf",schemaProp:r},n);t.ok(n),t.mergeEvaluated(s)})}};Mx.default=BSA});var pAA=Je(kx=>{"use strict";Object.defineProperty(kx,"__esModule",{value:!0});var $5=An(),mAA=bn(),ESA={message:({params:t})=>(0,$5.str)`must match "${t.ifClause}" schema`,params:({params:t})=>(0,$5._)`{failingKeyword: ${t.ifClause}}`},QSA={keyword:"if",schemaType:["object","boolean"],trackErrors:!0,error:ESA,code(t){let{gen:e,parentSchema:A,it:i}=t;A.then===void 0&&A.else===void 0&&(0,mAA.checkStrictMode)(i,'"if" without "then" and "else" is ignored');let n=fAA(i,"then"),o=fAA(i,"else");if(!n&&!o)return;let r=e.let("valid",!0),s=e.name("_valid");if(a(),t.reset(),n&&o){let l=e.let("ifClause");t.setParams({ifClause:l}),e.if(s,c("then",l),c("else",l))}else n?e.if(s,c("then")):e.if((0,$5.not)(s),c("else"));t.pass(r,()=>t.error(!0));function a(){let l=t.subschema({keyword:"if",compositeRule:!0,createErrors:!1,allErrors:!1},s);t.mergeEvaluated(l)}function c(l,I){return()=>{let C=t.subschema({keyword:l},s);e.assign(r,s),t.mergeValidEvaluated(C,r),I?e.assign(I,(0,$5._)`${l}`):t.setParams({ifClause:l})}}}};function fAA(t,e){let A=t.schema[e];return A!==void 0&&!(0,mAA.alwaysValidSchema)(t,A)}kx.default=QSA});var wAA=Je(Sx=>{"use strict";Object.defineProperty(Sx,"__esModule",{value:!0});var hSA=bn(),uSA={keyword:["then","else"],schemaType:["object","boolean"],code({keyword:t,parentSchema:e,it:A}){e.if===void 0&&(0,hSA.checkStrictMode)(A,`"${t}" without "if" is ignored`)}};Sx.default=uSA});var DAA=Je(Rx=>{"use strict";Object.defineProperty(Rx,"__esModule",{value:!0});var fSA=Cx(),mSA=AAA(),pSA=dx(),wSA=tAA(),DSA=iAA(),ySA=rAA(),vSA=aAA(),bSA=mx(),MSA=gAA(),kSA=BAA(),SSA=EAA(),RSA=QAA(),xSA=hAA(),LSA=uAA(),FSA=pAA(),NSA=wAA();function _SA(t=!1){let e=[SSA.default,RSA.default,xSA.default,LSA.default,FSA.default,NSA.default,vSA.default,bSA.default,ySA.default,MSA.default,kSA.default];return t?e.push(mSA.default,wSA.default):e.push(fSA.default,pSA.default),e.push(DSA.default),e}Rx.default=_SA});var yAA=Je(xx=>{"use strict";Object.defineProperty(xx,"__esModule",{value:!0});var Qr=An(),GSA={message:({schemaCode:t})=>(0,Qr.str)`must match format "${t}"`,params:({schemaCode:t})=>(0,Qr._)`{format: ${t}}`},USA={keyword:"format",type:["number","string"],schemaType:"string",$data:!0,error:GSA,code(t,e){let{gen:A,data:i,$data:n,schema:o,schemaCode:r,it:s}=t,{opts:a,errSchemaPath:c,schemaEnv:l,self:I}=s;if(!a.validateFormats)return;n?C():d();function C(){let B=A.scopeValue("formats",{ref:I.formats,code:a.code.formats}),E=A.const("fDef",(0,Qr._)`${B}[${r}]`),h=A.let("fType"),u=A.let("format");A.if((0,Qr._)`typeof ${E} == "object" && !(${E} instanceof RegExp)`,()=>A.assign(h,(0,Qr._)`${E}.type || "string"`).assign(u,(0,Qr._)`${E}.validate`),()=>A.assign(h,(0,Qr._)`"string"`).assign(u,E)),t.fail$data((0,Qr.or)(D(),L()));function D(){return a.strictSchema===!1?Qr.nil:(0,Qr._)`${r} && !${u}`}function L(){let R=l.$async?(0,Qr._)`(${E}.async ? await ${u}(${i}) : ${u}(${i}))`:(0,Qr._)`${u}(${i})`,w=(0,Qr._)`(typeof ${u} == "function" ? ${R} : ${u}.test(${i}))`;return(0,Qr._)`${u} && ${u} !== true && ${h} === ${e} && !${w}`}}function d(){let B=I.formats[o];if(!B){D();return}if(B===!0)return;let[E,h,u]=L(B);E===e&&t.pass(R());function D(){if(a.strictSchema===!1){I.logger.warn(w());return}throw new Error(w());function w(){return`unknown format "${o}" ignored in schema at path "${c}"`}}function L(w){let _=w instanceof RegExp?(0,Qr.regexpCode)(w):a.code.formats?(0,Qr._)`${a.code.formats}${(0,Qr.getProperty)(o)}`:void 0,K=A.scopeValue("formats",{key:o,ref:w,code:_});return typeof w=="object"&&!(w instanceof RegExp)?[w.type||"string",w.validate,(0,Qr._)`${K}.validate`]:["string",w,K]}function R(){if(typeof B=="object"&&!(B instanceof RegExp)&&B.async){if(!l.$async)throw new Error("async format in sync schema");return(0,Qr._)`await ${u}(${i})`}return typeof h=="function"?(0,Qr._)`${u}(${i})`:(0,Qr._)`${u}.test(${i})`}}}};xx.default=USA});var vAA=Je(Lx=>{"use strict";Object.defineProperty(Lx,"__esModule",{value:!0});var KSA=yAA(),YSA=[KSA.default];Lx.default=YSA});var bAA=Je(hE=>{"use strict";Object.defineProperty(hE,"__esModule",{value:!0});hE.contentVocabulary=hE.metadataVocabulary=void 0;hE.metadataVocabulary=["title","description","default","deprecated","readOnly","writeOnly","examples"];hE.contentVocabulary=["contentMediaType","contentEncoding","contentSchema"]});var kAA=Je(Fx=>{"use strict";Object.defineProperty(Fx,"__esModule",{value:!0});var JSA=_$(),TSA=Z$(),HSA=DAA(),zSA=vAA(),MAA=bAA(),OSA=[JSA.default,TSA.default,(0,HSA.default)(),zSA.default,MAA.metadataVocabulary,MAA.contentVocabulary];Fx.default=OSA});var RAA=Je(Aw=>{"use strict";Object.defineProperty(Aw,"__esModule",{value:!0});Aw.DiscrError=void 0;var SAA=function(t){return t.Tag="tag",t.Mapping="mapping",t}(SAA||(Aw.DiscrError=SAA={}))});var LAA=Je(_x=>{"use strict";Object.defineProperty(_x,"__esModule",{value:!0});var uE=An(),Nx=RAA(),xAA=_5(),PSA=w4(),jSA=bn(),qSA={message:({params:{discrError:t,tagName:e}})=>t===Nx.DiscrError.Tag?`tag "${e}" must be string`:`value of tag "${e}" must be in oneOf`,params:({params:{discrError:t,tag:e,tagName:A}})=>(0,uE._)`{error: ${t}, tag: ${A}, tagValue: ${e}}`},VSA={keyword:"discriminator",type:"object",schemaType:"object",error:qSA,code(t){let{gen:e,data:A,schema:i,parentSchema:n,it:o}=t,{oneOf:r}=n;if(!o.opts.discriminator)throw new Error("discriminator: requires discriminator option");let s=i.propertyName;if(typeof s!="string")throw new Error("discriminator: requires propertyName");if(i.mapping)throw new Error("discriminator: mapping is not supported");if(!r)throw new Error("discriminator: requires oneOf keyword");let a=e.let("valid",!1),c=e.const("tag",(0,uE._)`${A}${(0,uE.getProperty)(s)}`);e.if((0,uE._)`typeof ${c} == "string"`,()=>l(),()=>t.error(!1,{discrError:Nx.DiscrError.Tag,tag:c,tagName:s})),t.ok(a);function l(){let d=C();e.if(!1);for(let B in d)e.elseIf((0,uE._)`${c} === ${B}`),e.assign(a,I(d[B]));e.else(),t.error(!1,{discrError:Nx.DiscrError.Mapping,tag:c,tagName:s}),e.endIf()}function I(d){let B=e.name("valid"),E=t.subschema({keyword:"oneOf",schemaProp:d},B);return t.mergeEvaluated(E,uE.Name),B}function C(){var d;let B={},E=u(n),h=!0;for(let R=0;R<r.length;R++){let w=r[R];if(w?.$ref&&!(0,jSA.schemaHasRulesButRef)(w,o.self.RULES)){let K=w.$ref;if(w=xAA.resolveRef.call(o.self,o.schemaEnv.root,o.baseId,K),w instanceof xAA.SchemaEnv&&(w=w.schema),w===void 0)throw new PSA.default(o.opts.uriResolver,o.baseId,K)}let _=(d=w?.properties)===null||d===void 0?void 0:d[s];if(typeof _!="object")throw new Error(`discriminator: oneOf subschemas (or referenced schemas) must have "properties/${s}"`);h=h&&(E||u(w)),D(_,R)}if(!h)throw new Error(`discriminator: "${s}" must be required`);return B;function u({required:R}){return Array.isArray(R)&&R.includes(s)}function D(R,w){if(R.const)L(R.const,w);else if(R.enum)for(let _ of R.enum)L(_,w);else throw new Error(`discriminator: "properties/${s}" must have "const" or "enum"`)}function L(R,w){if(typeof R!="string"||R in B)throw new Error(`discriminator: "${s}" values must be unique strings`);B[R]=w}}}};_x.default=VSA});var FAA=Je((Ipe,ZSA)=>{ZSA.exports={$schema:"http://json-schema.org/draft-07/schema#",$id:"http://json-schema.org/draft-07/schema#",title:"Core schema meta-schema",definitions:{schemaArray:{type:"array",minItems:1,items:{$ref:"#"}},nonNegativeInteger:{type:"integer",minimum:0},nonNegativeIntegerDefault0:{allOf:[{$ref:"#/definitions/nonNegativeInteger"},{default:0}]},simpleTypes:{enum:["array","boolean","integer","null","number","object","string"]},stringArray:{type:"array",items:{type:"string"},uniqueItems:!0,default:[]}},type:["object","boolean"],properties:{$id:{type:"string",format:"uri-reference"},$schema:{type:"string",format:"uri"},$ref:{type:"string",format:"uri-reference"},$comment:{type:"string"},title:{type:"string"},description:{type:"string"},default:!0,readOnly:{type:"boolean",default:!1},examples:{type:"array",items:!0},multipleOf:{type:"number",exclusiveMinimum:0},maximum:{type:"number"},exclusiveMaximum:{type:"number"},minimum:{type:"number"},exclusiveMinimum:{type:"number"},maxLength:{$ref:"#/definitions/nonNegativeInteger"},minLength:{$ref:"#/definitions/nonNegativeIntegerDefault0"},pattern:{type:"string",format:"regex"},additionalItems:{$ref:"#"},items:{anyOf:[{$ref:"#"},{$ref:"#/definitions/schemaArray"}],default:!0},maxItems:{$ref:"#/definitions/nonNegativeInteger"},minItems:{$ref:"#/definitions/nonNegativeIntegerDefault0"},uniqueItems:{type:"boolean",default:!1},contains:{$ref:"#"},maxProperties:{$ref:"#/definitions/nonNegativeInteger"},minProperties:{$ref:"#/definitions/nonNegativeIntegerDefault0"},required:{$ref:"#/definitions/stringArray"},additionalProperties:{$ref:"#"},definitions:{type:"object",additionalProperties:{$ref:"#"},default:{}},properties:{type:"object",additionalProperties:{$ref:"#"},default:{}},patternProperties:{type:"object",additionalProperties:{$ref:"#"},propertyNames:{format:"regex"},default:{}},dependencies:{type:"object",additionalProperties:{anyOf:[{$ref:"#"},{$ref:"#/definitions/stringArray"}]}},propertyNames:{$ref:"#"},const:!0,enum:{type:"array",items:!0,minItems:1,uniqueItems:!0},type:{anyOf:[{$ref:"#/definitions/simpleTypes"},{type:"array",items:{$ref:"#/definitions/simpleTypes"},minItems:1,uniqueItems:!0}]},format:{type:"string"},contentMediaType:{type:"string"},contentEncoding:{type:"string"},if:{$ref:"#"},then:{$ref:"#"},else:{$ref:"#"},allOf:{$ref:"#/definitions/schemaArray"},anyOf:{$ref:"#/definitions/schemaArray"},oneOf:{$ref:"#/definitions/schemaArray"},not:{$ref:"#"}},default:!0}});var _AA=Je((Oo,Gx)=>{"use strict";Object.defineProperty(Oo,"__esModule",{value:!0});Oo.MissingRefError=Oo.ValidationError=Oo.CodeGen=Oo.Name=Oo.nil=Oo.stringify=Oo.str=Oo._=Oo.KeywordCxt=Oo.Ajv=void 0;var WSA=S$(),XSA=kAA(),$SA=LAA(),NAA=FAA(),ARA=["/properties"],ew="http://json-schema.org/draft-07/schema",fE=class extends WSA.default{_addVocabularies(){super._addVocabularies(),XSA.default.forEach(e=>this.addVocabulary(e)),this.opts.discriminator&&this.addKeyword($SA.default)}_addDefaultMetaSchema(){if(super._addDefaultMetaSchema(),!this.opts.meta)return;let e=this.opts.$data?this.$dataMetaSchema(NAA,ARA):NAA;this.addMetaSchema(e,ew,!1),this.refs["http://json-schema.org/schema"]=ew}defaultMeta(){return this.opts.defaultMeta=super.defaultMeta()||(this.getSchema(ew)?ew:void 0)}};Oo.Ajv=fE;Gx.exports=Oo=fE;Gx.exports.Ajv=fE;Object.defineProperty(Oo,"__esModule",{value:!0});Oo.default=fE;var eRA=p4();Object.defineProperty(Oo,"KeywordCxt",{enumerable:!0,get:function(){return eRA.KeywordCxt}});var mE=An();Object.defineProperty(Oo,"_",{enumerable:!0,get:function(){return mE._}});Object.defineProperty(Oo,"str",{enumerable:!0,get:function(){return mE.str}});Object.defineProperty(Oo,"stringify",{enumerable:!0,get:function(){return mE.stringify}});Object.defineProperty(Oo,"nil",{enumerable:!0,get:function(){return mE.nil}});Object.defineProperty(Oo,"Name",{enumerable:!0,get:function(){return mE.Name}});Object.defineProperty(Oo,"CodeGen",{enumerable:!0,get:function(){return mE.CodeGen}});var tRA=F5();Object.defineProperty(Oo,"ValidationError",{enumerable:!0,get:function(){return tRA.default}});var iRA=w4();Object.defineProperty(Oo,"MissingRefError",{enumerable:!0,get:function(){return iRA.default}})});var GAA=Je(tw=>{"use strict";(function(t){"use strict";function e(x){return x!==null?Object.prototype.toString.call(x)==="[object Array]":!1}function A(x){return x!==null?Object.prototype.toString.call(x)==="[object Object]":!1}function i(x,Y){if(x===Y)return!0;var P=Object.prototype.toString.call(x);if(P!==Object.prototype.toString.call(Y))return!1;if(e(x)===!0){if(x.length!==Y.length)return!1;for(var X=0;X<x.length;X++)if(i(x[X],Y[X])===!1)return!1;return!0}if(A(x)===!0){var bA={};for(var Be in x)if(hasOwnProperty.call(x,Be)){if(i(x[Be],Y[Be])===!1)return!1;bA[Be]=!0}for(var Ee in Y)if(hasOwnProperty.call(Y,Ee)&&bA[Ee]!==!0)return!1;return!0}return!1}function n(x){if(x===""||x===!1||x===null)return!0;if(e(x)&&x.length===0)return!0;if(A(x)){for(var Y in x)if(x.hasOwnProperty(Y))return!1;return!0}else return!1}function o(x){for(var Y=Object.keys(x),P=[],X=0;X<Y.length;X++)P.push(x[Y[X]]);return P}function r(x,Y){var P={};for(var X in x)P[X]=x[X];for(var bA in Y)P[bA]=Y[bA];return P}var s;typeof String.prototype.trimLeft=="function"?s=function(x){return x.trimLeft()}:s=function(x){return x.match(/^\s*(.*)/)[1]};var a=0,c=1,l=2,I=3,C=4,d=5,B=6,E=7,h=8,u=9,D={0:"number",1:"any",2:"string",3:"array",4:"object",5:"boolean",6:"expression",7:"null",8:"Array<number>",9:"Array<string>"},L="EOF",R="UnquotedIdentifier",w="QuotedIdentifier",_="Rbracket",K="Rparen",z="Comma",U="Colon",H="Rbrace",q="Number",j="Current",gA="Expref",QA="Pipe",BA="Or",lA="And",vA="EQ",tA="GT",cA="LT",pA="GTE",VA="LTE",oe="NE",KA="Flatten",CA="Star",TA="Filter",Ze="Dot",He="Not",uA="Lbrace",eA="Lbracket",UA="Lparen",aA="Literal",le={".":Ze,"*":CA,",":z,":":U,"{":uA,"}":H,"]":_,"(":UA,")":K,"@":j},SA={"<":!0,">":!0,"=":!0,"!":!0},Ue={" ":!0,"	":!0,"\n":!0};function mA(x){return x>="a"&&x<="z"||x>="A"&&x<="Z"||x==="_"}function sA(x){return x>="0"&&x<="9"||x==="-"}function xt(x){return x>="a"&&x<="z"||x>="A"&&x<="Z"||x>="0"&&x<="9"||x==="_"}function tt(){}tt.prototype={tokenize:function(x){var Y=[];this._current=0;for(var P,X,bA;this._current<x.length;)if(mA(x[this._current]))P=this._current,X=this._consumeUnquotedIdentifier(x),Y.push({type:R,value:X,start:P});else if(le[x[this._current]]!==void 0)Y.push({type:le[x[this._current]],value:x[this._current],start:this._current}),this._current++;else if(sA(x[this._current]))bA=this._consumeNumber(x),Y.push(bA);else if(x[this._current]==="[")bA=this._consumeLBracket(x),Y.push(bA);else if(x[this._current]==='"')P=this._current,X=this._consumeQuotedIdentifier(x),Y.push({type:w,value:X,start:P});else if(x[this._current]==="'")P=this._current,X=this._consumeRawStringLiteral(x),Y.push({type:aA,value:X,start:P});else if(x[this._current]==="`"){P=this._current;var Be=this._consumeLiteral(x);Y.push({type:aA,value:Be,start:P})}else if(SA[x[this._current]]!==void 0)Y.push(this._consumeOperator(x));else if(Ue[x[this._current]]!==void 0)this._current++;else if(x[this._current]==="&")P=this._current,this._current++,x[this._current]==="&"?(this._current++,Y.push({type:lA,value:"&&",start:P})):Y.push({type:gA,value:"&",start:P});else if(x[this._current]==="|")P=this._current,this._current++,x[this._current]==="|"?(this._current++,Y.push({type:BA,value:"||",start:P})):Y.push({type:QA,value:"|",start:P});else{var Ee=new Error("Unknown character:"+x[this._current]);throw Ee.name="LexerError",Ee}return Y},_consumeUnquotedIdentifier:function(x){var Y=this._current;for(this._current++;this._current<x.length&&xt(x[this._current]);)this._current++;return x.slice(Y,this._current)},_consumeQuotedIdentifier:function(x){var Y=this._current;this._current++;for(var P=x.length;x[this._current]!=='"'&&this._current<P;){var X=this._current;x[X]==="\\"&&(x[X+1]==="\\"||x[X+1]==='"')?X+=2:X++,this._current=X}return this._current++,JSON.parse(x.slice(Y,this._current))},_consumeRawStringLiteral:function(x){var Y=this._current;this._current++;for(var P=x.length;x[this._current]!=="'"&&this._current<P;){var X=this._current;x[X]==="\\"&&(x[X+1]==="\\"||x[X+1]==="'")?X+=2:X++,this._current=X}this._current++;var bA=x.slice(Y+1,this._current-1);return bA.replace("\\'","'")},_consumeNumber:function(x){var Y=this._current;this._current++;for(var P=x.length;sA(x[this._current])&&this._current<P;)this._current++;var X=parseInt(x.slice(Y,this._current));return{type:q,value:X,start:Y}},_consumeLBracket:function(x){var Y=this._current;return this._current++,x[this._current]==="?"?(this._current++,{type:TA,value:"[?",start:Y}):x[this._current]==="]"?(this._current++,{type:KA,value:"[]",start:Y}):{type:eA,value:"[",start:Y}},_consumeOperator:function(x){var Y=this._current,P=x[Y];if(this._current++,P==="!")return x[this._current]==="="?(this._current++,{type:oe,value:"!=",start:Y}):{type:He,value:"!",start:Y};if(P==="<")return x[this._current]==="="?(this._current++,{type:VA,value:"<=",start:Y}):{type:cA,value:"<",start:Y};if(P===">")return x[this._current]==="="?(this._current++,{type:pA,value:">=",start:Y}):{type:tA,value:">",start:Y};if(P==="="&&x[this._current]==="=")return this._current++,{type:vA,value:"==",start:Y}},_consumeLiteral:function(x){this._current++;for(var Y=this._current,P=x.length,X;x[this._current]!=="`"&&this._current<P;){var bA=this._current;x[bA]==="\\"&&(x[bA+1]==="\\"||x[bA+1]==="`")?bA+=2:bA++,this._current=bA}var Be=s(x.slice(Y,this._current));return Be=Be.replace("\\`","`"),this._looksLikeJSON(Be)?X=JSON.parse(Be):X=JSON.parse('"'+Be+'"'),this._current++,X},_looksLikeJSON:function(x){var Y='[{"',P=["true","false","null"],X="-0123456789";if(x==="")return!1;if(Y.indexOf(x[0])>=0)return!0;if(P.indexOf(x)>=0)return!0;if(X.indexOf(x[0])>=0)try{return JSON.parse(x),!0}catch{return!1}else return!1}};var de={};de[L]=0,de[R]=0,de[w]=0,de[_]=0,de[K]=0,de[z]=0,de[H]=0,de[q]=0,de[j]=0,de[gA]=0,de[QA]=1,de[BA]=2,de[lA]=3,de[vA]=5,de[tA]=5,de[cA]=5,de[pA]=5,de[VA]=5,de[oe]=5,de[KA]=9,de[CA]=20,de[TA]=21,de[Ze]=40,de[He]=45,de[uA]=50,de[eA]=55,de[UA]=60;function Dt(){}Dt.prototype={parse:function(x){this._loadTokens(x),this.index=0;var Y=this.expression(0);if(this._lookahead(0)!==L){var P=this._lookaheadToken(0),X=new Error("Unexpected token type: "+P.type+", value: "+P.value);throw X.name="ParserError",X}return Y},_loadTokens:function(x){var Y=new tt,P=Y.tokenize(x);P.push({type:L,value:"",start:x.length}),this.tokens=P},expression:function(x){var Y=this._lookaheadToken(0);this._advance();for(var P=this.nud(Y),X=this._lookahead(0);x<de[X];)this._advance(),P=this.led(X,P),X=this._lookahead(0);return P},_lookahead:function(x){return this.tokens[this.index+x].type},_lookaheadToken:function(x){return this.tokens[this.index+x]},_advance:function(){this.index++},nud:function(x){var Y,P,X;switch(x.type){case aA:return{type:"Literal",value:x.value};case R:return{type:"Field",name:x.value};case w:var bA={type:"Field",name:x.value};if(this._lookahead(0)===UA)throw new Error("Quoted identifier not allowed for function names.");return bA;case He:return P=this.expression(de.Not),{type:"NotExpression",children:[P]};case CA:return Y={type:"Identity"},P=null,this._lookahead(0)===_?P={type:"Identity"}:P=this._parseProjectionRHS(de.Star),{type:"ValueProjection",children:[Y,P]};case TA:return this.led(x.type,{type:"Identity"});case uA:return this._parseMultiselectHash();case KA:return Y={type:KA,children:[{type:"Identity"}]},P=this._parseProjectionRHS(de.Flatten),{type:"Projection",children:[Y,P]};case eA:return this._lookahead(0)===q||this._lookahead(0)===U?(P=this._parseIndexExpression(),this._projectIfSlice({type:"Identity"},P)):this._lookahead(0)===CA&&this._lookahead(1)===_?(this._advance(),this._advance(),P=this._parseProjectionRHS(de.Star),{type:"Projection",children:[{type:"Identity"},P]}):this._parseMultiselectList();case j:return{type:j};case gA:return X=this.expression(de.Expref),{type:"ExpressionReference",children:[X]};case UA:for(var Be=[];this._lookahead(0)!==K;)this._lookahead(0)===j?(X={type:j},this._advance()):X=this.expression(0),Be.push(X);return this._match(K),Be[0];default:this._errorToken(x)}},led:function(x,Y){var P;switch(x){case Ze:var X=de.Dot;return this._lookahead(0)!==CA?(P=this._parseDotRHS(X),{type:"Subexpression",children:[Y,P]}):(this._advance(),P=this._parseProjectionRHS(X),{type:"ValueProjection",children:[Y,P]});case QA:return P=this.expression(de.Pipe),{type:QA,children:[Y,P]};case BA:return P=this.expression(de.Or),{type:"OrExpression",children:[Y,P]};case lA:return P=this.expression(de.And),{type:"AndExpression",children:[Y,P]};case UA:for(var bA=Y.name,Be=[],Ee,kA;this._lookahead(0)!==K;)this._lookahead(0)===j?(Ee={type:j},this._advance()):Ee=this.expression(0),this._lookahead(0)===z&&this._match(z),Be.push(Ee);return this._match(K),kA={type:"Function",name:bA,children:Be},kA;case TA:var DA=this.expression(0);return this._match(_),this._lookahead(0)===KA?P={type:"Identity"}:P=this._parseProjectionRHS(de.Filter),{type:"FilterProjection",children:[Y,P,DA]};case KA:var gt={type:KA,children:[Y]},Ve=this._parseProjectionRHS(de.Flatten);return{type:"Projection",children:[gt,Ve]};case vA:case oe:case tA:case pA:case cA:case VA:return this._parseComparator(Y,x);case eA:var ZA=this._lookaheadToken(0);return ZA.type===q||ZA.type===U?(P=this._parseIndexExpression(),this._projectIfSlice(Y,P)):(this._match(CA),this._match(_),P=this._parseProjectionRHS(de.Star),{type:"Projection",children:[Y,P]});default:this._errorToken(this._lookaheadToken(0))}},_match:function(x){if(this._lookahead(0)===x)this._advance();else{var Y=this._lookaheadToken(0),P=new Error("Expected "+x+", got: "+Y.type);throw P.name="ParserError",P}},_errorToken:function(x){var Y=new Error("Invalid token ("+x.type+'): "'+x.value+'"');throw Y.name="ParserError",Y},_parseIndexExpression:function(){if(this._lookahead(0)===U||this._lookahead(1)===U)return this._parseSliceExpression();var x={type:"Index",value:this._lookaheadToken(0).value};return this._advance(),this._match(_),x},_projectIfSlice:function(x,Y){var P={type:"IndexExpression",children:[x,Y]};return Y.type==="Slice"?{type:"Projection",children:[P,this._parseProjectionRHS(de.Star)]}:P},_parseSliceExpression:function(){for(var x=[null,null,null],Y=0,P=this._lookahead(0);P!==_&&Y<3;){if(P===U)Y++,this._advance();else if(P===q)x[Y]=this._lookaheadToken(0).value,this._advance();else{var X=this._lookahead(0),bA=new Error("Syntax error, unexpected token: "+X.value+"("+X.type+")");throw bA.name="Parsererror",bA}P=this._lookahead(0)}return this._match(_),{type:"Slice",children:x}},_parseComparator:function(x,Y){var P=this.expression(de[Y]);return{type:"Comparator",name:Y,children:[x,P]}},_parseDotRHS:function(x){var Y=this._lookahead(0),P=[R,w,CA];if(P.indexOf(Y)>=0)return this.expression(x);if(Y===eA)return this._match(eA),this._parseMultiselectList();if(Y===uA)return this._match(uA),this._parseMultiselectHash()},_parseProjectionRHS:function(x){var Y;if(de[this._lookahead(0)]<10)Y={type:"Identity"};else if(this._lookahead(0)===eA)Y=this.expression(x);else if(this._lookahead(0)===TA)Y=this.expression(x);else if(this._lookahead(0)===Ze)this._match(Ze),Y=this._parseDotRHS(x);else{var P=this._lookaheadToken(0),X=new Error("Sytanx error, unexpected token: "+P.value+"("+P.type+")");throw X.name="ParserError",X}return Y},_parseMultiselectList:function(){for(var x=[];this._lookahead(0)!==_;){var Y=this.expression(0);if(x.push(Y),this._lookahead(0)===z&&(this._match(z),this._lookahead(0)===_))throw new Error("Unexpected token Rbracket")}return this._match(_),{type:"MultiSelectList",children:x}},_parseMultiselectHash:function(){for(var x=[],Y=[R,w],P,X,bA,Be;;){if(P=this._lookaheadToken(0),Y.indexOf(P.type)<0)throw new Error("Expecting an identifier token, got: "+P.type);if(X=P.value,this._advance(),this._match(U),bA=this.expression(0),Be={type:"KeyValuePair",name:X,value:bA},x.push(Be),this._lookahead(0)===z)this._match(z);else if(this._lookahead(0)===H){this._match(H);break}}return{type:"MultiSelectHash",children:x}}};function _e(x){this.runtime=x}_e.prototype={search:function(x,Y){return this.visit(x,Y)},visit:function(x,Y){var P,X,bA,Be,Ee,kA,DA,gt,Ve,ZA;switch(x.type){case"Field":return Y!==null&&A(Y)?(kA=Y[x.name],kA===void 0?null:kA):null;case"Subexpression":for(bA=this.visit(x.children[0],Y),ZA=1;ZA<x.children.length;ZA++)if(bA=this.visit(x.children[1],bA),bA===null)return null;return bA;case"IndexExpression":return DA=this.visit(x.children[0],Y),gt=this.visit(x.children[1],DA),gt;case"Index":if(!e(Y))return null;var rt=x.value;return rt<0&&(rt=Y.length+rt),bA=Y[rt],bA===void 0&&(bA=null),bA;case"Slice":if(!e(Y))return null;var Ei=x.children.slice(0),tn=this.computeSliceParams(Y.length,Ei),qi=tn[0],xe=tn[1],nn=tn[2];if(bA=[],nn>0)for(ZA=qi;ZA<xe;ZA+=nn)bA.push(Y[ZA]);else for(ZA=qi;ZA>xe;ZA+=nn)bA.push(Y[ZA]);return bA;case"Projection":var mi=this.visit(x.children[0],Y);if(!e(mi))return null;for(Ve=[],ZA=0;ZA<mi.length;ZA++)X=this.visit(x.children[1],mi[ZA]),X!==null&&Ve.push(X);return Ve;case"ValueProjection":if(mi=this.visit(x.children[0],Y),!A(mi))return null;Ve=[];var Ot=o(mi);for(ZA=0;ZA<Ot.length;ZA++)X=this.visit(x.children[1],Ot[ZA]),X!==null&&Ve.push(X);return Ve;case"FilterProjection":if(mi=this.visit(x.children[0],Y),!e(mi))return null;var Lt=[],ii=[];for(ZA=0;ZA<mi.length;ZA++)P=this.visit(x.children[2],mi[ZA]),n(P)||Lt.push(mi[ZA]);for(var _i=0;_i<Lt.length;_i++)X=this.visit(x.children[1],Lt[_i]),X!==null&&ii.push(X);return ii;case"Comparator":switch(Be=this.visit(x.children[0],Y),Ee=this.visit(x.children[1],Y),x.name){case vA:bA=i(Be,Ee);break;case oe:bA=!i(Be,Ee);break;case tA:bA=Be>Ee;break;case pA:bA=Be>=Ee;break;case cA:bA=Be<Ee;break;case VA:bA=Be<=Ee;break;default:throw new Error("Unknown comparator: "+x.name)}return bA;case KA:var Tt=this.visit(x.children[0],Y);if(!e(Tt))return null;var M=[];for(ZA=0;ZA<Tt.length;ZA++)X=Tt[ZA],e(X)?M.push.apply(M,X):M.push(X);return M;case"Identity":return Y;case"MultiSelectList":if(Y===null)return null;for(Ve=[],ZA=0;ZA<x.children.length;ZA++)Ve.push(this.visit(x.children[ZA],Y));return Ve;case"MultiSelectHash":if(Y===null)return null;Ve={};var We;for(ZA=0;ZA<x.children.length;ZA++)We=x.children[ZA],Ve[We.name]=this.visit(We.value,Y);return Ve;case"OrExpression":return P=this.visit(x.children[0],Y),n(P)&&(P=this.visit(x.children[1],Y)),P;case"AndExpression":return Be=this.visit(x.children[0],Y),n(Be)===!0?Be:this.visit(x.children[1],Y);case"NotExpression":return Be=this.visit(x.children[0],Y),n(Be);case"Literal":return x.value;case QA:return DA=this.visit(x.children[0],Y),this.visit(x.children[1],DA);case j:return Y;case"Function":var ni=[];for(ZA=0;ZA<x.children.length;ZA++)ni.push(this.visit(x.children[ZA],Y));return this.runtime.callFunction(x.name,ni);case"ExpressionReference":var pi=x.children[0];return pi.jmespathType=gA,pi;default:throw new Error("Unknown node type: "+x.type)}},computeSliceParams:function(x,Y){var P=Y[0],X=Y[1],bA=Y[2],Be=[null,null,null];if(bA===null)bA=1;else if(bA===0){var Ee=new Error("Invalid slice, step cannot be 0");throw Ee.name="RuntimeError",Ee}var kA=bA<0;return P===null?P=kA?x-1:0:P=this.capSliceRange(x,P,bA),X===null?X=kA?-1:x:X=this.capSliceRange(x,X,bA),Be[0]=P,Be[1]=X,Be[2]=bA,Be},capSliceRange:function(x,Y,P){return Y<0?(Y+=x,Y<0&&(Y=P<0?-1:0)):Y>=x&&(Y=P<0?x-1:x),Y}};function Le(x){this._interpreter=x,this.functionTable={abs:{_func:this._functionAbs,_signature:[{types:[a]}]},avg:{_func:this._functionAvg,_signature:[{types:[h]}]},ceil:{_func:this._functionCeil,_signature:[{types:[a]}]},contains:{_func:this._functionContains,_signature:[{types:[l,I]},{types:[c]}]},ends_with:{_func:this._functionEndsWith,_signature:[{types:[l]},{types:[l]}]},floor:{_func:this._functionFloor,_signature:[{types:[a]}]},length:{_func:this._functionLength,_signature:[{types:[l,I,C]}]},map:{_func:this._functionMap,_signature:[{types:[B]},{types:[I]}]},max:{_func:this._functionMax,_signature:[{types:[h,u]}]},merge:{_func:this._functionMerge,_signature:[{types:[C],variadic:!0}]},max_by:{_func:this._functionMaxBy,_signature:[{types:[I]},{types:[B]}]},sum:{_func:this._functionSum,_signature:[{types:[h]}]},starts_with:{_func:this._functionStartsWith,_signature:[{types:[l]},{types:[l]}]},min:{_func:this._functionMin,_signature:[{types:[h,u]}]},min_by:{_func:this._functionMinBy,_signature:[{types:[I]},{types:[B]}]},type:{_func:this._functionType,_signature:[{types:[c]}]},keys:{_func:this._functionKeys,_signature:[{types:[C]}]},values:{_func:this._functionValues,_signature:[{types:[C]}]},sort:{_func:this._functionSort,_signature:[{types:[u,h]}]},sort_by:{_func:this._functionSortBy,_signature:[{types:[I]},{types:[B]}]},join:{_func:this._functionJoin,_signature:[{types:[l]},{types:[u]}]},reverse:{_func:this._functionReverse,_signature:[{types:[l,I]}]},to_array:{_func:this._functionToArray,_signature:[{types:[c]}]},to_string:{_func:this._functionToString,_signature:[{types:[c]}]},to_number:{_func:this._functionToNumber,_signature:[{types:[c]}]},not_null:{_func:this._functionNotNull,_signature:[{types:[c],variadic:!0}]}}}Le.prototype={callFunction:function(x,Y){var P=this.functionTable[x];if(P===void 0)throw new Error("Unknown function: "+x+"()");return this._validateArgs(x,Y,P._signature),P._func.call(this,Y)},_validateArgs:function(x,Y,P){var X;if(P[P.length-1].variadic){if(Y.length<P.length)throw X=P.length===1?" argument":" arguments",new Error("ArgumentError: "+x+"() takes at least"+P.length+X+" but received "+Y.length)}else if(Y.length!==P.length)throw X=P.length===1?" argument":" arguments",new Error("ArgumentError: "+x+"() takes "+P.length+X+" but received "+Y.length);for(var bA,Be,Ee,kA=0;kA<P.length;kA++){Ee=!1,bA=P[kA].types,Be=this._getTypeName(Y[kA]);for(var DA=0;DA<bA.length;DA++)if(this._typeMatches(Be,bA[DA],Y[kA])){Ee=!0;break}if(!Ee){var gt=bA.map(function(Ve){return D[Ve]}).join(",");throw new Error("TypeError: "+x+"() expected argument "+(kA+1)+" to be type "+gt+" but received type "+D[Be]+" instead.")}}},_typeMatches:function(x,Y,P){if(Y===c)return!0;if(Y===u||Y===h||Y===I){if(Y===I)return x===I;if(x===I){var X;Y===h?X=a:Y===u&&(X=l);for(var bA=0;bA<P.length;bA++)if(!this._typeMatches(this._getTypeName(P[bA]),X,P[bA]))return!1;return!0}}else return x===Y},_getTypeName:function(x){switch(Object.prototype.toString.call(x)){case"[object String]":return l;case"[object Number]":return a;case"[object Array]":return I;case"[object Boolean]":return d;case"[object Null]":return E;case"[object Object]":return x.jmespathType===gA?B:C}},_functionStartsWith:function(x){return x[0].lastIndexOf(x[1])===0},_functionEndsWith:function(x){var Y=x[0],P=x[1];return Y.indexOf(P,Y.length-P.length)!==-1},_functionReverse:function(x){var Y=this._getTypeName(x[0]);if(Y===l){for(var P=x[0],X="",bA=P.length-1;bA>=0;bA--)X+=P[bA];return X}else{var Be=x[0].slice(0);return Be.reverse(),Be}},_functionAbs:function(x){return Math.abs(x[0])},_functionCeil:function(x){return Math.ceil(x[0])},_functionAvg:function(x){for(var Y=0,P=x[0],X=0;X<P.length;X++)Y+=P[X];return Y/P.length},_functionContains:function(x){return x[0].indexOf(x[1])>=0},_functionFloor:function(x){return Math.floor(x[0])},_functionLength:function(x){return A(x[0])?Object.keys(x[0]).length:x[0].length},_functionMap:function(x){for(var Y=[],P=this._interpreter,X=x[0],bA=x[1],Be=0;Be<bA.length;Be++)Y.push(P.visit(X,bA[Be]));return Y},_functionMerge:function(x){for(var Y={},P=0;P<x.length;P++){var X=x[P];for(var bA in X)Y[bA]=X[bA]}return Y},_functionMax:function(x){if(x[0].length>0){var Y=this._getTypeName(x[0][0]);if(Y===a)return Math.max.apply(Math,x[0]);for(var P=x[0],X=P[0],bA=1;bA<P.length;bA++)X.localeCompare(P[bA])<0&&(X=P[bA]);return X}else return null},_functionMin:function(x){if(x[0].length>0){var Y=this._getTypeName(x[0][0]);if(Y===a)return Math.min.apply(Math,x[0]);for(var P=x[0],X=P[0],bA=1;bA<P.length;bA++)P[bA].localeCompare(X)<0&&(X=P[bA]);return X}else return null},_functionSum:function(x){for(var Y=0,P=x[0],X=0;X<P.length;X++)Y+=P[X];return Y},_functionType:function(x){switch(this._getTypeName(x[0])){case a:return"number";case l:return"string";case I:return"array";case C:return"object";case d:return"boolean";case B:return"expref";case E:return"null"}},_functionKeys:function(x){return Object.keys(x[0])},_functionValues:function(x){for(var Y=x[0],P=Object.keys(Y),X=[],bA=0;bA<P.length;bA++)X.push(Y[P[bA]]);return X},_functionJoin:function(x){var Y=x[0],P=x[1];return P.join(Y)},_functionToArray:function(x){return this._getTypeName(x[0])===I?x[0]:[x[0]]},_functionToString:function(x){return this._getTypeName(x[0])===l?x[0]:JSON.stringify(x[0])},_functionToNumber:function(x){var Y=this._getTypeName(x[0]),P;return Y===a?x[0]:Y===l&&(P=+x[0],!isNaN(P))?P:null},_functionNotNull:function(x){for(var Y=0;Y<x.length;Y++)if(this._getTypeName(x[Y])!==E)return x[Y];return null},_functionSort:function(x){var Y=x[0].slice(0);return Y.sort(),Y},_functionSortBy:function(x){var Y=x[0].slice(0);if(Y.length===0)return Y;var P=this._interpreter,X=x[1],bA=this._getTypeName(P.visit(X,Y[0]));if([a,l].indexOf(bA)<0)throw new Error("TypeError");for(var Be=this,Ee=[],kA=0;kA<Y.length;kA++)Ee.push([kA,Y[kA]]);Ee.sort(function(gt,Ve){var ZA=P.visit(X,gt[1]),rt=P.visit(X,Ve[1]);if(Be._getTypeName(ZA)!==bA)throw new Error("TypeError: expected "+bA+", received "+Be._getTypeName(ZA));if(Be._getTypeName(rt)!==bA)throw new Error("TypeError: expected "+bA+", received "+Be._getTypeName(rt));return ZA>rt?1:ZA<rt?-1:gt[0]-Ve[0]});for(var DA=0;DA<Ee.length;DA++)Y[DA]=Ee[DA][1];return Y},_functionMaxBy:function(x){for(var Y=x[1],P=x[0],X=this.createKeyFunction(Y,[a,l]),bA=-1/0,Be,Ee,kA=0;kA<P.length;kA++)Ee=X(P[kA]),Ee>bA&&(bA=Ee,Be=P[kA]);return Be},_functionMinBy:function(x){for(var Y=x[1],P=x[0],X=this.createKeyFunction(Y,[a,l]),bA=1/0,Be,Ee,kA=0;kA<P.length;kA++)Ee=X(P[kA]),Ee<bA&&(bA=Ee,Be=P[kA]);return Be},createKeyFunction:function(x,Y){var P=this,X=this._interpreter,bA=function(Be){var Ee=X.visit(x,Be);if(Y.indexOf(P._getTypeName(Ee))<0){var kA="TypeError: expected one of "+Y+", received "+P._getTypeName(Ee);throw new Error(kA)}return Ee};return bA}};function bt(x){var Y=new Dt,P=Y.parse(x);return P}function Re(x){var Y=new tt;return Y.tokenize(x)}function $t(x,Y){var P=new Dt,X=new Le,bA=new _e(X);X._interpreter=bA;var Be=P.parse(Y);return bA.search(Be,x)}t.tokenize=Re,t.compile=bt,t.search=$t,t.strictDeepEqual=i})(typeof tw>"u"?tw.jmespath={}:tw)});function O7(t,e){return Object.is(t,e)}var Lr=null,zf=!1,P7=1,ia=Symbol("SIGNAL");function Ti(t){let e=Lr;return Lr=t,e}function j7(){return Lr}var Bd={version:0,lastCleanEpoch:0,dirty:!1,producerNode:void 0,producerLastReadVersion:void 0,producerIndexOfThis:void 0,nextProducerIndex:0,liveConsumerNode:void 0,liveConsumerIndexOfThis:void 0,consumerAllowSignalWrites:!1,consumerIsAlwaysLive:!1,kind:"unknown",producerMustRecompute:()=>!1,producerRecomputeValue:()=>{},consumerMarkedDirty:()=>{},consumerOnSignalRead:()=>{}};function VQ(t){if(zf)throw new Error("");if(Lr===null)return;Lr.consumerOnSignalRead(t);let e=Lr.nextProducerIndex++;if(Vf(Lr),e<Lr.producerNode.length&&Lr.producerNode[e]!==t&&qQ(Lr)){let A=Lr.producerNode[e];qf(A,Lr.producerIndexOfThis[e])}Lr.producerNode[e]!==t&&(Lr.producerNode[e]=t,Lr.producerIndexOfThis[e]=qQ(Lr)?VU(t,Lr,e):0),Lr.producerLastReadVersion[e]=t.version}function qU(){P7++}function q7(t){if(!(qQ(t)&&!t.dirty)&&!(!t.dirty&&t.lastCleanEpoch===P7)){if(!t.producerMustRecompute(t)&&!jf(t)){z7(t);return}t.producerRecomputeValue(t),z7(t)}}function V7(t){if(t.liveConsumerNode===void 0)return;let e=zf;zf=!0;try{for(let A of t.liveConsumerNode)A.dirty||KgA(A)}finally{zf=e}}function Z7(){return Lr?.consumerAllowSignalWrites!==!1}function KgA(t){t.dirty=!0,V7(t),t.consumerMarkedDirty?.(t)}function z7(t){t.dirty=!1,t.lastCleanEpoch=P7}function ZQ(t){return t&&(t.nextProducerIndex=0),Ti(t)}function Pf(t,e){if(Ti(e),!(!t||t.producerNode===void 0||t.producerIndexOfThis===void 0||t.producerLastReadVersion===void 0)){if(qQ(t))for(let A=t.nextProducerIndex;A<t.producerNode.length;A++)qf(t.producerNode[A],t.producerIndexOfThis[A]);for(;t.producerNode.length>t.nextProducerIndex;)t.producerNode.pop(),t.producerLastReadVersion.pop(),t.producerIndexOfThis.pop()}}function jf(t){Vf(t);for(let e=0;e<t.producerNode.length;e++){let A=t.producerNode[e],i=t.producerLastReadVersion[e];if(i!==A.version||(q7(A),i!==A.version))return!0}return!1}function WQ(t){if(Vf(t),qQ(t))for(let e=0;e<t.producerNode.length;e++)qf(t.producerNode[e],t.producerIndexOfThis[e]);t.producerNode.length=t.producerLastReadVersion.length=t.producerIndexOfThis.length=0,t.liveConsumerNode&&(t.liveConsumerNode.length=t.liveConsumerIndexOfThis.length=0)}function VU(t,e,A){if(ZU(t),t.liveConsumerNode.length===0&&WU(t))for(let i=0;i<t.producerNode.length;i++)t.producerIndexOfThis[i]=VU(t.producerNode[i],t,i);return t.liveConsumerIndexOfThis.push(A),t.liveConsumerNode.push(e)-1}function qf(t,e){if(ZU(t),t.liveConsumerNode.length===1&&WU(t))for(let i=0;i<t.producerNode.length;i++)qf(t.producerNode[i],t.producerIndexOfThis[i]);let A=t.liveConsumerNode.length-1;if(t.liveConsumerNode[e]=t.liveConsumerNode[A],t.liveConsumerIndexOfThis[e]=t.liveConsumerIndexOfThis[A],t.liveConsumerNode.length--,t.liveConsumerIndexOfThis.length--,e<t.liveConsumerNode.length){let i=t.liveConsumerIndexOfThis[e],n=t.liveConsumerNode[e];Vf(n),n.producerIndexOfThis[i]=e}}function qQ(t){return t.consumerIsAlwaysLive||(t?.liveConsumerNode?.length??0)>0}function Vf(t){t.producerNode??=[],t.producerIndexOfThis??=[],t.producerLastReadVersion??=[]}function ZU(t){t.liveConsumerNode??=[],t.liveConsumerIndexOfThis??=[]}function WU(t){return t.producerNode!==void 0}function Zf(t,e){let A=Object.create(YgA);A.computation=t,e!==void 0&&(A.equal=e);let i=()=>{if(q7(A),VQ(A),A.value===Of)throw A.error;return A.value};return i[ia]=A,i}var T7=Symbol("UNSET"),H7=Symbol("COMPUTING"),Of=Symbol("ERRORED"),YgA=Ye(rA({},Bd),{value:T7,dirty:!0,error:null,equal:O7,kind:"computed",producerMustRecompute(t){return t.value===T7||t.value===H7},producerRecomputeValue(t){if(t.value===H7)throw new Error("Detected cycle in computations.");let e=t.value;t.value=H7;let A=ZQ(t),i,n=!1;try{i=t.computation(),Ti(null),n=e!==T7&&e!==Of&&i!==Of&&t.equal(e,i)}catch(o){i=Of,t.error=o}finally{Pf(t,A)}if(n){t.value=e;return}t.value=i,t.version++}});function JgA(){throw new Error}var XU=JgA;function $U(t){XU(t)}function W7(t){XU=t}var TgA=null;function X7(t,e){let A=Object.create(Wf);A.value=t,e!==void 0&&(A.equal=e);let i=()=>(VQ(A),A.value);return i[ia]=A,i}function XQ(t,e){Z7()||$U(t),t.equal(t.value,e)||(t.value=e,HgA(t))}function $7(t,e){Z7()||$U(t),XQ(t,e(t.value))}var Wf=Ye(rA({},Bd),{equal:O7,value:void 0,kind:"signal"});function HgA(t){t.version++,qU(),V7(t),TgA?.()}function Av(t){let e=Ti(null);try{return t()}finally{Ti(e)}}var ev;function $Q(){return ev}function n0(t){let e=ev;return ev=t,e}var Xf=Symbol("NotFound");function Xt(t){return typeof t=="function"}function Ed(t){let A=t(i=>{Error.call(i),i.stack=new Error().stack});return A.prototype=Object.create(Error.prototype),A.prototype.constructor=A,A}var $f=Ed(t=>function(A){t(this),this.message=A?`${A.length} errors occurred during unsubscription:
${A.map((i,n)=>`${n+1}) ${i.toString()}`).join(`
  `)}`:"",this.name="UnsubscriptionError",this.errors=A});function cI(t,e){if(t){let A=t.indexOf(e);0<=A&&t.splice(A,1)}}var Kt=class t{constructor(e){this.initialTeardown=e,this.closed=!1,this._parentage=null,this._finalizers=null}unsubscribe(){let e;if(!this.closed){this.closed=!0;let{_parentage:A}=this;if(A)if(this._parentage=null,Array.isArray(A))for(let o of A)o.remove(this);else A.remove(this);let{initialTeardown:i}=this;if(Xt(i))try{i()}catch(o){e=o instanceof $f?o.errors:[o]}let{_finalizers:n}=this;if(n){this._finalizers=null;for(let o of n)try{AK(o)}catch(r){e=e??[],r instanceof $f?e=[...e,...r.errors]:e.push(r)}}if(e)throw new $f(e)}}add(e){var A;if(e&&e!==this)if(this.closed)AK(e);else{if(e instanceof t){if(e.closed||e._hasParent(this))return;e._addParent(this)}(this._finalizers=(A=this._finalizers)!==null&&A!==void 0?A:[]).push(e)}}_hasParent(e){let{_parentage:A}=this;return A===e||Array.isArray(A)&&A.includes(e)}_addParent(e){let{_parentage:A}=this;this._parentage=Array.isArray(A)?(A.push(e),A):A?[A,e]:e}_removeParent(e){let{_parentage:A}=this;A===e?this._parentage=null:Array.isArray(A)&&cI(A,e)}remove(e){let{_finalizers:A}=this;A&&cI(A,e),e instanceof t&&e._removeParent(this)}};Kt.EMPTY=(()=>{let t=new Kt;return t.closed=!0,t})();var tv=Kt.EMPTY;function Am(t){return t instanceof Kt||t&&"closed"in t&&Xt(t.remove)&&Xt(t.add)&&Xt(t.unsubscribe)}function AK(t){Xt(t)?t():t.unsubscribe()}var $c={onUnhandledError:null,onStoppedNotification:null,Promise:void 0,useDeprecatedSynchronousErrorHandling:!1,useDeprecatedNextContext:!1};var Qd={setTimeout(t,e,...A){let{delegate:i}=Qd;return i?.setTimeout?i.setTimeout(t,e,...A):setTimeout(t,e,...A)},clearTimeout(t){let{delegate:e}=Qd;return(e?.clearTimeout||clearTimeout)(t)},delegate:void 0};function em(t){Qd.setTimeout(()=>{let{onUnhandledError:e}=$c;if(e)e(t);else throw t})}function Ah(){}var eK=iv("C",void 0,void 0);function tK(t){return iv("E",void 0,t)}function iK(t){return iv("N",t,void 0)}function iv(t,e,A){return{kind:t,value:e,error:A}}var lI=null;function hd(t){if($c.useDeprecatedSynchronousErrorHandling){let e=!lI;if(e&&(lI={errorThrown:!1,error:null}),t(),e){let{errorThrown:A,error:i}=lI;if(lI=null,A)throw i}}else t()}function nK(t){$c.useDeprecatedSynchronousErrorHandling&&lI&&(lI.errorThrown=!0,lI.error=t)}var o0=class extends Kt{constructor(e){super(),this.isStopped=!1,e?(this.destination=e,Am(e)&&e.add(this)):this.destination=VgA}static create(e,A,i){return new r0(e,A,i)}next(e){this.isStopped?ov(iK(e),this):this._next(e)}error(e){this.isStopped?ov(tK(e),this):(this.isStopped=!0,this._error(e))}complete(){this.isStopped?ov(eK,this):(this.isStopped=!0,this._complete())}unsubscribe(){this.closed||(this.isStopped=!0,super.unsubscribe(),this.destination=null)}_next(e){this.destination.next(e)}_error(e){try{this.destination.error(e)}finally{this.unsubscribe()}}_complete(){try{this.destination.complete()}finally{this.unsubscribe()}}},jgA=Function.prototype.bind;function nv(t,e){return jgA.call(t,e)}var rv=class{constructor(e){this.partialObserver=e}next(e){let{partialObserver:A}=this;if(A.next)try{A.next(e)}catch(i){tm(i)}}error(e){let{partialObserver:A}=this;if(A.error)try{A.error(e)}catch(i){tm(i)}else tm(e)}complete(){let{partialObserver:e}=this;if(e.complete)try{e.complete()}catch(A){tm(A)}}},r0=class extends o0{constructor(e,A,i){super();let n;if(Xt(e)||!e)n={next:e??void 0,error:A??void 0,complete:i??void 0};else{let o;this&&$c.useDeprecatedNextContext?(o=Object.create(e),o.unsubscribe=()=>this.unsubscribe(),n={next:e.next&&nv(e.next,o),error:e.error&&nv(e.error,o),complete:e.complete&&nv(e.complete,o)}):n=e}this.destination=new rv(n)}};function tm(t){$c.useDeprecatedSynchronousErrorHandling?nK(t):em(t)}function qgA(t){throw t}function ov(t,e){let{onStoppedNotification:A}=$c;A&&Qd.setTimeout(()=>A(t,e))}var VgA={closed:!0,next:Ah,error:qgA,complete:Ah};var ud=typeof Symbol=="function"&&Symbol.observable||"@@observable";function Ys(t){return t}function sv(...t){return av(t)}function av(t){return t.length===0?Ys:t.length===1?t[0]:function(A){return t.reduce((i,n)=>n(i),A)}}var ct=(()=>{class t{constructor(A){A&&(this._subscribe=A)}lift(A){let i=new t;return i.source=this,i.operator=A,i}subscribe(A,i,n){let o=WgA(A)?A:new r0(A,i,n);return hd(()=>{let{operator:r,source:s}=this;o.add(r?r.call(o,s):s?this._subscribe(o):this._trySubscribe(o))}),o}_trySubscribe(A){try{return this._subscribe(A)}catch(i){A.error(i)}}forEach(A,i){return i=oK(i),new i((n,o)=>{let r=new r0({next:s=>{try{A(s)}catch(a){o(a),r.unsubscribe()}},error:o,complete:n});this.subscribe(r)})}_subscribe(A){var i;return(i=this.source)===null||i===void 0?void 0:i.subscribe(A)}[ud](){return this}pipe(...A){return av(A)(this)}toPromise(A){return A=oK(A),new A((i,n)=>{let o;this.subscribe(r=>o=r,r=>n(r),()=>i(o))})}}return t.create=e=>new t(e),t})();function oK(t){var e;return(e=t??$c.Promise)!==null&&e!==void 0?e:Promise}function ZgA(t){return t&&Xt(t.next)&&Xt(t.error)&&Xt(t.complete)}function WgA(t){return t&&t instanceof o0||ZgA(t)&&Am(t)}function cv(t){return Xt(t?.lift)}function li(t){return e=>{if(cv(e))return e.lift(function(A){try{return t(A,this)}catch(i){this.error(i)}});throw new TypeError("Unable to lift unknown Observable type")}}function ai(t,e,A,i,n){return new lv(t,e,A,i,n)}var lv=class extends o0{constructor(e,A,i,n,o,r){super(e),this.onFinalize=o,this.shouldUnsubscribe=r,this._next=A?function(s){try{A(s)}catch(a){e.error(a)}}:super._next,this._error=n?function(s){try{n(s)}catch(a){e.error(a)}finally{this.unsubscribe()}}:super._error,this._complete=i?function(){try{i()}catch(s){e.error(s)}finally{this.unsubscribe()}}:super._complete}unsubscribe(){var e;if(!this.shouldUnsubscribe||this.shouldUnsubscribe()){let{closed:A}=this;super.unsubscribe(),!A&&((e=this.onFinalize)===null||e===void 0||e.call(this))}}};function fd(){return li((t,e)=>{let A=null;t._refCount++;let i=ai(e,void 0,void 0,void 0,()=>{if(!t||t._refCount<=0||0<--t._refCount){A=null;return}let n=t._connection,o=A;A=null,n&&(!o||n===o)&&n.unsubscribe(),e.unsubscribe()});t.subscribe(i),i.closed||(A=t.connect())})}var l2=class extends ct{constructor(e,A){super(),this.source=e,this.subjectFactory=A,this._subject=null,this._refCount=0,this._connection=null,cv(e)&&(this.lift=e.lift)}_subscribe(e){return this.getSubject().subscribe(e)}getSubject(){let e=this._subject;return(!e||e.isStopped)&&(this._subject=this.subjectFactory()),this._subject}_teardown(){this._refCount=0;let{_connection:e}=this;this._subject=this._connection=null,e?.unsubscribe()}connect(){let e=this._connection;if(!e){e=this._connection=new Kt;let A=this.getSubject();e.add(this.source.subscribe(ai(A,void 0,()=>{this._teardown(),A.complete()},i=>{this._teardown(),A.error(i)},()=>this._teardown()))),e.closed&&(this._connection=null,e=Kt.EMPTY)}return e}refCount(){return fd()(this)}};var rK=Ed(t=>function(){t(this),this.name="ObjectUnsubscribedError",this.message="object unsubscribed"});var OA=(()=>{class t extends ct{constructor(){super(),this.closed=!1,this.currentObservers=null,this.observers=[],this.isStopped=!1,this.hasError=!1,this.thrownError=null}lift(A){let i=new md(this,this);return i.operator=A,i}_throwIfClosed(){if(this.closed)throw new rK}next(A){hd(()=>{if(this._throwIfClosed(),!this.isStopped){this.currentObservers||(this.currentObservers=Array.from(this.observers));for(let i of this.currentObservers)i.next(A)}})}error(A){hd(()=>{if(this._throwIfClosed(),!this.isStopped){this.hasError=this.isStopped=!0,this.thrownError=A;let{observers:i}=this;for(;i.length;)i.shift().error(A)}})}complete(){hd(()=>{if(this._throwIfClosed(),!this.isStopped){this.isStopped=!0;let{observers:A}=this;for(;A.length;)A.shift().complete()}})}unsubscribe(){this.isStopped=this.closed=!0,this.observers=this.currentObservers=null}get observed(){var A;return((A=this.observers)===null||A===void 0?void 0:A.length)>0}_trySubscribe(A){return this._throwIfClosed(),super._trySubscribe(A)}_subscribe(A){return this._throwIfClosed(),this._checkFinalizedStatuses(A),this._innerSubscribe(A)}_innerSubscribe(A){let{hasError:i,isStopped:n,observers:o}=this;return i||n?tv:(this.currentObservers=null,o.push(A),new Kt(()=>{this.currentObservers=null,cI(o,A)}))}_checkFinalizedStatuses(A){let{hasError:i,thrownError:n,isStopped:o}=this;i?A.error(n):o&&A.complete()}asObservable(){let A=new ct;return A.source=this,A}}return t.create=(e,A)=>new md(e,A),t})(),md=class extends OA{constructor(e,A){super(),this.destination=e,this.source=A}next(e){var A,i;(i=(A=this.destination)===null||A===void 0?void 0:A.next)===null||i===void 0||i.call(A,e)}error(e){var A,i;(i=(A=this.destination)===null||A===void 0?void 0:A.error)===null||i===void 0||i.call(A,e)}complete(){var e,A;(A=(e=this.destination)===null||e===void 0?void 0:e.complete)===null||A===void 0||A.call(e)}_subscribe(e){var A,i;return(i=(A=this.source)===null||A===void 0?void 0:A.subscribe(e))!==null&&i!==void 0?i:tv}};var Mi=class extends OA{constructor(e){super(),this._value=e}get value(){return this.getValue()}_subscribe(e){let A=super._subscribe(e);return!A.closed&&e.next(this._value),A}getValue(){let{hasError:e,thrownError:A,_value:i}=this;if(e)throw A;return this._throwIfClosed(),i}next(e){super.next(this._value=e)}};var eh={now(){return(eh.delegate||Date).now()},delegate:void 0};var Al=class extends OA{constructor(e=1/0,A=1/0,i=eh){super(),this._bufferSize=e,this._windowTime=A,this._timestampProvider=i,this._buffer=[],this._infiniteTimeWindow=!0,this._infiniteTimeWindow=A===1/0,this._bufferSize=Math.max(1,e),this._windowTime=Math.max(1,A)}next(e){let{isStopped:A,_buffer:i,_infiniteTimeWindow:n,_timestampProvider:o,_windowTime:r}=this;A||(i.push(e),!n&&i.push(o.now()+r)),this._trimBuffer(),super.next(e)}_subscribe(e){this._throwIfClosed(),this._trimBuffer();let A=this._innerSubscribe(e),{_infiniteTimeWindow:i,_buffer:n}=this,o=n.slice();for(let r=0;r<o.length&&!e.closed;r+=i?1:2)e.next(o[r]);return this._checkFinalizedStatuses(e),A}_trimBuffer(){let{_bufferSize:e,_timestampProvider:A,_buffer:i,_infiniteTimeWindow:n}=this,o=(n?1:2)*e;if(e<1/0&&o<i.length&&i.splice(0,i.length-o),!n){let r=A.now(),s=0;for(let a=1;a<i.length&&i[a]<=r;a+=2)s=a;s&&i.splice(0,s+1)}}};var im=class extends Kt{constructor(e,A){super()}schedule(e,A=0){return this}};var th={setInterval(t,e,...A){let{delegate:i}=th;return i?.setInterval?i.setInterval(t,e,...A):setInterval(t,e,...A)},clearInterval(t){let{delegate:e}=th;return(e?.clearInterval||clearInterval)(t)},delegate:void 0};var nm=class extends im{constructor(e,A){super(e,A),this.scheduler=e,this.work=A,this.pending=!1}schedule(e,A=0){var i;if(this.closed)return this;this.state=e;let n=this.id,o=this.scheduler;return n!=null&&(this.id=this.recycleAsyncId(o,n,A)),this.pending=!0,this.delay=A,this.id=(i=this.id)!==null&&i!==void 0?i:this.requestAsyncId(o,this.id,A),this}requestAsyncId(e,A,i=0){return th.setInterval(e.flush.bind(e,this),i)}recycleAsyncId(e,A,i=0){if(i!=null&&this.delay===i&&this.pending===!1)return A;A!=null&&th.clearInterval(A)}execute(e,A){if(this.closed)return new Error("executing a cancelled action");this.pending=!1;let i=this._execute(e,A);if(i)return i;this.pending===!1&&this.id!=null&&(this.id=this.recycleAsyncId(this.scheduler,this.id,null))}_execute(e,A){let i=!1,n;try{this.work(e)}catch(o){i=!0,n=o||new Error("Scheduled action threw falsy error")}if(i)return this.unsubscribe(),n}unsubscribe(){if(!this.closed){let{id:e,scheduler:A}=this,{actions:i}=A;this.work=this.state=this.scheduler=null,this.pending=!1,cI(i,this),e!=null&&(this.id=this.recycleAsyncId(A,e,null)),this.delay=null,super.unsubscribe()}}};var pd=class t{constructor(e,A=t.now){this.schedulerActionCtor=e,this.now=A}schedule(e,A=0,i){return new this.schedulerActionCtor(this,e).schedule(i,A)}};pd.now=eh.now;var om=class extends pd{constructor(e,A=pd.now){super(e,A),this.actions=[],this._active=!1}flush(e){let{actions:A}=this;if(this._active){A.push(e);return}let i;this._active=!0;do if(i=e.execute(e.state,e.delay))break;while(e=A.shift());if(this._active=!1,i){for(;e=A.shift();)e.unsubscribe();throw i}}};var ih=new om(nm),sK=ih;var sr=new ct(t=>t.complete());function rm(t){return t&&Xt(t.schedule)}function gv(t){return t[t.length-1]}function sm(t){return Xt(gv(t))?t.pop():void 0}function Pl(t){return rm(gv(t))?t.pop():void 0}function aK(t,e){return typeof gv(t)=="number"?t.pop():e}function lK(t,e,A,i){function n(o){return o instanceof A?o:new A(function(r){r(o)})}return new(A||(A=Promise))(function(o,r){function s(l){try{c(i.next(l))}catch(I){r(I)}}function a(l){try{c(i.throw(l))}catch(I){r(I)}}function c(l){l.done?o(l.value):n(l.value).then(s,a)}c((i=i.apply(t,e||[])).next())})}function cK(t){var e=typeof Symbol=="function"&&Symbol.iterator,A=e&&t[e],i=0;if(A)return A.call(t);if(t&&typeof t.length=="number")return{next:function(){return t&&i>=t.length&&(t=void 0),{value:t&&t[i++],done:!t}}};throw new TypeError(e?"Object is not iterable.":"Symbol.iterator is not defined.")}function gI(t){return this instanceof gI?(this.v=t,this):new gI(t)}function gK(t,e,A){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var i=A.apply(t,e||[]),n,o=[];return n=Object.create((typeof AsyncIterator=="function"?AsyncIterator:Object).prototype),s("next"),s("throw"),s("return",r),n[Symbol.asyncIterator]=function(){return this},n;function r(d){return function(B){return Promise.resolve(B).then(d,I)}}function s(d,B){i[d]&&(n[d]=function(E){return new Promise(function(h,u){o.push([d,E,h,u])>1||a(d,E)})},B&&(n[d]=B(n[d])))}function a(d,B){try{c(i[d](B))}catch(E){C(o[0][3],E)}}function c(d){d.value instanceof gI?Promise.resolve(d.value.v).then(l,I):C(o[0][2],d)}function l(d){a("next",d)}function I(d){a("throw",d)}function C(d,B){d(B),o.shift(),o.length&&a(o[0][0],o[0][1])}}function IK(t){if(!Symbol.asyncIterator)throw new TypeError("Symbol.asyncIterator is not defined.");var e=t[Symbol.asyncIterator],A;return e?e.call(t):(t=typeof cK=="function"?cK(t):t[Symbol.iterator](),A={},i("next"),i("throw"),i("return"),A[Symbol.asyncIterator]=function(){return this},A);function i(o){A[o]=t[o]&&function(r){return new Promise(function(s,a){r=t[o](r),n(s,a,r.done,r.value)})}}function n(o,r,s,a){Promise.resolve(a).then(function(c){o({value:c,done:s})},r)}}var wd=t=>t&&typeof t.length=="number"&&typeof t!="function";function am(t){return Xt(t?.then)}function cm(t){return Xt(t[ud])}function lm(t){return Symbol.asyncIterator&&Xt(t?.[Symbol.asyncIterator])}function gm(t){return new TypeError(`You provided ${t!==null&&typeof t=="object"?"an invalid object":`'${t}'`} where a stream was expected. You can provide an Observable, Promise, ReadableStream, Array, AsyncIterable, or Iterable.`)}function XgA(){return typeof Symbol!="function"||!Symbol.iterator?"@@iterator":Symbol.iterator}var Im=XgA();function Cm(t){return Xt(t?.[Im])}function dm(t){return gK(this,arguments,function*(){let A=t.getReader();try{for(;;){let{value:i,done:n}=yield gI(A.read());if(n)return yield gI(void 0);yield yield gI(i)}}finally{A.releaseLock()}})}function Bm(t){return Xt(t?.getReader)}function eo(t){if(t instanceof ct)return t;if(t!=null){if(cm(t))return $gA(t);if(wd(t))return A0A(t);if(am(t))return e0A(t);if(lm(t))return CK(t);if(Cm(t))return t0A(t);if(Bm(t))return i0A(t)}throw gm(t)}function $gA(t){return new ct(e=>{let A=t[ud]();if(Xt(A.subscribe))return A.subscribe(e);throw new TypeError("Provided object does not correctly implement Symbol.observable")})}function A0A(t){return new ct(e=>{for(let A=0;A<t.length&&!e.closed;A++)e.next(t[A]);e.complete()})}function e0A(t){return new ct(e=>{t.then(A=>{e.closed||(e.next(A),e.complete())},A=>e.error(A)).then(null,em)})}function t0A(t){return new ct(e=>{for(let A of t)if(e.next(A),e.closed)return;e.complete()})}function CK(t){return new ct(e=>{n0A(t,e).catch(A=>e.error(A))})}function i0A(t){return CK(dm(t))}function n0A(t,e){var A,i,n,o;return lK(this,void 0,void 0,function*(){try{for(A=IK(t);i=yield A.next(),!i.done;){let r=i.value;if(e.next(r),e.closed)return}}catch(r){n={error:r}}finally{try{i&&!i.done&&(o=A.return)&&(yield o.call(A))}finally{if(n)throw n.error}}e.complete()})}function na(t,e,A,i=0,n=!1){let o=e.schedule(function(){A(),n?t.add(this.schedule(null,i)):this.unsubscribe()},i);if(t.add(o),!n)return o}function Em(t,e=0){return li((A,i)=>{A.subscribe(ai(i,n=>na(i,t,()=>i.next(n),e),()=>na(i,t,()=>i.complete(),e),n=>na(i,t,()=>i.error(n),e)))})}function Qm(t,e=0){return li((A,i)=>{i.add(t.schedule(()=>A.subscribe(i),e))})}function dK(t,e){return eo(t).pipe(Qm(e),Em(e))}function BK(t,e){return eo(t).pipe(Qm(e),Em(e))}function EK(t,e){return new ct(A=>{let i=0;return e.schedule(function(){i===t.length?A.complete():(A.next(t[i++]),A.closed||this.schedule())})})}function QK(t,e){return new ct(A=>{let i;return na(A,e,()=>{i=t[Im](),na(A,e,()=>{let n,o;try{({value:n,done:o}=i.next())}catch(r){A.error(r);return}o?A.complete():A.next(n)},0,!0)}),()=>Xt(i?.return)&&i.return()})}function hm(t,e){if(!t)throw new Error("Iterable cannot be null");return new ct(A=>{na(A,e,()=>{let i=t[Symbol.asyncIterator]();na(A,e,()=>{i.next().then(n=>{n.done?A.complete():A.next(n.value)})},0,!0)})})}function hK(t,e){return hm(dm(t),e)}function uK(t,e){if(t!=null){if(cm(t))return dK(t,e);if(wd(t))return EK(t,e);if(am(t))return BK(t,e);if(lm(t))return hm(t,e);if(Cm(t))return QK(t,e);if(Bm(t))return hK(t,e)}throw gm(t)}function oo(t,e){return e?uK(t,e):eo(t)}function Me(...t){let e=Pl(t);return oo(t,e)}function g2(t,e){let A=Xt(t)?t:()=>t,i=n=>n.error(A());return new ct(e?n=>e.schedule(i,0,n):i)}function I2(t){return!!t&&(t instanceof ct||Xt(t.lift)&&Xt(t.subscribe))}var s0=Ed(t=>function(){t(this),this.name="EmptyError",this.message="no elements in sequence"});function fK(t){return t instanceof Date&&!isNaN(t)}function je(t,e){return li((A,i)=>{let n=0;A.subscribe(ai(i,o=>{i.next(t.call(e,o,n++))}))})}var{isArray:o0A}=Array;function r0A(t,e){return o0A(e)?t(...e):t(e)}function Dd(t){return je(e=>r0A(t,e))}var{isArray:s0A}=Array,{getPrototypeOf:a0A,prototype:c0A,keys:l0A}=Object;function um(t){if(t.length===1){let e=t[0];if(s0A(e))return{args:e,keys:null};if(g0A(e)){let A=l0A(e);return{args:A.map(i=>e[i]),keys:A}}}return{args:t,keys:null}}function g0A(t){return t&&typeof t=="object"&&a0A(t)===c0A}function fm(t,e){return t.reduce((A,i,n)=>(A[i]=e[n],A),{})}function Js(...t){let e=Pl(t),A=sm(t),{args:i,keys:n}=um(t);if(i.length===0)return oo([],e);let o=new ct(I0A(i,e,n?r=>fm(n,r):Ys));return A?o.pipe(Dd(A)):o}function I0A(t,e,A=Ys){return i=>{mK(e,()=>{let{length:n}=t,o=new Array(n),r=n,s=n;for(let a=0;a<n;a++)mK(e,()=>{let c=oo(t[a],e),l=!1;c.subscribe(ai(i,I=>{o[a]=I,l||(l=!0,s--),s||i.next(A(o.slice()))},()=>{--r||i.complete()}))},i)},i)}}function mK(t,e,A){t?na(A,t,e):e()}function pK(t,e,A,i,n,o,r,s){let a=[],c=0,l=0,I=!1,C=()=>{I&&!a.length&&!c&&e.complete()},d=E=>c<i?B(E):a.push(E),B=E=>{o&&e.next(E),c++;let h=!1;eo(A(E,l++)).subscribe(ai(e,u=>{n?.(u),o?d(u):e.next(u)},()=>{h=!0},void 0,()=>{if(h)try{for(c--;a.length&&c<i;){let u=a.shift();r?na(e,r,()=>B(u)):B(u)}C()}catch(u){e.error(u)}}))};return t.subscribe(ai(e,d,()=>{I=!0,C()})),()=>{s?.()}}function tr(t,e,A=1/0){return Xt(e)?tr((i,n)=>je((o,r)=>e(i,o,n,r))(eo(t(i,n))),A):(typeof e=="number"&&(A=e),li((i,n)=>pK(i,n,t,A)))}function C2(t=1/0){return tr(Ys,t)}function wK(){return C2(1)}function d2(...t){return wK()(oo(t,Pl(t)))}function jl(t){return new ct(e=>{eo(t()).subscribe(e)})}function nh(...t){let e=sm(t),{args:A,keys:i}=um(t),n=new ct(o=>{let{length:r}=A;if(!r){o.complete();return}let s=new Array(r),a=r,c=r;for(let l=0;l<r;l++){let I=!1;eo(A[l]).subscribe(ai(o,C=>{I||(I=!0,c--),s[l]=C},()=>a--,void 0,()=>{(!a||!I)&&(c||o.next(i?fm(i,s):s),o.complete())}))}});return e?n.pipe(Dd(e)):n}var C0A=["addListener","removeListener"],d0A=["addEventListener","removeEventListener"],B0A=["on","off"];function oh(t,e,A,i){if(Xt(A)&&(i=A,A=void 0),i)return oh(t,e,A).pipe(Dd(i));let[n,o]=h0A(t)?d0A.map(r=>s=>t[r](e,s,A)):E0A(t)?C0A.map(DK(t,e)):Q0A(t)?B0A.map(DK(t,e)):[];if(!n&&wd(t))return tr(r=>oh(r,e,A))(eo(t));if(!n)throw new TypeError("Invalid event target");return new ct(r=>{let s=(...a)=>r.next(1<a.length?a:a[0]);return n(s),()=>o(s)})}function DK(t,e){return A=>i=>t[A](e,i)}function E0A(t){return Xt(t.addListener)&&Xt(t.removeListener)}function Q0A(t){return Xt(t.on)&&Xt(t.off)}function h0A(t){return Xt(t.addEventListener)&&Xt(t.removeEventListener)}function II(t=0,e,A=sK){let i=-1;return e!=null&&(rm(e)?A=e:i=e),new ct(n=>{let o=fK(t)?+t-A.now():t;o<0&&(o=0);let r=0;return A.schedule(function(){n.closed||(n.next(r++),0<=i?this.schedule(void 0,i):n.complete())},o)})}function zn(...t){let e=Pl(t),A=aK(t,1/0),i=t;return i.length?i.length===1?eo(i[0]):C2(A)(oo(i,e)):sr}function kt(t,e){return li((A,i)=>{let n=0;A.subscribe(ai(i,o=>t.call(e,o,n++)&&i.next(o)))})}function yK(t){return li((e,A)=>{let i=!1,n=null,o=null,r=!1,s=()=>{if(o?.unsubscribe(),o=null,i){i=!1;let c=n;n=null,A.next(c)}r&&A.complete()},a=()=>{o=null,r&&A.complete()};e.subscribe(ai(A,c=>{i=!0,n=c,o||eo(t(c)).subscribe(o=ai(A,s,a))},()=>{r=!0,(!i||!o||o.closed)&&A.complete()}))})}function yd(t,e=ih){return yK(()=>II(t,e))}function mr(t){return li((e,A)=>{let i=null,n=!1,o;i=e.subscribe(ai(A,void 0,void 0,r=>{o=eo(t(r,mr(t)(e))),i?(i.unsubscribe(),i=null,o.subscribe(A)):n=!0})),n&&(i.unsubscribe(),i=null,o.subscribe(A))})}function vK(t,e,A,i,n){return(o,r)=>{let s=A,a=e,c=0;o.subscribe(ai(r,l=>{let I=c++;a=s?t(a,l,I):(s=!0,l),i&&r.next(a)},n&&(()=>{s&&r.next(a),r.complete()})))}}function ql(t,e){return Xt(e)?tr(t,e,1):tr(t,1)}function el(t,e=ih){return li((A,i)=>{let n=null,o=null,r=null,s=()=>{if(n){n.unsubscribe(),n=null;let c=o;o=null,i.next(c)}};function a(){let c=r+t,l=e.now();if(l<c){n=this.schedule(void 0,c-l),i.add(n);return}s()}A.subscribe(ai(i,c=>{o=c,r=e.now(),n||(n=e.schedule(a,t),i.add(n))},()=>{s(),i.complete()},void 0,()=>{o=n=null}))})}function B2(t){return li((e,A)=>{let i=!1;e.subscribe(ai(A,n=>{i=!0,A.next(n)},()=>{i||A.next(t),A.complete()}))})}function On(t){return t<=0?()=>sr:li((e,A)=>{let i=0;e.subscribe(ai(A,n=>{++i<=t&&(A.next(n),t<=i&&A.complete())}))})}function vd(t){return je(()=>t)}function tl(t,e=Ys){return t=t??u0A,li((A,i)=>{let n,o=!0;A.subscribe(ai(i,r=>{let s=e(r);(o||!t(n,s))&&(o=!1,n=s,i.next(r))}))})}function u0A(t,e){return t===e}function mm(t=f0A){return li((e,A)=>{let i=!1;e.subscribe(ai(A,n=>{i=!0,A.next(n)},()=>i?A.complete():A.error(t())))})}function f0A(){return new s0}function Vl(t){return li((e,A)=>{try{e.subscribe(A)}finally{A.add(t)}})}function Zl(t,e){let A=arguments.length>=2;return i=>i.pipe(t?kt((n,o)=>t(n,o,i)):Ys,On(1),A?B2(e):mm(()=>new s0))}function bd(t){return t<=0?()=>sr:li((e,A)=>{let i=[];e.subscribe(ai(A,n=>{i.push(n),t<i.length&&i.shift()},()=>{for(let n of i)A.next(n);A.complete()},void 0,()=>{i=null}))})}function Iv(t,e){let A=arguments.length>=2;return i=>i.pipe(t?kt((n,o)=>t(n,o,i)):Ys,bd(1),A?B2(e):mm(()=>new s0))}function pm(){return li((t,e)=>{let A,i=!1;t.subscribe(ai(e,n=>{let o=A;A=n,i&&e.next([o,n]),i=!0}))})}function Cv(t,e){return li(vK(t,e,arguments.length>=2,!0))}function rh(t={}){let{connector:e=()=>new OA,resetOnError:A=!0,resetOnComplete:i=!0,resetOnRefCountZero:n=!0}=t;return o=>{let r,s,a,c=0,l=!1,I=!1,C=()=>{s?.unsubscribe(),s=void 0},d=()=>{C(),r=a=void 0,l=I=!1},B=()=>{let E=r;d(),E?.unsubscribe()};return li((E,h)=>{c++,!I&&!l&&C();let u=a=a??e();h.add(()=>{c--,c===0&&!I&&!l&&(s=dv(B,n))}),u.subscribe(h),!r&&c>0&&(r=new r0({next:D=>u.next(D),error:D=>{I=!0,C(),s=dv(d,A,D),u.error(D)},complete:()=>{l=!0,C(),s=dv(d,i),u.complete()}}),eo(E).subscribe(r))})(o)}}function dv(t,e,...A){if(e===!0){t();return}if(e===!1)return;let i=new r0({next:()=>{i.unsubscribe(),t()}});return eo(e(...A)).subscribe(i)}function a0(t,e,A){let i,n=!1;return t&&typeof t=="object"?{bufferSize:i=1/0,windowTime:e=1/0,refCount:n=!1,scheduler:A}=t:i=t??1/0,rh({connector:()=>new Al(i,e,A),resetOnError:!0,resetOnComplete:!1,resetOnRefCountZero:n})}function CI(t){return kt((e,A)=>t<=A)}function Pn(...t){let e=Pl(t);return li((A,i)=>{(e?d2(t,A,e):d2(t,A)).subscribe(i)})}function jn(t,e){return li((A,i)=>{let n=null,o=0,r=!1,s=()=>r&&!n&&i.complete();A.subscribe(ai(i,a=>{n?.unsubscribe();let c=0,l=o++;eo(t(a,l)).subscribe(n=ai(i,I=>i.next(e?e(a,I,l,c++):I),()=>{n=null,s()}))},()=>{r=!0,s()}))})}function St(t){return li((e,A)=>{eo(t).subscribe(ai(A,()=>A.complete(),Ah)),!A.closed&&e.subscribe(A)})}function Bv(t,e=!1){return li((A,i)=>{let n=0;A.subscribe(ai(i,o=>{let r=t(o,n++);(r||e)&&i.next(o),!r&&i.complete()}))})}function Qo(t,e,A){let i=Xt(t)||e||A?{next:t,error:e,complete:A}:t;return i?li((n,o)=>{var r;(r=i.subscribe)===null||r===void 0||r.call(i);let s=!0;n.subscribe(ai(o,a=>{var c;(c=i.next)===null||c===void 0||c.call(i,a),o.next(a)},()=>{var a;s=!1,(a=i.complete)===null||a===void 0||a.call(i),o.complete()},a=>{var c;s=!1,(c=i.error)===null||c===void 0||c.call(i,a),o.error(a)},()=>{var a,c;s&&((a=i.unsubscribe)===null||a===void 0||a.call(i)),(c=i.finalize)===null||c===void 0||c.call(i)}))}):Ys}var uY="https://angular.dev/best-practices/security#preventing-cross-site-scripting-xss",Ae=class extends Error{code;constructor(e,A){super(K9(e,A)),this.code=e}};function m0A(t){return`NG0${Math.abs(t)}`}function K9(t,e){return`${m0A(t)}${e?": "+e:""}`}var fY=Symbol("InputSignalNode#UNSET"),p0A=Ye(rA({},Wf),{transformFn:void 0,applyValueToInputSignal(t,e){XQ(t,e)}});function mY(t,e){let A=Object.create(p0A);A.value=t,A.transformFn=e?.transform;function i(){if(VQ(A),A.value===fY){let n=null;throw new Ae(-950,n)}return A.value}return i[ia]=A,i}function uh(t){return{toString:t}.toString()}var wm="__parameters__";function w0A(t){return function(...A){if(t){let i=t(...A);for(let n in i)this[n]=i[n]}}}function pY(t,e,A){return uh(()=>{let i=w0A(e);function n(...o){if(this instanceof n)return i.apply(this,o),this;let r=new n(...o);return s.annotation=r,s;function s(a,c,l){let I=a.hasOwnProperty(wm)?a[wm]:Object.defineProperty(a,wm,{value:[]})[wm];for(;I.length<=l;)I.push(null);return(I[l]=I[l]||[]).push(r),a}}return n.prototype.ngMetadataName=t,n.annotationCls=n,n})}var sa=globalThis;function ho(t){for(let e in t)if(t[e]===ho)return e;throw Error("Could not find renamed property on target object.")}function D0A(t,e){for(let A in e)e.hasOwnProperty(A)&&!t.hasOwnProperty(A)&&(t[A]=e[A])}function ra(t){if(typeof t=="string")return t;if(Array.isArray(t))return`[${t.map(ra).join(", ")}]`;if(t==null)return""+t;let e=t.overriddenName||t.name;if(e)return`${e}`;let A=t.toString();if(A==null)return""+A;let i=A.indexOf(`
`);return i>=0?A.slice(0,i):A}function Sv(t,e){return t?e?`${t} ${e}`:t:e||""}var y0A=ho({__forward_ref__:ho});function Ir(t){return t.__forward_ref__=Ir,t.toString=function(){return ra(this())},t}function ns(t){return wY(t)?t():t}function wY(t){return typeof t=="function"&&t.hasOwnProperty(y0A)&&t.__forward_ref__===Ir}function NA(t){return{token:t.token,providedIn:t.providedIn||null,factory:t.factory,value:void 0}}function Ie(t){return{providers:t.providers||[],imports:t.imports||[]}}function lp(t){return bK(t,yY)||bK(t,vY)}function DY(t){return lp(t)!==null}function bK(t,e){return t.hasOwnProperty(e)?t[e]:null}function v0A(t){let e=t&&(t[yY]||t[vY]);return e||null}function MK(t){return t&&(t.hasOwnProperty(kK)||t.hasOwnProperty(b0A))?t[kK]:null}var yY=ho({\u0275prov:ho}),kK=ho({\u0275inj:ho}),vY=ho({ngInjectableDef:ho}),b0A=ho({ngInjectorDef:ho}),dA=class{_desc;ngMetadataName="InjectionToken";\u0275prov;constructor(e,A){this._desc=e,this.\u0275prov=void 0,typeof A=="number"?this.__NG_ELEMENT_ID__=A:A!==void 0&&(this.\u0275prov=NA({token:this,providedIn:A.providedIn||"root",factory:A.factory}))}get multi(){return this}toString(){return`InjectionToken ${this._desc}`}};function bY(t){return t&&!!t.\u0275providers}var M0A=ho({\u0275cmp:ho}),k0A=ho({\u0275dir:ho}),S0A=ho({\u0275pipe:ho}),R0A=ho({\u0275mod:ho}),Fm=ho({\u0275fac:ho}),lh=ho({__NG_ELEMENT_ID__:ho}),SK=ho({__NG_ENV_ID__:ho});function EI(t){return typeof t=="string"?t:t==null?"":String(t)}function x0A(t){return typeof t=="function"?t.name||t.toString():typeof t=="object"&&t!=null&&typeof t.type=="function"?t.type.name||t.type.toString():EI(t)}function MY(t,e){throw new Ae(-200,t)}function Y9(t,e){throw new Ae(-201,!1)}var Gi=function(t){return t[t.Default=0]="Default",t[t.Host=1]="Host",t[t.Self=2]="Self",t[t.SkipSelf=4]="SkipSelf",t[t.Optional=8]="Optional",t}(Gi||{}),Rv;function kY(){return Rv}function oa(t){let e=Rv;return Rv=t,e}function SY(t,e,A){let i=lp(t);if(i&&i.providedIn=="root")return i.value===void 0?i.value=i.factory():i.value;if(A&Gi.Optional)return null;if(e!==void 0)return e;Y9(t,"Injector")}var L0A={},dI=L0A,xv="__NG_DI_FLAG__",Nm=class{injector;constructor(e){this.injector=e}retrieve(e,A){let i=A;return this.injector.get(e,i.optional?Xf:dI,i)}},_m="ngTempTokenPath",F0A="ngTokenPath",N0A=/\n/gm,_0A="\u0275",RK="__source";function G0A(t,e=Gi.Default){if($Q()===void 0)throw new Ae(-203,!1);if($Q()===null)return SY(t,void 0,e);{let A=$Q(),i;return A instanceof Nm?i=A.injector:i=A,i.get(t,e&Gi.Optional?null:void 0,e)}}function we(t,e=Gi.Default){return(kY()||G0A)(ns(t),e)}function f(t,e=Gi.Default){return we(t,gp(e))}function gp(t){return typeof t>"u"||typeof t=="number"?t:0|(t.optional&&8)|(t.host&&1)|(t.self&&2)|(t.skipSelf&&4)}function Lv(t){let e=[];for(let A=0;A<t.length;A++){let i=ns(t[A]);if(Array.isArray(i)){if(i.length===0)throw new Ae(900,!1);let n,o=Gi.Default;for(let r=0;r<i.length;r++){let s=i[r],a=U0A(s);typeof a=="number"?a===-1?n=s.token:o|=a:n=s}e.push(we(n,o))}else e.push(we(i))}return e}function RY(t,e){return t[xv]=e,t.prototype[xv]=e,t}function U0A(t){return t[xv]}function K0A(t,e,A,i){let n=t[_m];throw e[RK]&&n.unshift(e[RK]),t.message=Y0A(`
`+t.message,n,A,i),t[F0A]=n,t[_m]=null,t}function Y0A(t,e,A,i=null){t=t&&t.charAt(0)===`
`&&t.charAt(1)==_0A?t.slice(2):t;let n=ra(e);if(Array.isArray(e))n=e.map(ra).join(" -> ");else if(typeof e=="object"){let o=[];for(let r in e)if(e.hasOwnProperty(r)){let s=e[r];o.push(r+":"+(typeof s=="string"?JSON.stringify(s):ra(s)))}n=`{${o.join(", ")}}`}return`${A}${i?"("+i+")":""}[${n}]: ${t.replace(N0A,`
  `)}`}var MI=RY(pY("Optional"),8);var fh=RY(pY("SkipSelf"),4);function QI(t,e){let A=t.hasOwnProperty(Fm);return A?t[Fm]:null}function J0A(t,e,A){if(t.length!==e.length)return!1;for(let i=0;i<t.length;i++){let n=t[i],o=e[i];if(A&&(n=A(n),o=A(o)),o!==n)return!1}return!0}function T0A(t){return t.flat(Number.POSITIVE_INFINITY)}function J9(t,e){t.forEach(A=>Array.isArray(A)?J9(A,e):e(A))}function xY(t,e,A){e>=t.length?t.push(A):t.splice(e,0,A)}function Gm(t,e){return e>=t.length-1?t.pop():t.splice(e,1)[0]}function H0A(t,e){let A=[];for(let i=0;i<t;i++)A.push(e);return A}function z0A(t,e,A,i){let n=t.length;if(n==e)t.push(A,i);else if(n===1)t.push(i,t[0]),t[0]=A;else{for(n--,t.push(t[n-1],t[n]);n>e;){let o=n-2;t[n]=t[o],n--}t[e]=A,t[e+1]=i}}function Ip(t,e,A){let i=mh(t,e);return i>=0?t[i|1]=A:(i=~i,z0A(t,i,e,A)),i}function Ev(t,e){let A=mh(t,e);if(A>=0)return t[A|1]}function mh(t,e){return O0A(t,e,1)}function O0A(t,e,A){let i=0,n=t.length>>A;for(;n!==i;){let o=i+(n-i>>1),r=t[o<<A];if(e===r)return o<<A;r>e?n=o:i=o+1}return~(n<<A)}var Xl={},Ts=[],Fd=new dA(""),LY=new dA("",-1),FY=new dA(""),Um=class{get(e,A=dI){if(A===dI){let i=new Error(`NullInjectorError: No provider for ${ra(e)}!`);throw i.name="NullInjectorError",i}return A}};function NY(t,e){let A=t[R0A]||null;if(!A&&e===!0)throw new Error(`Type ${ra(t)} does not have '\u0275mod' property.`);return A}function h2(t){return t[M0A]||null}function _Y(t){return t[k0A]||null}function P0A(t){return t[S0A]||null}function ph(t){return{\u0275providers:t}}function j0A(...t){return{\u0275providers:GY(!0,t),\u0275fromNgModule:!0}}function GY(t,...e){let A=[],i=new Set,n,o=r=>{A.push(r)};return J9(e,r=>{let s=r;Fv(s,o,[],i)&&(n||=[],n.push(s))}),n!==void 0&&UY(n,o),A}function UY(t,e){for(let A=0;A<t.length;A++){let{ngModule:i,providers:n}=t[A];T9(n,o=>{e(o,i)})}}function Fv(t,e,A,i){if(t=ns(t),!t)return!1;let n=null,o=MK(t),r=!o&&h2(t);if(!o&&!r){let a=t.ngModule;if(o=MK(a),o)n=a;else return!1}else{if(r&&!r.standalone)return!1;n=t}let s=i.has(n);if(r){if(s)return!1;if(i.add(n),r.dependencies){let a=typeof r.dependencies=="function"?r.dependencies():r.dependencies;for(let c of a)Fv(c,e,A,i)}}else if(o){if(o.imports!=null&&!s){i.add(n);let c;try{J9(o.imports,l=>{Fv(l,e,A,i)&&(c||=[],c.push(l))})}finally{}c!==void 0&&UY(c,e)}if(!s){let c=QI(n)||(()=>new n);e({provide:n,useFactory:c,deps:Ts},n),e({provide:FY,useValue:n,multi:!0},n),e({provide:Fd,useValue:()=>we(n),multi:!0},n)}let a=o.providers;if(a!=null&&!s){let c=t;T9(a,l=>{e(l,c)})}}else return!1;return n!==t&&t.providers!==void 0}function T9(t,e){for(let A of t)bY(A)&&(A=A.\u0275providers),Array.isArray(A)?T9(A,e):e(A)}var q0A=ho({provide:String,useValue:ho});function KY(t){return t!==null&&typeof t=="object"&&q0A in t}function V0A(t){return!!(t&&t.useExisting)}function Z0A(t){return!!(t&&t.useFactory)}function Nd(t){return typeof t=="function"}function W0A(t){return!!t.useClass}var Cp=new dA(""),km={},xK={},Qv;function dp(){return Qv===void 0&&(Qv=new Um),Qv}var pr=class{},Ih=class extends pr{parent;source;scopes;records=new Map;_ngOnDestroyHooks=new Set;_onDestroyHooks=[];get destroyed(){return this._destroyed}_destroyed=!1;injectorDefTypes;constructor(e,A,i,n){super(),this.parent=A,this.source=i,this.scopes=n,_v(e,r=>this.processProvider(r)),this.records.set(LY,Md(void 0,this)),n.has("environment")&&this.records.set(pr,Md(void 0,this));let o=this.records.get(Cp);o!=null&&typeof o.value=="string"&&this.scopes.add(o.value),this.injectorDefTypes=new Set(this.get(FY,Ts,Gi.Self))}retrieve(e,A){let i=A;return this.get(e,i.optional?Xf:dI,i)}destroy(){ah(this),this._destroyed=!0;let e=Ti(null);try{for(let i of this._ngOnDestroyHooks)i.ngOnDestroy();let A=this._onDestroyHooks;this._onDestroyHooks=[];for(let i of A)i()}finally{this.records.clear(),this._ngOnDestroyHooks.clear(),this.injectorDefTypes.clear(),Ti(e)}}onDestroy(e){return ah(this),this._onDestroyHooks.push(e),()=>this.removeOnDestroy(e)}runInContext(e){ah(this);let A=n0(this),i=oa(void 0),n;try{return e()}finally{n0(A),oa(i)}}get(e,A=dI,i=Gi.Default){if(ah(this),e.hasOwnProperty(SK))return e[SK](this);i=gp(i);let n,o=n0(this),r=oa(void 0);try{if(!(i&Gi.SkipSelf)){let a=this.records.get(e);if(a===void 0){let c=t2A(e)&&lp(e);c&&this.injectableDefInScope(c)?a=Md(Nv(e),km):a=null,this.records.set(e,a)}if(a!=null)return this.hydrate(e,a,i)}let s=i&Gi.Self?dp():this.parent;return A=i&Gi.Optional&&A===dI?null:A,s.get(e,A)}catch(s){if(s.name==="NullInjectorError"){if((s[_m]=s[_m]||[]).unshift(ra(e)),o)throw s;return K0A(s,e,"R3InjectorError",this.source)}else throw s}finally{oa(r),n0(o)}}resolveInjectorInitializers(){let e=Ti(null),A=n0(this),i=oa(void 0),n;try{let o=this.get(Fd,Ts,Gi.Self);for(let r of o)r()}finally{n0(A),oa(i),Ti(e)}}toString(){let e=[],A=this.records;for(let i of A.keys())e.push(ra(i));return`R3Injector[${e.join(", ")}]`}processProvider(e){e=ns(e);let A=Nd(e)?e:ns(e&&e.provide),i=$0A(e);if(!Nd(e)&&e.multi===!0){let n=this.records.get(A);n||(n=Md(void 0,km,!0),n.factory=()=>Lv(n.multi),this.records.set(A,n)),A=e,n.multi.push(e)}this.records.set(A,i)}hydrate(e,A,i){let n=Ti(null);try{return A.value===xK?MY(ra(e)):A.value===km&&(A.value=xK,A.value=A.factory(void 0,i)),typeof A.value=="object"&&A.value&&e2A(A.value)&&this._ngOnDestroyHooks.add(A.value),A.value}finally{Ti(n)}}injectableDefInScope(e){if(!e.providedIn)return!1;let A=ns(e.providedIn);return typeof A=="string"?A==="any"||this.scopes.has(A):this.injectorDefTypes.has(A)}removeOnDestroy(e){let A=this._onDestroyHooks.indexOf(e);A!==-1&&this._onDestroyHooks.splice(A,1)}};function Nv(t){let e=lp(t),A=e!==null?e.factory:QI(t);if(A!==null)return A;if(t instanceof dA)throw new Ae(204,!1);if(t instanceof Function)return X0A(t);throw new Ae(204,!1)}function X0A(t){if(t.length>0)throw new Ae(204,!1);let A=v0A(t);return A!==null?()=>A.factory(t):()=>new t}function $0A(t){if(KY(t))return Md(void 0,t.useValue);{let e=YY(t);return Md(e,km)}}function YY(t,e,A){let i;if(Nd(t)){let n=ns(t);return QI(n)||Nv(n)}else if(KY(t))i=()=>ns(t.useValue);else if(Z0A(t))i=()=>t.useFactory(...Lv(t.deps||[]));else if(V0A(t))i=(n,o)=>we(ns(t.useExisting),o!==void 0&&o&Gi.Optional?Gi.Optional:void 0);else{let n=ns(t&&(t.useClass||t.provide));if(A2A(t))i=()=>new n(...Lv(t.deps));else return QI(n)||Nv(n)}return i}function ah(t){if(t.destroyed)throw new Ae(205,!1)}function Md(t,e,A=!1){return{factory:t,value:e,multi:A?[]:void 0}}function A2A(t){return!!t.deps}function e2A(t){return t!==null&&typeof t=="object"&&typeof t.ngOnDestroy=="function"}function t2A(t){return typeof t=="function"||typeof t=="object"&&t instanceof dA}function _v(t,e){for(let A of t)Array.isArray(A)?_v(A,e):A&&bY(A)?_v(A.\u0275providers,e):e(A)}function ga(t,e){let A;t instanceof Ih?(ah(t),A=t):A=new Nm(t);let i,n=n0(A),o=oa(void 0);try{return e()}finally{n0(n),oa(o)}}function H9(){return kY()!==void 0||$Q()!=null}function z9(t){if(!H9())throw new Ae(-203,!1)}function i2A(t){let e=sa.ng;if(e&&e.\u0275compilerFacade)return e.\u0275compilerFacade;throw new Error("JIT compiler unavailable")}function n2A(t){return typeof t=="function"}var ig=0,ki=1,gi=2,ps=3,ol=4,Ia=5,_d=6,Km=7,Fr=8,Gd=9,c0=10,Wo=11,Ch=12,LK=13,Hd=14,Oa=15,hI=16,kd=17,l0=18,Bp=19,JY=20,E2=21,hv=22,uI=23,wc=24,xd=25,Nr=26,O9=1;var fI=7,Ym=8,Ud=9,ms=10;function Q2(t){return Array.isArray(t)&&typeof t[O9]=="object"}function C0(t){return Array.isArray(t)&&t[O9]===!0}function P9(t){return(t.flags&4)!==0}function zd(t){return t.componentOffset>-1}function Ep(t){return(t.flags&1)===1}function rl(t){return!!t.template}function Jm(t){return(t[gi]&512)!==0}function Od(t){return(t[gi]&256)===256}var Gv=class{previousValue;currentValue;firstChange;constructor(e,A,i){this.previousValue=e,this.currentValue=A,this.firstChange=i}isFirstChange(){return this.firstChange}};function TY(t,e,A,i){e!==null?e.applyValueToInputSignal(e,i):t[A]=i}var ti=(()=>{let t=()=>HY;return t.ngInherit=!0,t})();function HY(t){return t.type.prototype.ngOnChanges&&(t.setInput=r2A),o2A}function o2A(){let t=OY(this),e=t?.current;if(e){let A=t.previous;if(A===Xl)t.previous=e;else for(let i in e)A[i]=e[i];t.current=null,this.ngOnChanges(e)}}function r2A(t,e,A,i,n){let o=this.declaredInputs[i],r=OY(t)||s2A(t,{previous:Xl,current:null}),s=r.current||(r.current={}),a=r.previous,c=a[o];s[o]=new Gv(c&&c.currentValue,A,a===Xl),TY(t,e,n,A)}var zY="__ngSimpleChanges__";function OY(t){return t[zY]||null}function s2A(t,e){return t[zY]=e}var FK=null;var Ro=function(t,e=null,A){FK?.(t,e,A)},PY="svg",a2A="math";function $l(t){for(;Array.isArray(t);)t=t[ig];return t}function c2A(t){for(;Array.isArray(t);){if(typeof t[O9]=="object")return t;t=t[ig]}return null}function jY(t,e){return $l(e[t])}function ng(t,e){return $l(e[t.index])}function j9(t,e){return t.data[e]}function q9(t,e){return t[e]}function l2A(t,e,A,i){A>=t.data.length&&(t.data[A]=null,t.blueprint[A]=null),e[A]=i}function Ag(t,e){let A=e[t];return Q2(A)?A:A[ig]}function g2A(t){return(t[gi]&4)===4}function V9(t){return(t[gi]&128)===128}function I2A(t){return C0(t[ps])}function u2(t,e){return e==null?null:t[e]}function qY(t){t[kd]=0}function VY(t){t[gi]&1024||(t[gi]|=1024,V9(t)&&Pd(t))}function C2A(t,e){for(;t>0;)e=e[Hd],t--;return e}function Qp(t){return!!(t[gi]&9216||t[wc]?.dirty)}function Uv(t){t[c0].changeDetectionScheduler?.notify(8),t[gi]&64&&(t[gi]|=1024),Qp(t)&&Pd(t)}function Pd(t){t[c0].changeDetectionScheduler?.notify(0);let e=mI(t);for(;e!==null&&!(e[gi]&8192||(e[gi]|=8192,!V9(e)));)e=mI(e)}function ZY(t,e){if(Od(t))throw new Ae(911,!1);t[E2]===null&&(t[E2]=[]),t[E2].push(e)}function d2A(t,e){if(t[E2]===null)return;let A=t[E2].indexOf(e);A!==-1&&t[E2].splice(A,1)}function mI(t){let e=t[ps];return C0(e)?e[ps]:e}function Z9(t){return t[Km]??=[]}function W9(t){return t.cleanup??=[]}function B2A(t,e,A,i){let n=Z9(e);n.push(A),t.firstCreatePass&&W9(t).push(i,n.length-1)}var xi={lFrame:tJ(null),bindingsEnabled:!0,skipHydrationRootTNode:null};var Kv=!1;function E2A(){return xi.lFrame.elementDepthCount}function Q2A(){xi.lFrame.elementDepthCount++}function h2A(){xi.lFrame.elementDepthCount--}function X9(){return xi.bindingsEnabled}function WY(){return xi.skipHydrationRootTNode!==null}function u2A(t){return xi.skipHydrationRootTNode===t}function f2A(){xi.skipHydrationRootTNode=null}function ei(){return xi.lFrame.lView}function Yo(){return xi.lFrame.tView}function RA(t){return xi.lFrame.contextLView=t,t[Fr]}function xA(t){return xi.lFrame.contextLView=null,t}function os(){let t=XY();for(;t!==null&&t.type===64;)t=t.parent;return t}function XY(){return xi.lFrame.currentTNode}function m2A(){let t=xi.lFrame,e=t.currentTNode;return t.isParent?e:e.parent}function kI(t,e){let A=xi.lFrame;A.currentTNode=t,A.isParent=e}function $9(){return xi.lFrame.isParent}function Ab(){xi.lFrame.isParent=!1}function p2A(){return xi.lFrame.contextLView}function $Y(){return Kv}function Tm(t){let e=Kv;return Kv=t,e}function wh(){let t=xi.lFrame,e=t.bindingRootIndex;return e===-1&&(e=t.bindingRootIndex=t.tView.bindingStartIndex),e}function w2A(){return xi.lFrame.bindingIndex}function D2A(t){return xi.lFrame.bindingIndex=t}function f2(){return xi.lFrame.bindingIndex++}function eb(t){let e=xi.lFrame,A=e.bindingIndex;return e.bindingIndex=e.bindingIndex+t,A}function y2A(){return xi.lFrame.inI18n}function v2A(t,e){let A=xi.lFrame;A.bindingIndex=A.bindingRootIndex=t,Yv(e)}function b2A(){return xi.lFrame.currentDirectiveIndex}function Yv(t){xi.lFrame.currentDirectiveIndex=t}function tb(t){let e=xi.lFrame.currentDirectiveIndex;return e===-1?null:t[e]}function ib(){return xi.lFrame.currentQueryIndex}function hp(t){xi.lFrame.currentQueryIndex=t}function M2A(t){let e=t[ki];return e.type===2?e.declTNode:e.type===1?t[Ia]:null}function AJ(t,e,A){if(A&Gi.SkipSelf){let n=e,o=t;for(;n=n.parent,n===null&&!(A&Gi.Host);)if(n=M2A(o),n===null||(o=o[Hd],n.type&10))break;if(n===null)return!1;e=n,t=o}let i=xi.lFrame=eJ();return i.currentTNode=e,i.lView=t,!0}function nb(t){let e=eJ(),A=t[ki];xi.lFrame=e,e.currentTNode=A.firstChild,e.lView=t,e.tView=A,e.contextLView=t,e.bindingIndex=A.bindingStartIndex,e.inI18n=!1}function eJ(){let t=xi.lFrame,e=t===null?null:t.child;return e===null?tJ(t):e}function tJ(t){let e={currentTNode:null,isParent:!0,lView:null,tView:null,selectedIndex:-1,contextLView:null,elementDepthCount:0,currentNamespace:null,currentDirectiveIndex:-1,bindingRootIndex:-1,bindingIndex:-1,currentQueryIndex:0,parent:t,child:null,inI18n:!1};return t!==null&&(t.child=e),e}function iJ(){let t=xi.lFrame;return xi.lFrame=t.parent,t.currentTNode=null,t.lView=null,t}var nJ=iJ;function ob(){let t=iJ();t.isParent=!0,t.tView=null,t.selectedIndex=-1,t.contextLView=null,t.elementDepthCount=0,t.currentDirectiveIndex=-1,t.currentNamespace=null,t.bindingRootIndex=-1,t.bindingIndex=-1,t.currentQueryIndex=0}function k2A(t){return(xi.lFrame.contextLView=C2A(t,xi.lFrame.contextLView))[Fr]}function d0(){return xi.lFrame.selectedIndex}function pI(t){xi.lFrame.selectedIndex=t}function Dh(){let t=xi.lFrame;return j9(t.tView,t.selectedIndex)}function ar(){xi.lFrame.currentNamespace=PY}function SI(){S2A()}function S2A(){xi.lFrame.currentNamespace=null}function R2A(){return xi.lFrame.currentNamespace}var oJ=!0;function up(){return oJ}function fp(t){oJ=t}function x2A(t,e,A){let{ngOnChanges:i,ngOnInit:n,ngDoCheck:o}=e.type.prototype;if(i){let r=HY(e);(A.preOrderHooks??=[]).push(t,r),(A.preOrderCheckHooks??=[]).push(t,r)}n&&(A.preOrderHooks??=[]).push(0-t,n),o&&((A.preOrderHooks??=[]).push(t,o),(A.preOrderCheckHooks??=[]).push(t,o))}function rb(t,e){for(let A=e.directiveStart,i=e.directiveEnd;A<i;A++){let o=t.data[A].type.prototype,{ngAfterContentInit:r,ngAfterContentChecked:s,ngAfterViewInit:a,ngAfterViewChecked:c,ngOnDestroy:l}=o;r&&(t.contentHooks??=[]).push(-A,r),s&&((t.contentHooks??=[]).push(A,s),(t.contentCheckHooks??=[]).push(A,s)),a&&(t.viewHooks??=[]).push(-A,a),c&&((t.viewHooks??=[]).push(A,c),(t.viewCheckHooks??=[]).push(A,c)),l!=null&&(t.destroyHooks??=[]).push(A,l)}}function Sm(t,e,A){rJ(t,e,3,A)}function Rm(t,e,A,i){(t[gi]&3)===A&&rJ(t,e,A,i)}function uv(t,e){let A=t[gi];(A&3)===e&&(A&=16383,A+=1,t[gi]=A)}function rJ(t,e,A,i){let n=i!==void 0?t[kd]&65535:0,o=i??-1,r=e.length-1,s=0;for(let a=n;a<r;a++)if(typeof e[a+1]=="number"){if(s=e[a],i!=null&&s>=i)break}else e[a]<0&&(t[kd]+=65536),(s<o||o==-1)&&(L2A(t,A,e,a),t[kd]=(t[kd]&4294901760)+a+2),a++}function NK(t,e){Ro(4,t,e);let A=Ti(null);try{e.call(t)}finally{Ti(A),Ro(5,t,e)}}function L2A(t,e,A,i){let n=A[i]<0,o=A[i+1],r=n?-A[i]:A[i],s=t[r];n?t[gi]>>14<t[kd]>>16&&(t[gi]&3)===e&&(t[gi]+=16384,NK(s,o)):NK(s,o)}var Ld=-1,wI=class{factory;injectImpl;resolving=!1;canSeeViewProviders;multi;componentProviders;index;providerFactory;constructor(e,A,i){this.factory=e,this.canSeeViewProviders=A,this.injectImpl=i}};function F2A(t){return(t.flags&8)!==0}function N2A(t){return(t.flags&16)!==0}function _2A(t,e,A){let i=0;for(;i<A.length;){let n=A[i];if(typeof n=="number"){if(n!==0)break;i++;let o=A[i++],r=A[i++],s=A[i++];t.setAttribute(e,r,s,o)}else{let o=n,r=A[++i];G2A(o)?t.setProperty(e,o,r):t.setAttribute(e,o,r),i++}}return i}function sJ(t){return t===3||t===4||t===6}function G2A(t){return t.charCodeAt(0)===64}function Kd(t,e){if(!(e===null||e.length===0))if(t===null||t.length===0)t=e.slice();else{let A=-1;for(let i=0;i<e.length;i++){let n=e[i];typeof n=="number"?A=n:A===0||(A===-1||A===2?_K(t,A,n,null,e[++i]):_K(t,A,n,null,null))}}return t}function _K(t,e,A,i,n){let o=0,r=t.length;if(e===-1)r=-1;else for(;o<t.length;){let s=t[o++];if(typeof s=="number"){if(s===e){r=-1;break}else if(s>e){r=o-1;break}}}for(;o<t.length;){let s=t[o];if(typeof s=="number")break;if(s===A){n!==null&&(t[o+1]=n);return}o++,n!==null&&o++}r!==-1&&(t.splice(r,0,e),o=r+1),t.splice(o++,0,A),n!==null&&t.splice(o++,0,n)}function aJ(t){return t!==Ld}function Hm(t){return t&32767}function U2A(t){return t>>16}function zm(t,e){let A=U2A(t),i=e;for(;A>0;)i=i[Hd],A--;return i}var Jv=!0;function Om(t){let e=Jv;return Jv=t,e}var K2A=256,cJ=K2A-1,lJ=5,Y2A=0,Wl={};function J2A(t,e,A){let i;typeof A=="string"?i=A.charCodeAt(0)||0:A.hasOwnProperty(lh)&&(i=A[lh]),i==null&&(i=A[lh]=Y2A++);let n=i&cJ,o=1<<n;e.data[t+(n>>lJ)]|=o}function Pm(t,e){let A=gJ(t,e);if(A!==-1)return A;let i=e[ki];i.firstCreatePass&&(t.injectorIndex=e.length,fv(i.data,t),fv(e,null),fv(i.blueprint,null));let n=sb(t,e),o=t.injectorIndex;if(aJ(n)){let r=Hm(n),s=zm(n,e),a=s[ki].data;for(let c=0;c<8;c++)e[o+c]=s[r+c]|a[r+c]}return e[o+8]=n,o}function fv(t,e){t.push(0,0,0,0,0,0,0,0,e)}function gJ(t,e){return t.injectorIndex===-1||t.parent&&t.parent.injectorIndex===t.injectorIndex||e[t.injectorIndex+8]===null?-1:t.injectorIndex}function sb(t,e){if(t.parent&&t.parent.injectorIndex!==-1)return t.parent.injectorIndex;let A=0,i=null,n=e;for(;n!==null;){if(i=EJ(n),i===null)return Ld;if(A++,n=n[Hd],i.injectorIndex!==-1)return i.injectorIndex|A<<16}return Ld}function Tv(t,e,A){J2A(t,e,A)}function T2A(t,e){if(e==="class")return t.classes;if(e==="style")return t.styles;let A=t.attrs;if(A){let i=A.length,n=0;for(;n<i;){let o=A[n];if(sJ(o))break;if(o===0)n=n+2;else if(typeof o=="number")for(n++;n<i&&typeof A[n]=="string";)n++;else{if(o===e)return A[n+1];n=n+2}}}return null}function IJ(t,e,A){if(A&Gi.Optional||t!==void 0)return t;Y9(e,"NodeInjector")}function CJ(t,e,A,i){if(A&Gi.Optional&&i===void 0&&(i=null),(A&(Gi.Self|Gi.Host))===0){let n=t[Gd],o=oa(void 0);try{return n?n.get(e,i,A&Gi.Optional):SY(e,i,A&Gi.Optional)}finally{oa(o)}}return IJ(i,e,A)}function dJ(t,e,A,i=Gi.Default,n){if(t!==null){if(e[gi]&2048&&!(i&Gi.Self)){let r=P2A(t,e,A,i,Wl);if(r!==Wl)return r}let o=BJ(t,e,A,i,Wl);if(o!==Wl)return o}return CJ(e,A,i,n)}function BJ(t,e,A,i,n){let o=z2A(A);if(typeof o=="function"){if(!AJ(e,t,i))return i&Gi.Host?IJ(n,A,i):CJ(e,A,i,n);try{let r;if(r=o(i),r==null&&!(i&Gi.Optional))Y9(A);else return r}finally{nJ()}}else if(typeof o=="number"){let r=null,s=gJ(t,e),a=Ld,c=i&Gi.Host?e[Oa][Ia]:null;for((s===-1||i&Gi.SkipSelf)&&(a=s===-1?sb(t,e):e[s+8],a===Ld||!UK(i,!1)?s=-1:(r=e[ki],s=Hm(a),e=zm(a,e)));s!==-1;){let l=e[ki];if(GK(o,s,l.data)){let I=H2A(s,e,A,r,i,c);if(I!==Wl)return I}a=e[s+8],a!==Ld&&UK(i,e[ki].data[s+8]===c)&&GK(o,s,e)?(r=l,s=Hm(a),e=zm(a,e)):s=-1}}return n}function H2A(t,e,A,i,n,o){let r=e[ki],s=r.data[t+8],a=i==null?zd(s)&&Jv:i!=r&&(s.type&3)!==0,c=n&Gi.Host&&o===s,l=xm(s,r,A,a,c);return l!==null?dh(e,r,l,s,n):Wl}function xm(t,e,A,i,n){let o=t.providerIndexes,r=e.data,s=o&1048575,a=t.directiveStart,c=t.directiveEnd,l=o>>20,I=i?s:s+l,C=n?s+l:c;for(let d=I;d<C;d++){let B=r[d];if(d<a&&A===B||d>=a&&B.type===A)return d}if(n){let d=r[a];if(d&&rl(d)&&d.type===A)return a}return null}function dh(t,e,A,i,n){let o=t[A],r=e.data;if(o instanceof wI){let s=o;s.resolving&&MY(x0A(r[A]));let a=Om(s.canSeeViewProviders);s.resolving=!0;let c,l=s.injectImpl?oa(s.injectImpl):null,I=AJ(t,i,Gi.Default);try{o=t[A]=s.factory(void 0,n,r,t,i),e.firstCreatePass&&A>=i.directiveStart&&x2A(A,r[A],e)}finally{l!==null&&oa(l),Om(a),s.resolving=!1,nJ()}}return o}function z2A(t){if(typeof t=="string")return t.charCodeAt(0)||0;let e=t.hasOwnProperty(lh)?t[lh]:void 0;return typeof e=="number"?e>=0?e&cJ:O2A:e}function GK(t,e,A){let i=1<<t;return!!(A[e+(t>>lJ)]&i)}function UK(t,e){return!(t&Gi.Self)&&!(t&Gi.Host&&e)}var BI=class{_tNode;_lView;constructor(e,A){this._tNode=e,this._lView=A}get(e,A,i){return dJ(this._tNode,this._lView,e,gp(i),A)}};function O2A(){return new BI(os(),ei())}function Hi(t){return uh(()=>{let e=t.prototype.constructor,A=e[Fm]||Hv(e),i=Object.prototype,n=Object.getPrototypeOf(t.prototype).constructor;for(;n&&n!==i;){let o=n[Fm]||Hv(n);if(o&&o!==A)return o;n=Object.getPrototypeOf(n)}return o=>new o})}function Hv(t){return wY(t)?()=>{let e=Hv(ns(t));return e&&e()}:QI(t)}function P2A(t,e,A,i,n){let o=t,r=e;for(;o!==null&&r!==null&&r[gi]&2048&&!Jm(r);){let s=BJ(o,r,A,i|Gi.Self,Wl);if(s!==Wl)return s;let a=o.parent;if(!a){let c=r[JY];if(c){let l=c.get(A,Wl,i);if(l!==Wl)return l}a=EJ(r),r=r[Hd]}o=a}return n}function EJ(t){let e=t[ki],A=e.type;return A===2?e.declTNode:A===1?t[Ia]:null}function ab(t){return T2A(os(),t)}function KK(t,e=null,A=null,i){let n=QJ(t,e,A,i);return n.resolveInjectorInitializers(),n}function QJ(t,e=null,A=null,i,n=new Set){let o=[A||Ts,j0A(t)];return i=i||(typeof t=="object"?void 0:ra(t)),new Ih(o,e||dp(),i||null,n)}var Rt=class t{static THROW_IF_NOT_FOUND=dI;static NULL=new Um;static create(e,A){if(Array.isArray(e))return KK({name:""},A,e,"");{let i=e.name??"";return KK({name:i},e.parent,e.providers,i)}}static \u0275prov=NA({token:t,providedIn:"any",factory:()=>we(LY)});static __NG_ELEMENT_ID__=-1};var wr=class{attributeName;constructor(e){this.attributeName=e}__NG_ELEMENT_ID__=()=>ab(this.attributeName);toString(){return`HostAttributeToken ${this.attributeName}`}},j2A=new dA("");j2A.__NG_ELEMENT_ID__=t=>{let e=os();if(e===null)throw new Ae(204,!1);if(e.type&2)return e.value;if(t&Gi.Optional)return null;throw new Ae(204,!1)};var hJ=!1,m2=(()=>{class t{static __NG_ELEMENT_ID__=q2A;static __NG_ENV_ID__=A=>A}return t})(),jm=class extends m2{_lView;constructor(e){super(),this._lView=e}onDestroy(e){let A=this._lView;return Od(A)?(e(),()=>{}):(ZY(A,e),()=>d2A(A,e))}};function q2A(){return new jm(ei())}var DI=class{},cb=new dA("",{providedIn:"root",factory:()=>!1});var uJ=new dA(""),fJ=new dA(""),B0=(()=>{class t{taskId=0;pendingTasks=new Set;get _hasPendingTasks(){return this.hasPendingTasks.value}hasPendingTasks=new Mi(!1);add(){this._hasPendingTasks||this.hasPendingTasks.next(!0);let A=this.taskId++;return this.pendingTasks.add(A),A}has(A){return this.pendingTasks.has(A)}remove(A){this.pendingTasks.delete(A),this.pendingTasks.size===0&&this._hasPendingTasks&&this.hasPendingTasks.next(!1)}ngOnDestroy(){this.pendingTasks.clear(),this._hasPendingTasks&&this.hasPendingTasks.next(!1)}static \u0275prov=NA({token:t,providedIn:"root",factory:()=>new t})}return t})();var zv=class extends OA{__isAsync;destroyRef=void 0;pendingTasks=void 0;constructor(e=!1){super(),this.__isAsync=e,H9()&&(this.destroyRef=f(m2,{optional:!0})??void 0,this.pendingTasks=f(B0,{optional:!0})??void 0)}emit(e){let A=Ti(null);try{super.next(e)}finally{Ti(A)}}subscribe(e,A,i){let n=e,o=A||(()=>null),r=i;if(e&&typeof e=="object"){let a=e;n=a.next?.bind(a),o=a.error?.bind(a),r=a.complete?.bind(a)}this.__isAsync&&(o=this.wrapInTimeout(o),n&&(n=this.wrapInTimeout(n)),r&&(r=this.wrapInTimeout(r)));let s=super.subscribe({next:n,error:o,complete:r});return e instanceof Kt&&e.add(s),s}wrapInTimeout(e){return A=>{let i=this.pendingTasks?.add();setTimeout(()=>{try{e(A)}finally{i!==void 0&&this.pendingTasks?.remove(i)}})}}},WA=zv;function Bh(...t){}function mJ(t){let e,A;function i(){t=Bh;try{A!==void 0&&typeof cancelAnimationFrame=="function"&&cancelAnimationFrame(A),e!==void 0&&clearTimeout(e)}catch{}}return e=setTimeout(()=>{t(),i()}),typeof requestAnimationFrame=="function"&&(A=requestAnimationFrame(()=>{t(),i()})),()=>i()}function YK(t){return queueMicrotask(()=>t()),()=>{t=Bh}}var lb="isAngularZone",qm=lb+"_ID",V2A=0,Qe=class t{hasPendingMacrotasks=!1;hasPendingMicrotasks=!1;isStable=!0;onUnstable=new WA(!1);onMicrotaskEmpty=new WA(!1);onStable=new WA(!1);onError=new WA(!1);constructor(e){let{enableLongStackTrace:A=!1,shouldCoalesceEventChangeDetection:i=!1,shouldCoalesceRunChangeDetection:n=!1,scheduleInRootZone:o=hJ}=e;if(typeof Zone>"u")throw new Ae(908,!1);Zone.assertZonePatched();let r=this;r._nesting=0,r._outer=r._inner=Zone.current,Zone.TaskTrackingZoneSpec&&(r._inner=r._inner.fork(new Zone.TaskTrackingZoneSpec)),A&&Zone.longStackTraceZoneSpec&&(r._inner=r._inner.fork(Zone.longStackTraceZoneSpec)),r.shouldCoalesceEventChangeDetection=!n&&i,r.shouldCoalesceRunChangeDetection=n,r.callbackScheduled=!1,r.scheduleInRootZone=o,X2A(r)}static isInAngularZone(){return typeof Zone<"u"&&Zone.current.get(lb)===!0}static assertInAngularZone(){if(!t.isInAngularZone())throw new Ae(909,!1)}static assertNotInAngularZone(){if(t.isInAngularZone())throw new Ae(909,!1)}run(e,A,i){return this._inner.run(e,A,i)}runTask(e,A,i,n){let o=this._inner,r=o.scheduleEventTask("NgZoneEvent: "+n,e,Z2A,Bh,Bh);try{return o.runTask(r,A,i)}finally{o.cancelTask(r)}}runGuarded(e,A,i){return this._inner.runGuarded(e,A,i)}runOutsideAngular(e){return this._outer.run(e)}},Z2A={};function gb(t){if(t._nesting==0&&!t.hasPendingMicrotasks&&!t.isStable)try{t._nesting++,t.onMicrotaskEmpty.emit(null)}finally{if(t._nesting--,!t.hasPendingMicrotasks)try{t.runOutsideAngular(()=>t.onStable.emit(null))}finally{t.isStable=!0}}}function W2A(t){if(t.isCheckStableRunning||t.callbackScheduled)return;t.callbackScheduled=!0;function e(){mJ(()=>{t.callbackScheduled=!1,Ov(t),t.isCheckStableRunning=!0,gb(t),t.isCheckStableRunning=!1})}t.scheduleInRootZone?Zone.root.run(()=>{e()}):t._outer.run(()=>{e()}),Ov(t)}function X2A(t){let e=()=>{W2A(t)},A=V2A++;t._inner=t._inner.fork({name:"angular",properties:{[lb]:!0,[qm]:A,[qm+A]:!0},onInvokeTask:(i,n,o,r,s,a)=>{if($2A(a))return i.invokeTask(o,r,s,a);try{return JK(t),i.invokeTask(o,r,s,a)}finally{(t.shouldCoalesceEventChangeDetection&&r.type==="eventTask"||t.shouldCoalesceRunChangeDetection)&&e(),TK(t)}},onInvoke:(i,n,o,r,s,a,c)=>{try{return JK(t),i.invoke(o,r,s,a,c)}finally{t.shouldCoalesceRunChangeDetection&&!t.callbackScheduled&&!A1A(a)&&e(),TK(t)}},onHasTask:(i,n,o,r)=>{i.hasTask(o,r),n===o&&(r.change=="microTask"?(t._hasPendingMicrotasks=r.microTask,Ov(t),gb(t)):r.change=="macroTask"&&(t.hasPendingMacrotasks=r.macroTask))},onHandleError:(i,n,o,r)=>(i.handleError(o,r),t.runOutsideAngular(()=>t.onError.emit(r)),!1)})}function Ov(t){t._hasPendingMicrotasks||(t.shouldCoalesceEventChangeDetection||t.shouldCoalesceRunChangeDetection)&&t.callbackScheduled===!0?t.hasPendingMicrotasks=!0:t.hasPendingMicrotasks=!1}function JK(t){t._nesting++,t.isStable&&(t.isStable=!1,t.onUnstable.emit(null))}function TK(t){t._nesting--,gb(t)}var Vm=class{hasPendingMicrotasks=!1;hasPendingMacrotasks=!1;isStable=!0;onUnstable=new WA;onMicrotaskEmpty=new WA;onStable=new WA;onError=new WA;run(e,A,i){return e.apply(A,i)}runGuarded(e,A,i){return e.apply(A,i)}runOutsideAngular(e){return e()}runTask(e,A,i,n){return e.apply(A,i)}};function $2A(t){return pJ(t,"__ignore_ng_zone__")}function A1A(t){return pJ(t,"__scheduler_tick__")}function pJ(t,e){return!Array.isArray(t)||t.length!==1?!1:t[0]?.data?.[e]===!0}function e1A(t="zone.js",e){return t==="noop"?new Vm:t==="zone.js"?new Qe(e):t}var aa=class{_console=console;handleError(e){this._console.error("ERROR",e)}},t1A=new dA("",{providedIn:"root",factory:()=>{let t=f(Qe),e=f(aa);return A=>t.runOutsideAngular(()=>e.handleError(A))}});function HK(t,e){return mY(t,e)}function i1A(t){return mY(fY,t)}var wJ=(HK.required=i1A,HK);function n1A(){return jd(os(),ei())}function jd(t,e){return new ee(ng(t,e))}var ee=(()=>{class t{nativeElement;constructor(A){this.nativeElement=A}static __NG_ELEMENT_ID__=n1A}return t})();function DJ(t){return t instanceof ee?t.nativeElement:t}function p2(t){return typeof t=="function"&&t[ia]!==void 0}function Jo(t,e){let A=X7(t,e?.equal),i=A[ia];return A.set=n=>XQ(i,n),A.update=n=>$7(i,n),A.asReadonly=o1A.bind(A),A}function o1A(){let t=this[ia];if(t.readonlyFn===void 0){let e=()=>this();e[ia]=t,t.readonlyFn=e}return t.readonlyFn}function yJ(t){return p2(t)&&typeof t.set=="function"}function r1A(){return this._results[Symbol.iterator]()}var ca=class{_emitDistinctChangesOnly;dirty=!0;_onDirty=void 0;_results=[];_changesDetected=!1;_changes=void 0;length=0;first=void 0;last=void 0;get changes(){return this._changes??=new OA}constructor(e=!1){this._emitDistinctChangesOnly=e}get(e){return this._results[e]}map(e){return this._results.map(e)}filter(e){return this._results.filter(e)}find(e){return this._results.find(e)}reduce(e,A){return this._results.reduce(e,A)}forEach(e){this._results.forEach(e)}some(e){return this._results.some(e)}toArray(){return this._results.slice()}toString(){return this._results.toString()}reset(e,A){this.dirty=!1;let i=T0A(e);(this._changesDetected=!J0A(this._results,i,A))&&(this._results=i,this.length=i.length,this.last=i[this.length-1],this.first=i[0])}notifyOnChanges(){this._changes!==void 0&&(this._changesDetected||!this._emitDistinctChangesOnly)&&this._changes.next(this)}onDirty(e){this._onDirty=e}setDirty(){this.dirty=!0,this._onDirty?.()}destroy(){this._changes!==void 0&&(this._changes.complete(),this._changes.unsubscribe())}[Symbol.iterator]=r1A};function vJ(t){return(t.flags&128)===128}var bJ=function(t){return t[t.OnPush=0]="OnPush",t[t.Default=1]="Default",t}(bJ||{}),MJ=new Map,s1A=0;function a1A(){return s1A++}function c1A(t){MJ.set(t[Bp],t)}function Pv(t){MJ.delete(t[Bp])}var zK="__ngContext__";function qd(t,e){Q2(e)?(t[zK]=e[Bp],c1A(e)):t[zK]=e}function kJ(t){return RJ(t[Ch])}function SJ(t){return RJ(t[ol])}function RJ(t){for(;t!==null&&!C0(t);)t=t[ol];return t}var jv;function xJ(t){jv=t}function LJ(){if(jv!==void 0)return jv;if(typeof document<"u")return document;throw new Ae(210,!1)}var Vd=new dA("",{providedIn:"root",factory:()=>l1A}),l1A="ng",Ib=new dA(""),og=new dA("",{providedIn:"platform",factory:()=>"unknown"});var Si=new dA(""),yh=new dA("",{providedIn:"root",factory:()=>LJ().body?.querySelector("[ngCspNonce]")?.getAttribute("ngCspNonce")||null});var g1A="h",I1A="b";var FJ=!1,C1A=new dA("",{providedIn:"root",factory:()=>FJ});var Cb=function(t){return t[t.CHANGE_DETECTION=0]="CHANGE_DETECTION",t[t.AFTER_NEXT_RENDER=1]="AFTER_NEXT_RENDER",t}(Cb||{}),Zd=new dA(""),OK=new Set;function E0(t){OK.has(t)||(OK.add(t),performance?.mark?.("mark_feature_usage",{detail:{feature:t}}))}var db=(()=>{class t{view;node;constructor(A,i){this.view=A,this.node=i}static __NG_ELEMENT_ID__=d1A}return t})();function d1A(){return new db(ei(),os())}var Sd=function(t){return t[t.EarlyRead=0]="EarlyRead",t[t.Write=1]="Write",t[t.MixedReadWrite=2]="MixedReadWrite",t[t.Read=3]="Read",t}(Sd||{}),NJ=(()=>{class t{impl=null;execute(){this.impl?.execute()}static \u0275prov=NA({token:t,providedIn:"root",factory:()=>new t})}return t})(),B1A=[Sd.EarlyRead,Sd.Write,Sd.MixedReadWrite,Sd.Read],E1A=(()=>{class t{ngZone=f(Qe);scheduler=f(DI);errorHandler=f(aa,{optional:!0});sequences=new Set;deferredRegistrations=new Set;executing=!1;constructor(){f(Zd,{optional:!0})}execute(){let A=this.sequences.size>0;A&&Ro(16),this.executing=!0;for(let i of B1A)for(let n of this.sequences)if(!(n.erroredOrDestroyed||!n.hooks[i]))try{n.pipelinedValue=this.ngZone.runOutsideAngular(()=>this.maybeTrace(()=>{let o=n.hooks[i];return o(n.pipelinedValue)},n.snapshot))}catch(o){n.erroredOrDestroyed=!0,this.errorHandler?.handleError(o)}this.executing=!1;for(let i of this.sequences)i.afterRun(),i.once&&(this.sequences.delete(i),i.destroy());for(let i of this.deferredRegistrations)this.sequences.add(i);this.deferredRegistrations.size>0&&this.scheduler.notify(7),this.deferredRegistrations.clear(),A&&Ro(17)}register(A){let{view:i}=A;i!==void 0?((i[xd]??=[]).push(A),Pd(i),i[gi]|=8192):this.executing?this.deferredRegistrations.add(A):this.addSequence(A)}addSequence(A){this.sequences.add(A),this.scheduler.notify(7)}unregister(A){this.executing&&this.sequences.has(A)?(A.erroredOrDestroyed=!0,A.pipelinedValue=void 0,A.once=!0):(this.sequences.delete(A),this.deferredRegistrations.delete(A))}maybeTrace(A,i){return i?i.run(Cb.AFTER_NEXT_RENDER,A):A()}static \u0275prov=NA({token:t,providedIn:"root",factory:()=>new t})}return t})(),qv=class{impl;hooks;view;once;snapshot;erroredOrDestroyed=!1;pipelinedValue=void 0;unregisterOnDestroy;constructor(e,A,i,n,o,r=null){this.impl=e,this.hooks=A,this.view=i,this.once=n,this.snapshot=r,this.unregisterOnDestroy=o?.onDestroy(()=>this.destroy())}afterRun(){this.erroredOrDestroyed=!1,this.pipelinedValue=void 0,this.snapshot?.dispose(),this.snapshot=null}destroy(){this.impl.unregister(this),this.unregisterOnDestroy?.();let e=this.view?.[xd];e&&(this.view[xd]=e.filter(A=>A!==this))}};function vh(t,e){!e?.injector&&z9(vh);let A=e?.injector??f(Rt);return E0("NgAfterRender"),_J(t,A,e,!1)}function To(t,e){!e?.injector&&z9(To);let A=e?.injector??f(Rt);return E0("NgAfterNextRender"),_J(t,A,e,!0)}function Q1A(t,e){if(t instanceof Function){let A=[void 0,void 0,void 0,void 0];return A[e]=t,A}else return[t.earlyRead,t.write,t.mixedReadWrite,t.read]}function _J(t,e,A,i){let n=e.get(NJ);n.impl??=e.get(E1A);let o=e.get(Zd,null,{optional:!0}),r=A?.phase??Sd.MixedReadWrite,s=A?.manualCleanup!==!0?e.get(m2):null,a=e.get(db,null,{optional:!0}),c=new qv(n.impl,Q1A(t,r),a?.view,i,s,o?.snapshot(null));return n.impl.register(c),c}var h1A=(t,e,A,i)=>{};function u1A(t,e,A,i){h1A(t,e,A,i)}var f1A=()=>null;function GJ(t,e,A=!1){return f1A(t,e,A)}function UJ(t,e){let A=t.contentQueries;if(A!==null){let i=Ti(null);try{for(let n=0;n<A.length;n+=2){let o=A[n],r=A[n+1];if(r!==-1){let s=t.data[r];hp(o),s.contentQueries(2,e[r],r)}}}finally{Ti(i)}}}function Vv(t,e,A){hp(0);let i=Ti(null);try{e(t,A)}finally{Ti(i)}}function Bb(t,e,A){if(P9(e)){let i=Ti(null);try{let n=e.directiveStart,o=e.directiveEnd;for(let r=n;r<o;r++){let s=t.data[r];if(s.contentQueries){let a=A[r];s.contentQueries(1,a,r)}}}finally{Ti(i)}}}var eg=function(t){return t[t.Emulated=0]="Emulated",t[t.None=2]="None",t[t.ShadowDom=3]="ShadowDom",t}(eg||{}),Dm;function m1A(){if(Dm===void 0&&(Dm=null,sa.trustedTypes))try{Dm=sa.trustedTypes.createPolicy("angular",{createHTML:t=>t,createScript:t=>t,createScriptURL:t=>t})}catch{}return Dm}function mp(t){return m1A()?.createHTML(t)||t}var ym;function p1A(){if(ym===void 0&&(ym=null,sa.trustedTypes))try{ym=sa.trustedTypes.createPolicy("angular#unsafe-bypass",{createHTML:t=>t,createScript:t=>t,createScriptURL:t=>t})}catch{}return ym}function PK(t){return p1A()?.createHTML(t)||t}var g0=class{changingThisBreaksApplicationSecurity;constructor(e){this.changingThisBreaksApplicationSecurity=e}toString(){return`SafeValue must use [property]=binding: ${this.changingThisBreaksApplicationSecurity} (see ${uY})`}},Zv=class extends g0{getTypeName(){return"HTML"}},Wv=class extends g0{getTypeName(){return"Style"}},Xv=class extends g0{getTypeName(){return"Script"}},$v=class extends g0{getTypeName(){return"URL"}},A9=class extends g0{getTypeName(){return"ResourceURL"}};function sl(t){return t instanceof g0?t.changingThisBreaksApplicationSecurity:t}function w2(t,e){let A=w1A(t);if(A!=null&&A!==e){if(A==="ResourceURL"&&e==="URL")return!0;throw new Error(`Required a safe ${e}, got a ${A} (see ${uY})`)}return A===e}function w1A(t){return t instanceof g0&&t.getTypeName()||null}function KJ(t){return new Zv(t)}function YJ(t){return new Wv(t)}function JJ(t){return new Xv(t)}function TJ(t){return new $v(t)}function HJ(t){return new A9(t)}function D1A(t){let e=new t9(t);return y1A()?new e9(e):e}var e9=class{inertDocumentHelper;constructor(e){this.inertDocumentHelper=e}getInertBodyElement(e){e="<body><remove></remove>"+e;try{let A=new window.DOMParser().parseFromString(mp(e),"text/html").body;return A===null?this.inertDocumentHelper.getInertBodyElement(e):(A.firstChild?.remove(),A)}catch{return null}}},t9=class{defaultDoc;inertDocument;constructor(e){this.defaultDoc=e,this.inertDocument=this.defaultDoc.implementation.createHTMLDocument("sanitization-inert")}getInertBodyElement(e){let A=this.inertDocument.createElement("template");return A.innerHTML=mp(e),A}};function y1A(){try{return!!new window.DOMParser().parseFromString(mp(""),"text/html")}catch{return!1}}var v1A=/^(?!javascript:)(?:[a-z0-9+.-]+:|[^&:\/?#]*(?:[\/?#]|$))/i;function pp(t){return t=String(t),t.match(v1A)?t:"unsafe:"+t}function Q0(t){let e={};for(let A of t.split(","))e[A]=!0;return e}function bh(...t){let e={};for(let A of t)for(let i in A)A.hasOwnProperty(i)&&(e[i]=!0);return e}var zJ=Q0("area,br,col,hr,img,wbr"),OJ=Q0("colgroup,dd,dt,li,p,tbody,td,tfoot,th,thead,tr"),PJ=Q0("rp,rt"),b1A=bh(PJ,OJ),M1A=bh(OJ,Q0("address,article,aside,blockquote,caption,center,del,details,dialog,dir,div,dl,figure,figcaption,footer,h1,h2,h3,h4,h5,h6,header,hgroup,hr,ins,main,map,menu,nav,ol,pre,section,summary,table,ul")),k1A=bh(PJ,Q0("a,abbr,acronym,audio,b,bdi,bdo,big,br,cite,code,del,dfn,em,font,i,img,ins,kbd,label,map,mark,picture,q,ruby,rp,rt,s,samp,small,source,span,strike,strong,sub,sup,time,track,tt,u,var,video")),jK=bh(zJ,M1A,k1A,b1A),jJ=Q0("background,cite,href,itemtype,longdesc,poster,src,xlink:href"),S1A=Q0("abbr,accesskey,align,alt,autoplay,axis,bgcolor,border,cellpadding,cellspacing,class,clear,color,cols,colspan,compact,controls,coords,datetime,default,dir,download,face,headers,height,hidden,hreflang,hspace,ismap,itemscope,itemprop,kind,label,lang,language,loop,media,muted,nohref,nowrap,open,preload,rel,rev,role,rows,rowspan,rules,scope,scrolling,shape,size,sizes,span,srclang,srcset,start,summary,tabindex,target,title,translate,type,usemap,valign,value,vspace,width"),R1A=Q0("aria-activedescendant,aria-atomic,aria-autocomplete,aria-busy,aria-checked,aria-colcount,aria-colindex,aria-colspan,aria-controls,aria-current,aria-describedby,aria-details,aria-disabled,aria-dropeffect,aria-errormessage,aria-expanded,aria-flowto,aria-grabbed,aria-haspopup,aria-hidden,aria-invalid,aria-keyshortcuts,aria-label,aria-labelledby,aria-level,aria-live,aria-modal,aria-multiline,aria-multiselectable,aria-orientation,aria-owns,aria-placeholder,aria-posinset,aria-pressed,aria-readonly,aria-relevant,aria-required,aria-roledescription,aria-rowcount,aria-rowindex,aria-rowspan,aria-selected,aria-setsize,aria-sort,aria-valuemax,aria-valuemin,aria-valuenow,aria-valuetext"),x1A=bh(jJ,S1A,R1A),L1A=Q0("script,style,template"),i9=class{sanitizedSomething=!1;buf=[];sanitizeChildren(e){let A=e.firstChild,i=!0,n=[];for(;A;){if(A.nodeType===Node.ELEMENT_NODE?i=this.startElement(A):A.nodeType===Node.TEXT_NODE?this.chars(A.nodeValue):this.sanitizedSomething=!0,i&&A.firstChild){n.push(A),A=_1A(A);continue}for(;A;){A.nodeType===Node.ELEMENT_NODE&&this.endElement(A);let o=N1A(A);if(o){A=o;break}A=n.pop()}}return this.buf.join("")}startElement(e){let A=qK(e).toLowerCase();if(!jK.hasOwnProperty(A))return this.sanitizedSomething=!0,!L1A.hasOwnProperty(A);this.buf.push("<"),this.buf.push(A);let i=e.attributes;for(let n=0;n<i.length;n++){let o=i.item(n),r=o.name,s=r.toLowerCase();if(!x1A.hasOwnProperty(s)){this.sanitizedSomething=!0;continue}let a=o.value;jJ[s]&&(a=pp(a)),this.buf.push(" ",r,'="',VK(a),'"')}return this.buf.push(">"),!0}endElement(e){let A=qK(e).toLowerCase();jK.hasOwnProperty(A)&&!zJ.hasOwnProperty(A)&&(this.buf.push("</"),this.buf.push(A),this.buf.push(">"))}chars(e){this.buf.push(VK(e))}};function F1A(t,e){return(t.compareDocumentPosition(e)&Node.DOCUMENT_POSITION_CONTAINED_BY)!==Node.DOCUMENT_POSITION_CONTAINED_BY}function N1A(t){let e=t.nextSibling;if(e&&t!==e.previousSibling)throw qJ(e);return e}function _1A(t){let e=t.firstChild;if(e&&F1A(t,e))throw qJ(e);return e}function qK(t){let e=t.nodeName;return typeof e=="string"?e:"FORM"}function qJ(t){return new Error(`Failed to sanitize html because the element is clobbered: ${t.outerHTML}`)}var G1A=/[\uD800-\uDBFF][\uDC00-\uDFFF]/g,U1A=/([^\#-~ |!])/g;function VK(t){return t.replace(/&/g,"&amp;").replace(G1A,function(e){let A=e.charCodeAt(0),i=e.charCodeAt(1);return"&#"+((A-55296)*1024+(i-56320)+65536)+";"}).replace(U1A,function(e){return"&#"+e.charCodeAt(0)+";"}).replace(/</g,"&lt;").replace(/>/g,"&gt;")}var vm;function Eb(t,e){let A=null;try{vm=vm||D1A(t);let i=e?String(e):"";A=vm.getInertBodyElement(i);let n=5,o=i;do{if(n===0)throw new Error("Failed to sanitize html because the input is unstable");n--,i=o,o=A.innerHTML,A=vm.getInertBodyElement(i)}while(i!==o);let s=new i9().sanitizeChildren(ZK(A)||A);return mp(s)}finally{if(A){let i=ZK(A)||A;for(;i.firstChild;)i.firstChild.remove()}}}function ZK(t){return"content"in t&&K1A(t)?t.content:null}function K1A(t){return t.nodeType===Node.ELEMENT_NODE&&t.nodeName==="TEMPLATE"}var jr=function(t){return t[t.NONE=0]="NONE",t[t.HTML=1]="HTML",t[t.STYLE=2]="STYLE",t[t.SCRIPT=3]="SCRIPT",t[t.URL=4]="URL",t[t.RESOURCE_URL=5]="RESOURCE_URL",t}(jr||{});function RI(t){let e=VJ();return e?PK(e.sanitize(jr.HTML,t)||""):w2(t,"HTML")?PK(sl(t)):Eb(LJ(),EI(t))}function ja(t){let e=VJ();return e?e.sanitize(jr.URL,t)||"":w2(t,"URL")?sl(t):pp(EI(t))}function VJ(){let t=ei();return t&&t[c0].sanitizer}var Y1A=/^>|^->|<!--|-->|--!>|<!-$/g,J1A=/(<|>)/g,T1A="\u200B$1\u200B";function H1A(t){return t.replace(Y1A,e=>e.replace(J1A,T1A))}function wp(t){return t.ownerDocument.defaultView}function Wd(t){return t.ownerDocument}function ZJ(t){return t instanceof Function?t():t}function z1A(t,e,A){let i=t.length;for(;;){let n=t.indexOf(e,A);if(n===-1)return n;if(n===0||t.charCodeAt(n-1)<=32){let o=e.length;if(n+o===i||t.charCodeAt(n+o)<=32)return n}A=n+1}}var WJ="ng-template";function O1A(t,e,A,i){let n=0;if(i){for(;n<e.length&&typeof e[n]=="string";n+=2)if(e[n]==="class"&&z1A(e[n+1].toLowerCase(),A,0)!==-1)return!0}else if(Qb(t))return!1;if(n=e.indexOf(1,n),n>-1){let o;for(;++n<e.length&&typeof(o=e[n])=="string";)if(o.toLowerCase()===A)return!0}return!1}function Qb(t){return t.type===4&&t.value!==WJ}function P1A(t,e,A){let i=t.type===4&&!A?WJ:t.value;return e===i}function j1A(t,e,A){let i=4,n=t.attrs,o=n!==null?Z1A(n):0,r=!1;for(let s=0;s<e.length;s++){let a=e[s];if(typeof a=="number"){if(!r&&!il(i)&&!il(a))return!1;if(r&&il(a))continue;r=!1,i=a|i&1;continue}if(!r)if(i&4){if(i=2|i&1,a!==""&&!P1A(t,a,A)||a===""&&e.length===1){if(il(i))return!1;r=!0}}else if(i&8){if(n===null||!O1A(t,n,a,A)){if(il(i))return!1;r=!0}}else{let c=e[++s],l=q1A(a,n,Qb(t),A);if(l===-1){if(il(i))return!1;r=!0;continue}if(c!==""){let I;if(l>o?I="":I=n[l+1].toLowerCase(),i&2&&c!==I){if(il(i))return!1;r=!0}}}}return il(i)||r}function il(t){return(t&1)===0}function q1A(t,e,A,i){if(e===null)return-1;let n=0;if(i||!A){let o=!1;for(;n<e.length;){let r=e[n];if(r===t)return n;if(r===3||r===6)o=!0;else if(r===1||r===2){let s=e[++n];for(;typeof s=="string";)s=e[++n];continue}else{if(r===4)break;if(r===0){n+=4;continue}}n+=o?1:2}return-1}else return W1A(e,t)}function XJ(t,e,A=!1){for(let i=0;i<e.length;i++)if(j1A(t,e[i],A))return!0;return!1}function V1A(t){let e=t.attrs;if(e!=null){let A=e.indexOf(5);if((A&1)===0)return e[A+1]}return null}function Z1A(t){for(let e=0;e<t.length;e++){let A=t[e];if(sJ(A))return e}return t.length}function W1A(t,e){let A=t.indexOf(4);if(A>-1)for(A++;A<t.length;){let i=t[A];if(typeof i=="number")return-1;if(i===e)return A;A++}return-1}function X1A(t,e){A:for(let A=0;A<e.length;A++){let i=e[A];if(t.length===i.length){for(let n=0;n<t.length;n++)if(t[n]!==i[n])continue A;return!0}}return!1}function WK(t,e){return t?":not("+e.trim()+")":e}function $1A(t){let e=t[0],A=1,i=2,n="",o=!1;for(;A<t.length;){let r=t[A];if(typeof r=="string")if(i&2){let s=t[++A];n+="["+r+(s.length>0?'="'+s+'"':"")+"]"}else i&8?n+="."+r:i&4&&(n+=" "+r);else n!==""&&!il(r)&&(e+=WK(o,n),n=""),i=r,o=o||!il(i);A++}return n!==""&&(e+=WK(o,n)),e}function AIA(t){return t.map($1A).join(",")}function eIA(t){let e=[],A=[],i=1,n=2;for(;i<t.length;){let o=t[i];if(typeof o=="string")n===2?o!==""&&e.push(o,t[++i]):n===8&&A.push(o);else{if(!il(n))break;n=o}i++}return A.length&&e.push(1,...A),e}var qa={};function tIA(t,e){return t.createText(e)}function iIA(t,e,A){t.setValue(e,A)}function nIA(t,e){return t.createComment(H1A(e))}function $J(t,e,A){return t.createElement(e,A)}function Zm(t,e,A,i,n){t.insertBefore(e,A,i,n)}function AT(t,e,A){t.appendChild(e,A)}function XK(t,e,A,i,n){i!==null?Zm(t,e,A,i,n):AT(t,e,A)}function oIA(t,e,A){t.removeChild(null,e,A)}function rIA(t,e,A){t.setAttribute(e,"style",A)}function sIA(t,e,A){A===""?t.removeAttribute(e,"class"):t.setAttribute(e,"class",A)}function eT(t,e,A){let{mergedAttrs:i,classes:n,styles:o}=A;i!==null&&_2A(t,e,i),n!==null&&sIA(t,e,n),o!==null&&rIA(t,e,o)}function hb(t,e,A,i,n,o,r,s,a,c,l){let I=Nr+i,C=I+n,d=aIA(I,C),B=typeof c=="function"?c():c;return d[ki]={type:t,blueprint:d,template:A,queries:null,viewQuery:s,declTNode:e,data:d.slice().fill(null,I),bindingStartIndex:I,expandoStartIndex:C,hostBindingOpCodes:null,firstCreatePass:!0,firstUpdatePass:!0,staticViewQueries:!1,staticContentQueries:!1,preOrderHooks:null,preOrderCheckHooks:null,contentHooks:null,contentCheckHooks:null,viewHooks:null,viewCheckHooks:null,destroyHooks:null,cleanup:null,contentQueries:null,components:null,directiveRegistry:typeof o=="function"?o():o,pipeRegistry:typeof r=="function"?r():r,firstChild:null,schemas:a,consts:B,incompleteFirstPass:!1,ssrId:l}}function aIA(t,e){let A=[];for(let i=0;i<e;i++)A.push(i<t?null:qa);return A}function cIA(t){let e=t.tView;return e===null||e.incompleteFirstPass?t.tView=hb(1,null,t.template,t.decls,t.vars,t.directiveDefs,t.pipeDefs,t.viewQuery,t.schemas,t.consts,t.id):e}function ub(t,e,A,i,n,o,r,s,a,c,l){let I=e.blueprint.slice();return I[ig]=n,I[gi]=i|4|128|8|64|1024,(c!==null||t&&t[gi]&2048)&&(I[gi]|=2048),qY(I),I[ps]=I[Hd]=t,I[Fr]=A,I[c0]=r||t&&t[c0],I[Wo]=s||t&&t[Wo],I[Gd]=a||t&&t[Gd]||null,I[Ia]=o,I[Bp]=a1A(),I[_d]=l,I[JY]=c,I[Oa]=e.type==2?t[Oa]:I,I}function lIA(t,e,A){let i=ng(e,t),n=cIA(A),o=t[c0].rendererFactory,r=fb(t,ub(t,n,null,tT(A),i,e,null,o.createRenderer(i,A),null,null,null));return t[e.index]=r}function tT(t){let e=16;return t.signals?e=4096:t.onPush&&(e=64),e}function iT(t,e,A,i){if(A===0)return-1;let n=e.length;for(let o=0;o<A;o++)e.push(i),t.blueprint.push(i),t.data.push(null);return n}function fb(t,e){return t[Ch]?t[LK][ol]=e:t[Ch]=e,t[LK]=e,e}function G(t=1){nT(Yo(),ei(),d0()+t,!1)}function nT(t,e,A,i){if(!i)if((e[gi]&3)===3){let o=t.preOrderCheckHooks;o!==null&&Sm(e,o,A)}else{let o=t.preOrderHooks;o!==null&&Rm(e,o,0,A)}pI(A)}var Dp=function(t){return t[t.None=0]="None",t[t.SignalBased=1]="SignalBased",t[t.HasDecoratorInputTransform=2]="HasDecoratorInputTransform",t}(Dp||{});function n9(t,e,A,i){let n=Ti(null);try{let[o,r,s]=t.inputs[A],a=null;(r&Dp.SignalBased)!==0&&(a=e[o][ia]),a!==null&&a.transformFn!==void 0?i=a.transformFn(i):s!==null&&(i=s.call(e,i)),t.setInput!==null?t.setInput(e,a,i,A,o):TY(e,a,o,i)}finally{Ti(n)}}function oT(t,e,A,i,n){let o=d0(),r=i&2;try{pI(-1),r&&e.length>Nr&&nT(t,e,Nr,!1),Ro(r?2:0,n),A(i,n)}finally{pI(o),Ro(r?3:1,n)}}function yp(t,e,A){EIA(t,e,A),(A.flags&64)===64&&QIA(t,e,A)}function mb(t,e,A=ng){let i=e.localNames;if(i!==null){let n=e.index+1;for(let o=0;o<i.length;o+=2){let r=i[o+1],s=r===-1?A(e,t):t[r];t[n++]=s}}}function gIA(t,e,A,i){let o=i.get(C1A,FJ)||A===eg.ShadowDom,r=t.selectRootElement(e,o);return IIA(r),r}function IIA(t){CIA(t)}var CIA=()=>null;function dIA(t){return t==="class"?"className":t==="for"?"htmlFor":t==="formaction"?"formAction":t==="innerHtml"?"innerHTML":t==="readonly"?"readOnly":t==="tabindex"?"tabIndex":t}function vp(t,e,A,i,n,o,r,s){if(!s&&wb(e,t,A,i,n)){zd(e)&&BIA(A,e.index);return}if(e.type&3){let a=ng(e,A);i=dIA(i),n=r!=null?r(n,e.value||"",i):n,o.setProperty(a,i,n)}else e.type&12}function BIA(t,e){let A=Ag(e,t);A[gi]&16||(A[gi]|=64)}function EIA(t,e,A){let i=A.directiveStart,n=A.directiveEnd;zd(A)&&lIA(e,A,t.data[i+A.componentOffset]),t.firstCreatePass||Pm(A,e);let o=A.initialInputs;for(let r=i;r<n;r++){let s=t.data[r],a=dh(e,t,r,A);if(qd(a,e),o!==null&&mIA(e,r-i,a,s,A,o),rl(s)){let c=Ag(A.index,e);c[Fr]=dh(e,t,r,A)}}}function QIA(t,e,A){let i=A.directiveStart,n=A.directiveEnd,o=A.index,r=b2A();try{pI(o);for(let s=i;s<n;s++){let a=t.data[s],c=e[s];Yv(s),(a.hostBindings!==null||a.hostVars!==0||a.hostAttrs!==null)&&hIA(a,c)}}finally{pI(-1),Yv(r)}}function hIA(t,e){t.hostBindings!==null&&t.hostBindings(1,e)}function pb(t,e){let A=t.directiveRegistry,i=null;if(A)for(let n=0;n<A.length;n++){let o=A[n];XJ(e,o.selectors,!1)&&(i??=[],rl(o)?i.unshift(o):i.push(o))}return i}function uIA(t,e,A,i,n,o){let r=ng(t,e);fIA(e[Wo],r,o,t.value,A,i,n)}function fIA(t,e,A,i,n,o,r){if(o==null)t.removeAttribute(e,n,A);else{let s=r==null?EI(o):r(o,i||"",n);t.setAttribute(e,n,s,A)}}function mIA(t,e,A,i,n,o){let r=o[e];if(r!==null)for(let s=0;s<r.length;s+=2){let a=r[s],c=r[s+1];n9(i,A,a,c)}}function rT(t,e,A){return(t===null||rl(t))&&(A=c2A(A[e.index])),A[Wo]}function pIA(t,e){let A=t[Gd],i=A?A.get(aa,null):null;i&&i.handleError(e)}function wb(t,e,A,i,n){let o=t.inputs?.[i],r=t.hostDirectiveInputs?.[i],s=!1;if(r)for(let a=0;a<r.length;a+=2){let c=r[a],l=r[a+1],I=e.data[c];n9(I,A[c],l,n),s=!0}if(o)for(let a of o){let c=A[a],l=e.data[a];n9(l,c,i,n),s=!0}return s}function wIA(t,e){let A=Ag(e,t),i=A[ki];DIA(i,A);let n=A[ig];n!==null&&A[_d]===null&&(A[_d]=GJ(n,A[Gd])),Ro(18),Db(i,A,A[Fr]),Ro(19,A[Fr])}function DIA(t,e){for(let A=e.length;A<t.blueprint.length;A++)e.push(t.blueprint[A])}function Db(t,e,A){nb(e);try{let i=t.viewQuery;i!==null&&Vv(1,i,A);let n=t.template;n!==null&&oT(t,e,n,1,A),t.firstCreatePass&&(t.firstCreatePass=!1),e[l0]?.finishViewCreation(t),t.staticContentQueries&&UJ(t,e),t.staticViewQueries&&Vv(2,t.viewQuery,A);let o=t.components;o!==null&&yIA(e,o)}catch(i){throw t.firstCreatePass&&(t.incompleteFirstPass=!0,t.firstCreatePass=!1),i}finally{e[gi]&=-5,ob()}}function yIA(t,e){for(let A=0;A<e.length;A++)wIA(t,e[A])}function Mh(t,e,A,i){let n=Ti(null);try{let o=e.tView,s=t[gi]&4096?4096:16,a=ub(t,o,A,s,null,e,null,null,i?.injector??null,i?.embeddedViewInjector??null,i?.dehydratedView??null),c=t[e.index];a[hI]=c;let l=t[l0];return l!==null&&(a[l0]=l.createEmbeddedView(o)),Db(o,a,A),a}finally{Ti(n)}}function Yd(t,e){return!e||e.firstChild===null||vJ(t)}var vIA;function yb(t,e){return vIA(t,e)}var tg=function(t){return t[t.Important=1]="Important",t[t.DashCase=2]="DashCase",t}(tg||{});function vb(t){return(t.flags&32)===32}function Rd(t,e,A,i,n){if(i!=null){let o,r=!1;C0(i)?o=i:Q2(i)&&(r=!0,i=i[ig]);let s=$l(i);t===0&&A!==null?n==null?AT(e,A,s):Zm(e,A,s,n||null,!0):t===1&&A!==null?Zm(e,A,s,n||null,!0):t===2?oIA(e,s,r):t===3&&e.destroyNode(s),o!=null&&_IA(e,t,o,A,n)}}function bIA(t,e){sT(t,e),e[ig]=null,e[Ia]=null}function MIA(t,e,A,i,n,o){i[ig]=n,i[Ia]=e,kp(t,i,A,1,n,o)}function sT(t,e){e[c0].changeDetectionScheduler?.notify(9),kp(t,e,e[Wo],2,null,null)}function kIA(t){let e=t[Ch];if(!e)return mv(t[ki],t);for(;e;){let A=null;if(Q2(e))A=e[Ch];else{let i=e[ms];i&&(A=i)}if(!A){for(;e&&!e[ol]&&e!==t;)Q2(e)&&mv(e[ki],e),e=e[ps];e===null&&(e=t),Q2(e)&&mv(e[ki],e),A=e&&e[ol]}e=A}}function bb(t,e){let A=t[Ud],i=A.indexOf(e);A.splice(i,1)}function bp(t,e){if(Od(e))return;let A=e[Wo];A.destroyNode&&kp(t,e,A,3,null,null),kIA(e)}function mv(t,e){if(Od(e))return;let A=Ti(null);try{e[gi]&=-129,e[gi]|=256,e[wc]&&WQ(e[wc]),RIA(t,e),SIA(t,e),e[ki].type===1&&e[Wo].destroy();let i=e[hI];if(i!==null&&C0(e[ps])){i!==e[ps]&&bb(i,e);let n=e[l0];n!==null&&n.detachView(t)}Pv(e)}finally{Ti(A)}}function SIA(t,e){let A=t.cleanup,i=e[Km];if(A!==null)for(let r=0;r<A.length-1;r+=2)if(typeof A[r]=="string"){let s=A[r+3];s>=0?i[s]():i[-s].unsubscribe(),r+=2}else{let s=i[A[r+1]];A[r].call(s)}i!==null&&(e[Km]=null);let n=e[E2];if(n!==null){e[E2]=null;for(let r=0;r<n.length;r++){let s=n[r];s()}}let o=e[uI];if(o!==null){e[uI]=null;for(let r of o)r.destroy()}}function RIA(t,e){let A;if(t!=null&&(A=t.destroyHooks)!=null)for(let i=0;i<A.length;i+=2){let n=e[A[i]];if(!(n instanceof wI)){let o=A[i+1];if(Array.isArray(o))for(let r=0;r<o.length;r+=2){let s=n[o[r]],a=o[r+1];Ro(4,s,a);try{a.call(s)}finally{Ro(5,s,a)}}else{Ro(4,n,o);try{o.call(n)}finally{Ro(5,n,o)}}}}}function aT(t,e,A){return xIA(t,e.parent,A)}function xIA(t,e,A){let i=e;for(;i!==null&&i.type&168;)e=i,i=e.parent;if(i===null)return A[ig];if(zd(i)){let{encapsulation:n}=t.data[i.directiveStart+i.componentOffset];if(n===eg.None||n===eg.Emulated)return null}return ng(i,A)}function cT(t,e,A){return FIA(t,e,A)}function LIA(t,e,A){return t.type&40?ng(t,A):null}var FIA=LIA,$K;function Mp(t,e,A,i){let n=aT(t,i,e),o=e[Wo],r=i.parent||e[Ia],s=cT(r,i,e);if(n!=null)if(Array.isArray(A))for(let a=0;a<A.length;a++)XK(o,n,A[a],s,!1);else XK(o,n,A,s,!1);$K!==void 0&&$K(o,i,e,A,n)}function ch(t,e){if(e!==null){let A=e.type;if(A&3)return ng(e,t);if(A&4)return o9(-1,t[e.index]);if(A&8){let i=e.child;if(i!==null)return ch(t,i);{let n=t[e.index];return C0(n)?o9(-1,n):$l(n)}}else{if(A&128)return ch(t,e.next);if(A&32)return yb(e,t)()||$l(t[e.index]);{let i=lT(t,e);if(i!==null){if(Array.isArray(i))return i[0];let n=mI(t[Oa]);return ch(n,i)}else return ch(t,e.next)}}}return null}function lT(t,e){if(e!==null){let i=t[Oa][Ia],n=e.projection;return i.projection[n]}return null}function o9(t,e){let A=ms+t+1;if(A<e.length){let i=e[A],n=i[ki].firstChild;if(n!==null)return ch(i,n)}return e[fI]}function Mb(t,e,A,i,n,o,r){for(;A!=null;){if(A.type===128){A=A.next;continue}let s=i[A.index],a=A.type;if(r&&e===0&&(s&&qd($l(s),i),A.flags|=2),!vb(A))if(a&8)Mb(t,e,A.child,i,n,o,!1),Rd(e,t,n,s,o);else if(a&32){let c=yb(A,i),l;for(;l=c();)Rd(e,t,n,l,o);Rd(e,t,n,s,o)}else a&16?gT(t,e,i,A,n,o):Rd(e,t,n,s,o);A=r?A.projectionNext:A.next}}function kp(t,e,A,i,n,o){Mb(A,i,t.firstChild,e,n,o,!1)}function NIA(t,e,A){let i=e[Wo],n=aT(t,A,e),o=A.parent||e[Ia],r=cT(o,A,e);gT(i,0,e,A,n,r)}function gT(t,e,A,i,n,o){let r=A[Oa],a=r[Ia].projection[i.projection];if(Array.isArray(a))for(let c=0;c<a.length;c++){let l=a[c];Rd(e,t,n,l,o)}else{let c=a,l=r[ps];vJ(i)&&(c.flags|=128),Mb(t,e,c,l,n,o,!0)}}function _IA(t,e,A,i,n){let o=A[fI],r=$l(A);o!==r&&Rd(e,t,i,o,n);for(let s=ms;s<A.length;s++){let a=A[s];kp(a[ki],a,t,e,i,o)}}function GIA(t,e,A,i,n){if(e)n?t.addClass(A,i):t.removeClass(A,i);else{let o=i.indexOf("-")===-1?void 0:tg.DashCase;n==null?t.removeStyle(A,i,o):(typeof n=="string"&&n.endsWith("!important")&&(n=n.slice(0,-10),o|=tg.Important),t.setStyle(A,i,n,o))}}function Wm(t,e,A,i,n=!1){for(;A!==null;){if(A.type===128){A=n?A.projectionNext:A.next;continue}let o=e[A.index];o!==null&&i.push($l(o)),C0(o)&&UIA(o,i);let r=A.type;if(r&8)Wm(t,e,A.child,i);else if(r&32){let s=yb(A,e),a;for(;a=s();)i.push(a)}else if(r&16){let s=lT(e,A);if(Array.isArray(s))i.push(...s);else{let a=mI(e[Oa]);Wm(a[ki],a,s,i,!0)}}A=n?A.projectionNext:A.next}return i}function UIA(t,e){for(let A=ms;A<t.length;A++){let i=t[A],n=i[ki].firstChild;n!==null&&Wm(i[ki],i,n,e)}t[fI]!==t[ig]&&e.push(t[fI])}function IT(t){if(t[xd]!==null){for(let e of t[xd])e.impl.addSequence(e);t[xd].length=0}}var CT=[];function KIA(t){return t[wc]??YIA(t)}function YIA(t){let e=CT.pop()??Object.create(TIA);return e.lView=t,e}function JIA(t){t.lView[wc]!==t&&(t.lView=null,CT.push(t))}var TIA=Ye(rA({},Bd),{consumerIsAlwaysLive:!0,kind:"template",consumerMarkedDirty:t=>{Pd(t.lView)},consumerOnSignalRead(){this.lView[wc]=this}});function HIA(t){let e=t[wc]??Object.create(zIA);return e.lView=t,e}var zIA=Ye(rA({},Bd),{consumerIsAlwaysLive:!0,kind:"template",consumerMarkedDirty:t=>{let e=mI(t.lView);for(;e&&!dT(e[ki]);)e=mI(e);e&&VY(e)},consumerOnSignalRead(){this.lView[wc]=this}});function dT(t){return t.type!==2}function BT(t){if(t[uI]===null)return;let e=!0;for(;e;){let A=!1;for(let i of t[uI])i.dirty&&(A=!0,i.zone===null||Zone.current===i.zone?i.run():i.zone.run(()=>i.run()));e=A&&!!(t[gi]&8192)}}var OIA=100;function ET(t,e=!0,A=0){let n=t[c0].rendererFactory,o=!1;o||n.begin?.();try{PIA(t,A)}catch(r){throw e&&pIA(t,r),r}finally{o||n.end?.()}}function PIA(t,e){let A=$Y();try{Tm(!0),r9(t,e);let i=0;for(;Qp(t);){if(i===OIA)throw new Ae(103,!1);i++,r9(t,1)}}finally{Tm(A)}}function jIA(t,e,A,i){if(Od(e))return;let n=e[gi],o=!1,r=!1;nb(e);let s=!0,a=null,c=null;o||(dT(t)?(c=KIA(e),a=ZQ(c)):j7()===null?(s=!1,c=HIA(e),a=ZQ(c)):e[wc]&&(WQ(e[wc]),e[wc]=null));try{qY(e),D2A(t.bindingStartIndex),A!==null&&oT(t,e,A,2,i);let l=(n&3)===3;if(!o)if(l){let d=t.preOrderCheckHooks;d!==null&&Sm(e,d,null)}else{let d=t.preOrderHooks;d!==null&&Rm(e,d,0,null),uv(e,0)}if(r||qIA(e),BT(e),QT(e,0),t.contentQueries!==null&&UJ(t,e),!o)if(l){let d=t.contentCheckHooks;d!==null&&Sm(e,d)}else{let d=t.contentHooks;d!==null&&Rm(e,d,1),uv(e,1)}ZIA(t,e);let I=t.components;I!==null&&uT(e,I,0);let C=t.viewQuery;if(C!==null&&Vv(2,C,i),!o)if(l){let d=t.viewCheckHooks;d!==null&&Sm(e,d)}else{let d=t.viewHooks;d!==null&&Rm(e,d,2),uv(e,2)}if(t.firstUpdatePass===!0&&(t.firstUpdatePass=!1),e[hv]){for(let d of e[hv])d();e[hv]=null}o||(IT(e),e[gi]&=-73)}catch(l){throw o||Pd(e),l}finally{c!==null&&(Pf(c,a),s&&JIA(c)),ob()}}function QT(t,e){for(let A=kJ(t);A!==null;A=SJ(A))for(let i=ms;i<A.length;i++){let n=A[i];hT(n,e)}}function qIA(t){for(let e=kJ(t);e!==null;e=SJ(e)){if(!(e[gi]&2))continue;let A=e[Ud];for(let i=0;i<A.length;i++){let n=A[i];VY(n)}}}function VIA(t,e,A){Ro(18);let i=Ag(e,t);hT(i,A),Ro(19,i[Fr])}function hT(t,e){V9(t)&&r9(t,e)}function r9(t,e){let i=t[ki],n=t[gi],o=t[wc],r=!!(e===0&&n&16);if(r||=!!(n&64&&e===0),r||=!!(n&1024),r||=!!(o?.dirty&&jf(o)),r||=!1,o&&(o.dirty=!1),t[gi]&=-9217,r)jIA(i,t,i.template,t[Fr]);else if(n&8192){BT(t),QT(t,1);let s=i.components;s!==null&&uT(t,s,1),IT(t)}}function uT(t,e,A){for(let i=0;i<e.length;i++)VIA(t,e[i],A)}function ZIA(t,e){let A=t.hostBindingOpCodes;if(A!==null)try{for(let i=0;i<A.length;i++){let n=A[i];if(n<0)pI(~n);else{let o=n,r=A[++i],s=A[++i];v2A(r,o);let a=e[o];Ro(24,a),s(2,a),Ro(25,a)}}}finally{pI(-1)}}function kb(t,e){let A=$Y()?64:1088;for(t[c0].changeDetectionScheduler?.notify(e);t;){t[gi]|=A;let i=mI(t);if(Jm(t)&&!i)return t;t=i}return null}function fT(t,e,A,i){return[t,!0,0,e,null,i,null,A,null,null]}function mT(t,e){let A=ms+e;if(A<t.length)return t[A]}function kh(t,e,A,i=!0){let n=e[ki];if(WIA(n,e,t,A),i){let r=o9(A,t),s=e[Wo],a=s.parentNode(t[fI]);a!==null&&MIA(n,t[Ia],s,e,a,r)}let o=e[_d];o!==null&&o.firstChild!==null&&(o.firstChild=null)}function pT(t,e){let A=Eh(t,e);return A!==void 0&&bp(A[ki],A),A}function Eh(t,e){if(t.length<=ms)return;let A=ms+e,i=t[A];if(i){let n=i[hI];n!==null&&n!==t&&bb(n,i),e>0&&(t[A-1][ol]=i[ol]);let o=Gm(t,ms+e);bIA(i[ki],i);let r=o[l0];r!==null&&r.detachView(o[ki]),i[ps]=null,i[ol]=null,i[gi]&=-129}return i}function WIA(t,e,A,i){let n=ms+i,o=A.length;i>0&&(A[n-1][ol]=e),i<o-ms?(e[ol]=A[n],xY(A,ms+i,e)):(A.push(e),e[ol]=null),e[ps]=A;let r=e[hI];r!==null&&A!==r&&wT(r,e);let s=e[l0];s!==null&&s.insertView(t),Uv(e),e[gi]|=128}function wT(t,e){let A=t[Ud],i=e[ps];if(Q2(i))t[gi]|=2;else{let n=i[ps][Oa];e[Oa]!==n&&(t[gi]|=2)}A===null?t[Ud]=[e]:A.push(e)}var Qh=class{_lView;_cdRefInjectingView;notifyErrorHandler;_appRef=null;_attachedToViewContainer=!1;get rootNodes(){let e=this._lView,A=e[ki];return Wm(A,e,A.firstChild,[])}constructor(e,A,i=!0){this._lView=e,this._cdRefInjectingView=A,this.notifyErrorHandler=i}get context(){return this._lView[Fr]}set context(e){this._lView[Fr]=e}get destroyed(){return Od(this._lView)}destroy(){if(this._appRef)this._appRef.detachView(this);else if(this._attachedToViewContainer){let e=this._lView[ps];if(C0(e)){let A=e[Ym],i=A?A.indexOf(this):-1;i>-1&&(Eh(e,i),Gm(A,i))}this._attachedToViewContainer=!1}bp(this._lView[ki],this._lView)}onDestroy(e){ZY(this._lView,e)}markForCheck(){kb(this._cdRefInjectingView||this._lView,4)}detach(){this._lView[gi]&=-129}reattach(){Uv(this._lView),this._lView[gi]|=128}detectChanges(){this._lView[gi]|=1024,ET(this._lView,this.notifyErrorHandler)}checkNoChanges(){}attachToViewContainerRef(){if(this._appRef)throw new Ae(902,!1);this._attachedToViewContainer=!0}detachFromAppRef(){this._appRef=null;let e=Jm(this._lView),A=this._lView[hI];A!==null&&!e&&bb(A,this._lView),sT(this._lView[ki],this._lView)}attachToAppRef(e){if(this._attachedToViewContainer)throw new Ae(902,!1);this._appRef=e;let A=Jm(this._lView),i=this._lView[hI];i!==null&&!A&&wT(i,this._lView),Uv(this._lView)}};var wn=(()=>{class t{static __NG_ELEMENT_ID__=ACA}return t})(),XIA=wn,$IA=class extends XIA{_declarationLView;_declarationTContainer;elementRef;constructor(e,A,i){super(),this._declarationLView=e,this._declarationTContainer=A,this.elementRef=i}get ssrId(){return this._declarationTContainer.tView?.ssrId||null}createEmbeddedView(e,A){return this.createEmbeddedViewImpl(e,A)}createEmbeddedViewImpl(e,A,i){let n=Mh(this._declarationLView,this._declarationTContainer,e,{embeddedViewInjector:A,dehydratedView:i});return new Qh(n)}};function ACA(){return Sp(os(),ei())}function Sp(t,e){return t.type&4?new $IA(e,t,jd(t,e)):null}function Sh(t,e,A,i,n){let o=t.data[e];if(o===null)o=eCA(t,e,A,i,n),y2A()&&(o.flags|=32);else if(o.type&64){o.type=A,o.value=i,o.attrs=n;let r=m2A();o.injectorIndex=r===null?-1:r.injectorIndex}return kI(o,!0),o}function eCA(t,e,A,i,n){let o=XY(),r=$9(),s=r?o:o&&o.parent,a=t.data[e]=iCA(t,s,A,e,i,n);return tCA(t,a,o,r),a}function tCA(t,e,A,i){t.firstChild===null&&(t.firstChild=e),A!==null&&(i?A.child==null&&e.parent!==null&&(A.child=e):A.next===null&&(A.next=e,e.prev=A))}function iCA(t,e,A,i,n,o){let r=e?e.injectorIndex:-1,s=0;return WY()&&(s|=128),{type:A,index:i,insertBeforeIndex:null,injectorIndex:r,directiveStart:-1,directiveEnd:-1,directiveStylingLast:-1,componentOffset:-1,propertyBindings:null,flags:s,providerIndexes:0,value:n,attrs:o,mergedAttrs:null,localNames:null,initialInputs:null,inputs:null,hostDirectiveInputs:null,outputs:null,hostDirectiveOutputs:null,directiveToIndex:null,tView:null,next:null,prev:null,projectionNext:null,child:null,parent:e,projection:null,styles:null,stylesWithoutHost:null,residualStyles:void 0,classes:null,classesWithoutHost:null,residualClasses:void 0,classBindings:0,styleBindings:0}}var qie=new RegExp(`^(\\d+)*(${I1A}|${g1A})*(.*)`);var nCA=()=>null;function Jd(t,e){return nCA(t,e)}var oCA=class{},DT=class{},s9=class{resolveComponentFactory(e){throw Error(`No component factory found for ${ra(e)}.`)}},Rp=class{static NULL=new s9},ws=class{},Wi=(()=>{class t{destroyNode=null;static __NG_ELEMENT_ID__=()=>rCA()}return t})();function rCA(){let t=ei(),e=os(),A=Ag(e.index,t);return(Q2(A)?A:t)[Wo]}var sCA=(()=>{class t{static \u0275prov=NA({token:t,providedIn:"root",factory:()=>null})}return t})();var pv={},a9=class{injector;parentInjector;constructor(e,A){this.injector=e,this.parentInjector=A}get(e,A,i){i=gp(i);let n=this.injector.get(e,pv,i);return n!==pv||A===pv?n:this.parentInjector.get(e,A,i)}};function c9(t,e,A){let i=A?t.styles:null,n=A?t.classes:null,o=0;if(e!==null)for(let r=0;r<e.length;r++){let s=e[r];if(typeof s=="number")o=s;else if(o==1)n=Sv(n,s);else if(o==2){let a=s,c=e[++r];i=Sv(i,a+": "+c+";")}}A?t.styles=i:t.stylesWithoutHost=i,A?t.classes=n:t.classesWithoutHost=n}function PA(t,e=Gi.Default){let A=ei();if(A===null)return we(t,e);let i=os();return dJ(i,A,ns(t),e)}function yT(){let t="invalid";throw new Error(t)}function Sb(t,e,A,i,n){let o=i===null?null:{"":-1},r=n(t,A);if(r!==null){let s,a=null,c=null,l=cCA(r);l===null?s=r:[s,a,c]=l,ICA(t,e,A,s,o,a,c)}o!==null&&i!==null&&aCA(A,i,o)}function aCA(t,e,A){let i=t.localNames=[];for(let n=0;n<e.length;n+=2){let o=A[e[n+1]];if(o==null)throw new Ae(-301,!1);i.push(e[n],o)}}function cCA(t){let e=null,A=!1;for(let r=0;r<t.length;r++){let s=t[r];if(r===0&&rl(s)&&(e=s),s.findHostDirectiveDefs!==null){A=!0;break}}if(!A)return null;let i=null,n=null,o=null;for(let r of t)r.findHostDirectiveDefs!==null&&(i??=[],n??=new Map,o??=new Map,lCA(r,i,o,n)),r===e&&(i??=[],i.push(r));return i!==null?(i.push(...e===null?t:t.slice(1)),[i,n,o]):null}function lCA(t,e,A,i){let n=e.length;t.findHostDirectiveDefs(t,e,i),A.set(t,[n,e.length-1])}function gCA(t,e,A){e.componentOffset=A,(t.components??=[]).push(e.index)}function ICA(t,e,A,i,n,o,r){let s=i.length,a=!1;for(let C=0;C<s;C++){let d=i[C];!a&&rl(d)&&(a=!0,gCA(t,A,C)),Tv(Pm(A,e),t,d.type)}hCA(A,t.data.length,s);for(let C=0;C<s;C++){let d=i[C];d.providersResolver&&d.providersResolver(d)}let c=!1,l=!1,I=iT(t,e,s,null);s>0&&(A.directiveToIndex=new Map);for(let C=0;C<s;C++){let d=i[C];if(A.mergedAttrs=Kd(A.mergedAttrs,d.hostAttrs),dCA(t,A,e,I,d),QCA(I,d,n),r!==null&&r.has(d)){let[E,h]=r.get(d);A.directiveToIndex.set(d.type,[I,E+A.directiveStart,h+A.directiveStart])}else(o===null||!o.has(d))&&A.directiveToIndex.set(d.type,I);d.contentQueries!==null&&(A.flags|=4),(d.hostBindings!==null||d.hostAttrs!==null||d.hostVars!==0)&&(A.flags|=64);let B=d.type.prototype;!c&&(B.ngOnChanges||B.ngOnInit||B.ngDoCheck)&&((t.preOrderHooks??=[]).push(A.index),c=!0),!l&&(B.ngOnChanges||B.ngDoCheck)&&((t.preOrderCheckHooks??=[]).push(A.index),l=!0),I++}CCA(t,A,o)}function CCA(t,e,A){for(let i=e.directiveStart;i<e.directiveEnd;i++){let n=t.data[i];if(A===null||!A.has(n))AY(0,e,n,i),AY(1,e,n,i),tY(e,i,!1);else{let o=A.get(n);eY(0,e,o,i),eY(1,e,o,i),tY(e,i,!0)}}}function AY(t,e,A,i){let n=t===0?A.inputs:A.outputs;for(let o in n)if(n.hasOwnProperty(o)){let r;t===0?r=e.inputs??={}:r=e.outputs??={},r[o]??=[],r[o].push(i),vT(e,o)}}function eY(t,e,A,i){let n=t===0?A.inputs:A.outputs;for(let o in n)if(n.hasOwnProperty(o)){let r=n[o],s;t===0?s=e.hostDirectiveInputs??={}:s=e.hostDirectiveOutputs??={},s[r]??=[],s[r].push(i,o),vT(e,r)}}function vT(t,e){e==="class"?t.flags|=8:e==="style"&&(t.flags|=16)}function tY(t,e,A){let{attrs:i,inputs:n,hostDirectiveInputs:o}=t;if(i===null||!A&&n===null||A&&o===null||Qb(t)){t.initialInputs??=[],t.initialInputs.push(null);return}let r=null,s=0;for(;s<i.length;){let a=i[s];if(a===0){s+=4;continue}else if(a===5){s+=2;continue}else if(typeof a=="number")break;if(!A&&n.hasOwnProperty(a)){let c=n[a];for(let l of c)if(l===e){r??=[],r.push(a,i[s+1]);break}}else if(A&&o.hasOwnProperty(a)){let c=o[a];for(let l=0;l<c.length;l+=2)if(c[l]===e){r??=[],r.push(c[l+1],i[s+1]);break}}s+=2}t.initialInputs??=[],t.initialInputs.push(r)}function dCA(t,e,A,i,n){t.data[i]=n;let o=n.factory||(n.factory=QI(n.type,!0)),r=new wI(o,rl(n),PA);t.blueprint[i]=r,A[i]=r,BCA(t,e,i,iT(t,A,n.hostVars,qa),n)}function BCA(t,e,A,i,n){let o=n.hostBindings;if(o){let r=t.hostBindingOpCodes;r===null&&(r=t.hostBindingOpCodes=[]);let s=~e.index;ECA(r)!=s&&r.push(s),r.push(A,i,o)}}function ECA(t){let e=t.length;for(;e>0;){let A=t[--e];if(typeof A=="number"&&A<0)return A}return 0}function QCA(t,e,A){if(A){if(e.exportAs)for(let i=0;i<e.exportAs.length;i++)A[e.exportAs[i]]=t;rl(e)&&(A[""]=t)}}function hCA(t,e,A){t.flags|=1,t.directiveStart=e,t.directiveEnd=e+A,t.providerIndexes=e}function bT(t,e,A,i,n,o,r,s){let a=e.consts,c=u2(a,r),l=Sh(e,t,2,i,c);return o&&Sb(e,A,l,u2(a,s),n),l.mergedAttrs=Kd(l.mergedAttrs,l.attrs),l.attrs!==null&&c9(l,l.attrs,!1),l.mergedAttrs!==null&&c9(l,l.mergedAttrs,!0),e.queries!==null&&e.queries.elementStart(e,l),l}function MT(t,e){rb(t,e),P9(e)&&t.queries.elementEnd(e)}var Xm=class extends Rp{ngModule;constructor(e){super(),this.ngModule=e}resolveComponentFactory(e){let A=h2(e);return new yI(A,this.ngModule)}};function uCA(t){return Object.keys(t).map(e=>{let[A,i,n]=t[e],o={propName:A,templateName:e,isSignal:(i&Dp.SignalBased)!==0};return n&&(o.transform=n),o})}function fCA(t){return Object.keys(t).map(e=>({propName:t[e],templateName:e}))}function mCA(t,e,A){let i=e instanceof pr?e:e?.injector;return i&&t.getStandaloneInjector!==null&&(i=t.getStandaloneInjector(i)||i),i?new a9(A,i):A}function pCA(t){let e=t.get(ws,null);if(e===null)throw new Ae(407,!1);let A=t.get(sCA,null),i=t.get(DI,null);return{rendererFactory:e,sanitizer:A,changeDetectionScheduler:i}}function wCA(t,e){let A=(t.selectors[0][0]||"div").toLowerCase();return $J(e,A,A==="svg"?PY:A==="math"?a2A:null)}var yI=class extends DT{componentDef;ngModule;selector;componentType;ngContentSelectors;isBoundToModule;cachedInputs=null;cachedOutputs=null;get inputs(){return this.cachedInputs??=uCA(this.componentDef.inputs),this.cachedInputs}get outputs(){return this.cachedOutputs??=fCA(this.componentDef.outputs),this.cachedOutputs}constructor(e,A){super(),this.componentDef=e,this.ngModule=A,this.componentType=e.type,this.selector=AIA(e.selectors),this.ngContentSelectors=e.ngContentSelectors??[],this.isBoundToModule=!!A}create(e,A,i,n){Ro(22);let o=Ti(null);try{let r=this.componentDef,s=i?["ng-version","19.2.14"]:eIA(this.componentDef.selectors[0]),a=hb(0,null,null,1,0,null,null,null,null,[s],null),c=mCA(r,n||this.ngModule,e),l=pCA(c),I=l.rendererFactory.createRenderer(null,r),C=i?gIA(I,i,r.encapsulation,c):wCA(r,I),d=ub(null,a,null,512|tT(r),null,null,l,I,c,null,GJ(C,c,!0));d[Nr]=C,nb(d);let B=null;try{let E=bT(Nr,a,d,"#host",()=>[this.componentDef],!0,0);C&&(eT(I,C,E),qd(C,d)),yp(a,d,E),Bb(a,E,d),MT(a,E),A!==void 0&&DCA(E,this.ngContentSelectors,A),B=Ag(E.index,d),d[Fr]=B[Fr],Db(a,d,null)}catch(E){throw B!==null&&Pv(B),Pv(d),E}finally{Ro(23),ob()}return new l9(this.componentType,d)}finally{Ti(o)}}},l9=class extends oCA{_rootLView;instance;hostView;changeDetectorRef;componentType;location;previousInputValues=null;_tNode;constructor(e,A){super(),this._rootLView=A,this._tNode=j9(A[ki],Nr),this.location=jd(this._tNode,A),this.instance=Ag(this._tNode.index,A)[Fr],this.hostView=this.changeDetectorRef=new Qh(A,void 0,!1),this.componentType=e}setInput(e,A){let i=this._tNode;if(this.previousInputValues??=new Map,this.previousInputValues.has(e)&&Object.is(this.previousInputValues.get(e),A))return;let n=this._rootLView,o=wb(i,n[ki],n,e,A);this.previousInputValues.set(e,A);let r=Ag(i.index,n);kb(r,1)}get injector(){return new BI(this._tNode,this._rootLView)}destroy(){this.hostView.destroy()}onDestroy(e){this.hostView.onDestroy(e)}};function DCA(t,e,A){let i=t.projection=[];for(let n=0;n<e.length;n++){let o=A[n];i.push(o!=null&&o.length?Array.from(o):null)}}var Nn=(()=>{class t{static __NG_ELEMENT_ID__=yCA}return t})();function yCA(){let t=os();return ST(t,ei())}var vCA=Nn,kT=class extends vCA{_lContainer;_hostTNode;_hostLView;constructor(e,A,i){super(),this._lContainer=e,this._hostTNode=A,this._hostLView=i}get element(){return jd(this._hostTNode,this._hostLView)}get injector(){return new BI(this._hostTNode,this._hostLView)}get parentInjector(){let e=sb(this._hostTNode,this._hostLView);if(aJ(e)){let A=zm(e,this._hostLView),i=Hm(e),n=A[ki].data[i+8];return new BI(n,A)}else return new BI(null,this._hostLView)}clear(){for(;this.length>0;)this.remove(this.length-1)}get(e){let A=iY(this._lContainer);return A!==null&&A[e]||null}get length(){return this._lContainer.length-ms}createEmbeddedView(e,A,i){let n,o;typeof i=="number"?n=i:i!=null&&(n=i.index,o=i.injector);let r=Jd(this._lContainer,e.ssrId),s=e.createEmbeddedViewImpl(A||{},o,r);return this.insertImpl(s,n,Yd(this._hostTNode,r)),s}createComponent(e,A,i,n,o){let r=e&&!n2A(e),s;if(r)s=A;else{let B=A||{};s=B.index,i=B.injector,n=B.projectableNodes,o=B.environmentInjector||B.ngModuleRef}let a=r?e:new yI(h2(e)),c=i||this.parentInjector;if(!o&&a.ngModule==null){let E=(r?c:this.parentInjector).get(pr,null);E&&(o=E)}let l=h2(a.componentType??{}),I=Jd(this._lContainer,l?.id??null),C=I?.firstChild??null,d=a.create(c,n,C,o);return this.insertImpl(d.hostView,s,Yd(this._hostTNode,I)),d}insert(e,A){return this.insertImpl(e,A,!0)}insertImpl(e,A,i){let n=e._lView;if(I2A(n)){let s=this.indexOf(e);if(s!==-1)this.detach(s);else{let a=n[ps],c=new kT(a,a[Ia],a[ps]);c.detach(c.indexOf(e))}}let o=this._adjustIndex(A),r=this._lContainer;return kh(r,n,o,i),e.attachToViewContainerRef(),xY(wv(r),o,e),e}move(e,A){return this.insert(e,A)}indexOf(e){let A=iY(this._lContainer);return A!==null?A.indexOf(e):-1}remove(e){let A=this._adjustIndex(e,-1),i=Eh(this._lContainer,A);i&&(Gm(wv(this._lContainer),A),bp(i[ki],i))}detach(e){let A=this._adjustIndex(e,-1),i=Eh(this._lContainer,A);return i&&Gm(wv(this._lContainer),A)!=null?new Qh(i):null}_adjustIndex(e,A=0){return e??this.length+A}};function iY(t){return t[Ym]}function wv(t){return t[Ym]||(t[Ym]=[])}function ST(t,e){let A,i=e[t.index];return C0(i)?A=i:(A=fT(i,e,null,t),e[t.index]=A,fb(e,A)),MCA(A,e,t,i),new kT(A,t,e)}function bCA(t,e){let A=t[Wo],i=A.createComment(""),n=ng(e,t),o=A.parentNode(n);return Zm(A,o,i,A.nextSibling(n),!1),i}var MCA=RCA,kCA=()=>!1;function SCA(t,e,A){return kCA(t,e,A)}function RCA(t,e,A,i){if(t[fI])return;let n;A.type&8?n=$l(i):n=bCA(e,A),t[fI]=n}var g9=class t{queryList;matches=null;constructor(e){this.queryList=e}clone(){return new t(this.queryList)}setDirty(){this.queryList.setDirty()}},I9=class t{queries;constructor(e=[]){this.queries=e}createEmbeddedView(e){let A=e.queries;if(A!==null){let i=e.contentQueries!==null?e.contentQueries[0]:A.length,n=[];for(let o=0;o<i;o++){let r=A.getByIndex(o),s=this.queries[r.indexInDeclarationView];n.push(s.clone())}return new t(n)}return null}insertView(e){this.dirtyQueriesWithMatches(e)}detachView(e){this.dirtyQueriesWithMatches(e)}finishViewCreation(e){this.dirtyQueriesWithMatches(e)}dirtyQueriesWithMatches(e){for(let A=0;A<this.queries.length;A++)xb(e,A).matches!==null&&this.queries[A].setDirty()}},$m=class{flags;read;predicate;constructor(e,A,i=null){this.flags=A,this.read=i,typeof e=="string"?this.predicate=GCA(e):this.predicate=e}},C9=class t{queries;constructor(e=[]){this.queries=e}elementStart(e,A){for(let i=0;i<this.queries.length;i++)this.queries[i].elementStart(e,A)}elementEnd(e){for(let A=0;A<this.queries.length;A++)this.queries[A].elementEnd(e)}embeddedTView(e){let A=null;for(let i=0;i<this.length;i++){let n=A!==null?A.length:0,o=this.getByIndex(i).embeddedTView(e,n);o&&(o.indexInDeclarationView=i,A!==null?A.push(o):A=[o])}return A!==null?new t(A):null}template(e,A){for(let i=0;i<this.queries.length;i++)this.queries[i].template(e,A)}getByIndex(e){return this.queries[e]}get length(){return this.queries.length}track(e){this.queries.push(e)}},d9=class t{metadata;matches=null;indexInDeclarationView=-1;crossesNgTemplate=!1;_declarationNodeIndex;_appliesToNextNode=!0;constructor(e,A=-1){this.metadata=e,this._declarationNodeIndex=A}elementStart(e,A){this.isApplyingToNode(A)&&this.matchTNode(e,A)}elementEnd(e){this._declarationNodeIndex===e.index&&(this._appliesToNextNode=!1)}template(e,A){this.elementStart(e,A)}embeddedTView(e,A){return this.isApplyingToNode(e)?(this.crossesNgTemplate=!0,this.addMatch(-e.index,A),new t(this.metadata)):null}isApplyingToNode(e){if(this._appliesToNextNode&&(this.metadata.flags&1)!==1){let A=this._declarationNodeIndex,i=e.parent;for(;i!==null&&i.type&8&&i.index!==A;)i=i.parent;return A===(i!==null?i.index:-1)}return this._appliesToNextNode}matchTNode(e,A){let i=this.metadata.predicate;if(Array.isArray(i))for(let n=0;n<i.length;n++){let o=i[n];this.matchTNodeWithReadOption(e,A,xCA(A,o)),this.matchTNodeWithReadOption(e,A,xm(A,e,o,!1,!1))}else i===wn?A.type&4&&this.matchTNodeWithReadOption(e,A,-1):this.matchTNodeWithReadOption(e,A,xm(A,e,i,!1,!1))}matchTNodeWithReadOption(e,A,i){if(i!==null){let n=this.metadata.read;if(n!==null)if(n===ee||n===Nn||n===wn&&A.type&4)this.addMatch(A.index,-2);else{let o=xm(A,e,n,!1,!1);o!==null&&this.addMatch(A.index,o)}else this.addMatch(A.index,i)}}addMatch(e,A){this.matches===null?this.matches=[e,A]:this.matches.push(e,A)}};function xCA(t,e){let A=t.localNames;if(A!==null){for(let i=0;i<A.length;i+=2)if(A[i]===e)return A[i+1]}return null}function LCA(t,e){return t.type&11?jd(t,e):t.type&4?Sp(t,e):null}function FCA(t,e,A,i){return A===-1?LCA(e,t):A===-2?NCA(t,e,i):dh(t,t[ki],A,e)}function NCA(t,e,A){if(A===ee)return jd(e,t);if(A===wn)return Sp(e,t);if(A===Nn)return ST(e,t)}function RT(t,e,A,i){let n=e[l0].queries[i];if(n.matches===null){let o=t.data,r=A.matches,s=[];for(let a=0;r!==null&&a<r.length;a+=2){let c=r[a];if(c<0)s.push(null);else{let l=o[c];s.push(FCA(e,l,r[a+1],A.metadata.read))}}n.matches=s}return n.matches}function B9(t,e,A,i){let n=t.queries.getByIndex(A),o=n.matches;if(o!==null){let r=RT(t,e,n,A);for(let s=0;s<o.length;s+=2){let a=o[s];if(a>0)i.push(r[s/2]);else{let c=o[s+1],l=e[-a];for(let I=ms;I<l.length;I++){let C=l[I];C[hI]===C[ps]&&B9(C[ki],C,c,i)}if(l[Ud]!==null){let I=l[Ud];for(let C=0;C<I.length;C++){let d=I[C];B9(d[ki],d,c,i)}}}}}return i}function Rb(t,e){return t[l0].queries[e].queryList}function xT(t,e,A){let i=new ca((A&4)===4);return B2A(t,e,i,i.destroy),(e[l0]??=new I9).queries.push(new g9(i))-1}function _CA(t,e,A){let i=Yo();return i.firstCreatePass&&(FT(i,new $m(t,e,A),-1),(e&2)===2&&(i.staticViewQueries=!0)),xT(i,ei(),e)}function LT(t,e,A,i){let n=Yo();if(n.firstCreatePass){let o=os();FT(n,new $m(e,A,i),o.index),UCA(n,t),(A&2)===2&&(n.staticContentQueries=!0)}return xT(n,ei(),A)}function GCA(t){return t.split(",").map(e=>e.trim())}function FT(t,e,A){t.queries===null&&(t.queries=new C9),t.queries.track(new d9(e,A))}function UCA(t,e){let A=t.contentQueries||(t.contentQueries=[]),i=A.length?A[A.length-1]:-1;e!==i&&A.push(t.queries.length-1,e)}function xb(t,e){return t.queries.getByIndex(e)}function NT(t,e){let A=t[ki],i=xb(A,e);return i.crossesNgTemplate?B9(A,t,e,[]):RT(A,t,i,e)}function _T(t,e,A){let i,n=Zf(()=>{i._dirtyCounter();let o=TCA(i,t);if(e&&o===void 0)throw new Ae(-951,!1);return o});return i=n[ia],i._dirtyCounter=Jo(0),i._flatValue=void 0,n}function KCA(t){return _T(!0,!1,t)}function YCA(t){return _T(!0,!0,t)}function JCA(t,e){let A=t[ia];A._lView=ei(),A._queryIndex=e,A._queryList=Rb(A._lView,e),A._queryList.onDirty(()=>A._dirtyCounter.update(i=>i+1))}function TCA(t,e){let A=t._lView,i=t._queryIndex;if(A===void 0||i===void 0||A[gi]&4)return e?void 0:Ts;let n=Rb(A,i),o=NT(A,i);return n.reset(o,DJ),e?n.first:n._changesDetected||t._flatValue===void 0?t._flatValue=n.toArray():t._flatValue}function nY(t,e){return KCA(e)}function HCA(t,e){return YCA(e)}var GT=(nY.required=HCA,nY);function zCA(t){let e=[],A=new Map;function i(n){let o=A.get(n);if(!o){let r=t(n);A.set(n,o=r.then(qCA))}return o}return Ap.forEach((n,o)=>{let r=[];n.templateUrl&&r.push(i(n.templateUrl).then(c=>{n.template=c}));let s=typeof n.styles=="string"?[n.styles]:n.styles||[];if(n.styles=s,n.styleUrl&&n.styleUrls?.length)throw new Error("@Component cannot define both `styleUrl` and `styleUrls`. Use `styleUrl` if the component has one stylesheet, or `styleUrls` if it has multiple");if(n.styleUrls?.length){let c=n.styles.length,l=n.styleUrls;n.styleUrls.forEach((I,C)=>{s.push(""),r.push(i(I).then(d=>{s[c+C]=d,l.splice(l.indexOf(I),1),l.length==0&&(n.styleUrls=void 0)}))})}else n.styleUrl&&r.push(i(n.styleUrl).then(c=>{s.push(c),n.styleUrl=void 0}));let a=Promise.all(r).then(()=>VCA(o));e.push(a)}),PCA(),Promise.all(e).then(()=>{})}var Ap=new Map,OCA=new Set;function PCA(){let t=Ap;return Ap=new Map,t}function jCA(){return Ap.size===0}function qCA(t){return typeof t=="string"?t:t.text()}function VCA(t){OCA.delete(t)}var I0=class{},Lb=class{};var ep=class extends I0{ngModuleType;_parent;_bootstrapComponents=[];_r3Injector;instance;destroyCbs=[];componentFactoryResolver=new Xm(this);constructor(e,A,i,n=!0){super(),this.ngModuleType=e,this._parent=A;let o=NY(e);this._bootstrapComponents=ZJ(o.bootstrap),this._r3Injector=QJ(e,A,[{provide:I0,useValue:this},{provide:Rp,useValue:this.componentFactoryResolver},...i],ra(e),new Set(["environment"])),n&&this.resolveInjectorInitializers()}resolveInjectorInitializers(){this._r3Injector.resolveInjectorInitializers(),this.instance=this._r3Injector.get(this.ngModuleType)}get injector(){return this._r3Injector}destroy(){let e=this._r3Injector;!e.destroyed&&e.destroy(),this.destroyCbs.forEach(A=>A()),this.destroyCbs=null}onDestroy(e){this.destroyCbs.push(e)}},tp=class extends Lb{moduleType;constructor(e){super(),this.moduleType=e}create(e){return new ep(this.moduleType,e,[])}};function ZCA(t,e,A){return new ep(t,e,A,!1)}var E9=class extends I0{injector;componentFactoryResolver=new Xm(this);instance=null;constructor(e){super();let A=new Ih([...e.providers,{provide:I0,useValue:this},{provide:Rp,useValue:this.componentFactoryResolver}],e.parent||dp(),e.debugName,new Set(["environment"]));this.injector=A,e.runEnvironmentInitializers&&A.resolveInjectorInitializers()}destroy(){this.injector.destroy()}onDestroy(e){this.injector.onDestroy(e)}};function Rh(t,e,A=null){return new E9({providers:t,parent:e,debugName:A,runEnvironmentInitializers:!0}).injector}var WCA=(()=>{class t{_injector;cachedInjectors=new Map;constructor(A){this._injector=A}getOrCreateStandaloneInjector(A){if(!A.standalone)return null;if(!this.cachedInjectors.has(A)){let i=GY(!1,A.type),n=i.length>0?Rh([i],this._injector,`Standalone[${A.type.name}]`):null;this.cachedInjectors.set(A,n)}return this.cachedInjectors.get(A)}ngOnDestroy(){try{for(let A of this.cachedInjectors.values())A!==null&&A.destroy()}finally{this.cachedInjectors.clear()}}static \u0275prov=NA({token:t,providedIn:"environment",factory:()=>new t(we(pr))})}return t})();function zA(t){return uh(()=>{let e=UT(t),A=Ye(rA({},e),{decls:t.decls,vars:t.vars,template:t.template,consts:t.consts||null,ngContentSelectors:t.ngContentSelectors,onPush:t.changeDetection===bJ.OnPush,directiveDefs:null,pipeDefs:null,dependencies:e.standalone&&t.dependencies||null,getStandaloneInjector:e.standalone?n=>n.get(WCA).getOrCreateStandaloneInjector(A):null,getExternalStyles:null,signals:t.signals??!1,data:t.data||{},encapsulation:t.encapsulation||eg.Emulated,styles:t.styles||Ts,_:null,schemas:t.schemas||null,tView:null,id:""});e.standalone&&E0("NgStandalone"),KT(A);let i=t.dependencies;return A.directiveDefs=oY(i,!1),A.pipeDefs=oY(i,!0),A.id=tdA(A),A})}function XCA(t){return h2(t)||_Y(t)}function $CA(t){return t!==null}function Ce(t){return uh(()=>({type:t.type,bootstrap:t.bootstrap||Ts,declarations:t.declarations||Ts,imports:t.imports||Ts,exports:t.exports||Ts,transitiveCompileScopes:null,schemas:t.schemas||null,id:t.id||null}))}function AdA(t,e){if(t==null)return Xl;let A={};for(let i in t)if(t.hasOwnProperty(i)){let n=t[i],o,r,s,a;Array.isArray(n)?(s=n[0],o=n[1],r=n[2]??o,a=n[3]||null):(o=n,r=n,s=Dp.None,a=null),A[o]=[i,s,a],e[o]=r}return A}function edA(t){if(t==null)return Xl;let e={};for(let A in t)t.hasOwnProperty(A)&&(e[t[A]]=A);return e}function jA(t){return uh(()=>{let e=UT(t);return KT(e),e})}function xp(t){return{type:t.type,name:t.name,factory:null,pure:t.pure!==!1,standalone:t.standalone??!0,onDestroy:t.type.prototype.ngOnDestroy||null}}function UT(t){let e={};return{type:t.type,providersResolver:null,factory:null,hostBindings:t.hostBindings||null,hostVars:t.hostVars||0,hostAttrs:t.hostAttrs||null,contentQueries:t.contentQueries||null,declaredInputs:e,inputConfig:t.inputs||Xl,exportAs:t.exportAs||null,standalone:t.standalone??!0,signals:t.signals===!0,selectors:t.selectors||Ts,viewQuery:t.viewQuery||null,features:t.features||null,setInput:null,findHostDirectiveDefs:null,hostDirectives:null,inputs:AdA(t.inputs,e),outputs:edA(t.outputs),debugInfo:null}}function KT(t){t.features?.forEach(e=>e(t))}function oY(t,e){if(!t)return null;let A=e?P0A:XCA;return()=>(typeof t=="function"?t():t).map(i=>A(i)).filter($CA)}function tdA(t){let e=0,A=typeof t.consts=="function"?"":t.consts,i=[t.selectors,t.ngContentSelectors,t.hostVars,t.hostAttrs,A,t.vars,t.decls,t.encapsulation,t.standalone,t.signals,t.exportAs,JSON.stringify(t.inputs),JSON.stringify(t.outputs),Object.getOwnPropertyNames(t.type.prototype),!!t.contentQueries,!!t.viewQuery];for(let o of i.join("|"))e=Math.imul(31,e)+o.charCodeAt(0)<<0;return e+=2147483648,"c"+e}function idA(t){return Object.getPrototypeOf(t.prototype).constructor}function lt(t){let e=idA(t.type),A=!0,i=[t];for(;e;){let n;if(rl(t))n=e.\u0275cmp||e.\u0275dir;else{if(e.\u0275cmp)throw new Ae(903,!1);n=e.\u0275dir}if(n){if(A){i.push(n);let r=t;r.inputs=Dv(t.inputs),r.declaredInputs=Dv(t.declaredInputs),r.outputs=Dv(t.outputs);let s=n.hostBindings;s&&adA(t,s);let a=n.viewQuery,c=n.contentQueries;if(a&&rdA(t,a),c&&sdA(t,c),ndA(t,n),D0A(t.outputs,n.outputs),rl(n)&&n.data.animation){let l=t.data;l.animation=(l.animation||[]).concat(n.data.animation)}}let o=n.features;if(o)for(let r=0;r<o.length;r++){let s=o[r];s&&s.ngInherit&&s(t),s===lt&&(A=!1)}}e=Object.getPrototypeOf(e)}odA(i)}function ndA(t,e){for(let A in e.inputs){if(!e.inputs.hasOwnProperty(A)||t.inputs.hasOwnProperty(A))continue;let i=e.inputs[A];i!==void 0&&(t.inputs[A]=i,t.declaredInputs[A]=e.declaredInputs[A])}}function odA(t){let e=0,A=null;for(let i=t.length-1;i>=0;i--){let n=t[i];n.hostVars=e+=n.hostVars,n.hostAttrs=Kd(n.hostAttrs,A=Kd(A,n.hostAttrs))}}function Dv(t){return t===Xl?{}:t===Ts?[]:t}function rdA(t,e){let A=t.viewQuery;A?t.viewQuery=(i,n)=>{e(i,n),A(i,n)}:t.viewQuery=e}function sdA(t,e){let A=t.contentQueries;A?t.contentQueries=(i,n,o)=>{e(i,n,o),A(i,n,o)}:t.contentQueries=e}function adA(t,e){let A=t.hostBindings;A?t.hostBindings=(i,n)=>{e(i,n),A(i,n)}:t.hostBindings=e}function YT(t){let e=A=>{let i=Array.isArray(t);A.hostDirectives===null?(A.findHostDirectiveDefs=JT,A.hostDirectives=i?t.map(Q9):[t]):i?A.hostDirectives.unshift(...t.map(Q9)):A.hostDirectives.unshift(t)};return e.ngInherit=!0,e}function JT(t,e,A){if(t.hostDirectives!==null)for(let i of t.hostDirectives)if(typeof i=="function"){let n=i();for(let o of n)rY(Q9(o),e,A)}else rY(i,e,A)}function rY(t,e,A){let i=_Y(t.directive);cdA(i.declaredInputs,t.inputs),JT(i,e,A),A.set(i,t),e.push(i)}function Q9(t){return typeof t=="function"?{directive:ns(t),inputs:Xl,outputs:Xl}:{directive:ns(t.directive),inputs:sY(t.inputs),outputs:sY(t.outputs)}}function sY(t){if(t===void 0||t.length===0)return Xl;let e={};for(let A=0;A<t.length;A+=2)e[t[A]]=t[A+1];return e}function cdA(t,e){for(let A in e)if(e.hasOwnProperty(A)){let i=e[A],n=t[A];t[i]=n}}function TT(t){return Fb(t)?Array.isArray(t)||!(t instanceof Map)&&Symbol.iterator in t:!1}function ldA(t,e){if(Array.isArray(t))for(let A=0;A<t.length;A++)e(t[A]);else{let A=t[Symbol.iterator](),i;for(;!(i=A.next()).done;)e(i.value)}}function Fb(t){return t!==null&&(typeof t=="function"||typeof t=="object")}function Nb(t,e,A){return t[e]=A}function gdA(t,e){return t[e]}function Pa(t,e,A){let i=t[e];return Object.is(i,A)?!1:(t[e]=A,!0)}function HT(t,e,A,i){let n=Pa(t,e,A);return Pa(t,e+1,i)||n}function IdA(t,e,A,i,n,o,r,s,a){let c=e.consts,l=Sh(e,t,4,r||null,s||null);X9()&&Sb(e,A,l,u2(c,a),pb),l.mergedAttrs=Kd(l.mergedAttrs,l.attrs),rb(e,l);let I=l.tView=hb(2,l,i,n,o,e.directiveRegistry,e.pipeRegistry,null,e.schemas,c,null);return e.queries!==null&&(e.queries.template(e,l),I.queries=e.queries.embeddedTView(l)),l}function ip(t,e,A,i,n,o,r,s,a,c){let l=A+Nr,I=e.firstCreatePass?IdA(l,e,t,i,n,o,r,s,a):e.data[l];kI(I,!1);let C=CdA(e,t,I,A);up()&&Mp(e,t,C,I),qd(C,t);let d=fT(C,t,C,I);return t[l]=d,fb(t,d),SCA(d,I,t),Ep(I)&&yp(e,t,I),a!=null&&mb(t,I,c),I}function _A(t,e,A,i,n,o,r,s){let a=ei(),c=Yo(),l=u2(c.consts,o);return ip(a,c,t,e,A,i,n,l,r,s),_A}var CdA=ddA;function ddA(t,e,A,i){return fp(!0),e[Wo].createComment("")}var _b=(()=>{class t{log(A){console.log(A)}warn(A){console.warn(A)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"platform"})}return t})();var Gb=new dA(""),xh=new dA(""),Lp=(()=>{class t{_ngZone;registry;_isZoneStable=!0;_callbacks=[];_taskTrackingZone=null;_destroyRef;constructor(A,i,n){this._ngZone=A,this.registry=i,H9()&&(this._destroyRef=f(m2,{optional:!0})??void 0),Ub||(BdA(n),n.addToWindow(i)),this._watchAngularEvents(),A.run(()=>{this._taskTrackingZone=typeof Zone>"u"?null:Zone.current.get("TaskTrackingZone")})}_watchAngularEvents(){let A=this._ngZone.onUnstable.subscribe({next:()=>{this._isZoneStable=!1}}),i=this._ngZone.runOutsideAngular(()=>this._ngZone.onStable.subscribe({next:()=>{Qe.assertNotInAngularZone(),queueMicrotask(()=>{this._isZoneStable=!0,this._runCallbacksIfReady()})}}));this._destroyRef?.onDestroy(()=>{A.unsubscribe(),i.unsubscribe()})}isStable(){return this._isZoneStable&&!this._ngZone.hasPendingMacrotasks}_runCallbacksIfReady(){if(this.isStable())queueMicrotask(()=>{for(;this._callbacks.length!==0;){let A=this._callbacks.pop();clearTimeout(A.timeoutId),A.doneCb()}});else{let A=this.getPendingTasks();this._callbacks=this._callbacks.filter(i=>i.updateCb&&i.updateCb(A)?(clearTimeout(i.timeoutId),!1):!0)}}getPendingTasks(){return this._taskTrackingZone?this._taskTrackingZone.macroTasks.map(A=>({source:A.source,creationLocation:A.creationLocation,data:A.data})):[]}addCallback(A,i,n){let o=-1;i&&i>0&&(o=setTimeout(()=>{this._callbacks=this._callbacks.filter(r=>r.timeoutId!==o),A()},i)),this._callbacks.push({doneCb:A,timeoutId:o,updateCb:n})}whenStable(A,i,n){if(n&&!this._taskTrackingZone)throw new Error('Task tracking zone is required when passing an update callback to whenStable(). Is "zone.js/plugins/task-tracking" loaded?');this.addCallback(A,i,n),this._runCallbacksIfReady()}registerApplication(A){this.registry.registerApplication(A,this)}unregisterApplication(A){this.registry.unregisterApplication(A)}findProviders(A,i,n){return[]}static \u0275fac=function(i){return new(i||t)(we(Qe),we(Fp),we(xh))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),Fp=(()=>{class t{_applications=new Map;registerApplication(A,i){this._applications.set(A,i)}unregisterApplication(A){this._applications.delete(A)}unregisterAllApplications(){this._applications.clear()}getTestability(A){return this._applications.get(A)||null}getAllTestabilities(){return Array.from(this._applications.values())}getAllRootElements(){return Array.from(this._applications.keys())}findTestabilityInTree(A,i=!0){return Ub?.findTestabilityInTree(this,A,i)??null}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"platform"})}return t})();function BdA(t){Ub=t}var Ub,zT=(()=>{class t{static \u0275prov=NA({token:t,providedIn:"root",factory:()=>new h9})}return t})(),h9=class{queuedEffectCount=0;queues=new Map;schedule(e){this.enqueue(e)}remove(e){let A=e.zone,i=this.queues.get(A);i.has(e)&&(i.delete(e),this.queuedEffectCount--)}enqueue(e){let A=e.zone;this.queues.has(A)||this.queues.set(A,new Set);let i=this.queues.get(A);i.has(e)||(this.queuedEffectCount++,i.add(e))}flush(){for(;this.queuedEffectCount>0;)for(let[e,A]of this.queues)e===null?this.flushQueue(A):e.run(()=>this.flushQueue(A))}flushQueue(e){for(let A of e)e.delete(A),this.queuedEffectCount--,A.run()}};function D2(t){return!!t&&typeof t.then=="function"}function Kb(t){return!!t&&typeof t.subscribe=="function"}var OT=new dA("");function Yb(t){return ph([{provide:OT,multi:!0,useValue:t}])}var PT=(()=>{class t{resolve;reject;initialized=!1;done=!1;donePromise=new Promise((A,i)=>{this.resolve=A,this.reject=i});appInits=f(OT,{optional:!0})??[];injector=f(Rt);constructor(){}runInitializers(){if(this.initialized)return;let A=[];for(let n of this.appInits){let o=ga(this.injector,n);if(D2(o))A.push(o);else if(Kb(o)){let r=new Promise((s,a)=>{o.subscribe({complete:s,error:a})});A.push(r)}}let i=()=>{this.done=!0,this.resolve()};Promise.all(A).then(()=>{i()}).catch(n=>{this.reject(n)}),A.length===0&&i(),this.initialized=!0}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),Jb=new dA("");function EdA(){W7(()=>{throw new Ae(600,!1)})}function QdA(t){return t.isBoundToModule}var hdA=10;function jT(t,e){return Array.isArray(e)?e.reduce(jT,t):rA(rA({},t),e)}var la=(()=>{class t{_runningTick=!1;_destroyed=!1;_destroyListeners=[];_views=[];internalErrorHandler=f(t1A);afterRenderManager=f(NJ);zonelessEnabled=f(cb);rootEffectScheduler=f(zT);dirtyFlags=0;tracingSnapshot=null;externalTestViews=new Set;afterTick=new OA;get allViews(){return[...this.externalTestViews.keys(),...this._views]}get destroyed(){return this._destroyed}componentTypes=[];components=[];isStable=f(B0).hasPendingTasks.pipe(je(A=>!A));constructor(){f(Zd,{optional:!0})}whenStable(){let A;return new Promise(i=>{A=this.isStable.subscribe({next:n=>{n&&i()}})}).finally(()=>{A.unsubscribe()})}_injector=f(pr);_rendererFactory=null;get injector(){return this._injector}bootstrap(A,i){return this.bootstrapImpl(A,i)}bootstrapImpl(A,i,n=Rt.NULL){Ro(10);let o=A instanceof DT;if(!this._injector.get(PT).done){let d="";throw new Ae(405,d)}let s;o?s=A:s=this._injector.get(Rp).resolveComponentFactory(A),this.componentTypes.push(s.componentType);let a=QdA(s)?void 0:this._injector.get(I0),c=i||s.selector,l=s.create(n,[],c,a),I=l.location.nativeElement,C=l.injector.get(Gb,null);return C?.registerApplication(I),l.onDestroy(()=>{this.detachView(l.hostView),Lm(this.components,l),C?.unregisterApplication(I)}),this._loadComponent(l),Ro(11,l),l}tick(){this.zonelessEnabled||(this.dirtyFlags|=1),this._tick()}_tick(){Ro(12),this.tracingSnapshot!==null?this.tracingSnapshot.run(Cb.CHANGE_DETECTION,this.tickImpl):this.tickImpl()}tickImpl=()=>{if(this._runningTick)throw new Ae(101,!1);let A=Ti(null);try{this._runningTick=!0,this.synchronize()}catch(i){this.internalErrorHandler(i)}finally{this._runningTick=!1,this.tracingSnapshot?.dispose(),this.tracingSnapshot=null,Ti(A),this.afterTick.next(),Ro(13)}};synchronize(){this._rendererFactory===null&&!this._injector.destroyed&&(this._rendererFactory=this._injector.get(ws,null,{optional:!0}));let A=0;for(;this.dirtyFlags!==0&&A++<hdA;)Ro(14),this.synchronizeOnce(),Ro(15)}synchronizeOnce(){if(this.dirtyFlags&16&&(this.dirtyFlags&=-17,this.rootEffectScheduler.flush()),this.dirtyFlags&7){let A=!!(this.dirtyFlags&1);this.dirtyFlags&=-8,this.dirtyFlags|=8;for(let{_lView:i,notifyErrorHandler:n}of this.allViews)udA(i,n,A,this.zonelessEnabled);if(this.dirtyFlags&=-5,this.syncDirtyFlagsWithViews(),this.dirtyFlags&23)return}else this._rendererFactory?.begin?.(),this._rendererFactory?.end?.();this.dirtyFlags&8&&(this.dirtyFlags&=-9,this.afterRenderManager.execute()),this.syncDirtyFlagsWithViews()}syncDirtyFlagsWithViews(){if(this.allViews.some(({_lView:A})=>Qp(A))){this.dirtyFlags|=2;return}else this.dirtyFlags&=-8}attachView(A){let i=A;this._views.push(i),i.attachToAppRef(this)}detachView(A){let i=A;Lm(this._views,i),i.detachFromAppRef()}_loadComponent(A){this.attachView(A.hostView),this.tick(),this.components.push(A),this._injector.get(Jb,[]).forEach(n=>n(A))}ngOnDestroy(){if(!this._destroyed)try{this._destroyListeners.forEach(A=>A()),this._views.slice().forEach(A=>A.destroy())}finally{this._destroyed=!0,this._views=[],this._destroyListeners=[]}}onDestroy(A){return this._destroyListeners.push(A),()=>Lm(this._destroyListeners,A)}destroy(){if(this._destroyed)throw new Ae(406,!1);let A=this._injector;A.destroy&&!A.destroyed&&A.destroy()}get viewCount(){return this._views.length}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function Lm(t,e){let A=t.indexOf(e);A>-1&&t.splice(A,1)}function udA(t,e,A,i){if(!A&&!Qp(t))return;ET(t,e,A&&!i?0:1)}function Ne(t,e,A,i){let n=ei(),o=f2();if(Pa(n,o,e)){let r=Yo(),s=Dh();uIA(s,n,t,e,A,i)}return Ne}function qT(t,e,A,i){return Pa(t,f2(),A)?e+EI(A)+i:qa}function fdA(t,e,A,i,n,o){let r=w2A(),s=HT(t,r,A,n);return eb(2),s?e+EI(A)+i+EI(n)+o:qa}function bm(t,e){return t<<17|e<<2}function vI(t){return t>>17&32767}function mdA(t){return(t&2)==2}function pdA(t,e){return t&131071|e<<17}function u9(t){return t|2}function Td(t){return(t&131068)>>2}function yv(t,e){return t&-131069|e<<2}function wdA(t){return(t&1)===1}function f9(t){return t|1}function DdA(t,e,A,i,n,o){let r=o?e.classBindings:e.styleBindings,s=vI(r),a=Td(r);t[i]=A;let c=!1,l;if(Array.isArray(A)){let I=A;l=I[1],(l===null||mh(I,l)>0)&&(c=!0)}else l=A;if(n)if(a!==0){let C=vI(t[s+1]);t[i+1]=bm(C,s),C!==0&&(t[C+1]=yv(t[C+1],i)),t[s+1]=pdA(t[s+1],i)}else t[i+1]=bm(s,0),s!==0&&(t[s+1]=yv(t[s+1],i)),s=i;else t[i+1]=bm(a,0),s===0?s=i:t[a+1]=yv(t[a+1],i),a=i;c&&(t[i+1]=u9(t[i+1])),aY(t,l,i,!0),aY(t,l,i,!1),ydA(e,l,t,i,o),r=bm(s,a),o?e.classBindings=r:e.styleBindings=r}function ydA(t,e,A,i,n){let o=n?t.residualClasses:t.residualStyles;o!=null&&typeof e=="string"&&mh(o,e)>=0&&(A[i+1]=f9(A[i+1]))}function aY(t,e,A,i){let n=t[A+1],o=e===null,r=i?vI(n):Td(n),s=!1;for(;r!==0&&(s===!1||o);){let a=t[r],c=t[r+1];vdA(a,e)&&(s=!0,t[r+1]=i?f9(c):u9(c)),r=i?vI(c):Td(c)}s&&(t[A+1]=i?u9(n):f9(n))}function vdA(t,e){return t===null||e==null||(Array.isArray(t)?t[1]:t)===e?!0:Array.isArray(t)&&typeof e=="string"?mh(t,e)>=0:!1}var nl={textEnd:0,key:0,keyEnd:0,value:0,valueEnd:0};function bdA(t){return t.substring(nl.key,nl.keyEnd)}function MdA(t){return kdA(t),VT(t,ZT(t,0,nl.textEnd))}function VT(t,e){let A=nl.textEnd;return A===e?-1:(e=nl.keyEnd=SdA(t,nl.key=e,A),ZT(t,e,A))}function kdA(t){nl.key=0,nl.keyEnd=0,nl.value=0,nl.valueEnd=0,nl.textEnd=t.length}function ZT(t,e,A){for(;e<A&&t.charCodeAt(e)<=32;)e++;return e}function SdA(t,e,A){for(;e<A&&t.charCodeAt(e)>32;)e++;return e}function yA(t,e,A){let i=ei(),n=f2();if(Pa(i,n,e)){let o=Yo(),r=Dh();vp(o,r,i,t,e,i[Wo],A,!1)}return yA}function m9(t,e,A,i,n){wb(e,t,A,n?"class":"style",i)}function uo(t,e,A){return XT(t,e,A,!1),uo}function ue(t,e){return XT(t,e,null,!0),ue}function fo(t){$T(_dA,WT,t,!0)}function WT(t,e){for(let A=MdA(e);A>=0;A=VT(e,A))Ip(t,bdA(e),!0)}function XT(t,e,A,i){let n=ei(),o=Yo(),r=eb(2);if(o.firstUpdatePass&&eH(o,t,r,i),e!==qa&&Pa(n,r,e)){let s=o.data[d0()];tH(o,s,n,n[Wo],t,n[r+1]=UdA(e,A),i,r)}}function $T(t,e,A,i){let n=Yo(),o=eb(2);n.firstUpdatePass&&eH(n,null,o,i);let r=ei();if(A!==qa&&Pa(r,o,A)){let s=n.data[d0()];if(iH(s,i)&&!AH(n,o)){let a=i?s.classesWithoutHost:s.stylesWithoutHost;a!==null&&(A=Sv(a,A||"")),m9(n,s,r,A,i)}else GdA(n,s,r,r[Wo],r[o+1],r[o+1]=NdA(t,e,A),i,o)}}function AH(t,e){return e>=t.expandoStartIndex}function eH(t,e,A,i){let n=t.data;if(n[A+1]===null){let o=n[d0()],r=AH(t,A);iH(o,i)&&e===null&&!r&&(e=!1),e=RdA(n,o,e,i),DdA(n,o,e,A,r,i)}}function RdA(t,e,A,i){let n=tb(t),o=i?e.residualClasses:e.residualStyles;if(n===null)(i?e.classBindings:e.styleBindings)===0&&(A=vv(null,t,e,A,i),A=hh(A,e.attrs,i),o=null);else{let r=e.directiveStylingLast;if(r===-1||t[r]!==n)if(A=vv(n,t,e,A,i),o===null){let a=xdA(t,e,i);a!==void 0&&Array.isArray(a)&&(a=vv(null,t,e,a[1],i),a=hh(a,e.attrs,i),LdA(t,e,i,a))}else o=FdA(t,e,i)}return o!==void 0&&(i?e.residualClasses=o:e.residualStyles=o),A}function xdA(t,e,A){let i=A?e.classBindings:e.styleBindings;if(Td(i)!==0)return t[vI(i)]}function LdA(t,e,A,i){let n=A?e.classBindings:e.styleBindings;t[vI(n)]=i}function FdA(t,e,A){let i,n=e.directiveEnd;for(let o=1+e.directiveStylingLast;o<n;o++){let r=t[o].hostAttrs;i=hh(i,r,A)}return hh(i,e.attrs,A)}function vv(t,e,A,i,n){let o=null,r=A.directiveEnd,s=A.directiveStylingLast;for(s===-1?s=A.directiveStart:s++;s<r&&(o=e[s],i=hh(i,o.hostAttrs,n),o!==t);)s++;return t!==null&&(A.directiveStylingLast=s),i}function hh(t,e,A){let i=A?1:2,n=-1;if(e!==null)for(let o=0;o<e.length;o++){let r=e[o];typeof r=="number"?n=r:n===i&&(Array.isArray(t)||(t=t===void 0?[]:["",t]),Ip(t,r,A?!0:e[++o]))}return t===void 0?null:t}function NdA(t,e,A){if(A==null||A==="")return Ts;let i=[],n=sl(A);if(Array.isArray(n))for(let o=0;o<n.length;o++)t(i,n[o],!0);else if(typeof n=="object")for(let o in n)n.hasOwnProperty(o)&&t(i,o,n[o]);else typeof n=="string"&&e(i,n);return i}function _dA(t,e,A){let i=String(e);i!==""&&!i.includes(" ")&&Ip(t,i,A)}function GdA(t,e,A,i,n,o,r,s){n===qa&&(n=Ts);let a=0,c=0,l=0<n.length?n[0]:null,I=0<o.length?o[0]:null;for(;l!==null||I!==null;){let C=a<n.length?n[a+1]:void 0,d=c<o.length?o[c+1]:void 0,B=null,E;l===I?(a+=2,c+=2,C!==d&&(B=I,E=d)):I===null||l!==null&&l<I?(a+=2,B=l):(c+=2,B=I,E=d),B!==null&&tH(t,e,A,i,B,E,r,s),l=a<n.length?n[a]:null,I=c<o.length?o[c]:null}}function tH(t,e,A,i,n,o,r,s){if(!(e.type&3))return;let a=t.data,c=a[s+1],l=wdA(c)?cY(a,e,A,n,Td(c),r):void 0;if(!np(l)){np(o)||mdA(c)&&(o=cY(a,null,A,n,s,r));let I=jY(d0(),A);GIA(i,r,I,n,o)}}function cY(t,e,A,i,n,o){let r=e===null,s;for(;n>0;){let a=t[n],c=Array.isArray(a),l=c?a[1]:a,I=l===null,C=A[n+1];C===qa&&(C=I?Ts:void 0);let d=I?Ev(C,i):l===i?C:void 0;if(c&&!np(d)&&(d=Ev(a,i)),np(d)&&(s=d,r))return s;let B=t[n+1];n=r?vI(B):Td(B)}if(e!==null){let a=o?e.residualClasses:e.residualStyles;a!=null&&(s=Ev(a,i))}return s}function np(t){return t!==void 0}function UdA(t,e){return t==null||t===""||(typeof e=="string"?t=t+e:typeof t=="object"&&(t=ra(sl(t)))),t}function iH(t,e){return(t.flags&(e?8:16))!==0}function nH(t,e,A){let i=ei(),n=qT(i,t,e,A);$T(Ip,WT,n,!0)}var p9=class{destroy(e){}updateValue(e,A){}swap(e,A){let i=Math.min(e,A),n=Math.max(e,A),o=this.detach(n);if(n-i>1){let r=this.detach(i);this.attach(i,o),this.attach(n,r)}else this.attach(i,o)}move(e,A){this.attach(A,this.detach(e))}};function bv(t,e,A,i,n){return t===A&&Object.is(e,i)?1:Object.is(n(t,e),n(A,i))?-1:0}function KdA(t,e,A){let i,n,o=0,r=t.length-1,s=void 0;if(Array.isArray(e)){let a=e.length-1;for(;o<=r&&o<=a;){let c=t.at(o),l=e[o],I=bv(o,c,o,l,A);if(I!==0){I<0&&t.updateValue(o,l),o++;continue}let C=t.at(r),d=e[a],B=bv(r,C,a,d,A);if(B!==0){B<0&&t.updateValue(r,d),r--,a--;continue}let E=A(o,c),h=A(r,C),u=A(o,l);if(Object.is(u,h)){let D=A(a,d);Object.is(D,E)?(t.swap(o,r),t.updateValue(r,d),a--,r--):t.move(r,o),t.updateValue(o,l),o++;continue}if(i??=new op,n??=gY(t,o,r,A),w9(t,i,o,u))t.updateValue(o,l),o++,r++;else if(n.has(u))i.set(E,t.detach(o)),r--;else{let D=t.create(o,e[o]);t.attach(o,D),o++,r++}}for(;o<=a;)lY(t,i,A,o,e[o]),o++}else if(e!=null){let a=e[Symbol.iterator](),c=a.next();for(;!c.done&&o<=r;){let l=t.at(o),I=c.value,C=bv(o,l,o,I,A);if(C!==0)C<0&&t.updateValue(o,I),o++,c=a.next();else{i??=new op,n??=gY(t,o,r,A);let d=A(o,I);if(w9(t,i,o,d))t.updateValue(o,I),o++,r++,c=a.next();else if(!n.has(d))t.attach(o,t.create(o,I)),o++,r++,c=a.next();else{let B=A(o,l);i.set(B,t.detach(o)),r--}}}for(;!c.done;)lY(t,i,A,t.length,c.value),c=a.next()}for(;o<=r;)t.destroy(t.detach(r--));i?.forEach(a=>{t.destroy(a)})}function w9(t,e,A,i){return e!==void 0&&e.has(i)?(t.attach(A,e.get(i)),e.delete(i),!0):!1}function lY(t,e,A,i,n){if(w9(t,e,i,A(i,n)))t.updateValue(i,n);else{let o=t.create(i,n);t.attach(i,o)}}function gY(t,e,A,i){let n=new Set;for(let o=e;o<=A;o++)n.add(i(o,t.at(o)));return n}var op=class{kvMap=new Map;_vMap=void 0;has(e){return this.kvMap.has(e)}delete(e){if(!this.has(e))return!1;let A=this.kvMap.get(e);return this._vMap!==void 0&&this._vMap.has(A)?(this.kvMap.set(e,this._vMap.get(A)),this._vMap.delete(A)):this.kvMap.delete(e),!0}get(e){return this.kvMap.get(e)}set(e,A){if(this.kvMap.has(e)){let i=this.kvMap.get(e);this._vMap===void 0&&(this._vMap=new Map);let n=this._vMap;for(;n.has(i);)i=n.get(i);n.set(i,A)}else this.kvMap.set(e,A)}forEach(e){for(let[A,i]of this.kvMap)if(e(i,A),this._vMap!==void 0){let n=this._vMap;for(;n.has(i);)i=n.get(i),e(i,A)}}};function GA(t,e){E0("NgControlFlow");let A=ei(),i=f2(),n=A[i]!==qa?A[i]:-1,o=n!==-1?rp(A,Nr+n):void 0,r=0;if(Pa(A,i,t)){let s=Ti(null);try{if(o!==void 0&&pT(o,r),t!==-1){let a=Nr+t,c=rp(A,a),l=b9(A[ki],a),I=Jd(c,l.tView.ssrId),C=Mh(A,l,e,{dehydratedView:I});kh(c,C,r,Yd(l,I))}}finally{Ti(s)}}else if(o!==void 0){let s=mT(o,r);s!==void 0&&(s[Fr]=e)}}var D9=class{lContainer;$implicit;$index;constructor(e,A,i){this.lContainer=e,this.$implicit=A,this.$index=i}get $count(){return this.lContainer.length-ms}};function Xd(t){return t}function to(t,e){return e}var y9=class{hasEmptyBlock;trackByFn;liveCollection;constructor(e,A,i){this.hasEmptyBlock=e,this.trackByFn=A,this.liveCollection=i}};function Dn(t,e,A,i,n,o,r,s,a,c,l,I,C){E0("NgControlFlow");let d=ei(),B=Yo(),E=a!==void 0,h=ei(),u=s?r.bind(h[Oa][Fr]):r,D=new y9(E,u);h[Nr+t]=D,ip(d,B,t+1,e,A,i,n,u2(B.consts,o)),E&&ip(d,B,t+2,a,c,l,I,u2(B.consts,C))}var v9=class extends p9{lContainer;hostLView;templateTNode;operationsCounter=void 0;needsIndexUpdate=!1;constructor(e,A,i){super(),this.lContainer=e,this.hostLView=A,this.templateTNode=i}get length(){return this.lContainer.length-ms}at(e){return this.getLView(e)[Fr].$implicit}attach(e,A){let i=A[_d];this.needsIndexUpdate||=e!==this.length,kh(this.lContainer,A,e,Yd(this.templateTNode,i))}detach(e){return this.needsIndexUpdate||=e!==this.length-1,YdA(this.lContainer,e)}create(e,A){let i=Jd(this.lContainer,this.templateTNode.tView.ssrId),n=Mh(this.hostLView,this.templateTNode,new D9(this.lContainer,A,e),{dehydratedView:i});return this.operationsCounter?.recordCreate(),n}destroy(e){bp(e[ki],e),this.operationsCounter?.recordDestroy()}updateValue(e,A){this.getLView(e)[Fr].$implicit=A}reset(){this.needsIndexUpdate=!1,this.operationsCounter?.reset()}updateIndexes(){if(this.needsIndexUpdate)for(let e=0;e<this.length;e++)this.getLView(e)[Fr].$index=e}getLView(e){return JdA(this.lContainer,e)}};function yn(t){let e=Ti(null),A=d0();try{let i=ei(),n=i[ki],o=i[A],r=A+1,s=rp(i,r);if(o.liveCollection===void 0){let c=b9(n,r);o.liveCollection=new v9(s,i,c)}else o.liveCollection.reset();let a=o.liveCollection;if(KdA(a,t,o.trackByFn),a.updateIndexes(),o.hasEmptyBlock){let c=f2(),l=a.length===0;if(Pa(i,c,l)){let I=A+2,C=rp(i,I);if(l){let d=b9(n,I),B=Jd(C,d.tView.ssrId),E=Mh(i,d,void 0,{dehydratedView:B});kh(C,E,0,Yd(d,B))}else pT(C,0)}}}finally{Ti(e)}}function rp(t,e){return t[e]}function YdA(t,e){return Eh(t,e)}function JdA(t,e){return mT(t,e)}function b9(t,e){return j9(t,e)}function S(t,e,A,i){let n=ei(),o=Yo(),r=Nr+t,s=n[Wo],a=o.firstCreatePass?bT(r,o,n,e,pb,X9(),A,i):o.data[r],c=TdA(o,n,a,s,e,t);n[r]=c;let l=Ep(a);return kI(a,!0),eT(s,c,a),!vb(a)&&up()&&Mp(o,n,c,a),(E2A()===0||l)&&qd(c,n),Q2A(),l&&(yp(o,n,a),Bb(o,a,n)),i!==null&&mb(n,a),S}function F(){let t=os();$9()?Ab():(t=t.parent,kI(t,!1));let e=t;u2A(e)&&f2A(),h2A();let A=Yo();return A.firstCreatePass&&MT(A,e),e.classesWithoutHost!=null&&F2A(e)&&m9(A,e,ei(),e.classesWithoutHost,!0),e.stylesWithoutHost!=null&&N2A(e)&&m9(A,e,ei(),e.stylesWithoutHost,!1),F}function JA(t,e,A,i){return S(t,e,A,i),F(),JA}var TdA=(t,e,A,i,n,o)=>(fp(!0),$J(i,n,R2A()));function HdA(t,e,A,i,n){let o=e.consts,r=u2(o,i),s=Sh(e,t,8,"ng-container",r);r!==null&&c9(s,r,!0);let a=u2(o,n);return X9()&&Sb(e,A,s,a,pb),s.mergedAttrs=Kd(s.mergedAttrs,s.attrs),e.queries!==null&&e.queries.elementStart(e,s),s}function y2(t,e,A){let i=ei(),n=Yo(),o=t+Nr,r=n.firstCreatePass?HdA(o,n,i,e,A):n.data[o];kI(r,!0);let s=zdA(n,i,r,t);return i[o]=s,up()&&Mp(n,i,s,r),qd(s,i),Ep(r)&&(yp(n,i,r),Bb(n,r,i)),A!=null&&mb(i,r),y2}function v2(){let t=os(),e=Yo();return $9()?Ab():(t=t.parent,kI(t,!1)),e.firstCreatePass&&(rb(e,t),P9(t)&&e.queries.elementEnd(t)),v2}function _r(t,e,A){return y2(t,e,A),v2(),_r}var zdA=(t,e,A,i)=>(fp(!0),nIA(e[Wo],""));function be(){return ei()}function Hs(t,e,A){let i=ei(),n=f2();if(Pa(i,n,e)){let o=Yo(),r=Dh();vp(o,r,i,t,e,i[Wo],A,!0)}return Hs}function Tb(t,e,A){let i=ei(),n=f2();if(Pa(i,n,e)){let o=Yo(),r=Dh(),s=tb(o.data),a=rT(s,r,i);vp(o,r,i,t,e,a,A,!0)}return Tb}var sp="en-US";var OdA=sp;function PdA(t){typeof t=="string"&&(OdA=t.toLowerCase().replace(/_/g,"-"))}function IY(t,e,A){return function i(n){if(n===Function)return A;let o=zd(t)?Ag(t.index,e):e;kb(o,5);let r=e[Fr],s=CY(e,r,A,n),a=i.__ngNextListenerFn__;for(;a;)s=CY(e,r,a,n)&&s,a=a.__ngNextListenerFn__;return s}}function CY(t,e,A,i){let n=Ti(null);try{return Ro(6,e,A),A(i)!==!1}catch(o){return jdA(t,o),!1}finally{Ro(7,e,A),Ti(n)}}function jdA(t,e){let A=t[Gd],i=A?A.get(aa,null):null;i&&i.handleError(e)}function dY(t,e,A,i,n,o){let r=e[A],s=e[ki],c=s.data[A].outputs[i],l=r[c],I=s.firstCreatePass?W9(s):null,C=Z9(e),d=l.subscribe(o),B=C.length;C.push(o,d),I&&I.push(n,t.index,B,-(B+1))}function hA(t,e,A,i){let n=ei(),o=Yo(),r=os();return zb(o,n,n[Wo],r,t,e,i),hA}function Hb(t,e){let A=os(),i=ei(),n=Yo(),o=tb(n.data),r=rT(o,A,i);return zb(n,i,r,A,t,e),Hb}function qdA(t,e,A,i){let n=t.cleanup;if(n!=null)for(let o=0;o<n.length-1;o+=2){let r=n[o];if(r===A&&n[o+1]===i){let s=e[Km],a=n[o+2];return s.length>a?s[a]:null}typeof r=="string"&&(o+=2)}return null}function zb(t,e,A,i,n,o,r){let s=Ep(i),c=t.firstCreatePass?W9(t):null,l=Z9(e),I=!0;if(i.type&3||r){let C=ng(i,e),d=r?r(C):C,B=l.length,E=r?u=>r($l(u[i.index])):i.index,h=null;if(!r&&s&&(h=qdA(t,e,n,i.index)),h!==null){let u=h.__ngLastListenerFn__||h;u.__ngNextListenerFn__=o,h.__ngLastListenerFn__=o,I=!1}else{o=IY(i,e,o),u1A(e,d,n,o);let u=A.listen(d,n,o);l.push(o,u),c&&c.push(n,E,B,B+1)}}else o=IY(i,e,o);if(I){let C=i.outputs?.[n],d=i.hostDirectiveOutputs?.[n];if(d&&d.length)for(let B=0;B<d.length;B+=2){let E=d[B],h=d[B+1];dY(i,e,E,h,n,o)}if(C&&C.length)for(let B of C)dY(i,e,B,n,n,o)}}function O(t=1){return k2A(t)}function VdA(t,e){let A=null,i=V1A(t);for(let n=0;n<e.length;n++){let o=e[n];if(o==="*"){A=n;continue}if(i===null?XJ(t,o,!0):X1A(i,o))return n}return A}function jt(t){let e=ei()[Oa][Ia];if(!e.projection){let A=t?t.length:1,i=e.projection=H0A(A,null),n=i.slice(),o=e.child;for(;o!==null;){if(o.type!==128){let r=t?VdA(o,t):0;r!==null&&(n[r]?n[r].projectionNext=o:i[r]=o,n[r]=o)}o=o.next}}}function Fe(t,e=0,A,i,n,o){let r=ei(),s=Yo(),a=i?t+1:null;a!==null&&ip(r,s,a,i,n,o,null,A);let c=Sh(s,Nr+t,16,null,A||null);c.projection===null&&(c.projection=e),Ab();let I=!r[_d]||WY();r[Oa][Ia].projection[c.projection]===null&&a!==null?ZdA(r,s,a):I&&!vb(c)&&NIA(s,r,c)}function ZdA(t,e,A){let i=Nr+A,n=e.data[i],o=t[i],r=Jd(o,n.tView.ssrId),s=Mh(t,n,void 0,{dehydratedView:r});kh(o,s,0,Yd(n,r))}function ci(t,e,A,i){LT(t,e,A,i)}function Te(t,e,A){_CA(t,e,A)}function XA(t){let e=ei(),A=Yo(),i=ib();hp(i+1);let n=xb(A,i);if(t.dirty&&g2A(e)===((n.metadata.flags&2)===2)){if(n.matches===null)t.reset([]);else{let o=NT(e,i);t.reset(o,DJ),t.notifyOnChanges()}return!0}return!1}function $A(){return Rb(ei(),ib())}function oH(t,e,A,i,n){JCA(e,LT(t,A,i,n))}function rH(t=1){hp(ib()+t)}function cr(t){let e=p2A();return q9(e,Nr+t)}function AA(t,e=""){let A=ei(),i=Yo(),n=t+Nr,o=i.firstCreatePass?Sh(i,n,1,e,null):i.data[n],r=WdA(i,A,o,e,t);A[n]=r,up()&&Mp(i,A,r,o),kI(o,!1)}var WdA=(t,e,A,i,n)=>(fp(!0),tIA(e[Wo],i));function Gt(t){return Et("",t,""),Gt}function Et(t,e,A){let i=ei(),n=qT(i,t,e,A);return n!==qa&&sH(i,d0(),n),Et}function Ob(t,e,A,i,n){let o=ei(),r=fdA(o,t,e,A,i,n);return r!==qa&&sH(o,d0(),r),Ob}function sH(t,e,A){let i=jY(e,t);iIA(t[Wo],i,A)}function Ca(t,e,A){yJ(e)&&(e=e());let i=ei(),n=f2();if(Pa(i,n,e)){let o=Yo(),r=Dh();vp(o,r,i,t,e,i[Wo],A,!1)}return Ca}function Va(t,e){let A=yJ(t);return A&&t.set(e),A}function da(t,e){let A=ei(),i=Yo(),n=os();return zb(i,A,A[Wo],n,t,e),da}function XdA(t,e,A){let i=Yo();if(i.firstCreatePass){let n=rl(t);M9(A,i.data,i.blueprint,n,!0),M9(e,i.data,i.blueprint,n,!1)}}function M9(t,e,A,i,n){if(t=ns(t),Array.isArray(t))for(let o=0;o<t.length;o++)M9(t[o],e,A,i,n);else{let o=Yo(),r=ei(),s=os(),a=Nd(t)?t:ns(t.provide),c=YY(t),l=s.providerIndexes&1048575,I=s.directiveStart,C=s.providerIndexes>>20;if(Nd(t)||!t.multi){let d=new wI(c,n,PA),B=kv(a,e,n?l:l+C,I);B===-1?(Tv(Pm(s,r),o,a),Mv(o,t,e.length),e.push(a),s.directiveStart++,s.directiveEnd++,n&&(s.providerIndexes+=1048576),A.push(d),r.push(d)):(A[B]=d,r[B]=d)}else{let d=kv(a,e,l+C,I),B=kv(a,e,l,l+C),E=d>=0&&A[d],h=B>=0&&A[B];if(n&&!h||!n&&!E){Tv(Pm(s,r),o,a);let u=eBA(n?ABA:$dA,A.length,n,i,c);!n&&h&&(A[B].providerFactory=u),Mv(o,t,e.length,0),e.push(a),s.directiveStart++,s.directiveEnd++,n&&(s.providerIndexes+=1048576),A.push(u),r.push(u)}else{let u=aH(A[n?B:d],c,!n&&i);Mv(o,t,d>-1?d:B,u)}!n&&i&&h&&A[B].componentProviders++}}}function Mv(t,e,A,i){let n=Nd(e),o=W0A(e);if(n||o){let a=(o?ns(e.useClass):e).prototype.ngOnDestroy;if(a){let c=t.destroyHooks||(t.destroyHooks=[]);if(!n&&e.multi){let l=c.indexOf(A);l===-1?c.push(A,[i,a]):c[l+1].push(i,a)}else c.push(A,a)}}}function aH(t,e,A){return A&&t.componentProviders++,t.multi.push(e)-1}function kv(t,e,A,i){for(let n=A;n<i;n++)if(e[n]===t)return n;return-1}function $dA(t,e,A,i,n){return k9(this.multi,[])}function ABA(t,e,A,i,n){let o=this.multi,r;if(this.providerFactory){let s=this.providerFactory.componentProviders,a=dh(i,i[ki],this.providerFactory.index,n);r=a.slice(0,s),k9(o,r);for(let c=s;c<a.length;c++)r.push(a[c])}else r=[],k9(o,r);return r}function k9(t,e){for(let A=0;A<t.length;A++){let i=t[A];e.push(i())}return e}function eBA(t,e,A,i,n){let o=new wI(t,A,PA);return o.multi=[],o.index=e,o.componentProviders=0,aH(o,n,i&&!A),o}function ht(t,e=[]){return A=>{A.providersResolver=(i,n)=>XdA(i,n?n(t):t,e)}}function cH(t,e,A){let i=wh()+t,n=ei();return n[i]===qa?Nb(n,i,A?e.call(A):e()):gdA(n,i)}function qr(t,e,A,i){return gH(ei(),wh(),t,e,A,i)}function b2(t,e,A,i,n){return IH(ei(),wh(),t,e,A,i,n)}function lH(t,e){let A=t[e];return A===qa?void 0:A}function gH(t,e,A,i,n,o){let r=e+A;return Pa(t,r,n)?Nb(t,r+1,o?i.call(o,n):i(n)):lH(t,r+1)}function IH(t,e,A,i,n,o,r){let s=e+A;return HT(t,s,n,o)?Nb(t,s+2,r?i.call(r,n,o):i(n,o)):lH(t,s+2)}function Za(t,e){let A=Yo(),i,n=t+Nr;A.firstCreatePass?(i=tBA(e,A.pipeRegistry),A.data[n]=i,i.onDestroy&&(A.destroyHooks??=[]).push(n,i.onDestroy)):i=A.data[n];let o=i.factory||(i.factory=QI(i.type,!0)),r,s=oa(PA);try{let a=Om(!1),c=o();return Om(a),l2A(A,ei(),n,c),c}finally{oa(s)}}function tBA(t,e){if(e)for(let A=e.length-1;A>=0;A--){let i=e[A];if(t===i.name)return i}}function M2(t,e,A){let i=t+Nr,n=ei(),o=q9(n,i);return CH(n,i)?gH(n,wh(),e,o.transform,A,o):o.transform(A)}function Lh(t,e,A,i){let n=t+Nr,o=ei(),r=q9(o,n);return CH(o,n)?IH(o,wh(),e,r.transform,A,i,r):r.transform(A,i)}function CH(t,e){return t[ki].data[e].pure}function Fh(t,e){return Sp(t,e)}var Mm=null;function iBA(t){Mm!==null&&(t.defaultEncapsulation!==Mm.defaultEncapsulation||t.preserveWhitespaces!==Mm.preserveWhitespaces)||(Mm=t)}var bI=class{full;major;minor;patch;constructor(e){this.full=e;let A=e.split(".");this.major=A[0],this.minor=A[1],this.patch=A.slice(2).join(".")}},Pb=new bI("19.2.14"),S9=class{ngModuleFactory;componentFactories;constructor(e,A){this.ngModuleFactory=e,this.componentFactories=A}},dH=(()=>{class t{compileModuleSync(A){return new tp(A)}compileModuleAsync(A){return Promise.resolve(this.compileModuleSync(A))}compileModuleAndAllComponentsSync(A){let i=this.compileModuleSync(A),n=NY(A),o=ZJ(n.declarations).reduce((r,s)=>{let a=h2(s);return a&&r.push(new yI(a)),r},[]);return new S9(i,o)}compileModuleAndAllComponentsAsync(A){return Promise.resolve(this.compileModuleAndAllComponentsSync(A))}clearCache(){}clearCacheFor(A){}getModuleId(A){}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),nBA=new dA("");function oBA(t,e,A){let i=new tp(A);return Promise.resolve(i)}function BY(t){for(let e=t.length-1;e>=0;e--)if(t[e]!==void 0)return t[e]}var rBA=(()=>{class t{zone=f(Qe);changeDetectionScheduler=f(DI);applicationRef=f(la);_onMicrotaskEmptySubscription;initialize(){this._onMicrotaskEmptySubscription||(this._onMicrotaskEmptySubscription=this.zone.onMicrotaskEmpty.subscribe({next:()=>{this.changeDetectionScheduler.runningTick||this.zone.run(()=>{this.applicationRef.tick()})}}))}ngOnDestroy(){this._onMicrotaskEmptySubscription?.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function sBA({ngZoneFactory:t,ignoreChangesOutsideZone:e,scheduleInRootZone:A}){return t??=()=>new Qe(Ye(rA({},BH()),{scheduleInRootZone:A})),[{provide:Qe,useFactory:t},{provide:Fd,multi:!0,useFactory:()=>{let i=f(rBA,{optional:!0});return()=>i.initialize()}},{provide:Fd,multi:!0,useFactory:()=>{let i=f(aBA);return()=>{i.initialize()}}},e===!0?{provide:uJ,useValue:!0}:[],{provide:fJ,useValue:A??hJ}]}function BH(t){return{enableLongStackTrace:!1,shouldCoalesceEventChangeDetection:t?.eventCoalescing??!1,shouldCoalesceRunChangeDetection:t?.runCoalescing??!1}}var aBA=(()=>{class t{subscription=new Kt;initialized=!1;zone=f(Qe);pendingTasks=f(B0);initialize(){if(this.initialized)return;this.initialized=!0;let A=null;!this.zone.isStable&&!this.zone.hasPendingMacrotasks&&!this.zone.hasPendingMicrotasks&&(A=this.pendingTasks.add()),this.zone.runOutsideAngular(()=>{this.subscription.add(this.zone.onStable.subscribe(()=>{Qe.assertNotInAngularZone(),queueMicrotask(()=>{A!==null&&!this.zone.hasPendingMacrotasks&&!this.zone.hasPendingMicrotasks&&(this.pendingTasks.remove(A),A=null)})}))}),this.subscription.add(this.zone.onUnstable.subscribe(()=>{Qe.assertInAngularZone(),A??=this.pendingTasks.add()}))}ngOnDestroy(){this.subscription.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var cBA=(()=>{class t{appRef=f(la);taskService=f(B0);ngZone=f(Qe);zonelessEnabled=f(cb);tracing=f(Zd,{optional:!0});disableScheduling=f(uJ,{optional:!0})??!1;zoneIsDefined=typeof Zone<"u"&&!!Zone.root.run;schedulerTickApplyArgs=[{data:{__scheduler_tick__:!0}}];subscriptions=new Kt;angularZoneId=this.zoneIsDefined?this.ngZone._inner?.get(qm):null;scheduleInRootZone=!this.zonelessEnabled&&this.zoneIsDefined&&(f(fJ,{optional:!0})??!1);cancelScheduledCallback=null;useMicrotaskScheduler=!1;runningTick=!1;pendingRenderTaskId=null;constructor(){this.subscriptions.add(this.appRef.afterTick.subscribe(()=>{this.runningTick||this.cleanup()})),this.subscriptions.add(this.ngZone.onUnstable.subscribe(()=>{this.runningTick||this.cleanup()})),this.disableScheduling||=!this.zonelessEnabled&&(this.ngZone instanceof Vm||!this.zoneIsDefined)}notify(A){if(!this.zonelessEnabled&&A===5)return;let i=!1;switch(A){case 0:{this.appRef.dirtyFlags|=2;break}case 3:case 2:case 4:case 5:case 1:{this.appRef.dirtyFlags|=4;break}case 6:{this.appRef.dirtyFlags|=2,i=!0;break}case 12:{this.appRef.dirtyFlags|=16,i=!0;break}case 13:{this.appRef.dirtyFlags|=2,i=!0;break}case 11:{i=!0;break}case 9:case 8:case 7:case 10:default:this.appRef.dirtyFlags|=8}if(this.appRef.tracingSnapshot=this.tracing?.snapshot(this.appRef.tracingSnapshot)??null,!this.shouldScheduleTick(i))return;let n=this.useMicrotaskScheduler?YK:mJ;this.pendingRenderTaskId=this.taskService.add(),this.scheduleInRootZone?this.cancelScheduledCallback=Zone.root.run(()=>n(()=>this.tick())):this.cancelScheduledCallback=this.ngZone.runOutsideAngular(()=>n(()=>this.tick()))}shouldScheduleTick(A){return!(this.disableScheduling&&!A||this.appRef.destroyed||this.pendingRenderTaskId!==null||this.runningTick||this.appRef._runningTick||!this.zonelessEnabled&&this.zoneIsDefined&&Zone.current.get(qm+this.angularZoneId))}tick(){if(this.runningTick||this.appRef.destroyed)return;if(this.appRef.dirtyFlags===0){this.cleanup();return}!this.zonelessEnabled&&this.appRef.dirtyFlags&7&&(this.appRef.dirtyFlags|=1);let A=this.taskService.add();try{this.ngZone.run(()=>{this.runningTick=!0,this.appRef._tick()},void 0,this.schedulerTickApplyArgs)}catch(i){throw this.taskService.remove(A),i}finally{this.cleanup()}this.useMicrotaskScheduler=!0,YK(()=>{this.useMicrotaskScheduler=!1,this.taskService.remove(A)})}ngOnDestroy(){this.subscriptions.unsubscribe(),this.cleanup()}cleanup(){if(this.runningTick=!1,this.cancelScheduledCallback?.(),this.cancelScheduledCallback=null,this.pendingRenderTaskId!==null){let A=this.pendingRenderTaskId;this.pendingRenderTaskId=null,this.taskService.remove(A)}}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function lBA(){return typeof $localize<"u"&&$localize.locale||sp}var Np=new dA("",{providedIn:"root",factory:()=>f(Np,Gi.Optional|Gi.SkipSelf)||lBA()});var ap=new dA(""),gBA=new dA("");function sh(t){return!t.moduleRef}function IBA(t){let e=sh(t)?t.r3Injector:t.moduleRef.injector,A=e.get(Qe);return A.run(()=>{sh(t)?t.r3Injector.resolveInjectorInitializers():t.moduleRef.resolveInjectorInitializers();let i=e.get(aa,null),n;if(A.runOutsideAngular(()=>{n=A.onError.subscribe({next:o=>{i.handleError(o)}})}),sh(t)){let o=()=>e.destroy(),r=t.platformInjector.get(ap);r.add(o),e.onDestroy(()=>{n.unsubscribe(),r.delete(o)})}else{let o=()=>t.moduleRef.destroy(),r=t.platformInjector.get(ap);r.add(o),t.moduleRef.onDestroy(()=>{Lm(t.allPlatformModules,t.moduleRef),n.unsubscribe(),r.delete(o)})}return dBA(i,A,()=>{let o=e.get(PT);return o.runInitializers(),o.donePromise.then(()=>{let r=e.get(Np,sp);if(PdA(r||sp),!e.get(gBA,!0))return sh(t)?e.get(la):(t.allPlatformModules.push(t.moduleRef),t.moduleRef);if(sh(t)){let a=e.get(la);return t.rootComponent!==void 0&&a.bootstrap(t.rootComponent),a}else return CBA(t.moduleRef,t.allPlatformModules),t.moduleRef})})})}function CBA(t,e){let A=t.injector.get(la);if(t._bootstrapComponents.length>0)t._bootstrapComponents.forEach(i=>A.bootstrap(i));else if(t.instance.ngDoBootstrap)t.instance.ngDoBootstrap(A);else throw new Ae(-403,!1);e.push(t)}function dBA(t,e,A){try{let i=A();return D2(i)?i.catch(n=>{throw e.runOutsideAngular(()=>t.handleError(n)),n}):i}catch(i){throw e.runOutsideAngular(()=>t.handleError(i)),i}}var EH=(()=>{class t{_injector;_modules=[];_destroyListeners=[];_destroyed=!1;constructor(A){this._injector=A}bootstrapModuleFactory(A,i){let n=i?.scheduleInRootZone,o=()=>e1A(i?.ngZone,Ye(rA({},BH({eventCoalescing:i?.ngZoneEventCoalescing,runCoalescing:i?.ngZoneRunCoalescing})),{scheduleInRootZone:n})),r=i?.ignoreChangesOutsideZone,s=[sBA({ngZoneFactory:o,ignoreChangesOutsideZone:r}),{provide:DI,useExisting:cBA}],a=ZCA(A.moduleType,this.injector,s);return IBA({moduleRef:a,allPlatformModules:this._modules,platformInjector:this.injector})}bootstrapModule(A,i=[]){let n=jT({},i);return oBA(this.injector,n,A).then(o=>this.bootstrapModuleFactory(o,n))}onDestroy(A){this._destroyListeners.push(A)}get injector(){return this._injector}destroy(){if(this._destroyed)throw new Ae(404,!1);this._modules.slice().forEach(i=>i.destroy()),this._destroyListeners.forEach(i=>i());let A=this._injector.get(ap,null);A&&(A.forEach(i=>i()),A.clear()),this._destroyed=!0}get destroyed(){return this._destroyed}static \u0275fac=function(i){return new(i||t)(we(Rt))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"platform"})}return t})(),gh=null,QH=new dA("");function BBA(t){if(gh&&!gh.get(QH,!1))throw new Ae(400,!1);EdA(),gh=t;let e=t.get(EH);return hBA(t),e}function jb(t,e,A=[]){let i=`Platform: ${e}`,n=new dA(i);return(o=[])=>{let r=hH();if(!r||r.injector.get(QH,!1)){let s=[...A,...o,{provide:n,useValue:!0}];t?t(s):BBA(EBA(s,i))}return QBA(n)}}function EBA(t=[],e){return Rt.create({name:e,providers:[{provide:Cp,useValue:"platform"},{provide:ap,useValue:new Set([()=>gh=null])},...t]})}function QBA(t){let e=hH();if(!e)throw new Ae(401,!1);return e}function hH(){return gh?.get(EH)??null}function hBA(t){let e=t.get(Ib,null);ga(t,()=>{e?.forEach(A=>A())})}var It=(()=>{class t{static __NG_ELEMENT_ID__=uBA}return t})();function uBA(t){return fBA(os(),ei(),(t&16)===16)}function fBA(t,e,A){if(zd(t)&&!A){let i=Ag(t.index,e);return new Qh(i,i)}else if(t.type&175){let i=e[Oa];return new Qh(i,e)}return null}var R9=class{constructor(){}supports(e){return TT(e)}create(e){return new x9(e)}},mBA=(t,e)=>e,x9=class{length=0;collection;_linkedRecords=null;_unlinkedRecords=null;_previousItHead=null;_itHead=null;_itTail=null;_additionsHead=null;_additionsTail=null;_movesHead=null;_movesTail=null;_removalsHead=null;_removalsTail=null;_identityChangesHead=null;_identityChangesTail=null;_trackByFn;constructor(e){this._trackByFn=e||mBA}forEachItem(e){let A;for(A=this._itHead;A!==null;A=A._next)e(A)}forEachOperation(e){let A=this._itHead,i=this._removalsHead,n=0,o=null;for(;A||i;){let r=!i||A&&A.currentIndex<EY(i,n,o)?A:i,s=EY(r,n,o),a=r.currentIndex;if(r===i)n--,i=i._nextRemoved;else if(A=A._next,r.previousIndex==null)n++;else{o||(o=[]);let c=s-n,l=a-n;if(c!=l){for(let C=0;C<c;C++){let d=C<o.length?o[C]:o[C]=0,B=d+C;l<=B&&B<c&&(o[C]=d+1)}let I=r.previousIndex;o[I]=l-c}}s!==a&&e(r,s,a)}}forEachPreviousItem(e){let A;for(A=this._previousItHead;A!==null;A=A._nextPrevious)e(A)}forEachAddedItem(e){let A;for(A=this._additionsHead;A!==null;A=A._nextAdded)e(A)}forEachMovedItem(e){let A;for(A=this._movesHead;A!==null;A=A._nextMoved)e(A)}forEachRemovedItem(e){let A;for(A=this._removalsHead;A!==null;A=A._nextRemoved)e(A)}forEachIdentityChange(e){let A;for(A=this._identityChangesHead;A!==null;A=A._nextIdentityChange)e(A)}diff(e){if(e==null&&(e=[]),!TT(e))throw new Ae(900,!1);return this.check(e)?this:null}onDestroy(){}check(e){this._reset();let A=this._itHead,i=!1,n,o,r;if(Array.isArray(e)){this.length=e.length;for(let s=0;s<this.length;s++)o=e[s],r=this._trackByFn(s,o),A===null||!Object.is(A.trackById,r)?(A=this._mismatch(A,o,r,s),i=!0):(i&&(A=this._verifyReinsertion(A,o,r,s)),Object.is(A.item,o)||this._addIdentityChange(A,o)),A=A._next}else n=0,ldA(e,s=>{r=this._trackByFn(n,s),A===null||!Object.is(A.trackById,r)?(A=this._mismatch(A,s,r,n),i=!0):(i&&(A=this._verifyReinsertion(A,s,r,n)),Object.is(A.item,s)||this._addIdentityChange(A,s)),A=A._next,n++}),this.length=n;return this._truncate(A),this.collection=e,this.isDirty}get isDirty(){return this._additionsHead!==null||this._movesHead!==null||this._removalsHead!==null||this._identityChangesHead!==null}_reset(){if(this.isDirty){let e;for(e=this._previousItHead=this._itHead;e!==null;e=e._next)e._nextPrevious=e._next;for(e=this._additionsHead;e!==null;e=e._nextAdded)e.previousIndex=e.currentIndex;for(this._additionsHead=this._additionsTail=null,e=this._movesHead;e!==null;e=e._nextMoved)e.previousIndex=e.currentIndex;this._movesHead=this._movesTail=null,this._removalsHead=this._removalsTail=null,this._identityChangesHead=this._identityChangesTail=null}}_mismatch(e,A,i,n){let o;return e===null?o=this._itTail:(o=e._prev,this._remove(e)),e=this._unlinkedRecords===null?null:this._unlinkedRecords.get(i,null),e!==null?(Object.is(e.item,A)||this._addIdentityChange(e,A),this._reinsertAfter(e,o,n)):(e=this._linkedRecords===null?null:this._linkedRecords.get(i,n),e!==null?(Object.is(e.item,A)||this._addIdentityChange(e,A),this._moveAfter(e,o,n)):e=this._addAfter(new L9(A,i),o,n)),e}_verifyReinsertion(e,A,i,n){let o=this._unlinkedRecords===null?null:this._unlinkedRecords.get(i,null);return o!==null?e=this._reinsertAfter(o,e._prev,n):e.currentIndex!=n&&(e.currentIndex=n,this._addToMoves(e,n)),e}_truncate(e){for(;e!==null;){let A=e._next;this._addToRemovals(this._unlink(e)),e=A}this._unlinkedRecords!==null&&this._unlinkedRecords.clear(),this._additionsTail!==null&&(this._additionsTail._nextAdded=null),this._movesTail!==null&&(this._movesTail._nextMoved=null),this._itTail!==null&&(this._itTail._next=null),this._removalsTail!==null&&(this._removalsTail._nextRemoved=null),this._identityChangesTail!==null&&(this._identityChangesTail._nextIdentityChange=null)}_reinsertAfter(e,A,i){this._unlinkedRecords!==null&&this._unlinkedRecords.remove(e);let n=e._prevRemoved,o=e._nextRemoved;return n===null?this._removalsHead=o:n._nextRemoved=o,o===null?this._removalsTail=n:o._prevRemoved=n,this._insertAfter(e,A,i),this._addToMoves(e,i),e}_moveAfter(e,A,i){return this._unlink(e),this._insertAfter(e,A,i),this._addToMoves(e,i),e}_addAfter(e,A,i){return this._insertAfter(e,A,i),this._additionsTail===null?this._additionsTail=this._additionsHead=e:this._additionsTail=this._additionsTail._nextAdded=e,e}_insertAfter(e,A,i){let n=A===null?this._itHead:A._next;return e._next=n,e._prev=A,n===null?this._itTail=e:n._prev=e,A===null?this._itHead=e:A._next=e,this._linkedRecords===null&&(this._linkedRecords=new cp),this._linkedRecords.put(e),e.currentIndex=i,e}_remove(e){return this._addToRemovals(this._unlink(e))}_unlink(e){this._linkedRecords!==null&&this._linkedRecords.remove(e);let A=e._prev,i=e._next;return A===null?this._itHead=i:A._next=i,i===null?this._itTail=A:i._prev=A,e}_addToMoves(e,A){return e.previousIndex===A||(this._movesTail===null?this._movesTail=this._movesHead=e:this._movesTail=this._movesTail._nextMoved=e),e}_addToRemovals(e){return this._unlinkedRecords===null&&(this._unlinkedRecords=new cp),this._unlinkedRecords.put(e),e.currentIndex=null,e._nextRemoved=null,this._removalsTail===null?(this._removalsTail=this._removalsHead=e,e._prevRemoved=null):(e._prevRemoved=this._removalsTail,this._removalsTail=this._removalsTail._nextRemoved=e),e}_addIdentityChange(e,A){return e.item=A,this._identityChangesTail===null?this._identityChangesTail=this._identityChangesHead=e:this._identityChangesTail=this._identityChangesTail._nextIdentityChange=e,e}},L9=class{item;trackById;currentIndex=null;previousIndex=null;_nextPrevious=null;_prev=null;_next=null;_prevDup=null;_nextDup=null;_prevRemoved=null;_nextRemoved=null;_nextAdded=null;_nextMoved=null;_nextIdentityChange=null;constructor(e,A){this.item=e,this.trackById=A}},F9=class{_head=null;_tail=null;add(e){this._head===null?(this._head=this._tail=e,e._nextDup=null,e._prevDup=null):(this._tail._nextDup=e,e._prevDup=this._tail,e._nextDup=null,this._tail=e)}get(e,A){let i;for(i=this._head;i!==null;i=i._nextDup)if((A===null||A<=i.currentIndex)&&Object.is(i.trackById,e))return i;return null}remove(e){let A=e._prevDup,i=e._nextDup;return A===null?this._head=i:A._nextDup=i,i===null?this._tail=A:i._prevDup=A,this._head===null}},cp=class{map=new Map;put(e){let A=e.trackById,i=this.map.get(A);i||(i=new F9,this.map.set(A,i)),i.add(e)}get(e,A){let i=e,n=this.map.get(i);return n?n.get(e,A):null}remove(e){let A=e.trackById;return this.map.get(A).remove(e)&&this.map.delete(A),e}get isEmpty(){return this.map.size===0}clear(){this.map.clear()}};function EY(t,e,A){let i=t.previousIndex;if(i===null)return i;let n=0;return A&&i<A.length&&(n=A[i]),i+e+n}var N9=class{constructor(){}supports(e){return e instanceof Map||Fb(e)}create(){return new _9}},_9=class{_records=new Map;_mapHead=null;_appendAfter=null;_previousMapHead=null;_changesHead=null;_changesTail=null;_additionsHead=null;_additionsTail=null;_removalsHead=null;_removalsTail=null;get isDirty(){return this._additionsHead!==null||this._changesHead!==null||this._removalsHead!==null}forEachItem(e){let A;for(A=this._mapHead;A!==null;A=A._next)e(A)}forEachPreviousItem(e){let A;for(A=this._previousMapHead;A!==null;A=A._nextPrevious)e(A)}forEachChangedItem(e){let A;for(A=this._changesHead;A!==null;A=A._nextChanged)e(A)}forEachAddedItem(e){let A;for(A=this._additionsHead;A!==null;A=A._nextAdded)e(A)}forEachRemovedItem(e){let A;for(A=this._removalsHead;A!==null;A=A._nextRemoved)e(A)}diff(e){if(!e)e=new Map;else if(!(e instanceof Map||Fb(e)))throw new Ae(900,!1);return this.check(e)?this:null}onDestroy(){}check(e){this._reset();let A=this._mapHead;if(this._appendAfter=null,this._forEach(e,(i,n)=>{if(A&&A.key===n)this._maybeAddToChanges(A,i),this._appendAfter=A,A=A._next;else{let o=this._getOrCreateRecordForKey(n,i);A=this._insertBeforeOrAppend(A,o)}}),A){A._prev&&(A._prev._next=null),this._removalsHead=A;for(let i=A;i!==null;i=i._nextRemoved)i===this._mapHead&&(this._mapHead=null),this._records.delete(i.key),i._nextRemoved=i._next,i.previousValue=i.currentValue,i.currentValue=null,i._prev=null,i._next=null}return this._changesTail&&(this._changesTail._nextChanged=null),this._additionsTail&&(this._additionsTail._nextAdded=null),this.isDirty}_insertBeforeOrAppend(e,A){if(e){let i=e._prev;return A._next=e,A._prev=i,e._prev=A,i&&(i._next=A),e===this._mapHead&&(this._mapHead=A),this._appendAfter=e,e}return this._appendAfter?(this._appendAfter._next=A,A._prev=this._appendAfter):this._mapHead=A,this._appendAfter=A,null}_getOrCreateRecordForKey(e,A){if(this._records.has(e)){let n=this._records.get(e);this._maybeAddToChanges(n,A);let o=n._prev,r=n._next;return o&&(o._next=r),r&&(r._prev=o),n._next=null,n._prev=null,n}let i=new G9(e);return this._records.set(e,i),i.currentValue=A,this._addToAdditions(i),i}_reset(){if(this.isDirty){let e;for(this._previousMapHead=this._mapHead,e=this._previousMapHead;e!==null;e=e._next)e._nextPrevious=e._next;for(e=this._changesHead;e!==null;e=e._nextChanged)e.previousValue=e.currentValue;for(e=this._additionsHead;e!=null;e=e._nextAdded)e.previousValue=e.currentValue;this._changesHead=this._changesTail=null,this._additionsHead=this._additionsTail=null,this._removalsHead=null}}_maybeAddToChanges(e,A){Object.is(A,e.currentValue)||(e.previousValue=e.currentValue,e.currentValue=A,this._addToChanges(e))}_addToAdditions(e){this._additionsHead===null?this._additionsHead=this._additionsTail=e:(this._additionsTail._nextAdded=e,this._additionsTail=e)}_addToChanges(e){this._changesHead===null?this._changesHead=this._changesTail=e:(this._changesTail._nextChanged=e,this._changesTail=e)}_forEach(e,A){e instanceof Map?e.forEach(A):Object.keys(e).forEach(i=>A(e[i],i))}},G9=class{key;previousValue=null;currentValue=null;_nextPrevious=null;_next=null;_prev=null;_nextAdded=null;_nextRemoved=null;_nextChanged=null;constructor(e){this.key=e}};function QY(){return new rg([new R9])}var rg=(()=>{class t{factories;static \u0275prov=NA({token:t,providedIn:"root",factory:QY});constructor(A){this.factories=A}static create(A,i){if(i!=null){let n=i.factories.slice();A=A.concat(n)}return new t(A)}static extend(A){return{provide:t,useFactory:i=>t.create(A,i||QY()),deps:[[t,new fh,new MI]]}}find(A){let i=this.factories.find(n=>n.supports(A));if(i!=null)return i;throw new Ae(901,!1)}}return t})();function hY(){return new _p([new N9])}var _p=(()=>{class t{static \u0275prov=NA({token:t,providedIn:"root",factory:hY});factories;constructor(A){this.factories=A}static create(A,i){if(i){let n=i.factories.slice();A=A.concat(n)}return new t(A)}static extend(A){return{provide:t,useFactory:i=>t.create(A,i||hY()),deps:[[t,new fh,new MI]]}}find(A){let i=this.factories.find(n=>n.supports(A));if(i)return i;throw new Ae(901,!1)}}return t})();var uH=jb(null,"core",[]),fH=(()=>{class t{constructor(A){}static \u0275fac=function(i){return new(i||t)(we(la))};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();function ie(t){return typeof t=="boolean"?t:t!=null&&t!=="false"}function zi(t,e=NaN){return!isNaN(parseFloat(t))&&!isNaN(Number(t))?Number(t):e}function Ba(t){return Av(t)}function h0(t,e){return Zf(t,e?.equal)}var U9=class{[ia];constructor(e){this[ia]=e}destroy(){this[ia].destroy()}};function Nh(t,e){!e?.injector&&z9(Nh);let A=e?.injector??f(Rt),i=e?.manualCleanup!==!0?A.get(m2):null,n,o=A.get(db,null,{optional:!0}),r=A.get(DI);return o!==null&&!e?.forceRoot?(n=DBA(o.view,r,t),i instanceof jm&&i._lView===o.view&&(i=null)):n=yBA(t,A.get(zT),r),n.injector=A,i!==null&&(n.onDestroyFn=i.onDestroy(()=>n.destroy())),new U9(n)}var mH=Ye(rA({},Bd),{consumerIsAlwaysLive:!0,consumerAllowSignalWrites:!0,dirty:!0,hasRun:!1,cleanupFns:void 0,zone:null,kind:"effect",onDestroyFn:Bh,run(){if(this.dirty=!1,this.hasRun&&!jf(this))return;this.hasRun=!0;let t=i=>(this.cleanupFns??=[]).push(i),e=ZQ(this),A=Tm(!1);try{this.maybeCleanup(),this.fn(t)}finally{Tm(A),Pf(this,e)}},maybeCleanup(){if(this.cleanupFns?.length)try{for(;this.cleanupFns.length;)this.cleanupFns.pop()()}finally{this.cleanupFns=[]}}}),pBA=Ye(rA({},mH),{consumerMarkedDirty(){this.scheduler.schedule(this),this.notifier.notify(12)},destroy(){WQ(this),this.onDestroyFn(),this.maybeCleanup(),this.scheduler.remove(this)}}),wBA=Ye(rA({},mH),{consumerMarkedDirty(){this.view[gi]|=8192,Pd(this.view),this.notifier.notify(13)},destroy(){WQ(this),this.onDestroyFn(),this.maybeCleanup(),this.view[uI]?.delete(this)}});function DBA(t,e,A){let i=Object.create(wBA);return i.view=t,i.zone=typeof Zone<"u"?Zone.current:null,i.notifier=e,i.fn=A,t[uI]??=new Set,t[uI].add(i),i.consumerMarkedDirty(i),i}function yBA(t,e,A){let i=Object.create(pBA);return i.fn=t,i.scheduler=e,i.notifier=A,i.zone=typeof Zone<"u"?Zone.current:null,i.scheduler.schedule(i),i.notifier.notify(12),i}function Gp(t,e){let A=h2(t),i=e.elementInjector||dp();return new yI(A).create(i,e.projectableNodes,e.hostElement,e.environmentInjector)}function pH(t){let e=h2(t);if(!e)return null;let A=new yI(e);return{get selector(){return A.selector},get type(){return A.componentType},get inputs(){return A.inputs},get outputs(){return A.outputs},get ngContentSelectors(){return A.ngContentSelectors},get isStandalone(){return e.standalone},get isSignal(){return e.signals}}}var at=new dA("");var yH=null;function Wa(){return yH}function qb(t){yH??=t}var _h=class{},Gh=(()=>{class t{historyGo(A){throw new Error("")}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(vH),providedIn:"platform"})}return t})(),Vb=new dA(""),vH=(()=>{class t extends Gh{_location;_history;_doc=f(at);constructor(){super(),this._location=window.location,this._history=window.history}getBaseHrefFromDOM(){return Wa().getBaseHref(this._doc)}onPopState(A){let i=Wa().getGlobalEventTarget(this._doc,"window");return i.addEventListener("popstate",A,!1),()=>i.removeEventListener("popstate",A)}onHashChange(A){let i=Wa().getGlobalEventTarget(this._doc,"window");return i.addEventListener("hashchange",A,!1),()=>i.removeEventListener("hashchange",A)}get href(){return this._location.href}get protocol(){return this._location.protocol}get hostname(){return this._location.hostname}get port(){return this._location.port}get pathname(){return this._location.pathname}get search(){return this._location.search}get hash(){return this._location.hash}set pathname(A){this._location.pathname=A}pushState(A,i,n){this._history.pushState(A,i,n)}replaceState(A,i,n){this._history.replaceState(A,i,n)}forward(){this._history.forward()}back(){this._history.back()}historyGo(A=0){this._history.go(A)}getState(){return this._history.state}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>new t,providedIn:"platform"})}return t})();function Up(t,e){return t?e?t.endsWith("/")?e.startsWith("/")?t+e.slice(1):t+e:e.startsWith("/")?t+e:`${t}/${e}`:t:e}function wH(t){let e=t.search(/#|\?|$/);return t[e-1]==="/"?t.slice(0,e-1)+t.slice(e):t}function al(t){return t&&t[0]!=="?"?`?${t}`:t}var u0=(()=>{class t{historyGo(A){throw new Error("")}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(Yp),providedIn:"root"})}return t})(),Kp=new dA(""),Yp=(()=>{class t extends u0{_platformLocation;_baseHref;_removeListenerFns=[];constructor(A,i){super(),this._platformLocation=A,this._baseHref=i??this._platformLocation.getBaseHrefFromDOM()??f(at).location?.origin??""}ngOnDestroy(){for(;this._removeListenerFns.length;)this._removeListenerFns.pop()()}onPopState(A){this._removeListenerFns.push(this._platformLocation.onPopState(A),this._platformLocation.onHashChange(A))}getBaseHref(){return this._baseHref}prepareExternalUrl(A){return Up(this._baseHref,A)}path(A=!1){let i=this._platformLocation.pathname+al(this._platformLocation.search),n=this._platformLocation.hash;return n&&A?`${i}${n}`:i}pushState(A,i,n,o){let r=this.prepareExternalUrl(n+al(o));this._platformLocation.pushState(A,i,r)}replaceState(A,i,n,o){let r=this.prepareExternalUrl(n+al(o));this._platformLocation.replaceState(A,i,r)}forward(){this._platformLocation.forward()}back(){this._platformLocation.back()}getState(){return this._platformLocation.getState()}historyGo(A=0){this._platformLocation.historyGo?.(A)}static \u0275fac=function(i){return new(i||t)(we(Gh),we(Kp,8))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),Dc=(()=>{class t{_subject=new OA;_basePath;_locationStrategy;_urlChangeListeners=[];_urlChangeSubscription=null;constructor(A){this._locationStrategy=A;let i=this._locationStrategy.getBaseHref();this._basePath=MBA(wH(DH(i))),this._locationStrategy.onPopState(n=>{this._subject.next({url:this.path(!0),pop:!0,state:n.state,type:n.type})})}ngOnDestroy(){this._urlChangeSubscription?.unsubscribe(),this._urlChangeListeners=[]}path(A=!1){return this.normalize(this._locationStrategy.path(A))}getState(){return this._locationStrategy.getState()}isCurrentPathEqualTo(A,i=""){return this.path()==this.normalize(A+al(i))}normalize(A){return t.stripTrailingSlash(bBA(this._basePath,DH(A)))}prepareExternalUrl(A){return A&&A[0]!=="/"&&(A="/"+A),this._locationStrategy.prepareExternalUrl(A)}go(A,i="",n=null){this._locationStrategy.pushState(n,"",A,i),this._notifyUrlChangeListeners(this.prepareExternalUrl(A+al(i)),n)}replaceState(A,i="",n=null){this._locationStrategy.replaceState(n,"",A,i),this._notifyUrlChangeListeners(this.prepareExternalUrl(A+al(i)),n)}forward(){this._locationStrategy.forward()}back(){this._locationStrategy.back()}historyGo(A=0){this._locationStrategy.historyGo?.(A)}onUrlChange(A){return this._urlChangeListeners.push(A),this._urlChangeSubscription??=this.subscribe(i=>{this._notifyUrlChangeListeners(i.url,i.state)}),()=>{let i=this._urlChangeListeners.indexOf(A);this._urlChangeListeners.splice(i,1),this._urlChangeListeners.length===0&&(this._urlChangeSubscription?.unsubscribe(),this._urlChangeSubscription=null)}}_notifyUrlChangeListeners(A="",i){this._urlChangeListeners.forEach(n=>n(A,i))}subscribe(A,i,n){return this._subject.subscribe({next:A,error:i??void 0,complete:n??void 0})}static normalizeQueryParams=al;static joinWithSlash=Up;static stripTrailingSlash=wH;static \u0275fac=function(i){return new(i||t)(we(u0))};static \u0275prov=NA({token:t,factory:()=>vBA(),providedIn:"root"})}return t})();function vBA(){return new Dc(we(u0))}function bBA(t,e){if(!t||!e.startsWith(t))return e;let A=e.substring(t.length);return A===""||["/",";","?","#"].includes(A[0])?A:e}function DH(t){return t.replace(/\/index.html$/,"")}function MBA(t){if(new RegExp("^(https?:)?//").test(t)){let[,A]=t.split(/\/\/[^\/]+/);return A}return t}var $b=(()=>{class t extends u0{_platformLocation;_baseHref="";_removeListenerFns=[];constructor(A,i){super(),this._platformLocation=A,i!=null&&(this._baseHref=i)}ngOnDestroy(){for(;this._removeListenerFns.length;)this._removeListenerFns.pop()()}onPopState(A){this._removeListenerFns.push(this._platformLocation.onPopState(A),this._platformLocation.onHashChange(A))}getBaseHref(){return this._baseHref}path(A=!1){let i=this._platformLocation.hash??"#";return i.length>0?i.substring(1):i}prepareExternalUrl(A){let i=Up(this._baseHref,A);return i.length>0?"#"+i:i}pushState(A,i,n,o){let r=this.prepareExternalUrl(n+al(o))||this._platformLocation.pathname;this._platformLocation.pushState(A,i,r)}replaceState(A,i,n,o){let r=this.prepareExternalUrl(n+al(o))||this._platformLocation.pathname;this._platformLocation.replaceState(A,i,r)}forward(){this._platformLocation.forward()}back(){this._platformLocation.back()}getState(){return this._platformLocation.getState()}historyGo(A=0){this._platformLocation.historyGo?.(A)}static \u0275fac=function(i){return new(i||t)(we(Gh),we(Kp,8))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();var Zb=/\s+/,bH=[],Xa=(()=>{class t{_ngEl;_renderer;initialClasses=bH;rawClass;stateMap=new Map;constructor(A,i){this._ngEl=A,this._renderer=i}set klass(A){this.initialClasses=A!=null?A.trim().split(Zb):bH}set ngClass(A){this.rawClass=typeof A=="string"?A.trim().split(Zb):A}ngDoCheck(){for(let i of this.initialClasses)this._updateState(i,!0);let A=this.rawClass;if(Array.isArray(A)||A instanceof Set)for(let i of A)this._updateState(i,!0);else if(A!=null)for(let i of Object.keys(A))this._updateState(i,!!A[i]);this._applyStateDiff()}_updateState(A,i){let n=this.stateMap.get(A);n!==void 0?(n.enabled!==i&&(n.changed=!0,n.enabled=i),n.touched=!0):this.stateMap.set(A,{enabled:i,changed:!0,touched:!0})}_applyStateDiff(){for(let A of this.stateMap){let i=A[0],n=A[1];n.changed?(this._toggleClass(i,n.enabled),n.changed=!1):n.touched||(n.enabled&&this._toggleClass(i,!1),this.stateMap.delete(i)),n.touched=!1}}_toggleClass(A,i){A=A.trim(),A.length>0&&A.split(Zb).forEach(n=>{i?this._renderer.addClass(this._ngEl.nativeElement,n):this._renderer.removeClass(this._ngEl.nativeElement,n)})}static \u0275fac=function(i){return new(i||t)(PA(ee),PA(Wi))};static \u0275dir=jA({type:t,selectors:[["","ngClass",""]],inputs:{klass:[0,"class","klass"],ngClass:"ngClass"}})}return t})();var Jp=class{$implicit;ngForOf;index;count;constructor(e,A,i,n){this.$implicit=e,this.ngForOf=A,this.index=i,this.count=n}get first(){return this.index===0}get last(){return this.index===this.count-1}get even(){return this.index%2===0}get odd(){return!this.even}},Hp=(()=>{class t{_viewContainer;_template;_differs;set ngForOf(A){this._ngForOf=A,this._ngForOfDirty=!0}set ngForTrackBy(A){this._trackByFn=A}get ngForTrackBy(){return this._trackByFn}_ngForOf=null;_ngForOfDirty=!0;_differ=null;_trackByFn;constructor(A,i,n){this._viewContainer=A,this._template=i,this._differs=n}set ngForTemplate(A){A&&(this._template=A)}ngDoCheck(){if(this._ngForOfDirty){this._ngForOfDirty=!1;let A=this._ngForOf;!this._differ&&A&&(this._differ=this._differs.find(A).create(this.ngForTrackBy))}if(this._differ){let A=this._differ.diff(this._ngForOf);A&&this._applyChanges(A)}}_applyChanges(A){let i=this._viewContainer;A.forEachOperation((n,o,r)=>{if(n.previousIndex==null)i.createEmbeddedView(this._template,new Jp(n.item,this._ngForOf,-1,-1),r===null?void 0:r);else if(r==null)i.remove(o===null?void 0:o);else if(o!==null){let s=i.get(o);i.move(s,r),MH(s,n)}});for(let n=0,o=i.length;n<o;n++){let s=i.get(n).context;s.index=n,s.count=o,s.ngForOf=this._ngForOf}A.forEachIdentityChange(n=>{let o=i.get(n.currentIndex);MH(o,n)})}static ngTemplateContextGuard(A,i){return!0}static \u0275fac=function(i){return new(i||t)(PA(Nn),PA(wn),PA(rg))};static \u0275dir=jA({type:t,selectors:[["","ngFor","","ngForOf",""]],inputs:{ngForOf:"ngForOf",ngForTrackBy:"ngForTrackBy",ngForTemplate:"ngForTemplate"}})}return t})();function MH(t,e){t.context.$implicit=e.item}var Uh=(()=>{class t{_viewContainer;_context=new Tp;_thenTemplateRef=null;_elseTemplateRef=null;_thenViewRef=null;_elseViewRef=null;constructor(A,i){this._viewContainer=A,this._thenTemplateRef=i}set ngIf(A){this._context.$implicit=this._context.ngIf=A,this._updateView()}set ngIfThen(A){kH(A,!1),this._thenTemplateRef=A,this._thenViewRef=null,this._updateView()}set ngIfElse(A){kH(A,!1),this._elseTemplateRef=A,this._elseViewRef=null,this._updateView()}_updateView(){this._context.$implicit?this._thenViewRef||(this._viewContainer.clear(),this._elseViewRef=null,this._thenTemplateRef&&(this._thenViewRef=this._viewContainer.createEmbeddedView(this._thenTemplateRef,this._context))):this._elseViewRef||(this._viewContainer.clear(),this._thenViewRef=null,this._elseTemplateRef&&(this._elseViewRef=this._viewContainer.createEmbeddedView(this._elseTemplateRef,this._context)))}static ngIfUseIfTypeGuard;static ngTemplateGuard_ngIf;static ngTemplateContextGuard(A,i){return!0}static \u0275fac=function(i){return new(i||t)(PA(Nn),PA(wn))};static \u0275dir=jA({type:t,selectors:[["","ngIf",""]],inputs:{ngIf:"ngIf",ngIfThen:"ngIfThen",ngIfElse:"ngIfElse"}})}return t})(),Tp=class{$implicit=null;ngIf=null};function kH(t,e){if(t&&!t.createEmbeddedView)throw new Ae(2020,!1)}var Kh=(()=>{class t{_ngEl;_differs;_renderer;_ngStyle=null;_differ=null;constructor(A,i,n){this._ngEl=A,this._differs=i,this._renderer=n}set ngStyle(A){this._ngStyle=A,!this._differ&&A&&(this._differ=this._differs.find(A).create())}ngDoCheck(){if(this._differ){let A=this._differ.diff(this._ngStyle);A&&this._applyChanges(A)}}_setStyle(A,i){let[n,o]=A.split("."),r=n.indexOf("-")===-1?void 0:tg.DashCase;i!=null?this._renderer.setStyle(this._ngEl.nativeElement,n,o?`${i}${o}`:i,r):this._renderer.removeStyle(this._ngEl.nativeElement,n,r)}_applyChanges(A){A.forEachRemovedItem(i=>this._setStyle(i.key,null)),A.forEachAddedItem(i=>this._setStyle(i.key,i.currentValue)),A.forEachChangedItem(i=>this._setStyle(i.key,i.currentValue))}static \u0275fac=function(i){return new(i||t)(PA(ee),PA(_p),PA(Wi))};static \u0275dir=jA({type:t,selectors:[["","ngStyle",""]],inputs:{ngStyle:"ngStyle"}})}return t})(),Yh=(()=>{class t{_viewContainerRef;_viewRef=null;ngTemplateOutletContext=null;ngTemplateOutlet=null;ngTemplateOutletInjector=null;constructor(A){this._viewContainerRef=A}ngOnChanges(A){if(this._shouldRecreateView(A)){let i=this._viewContainerRef;if(this._viewRef&&i.remove(i.indexOf(this._viewRef)),!this.ngTemplateOutlet){this._viewRef=null;return}let n=this._createContextForwardProxy();this._viewRef=i.createEmbeddedView(this.ngTemplateOutlet,n,{injector:this.ngTemplateOutletInjector??void 0})}}_shouldRecreateView(A){return!!A.ngTemplateOutlet||!!A.ngTemplateOutletInjector}_createContextForwardProxy(){return new Proxy({},{set:(A,i,n)=>this.ngTemplateOutletContext?Reflect.set(this.ngTemplateOutletContext,i,n):!1,get:(A,i,n)=>{if(this.ngTemplateOutletContext)return Reflect.get(this.ngTemplateOutletContext,i,n)}})}static \u0275fac=function(i){return new(i||t)(PA(Nn))};static \u0275dir=jA({type:t,selectors:[["","ngTemplateOutlet",""]],inputs:{ngTemplateOutletContext:"ngTemplateOutletContext",ngTemplateOutlet:"ngTemplateOutlet",ngTemplateOutletInjector:"ngTemplateOutletInjector"},features:[ti]})}return t})();function kBA(t,e){return new Ae(2100,!1)}var Wb=class{createSubscription(e,A){return Ba(()=>e.subscribe({next:A,error:i=>{throw i}}))}dispose(e){Ba(()=>e.unsubscribe())}},Xb=class{createSubscription(e,A){return e.then(i=>A?.(i),i=>{throw i}),{unsubscribe:()=>{A=null}}}dispose(e){e.unsubscribe()}},SBA=new Xb,RBA=new Wb,Jh=(()=>{class t{_ref;_latestValue=null;markForCheckOnValueUpdate=!0;_subscription=null;_obj=null;_strategy=null;constructor(A){this._ref=A}ngOnDestroy(){this._subscription&&this._dispose(),this._ref=null}transform(A){if(!this._obj){if(A)try{this.markForCheckOnValueUpdate=!1,this._subscribe(A)}finally{this.markForCheckOnValueUpdate=!0}return this._latestValue}return A!==this._obj?(this._dispose(),this.transform(A)):this._latestValue}_subscribe(A){this._obj=A,this._strategy=this._selectStrategy(A),this._subscription=this._strategy.createSubscription(A,i=>this._updateLatestValue(A,i))}_selectStrategy(A){if(D2(A))return SBA;if(Kb(A))return RBA;throw kBA(t,A)}_dispose(){this._strategy.dispose(this._subscription),this._latestValue=null,this._subscription=null,this._obj=null}_updateLatestValue(A,i){A===this._obj&&(this._latestValue=i,this.markForCheckOnValueUpdate&&this._ref?.markForCheck())}static \u0275fac=function(i){return new(i||t)(PA(It,16))};static \u0275pipe=xp({name:"async",type:t,pure:!1})}return t})();function xBA(t,e){return{key:t,value:e}}var Th=(()=>{class t{differs;constructor(A){this.differs=A}differ;keyValues=[];compareFn=SH;transform(A,i=SH){if(!A||!(A instanceof Map)&&typeof A!="object")return null;this.differ??=this.differs.find(A).create();let n=this.differ.diff(A),o=i!==this.compareFn;return n&&(this.keyValues=[],n.forEachItem(r=>{this.keyValues.push(xBA(r.key,r.currentValue))})),(n||o)&&(i&&this.keyValues.sort(i),this.compareFn=i),this.keyValues}static \u0275fac=function(i){return new(i||t)(PA(_p,16))};static \u0275pipe=xp({name:"keyvalue",type:t,pure:!1})}return t})();function SH(t,e){let A=t.key,i=e.key;if(A===i)return 0;if(A==null)return 1;if(i==null)return-1;if(typeof A=="string"&&typeof i=="string")return A<i?-1:1;if(typeof A=="number"&&typeof i=="number")return A-i;if(typeof A=="boolean"&&typeof i=="boolean")return A<i?-1:1;let n=String(A),o=String(i);return n==o?0:n<o?-1:1}var f0=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();function Hh(t,e){e=encodeURIComponent(e);for(let A of t.split(";")){let i=A.indexOf("="),[n,o]=i==-1?[A,""]:[A.slice(0,i),A.slice(i+1)];if(n.trim()===e)return decodeURIComponent(o)}return null}var zp="browser",RH="server";function sg(t){return t===zp}function Op(t){return t===RH}var xI=class{};var xH=(()=>{class t{static \u0275prov=NA({token:t,providedIn:"root",factory:()=>new AM(f(at),window)})}return t})(),AM=class{document;window;offset=()=>[0,0];constructor(e,A){this.document=e,this.window=A}setOffset(e){Array.isArray(e)?this.offset=()=>e:this.offset=e}getScrollPosition(){return[this.window.scrollX,this.window.scrollY]}scrollToPosition(e){this.window.scrollTo(e[0],e[1])}scrollToAnchor(e){let A=LBA(this.document,e);A&&(this.scrollToElement(A),A.focus())}setHistoryScrollRestoration(e){this.window.history.scrollRestoration=e}scrollToElement(e){let A=e.getBoundingClientRect(),i=A.left+this.window.pageXOffset,n=A.top+this.window.pageYOffset,o=this.offset();this.window.scrollTo(i-o[0],n-o[1])}};function LBA(t,e){let A=t.getElementById(e)||t.getElementsByName(e)[0];if(A)return A;if(typeof t.createTreeWalker=="function"&&t.body&&typeof t.body.attachShadow=="function"){let i=t.createTreeWalker(t.body,NodeFilter.SHOW_ELEMENT),n=i.currentNode;for(;n;){let o=n.shadowRoot;if(o){let r=o.getElementById(e)||o.querySelector(`[name="${e}"]`);if(r)return r}n=i.nextNode()}}return null}var qp=new dA(""),nM=(()=>{class t{_zone;_plugins;_eventNameToPlugin=new Map;constructor(A,i){this._zone=i,A.forEach(n=>{n.manager=this}),this._plugins=A.slice().reverse()}addEventListener(A,i,n,o){return this._findPluginFor(i).addEventListener(A,i,n,o)}getZone(){return this._zone}_findPluginFor(A){let i=this._eventNameToPlugin.get(A);if(i)return i;if(i=this._plugins.find(o=>o.supports(A)),!i)throw new Ae(5101,!1);return this._eventNameToPlugin.set(A,i),i}static \u0275fac=function(i){return new(i||t)(we(qp),we(Qe))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),zh=class{_doc;constructor(e){this._doc=e}manager},Pp="ng-app-id";function LH(t){for(let e of t)e.remove()}function FH(t,e){let A=e.createElement("style");return A.textContent=t,A}function FBA(t,e,A,i){let n=t.head?.querySelectorAll(`style[${Pp}="${e}"],link[${Pp}="${e}"]`);if(n)for(let o of n)o.removeAttribute(Pp),o instanceof HTMLLinkElement?i.set(o.href.slice(o.href.lastIndexOf("/")+1),{usage:0,elements:[o]}):o.textContent&&A.set(o.textContent,{usage:0,elements:[o]})}function tM(t,e){let A=e.createElement("link");return A.setAttribute("rel","stylesheet"),A.setAttribute("href",t),A}var oM=(()=>{class t{doc;appId;nonce;inline=new Map;external=new Map;hosts=new Set;isServer;constructor(A,i,n,o={}){this.doc=A,this.appId=i,this.nonce=n,this.isServer=Op(o),FBA(A,i,this.inline,this.external),this.hosts.add(A.head)}addStyles(A,i){for(let n of A)this.addUsage(n,this.inline,FH);i?.forEach(n=>this.addUsage(n,this.external,tM))}removeStyles(A,i){for(let n of A)this.removeUsage(n,this.inline);i?.forEach(n=>this.removeUsage(n,this.external))}addUsage(A,i,n){let o=i.get(A);o?o.usage++:i.set(A,{usage:1,elements:[...this.hosts].map(r=>this.addElement(r,n(A,this.doc)))})}removeUsage(A,i){let n=i.get(A);n&&(n.usage--,n.usage<=0&&(LH(n.elements),i.delete(A)))}ngOnDestroy(){for(let[,{elements:A}]of[...this.inline,...this.external])LH(A);this.hosts.clear()}addHost(A){this.hosts.add(A);for(let[i,{elements:n}]of this.inline)n.push(this.addElement(A,FH(i,this.doc)));for(let[i,{elements:n}]of this.external)n.push(this.addElement(A,tM(i,this.doc)))}removeHost(A){this.hosts.delete(A)}addElement(A,i){return this.nonce&&i.setAttribute("nonce",this.nonce),this.isServer&&i.setAttribute(Pp,this.appId),A.appendChild(i)}static \u0275fac=function(i){return new(i||t)(we(at),we(Vd),we(yh,8),we(og))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),eM={svg:"http://www.w3.org/2000/svg",xhtml:"http://www.w3.org/1999/xhtml",xlink:"http://www.w3.org/1999/xlink",xml:"http://www.w3.org/XML/1998/namespace",xmlns:"http://www.w3.org/2000/xmlns/",math:"http://www.w3.org/1998/Math/MathML"},rM=/%COMP%/g;var _H="%COMP%",NBA=`_nghost-${_H}`,_BA=`_ngcontent-${_H}`,GBA=!0,UBA=new dA("",{providedIn:"root",factory:()=>GBA});function KBA(t){return _BA.replace(rM,t)}function YBA(t){return NBA.replace(rM,t)}function GH(t,e){return e.map(A=>A.replace(rM,t))}var jh=(()=>{class t{eventManager;sharedStylesHost;appId;removeStylesOnCompDestroy;doc;platformId;ngZone;nonce;tracingService;rendererByCompId=new Map;defaultRenderer;platformIsServer;constructor(A,i,n,o,r,s,a,c=null,l=null){this.eventManager=A,this.sharedStylesHost=i,this.appId=n,this.removeStylesOnCompDestroy=o,this.doc=r,this.platformId=s,this.ngZone=a,this.nonce=c,this.tracingService=l,this.platformIsServer=Op(s),this.defaultRenderer=new Oh(A,r,a,this.platformIsServer,this.tracingService)}createRenderer(A,i){if(!A||!i)return this.defaultRenderer;this.platformIsServer&&i.encapsulation===eg.ShadowDom&&(i=Ye(rA({},i),{encapsulation:eg.Emulated}));let n=this.getOrCreateRenderer(A,i);return n instanceof jp?n.applyToHost(A):n instanceof Ph&&n.applyStyles(),n}getOrCreateRenderer(A,i){let n=this.rendererByCompId,o=n.get(i.id);if(!o){let r=this.doc,s=this.ngZone,a=this.eventManager,c=this.sharedStylesHost,l=this.removeStylesOnCompDestroy,I=this.platformIsServer,C=this.tracingService;switch(i.encapsulation){case eg.Emulated:o=new jp(a,c,i,this.appId,l,r,s,I,C);break;case eg.ShadowDom:return new iM(a,c,A,i,r,s,this.nonce,I,C);default:o=new Ph(a,c,i,l,r,s,I,C);break}n.set(i.id,o)}return o}ngOnDestroy(){this.rendererByCompId.clear()}componentReplaced(A){this.rendererByCompId.delete(A)}static \u0275fac=function(i){return new(i||t)(we(nM),we(oM),we(Vd),we(UBA),we(at),we(og),we(Qe),we(yh),we(Zd,8))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),Oh=class{eventManager;doc;ngZone;platformIsServer;tracingService;data=Object.create(null);throwOnSyntheticProps=!0;constructor(e,A,i,n,o){this.eventManager=e,this.doc=A,this.ngZone=i,this.platformIsServer=n,this.tracingService=o}destroy(){}destroyNode=null;createElement(e,A){return A?this.doc.createElementNS(eM[A]||A,e):this.doc.createElement(e)}createComment(e){return this.doc.createComment(e)}createText(e){return this.doc.createTextNode(e)}appendChild(e,A){(NH(e)?e.content:e).appendChild(A)}insertBefore(e,A,i){e&&(NH(e)?e.content:e).insertBefore(A,i)}removeChild(e,A){A.remove()}selectRootElement(e,A){let i=typeof e=="string"?this.doc.querySelector(e):e;if(!i)throw new Ae(-5104,!1);return A||(i.textContent=""),i}parentNode(e){return e.parentNode}nextSibling(e){return e.nextSibling}setAttribute(e,A,i,n){if(n){A=n+":"+A;let o=eM[n];o?e.setAttributeNS(o,A,i):e.setAttribute(A,i)}else e.setAttribute(A,i)}removeAttribute(e,A,i){if(i){let n=eM[i];n?e.removeAttributeNS(n,A):e.removeAttribute(`${i}:${A}`)}else e.removeAttribute(A)}addClass(e,A){e.classList.add(A)}removeClass(e,A){e.classList.remove(A)}setStyle(e,A,i,n){n&(tg.DashCase|tg.Important)?e.style.setProperty(A,i,n&tg.Important?"important":""):e.style[A]=i}removeStyle(e,A,i){i&tg.DashCase?e.style.removeProperty(A):e.style[A]=""}setProperty(e,A,i){e!=null&&(e[A]=i)}setValue(e,A){e.nodeValue=A}listen(e,A,i,n){if(typeof e=="string"&&(e=Wa().getGlobalEventTarget(this.doc,e),!e))throw new Ae(5102,!1);let o=this.decoratePreventDefault(i);return this.tracingService?.wrapEventListener&&(o=this.tracingService.wrapEventListener(e,A,o)),this.eventManager.addEventListener(e,A,o,n)}decoratePreventDefault(e){return A=>{if(A==="__ngUnwrap__")return e;(this.platformIsServer?this.ngZone.runGuarded(()=>e(A)):e(A))===!1&&A.preventDefault()}}};function NH(t){return t.tagName==="TEMPLATE"&&t.content!==void 0}var iM=class extends Oh{sharedStylesHost;hostEl;shadowRoot;constructor(e,A,i,n,o,r,s,a,c){super(e,o,r,a,c),this.sharedStylesHost=A,this.hostEl=i,this.shadowRoot=i.attachShadow({mode:"open"}),this.sharedStylesHost.addHost(this.shadowRoot);let l=n.styles;l=GH(n.id,l);for(let C of l){let d=document.createElement("style");s&&d.setAttribute("nonce",s),d.textContent=C,this.shadowRoot.appendChild(d)}let I=n.getExternalStyles?.();if(I)for(let C of I){let d=tM(C,o);s&&d.setAttribute("nonce",s),this.shadowRoot.appendChild(d)}}nodeOrShadowRoot(e){return e===this.hostEl?this.shadowRoot:e}appendChild(e,A){return super.appendChild(this.nodeOrShadowRoot(e),A)}insertBefore(e,A,i){return super.insertBefore(this.nodeOrShadowRoot(e),A,i)}removeChild(e,A){return super.removeChild(null,A)}parentNode(e){return this.nodeOrShadowRoot(super.parentNode(this.nodeOrShadowRoot(e)))}destroy(){this.sharedStylesHost.removeHost(this.shadowRoot)}},Ph=class extends Oh{sharedStylesHost;removeStylesOnCompDestroy;styles;styleUrls;constructor(e,A,i,n,o,r,s,a,c){super(e,o,r,s,a),this.sharedStylesHost=A,this.removeStylesOnCompDestroy=n;let l=i.styles;this.styles=c?GH(c,l):l,this.styleUrls=i.getExternalStyles?.(c)}applyStyles(){this.sharedStylesHost.addStyles(this.styles,this.styleUrls)}destroy(){this.removeStylesOnCompDestroy&&this.sharedStylesHost.removeStyles(this.styles,this.styleUrls)}},jp=class extends Ph{contentAttr;hostAttr;constructor(e,A,i,n,o,r,s,a,c){let l=n+"-"+i.id;super(e,A,i,o,r,s,a,c,l),this.contentAttr=KBA(l),this.hostAttr=YBA(l)}applyToHost(e){this.applyStyles(),this.setAttribute(e,this.hostAttr,"")}createElement(e,A){let i=super.createElement(e,A);return super.setAttribute(i,this.contentAttr,""),i}};var Vp=class t extends _h{supportsDOMEvents=!0;static makeCurrent(){qb(new t)}onAndCancel(e,A,i,n){return e.addEventListener(A,i,n),()=>{e.removeEventListener(A,i,n)}}dispatchEvent(e,A){e.dispatchEvent(A)}remove(e){e.remove()}createElement(e,A){return A=A||this.getDefaultDocument(),A.createElement(e)}createHtmlDocument(){return document.implementation.createHTMLDocument("fakeTitle")}getDefaultDocument(){return document}isElementNode(e){return e.nodeType===Node.ELEMENT_NODE}isShadowRoot(e){return e instanceof DocumentFragment}getGlobalEventTarget(e,A){return A==="window"?window:A==="document"?e:A==="body"?e.body:null}getBaseHref(e){let A=JBA();return A==null?null:TBA(A)}resetBaseElement(){qh=null}getUserAgent(){return window.navigator.userAgent}getCookie(e){return Hh(document.cookie,e)}},qh=null;function JBA(){return qh=qh||document.head.querySelector("base"),qh?qh.getAttribute("href"):null}function TBA(t){return new URL(t,document.baseURI).pathname}var Zp=class{addToWindow(e){sa.getAngularTestability=(i,n=!0)=>{let o=e.findTestabilityInTree(i,n);if(o==null)throw new Ae(5103,!1);return o},sa.getAllAngularTestabilities=()=>e.getAllTestabilities(),sa.getAllAngularRootElements=()=>e.getAllRootElements();let A=i=>{let n=sa.getAllAngularTestabilities(),o=n.length,r=function(){o--,o==0&&i()};n.forEach(s=>{s.whenStable(r)})};sa.frameworkStabilizers||(sa.frameworkStabilizers=[]),sa.frameworkStabilizers.push(A)}findTestabilityInTree(e,A,i){if(A==null)return null;let n=e.getTestability(A);return n??(i?Wa().isShadowRoot(A)?this.findTestabilityInTree(e,A.host,!0):this.findTestabilityInTree(e,A.parentElement,!0):null)}},HBA=(()=>{class t{build(){return new XMLHttpRequest}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),KH=(()=>{class t extends zh{constructor(A){super(A)}supports(A){return!0}addEventListener(A,i,n,o){return A.addEventListener(i,n,o),()=>this.removeEventListener(A,i,n,o)}removeEventListener(A,i,n,o){return A.removeEventListener(i,n,o)}static \u0275fac=function(i){return new(i||t)(we(at))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),UH=["alt","control","meta","shift"],zBA={"\b":"Backspace","	":"Tab","\x7F":"Delete","\x1B":"Escape",Del:"Delete",Esc:"Escape",Left:"ArrowLeft",Right:"ArrowRight",Up:"ArrowUp",Down:"ArrowDown",Menu:"ContextMenu",Scroll:"ScrollLock",Win:"OS"},OBA={alt:t=>t.altKey,control:t=>t.ctrlKey,meta:t=>t.metaKey,shift:t=>t.shiftKey},YH=(()=>{class t extends zh{constructor(A){super(A)}supports(A){return t.parseEventName(A)!=null}addEventListener(A,i,n,o){let r=t.parseEventName(i),s=t.eventCallback(r.fullKey,n,this.manager.getZone());return this.manager.getZone().runOutsideAngular(()=>Wa().onAndCancel(A,r.domEventName,s,o))}static parseEventName(A){let i=A.toLowerCase().split("."),n=i.shift();if(i.length===0||!(n==="keydown"||n==="keyup"))return null;let o=t._normalizeKey(i.pop()),r="",s=i.indexOf("code");if(s>-1&&(i.splice(s,1),r="code."),UH.forEach(c=>{let l=i.indexOf(c);l>-1&&(i.splice(l,1),r+=c+".")}),r+=o,i.length!=0||o.length===0)return null;let a={};return a.domEventName=n,a.fullKey=r,a}static matchEventFullKeyCode(A,i){let n=zBA[A.key]||A.key,o="";return i.indexOf("code.")>-1&&(n=A.code,o="code."),n==null||!n?!1:(n=n.toLowerCase(),n===" "?n="space":n==="."&&(n="dot"),UH.forEach(r=>{if(r!==n){let s=OBA[r];s(A)&&(o+=r+".")}}),o+=n,o===i)}static eventCallback(A,i,n){return o=>{t.matchEventFullKeyCode(o,A)&&n.runGuarded(()=>i(o))}}static _normalizeKey(A){return A==="esc"?"escape":A}static \u0275fac=function(i){return new(i||t)(we(at))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();function PBA(){Vp.makeCurrent()}function jBA(){return new aa}function qBA(){return xJ(document),document}var VBA=[{provide:og,useValue:zp},{provide:Ib,useValue:PBA,multi:!0},{provide:at,useFactory:qBA}],Wp=jb(uH,"browser",VBA);var ZBA=[{provide:xh,useClass:Zp},{provide:Gb,useClass:Lp,deps:[Qe,Fp,xh]},{provide:Lp,useClass:Lp,deps:[Qe,Fp,xh]}],WBA=[{provide:Cp,useValue:"root"},{provide:aa,useFactory:jBA},{provide:qp,useClass:KH,multi:!0,deps:[at]},{provide:qp,useClass:YH,multi:!0,deps:[at]},jh,oM,nM,{provide:ws,useExisting:jh},{provide:xI,useClass:HBA},[]],Vh=(()=>{class t{constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[...WBA,...ZBA],imports:[f0,fH]})}return t})();var AB=class{},Zh=class{},S2=class t{headers;normalizedNames=new Map;lazyInit;lazyUpdate=null;constructor(e){e?typeof e=="string"?this.lazyInit=()=>{this.headers=new Map,e.split(`
`).forEach(A=>{let i=A.indexOf(":");if(i>0){let n=A.slice(0,i),o=A.slice(i+1).trim();this.addHeaderEntry(n,o)}})}:typeof Headers<"u"&&e instanceof Headers?(this.headers=new Map,e.forEach((A,i)=>{this.addHeaderEntry(i,A)})):this.lazyInit=()=>{this.headers=new Map,Object.entries(e).forEach(([A,i])=>{this.setHeaderEntries(A,i)})}:this.headers=new Map}has(e){return this.init(),this.headers.has(e.toLowerCase())}get(e){this.init();let A=this.headers.get(e.toLowerCase());return A&&A.length>0?A[0]:null}keys(){return this.init(),Array.from(this.normalizedNames.values())}getAll(e){return this.init(),this.headers.get(e.toLowerCase())||null}append(e,A){return this.clone({name:e,value:A,op:"a"})}set(e,A){return this.clone({name:e,value:A,op:"s"})}delete(e,A){return this.clone({name:e,value:A,op:"d"})}maybeSetNormalizedName(e,A){this.normalizedNames.has(A)||this.normalizedNames.set(A,e)}init(){this.lazyInit&&(this.lazyInit instanceof t?this.copyFrom(this.lazyInit):this.lazyInit(),this.lazyInit=null,this.lazyUpdate&&(this.lazyUpdate.forEach(e=>this.applyUpdate(e)),this.lazyUpdate=null))}copyFrom(e){e.init(),Array.from(e.headers.keys()).forEach(A=>{this.headers.set(A,e.headers.get(A)),this.normalizedNames.set(A,e.normalizedNames.get(A))})}clone(e){let A=new t;return A.lazyInit=this.lazyInit&&this.lazyInit instanceof t?this.lazyInit:this,A.lazyUpdate=(this.lazyUpdate||[]).concat([e]),A}applyUpdate(e){let A=e.name.toLowerCase();switch(e.op){case"a":case"s":let i=e.value;if(typeof i=="string"&&(i=[i]),i.length===0)return;this.maybeSetNormalizedName(e.name,A);let n=(e.op==="a"?this.headers.get(A):void 0)||[];n.push(...i),this.headers.set(A,n);break;case"d":let o=e.value;if(!o)this.headers.delete(A),this.normalizedNames.delete(A);else{let r=this.headers.get(A);if(!r)return;r=r.filter(s=>o.indexOf(s)===-1),r.length===0?(this.headers.delete(A),this.normalizedNames.delete(A)):this.headers.set(A,r)}break}}addHeaderEntry(e,A){let i=e.toLowerCase();this.maybeSetNormalizedName(e,i),this.headers.has(i)?this.headers.get(i).push(A):this.headers.set(i,[A])}setHeaderEntries(e,A){let i=(Array.isArray(A)?A:[A]).map(o=>o.toString()),n=e.toLowerCase();this.headers.set(n,i),this.maybeSetNormalizedName(e,n)}forEach(e){this.init(),Array.from(this.normalizedNames.keys()).forEach(A=>e(this.normalizedNames.get(A),this.headers.get(A)))}};var $p=class{encodeKey(e){return JH(e)}encodeValue(e){return JH(e)}decodeKey(e){return decodeURIComponent(e)}decodeValue(e){return decodeURIComponent(e)}};function XBA(t,e){let A=new Map;return t.length>0&&t.replace(/^\?/,"").split("&").forEach(n=>{let o=n.indexOf("="),[r,s]=o==-1?[e.decodeKey(n),""]:[e.decodeKey(n.slice(0,o)),e.decodeValue(n.slice(o+1))],a=A.get(r)||[];a.push(s),A.set(r,a)}),A}var $BA=/%(\d[a-f0-9])/gi,AEA={40:"@","3A":":",24:"$","2C":",","3B":";","3D":"=","3F":"?","2F":"/"};function JH(t){return encodeURIComponent(t).replace($BA,(e,A)=>AEA[A]??e)}function Xp(t){return`${t}`}var m0=class t{map;encoder;updates=null;cloneFrom=null;constructor(e={}){if(this.encoder=e.encoder||new $p,e.fromString){if(e.fromObject)throw new Ae(2805,!1);this.map=XBA(e.fromString,this.encoder)}else e.fromObject?(this.map=new Map,Object.keys(e.fromObject).forEach(A=>{let i=e.fromObject[A],n=Array.isArray(i)?i.map(Xp):[Xp(i)];this.map.set(A,n)})):this.map=null}has(e){return this.init(),this.map.has(e)}get(e){this.init();let A=this.map.get(e);return A?A[0]:null}getAll(e){return this.init(),this.map.get(e)||null}keys(){return this.init(),Array.from(this.map.keys())}append(e,A){return this.clone({param:e,value:A,op:"a"})}appendAll(e){let A=[];return Object.keys(e).forEach(i=>{let n=e[i];Array.isArray(n)?n.forEach(o=>{A.push({param:i,value:o,op:"a"})}):A.push({param:i,value:n,op:"a"})}),this.clone(A)}set(e,A){return this.clone({param:e,value:A,op:"s"})}delete(e,A){return this.clone({param:e,value:A,op:"d"})}toString(){return this.init(),this.keys().map(e=>{let A=this.encoder.encodeKey(e);return this.map.get(e).map(i=>A+"="+this.encoder.encodeValue(i)).join("&")}).filter(e=>e!=="").join("&")}clone(e){let A=new t({encoder:this.encoder});return A.cloneFrom=this.cloneFrom||this,A.updates=(this.updates||[]).concat(e),A}init(){this.map===null&&(this.map=new Map),this.cloneFrom!==null&&(this.cloneFrom.init(),this.cloneFrom.keys().forEach(e=>this.map.set(e,this.cloneFrom.map.get(e))),this.updates.forEach(e=>{switch(e.op){case"a":case"s":let A=(e.op==="a"?this.map.get(e.param):void 0)||[];A.push(Xp(e.value)),this.map.set(e.param,A);break;case"d":if(e.value!==void 0){let i=this.map.get(e.param)||[],n=i.indexOf(Xp(e.value));n!==-1&&i.splice(n,1),i.length>0?this.map.set(e.param,i):this.map.delete(e.param)}else{this.map.delete(e.param);break}}}),this.cloneFrom=this.updates=null)}};var A6=class{map=new Map;set(e,A){return this.map.set(e,A),this}get(e){return this.map.has(e)||this.map.set(e,e.defaultValue()),this.map.get(e)}delete(e){return this.map.delete(e),this}has(e){return this.map.has(e)}keys(){return this.map.keys()}};function eEA(t){switch(t){case"DELETE":case"GET":case"HEAD":case"OPTIONS":case"JSONP":return!1;default:return!0}}function TH(t){return typeof ArrayBuffer<"u"&&t instanceof ArrayBuffer}function HH(t){return typeof Blob<"u"&&t instanceof Blob}function zH(t){return typeof FormData<"u"&&t instanceof FormData}function tEA(t){return typeof URLSearchParams<"u"&&t instanceof URLSearchParams}var OH="Content-Type",PH="Accept",qH="X-Request-URL",VH="text/plain",ZH="application/json",iEA=`${ZH}, ${VH}, */*`,$d=class t{url;body=null;headers;context;reportProgress=!1;withCredentials=!1;responseType="json";method;params;urlWithParams;transferCache;constructor(e,A,i,n){this.url=A,this.method=e.toUpperCase();let o;if(eEA(this.method)||n?(this.body=i!==void 0?i:null,o=n):o=i,o&&(this.reportProgress=!!o.reportProgress,this.withCredentials=!!o.withCredentials,o.responseType&&(this.responseType=o.responseType),o.headers&&(this.headers=o.headers),o.context&&(this.context=o.context),o.params&&(this.params=o.params),this.transferCache=o.transferCache),this.headers??=new S2,this.context??=new A6,!this.params)this.params=new m0,this.urlWithParams=A;else{let r=this.params.toString();if(r.length===0)this.urlWithParams=A;else{let s=A.indexOf("?"),a=s===-1?"?":s<A.length-1?"&":"";this.urlWithParams=A+a+r}}}serializeBody(){return this.body===null?null:typeof this.body=="string"||TH(this.body)||HH(this.body)||zH(this.body)||tEA(this.body)?this.body:this.body instanceof m0?this.body.toString():typeof this.body=="object"||typeof this.body=="boolean"||Array.isArray(this.body)?JSON.stringify(this.body):this.body.toString()}detectContentTypeHeader(){return this.body===null||zH(this.body)?null:HH(this.body)?this.body.type||null:TH(this.body)?null:typeof this.body=="string"?VH:this.body instanceof m0?"application/x-www-form-urlencoded;charset=UTF-8":typeof this.body=="object"||typeof this.body=="number"||typeof this.body=="boolean"?ZH:null}clone(e={}){let A=e.method||this.method,i=e.url||this.url,n=e.responseType||this.responseType,o=e.transferCache??this.transferCache,r=e.body!==void 0?e.body:this.body,s=e.withCredentials??this.withCredentials,a=e.reportProgress??this.reportProgress,c=e.headers||this.headers,l=e.params||this.params,I=e.context??this.context;return e.setHeaders!==void 0&&(c=Object.keys(e.setHeaders).reduce((C,d)=>C.set(d,e.setHeaders[d]),c)),e.setParams&&(l=Object.keys(e.setParams).reduce((C,d)=>C.set(d,e.setParams[d]),l)),new t(A,i,r,{params:l,headers:c,context:I,reportProgress:a,responseType:n,withCredentials:s,transferCache:o})}},LI=function(t){return t[t.Sent=0]="Sent",t[t.UploadProgress=1]="UploadProgress",t[t.ResponseHeader=2]="ResponseHeader",t[t.DownloadProgress=3]="DownloadProgress",t[t.Response=4]="Response",t[t.User=5]="User",t}(LI||{}),eB=class{headers;status;statusText;url;ok;type;constructor(e,A=200,i="OK"){this.headers=e.headers||new S2,this.status=e.status!==void 0?e.status:A,this.statusText=e.statusText||i,this.url=e.url||null,this.ok=this.status>=200&&this.status<300}},e6=class t extends eB{constructor(e={}){super(e)}type=LI.ResponseHeader;clone(e={}){return new t({headers:e.headers||this.headers,status:e.status!==void 0?e.status:this.status,statusText:e.statusText||this.statusText,url:e.url||this.url||void 0})}},Wh=class t extends eB{body;constructor(e={}){super(e),this.body=e.body!==void 0?e.body:null}type=LI.Response;clone(e={}){return new t({body:e.body!==void 0?e.body:this.body,headers:e.headers||this.headers,status:e.status!==void 0?e.status:this.status,statusText:e.statusText||this.statusText,url:e.url||this.url||void 0})}},Xh=class extends eB{name="HttpErrorResponse";message;error;ok=!1;constructor(e){super(e,0,"Unknown Error"),this.status>=200&&this.status<300?this.message=`Http failure during parsing for ${e.url||"(unknown url)"}`:this.message=`Http failure response for ${e.url||"(unknown url)"}: ${e.status} ${e.statusText}`,this.error=e.error||null}},nEA=200,oEA=204;function sM(t,e){return{body:e,headers:t.headers,context:t.context,observe:t.observe,params:t.params,reportProgress:t.reportProgress,responseType:t.responseType,withCredentials:t.withCredentials,transferCache:t.transferCache}}var Ds=(()=>{class t{handler;constructor(A){this.handler=A}request(A,i,n={}){let o;if(A instanceof $d)o=A;else{let a;n.headers instanceof S2?a=n.headers:a=new S2(n.headers);let c;n.params&&(n.params instanceof m0?c=n.params:c=new m0({fromObject:n.params})),o=new $d(A,i,n.body!==void 0?n.body:null,{headers:a,context:n.context,params:c,reportProgress:n.reportProgress,responseType:n.responseType||"json",withCredentials:n.withCredentials,transferCache:n.transferCache})}let r=Me(o).pipe(ql(a=>this.handler.handle(a)));if(A instanceof $d||n.observe==="events")return r;let s=r.pipe(kt(a=>a instanceof Wh));switch(n.observe||"body"){case"body":switch(o.responseType){case"arraybuffer":return s.pipe(je(a=>{if(a.body!==null&&!(a.body instanceof ArrayBuffer))throw new Ae(2806,!1);return a.body}));case"blob":return s.pipe(je(a=>{if(a.body!==null&&!(a.body instanceof Blob))throw new Ae(2807,!1);return a.body}));case"text":return s.pipe(je(a=>{if(a.body!==null&&typeof a.body!="string")throw new Ae(2808,!1);return a.body}));case"json":default:return s.pipe(je(a=>a.body))}case"response":return s;default:throw new Ae(2809,!1)}}delete(A,i={}){return this.request("DELETE",A,i)}get(A,i={}){return this.request("GET",A,i)}head(A,i={}){return this.request("HEAD",A,i)}jsonp(A,i){return this.request("JSONP",A,{params:new m0().append(i,"JSONP_CALLBACK"),observe:"body",responseType:"json"})}options(A,i={}){return this.request("OPTIONS",A,i)}patch(A,i,n={}){return this.request("PATCH",A,sM(n,i))}post(A,i,n={}){return this.request("POST",A,sM(n,i))}put(A,i,n={}){return this.request("PUT",A,sM(n,i))}static \u0275fac=function(i){return new(i||t)(we(AB))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();var rEA=new dA("");function WH(t,e){return e(t)}function sEA(t,e){return(A,i)=>e.intercept(A,{handle:n=>t(n,i)})}function aEA(t,e,A){return(i,n)=>ga(A,()=>e(i,o=>t(o,n)))}var XH=new dA(""),cM=new dA(""),$H=new dA(""),lM=new dA("",{providedIn:"root",factory:()=>!0});function cEA(){let t=null;return(e,A)=>{t===null&&(t=(f(XH,{optional:!0})??[]).reduceRight(sEA,WH));let i=f(B0);if(f(lM)){let o=i.add();return t(e,A).pipe(Vl(()=>i.remove(o)))}else return t(e,A)}}var t6=(()=>{class t extends AB{backend;injector;chain=null;pendingTasks=f(B0);contributeToStability=f(lM);constructor(A,i){super(),this.backend=A,this.injector=i}handle(A){if(this.chain===null){let i=Array.from(new Set([...this.injector.get(cM),...this.injector.get($H,[])]));this.chain=i.reduceRight((n,o)=>aEA(n,o,this.injector),WH)}if(this.contributeToStability){let i=this.pendingTasks.add();return this.chain(A,n=>this.backend.handle(n)).pipe(Vl(()=>this.pendingTasks.remove(i)))}else return this.chain(A,i=>this.backend.handle(i))}static \u0275fac=function(i){return new(i||t)(we(Zh),we(pr))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();var lEA=/^\)\]\}',?\n/,gEA=RegExp(`^${qH}:`,"m");function IEA(t){return"responseURL"in t&&t.responseURL?t.responseURL:gEA.test(t.getAllResponseHeaders())?t.getResponseHeader(qH):null}var aM=(()=>{class t{xhrFactory;constructor(A){this.xhrFactory=A}handle(A){if(A.method==="JSONP")throw new Ae(-2800,!1);let i=this.xhrFactory;return(i.\u0275loadImpl?oo(i.\u0275loadImpl()):Me(null)).pipe(jn(()=>new ct(o=>{let r=i.build();if(r.open(A.method,A.urlWithParams),A.withCredentials&&(r.withCredentials=!0),A.headers.forEach((E,h)=>r.setRequestHeader(E,h.join(","))),A.headers.has(PH)||r.setRequestHeader(PH,iEA),!A.headers.has(OH)){let E=A.detectContentTypeHeader();E!==null&&r.setRequestHeader(OH,E)}if(A.responseType){let E=A.responseType.toLowerCase();r.responseType=E!=="json"?E:"text"}let s=A.serializeBody(),a=null,c=()=>{if(a!==null)return a;let E=r.statusText||"OK",h=new S2(r.getAllResponseHeaders()),u=IEA(r)||A.url;return a=new e6({headers:h,status:r.status,statusText:E,url:u}),a},l=()=>{let{headers:E,status:h,statusText:u,url:D}=c(),L=null;h!==oEA&&(L=typeof r.response>"u"?r.responseText:r.response),h===0&&(h=L?nEA:0);let R=h>=200&&h<300;if(A.responseType==="json"&&typeof L=="string"){let w=L;L=L.replace(lEA,"");try{L=L!==""?JSON.parse(L):null}catch(_){L=w,R&&(R=!1,L={error:_,text:L})}}R?(o.next(new Wh({body:L,headers:E,status:h,statusText:u,url:D||void 0})),o.complete()):o.error(new Xh({error:L,headers:E,status:h,statusText:u,url:D||void 0}))},I=E=>{let{url:h}=c(),u=new Xh({error:E,status:r.status||0,statusText:r.statusText||"Unknown Error",url:h||void 0});o.error(u)},C=!1,d=E=>{C||(o.next(c()),C=!0);let h={type:LI.DownloadProgress,loaded:E.loaded};E.lengthComputable&&(h.total=E.total),A.responseType==="text"&&r.responseText&&(h.partialText=r.responseText),o.next(h)},B=E=>{let h={type:LI.UploadProgress,loaded:E.loaded};E.lengthComputable&&(h.total=E.total),o.next(h)};return r.addEventListener("load",l),r.addEventListener("error",I),r.addEventListener("timeout",I),r.addEventListener("abort",I),A.reportProgress&&(r.addEventListener("progress",d),s!==null&&r.upload&&r.upload.addEventListener("progress",B)),r.send(s),o.next({type:LI.Sent}),()=>{r.removeEventListener("error",I),r.removeEventListener("abort",I),r.removeEventListener("load",l),r.removeEventListener("timeout",I),A.reportProgress&&(r.removeEventListener("progress",d),s!==null&&r.upload&&r.upload.removeEventListener("progress",B)),r.readyState!==r.DONE&&r.abort()}})))}static \u0275fac=function(i){return new(i||t)(we(xI))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),Az=new dA(""),CEA="XSRF-TOKEN",dEA=new dA("",{providedIn:"root",factory:()=>CEA}),BEA="X-XSRF-TOKEN",EEA=new dA("",{providedIn:"root",factory:()=>BEA}),$h=class{},QEA=(()=>{class t{doc;cookieName;lastCookieString="";lastToken=null;parseCount=0;constructor(A,i){this.doc=A,this.cookieName=i}getToken(){let A=this.doc.cookie||"";return A!==this.lastCookieString&&(this.parseCount++,this.lastToken=Hh(A,this.cookieName),this.lastCookieString=A),this.lastToken}static \u0275fac=function(i){return new(i||t)(we(at),we(dEA))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();function hEA(t,e){let A=t.url.toLowerCase();if(!f(Az)||t.method==="GET"||t.method==="HEAD"||A.startsWith("http://")||A.startsWith("https://"))return e(t);let i=f($h).getToken(),n=f(EEA);return i!=null&&!t.headers.has(n)&&(t=t.clone({headers:t.headers.set(n,i)})),e(t)}var gM=function(t){return t[t.Interceptors=0]="Interceptors",t[t.LegacyInterceptors=1]="LegacyInterceptors",t[t.CustomXsrfConfiguration=2]="CustomXsrfConfiguration",t[t.NoXsrfProtection=3]="NoXsrfProtection",t[t.JsonpSupport=4]="JsonpSupport",t[t.RequestsMadeViaParent=5]="RequestsMadeViaParent",t[t.Fetch=6]="Fetch",t}(gM||{});function uEA(t,e){return{\u0275kind:t,\u0275providers:e}}function ez(...t){let e=[Ds,aM,t6,{provide:AB,useExisting:t6},{provide:Zh,useFactory:()=>f(rEA,{optional:!0})??f(aM)},{provide:cM,useValue:hEA,multi:!0},{provide:Az,useValue:!0},{provide:$h,useClass:QEA}];for(let A of t)e.push(...A.\u0275providers);return ph(e)}var jH=new dA("");function tz(){return uEA(gM.LegacyInterceptors,[{provide:jH,useFactory:cEA},{provide:cM,useExisting:jH,multi:!0}])}var IM=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[ez(tz())]})}return t})();var iz=(()=>{class t{_doc;constructor(A){this._doc=A}getTitle(){return this._doc.title}setTitle(A){this._doc.title=A||""}static \u0275fac=function(i){return new(i||t)(we(at))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var cl=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:function(i){let n=null;return i?n=new(i||t):n=we(fEA),n},providedIn:"root"})}return t})(),fEA=(()=>{class t extends cl{_doc;constructor(A){super(),this._doc=A}sanitize(A,i){if(i==null)return null;switch(A){case jr.NONE:return i;case jr.HTML:return w2(i,"HTML")?sl(i):Eb(this._doc,String(i)).toString();case jr.STYLE:return w2(i,"Style")?sl(i):i;case jr.SCRIPT:if(w2(i,"Script"))return sl(i);throw new Ae(5200,!1);case jr.URL:return w2(i,"URL")?sl(i):pp(String(i));case jr.RESOURCE_URL:if(w2(i,"ResourceURL"))return sl(i);throw new Ae(5201,!1);default:throw new Ae(5202,!1)}}bypassSecurityTrustHtml(A){return KJ(A)}bypassSecurityTrustStyle(A){return YJ(A)}bypassSecurityTrustScript(A){return JJ(A)}bypassSecurityTrustUrl(A){return TJ(A)}bypassSecurityTrustResourceUrl(A){return HJ(A)}static \u0275fac=function(i){return new(i||t)(we(at))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var gz=(()=>{class t{_renderer;_elementRef;onChange=A=>{};onTouched=()=>{};constructor(A,i){this._renderer=A,this._elementRef=i}setProperty(A,i){this._renderer.setProperty(this._elementRef.nativeElement,A,i)}registerOnTouched(A){this.onTouched=A}registerOnChange(A){this.onChange=A}setDisabledState(A){this.setProperty("disabled",A)}static \u0275fac=function(i){return new(i||t)(PA(Wi),PA(ee))};static \u0275dir=jA({type:t})}return t})(),mEA=(()=>{class t extends gz{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,features:[lt]})}return t})(),yc=new dA("");var pEA={provide:yc,useExisting:Ir(()=>vc),multi:!0};function wEA(){let t=Wa()?Wa().getUserAgent():"";return/android (\d+)/.test(t.toLowerCase())}var DEA=new dA(""),vc=(()=>{class t extends gz{_compositionMode;_composing=!1;constructor(A,i,n){super(A,i),this._compositionMode=n,this._compositionMode==null&&(this._compositionMode=!wEA())}writeValue(A){let i=A??"";this.setProperty("value",i)}_handleInput(A){(!this._compositionMode||this._compositionMode&&!this._composing)&&this.onChange(A)}_compositionStart(){this._composing=!0}_compositionEnd(A){this._composing=!1,this._compositionMode&&this.onChange(A)}static \u0275fac=function(i){return new(i||t)(PA(Wi),PA(ee),PA(DEA,8))};static \u0275dir=jA({type:t,selectors:[["input","formControlName","",3,"type","checkbox"],["textarea","formControlName",""],["input","formControl","",3,"type","checkbox"],["textarea","formControl",""],["input","ngModel","",3,"type","checkbox"],["textarea","ngModel",""],["","ngDefaultControl",""]],hostBindings:function(i,n){i&1&&hA("input",function(r){return n._handleInput(r.target.value)})("blur",function(){return n.onTouched()})("compositionstart",function(){return n._compositionStart()})("compositionend",function(r){return n._compositionEnd(r.target.value)})},standalone:!1,features:[ht([pEA]),lt]})}return t})();function EM(t){return t==null||QM(t)===0}function QM(t){return t==null?null:Array.isArray(t)||typeof t=="string"?t.length:t instanceof Set?t.size:null}var w0=new dA(""),ru=new dA(""),yEA=/^(?=.{1,254}$)(?=.{1,64}@)[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+(?:\.[a-zA-Z0-9!#$%&'*+/=?^_`{|}~-]+)*@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$/,$a=class{static min(e){return vEA(e)}static max(e){return bEA(e)}static required(e){return MEA(e)}static requiredTrue(e){return kEA(e)}static email(e){return SEA(e)}static minLength(e){return REA(e)}static maxLength(e){return xEA(e)}static pattern(e){return LEA(e)}static nullValidator(e){return Iz()}static compose(e){return hz(e)}static composeAsync(e){return uz(e)}};function vEA(t){return e=>{if(e.value==null||t==null)return null;let A=parseFloat(e.value);return!isNaN(A)&&A<t?{min:{min:t,actual:e.value}}:null}}function bEA(t){return e=>{if(e.value==null||t==null)return null;let A=parseFloat(e.value);return!isNaN(A)&&A>t?{max:{max:t,actual:e.value}}:null}}function MEA(t){return EM(t.value)?{required:!0}:null}function kEA(t){return t.value===!0?null:{required:!0}}function SEA(t){return EM(t.value)||yEA.test(t.value)?null:{email:!0}}function REA(t){return e=>{let A=e.value?.length??QM(e.value);return A===null||A===0?null:A<t?{minlength:{requiredLength:t,actualLength:A}}:null}}function xEA(t){return e=>{let A=e.value?.length??QM(e.value);return A!==null&&A>t?{maxlength:{requiredLength:t,actualLength:A}}:null}}function LEA(t){if(!t)return Iz;let e,A;return typeof t=="string"?(A="",t.charAt(0)!=="^"&&(A+="^"),A+=t,t.charAt(t.length-1)!=="$"&&(A+="$"),e=new RegExp(A)):(A=t.toString(),e=t),i=>{if(EM(i.value))return null;let n=i.value;return e.test(n)?null:{pattern:{requiredPattern:A,actualValue:n}}}}function Iz(t){return null}function Cz(t){return t!=null}function dz(t){return D2(t)?oo(t):t}function Bz(t){let e={};return t.forEach(A=>{e=A!=null?rA(rA({},e),A):e}),Object.keys(e).length===0?null:e}function Ez(t,e){return e.map(A=>A(t))}function FEA(t){return!t.validate}function Qz(t){return t.map(e=>FEA(e)?e:A=>e.validate(A))}function hz(t){if(!t)return null;let e=t.filter(Cz);return e.length==0?null:function(A){return Bz(Ez(A,e))}}function hM(t){return t!=null?hz(Qz(t)):null}function uz(t){if(!t)return null;let e=t.filter(Cz);return e.length==0?null:function(A){let i=Ez(A,e).map(dz);return nh(i).pipe(je(Bz))}}function uM(t){return t!=null?uz(Qz(t)):null}function nz(t,e){return t===null?[e]:Array.isArray(t)?[...t,e]:[t,e]}function fz(t){return t._rawValidators}function mz(t){return t._rawAsyncValidators}function CM(t){return t?Array.isArray(t)?t:[t]:[]}function n6(t,e){return Array.isArray(t)?t.includes(e):t===e}function oz(t,e){let A=CM(e);return CM(t).forEach(n=>{n6(A,n)||A.push(n)}),A}function rz(t,e){return CM(e).filter(A=>!n6(t,A))}var o6=class{get value(){return this.control?this.control.value:null}get valid(){return this.control?this.control.valid:null}get invalid(){return this.control?this.control.invalid:null}get pending(){return this.control?this.control.pending:null}get disabled(){return this.control?this.control.disabled:null}get enabled(){return this.control?this.control.enabled:null}get errors(){return this.control?this.control.errors:null}get pristine(){return this.control?this.control.pristine:null}get dirty(){return this.control?this.control.dirty:null}get touched(){return this.control?this.control.touched:null}get status(){return this.control?this.control.status:null}get untouched(){return this.control?this.control.untouched:null}get statusChanges(){return this.control?this.control.statusChanges:null}get valueChanges(){return this.control?this.control.valueChanges:null}get path(){return null}_composedValidatorFn;_composedAsyncValidatorFn;_rawValidators=[];_rawAsyncValidators=[];_setValidators(e){this._rawValidators=e||[],this._composedValidatorFn=hM(this._rawValidators)}_setAsyncValidators(e){this._rawAsyncValidators=e||[],this._composedAsyncValidatorFn=uM(this._rawAsyncValidators)}get validator(){return this._composedValidatorFn||null}get asyncValidator(){return this._composedAsyncValidatorFn||null}_onDestroyCallbacks=[];_registerOnDestroy(e){this._onDestroyCallbacks.push(e)}_invokeOnDestroyCallbacks(){this._onDestroyCallbacks.forEach(e=>e()),this._onDestroyCallbacks=[]}reset(e=void 0){this.control&&this.control.reset(e)}hasError(e,A){return this.control?this.control.hasError(e,A):!1}getError(e,A){return this.control?this.control.getError(e,A):null}},p0=class extends o6{name;get formDirective(){return null}get path(){return null}},Ac=class extends o6{_parent=null;name=null;valueAccessor=null},r6=class{_cd;constructor(e){this._cd=e}get isTouched(){return this._cd?.control?._touched?.(),!!this._cd?.control?.touched}get isUntouched(){return!!this._cd?.control?.untouched}get isPristine(){return this._cd?.control?._pristine?.(),!!this._cd?.control?.pristine}get isDirty(){return!!this._cd?.control?.dirty}get isValid(){return this._cd?.control?._status?.(),!!this._cd?.control?.valid}get isInvalid(){return!!this._cd?.control?.invalid}get isPending(){return!!this._cd?.control?.pending}get isSubmitted(){return this._cd?._submitted?.(),!!this._cd?.submitted}},NEA={"[class.ng-untouched]":"isUntouched","[class.ng-touched]":"isTouched","[class.ng-pristine]":"isPristine","[class.ng-dirty]":"isDirty","[class.ng-valid]":"isValid","[class.ng-invalid]":"isInvalid","[class.ng-pending]":"isPending"},Bse=Ye(rA({},NEA),{"[class.ng-submitted]":"isSubmitted"}),Ea=(()=>{class t extends r6{constructor(A){super(A)}static \u0275fac=function(i){return new(i||t)(PA(Ac,2))};static \u0275dir=jA({type:t,selectors:[["","formControlName",""],["","ngModel",""],["","formControl",""]],hostVars:14,hostBindings:function(i,n){i&2&&ue("ng-untouched",n.isUntouched)("ng-touched",n.isTouched)("ng-pristine",n.isPristine)("ng-dirty",n.isDirty)("ng-valid",n.isValid)("ng-invalid",n.isInvalid)("ng-pending",n.isPending)},standalone:!1,features:[lt]})}return t})(),pz=(()=>{class t extends r6{constructor(A){super(A)}static \u0275fac=function(i){return new(i||t)(PA(p0,10))};static \u0275dir=jA({type:t,selectors:[["","formGroupName",""],["","formArrayName",""],["","ngModelGroup",""],["","formGroup",""],["form",3,"ngNoForm",""],["","ngForm",""]],hostVars:16,hostBindings:function(i,n){i&2&&ue("ng-untouched",n.isUntouched)("ng-touched",n.isTouched)("ng-pristine",n.isPristine)("ng-dirty",n.isDirty)("ng-valid",n.isValid)("ng-invalid",n.isInvalid)("ng-pending",n.isPending)("ng-submitted",n.isSubmitted)},standalone:!1,features:[lt]})}return t})();var Au="VALID",i6="INVALID",tB="PENDING",eu="DISABLED",R2=class{},s6=class extends R2{value;source;constructor(e,A){super(),this.value=e,this.source=A}},iu=class extends R2{pristine;source;constructor(e,A){super(),this.pristine=e,this.source=A}},nu=class extends R2{touched;source;constructor(e,A){super(),this.touched=e,this.source=A}},iB=class extends R2{status;source;constructor(e,A){super(),this.status=e,this.source=A}},a6=class extends R2{source;constructor(e){super(),this.source=e}},c6=class extends R2{source;constructor(e){super(),this.source=e}};function fM(t){return(C6(t)?t.validators:t)||null}function _EA(t){return Array.isArray(t)?hM(t):t||null}function mM(t,e){return(C6(e)?e.asyncValidators:t)||null}function GEA(t){return Array.isArray(t)?uM(t):t||null}function C6(t){return t!=null&&!Array.isArray(t)&&typeof t=="object"}function wz(t,e,A){let i=t.controls;if(!(e?Object.keys(i):i).length)throw new Ae(1e3,"");if(!i[A])throw new Ae(1001,"")}function Dz(t,e,A){t._forEachChild((i,n)=>{if(A[n]===void 0)throw new Ae(1002,"")})}var nB=class{_pendingDirty=!1;_hasOwnPendingAsyncValidator=null;_pendingTouched=!1;_onCollectionChange=()=>{};_updateOn;_parent=null;_asyncValidationSubscription;_composedValidatorFn;_composedAsyncValidatorFn;_rawValidators;_rawAsyncValidators;value;constructor(e,A){this._assignValidators(e),this._assignAsyncValidators(A)}get validator(){return this._composedValidatorFn}set validator(e){this._rawValidators=this._composedValidatorFn=e}get asyncValidator(){return this._composedAsyncValidatorFn}set asyncValidator(e){this._rawAsyncValidators=this._composedAsyncValidatorFn=e}get parent(){return this._parent}get status(){return Ba(this.statusReactive)}set status(e){Ba(()=>this.statusReactive.set(e))}_status=h0(()=>this.statusReactive());statusReactive=Jo(void 0);get valid(){return this.status===Au}get invalid(){return this.status===i6}get pending(){return this.status==tB}get disabled(){return this.status===eu}get enabled(){return this.status!==eu}errors;get pristine(){return Ba(this.pristineReactive)}set pristine(e){Ba(()=>this.pristineReactive.set(e))}_pristine=h0(()=>this.pristineReactive());pristineReactive=Jo(!0);get dirty(){return!this.pristine}get touched(){return Ba(this.touchedReactive)}set touched(e){Ba(()=>this.touchedReactive.set(e))}_touched=h0(()=>this.touchedReactive());touchedReactive=Jo(!1);get untouched(){return!this.touched}_events=new OA;events=this._events.asObservable();valueChanges;statusChanges;get updateOn(){return this._updateOn?this._updateOn:this.parent?this.parent.updateOn:"change"}setValidators(e){this._assignValidators(e)}setAsyncValidators(e){this._assignAsyncValidators(e)}addValidators(e){this.setValidators(oz(e,this._rawValidators))}addAsyncValidators(e){this.setAsyncValidators(oz(e,this._rawAsyncValidators))}removeValidators(e){this.setValidators(rz(e,this._rawValidators))}removeAsyncValidators(e){this.setAsyncValidators(rz(e,this._rawAsyncValidators))}hasValidator(e){return n6(this._rawValidators,e)}hasAsyncValidator(e){return n6(this._rawAsyncValidators,e)}clearValidators(){this.validator=null}clearAsyncValidators(){this.asyncValidator=null}markAsTouched(e={}){let A=this.touched===!1;this.touched=!0;let i=e.sourceControl??this;this._parent&&!e.onlySelf&&this._parent.markAsTouched(Ye(rA({},e),{sourceControl:i})),A&&e.emitEvent!==!1&&this._events.next(new nu(!0,i))}markAllAsTouched(e={}){this.markAsTouched({onlySelf:!0,emitEvent:e.emitEvent,sourceControl:this}),this._forEachChild(A=>A.markAllAsTouched(e))}markAsUntouched(e={}){let A=this.touched===!0;this.touched=!1,this._pendingTouched=!1;let i=e.sourceControl??this;this._forEachChild(n=>{n.markAsUntouched({onlySelf:!0,emitEvent:e.emitEvent,sourceControl:i})}),this._parent&&!e.onlySelf&&this._parent._updateTouched(e,i),A&&e.emitEvent!==!1&&this._events.next(new nu(!1,i))}markAsDirty(e={}){let A=this.pristine===!0;this.pristine=!1;let i=e.sourceControl??this;this._parent&&!e.onlySelf&&this._parent.markAsDirty(Ye(rA({},e),{sourceControl:i})),A&&e.emitEvent!==!1&&this._events.next(new iu(!1,i))}markAsPristine(e={}){let A=this.pristine===!1;this.pristine=!0,this._pendingDirty=!1;let i=e.sourceControl??this;this._forEachChild(n=>{n.markAsPristine({onlySelf:!0,emitEvent:e.emitEvent})}),this._parent&&!e.onlySelf&&this._parent._updatePristine(e,i),A&&e.emitEvent!==!1&&this._events.next(new iu(!0,i))}markAsPending(e={}){this.status=tB;let A=e.sourceControl??this;e.emitEvent!==!1&&(this._events.next(new iB(this.status,A)),this.statusChanges.emit(this.status)),this._parent&&!e.onlySelf&&this._parent.markAsPending(Ye(rA({},e),{sourceControl:A}))}disable(e={}){let A=this._parentMarkedDirty(e.onlySelf);this.status=eu,this.errors=null,this._forEachChild(n=>{n.disable(Ye(rA({},e),{onlySelf:!0}))}),this._updateValue();let i=e.sourceControl??this;e.emitEvent!==!1&&(this._events.next(new s6(this.value,i)),this._events.next(new iB(this.status,i)),this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._updateAncestors(Ye(rA({},e),{skipPristineCheck:A}),this),this._onDisabledChange.forEach(n=>n(!0))}enable(e={}){let A=this._parentMarkedDirty(e.onlySelf);this.status=Au,this._forEachChild(i=>{i.enable(Ye(rA({},e),{onlySelf:!0}))}),this.updateValueAndValidity({onlySelf:!0,emitEvent:e.emitEvent}),this._updateAncestors(Ye(rA({},e),{skipPristineCheck:A}),this),this._onDisabledChange.forEach(i=>i(!1))}_updateAncestors(e,A){this._parent&&!e.onlySelf&&(this._parent.updateValueAndValidity(e),e.skipPristineCheck||this._parent._updatePristine({},A),this._parent._updateTouched({},A))}setParent(e){this._parent=e}getRawValue(){return this.value}updateValueAndValidity(e={}){if(this._setInitialStatus(),this._updateValue(),this.enabled){let i=this._cancelExistingSubscription();this.errors=this._runValidator(),this.status=this._calculateStatus(),(this.status===Au||this.status===tB)&&this._runAsyncValidator(i,e.emitEvent)}let A=e.sourceControl??this;e.emitEvent!==!1&&(this._events.next(new s6(this.value,A)),this._events.next(new iB(this.status,A)),this.valueChanges.emit(this.value),this.statusChanges.emit(this.status)),this._parent&&!e.onlySelf&&this._parent.updateValueAndValidity(Ye(rA({},e),{sourceControl:A}))}_updateTreeValidity(e={emitEvent:!0}){this._forEachChild(A=>A._updateTreeValidity(e)),this.updateValueAndValidity({onlySelf:!0,emitEvent:e.emitEvent})}_setInitialStatus(){this.status=this._allControlsDisabled()?eu:Au}_runValidator(){return this.validator?this.validator(this):null}_runAsyncValidator(e,A){if(this.asyncValidator){this.status=tB,this._hasOwnPendingAsyncValidator={emitEvent:A!==!1};let i=dz(this.asyncValidator(this));this._asyncValidationSubscription=i.subscribe(n=>{this._hasOwnPendingAsyncValidator=null,this.setErrors(n,{emitEvent:A,shouldHaveEmitted:e})})}}_cancelExistingSubscription(){if(this._asyncValidationSubscription){this._asyncValidationSubscription.unsubscribe();let e=this._hasOwnPendingAsyncValidator?.emitEvent??!1;return this._hasOwnPendingAsyncValidator=null,e}return!1}setErrors(e,A={}){this.errors=e,this._updateControlsErrors(A.emitEvent!==!1,this,A.shouldHaveEmitted)}get(e){let A=e;return A==null||(Array.isArray(A)||(A=A.split(".")),A.length===0)?null:A.reduce((i,n)=>i&&i._find(n),this)}getError(e,A){let i=A?this.get(A):this;return i&&i.errors?i.errors[e]:null}hasError(e,A){return!!this.getError(e,A)}get root(){let e=this;for(;e._parent;)e=e._parent;return e}_updateControlsErrors(e,A,i){this.status=this._calculateStatus(),e&&this.statusChanges.emit(this.status),(e||i)&&this._events.next(new iB(this.status,A)),this._parent&&this._parent._updateControlsErrors(e,A,i)}_initObservables(){this.valueChanges=new WA,this.statusChanges=new WA}_calculateStatus(){return this._allControlsDisabled()?eu:this.errors?i6:this._hasOwnPendingAsyncValidator||this._anyControlsHaveStatus(tB)?tB:this._anyControlsHaveStatus(i6)?i6:Au}_anyControlsHaveStatus(e){return this._anyControls(A=>A.status===e)}_anyControlsDirty(){return this._anyControls(e=>e.dirty)}_anyControlsTouched(){return this._anyControls(e=>e.touched)}_updatePristine(e,A){let i=!this._anyControlsDirty(),n=this.pristine!==i;this.pristine=i,this._parent&&!e.onlySelf&&this._parent._updatePristine(e,A),n&&this._events.next(new iu(this.pristine,A))}_updateTouched(e={},A){this.touched=this._anyControlsTouched(),this._events.next(new nu(this.touched,A)),this._parent&&!e.onlySelf&&this._parent._updateTouched(e,A)}_onDisabledChange=[];_registerOnCollectionChange(e){this._onCollectionChange=e}_setUpdateStrategy(e){C6(e)&&e.updateOn!=null&&(this._updateOn=e.updateOn)}_parentMarkedDirty(e){let A=this._parent&&this._parent.dirty;return!e&&!!A&&!this._parent._anyControlsDirty()}_find(e){return null}_assignValidators(e){this._rawValidators=Array.isArray(e)?e.slice():e,this._composedValidatorFn=_EA(this._rawValidators)}_assignAsyncValidators(e){this._rawAsyncValidators=Array.isArray(e)?e.slice():e,this._composedAsyncValidatorFn=GEA(this._rawAsyncValidators)}},oB=class extends nB{constructor(e,A,i){super(fM(A),mM(i,A)),this.controls=e,this._initObservables(),this._setUpdateStrategy(A),this._setUpControls(),this.updateValueAndValidity({onlySelf:!0,emitEvent:!!this.asyncValidator})}controls;registerControl(e,A){return this.controls[e]?this.controls[e]:(this.controls[e]=A,A.setParent(this),A._registerOnCollectionChange(this._onCollectionChange),A)}addControl(e,A,i={}){this.registerControl(e,A),this.updateValueAndValidity({emitEvent:i.emitEvent}),this._onCollectionChange()}removeControl(e,A={}){this.controls[e]&&this.controls[e]._registerOnCollectionChange(()=>{}),delete this.controls[e],this.updateValueAndValidity({emitEvent:A.emitEvent}),this._onCollectionChange()}setControl(e,A,i={}){this.controls[e]&&this.controls[e]._registerOnCollectionChange(()=>{}),delete this.controls[e],A&&this.registerControl(e,A),this.updateValueAndValidity({emitEvent:i.emitEvent}),this._onCollectionChange()}contains(e){return this.controls.hasOwnProperty(e)&&this.controls[e].enabled}setValue(e,A={}){Dz(this,!0,e),Object.keys(e).forEach(i=>{wz(this,!0,i),this.controls[i].setValue(e[i],{onlySelf:!0,emitEvent:A.emitEvent})}),this.updateValueAndValidity(A)}patchValue(e,A={}){e!=null&&(Object.keys(e).forEach(i=>{let n=this.controls[i];n&&n.patchValue(e[i],{onlySelf:!0,emitEvent:A.emitEvent})}),this.updateValueAndValidity(A))}reset(e={},A={}){this._forEachChild((i,n)=>{i.reset(e?e[n]:null,{onlySelf:!0,emitEvent:A.emitEvent})}),this._updatePristine(A,this),this._updateTouched(A,this),this.updateValueAndValidity(A)}getRawValue(){return this._reduceChildren({},(e,A,i)=>(e[i]=A.getRawValue(),e))}_syncPendingControls(){let e=this._reduceChildren(!1,(A,i)=>i._syncPendingControls()?!0:A);return e&&this.updateValueAndValidity({onlySelf:!0}),e}_forEachChild(e){Object.keys(this.controls).forEach(A=>{let i=this.controls[A];i&&e(i,A)})}_setUpControls(){this._forEachChild(e=>{e.setParent(this),e._registerOnCollectionChange(this._onCollectionChange)})}_updateValue(){this.value=this._reduceValue()}_anyControls(e){for(let[A,i]of Object.entries(this.controls))if(this.contains(A)&&e(i))return!0;return!1}_reduceValue(){let e={};return this._reduceChildren(e,(A,i,n)=>((i.enabled||this.disabled)&&(A[n]=i.value),A))}_reduceChildren(e,A){let i=e;return this._forEachChild((n,o)=>{i=A(i,n,o)}),i}_allControlsDisabled(){for(let e of Object.keys(this.controls))if(this.controls[e].enabled)return!1;return Object.keys(this.controls).length>0||this.disabled}_find(e){return this.controls.hasOwnProperty(e)?this.controls[e]:null}};var dM=class extends oB{};var rB=new dA("",{providedIn:"root",factory:()=>d6}),d6="always";function yz(t,e){return[...e.path,t]}function ou(t,e,A=d6){pM(t,e),e.valueAccessor.writeValue(t.value),(t.disabled||A==="always")&&e.valueAccessor.setDisabledState?.(t.disabled),KEA(t,e),JEA(t,e),YEA(t,e),UEA(t,e)}function l6(t,e,A=!0){let i=()=>{};e.valueAccessor&&(e.valueAccessor.registerOnChange(i),e.valueAccessor.registerOnTouched(i)),I6(t,e),t&&(e._invokeOnDestroyCallbacks(),t._registerOnCollectionChange(()=>{}))}function g6(t,e){t.forEach(A=>{A.registerOnValidatorChange&&A.registerOnValidatorChange(e)})}function UEA(t,e){if(e.valueAccessor.setDisabledState){let A=i=>{e.valueAccessor.setDisabledState(i)};t.registerOnDisabledChange(A),e._registerOnDestroy(()=>{t._unregisterOnDisabledChange(A)})}}function pM(t,e){let A=fz(t);e.validator!==null?t.setValidators(nz(A,e.validator)):typeof A=="function"&&t.setValidators([A]);let i=mz(t);e.asyncValidator!==null?t.setAsyncValidators(nz(i,e.asyncValidator)):typeof i=="function"&&t.setAsyncValidators([i]);let n=()=>t.updateValueAndValidity();g6(e._rawValidators,n),g6(e._rawAsyncValidators,n)}function I6(t,e){let A=!1;if(t!==null){if(e.validator!==null){let n=fz(t);if(Array.isArray(n)&&n.length>0){let o=n.filter(r=>r!==e.validator);o.length!==n.length&&(A=!0,t.setValidators(o))}}if(e.asyncValidator!==null){let n=mz(t);if(Array.isArray(n)&&n.length>0){let o=n.filter(r=>r!==e.asyncValidator);o.length!==n.length&&(A=!0,t.setAsyncValidators(o))}}}let i=()=>{};return g6(e._rawValidators,i),g6(e._rawAsyncValidators,i),A}function KEA(t,e){e.valueAccessor.registerOnChange(A=>{t._pendingValue=A,t._pendingChange=!0,t._pendingDirty=!0,t.updateOn==="change"&&vz(t,e)})}function YEA(t,e){e.valueAccessor.registerOnTouched(()=>{t._pendingTouched=!0,t.updateOn==="blur"&&t._pendingChange&&vz(t,e),t.updateOn!=="submit"&&t.markAsTouched()})}function vz(t,e){t._pendingDirty&&t.markAsDirty(),t.setValue(t._pendingValue,{emitModelToViewChange:!1}),e.viewToModelUpdate(t._pendingValue),t._pendingChange=!1}function JEA(t,e){let A=(i,n)=>{e.valueAccessor.writeValue(i),n&&e.viewToModelUpdate(i)};t.registerOnChange(A),e._registerOnDestroy(()=>{t._unregisterOnChange(A)})}function bz(t,e){t==null,pM(t,e)}function TEA(t,e){return I6(t,e)}function wM(t,e){if(!t.hasOwnProperty("model"))return!1;let A=t.model;return A.isFirstChange()?!0:!Object.is(e,A.currentValue)}function HEA(t){return Object.getPrototypeOf(t.constructor)===mEA}function Mz(t,e){t._syncPendingControls(),e.forEach(A=>{let i=A.control;i.updateOn==="submit"&&i._pendingChange&&(A.viewToModelUpdate(i._pendingValue),i._pendingChange=!1)})}function DM(t,e){if(!e)return null;Array.isArray(e);let A,i,n;return e.forEach(o=>{o.constructor===vc?A=o:HEA(o)?i=o:n=o}),n||i||A||null}function zEA(t,e){let A=t.indexOf(e);A>-1&&t.splice(A,1)}var OEA={provide:p0,useExisting:Ir(()=>su)},tu=Promise.resolve(),su=(()=>{class t extends p0{callSetDisabledState;get submitted(){return Ba(this.submittedReactive)}_submitted=h0(()=>this.submittedReactive());submittedReactive=Jo(!1);_directives=new Set;form;ngSubmit=new WA;options;constructor(A,i,n){super(),this.callSetDisabledState=n,this.form=new oB({},hM(A),uM(i))}ngAfterViewInit(){this._setUpdateStrategy()}get formDirective(){return this}get control(){return this.form}get path(){return[]}get controls(){return this.form.controls}addControl(A){tu.then(()=>{let i=this._findContainer(A.path);A.control=i.registerControl(A.name,A.control),ou(A.control,A,this.callSetDisabledState),A.control.updateValueAndValidity({emitEvent:!1}),this._directives.add(A)})}getControl(A){return this.form.get(A.path)}removeControl(A){tu.then(()=>{let i=this._findContainer(A.path);i&&i.removeControl(A.name),this._directives.delete(A)})}addFormGroup(A){tu.then(()=>{let i=this._findContainer(A.path),n=new oB({});bz(n,A),i.registerControl(A.name,n),n.updateValueAndValidity({emitEvent:!1})})}removeFormGroup(A){tu.then(()=>{let i=this._findContainer(A.path);i&&i.removeControl(A.name)})}getFormGroup(A){return this.form.get(A.path)}updateModel(A,i){tu.then(()=>{this.form.get(A.path).setValue(i)})}setValue(A){this.control.setValue(A)}onSubmit(A){return this.submittedReactive.set(!0),Mz(this.form,this._directives),this.ngSubmit.emit(A),this.form._events.next(new a6(this.control)),A?.target?.method==="dialog"}onReset(){this.resetForm()}resetForm(A=void 0){this.form.reset(A),this.submittedReactive.set(!1),this.form._events.next(new c6(this.form))}_setUpdateStrategy(){this.options&&this.options.updateOn!=null&&(this.form._updateOn=this.options.updateOn)}_findContainer(A){return A.pop(),A.length?this.form.get(A):this.form}static \u0275fac=function(i){return new(i||t)(PA(w0,10),PA(ru,10),PA(rB,8))};static \u0275dir=jA({type:t,selectors:[["form",3,"ngNoForm","",3,"formGroup",""],["ng-form"],["","ngForm",""]],hostBindings:function(i,n){i&1&&hA("submit",function(r){return n.onSubmit(r)})("reset",function(){return n.onReset()})},inputs:{options:[0,"ngFormOptions","options"]},outputs:{ngSubmit:"ngSubmit"},exportAs:["ngForm"],standalone:!1,features:[ht([OEA]),lt]})}return t})();function sz(t,e){let A=t.indexOf(e);A>-1&&t.splice(A,1)}function az(t){return typeof t=="object"&&t!==null&&Object.keys(t).length===2&&"value"in t&&"disabled"in t}var _I=class extends nB{defaultValue=null;_onChange=[];_pendingValue;_pendingChange=!1;constructor(e=null,A,i){super(fM(A),mM(i,A)),this._applyFormState(e),this._setUpdateStrategy(A),this._initObservables(),this.updateValueAndValidity({onlySelf:!0,emitEvent:!!this.asyncValidator}),C6(A)&&(A.nonNullable||A.initialValueIsDefault)&&(az(e)?this.defaultValue=e.value:this.defaultValue=e)}setValue(e,A={}){this.value=this._pendingValue=e,this._onChange.length&&A.emitModelToViewChange!==!1&&this._onChange.forEach(i=>i(this.value,A.emitViewToModelChange!==!1)),this.updateValueAndValidity(A)}patchValue(e,A={}){this.setValue(e,A)}reset(e=this.defaultValue,A={}){this._applyFormState(e),this.markAsPristine(A),this.markAsUntouched(A),this.setValue(this.value,A),this._pendingChange=!1}_updateValue(){}_anyControls(e){return!1}_allControlsDisabled(){return this.disabled}registerOnChange(e){this._onChange.push(e)}_unregisterOnChange(e){sz(this._onChange,e)}registerOnDisabledChange(e){this._onDisabledChange.push(e)}_unregisterOnDisabledChange(e){sz(this._onDisabledChange,e)}_forEachChild(e){}_syncPendingControls(){return this.updateOn==="submit"&&(this._pendingDirty&&this.markAsDirty(),this._pendingTouched&&this.markAsTouched(),this._pendingChange)?(this.setValue(this._pendingValue,{onlySelf:!0,emitModelToViewChange:!1}),!0):!1}_applyFormState(e){az(e)?(this.value=this._pendingValue=e.value,e.disabled?this.disable({onlySelf:!0,emitEvent:!1}):this.enable({onlySelf:!0,emitEvent:!1})):this.value=this._pendingValue=e}};var PEA=t=>t instanceof _I;var jEA={provide:Ac,useExisting:Ir(()=>ec)},cz=Promise.resolve(),ec=(()=>{class t extends Ac{_changeDetectorRef;callSetDisabledState;control=new _I;static ngAcceptInputType_isDisabled;_registered=!1;viewModel;name="";isDisabled;model;options;update=new WA;constructor(A,i,n,o,r,s){super(),this._changeDetectorRef=r,this.callSetDisabledState=s,this._parent=A,this._setValidators(i),this._setAsyncValidators(n),this.valueAccessor=DM(this,o)}ngOnChanges(A){if(this._checkForErrors(),!this._registered||"name"in A){if(this._registered&&(this._checkName(),this.formDirective)){let i=A.name.previousValue;this.formDirective.removeControl({name:i,path:this._getPath(i)})}this._setUpControl()}"isDisabled"in A&&this._updateDisabled(A),wM(A,this.viewModel)&&(this._updateValue(this.model),this.viewModel=this.model)}ngOnDestroy(){this.formDirective&&this.formDirective.removeControl(this)}get path(){return this._getPath(this.name)}get formDirective(){return this._parent?this._parent.formDirective:null}viewToModelUpdate(A){this.viewModel=A,this.update.emit(A)}_setUpControl(){this._setUpdateStrategy(),this._isStandalone()?this._setUpStandalone():this.formDirective.addControl(this),this._registered=!0}_setUpdateStrategy(){this.options&&this.options.updateOn!=null&&(this.control._updateOn=this.options.updateOn)}_isStandalone(){return!this._parent||!!(this.options&&this.options.standalone)}_setUpStandalone(){ou(this.control,this,this.callSetDisabledState),this.control.updateValueAndValidity({emitEvent:!1})}_checkForErrors(){this._checkName()}_checkName(){this.options&&this.options.name&&(this.name=this.options.name),!this._isStandalone()&&this.name}_updateValue(A){cz.then(()=>{this.control.setValue(A,{emitViewToModelChange:!1}),this._changeDetectorRef?.markForCheck()})}_updateDisabled(A){let i=A.isDisabled.currentValue,n=i!==0&&ie(i);cz.then(()=>{n&&!this.control.disabled?this.control.disable():!n&&this.control.disabled&&this.control.enable(),this._changeDetectorRef?.markForCheck()})}_getPath(A){return this._parent?yz(A,this._parent):[A]}static \u0275fac=function(i){return new(i||t)(PA(p0,9),PA(w0,10),PA(ru,10),PA(yc,10),PA(It,8),PA(rB,8))};static \u0275dir=jA({type:t,selectors:[["","ngModel","",3,"formControlName","",3,"formControl",""]],inputs:{name:"name",isDisabled:[0,"disabled","isDisabled"],model:[0,"ngModel","model"],options:[0,"ngModelOptions","options"]},outputs:{update:"ngModelChange"},exportAs:["ngModel"],standalone:!1,features:[ht([jEA]),lt,ti]})}return t})();var kz=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["form",3,"ngNoForm","",3,"ngNativeValidate",""]],hostAttrs:["novalidate",""],standalone:!1})}return t})();var yM=new dA(""),qEA={provide:Ac,useExisting:Ir(()=>vM)},vM=(()=>{class t extends Ac{_ngModelWarningConfig;callSetDisabledState;viewModel;form;set isDisabled(A){}model;update=new WA;static _ngModelWarningSentOnce=!1;_ngModelWarningSent=!1;constructor(A,i,n,o,r){super(),this._ngModelWarningConfig=o,this.callSetDisabledState=r,this._setValidators(A),this._setAsyncValidators(i),this.valueAccessor=DM(this,n)}ngOnChanges(A){if(this._isControlChanged(A)){let i=A.form.previousValue;i&&l6(i,this,!1),ou(this.form,this,this.callSetDisabledState),this.form.updateValueAndValidity({emitEvent:!1})}wM(A,this.viewModel)&&(this.form.setValue(this.model),this.viewModel=this.model)}ngOnDestroy(){this.form&&l6(this.form,this,!1)}get path(){return[]}get control(){return this.form}viewToModelUpdate(A){this.viewModel=A,this.update.emit(A)}_isControlChanged(A){return A.hasOwnProperty("form")}static \u0275fac=function(i){return new(i||t)(PA(w0,10),PA(ru,10),PA(yc,10),PA(yM,8),PA(rB,8))};static \u0275dir=jA({type:t,selectors:[["","formControl",""]],inputs:{form:[0,"formControl","form"],isDisabled:[0,"disabled","isDisabled"],model:[0,"ngModel","model"]},outputs:{update:"ngModelChange"},exportAs:["ngForm"],standalone:!1,features:[ht([qEA]),lt,ti]})}return t})(),VEA={provide:p0,useExisting:Ir(()=>GI)},GI=(()=>{class t extends p0{callSetDisabledState;get submitted(){return Ba(this._submittedReactive)}set submitted(A){this._submittedReactive.set(A)}_submitted=h0(()=>this._submittedReactive());_submittedReactive=Jo(!1);_oldForm;_onCollectionChange=()=>this._updateDomValue();directives=[];form=null;ngSubmit=new WA;constructor(A,i,n){super(),this.callSetDisabledState=n,this._setValidators(A),this._setAsyncValidators(i)}ngOnChanges(A){A.hasOwnProperty("form")&&(this._updateValidators(),this._updateDomValue(),this._updateRegistrations(),this._oldForm=this.form)}ngOnDestroy(){this.form&&(I6(this.form,this),this.form._onCollectionChange===this._onCollectionChange&&this.form._registerOnCollectionChange(()=>{}))}get formDirective(){return this}get control(){return this.form}get path(){return[]}addControl(A){let i=this.form.get(A.path);return ou(i,A,this.callSetDisabledState),i.updateValueAndValidity({emitEvent:!1}),this.directives.push(A),i}getControl(A){return this.form.get(A.path)}removeControl(A){l6(A.control||null,A,!1),zEA(this.directives,A)}addFormGroup(A){this._setUpFormContainer(A)}removeFormGroup(A){this._cleanUpFormContainer(A)}getFormGroup(A){return this.form.get(A.path)}addFormArray(A){this._setUpFormContainer(A)}removeFormArray(A){this._cleanUpFormContainer(A)}getFormArray(A){return this.form.get(A.path)}updateModel(A,i){this.form.get(A.path).setValue(i)}onSubmit(A){return this._submittedReactive.set(!0),Mz(this.form,this.directives),this.ngSubmit.emit(A),this.form._events.next(new a6(this.control)),A?.target?.method==="dialog"}onReset(){this.resetForm()}resetForm(A=void 0){this.form.reset(A),this._submittedReactive.set(!1),this.form._events.next(new c6(this.form))}_updateDomValue(){this.directives.forEach(A=>{let i=A.control,n=this.form.get(A.path);i!==n&&(l6(i||null,A),PEA(n)&&(ou(n,A,this.callSetDisabledState),A.control=n))}),this.form._updateTreeValidity({emitEvent:!1})}_setUpFormContainer(A){let i=this.form.get(A.path);bz(i,A),i.updateValueAndValidity({emitEvent:!1})}_cleanUpFormContainer(A){if(this.form){let i=this.form.get(A.path);i&&TEA(i,A)&&i.updateValueAndValidity({emitEvent:!1})}}_updateRegistrations(){this.form._registerOnCollectionChange(this._onCollectionChange),this._oldForm&&this._oldForm._registerOnCollectionChange(()=>{})}_updateValidators(){pM(this.form,this),this._oldForm&&I6(this._oldForm,this)}static \u0275fac=function(i){return new(i||t)(PA(w0,10),PA(ru,10),PA(rB,8))};static \u0275dir=jA({type:t,selectors:[["","formGroup",""]],hostBindings:function(i,n){i&1&&hA("submit",function(r){return n.onSubmit(r)})("reset",function(){return n.onReset()})},inputs:{form:[0,"formGroup","form"]},outputs:{ngSubmit:"ngSubmit"},exportAs:["ngForm"],standalone:!1,features:[ht([VEA]),lt,ti]})}return t})();var ZEA={provide:Ac,useExisting:Ir(()=>bM)},bM=(()=>{class t extends Ac{_ngModelWarningConfig;_added=!1;viewModel;control;name=null;set isDisabled(A){}model;update=new WA;static _ngModelWarningSentOnce=!1;_ngModelWarningSent=!1;constructor(A,i,n,o,r){super(),this._ngModelWarningConfig=r,this._parent=A,this._setValidators(i),this._setAsyncValidators(n),this.valueAccessor=DM(this,o)}ngOnChanges(A){this._added||this._setUpControl(),wM(A,this.viewModel)&&(this.viewModel=this.model,this.formDirective.updateModel(this,this.model))}ngOnDestroy(){this.formDirective&&this.formDirective.removeControl(this)}viewToModelUpdate(A){this.viewModel=A,this.update.emit(A)}get path(){return yz(this.name==null?this.name:this.name.toString(),this._parent)}get formDirective(){return this._parent?this._parent.formDirective:null}_setUpControl(){this.control=this.formDirective.addControl(this),this._added=!0}static \u0275fac=function(i){return new(i||t)(PA(p0,13),PA(w0,10),PA(ru,10),PA(yc,10),PA(yM,8))};static \u0275dir=jA({type:t,selectors:[["","formControlName",""]],inputs:{name:[0,"formControlName","name"],isDisabled:[0,"disabled","isDisabled"],model:[0,"ngModel","model"]},outputs:{update:"ngModelChange"},standalone:!1,features:[ht([ZEA]),lt,ti]})}return t})();var Sz=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})(),BM=class extends nB{constructor(e,A,i){super(fM(A),mM(i,A)),this.controls=e,this._initObservables(),this._setUpdateStrategy(A),this._setUpControls(),this.updateValueAndValidity({onlySelf:!0,emitEvent:!!this.asyncValidator})}controls;at(e){return this.controls[this._adjustIndex(e)]}push(e,A={}){this.controls.push(e),this._registerControl(e),this.updateValueAndValidity({emitEvent:A.emitEvent}),this._onCollectionChange()}insert(e,A,i={}){this.controls.splice(e,0,A),this._registerControl(A),this.updateValueAndValidity({emitEvent:i.emitEvent})}removeAt(e,A={}){let i=this._adjustIndex(e);i<0&&(i=0),this.controls[i]&&this.controls[i]._registerOnCollectionChange(()=>{}),this.controls.splice(i,1),this.updateValueAndValidity({emitEvent:A.emitEvent})}setControl(e,A,i={}){let n=this._adjustIndex(e);n<0&&(n=0),this.controls[n]&&this.controls[n]._registerOnCollectionChange(()=>{}),this.controls.splice(n,1),A&&(this.controls.splice(n,0,A),this._registerControl(A)),this.updateValueAndValidity({emitEvent:i.emitEvent}),this._onCollectionChange()}get length(){return this.controls.length}setValue(e,A={}){Dz(this,!1,e),e.forEach((i,n)=>{wz(this,!1,n),this.at(n).setValue(i,{onlySelf:!0,emitEvent:A.emitEvent})}),this.updateValueAndValidity(A)}patchValue(e,A={}){e!=null&&(e.forEach((i,n)=>{this.at(n)&&this.at(n).patchValue(i,{onlySelf:!0,emitEvent:A.emitEvent})}),this.updateValueAndValidity(A))}reset(e=[],A={}){this._forEachChild((i,n)=>{i.reset(e[n],{onlySelf:!0,emitEvent:A.emitEvent})}),this._updatePristine(A,this),this._updateTouched(A,this),this.updateValueAndValidity(A)}getRawValue(){return this.controls.map(e=>e.getRawValue())}clear(e={}){this.controls.length<1||(this._forEachChild(A=>A._registerOnCollectionChange(()=>{})),this.controls.splice(0),this.updateValueAndValidity({emitEvent:e.emitEvent}))}_adjustIndex(e){return e<0?e+this.length:e}_syncPendingControls(){let e=this.controls.reduce((A,i)=>i._syncPendingControls()?!0:A,!1);return e&&this.updateValueAndValidity({onlySelf:!0}),e}_forEachChild(e){this.controls.forEach((A,i)=>{e(A,i)})}_updateValue(){this.value=this.controls.filter(e=>e.enabled||this.disabled).map(e=>e.value)}_anyControls(e){return this.controls.some(A=>A.enabled&&e(A))}_setUpControls(){this._forEachChild(e=>this._registerControl(e))}_allControlsDisabled(){for(let e of this.controls)if(e.enabled)return!1;return this.controls.length>0||this.disabled}_registerControl(e){e.setParent(this),e._registerOnCollectionChange(this._onCollectionChange)}_find(e){return this.at(e)??null}};function lz(t){return!!t&&(t.asyncValidators!==void 0||t.validators!==void 0||t.updateOn!==void 0)}var Rz=(()=>{class t{useNonNullable=!1;get nonNullable(){let A=new t;return A.useNonNullable=!0,A}group(A,i=null){let n=this._reduceControls(A),o={};return lz(i)?o=i:i!==null&&(o.validators=i.validator,o.asyncValidators=i.asyncValidator),new oB(n,o)}record(A,i=null){let n=this._reduceControls(A);return new dM(n,i)}control(A,i,n){let o={};return this.useNonNullable?(lz(i)?o=i:(o.validators=i,o.asyncValidators=n),new _I(A,Ye(rA({},o),{nonNullable:!0}))):new _I(A,i,n)}array(A,i,n){let o=A.map(r=>this._createControl(r));return new BM(o,i,n)}_reduceControls(A){let i={};return Object.keys(A).forEach(n=>{i[n]=this._createControl(A[n])}),i}_createControl(A){if(A instanceof _I)return A;if(A instanceof nB)return A;if(Array.isArray(A)){let i=A[0],n=A.length>1?A[1]:null,o=A.length>2?A[2]:null;return this.control(i,n,o)}else return this.control(A)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var B6=(()=>{class t{static withConfig(A){return{ngModule:t,providers:[{provide:rB,useValue:A.callSetDisabledState??d6}]}}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[Sz]})}return t})(),xz=(()=>{class t{static withConfig(A){return{ngModule:t,providers:[{provide:yM,useValue:A.warnOnNgModelWithFormControl??"always"},{provide:rB,useValue:A.callSetDisabledState??d6}]}}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[Sz]})}return t})();var Li="primary",fu=Symbol("RouteTitle"),xM=class{params;constructor(e){this.params=e||{}}has(e){return Object.prototype.hasOwnProperty.call(this.params,e)}get(e){if(this.has(e)){let A=this.params[e];return Array.isArray(A)?A[0]:A}return null}getAll(e){if(this.has(e)){let A=this.params[e];return Array.isArray(A)?A:[A]}return[]}get keys(){return Object.keys(this.params)}};function JI(t){return new xM(t)}function Yz(t,e,A){let i=A.path.split("/");if(i.length>t.length||A.pathMatch==="full"&&(e.hasChildren()||i.length<t.length))return null;let n={};for(let o=0;o<i.length;o++){let r=i[o],s=t[o];if(r[0]===":")n[r.substring(1)]=s;else if(r!==s.path)return null}return{consumed:t.slice(0,i.length),posParams:n}}function WEA(t,e){if(t.length!==e.length)return!1;for(let A=0;A<t.length;++A)if(!ag(t[A],e[A]))return!1;return!0}function ag(t,e){let A=t?LM(t):void 0,i=e?LM(e):void 0;if(!A||!i||A.length!=i.length)return!1;let n;for(let o=0;o<A.length;o++)if(n=A[o],!Jz(t[n],e[n]))return!1;return!0}function LM(t){return[...Object.keys(t),...Object.getOwnPropertySymbols(t)]}function Jz(t,e){if(Array.isArray(t)&&Array.isArray(e)){if(t.length!==e.length)return!1;let A=[...t].sort(),i=[...e].sort();return A.every((n,o)=>i[o]===n)}else return t===e}function Tz(t){return t.length>0?t[t.length-1]:null}function N2(t){return I2(t)?t:D2(t)?oo(Promise.resolve(t)):Me(t)}var XEA={exact:zz,subset:Oz},Hz={exact:$EA,subset:AQA,ignored:()=>!0};function Lz(t,e,A){return XEA[A.paths](t.root,e.root,A.matrixParams)&&Hz[A.queryParams](t.queryParams,e.queryParams)&&!(A.fragment==="exact"&&t.fragment!==e.fragment)}function $EA(t,e){return ag(t,e)}function zz(t,e,A){if(!KI(t.segments,e.segments)||!h6(t.segments,e.segments,A)||t.numberOfChildren!==e.numberOfChildren)return!1;for(let i in e.children)if(!t.children[i]||!zz(t.children[i],e.children[i],A))return!1;return!0}function AQA(t,e){return Object.keys(e).length<=Object.keys(t).length&&Object.keys(e).every(A=>Jz(t[A],e[A]))}function Oz(t,e,A){return Pz(t,e,e.segments,A)}function Pz(t,e,A,i){if(t.segments.length>A.length){let n=t.segments.slice(0,A.length);return!(!KI(n,A)||e.hasChildren()||!h6(n,A,i))}else if(t.segments.length===A.length){if(!KI(t.segments,A)||!h6(t.segments,A,i))return!1;for(let n in e.children)if(!t.children[n]||!Oz(t.children[n],e.children[n],i))return!1;return!0}else{let n=A.slice(0,t.segments.length),o=A.slice(t.segments.length);return!KI(t.segments,n)||!h6(t.segments,n,i)||!t.children[Li]?!1:Pz(t.children[Li],e,o,i)}}function h6(t,e,A){return e.every((i,n)=>Hz[A](t[n].parameters,i.parameters))}var lg=class{root;queryParams;fragment;_queryParamMap;constructor(e=new _n([],{}),A={},i=null){this.root=e,this.queryParams=A,this.fragment=i}get queryParamMap(){return this._queryParamMap??=JI(this.queryParams),this._queryParamMap}toString(){return iQA.serialize(this)}},_n=class{segments;children;parent=null;constructor(e,A){this.segments=e,this.children=A,Object.values(A).forEach(i=>i.parent=this)}hasChildren(){return this.numberOfChildren>0}get numberOfChildren(){return Object.keys(this.children).length}toString(){return u6(this)}},x2=class{path;parameters;_parameterMap;constructor(e,A){this.path=e,this.parameters=A}get parameterMap(){return this._parameterMap??=JI(this.parameters),this._parameterMap}toString(){return qz(this)}};function eQA(t,e){return KI(t,e)&&t.every((A,i)=>ag(A.parameters,e[i].parameters))}function KI(t,e){return t.length!==e.length?!1:t.every((A,i)=>A.path===e[i].path)}function tQA(t,e){let A=[];return Object.entries(t.children).forEach(([i,n])=>{i===Li&&(A=A.concat(e(n,i)))}),Object.entries(t.children).forEach(([i,n])=>{i!==Li&&(A=A.concat(e(n,i)))}),A}var TI=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>new L2,providedIn:"root"})}return t})(),L2=class{parse(e){let A=new NM(e);return new lg(A.parseRootSegment(),A.parseQueryParams(),A.parseFragment())}serialize(e){let A=`/${au(e.root,!0)}`,i=rQA(e.queryParams),n=typeof e.fragment=="string"?`#${nQA(e.fragment)}`:"";return`${A}${i}${n}`}},iQA=new L2;function u6(t){return t.segments.map(e=>qz(e)).join("/")}function au(t,e){if(!t.hasChildren())return u6(t);if(e){let A=t.children[Li]?au(t.children[Li],!1):"",i=[];return Object.entries(t.children).forEach(([n,o])=>{n!==Li&&i.push(`${n}:${au(o,!1)}`)}),i.length>0?`${A}(${i.join("//")})`:A}else{let A=tQA(t,(i,n)=>n===Li?[au(t.children[Li],!1)]:[`${n}:${au(i,!1)}`]);return Object.keys(t.children).length===1&&t.children[Li]!=null?`${u6(t)}/${A[0]}`:`${u6(t)}/(${A.join("//")})`}}function jz(t){return encodeURIComponent(t).replace(/%40/g,"@").replace(/%3A/gi,":").replace(/%24/g,"$").replace(/%2C/gi,",")}function E6(t){return jz(t).replace(/%3B/gi,";")}function nQA(t){return encodeURI(t)}function FM(t){return jz(t).replace(/\(/g,"%28").replace(/\)/g,"%29").replace(/%26/gi,"&")}function f6(t){return decodeURIComponent(t)}function Fz(t){return f6(t.replace(/\+/g,"%20"))}function qz(t){return`${FM(t.path)}${oQA(t.parameters)}`}function oQA(t){return Object.entries(t).map(([e,A])=>`;${FM(e)}=${FM(A)}`).join("")}function rQA(t){let e=Object.entries(t).map(([A,i])=>Array.isArray(i)?i.map(n=>`${E6(A)}=${E6(n)}`).join("&"):`${E6(A)}=${E6(i)}`).filter(A=>A);return e.length?`?${e.join("&")}`:""}var sQA=/^[^\/()?;#]+/;function MM(t){let e=t.match(sQA);return e?e[0]:""}var aQA=/^[^\/()?;=#]+/;function cQA(t){let e=t.match(aQA);return e?e[0]:""}var lQA=/^[^=?&#]+/;function gQA(t){let e=t.match(lQA);return e?e[0]:""}var IQA=/^[^&#]+/;function CQA(t){let e=t.match(IQA);return e?e[0]:""}var NM=class{url;remaining;constructor(e){this.url=e,this.remaining=e}parseRootSegment(){return this.consumeOptional("/"),this.remaining===""||this.peekStartsWith("?")||this.peekStartsWith("#")?new _n([],{}):new _n([],this.parseChildren())}parseQueryParams(){let e={};if(this.consumeOptional("?"))do this.parseQueryParam(e);while(this.consumeOptional("&"));return e}parseFragment(){return this.consumeOptional("#")?decodeURIComponent(this.remaining):null}parseChildren(){if(this.remaining==="")return{};this.consumeOptional("/");let e=[];for(this.peekStartsWith("(")||e.push(this.parseSegment());this.peekStartsWith("/")&&!this.peekStartsWith("//")&&!this.peekStartsWith("/(");)this.capture("/"),e.push(this.parseSegment());let A={};this.peekStartsWith("/(")&&(this.capture("/"),A=this.parseParens(!0));let i={};return this.peekStartsWith("(")&&(i=this.parseParens(!1)),(e.length>0||Object.keys(A).length>0)&&(i[Li]=new _n(e,A)),i}parseSegment(){let e=MM(this.remaining);if(e===""&&this.peekStartsWith(";"))throw new Ae(4009,!1);return this.capture(e),new x2(f6(e),this.parseMatrixParams())}parseMatrixParams(){let e={};for(;this.consumeOptional(";");)this.parseParam(e);return e}parseParam(e){let A=cQA(this.remaining);if(!A)return;this.capture(A);let i="";if(this.consumeOptional("=")){let n=MM(this.remaining);n&&(i=n,this.capture(i))}e[f6(A)]=f6(i)}parseQueryParam(e){let A=gQA(this.remaining);if(!A)return;this.capture(A);let i="";if(this.consumeOptional("=")){let r=CQA(this.remaining);r&&(i=r,this.capture(i))}let n=Fz(A),o=Fz(i);if(e.hasOwnProperty(n)){let r=e[n];Array.isArray(r)||(r=[r],e[n]=r),r.push(o)}else e[n]=o}parseParens(e){let A={};for(this.capture("(");!this.consumeOptional(")")&&this.remaining.length>0;){let i=MM(this.remaining),n=this.remaining[i.length];if(n!=="/"&&n!==")"&&n!==";")throw new Ae(4010,!1);let o;i.indexOf(":")>-1?(o=i.slice(0,i.indexOf(":")),this.capture(o),this.capture(":")):e&&(o=Li);let r=this.parseChildren();A[o]=Object.keys(r).length===1?r[Li]:new _n([],r),this.consumeOptional("//")}return A}peekStartsWith(e){return this.remaining.startsWith(e)}consumeOptional(e){return this.peekStartsWith(e)?(this.remaining=this.remaining.substring(e.length),!0):!1}capture(e){if(!this.consumeOptional(e))throw new Ae(4011,!1)}};function Vz(t){return t.segments.length>0?new _n([],{[Li]:t}):t}function Zz(t){let e={};for(let[i,n]of Object.entries(t.children)){let o=Zz(n);if(i===Li&&o.segments.length===0&&o.hasChildren())for(let[r,s]of Object.entries(o.children))e[r]=s;else(o.segments.length>0||o.hasChildren())&&(e[i]=o)}let A=new _n(t.segments,e);return dQA(A)}function dQA(t){if(t.numberOfChildren===1&&t.children[Li]){let e=t.children[Li];return new _n(t.segments.concat(e.segments),e.children)}return t}function gB(t){return t instanceof lg}function Wz(t,e,A=null,i=null){let n=Xz(t);return $z(n,e,A,i)}function Xz(t){let e;function A(o){let r={};for(let a of o.children){let c=A(a);r[a.outlet]=c}let s=new _n(o.url,r);return o===t&&(e=s),s}let i=A(t.root),n=Vz(i);return e??n}function $z(t,e,A,i){let n=t;for(;n.parent;)n=n.parent;if(e.length===0)return kM(n,n,n,A,i);let o=BQA(e);if(o.toRoot())return kM(n,n,new _n([],{}),A,i);let r=EQA(o,n,t),s=r.processChildren?lu(r.segmentGroup,r.index,o.commands):eO(r.segmentGroup,r.index,o.commands);return kM(n,r.segmentGroup,s,A,i)}function p6(t){return typeof t=="object"&&t!=null&&!t.outlets&&!t.segmentPath}function Iu(t){return typeof t=="object"&&t!=null&&t.outlets}function kM(t,e,A,i,n){let o={};i&&Object.entries(i).forEach(([a,c])=>{o[a]=Array.isArray(c)?c.map(l=>`${l}`):`${c}`});let r;t===e?r=A:r=AO(t,e,A);let s=Vz(Zz(r));return new lg(s,o,n)}function AO(t,e,A){let i={};return Object.entries(t.children).forEach(([n,o])=>{o===e?i[n]=A:i[n]=AO(o,e,A)}),new _n(t.segments,i)}var w6=class{isAbsolute;numberOfDoubleDots;commands;constructor(e,A,i){if(this.isAbsolute=e,this.numberOfDoubleDots=A,this.commands=i,e&&i.length>0&&p6(i[0]))throw new Ae(4003,!1);let n=i.find(Iu);if(n&&n!==Tz(i))throw new Ae(4004,!1)}toRoot(){return this.isAbsolute&&this.commands.length===1&&this.commands[0]=="/"}};function BQA(t){if(typeof t[0]=="string"&&t.length===1&&t[0]==="/")return new w6(!0,0,t);let e=0,A=!1,i=t.reduce((n,o,r)=>{if(typeof o=="object"&&o!=null){if(o.outlets){let s={};return Object.entries(o.outlets).forEach(([a,c])=>{s[a]=typeof c=="string"?c.split("/"):c}),[...n,{outlets:s}]}if(o.segmentPath)return[...n,o.segmentPath]}return typeof o!="string"?[...n,o]:r===0?(o.split("/").forEach((s,a)=>{a==0&&s==="."||(a==0&&s===""?A=!0:s===".."?e++:s!=""&&n.push(s))}),n):[...n,o]},[]);return new w6(A,e,i)}var cB=class{segmentGroup;processChildren;index;constructor(e,A,i){this.segmentGroup=e,this.processChildren=A,this.index=i}};function EQA(t,e,A){if(t.isAbsolute)return new cB(e,!0,0);if(!A)return new cB(e,!1,NaN);if(A.parent===null)return new cB(A,!0,0);let i=p6(t.commands[0])?0:1,n=A.segments.length-1+i;return QQA(A,n,t.numberOfDoubleDots)}function QQA(t,e,A){let i=t,n=e,o=A;for(;o>n;){if(o-=n,i=i.parent,!i)throw new Ae(4005,!1);n=i.segments.length}return new cB(i,!1,n-o)}function hQA(t){return Iu(t[0])?t[0].outlets:{[Li]:t}}function eO(t,e,A){if(t??=new _n([],{}),t.segments.length===0&&t.hasChildren())return lu(t,e,A);let i=uQA(t,e,A),n=A.slice(i.commandIndex);if(i.match&&i.pathIndex<t.segments.length){let o=new _n(t.segments.slice(0,i.pathIndex),{});return o.children[Li]=new _n(t.segments.slice(i.pathIndex),t.children),lu(o,0,n)}else return i.match&&n.length===0?new _n(t.segments,{}):i.match&&!t.hasChildren()?_M(t,e,A):i.match?lu(t,0,n):_M(t,e,A)}function lu(t,e,A){if(A.length===0)return new _n(t.segments,{});{let i=hQA(A),n={};if(Object.keys(i).some(o=>o!==Li)&&t.children[Li]&&t.numberOfChildren===1&&t.children[Li].segments.length===0){let o=lu(t.children[Li],e,A);return new _n(t.segments,o.children)}return Object.entries(i).forEach(([o,r])=>{typeof r=="string"&&(r=[r]),r!==null&&(n[o]=eO(t.children[o],e,r))}),Object.entries(t.children).forEach(([o,r])=>{i[o]===void 0&&(n[o]=r)}),new _n(t.segments,n)}}function uQA(t,e,A){let i=0,n=e,o={match:!1,pathIndex:0,commandIndex:0};for(;n<t.segments.length;){if(i>=A.length)return o;let r=t.segments[n],s=A[i];if(Iu(s))break;let a=`${s}`,c=i<A.length-1?A[i+1]:null;if(n>0&&a===void 0)break;if(a&&c&&typeof c=="object"&&c.outlets===void 0){if(!_z(a,c,r))return o;i+=2}else{if(!_z(a,{},r))return o;i++}n++}return{match:!0,pathIndex:n,commandIndex:i}}function _M(t,e,A){let i=t.segments.slice(0,e),n=0;for(;n<A.length;){let o=A[n];if(Iu(o)){let a=fQA(o.outlets);return new _n(i,a)}if(n===0&&p6(A[0])){let a=t.segments[e];i.push(new x2(a.path,Nz(A[0]))),n++;continue}let r=Iu(o)?o.outlets[Li]:`${o}`,s=n<A.length-1?A[n+1]:null;r&&s&&p6(s)?(i.push(new x2(r,Nz(s))),n+=2):(i.push(new x2(r,{})),n++)}return new _n(i,{})}function fQA(t){let e={};return Object.entries(t).forEach(([A,i])=>{typeof i=="string"&&(i=[i]),i!==null&&(e[A]=_M(new _n([],{}),0,i))}),e}function Nz(t){let e={};return Object.entries(t).forEach(([A,i])=>e[A]=`${i}`),e}function _z(t,e,A){return t==A.path&&ag(e,A.parameters)}var m6="imperative",Gr=function(t){return t[t.NavigationStart=0]="NavigationStart",t[t.NavigationEnd=1]="NavigationEnd",t[t.NavigationCancel=2]="NavigationCancel",t[t.NavigationError=3]="NavigationError",t[t.RoutesRecognized=4]="RoutesRecognized",t[t.ResolveStart=5]="ResolveStart",t[t.ResolveEnd=6]="ResolveEnd",t[t.GuardsCheckStart=7]="GuardsCheckStart",t[t.GuardsCheckEnd=8]="GuardsCheckEnd",t[t.RouteConfigLoadStart=9]="RouteConfigLoadStart",t[t.RouteConfigLoadEnd=10]="RouteConfigLoadEnd",t[t.ChildActivationStart=11]="ChildActivationStart",t[t.ChildActivationEnd=12]="ChildActivationEnd",t[t.ActivationStart=13]="ActivationStart",t[t.ActivationEnd=14]="ActivationEnd",t[t.Scroll=15]="Scroll",t[t.NavigationSkipped=16]="NavigationSkipped",t}(Gr||{}),ic=class{id;url;constructor(e,A){this.id=e,this.url=A}},F2=class extends ic{type=Gr.NavigationStart;navigationTrigger;restoredState;constructor(e,A,i="imperative",n=null){super(e,A),this.navigationTrigger=i,this.restoredState=n}toString(){return`NavigationStart(id: ${this.id}, url: '${this.url}')`}},nc=class extends ic{urlAfterRedirects;type=Gr.NavigationEnd;constructor(e,A,i){super(e,A),this.urlAfterRedirects=i}toString(){return`NavigationEnd(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}')`}},Qa=function(t){return t[t.Redirect=0]="Redirect",t[t.SupersededByNewNavigation=1]="SupersededByNewNavigation",t[t.NoDataFromResolver=2]="NoDataFromResolver",t[t.GuardRejected=3]="GuardRejected",t}(Qa||{}),IB=function(t){return t[t.IgnoredSameUrlNavigation=0]="IgnoredSameUrlNavigation",t[t.IgnoredByUrlHandlingStrategy=1]="IgnoredByUrlHandlingStrategy",t}(IB||{}),cg=class extends ic{reason;code;type=Gr.NavigationCancel;constructor(e,A,i,n){super(e,A),this.reason=i,this.code=n}toString(){return`NavigationCancel(id: ${this.id}, url: '${this.url}')`}},gg=class extends ic{reason;code;type=Gr.NavigationSkipped;constructor(e,A,i,n){super(e,A),this.reason=i,this.code=n}},CB=class extends ic{error;target;type=Gr.NavigationError;constructor(e,A,i,n){super(e,A),this.error=i,this.target=n}toString(){return`NavigationError(id: ${this.id}, url: '${this.url}', error: ${this.error})`}},Cu=class extends ic{urlAfterRedirects;state;type=Gr.RoutesRecognized;constructor(e,A,i,n){super(e,A),this.urlAfterRedirects=i,this.state=n}toString(){return`RoutesRecognized(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}', state: ${this.state})`}},D6=class extends ic{urlAfterRedirects;state;type=Gr.GuardsCheckStart;constructor(e,A,i,n){super(e,A),this.urlAfterRedirects=i,this.state=n}toString(){return`GuardsCheckStart(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}', state: ${this.state})`}},y6=class extends ic{urlAfterRedirects;state;shouldActivate;type=Gr.GuardsCheckEnd;constructor(e,A,i,n,o){super(e,A),this.urlAfterRedirects=i,this.state=n,this.shouldActivate=o}toString(){return`GuardsCheckEnd(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}', state: ${this.state}, shouldActivate: ${this.shouldActivate})`}},v6=class extends ic{urlAfterRedirects;state;type=Gr.ResolveStart;constructor(e,A,i,n){super(e,A),this.urlAfterRedirects=i,this.state=n}toString(){return`ResolveStart(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}', state: ${this.state})`}},b6=class extends ic{urlAfterRedirects;state;type=Gr.ResolveEnd;constructor(e,A,i,n){super(e,A),this.urlAfterRedirects=i,this.state=n}toString(){return`ResolveEnd(id: ${this.id}, url: '${this.url}', urlAfterRedirects: '${this.urlAfterRedirects}', state: ${this.state})`}},M6=class{route;type=Gr.RouteConfigLoadStart;constructor(e){this.route=e}toString(){return`RouteConfigLoadStart(path: ${this.route.path})`}},k6=class{route;type=Gr.RouteConfigLoadEnd;constructor(e){this.route=e}toString(){return`RouteConfigLoadEnd(path: ${this.route.path})`}},S6=class{snapshot;type=Gr.ChildActivationStart;constructor(e){this.snapshot=e}toString(){return`ChildActivationStart(path: '${this.snapshot.routeConfig&&this.snapshot.routeConfig.path||""}')`}},R6=class{snapshot;type=Gr.ChildActivationEnd;constructor(e){this.snapshot=e}toString(){return`ChildActivationEnd(path: '${this.snapshot.routeConfig&&this.snapshot.routeConfig.path||""}')`}},x6=class{snapshot;type=Gr.ActivationStart;constructor(e){this.snapshot=e}toString(){return`ActivationStart(path: '${this.snapshot.routeConfig&&this.snapshot.routeConfig.path||""}')`}},L6=class{snapshot;type=Gr.ActivationEnd;constructor(e){this.snapshot=e}toString(){return`ActivationEnd(path: '${this.snapshot.routeConfig&&this.snapshot.routeConfig.path||""}')`}},dB=class{routerEvent;position;anchor;type=Gr.Scroll;constructor(e,A,i){this.routerEvent=e,this.position=A,this.anchor=i}toString(){let e=this.position?`${this.position[0]}, ${this.position[1]}`:null;return`Scroll(anchor: '${this.anchor}', position: '${e}')`}},du=class{},BB=class{url;navigationBehaviorOptions;constructor(e,A){this.url=e,this.navigationBehaviorOptions=A}};function mQA(t,e){return t.providers&&!t._injector&&(t._injector=Rh(t.providers,e,`Route: ${t.path}`)),t._injector??e}function ll(t){return t.outlet||Li}function pQA(t,e){let A=t.filter(i=>ll(i)===e);return A.push(...t.filter(i=>ll(i)!==e)),A}function mu(t){if(!t)return null;if(t.routeConfig?._injector)return t.routeConfig._injector;for(let e=t.parent;e;e=e.parent){let A=e.routeConfig;if(A?._loadedInjector)return A._loadedInjector;if(A?._injector)return A._injector}return null}var F6=class{rootInjector;outlet=null;route=null;children;attachRef=null;get injector(){return mu(this.route?.snapshot)??this.rootInjector}constructor(e){this.rootInjector=e,this.children=new HI(this.rootInjector)}},HI=(()=>{class t{rootInjector;contexts=new Map;constructor(A){this.rootInjector=A}onChildOutletCreated(A,i){let n=this.getOrCreateContext(A);n.outlet=i,this.contexts.set(A,n)}onChildOutletDestroyed(A){let i=this.getContext(A);i&&(i.outlet=null,i.attachRef=null)}onOutletDeactivated(){let A=this.contexts;return this.contexts=new Map,A}onOutletReAttached(A){this.contexts=A}getOrCreateContext(A){let i=this.getContext(A);return i||(i=new F6(this.rootInjector),this.contexts.set(A,i)),i}getContext(A){return this.contexts.get(A)||null}static \u0275fac=function(i){return new(i||t)(we(pr))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),N6=class{_root;constructor(e){this._root=e}get root(){return this._root.value}parent(e){let A=this.pathFromRoot(e);return A.length>1?A[A.length-2]:null}children(e){let A=GM(e,this._root);return A?A.children.map(i=>i.value):[]}firstChild(e){let A=GM(e,this._root);return A&&A.children.length>0?A.children[0].value:null}siblings(e){let A=UM(e,this._root);return A.length<2?[]:A[A.length-2].children.map(n=>n.value).filter(n=>n!==e)}pathFromRoot(e){return UM(e,this._root).map(A=>A.value)}};function GM(t,e){if(t===e.value)return e;for(let A of e.children){let i=GM(t,A);if(i)return i}return null}function UM(t,e){if(t===e.value)return[e];for(let A of e.children){let i=UM(t,A);if(i.length)return i.unshift(e),i}return[]}var tc=class{value;children;constructor(e,A){this.value=e,this.children=A}toString(){return`TreeNode(${this.value})`}};function aB(t){let e={};return t&&t.children.forEach(A=>e[A.value.outlet]=A),e}var Bu=class extends N6{snapshot;constructor(e,A){super(e),this.snapshot=A,PM(this,e)}toString(){return this.snapshot.toString()}};function tO(t){let e=wQA(t),A=new Mi([new x2("",{})]),i=new Mi({}),n=new Mi({}),o=new Mi({}),r=new Mi(""),s=new ha(A,i,o,r,n,Li,t,e.root);return s.snapshot=e.root,new Bu(new tc(s,[]),e)}function wQA(t){let e={},A={},i={},n="",o=new YI([],e,i,n,A,Li,t,null,{});return new Eu("",new tc(o,[]))}var ha=class{urlSubject;paramsSubject;queryParamsSubject;fragmentSubject;dataSubject;outlet;component;snapshot;_futureSnapshot;_routerState;_paramMap;_queryParamMap;title;url;params;queryParams;fragment;data;constructor(e,A,i,n,o,r,s,a){this.urlSubject=e,this.paramsSubject=A,this.queryParamsSubject=i,this.fragmentSubject=n,this.dataSubject=o,this.outlet=r,this.component=s,this._futureSnapshot=a,this.title=this.dataSubject?.pipe(je(c=>c[fu]))??Me(void 0),this.url=e,this.params=A,this.queryParams=i,this.fragment=n,this.data=o}get routeConfig(){return this._futureSnapshot.routeConfig}get root(){return this._routerState.root}get parent(){return this._routerState.parent(this)}get firstChild(){return this._routerState.firstChild(this)}get children(){return this._routerState.children(this)}get pathFromRoot(){return this._routerState.pathFromRoot(this)}get paramMap(){return this._paramMap??=this.params.pipe(je(e=>JI(e))),this._paramMap}get queryParamMap(){return this._queryParamMap??=this.queryParams.pipe(je(e=>JI(e))),this._queryParamMap}toString(){return this.snapshot?this.snapshot.toString():`Future(${this._futureSnapshot})`}};function _6(t,e,A="emptyOnly"){let i,{routeConfig:n}=t;return e!==null&&(A==="always"||n?.path===""||!e.component&&!e.routeConfig?.loadComponent)?i={params:rA(rA({},e.params),t.params),data:rA(rA({},e.data),t.data),resolve:rA(rA(rA(rA({},t.data),e.data),n?.data),t._resolvedData)}:i={params:rA({},t.params),data:rA({},t.data),resolve:rA(rA({},t.data),t._resolvedData??{})},n&&nO(n)&&(i.resolve[fu]=n.title),i}var YI=class{url;params;queryParams;fragment;data;outlet;component;routeConfig;_resolve;_resolvedData;_routerState;_paramMap;_queryParamMap;get title(){return this.data?.[fu]}constructor(e,A,i,n,o,r,s,a,c){this.url=e,this.params=A,this.queryParams=i,this.fragment=n,this.data=o,this.outlet=r,this.component=s,this.routeConfig=a,this._resolve=c}get root(){return this._routerState.root}get parent(){return this._routerState.parent(this)}get firstChild(){return this._routerState.firstChild(this)}get children(){return this._routerState.children(this)}get pathFromRoot(){return this._routerState.pathFromRoot(this)}get paramMap(){return this._paramMap??=JI(this.params),this._paramMap}get queryParamMap(){return this._queryParamMap??=JI(this.queryParams),this._queryParamMap}toString(){let e=this.url.map(i=>i.toString()).join("/"),A=this.routeConfig?this.routeConfig.path:"";return`Route(url:'${e}', path:'${A}')`}},Eu=class extends N6{url;constructor(e,A){super(A),this.url=e,PM(this,A)}toString(){return iO(this._root)}};function PM(t,e){e.value._routerState=t,e.children.forEach(A=>PM(t,A))}function iO(t){let e=t.children.length>0?` { ${t.children.map(iO).join(", ")} } `:"";return`${t.value}${e}`}function SM(t){if(t.snapshot){let e=t.snapshot,A=t._futureSnapshot;t.snapshot=A,ag(e.queryParams,A.queryParams)||t.queryParamsSubject.next(A.queryParams),e.fragment!==A.fragment&&t.fragmentSubject.next(A.fragment),ag(e.params,A.params)||t.paramsSubject.next(A.params),WEA(e.url,A.url)||t.urlSubject.next(A.url),ag(e.data,A.data)||t.dataSubject.next(A.data)}else t.snapshot=t._futureSnapshot,t.dataSubject.next(t._futureSnapshot.data)}function KM(t,e){let A=ag(t.params,e.params)&&eQA(t.url,e.url),i=!t.parent!=!e.parent;return A&&!i&&(!t.parent||KM(t.parent,e.parent))}function nO(t){return typeof t.title=="string"||t.title===null}var oO=new dA(""),jM=(()=>{class t{activated=null;get activatedComponentRef(){return this.activated}_activatedRoute=null;name=Li;activateEvents=new WA;deactivateEvents=new WA;attachEvents=new WA;detachEvents=new WA;routerOutletData=wJ(void 0);parentContexts=f(HI);location=f(Nn);changeDetector=f(It);inputBinder=f(pu,{optional:!0});supportsBindingToComponentInputs=!0;ngOnChanges(A){if(A.name){let{firstChange:i,previousValue:n}=A.name;if(i)return;this.isTrackedInParentContexts(n)&&(this.deactivate(),this.parentContexts.onChildOutletDestroyed(n)),this.initializeOutletWithName()}}ngOnDestroy(){this.isTrackedInParentContexts(this.name)&&this.parentContexts.onChildOutletDestroyed(this.name),this.inputBinder?.unsubscribeFromRouteData(this)}isTrackedInParentContexts(A){return this.parentContexts.getContext(A)?.outlet===this}ngOnInit(){this.initializeOutletWithName()}initializeOutletWithName(){if(this.parentContexts.onChildOutletCreated(this.name,this),this.activated)return;let A=this.parentContexts.getContext(this.name);A?.route&&(A.attachRef?this.attach(A.attachRef,A.route):this.activateWith(A.route,A.injector))}get isActivated(){return!!this.activated}get component(){if(!this.activated)throw new Ae(4012,!1);return this.activated.instance}get activatedRoute(){if(!this.activated)throw new Ae(4012,!1);return this._activatedRoute}get activatedRouteData(){return this._activatedRoute?this._activatedRoute.snapshot.data:{}}detach(){if(!this.activated)throw new Ae(4012,!1);this.location.detach();let A=this.activated;return this.activated=null,this._activatedRoute=null,this.detachEvents.emit(A.instance),A}attach(A,i){this.activated=A,this._activatedRoute=i,this.location.insert(A.hostView),this.inputBinder?.bindActivatedRouteToOutletComponent(this),this.attachEvents.emit(A.instance)}deactivate(){if(this.activated){let A=this.component;this.activated.destroy(),this.activated=null,this._activatedRoute=null,this.deactivateEvents.emit(A)}}activateWith(A,i){if(this.isActivated)throw new Ae(4013,!1);this._activatedRoute=A;let n=this.location,r=A.snapshot.component,s=this.parentContexts.getOrCreateContext(this.name).children,a=new YM(A,s,n.injector,this.routerOutletData);this.activated=n.createComponent(r,{index:n.length,injector:a,environmentInjector:i}),this.changeDetector.markForCheck(),this.inputBinder?.bindActivatedRouteToOutletComponent(this),this.activateEvents.emit(this.activated.instance)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["router-outlet"]],inputs:{name:"name",routerOutletData:[1,"routerOutletData"]},outputs:{activateEvents:"activate",deactivateEvents:"deactivate",attachEvents:"attach",detachEvents:"detach"},exportAs:["outlet"],features:[ti]})}return t})(),YM=class{route;childContexts;parent;outletData;constructor(e,A,i,n){this.route=e,this.childContexts=A,this.parent=i,this.outletData=n}get(e,A){return e===ha?this.route:e===HI?this.childContexts:e===oO?this.outletData:this.parent.get(e,A)}},pu=new dA(""),qM=(()=>{class t{outletDataSubscriptions=new Map;bindActivatedRouteToOutletComponent(A){this.unsubscribeFromRouteData(A),this.subscribeToRouteData(A)}unsubscribeFromRouteData(A){this.outletDataSubscriptions.get(A)?.unsubscribe(),this.outletDataSubscriptions.delete(A)}subscribeToRouteData(A){let{activatedRoute:i}=A,n=Js([i.queryParams,i.params,i.data]).pipe(jn(([o,r,s],a)=>(s=rA(rA(rA({},o),r),s),a===0?Me(s):Promise.resolve(s)))).subscribe(o=>{if(!A.isActivated||!A.activatedComponentRef||A.activatedRoute!==i||i.component===null){this.unsubscribeFromRouteData(A);return}let r=pH(i.component);if(!r){this.unsubscribeFromRouteData(A);return}for(let{templateName:s}of r.inputs)A.activatedComponentRef.setInput(s,o[s])});this.outletDataSubscriptions.set(A,n)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),VM=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["ng-component"]],exportAs:["emptyRouterOutlet"],decls:1,vars:0,template:function(i,n){i&1&&JA(0,"router-outlet")},dependencies:[jM],encapsulation:2})}return t})();function ZM(t){let e=t.children&&t.children.map(ZM),A=e?Ye(rA({},t),{children:e}):rA({},t);return!A.component&&!A.loadComponent&&(e||A.loadChildren)&&A.outlet&&A.outlet!==Li&&(A.component=VM),A}function DQA(t,e,A){let i=Qu(t,e._root,A?A._root:void 0);return new Bu(i,e)}function Qu(t,e,A){if(A&&t.shouldReuseRoute(e.value,A.value.snapshot)){let i=A.value;i._futureSnapshot=e.value;let n=yQA(t,e,A);return new tc(i,n)}else{if(t.shouldAttach(e.value)){let o=t.retrieve(e.value);if(o!==null){let r=o.route;return r.value._futureSnapshot=e.value,r.children=e.children.map(s=>Qu(t,s)),r}}let i=vQA(e.value),n=e.children.map(o=>Qu(t,o));return new tc(i,n)}}function yQA(t,e,A){return e.children.map(i=>{for(let n of A.children)if(t.shouldReuseRoute(i.value,n.value.snapshot))return Qu(t,i,n);return Qu(t,i)})}function vQA(t){return new ha(new Mi(t.url),new Mi(t.params),new Mi(t.queryParams),new Mi(t.fragment),new Mi(t.data),t.outlet,t.component,t)}var EB=class{redirectTo;navigationBehaviorOptions;constructor(e,A){this.redirectTo=e,this.navigationBehaviorOptions=A}},rO="ngNavigationCancelingError";function G6(t,e){let{redirectTo:A,navigationBehaviorOptions:i}=gB(e)?{redirectTo:e,navigationBehaviorOptions:void 0}:e,n=sO(!1,Qa.Redirect);return n.url=A,n.navigationBehaviorOptions=i,n}function sO(t,e){let A=new Error(`NavigationCancelingError: ${t||""}`);return A[rO]=!0,A.cancellationCode=e,A}function bQA(t){return aO(t)&&gB(t.url)}function aO(t){return!!t&&t[rO]}var MQA=(t,e,A,i)=>je(n=>(new JM(e,n.targetRouterState,n.currentRouterState,A,i).activate(t),n)),JM=class{routeReuseStrategy;futureState;currState;forwardEvent;inputBindingEnabled;constructor(e,A,i,n,o){this.routeReuseStrategy=e,this.futureState=A,this.currState=i,this.forwardEvent=n,this.inputBindingEnabled=o}activate(e){let A=this.futureState._root,i=this.currState?this.currState._root:null;this.deactivateChildRoutes(A,i,e),SM(this.futureState.root),this.activateChildRoutes(A,i,e)}deactivateChildRoutes(e,A,i){let n=aB(A);e.children.forEach(o=>{let r=o.value.outlet;this.deactivateRoutes(o,n[r],i),delete n[r]}),Object.values(n).forEach(o=>{this.deactivateRouteAndItsChildren(o,i)})}deactivateRoutes(e,A,i){let n=e.value,o=A?A.value:null;if(n===o)if(n.component){let r=i.getContext(n.outlet);r&&this.deactivateChildRoutes(e,A,r.children)}else this.deactivateChildRoutes(e,A,i);else o&&this.deactivateRouteAndItsChildren(A,i)}deactivateRouteAndItsChildren(e,A){e.value.component&&this.routeReuseStrategy.shouldDetach(e.value.snapshot)?this.detachAndStoreRouteSubtree(e,A):this.deactivateRouteAndOutlet(e,A)}detachAndStoreRouteSubtree(e,A){let i=A.getContext(e.value.outlet),n=i&&e.value.component?i.children:A,o=aB(e);for(let r of Object.values(o))this.deactivateRouteAndItsChildren(r,n);if(i&&i.outlet){let r=i.outlet.detach(),s=i.children.onOutletDeactivated();this.routeReuseStrategy.store(e.value.snapshot,{componentRef:r,route:e,contexts:s})}}deactivateRouteAndOutlet(e,A){let i=A.getContext(e.value.outlet),n=i&&e.value.component?i.children:A,o=aB(e);for(let r of Object.values(o))this.deactivateRouteAndItsChildren(r,n);i&&(i.outlet&&(i.outlet.deactivate(),i.children.onOutletDeactivated()),i.attachRef=null,i.route=null)}activateChildRoutes(e,A,i){let n=aB(A);e.children.forEach(o=>{this.activateRoutes(o,n[o.value.outlet],i),this.forwardEvent(new L6(o.value.snapshot))}),e.children.length&&this.forwardEvent(new R6(e.value.snapshot))}activateRoutes(e,A,i){let n=e.value,o=A?A.value:null;if(SM(n),n===o)if(n.component){let r=i.getOrCreateContext(n.outlet);this.activateChildRoutes(e,A,r.children)}else this.activateChildRoutes(e,A,i);else if(n.component){let r=i.getOrCreateContext(n.outlet);if(this.routeReuseStrategy.shouldAttach(n.snapshot)){let s=this.routeReuseStrategy.retrieve(n.snapshot);this.routeReuseStrategy.store(n.snapshot,null),r.children.onOutletReAttached(s.contexts),r.attachRef=s.componentRef,r.route=s.route.value,r.outlet&&r.outlet.attach(s.componentRef,s.route.value),SM(s.route.value),this.activateChildRoutes(e,null,r.children)}else r.attachRef=null,r.route=n,r.outlet&&r.outlet.activateWith(n,r.injector),this.activateChildRoutes(e,null,r.children)}else this.activateChildRoutes(e,null,i)}},U6=class{path;route;constructor(e){this.path=e,this.route=this.path[this.path.length-1]}},lB=class{component;route;constructor(e,A){this.component=e,this.route=A}};function kQA(t,e,A){let i=t._root,n=e?e._root:null;return cu(i,n,A,[i.value])}function SQA(t){let e=t.routeConfig?t.routeConfig.canActivateChild:null;return!e||e.length===0?null:{node:t,guards:e}}function hB(t,e){let A=Symbol(),i=e.get(t,A);return i===A?typeof t=="function"&&!DY(t)?t:e.get(t):i}function cu(t,e,A,i,n={canDeactivateChecks:[],canActivateChecks:[]}){let o=aB(e);return t.children.forEach(r=>{RQA(r,o[r.value.outlet],A,i.concat([r.value]),n),delete o[r.value.outlet]}),Object.entries(o).forEach(([r,s])=>gu(s,A.getContext(r),n)),n}function RQA(t,e,A,i,n={canDeactivateChecks:[],canActivateChecks:[]}){let o=t.value,r=e?e.value:null,s=A?A.getContext(t.value.outlet):null;if(r&&o.routeConfig===r.routeConfig){let a=xQA(r,o,o.routeConfig.runGuardsAndResolvers);a?n.canActivateChecks.push(new U6(i)):(o.data=r.data,o._resolvedData=r._resolvedData),o.component?cu(t,e,s?s.children:null,i,n):cu(t,e,A,i,n),a&&s&&s.outlet&&s.outlet.isActivated&&n.canDeactivateChecks.push(new lB(s.outlet.component,r))}else r&&gu(e,s,n),n.canActivateChecks.push(new U6(i)),o.component?cu(t,null,s?s.children:null,i,n):cu(t,null,A,i,n);return n}function xQA(t,e,A){if(typeof A=="function")return A(t,e);switch(A){case"pathParamsChange":return!KI(t.url,e.url);case"pathParamsOrQueryParamsChange":return!KI(t.url,e.url)||!ag(t.queryParams,e.queryParams);case"always":return!0;case"paramsOrQueryParamsChange":return!KM(t,e)||!ag(t.queryParams,e.queryParams);case"paramsChange":default:return!KM(t,e)}}function gu(t,e,A){let i=aB(t),n=t.value;Object.entries(i).forEach(([o,r])=>{n.component?e?gu(r,e.children.getContext(o),A):gu(r,null,A):gu(r,e,A)}),n.component?e&&e.outlet&&e.outlet.isActivated?A.canDeactivateChecks.push(new lB(e.outlet.component,n)):A.canDeactivateChecks.push(new lB(null,n)):A.canDeactivateChecks.push(new lB(null,n))}function wu(t){return typeof t=="function"}function LQA(t){return typeof t=="boolean"}function FQA(t){return t&&wu(t.canLoad)}function NQA(t){return t&&wu(t.canActivate)}function _QA(t){return t&&wu(t.canActivateChild)}function GQA(t){return t&&wu(t.canDeactivate)}function UQA(t){return t&&wu(t.canMatch)}function cO(t){return t instanceof s0||t?.name==="EmptyError"}var Q6=Symbol("INITIAL_VALUE");function QB(){return jn(t=>Js(t.map(e=>e.pipe(On(1),Pn(Q6)))).pipe(je(e=>{for(let A of e)if(A!==!0){if(A===Q6)return Q6;if(A===!1||KQA(A))return A}return!0}),kt(e=>e!==Q6),On(1)))}function KQA(t){return gB(t)||t instanceof EB}function YQA(t,e){return tr(A=>{let{targetSnapshot:i,currentSnapshot:n,guards:{canActivateChecks:o,canDeactivateChecks:r}}=A;return r.length===0&&o.length===0?Me(Ye(rA({},A),{guardsResult:!0})):JQA(r,i,n,t).pipe(tr(s=>s&&LQA(s)?TQA(i,o,t,e):Me(s)),je(s=>Ye(rA({},A),{guardsResult:s})))})}function JQA(t,e,A,i){return oo(t).pipe(tr(n=>jQA(n.component,n.route,A,e,i)),Zl(n=>n!==!0,!0))}function TQA(t,e,A,i){return oo(e).pipe(ql(n=>d2(zQA(n.route.parent,i),HQA(n.route,i),PQA(t,n.path,A),OQA(t,n.route,A))),Zl(n=>n!==!0,!0))}function HQA(t,e){return t!==null&&e&&e(new x6(t)),Me(!0)}function zQA(t,e){return t!==null&&e&&e(new S6(t)),Me(!0)}function OQA(t,e,A){let i=e.routeConfig?e.routeConfig.canActivate:null;if(!i||i.length===0)return Me(!0);let n=i.map(o=>jl(()=>{let r=mu(e)??A,s=hB(o,r),a=NQA(s)?s.canActivate(e,t):ga(r,()=>s(e,t));return N2(a).pipe(Zl())}));return Me(n).pipe(QB())}function PQA(t,e,A){let i=e[e.length-1],o=e.slice(0,e.length-1).reverse().map(r=>SQA(r)).filter(r=>r!==null).map(r=>jl(()=>{let s=r.guards.map(a=>{let c=mu(r.node)??A,l=hB(a,c),I=_QA(l)?l.canActivateChild(i,t):ga(c,()=>l(i,t));return N2(I).pipe(Zl())});return Me(s).pipe(QB())}));return Me(o).pipe(QB())}function jQA(t,e,A,i,n){let o=e&&e.routeConfig?e.routeConfig.canDeactivate:null;if(!o||o.length===0)return Me(!0);let r=o.map(s=>{let a=mu(e)??n,c=hB(s,a),l=GQA(c)?c.canDeactivate(t,e,A,i):ga(a,()=>c(t,e,A,i));return N2(l).pipe(Zl())});return Me(r).pipe(QB())}function qQA(t,e,A,i){let n=e.canLoad;if(n===void 0||n.length===0)return Me(!0);let o=n.map(r=>{let s=hB(r,t),a=FQA(s)?s.canLoad(e,A):ga(t,()=>s(e,A));return N2(a)});return Me(o).pipe(QB(),lO(i))}function lO(t){return sv(Qo(e=>{if(typeof e!="boolean")throw G6(t,e)}),je(e=>e===!0))}function VQA(t,e,A,i){let n=e.canMatch;if(!n||n.length===0)return Me(!0);let o=n.map(r=>{let s=hB(r,t),a=UQA(s)?s.canMatch(e,A):ga(t,()=>s(e,A));return N2(a)});return Me(o).pipe(QB(),lO(i))}var hu=class{segmentGroup;constructor(e){this.segmentGroup=e||null}},uu=class extends Error{urlTree;constructor(e){super(),this.urlTree=e}};function sB(t){return g2(new hu(t))}function ZQA(t){return g2(new Ae(4e3,!1))}function WQA(t){return g2(sO(!1,Qa.GuardRejected))}var TM=class{urlSerializer;urlTree;constructor(e,A){this.urlSerializer=e,this.urlTree=A}lineralizeSegments(e,A){let i=[],n=A.root;for(;;){if(i=i.concat(n.segments),n.numberOfChildren===0)return Me(i);if(n.numberOfChildren>1||!n.children[Li])return ZQA(`${e.redirectTo}`);n=n.children[Li]}}applyRedirectCommands(e,A,i,n,o){if(typeof A!="string"){let s=A,{queryParams:a,fragment:c,routeConfig:l,url:I,outlet:C,params:d,data:B,title:E}=n,h=ga(o,()=>s({params:d,data:B,queryParams:a,fragment:c,routeConfig:l,url:I,outlet:C,title:E}));if(h instanceof lg)throw new uu(h);A=h}let r=this.applyRedirectCreateUrlTree(A,this.urlSerializer.parse(A),e,i);if(A[0]==="/")throw new uu(r);return r}applyRedirectCreateUrlTree(e,A,i,n){let o=this.createSegmentGroup(e,A.root,i,n);return new lg(o,this.createQueryParams(A.queryParams,this.urlTree.queryParams),A.fragment)}createQueryParams(e,A){let i={};return Object.entries(e).forEach(([n,o])=>{if(typeof o=="string"&&o[0]===":"){let s=o.substring(1);i[n]=A[s]}else i[n]=o}),i}createSegmentGroup(e,A,i,n){let o=this.createSegments(e,A.segments,i,n),r={};return Object.entries(A.children).forEach(([s,a])=>{r[s]=this.createSegmentGroup(e,a,i,n)}),new _n(o,r)}createSegments(e,A,i,n){return A.map(o=>o.path[0]===":"?this.findPosParam(e,o,n):this.findOrReturn(o,i))}findPosParam(e,A,i){let n=i[A.path.substring(1)];if(!n)throw new Ae(4001,!1);return n}findOrReturn(e,A){let i=0;for(let n of A){if(n.path===e.path)return A.splice(i),n;i++}return e}},HM={matched:!1,consumedSegments:[],remainingSegments:[],parameters:{},positionalParamSegments:{}};function XQA(t,e,A,i,n){let o=gO(t,e,A);return o.matched?(i=mQA(e,i),VQA(i,e,A,n).pipe(je(r=>r===!0?o:rA({},HM)))):Me(o)}function gO(t,e,A){if(e.path==="**")return $QA(A);if(e.path==="")return e.pathMatch==="full"&&(t.hasChildren()||A.length>0)?rA({},HM):{matched:!0,consumedSegments:[],remainingSegments:A,parameters:{},positionalParamSegments:{}};let n=(e.matcher||Yz)(A,t,e);if(!n)return rA({},HM);let o={};Object.entries(n.posParams??{}).forEach(([s,a])=>{o[s]=a.path});let r=n.consumed.length>0?rA(rA({},o),n.consumed[n.consumed.length-1].parameters):o;return{matched:!0,consumedSegments:n.consumed,remainingSegments:A.slice(n.consumed.length),parameters:r,positionalParamSegments:n.posParams??{}}}function $QA(t){return{matched:!0,parameters:t.length>0?Tz(t).parameters:{},consumedSegments:t,remainingSegments:[],positionalParamSegments:{}}}function Gz(t,e,A,i){return A.length>0&&thA(t,A,i)?{segmentGroup:new _n(e,ehA(i,new _n(A,t.children))),slicedSegments:[]}:A.length===0&&ihA(t,A,i)?{segmentGroup:new _n(t.segments,AhA(t,A,i,t.children)),slicedSegments:A}:{segmentGroup:new _n(t.segments,t.children),slicedSegments:A}}function AhA(t,e,A,i){let n={};for(let o of A)if(Y6(t,e,o)&&!i[ll(o)]){let r=new _n([],{});n[ll(o)]=r}return rA(rA({},i),n)}function ehA(t,e){let A={};A[Li]=e;for(let i of t)if(i.path===""&&ll(i)!==Li){let n=new _n([],{});A[ll(i)]=n}return A}function thA(t,e,A){return A.some(i=>Y6(t,e,i)&&ll(i)!==Li)}function ihA(t,e,A){return A.some(i=>Y6(t,e,i))}function Y6(t,e,A){return(t.hasChildren()||e.length>0)&&A.pathMatch==="full"?!1:A.path===""}function nhA(t,e,A){return e.length===0&&!t.children[A]}var zM=class{};function ohA(t,e,A,i,n,o,r="emptyOnly"){return new OM(t,e,A,i,n,r,o).recognize()}var rhA=31,OM=class{injector;configLoader;rootComponentType;config;urlTree;paramsInheritanceStrategy;urlSerializer;applyRedirects;absoluteRedirectCount=0;allowRedirects=!0;constructor(e,A,i,n,o,r,s){this.injector=e,this.configLoader=A,this.rootComponentType=i,this.config=n,this.urlTree=o,this.paramsInheritanceStrategy=r,this.urlSerializer=s,this.applyRedirects=new TM(this.urlSerializer,this.urlTree)}noMatchError(e){return new Ae(4002,`'${e.segmentGroup}'`)}recognize(){let e=Gz(this.urlTree.root,[],[],this.config).segmentGroup;return this.match(e).pipe(je(({children:A,rootSnapshot:i})=>{let n=new tc(i,A),o=new Eu("",n),r=Wz(i,[],this.urlTree.queryParams,this.urlTree.fragment);return r.queryParams=this.urlTree.queryParams,o.url=this.urlSerializer.serialize(r),{state:o,tree:r}}))}match(e){let A=new YI([],Object.freeze({}),Object.freeze(rA({},this.urlTree.queryParams)),this.urlTree.fragment,Object.freeze({}),Li,this.rootComponentType,null,{});return this.processSegmentGroup(this.injector,this.config,e,Li,A).pipe(je(i=>({children:i,rootSnapshot:A})),mr(i=>{if(i instanceof uu)return this.urlTree=i.urlTree,this.match(i.urlTree.root);throw i instanceof hu?this.noMatchError(i):i}))}processSegmentGroup(e,A,i,n,o){return i.segments.length===0&&i.hasChildren()?this.processChildren(e,A,i,o):this.processSegment(e,A,i,i.segments,n,!0,o).pipe(je(r=>r instanceof tc?[r]:[]))}processChildren(e,A,i,n){let o=[];for(let r of Object.keys(i.children))r==="primary"?o.unshift(r):o.push(r);return oo(o).pipe(ql(r=>{let s=i.children[r],a=pQA(A,r);return this.processSegmentGroup(e,a,s,r,n)}),Cv((r,s)=>(r.push(...s),r)),B2(null),Iv(),tr(r=>{if(r===null)return sB(i);let s=IO(r);return shA(s),Me(s)}))}processSegment(e,A,i,n,o,r,s){return oo(A).pipe(ql(a=>this.processSegmentAgainstRoute(a._injector??e,A,a,i,n,o,r,s).pipe(mr(c=>{if(c instanceof hu)return Me(null);throw c}))),Zl(a=>!!a),mr(a=>{if(cO(a))return nhA(i,n,o)?Me(new zM):sB(i);throw a}))}processSegmentAgainstRoute(e,A,i,n,o,r,s,a){return ll(i)!==r&&(r===Li||!Y6(n,o,i))?sB(n):i.redirectTo===void 0?this.matchSegmentAgainstRoute(e,n,i,o,r,a):this.allowRedirects&&s?this.expandSegmentAgainstRouteUsingRedirect(e,n,A,i,o,r,a):sB(n)}expandSegmentAgainstRouteUsingRedirect(e,A,i,n,o,r,s){let{matched:a,parameters:c,consumedSegments:l,positionalParamSegments:I,remainingSegments:C}=gO(A,n,o);if(!a)return sB(A);typeof n.redirectTo=="string"&&n.redirectTo[0]==="/"&&(this.absoluteRedirectCount++,this.absoluteRedirectCount>rhA&&(this.allowRedirects=!1));let d=new YI(o,c,Object.freeze(rA({},this.urlTree.queryParams)),this.urlTree.fragment,Uz(n),ll(n),n.component??n._loadedComponent??null,n,Kz(n)),B=_6(d,s,this.paramsInheritanceStrategy);d.params=Object.freeze(B.params),d.data=Object.freeze(B.data);let E=this.applyRedirects.applyRedirectCommands(l,n.redirectTo,I,d,e);return this.applyRedirects.lineralizeSegments(n,E).pipe(tr(h=>this.processSegment(e,i,A,h.concat(C),r,!1,s)))}matchSegmentAgainstRoute(e,A,i,n,o,r){let s=XQA(A,i,n,e,this.urlSerializer);return i.path==="**"&&(A.children={}),s.pipe(jn(a=>a.matched?(e=i._injector??e,this.getChildConfig(e,i,n).pipe(jn(({routes:c})=>{let l=i._loadedInjector??e,{parameters:I,consumedSegments:C,remainingSegments:d}=a,B=new YI(C,I,Object.freeze(rA({},this.urlTree.queryParams)),this.urlTree.fragment,Uz(i),ll(i),i.component??i._loadedComponent??null,i,Kz(i)),E=_6(B,r,this.paramsInheritanceStrategy);B.params=Object.freeze(E.params),B.data=Object.freeze(E.data);let{segmentGroup:h,slicedSegments:u}=Gz(A,C,d,c);if(u.length===0&&h.hasChildren())return this.processChildren(l,c,h,B).pipe(je(L=>new tc(B,L)));if(c.length===0&&u.length===0)return Me(new tc(B,[]));let D=ll(i)===o;return this.processSegment(l,c,h,u,D?Li:o,!0,B).pipe(je(L=>new tc(B,L instanceof tc?[L]:[])))}))):sB(A)))}getChildConfig(e,A,i){return A.children?Me({routes:A.children,injector:e}):A.loadChildren?A._loadedRoutes!==void 0?Me({routes:A._loadedRoutes,injector:A._loadedInjector}):qQA(e,A,i,this.urlSerializer).pipe(tr(n=>n?this.configLoader.loadChildren(e,A).pipe(Qo(o=>{A._loadedRoutes=o.routes,A._loadedInjector=o.injector})):WQA(A))):Me({routes:[],injector:e})}};function shA(t){t.sort((e,A)=>e.value.outlet===Li?-1:A.value.outlet===Li?1:e.value.outlet.localeCompare(A.value.outlet))}function ahA(t){let e=t.value.routeConfig;return e&&e.path===""}function IO(t){let e=[],A=new Set;for(let i of t){if(!ahA(i)){e.push(i);continue}let n=e.find(o=>i.value.routeConfig===o.value.routeConfig);n!==void 0?(n.children.push(...i.children),A.add(n)):e.push(i)}for(let i of A){let n=IO(i.children);e.push(new tc(i.value,n))}return e.filter(i=>!A.has(i))}function Uz(t){return t.data||{}}function Kz(t){return t.resolve||{}}function chA(t,e,A,i,n,o){return tr(r=>ohA(t,e,A,i,r.extractedUrl,n,o).pipe(je(({state:s,tree:a})=>Ye(rA({},r),{targetSnapshot:s,urlAfterRedirects:a}))))}function lhA(t,e){return tr(A=>{let{targetSnapshot:i,guards:{canActivateChecks:n}}=A;if(!n.length)return Me(A);let o=new Set(n.map(a=>a.route)),r=new Set;for(let a of o)if(!r.has(a))for(let c of CO(a))r.add(c);let s=0;return oo(r).pipe(ql(a=>o.has(a)?ghA(a,i,t,e):(a.data=_6(a,a.parent,t).resolve,Me(void 0))),Qo(()=>s++),bd(1),tr(a=>s===r.size?Me(A):sr))})}function CO(t){let e=t.children.map(A=>CO(A)).flat();return[t,...e]}function ghA(t,e,A,i){let n=t.routeConfig,o=t._resolve;return n?.title!==void 0&&!nO(n)&&(o[fu]=n.title),IhA(o,t,e,i).pipe(je(r=>(t._resolvedData=r,t.data=_6(t,t.parent,A).resolve,null)))}function IhA(t,e,A,i){let n=LM(t);if(n.length===0)return Me({});let o={};return oo(n).pipe(tr(r=>ChA(t[r],e,A,i).pipe(Zl(),Qo(s=>{if(s instanceof EB)throw G6(new L2,s);o[r]=s}))),bd(1),je(()=>o),mr(r=>cO(r)?sr:g2(r)))}function ChA(t,e,A,i){let n=mu(e)??i,o=hB(t,n),r=o.resolve?o.resolve(e,A):ga(n,()=>o(e,A));return N2(r)}function RM(t){return jn(e=>{let A=t(e);return A?oo(A).pipe(je(()=>e)):Me(e)})}var WM=(()=>{class t{buildTitle(A){let i,n=A.root;for(;n!==void 0;)i=this.getResolvedTitleForRoute(n)??i,n=n.children.find(o=>o.outlet===Li);return i}getResolvedTitleForRoute(A){return A.data[fu]}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(dO),providedIn:"root"})}return t})(),dO=(()=>{class t extends WM{title;constructor(A){super(),this.title=A}updateTitle(A){let i=this.buildTitle(A);i!==void 0&&this.title.setTitle(i)}static \u0275fac=function(i){return new(i||t)(we(iz))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),zI=new dA("",{providedIn:"root",factory:()=>({})}),uB=new dA(""),J6=(()=>{class t{componentLoaders=new WeakMap;childrenLoaders=new WeakMap;onLoadStartListener;onLoadEndListener;compiler=f(dH);loadComponent(A){if(this.componentLoaders.get(A))return this.componentLoaders.get(A);if(A._loadedComponent)return Me(A._loadedComponent);this.onLoadStartListener&&this.onLoadStartListener(A);let i=N2(A.loadComponent()).pipe(je(EO),Qo(o=>{this.onLoadEndListener&&this.onLoadEndListener(A),A._loadedComponent=o}),Vl(()=>{this.componentLoaders.delete(A)})),n=new l2(i,()=>new OA).pipe(fd());return this.componentLoaders.set(A,n),n}loadChildren(A,i){if(this.childrenLoaders.get(i))return this.childrenLoaders.get(i);if(i._loadedRoutes)return Me({routes:i._loadedRoutes,injector:i._loadedInjector});this.onLoadStartListener&&this.onLoadStartListener(i);let o=BO(i,this.compiler,A,this.onLoadEndListener).pipe(Vl(()=>{this.childrenLoaders.delete(i)})),r=new l2(o,()=>new OA).pipe(fd());return this.childrenLoaders.set(i,r),r}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function BO(t,e,A,i){return N2(t.loadChildren()).pipe(je(EO),tr(n=>n instanceof Lb||Array.isArray(n)?Me(n):oo(e.compileModuleAsync(n))),je(n=>{i&&i(t);let o,r,s=!1;return Array.isArray(n)?(r=n,s=!0):(o=n.create(A).injector,r=o.get(uB,[],{optional:!0,self:!0}).flat()),{routes:r.map(ZM),injector:o}}))}function dhA(t){return t&&typeof t=="object"&&"default"in t}function EO(t){return dhA(t)?t.default:t}var T6=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(BhA),providedIn:"root"})}return t})(),BhA=(()=>{class t{shouldProcessUrl(A){return!0}extract(A){return A}merge(A,i){return A}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),XM=new dA(""),$M=new dA("");function QO(t,e,A){let i=t.get($M),n=t.get(at);return t.get(Qe).runOutsideAngular(()=>{if(!n.startViewTransition||i.skipNextTransition)return i.skipNextTransition=!1,new Promise(c=>setTimeout(c));let o,r=new Promise(c=>{o=c}),s=n.startViewTransition(()=>(o(),EhA(t))),{onViewTransitionCreated:a}=i;return a&&ga(t,()=>a({transition:s,from:e,to:A})),r})}function EhA(t){return new Promise(e=>{To({read:()=>setTimeout(e)},{injector:t})})}var Ak=new dA(""),H6=(()=>{class t{currentNavigation=null;currentTransition=null;lastSuccessfulNavigation=null;events=new OA;transitionAbortSubject=new OA;configLoader=f(J6);environmentInjector=f(pr);destroyRef=f(m2);urlSerializer=f(TI);rootContexts=f(HI);location=f(Dc);inputBindingEnabled=f(pu,{optional:!0})!==null;titleStrategy=f(WM);options=f(zI,{optional:!0})||{};paramsInheritanceStrategy=this.options.paramsInheritanceStrategy||"emptyOnly";urlHandlingStrategy=f(T6);createViewTransition=f(XM,{optional:!0});navigationErrorHandler=f(Ak,{optional:!0});navigationId=0;get hasRequestedNavigation(){return this.navigationId!==0}transitions;afterPreactivation=()=>Me(void 0);rootComponentType=null;destroyed=!1;constructor(){let A=n=>this.events.next(new M6(n)),i=n=>this.events.next(new k6(n));this.configLoader.onLoadEndListener=i,this.configLoader.onLoadStartListener=A,this.destroyRef.onDestroy(()=>{this.destroyed=!0})}complete(){this.transitions?.complete()}handleNavigationRequest(A){let i=++this.navigationId;this.transitions?.next(Ye(rA({},A),{extractedUrl:this.urlHandlingStrategy.extract(A.rawUrl),targetSnapshot:null,targetRouterState:null,guards:{canActivateChecks:[],canDeactivateChecks:[]},guardsResult:null,id:i}))}setupNavigations(A){return this.transitions=new Mi(null),this.transitions.pipe(kt(i=>i!==null),jn(i=>{let n=!1,o=!1;return Me(i).pipe(jn(r=>{if(this.navigationId>i.id)return this.cancelNavigationTransition(i,"",Qa.SupersededByNewNavigation),sr;this.currentTransition=i,this.currentNavigation={id:r.id,initialUrl:r.rawUrl,extractedUrl:r.extractedUrl,targetBrowserUrl:typeof r.extras.browserUrl=="string"?this.urlSerializer.parse(r.extras.browserUrl):r.extras.browserUrl,trigger:r.source,extras:r.extras,previousNavigation:this.lastSuccessfulNavigation?Ye(rA({},this.lastSuccessfulNavigation),{previousNavigation:null}):null};let s=!A.navigated||this.isUpdatingInternalState()||this.isUpdatedBrowserUrl(),a=r.extras.onSameUrlNavigation??A.onSameUrlNavigation;if(!s&&a!=="reload"){let c="";return this.events.next(new gg(r.id,this.urlSerializer.serialize(r.rawUrl),c,IB.IgnoredSameUrlNavigation)),r.resolve(!1),sr}if(this.urlHandlingStrategy.shouldProcessUrl(r.rawUrl))return Me(r).pipe(jn(c=>(this.events.next(new F2(c.id,this.urlSerializer.serialize(c.extractedUrl),c.source,c.restoredState)),c.id!==this.navigationId?sr:Promise.resolve(c))),chA(this.environmentInjector,this.configLoader,this.rootComponentType,A.config,this.urlSerializer,this.paramsInheritanceStrategy),Qo(c=>{i.targetSnapshot=c.targetSnapshot,i.urlAfterRedirects=c.urlAfterRedirects,this.currentNavigation=Ye(rA({},this.currentNavigation),{finalUrl:c.urlAfterRedirects});let l=new Cu(c.id,this.urlSerializer.serialize(c.extractedUrl),this.urlSerializer.serialize(c.urlAfterRedirects),c.targetSnapshot);this.events.next(l)}));if(s&&this.urlHandlingStrategy.shouldProcessUrl(r.currentRawUrl)){let{id:c,extractedUrl:l,source:I,restoredState:C,extras:d}=r,B=new F2(c,this.urlSerializer.serialize(l),I,C);this.events.next(B);let E=tO(this.rootComponentType).snapshot;return this.currentTransition=i=Ye(rA({},r),{targetSnapshot:E,urlAfterRedirects:l,extras:Ye(rA({},d),{skipLocationChange:!1,replaceUrl:!1})}),this.currentNavigation.finalUrl=l,Me(i)}else{let c="";return this.events.next(new gg(r.id,this.urlSerializer.serialize(r.extractedUrl),c,IB.IgnoredByUrlHandlingStrategy)),r.resolve(!1),sr}}),Qo(r=>{let s=new D6(r.id,this.urlSerializer.serialize(r.extractedUrl),this.urlSerializer.serialize(r.urlAfterRedirects),r.targetSnapshot);this.events.next(s)}),je(r=>(this.currentTransition=i=Ye(rA({},r),{guards:kQA(r.targetSnapshot,r.currentSnapshot,this.rootContexts)}),i)),YQA(this.environmentInjector,r=>this.events.next(r)),Qo(r=>{if(i.guardsResult=r.guardsResult,r.guardsResult&&typeof r.guardsResult!="boolean")throw G6(this.urlSerializer,r.guardsResult);let s=new y6(r.id,this.urlSerializer.serialize(r.extractedUrl),this.urlSerializer.serialize(r.urlAfterRedirects),r.targetSnapshot,!!r.guardsResult);this.events.next(s)}),kt(r=>r.guardsResult?!0:(this.cancelNavigationTransition(r,"",Qa.GuardRejected),!1)),RM(r=>{if(r.guards.canActivateChecks.length!==0)return Me(r).pipe(Qo(s=>{let a=new v6(s.id,this.urlSerializer.serialize(s.extractedUrl),this.urlSerializer.serialize(s.urlAfterRedirects),s.targetSnapshot);this.events.next(a)}),jn(s=>{let a=!1;return Me(s).pipe(lhA(this.paramsInheritanceStrategy,this.environmentInjector),Qo({next:()=>a=!0,complete:()=>{a||this.cancelNavigationTransition(s,"",Qa.NoDataFromResolver)}}))}),Qo(s=>{let a=new b6(s.id,this.urlSerializer.serialize(s.extractedUrl),this.urlSerializer.serialize(s.urlAfterRedirects),s.targetSnapshot);this.events.next(a)}))}),RM(r=>{let s=a=>{let c=[];a.routeConfig?.loadComponent&&!a.routeConfig._loadedComponent&&c.push(this.configLoader.loadComponent(a.routeConfig).pipe(Qo(l=>{a.component=l}),je(()=>{})));for(let l of a.children)c.push(...s(l));return c};return Js(s(r.targetSnapshot.root)).pipe(B2(null),On(1))}),RM(()=>this.afterPreactivation()),jn(()=>{let{currentSnapshot:r,targetSnapshot:s}=i,a=this.createViewTransition?.(this.environmentInjector,r.root,s.root);return a?oo(a).pipe(je(()=>i)):Me(i)}),je(r=>{let s=DQA(A.routeReuseStrategy,r.targetSnapshot,r.currentRouterState);return this.currentTransition=i=Ye(rA({},r),{targetRouterState:s}),this.currentNavigation.targetRouterState=s,i}),Qo(()=>{this.events.next(new du)}),MQA(this.rootContexts,A.routeReuseStrategy,r=>this.events.next(r),this.inputBindingEnabled),On(1),Qo({next:r=>{n=!0,this.lastSuccessfulNavigation=this.currentNavigation,this.events.next(new nc(r.id,this.urlSerializer.serialize(r.extractedUrl),this.urlSerializer.serialize(r.urlAfterRedirects))),this.titleStrategy?.updateTitle(r.targetRouterState.snapshot),r.resolve(!0)},complete:()=>{n=!0}}),St(this.transitionAbortSubject.pipe(Qo(r=>{throw r}))),Vl(()=>{!n&&!o&&this.cancelNavigationTransition(i,"",Qa.SupersededByNewNavigation),this.currentTransition?.id===i.id&&(this.currentNavigation=null,this.currentTransition=null)}),mr(r=>{if(this.destroyed)return i.resolve(!1),sr;if(o=!0,aO(r))this.events.next(new cg(i.id,this.urlSerializer.serialize(i.extractedUrl),r.message,r.cancellationCode)),bQA(r)?this.events.next(new BB(r.url,r.navigationBehaviorOptions)):i.resolve(!1);else{let s=new CB(i.id,this.urlSerializer.serialize(i.extractedUrl),r,i.targetSnapshot??void 0);try{let a=ga(this.environmentInjector,()=>this.navigationErrorHandler?.(s));if(a instanceof EB){let{message:c,cancellationCode:l}=G6(this.urlSerializer,a);this.events.next(new cg(i.id,this.urlSerializer.serialize(i.extractedUrl),c,l)),this.events.next(new BB(a.redirectTo,a.navigationBehaviorOptions))}else throw this.events.next(s),r}catch(a){this.options.resolveNavigationPromiseOnError?i.resolve(!1):i.reject(a)}}return sr}))}))}cancelNavigationTransition(A,i,n){let o=new cg(A.id,this.urlSerializer.serialize(A.extractedUrl),i,n);this.events.next(o),A.resolve(!1)}isUpdatingInternalState(){return this.currentTransition?.extractedUrl.toString()!==this.currentTransition?.currentUrlTree.toString()}isUpdatedBrowserUrl(){let A=this.urlHandlingStrategy.extract(this.urlSerializer.parse(this.location.path(!0))),i=this.currentNavigation?.targetBrowserUrl??this.currentNavigation?.extractedUrl;return A.toString()!==i?.toString()&&!this.currentNavigation?.extras.skipLocationChange}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function QhA(t){return t!==m6}var hO=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(hhA),providedIn:"root"})}return t})(),K6=class{shouldDetach(e){return!1}store(e,A){}shouldAttach(e){return!1}retrieve(e){return null}shouldReuseRoute(e,A){return e.routeConfig===A.routeConfig}},hhA=(()=>{class t extends K6{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),uO=(()=>{class t{urlSerializer=f(TI);options=f(zI,{optional:!0})||{};canceledNavigationResolution=this.options.canceledNavigationResolution||"replace";location=f(Dc);urlHandlingStrategy=f(T6);urlUpdateStrategy=this.options.urlUpdateStrategy||"deferred";currentUrlTree=new lg;getCurrentUrlTree(){return this.currentUrlTree}rawUrlTree=this.currentUrlTree;getRawUrlTree(){return this.rawUrlTree}createBrowserPath({finalUrl:A,initialUrl:i,targetBrowserUrl:n}){let o=A!==void 0?this.urlHandlingStrategy.merge(A,i):i,r=n??o;return r instanceof lg?this.urlSerializer.serialize(r):r}commitTransition({targetRouterState:A,finalUrl:i,initialUrl:n}){i&&A?(this.currentUrlTree=i,this.rawUrlTree=this.urlHandlingStrategy.merge(i,n),this.routerState=A):this.rawUrlTree=n}routerState=tO(null);getRouterState(){return this.routerState}stateMemento=this.createStateMemento();updateStateMemento(){this.stateMemento=this.createStateMemento()}createStateMemento(){return{rawUrlTree:this.rawUrlTree,currentUrlTree:this.currentUrlTree,routerState:this.routerState}}resetInternalState({finalUrl:A}){this.routerState=this.stateMemento.routerState,this.currentUrlTree=this.stateMemento.currentUrlTree,this.rawUrlTree=this.urlHandlingStrategy.merge(this.currentUrlTree,A??this.rawUrlTree)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:()=>f(uhA),providedIn:"root"})}return t})(),uhA=(()=>{class t extends uO{currentPageId=0;lastSuccessfulId=-1;restoredState(){return this.location.getState()}get browserPageId(){return this.canceledNavigationResolution!=="computed"?this.currentPageId:this.restoredState()?.\u0275routerPageId??this.currentPageId}registerNonRouterCurrentEntryChangeListener(A){return this.location.subscribe(i=>{i.type==="popstate"&&setTimeout(()=>{A(i.url,i.state,"popstate")})})}handleRouterEvent(A,i){A instanceof F2?this.updateStateMemento():A instanceof gg?this.commitTransition(i):A instanceof Cu?this.urlUpdateStrategy==="eager"&&(i.extras.skipLocationChange||this.setBrowserUrl(this.createBrowserPath(i),i)):A instanceof du?(this.commitTransition(i),this.urlUpdateStrategy==="deferred"&&!i.extras.skipLocationChange&&this.setBrowserUrl(this.createBrowserPath(i),i)):A instanceof cg&&(A.code===Qa.GuardRejected||A.code===Qa.NoDataFromResolver)?this.restoreHistory(i):A instanceof CB?this.restoreHistory(i,!0):A instanceof nc&&(this.lastSuccessfulId=A.id,this.currentPageId=this.browserPageId)}setBrowserUrl(A,{extras:i,id:n}){let{replaceUrl:o,state:r}=i;if(this.location.isCurrentPathEqualTo(A)||o){let s=this.browserPageId,a=rA(rA({},r),this.generateNgRouterState(n,s));this.location.replaceState(A,"",a)}else{let s=rA(rA({},r),this.generateNgRouterState(n,this.browserPageId+1));this.location.go(A,"",s)}}restoreHistory(A,i=!1){if(this.canceledNavigationResolution==="computed"){let n=this.browserPageId,o=this.currentPageId-n;o!==0?this.location.historyGo(o):this.getCurrentUrlTree()===A.finalUrl&&o===0&&(this.resetInternalState(A),this.resetUrlToCurrentUrlTree())}else this.canceledNavigationResolution==="replace"&&(i&&this.resetInternalState(A),this.resetUrlToCurrentUrlTree())}resetUrlToCurrentUrlTree(){this.location.replaceState(this.urlSerializer.serialize(this.getRawUrlTree()),"",this.generateNgRouterState(this.lastSuccessfulId,this.currentPageId))}generateNgRouterState(A,i){return this.canceledNavigationResolution==="computed"?{navigationId:A,\u0275routerPageId:i}:{navigationId:A}}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function z6(t,e){t.events.pipe(kt(A=>A instanceof nc||A instanceof cg||A instanceof CB||A instanceof gg),je(A=>A instanceof nc||A instanceof gg?0:(A instanceof cg?A.code===Qa.Redirect||A.code===Qa.SupersededByNewNavigation:!1)?2:1),kt(A=>A!==2),On(1)).subscribe(()=>{e()})}var fhA={paths:"exact",fragment:"ignored",matrixParams:"ignored",queryParams:"exact"},mhA={paths:"subset",fragment:"ignored",matrixParams:"ignored",queryParams:"subset"},Ig=(()=>{class t{get currentUrlTree(){return this.stateManager.getCurrentUrlTree()}get rawUrlTree(){return this.stateManager.getRawUrlTree()}disposed=!1;nonRouterCurrentEntryChangeSubscription;console=f(_b);stateManager=f(uO);options=f(zI,{optional:!0})||{};pendingTasks=f(B0);urlUpdateStrategy=this.options.urlUpdateStrategy||"deferred";navigationTransitions=f(H6);urlSerializer=f(TI);location=f(Dc);urlHandlingStrategy=f(T6);_events=new OA;get events(){return this._events}get routerState(){return this.stateManager.getRouterState()}navigated=!1;routeReuseStrategy=f(hO);onSameUrlNavigation=this.options.onSameUrlNavigation||"ignore";config=f(uB,{optional:!0})?.flat()??[];componentInputBindingEnabled=!!f(pu,{optional:!0});constructor(){this.resetConfig(this.config),this.navigationTransitions.setupNavigations(this).subscribe({error:A=>{this.console.warn(A)}}),this.subscribeToNavigationEvents()}eventsSubscription=new Kt;subscribeToNavigationEvents(){let A=this.navigationTransitions.events.subscribe(i=>{try{let n=this.navigationTransitions.currentTransition,o=this.navigationTransitions.currentNavigation;if(n!==null&&o!==null){if(this.stateManager.handleRouterEvent(i,o),i instanceof cg&&i.code!==Qa.Redirect&&i.code!==Qa.SupersededByNewNavigation)this.navigated=!0;else if(i instanceof nc)this.navigated=!0;else if(i instanceof BB){let r=i.navigationBehaviorOptions,s=this.urlHandlingStrategy.merge(i.url,n.currentRawUrl),a=rA({browserUrl:n.extras.browserUrl,info:n.extras.info,skipLocationChange:n.extras.skipLocationChange,replaceUrl:n.extras.replaceUrl||this.urlUpdateStrategy==="eager"||QhA(n.source)},r);this.scheduleNavigation(s,m6,null,a,{resolve:n.resolve,reject:n.reject,promise:n.promise})}}whA(i)&&this._events.next(i)}catch(n){this.navigationTransitions.transitionAbortSubject.next(n)}});this.eventsSubscription.add(A)}resetRootComponentType(A){this.routerState.root.component=A,this.navigationTransitions.rootComponentType=A}initialNavigation(){this.setUpLocationChangeListener(),this.navigationTransitions.hasRequestedNavigation||this.navigateToSyncWithBrowser(this.location.path(!0),m6,this.stateManager.restoredState())}setUpLocationChangeListener(){this.nonRouterCurrentEntryChangeSubscription??=this.stateManager.registerNonRouterCurrentEntryChangeListener((A,i,n)=>{this.navigateToSyncWithBrowser(A,n,i)})}navigateToSyncWithBrowser(A,i,n){let o={replaceUrl:!0},r=n?.navigationId?n:null;if(n){let a=rA({},n);delete a.navigationId,delete a.\u0275routerPageId,Object.keys(a).length!==0&&(o.state=a)}let s=this.parseUrl(A);this.scheduleNavigation(s,i,r,o)}get url(){return this.serializeUrl(this.currentUrlTree)}getCurrentNavigation(){return this.navigationTransitions.currentNavigation}get lastSuccessfulNavigation(){return this.navigationTransitions.lastSuccessfulNavigation}resetConfig(A){this.config=A.map(ZM),this.navigated=!1}ngOnDestroy(){this.dispose()}dispose(){this._events.unsubscribe(),this.navigationTransitions.complete(),this.nonRouterCurrentEntryChangeSubscription&&(this.nonRouterCurrentEntryChangeSubscription.unsubscribe(),this.nonRouterCurrentEntryChangeSubscription=void 0),this.disposed=!0,this.eventsSubscription.unsubscribe()}createUrlTree(A,i={}){let{relativeTo:n,queryParams:o,fragment:r,queryParamsHandling:s,preserveFragment:a}=i,c=a?this.currentUrlTree.fragment:r,l=null;switch(s??this.options.defaultQueryParamsHandling){case"merge":l=rA(rA({},this.currentUrlTree.queryParams),o);break;case"preserve":l=this.currentUrlTree.queryParams;break;default:l=o||null}l!==null&&(l=this.removeEmptyProps(l));let I;try{let C=n?n.snapshot:this.routerState.snapshot.root;I=Xz(C)}catch{(typeof A[0]!="string"||A[0][0]!=="/")&&(A=[]),I=this.currentUrlTree.root}return $z(I,A,l,c??null)}navigateByUrl(A,i={skipLocationChange:!1}){let n=gB(A)?A:this.parseUrl(A),o=this.urlHandlingStrategy.merge(n,this.rawUrlTree);return this.scheduleNavigation(o,m6,null,i)}navigate(A,i={skipLocationChange:!1}){return phA(A),this.navigateByUrl(this.createUrlTree(A,i),i)}serializeUrl(A){return this.urlSerializer.serialize(A)}parseUrl(A){try{return this.urlSerializer.parse(A)}catch{return this.urlSerializer.parse("/")}}isActive(A,i){let n;if(i===!0?n=rA({},fhA):i===!1?n=rA({},mhA):n=i,gB(A))return Lz(this.currentUrlTree,A,n);let o=this.parseUrl(A);return Lz(this.currentUrlTree,o,n)}removeEmptyProps(A){return Object.entries(A).reduce((i,[n,o])=>(o!=null&&(i[n]=o),i),{})}scheduleNavigation(A,i,n,o,r){if(this.disposed)return Promise.resolve(!1);let s,a,c;r?(s=r.resolve,a=r.reject,c=r.promise):c=new Promise((I,C)=>{s=I,a=C});let l=this.pendingTasks.add();return z6(this,()=>{queueMicrotask(()=>this.pendingTasks.remove(l))}),this.navigationTransitions.handleNavigationRequest({source:i,restoredState:n,currentUrlTree:this.currentUrlTree,currentRawUrl:this.currentUrlTree,rawUrl:A,extras:o,resolve:s,reject:a,promise:c,currentSnapshot:this.routerState.snapshot,currentRouterState:this.routerState}),c.catch(I=>Promise.reject(I))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function phA(t){for(let e=0;e<t.length;e++)if(t[e]==null)throw new Ae(4008,!1)}function whA(t){return!(t instanceof du)&&!(t instanceof BB)}var Du=class{};var fO=(()=>{class t{router;injector;preloadingStrategy;loader;subscription;constructor(A,i,n,o){this.router=A,this.injector=i,this.preloadingStrategy=n,this.loader=o}setUpPreloading(){this.subscription=this.router.events.pipe(kt(A=>A instanceof nc),ql(()=>this.preload())).subscribe(()=>{})}preload(){return this.processRoutes(this.injector,this.router.config)}ngOnDestroy(){this.subscription&&this.subscription.unsubscribe()}processRoutes(A,i){let n=[];for(let o of i){o.providers&&!o._injector&&(o._injector=Rh(o.providers,A,`Route: ${o.path}`));let r=o._injector??A,s=o._loadedInjector??r;(o.loadChildren&&!o._loadedRoutes&&o.canLoad===void 0||o.loadComponent&&!o._loadedComponent)&&n.push(this.preloadConfig(r,o)),(o.children||o._loadedRoutes)&&n.push(this.processRoutes(s,o.children??o._loadedRoutes))}return oo(n).pipe(C2())}preloadConfig(A,i){return this.preloadingStrategy.preload(i,()=>{let n;i.loadChildren&&i.canLoad===void 0?n=this.loader.loadChildren(A,i):n=Me(null);let o=n.pipe(tr(r=>r===null?Me(void 0):(i._loadedRoutes=r.routes,i._loadedInjector=r.injector,this.processRoutes(r.injector??A,r.routes))));if(i.loadComponent&&!i._loadedComponent){let r=this.loader.loadComponent(i);return oo([o,r]).pipe(C2())}else return o})}static \u0275fac=function(i){return new(i||t)(we(Ig),we(pr),we(Du),we(J6))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),mO=new dA(""),DhA=(()=>{class t{urlSerializer;transitions;viewportScroller;zone;options;routerEventsSubscription;scrollEventsSubscription;lastId=0;lastSource="imperative";restoredId=0;store={};constructor(A,i,n,o,r={}){this.urlSerializer=A,this.transitions=i,this.viewportScroller=n,this.zone=o,this.options=r,r.scrollPositionRestoration||="disabled",r.anchorScrolling||="disabled"}init(){this.options.scrollPositionRestoration!=="disabled"&&this.viewportScroller.setHistoryScrollRestoration("manual"),this.routerEventsSubscription=this.createScrollEvents(),this.scrollEventsSubscription=this.consumeScrollEvents()}createScrollEvents(){return this.transitions.events.subscribe(A=>{A instanceof F2?(this.store[this.lastId]=this.viewportScroller.getScrollPosition(),this.lastSource=A.navigationTrigger,this.restoredId=A.restoredState?A.restoredState.navigationId:0):A instanceof nc?(this.lastId=A.id,this.scheduleScrollEvent(A,this.urlSerializer.parse(A.urlAfterRedirects).fragment)):A instanceof gg&&A.code===IB.IgnoredSameUrlNavigation&&(this.lastSource=void 0,this.restoredId=0,this.scheduleScrollEvent(A,this.urlSerializer.parse(A.url).fragment))})}consumeScrollEvents(){return this.transitions.events.subscribe(A=>{A instanceof dB&&(A.position?this.options.scrollPositionRestoration==="top"?this.viewportScroller.scrollToPosition([0,0]):this.options.scrollPositionRestoration==="enabled"&&this.viewportScroller.scrollToPosition(A.position):A.anchor&&this.options.anchorScrolling==="enabled"?this.viewportScroller.scrollToAnchor(A.anchor):this.options.scrollPositionRestoration!=="disabled"&&this.viewportScroller.scrollToPosition([0,0]))})}scheduleScrollEvent(A,i){this.zone.runOutsideAngular(()=>{setTimeout(()=>{this.zone.run(()=>{this.transitions.events.next(new dB(A,this.lastSource==="popstate"?this.store[this.restoredId]:null,i))})},0)})}ngOnDestroy(){this.routerEventsSubscription?.unsubscribe(),this.scrollEventsSubscription?.unsubscribe()}static \u0275fac=function(i){yT()};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();function yhA(t){return t.routerState.root}function yu(t,e){return{\u0275kind:t,\u0275providers:e}}function vhA(){let t=f(Rt);return e=>{let A=t.get(la);if(e!==A.components[0])return;let i=t.get(Ig),n=t.get(pO);t.get(tk)===1&&i.initialNavigation(),t.get(yO,null,Gi.Optional)?.setUpPreloading(),t.get(mO,null,Gi.Optional)?.init(),i.resetRootComponentType(A.componentTypes[0]),n.closed||(n.next(),n.complete(),n.unsubscribe())}}var pO=new dA("",{factory:()=>new OA}),tk=new dA("",{providedIn:"root",factory:()=>1});function wO(){let t=[{provide:tk,useValue:0},Yb(()=>{let e=f(Rt);return e.get(Vb,Promise.resolve()).then(()=>new Promise(i=>{let n=e.get(Ig),o=e.get(pO);z6(n,()=>{i(!0)}),e.get(H6).afterPreactivation=()=>(i(!0),o.closed?Me(void 0):o),n.initialNavigation()}))})];return yu(2,t)}function DO(){let t=[Yb(()=>{f(Ig).setUpLocationChangeListener()}),{provide:tk,useValue:2}];return yu(3,t)}var yO=new dA("");function vO(t){return yu(0,[{provide:yO,useExisting:fO},{provide:Du,useExisting:t}])}function bO(){return yu(8,[qM,{provide:pu,useExisting:qM}])}function MO(t){E0("NgRouterViewTransitions");let e=[{provide:XM,useValue:QO},{provide:$M,useValue:rA({skipNextTransition:!!t?.skipInitialTransition},t)}];return yu(9,e)}var kO=[Dc,{provide:TI,useClass:L2},Ig,HI,{provide:ha,useFactory:yhA,deps:[Ig]},J6,[]],O6=(()=>{class t{constructor(){}static forRoot(A,i){return{ngModule:t,providers:[kO,[],{provide:uB,multi:!0,useValue:A},[],i?.errorHandler?{provide:Ak,useValue:i.errorHandler}:[],{provide:zI,useValue:i||{}},i?.useHash?MhA():khA(),bhA(),i?.preloadingStrategy?vO(i.preloadingStrategy).\u0275providers:[],i?.initialNavigation?ShA(i):[],i?.bindToComponentInputs?bO().\u0275providers:[],i?.enableViewTransitions?MO().\u0275providers:[],RhA()]}}static forChild(A){return{ngModule:t,providers:[{provide:uB,multi:!0,useValue:A}]}}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();function bhA(){return{provide:mO,useFactory:()=>{let t=f(xH),e=f(Qe),A=f(zI),i=f(H6),n=f(TI);return A.scrollOffset&&t.setOffset(A.scrollOffset),new DhA(n,i,t,e,A)}}}function MhA(){return{provide:u0,useClass:$b}}function khA(){return{provide:u0,useClass:Yp}}function ShA(t){return[t.initialNavigation==="disabled"?DO().\u0275providers:[],t.initialNavigation==="enabledBlocking"?wO().\u0275providers:[]]}var ek=new dA("");function RhA(){return[{provide:ek,useFactory:vhA},{provide:Jb,multi:!0,useExisting:ek}]}var nk;try{nk=typeof Intl<"u"&&Intl.v8BreakIterator}catch{nk=!1}var Ii=(()=>{class t{_platformId=f(og);isBrowser=this._platformId?sg(this._platformId):typeof document=="object"&&!!document;EDGE=this.isBrowser&&/(edge)/i.test(navigator.userAgent);TRIDENT=this.isBrowser&&/(msie|trident)/i.test(navigator.userAgent);BLINK=this.isBrowser&&!!(window.chrome||nk)&&typeof CSS<"u"&&!this.EDGE&&!this.TRIDENT;WEBKIT=this.isBrowser&&/AppleWebKit/i.test(navigator.userAgent)&&!this.BLINK&&!this.EDGE&&!this.TRIDENT;IOS=this.isBrowser&&/iPad|iPhone|iPod/.test(navigator.userAgent)&&!("MSStream"in window);FIREFOX=this.isBrowser&&/(firefox|minefield)/i.test(navigator.userAgent);ANDROID=this.isBrowser&&/android/i.test(navigator.userAgent)&&!this.TRIDENT;SAFARI=this.isBrowser&&/safari/i.test(navigator.userAgent)&&this.WEBKIT;constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var fB,SO=["color","button","checkbox","date","datetime-local","email","file","hidden","image","month","number","password","radio","range","reset","search","submit","tel","text","time","url","week"];function ok(){if(fB)return fB;if(typeof document!="object"||!document)return fB=new Set(SO),fB;let t=document.createElement("input");return fB=new Set(SO.filter(e=>(t.setAttribute("type",e),t.type===e))),fB}var vu;function FhA(){if(vu==null&&typeof window<"u")try{window.addEventListener("test",null,Object.defineProperty({},"passive",{get:()=>vu=!0}))}finally{vu=vu||!1}return vu}function bc(t){return FhA()?t:!!t.capture}var gl=function(t){return t[t.NORMAL=0]="NORMAL",t[t.NEGATED=1]="NEGATED",t[t.INVERTED=2]="INVERTED",t}(gl||{}),P6,OI;function j6(){if(OI==null){if(typeof document!="object"||!document||typeof Element!="function"||!Element)return OI=!1,OI;if("scrollBehavior"in document.documentElement.style)OI=!0;else{let t=Element.prototype.scrollTo;t?OI=!/\{\s*\[native code\]\s*\}/.test(t.toString()):OI=!1}}return OI}function mB(){if(typeof document!="object"||!document)return gl.NORMAL;if(P6==null){let t=document.createElement("div"),e=t.style;t.dir="rtl",e.width="1px",e.overflow="auto",e.visibility="hidden",e.pointerEvents="none",e.position="absolute";let A=document.createElement("div"),i=A.style;i.width="2px",i.height="1px",t.appendChild(A),document.body.appendChild(t),P6=gl.NORMAL,t.scrollLeft===0&&(t.scrollLeft=1,P6=t.scrollLeft===0?gl.NEGATED:gl.INVERTED),t.remove()}return P6}var ik;function NhA(){if(ik==null){let t=typeof document<"u"?document.head:null;ik=!!(t&&(t.createShadowRoot||t.attachShadow))}return ik}function RO(t){if(NhA()){let e=t.getRootNode?t.getRootNode():null;if(typeof ShadowRoot<"u"&&ShadowRoot&&e instanceof ShadowRoot)return e}return null}function pB(){let t=typeof document<"u"&&document?document.activeElement:null;for(;t&&t.shadowRoot;){let e=t.shadowRoot.activeElement;if(e===t)break;t=e}return t}function oc(t){return t.composedPath?t.composedPath()[0]:t.target}function rk(){return typeof __karma__<"u"&&!!__karma__||typeof jasmine<"u"&&!!jasmine||typeof jest<"u"&&!!jest||typeof Mocha<"u"&&!!Mocha}function sk(t,e,A,i,n){let o=parseInt(Pb.major),r=parseInt(Pb.minor);return o>19||o===19&&r>0||o===0&&r===0?t.listen(e,A,i,n):(e.addEventListener(A,i,n),()=>{e.removeEventListener(A,i,n)})}var q6=new WeakMap,Rn=(()=>{class t{_appRef;_injector=f(Rt);_environmentInjector=f(pr);load(A){let i=this._appRef=this._appRef||this._injector.get(la),n=q6.get(i);n||(n={loaders:new Set,refs:[]},q6.set(i,n),i.onDestroy(()=>{q6.get(i)?.refs.forEach(o=>o.destroy()),q6.delete(i)})),n.loaders.has(A)||(n.loaders.add(A),n.refs.push(Gp(A,{environmentInjector:this._environmentInjector})))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),bu=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["ng-component"]],exportAs:["cdkVisuallyHidden"],decls:0,vars:0,template:function(i,n){},styles:[".cdk-visually-hidden{border:0;clip:rect(0 0 0 0);height:1px;margin:-1px;overflow:hidden;padding:0;position:absolute;width:1px;white-space:nowrap;outline:0;-webkit-appearance:none;-moz-appearance:none;left:0}[dir=rtl] .cdk-visually-hidden{left:auto;right:0}"],encapsulation:2,changeDetection:0})}return t})();function ir(t,...e){return e.length?e.some(A=>t[A]):t.altKey||t.shiftKey||t.ctrlKey||t.metaKey}function Xo(t){return t!=null&&`${t}`!="false"}function zs(t,e=0){return ak(t)?Number(t):arguments.length===2?e:0}function ak(t){return!isNaN(parseFloat(t))&&!isNaN(Number(t))}function wB(t){return Array.isArray(t)?t:[t]}function Cr(t){return t==null?"":typeof t=="string"?t:`${t}px`}function ua(t){return t instanceof ee?t.nativeElement:t}function _hA(t){if(t.type==="characterData"&&t.target instanceof Comment)return!0;if(t.type==="childList"){for(let e=0;e<t.addedNodes.length;e++)if(!(t.addedNodes[e]instanceof Comment))return!1;for(let e=0;e<t.removedNodes.length;e++)if(!(t.removedNodes[e]instanceof Comment))return!1;return!0}return!1}var xO=(()=>{class t{create(A){return typeof MutationObserver>"u"?null:new MutationObserver(A)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),LO=(()=>{class t{_mutationObserverFactory=f(xO);_observedElements=new Map;_ngZone=f(Qe);constructor(){}ngOnDestroy(){this._observedElements.forEach((A,i)=>this._cleanupObserver(i))}observe(A){let i=ua(A);return new ct(n=>{let r=this._observeElement(i).pipe(je(s=>s.filter(a=>!_hA(a))),kt(s=>!!s.length)).subscribe(s=>{this._ngZone.run(()=>{n.next(s)})});return()=>{r.unsubscribe(),this._unobserveElement(i)}})}_observeElement(A){return this._ngZone.runOutsideAngular(()=>{if(this._observedElements.has(A))this._observedElements.get(A).count++;else{let i=new OA,n=this._mutationObserverFactory.create(o=>i.next(o));n&&n.observe(A,{characterData:!0,childList:!0,subtree:!0}),this._observedElements.set(A,{observer:n,stream:i,count:1})}return this._observedElements.get(A).stream})}_unobserveElement(A){this._observedElements.has(A)&&(this._observedElements.get(A).count--,this._observedElements.get(A).count||this._cleanupObserver(A))}_cleanupObserver(A){if(this._observedElements.has(A)){let{observer:i,stream:n}=this._observedElements.get(A);i&&i.disconnect(),n.complete(),this._observedElements.delete(A)}}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),V6=(()=>{class t{_contentObserver=f(LO);_elementRef=f(ee);event=new WA;get disabled(){return this._disabled}set disabled(A){this._disabled=A,this._disabled?this._unsubscribe():this._subscribe()}_disabled=!1;get debounce(){return this._debounce}set debounce(A){this._debounce=zs(A),this._subscribe()}_debounce;_currentSubscription=null;constructor(){}ngAfterContentInit(){!this._currentSubscription&&!this.disabled&&this._subscribe()}ngOnDestroy(){this._unsubscribe()}_subscribe(){this._unsubscribe();let A=this._contentObserver.observe(this._elementRef);this._currentSubscription=(this.debounce?A.pipe(el(this.debounce)):A).subscribe(this.event)}_unsubscribe(){this._currentSubscription?.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkObserveContent",""]],inputs:{disabled:[2,"cdkObserveContentDisabled","disabled",ie],debounce:"debounce"},outputs:{event:"cdkObserveContent"},exportAs:["cdkObserveContent"]})}return t})(),DB=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[xO]})}return t})();var FO=new Set,PI,GhA=(()=>{class t{_platform=f(Ii);_nonce=f(yh,{optional:!0});_matchMedia;constructor(){this._matchMedia=this._platform.isBrowser&&window.matchMedia?window.matchMedia.bind(window):KhA}matchMedia(A){return(this._platform.WEBKIT||this._platform.BLINK)&&UhA(A,this._nonce),this._matchMedia(A)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function UhA(t,e){if(!FO.has(t))try{PI||(PI=document.createElement("style"),e&&PI.setAttribute("nonce",e),PI.setAttribute("type","text/css"),document.head.appendChild(PI)),PI.sheet&&(PI.sheet.insertRule(`@media ${t} {body{ }}`,0),FO.add(t))}catch(A){console.error(A)}}function KhA(t){return{matches:t==="all"||t==="",media:t,addListener:()=>{},removeListener:()=>{}}}var Z6=(()=>{class t{_mediaMatcher=f(GhA);_zone=f(Qe);_queries=new Map;_destroySubject=new OA;constructor(){}ngOnDestroy(){this._destroySubject.next(),this._destroySubject.complete()}isMatched(A){return NO(wB(A)).some(n=>this._registerQuery(n).mql.matches)}observe(A){let n=NO(wB(A)).map(r=>this._registerQuery(r).observable),o=Js(n);return o=d2(o.pipe(On(1)),o.pipe(CI(1),el(0))),o.pipe(je(r=>{let s={matches:!1,breakpoints:{}};return r.forEach(({matches:a,query:c})=>{s.matches=s.matches||a,s.breakpoints[c]=a}),s}))}_registerQuery(A){if(this._queries.has(A))return this._queries.get(A);let i=this._mediaMatcher.matchMedia(A),o={observable:new ct(r=>{let s=a=>this._zone.run(()=>r.next(a));return i.addListener(s),()=>{i.removeListener(s)}}).pipe(Pn(i),je(({matches:r})=>({query:A,matches:r})),St(this._destroySubject)),mql:i};return this._queries.set(A,o),o}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function NO(t){return t.map(e=>e.split(",")).reduce((e,A)=>e.concat(A)).map(e=>e.trim())}var _O={XSmall:"(max-width: 599.98px)",Small:"(min-width: 600px) and (max-width: 959.98px)",Medium:"(min-width: 960px) and (max-width: 1279.98px)",Large:"(min-width: 1280px) and (max-width: 1919.98px)",XLarge:"(min-width: 1920px)",Handset:"(max-width: 599.98px) and (orientation: portrait), (max-width: 959.98px) and (orientation: landscape)",Tablet:"(min-width: 600px) and (max-width: 839.98px) and (orientation: portrait), (min-width: 960px) and (max-width: 1279.98px) and (orientation: landscape)",Web:"(min-width: 840px) and (orientation: portrait), (min-width: 1280px) and (orientation: landscape)",HandsetPortrait:"(max-width: 599.98px) and (orientation: portrait)",TabletPortrait:"(min-width: 600px) and (max-width: 839.98px) and (orientation: portrait)",WebPortrait:"(min-width: 840px) and (orientation: portrait)",HandsetLandscape:"(max-width: 959.98px) and (orientation: landscape)",TabletLandscape:"(min-width: 960px) and (max-width: 1279.98px) and (orientation: landscape)",WebLandscape:"(min-width: 1280px) and (orientation: landscape)"};var JO=" ";function Ek(t,e,A){let i=A8(t,e);A=A.trim(),!i.some(n=>n.trim()===A)&&(i.push(A),t.setAttribute(e,i.join(JO)))}function i8(t,e,A){let i=A8(t,e);A=A.trim();let n=i.filter(o=>o!==A);n.length?t.setAttribute(e,n.join(JO)):t.removeAttribute(e)}function A8(t,e){return t.getAttribute(e)?.match(/\S+/g)??[]}var TO="cdk-describedby-message",W6="cdk-describedby-host",Ik=0,HO=(()=>{class t{_platform=f(Ii);_document=f(at);_messageRegistry=new Map;_messagesContainer=null;_id=`${Ik++}`;constructor(){f(Rn).load(bu),this._id=f(Vd)+"-"+Ik++}describe(A,i,n){if(!this._canBeDescribed(A,i))return;let o=ck(i,n);typeof i!="string"?(GO(i,this._id),this._messageRegistry.set(o,{messageElement:i,referenceCount:0})):this._messageRegistry.has(o)||this._createMessageElement(i,n),this._isElementDescribedByMessage(A,o)||this._addMessageReference(A,o)}removeDescription(A,i,n){if(!i||!this._isElementNode(A))return;let o=ck(i,n);if(this._isElementDescribedByMessage(A,o)&&this._removeMessageReference(A,o),typeof i=="string"){let r=this._messageRegistry.get(o);r&&r.referenceCount===0&&this._deleteMessageElement(o)}this._messagesContainer?.childNodes.length===0&&(this._messagesContainer.remove(),this._messagesContainer=null)}ngOnDestroy(){let A=this._document.querySelectorAll(`[${W6}="${this._id}"]`);for(let i=0;i<A.length;i++)this._removeCdkDescribedByReferenceIds(A[i]),A[i].removeAttribute(W6);this._messagesContainer?.remove(),this._messagesContainer=null,this._messageRegistry.clear()}_createMessageElement(A,i){let n=this._document.createElement("div");GO(n,this._id),n.textContent=A,i&&n.setAttribute("role",i),this._createMessagesContainer(),this._messagesContainer.appendChild(n),this._messageRegistry.set(ck(A,i),{messageElement:n,referenceCount:0})}_deleteMessageElement(A){this._messageRegistry.get(A)?.messageElement?.remove(),this._messageRegistry.delete(A)}_createMessagesContainer(){if(this._messagesContainer)return;let A="cdk-describedby-message-container",i=this._document.querySelectorAll(`.${A}[platform="server"]`);for(let o=0;o<i.length;o++)i[o].remove();let n=this._document.createElement("div");n.style.visibility="hidden",n.classList.add(A),n.classList.add("cdk-visually-hidden"),this._platform.isBrowser||n.setAttribute("platform","server"),this._document.body.appendChild(n),this._messagesContainer=n}_removeCdkDescribedByReferenceIds(A){let i=A8(A,"aria-describedby").filter(n=>n.indexOf(TO)!=0);A.setAttribute("aria-describedby",i.join(" "))}_addMessageReference(A,i){let n=this._messageRegistry.get(i);Ek(A,"aria-describedby",n.messageElement.id),A.setAttribute(W6,this._id),n.referenceCount++}_removeMessageReference(A,i){let n=this._messageRegistry.get(i);n.referenceCount--,i8(A,"aria-describedby",n.messageElement.id),A.removeAttribute(W6)}_isElementDescribedByMessage(A,i){let n=A8(A,"aria-describedby"),o=this._messageRegistry.get(i),r=o&&o.messageElement.id;return!!r&&n.indexOf(r)!=-1}_canBeDescribed(A,i){if(!this._isElementNode(A))return!1;if(i&&typeof i=="object")return!0;let n=i==null?"":`${i}`.trim(),o=A.getAttribute("aria-label");return n?!o||o.trim()!==n:!1}_isElementNode(A){return A.nodeType===this._document.ELEMENT_NODE}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function ck(t,e){return typeof t=="string"?`${e||""}/${t}`:t}function GO(t,e){t.id||(t.id=`${TO}-${e}-${Ik++}`)}var $hA=200,Ck=class{_letterKeyStream=new OA;_items=[];_selectedItemIndex=-1;_pressedLetters=[];_skipPredicateFn;_selectedItem=new OA;selectedItem=this._selectedItem;constructor(e,A){let i=typeof A?.debounceInterval=="number"?A.debounceInterval:$hA;A?.skipPredicate&&(this._skipPredicateFn=A.skipPredicate),this.setItems(e),this._setupKeyHandler(i)}destroy(){this._pressedLetters=[],this._letterKeyStream.complete(),this._selectedItem.complete()}setCurrentSelectedItemIndex(e){this._selectedItemIndex=e}setItems(e){this._items=e}handleKey(e){let A=e.keyCode;e.key&&e.key.length===1?this._letterKeyStream.next(e.key.toLocaleUpperCase()):(A>=65&&A<=90||A>=48&&A<=57)&&this._letterKeyStream.next(String.fromCharCode(A))}isTyping(){return this._pressedLetters.length>0}reset(){this._pressedLetters=[]}_setupKeyHandler(e){this._letterKeyStream.pipe(Qo(A=>this._pressedLetters.push(A)),el(e),kt(()=>this._pressedLetters.length>0),je(()=>this._pressedLetters.join("").toLocaleUpperCase())).subscribe(A=>{for(let i=1;i<this._items.length+1;i++){let n=(this._selectedItemIndex+i)%this._items.length,o=this._items[n];if(!this._skipPredicateFn?.(o)&&o.getLabel?.().toLocaleUpperCase().trim().indexOf(A)===0){this._selectedItem.next(o);break}}this._pressedLetters=[]})}},e8=class{_items;_activeItemIndex=-1;_activeItem=Jo(null);_wrap=!1;_typeaheadSubscription=Kt.EMPTY;_itemChangesSubscription;_vertical=!0;_horizontal;_allowedModifierKeys=[];_homeAndEnd=!1;_pageUpAndDown={enabled:!1,delta:10};_effectRef;_typeahead;_skipPredicateFn=e=>e.disabled;constructor(e,A){this._items=e,e instanceof ca?this._itemChangesSubscription=e.changes.subscribe(i=>this._itemsChanged(i.toArray())):p2(e)&&(this._effectRef=Nh(()=>this._itemsChanged(e()),{injector:A}))}tabOut=new OA;change=new OA;skipPredicate(e){return this._skipPredicateFn=e,this}withWrap(e=!0){return this._wrap=e,this}withVerticalOrientation(e=!0){return this._vertical=e,this}withHorizontalOrientation(e){return this._horizontal=e,this}withAllowedModifierKeys(e){return this._allowedModifierKeys=e,this}withTypeAhead(e=200){this._typeaheadSubscription.unsubscribe();let A=this._getItemsArray();return this._typeahead=new Ck(A,{debounceInterval:typeof e=="number"?e:void 0,skipPredicate:i=>this._skipPredicateFn(i)}),this._typeaheadSubscription=this._typeahead.selectedItem.subscribe(i=>{this.setActiveItem(i)}),this}cancelTypeahead(){return this._typeahead?.reset(),this}withHomeAndEnd(e=!0){return this._homeAndEnd=e,this}withPageUpDown(e=!0,A=10){return this._pageUpAndDown={enabled:e,delta:A},this}setActiveItem(e){let A=this._activeItem();this.updateActiveItem(e),this._activeItem()!==A&&this.change.next(this._activeItemIndex)}onKeydown(e){let A=e.keyCode,n=["altKey","ctrlKey","metaKey","shiftKey"].every(o=>!e[o]||this._allowedModifierKeys.indexOf(o)>-1);switch(A){case 9:this.tabOut.next();return;case 40:if(this._vertical&&n){this.setNextItemActive();break}else return;case 38:if(this._vertical&&n){this.setPreviousItemActive();break}else return;case 39:if(this._horizontal&&n){this._horizontal==="rtl"?this.setPreviousItemActive():this.setNextItemActive();break}else return;case 37:if(this._horizontal&&n){this._horizontal==="rtl"?this.setNextItemActive():this.setPreviousItemActive();break}else return;case 36:if(this._homeAndEnd&&n){this.setFirstItemActive();break}else return;case 35:if(this._homeAndEnd&&n){this.setLastItemActive();break}else return;case 33:if(this._pageUpAndDown.enabled&&n){let o=this._activeItemIndex-this._pageUpAndDown.delta;this._setActiveItemByIndex(o>0?o:0,1);break}else return;case 34:if(this._pageUpAndDown.enabled&&n){let o=this._activeItemIndex+this._pageUpAndDown.delta,r=this._getItemsArray().length;this._setActiveItemByIndex(o<r?o:r-1,-1);break}else return;default:(n||ir(e,"shiftKey"))&&this._typeahead?.handleKey(e);return}this._typeahead?.reset(),e.preventDefault()}get activeItemIndex(){return this._activeItemIndex}get activeItem(){return this._activeItem()}isTyping(){return!!this._typeahead&&this._typeahead.isTyping()}setFirstItemActive(){this._setActiveItemByIndex(0,1)}setLastItemActive(){this._setActiveItemByIndex(this._getItemsArray().length-1,-1)}setNextItemActive(){this._activeItemIndex<0?this.setFirstItemActive():this._setActiveItemByDelta(1)}setPreviousItemActive(){this._activeItemIndex<0&&this._wrap?this.setLastItemActive():this._setActiveItemByDelta(-1)}updateActiveItem(e){let A=this._getItemsArray(),i=typeof e=="number"?e:A.indexOf(e),n=A[i];this._activeItem.set(n??null),this._activeItemIndex=i,this._typeahead?.setCurrentSelectedItemIndex(i)}destroy(){this._typeaheadSubscription.unsubscribe(),this._itemChangesSubscription?.unsubscribe(),this._effectRef?.destroy(),this._typeahead?.destroy(),this.tabOut.complete(),this.change.complete()}_setActiveItemByDelta(e){this._wrap?this._setActiveInWrapMode(e):this._setActiveInDefaultMode(e)}_setActiveInWrapMode(e){let A=this._getItemsArray();for(let i=1;i<=A.length;i++){let n=(this._activeItemIndex+e*i+A.length)%A.length,o=A[n];if(!this._skipPredicateFn(o)){this.setActiveItem(n);return}}}_setActiveInDefaultMode(e){this._setActiveItemByIndex(this._activeItemIndex+e,e)}_setActiveItemByIndex(e,A){let i=this._getItemsArray();if(i[e]){for(;this._skipPredicateFn(i[e]);)if(e+=A,!i[e])return;this.setActiveItem(e)}}_getItemsArray(){return p2(this._items)?this._items():this._items instanceof ca?this._items.toArray():this._items}_itemsChanged(e){this._typeahead?.setItems(e);let A=this._activeItem();if(A){let i=e.indexOf(A);i>-1&&i!==this._activeItemIndex&&(this._activeItemIndex=i,this._typeahead?.setCurrentSelectedItemIndex(i))}}},t8=class extends e8{setActiveItem(e){this.activeItem&&this.activeItem.setInactiveStyles(),super.setActiveItem(e),this.activeItem&&this.activeItem.setActiveStyles()}},qI=class extends e8{_origin="program";setFocusOrigin(e){return this._origin=e,this}setActiveItem(e){super.setActiveItem(e),this.activeItem&&this.activeItem.focus(this._origin)}};var Mu=(()=>{class t{_platform=f(Ii);constructor(){}isDisabled(A){return A.hasAttribute("disabled")}isVisible(A){return euA(A)&&getComputedStyle(A).visibility==="visible"}isTabbable(A){if(!this._platform.isBrowser)return!1;let i=AuA(cuA(A));if(i&&(UO(i)===-1||!this.isVisible(i)))return!1;let n=A.nodeName.toLowerCase(),o=UO(A);return A.hasAttribute("contenteditable")?o!==-1:n==="iframe"||n==="object"||this._platform.WEBKIT&&this._platform.IOS&&!suA(A)?!1:n==="audio"?A.hasAttribute("controls")?o!==-1:!1:n==="video"?o===-1?!1:o!==null?!0:this._platform.FIREFOX||A.hasAttribute("controls"):A.tabIndex>=0}isFocusable(A,i){return auA(A)&&!this.isDisabled(A)&&(i?.ignoreVisibility||this.isVisible(A))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function AuA(t){try{return t.frameElement}catch{return null}}function euA(t){return!!(t.offsetWidth||t.offsetHeight||typeof t.getClientRects=="function"&&t.getClientRects().length)}function tuA(t){let e=t.nodeName.toLowerCase();return e==="input"||e==="select"||e==="button"||e==="textarea"}function iuA(t){return ouA(t)&&t.type=="hidden"}function nuA(t){return ruA(t)&&t.hasAttribute("href")}function ouA(t){return t.nodeName.toLowerCase()=="input"}function ruA(t){return t.nodeName.toLowerCase()=="a"}function zO(t){if(!t.hasAttribute("tabindex")||t.tabIndex===void 0)return!1;let e=t.getAttribute("tabindex");return!!(e&&!isNaN(parseInt(e,10)))}function UO(t){if(!zO(t))return null;let e=parseInt(t.getAttribute("tabindex")||"",10);return isNaN(e)?-1:e}function suA(t){let e=t.nodeName.toLowerCase(),A=e==="input"&&t.type;return A==="text"||A==="password"||e==="select"||e==="textarea"}function auA(t){return iuA(t)?!1:tuA(t)||nuA(t)||t.hasAttribute("contenteditable")||zO(t)}function cuA(t){return t.ownerDocument&&t.ownerDocument.defaultView||window}var dk=class{_element;_checker;_ngZone;_document;_injector;_startAnchor;_endAnchor;_hasAttached=!1;startAnchorListener=()=>this.focusLastTabbableElement();endAnchorListener=()=>this.focusFirstTabbableElement();get enabled(){return this._enabled}set enabled(e){this._enabled=e,this._startAnchor&&this._endAnchor&&(this._toggleAnchorTabIndex(e,this._startAnchor),this._toggleAnchorTabIndex(e,this._endAnchor))}_enabled=!0;constructor(e,A,i,n,o=!1,r){this._element=e,this._checker=A,this._ngZone=i,this._document=n,this._injector=r,o||this.attachAnchors()}destroy(){let e=this._startAnchor,A=this._endAnchor;e&&(e.removeEventListener("focus",this.startAnchorListener),e.remove()),A&&(A.removeEventListener("focus",this.endAnchorListener),A.remove()),this._startAnchor=this._endAnchor=null,this._hasAttached=!1}attachAnchors(){return this._hasAttached?!0:(this._ngZone.runOutsideAngular(()=>{this._startAnchor||(this._startAnchor=this._createAnchor(),this._startAnchor.addEventListener("focus",this.startAnchorListener)),this._endAnchor||(this._endAnchor=this._createAnchor(),this._endAnchor.addEventListener("focus",this.endAnchorListener))}),this._element.parentNode&&(this._element.parentNode.insertBefore(this._startAnchor,this._element),this._element.parentNode.insertBefore(this._endAnchor,this._element.nextSibling),this._hasAttached=!0),this._hasAttached)}focusInitialElementWhenReady(e){return new Promise(A=>{this._executeOnStable(()=>A(this.focusInitialElement(e)))})}focusFirstTabbableElementWhenReady(e){return new Promise(A=>{this._executeOnStable(()=>A(this.focusFirstTabbableElement(e)))})}focusLastTabbableElementWhenReady(e){return new Promise(A=>{this._executeOnStable(()=>A(this.focusLastTabbableElement(e)))})}_getRegionBoundary(e){let A=this._element.querySelectorAll(`[cdk-focus-region-${e}], [cdkFocusRegion${e}], [cdk-focus-${e}]`);return e=="start"?A.length?A[0]:this._getFirstTabbableElement(this._element):A.length?A[A.length-1]:this._getLastTabbableElement(this._element)}focusInitialElement(e){let A=this._element.querySelector("[cdk-focus-initial], [cdkFocusInitial]");if(A){if(!this._checker.isFocusable(A)){let i=this._getFirstTabbableElement(A);return i?.focus(e),!!i}return A.focus(e),!0}return this.focusFirstTabbableElement(e)}focusFirstTabbableElement(e){let A=this._getRegionBoundary("start");return A&&A.focus(e),!!A}focusLastTabbableElement(e){let A=this._getRegionBoundary("end");return A&&A.focus(e),!!A}hasAttached(){return this._hasAttached}_getFirstTabbableElement(e){if(this._checker.isFocusable(e)&&this._checker.isTabbable(e))return e;let A=e.children;for(let i=0;i<A.length;i++){let n=A[i].nodeType===this._document.ELEMENT_NODE?this._getFirstTabbableElement(A[i]):null;if(n)return n}return null}_getLastTabbableElement(e){if(this._checker.isFocusable(e)&&this._checker.isTabbable(e))return e;let A=e.children;for(let i=A.length-1;i>=0;i--){let n=A[i].nodeType===this._document.ELEMENT_NODE?this._getLastTabbableElement(A[i]):null;if(n)return n}return null}_createAnchor(){let e=this._document.createElement("div");return this._toggleAnchorTabIndex(this._enabled,e),e.classList.add("cdk-visually-hidden"),e.classList.add("cdk-focus-trap-anchor"),e.setAttribute("aria-hidden","true"),e}_toggleAnchorTabIndex(e,A){e?A.setAttribute("tabindex","0"):A.removeAttribute("tabindex")}toggleAnchors(e){this._startAnchor&&this._endAnchor&&(this._toggleAnchorTabIndex(e,this._startAnchor),this._toggleAnchorTabIndex(e,this._endAnchor))}_executeOnStable(e){this._injector?To(e,{injector:this._injector}):setTimeout(e)}},n8=(()=>{class t{_checker=f(Mu);_ngZone=f(Qe);_document=f(at);_injector=f(Rt);constructor(){f(Rn).load(bu)}create(A,i=!1){return new dk(A,this._checker,this._ngZone,this._document,i,this._injector)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function ku(t){return t.buttons===0||t.detail===0}function Su(t){let e=t.touches&&t.touches[0]||t.changedTouches&&t.changedTouches[0];return!!e&&e.identifier===-1&&(e.radiusX==null||e.radiusX===1)&&(e.radiusY==null||e.radiusY===1)}var luA=new dA("cdk-input-modality-detector-options"),guA={ignoreKeys:[18,17,224,91,16]},OO=650,yB=bc({passive:!0,capture:!0}),IuA=(()=>{class t{_platform=f(Ii);modalityDetected;modalityChanged;get mostRecentModality(){return this._modality.value}_mostRecentTarget=null;_modality=new Mi(null);_options;_lastTouchMs=0;_onKeydown=A=>{this._options?.ignoreKeys?.some(i=>i===A.keyCode)||(this._modality.next("keyboard"),this._mostRecentTarget=oc(A))};_onMousedown=A=>{Date.now()-this._lastTouchMs<OO||(this._modality.next(ku(A)?"keyboard":"mouse"),this._mostRecentTarget=oc(A))};_onTouchstart=A=>{if(Su(A)){this._modality.next("keyboard");return}this._lastTouchMs=Date.now(),this._modality.next("touch"),this._mostRecentTarget=oc(A)};constructor(){let A=f(Qe),i=f(at),n=f(luA,{optional:!0});this._options=rA(rA({},guA),n),this.modalityDetected=this._modality.pipe(CI(1)),this.modalityChanged=this.modalityDetected.pipe(tl()),this._platform.isBrowser&&A.runOutsideAngular(()=>{i.addEventListener("keydown",this._onKeydown,yB),i.addEventListener("mousedown",this._onMousedown,yB),i.addEventListener("touchstart",this._onTouchstart,yB)})}ngOnDestroy(){this._modality.complete(),this._platform.isBrowser&&(document.removeEventListener("keydown",this._onKeydown,yB),document.removeEventListener("mousedown",this._onMousedown,yB),document.removeEventListener("touchstart",this._onTouchstart,yB))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),CuA=new dA("liveAnnouncerElement",{providedIn:"root",factory:duA});function duA(){return null}var BuA=new dA("LIVE_ANNOUNCER_DEFAULT_OPTIONS"),EuA=0,o8=(()=>{class t{_ngZone=f(Qe);_defaultOptions=f(BuA,{optional:!0});_liveElement;_document=f(at);_previousTimeout;_currentPromise;_currentResolve;constructor(){let A=f(CuA,{optional:!0});this._liveElement=A||this._createLiveElement()}announce(A,...i){let n=this._defaultOptions,o,r;return i.length===1&&typeof i[0]=="number"?r=i[0]:[o,r]=i,this.clear(),clearTimeout(this._previousTimeout),o||(o=n&&n.politeness?n.politeness:"polite"),r==null&&n&&(r=n.duration),this._liveElement.setAttribute("aria-live",o),this._liveElement.id&&this._exposeAnnouncerToModals(this._liveElement.id),this._ngZone.runOutsideAngular(()=>(this._currentPromise||(this._currentPromise=new Promise(s=>this._currentResolve=s)),clearTimeout(this._previousTimeout),this._previousTimeout=setTimeout(()=>{this._liveElement.textContent=A,typeof r=="number"&&(this._previousTimeout=setTimeout(()=>this.clear(),r)),this._currentResolve?.(),this._currentPromise=this._currentResolve=void 0},100),this._currentPromise))}clear(){this._liveElement&&(this._liveElement.textContent="")}ngOnDestroy(){clearTimeout(this._previousTimeout),this._liveElement?.remove(),this._liveElement=null,this._currentResolve?.(),this._currentPromise=this._currentResolve=void 0}_createLiveElement(){let A="cdk-live-announcer-element",i=this._document.getElementsByClassName(A),n=this._document.createElement("div");for(let o=0;o<i.length;o++)i[o].remove();return n.classList.add(A),n.classList.add("cdk-visually-hidden"),n.setAttribute("aria-atomic","true"),n.setAttribute("aria-live","polite"),n.id=`cdk-live-announcer-${EuA++}`,this._document.body.appendChild(n),n}_exposeAnnouncerToModals(A){let i=this._document.querySelectorAll('body > .cdk-overlay-container [aria-modal="true"]');for(let n=0;n<i.length;n++){let o=i[n],r=o.getAttribute("aria-owns");r?r.indexOf(A)===-1&&o.setAttribute("aria-owns",r+" "+A):o.setAttribute("aria-owns",A)}}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var $6=function(t){return t[t.IMMEDIATE=0]="IMMEDIATE",t[t.EVENTUAL=1]="EVENTUAL",t}($6||{}),QuA=new dA("cdk-focus-monitor-default-options"),X6=bc({passive:!0,capture:!0}),dr=(()=>{class t{_ngZone=f(Qe);_platform=f(Ii);_inputModalityDetector=f(IuA);_origin=null;_lastFocusOrigin;_windowFocused=!1;_windowFocusTimeoutId;_originTimeoutId;_originFromTouchInteraction=!1;_elementInfo=new Map;_monitoredElementCount=0;_rootNodeFocusListenerCount=new Map;_detectionMode;_windowFocusListener=()=>{this._windowFocused=!0,this._windowFocusTimeoutId=setTimeout(()=>this._windowFocused=!1)};_document=f(at,{optional:!0});_stopInputModalityDetector=new OA;constructor(){let A=f(QuA,{optional:!0});this._detectionMode=A?.detectionMode||$6.IMMEDIATE}_rootNodeFocusAndBlurListener=A=>{let i=oc(A);for(let n=i;n;n=n.parentElement)A.type==="focus"?this._onFocus(A,n):this._onBlur(A,n)};monitor(A,i=!1){let n=ua(A);if(!this._platform.isBrowser||n.nodeType!==1)return Me();let o=RO(n)||this._getDocument(),r=this._elementInfo.get(n);if(r)return i&&(r.checkChildren=!0),r.subject;let s={checkChildren:i,subject:new OA,rootNode:o};return this._elementInfo.set(n,s),this._registerGlobalListeners(s),s.subject}stopMonitoring(A){let i=ua(A),n=this._elementInfo.get(i);n&&(n.subject.complete(),this._setClasses(i),this._elementInfo.delete(i),this._removeGlobalListeners(n))}focusVia(A,i,n){let o=ua(A),r=this._getDocument().activeElement;o===r?this._getClosestElementsInfo(o).forEach(([s,a])=>this._originChanged(s,i,a)):(this._setOrigin(i),typeof o.focus=="function"&&o.focus(n))}ngOnDestroy(){this._elementInfo.forEach((A,i)=>this.stopMonitoring(i))}_getDocument(){return this._document||document}_getWindow(){return this._getDocument().defaultView||window}_getFocusOrigin(A){return this._origin?this._originFromTouchInteraction?this._shouldBeAttributedToTouch(A)?"touch":"program":this._origin:this._windowFocused&&this._lastFocusOrigin?this._lastFocusOrigin:A&&this._isLastInteractionFromInputLabel(A)?"mouse":"program"}_shouldBeAttributedToTouch(A){return this._detectionMode===$6.EVENTUAL||!!A?.contains(this._inputModalityDetector._mostRecentTarget)}_setClasses(A,i){A.classList.toggle("cdk-focused",!!i),A.classList.toggle("cdk-touch-focused",i==="touch"),A.classList.toggle("cdk-keyboard-focused",i==="keyboard"),A.classList.toggle("cdk-mouse-focused",i==="mouse"),A.classList.toggle("cdk-program-focused",i==="program")}_setOrigin(A,i=!1){this._ngZone.runOutsideAngular(()=>{if(this._origin=A,this._originFromTouchInteraction=A==="touch"&&i,this._detectionMode===$6.IMMEDIATE){clearTimeout(this._originTimeoutId);let n=this._originFromTouchInteraction?OO:1;this._originTimeoutId=setTimeout(()=>this._origin=null,n)}})}_onFocus(A,i){let n=this._elementInfo.get(i),o=oc(A);!n||!n.checkChildren&&i!==o||this._originChanged(i,this._getFocusOrigin(o),n)}_onBlur(A,i){let n=this._elementInfo.get(i);!n||n.checkChildren&&A.relatedTarget instanceof Node&&i.contains(A.relatedTarget)||(this._setClasses(i),this._emitOrigin(n,null))}_emitOrigin(A,i){A.subject.observers.length&&this._ngZone.run(()=>A.subject.next(i))}_registerGlobalListeners(A){if(!this._platform.isBrowser)return;let i=A.rootNode,n=this._rootNodeFocusListenerCount.get(i)||0;n||this._ngZone.runOutsideAngular(()=>{i.addEventListener("focus",this._rootNodeFocusAndBlurListener,X6),i.addEventListener("blur",this._rootNodeFocusAndBlurListener,X6)}),this._rootNodeFocusListenerCount.set(i,n+1),++this._monitoredElementCount===1&&(this._ngZone.runOutsideAngular(()=>{this._getWindow().addEventListener("focus",this._windowFocusListener)}),this._inputModalityDetector.modalityDetected.pipe(St(this._stopInputModalityDetector)).subscribe(o=>{this._setOrigin(o,!0)}))}_removeGlobalListeners(A){let i=A.rootNode;if(this._rootNodeFocusListenerCount.has(i)){let n=this._rootNodeFocusListenerCount.get(i);n>1?this._rootNodeFocusListenerCount.set(i,n-1):(i.removeEventListener("focus",this._rootNodeFocusAndBlurListener,X6),i.removeEventListener("blur",this._rootNodeFocusAndBlurListener,X6),this._rootNodeFocusListenerCount.delete(i))}--this._monitoredElementCount||(this._getWindow().removeEventListener("focus",this._windowFocusListener),this._stopInputModalityDetector.next(),clearTimeout(this._windowFocusTimeoutId),clearTimeout(this._originTimeoutId))}_originChanged(A,i,n){this._setClasses(A,i),this._emitOrigin(n,i),this._lastFocusOrigin=i}_getClosestElementsInfo(A){let i=[];return this._elementInfo.forEach((n,o)=>{(o===A||n.checkChildren&&o.contains(A))&&i.push([o,n])}),i}_isLastInteractionFromInputLabel(A){let{_mostRecentTarget:i,mostRecentModality:n}=this._inputModalityDetector;if(n!=="mouse"||!i||i===A||A.nodeName!=="INPUT"&&A.nodeName!=="TEXTAREA"||A.disabled)return!1;let o=A.labels;if(o){for(let r=0;r<o.length;r++)if(o[r].contains(i))return!0}return!1}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),PO=(()=>{class t{_elementRef=f(ee);_focusMonitor=f(dr);_monitorSubscription;_focusOrigin=null;cdkFocusChange=new WA;constructor(){}get focusOrigin(){return this._focusOrigin}ngAfterViewInit(){let A=this._elementRef.nativeElement;this._monitorSubscription=this._focusMonitor.monitor(A,A.nodeType===1&&A.hasAttribute("cdkMonitorSubtreeFocus")).subscribe(i=>{this._focusOrigin=i,this.cdkFocusChange.emit(i)})}ngOnDestroy(){this._focusMonitor.stopMonitoring(this._elementRef),this._monitorSubscription&&this._monitorSubscription.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkMonitorElementFocus",""],["","cdkMonitorSubtreeFocus",""]],outputs:{cdkFocusChange:"cdkFocusChange"},exportAs:["cdkMonitorFocus"]})}return t})(),jI=function(t){return t[t.NONE=0]="NONE",t[t.BLACK_ON_WHITE=1]="BLACK_ON_WHITE",t[t.WHITE_ON_BLACK=2]="WHITE_ON_BLACK",t}(jI||{}),KO="cdk-high-contrast-black-on-white",YO="cdk-high-contrast-white-on-black",lk="cdk-high-contrast-active",Qk=(()=>{class t{_platform=f(Ii);_hasCheckedHighContrastMode;_document=f(at);_breakpointSubscription;constructor(){this._breakpointSubscription=f(Z6).observe("(forced-colors: active)").subscribe(()=>{this._hasCheckedHighContrastMode&&(this._hasCheckedHighContrastMode=!1,this._applyBodyHighContrastModeCssClasses())})}getHighContrastMode(){if(!this._platform.isBrowser)return jI.NONE;let A=this._document.createElement("div");A.style.backgroundColor="rgb(1,2,3)",A.style.position="absolute",this._document.body.appendChild(A);let i=this._document.defaultView||window,n=i&&i.getComputedStyle?i.getComputedStyle(A):null,o=(n&&n.backgroundColor||"").replace(/ /g,"");switch(A.remove(),o){case"rgb(0,0,0)":case"rgb(45,50,54)":case"rgb(32,32,32)":return jI.WHITE_ON_BLACK;case"rgb(255,255,255)":case"rgb(255,250,239)":return jI.BLACK_ON_WHITE}return jI.NONE}ngOnDestroy(){this._breakpointSubscription.unsubscribe()}_applyBodyHighContrastModeCssClasses(){if(!this._hasCheckedHighContrastMode&&this._platform.isBrowser&&this._document.body){let A=this._document.body.classList;A.remove(lk,KO,YO),this._hasCheckedHighContrastMode=!0;let i=this.getHighContrastMode();i===jI.BLACK_ON_WHITE?A.add(lk,KO):i===jI.WHITE_ON_BLACK&&A.add(lk,YO)}}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),r8=(()=>{class t{constructor(){f(Qk)._applyBodyHighContrastModeCssClasses()}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[DB]})}return t})(),gk={},sn=(()=>{class t{_appId=f(Vd);getId(A){return this._appId!=="ng"&&(A+=this._appId),gk.hasOwnProperty(A)||(gk[A]=0),`${A}${gk[A]++}`}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var huA=new dA("cdk-dir-doc",{providedIn:"root",factory:uuA});function uuA(){return f(at)}var fuA=/^(ar|ckb|dv|he|iw|fa|nqo|ps|sd|ug|ur|yi|.*[-_](Adlm|Arab|Hebr|Nkoo|Rohg|Thaa))(?!.*[-_](Latn|Cyrl)($|-|_))($|-|_)/i;function muA(t){let e=t?.toLowerCase()||"";return e==="auto"&&typeof navigator<"u"&&navigator?.language?fuA.test(navigator.language)?"rtl":"ltr":e==="rtl"?"rtl":"ltr"}var mo=(()=>{class t{value="ltr";change=new WA;constructor(){let A=f(huA,{optional:!0});if(A){let i=A.body?A.body.dir:null,n=A.documentElement?A.documentElement.dir:null;this.value=muA(i||n||"ltr")}}ngOnDestroy(){this.change.complete()}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var _2=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();var puA=["text"],wuA=[[["mat-icon"]],"*"],DuA=["mat-icon","*"];function yuA(t,e){if(t&1&&JA(0,"mat-pseudo-checkbox",1),t&2){let A=O();yA("disabled",A.disabled)("state",A.selected?"checked":"unchecked")}}function vuA(t,e){if(t&1&&JA(0,"mat-pseudo-checkbox",3),t&2){let A=O();yA("disabled",A.disabled)}}function buA(t,e){if(t&1&&(S(0,"span",4),AA(1),F()),t&2){let A=O();G(),Et("(",A.group.label,")")}}var MuA=["mat-internal-form-field",""],kuA=["*"];var it=(()=>{class t{constructor(){f(Qk)._applyBodyHighContrastModeCssClasses()}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[_2,_2]})}return t})(),$I=class{_defaultMatcher;ngControl;_parentFormGroup;_parentForm;_stateChanges;errorState=!1;matcher;constructor(e,A,i,n,o){this._defaultMatcher=e,this.ngControl=A,this._parentFormGroup=i,this._parentForm=n,this._stateChanges=o}updateErrorState(){let e=this.errorState,A=this._parentFormGroup||this._parentForm,i=this.matcher||this._defaultMatcher,n=this.ngControl?this.ngControl.control:null,o=i?.isErrorState(n,A)??!1;o!==e&&(this.errorState=o,this._stateChanges.next())}};var bB=(()=>{class t{isErrorState(A,i){return!!(A&&A.invalid&&(A.touched||i&&i.submitted))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),lr=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["structural-styles"]],decls:0,vars:0,template:function(i,n){},styles:['.mat-focus-indicator{position:relative}.mat-focus-indicator::before{top:0;left:0;right:0;bottom:0;position:absolute;box-sizing:border-box;pointer-events:none;display:var(--mat-focus-indicator-display, none);border-width:var(--mat-focus-indicator-border-width, 3px);border-style:var(--mat-focus-indicator-border-style, solid);border-color:var(--mat-focus-indicator-border-color, transparent);border-radius:var(--mat-focus-indicator-border-radius, 4px)}.mat-focus-indicator:focus::before{content:""}@media(forced-colors: active){html{--mat-focus-indicator-display: block}}'],encapsulation:2,changeDetection:0})}return t})();var Os=function(t){return t[t.FADING_IN=0]="FADING_IN",t[t.VISIBLE=1]="VISIBLE",t[t.FADING_OUT=2]="FADING_OUT",t[t.HIDDEN=3]="HIDDEN",t}(Os||{}),fk=class{_renderer;element;config;_animationForciblyDisabledThroughCss;state=Os.HIDDEN;constructor(e,A,i,n=!1){this._renderer=e,this.element=A,this.config=i,this._animationForciblyDisabledThroughCss=n}fadeOut(){this._renderer.fadeOutRipple(this)}},jO=bc({passive:!0,capture:!0}),mk=class{_events=new Map;addHandler(e,A,i,n){let o=this._events.get(A);if(o){let r=o.get(i);r?r.add(n):o.set(i,new Set([n]))}else this._events.set(A,new Map([[i,new Set([n])]])),e.runOutsideAngular(()=>{document.addEventListener(A,this._delegateEventHandler,jO)})}removeHandler(e,A,i){let n=this._events.get(e);if(!n)return;let o=n.get(A);o&&(o.delete(i),o.size===0&&n.delete(A),n.size===0&&(this._events.delete(e),document.removeEventListener(e,this._delegateEventHandler,jO)))}_delegateEventHandler=e=>{let A=oc(e);A&&this._events.get(e.type)?.forEach((i,n)=>{(n===A||n.contains(A))&&i.forEach(o=>o.handleEvent(e))})}},a8={enterDuration:225,exitDuration:150},SuA=800,qO=bc({passive:!0,capture:!0}),VO=["mousedown","touchstart"],ZO=["mouseup","mouseleave","touchend","touchcancel"],RuA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["ng-component"]],hostAttrs:["mat-ripple-style-loader",""],decls:0,vars:0,template:function(i,n){},styles:[".mat-ripple{overflow:hidden;position:relative}.mat-ripple:not(:empty){transform:translateZ(0)}.mat-ripple.mat-ripple-unbounded{overflow:visible}.mat-ripple-element{position:absolute;border-radius:50%;pointer-events:none;transition:opacity,transform 0ms cubic-bezier(0, 0, 0.2, 1);transform:scale3d(0, 0, 0);background-color:var(--mat-ripple-color, color-mix(in srgb, var(--mat-sys-on-surface) 10%, transparent))}@media(forced-colors: active){.mat-ripple-element{display:none}}.cdk-drag-preview .mat-ripple-element,.cdk-drag-placeholder .mat-ripple-element{display:none}"],encapsulation:2,changeDetection:0})}return t})(),vB=class t{_target;_ngZone;_platform;_containerElement;_triggerElement;_isPointerDown=!1;_activeRipples=new Map;_mostRecentTransientRipple;_lastTouchStartEvent;_pointerUpEventsRegistered=!1;_containerRect;static _eventManager=new mk;constructor(e,A,i,n,o){this._target=e,this._ngZone=A,this._platform=n,n.isBrowser&&(this._containerElement=ua(i)),o&&o.get(Rn).load(RuA)}fadeInRipple(e,A,i={}){let n=this._containerRect=this._containerRect||this._containerElement.getBoundingClientRect(),o=rA(rA({},a8),i.animation);i.centered&&(e=n.left+n.width/2,A=n.top+n.height/2);let r=i.radius||xuA(e,A,n),s=e-n.left,a=A-n.top,c=o.enterDuration,l=document.createElement("div");l.classList.add("mat-ripple-element"),l.style.left=`${s-r}px`,l.style.top=`${a-r}px`,l.style.height=`${r*2}px`,l.style.width=`${r*2}px`,i.color!=null&&(l.style.backgroundColor=i.color),l.style.transitionDuration=`${c}ms`,this._containerElement.appendChild(l);let I=window.getComputedStyle(l),C=I.transitionProperty,d=I.transitionDuration,B=C==="none"||d==="0s"||d==="0s, 0s"||n.width===0&&n.height===0,E=new fk(this,l,i,B);l.style.transform="scale3d(1, 1, 1)",E.state=Os.FADING_IN,i.persistent||(this._mostRecentTransientRipple=E);let h=null;return!B&&(c||o.exitDuration)&&this._ngZone.runOutsideAngular(()=>{let u=()=>{h&&(h.fallbackTimer=null),clearTimeout(L),this._finishRippleTransition(E)},D=()=>this._destroyRipple(E),L=setTimeout(D,c+100);l.addEventListener("transitionend",u),l.addEventListener("transitioncancel",D),h={onTransitionEnd:u,onTransitionCancel:D,fallbackTimer:L}}),this._activeRipples.set(E,h),(B||!c)&&this._finishRippleTransition(E),E}fadeOutRipple(e){if(e.state===Os.FADING_OUT||e.state===Os.HIDDEN)return;let A=e.element,i=rA(rA({},a8),e.config.animation);A.style.transitionDuration=`${i.exitDuration}ms`,A.style.opacity="0",e.state=Os.FADING_OUT,(e._animationForciblyDisabledThroughCss||!i.exitDuration)&&this._finishRippleTransition(e)}fadeOutAll(){this._getActiveRipples().forEach(e=>e.fadeOut())}fadeOutAllNonPersistent(){this._getActiveRipples().forEach(e=>{e.config.persistent||e.fadeOut()})}setupTriggerEvents(e){let A=ua(e);!this._platform.isBrowser||!A||A===this._triggerElement||(this._removeTriggerEvents(),this._triggerElement=A,VO.forEach(i=>{t._eventManager.addHandler(this._ngZone,i,A,this)}))}handleEvent(e){e.type==="mousedown"?this._onMousedown(e):e.type==="touchstart"?this._onTouchStart(e):this._onPointerUp(),this._pointerUpEventsRegistered||(this._ngZone.runOutsideAngular(()=>{ZO.forEach(A=>{this._triggerElement.addEventListener(A,this,qO)})}),this._pointerUpEventsRegistered=!0)}_finishRippleTransition(e){e.state===Os.FADING_IN?this._startFadeOutTransition(e):e.state===Os.FADING_OUT&&this._destroyRipple(e)}_startFadeOutTransition(e){let A=e===this._mostRecentTransientRipple,{persistent:i}=e.config;e.state=Os.VISIBLE,!i&&(!A||!this._isPointerDown)&&e.fadeOut()}_destroyRipple(e){let A=this._activeRipples.get(e)??null;this._activeRipples.delete(e),this._activeRipples.size||(this._containerRect=null),e===this._mostRecentTransientRipple&&(this._mostRecentTransientRipple=null),e.state=Os.HIDDEN,A!==null&&(e.element.removeEventListener("transitionend",A.onTransitionEnd),e.element.removeEventListener("transitioncancel",A.onTransitionCancel),A.fallbackTimer!==null&&clearTimeout(A.fallbackTimer)),e.element.remove()}_onMousedown(e){let A=ku(e),i=this._lastTouchStartEvent&&Date.now()<this._lastTouchStartEvent+SuA;!this._target.rippleDisabled&&!A&&!i&&(this._isPointerDown=!0,this.fadeInRipple(e.clientX,e.clientY,this._target.rippleConfig))}_onTouchStart(e){if(!this._target.rippleDisabled&&!Su(e)){this._lastTouchStartEvent=Date.now(),this._isPointerDown=!0;let A=e.changedTouches;if(A)for(let i=0;i<A.length;i++)this.fadeInRipple(A[i].clientX,A[i].clientY,this._target.rippleConfig)}}_onPointerUp(){this._isPointerDown&&(this._isPointerDown=!1,this._getActiveRipples().forEach(e=>{let A=e.state===Os.VISIBLE||e.config.terminateOnPointerUp&&e.state===Os.FADING_IN;!e.config.persistent&&A&&e.fadeOut()}))}_getActiveRipples(){return Array.from(this._activeRipples.keys())}_removeTriggerEvents(){let e=this._triggerElement;e&&(VO.forEach(A=>t._eventManager.removeHandler(A,e,this)),this._pointerUpEventsRegistered&&(ZO.forEach(A=>e.removeEventListener(A,this,qO)),this._pointerUpEventsRegistered=!1))}};function xuA(t,e,A){let i=Math.max(Math.abs(t-A.left),Math.abs(t-A.right)),n=Math.max(Math.abs(e-A.top),Math.abs(e-A.bottom));return Math.sqrt(i*i+n*n)}var G2=new dA("mat-ripple-global-options"),rs=(()=>{class t{_elementRef=f(ee);_animationMode=f(Si,{optional:!0});color;unbounded;centered;radius=0;animation;get disabled(){return this._disabled}set disabled(A){A&&this.fadeOutAllNonPersistent(),this._disabled=A,this._setupTriggerEventsIfEnabled()}_disabled=!1;get trigger(){return this._trigger||this._elementRef.nativeElement}set trigger(A){this._trigger=A,this._setupTriggerEventsIfEnabled()}_trigger;_rippleRenderer;_globalOptions;_isInitialized=!1;constructor(){let A=f(Qe),i=f(Ii),n=f(G2,{optional:!0}),o=f(Rt);this._globalOptions=n||{},this._rippleRenderer=new vB(this,A,this._elementRef,i,o)}ngOnInit(){this._isInitialized=!0,this._setupTriggerEventsIfEnabled()}ngOnDestroy(){this._rippleRenderer._removeTriggerEvents()}fadeOutAll(){this._rippleRenderer.fadeOutAll()}fadeOutAllNonPersistent(){this._rippleRenderer.fadeOutAllNonPersistent()}get rippleConfig(){return{centered:this.centered,radius:this.radius,color:this.color,animation:rA(rA(rA({},this._globalOptions.animation),this._animationMode==="NoopAnimations"?{enterDuration:0,exitDuration:0}:{}),this.animation),terminateOnPointerUp:this._globalOptions.terminateOnPointerUp}}get rippleDisabled(){return this.disabled||!!this._globalOptions.disabled}_setupTriggerEventsIfEnabled(){!this.disabled&&this._isInitialized&&this._rippleRenderer.setupTriggerEvents(this.trigger)}launch(A,i=0,n){return typeof A=="number"?this._rippleRenderer.fadeInRipple(A,i,rA(rA({},this.rippleConfig),n)):this._rippleRenderer.fadeInRipple(0,0,rA(rA({},this.rippleConfig),A))}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","mat-ripple",""],["","matRipple",""]],hostAttrs:[1,"mat-ripple"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mat-ripple-unbounded",n.unbounded)},inputs:{color:[0,"matRippleColor","color"],unbounded:[0,"matRippleUnbounded","unbounded"],centered:[0,"matRippleCentered","centered"],radius:[0,"matRippleRadius","radius"],animation:[0,"matRippleAnimation","animation"],disabled:[0,"matRippleDisabled","disabled"],trigger:[0,"matRippleTrigger","trigger"]},exportAs:["matRipple"]})}return t})(),Ps=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,it]})}return t})(),wk=(()=>{class t{_animationMode=f(Si,{optional:!0});state="unchecked";disabled=!1;appearance="full";constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-pseudo-checkbox"]],hostAttrs:[1,"mat-pseudo-checkbox"],hostVars:12,hostBindings:function(i,n){i&2&&ue("mat-pseudo-checkbox-indeterminate",n.state==="indeterminate")("mat-pseudo-checkbox-checked",n.state==="checked")("mat-pseudo-checkbox-disabled",n.disabled)("mat-pseudo-checkbox-minimal",n.appearance==="minimal")("mat-pseudo-checkbox-full",n.appearance==="full")("_mat-animation-noopable",n._animationMode==="NoopAnimations")},inputs:{state:"state",disabled:"disabled",appearance:"appearance"},decls:0,vars:0,template:function(i,n){},styles:['.mat-pseudo-checkbox{border-radius:2px;cursor:pointer;display:inline-block;vertical-align:middle;box-sizing:border-box;position:relative;flex-shrink:0;transition:border-color 90ms cubic-bezier(0, 0, 0.2, 0.1),background-color 90ms cubic-bezier(0, 0, 0.2, 0.1)}.mat-pseudo-checkbox::after{position:absolute;opacity:0;content:"";border-bottom:2px solid currentColor;transition:opacity 90ms cubic-bezier(0, 0, 0.2, 0.1)}.mat-pseudo-checkbox._mat-animation-noopable{transition:none !important;animation:none !important}.mat-pseudo-checkbox._mat-animation-noopable::after{transition:none}.mat-pseudo-checkbox-disabled{cursor:default}.mat-pseudo-checkbox-indeterminate::after{left:1px;opacity:1;border-radius:2px}.mat-pseudo-checkbox-checked::after{left:1px;border-left:2px solid currentColor;transform:rotate(-45deg);opacity:1;box-sizing:content-box}.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-checked::after,.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-indeterminate::after{color:var(--mat-minimal-pseudo-checkbox-selected-checkmark-color, var(--mat-sys-primary))}.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled::after,.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled::after{color:var(--mat-minimal-pseudo-checkbox-disabled-selected-checkmark-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-pseudo-checkbox-full{border-color:var(--mat-full-pseudo-checkbox-unselected-icon-color, var(--mat-sys-on-surface-variant));border-width:2px;border-style:solid}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-disabled{border-color:var(--mat-full-pseudo-checkbox-disabled-unselected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-checked,.mat-pseudo-checkbox-full.mat-pseudo-checkbox-indeterminate{background-color:var(--mat-full-pseudo-checkbox-selected-icon-color, var(--mat-sys-primary));border-color:rgba(0,0,0,0)}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-checked::after,.mat-pseudo-checkbox-full.mat-pseudo-checkbox-indeterminate::after{color:var(--mat-full-pseudo-checkbox-selected-checkmark-color, var(--mat-sys-on-primary))}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled,.mat-pseudo-checkbox-full.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled{background-color:var(--mat-full-pseudo-checkbox-disabled-selected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-checked.mat-pseudo-checkbox-disabled::after,.mat-pseudo-checkbox-full.mat-pseudo-checkbox-indeterminate.mat-pseudo-checkbox-disabled::after{color:var(--mat-full-pseudo-checkbox-disabled-selected-checkmark-color, var(--mat-sys-surface))}.mat-pseudo-checkbox{width:18px;height:18px}.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-checked::after{width:14px;height:6px;transform-origin:center;top:-4.2426406871px;left:0;bottom:0;right:0;margin:auto}.mat-pseudo-checkbox-minimal.mat-pseudo-checkbox-indeterminate::after{top:8px;width:16px}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-checked::after{width:10px;height:4px;transform-origin:center;top:-2.8284271247px;left:0;bottom:0;right:0;margin:auto}.mat-pseudo-checkbox-full.mat-pseudo-checkbox-indeterminate::after{top:6px;width:12px}'],encapsulation:2,changeDetection:0})}return t})(),Dk=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it]})}return t})(),yk=new dA("MAT_OPTION_PARENT_COMPONENT"),vk=new dA("MatOptgroup");var pk=class{source;isUserInput;constructor(e,A=!1){this.source=e,this.isUserInput=A}},U2=(()=>{class t{_element=f(ee);_changeDetectorRef=f(It);_parent=f(yk,{optional:!0});group=f(vk,{optional:!0});_signalDisableRipple=!1;_selected=!1;_active=!1;_disabled=!1;_mostRecentViewValue="";get multiple(){return this._parent&&this._parent.multiple}get selected(){return this._selected}value;id=f(sn).getId("mat-option-");get disabled(){return this.group&&this.group.disabled||this._disabled}set disabled(A){this._disabled=A}get disableRipple(){return this._signalDisableRipple?this._parent.disableRipple():!!this._parent?.disableRipple}get hideSingleSelectionIndicator(){return!!(this._parent&&this._parent.hideSingleSelectionIndicator)}onSelectionChange=new WA;_text;_stateChanges=new OA;constructor(){let A=f(Rn);A.load(lr),A.load(bu),this._signalDisableRipple=!!this._parent&&p2(this._parent.disableRipple)}get active(){return this._active}get viewValue(){return(this._text?.nativeElement.textContent||"").trim()}select(A=!0){this._selected||(this._selected=!0,this._changeDetectorRef.markForCheck(),A&&this._emitSelectionChangeEvent())}deselect(A=!0){this._selected&&(this._selected=!1,this._changeDetectorRef.markForCheck(),A&&this._emitSelectionChangeEvent())}focus(A,i){let n=this._getHostElement();typeof n.focus=="function"&&n.focus(i)}setActiveStyles(){this._active||(this._active=!0,this._changeDetectorRef.markForCheck())}setInactiveStyles(){this._active&&(this._active=!1,this._changeDetectorRef.markForCheck())}getLabel(){return this.viewValue}_handleKeydown(A){(A.keyCode===13||A.keyCode===32)&&!ir(A)&&(this._selectViaInteraction(),A.preventDefault())}_selectViaInteraction(){this.disabled||(this._selected=this.multiple?!this._selected:!0,this._changeDetectorRef.markForCheck(),this._emitSelectionChangeEvent(!0))}_getTabIndex(){return this.disabled?"-1":"0"}_getHostElement(){return this._element.nativeElement}ngAfterViewChecked(){if(this._selected){let A=this.viewValue;A!==this._mostRecentViewValue&&(this._mostRecentViewValue&&this._stateChanges.next(),this._mostRecentViewValue=A)}}ngOnDestroy(){this._stateChanges.complete()}_emitSelectionChangeEvent(A=!1){this.onSelectionChange.emit(new pk(this,A))}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-option"]],viewQuery:function(i,n){if(i&1&&Te(puA,7),i&2){let o;XA(o=$A())&&(n._text=o.first)}},hostAttrs:["role","option",1,"mat-mdc-option","mdc-list-item"],hostVars:11,hostBindings:function(i,n){i&1&&hA("click",function(){return n._selectViaInteraction()})("keydown",function(r){return n._handleKeydown(r)}),i&2&&(Hs("id",n.id),Ne("aria-selected",n.selected)("aria-disabled",n.disabled.toString()),ue("mdc-list-item--selected",n.selected)("mat-mdc-option-multiple",n.multiple)("mat-mdc-option-active",n.active)("mdc-list-item--disabled",n.disabled))},inputs:{value:"value",id:"id",disabled:[2,"disabled","disabled",ie]},outputs:{onSelectionChange:"onSelectionChange"},exportAs:["matOption"],ngContentSelectors:DuA,decls:8,vars:5,consts:[["text",""],["aria-hidden","true",1,"mat-mdc-option-pseudo-checkbox",3,"disabled","state"],[1,"mdc-list-item__primary-text"],["state","checked","aria-hidden","true","appearance","minimal",1,"mat-mdc-option-pseudo-checkbox",3,"disabled"],[1,"cdk-visually-hidden"],["aria-hidden","true","mat-ripple","",1,"mat-mdc-option-ripple","mat-focus-indicator",3,"matRippleTrigger","matRippleDisabled"]],template:function(i,n){i&1&&(jt(wuA),_A(0,yuA,1,2,"mat-pseudo-checkbox",1),Fe(1),S(2,"span",2,0),Fe(4,1),F(),_A(5,vuA,1,1,"mat-pseudo-checkbox",3)(6,buA,2,1,"span",4),JA(7,"div",5)),i&2&&(GA(n.multiple?0:-1),G(5),GA(!n.multiple&&n.selected&&!n.hideSingleSelectionIndicator?5:-1),G(),GA(n.group&&n.group._inert?6:-1),G(),yA("matRippleTrigger",n._getHostElement())("matRippleDisabled",n.disabled||n.disableRipple))},dependencies:[wk,rs],styles:['.mat-mdc-option{-webkit-user-select:none;user-select:none;-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;display:flex;position:relative;align-items:center;justify-content:flex-start;overflow:hidden;min-height:48px;padding:0 16px;cursor:pointer;-webkit-tap-highlight-color:rgba(0,0,0,0);color:var(--mat-option-label-text-color, var(--mat-sys-on-surface));font-family:var(--mat-option-label-text-font, var(--mat-sys-label-large-font));line-height:var(--mat-option-label-text-line-height, var(--mat-sys-label-large-line-height));font-size:var(--mat-option-label-text-size, var(--mat-sys-body-large-size));letter-spacing:var(--mat-option-label-text-tracking, var(--mat-sys-label-large-tracking));font-weight:var(--mat-option-label-text-weight, var(--mat-sys-body-large-weight))}.mat-mdc-option:hover:not(.mdc-list-item--disabled){background-color:var(--mat-option-hover-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-hover-state-layer-opacity) * 100%), transparent))}.mat-mdc-option:focus.mdc-list-item,.mat-mdc-option.mat-mdc-option-active.mdc-list-item{background-color:var(--mat-option-focus-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-focus-state-layer-opacity) * 100%), transparent));outline:0}.mat-mdc-option.mdc-list-item--selected:not(.mdc-list-item--disabled):not(.mat-mdc-option-multiple){background-color:var(--mat-option-selected-state-layer-color, var(--mat-sys-secondary-container))}.mat-mdc-option.mdc-list-item--selected:not(.mdc-list-item--disabled):not(.mat-mdc-option-multiple) .mdc-list-item__primary-text{color:var(--mat-option-selected-state-label-text-color, var(--mat-sys-on-secondary-container))}.mat-mdc-option .mat-pseudo-checkbox{--mat-minimal-pseudo-checkbox-selected-checkmark-color: var(--mat-option-selected-state-label-text-color, var(--mat-sys-on-secondary-container))}.mat-mdc-option.mdc-list-item{align-items:center;background:rgba(0,0,0,0)}.mat-mdc-option.mdc-list-item--disabled{cursor:default;pointer-events:none}.mat-mdc-option.mdc-list-item--disabled .mat-mdc-option-pseudo-checkbox,.mat-mdc-option.mdc-list-item--disabled .mdc-list-item__primary-text,.mat-mdc-option.mdc-list-item--disabled>mat-icon{opacity:.38}.mat-mdc-optgroup .mat-mdc-option:not(.mat-mdc-option-multiple){padding-left:32px}[dir=rtl] .mat-mdc-optgroup .mat-mdc-option:not(.mat-mdc-option-multiple){padding-left:16px;padding-right:32px}.mat-mdc-option .mat-icon,.mat-mdc-option .mat-pseudo-checkbox-full{margin-right:16px;flex-shrink:0}[dir=rtl] .mat-mdc-option .mat-icon,[dir=rtl] .mat-mdc-option .mat-pseudo-checkbox-full{margin-right:0;margin-left:16px}.mat-mdc-option .mat-pseudo-checkbox-minimal{margin-left:16px;flex-shrink:0}[dir=rtl] .mat-mdc-option .mat-pseudo-checkbox-minimal{margin-right:16px;margin-left:0}.mat-mdc-option .mat-mdc-option-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-mdc-option .mdc-list-item__primary-text{white-space:normal;font-size:inherit;font-weight:inherit;letter-spacing:inherit;line-height:inherit;font-family:inherit;text-decoration:inherit;text-transform:inherit;margin-right:auto}[dir=rtl] .mat-mdc-option .mdc-list-item__primary-text{margin-right:0;margin-left:auto}@media(forced-colors: active){.mat-mdc-option.mdc-list-item--selected:not(:has(.mat-mdc-option-pseudo-checkbox))::after{content:"";position:absolute;top:50%;right:16px;transform:translateY(-50%);width:10px;height:0;border-bottom:solid 10px;border-radius:10px}[dir=rtl] .mat-mdc-option.mdc-list-item--selected:not(:has(.mat-mdc-option-pseudo-checkbox))::after{right:auto;left:16px}}.mat-mdc-option-multiple{--mdc-list-list-item-selected-container-color:var(--mdc-list-list-item-container-color, transparent)}.mat-mdc-option-active .mat-focus-indicator::before{content:""}'],encapsulation:2,changeDetection:0})}return t})();function AP(t,e,A){if(A.length){let i=e.toArray(),n=A.toArray(),o=0;for(let r=0;r<t+1;r++)i[r].group&&i[r].group===n[o]&&o++;return o}return 0}function eP(t,e,A,i){return t<A?t:t+e>A+i?Math.max(0,t-i+e):A}var bk=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[Ps,it,Dk]})}return t})(),WO={capture:!0},XO=["focus","mousedown","mouseenter","touchstart"],hk="mat-ripple-loader-uninitialized",uk="mat-ripple-loader-class-name",$O="mat-ripple-loader-centered",s8="mat-ripple-loader-disabled",Mk=(()=>{class t{_document=f(at,{optional:!0});_animationMode=f(Si,{optional:!0});_globalRippleOptions=f(G2,{optional:!0});_platform=f(Ii);_ngZone=f(Qe);_injector=f(Rt);_hosts=new Map;constructor(){this._ngZone.runOutsideAngular(()=>{for(let A of XO)this._document?.addEventListener(A,this._onInteraction,WO)})}ngOnDestroy(){let A=this._hosts.keys();for(let i of A)this.destroyRipple(i);for(let i of XO)this._document?.removeEventListener(i,this._onInteraction,WO)}configureRipple(A,i){A.setAttribute(hk,this._globalRippleOptions?.namespace??""),(i.className||!A.hasAttribute(uk))&&A.setAttribute(uk,i.className||""),i.centered&&A.setAttribute($O,""),i.disabled&&A.setAttribute(s8,"")}setDisabled(A,i){let n=this._hosts.get(A);n?(n.target.rippleDisabled=i,!i&&!n.hasSetUpEvents&&(n.hasSetUpEvents=!0,n.renderer.setupTriggerEvents(A))):i?A.setAttribute(s8,""):A.removeAttribute(s8)}_onInteraction=A=>{let i=oc(A);if(i instanceof HTMLElement){let n=i.closest(`[${hk}="${this._globalRippleOptions?.namespace??""}"]`);n&&this._createRipple(n)}};_createRipple(A){if(!this._document||this._hosts.has(A))return;A.querySelector(".mat-ripple")?.remove();let i=this._document.createElement("span");i.classList.add("mat-ripple",A.getAttribute(uk)),A.append(i);let n=this._animationMode==="NoopAnimations",o=this._globalRippleOptions,r=n?0:o?.animation?.enterDuration??a8.enterDuration,s=n?0:o?.animation?.exitDuration??a8.exitDuration,a={rippleDisabled:n||o?.disabled||A.hasAttribute(s8),rippleConfig:{centered:A.hasAttribute($O),terminateOnPointerUp:o?.terminateOnPointerUp,animation:{enterDuration:r,exitDuration:s}}},c=new vB(a,this._ngZone,i,this._platform,this._injector),l=!a.rippleDisabled;l&&c.setupTriggerEvents(A),this._hosts.set(A,{target:a,renderer:c,hasSetUpEvents:l}),A.removeAttribute(hk)}destroyRipple(A){let i=this._hosts.get(A);i&&(i.renderer._removeTriggerEvents(),this._hosts.delete(A))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),MB=(()=>{class t{labelPosition;static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["div","mat-internal-form-field",""]],hostAttrs:[1,"mdc-form-field","mat-internal-form-field"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mdc-form-field--align-end",n.labelPosition==="before")},inputs:{labelPosition:"labelPosition"},attrs:MuA,ngContentSelectors:kuA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},styles:[".mat-internal-form-field{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;display:inline-flex;align-items:center;vertical-align:middle}.mat-internal-form-field>label{margin-left:0;margin-right:auto;padding-left:4px;padding-right:0;order:0}[dir=rtl] .mat-internal-form-field>label{margin-left:auto;margin-right:0;padding-left:0;padding-right:4px}.mdc-form-field--align-end>label{margin-left:auto;margin-right:0;padding-left:0;padding-right:4px;order:-1}[dir=rtl] .mdc-form-field--align-end .mdc-form-field--align-end label{margin-left:0;margin-right:auto;padding-left:4px;padding-right:0}"],encapsulation:2,changeDetection:0})}return t})();var LuA=["mat-button",""],kk=[[["",8,"material-icons",3,"iconPositionEnd",""],["mat-icon",3,"iconPositionEnd",""],["","matButtonIcon","",3,"iconPositionEnd",""]],"*",[["","iconPositionEnd","",8,"material-icons"],["mat-icon","iconPositionEnd",""],["","matButtonIcon","","iconPositionEnd",""]]],Sk=[".material-icons:not([iconPositionEnd]), mat-icon:not([iconPositionEnd]), [matButtonIcon]:not([iconPositionEnd])","*",".material-icons[iconPositionEnd], mat-icon[iconPositionEnd], [matButtonIcon][iconPositionEnd]"];var FuA="@media(forced-colors: active){.mat-mdc-button:not(.mdc-button--outlined),.mat-mdc-unelevated-button:not(.mdc-button--outlined),.mat-mdc-raised-button:not(.mdc-button--outlined),.mat-mdc-outlined-button:not(.mdc-button--outlined),.mat-mdc-icon-button.mat-mdc-icon-button{outline:solid 1px}}",NuA=["mat-fab",""],_uA=["mat-mini-fab",""],GuA='.mat-mdc-fab-base{-webkit-user-select:none;user-select:none;position:relative;display:inline-flex;align-items:center;justify-content:center;box-sizing:border-box;width:56px;height:56px;padding:0;border:none;fill:currentColor;text-decoration:none;cursor:pointer;-moz-appearance:none;-webkit-appearance:none;overflow:visible;transition:box-shadow 280ms cubic-bezier(0.4, 0, 0.2, 1),opacity 15ms linear 30ms,transform 270ms 0ms cubic-bezier(0, 0, 0.2, 1);flex-shrink:0;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-fab-base .mat-mdc-button-ripple,.mat-mdc-fab-base .mat-mdc-button-persistent-ripple,.mat-mdc-fab-base .mat-mdc-button-persistent-ripple::before{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-mdc-fab-base .mat-mdc-button-ripple{overflow:hidden}.mat-mdc-fab-base .mat-mdc-button-persistent-ripple::before{content:"";opacity:0}.mat-mdc-fab-base .mdc-button__label,.mat-mdc-fab-base .mat-icon{z-index:1;position:relative}.mat-mdc-fab-base .mat-focus-indicator{top:0;left:0;right:0;bottom:0;position:absolute}.mat-mdc-fab-base:focus>.mat-focus-indicator::before{content:""}.mat-mdc-fab-base._mat-animation-noopable{transition:none !important;animation:none !important}.mat-mdc-fab-base::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:1px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}.mat-mdc-fab-base[hidden]{display:none}.mat-mdc-fab-base::-moz-focus-inner{padding:0;border:0}.mat-mdc-fab-base:active,.mat-mdc-fab-base:focus{outline:none}.mat-mdc-fab-base:hover{cursor:pointer}.mat-mdc-fab-base>svg{width:100%}.mat-mdc-fab-base .mat-icon,.mat-mdc-fab-base .material-icons{transition:transform 180ms 90ms cubic-bezier(0, 0, 0.2, 1);fill:currentColor;will-change:transform}.mat-mdc-fab-base .mat-focus-indicator::before{margin:calc(calc(var(--mat-focus-indicator-border-width, 3px) + 2px)*-1)}.mat-mdc-fab-base[disabled],.mat-mdc-fab-base.mat-mdc-button-disabled{cursor:default;pointer-events:none}.mat-mdc-fab-base[disabled],.mat-mdc-fab-base[disabled]:focus,.mat-mdc-fab-base.mat-mdc-button-disabled,.mat-mdc-fab-base.mat-mdc-button-disabled:focus{box-shadow:none}.mat-mdc-fab-base.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-fab{background-color:var(--mdc-fab-container-color, var(--mat-sys-primary-container));border-radius:var(--mdc-fab-container-shape, var(--mat-sys-corner-large));color:var(--mat-fab-foreground-color, var(--mat-sys-on-primary-container, inherit));box-shadow:var(--mdc-fab-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab:hover{box-shadow:var(--mdc-fab-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-fab:focus{box-shadow:var(--mdc-fab-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab:active,.mat-mdc-fab:focus:active{box-shadow:var(--mdc-fab-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab[disabled],.mat-mdc-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mat-fab-disabled-state-foreground-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-fab-disabled-state-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-fab .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:50%;width:48px;transform:translate(-50%, -50%);display:var(--mat-fab-touch-target-display, block)}.mat-mdc-fab .mat-ripple-element{background-color:var(--mat-fab-ripple-color, color-mix(in srgb, var(--mat-sys-on-primary-container) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-fab .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-state-layer-color, var(--mat-sys-on-primary-container))}.mat-mdc-fab.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-disabled-state-layer-color)}.mat-mdc-fab:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-fab.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-fab.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-fab.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-fab:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-mini-fab{width:40px;height:40px;background-color:var(--mdc-fab-small-container-color, var(--mat-sys-primary-container));border-radius:var(--mdc-fab-small-container-shape, var(--mat-sys-corner-medium));color:var(--mat-fab-small-foreground-color, var(--mat-sys-on-primary-container, inherit));box-shadow:var(--mdc-fab-small-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab:hover{box-shadow:var(--mdc-fab-small-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-mini-fab:focus{box-shadow:var(--mdc-fab-small-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab:active,.mat-mdc-mini-fab:focus:active{box-shadow:var(--mdc-fab-small-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab[disabled],.mat-mdc-mini-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mat-fab-small-disabled-state-foreground-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-fab-small-disabled-state-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-mini-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-mini-fab .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:50%;width:48px;transform:translate(-50%, -50%);display:var(--mat-fab-small-touch-target-display)}.mat-mdc-mini-fab .mat-ripple-element{background-color:var(--mat-fab-small-ripple-color, color-mix(in srgb, var(--mat-sys-on-primary-container) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-mini-fab .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-small-state-layer-color, var(--mat-sys-on-primary-container))}.mat-mdc-mini-fab.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-small-disabled-state-layer-color)}.mat-mdc-mini-fab:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-mini-fab.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-mini-fab.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-mini-fab.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-mini-fab:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-extended-fab{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;border-radius:24px;padding-left:20px;padding-right:20px;width:auto;max-width:100%;line-height:normal;height:var(--mdc-extended-fab-container-height, 56px);border-radius:var(--mdc-extended-fab-container-shape, var(--mat-sys-corner-large));font-family:var(--mdc-extended-fab-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-extended-fab-label-text-size, var(--mat-sys-label-large-size));font-weight:var(--mdc-extended-fab-label-text-weight, var(--mat-sys-label-large-weight));letter-spacing:var(--mdc-extended-fab-label-text-tracking, var(--mat-sys-label-large-tracking));box-shadow:var(--mdc-extended-fab-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab:hover{box-shadow:var(--mdc-extended-fab-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-extended-fab:focus{box-shadow:var(--mdc-extended-fab-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab:active,.mat-mdc-extended-fab:focus:active{box-shadow:var(--mdc-extended-fab-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab[disabled],.mat-mdc-extended-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none}.mat-mdc-extended-fab[disabled],.mat-mdc-extended-fab[disabled]:focus,.mat-mdc-extended-fab.mat-mdc-button-disabled,.mat-mdc-extended-fab.mat-mdc-button-disabled:focus{box-shadow:none}.mat-mdc-extended-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}[dir=rtl] .mat-mdc-extended-fab .mdc-button__label+.mat-icon,[dir=rtl] .mat-mdc-extended-fab .mdc-button__label+.material-icons,.mat-mdc-extended-fab>.mat-icon,.mat-mdc-extended-fab>.material-icons{margin-left:-8px;margin-right:12px}.mat-mdc-extended-fab .mdc-button__label+.mat-icon,.mat-mdc-extended-fab .mdc-button__label+.material-icons,[dir=rtl] .mat-mdc-extended-fab>.mat-icon,[dir=rtl] .mat-mdc-extended-fab>.material-icons{margin-left:12px;margin-right:-8px}.mat-mdc-extended-fab .mat-mdc-button-touch-target{width:100%}',UuA=["mat-icon-button",""],KuA=["*"];var YuA=new dA("MAT_BUTTON_CONFIG");var JuA=[{attribute:"mat-button",mdcClasses:["mdc-button","mat-mdc-button"]},{attribute:"mat-flat-button",mdcClasses:["mdc-button","mdc-button--unelevated","mat-mdc-unelevated-button"]},{attribute:"mat-raised-button",mdcClasses:["mdc-button","mdc-button--raised","mat-mdc-raised-button"]},{attribute:"mat-stroked-button",mdcClasses:["mdc-button","mdc-button--outlined","mat-mdc-outlined-button"]},{attribute:"mat-fab",mdcClasses:["mdc-fab","mat-mdc-fab-base","mat-mdc-fab"]},{attribute:"mat-mini-fab",mdcClasses:["mdc-fab","mat-mdc-fab-base","mdc-fab--mini","mat-mdc-mini-fab"]},{attribute:"mat-icon-button",mdcClasses:["mdc-icon-button","mat-mdc-icon-button"]}],l8=(()=>{class t{_elementRef=f(ee);_ngZone=f(Qe);_animationMode=f(Si,{optional:!0});_focusMonitor=f(dr);_rippleLoader=f(Mk);_isFab=!1;color;get disableRipple(){return this._disableRipple}set disableRipple(A){this._disableRipple=A,this._updateRippleDisabled()}_disableRipple=!1;get disabled(){return this._disabled}set disabled(A){this._disabled=A,this._updateRippleDisabled()}_disabled=!1;ariaDisabled;disabledInteractive;constructor(){f(Rn).load(lr);let A=f(YuA,{optional:!0}),i=this._elementRef.nativeElement,n=i.classList;this.disabledInteractive=A?.disabledInteractive??!1,this.color=A?.color??null,this._rippleLoader?.configureRipple(i,{className:"mat-mdc-button-ripple"});for(let{attribute:o,mdcClasses:r}of JuA)i.hasAttribute(o)&&n.add(...r)}ngAfterViewInit(){this._focusMonitor.monitor(this._elementRef,!0)}ngOnDestroy(){this._focusMonitor.stopMonitoring(this._elementRef),this._rippleLoader?.destroyRipple(this._elementRef.nativeElement)}focus(A="program",i){A?this._focusMonitor.focusVia(this._elementRef.nativeElement,A,i):this._elementRef.nativeElement.focus(i)}_getAriaDisabled(){return this.ariaDisabled!=null?this.ariaDisabled:this.disabled&&this.disabledInteractive?!0:null}_getDisabledAttribute(){return this.disabledInteractive||!this.disabled?null:!0}_updateRippleDisabled(){this._rippleLoader?.setDisabled(this._elementRef.nativeElement,this.disableRipple||this.disabled)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,inputs:{color:"color",disableRipple:[2,"disableRipple","disableRipple",ie],disabled:[2,"disabled","disabled",ie],ariaDisabled:[2,"aria-disabled","ariaDisabled",ie],disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie]}})}return t})();var Dr=(()=>{class t extends l8{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["button","mat-button",""],["button","mat-raised-button",""],["button","mat-flat-button",""],["button","mat-stroked-button",""]],hostVars:14,hostBindings:function(i,n){i&2&&(Ne("disabled",n._getDisabledAttribute())("aria-disabled",n._getAriaDisabled()),fo(n.color?"mat-"+n.color:""),ue("mat-mdc-button-disabled",n.disabled)("mat-mdc-button-disabled-interactive",n.disabledInteractive)("_mat-animation-noopable",n._animationMode==="NoopAnimations")("mat-unthemed",!n.color)("mat-mdc-button-base",!0))},exportAs:["matButton"],features:[lt],attrs:LuA,ngContentSelectors:Sk,decls:7,vars:4,consts:[[1,"mat-mdc-button-persistent-ripple"],[1,"mdc-button__label"],[1,"mat-focus-indicator"],[1,"mat-mdc-button-touch-target"]],template:function(i,n){i&1&&(jt(kk),JA(0,"span",0),Fe(1),S(2,"span",1),Fe(3,1),F(),Fe(4,2),JA(5,"span",2)(6,"span",3)),i&2&&ue("mdc-button__ripple",!n._isFab)("mdc-fab__ripple",n._isFab)},styles:['.mat-mdc-button-base{text-decoration:none}.mdc-button{-webkit-user-select:none;user-select:none;position:relative;display:inline-flex;align-items:center;justify-content:center;box-sizing:border-box;min-width:64px;border:none;outline:none;line-height:inherit;-webkit-appearance:none;overflow:visible;vertical-align:middle;background:rgba(0,0,0,0);padding:0 8px}.mdc-button::-moz-focus-inner{padding:0;border:0}.mdc-button:active{outline:none}.mdc-button:hover{cursor:pointer}.mdc-button:disabled{cursor:default;pointer-events:none}.mdc-button[hidden]{display:none}.mdc-button .mdc-button__label{position:relative}.mat-mdc-button{padding:0 var(--mat-text-button-horizontal-padding, 12px);height:var(--mdc-text-button-container-height, 40px);font-family:var(--mdc-text-button-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-text-button-label-text-size, var(--mat-sys-label-large-size));letter-spacing:var(--mdc-text-button-label-text-tracking, var(--mat-sys-label-large-tracking));text-transform:var(--mdc-text-button-label-text-transform);font-weight:var(--mdc-text-button-label-text-weight, var(--mat-sys-label-large-weight))}.mat-mdc-button,.mat-mdc-button .mdc-button__ripple{border-radius:var(--mdc-text-button-container-shape, var(--mat-sys-corner-full))}.mat-mdc-button:not(:disabled){color:var(--mdc-text-button-label-text-color, var(--mat-sys-primary))}.mat-mdc-button[disabled],.mat-mdc-button.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mdc-text-button-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-button.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-button:has(.material-icons,mat-icon,[matButtonIcon]){padding:0 var(--mat-text-button-with-icon-horizontal-padding, 16px)}.mat-mdc-button>.mat-icon{margin-right:var(--mat-text-button-icon-spacing, 8px);margin-left:var(--mat-text-button-icon-offset, -4px)}[dir=rtl] .mat-mdc-button>.mat-icon{margin-right:var(--mat-text-button-icon-offset, -4px);margin-left:var(--mat-text-button-icon-spacing, 8px)}.mat-mdc-button .mdc-button__label+.mat-icon{margin-right:var(--mat-text-button-icon-offset, -4px);margin-left:var(--mat-text-button-icon-spacing, 8px)}[dir=rtl] .mat-mdc-button .mdc-button__label+.mat-icon{margin-right:var(--mat-text-button-icon-spacing, 8px);margin-left:var(--mat-text-button-icon-offset, -4px)}.mat-mdc-button .mat-ripple-element{background-color:var(--mat-text-button-ripple-color, color-mix(in srgb, var(--mat-sys-primary) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-button .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-text-button-state-layer-color, var(--mat-sys-primary))}.mat-mdc-button.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-text-button-disabled-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-button:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-text-button-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-button.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-button.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-button.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-text-button-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-button:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-text-button-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-button .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:0;right:0;transform:translateY(-50%);display:var(--mat-text-button-touch-target-display, block)}.mat-mdc-unelevated-button{transition:box-shadow 280ms cubic-bezier(0.4, 0, 0.2, 1);height:var(--mdc-filled-button-container-height, 40px);font-family:var(--mdc-filled-button-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-filled-button-label-text-size, var(--mat-sys-label-large-size));letter-spacing:var(--mdc-filled-button-label-text-tracking, var(--mat-sys-label-large-tracking));text-transform:var(--mdc-filled-button-label-text-transform);font-weight:var(--mdc-filled-button-label-text-weight, var(--mat-sys-label-large-weight));padding:0 var(--mat-filled-button-horizontal-padding, 24px)}.mat-mdc-unelevated-button>.mat-icon{margin-right:var(--mat-filled-button-icon-spacing, 8px);margin-left:var(--mat-filled-button-icon-offset, -8px)}[dir=rtl] .mat-mdc-unelevated-button>.mat-icon{margin-right:var(--mat-filled-button-icon-offset, -8px);margin-left:var(--mat-filled-button-icon-spacing, 8px)}.mat-mdc-unelevated-button .mdc-button__label+.mat-icon{margin-right:var(--mat-filled-button-icon-offset, -8px);margin-left:var(--mat-filled-button-icon-spacing, 8px)}[dir=rtl] .mat-mdc-unelevated-button .mdc-button__label+.mat-icon{margin-right:var(--mat-filled-button-icon-spacing, 8px);margin-left:var(--mat-filled-button-icon-offset, -8px)}.mat-mdc-unelevated-button .mat-ripple-element{background-color:var(--mat-filled-button-ripple-color, color-mix(in srgb, var(--mat-sys-on-primary) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-unelevated-button .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-filled-button-state-layer-color, var(--mat-sys-on-primary))}.mat-mdc-unelevated-button.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-filled-button-disabled-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-unelevated-button:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-filled-button-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-unelevated-button.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-unelevated-button.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-unelevated-button.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-filled-button-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-unelevated-button:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-filled-button-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-unelevated-button .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:0;right:0;transform:translateY(-50%);display:var(--mat-filled-button-touch-target-display, block)}.mat-mdc-unelevated-button:not(:disabled){color:var(--mdc-filled-button-label-text-color, var(--mat-sys-on-primary));background-color:var(--mdc-filled-button-container-color, var(--mat-sys-primary))}.mat-mdc-unelevated-button,.mat-mdc-unelevated-button .mdc-button__ripple{border-radius:var(--mdc-filled-button-container-shape, var(--mat-sys-corner-full))}.mat-mdc-unelevated-button[disabled],.mat-mdc-unelevated-button.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mdc-filled-button-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mdc-filled-button-disabled-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-unelevated-button.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-raised-button{transition:box-shadow 280ms cubic-bezier(0.4, 0, 0.2, 1);box-shadow:var(--mdc-protected-button-container-elevation-shadow, var(--mat-sys-level1));height:var(--mdc-protected-button-container-height, 40px);font-family:var(--mdc-protected-button-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-protected-button-label-text-size, var(--mat-sys-label-large-size));letter-spacing:var(--mdc-protected-button-label-text-tracking, var(--mat-sys-label-large-tracking));text-transform:var(--mdc-protected-button-label-text-transform);font-weight:var(--mdc-protected-button-label-text-weight, var(--mat-sys-label-large-weight));padding:0 var(--mat-protected-button-horizontal-padding, 24px)}.mat-mdc-raised-button>.mat-icon{margin-right:var(--mat-protected-button-icon-spacing, 8px);margin-left:var(--mat-protected-button-icon-offset, -8px)}[dir=rtl] .mat-mdc-raised-button>.mat-icon{margin-right:var(--mat-protected-button-icon-offset, -8px);margin-left:var(--mat-protected-button-icon-spacing, 8px)}.mat-mdc-raised-button .mdc-button__label+.mat-icon{margin-right:var(--mat-protected-button-icon-offset, -8px);margin-left:var(--mat-protected-button-icon-spacing, 8px)}[dir=rtl] .mat-mdc-raised-button .mdc-button__label+.mat-icon{margin-right:var(--mat-protected-button-icon-spacing, 8px);margin-left:var(--mat-protected-button-icon-offset, -8px)}.mat-mdc-raised-button .mat-ripple-element{background-color:var(--mat-protected-button-ripple-color, color-mix(in srgb, var(--mat-sys-primary) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-raised-button .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-protected-button-state-layer-color, var(--mat-sys-primary))}.mat-mdc-raised-button.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-protected-button-disabled-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-raised-button:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-protected-button-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-raised-button.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-raised-button.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-raised-button.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-protected-button-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-raised-button:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-protected-button-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-raised-button .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:0;right:0;transform:translateY(-50%);display:var(--mat-protected-button-touch-target-display, block)}.mat-mdc-raised-button:not(:disabled){color:var(--mdc-protected-button-label-text-color, var(--mat-sys-primary));background-color:var(--mdc-protected-button-container-color, var(--mat-sys-surface))}.mat-mdc-raised-button,.mat-mdc-raised-button .mdc-button__ripple{border-radius:var(--mdc-protected-button-container-shape, var(--mat-sys-corner-full))}.mat-mdc-raised-button:hover{box-shadow:var(--mdc-protected-button-hover-container-elevation-shadow, var(--mat-sys-level2))}.mat-mdc-raised-button:focus{box-shadow:var(--mdc-protected-button-focus-container-elevation-shadow, var(--mat-sys-level1))}.mat-mdc-raised-button:active,.mat-mdc-raised-button:focus:active{box-shadow:var(--mdc-protected-button-pressed-container-elevation-shadow, var(--mat-sys-level1))}.mat-mdc-raised-button[disabled],.mat-mdc-raised-button.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mdc-protected-button-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mdc-protected-button-disabled-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-raised-button[disabled].mat-mdc-button-disabled,.mat-mdc-raised-button.mat-mdc-button-disabled.mat-mdc-button-disabled{box-shadow:var(--mdc-protected-button-disabled-container-elevation-shadow, var(--mat-sys-level0))}.mat-mdc-raised-button.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-outlined-button{border-style:solid;transition:border 280ms cubic-bezier(0.4, 0, 0.2, 1);height:var(--mdc-outlined-button-container-height, 40px);font-family:var(--mdc-outlined-button-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-outlined-button-label-text-size, var(--mat-sys-label-large-size));letter-spacing:var(--mdc-outlined-button-label-text-tracking, var(--mat-sys-label-large-tracking));text-transform:var(--mdc-outlined-button-label-text-transform);font-weight:var(--mdc-outlined-button-label-text-weight, var(--mat-sys-label-large-weight));border-radius:var(--mdc-outlined-button-container-shape, var(--mat-sys-corner-full));border-width:var(--mdc-outlined-button-outline-width, 1px);padding:0 var(--mat-outlined-button-horizontal-padding, 24px)}.mat-mdc-outlined-button>.mat-icon{margin-right:var(--mat-outlined-button-icon-spacing, 8px);margin-left:var(--mat-outlined-button-icon-offset, -8px)}[dir=rtl] .mat-mdc-outlined-button>.mat-icon{margin-right:var(--mat-outlined-button-icon-offset, -8px);margin-left:var(--mat-outlined-button-icon-spacing, 8px)}.mat-mdc-outlined-button .mdc-button__label+.mat-icon{margin-right:var(--mat-outlined-button-icon-offset, -8px);margin-left:var(--mat-outlined-button-icon-spacing, 8px)}[dir=rtl] .mat-mdc-outlined-button .mdc-button__label+.mat-icon{margin-right:var(--mat-outlined-button-icon-spacing, 8px);margin-left:var(--mat-outlined-button-icon-offset, -8px)}.mat-mdc-outlined-button .mat-ripple-element{background-color:var(--mat-outlined-button-ripple-color, color-mix(in srgb, var(--mat-sys-primary) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-outlined-button .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-outlined-button-state-layer-color, var(--mat-sys-primary))}.mat-mdc-outlined-button.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-outlined-button-disabled-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-outlined-button:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-outlined-button-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-outlined-button.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-outlined-button.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-outlined-button.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-outlined-button-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-outlined-button:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-outlined-button-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-outlined-button .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:0;right:0;transform:translateY(-50%);display:var(--mat-outlined-button-touch-target-display, block)}.mat-mdc-outlined-button:not(:disabled){color:var(--mdc-outlined-button-label-text-color, var(--mat-sys-primary));border-color:var(--mdc-outlined-button-outline-color, var(--mat-sys-outline))}.mat-mdc-outlined-button[disabled],.mat-mdc-outlined-button.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mdc-outlined-button-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));border-color:var(--mdc-outlined-button-disabled-outline-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-outlined-button.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-outlined-button .mdc-button__ripple{border-width:var(--mdc-outlined-button-outline-width, 1px);border-style:solid;border-color:rgba(0,0,0,0)}.mat-mdc-button,.mat-mdc-unelevated-button,.mat-mdc-raised-button,.mat-mdc-outlined-button{-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-button .mat-mdc-button-ripple,.mat-mdc-button .mat-mdc-button-persistent-ripple,.mat-mdc-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-unelevated-button .mat-mdc-button-ripple,.mat-mdc-unelevated-button .mat-mdc-button-persistent-ripple,.mat-mdc-unelevated-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-raised-button .mat-mdc-button-ripple,.mat-mdc-raised-button .mat-mdc-button-persistent-ripple,.mat-mdc-raised-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-outlined-button .mat-mdc-button-ripple,.mat-mdc-outlined-button .mat-mdc-button-persistent-ripple,.mat-mdc-outlined-button .mat-mdc-button-persistent-ripple::before{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-mdc-button .mat-mdc-button-ripple,.mat-mdc-unelevated-button .mat-mdc-button-ripple,.mat-mdc-raised-button .mat-mdc-button-ripple,.mat-mdc-outlined-button .mat-mdc-button-ripple{overflow:hidden}.mat-mdc-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-unelevated-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-raised-button .mat-mdc-button-persistent-ripple::before,.mat-mdc-outlined-button .mat-mdc-button-persistent-ripple::before{content:"";opacity:0}.mat-mdc-button .mdc-button__label,.mat-mdc-button .mat-icon,.mat-mdc-unelevated-button .mdc-button__label,.mat-mdc-unelevated-button .mat-icon,.mat-mdc-raised-button .mdc-button__label,.mat-mdc-raised-button .mat-icon,.mat-mdc-outlined-button .mdc-button__label,.mat-mdc-outlined-button .mat-icon{z-index:1;position:relative}.mat-mdc-button .mat-focus-indicator,.mat-mdc-unelevated-button .mat-focus-indicator,.mat-mdc-raised-button .mat-focus-indicator,.mat-mdc-outlined-button .mat-focus-indicator{top:0;left:0;right:0;bottom:0;position:absolute}.mat-mdc-button:focus>.mat-focus-indicator::before,.mat-mdc-unelevated-button:focus>.mat-focus-indicator::before,.mat-mdc-raised-button:focus>.mat-focus-indicator::before,.mat-mdc-outlined-button:focus>.mat-focus-indicator::before{content:""}.mat-mdc-button._mat-animation-noopable,.mat-mdc-unelevated-button._mat-animation-noopable,.mat-mdc-raised-button._mat-animation-noopable,.mat-mdc-outlined-button._mat-animation-noopable{transition:none !important;animation:none !important}.mat-mdc-button>.mat-icon,.mat-mdc-unelevated-button>.mat-icon,.mat-mdc-raised-button>.mat-icon,.mat-mdc-outlined-button>.mat-icon{display:inline-block;position:relative;vertical-align:top;font-size:1.125rem;height:1.125rem;width:1.125rem}.mat-mdc-outlined-button .mat-mdc-button-ripple,.mat-mdc-outlined-button .mdc-button__ripple{top:-1px;left:-1px;bottom:-1px;right:-1px}.mat-mdc-unelevated-button .mat-focus-indicator::before,.mat-mdc-raised-button .mat-focus-indicator::before{margin:calc(calc(var(--mat-focus-indicator-border-width, 3px) + 2px)*-1)}.mat-mdc-outlined-button .mat-focus-indicator::before{margin:calc(calc(var(--mat-focus-indicator-border-width, 3px) + 3px)*-1)}',"@media(forced-colors: active){.mat-mdc-button:not(.mdc-button--outlined),.mat-mdc-unelevated-button:not(.mdc-button--outlined),.mat-mdc-raised-button:not(.mdc-button--outlined),.mat-mdc-outlined-button:not(.mdc-button--outlined),.mat-mdc-icon-button.mat-mdc-icon-button{outline:solid 1px}}"],encapsulation:2,changeDetection:0})}return t})();var iP=new dA("mat-mdc-fab-default-options",{providedIn:"root",factory:nP});function nP(){return{color:"accent"}}var c8=nP(),oP=(()=>{class t extends l8{_options=f(iP,{optional:!0});_isFab=!0;extended;constructor(){super(),this._options=this._options||c8,this.color=this._options.color||c8.color}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["button","mat-fab",""]],hostVars:18,hostBindings:function(i,n){i&2&&(Ne("disabled",n._getDisabledAttribute())("aria-disabled",n._getAriaDisabled()),fo(n.color?"mat-"+n.color:""),ue("mat-mdc-button-disabled",n.disabled)("mat-mdc-button-disabled-interactive",n.disabledInteractive)("_mat-animation-noopable",n._animationMode==="NoopAnimations")("mat-unthemed",!n.color)("mat-mdc-button-base",!0)("mdc-fab--extended",n.extended)("mat-mdc-extended-fab",n.extended))},inputs:{extended:[2,"extended","extended",ie]},exportAs:["matButton"],features:[lt],attrs:NuA,ngContentSelectors:Sk,decls:7,vars:4,consts:[[1,"mat-mdc-button-persistent-ripple"],[1,"mdc-button__label"],[1,"mat-focus-indicator"],[1,"mat-mdc-button-touch-target"]],template:function(i,n){i&1&&(jt(kk),JA(0,"span",0),Fe(1),S(2,"span",1),Fe(3,1),F(),Fe(4,2),JA(5,"span",2)(6,"span",3)),i&2&&ue("mdc-button__ripple",!n._isFab)("mdc-fab__ripple",n._isFab)},styles:['.mat-mdc-fab-base{-webkit-user-select:none;user-select:none;position:relative;display:inline-flex;align-items:center;justify-content:center;box-sizing:border-box;width:56px;height:56px;padding:0;border:none;fill:currentColor;text-decoration:none;cursor:pointer;-moz-appearance:none;-webkit-appearance:none;overflow:visible;transition:box-shadow 280ms cubic-bezier(0.4, 0, 0.2, 1),opacity 15ms linear 30ms,transform 270ms 0ms cubic-bezier(0, 0, 0.2, 1);flex-shrink:0;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-fab-base .mat-mdc-button-ripple,.mat-mdc-fab-base .mat-mdc-button-persistent-ripple,.mat-mdc-fab-base .mat-mdc-button-persistent-ripple::before{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-mdc-fab-base .mat-mdc-button-ripple{overflow:hidden}.mat-mdc-fab-base .mat-mdc-button-persistent-ripple::before{content:"";opacity:0}.mat-mdc-fab-base .mdc-button__label,.mat-mdc-fab-base .mat-icon{z-index:1;position:relative}.mat-mdc-fab-base .mat-focus-indicator{top:0;left:0;right:0;bottom:0;position:absolute}.mat-mdc-fab-base:focus>.mat-focus-indicator::before{content:""}.mat-mdc-fab-base._mat-animation-noopable{transition:none !important;animation:none !important}.mat-mdc-fab-base::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:1px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}.mat-mdc-fab-base[hidden]{display:none}.mat-mdc-fab-base::-moz-focus-inner{padding:0;border:0}.mat-mdc-fab-base:active,.mat-mdc-fab-base:focus{outline:none}.mat-mdc-fab-base:hover{cursor:pointer}.mat-mdc-fab-base>svg{width:100%}.mat-mdc-fab-base .mat-icon,.mat-mdc-fab-base .material-icons{transition:transform 180ms 90ms cubic-bezier(0, 0, 0.2, 1);fill:currentColor;will-change:transform}.mat-mdc-fab-base .mat-focus-indicator::before{margin:calc(calc(var(--mat-focus-indicator-border-width, 3px) + 2px)*-1)}.mat-mdc-fab-base[disabled],.mat-mdc-fab-base.mat-mdc-button-disabled{cursor:default;pointer-events:none}.mat-mdc-fab-base[disabled],.mat-mdc-fab-base[disabled]:focus,.mat-mdc-fab-base.mat-mdc-button-disabled,.mat-mdc-fab-base.mat-mdc-button-disabled:focus{box-shadow:none}.mat-mdc-fab-base.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-fab{background-color:var(--mdc-fab-container-color, var(--mat-sys-primary-container));border-radius:var(--mdc-fab-container-shape, var(--mat-sys-corner-large));color:var(--mat-fab-foreground-color, var(--mat-sys-on-primary-container, inherit));box-shadow:var(--mdc-fab-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab:hover{box-shadow:var(--mdc-fab-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-fab:focus{box-shadow:var(--mdc-fab-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab:active,.mat-mdc-fab:focus:active{box-shadow:var(--mdc-fab-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-fab[disabled],.mat-mdc-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mat-fab-disabled-state-foreground-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-fab-disabled-state-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-fab .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:50%;width:48px;transform:translate(-50%, -50%);display:var(--mat-fab-touch-target-display, block)}.mat-mdc-fab .mat-ripple-element{background-color:var(--mat-fab-ripple-color, color-mix(in srgb, var(--mat-sys-on-primary-container) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-fab .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-state-layer-color, var(--mat-sys-on-primary-container))}.mat-mdc-fab.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-disabled-state-layer-color)}.mat-mdc-fab:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-fab.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-fab.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-fab.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-fab:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-mini-fab{width:40px;height:40px;background-color:var(--mdc-fab-small-container-color, var(--mat-sys-primary-container));border-radius:var(--mdc-fab-small-container-shape, var(--mat-sys-corner-medium));color:var(--mat-fab-small-foreground-color, var(--mat-sys-on-primary-container, inherit));box-shadow:var(--mdc-fab-small-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab:hover{box-shadow:var(--mdc-fab-small-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-mini-fab:focus{box-shadow:var(--mdc-fab-small-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab:active,.mat-mdc-mini-fab:focus:active{box-shadow:var(--mdc-fab-small-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-mini-fab[disabled],.mat-mdc-mini-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mat-fab-small-disabled-state-foreground-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-fab-small-disabled-state-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-mdc-mini-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-mini-fab .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:50%;width:48px;transform:translate(-50%, -50%);display:var(--mat-fab-small-touch-target-display)}.mat-mdc-mini-fab .mat-ripple-element{background-color:var(--mat-fab-small-ripple-color, color-mix(in srgb, var(--mat-sys-on-primary-container) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-mini-fab .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-small-state-layer-color, var(--mat-sys-on-primary-container))}.mat-mdc-mini-fab.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-fab-small-disabled-state-layer-color)}.mat-mdc-mini-fab:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-mini-fab.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-mini-fab.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-mini-fab.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-mini-fab:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-fab-small-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-extended-fab{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;border-radius:24px;padding-left:20px;padding-right:20px;width:auto;max-width:100%;line-height:normal;height:var(--mdc-extended-fab-container-height, 56px);border-radius:var(--mdc-extended-fab-container-shape, var(--mat-sys-corner-large));font-family:var(--mdc-extended-fab-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mdc-extended-fab-label-text-size, var(--mat-sys-label-large-size));font-weight:var(--mdc-extended-fab-label-text-weight, var(--mat-sys-label-large-weight));letter-spacing:var(--mdc-extended-fab-label-text-tracking, var(--mat-sys-label-large-tracking));box-shadow:var(--mdc-extended-fab-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab:hover{box-shadow:var(--mdc-extended-fab-hover-container-elevation-shadow, var(--mat-sys-level4))}.mat-mdc-extended-fab:focus{box-shadow:var(--mdc-extended-fab-focus-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab:active,.mat-mdc-extended-fab:focus:active{box-shadow:var(--mdc-extended-fab-pressed-container-elevation-shadow, var(--mat-sys-level3))}.mat-mdc-extended-fab[disabled],.mat-mdc-extended-fab.mat-mdc-button-disabled{cursor:default;pointer-events:none}.mat-mdc-extended-fab[disabled],.mat-mdc-extended-fab[disabled]:focus,.mat-mdc-extended-fab.mat-mdc-button-disabled,.mat-mdc-extended-fab.mat-mdc-button-disabled:focus{box-shadow:none}.mat-mdc-extended-fab.mat-mdc-button-disabled-interactive{pointer-events:auto}[dir=rtl] .mat-mdc-extended-fab .mdc-button__label+.mat-icon,[dir=rtl] .mat-mdc-extended-fab .mdc-button__label+.material-icons,.mat-mdc-extended-fab>.mat-icon,.mat-mdc-extended-fab>.material-icons{margin-left:-8px;margin-right:12px}.mat-mdc-extended-fab .mdc-button__label+.mat-icon,.mat-mdc-extended-fab .mdc-button__label+.material-icons,[dir=rtl] .mat-mdc-extended-fab>.mat-icon,[dir=rtl] .mat-mdc-extended-fab>.material-icons{margin-left:12px;margin-right:-8px}.mat-mdc-extended-fab .mat-mdc-button-touch-target{width:100%}'],encapsulation:2,changeDetection:0})}return t})(),rP=(()=>{class t extends l8{_options=f(iP,{optional:!0});_isFab=!0;constructor(){super(),this._options=this._options||c8,this.color=this._options.color||c8.color}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["button","mat-mini-fab",""]],hostVars:14,hostBindings:function(i,n){i&2&&(Ne("disabled",n._getDisabledAttribute())("aria-disabled",n._getAriaDisabled()),fo(n.color?"mat-"+n.color:""),ue("mat-mdc-button-disabled",n.disabled)("mat-mdc-button-disabled-interactive",n.disabledInteractive)("_mat-animation-noopable",n._animationMode==="NoopAnimations")("mat-unthemed",!n.color)("mat-mdc-button-base",!0))},exportAs:["matButton"],features:[lt],attrs:_uA,ngContentSelectors:Sk,decls:7,vars:4,consts:[[1,"mat-mdc-button-persistent-ripple"],[1,"mdc-button__label"],[1,"mat-focus-indicator"],[1,"mat-mdc-button-touch-target"]],template:function(i,n){i&1&&(jt(kk),JA(0,"span",0),Fe(1),S(2,"span",1),Fe(3,1),F(),Fe(4,2),JA(5,"span",2)(6,"span",3)),i&2&&ue("mdc-button__ripple",!n._isFab)("mdc-fab__ripple",n._isFab)},styles:[GuA],encapsulation:2,changeDetection:0})}return t})();var kB=(()=>{class t extends l8{constructor(){super(),this._rippleLoader.configureRipple(this._elementRef.nativeElement,{centered:!0})}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["button","mat-icon-button",""]],hostVars:14,hostBindings:function(i,n){i&2&&(Ne("disabled",n._getDisabledAttribute())("aria-disabled",n._getAriaDisabled()),fo(n.color?"mat-"+n.color:""),ue("mat-mdc-button-disabled",n.disabled)("mat-mdc-button-disabled-interactive",n.disabledInteractive)("_mat-animation-noopable",n._animationMode==="NoopAnimations")("mat-unthemed",!n.color)("mat-mdc-button-base",!0))},exportAs:["matButton"],features:[lt],attrs:UuA,ngContentSelectors:KuA,decls:4,vars:0,consts:[[1,"mat-mdc-button-persistent-ripple","mdc-icon-button__ripple"],[1,"mat-focus-indicator"],[1,"mat-mdc-button-touch-target"]],template:function(i,n){i&1&&(jt(),JA(0,"span",0),Fe(1),JA(2,"span",1)(3,"span",2))},styles:['.mat-mdc-icon-button{-webkit-user-select:none;user-select:none;display:inline-block;position:relative;box-sizing:border-box;border:none;outline:none;background-color:rgba(0,0,0,0);fill:currentColor;color:inherit;text-decoration:none;cursor:pointer;z-index:0;overflow:visible;border-radius:50%;flex-shrink:0;text-align:center;width:var(--mdc-icon-button-state-layer-size, 40px);height:var(--mdc-icon-button-state-layer-size, 40px);padding:calc(calc(var(--mdc-icon-button-state-layer-size, 40px) - var(--mdc-icon-button-icon-size, 24px)) / 2);font-size:var(--mdc-icon-button-icon-size, 24px);color:var(--mdc-icon-button-icon-color, var(--mat-sys-on-surface-variant));-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-icon-button .mat-mdc-button-ripple,.mat-mdc-icon-button .mat-mdc-button-persistent-ripple,.mat-mdc-icon-button .mat-mdc-button-persistent-ripple::before{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:inherit}.mat-mdc-icon-button .mat-mdc-button-ripple{overflow:hidden}.mat-mdc-icon-button .mat-mdc-button-persistent-ripple::before{content:"";opacity:0}.mat-mdc-icon-button .mdc-button__label,.mat-mdc-icon-button .mat-icon{z-index:1;position:relative}.mat-mdc-icon-button .mat-focus-indicator{top:0;left:0;right:0;bottom:0;position:absolute}.mat-mdc-icon-button:focus>.mat-focus-indicator::before{content:""}.mat-mdc-icon-button .mat-ripple-element{background-color:var(--mat-icon-button-ripple-color, color-mix(in srgb, var(--mat-sys-on-surface-variant) calc(var(--mat-sys-pressed-state-layer-opacity) * 100%), transparent))}.mat-mdc-icon-button .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-icon-button-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-icon-button.mat-mdc-button-disabled .mat-mdc-button-persistent-ripple::before{background-color:var(--mat-icon-button-disabled-state-layer-color, var(--mat-sys-on-surface-variant))}.mat-mdc-icon-button:hover>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-icon-button-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-icon-button.cdk-program-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-icon-button.cdk-keyboard-focused>.mat-mdc-button-persistent-ripple::before,.mat-mdc-icon-button.mat-mdc-button-disabled-interactive:focus>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-icon-button-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mat-mdc-icon-button:active>.mat-mdc-button-persistent-ripple::before{opacity:var(--mat-icon-button-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity))}.mat-mdc-icon-button .mat-mdc-button-touch-target{position:absolute;top:50%;height:48px;left:50%;width:48px;transform:translate(-50%, -50%);display:var(--mat-icon-button-touch-target-display, block)}.mat-mdc-icon-button._mat-animation-noopable{transition:none !important;animation:none !important}.mat-mdc-icon-button[disabled],.mat-mdc-icon-button.mat-mdc-button-disabled{cursor:default;pointer-events:none;color:var(--mdc-icon-button-disabled-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-icon-button.mat-mdc-button-disabled-interactive{pointer-events:auto}.mat-mdc-icon-button img,.mat-mdc-icon-button svg{width:var(--mdc-icon-button-icon-size, 24px);height:var(--mdc-icon-button-icon-size, 24px);vertical-align:baseline}.mat-mdc-icon-button .mat-mdc-button-persistent-ripple{border-radius:50%}.mat-mdc-icon-button[hidden]{display:none}.mat-mdc-icon-button.mat-unthemed:not(.mdc-ripple-upgraded):focus::before,.mat-mdc-icon-button.mat-primary:not(.mdc-ripple-upgraded):focus::before,.mat-mdc-icon-button.mat-accent:not(.mdc-ripple-upgraded):focus::before,.mat-mdc-icon-button.mat-warn:not(.mdc-ripple-upgraded):focus::before{background:rgba(0,0,0,0);opacity:1}',FuA],encapsulation:2,changeDetection:0})}return t})();var AC=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,Ps,it]})}return t})();var g8=class{};function I8(t){return t&&typeof t.connect=="function"&&!(t instanceof l2)}var SB=function(t){return t[t.REPLACED=0]="REPLACED",t[t.INSERTED=1]="INSERTED",t[t.MOVED=2]="MOVED",t[t.REMOVED=3]="REMOVED",t}(SB||{}),Ru=new dA("_ViewRepeater"),RB=class{applyChanges(e,A,i,n,o){e.forEachOperation((r,s,a)=>{let c,l;if(r.previousIndex==null){let I=i(r,s,a);c=A.createEmbeddedView(I.templateRef,I.context,I.index),l=SB.INSERTED}else a==null?(A.remove(s),l=SB.REMOVED):(c=A.get(s),A.move(c,a),l=SB.MOVED);o&&o({context:c?.context,operation:l,record:r})})}detach(){}};var K2=class{_multiple;_emitChanges;compareWith;_selection=new Set;_deselectedToEmit=[];_selectedToEmit=[];_selected;get selected(){return this._selected||(this._selected=Array.from(this._selection.values())),this._selected}changed=new OA;constructor(e=!1,A,i=!0,n){this._multiple=e,this._emitChanges=i,this.compareWith=n,A&&A.length&&(e?A.forEach(o=>this._markSelected(o)):this._markSelected(A[0]),this._selectedToEmit.length=0)}select(...e){this._verifyValueAssignment(e),e.forEach(i=>this._markSelected(i));let A=this._hasQueuedChanges();return this._emitChangeEvent(),A}deselect(...e){this._verifyValueAssignment(e),e.forEach(i=>this._unmarkSelected(i));let A=this._hasQueuedChanges();return this._emitChangeEvent(),A}setSelection(...e){this._verifyValueAssignment(e);let A=this.selected,i=new Set(e);e.forEach(o=>this._markSelected(o)),A.filter(o=>!i.has(this._getConcreteValue(o,i))).forEach(o=>this._unmarkSelected(o));let n=this._hasQueuedChanges();return this._emitChangeEvent(),n}toggle(e){return this.isSelected(e)?this.deselect(e):this.select(e)}clear(e=!0){this._unmarkAll();let A=this._hasQueuedChanges();return e&&this._emitChangeEvent(),A}isSelected(e){return this._selection.has(this._getConcreteValue(e))}isEmpty(){return this._selection.size===0}hasValue(){return!this.isEmpty()}sort(e){this._multiple&&this.selected&&this._selected.sort(e)}isMultipleSelection(){return this._multiple}_emitChangeEvent(){this._selected=null,(this._selectedToEmit.length||this._deselectedToEmit.length)&&(this.changed.next({source:this,added:this._selectedToEmit,removed:this._deselectedToEmit}),this._deselectedToEmit=[],this._selectedToEmit=[])}_markSelected(e){e=this._getConcreteValue(e),this.isSelected(e)||(this._multiple||this._unmarkAll(),this.isSelected(e)||this._selection.add(e),this._emitChanges&&this._selectedToEmit.push(e))}_unmarkSelected(e){e=this._getConcreteValue(e),this.isSelected(e)&&(this._selection.delete(e),this._emitChanges&&this._deselectedToEmit.push(e))}_unmarkAll(){this.isEmpty()||this._selection.forEach(e=>this._unmarkSelected(e))}_verifyValueAssignment(e){e.length>1&&this._multiple}_hasQueuedChanges(){return!!(this._deselectedToEmit.length||this._selectedToEmit.length)}_getConcreteValue(e,A){if(this.compareWith){A=A??this._selection;for(let i of A)if(this.compareWith(e,i))return i;return e}else return e}};var xB=(()=>{class t{_listeners=[];notify(A,i){for(let n of this._listeners)n(A,i)}listen(A){return this._listeners.push(A),()=>{this._listeners=this._listeners.filter(i=>A!==i)}}ngOnDestroy(){this._listeners=[]}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var TuA=20,Y2=(()=>{class t{_ngZone=f(Qe);_platform=f(Ii);_renderer=f(ws).createRenderer(null,null);_cleanupGlobalListener;constructor(){}_scrolled=new OA;_scrolledCount=0;scrollContainers=new Map;register(A){this.scrollContainers.has(A)||this.scrollContainers.set(A,A.elementScrolled().subscribe(()=>this._scrolled.next(A)))}deregister(A){let i=this.scrollContainers.get(A);i&&(i.unsubscribe(),this.scrollContainers.delete(A))}scrolled(A=TuA){return this._platform.isBrowser?new ct(i=>{this._cleanupGlobalListener||(this._cleanupGlobalListener=this._ngZone.runOutsideAngular(()=>this._renderer.listen("document","scroll",()=>this._scrolled.next())));let n=A>0?this._scrolled.pipe(yd(A)).subscribe(i):this._scrolled.subscribe(i);return this._scrolledCount++,()=>{n.unsubscribe(),this._scrolledCount--,this._scrolledCount||(this._cleanupGlobalListener?.(),this._cleanupGlobalListener=void 0)}}):Me()}ngOnDestroy(){this._cleanupGlobalListener?.(),this._cleanupGlobalListener=void 0,this.scrollContainers.forEach((A,i)=>this.deregister(i)),this._scrolled.complete()}ancestorScrolled(A,i){let n=this.getAncestorScrollContainers(A);return this.scrolled(i).pipe(kt(o=>!o||n.indexOf(o)>-1))}getAncestorScrollContainers(A){let i=[];return this.scrollContainers.forEach((n,o)=>{this._scrollableContainsElement(o,A)&&i.push(o)}),i}_scrollableContainsElement(A,i){let n=ua(i),o=A.getElementRef().nativeElement;do if(n==o)return!0;while(n=n.parentElement);return!1}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),D0=(()=>{class t{elementRef=f(ee);scrollDispatcher=f(Y2);ngZone=f(Qe);dir=f(mo,{optional:!0});_scrollElement=this.elementRef.nativeElement;_destroyed=new OA;_renderer=f(Wi);_cleanupScroll;_elementScrolled=new OA;constructor(){}ngOnInit(){this._cleanupScroll=this.ngZone.runOutsideAngular(()=>this._renderer.listen(this._scrollElement,"scroll",A=>this._elementScrolled.next(A))),this.scrollDispatcher.register(this)}ngOnDestroy(){this._cleanupScroll?.(),this._elementScrolled.complete(),this.scrollDispatcher.deregister(this),this._destroyed.next(),this._destroyed.complete()}elementScrolled(){return this._elementScrolled}getElementRef(){return this.elementRef}scrollTo(A){let i=this.elementRef.nativeElement,n=this.dir&&this.dir.value=="rtl";A.left==null&&(A.left=n?A.end:A.start),A.right==null&&(A.right=n?A.start:A.end),A.bottom!=null&&(A.top=i.scrollHeight-i.clientHeight-A.bottom),n&&mB()!=gl.NORMAL?(A.left!=null&&(A.right=i.scrollWidth-i.clientWidth-A.left),mB()==gl.INVERTED?A.left=A.right:mB()==gl.NEGATED&&(A.left=A.right?-A.right:A.right)):A.right!=null&&(A.left=i.scrollWidth-i.clientWidth-A.right),this._applyScrollToOptions(A)}_applyScrollToOptions(A){let i=this.elementRef.nativeElement;j6()?i.scrollTo(A):(A.top!=null&&(i.scrollTop=A.top),A.left!=null&&(i.scrollLeft=A.left))}measureScrollOffset(A){let i="left",n="right",o=this.elementRef.nativeElement;if(A=="top")return o.scrollTop;if(A=="bottom")return o.scrollHeight-o.clientHeight-o.scrollTop;let r=this.dir&&this.dir.value=="rtl";return A=="start"?A=r?n:i:A=="end"&&(A=r?i:n),r&&mB()==gl.INVERTED?A==i?o.scrollWidth-o.clientWidth-o.scrollLeft:o.scrollLeft:r&&mB()==gl.NEGATED?A==i?o.scrollLeft+o.scrollWidth-o.clientWidth:-o.scrollLeft:A==i?o.scrollLeft:o.scrollWidth-o.clientWidth-o.scrollLeft}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdk-scrollable",""],["","cdkScrollable",""]]})}return t})(),HuA=20,Mc=(()=>{class t{_platform=f(Ii);_listeners;_viewportSize;_change=new OA;_document=f(at,{optional:!0});constructor(){let A=f(Qe),i=f(ws).createRenderer(null,null);A.runOutsideAngular(()=>{if(this._platform.isBrowser){let n=o=>this._change.next(o);this._listeners=[i.listen("window","resize",n),i.listen("window","orientationchange",n)]}this.change().subscribe(()=>this._viewportSize=null)})}ngOnDestroy(){this._listeners?.forEach(A=>A()),this._change.complete()}getViewportSize(){this._viewportSize||this._updateViewportSize();let A={width:this._viewportSize.width,height:this._viewportSize.height};return this._platform.isBrowser||(this._viewportSize=null),A}getViewportRect(){let A=this.getViewportScrollPosition(),{width:i,height:n}=this.getViewportSize();return{top:A.top,left:A.left,bottom:A.top+n,right:A.left+i,height:n,width:i}}getViewportScrollPosition(){if(!this._platform.isBrowser)return{top:0,left:0};let A=this._document,i=this._getWindow(),n=A.documentElement,o=n.getBoundingClientRect(),r=-o.top||A.body.scrollTop||i.scrollY||n.scrollTop||0,s=-o.left||A.body.scrollLeft||i.scrollX||n.scrollLeft||0;return{top:r,left:s}}change(A=HuA){return A>0?this._change.pipe(yd(A)):this._change}_getWindow(){return this._document.defaultView||window}_updateViewportSize(){let A=this._getWindow();this._viewportSize=this._platform.isBrowser?{width:A.innerWidth,height:A.innerHeight}:{width:0,height:0}}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var Cl=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})(),xu=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[_2,Cl,_2,Cl]})}return t})();var Lu=class{_attachedHost;attach(e){return this._attachedHost=e,e.attach(this)}detach(){let e=this._attachedHost;e!=null&&(this._attachedHost=null,e.detach())}get isAttached(){return this._attachedHost!=null}setAttachedHost(e){this._attachedHost=e}},dl=class extends Lu{component;viewContainerRef;injector;componentFactoryResolver;projectableNodes;constructor(e,A,i,n,o){super(),this.component=e,this.viewContainerRef=A,this.injector=i,this.projectableNodes=o}},ys=class extends Lu{templateRef;viewContainerRef;context;injector;constructor(e,A,i,n){super(),this.templateRef=e,this.viewContainerRef=A,this.context=i,this.injector=n}get origin(){return this.templateRef.elementRef}attach(e,A=this.context){return this.context=A,super.attach(e)}detach(){return this.context=void 0,super.detach()}},Rk=class extends Lu{element;constructor(e){super(),this.element=e instanceof ee?e.nativeElement:e}},J2=class{_attachedPortal;_disposeFn;_isDisposed=!1;hasAttached(){return!!this._attachedPortal}attach(e){if(e instanceof dl)return this._attachedPortal=e,this.attachComponentPortal(e);if(e instanceof ys)return this._attachedPortal=e,this.attachTemplatePortal(e);if(this.attachDomPortal&&e instanceof Rk)return this._attachedPortal=e,this.attachDomPortal(e)}attachDomPortal=null;detach(){this._attachedPortal&&(this._attachedPortal.setAttachedHost(null),this._attachedPortal=null),this._invokeDisposeFn()}dispose(){this.hasAttached()&&this.detach(),this._invokeDisposeFn(),this._isDisposed=!0}setDisposeFn(e){this._disposeFn=e}_invokeDisposeFn(){this._disposeFn&&(this._disposeFn(),this._disposeFn=null)}};var Fu=class extends J2{outletElement;_appRef;_defaultInjector;_document;constructor(e,A,i,n,o){super(),this.outletElement=e,this._appRef=i,this._defaultInjector=n,this._document=o}attachComponentPortal(e){let A;if(e.viewContainerRef){let i=e.injector||e.viewContainerRef.injector,n=i.get(I0,null,{optional:!0})||void 0;A=e.viewContainerRef.createComponent(e.component,{index:e.viewContainerRef.length,injector:i,ngModuleRef:n,projectableNodes:e.projectableNodes||void 0}),this.setDisposeFn(()=>A.destroy())}else A=Gp(e.component,{elementInjector:e.injector||this._defaultInjector||Rt.NULL,environmentInjector:this._appRef.injector,projectableNodes:e.projectableNodes||void 0}),this._appRef.attachView(A.hostView),this.setDisposeFn(()=>{this._appRef.viewCount>0&&this._appRef.detachView(A.hostView),A.destroy()});return this.outletElement.appendChild(this._getComponentRootNode(A)),this._attachedPortal=e,A}attachTemplatePortal(e){let A=e.viewContainerRef,i=A.createEmbeddedView(e.templateRef,e.context,{injector:e.injector});return i.rootNodes.forEach(n=>this.outletElement.appendChild(n)),i.detectChanges(),this.setDisposeFn(()=>{let n=A.indexOf(i);n!==-1&&A.remove(n)}),this._attachedPortal=e,i}attachDomPortal=e=>{let A=e.element;A.parentNode;let i=this._document.createComment("dom-portal");A.parentNode.insertBefore(i,A),this.outletElement.appendChild(A),this._attachedPortal=e,super.setDisposeFn(()=>{i.parentNode&&i.parentNode.replaceChild(A,i)})};dispose(){super.dispose(),this.outletElement.remove()}_getComponentRootNode(e){return e.hostView.rootNodes[0]}};var sP=(()=>{class t extends ys{constructor(){let A=f(wn),i=f(Nn);super(A,i)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkPortal",""]],exportAs:["cdkPortal"],features:[lt]})}return t})();var fa=(()=>{class t extends J2{_moduleRef=f(I0,{optional:!0});_document=f(at);_viewContainerRef=f(Nn);_isInitialized=!1;_attachedRef;constructor(){super()}get portal(){return this._attachedPortal}set portal(A){this.hasAttached()&&!A&&!this._isInitialized||(this.hasAttached()&&super.detach(),A&&super.attach(A),this._attachedPortal=A||null)}attached=new WA;get attachedRef(){return this._attachedRef}ngOnInit(){this._isInitialized=!0}ngOnDestroy(){super.dispose(),this._attachedRef=this._attachedPortal=null}attachComponentPortal(A){A.setAttachedHost(this);let i=A.viewContainerRef!=null?A.viewContainerRef:this._viewContainerRef,n=i.createComponent(A.component,{index:i.length,injector:A.injector||i.injector,projectableNodes:A.projectableNodes||void 0,ngModuleRef:this._moduleRef||void 0});return i!==this._viewContainerRef&&this._getRootNode().appendChild(n.hostView.rootNodes[0]),super.setDisposeFn(()=>n.destroy()),this._attachedPortal=A,this._attachedRef=n,this.attached.emit(n),n}attachTemplatePortal(A){A.setAttachedHost(this);let i=this._viewContainerRef.createEmbeddedView(A.templateRef,A.context,{injector:A.injector});return super.setDisposeFn(()=>this._viewContainerRef.clear()),this._attachedPortal=A,this._attachedRef=i,this.attached.emit(i),i}attachDomPortal=A=>{let i=A.element;i.parentNode;let n=this._document.createComment("dom-portal");A.setAttachedHost(this),i.parentNode.insertBefore(n,i),this._getRootNode().appendChild(i),this._attachedPortal=A,super.setDisposeFn(()=>{n.parentNode&&n.parentNode.replaceChild(i,n)})};_getRootNode(){let A=this._viewContainerRef.element.nativeElement;return A.nodeType===A.ELEMENT_NODE?A:A.parentNode}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkPortalOutlet",""]],inputs:{portal:[0,"cdkPortalOutlet","portal"]},outputs:{attached:"attached"},exportAs:["cdkPortalOutlet"],features:[lt]})}return t})();var dg=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();var aP=j6(),xk=class{_viewportRuler;_previousHTMLStyles={top:"",left:""};_previousScrollPosition;_isEnabled=!1;_document;constructor(e,A){this._viewportRuler=e,this._document=A}attach(){}enable(){if(this._canBeEnabled()){let e=this._document.documentElement;this._previousScrollPosition=this._viewportRuler.getViewportScrollPosition(),this._previousHTMLStyles.left=e.style.left||"",this._previousHTMLStyles.top=e.style.top||"",e.style.left=Cr(-this._previousScrollPosition.left),e.style.top=Cr(-this._previousScrollPosition.top),e.classList.add("cdk-global-scrollblock"),this._isEnabled=!0}}disable(){if(this._isEnabled){let e=this._document.documentElement,A=this._document.body,i=e.style,n=A.style,o=i.scrollBehavior||"",r=n.scrollBehavior||"";this._isEnabled=!1,i.left=this._previousHTMLStyles.left,i.top=this._previousHTMLStyles.top,e.classList.remove("cdk-global-scrollblock"),aP&&(i.scrollBehavior=n.scrollBehavior="auto"),window.scroll(this._previousScrollPosition.left,this._previousScrollPosition.top),aP&&(i.scrollBehavior=o,n.scrollBehavior=r)}}_canBeEnabled(){if(this._document.documentElement.classList.contains("cdk-global-scrollblock")||this._isEnabled)return!1;let A=this._document.body,i=this._viewportRuler.getViewportSize();return A.scrollHeight>i.height||A.scrollWidth>i.width}};var Lk=class{_scrollDispatcher;_ngZone;_viewportRuler;_config;_scrollSubscription=null;_overlayRef;_initialScrollPosition;constructor(e,A,i,n){this._scrollDispatcher=e,this._ngZone=A,this._viewportRuler=i,this._config=n}attach(e){this._overlayRef,this._overlayRef=e}enable(){if(this._scrollSubscription)return;let e=this._scrollDispatcher.scrolled(0).pipe(kt(A=>!A||!this._overlayRef.overlayElement.contains(A.getElementRef().nativeElement)));this._config&&this._config.threshold&&this._config.threshold>1?(this._initialScrollPosition=this._viewportRuler.getViewportScrollPosition().top,this._scrollSubscription=e.subscribe(()=>{let A=this._viewportRuler.getViewportScrollPosition().top;Math.abs(A-this._initialScrollPosition)>this._config.threshold?this._detach():this._overlayRef.updatePosition()})):this._scrollSubscription=e.subscribe(this._detach)}disable(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)}detach(){this.disable(),this._overlayRef=null}_detach=()=>{this.disable(),this._overlayRef.hasAttached()&&this._ngZone.run(()=>this._overlayRef.detach())}},C8=class{enable(){}disable(){}attach(){}};function Fk(t,e){return e.some(A=>{let i=t.bottom<A.top,n=t.top>A.bottom,o=t.right<A.left,r=t.left>A.right;return i||n||o||r})}function cP(t,e){return e.some(A=>{let i=t.top<A.top,n=t.bottom>A.bottom,o=t.left<A.left,r=t.right>A.right;return i||n||o||r})}var Nk=class{_scrollDispatcher;_viewportRuler;_ngZone;_config;_scrollSubscription=null;_overlayRef;constructor(e,A,i,n){this._scrollDispatcher=e,this._viewportRuler=A,this._ngZone=i,this._config=n}attach(e){this._overlayRef,this._overlayRef=e}enable(){if(!this._scrollSubscription){let e=this._config?this._config.scrollThrottle:0;this._scrollSubscription=this._scrollDispatcher.scrolled(e).subscribe(()=>{if(this._overlayRef.updatePosition(),this._config&&this._config.autoClose){let A=this._overlayRef.overlayElement.getBoundingClientRect(),{width:i,height:n}=this._viewportRuler.getViewportSize();Fk(A,[{width:i,height:n,bottom:n,right:i,top:0,left:0}])&&(this.disable(),this._ngZone.run(()=>this._overlayRef.detach()))}})}}disable(){this._scrollSubscription&&(this._scrollSubscription.unsubscribe(),this._scrollSubscription=null)}detach(){this.disable(),this._overlayRef=null}},OuA=(()=>{class t{_scrollDispatcher=f(Y2);_viewportRuler=f(Mc);_ngZone=f(Qe);_document=f(at);constructor(){}noop=()=>new C8;close=A=>new Lk(this._scrollDispatcher,this._ngZone,this._viewportRuler,A);block=()=>new xk(this._viewportRuler,this._document);reposition=A=>new Nk(this._scrollDispatcher,this._viewportRuler,this._ngZone,A);static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),Bg=class{positionStrategy;scrollStrategy=new C8;panelClass="";hasBackdrop=!1;backdropClass="cdk-overlay-dark-backdrop";width;height;minWidth;minHeight;maxWidth;maxHeight;direction;disposeOnNavigation=!1;constructor(e){if(e){let A=Object.keys(e);for(let i of A)e[i]!==void 0&&(this[i]=e[i])}}};var _k=class{connectionPair;scrollableViewProperties;constructor(e,A){this.connectionPair=e,this.scrollableViewProperties=A}};var BP=(()=>{class t{_attachedOverlays=[];_document=f(at);_isAttached;constructor(){}ngOnDestroy(){this.detach()}add(A){this.remove(A),this._attachedOverlays.push(A)}remove(A){let i=this._attachedOverlays.indexOf(A);i>-1&&this._attachedOverlays.splice(i,1),this._attachedOverlays.length===0&&this.detach()}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),PuA=(()=>{class t extends BP{_ngZone=f(Qe);_renderer=f(ws).createRenderer(null,null);_cleanupKeydown;add(A){super.add(A),this._isAttached||(this._ngZone.runOutsideAngular(()=>{this._cleanupKeydown=this._renderer.listen("body","keydown",this._keydownListener)}),this._isAttached=!0)}detach(){this._isAttached&&(this._cleanupKeydown?.(),this._isAttached=!1)}_keydownListener=A=>{let i=this._attachedOverlays;for(let n=i.length-1;n>-1;n--)if(i[n]._keydownEvents.observers.length>0){this._ngZone.run(()=>i[n]._keydownEvents.next(A));break}};static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),juA=(()=>{class t extends BP{_platform=f(Ii);_ngZone=f(Qe,{optional:!0});_cursorOriginalValue;_cursorStyleIsSet=!1;_pointerDownEventTarget;add(A){if(super.add(A),!this._isAttached){let i=this._document.body;this._ngZone?this._ngZone.runOutsideAngular(()=>this._addEventListeners(i)):this._addEventListeners(i),this._platform.IOS&&!this._cursorStyleIsSet&&(this._cursorOriginalValue=i.style.cursor,i.style.cursor="pointer",this._cursorStyleIsSet=!0),this._isAttached=!0}}detach(){if(this._isAttached){let A=this._document.body;A.removeEventListener("pointerdown",this._pointerDownListener,!0),A.removeEventListener("click",this._clickListener,!0),A.removeEventListener("auxclick",this._clickListener,!0),A.removeEventListener("contextmenu",this._clickListener,!0),this._platform.IOS&&this._cursorStyleIsSet&&(A.style.cursor=this._cursorOriginalValue,this._cursorStyleIsSet=!1),this._isAttached=!1}}_addEventListeners(A){A.addEventListener("pointerdown",this._pointerDownListener,!0),A.addEventListener("click",this._clickListener,!0),A.addEventListener("auxclick",this._clickListener,!0),A.addEventListener("contextmenu",this._clickListener,!0)}_pointerDownListener=A=>{this._pointerDownEventTarget=oc(A)};_clickListener=A=>{let i=oc(A),n=A.type==="click"&&this._pointerDownEventTarget?this._pointerDownEventTarget:i;this._pointerDownEventTarget=null;let o=this._attachedOverlays.slice();for(let r=o.length-1;r>-1;r--){let s=o[r];if(s._outsidePointerEvents.observers.length<1||!s.hasAttached())continue;if(lP(s.overlayElement,i)||lP(s.overlayElement,n))break;let a=s._outsidePointerEvents;this._ngZone?this._ngZone.run(()=>a.next(A)):a.next(A)}};static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function lP(t,e){let A=typeof ShadowRoot<"u"&&ShadowRoot,i=e;for(;i;){if(i===t)return!0;i=A&&i instanceof ShadowRoot?i.host:i.parentNode}return!1}var EP=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["ng-component"]],hostAttrs:["cdk-overlay-style-loader",""],decls:0,vars:0,template:function(i,n){},styles:[".cdk-overlay-container,.cdk-global-overlay-wrapper{pointer-events:none;top:0;left:0;height:100%;width:100%}.cdk-overlay-container{position:fixed}@layer cdk-overlay{.cdk-overlay-container{z-index:1000}}.cdk-overlay-container:empty{display:none}.cdk-global-overlay-wrapper{display:flex;position:absolute}@layer cdk-overlay{.cdk-global-overlay-wrapper{z-index:1000}}.cdk-overlay-pane{position:absolute;pointer-events:auto;box-sizing:border-box;display:flex;max-width:100%;max-height:100%}@layer cdk-overlay{.cdk-overlay-pane{z-index:1000}}.cdk-overlay-backdrop{position:absolute;top:0;bottom:0;left:0;right:0;pointer-events:auto;-webkit-tap-highlight-color:rgba(0,0,0,0);opacity:0}@layer cdk-overlay{.cdk-overlay-backdrop{z-index:1000;transition:opacity 400ms cubic-bezier(0.25, 0.8, 0.25, 1)}}.cdk-overlay-backdrop-showing{opacity:1}@media(forced-colors: active){.cdk-overlay-backdrop-showing{opacity:.6}}@layer cdk-overlay{.cdk-overlay-dark-backdrop{background:rgba(0,0,0,.32)}}.cdk-overlay-transparent-backdrop{transition:visibility 1ms linear,opacity 1ms linear;visibility:hidden;opacity:1}.cdk-overlay-transparent-backdrop.cdk-overlay-backdrop-showing,.cdk-high-contrast-active .cdk-overlay-transparent-backdrop{opacity:0;visibility:visible}.cdk-overlay-backdrop-noop-animation{transition:none}.cdk-overlay-connected-position-bounding-box{position:absolute;display:flex;flex-direction:column;min-width:1px;min-height:1px}@layer cdk-overlay{.cdk-overlay-connected-position-bounding-box{z-index:1000}}.cdk-global-scrollblock{position:fixed;width:100%;overflow-y:scroll}"],encapsulation:2,changeDetection:0})}return t})(),d8=(()=>{class t{_platform=f(Ii);_containerElement;_document=f(at);_styleLoader=f(Rn);constructor(){}ngOnDestroy(){this._containerElement?.remove()}getContainerElement(){return this._loadStyles(),this._containerElement||this._createContainer(),this._containerElement}_createContainer(){let A="cdk-overlay-container";if(this._platform.isBrowser||rk()){let n=this._document.querySelectorAll(`.${A}[platform="server"], .${A}[platform="test"]`);for(let o=0;o<n.length;o++)n[o].remove()}let i=this._document.createElement("div");i.classList.add(A),rk()?i.setAttribute("platform","test"):this._platform.isBrowser||i.setAttribute("platform","server"),this._document.body.appendChild(i),this._containerElement=i}_loadStyles(){this._styleLoader.load(EP)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),Gk=class{_renderer;_ngZone;element;_cleanupClick;_cleanupTransitionEnd;_fallbackTimeout;constructor(e,A,i,n){this._renderer=A,this._ngZone=i,this.element=e.createElement("div"),this.element.classList.add("cdk-overlay-backdrop"),this._cleanupClick=A.listen(this.element,"click",n)}detach(){this._ngZone.runOutsideAngular(()=>{let e=this.element;clearTimeout(this._fallbackTimeout),this._cleanupTransitionEnd?.(),this._cleanupTransitionEnd=this._renderer.listen(e,"transitionend",this.dispose),this._fallbackTimeout=setTimeout(this.dispose,500),e.style.pointerEvents="none",e.classList.remove("cdk-overlay-backdrop-showing")})}dispose=()=>{clearTimeout(this._fallbackTimeout),this._cleanupClick?.(),this._cleanupTransitionEnd?.(),this._cleanupClick=this._cleanupTransitionEnd=this._fallbackTimeout=void 0,this.element.remove()}},LB=class{_portalOutlet;_host;_pane;_config;_ngZone;_keyboardDispatcher;_document;_location;_outsideClickDispatcher;_animationsDisabled;_injector;_renderer;_backdropClick=new OA;_attachments=new OA;_detachments=new OA;_positionStrategy;_scrollStrategy;_locationChanges=Kt.EMPTY;_backdropRef=null;_previousHostParent;_keydownEvents=new OA;_outsidePointerEvents=new OA;_renders=new OA;_afterRenderRef;_afterNextRenderRef;constructor(e,A,i,n,o,r,s,a,c,l=!1,I,C){this._portalOutlet=e,this._host=A,this._pane=i,this._config=n,this._ngZone=o,this._keyboardDispatcher=r,this._document=s,this._location=a,this._outsideClickDispatcher=c,this._animationsDisabled=l,this._injector=I,this._renderer=C,n.scrollStrategy&&(this._scrollStrategy=n.scrollStrategy,this._scrollStrategy.attach(this)),this._positionStrategy=n.positionStrategy,this._afterRenderRef=Ba(()=>vh(()=>{this._renders.next()},{injector:this._injector}))}get overlayElement(){return this._pane}get backdropElement(){return this._backdropRef?.element||null}get hostElement(){return this._host}attach(e){!this._host.parentElement&&this._previousHostParent&&this._previousHostParent.appendChild(this._host);let A=this._portalOutlet.attach(e);return this._positionStrategy&&this._positionStrategy.attach(this),this._updateStackingOrder(),this._updateElementSize(),this._updateElementDirection(),this._scrollStrategy&&this._scrollStrategy.enable(),this._afterNextRenderRef?.destroy(),this._afterNextRenderRef=To(()=>{this.hasAttached()&&this.updatePosition()},{injector:this._injector}),this._togglePointerEvents(!0),this._config.hasBackdrop&&this._attachBackdrop(),this._config.panelClass&&this._toggleClasses(this._pane,this._config.panelClass,!0),this._attachments.next(),this._keyboardDispatcher.add(this),this._config.disposeOnNavigation&&(this._locationChanges=this._location.subscribe(()=>this.dispose())),this._outsideClickDispatcher.add(this),typeof A?.onDestroy=="function"&&A.onDestroy(()=>{this.hasAttached()&&this._ngZone.runOutsideAngular(()=>Promise.resolve().then(()=>this.detach()))}),A}detach(){if(!this.hasAttached())return;this.detachBackdrop(),this._togglePointerEvents(!1),this._positionStrategy&&this._positionStrategy.detach&&this._positionStrategy.detach(),this._scrollStrategy&&this._scrollStrategy.disable();let e=this._portalOutlet.detach();return this._detachments.next(),this._keyboardDispatcher.remove(this),this._detachContentWhenEmpty(),this._locationChanges.unsubscribe(),this._outsideClickDispatcher.remove(this),e}dispose(){let e=this.hasAttached();this._positionStrategy&&this._positionStrategy.dispose(),this._disposeScrollStrategy(),this._backdropRef?.dispose(),this._locationChanges.unsubscribe(),this._keyboardDispatcher.remove(this),this._portalOutlet.dispose(),this._attachments.complete(),this._backdropClick.complete(),this._keydownEvents.complete(),this._outsidePointerEvents.complete(),this._outsideClickDispatcher.remove(this),this._host?.remove(),this._afterNextRenderRef?.destroy(),this._previousHostParent=this._pane=this._host=this._backdropRef=null,e&&this._detachments.next(),this._detachments.complete(),this._afterRenderRef.destroy(),this._renders.complete()}hasAttached(){return this._portalOutlet.hasAttached()}backdropClick(){return this._backdropClick}attachments(){return this._attachments}detachments(){return this._detachments}keydownEvents(){return this._keydownEvents}outsidePointerEvents(){return this._outsidePointerEvents}getConfig(){return this._config}updatePosition(){this._positionStrategy&&this._positionStrategy.apply()}updatePositionStrategy(e){e!==this._positionStrategy&&(this._positionStrategy&&this._positionStrategy.dispose(),this._positionStrategy=e,this.hasAttached()&&(e.attach(this),this.updatePosition()))}updateSize(e){this._config=rA(rA({},this._config),e),this._updateElementSize()}setDirection(e){this._config=Ye(rA({},this._config),{direction:e}),this._updateElementDirection()}addPanelClass(e){this._pane&&this._toggleClasses(this._pane,e,!0)}removePanelClass(e){this._pane&&this._toggleClasses(this._pane,e,!1)}getDirection(){let e=this._config.direction;return e?typeof e=="string"?e:e.value:"ltr"}updateScrollStrategy(e){e!==this._scrollStrategy&&(this._disposeScrollStrategy(),this._scrollStrategy=e,this.hasAttached()&&(e.attach(this),e.enable()))}_updateElementDirection(){this._host.setAttribute("dir",this.getDirection())}_updateElementSize(){if(!this._pane)return;let e=this._pane.style;e.width=Cr(this._config.width),e.height=Cr(this._config.height),e.minWidth=Cr(this._config.minWidth),e.minHeight=Cr(this._config.minHeight),e.maxWidth=Cr(this._config.maxWidth),e.maxHeight=Cr(this._config.maxHeight)}_togglePointerEvents(e){this._pane.style.pointerEvents=e?"":"none"}_attachBackdrop(){let e="cdk-overlay-backdrop-showing";this._backdropRef?.dispose(),this._backdropRef=new Gk(this._document,this._renderer,this._ngZone,A=>{this._backdropClick.next(A)}),this._animationsDisabled&&this._backdropRef.element.classList.add("cdk-overlay-backdrop-noop-animation"),this._config.backdropClass&&this._toggleClasses(this._backdropRef.element,this._config.backdropClass,!0),this._host.parentElement.insertBefore(this._backdropRef.element,this._host),!this._animationsDisabled&&typeof requestAnimationFrame<"u"?this._ngZone.runOutsideAngular(()=>{requestAnimationFrame(()=>this._backdropRef?.element.classList.add(e))}):this._backdropRef.element.classList.add(e)}_updateStackingOrder(){this._host.nextSibling&&this._host.parentNode.appendChild(this._host)}detachBackdrop(){this._animationsDisabled?(this._backdropRef?.dispose(),this._backdropRef=null):this._backdropRef?.detach()}_toggleClasses(e,A,i){let n=wB(A||[]).filter(o=>!!o);n.length&&(i?e.classList.add(...n):e.classList.remove(...n))}_detachContentWhenEmpty(){this._ngZone.runOutsideAngular(()=>{let e=this._renders.pipe(St(zn(this._attachments,this._detachments))).subscribe(()=>{(!this._pane||!this._host||this._pane.children.length===0)&&(this._pane&&this._config.panelClass&&this._toggleClasses(this._pane,this._config.panelClass,!1),this._host&&this._host.parentElement&&(this._previousHostParent=this._host.parentElement,this._host.remove()),e.unsubscribe())})})}_disposeScrollStrategy(){let e=this._scrollStrategy;e?.disable(),e?.detach?.()}},gP="cdk-overlay-connected-position-bounding-box",quA=/([A-Za-z%]+)$/,Uk=class{_viewportRuler;_document;_platform;_overlayContainer;_overlayRef;_isInitialRender;_lastBoundingBoxSize={width:0,height:0};_isPushed=!1;_canPush=!0;_growAfterOpen=!1;_hasFlexibleDimensions=!0;_positionLocked=!1;_originRect;_overlayRect;_viewportRect;_containerRect;_viewportMargin=0;_scrollables=[];_preferredPositions=[];_origin;_pane;_isDisposed;_boundingBox;_lastPosition;_lastScrollVisibility;_positionChanges=new OA;_resizeSubscription=Kt.EMPTY;_offsetX=0;_offsetY=0;_transformOriginSelector;_appliedPanelClasses=[];_previousPushAmount;positionChanges=this._positionChanges;get positions(){return this._preferredPositions}constructor(e,A,i,n,o){this._viewportRuler=A,this._document=i,this._platform=n,this._overlayContainer=o,this.setOrigin(e)}attach(e){this._overlayRef&&this._overlayRef,this._validatePositions(),e.hostElement.classList.add(gP),this._overlayRef=e,this._boundingBox=e.hostElement,this._pane=e.overlayElement,this._isDisposed=!1,this._isInitialRender=!0,this._lastPosition=null,this._resizeSubscription.unsubscribe(),this._resizeSubscription=this._viewportRuler.change().subscribe(()=>{this._isInitialRender=!0,this.apply()})}apply(){if(this._isDisposed||!this._platform.isBrowser)return;if(!this._isInitialRender&&this._positionLocked&&this._lastPosition){this.reapplyLastPosition();return}this._clearPanelClasses(),this._resetOverlayElementStyles(),this._resetBoundingBoxStyles(),this._viewportRect=this._getNarrowedViewportRect(),this._originRect=this._getOriginRect(),this._overlayRect=this._pane.getBoundingClientRect(),this._containerRect=this._overlayContainer.getContainerElement().getBoundingClientRect();let e=this._originRect,A=this._overlayRect,i=this._viewportRect,n=this._containerRect,o=[],r;for(let s of this._preferredPositions){let a=this._getOriginPoint(e,n,s),c=this._getOverlayPoint(a,A,s),l=this._getOverlayFit(c,A,i,s);if(l.isCompletelyWithinViewport){this._isPushed=!1,this._applyPosition(s,a);return}if(this._canFitWithFlexibleDimensions(l,c,i)){o.push({position:s,origin:a,overlayRect:A,boundingBoxRect:this._calculateBoundingBoxRect(a,s)});continue}(!r||r.overlayFit.visibleArea<l.visibleArea)&&(r={overlayFit:l,overlayPoint:c,originPoint:a,position:s,overlayRect:A})}if(o.length){let s=null,a=-1;for(let c of o){let l=c.boundingBoxRect.width*c.boundingBoxRect.height*(c.position.weight||1);l>a&&(a=l,s=c)}this._isPushed=!1,this._applyPosition(s.position,s.origin);return}if(this._canPush){this._isPushed=!0,this._applyPosition(r.position,r.originPoint);return}this._applyPosition(r.position,r.originPoint)}detach(){this._clearPanelClasses(),this._lastPosition=null,this._previousPushAmount=null,this._resizeSubscription.unsubscribe()}dispose(){this._isDisposed||(this._boundingBox&&eC(this._boundingBox.style,{top:"",left:"",right:"",bottom:"",height:"",width:"",alignItems:"",justifyContent:""}),this._pane&&this._resetOverlayElementStyles(),this._overlayRef&&this._overlayRef.hostElement.classList.remove(gP),this.detach(),this._positionChanges.complete(),this._overlayRef=this._boundingBox=null,this._isDisposed=!0)}reapplyLastPosition(){if(this._isDisposed||!this._platform.isBrowser)return;let e=this._lastPosition;if(e){this._originRect=this._getOriginRect(),this._overlayRect=this._pane.getBoundingClientRect(),this._viewportRect=this._getNarrowedViewportRect(),this._containerRect=this._overlayContainer.getContainerElement().getBoundingClientRect();let A=this._getOriginPoint(this._originRect,this._containerRect,e);this._applyPosition(e,A)}else this.apply()}withScrollableContainers(e){return this._scrollables=e,this}withPositions(e){return this._preferredPositions=e,e.indexOf(this._lastPosition)===-1&&(this._lastPosition=null),this._validatePositions(),this}withViewportMargin(e){return this._viewportMargin=e,this}withFlexibleDimensions(e=!0){return this._hasFlexibleDimensions=e,this}withGrowAfterOpen(e=!0){return this._growAfterOpen=e,this}withPush(e=!0){return this._canPush=e,this}withLockedPosition(e=!0){return this._positionLocked=e,this}setOrigin(e){return this._origin=e,this}withDefaultOffsetX(e){return this._offsetX=e,this}withDefaultOffsetY(e){return this._offsetY=e,this}withTransformOriginOn(e){return this._transformOriginSelector=e,this}_getOriginPoint(e,A,i){let n;if(i.originX=="center")n=e.left+e.width/2;else{let r=this._isRtl()?e.right:e.left,s=this._isRtl()?e.left:e.right;n=i.originX=="start"?r:s}A.left<0&&(n-=A.left);let o;return i.originY=="center"?o=e.top+e.height/2:o=i.originY=="top"?e.top:e.bottom,A.top<0&&(o-=A.top),{x:n,y:o}}_getOverlayPoint(e,A,i){let n;i.overlayX=="center"?n=-A.width/2:i.overlayX==="start"?n=this._isRtl()?-A.width:0:n=this._isRtl()?0:-A.width;let o;return i.overlayY=="center"?o=-A.height/2:o=i.overlayY=="top"?0:-A.height,{x:e.x+n,y:e.y+o}}_getOverlayFit(e,A,i,n){let o=CP(A),{x:r,y:s}=e,a=this._getOffset(n,"x"),c=this._getOffset(n,"y");a&&(r+=a),c&&(s+=c);let l=0-r,I=r+o.width-i.width,C=0-s,d=s+o.height-i.height,B=this._subtractOverflows(o.width,l,I),E=this._subtractOverflows(o.height,C,d),h=B*E;return{visibleArea:h,isCompletelyWithinViewport:o.width*o.height===h,fitsInViewportVertically:E===o.height,fitsInViewportHorizontally:B==o.width}}_canFitWithFlexibleDimensions(e,A,i){if(this._hasFlexibleDimensions){let n=i.bottom-A.y,o=i.right-A.x,r=IP(this._overlayRef.getConfig().minHeight),s=IP(this._overlayRef.getConfig().minWidth),a=e.fitsInViewportVertically||r!=null&&r<=n,c=e.fitsInViewportHorizontally||s!=null&&s<=o;return a&&c}return!1}_pushOverlayOnScreen(e,A,i){if(this._previousPushAmount&&this._positionLocked)return{x:e.x+this._previousPushAmount.x,y:e.y+this._previousPushAmount.y};let n=CP(A),o=this._viewportRect,r=Math.max(e.x+n.width-o.width,0),s=Math.max(e.y+n.height-o.height,0),a=Math.max(o.top-i.top-e.y,0),c=Math.max(o.left-i.left-e.x,0),l=0,I=0;return n.width<=o.width?l=c||-r:l=e.x<this._viewportMargin?o.left-i.left-e.x:0,n.height<=o.height?I=a||-s:I=e.y<this._viewportMargin?o.top-i.top-e.y:0,this._previousPushAmount={x:l,y:I},{x:e.x+l,y:e.y+I}}_applyPosition(e,A){if(this._setTransformOrigin(e),this._setOverlayElementStyles(A,e),this._setBoundingBoxStyles(A,e),e.panelClass&&this._addPanelClasses(e.panelClass),this._positionChanges.observers.length){let i=this._getScrollVisibility();if(e!==this._lastPosition||!this._lastScrollVisibility||!VuA(this._lastScrollVisibility,i)){let n=new _k(e,i);this._positionChanges.next(n)}this._lastScrollVisibility=i}this._lastPosition=e,this._isInitialRender=!1}_setTransformOrigin(e){if(!this._transformOriginSelector)return;let A=this._boundingBox.querySelectorAll(this._transformOriginSelector),i,n=e.overlayY;e.overlayX==="center"?i="center":this._isRtl()?i=e.overlayX==="start"?"right":"left":i=e.overlayX==="start"?"left":"right";for(let o=0;o<A.length;o++)A[o].style.transformOrigin=`${i} ${n}`}_calculateBoundingBoxRect(e,A){let i=this._viewportRect,n=this._isRtl(),o,r,s;if(A.overlayY==="top")r=e.y,o=i.height-r+this._viewportMargin;else if(A.overlayY==="bottom")s=i.height-e.y+this._viewportMargin*2,o=i.height-s+this._viewportMargin;else{let d=Math.min(i.bottom-e.y+i.top,e.y),B=this._lastBoundingBoxSize.height;o=d*2,r=e.y-d,o>B&&!this._isInitialRender&&!this._growAfterOpen&&(r=e.y-B/2)}let a=A.overlayX==="start"&&!n||A.overlayX==="end"&&n,c=A.overlayX==="end"&&!n||A.overlayX==="start"&&n,l,I,C;if(c)C=i.width-e.x+this._viewportMargin*2,l=e.x-this._viewportMargin;else if(a)I=e.x,l=i.right-e.x;else{let d=Math.min(i.right-e.x+i.left,e.x),B=this._lastBoundingBoxSize.width;l=d*2,I=e.x-d,l>B&&!this._isInitialRender&&!this._growAfterOpen&&(I=e.x-B/2)}return{top:r,left:I,bottom:s,right:C,width:l,height:o}}_setBoundingBoxStyles(e,A){let i=this._calculateBoundingBoxRect(e,A);!this._isInitialRender&&!this._growAfterOpen&&(i.height=Math.min(i.height,this._lastBoundingBoxSize.height),i.width=Math.min(i.width,this._lastBoundingBoxSize.width));let n={};if(this._hasExactPosition())n.top=n.left="0",n.bottom=n.right=n.maxHeight=n.maxWidth="",n.width=n.height="100%";else{let o=this._overlayRef.getConfig().maxHeight,r=this._overlayRef.getConfig().maxWidth;n.height=Cr(i.height),n.top=Cr(i.top),n.bottom=Cr(i.bottom),n.width=Cr(i.width),n.left=Cr(i.left),n.right=Cr(i.right),A.overlayX==="center"?n.alignItems="center":n.alignItems=A.overlayX==="end"?"flex-end":"flex-start",A.overlayY==="center"?n.justifyContent="center":n.justifyContent=A.overlayY==="bottom"?"flex-end":"flex-start",o&&(n.maxHeight=Cr(o)),r&&(n.maxWidth=Cr(r))}this._lastBoundingBoxSize=i,eC(this._boundingBox.style,n)}_resetBoundingBoxStyles(){eC(this._boundingBox.style,{top:"0",left:"0",right:"0",bottom:"0",height:"",width:"",alignItems:"",justifyContent:""})}_resetOverlayElementStyles(){eC(this._pane.style,{top:"",left:"",bottom:"",right:"",position:"",transform:""})}_setOverlayElementStyles(e,A){let i={},n=this._hasExactPosition(),o=this._hasFlexibleDimensions,r=this._overlayRef.getConfig();if(n){let l=this._viewportRuler.getViewportScrollPosition();eC(i,this._getExactOverlayY(A,e,l)),eC(i,this._getExactOverlayX(A,e,l))}else i.position="static";let s="",a=this._getOffset(A,"x"),c=this._getOffset(A,"y");a&&(s+=`translateX(${a}px) `),c&&(s+=`translateY(${c}px)`),i.transform=s.trim(),r.maxHeight&&(n?i.maxHeight=Cr(r.maxHeight):o&&(i.maxHeight="")),r.maxWidth&&(n?i.maxWidth=Cr(r.maxWidth):o&&(i.maxWidth="")),eC(this._pane.style,i)}_getExactOverlayY(e,A,i){let n={top:"",bottom:""},o=this._getOverlayPoint(A,this._overlayRect,e);if(this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,i)),e.overlayY==="bottom"){let r=this._document.documentElement.clientHeight;n.bottom=`${r-(o.y+this._overlayRect.height)}px`}else n.top=Cr(o.y);return n}_getExactOverlayX(e,A,i){let n={left:"",right:""},o=this._getOverlayPoint(A,this._overlayRect,e);this._isPushed&&(o=this._pushOverlayOnScreen(o,this._overlayRect,i));let r;if(this._isRtl()?r=e.overlayX==="end"?"left":"right":r=e.overlayX==="end"?"right":"left",r==="right"){let s=this._document.documentElement.clientWidth;n.right=`${s-(o.x+this._overlayRect.width)}px`}else n.left=Cr(o.x);return n}_getScrollVisibility(){let e=this._getOriginRect(),A=this._pane.getBoundingClientRect(),i=this._scrollables.map(n=>n.getElementRef().nativeElement.getBoundingClientRect());return{isOriginClipped:cP(e,i),isOriginOutsideView:Fk(e,i),isOverlayClipped:cP(A,i),isOverlayOutsideView:Fk(A,i)}}_subtractOverflows(e,...A){return A.reduce((i,n)=>i-Math.max(n,0),e)}_getNarrowedViewportRect(){let e=this._document.documentElement.clientWidth,A=this._document.documentElement.clientHeight,i=this._viewportRuler.getViewportScrollPosition();return{top:i.top+this._viewportMargin,left:i.left+this._viewportMargin,right:i.left+e-this._viewportMargin,bottom:i.top+A-this._viewportMargin,width:e-2*this._viewportMargin,height:A-2*this._viewportMargin}}_isRtl(){return this._overlayRef.getDirection()==="rtl"}_hasExactPosition(){return!this._hasFlexibleDimensions||this._isPushed}_getOffset(e,A){return A==="x"?e.offsetX==null?this._offsetX:e.offsetX:e.offsetY==null?this._offsetY:e.offsetY}_validatePositions(){}_addPanelClasses(e){this._pane&&wB(e).forEach(A=>{A!==""&&this._appliedPanelClasses.indexOf(A)===-1&&(this._appliedPanelClasses.push(A),this._pane.classList.add(A))})}_clearPanelClasses(){this._pane&&(this._appliedPanelClasses.forEach(e=>{this._pane.classList.remove(e)}),this._appliedPanelClasses=[])}_getOriginRect(){let e=this._origin;if(e instanceof ee)return e.nativeElement.getBoundingClientRect();if(e instanceof Element)return e.getBoundingClientRect();let A=e.width||0,i=e.height||0;return{top:e.y,bottom:e.y+i,left:e.x,right:e.x+A,height:i,width:A}}};function eC(t,e){for(let A in e)e.hasOwnProperty(A)&&(t[A]=e[A]);return t}function IP(t){if(typeof t!="number"&&t!=null){let[e,A]=t.split(quA);return!A||A==="px"?parseFloat(e):null}return t||null}function CP(t){return{top:Math.floor(t.top),right:Math.floor(t.right),bottom:Math.floor(t.bottom),left:Math.floor(t.left),width:Math.floor(t.width),height:Math.floor(t.height)}}function VuA(t,e){return t===e?!0:t.isOriginClipped===e.isOriginClipped&&t.isOriginOutsideView===e.isOriginOutsideView&&t.isOverlayClipped===e.isOverlayClipped&&t.isOverlayOutsideView===e.isOverlayOutsideView}var dP="cdk-global-overlay-wrapper",Kk=class{_overlayRef;_cssPosition="static";_topOffset="";_bottomOffset="";_alignItems="";_xPosition="";_xOffset="";_width="";_height="";_isDisposed=!1;attach(e){let A=e.getConfig();this._overlayRef=e,this._width&&!A.width&&e.updateSize({width:this._width}),this._height&&!A.height&&e.updateSize({height:this._height}),e.hostElement.classList.add(dP),this._isDisposed=!1}top(e=""){return this._bottomOffset="",this._topOffset=e,this._alignItems="flex-start",this}left(e=""){return this._xOffset=e,this._xPosition="left",this}bottom(e=""){return this._topOffset="",this._bottomOffset=e,this._alignItems="flex-end",this}right(e=""){return this._xOffset=e,this._xPosition="right",this}start(e=""){return this._xOffset=e,this._xPosition="start",this}end(e=""){return this._xOffset=e,this._xPosition="end",this}width(e=""){return this._overlayRef?this._overlayRef.updateSize({width:e}):this._width=e,this}height(e=""){return this._overlayRef?this._overlayRef.updateSize({height:e}):this._height=e,this}centerHorizontally(e=""){return this.left(e),this._xPosition="center",this}centerVertically(e=""){return this.top(e),this._alignItems="center",this}apply(){if(!this._overlayRef||!this._overlayRef.hasAttached())return;let e=this._overlayRef.overlayElement.style,A=this._overlayRef.hostElement.style,i=this._overlayRef.getConfig(),{width:n,height:o,maxWidth:r,maxHeight:s}=i,a=(n==="100%"||n==="100vw")&&(!r||r==="100%"||r==="100vw"),c=(o==="100%"||o==="100vh")&&(!s||s==="100%"||s==="100vh"),l=this._xPosition,I=this._xOffset,C=this._overlayRef.getConfig().direction==="rtl",d="",B="",E="";a?E="flex-start":l==="center"?(E="center",C?B=I:d=I):C?l==="left"||l==="end"?(E="flex-end",d=I):(l==="right"||l==="start")&&(E="flex-start",B=I):l==="left"||l==="start"?(E="flex-start",d=I):(l==="right"||l==="end")&&(E="flex-end",B=I),e.position=this._cssPosition,e.marginLeft=a?"0":d,e.marginTop=c?"0":this._topOffset,e.marginBottom=this._bottomOffset,e.marginRight=a?"0":B,A.justifyContent=E,A.alignItems=c?"flex-start":this._alignItems}dispose(){if(this._isDisposed||!this._overlayRef)return;let e=this._overlayRef.overlayElement.style,A=this._overlayRef.hostElement,i=A.style;A.classList.remove(dP),i.justifyContent=i.alignItems=e.marginTop=e.marginBottom=e.marginLeft=e.marginRight=e.position="",this._overlayRef=null,this._isDisposed=!0}},ZuA=(()=>{class t{_viewportRuler=f(Mc);_document=f(at);_platform=f(Ii);_overlayContainer=f(d8);constructor(){}global(){return new Kk}flexibleConnectedTo(A){return new Uk(A,this._viewportRuler,this._document,this._platform,this._overlayContainer)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),nr=(()=>{class t{scrollStrategies=f(OuA);_overlayContainer=f(d8);_positionBuilder=f(ZuA);_keyboardDispatcher=f(PuA);_injector=f(Rt);_ngZone=f(Qe);_document=f(at);_directionality=f(mo);_location=f(Dc);_outsideClickDispatcher=f(juA);_animationsModuleType=f(Si,{optional:!0});_idGenerator=f(sn);_renderer=f(ws).createRenderer(null,null);_appRef;_styleLoader=f(Rn);constructor(){}create(A){this._styleLoader.load(EP);let i=this._createHostElement(),n=this._createPaneElement(i),o=this._createPortalOutlet(n),r=new Bg(A);return r.direction=r.direction||this._directionality.value,new LB(o,i,n,r,this._ngZone,this._keyboardDispatcher,this._document,this._location,this._outsideClickDispatcher,this._animationsModuleType==="NoopAnimations",this._injector.get(pr),this._renderer)}position(){return this._positionBuilder}_createPaneElement(A){let i=this._document.createElement("div");return i.id=this._idGenerator.getId("cdk-overlay-"),i.classList.add("cdk-overlay-pane"),A.appendChild(i),i}_createHostElement(){let A=this._document.createElement("div");return this._overlayContainer.getContainerElement().appendChild(A),A}_createPortalOutlet(A){return this._appRef||(this._appRef=this._injector.get(la)),new Fu(A,null,this._appRef,this._injector,this._document)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),WuA=[{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"}],QP=new dA("cdk-connected-overlay-scroll-strategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.reposition()}}),Nu=(()=>{class t{elementRef=f(ee);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdk-overlay-origin",""],["","overlay-origin",""],["","cdkOverlayOrigin",""]],exportAs:["cdkOverlayOrigin"]})}return t})(),Yk=(()=>{class t{_overlay=f(nr);_dir=f(mo,{optional:!0});_overlayRef;_templatePortal;_backdropSubscription=Kt.EMPTY;_attachSubscription=Kt.EMPTY;_detachSubscription=Kt.EMPTY;_positionSubscription=Kt.EMPTY;_offsetX;_offsetY;_position;_scrollStrategyFactory=f(QP);_disposeOnNavigation=!1;_ngZone=f(Qe);origin;positions;positionStrategy;get offsetX(){return this._offsetX}set offsetX(A){this._offsetX=A,this._position&&this._updatePositionStrategy(this._position)}get offsetY(){return this._offsetY}set offsetY(A){this._offsetY=A,this._position&&this._updatePositionStrategy(this._position)}width;height;minWidth;minHeight;backdropClass;panelClass;viewportMargin=0;scrollStrategy;open=!1;disableClose=!1;transformOriginSelector;hasBackdrop=!1;lockPosition=!1;flexibleDimensions=!1;growAfterOpen=!1;push=!1;get disposeOnNavigation(){return this._disposeOnNavigation}set disposeOnNavigation(A){this._disposeOnNavigation=A}backdropClick=new WA;positionChange=new WA;attach=new WA;detach=new WA;overlayKeydown=new WA;overlayOutsideClick=new WA;constructor(){let A=f(wn),i=f(Nn);this._templatePortal=new ys(A,i),this.scrollStrategy=this._scrollStrategyFactory()}get overlayRef(){return this._overlayRef}get dir(){return this._dir?this._dir.value:"ltr"}ngOnDestroy(){this._attachSubscription.unsubscribe(),this._detachSubscription.unsubscribe(),this._backdropSubscription.unsubscribe(),this._positionSubscription.unsubscribe(),this._overlayRef&&this._overlayRef.dispose()}ngOnChanges(A){this._position&&(this._updatePositionStrategy(this._position),this._overlayRef.updateSize({width:this.width,minWidth:this.minWidth,height:this.height,minHeight:this.minHeight}),A.origin&&this.open&&this._position.apply()),A.open&&(this.open?this._attachOverlay():this._detachOverlay())}_createOverlay(){(!this.positions||!this.positions.length)&&(this.positions=WuA);let A=this._overlayRef=this._overlay.create(this._buildConfig());this._attachSubscription=A.attachments().subscribe(()=>this.attach.emit()),this._detachSubscription=A.detachments().subscribe(()=>this.detach.emit()),A.keydownEvents().subscribe(i=>{this.overlayKeydown.next(i),i.keyCode===27&&!this.disableClose&&!ir(i)&&(i.preventDefault(),this._detachOverlay())}),this._overlayRef.outsidePointerEvents().subscribe(i=>{let n=this._getOriginElement(),o=oc(i);(!n||n!==o&&!n.contains(o))&&this.overlayOutsideClick.next(i)})}_buildConfig(){let A=this._position=this.positionStrategy||this._createPositionStrategy(),i=new Bg({direction:this._dir||"ltr",positionStrategy:A,scrollStrategy:this.scrollStrategy,hasBackdrop:this.hasBackdrop,disposeOnNavigation:this.disposeOnNavigation});return(this.width||this.width===0)&&(i.width=this.width),(this.height||this.height===0)&&(i.height=this.height),(this.minWidth||this.minWidth===0)&&(i.minWidth=this.minWidth),(this.minHeight||this.minHeight===0)&&(i.minHeight=this.minHeight),this.backdropClass&&(i.backdropClass=this.backdropClass),this.panelClass&&(i.panelClass=this.panelClass),i}_updatePositionStrategy(A){let i=this.positions.map(n=>({originX:n.originX,originY:n.originY,overlayX:n.overlayX,overlayY:n.overlayY,offsetX:n.offsetX||this.offsetX,offsetY:n.offsetY||this.offsetY,panelClass:n.panelClass||void 0}));return A.setOrigin(this._getOrigin()).withPositions(i).withFlexibleDimensions(this.flexibleDimensions).withPush(this.push).withGrowAfterOpen(this.growAfterOpen).withViewportMargin(this.viewportMargin).withLockedPosition(this.lockPosition).withTransformOriginOn(this.transformOriginSelector)}_createPositionStrategy(){let A=this._overlay.position().flexibleConnectedTo(this._getOrigin());return this._updatePositionStrategy(A),A}_getOrigin(){return this.origin instanceof Nu?this.origin.elementRef:this.origin}_getOriginElement(){return this.origin instanceof Nu?this.origin.elementRef.nativeElement:this.origin instanceof ee?this.origin.nativeElement:typeof Element<"u"&&this.origin instanceof Element?this.origin:null}_attachOverlay(){this._overlayRef?this._overlayRef.getConfig().hasBackdrop=this.hasBackdrop:this._createOverlay(),this._overlayRef.hasAttached()||this._overlayRef.attach(this._templatePortal),this.hasBackdrop?this._backdropSubscription=this._overlayRef.backdropClick().subscribe(A=>{this.backdropClick.emit(A)}):this._backdropSubscription.unsubscribe(),this._positionSubscription.unsubscribe(),this.positionChange.observers.length>0&&(this._positionSubscription=this._position.positionChanges.pipe(Bv(()=>this.positionChange.observers.length>0)).subscribe(A=>{this._ngZone.run(()=>this.positionChange.emit(A)),this.positionChange.observers.length===0&&this._positionSubscription.unsubscribe()}))}_detachOverlay(){this._overlayRef&&this._overlayRef.detach(),this._backdropSubscription.unsubscribe(),this._positionSubscription.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdk-connected-overlay",""],["","connected-overlay",""],["","cdkConnectedOverlay",""]],inputs:{origin:[0,"cdkConnectedOverlayOrigin","origin"],positions:[0,"cdkConnectedOverlayPositions","positions"],positionStrategy:[0,"cdkConnectedOverlayPositionStrategy","positionStrategy"],offsetX:[0,"cdkConnectedOverlayOffsetX","offsetX"],offsetY:[0,"cdkConnectedOverlayOffsetY","offsetY"],width:[0,"cdkConnectedOverlayWidth","width"],height:[0,"cdkConnectedOverlayHeight","height"],minWidth:[0,"cdkConnectedOverlayMinWidth","minWidth"],minHeight:[0,"cdkConnectedOverlayMinHeight","minHeight"],backdropClass:[0,"cdkConnectedOverlayBackdropClass","backdropClass"],panelClass:[0,"cdkConnectedOverlayPanelClass","panelClass"],viewportMargin:[0,"cdkConnectedOverlayViewportMargin","viewportMargin"],scrollStrategy:[0,"cdkConnectedOverlayScrollStrategy","scrollStrategy"],open:[0,"cdkConnectedOverlayOpen","open"],disableClose:[0,"cdkConnectedOverlayDisableClose","disableClose"],transformOriginSelector:[0,"cdkConnectedOverlayTransformOriginOn","transformOriginSelector"],hasBackdrop:[2,"cdkConnectedOverlayHasBackdrop","hasBackdrop",ie],lockPosition:[2,"cdkConnectedOverlayLockPosition","lockPosition",ie],flexibleDimensions:[2,"cdkConnectedOverlayFlexibleDimensions","flexibleDimensions",ie],growAfterOpen:[2,"cdkConnectedOverlayGrowAfterOpen","growAfterOpen",ie],push:[2,"cdkConnectedOverlayPush","push",ie],disposeOnNavigation:[2,"cdkConnectedOverlayDisposeOnNavigation","disposeOnNavigation",ie]},outputs:{backdropClick:"backdropClick",positionChange:"positionChange",attach:"attach",detach:"detach",overlayKeydown:"overlayKeydown",overlayOutsideClick:"overlayOutsideClick"},exportAs:["cdkConnectedOverlay"],features:[ti]})}return t})();function XuA(t){return()=>t.scrollStrategies.reposition()}var $uA={provide:QP,deps:[nr],useFactory:XuA},El=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[nr,$uA],imports:[_2,dg,xu,xu]})}return t})();var Jk=class{_box;_destroyed=new OA;_resizeSubject=new OA;_resizeObserver;_elementObservables=new Map;constructor(e){this._box=e,typeof ResizeObserver<"u"&&(this._resizeObserver=new ResizeObserver(A=>this._resizeSubject.next(A)))}observe(e){return this._elementObservables.has(e)||this._elementObservables.set(e,new ct(A=>{let i=this._resizeSubject.subscribe(A);return this._resizeObserver?.observe(e,{box:this._box}),()=>{this._resizeObserver?.unobserve(e),i.unsubscribe(),this._elementObservables.delete(e)}}).pipe(kt(A=>A.some(i=>i.target===e)),a0({bufferSize:1,refCount:!0}),St(this._destroyed))),this._elementObservables.get(e)}destroy(){this._destroyed.next(),this._destroyed.complete(),this._resizeSubject.complete(),this._elementObservables.clear()}},B8=(()=>{class t{_cleanupErrorListener;_observers=new Map;_ngZone=f(Qe);constructor(){typeof ResizeObserver<"u"}ngOnDestroy(){for(let[,A]of this._observers)A.destroy();this._observers.clear(),this._cleanupErrorListener?.()}observe(A,i){let n=i?.box||"content-box";return this._observers.has(n)||this._observers.set(n,new Jk(n)),this._observers.get(n).observe(A)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var Ci=function(t){return t[t.State=0]="State",t[t.Transition=1]="Transition",t[t.Sequence=2]="Sequence",t[t.Group=3]="Group",t[t.Animate=4]="Animate",t[t.Keyframes=5]="Keyframes",t[t.Style=6]="Style",t[t.Trigger=7]="Trigger",t[t.Reference=8]="Reference",t[t.AnimateChild=9]="AnimateChild",t[t.AnimateRef=10]="AnimateRef",t[t.Query=11]="Query",t[t.Stagger=12]="Stagger",t}(Ci||{}),kc="*";function sc(t,e){return{type:Ci.Trigger,name:t,definitions:e,options:{}}}function ss(t,e=null){return{type:Ci.Animate,styles:e,timings:t}}function hP(t,e=null){return{type:Ci.Sequence,steps:t,options:e}}function po(t){return{type:Ci.Style,styles:t,offset:null}}function js(t,e,A){return{type:Ci.State,name:t,styles:e,options:A}}function Vr(t,e,A=null){return{type:Ci.Transition,expr:t,animation:e,options:A}}function Tk(t=null){return{type:Ci.AnimateChild,options:t}}function Hk(t,e,A=null){return{type:Ci.Query,selector:t,animation:e,options:A}}var Eg=class{_onDoneFns=[];_onStartFns=[];_onDestroyFns=[];_originalOnDoneFns=[];_originalOnStartFns=[];_started=!1;_destroyed=!1;_finished=!1;_position=0;parentPlayer=null;totalTime;constructor(e=0,A=0){this.totalTime=e+A}_onFinish(){this._finished||(this._finished=!0,this._onDoneFns.forEach(e=>e()),this._onDoneFns=[])}onStart(e){this._originalOnStartFns.push(e),this._onStartFns.push(e)}onDone(e){this._originalOnDoneFns.push(e),this._onDoneFns.push(e)}onDestroy(e){this._onDestroyFns.push(e)}hasStarted(){return this._started}init(){}play(){this.hasStarted()||(this._onStart(),this.triggerMicrotask()),this._started=!0}triggerMicrotask(){queueMicrotask(()=>this._onFinish())}_onStart(){this._onStartFns.forEach(e=>e()),this._onStartFns=[]}pause(){}restart(){}finish(){this._onFinish()}destroy(){this._destroyed||(this._destroyed=!0,this.hasStarted()||this._onStart(),this.finish(),this._onDestroyFns.forEach(e=>e()),this._onDestroyFns=[])}reset(){this._started=!1,this._finished=!1,this._onStartFns=this._originalOnStartFns,this._onDoneFns=this._originalOnDoneFns}setPosition(e){this._position=this.totalTime?e*this.totalTime:1}getPosition(){return this.totalTime?this._position/this.totalTime:1}triggerCallback(e){let A=e=="start"?this._onStartFns:this._onDoneFns;A.forEach(i=>i()),A.length=0}},tC=class{_onDoneFns=[];_onStartFns=[];_finished=!1;_started=!1;_destroyed=!1;_onDestroyFns=[];parentPlayer=null;totalTime=0;players;constructor(e){this.players=e;let A=0,i=0,n=0,o=this.players.length;o==0?queueMicrotask(()=>this._onFinish()):this.players.forEach(r=>{r.onDone(()=>{++A==o&&this._onFinish()}),r.onDestroy(()=>{++i==o&&this._onDestroy()}),r.onStart(()=>{++n==o&&this._onStart()})}),this.totalTime=this.players.reduce((r,s)=>Math.max(r,s.totalTime),0)}_onFinish(){this._finished||(this._finished=!0,this._onDoneFns.forEach(e=>e()),this._onDoneFns=[])}init(){this.players.forEach(e=>e.init())}onStart(e){this._onStartFns.push(e)}_onStart(){this.hasStarted()||(this._started=!0,this._onStartFns.forEach(e=>e()),this._onStartFns=[])}onDone(e){this._onDoneFns.push(e)}onDestroy(e){this._onDestroyFns.push(e)}hasStarted(){return this._started}play(){this.parentPlayer||this.init(),this._onStart(),this.players.forEach(e=>e.play())}pause(){this.players.forEach(e=>e.pause())}restart(){this.players.forEach(e=>e.restart())}finish(){this._onFinish(),this.players.forEach(e=>e.finish())}destroy(){this._onDestroy()}_onDestroy(){this._destroyed||(this._destroyed=!0,this._onFinish(),this.players.forEach(e=>e.destroy()),this._onDestroyFns.forEach(e=>e()),this._onDestroyFns=[])}reset(){this.players.forEach(e=>e.reset()),this._destroyed=!1,this._finished=!1,this._started=!1}setPosition(e){let A=e*this.totalTime;this.players.forEach(i=>{let n=i.totalTime?Math.min(1,A/i.totalTime):1;i.setPosition(n)})}getPosition(){let e=this.players.reduce((A,i)=>A===null||i.totalTime>A.totalTime?i:A,null);return e!=null?e.getPosition():0}beforeDestroy(){this.players.forEach(e=>{e.beforeDestroy&&e.beforeDestroy()})}triggerCallback(e){let A=e=="start"?this._onStartFns:this._onDoneFns;A.forEach(i=>i()),A.length=0}},FB="!";var A4A=["notch"],e4A=["matFormFieldNotchedOutline",""],t4A=["*"],i4A=["textField"],n4A=["iconPrefixContainer"],o4A=["textPrefixContainer"],r4A=["iconSuffixContainer"],s4A=["textSuffixContainer"],a4A=["*",[["mat-label"]],[["","matPrefix",""],["","matIconPrefix",""]],[["","matTextPrefix",""]],[["","matTextSuffix",""]],[["","matSuffix",""],["","matIconSuffix",""]],[["mat-error"],["","matError",""]],[["mat-hint",3,"align","end"]],[["mat-hint","align","end"]]],c4A=["*","mat-label","[matPrefix], [matIconPrefix]","[matTextPrefix]","[matTextSuffix]","[matSuffix], [matIconSuffix]","mat-error, [matError]","mat-hint:not([align='end'])","mat-hint[align='end']"];function l4A(t,e){t&1&&JA(0,"span",21)}function g4A(t,e){if(t&1&&(S(0,"label",20),Fe(1,1),_A(2,l4A,1,0,"span",21),F()),t&2){let A=O(2);yA("floating",A._shouldLabelFloat())("monitorResize",A._hasOutline())("id",A._labelId),Ne("for",A._control.disableAutomaticLabeling?null:A._control.id),G(2),GA(!A.hideRequiredMarker&&A._control.required?2:-1)}}function I4A(t,e){if(t&1&&_A(0,g4A,3,5,"label",20),t&2){let A=O();GA(A._hasFloatingLabel()?0:-1)}}function C4A(t,e){t&1&&JA(0,"div",7)}function d4A(t,e){}function B4A(t,e){if(t&1&&_A(0,d4A,0,0,"ng-template",13),t&2){O(2);let A=cr(1);yA("ngTemplateOutlet",A)}}function E4A(t,e){if(t&1&&(S(0,"div",9),_A(1,B4A,1,1,null,13),F()),t&2){let A=O();yA("matFormFieldNotchedOutlineOpen",A._shouldLabelFloat()),G(),GA(A._forceDisplayInfixLabel()?-1:1)}}function Q4A(t,e){t&1&&(S(0,"div",10,2),Fe(2,2),F())}function h4A(t,e){t&1&&(S(0,"div",11,3),Fe(2,3),F())}function u4A(t,e){}function f4A(t,e){if(t&1&&_A(0,u4A,0,0,"ng-template",13),t&2){O();let A=cr(1);yA("ngTemplateOutlet",A)}}function m4A(t,e){t&1&&(S(0,"div",14,4),Fe(2,4),F())}function p4A(t,e){t&1&&(S(0,"div",15,5),Fe(2,5),F())}function w4A(t,e){t&1&&JA(0,"div",16)}function D4A(t,e){if(t&1&&(S(0,"div",18),Fe(1,6),F()),t&2){let A=O();yA("@transitionMessages",A._subscriptAnimationState)}}function y4A(t,e){if(t&1&&(S(0,"mat-hint",22),AA(1),F()),t&2){let A=O(2);yA("id",A._hintLabelId),G(),Gt(A.hintLabel)}}function v4A(t,e){if(t&1&&(S(0,"div",19),_A(1,y4A,2,2,"mat-hint",22),Fe(2,7),JA(3,"div",23),Fe(4,8),F()),t&2){let A=O();yA("@transitionMessages",A._subscriptAnimationState),G(),GA(A.hintLabel?1:-1)}}var Q8=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["mat-label"]]})}return t})(),b4A=new dA("MatError");var uP=(()=>{class t{align="start";id=f(sn).getId("mat-mdc-hint-");static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["mat-hint"]],hostAttrs:[1,"mat-mdc-form-field-hint","mat-mdc-form-field-bottom-align"],hostVars:4,hostBindings:function(i,n){i&2&&(Hs("id",n.id),Ne("align",null),ue("mat-mdc-form-field-hint-end",n.align==="end"))},inputs:{align:"align",id:"id"}})}return t})(),M4A=new dA("MatPrefix");var vP=new dA("MatSuffix"),bP=(()=>{class t{set _isTextSelector(A){this._isText=!0}_isText=!1;static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matSuffix",""],["","matIconSuffix",""],["","matTextSuffix",""]],inputs:{_isTextSelector:[0,"matTextSuffix","_isTextSelector"]},features:[ht([{provide:vP,useExisting:t}])]})}return t})(),MP=new dA("FloatingLabelParent"),fP=(()=>{class t{_elementRef=f(ee);get floating(){return this._floating}set floating(A){this._floating=A,this.monitorResize&&this._handleResize()}_floating=!1;get monitorResize(){return this._monitorResize}set monitorResize(A){this._monitorResize=A,this._monitorResize?this._subscribeToResize():this._resizeSubscription.unsubscribe()}_monitorResize=!1;_resizeObserver=f(B8);_ngZone=f(Qe);_parent=f(MP);_resizeSubscription=new Kt;constructor(){}ngOnDestroy(){this._resizeSubscription.unsubscribe()}getWidth(){return k4A(this._elementRef.nativeElement)}get element(){return this._elementRef.nativeElement}_handleResize(){setTimeout(()=>this._parent._handleLabelResized())}_subscribeToResize(){this._resizeSubscription.unsubscribe(),this._ngZone.runOutsideAngular(()=>{this._resizeSubscription=this._resizeObserver.observe(this._elementRef.nativeElement,{box:"border-box"}).subscribe(()=>this._handleResize())})}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["label","matFormFieldFloatingLabel",""]],hostAttrs:[1,"mdc-floating-label","mat-mdc-floating-label"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mdc-floating-label--float-above",n.floating)},inputs:{floating:"floating",monitorResize:"monitorResize"}})}return t})();function k4A(t){let e=t;if(e.offsetParent!==null)return e.scrollWidth;let A=e.cloneNode(!0);A.style.setProperty("position","absolute"),A.style.setProperty("transform","translate(-9999px, -9999px)"),document.documentElement.appendChild(A);let i=A.scrollWidth;return A.remove(),i}var mP="mdc-line-ripple--active",E8="mdc-line-ripple--deactivating",pP=(()=>{class t{_elementRef=f(ee);_cleanupTransitionEnd;constructor(){let A=f(Qe),i=f(Wi);A.runOutsideAngular(()=>{this._cleanupTransitionEnd=i.listen(this._elementRef.nativeElement,"transitionend",this._handleTransitionEnd)})}activate(){let A=this._elementRef.nativeElement.classList;A.remove(E8),A.add(mP)}deactivate(){this._elementRef.nativeElement.classList.add(E8)}_handleTransitionEnd=A=>{let i=this._elementRef.nativeElement.classList,n=i.contains(E8);A.propertyName==="opacity"&&n&&i.remove(mP,E8)};ngOnDestroy(){this._cleanupTransitionEnd()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["div","matFormFieldLineRipple",""]],hostAttrs:[1,"mdc-line-ripple"]})}return t})(),wP=(()=>{class t{_elementRef=f(ee);_ngZone=f(Qe);open=!1;_notch;constructor(){}ngAfterViewInit(){let A=this._elementRef.nativeElement.querySelector(".mdc-floating-label");A?(this._elementRef.nativeElement.classList.add("mdc-notched-outline--upgraded"),typeof requestAnimationFrame=="function"&&(A.style.transitionDuration="0s",this._ngZone.runOutsideAngular(()=>{requestAnimationFrame(()=>A.style.transitionDuration="")}))):this._elementRef.nativeElement.classList.add("mdc-notched-outline--no-label")}_setNotchWidth(A){!this.open||!A?this._notch.nativeElement.style.width="":this._notch.nativeElement.style.width=`calc(${A}px * var(--mat-mdc-form-field-floating-label-scale, 0.75) + 9px)`}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["div","matFormFieldNotchedOutline",""]],viewQuery:function(i,n){if(i&1&&Te(A4A,5),i&2){let o;XA(o=$A())&&(n._notch=o.first)}},hostAttrs:[1,"mdc-notched-outline"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mdc-notched-outline--notched",n.open)},inputs:{open:[0,"matFormFieldNotchedOutlineOpen","open"]},attrs:e4A,ngContentSelectors:t4A,decls:5,vars:0,consts:[["notch",""],[1,"mat-mdc-notch-piece","mdc-notched-outline__leading"],[1,"mat-mdc-notch-piece","mdc-notched-outline__notch"],[1,"mat-mdc-notch-piece","mdc-notched-outline__trailing"]],template:function(i,n){i&1&&(jt(),JA(0,"div",1),S(1,"div",2,0),Fe(3),F(),JA(4,"div",3))},encapsulation:2,changeDetection:0})}return t})(),S4A={transitionMessages:sc("transitionMessages",[js("enter",po({opacity:1,transform:"translateY(0%)"})),Vr("void => enter",[po({opacity:0,transform:"translateY(-5px)"}),ss("300ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},_u=(()=>{class t{value;stateChanges;id;placeholder;ngControl;focused;empty;shouldLabelFloat;required;disabled;errorState;controlType;autofilled;userAriaDescribedBy;disableAutomaticLabeling;static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t})}return t})();var Gu=new dA("MatFormField"),R4A=new dA("MAT_FORM_FIELD_DEFAULT_OPTIONS"),DP="fill",x4A="auto",yP="fixed",L4A="translateY(-50%)",Qg=(()=>{class t{_elementRef=f(ee);_changeDetectorRef=f(It);_dir=f(mo);_platform=f(Ii);_idGenerator=f(sn);_defaults=f(R4A,{optional:!0});_animationMode=f(Si,{optional:!0});_textField;_iconPrefixContainer;_textPrefixContainer;_iconSuffixContainer;_textSuffixContainer;_floatingLabel;_notchedOutline;_lineRipple;_formFieldControl;_prefixChildren;_suffixChildren;_errorChildren;_hintChildren;_labelChild=GT(Q8);get hideRequiredMarker(){return this._hideRequiredMarker}set hideRequiredMarker(A){this._hideRequiredMarker=Xo(A)}_hideRequiredMarker=!1;color="primary";get floatLabel(){return this._floatLabel||this._defaults?.floatLabel||x4A}set floatLabel(A){A!==this._floatLabel&&(this._floatLabel=A,this._changeDetectorRef.markForCheck())}_floatLabel;get appearance(){return this._appearance}set appearance(A){let i=this._appearance,n=A||this._defaults?.appearance||DP;this._appearance=n,this._appearance==="outline"&&this._appearance!==i&&(this._needsOutlineLabelOffsetUpdate=!0)}_appearance=DP;get subscriptSizing(){return this._subscriptSizing||this._defaults?.subscriptSizing||yP}set subscriptSizing(A){this._subscriptSizing=A||this._defaults?.subscriptSizing||yP}_subscriptSizing=null;get hintLabel(){return this._hintLabel}set hintLabel(A){this._hintLabel=A,this._processHints()}_hintLabel="";_hasIconPrefix=!1;_hasTextPrefix=!1;_hasIconSuffix=!1;_hasTextSuffix=!1;_labelId=this._idGenerator.getId("mat-mdc-form-field-label-");_hintLabelId=this._idGenerator.getId("mat-mdc-hint-");_subscriptAnimationState="";get _control(){return this._explicitFormFieldControl||this._formFieldControl}set _control(A){this._explicitFormFieldControl=A}_destroyed=new OA;_isFocused=null;_explicitFormFieldControl;_needsOutlineLabelOffsetUpdate=!1;_previousControl=null;_stateChanges;_valueChanges;_describedByChanges;_injector=f(Rt);constructor(){let A=this._defaults;A&&(A.appearance&&(this.appearance=A.appearance),this._hideRequiredMarker=!!A?.hideRequiredMarker,A.color&&(this.color=A.color))}ngAfterViewInit(){this._updateFocusState(),this._subscriptAnimationState="enter",this._changeDetectorRef.detectChanges()}ngAfterContentInit(){this._assertFormFieldControl(),this._initializeSubscript(),this._initializePrefixAndSuffix(),this._initializeOutlineLabelOffsetSubscriptions()}ngAfterContentChecked(){this._assertFormFieldControl(),this._control!==this._previousControl&&(this._initializeControl(this._previousControl),this._previousControl=this._control)}ngOnDestroy(){this._stateChanges?.unsubscribe(),this._valueChanges?.unsubscribe(),this._describedByChanges?.unsubscribe(),this._destroyed.next(),this._destroyed.complete()}getLabelId=h0(()=>this._hasFloatingLabel()?this._labelId:null);getConnectedOverlayOrigin(){return this._textField||this._elementRef}_animateAndLockLabel(){this._hasFloatingLabel()&&(this.floatLabel="always")}_initializeControl(A){let i=this._control,n="mat-mdc-form-field-type-";A&&this._elementRef.nativeElement.classList.remove(n+A.controlType),i.controlType&&this._elementRef.nativeElement.classList.add(n+i.controlType),this._stateChanges?.unsubscribe(),this._stateChanges=i.stateChanges.subscribe(()=>{this._updateFocusState(),this._changeDetectorRef.markForCheck()}),this._describedByChanges?.unsubscribe(),this._describedByChanges=i.stateChanges.pipe(Pn([void 0,void 0]),je(()=>[i.errorState,i.userAriaDescribedBy]),pm(),kt(([[o,r],[s,a]])=>o!==s||r!==a)).subscribe(()=>this._syncDescribedByIds()),this._valueChanges?.unsubscribe(),i.ngControl&&i.ngControl.valueChanges&&(this._valueChanges=i.ngControl.valueChanges.pipe(St(this._destroyed)).subscribe(()=>this._changeDetectorRef.markForCheck()))}_checkPrefixAndSuffixTypes(){this._hasIconPrefix=!!this._prefixChildren.find(A=>!A._isText),this._hasTextPrefix=!!this._prefixChildren.find(A=>A._isText),this._hasIconSuffix=!!this._suffixChildren.find(A=>!A._isText),this._hasTextSuffix=!!this._suffixChildren.find(A=>A._isText)}_initializePrefixAndSuffix(){this._checkPrefixAndSuffixTypes(),zn(this._prefixChildren.changes,this._suffixChildren.changes).subscribe(()=>{this._checkPrefixAndSuffixTypes(),this._changeDetectorRef.markForCheck()})}_initializeSubscript(){this._hintChildren.changes.subscribe(()=>{this._processHints(),this._changeDetectorRef.markForCheck()}),this._errorChildren.changes.subscribe(()=>{this._syncDescribedByIds(),this._changeDetectorRef.markForCheck()}),this._validateHints(),this._syncDescribedByIds()}_assertFormFieldControl(){this._control}_updateFocusState(){this._control.focused&&!this._isFocused?(this._isFocused=!0,this._lineRipple?.activate()):!this._control.focused&&(this._isFocused||this._isFocused===null)&&(this._isFocused=!1,this._lineRipple?.deactivate()),this._textField?.nativeElement.classList.toggle("mdc-text-field--focused",this._control.focused)}_initializeOutlineLabelOffsetSubscriptions(){this._prefixChildren.changes.subscribe(()=>this._needsOutlineLabelOffsetUpdate=!0),vh(()=>{this._needsOutlineLabelOffsetUpdate&&(this._needsOutlineLabelOffsetUpdate=!1,this._updateOutlineLabelOffset())},{injector:this._injector}),this._dir.change.pipe(St(this._destroyed)).subscribe(()=>this._needsOutlineLabelOffsetUpdate=!0)}_shouldAlwaysFloat(){return this.floatLabel==="always"}_hasOutline(){return this.appearance==="outline"}_forceDisplayInfixLabel(){return!this._platform.isBrowser&&this._prefixChildren.length&&!this._shouldLabelFloat()}_hasFloatingLabel=h0(()=>!!this._labelChild());_shouldLabelFloat(){return this._hasFloatingLabel()?this._control.shouldLabelFloat||this._shouldAlwaysFloat():!1}_shouldForward(A){let i=this._control?this._control.ngControl:null;return i&&i[A]}_getDisplayedMessages(){return this._errorChildren&&this._errorChildren.length>0&&this._control.errorState?"error":"hint"}_handleLabelResized(){this._refreshOutlineNotchWidth()}_refreshOutlineNotchWidth(){!this._hasOutline()||!this._floatingLabel||!this._shouldLabelFloat()?this._notchedOutline?._setNotchWidth(0):this._notchedOutline?._setNotchWidth(this._floatingLabel.getWidth())}_processHints(){this._validateHints(),this._syncDescribedByIds()}_validateHints(){this._hintChildren}_syncDescribedByIds(){if(this._control){let A=[];if(this._control.userAriaDescribedBy&&typeof this._control.userAriaDescribedBy=="string"&&A.push(...this._control.userAriaDescribedBy.split(" ")),this._getDisplayedMessages()==="hint"){let i=this._hintChildren?this._hintChildren.find(o=>o.align==="start"):null,n=this._hintChildren?this._hintChildren.find(o=>o.align==="end"):null;i?A.push(i.id):this._hintLabel&&A.push(this._hintLabelId),n&&A.push(n.id)}else this._errorChildren&&A.push(...this._errorChildren.map(i=>i.id));this._control.setDescribedByIds(A)}}_updateOutlineLabelOffset(){if(!this._hasOutline()||!this._floatingLabel)return;let A=this._floatingLabel.element;if(!(this._iconPrefixContainer||this._textPrefixContainer)){A.style.transform="";return}if(!this._isAttachedToDom()){this._needsOutlineLabelOffsetUpdate=!0;return}let i=this._iconPrefixContainer?.nativeElement,n=this._textPrefixContainer?.nativeElement,o=this._iconSuffixContainer?.nativeElement,r=this._textSuffixContainer?.nativeElement,s=i?.getBoundingClientRect().width??0,a=n?.getBoundingClientRect().width??0,c=o?.getBoundingClientRect().width??0,l=r?.getBoundingClientRect().width??0,I=this._dir.value==="rtl"?"-1":"1",C=`${s+a}px`,B=`calc(${I} * (${C} + var(--mat-mdc-form-field-label-offset-x, 0px)))`;A.style.transform=`var(
        --mat-mdc-form-field-label-transform,
        ${L4A} translateX(${B})
    )`;let E=s+a+c+l;this._elementRef.nativeElement.style.setProperty("--mat-form-field-notch-max-width",`calc(100% - ${E}px)`)}_isAttachedToDom(){let A=this._elementRef.nativeElement;if(A.getRootNode){let i=A.getRootNode();return i&&i!==A}return document.documentElement.contains(A)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-form-field"]],contentQueries:function(i,n,o){if(i&1&&(oH(o,n._labelChild,Q8,5),ci(o,_u,5),ci(o,M4A,5),ci(o,vP,5),ci(o,b4A,5),ci(o,uP,5)),i&2){rH();let r;XA(r=$A())&&(n._formFieldControl=r.first),XA(r=$A())&&(n._prefixChildren=r),XA(r=$A())&&(n._suffixChildren=r),XA(r=$A())&&(n._errorChildren=r),XA(r=$A())&&(n._hintChildren=r)}},viewQuery:function(i,n){if(i&1&&(Te(i4A,5),Te(n4A,5),Te(o4A,5),Te(r4A,5),Te(s4A,5),Te(fP,5),Te(wP,5),Te(pP,5)),i&2){let o;XA(o=$A())&&(n._textField=o.first),XA(o=$A())&&(n._iconPrefixContainer=o.first),XA(o=$A())&&(n._textPrefixContainer=o.first),XA(o=$A())&&(n._iconSuffixContainer=o.first),XA(o=$A())&&(n._textSuffixContainer=o.first),XA(o=$A())&&(n._floatingLabel=o.first),XA(o=$A())&&(n._notchedOutline=o.first),XA(o=$A())&&(n._lineRipple=o.first)}},hostAttrs:[1,"mat-mdc-form-field"],hostVars:42,hostBindings:function(i,n){i&2&&ue("mat-mdc-form-field-label-always-float",n._shouldAlwaysFloat())("mat-mdc-form-field-has-icon-prefix",n._hasIconPrefix)("mat-mdc-form-field-has-icon-suffix",n._hasIconSuffix)("mat-form-field-invalid",n._control.errorState)("mat-form-field-disabled",n._control.disabled)("mat-form-field-autofilled",n._control.autofilled)("mat-form-field-no-animations",n._animationMode==="NoopAnimations")("mat-form-field-appearance-fill",n.appearance=="fill")("mat-form-field-appearance-outline",n.appearance=="outline")("mat-form-field-hide-placeholder",n._hasFloatingLabel()&&!n._shouldLabelFloat())("mat-focused",n._control.focused)("mat-primary",n.color!=="accent"&&n.color!=="warn")("mat-accent",n.color==="accent")("mat-warn",n.color==="warn")("ng-untouched",n._shouldForward("untouched"))("ng-touched",n._shouldForward("touched"))("ng-pristine",n._shouldForward("pristine"))("ng-dirty",n._shouldForward("dirty"))("ng-valid",n._shouldForward("valid"))("ng-invalid",n._shouldForward("invalid"))("ng-pending",n._shouldForward("pending"))},inputs:{hideRequiredMarker:"hideRequiredMarker",color:"color",floatLabel:"floatLabel",appearance:"appearance",subscriptSizing:"subscriptSizing",hintLabel:"hintLabel"},exportAs:["matFormField"],features:[ht([{provide:Gu,useExisting:t},{provide:MP,useExisting:t}])],ngContentSelectors:c4A,decls:18,vars:21,consts:[["labelTemplate",""],["textField",""],["iconPrefixContainer",""],["textPrefixContainer",""],["textSuffixContainer",""],["iconSuffixContainer",""],[1,"mat-mdc-text-field-wrapper","mdc-text-field",3,"click"],[1,"mat-mdc-form-field-focus-overlay"],[1,"mat-mdc-form-field-flex"],["matFormFieldNotchedOutline","",3,"matFormFieldNotchedOutlineOpen"],[1,"mat-mdc-form-field-icon-prefix"],[1,"mat-mdc-form-field-text-prefix"],[1,"mat-mdc-form-field-infix"],[3,"ngTemplateOutlet"],[1,"mat-mdc-form-field-text-suffix"],[1,"mat-mdc-form-field-icon-suffix"],["matFormFieldLineRipple",""],[1,"mat-mdc-form-field-subscript-wrapper","mat-mdc-form-field-bottom-align"],[1,"mat-mdc-form-field-error-wrapper"],[1,"mat-mdc-form-field-hint-wrapper"],["matFormFieldFloatingLabel","",3,"floating","monitorResize","id"],["aria-hidden","true",1,"mat-mdc-form-field-required-marker","mdc-floating-label--required"],[3,"id"],[1,"mat-mdc-form-field-hint-spacer"]],template:function(i,n){if(i&1){let o=be();jt(a4A),_A(0,I4A,1,1,"ng-template",null,0,Fh),S(2,"div",6,1),hA("click",function(s){return RA(o),xA(n._control.onContainerClick(s))}),_A(4,C4A,1,0,"div",7),S(5,"div",8),_A(6,E4A,2,2,"div",9)(7,Q4A,3,0,"div",10)(8,h4A,3,0,"div",11),S(9,"div",12),_A(10,f4A,1,1,null,13),Fe(11),F(),_A(12,m4A,3,0,"div",14)(13,p4A,3,0,"div",15),F(),_A(14,w4A,1,0,"div",16),F(),S(15,"div",17),_A(16,D4A,2,1,"div",18)(17,v4A,5,2,"div",19),F()}if(i&2){let o;G(2),ue("mdc-text-field--filled",!n._hasOutline())("mdc-text-field--outlined",n._hasOutline())("mdc-text-field--no-label",!n._hasFloatingLabel())("mdc-text-field--disabled",n._control.disabled)("mdc-text-field--invalid",n._control.errorState),G(2),GA(!n._hasOutline()&&!n._control.disabled?4:-1),G(2),GA(n._hasOutline()?6:-1),G(),GA(n._hasIconPrefix?7:-1),G(),GA(n._hasTextPrefix?8:-1),G(2),GA(!n._hasOutline()||n._forceDisplayInfixLabel()?10:-1),G(2),GA(n._hasTextSuffix?12:-1),G(),GA(n._hasIconSuffix?13:-1),G(),GA(n._hasOutline()?-1:14),G(),ue("mat-mdc-form-field-subscript-dynamic-size",n.subscriptSizing==="dynamic"),G(),GA((o=n._getDisplayedMessages())==="error"?16:o==="hint"?17:-1)}},dependencies:[fP,wP,Yh,pP,uP],styles:['.mdc-text-field{display:inline-flex;align-items:baseline;padding:0 16px;position:relative;box-sizing:border-box;overflow:hidden;will-change:opacity,transform,color;border-top-left-radius:4px;border-top-right-radius:4px;border-bottom-right-radius:0;border-bottom-left-radius:0}.mdc-text-field__input{width:100%;min-width:0;border:none;border-radius:0;background:none;padding:0;-moz-appearance:none;-webkit-appearance:none;height:28px}.mdc-text-field__input::-webkit-calendar-picker-indicator{display:none}.mdc-text-field__input::-ms-clear{display:none}.mdc-text-field__input:focus{outline:none}.mdc-text-field__input:invalid{box-shadow:none}.mdc-text-field__input::placeholder{opacity:0}.mdc-text-field__input::-moz-placeholder{opacity:0}.mdc-text-field__input::-webkit-input-placeholder{opacity:0}.mdc-text-field__input:-ms-input-placeholder{opacity:0}.mdc-text-field--no-label .mdc-text-field__input::placeholder,.mdc-text-field--focused .mdc-text-field__input::placeholder{opacity:1}.mdc-text-field--no-label .mdc-text-field__input::-moz-placeholder,.mdc-text-field--focused .mdc-text-field__input::-moz-placeholder{opacity:1}.mdc-text-field--no-label .mdc-text-field__input::-webkit-input-placeholder,.mdc-text-field--focused .mdc-text-field__input::-webkit-input-placeholder{opacity:1}.mdc-text-field--no-label .mdc-text-field__input:-ms-input-placeholder,.mdc-text-field--focused .mdc-text-field__input:-ms-input-placeholder{opacity:1}.mdc-text-field--disabled:not(.mdc-text-field--no-label) .mdc-text-field__input.mat-mdc-input-disabled-interactive::placeholder{opacity:0}.mdc-text-field--disabled:not(.mdc-text-field--no-label) .mdc-text-field__input.mat-mdc-input-disabled-interactive::-moz-placeholder{opacity:0}.mdc-text-field--disabled:not(.mdc-text-field--no-label) .mdc-text-field__input.mat-mdc-input-disabled-interactive::-webkit-input-placeholder{opacity:0}.mdc-text-field--disabled:not(.mdc-text-field--no-label) .mdc-text-field__input.mat-mdc-input-disabled-interactive:-ms-input-placeholder{opacity:0}.mdc-text-field--outlined .mdc-text-field__input,.mdc-text-field--filled.mdc-text-field--no-label .mdc-text-field__input{height:100%}.mdc-text-field--outlined .mdc-text-field__input{display:flex;border:none !important;background-color:rgba(0,0,0,0)}.mdc-text-field--disabled .mdc-text-field__input{pointer-events:auto}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-text-field__input{color:var(--mdc-filled-text-field-input-text-color, var(--mat-sys-on-surface));caret-color:var(--mdc-filled-text-field-caret-color, var(--mat-sys-primary))}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-text-field__input::placeholder{color:var(--mdc-filled-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-text-field__input::-moz-placeholder{color:var(--mdc-filled-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-text-field__input::-webkit-input-placeholder{color:var(--mdc-filled-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-text-field__input:-ms-input-placeholder{color:var(--mdc-filled-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled.mdc-text-field--invalid:not(.mdc-text-field--disabled) .mdc-text-field__input{caret-color:var(--mdc-filled-text-field-error-caret-color)}.mdc-text-field--filled.mdc-text-field--disabled .mdc-text-field__input{color:var(--mdc-filled-text-field-disabled-input-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-text-field__input{color:var(--mdc-outlined-text-field-input-text-color, var(--mat-sys-on-surface));caret-color:var(--mdc-outlined-text-field-caret-color, var(--mat-sys-primary))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-text-field__input::placeholder{color:var(--mdc-outlined-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-text-field__input::-moz-placeholder{color:var(--mdc-outlined-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-text-field__input::-webkit-input-placeholder{color:var(--mdc-outlined-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-text-field__input:-ms-input-placeholder{color:var(--mdc-outlined-text-field-input-text-placeholder-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--outlined.mdc-text-field--invalid:not(.mdc-text-field--disabled) .mdc-text-field__input{caret-color:var(--mdc-outlined-text-field-error-caret-color)}.mdc-text-field--outlined.mdc-text-field--disabled .mdc-text-field__input{color:var(--mdc-outlined-text-field-disabled-input-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}@media(forced-colors: active){.mdc-text-field--disabled .mdc-text-field__input{background-color:Window}}.mdc-text-field--filled{height:56px;border-bottom-right-radius:0;border-bottom-left-radius:0;border-top-left-radius:var(--mdc-filled-text-field-container-shape, var(--mat-sys-corner-extra-small));border-top-right-radius:var(--mdc-filled-text-field-container-shape, var(--mat-sys-corner-extra-small))}.mdc-text-field--filled:not(.mdc-text-field--disabled){background-color:var(--mdc-filled-text-field-container-color, var(--mat-sys-surface-variant))}.mdc-text-field--filled.mdc-text-field--disabled{background-color:var(--mdc-filled-text-field-disabled-container-color, color-mix(in srgb, var(--mat-sys-on-surface) 4%, transparent))}.mdc-text-field--outlined{height:56px;overflow:visible;padding-right:max(16px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)));padding-left:max(16px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)) + 4px)}[dir=rtl] .mdc-text-field--outlined{padding-right:max(16px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)) + 4px);padding-left:max(16px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)))}.mdc-floating-label{position:absolute;left:0;transform-origin:left top;line-height:1.15rem;text-align:left;text-overflow:ellipsis;white-space:nowrap;cursor:text;overflow:hidden;will-change:transform}[dir=rtl] .mdc-floating-label{right:0;left:auto;transform-origin:right top;text-align:right}.mdc-text-field .mdc-floating-label{top:50%;transform:translateY(-50%);pointer-events:none}.mdc-notched-outline .mdc-floating-label{display:inline-block;position:relative;max-width:100%}.mdc-text-field--outlined .mdc-floating-label{left:4px;right:auto}[dir=rtl] .mdc-text-field--outlined .mdc-floating-label{left:auto;right:4px}.mdc-text-field--filled .mdc-floating-label{left:16px;right:auto}[dir=rtl] .mdc-text-field--filled .mdc-floating-label{left:auto;right:16px}.mdc-text-field--disabled .mdc-floating-label{cursor:default}@media(forced-colors: active){.mdc-text-field--disabled .mdc-floating-label{z-index:1}}.mdc-text-field--filled.mdc-text-field--no-label .mdc-floating-label{display:none}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-floating-label{color:var(--mdc-filled-text-field-label-text-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--focused .mdc-floating-label{color:var(--mdc-filled-text-field-focus-label-text-color, var(--mat-sys-primary))}.mdc-text-field--filled:not(.mdc-text-field--disabled):not(.mdc-text-field--focused):hover .mdc-floating-label{color:var(--mdc-filled-text-field-hover-label-text-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled.mdc-text-field--disabled .mdc-floating-label{color:var(--mdc-filled-text-field-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--invalid .mdc-floating-label{color:var(--mdc-filled-text-field-error-label-text-color, var(--mat-sys-error))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--invalid.mdc-text-field--focused .mdc-floating-label{color:var(--mdc-filled-text-field-error-focus-label-text-color, var(--mat-sys-error))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--invalid:not(.mdc-text-field--disabled):hover .mdc-floating-label{color:var(--mdc-filled-text-field-error-hover-label-text-color, var(--mat-sys-on-error-container))}.mdc-text-field--filled .mdc-floating-label{font-family:var(--mdc-filled-text-field-label-text-font, var(--mat-sys-body-large-font));font-size:var(--mdc-filled-text-field-label-text-size, var(--mat-sys-body-large-size));font-weight:var(--mdc-filled-text-field-label-text-weight, var(--mat-sys-body-large-weight));letter-spacing:var(--mdc-filled-text-field-label-text-tracking, var(--mat-sys-body-large-tracking))}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mdc-floating-label{color:var(--mdc-outlined-text-field-label-text-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--focused .mdc-floating-label{color:var(--mdc-outlined-text-field-focus-label-text-color, var(--mat-sys-primary))}.mdc-text-field--outlined:not(.mdc-text-field--disabled):not(.mdc-text-field--focused):hover .mdc-floating-label{color:var(--mdc-outlined-text-field-hover-label-text-color, var(--mat-sys-on-surface))}.mdc-text-field--outlined.mdc-text-field--disabled .mdc-floating-label{color:var(--mdc-outlined-text-field-disabled-label-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid .mdc-floating-label{color:var(--mdc-outlined-text-field-error-label-text-color, var(--mat-sys-error))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid.mdc-text-field--focused .mdc-floating-label{color:var(--mdc-outlined-text-field-error-focus-label-text-color, var(--mat-sys-error))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid:not(.mdc-text-field--disabled):hover .mdc-floating-label{color:var(--mdc-outlined-text-field-error-hover-label-text-color, var(--mat-sys-on-error-container))}.mdc-text-field--outlined .mdc-floating-label{font-family:var(--mdc-outlined-text-field-label-text-font, var(--mat-sys-body-large-font));font-size:var(--mdc-outlined-text-field-label-text-size, var(--mat-sys-body-large-size));font-weight:var(--mdc-outlined-text-field-label-text-weight, var(--mat-sys-body-large-weight));letter-spacing:var(--mdc-outlined-text-field-label-text-tracking, var(--mat-sys-body-large-tracking))}.mdc-floating-label--float-above{cursor:auto;transform:translateY(-106%) scale(0.75)}.mdc-text-field--filled .mdc-floating-label--float-above{transform:translateY(-106%) scale(0.75)}.mdc-text-field--outlined .mdc-floating-label--float-above{transform:translateY(-37.25px) scale(1);font-size:.75rem}.mdc-notched-outline .mdc-floating-label--float-above{text-overflow:clip}.mdc-notched-outline--upgraded .mdc-floating-label--float-above{max-width:133.3333333333%}.mdc-text-field--outlined.mdc-notched-outline--upgraded .mdc-floating-label--float-above,.mdc-text-field--outlined .mdc-notched-outline--upgraded .mdc-floating-label--float-above{transform:translateY(-34.75px) scale(0.75)}.mdc-text-field--outlined.mdc-notched-outline--upgraded .mdc-floating-label--float-above,.mdc-text-field--outlined .mdc-notched-outline--upgraded .mdc-floating-label--float-above{font-size:1rem}.mdc-floating-label--required:not(.mdc-floating-label--hide-required-marker)::after{margin-left:1px;margin-right:0;content:"*"}[dir=rtl] .mdc-floating-label--required:not(.mdc-floating-label--hide-required-marker)::after{margin-left:0;margin-right:1px}.mdc-notched-outline{display:flex;position:absolute;top:0;right:0;left:0;box-sizing:border-box;width:100%;max-width:100%;height:100%;text-align:left;pointer-events:none}[dir=rtl] .mdc-notched-outline{text-align:right}.mdc-text-field--outlined .mdc-notched-outline{z-index:1}.mat-mdc-notch-piece{box-sizing:border-box;height:100%;pointer-events:none;border-top:1px solid;border-bottom:1px solid}.mdc-text-field--focused .mat-mdc-notch-piece{border-width:2px}.mdc-text-field--outlined:not(.mdc-text-field--disabled) .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-outline-color, var(--mat-sys-outline));border-width:var(--mdc-outlined-text-field-outline-width, 1px)}.mdc-text-field--outlined:not(.mdc-text-field--disabled):not(.mdc-text-field--focused):hover .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-hover-outline-color, var(--mat-sys-on-surface))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--focused .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-focus-outline-color, var(--mat-sys-primary))}.mdc-text-field--outlined.mdc-text-field--disabled .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-disabled-outline-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-error-outline-color, var(--mat-sys-error))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid:not(.mdc-text-field--focused):hover .mdc-notched-outline .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-error-hover-outline-color, var(--mat-sys-on-error-container))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--invalid.mdc-text-field--focused .mat-mdc-notch-piece{border-color:var(--mdc-outlined-text-field-error-focus-outline-color, var(--mat-sys-error))}.mdc-text-field--outlined:not(.mdc-text-field--disabled).mdc-text-field--focused .mdc-notched-outline .mat-mdc-notch-piece{border-width:var(--mdc-outlined-text-field-focus-outline-width, 2px)}.mdc-notched-outline__leading{border-left:1px solid;border-right:none;border-top-right-radius:0;border-bottom-right-radius:0;border-top-left-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small));border-bottom-left-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small))}.mdc-text-field--outlined .mdc-notched-outline .mdc-notched-outline__leading{width:max(12px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)))}[dir=rtl] .mdc-notched-outline__leading{border-left:none;border-right:1px solid;border-bottom-left-radius:0;border-top-left-radius:0;border-top-right-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small));border-bottom-right-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small))}.mdc-notched-outline__trailing{flex-grow:1;border-left:none;border-right:1px solid;border-top-left-radius:0;border-bottom-left-radius:0;border-top-right-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small));border-bottom-right-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small))}[dir=rtl] .mdc-notched-outline__trailing{border-left:1px solid;border-right:none;border-top-right-radius:0;border-bottom-right-radius:0;border-top-left-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small));border-bottom-left-radius:var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small))}.mdc-notched-outline__notch{flex:0 0 auto;width:auto}.mdc-text-field--outlined .mdc-notched-outline .mdc-notched-outline__notch{max-width:min(var(--mat-form-field-notch-max-width, 100%),100% - max(12px,var(--mdc-outlined-text-field-container-shape, var(--mat-sys-corner-extra-small)))*2)}.mdc-text-field--outlined .mdc-notched-outline--notched .mdc-notched-outline__notch{padding-top:1px}.mdc-text-field--focused.mdc-text-field--outlined .mdc-notched-outline--notched .mdc-notched-outline__notch{padding-top:2px}.mdc-notched-outline--notched .mdc-notched-outline__notch{padding-left:0;padding-right:8px;border-top:none;--mat-form-field-notch-max-width: 100%}[dir=rtl] .mdc-notched-outline--notched .mdc-notched-outline__notch{padding-left:8px;padding-right:0}.mdc-notched-outline--no-label .mdc-notched-outline__notch{display:none}.mdc-line-ripple::before,.mdc-line-ripple::after{position:absolute;bottom:0;left:0;width:100%;border-bottom-style:solid;content:""}.mdc-line-ripple::before{z-index:1;border-bottom-width:var(--mdc-filled-text-field-active-indicator-height, 1px)}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-line-ripple::before{border-bottom-color:var(--mdc-filled-text-field-active-indicator-color, var(--mat-sys-on-surface-variant))}.mdc-text-field--filled:not(.mdc-text-field--disabled):not(.mdc-text-field--focused):hover .mdc-line-ripple::before{border-bottom-color:var(--mdc-filled-text-field-hover-active-indicator-color, var(--mat-sys-on-surface))}.mdc-text-field--filled.mdc-text-field--disabled .mdc-line-ripple::before{border-bottom-color:var(--mdc-filled-text-field-disabled-active-indicator-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--invalid .mdc-line-ripple::before{border-bottom-color:var(--mdc-filled-text-field-error-active-indicator-color, var(--mat-sys-error))}.mdc-text-field--filled:not(.mdc-text-field--disabled).mdc-text-field--invalid:not(.mdc-text-field--focused):hover .mdc-line-ripple::before{border-bottom-color:var(--mdc-filled-text-field-error-hover-active-indicator-color, var(--mat-sys-on-error-container))}.mdc-line-ripple::after{transform:scaleX(0);opacity:0;z-index:2}.mdc-text-field--filled .mdc-line-ripple::after{border-bottom-width:var(--mdc-filled-text-field-focus-active-indicator-height, 2px)}.mdc-text-field--filled:not(.mdc-text-field--disabled) .mdc-line-ripple::after{border-bottom-color:var(--mdc-filled-text-field-focus-active-indicator-color, var(--mat-sys-primary))}.mdc-text-field--filled.mdc-text-field--invalid:not(.mdc-text-field--disabled) .mdc-line-ripple::after{border-bottom-color:var(--mdc-filled-text-field-error-focus-active-indicator-color, var(--mat-sys-error))}.mdc-line-ripple--active::after{transform:scaleX(1);opacity:1}.mdc-line-ripple--deactivating::after{opacity:0}.mdc-text-field--disabled{pointer-events:none}.mat-mdc-form-field-textarea-control{vertical-align:middle;resize:vertical;box-sizing:border-box;height:auto;margin:0;padding:0;border:none;overflow:auto}.mat-mdc-form-field-input-control.mat-mdc-form-field-input-control{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;font:inherit;letter-spacing:inherit;text-decoration:inherit;text-transform:inherit;border:none}.mat-mdc-form-field .mat-mdc-floating-label.mdc-floating-label{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;line-height:normal;pointer-events:all;will-change:auto}.mat-mdc-form-field:not(.mat-form-field-disabled) .mat-mdc-floating-label.mdc-floating-label{cursor:inherit}.mdc-text-field--no-label:not(.mdc-text-field--textarea) .mat-mdc-form-field-input-control.mdc-text-field__input,.mat-mdc-text-field-wrapper .mat-mdc-form-field-input-control{height:auto}.mat-mdc-text-field-wrapper .mat-mdc-form-field-input-control.mdc-text-field__input[type=color]{height:23px}.mat-mdc-text-field-wrapper{height:auto;flex:auto;will-change:auto}.mat-mdc-form-field-has-icon-prefix .mat-mdc-text-field-wrapper{padding-left:0;--mat-mdc-form-field-label-offset-x: -16px}.mat-mdc-form-field-has-icon-suffix .mat-mdc-text-field-wrapper{padding-right:0}[dir=rtl] .mat-mdc-text-field-wrapper{padding-left:16px;padding-right:16px}[dir=rtl] .mat-mdc-form-field-has-icon-suffix .mat-mdc-text-field-wrapper{padding-left:0}[dir=rtl] .mat-mdc-form-field-has-icon-prefix .mat-mdc-text-field-wrapper{padding-right:0}.mat-form-field-disabled .mdc-text-field__input::placeholder{color:var(--mat-form-field-disabled-input-text-placeholder-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-form-field-disabled .mdc-text-field__input::-moz-placeholder{color:var(--mat-form-field-disabled-input-text-placeholder-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-form-field-disabled .mdc-text-field__input::-webkit-input-placeholder{color:var(--mat-form-field-disabled-input-text-placeholder-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-form-field-disabled .mdc-text-field__input:-ms-input-placeholder{color:var(--mat-form-field-disabled-input-text-placeholder-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-form-field-label-always-float .mdc-text-field__input::placeholder{transition-delay:40ms;transition-duration:110ms;opacity:1}.mat-mdc-text-field-wrapper .mat-mdc-form-field-infix .mat-mdc-floating-label{left:auto;right:auto}.mat-mdc-text-field-wrapper.mdc-text-field--outlined .mdc-text-field__input{display:inline-block}.mat-mdc-form-field .mat-mdc-text-field-wrapper.mdc-text-field .mdc-notched-outline__notch{padding-top:0}.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field .mdc-notched-outline__notch{border-left:1px solid rgba(0,0,0,0)}[dir=rtl] .mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field.mat-mdc-form-field .mdc-notched-outline__notch{border-left:none;border-right:1px solid rgba(0,0,0,0)}.mat-mdc-form-field-infix{min-height:var(--mat-form-field-container-height, 56px);padding-top:var(--mat-form-field-filled-with-label-container-padding-top, 24px);padding-bottom:var(--mat-form-field-filled-with-label-container-padding-bottom, 8px)}.mdc-text-field--outlined .mat-mdc-form-field-infix,.mdc-text-field--no-label .mat-mdc-form-field-infix{padding-top:var(--mat-form-field-container-vertical-padding, 16px);padding-bottom:var(--mat-form-field-container-vertical-padding, 16px)}.mat-mdc-text-field-wrapper .mat-mdc-form-field-flex .mat-mdc-floating-label{top:calc(var(--mat-form-field-container-height, 56px)/2)}.mdc-text-field--filled .mat-mdc-floating-label{display:var(--mat-form-field-filled-label-display, block)}.mat-mdc-text-field-wrapper.mdc-text-field--outlined .mdc-notched-outline--upgraded .mdc-floating-label--float-above{--mat-mdc-form-field-label-transform: translateY(calc(calc(6.75px + var(--mat-form-field-container-height, 56px) / 2) * -1)) scale(var(--mat-mdc-form-field-floating-label-scale, 0.75));transform:var(--mat-mdc-form-field-label-transform)}.mat-mdc-form-field-subscript-wrapper{box-sizing:border-box;width:100%;position:relative}.mat-mdc-form-field-hint-wrapper,.mat-mdc-form-field-error-wrapper{position:absolute;top:0;left:0;right:0;padding:0 16px}.mat-mdc-form-field-subscript-dynamic-size .mat-mdc-form-field-hint-wrapper,.mat-mdc-form-field-subscript-dynamic-size .mat-mdc-form-field-error-wrapper{position:static}.mat-mdc-form-field-bottom-align::before{content:"";display:inline-block;height:16px}.mat-mdc-form-field-bottom-align.mat-mdc-form-field-subscript-dynamic-size::before{content:unset}.mat-mdc-form-field-hint-end{order:1}.mat-mdc-form-field-hint-wrapper{display:flex}.mat-mdc-form-field-hint-spacer{flex:1 0 1em}.mat-mdc-form-field-error{display:block;color:var(--mat-form-field-error-text-color, var(--mat-sys-error))}.mat-mdc-form-field-subscript-wrapper,.mat-mdc-form-field-bottom-align::before{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;font-family:var(--mat-form-field-subscript-text-font, var(--mat-sys-body-small-font));line-height:var(--mat-form-field-subscript-text-line-height, var(--mat-sys-body-small-line-height));font-size:var(--mat-form-field-subscript-text-size, var(--mat-sys-body-small-size));letter-spacing:var(--mat-form-field-subscript-text-tracking, var(--mat-sys-body-small-tracking));font-weight:var(--mat-form-field-subscript-text-weight, var(--mat-sys-body-small-weight))}.mat-mdc-form-field-focus-overlay{top:0;left:0;right:0;bottom:0;position:absolute;opacity:0;pointer-events:none;background-color:var(--mat-form-field-state-layer-color, var(--mat-sys-on-surface))}.mat-mdc-text-field-wrapper:hover .mat-mdc-form-field-focus-overlay{opacity:var(--mat-form-field-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-mdc-form-field.mat-focused .mat-mdc-form-field-focus-overlay{opacity:var(--mat-form-field-focus-state-layer-opacity, 0)}select.mat-mdc-form-field-input-control{-moz-appearance:none;-webkit-appearance:none;background-color:rgba(0,0,0,0);display:inline-flex;box-sizing:border-box}select.mat-mdc-form-field-input-control:not(:disabled){cursor:pointer}select.mat-mdc-form-field-input-control:not(.mat-mdc-native-select-inline) option{color:var(--mat-form-field-select-option-text-color, var(--mat-sys-neutral10))}select.mat-mdc-form-field-input-control:not(.mat-mdc-native-select-inline) option:disabled{color:var(--mat-form-field-select-disabled-option-text-color, color-mix(in srgb, var(--mat-sys-neutral10) 38%, transparent))}.mat-mdc-form-field-type-mat-native-select .mat-mdc-form-field-infix::after{content:"";width:0;height:0;border-left:5px solid rgba(0,0,0,0);border-right:5px solid rgba(0,0,0,0);border-top:5px solid;position:absolute;right:0;top:50%;margin-top:-2.5px;pointer-events:none;color:var(--mat-form-field-enabled-select-arrow-color, var(--mat-sys-on-surface-variant))}[dir=rtl] .mat-mdc-form-field-type-mat-native-select .mat-mdc-form-field-infix::after{right:auto;left:0}.mat-mdc-form-field-type-mat-native-select.mat-focused .mat-mdc-form-field-infix::after{color:var(--mat-form-field-focus-select-arrow-color, var(--mat-sys-primary))}.mat-mdc-form-field-type-mat-native-select.mat-form-field-disabled .mat-mdc-form-field-infix::after{color:var(--mat-form-field-disabled-select-arrow-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-form-field-type-mat-native-select .mat-mdc-form-field-input-control{padding-right:15px}[dir=rtl] .mat-mdc-form-field-type-mat-native-select .mat-mdc-form-field-input-control{padding-right:0;padding-left:15px}@media(forced-colors: active){.mat-form-field-appearance-fill .mat-mdc-text-field-wrapper{outline:solid 1px}}@media(forced-colors: active){.mat-form-field-appearance-fill.mat-form-field-disabled .mat-mdc-text-field-wrapper{outline-color:GrayText}}@media(forced-colors: active){.mat-form-field-appearance-fill.mat-focused .mat-mdc-text-field-wrapper{outline:dashed 3px}}@media(forced-colors: active){.mat-mdc-form-field.mat-focused .mdc-notched-outline{border:dashed 3px}}.mat-mdc-form-field-input-control[type=date],.mat-mdc-form-field-input-control[type=datetime],.mat-mdc-form-field-input-control[type=datetime-local],.mat-mdc-form-field-input-control[type=month],.mat-mdc-form-field-input-control[type=week],.mat-mdc-form-field-input-control[type=time]{line-height:1}.mat-mdc-form-field-input-control::-webkit-datetime-edit{line-height:1;padding:0;margin-bottom:-2px}.mat-mdc-form-field{--mat-mdc-form-field-floating-label-scale: 0.75;display:inline-flex;flex-direction:column;min-width:0;text-align:left;-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;font-family:var(--mat-form-field-container-text-font, var(--mat-sys-body-large-font));line-height:var(--mat-form-field-container-text-line-height, var(--mat-sys-body-large-line-height));font-size:var(--mat-form-field-container-text-size, var(--mat-sys-body-large-size));letter-spacing:var(--mat-form-field-container-text-tracking, var(--mat-sys-body-large-tracking));font-weight:var(--mat-form-field-container-text-weight, var(--mat-sys-body-large-weight))}.mat-mdc-form-field .mdc-text-field--outlined .mdc-floating-label--float-above{font-size:calc(var(--mat-form-field-outlined-label-text-populated-size)*var(--mat-mdc-form-field-floating-label-scale))}.mat-mdc-form-field .mdc-text-field--outlined .mdc-notched-outline--upgraded .mdc-floating-label--float-above{font-size:var(--mat-form-field-outlined-label-text-populated-size)}[dir=rtl] .mat-mdc-form-field{text-align:right}.mat-mdc-form-field-flex{display:inline-flex;align-items:baseline;box-sizing:border-box;width:100%}.mat-mdc-text-field-wrapper{width:100%;z-index:0}.mat-mdc-form-field-icon-prefix,.mat-mdc-form-field-icon-suffix{align-self:center;line-height:0;pointer-events:auto;position:relative;z-index:1}.mat-mdc-form-field-icon-prefix>.mat-icon,.mat-mdc-form-field-icon-suffix>.mat-icon{padding:0 12px;box-sizing:content-box}.mat-mdc-form-field-icon-prefix{color:var(--mat-form-field-leading-icon-color, var(--mat-sys-on-surface-variant))}.mat-form-field-disabled .mat-mdc-form-field-icon-prefix{color:var(--mat-form-field-disabled-leading-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-form-field-icon-suffix{color:var(--mat-form-field-trailing-icon-color, var(--mat-sys-on-surface-variant))}.mat-form-field-disabled .mat-mdc-form-field-icon-suffix{color:var(--mat-form-field-disabled-trailing-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-form-field-invalid .mat-mdc-form-field-icon-suffix{color:var(--mat-form-field-error-trailing-icon-color, var(--mat-sys-error))}.mat-form-field-invalid:not(.mat-focused):not(.mat-form-field-disabled) .mat-mdc-text-field-wrapper:hover .mat-mdc-form-field-icon-suffix{color:var(--mat-form-field-error-hover-trailing-icon-color, var(--mat-sys-on-error-container))}.mat-form-field-invalid.mat-focused .mat-mdc-text-field-wrapper .mat-mdc-form-field-icon-suffix{color:var(--mat-form-field-error-focus-trailing-icon-color, var(--mat-sys-error))}.mat-mdc-form-field-icon-prefix,[dir=rtl] .mat-mdc-form-field-icon-suffix{padding:0 4px 0 0}.mat-mdc-form-field-icon-suffix,[dir=rtl] .mat-mdc-form-field-icon-prefix{padding:0 0 0 4px}.mat-mdc-form-field-subscript-wrapper .mat-icon,.mat-mdc-form-field label .mat-icon{width:1em;height:1em;font-size:inherit}.mat-mdc-form-field-infix{flex:auto;min-width:0;width:180px;position:relative;box-sizing:border-box}.mat-mdc-form-field-infix:has(textarea[cols]){width:auto}.mat-mdc-form-field .mdc-notched-outline__notch{margin-left:-1px;-webkit-clip-path:inset(-9em -999em -9em 1px);clip-path:inset(-9em -999em -9em 1px)}[dir=rtl] .mat-mdc-form-field .mdc-notched-outline__notch{margin-left:0;margin-right:-1px;-webkit-clip-path:inset(-9em 1px -9em -999em);clip-path:inset(-9em 1px -9em -999em)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-floating-label{transition:transform 150ms cubic-bezier(0.4, 0, 0.2, 1),color 150ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field__input{transition:opacity 150ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field__input::placeholder{transition:opacity 67ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field__input::-moz-placeholder{transition:opacity 67ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field__input::-webkit-input-placeholder{transition:opacity 67ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field__input:-ms-input-placeholder{transition:opacity 67ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--no-label .mdc-text-field__input::placeholder,.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--focused .mdc-text-field__input::placeholder{transition-delay:40ms;transition-duration:110ms}.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--no-label .mdc-text-field__input::-moz-placeholder,.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--focused .mdc-text-field__input::-moz-placeholder{transition-delay:40ms;transition-duration:110ms}.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--no-label .mdc-text-field__input::-webkit-input-placeholder,.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--focused .mdc-text-field__input::-webkit-input-placeholder{transition-delay:40ms;transition-duration:110ms}.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--no-label .mdc-text-field__input:-ms-input-placeholder,.mat-mdc-form-field:not(.mat-form-field-no-animations).mdc-text-field--focused .mdc-text-field__input:-ms-input-placeholder{transition-delay:40ms;transition-duration:110ms}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-text-field--filled:not(.mdc-ripple-upgraded):focus .mdc-text-field__ripple::before{transition-duration:75ms}.mat-mdc-form-field:not(.mat-form-field-no-animations) .mdc-line-ripple::after{transition:transform 180ms cubic-bezier(0.4, 0, 0.2, 1),opacity 180ms cubic-bezier(0.4, 0, 0.2, 1)}.mdc-notched-outline .mdc-floating-label{max-width:calc(100% + 1px)}.mdc-notched-outline--upgraded .mdc-floating-label--float-above{max-width:calc(133.3333333333% + 1px)}'],encapsulation:2,data:{animation:[S4A.transitionMessages]},changeDetection:0})}return t})(),y0=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,DB,it]})}return t})();var F4A=["trigger"],N4A=["panel"],_4A=[[["mat-select-trigger"]],"*"],G4A=["mat-select-trigger","*"];function U4A(t,e){if(t&1&&(S(0,"span",4),AA(1),F()),t&2){let A=O();G(),Gt(A.placeholder)}}function K4A(t,e){t&1&&Fe(0)}function Y4A(t,e){if(t&1&&(S(0,"span",11),AA(1),F()),t&2){let A=O(2);G(),Gt(A.triggerValue)}}function J4A(t,e){if(t&1&&(S(0,"span",5),_A(1,K4A,1,0)(2,Y4A,2,1,"span",11),F()),t&2){let A=O();G(),GA(A.customTrigger?1:2)}}function T4A(t,e){if(t&1){let A=be();S(0,"div",12,1),hA("@transformPanel.done",function(n){RA(A);let o=O();return xA(o._panelDoneAnimatingStream.next(n.toState))})("keydown",function(n){RA(A);let o=O();return xA(o._handleKeydown(n))}),Fe(2,1),F()}if(t&2){let A=O();nH("mat-mdc-select-panel mdc-menu-surface mdc-menu-surface--open ",A._getPanelTheme(),""),yA("ngClass",A.panelClass)("@transformPanel","showing"),Ne("id",A.id+"-panel")("aria-multiselectable",A.multiple)("aria-label",A.ariaLabel||null)("aria-labelledby",A._getPanelAriaLabelledby())}}var H4A={transformPanelWrap:sc("transformPanelWrap",[Vr("* => void",Hk("@transformPanel",[Tk()],{optional:!0}))]),transformPanel:sc("transformPanel",[js("void",po({opacity:0,transform:"scale(1, 0.8)"})),Vr("void => showing",ss("120ms cubic-bezier(0, 0, 0.2, 1)",po({opacity:1,transform:"scale(1, 1)"}))),Vr("* => void",ss("100ms linear",po({opacity:0})))])};var kP=new dA("mat-select-scroll-strategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.reposition()}});function z4A(t){return()=>t.scrollStrategies.reposition()}var O4A=new dA("MAT_SELECT_CONFIG"),P4A={provide:kP,deps:[nr],useFactory:z4A},j4A=new dA("MatSelectTrigger"),zk=class{source;value;constructor(e,A){this.source=e,this.value=A}},NB=(()=>{class t{_viewportRuler=f(Mc);_changeDetectorRef=f(It);_elementRef=f(ee);_dir=f(mo,{optional:!0});_idGenerator=f(sn);_parentFormField=f(Gu,{optional:!0});ngControl=f(Ac,{self:!0,optional:!0});_liveAnnouncer=f(o8);_defaultOptions=f(O4A,{optional:!0});_initialized=new OA;options;optionGroups;customTrigger;_positions=[{originX:"start",originY:"bottom",overlayX:"start",overlayY:"top"},{originX:"end",originY:"bottom",overlayX:"end",overlayY:"top"},{originX:"start",originY:"top",overlayX:"start",overlayY:"bottom",panelClass:"mat-mdc-select-panel-above"},{originX:"end",originY:"top",overlayX:"end",overlayY:"bottom",panelClass:"mat-mdc-select-panel-above"}];_scrollOptionIntoView(A){let i=this.options.toArray()[A];if(i){let n=this.panel.nativeElement,o=AP(A,this.options,this.optionGroups),r=i._getHostElement();A===0&&o===1?n.scrollTop=0:n.scrollTop=eP(r.offsetTop,r.offsetHeight,n.scrollTop,n.offsetHeight)}}_positioningSettled(){this._scrollOptionIntoView(this._keyManager.activeItemIndex||0)}_getChangeEvent(A){return new zk(this,A)}_scrollStrategyFactory=f(kP);_panelOpen=!1;_compareWith=(A,i)=>A===i;_uid=this._idGenerator.getId("mat-select-");_triggerAriaLabelledBy=null;_previousControl;_destroy=new OA;_errorStateTracker;stateChanges=new OA;disableAutomaticLabeling=!0;userAriaDescribedBy;_selectionModel;_keyManager;_preferredOverlayOrigin;_overlayWidth;_onChange=()=>{};_onTouched=()=>{};_valueId=this._idGenerator.getId("mat-select-value-");_panelDoneAnimatingStream=new OA;_scrollStrategy;_overlayPanelClass=this._defaultOptions?.overlayPanelClass||"";get focused(){return this._focused||this._panelOpen}_focused=!1;controlType="mat-select";trigger;panel;_overlayDir;panelClass;disabled=!1;disableRipple=!1;tabIndex=0;get hideSingleSelectionIndicator(){return this._hideSingleSelectionIndicator}set hideSingleSelectionIndicator(A){this._hideSingleSelectionIndicator=A,this._syncParentProperties()}_hideSingleSelectionIndicator=this._defaultOptions?.hideSingleSelectionIndicator??!1;get placeholder(){return this._placeholder}set placeholder(A){this._placeholder=A,this.stateChanges.next()}_placeholder;get required(){return this._required??this.ngControl?.control?.hasValidator($a.required)??!1}set required(A){this._required=A,this.stateChanges.next()}_required;get multiple(){return this._multiple}set multiple(A){this._selectionModel,this._multiple=A}_multiple=!1;disableOptionCentering=this._defaultOptions?.disableOptionCentering??!1;get compareWith(){return this._compareWith}set compareWith(A){this._compareWith=A,this._selectionModel&&this._initializeSelection()}get value(){return this._value}set value(A){this._assignValue(A)&&this._onChange(A)}_value;ariaLabel="";ariaLabelledby;get errorStateMatcher(){return this._errorStateTracker.matcher}set errorStateMatcher(A){this._errorStateTracker.matcher=A}typeaheadDebounceInterval;sortComparator;get id(){return this._id}set id(A){this._id=A||this._uid,this.stateChanges.next()}_id;get errorState(){return this._errorStateTracker.errorState}set errorState(A){this._errorStateTracker.errorState=A}panelWidth=this._defaultOptions&&typeof this._defaultOptions.panelWidth<"u"?this._defaultOptions.panelWidth:"auto";canSelectNullableOptions=this._defaultOptions?.canSelectNullableOptions??!1;optionSelectionChanges=jl(()=>{let A=this.options;return A?A.changes.pipe(Pn(A),jn(()=>zn(...A.map(i=>i.onSelectionChange)))):this._initialized.pipe(jn(()=>this.optionSelectionChanges))});openedChange=new WA;_openedStream=this.openedChange.pipe(kt(A=>A),je(()=>{}));_closedStream=this.openedChange.pipe(kt(A=>!A),je(()=>{}));selectionChange=new WA;valueChange=new WA;constructor(){let A=f(bB),i=f(su,{optional:!0}),n=f(GI,{optional:!0}),o=f(new wr("tabindex"),{optional:!0});this.ngControl&&(this.ngControl.valueAccessor=this),this._defaultOptions?.typeaheadDebounceInterval!=null&&(this.typeaheadDebounceInterval=this._defaultOptions.typeaheadDebounceInterval),this._errorStateTracker=new $I(A,this.ngControl,n,i,this.stateChanges),this._scrollStrategy=this._scrollStrategyFactory(),this.tabIndex=o==null?0:parseInt(o)||0,this.id=this.id}ngOnInit(){this._selectionModel=new K2(this.multiple),this.stateChanges.next(),this._panelDoneAnimatingStream.pipe(tl(),St(this._destroy)).subscribe(()=>this._panelDoneAnimating(this.panelOpen)),this._viewportRuler.change().pipe(St(this._destroy)).subscribe(()=>{this.panelOpen&&(this._overlayWidth=this._getOverlayWidth(this._preferredOverlayOrigin),this._changeDetectorRef.detectChanges())})}ngAfterContentInit(){this._initialized.next(),this._initialized.complete(),this._initKeyManager(),this._selectionModel.changed.pipe(St(this._destroy)).subscribe(A=>{A.added.forEach(i=>i.select()),A.removed.forEach(i=>i.deselect())}),this.options.changes.pipe(Pn(null),St(this._destroy)).subscribe(()=>{this._resetOptions(),this._initializeSelection()})}ngDoCheck(){let A=this._getTriggerAriaLabelledby(),i=this.ngControl;if(A!==this._triggerAriaLabelledBy){let n=this._elementRef.nativeElement;this._triggerAriaLabelledBy=A,A?n.setAttribute("aria-labelledby",A):n.removeAttribute("aria-labelledby")}i&&(this._previousControl!==i.control&&(this._previousControl!==void 0&&i.disabled!==null&&i.disabled!==this.disabled&&(this.disabled=i.disabled),this._previousControl=i.control),this.updateErrorState())}ngOnChanges(A){(A.disabled||A.userAriaDescribedBy)&&this.stateChanges.next(),A.typeaheadDebounceInterval&&this._keyManager&&this._keyManager.withTypeAhead(this.typeaheadDebounceInterval)}ngOnDestroy(){this._keyManager?.destroy(),this._destroy.next(),this._destroy.complete(),this.stateChanges.complete(),this._clearFromModal()}toggle(){this.panelOpen?this.close():this.open()}open(){this._canOpen()&&(this._parentFormField&&(this._preferredOverlayOrigin=this._parentFormField.getConnectedOverlayOrigin()),this._overlayWidth=this._getOverlayWidth(this._preferredOverlayOrigin),this._applyModalPanelOwnership(),this._panelOpen=!0,this._keyManager.withHorizontalOrientation(null),this._highlightCorrectOption(),this._changeDetectorRef.markForCheck(),this.stateChanges.next())}_trackedModal=null;_applyModalPanelOwnership(){let A=this._elementRef.nativeElement.closest('body > .cdk-overlay-container [aria-modal="true"]');if(!A)return;let i=`${this.id}-panel`;this._trackedModal&&i8(this._trackedModal,"aria-owns",i),Ek(A,"aria-owns",i),this._trackedModal=A}_clearFromModal(){if(!this._trackedModal)return;let A=`${this.id}-panel`;i8(this._trackedModal,"aria-owns",A),this._trackedModal=null}close(){this._panelOpen&&(this._panelOpen=!1,this._keyManager.withHorizontalOrientation(this._isRtl()?"rtl":"ltr"),this._changeDetectorRef.markForCheck(),this._onTouched(),this.stateChanges.next())}writeValue(A){this._assignValue(A)}registerOnChange(A){this._onChange=A}registerOnTouched(A){this._onTouched=A}setDisabledState(A){this.disabled=A,this._changeDetectorRef.markForCheck(),this.stateChanges.next()}get panelOpen(){return this._panelOpen}get selected(){return this.multiple?this._selectionModel?.selected||[]:this._selectionModel?.selected[0]}get triggerValue(){if(this.empty)return"";if(this._multiple){let A=this._selectionModel.selected.map(i=>i.viewValue);return this._isRtl()&&A.reverse(),A.join(", ")}return this._selectionModel.selected[0].viewValue}updateErrorState(){this._errorStateTracker.updateErrorState()}_isRtl(){return this._dir?this._dir.value==="rtl":!1}_handleKeydown(A){this.disabled||(this.panelOpen?this._handleOpenKeydown(A):this._handleClosedKeydown(A))}_handleClosedKeydown(A){let i=A.keyCode,n=i===40||i===38||i===37||i===39,o=i===13||i===32,r=this._keyManager;if(!r.isTyping()&&o&&!ir(A)||(this.multiple||A.altKey)&&n)A.preventDefault(),this.open();else if(!this.multiple){let s=this.selected;r.onKeydown(A);let a=this.selected;a&&s!==a&&this._liveAnnouncer.announce(a.viewValue,1e4)}}_handleOpenKeydown(A){let i=this._keyManager,n=A.keyCode,o=n===40||n===38,r=i.isTyping();if(o&&A.altKey)A.preventDefault(),this.close();else if(!r&&(n===13||n===32)&&i.activeItem&&!ir(A))A.preventDefault(),i.activeItem._selectViaInteraction();else if(!r&&this._multiple&&n===65&&A.ctrlKey){A.preventDefault();let s=this.options.some(a=>!a.disabled&&!a.selected);this.options.forEach(a=>{a.disabled||(s?a.select():a.deselect())})}else{let s=i.activeItemIndex;i.onKeydown(A),this._multiple&&o&&A.shiftKey&&i.activeItem&&i.activeItemIndex!==s&&i.activeItem._selectViaInteraction()}}_onFocus(){this.disabled||(this._focused=!0,this.stateChanges.next())}_onBlur(){this._focused=!1,this._keyManager?.cancelTypeahead(),!this.disabled&&!this.panelOpen&&(this._onTouched(),this._changeDetectorRef.markForCheck(),this.stateChanges.next())}_onAttached(){this._overlayDir.positionChange.pipe(On(1)).subscribe(()=>{this._changeDetectorRef.detectChanges(),this._positioningSettled()})}_getPanelTheme(){return this._parentFormField?`mat-${this._parentFormField.color}`:""}get empty(){return!this._selectionModel||this._selectionModel.isEmpty()}_initializeSelection(){Promise.resolve().then(()=>{this.ngControl&&(this._value=this.ngControl.value),this._setSelectionByValue(this._value),this.stateChanges.next()})}_setSelectionByValue(A){if(this.options.forEach(i=>i.setInactiveStyles()),this._selectionModel.clear(),this.multiple&&A)Array.isArray(A),A.forEach(i=>this._selectOptionByValue(i)),this._sortValues();else{let i=this._selectOptionByValue(A);i?this._keyManager.updateActiveItem(i):this.panelOpen||this._keyManager.updateActiveItem(-1)}this._changeDetectorRef.markForCheck()}_selectOptionByValue(A){let i=this.options.find(n=>{if(this._selectionModel.isSelected(n))return!1;try{return(n.value!=null||this.canSelectNullableOptions)&&this._compareWith(n.value,A)}catch{return!1}});return i&&this._selectionModel.select(i),i}_assignValue(A){return A!==this._value||this._multiple&&Array.isArray(A)?(this.options&&this._setSelectionByValue(A),this._value=A,!0):!1}_skipPredicate=A=>this.panelOpen?!1:A.disabled;_getOverlayWidth(A){return this.panelWidth==="auto"?(A instanceof Nu?A.elementRef:A||this._elementRef).nativeElement.getBoundingClientRect().width:this.panelWidth===null?"":this.panelWidth}_syncParentProperties(){if(this.options)for(let A of this.options)A._changeDetectorRef.markForCheck()}_initKeyManager(){this._keyManager=new t8(this.options).withTypeAhead(this.typeaheadDebounceInterval).withVerticalOrientation().withHorizontalOrientation(this._isRtl()?"rtl":"ltr").withHomeAndEnd().withPageUpDown().withAllowedModifierKeys(["shiftKey"]).skipPredicate(this._skipPredicate),this._keyManager.tabOut.subscribe(()=>{this.panelOpen&&(!this.multiple&&this._keyManager.activeItem&&this._keyManager.activeItem._selectViaInteraction(),this.focus(),this.close())}),this._keyManager.change.subscribe(()=>{this._panelOpen&&this.panel?this._scrollOptionIntoView(this._keyManager.activeItemIndex||0):!this._panelOpen&&!this.multiple&&this._keyManager.activeItem&&this._keyManager.activeItem._selectViaInteraction()})}_resetOptions(){let A=zn(this.options.changes,this._destroy);this.optionSelectionChanges.pipe(St(A)).subscribe(i=>{this._onSelect(i.source,i.isUserInput),i.isUserInput&&!this.multiple&&this._panelOpen&&(this.close(),this.focus())}),zn(...this.options.map(i=>i._stateChanges)).pipe(St(A)).subscribe(()=>{this._changeDetectorRef.detectChanges(),this.stateChanges.next()})}_onSelect(A,i){let n=this._selectionModel.isSelected(A);!this.canSelectNullableOptions&&A.value==null&&!this._multiple?(A.deselect(),this._selectionModel.clear(),this.value!=null&&this._propagateChanges(A.value)):(n!==A.selected&&(A.selected?this._selectionModel.select(A):this._selectionModel.deselect(A)),i&&this._keyManager.setActiveItem(A),this.multiple&&(this._sortValues(),i&&this.focus())),n!==this._selectionModel.isSelected(A)&&this._propagateChanges(),this.stateChanges.next()}_sortValues(){if(this.multiple){let A=this.options.toArray();this._selectionModel.sort((i,n)=>this.sortComparator?this.sortComparator(i,n,A):A.indexOf(i)-A.indexOf(n)),this.stateChanges.next()}}_propagateChanges(A){let i;this.multiple?i=this.selected.map(n=>n.value):i=this.selected?this.selected.value:A,this._value=i,this.valueChange.emit(i),this._onChange(i),this.selectionChange.emit(this._getChangeEvent(i)),this._changeDetectorRef.markForCheck()}_highlightCorrectOption(){if(this._keyManager)if(this.empty){let A=-1;for(let i=0;i<this.options.length;i++)if(!this.options.get(i).disabled){A=i;break}this._keyManager.setActiveItem(A)}else this._keyManager.setActiveItem(this._selectionModel.selected[0])}_canOpen(){return!this._panelOpen&&!this.disabled&&this.options?.length>0}focus(A){this._elementRef.nativeElement.focus(A)}_getPanelAriaLabelledby(){if(this.ariaLabel)return null;let A=this._parentFormField?.getLabelId()||null,i=A?A+" ":"";return this.ariaLabelledby?i+this.ariaLabelledby:A}_getAriaActiveDescendant(){return this.panelOpen&&this._keyManager&&this._keyManager.activeItem?this._keyManager.activeItem.id:null}_getTriggerAriaLabelledby(){if(this.ariaLabel)return null;let A=this._parentFormField?.getLabelId(),i=(A?A+" ":"")+this._valueId;return this.ariaLabelledby&&(i+=" "+this.ariaLabelledby),i}_panelDoneAnimating(A){this.openedChange.emit(A)}setDescribedByIds(A){A.length?this._elementRef.nativeElement.setAttribute("aria-describedby",A.join(" ")):this._elementRef.nativeElement.removeAttribute("aria-describedby")}onContainerClick(){this.focus(),this.open()}get shouldLabelFloat(){return this.panelOpen||!this.empty||this.focused&&!!this.placeholder}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-select"]],contentQueries:function(i,n,o){if(i&1&&(ci(o,j4A,5),ci(o,U2,5),ci(o,vk,5)),i&2){let r;XA(r=$A())&&(n.customTrigger=r.first),XA(r=$A())&&(n.options=r),XA(r=$A())&&(n.optionGroups=r)}},viewQuery:function(i,n){if(i&1&&(Te(F4A,5),Te(N4A,5),Te(Yk,5)),i&2){let o;XA(o=$A())&&(n.trigger=o.first),XA(o=$A())&&(n.panel=o.first),XA(o=$A())&&(n._overlayDir=o.first)}},hostAttrs:["role","combobox","aria-haspopup","listbox",1,"mat-mdc-select"],hostVars:19,hostBindings:function(i,n){i&1&&hA("keydown",function(r){return n._handleKeydown(r)})("focus",function(){return n._onFocus()})("blur",function(){return n._onBlur()}),i&2&&(Ne("id",n.id)("tabindex",n.disabled?-1:n.tabIndex)("aria-controls",n.panelOpen?n.id+"-panel":null)("aria-expanded",n.panelOpen)("aria-label",n.ariaLabel||null)("aria-required",n.required.toString())("aria-disabled",n.disabled.toString())("aria-invalid",n.errorState)("aria-activedescendant",n._getAriaActiveDescendant()),ue("mat-mdc-select-disabled",n.disabled)("mat-mdc-select-invalid",n.errorState)("mat-mdc-select-required",n.required)("mat-mdc-select-empty",n.empty)("mat-mdc-select-multiple",n.multiple))},inputs:{userAriaDescribedBy:[0,"aria-describedby","userAriaDescribedBy"],panelClass:"panelClass",disabled:[2,"disabled","disabled",ie],disableRipple:[2,"disableRipple","disableRipple",ie],tabIndex:[2,"tabIndex","tabIndex",A=>A==null?0:zi(A)],hideSingleSelectionIndicator:[2,"hideSingleSelectionIndicator","hideSingleSelectionIndicator",ie],placeholder:"placeholder",required:[2,"required","required",ie],multiple:[2,"multiple","multiple",ie],disableOptionCentering:[2,"disableOptionCentering","disableOptionCentering",ie],compareWith:"compareWith",value:"value",ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],errorStateMatcher:"errorStateMatcher",typeaheadDebounceInterval:[2,"typeaheadDebounceInterval","typeaheadDebounceInterval",zi],sortComparator:"sortComparator",id:"id",panelWidth:"panelWidth",canSelectNullableOptions:[2,"canSelectNullableOptions","canSelectNullableOptions",ie]},outputs:{openedChange:"openedChange",_openedStream:"opened",_closedStream:"closed",selectionChange:"selectionChange",valueChange:"valueChange"},exportAs:["matSelect"],features:[ht([{provide:_u,useExisting:t},{provide:yk,useExisting:t}]),ti],ngContentSelectors:G4A,decls:11,vars:8,consts:[["fallbackOverlayOrigin","cdkOverlayOrigin","trigger",""],["panel",""],["cdk-overlay-origin","",1,"mat-mdc-select-trigger",3,"click"],[1,"mat-mdc-select-value"],[1,"mat-mdc-select-placeholder","mat-mdc-select-min-line"],[1,"mat-mdc-select-value-text"],[1,"mat-mdc-select-arrow-wrapper"],[1,"mat-mdc-select-arrow"],["viewBox","0 0 24 24","width","24px","height","24px","focusable","false","aria-hidden","true"],["d","M7 10l5 5 5-5z"],["cdk-connected-overlay","","cdkConnectedOverlayLockPosition","","cdkConnectedOverlayHasBackdrop","","cdkConnectedOverlayBackdropClass","cdk-overlay-transparent-backdrop",3,"backdropClick","attach","detach","cdkConnectedOverlayPanelClass","cdkConnectedOverlayScrollStrategy","cdkConnectedOverlayOrigin","cdkConnectedOverlayOpen","cdkConnectedOverlayPositions","cdkConnectedOverlayWidth"],[1,"mat-mdc-select-min-line"],["role","listbox","tabindex","-1",3,"keydown","ngClass"]],template:function(i,n){if(i&1){let o=be();jt(_4A),S(0,"div",2,0),hA("click",function(){return RA(o),xA(n.open())}),S(3,"div",3),_A(4,U4A,2,1,"span",4)(5,J4A,3,1,"span",5),F(),S(6,"div",6)(7,"div",7),ar(),S(8,"svg",8),JA(9,"path",9),F()()()(),_A(10,T4A,3,9,"ng-template",10),hA("backdropClick",function(){return RA(o),xA(n.close())})("attach",function(){return RA(o),xA(n._onAttached())})("detach",function(){return RA(o),xA(n.close())})}if(i&2){let o=cr(1);G(3),Ne("id",n._valueId),G(),GA(n.empty?4:5),G(6),yA("cdkConnectedOverlayPanelClass",n._overlayPanelClass)("cdkConnectedOverlayScrollStrategy",n._scrollStrategy)("cdkConnectedOverlayOrigin",n._preferredOverlayOrigin||o)("cdkConnectedOverlayOpen",n.panelOpen)("cdkConnectedOverlayPositions",n._positions)("cdkConnectedOverlayWidth",n._overlayWidth)}},dependencies:[Nu,Yk,Xa],styles:['.mat-mdc-select{display:inline-block;width:100%;outline:none;-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;color:var(--mat-select-enabled-trigger-text-color, var(--mat-sys-on-surface));font-family:var(--mat-select-trigger-text-font, var(--mat-sys-body-large-font));line-height:var(--mat-select-trigger-text-line-height, var(--mat-sys-body-large-line-height));font-size:var(--mat-select-trigger-text-size, var(--mat-sys-body-large-size));font-weight:var(--mat-select-trigger-text-weight, var(--mat-sys-body-large-weight));letter-spacing:var(--mat-select-trigger-text-tracking, var(--mat-sys-body-large-tracking))}div.mat-mdc-select-panel{box-shadow:var(--mat-select-container-elevation-shadow, 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12))}.mat-mdc-select-disabled{color:var(--mat-select-disabled-trigger-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-select-disabled .mat-mdc-select-placeholder{color:var(--mat-select-disabled-trigger-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-select-trigger{display:inline-flex;align-items:center;cursor:pointer;position:relative;box-sizing:border-box;width:100%}.mat-mdc-select-disabled .mat-mdc-select-trigger{-webkit-user-select:none;user-select:none;cursor:default}.mat-mdc-select-value{width:100%;overflow:hidden;text-overflow:ellipsis;white-space:nowrap}.mat-mdc-select-value-text{white-space:nowrap;overflow:hidden;text-overflow:ellipsis}.mat-mdc-select-arrow-wrapper{height:24px;flex-shrink:0;display:inline-flex;align-items:center}.mat-form-field-appearance-fill .mdc-text-field--no-label .mat-mdc-select-arrow-wrapper{transform:none}.mat-mdc-form-field .mat-mdc-select.mat-mdc-select-invalid .mat-mdc-select-arrow,.mat-form-field-invalid:not(.mat-form-field-disabled) .mat-mdc-form-field-infix::after{color:var(--mat-select-invalid-arrow-color, var(--mat-sys-error))}.mat-mdc-select-arrow{width:10px;height:5px;position:relative;color:var(--mat-select-enabled-arrow-color, var(--mat-sys-on-surface-variant))}.mat-mdc-form-field.mat-focused .mat-mdc-select-arrow{color:var(--mat-select-focused-arrow-color, var(--mat-sys-primary))}.mat-mdc-form-field .mat-mdc-select.mat-mdc-select-disabled .mat-mdc-select-arrow{color:var(--mat-select-disabled-arrow-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-select-arrow svg{fill:currentColor;position:absolute;top:50%;left:50%;transform:translate(-50%, -50%)}@media(forced-colors: active){.mat-mdc-select-arrow svg{fill:CanvasText}.mat-mdc-select-disabled .mat-mdc-select-arrow svg{fill:GrayText}}div.mat-mdc-select-panel{width:100%;max-height:275px;outline:0;overflow:auto;padding:8px 0;border-radius:4px;box-sizing:border-box;position:static;background-color:var(--mat-select-panel-background-color, var(--mat-sys-surface-container))}@media(forced-colors: active){div.mat-mdc-select-panel{outline:solid 1px}}.cdk-overlay-pane:not(.mat-mdc-select-panel-above) div.mat-mdc-select-panel{border-top-left-radius:0;border-top-right-radius:0;transform-origin:top center}.mat-mdc-select-panel-above div.mat-mdc-select-panel{border-bottom-left-radius:0;border-bottom-right-radius:0;transform-origin:bottom center}div.mat-mdc-select-panel .mat-mdc-option{--mdc-list-list-item-container-color: var(--mat-select-panel-background-color)}.mat-mdc-select-placeholder{transition:color 400ms 133.3333333333ms cubic-bezier(0.25, 0.8, 0.25, 1);color:var(--mat-select-placeholder-text-color, var(--mat-sys-on-surface-variant))}.mat-form-field-no-animations .mat-mdc-select-placeholder,._mat-animation-noopable .mat-mdc-select-placeholder{transition:none}.mat-form-field-hide-placeholder .mat-mdc-select-placeholder{color:rgba(0,0,0,0);-webkit-text-fill-color:rgba(0,0,0,0);transition:none;display:block}.mat-mdc-form-field-type-mat-select:not(.mat-form-field-disabled) .mat-mdc-text-field-wrapper{cursor:pointer}.mat-mdc-form-field-type-mat-select.mat-form-field-appearance-fill .mat-mdc-floating-label{max-width:calc(100% - 18px)}.mat-mdc-form-field-type-mat-select.mat-form-field-appearance-fill .mdc-floating-label--float-above{max-width:calc(100%/0.75 - 24px)}.mat-mdc-form-field-type-mat-select.mat-form-field-appearance-outline .mdc-notched-outline__notch{max-width:calc(100% - 60px)}.mat-mdc-form-field-type-mat-select.mat-form-field-appearance-outline .mdc-text-field--label-floating .mdc-notched-outline__notch{max-width:calc(100% - 24px)}.mat-mdc-select-min-line:empty::before{content:" ";white-space:pre;width:1px;display:inline-block;visibility:hidden}.mat-form-field-appearance-fill .mat-mdc-select-arrow-wrapper{transform:var(--mat-select-arrow-transform, translateY(-8px))}'],encapsulation:2,data:{animation:[H4A.transformPanel]},changeDetection:0})}return t})();var u8=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[P4A],imports:[El,bk,it,Cl,y0,bk,it]})}return t})();var q4A=["tooltip"],LP=20;var FP=new dA("mat-tooltip-scroll-strategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.reposition({scrollThrottle:LP})}});function V4A(t){return()=>t.scrollStrategies.reposition({scrollThrottle:LP})}var Z4A={provide:FP,deps:[nr],useFactory:V4A};function W4A(){return{showDelay:0,hideDelay:0,touchendHideDelay:1500}}var X4A=new dA("mat-tooltip-default-options",{providedIn:"root",factory:W4A});var RP="tooltip-panel",xP=bc({passive:!0}),$4A=8,A3A=8,e3A=24,t3A=200,_B=(()=>{class t{_elementRef=f(ee);_ngZone=f(Qe);_platform=f(Ii);_ariaDescriber=f(HO);_focusMonitor=f(dr);_dir=f(mo);_injector=f(Rt);_defaultOptions=f(X4A,{optional:!0});_overlayRef;_tooltipInstance;_portal;_position="below";_positionAtOrigin=!1;_disabled=!1;_tooltipClass;_viewInitialized=!1;_pointerExitEventsInitialized=!1;_tooltipComponent=i3A;_viewportMargin=8;_currentPosition;_cssClassPrefix="mat-mdc";_ariaDescriptionPending;_dirSubscribed=!1;get position(){return this._position}set position(A){A!==this._position&&(this._position=A,this._overlayRef&&(this._updatePosition(this._overlayRef),this._tooltipInstance?.show(0),this._overlayRef.updatePosition()))}get positionAtOrigin(){return this._positionAtOrigin}set positionAtOrigin(A){this._positionAtOrigin=Xo(A),this._detach(),this._overlayRef=null}get disabled(){return this._disabled}set disabled(A){let i=Xo(A);this._disabled!==i&&(this._disabled=i,i?this.hide(0):this._setupPointerEnterEventsIfNeeded(),this._syncAriaDescription(this.message))}get showDelay(){return this._showDelay}set showDelay(A){this._showDelay=zs(A)}_showDelay;get hideDelay(){return this._hideDelay}set hideDelay(A){this._hideDelay=zs(A),this._tooltipInstance&&(this._tooltipInstance._mouseLeaveHideDelay=this._hideDelay)}_hideDelay;touchGestures="auto";get message(){return this._message}set message(A){let i=this._message;this._message=A!=null?String(A).trim():"",!this._message&&this._isTooltipVisible()?this.hide(0):(this._setupPointerEnterEventsIfNeeded(),this._updateTooltipMessage()),this._syncAriaDescription(i)}_message="";get tooltipClass(){return this._tooltipClass}set tooltipClass(A){this._tooltipClass=A,this._tooltipInstance&&this._setTooltipClass(this._tooltipClass)}_passiveListeners=[];_touchstartTimeout=null;_destroyed=new OA;_isDestroyed=!1;constructor(){let A=this._defaultOptions;A&&(this._showDelay=A.showDelay,this._hideDelay=A.hideDelay,A.position&&(this.position=A.position),A.positionAtOrigin&&(this.positionAtOrigin=A.positionAtOrigin),A.touchGestures&&(this.touchGestures=A.touchGestures),A.tooltipClass&&(this.tooltipClass=A.tooltipClass)),this._viewportMargin=$4A}ngAfterViewInit(){this._viewInitialized=!0,this._setupPointerEnterEventsIfNeeded(),this._focusMonitor.monitor(this._elementRef).pipe(St(this._destroyed)).subscribe(A=>{A?A==="keyboard"&&this._ngZone.run(()=>this.show()):this._ngZone.run(()=>this.hide(0))})}ngOnDestroy(){let A=this._elementRef.nativeElement;this._touchstartTimeout&&clearTimeout(this._touchstartTimeout),this._overlayRef&&(this._overlayRef.dispose(),this._tooltipInstance=null),this._passiveListeners.forEach(([i,n])=>{A.removeEventListener(i,n,xP)}),this._passiveListeners.length=0,this._destroyed.next(),this._destroyed.complete(),this._isDestroyed=!0,this._ariaDescriber.removeDescription(A,this.message,"tooltip"),this._focusMonitor.stopMonitoring(A)}show(A=this.showDelay,i){if(this.disabled||!this.message||this._isTooltipVisible()){this._tooltipInstance?._cancelPendingAnimations();return}let n=this._createOverlay(i);this._detach(),this._portal=this._portal||new dl(this._tooltipComponent,this._injector.get(Nn));let o=this._tooltipInstance=n.attach(this._portal).instance;o._triggerElement=this._elementRef.nativeElement,o._mouseLeaveHideDelay=this._hideDelay,o.afterHidden().pipe(St(this._destroyed)).subscribe(()=>this._detach()),this._setTooltipClass(this._tooltipClass),this._updateTooltipMessage(),o.show(A)}hide(A=this.hideDelay){let i=this._tooltipInstance;i&&(i.isVisible()?i.hide(A):(i._cancelPendingAnimations(),this._detach()))}toggle(A){this._isTooltipVisible()?this.hide():this.show(void 0,A)}_isTooltipVisible(){return!!this._tooltipInstance&&this._tooltipInstance.isVisible()}_createOverlay(A){if(this._overlayRef){let r=this._overlayRef.getConfig().positionStrategy;if((!this.positionAtOrigin||!A)&&r._origin instanceof ee)return this._overlayRef;this._detach()}let i=this._injector.get(Y2).getAncestorScrollContainers(this._elementRef),n=this._injector.get(nr),o=n.position().flexibleConnectedTo(this.positionAtOrigin?A||this._elementRef:this._elementRef).withTransformOriginOn(`.${this._cssClassPrefix}-tooltip`).withFlexibleDimensions(!1).withViewportMargin(this._viewportMargin).withScrollableContainers(i);return o.positionChanges.pipe(St(this._destroyed)).subscribe(r=>{this._updateCurrentPositionClass(r.connectionPair),this._tooltipInstance&&r.scrollableViewProperties.isOverlayClipped&&this._tooltipInstance.isVisible()&&this._ngZone.run(()=>this.hide(0))}),this._overlayRef=n.create({direction:this._dir,positionStrategy:o,panelClass:`${this._cssClassPrefix}-${RP}`,scrollStrategy:this._injector.get(FP)()}),this._updatePosition(this._overlayRef),this._overlayRef.detachments().pipe(St(this._destroyed)).subscribe(()=>this._detach()),this._overlayRef.outsidePointerEvents().pipe(St(this._destroyed)).subscribe(()=>this._tooltipInstance?._handleBodyInteraction()),this._overlayRef.keydownEvents().pipe(St(this._destroyed)).subscribe(r=>{this._isTooltipVisible()&&r.keyCode===27&&!ir(r)&&(r.preventDefault(),r.stopPropagation(),this._ngZone.run(()=>this.hide(0)))}),this._defaultOptions?.disableTooltipInteractivity&&this._overlayRef.addPanelClass(`${this._cssClassPrefix}-tooltip-panel-non-interactive`),this._dirSubscribed||(this._dirSubscribed=!0,this._dir.change.pipe(St(this._destroyed)).subscribe(()=>{this._overlayRef&&this._updatePosition(this._overlayRef)})),this._overlayRef}_detach(){this._overlayRef&&this._overlayRef.hasAttached()&&this._overlayRef.detach(),this._tooltipInstance=null}_updatePosition(A){let i=A.getConfig().positionStrategy,n=this._getOrigin(),o=this._getOverlayPosition();i.withPositions([this._addOffset(rA(rA({},n.main),o.main)),this._addOffset(rA(rA({},n.fallback),o.fallback))])}_addOffset(A){let i=A3A,n=!this._dir||this._dir.value=="ltr";return A.originY==="top"?A.offsetY=-i:A.originY==="bottom"?A.offsetY=i:A.originX==="start"?A.offsetX=n?-i:i:A.originX==="end"&&(A.offsetX=n?i:-i),A}_getOrigin(){let A=!this._dir||this._dir.value=="ltr",i=this.position,n;i=="above"||i=="below"?n={originX:"center",originY:i=="above"?"top":"bottom"}:i=="before"||i=="left"&&A||i=="right"&&!A?n={originX:"start",originY:"center"}:(i=="after"||i=="right"&&A||i=="left"&&!A)&&(n={originX:"end",originY:"center"});let{x:o,y:r}=this._invertPosition(n.originX,n.originY);return{main:n,fallback:{originX:o,originY:r}}}_getOverlayPosition(){let A=!this._dir||this._dir.value=="ltr",i=this.position,n;i=="above"?n={overlayX:"center",overlayY:"bottom"}:i=="below"?n={overlayX:"center",overlayY:"top"}:i=="before"||i=="left"&&A||i=="right"&&!A?n={overlayX:"end",overlayY:"center"}:(i=="after"||i=="right"&&A||i=="left"&&!A)&&(n={overlayX:"start",overlayY:"center"});let{x:o,y:r}=this._invertPosition(n.overlayX,n.overlayY);return{main:n,fallback:{overlayX:o,overlayY:r}}}_updateTooltipMessage(){this._tooltipInstance&&(this._tooltipInstance.message=this.message,this._tooltipInstance._markForCheck(),To(()=>{this._tooltipInstance&&this._overlayRef.updatePosition()},{injector:this._injector}))}_setTooltipClass(A){this._tooltipInstance&&(this._tooltipInstance.tooltipClass=A,this._tooltipInstance._markForCheck())}_invertPosition(A,i){return this.position==="above"||this.position==="below"?i==="top"?i="bottom":i==="bottom"&&(i="top"):A==="end"?A="start":A==="start"&&(A="end"),{x:A,y:i}}_updateCurrentPositionClass(A){let{overlayY:i,originX:n,originY:o}=A,r;if(i==="center"?this._dir&&this._dir.value==="rtl"?r=n==="end"?"left":"right":r=n==="start"?"left":"right":r=i==="bottom"&&o==="top"?"above":"below",r!==this._currentPosition){let s=this._overlayRef;if(s){let a=`${this._cssClassPrefix}-${RP}-`;s.removePanelClass(a+this._currentPosition),s.addPanelClass(a+r)}this._currentPosition=r}}_setupPointerEnterEventsIfNeeded(){this._disabled||!this.message||!this._viewInitialized||this._passiveListeners.length||(this._platformSupportsMouseEvents()?this._passiveListeners.push(["mouseenter",A=>{this._setupPointerExitEventsIfNeeded();let i;A.x!==void 0&&A.y!==void 0&&(i=A),this.show(void 0,i)}]):this.touchGestures!=="off"&&(this._disableNativeGesturesIfNecessary(),this._passiveListeners.push(["touchstart",A=>{let i=A.targetTouches?.[0],n=i?{x:i.clientX,y:i.clientY}:void 0;this._setupPointerExitEventsIfNeeded(),this._touchstartTimeout&&clearTimeout(this._touchstartTimeout);let o=500;this._touchstartTimeout=setTimeout(()=>{this._touchstartTimeout=null,this.show(void 0,n)},this._defaultOptions?.touchLongPressShowDelay??o)}])),this._addListeners(this._passiveListeners))}_setupPointerExitEventsIfNeeded(){if(this._pointerExitEventsInitialized)return;this._pointerExitEventsInitialized=!0;let A=[];if(this._platformSupportsMouseEvents())A.push(["mouseleave",i=>{let n=i.relatedTarget;(!n||!this._overlayRef?.overlayElement.contains(n))&&this.hide()}],["wheel",i=>this._wheelListener(i)]);else if(this.touchGestures!=="off"){this._disableNativeGesturesIfNecessary();let i=()=>{this._touchstartTimeout&&clearTimeout(this._touchstartTimeout),this.hide(this._defaultOptions?.touchendHideDelay)};A.push(["touchend",i],["touchcancel",i])}this._addListeners(A),this._passiveListeners.push(...A)}_addListeners(A){A.forEach(([i,n])=>{this._elementRef.nativeElement.addEventListener(i,n,xP)})}_platformSupportsMouseEvents(){return!this._platform.IOS&&!this._platform.ANDROID}_wheelListener(A){if(this._isTooltipVisible()){let i=this._injector.get(at).elementFromPoint(A.clientX,A.clientY),n=this._elementRef.nativeElement;i!==n&&!n.contains(i)&&this.hide()}}_disableNativeGesturesIfNecessary(){let A=this.touchGestures;if(A!=="off"){let i=this._elementRef.nativeElement,n=i.style;(A==="on"||i.nodeName!=="INPUT"&&i.nodeName!=="TEXTAREA")&&(n.userSelect=n.msUserSelect=n.webkitUserSelect=n.MozUserSelect="none"),(A==="on"||!i.draggable)&&(n.webkitUserDrag="none"),n.touchAction="none",n.webkitTapHighlightColor="transparent"}}_syncAriaDescription(A){this._ariaDescriptionPending||(this._ariaDescriptionPending=!0,this._ariaDescriber.removeDescription(this._elementRef.nativeElement,A,"tooltip"),this._isDestroyed||To({write:()=>{this._ariaDescriptionPending=!1,this.message&&!this.disabled&&this._ariaDescriber.describe(this._elementRef.nativeElement,this.message,"tooltip")}},{injector:this._injector}))}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matTooltip",""]],hostAttrs:[1,"mat-mdc-tooltip-trigger"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mat-mdc-tooltip-disabled",n.disabled)},inputs:{position:[0,"matTooltipPosition","position"],positionAtOrigin:[0,"matTooltipPositionAtOrigin","positionAtOrigin"],disabled:[0,"matTooltipDisabled","disabled"],showDelay:[0,"matTooltipShowDelay","showDelay"],hideDelay:[0,"matTooltipHideDelay","hideDelay"],touchGestures:[0,"matTooltipTouchGestures","touchGestures"],message:[0,"matTooltip","message"],tooltipClass:[0,"matTooltipClass","tooltipClass"]},exportAs:["matTooltip"]})}return t})(),i3A=(()=>{class t{_changeDetectorRef=f(It);_elementRef=f(ee);_isMultiline=!1;message;tooltipClass;_showTimeoutId;_hideTimeoutId;_triggerElement;_mouseLeaveHideDelay;_animationsDisabled;_tooltip;_closeOnInteraction=!1;_isVisible=!1;_onHide=new OA;_showAnimation="mat-mdc-tooltip-show";_hideAnimation="mat-mdc-tooltip-hide";constructor(){let A=f(Si,{optional:!0});this._animationsDisabled=A==="NoopAnimations"}show(A){this._hideTimeoutId!=null&&clearTimeout(this._hideTimeoutId),this._showTimeoutId=setTimeout(()=>{this._toggleVisibility(!0),this._showTimeoutId=void 0},A)}hide(A){this._showTimeoutId!=null&&clearTimeout(this._showTimeoutId),this._hideTimeoutId=setTimeout(()=>{this._toggleVisibility(!1),this._hideTimeoutId=void 0},A)}afterHidden(){return this._onHide}isVisible(){return this._isVisible}ngOnDestroy(){this._cancelPendingAnimations(),this._onHide.complete(),this._triggerElement=null}_handleBodyInteraction(){this._closeOnInteraction&&this.hide(0)}_markForCheck(){this._changeDetectorRef.markForCheck()}_handleMouseLeave({relatedTarget:A}){(!A||!this._triggerElement.contains(A))&&(this.isVisible()?this.hide(this._mouseLeaveHideDelay):this._finalizeAnimation(!1))}_onShow(){this._isMultiline=this._isTooltipMultiline(),this._markForCheck()}_isTooltipMultiline(){let A=this._elementRef.nativeElement.getBoundingClientRect();return A.height>e3A&&A.width>=t3A}_handleAnimationEnd({animationName:A}){(A===this._showAnimation||A===this._hideAnimation)&&this._finalizeAnimation(A===this._showAnimation)}_cancelPendingAnimations(){this._showTimeoutId!=null&&clearTimeout(this._showTimeoutId),this._hideTimeoutId!=null&&clearTimeout(this._hideTimeoutId),this._showTimeoutId=this._hideTimeoutId=void 0}_finalizeAnimation(A){A?this._closeOnInteraction=!0:this.isVisible()||this._onHide.next()}_toggleVisibility(A){let i=this._tooltip.nativeElement,n=this._showAnimation,o=this._hideAnimation;if(i.classList.remove(A?o:n),i.classList.add(A?n:o),this._isVisible!==A&&(this._isVisible=A,this._changeDetectorRef.markForCheck()),A&&!this._animationsDisabled&&typeof getComputedStyle=="function"){let r=getComputedStyle(i);(r.getPropertyValue("animation-duration")==="0s"||r.getPropertyValue("animation-name")==="none")&&(this._animationsDisabled=!0)}A&&this._onShow(),this._animationsDisabled&&(i.classList.add("_mat-animation-noopable"),this._finalizeAnimation(A))}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-tooltip-component"]],viewQuery:function(i,n){if(i&1&&Te(q4A,7),i&2){let o;XA(o=$A())&&(n._tooltip=o.first)}},hostAttrs:["aria-hidden","true"],hostBindings:function(i,n){i&1&&hA("mouseleave",function(r){return n._handleMouseLeave(r)})},decls:4,vars:4,consts:[["tooltip",""],[1,"mdc-tooltip","mat-mdc-tooltip",3,"animationend","ngClass"],[1,"mat-mdc-tooltip-surface","mdc-tooltip__surface"]],template:function(i,n){if(i&1){let o=be();S(0,"div",1,0),hA("animationend",function(s){return RA(o),xA(n._handleAnimationEnd(s))}),S(2,"div",2),AA(3),F()()}i&2&&(ue("mdc-tooltip--multiline",n._isMultiline),yA("ngClass",n.tooltipClass),G(3),Gt(n.message))},dependencies:[Xa],styles:['.mat-mdc-tooltip{position:relative;transform:scale(0);display:inline-flex}.mat-mdc-tooltip::before{content:"";top:0;right:0;bottom:0;left:0;z-index:-1;position:absolute}.mat-mdc-tooltip-panel-below .mat-mdc-tooltip::before{top:-8px}.mat-mdc-tooltip-panel-above .mat-mdc-tooltip::before{bottom:-8px}.mat-mdc-tooltip-panel-right .mat-mdc-tooltip::before{left:-8px}.mat-mdc-tooltip-panel-left .mat-mdc-tooltip::before{right:-8px}.mat-mdc-tooltip._mat-animation-noopable{animation:none;transform:scale(1)}.mat-mdc-tooltip-surface{word-break:normal;overflow-wrap:anywhere;padding:4px 8px;min-width:40px;max-width:200px;min-height:24px;max-height:40vh;box-sizing:border-box;overflow:hidden;text-align:center;will-change:transform,opacity;background-color:var(--mdc-plain-tooltip-container-color, var(--mat-sys-inverse-surface));color:var(--mdc-plain-tooltip-supporting-text-color, var(--mat-sys-inverse-on-surface));border-radius:var(--mdc-plain-tooltip-container-shape, var(--mat-sys-corner-extra-small));font-family:var(--mdc-plain-tooltip-supporting-text-font, var(--mat-sys-body-small-font));font-size:var(--mdc-plain-tooltip-supporting-text-size, var(--mat-sys-body-small-size));font-weight:var(--mdc-plain-tooltip-supporting-text-weight, var(--mat-sys-body-small-weight));line-height:var(--mdc-plain-tooltip-supporting-text-line-height, var(--mat-sys-body-small-line-height));letter-spacing:var(--mdc-plain-tooltip-supporting-text-tracking, var(--mat-sys-body-small-tracking))}.mat-mdc-tooltip-surface::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:1px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}.mdc-tooltip--multiline .mat-mdc-tooltip-surface{text-align:left}[dir=rtl] .mdc-tooltip--multiline .mat-mdc-tooltip-surface{text-align:right}.mat-mdc-tooltip-panel{line-height:normal}.mat-mdc-tooltip-panel.mat-mdc-tooltip-panel-non-interactive{pointer-events:none}@keyframes mat-mdc-tooltip-show{0%{opacity:0;transform:scale(0.8)}100%{opacity:1;transform:scale(1)}}@keyframes mat-mdc-tooltip-hide{0%{opacity:1;transform:scale(1)}100%{opacity:0;transform:scale(0.8)}}.mat-mdc-tooltip-show{animation:mat-mdc-tooltip-show 150ms cubic-bezier(0, 0, 0.2, 1) forwards}.mat-mdc-tooltip-hide{animation:mat-mdc-tooltip-hide 75ms cubic-bezier(0.4, 0, 1, 1) forwards}'],encapsulation:2,changeDetection:0})}return t})();var f8=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[Z4A],imports:[r8,El,it,it,Cl]})}return t})();function n3A(t,e){if(t&1&&(S(0,"mat-option",17),AA(1),F()),t&2){let A=e.$implicit;yA("value",A),G(),Et(" ",A," ")}}function o3A(t,e){if(t&1){let A=be();S(0,"mat-form-field",14)(1,"mat-select",16,0),hA("selectionChange",function(n){RA(A);let o=O(2);return xA(o._changePageSize(n.value))}),Dn(3,n3A,2,2,"mat-option",17,to),F(),S(5,"div",18),hA("click",function(){RA(A);let n=cr(2);return xA(n.open())}),F()()}if(t&2){let A=O(2);yA("appearance",A._formFieldAppearance)("color",A.color),G(),yA("value",A.pageSize)("disabled",A.disabled)("aria-labelledby",A._pageSizeLabelId)("panelClass",A.selectConfig.panelClass||"")("disableOptionCentering",A.selectConfig.disableOptionCentering),G(2),yn(A._displayedPageSizeOptions)}}function r3A(t,e){if(t&1&&(S(0,"div",15),AA(1),F()),t&2){let A=O(2);G(),Gt(A.pageSize)}}function s3A(t,e){if(t&1&&(S(0,"div",3)(1,"div",13),AA(2),F(),_A(3,o3A,6,7,"mat-form-field",14)(4,r3A,2,1,"div",15),F()),t&2){let A=O();G(),Ne("id",A._pageSizeLabelId),G(),Et(" ",A._intl.itemsPerPageLabel," "),G(),GA(A._displayedPageSizeOptions.length>1?3:-1),G(),GA(A._displayedPageSizeOptions.length<=1?4:-1)}}function a3A(t,e){if(t&1){let A=be();S(0,"button",19),hA("click",function(){RA(A);let n=O();return xA(n._buttonClicked(0,n._previousButtonsDisabled()))}),ar(),S(1,"svg",8),JA(2,"path",20),F()()}if(t&2){let A=O();yA("matTooltip",A._intl.firstPageLabel)("matTooltipDisabled",A._previousButtonsDisabled())("disabled",A._previousButtonsDisabled()),Ne("aria-label",A._intl.firstPageLabel)}}function c3A(t,e){if(t&1){let A=be();S(0,"button",21),hA("click",function(){RA(A);let n=O();return xA(n._buttonClicked(n.getNumberOfPages()-1,n._nextButtonsDisabled()))}),ar(),S(1,"svg",8),JA(2,"path",22),F()()}if(t&2){let A=O();yA("matTooltip",A._intl.lastPageLabel)("matTooltipDisabled",A._nextButtonsDisabled())("disabled",A._nextButtonsDisabled()),Ne("aria-label",A._intl.lastPageLabel)}}var iC=(()=>{class t{changes=new OA;itemsPerPageLabel="Items per page:";nextPageLabel="Next page";previousPageLabel="Previous page";firstPageLabel="First page";lastPageLabel="Last page";getRangeLabel=(A,i,n)=>{if(n==0||i==0)return`0 of ${n}`;n=Math.max(n,0);let o=A*i,r=o<n?Math.min(o+i,n):o+i;return`${o+1} \u2013 ${r} of ${n}`};static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function l3A(t){return t||new iC}var g3A={provide:iC,deps:[[new MI,new fh,iC]],useFactory:l3A},I3A=50;var C3A=new dA("MAT_PAGINATOR_DEFAULT_OPTIONS"),Ok=(()=>{class t{_intl=f(iC);_changeDetectorRef=f(It);_formFieldAppearance;_pageSizeLabelId=f(sn).getId("mat-paginator-page-size-label-");_intlChanges;_isInitialized=!1;_initializedStream=new Al(1);color;get pageIndex(){return this._pageIndex}set pageIndex(A){this._pageIndex=Math.max(A||0,0),this._changeDetectorRef.markForCheck()}_pageIndex=0;get length(){return this._length}set length(A){this._length=A||0,this._changeDetectorRef.markForCheck()}_length=0;get pageSize(){return this._pageSize}set pageSize(A){this._pageSize=Math.max(A||0,0),this._updateDisplayedPageSizeOptions()}_pageSize;get pageSizeOptions(){return this._pageSizeOptions}set pageSizeOptions(A){this._pageSizeOptions=(A||[]).map(i=>zi(i,0)),this._updateDisplayedPageSizeOptions()}_pageSizeOptions=[];hidePageSize=!1;showFirstLastButtons=!1;selectConfig={};disabled=!1;page=new WA;_displayedPageSizeOptions;initialized=this._initializedStream;constructor(){let A=this._intl,i=f(C3A,{optional:!0});if(this._intlChanges=A.changes.subscribe(()=>this._changeDetectorRef.markForCheck()),i){let{pageSize:n,pageSizeOptions:o,hidePageSize:r,showFirstLastButtons:s}=i;n!=null&&(this._pageSize=n),o!=null&&(this._pageSizeOptions=o),r!=null&&(this.hidePageSize=r),s!=null&&(this.showFirstLastButtons=s)}this._formFieldAppearance=i?.formFieldAppearance||"outline"}ngOnInit(){this._isInitialized=!0,this._updateDisplayedPageSizeOptions(),this._initializedStream.next()}ngOnDestroy(){this._initializedStream.complete(),this._intlChanges.unsubscribe()}nextPage(){this.hasNextPage()&&this._navigate(this.pageIndex+1)}previousPage(){this.hasPreviousPage()&&this._navigate(this.pageIndex-1)}firstPage(){this.hasPreviousPage()&&this._navigate(0)}lastPage(){this.hasNextPage()&&this._navigate(this.getNumberOfPages()-1)}hasPreviousPage(){return this.pageIndex>=1&&this.pageSize!=0}hasNextPage(){let A=this.getNumberOfPages()-1;return this.pageIndex<A&&this.pageSize!=0}getNumberOfPages(){return this.pageSize?Math.ceil(this.length/this.pageSize):0}_changePageSize(A){let i=this.pageIndex*this.pageSize,n=this.pageIndex;this.pageIndex=Math.floor(i/A)||0,this.pageSize=A,this._emitPageEvent(n)}_nextButtonsDisabled(){return this.disabled||!this.hasNextPage()}_previousButtonsDisabled(){return this.disabled||!this.hasPreviousPage()}_updateDisplayedPageSizeOptions(){this._isInitialized&&(this.pageSize||(this._pageSize=this.pageSizeOptions.length!=0?this.pageSizeOptions[0]:I3A),this._displayedPageSizeOptions=this.pageSizeOptions.slice(),this._displayedPageSizeOptions.indexOf(this.pageSize)===-1&&this._displayedPageSizeOptions.push(this.pageSize),this._displayedPageSizeOptions.sort((A,i)=>A-i),this._changeDetectorRef.markForCheck())}_emitPageEvent(A){this.page.emit({previousPageIndex:A,pageIndex:this.pageIndex,pageSize:this.pageSize,length:this.length})}_navigate(A){let i=this.pageIndex;A!==i&&(this.pageIndex=A,this._emitPageEvent(i))}_buttonClicked(A,i){i||this._navigate(A)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-paginator"]],hostAttrs:["role","group",1,"mat-mdc-paginator"],inputs:{color:"color",pageIndex:[2,"pageIndex","pageIndex",zi],length:[2,"length","length",zi],pageSize:[2,"pageSize","pageSize",zi],pageSizeOptions:"pageSizeOptions",hidePageSize:[2,"hidePageSize","hidePageSize",ie],showFirstLastButtons:[2,"showFirstLastButtons","showFirstLastButtons",ie],selectConfig:"selectConfig",disabled:[2,"disabled","disabled",ie]},outputs:{page:"page"},exportAs:["matPaginator"],decls:14,vars:12,consts:[["selectRef",""],[1,"mat-mdc-paginator-outer-container"],[1,"mat-mdc-paginator-container"],[1,"mat-mdc-paginator-page-size"],[1,"mat-mdc-paginator-range-actions"],["aria-live","polite",1,"mat-mdc-paginator-range-label"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-first",3,"matTooltip","matTooltipDisabled","disabled"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-previous",3,"click","matTooltip","matTooltipDisabled","disabled"],["viewBox","0 0 24 24","focusable","false","aria-hidden","true",1,"mat-mdc-paginator-icon"],["d","M15.41 7.41L14 6l-6 6 6 6 1.41-1.41L10.83 12z"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-next",3,"click","matTooltip","matTooltipDisabled","disabled"],["d","M10 6L8.59 7.41 13.17 12l-4.58 4.59L10 18l6-6z"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-last",3,"matTooltip","matTooltipDisabled","disabled"],[1,"mat-mdc-paginator-page-size-label"],[1,"mat-mdc-paginator-page-size-select",3,"appearance","color"],[1,"mat-mdc-paginator-page-size-value"],["hideSingleSelectionIndicator","",3,"selectionChange","value","disabled","aria-labelledby","panelClass","disableOptionCentering"],[3,"value"],[1,"mat-mdc-paginator-touch-target",3,"click"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-first",3,"click","matTooltip","matTooltipDisabled","disabled"],["d","M18.41 16.59L13.82 12l4.59-4.59L17 6l-6 6 6 6zM6 6h2v12H6z"],["mat-icon-button","","type","button","matTooltipPosition","above","disabledInteractive","",1,"mat-mdc-paginator-navigation-last",3,"click","matTooltip","matTooltipDisabled","disabled"],["d","M5.59 7.41L10.18 12l-4.59 4.59L7 18l6-6-6-6zM16 6h2v12h-2z"]],template:function(i,n){i&1&&(S(0,"div",1)(1,"div",2),_A(2,s3A,5,4,"div",3),S(3,"div",4)(4,"div",5),AA(5),F(),_A(6,a3A,3,4,"button",6),S(7,"button",7),hA("click",function(){return n._buttonClicked(n.pageIndex-1,n._previousButtonsDisabled())}),ar(),S(8,"svg",8),JA(9,"path",9),F()(),SI(),S(10,"button",10),hA("click",function(){return n._buttonClicked(n.pageIndex+1,n._nextButtonsDisabled())}),ar(),S(11,"svg",8),JA(12,"path",11),F()(),_A(13,c3A,3,4,"button",12),F()()()),i&2&&(G(2),GA(n.hidePageSize?-1:2),G(3),Et(" ",n._intl.getRangeLabel(n.pageIndex,n.pageSize,n.length)," "),G(),GA(n.showFirstLastButtons?6:-1),G(),yA("matTooltip",n._intl.previousPageLabel)("matTooltipDisabled",n._previousButtonsDisabled())("disabled",n._previousButtonsDisabled()),Ne("aria-label",n._intl.previousPageLabel),G(3),yA("matTooltip",n._intl.nextPageLabel)("matTooltipDisabled",n._nextButtonsDisabled())("disabled",n._nextButtonsDisabled()),Ne("aria-label",n._intl.nextPageLabel),G(3),GA(n.showFirstLastButtons?13:-1))},dependencies:[Qg,NB,U2,kB,_B],styles:[".mat-mdc-paginator{display:block;-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;color:var(--mat-paginator-container-text-color, var(--mat-sys-on-surface));background-color:var(--mat-paginator-container-background-color, var(--mat-sys-surface));font-family:var(--mat-paginator-container-text-font, var(--mat-sys-body-small-font));line-height:var(--mat-paginator-container-text-line-height, var(--mat-sys-body-small-line-height));font-size:var(--mat-paginator-container-text-size, var(--mat-sys-body-small-size));font-weight:var(--mat-paginator-container-text-weight, var(--mat-sys-body-small-weight));letter-spacing:var(--mat-paginator-container-text-tracking, var(--mat-sys-body-small-tracking));--mat-form-field-container-height:var(--mat-paginator-form-field-container-height, 40px);--mat-form-field-container-vertical-padding:var(--mat-paginator-form-field-container-vertical-padding, 8px)}.mat-mdc-paginator .mat-mdc-select-value{font-size:var(--mat-paginator-select-trigger-text-size, var(--mat-sys-body-small-size))}.mat-mdc-paginator .mat-mdc-form-field-subscript-wrapper{display:none}.mat-mdc-paginator .mat-mdc-select{line-height:1.5}.mat-mdc-paginator-outer-container{display:flex}.mat-mdc-paginator-container{display:flex;align-items:center;justify-content:flex-end;padding:0 8px;flex-wrap:wrap;width:100%;min-height:var(--mat-paginator-container-size, 56px)}.mat-mdc-paginator-page-size{display:flex;align-items:baseline;margin-right:8px}[dir=rtl] .mat-mdc-paginator-page-size{margin-right:0;margin-left:8px}.mat-mdc-paginator-page-size-label{margin:0 4px}.mat-mdc-paginator-page-size-select{margin:0 4px;width:84px}.mat-mdc-paginator-range-label{margin:0 32px 0 24px}.mat-mdc-paginator-range-actions{display:flex;align-items:center}.mat-mdc-paginator-icon{display:inline-block;width:28px;fill:var(--mat-paginator-enabled-icon-color, var(--mat-sys-on-surface-variant))}.mat-mdc-icon-button[aria-disabled] .mat-mdc-paginator-icon{fill:var(--mat-paginator-disabled-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}[dir=rtl] .mat-mdc-paginator-icon{transform:rotate(180deg)}@media(forced-colors: active){.mat-mdc-icon-button[disabled] .mat-mdc-paginator-icon,.mat-mdc-paginator-icon{fill:currentColor;fill:CanvasText}.mat-mdc-paginator-range-actions .mat-mdc-icon-button{outline:solid 1px}}.mat-mdc-paginator-touch-target{display:var(--mat-paginator-touch-target-display, block);position:absolute;top:50%;left:50%;width:84px;height:48px;background-color:rgba(0,0,0,0);transform:translate(-50%, -50%);cursor:pointer}"],encapsulation:2,changeDetection:0})}return t})(),_P=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[g3A],imports:[AC,u8,f8,Ok]})}return t})();function B3A(t,e){if(t&1){let A=be();S(0,"div",1)(1,"button",2),hA("click",function(){RA(A);let n=O();return xA(n.action())}),AA(2),F()()}if(t&2){let A=O();G(2),Et(" ",A.data.action," ")}}var E3A=["label"];function Q3A(t,e){}var h3A=Math.pow(2,31)-1,Uu=class{_overlayRef;instance;containerInstance;_afterDismissed=new OA;_afterOpened=new OA;_onAction=new OA;_durationTimeoutId;_dismissedByAction=!1;constructor(e,A){this._overlayRef=A,this.containerInstance=e,e._onExit.subscribe(()=>this._finishDismiss())}dismiss(){this._afterDismissed.closed||this.containerInstance.exit(),clearTimeout(this._durationTimeoutId)}dismissWithAction(){this._onAction.closed||(this._dismissedByAction=!0,this._onAction.next(),this._onAction.complete(),this.dismiss()),clearTimeout(this._durationTimeoutId)}closeWithAction(){this.dismissWithAction()}_dismissAfter(e){this._durationTimeoutId=setTimeout(()=>this.dismiss(),Math.min(e,h3A))}_open(){this._afterOpened.closed||(this._afterOpened.next(),this._afterOpened.complete())}_finishDismiss(){this._overlayRef.dispose(),this._onAction.closed||this._onAction.complete(),this._afterDismissed.next({dismissedByAction:this._dismissedByAction}),this._afterDismissed.complete(),this._dismissedByAction=!1}afterDismissed(){return this._afterDismissed}afterOpened(){return this.containerInstance._onEnter}onAction(){return this._onAction}},GP=new dA("MatSnackBarData"),GB=class{politeness="assertive";announcementMessage="";viewContainerRef;duration=0;panelClass;direction;data=null;horizontalPosition="center";verticalPosition="bottom"},u3A=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matSnackBarLabel",""]],hostAttrs:[1,"mat-mdc-snack-bar-label","mdc-snackbar__label"]})}return t})(),f3A=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matSnackBarActions",""]],hostAttrs:[1,"mat-mdc-snack-bar-actions","mdc-snackbar__actions"]})}return t})(),m3A=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matSnackBarAction",""]],hostAttrs:[1,"mat-mdc-snack-bar-action","mdc-snackbar__action"]})}return t})(),p3A=(()=>{class t{snackBarRef=f(Uu);data=f(GP);constructor(){}action(){this.snackBarRef.dismissWithAction()}get hasAction(){return!!this.data.action}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["simple-snack-bar"]],hostAttrs:[1,"mat-mdc-simple-snack-bar"],exportAs:["matSnackBar"],decls:3,vars:2,consts:[["matSnackBarLabel",""],["matSnackBarActions",""],["mat-button","","matSnackBarAction","",3,"click"]],template:function(i,n){i&1&&(S(0,"div",0),AA(1),F(),_A(2,B3A,3,1,"div",1)),i&2&&(G(),Et(" ",n.data.message,`
`),G(),GA(n.hasAction?2:-1))},dependencies:[Dr,u3A,f3A,m3A],styles:[".mat-mdc-simple-snack-bar{display:flex}"],encapsulation:2,changeDetection:0})}return t})(),w3A={snackBarState:sc("state",[js("void, hidden",po({transform:"scale(0.8)",opacity:0})),js("visible",po({transform:"scale(1)",opacity:1})),Vr("* => visible",ss("150ms cubic-bezier(0, 0, 0.2, 1)")),Vr("* => void, * => hidden",ss("75ms cubic-bezier(0.4, 0.0, 1, 1)",po({opacity:0})))])},D3A=(()=>{class t extends J2{_ngZone=f(Qe);_elementRef=f(ee);_changeDetectorRef=f(It);_platform=f(Ii);snackBarConfig=f(GB);_document=f(at);_trackedModals=new Set;_announceDelay=150;_announceTimeoutId;_destroyed=!1;_portalOutlet;_onAnnounce=new OA;_onExit=new OA;_onEnter=new OA;_animationState="void";_live;_label;_role;_liveElementId=f(sn).getId("mat-snack-bar-container-live-");constructor(){super();let A=this.snackBarConfig;A.politeness==="assertive"&&!A.announcementMessage?this._live="assertive":A.politeness==="off"?this._live="off":this._live="polite",this._platform.FIREFOX&&(this._live==="polite"&&(this._role="status"),this._live==="assertive"&&(this._role="alert"))}attachComponentPortal(A){this._assertNotAttached();let i=this._portalOutlet.attachComponentPortal(A);return this._afterPortalAttached(),i}attachTemplatePortal(A){this._assertNotAttached();let i=this._portalOutlet.attachTemplatePortal(A);return this._afterPortalAttached(),i}attachDomPortal=A=>{this._assertNotAttached();let i=this._portalOutlet.attachDomPortal(A);return this._afterPortalAttached(),i};onAnimationEnd(A){let{fromState:i,toState:n}=A;if((n==="void"&&i!=="void"||n==="hidden")&&this._completeExit(),n==="visible"){let o=this._onEnter;this._ngZone.run(()=>{o.next(),o.complete()})}}enter(){this._destroyed||(this._animationState="visible",this._changeDetectorRef.markForCheck(),this._changeDetectorRef.detectChanges(),this._screenReaderAnnounce())}exit(){return this._ngZone.run(()=>{this._animationState="hidden",this._changeDetectorRef.markForCheck(),this._elementRef.nativeElement.setAttribute("mat-exit",""),clearTimeout(this._announceTimeoutId)}),this._onExit}ngOnDestroy(){this._destroyed=!0,this._clearFromModals(),this._completeExit()}_completeExit(){queueMicrotask(()=>{this._onExit.next(),this._onExit.complete()})}_afterPortalAttached(){let A=this._elementRef.nativeElement,i=this.snackBarConfig.panelClass;i&&(Array.isArray(i)?i.forEach(r=>A.classList.add(r)):A.classList.add(i)),this._exposeToModals();let n=this._label.nativeElement,o="mdc-snackbar__label";n.classList.toggle(o,!n.querySelector(`.${o}`))}_exposeToModals(){let A=this._liveElementId,i=this._document.querySelectorAll('body > .cdk-overlay-container [aria-modal="true"]');for(let n=0;n<i.length;n++){let o=i[n],r=o.getAttribute("aria-owns");this._trackedModals.add(o),r?r.indexOf(A)===-1&&o.setAttribute("aria-owns",r+" "+A):o.setAttribute("aria-owns",A)}}_clearFromModals(){this._trackedModals.forEach(A=>{let i=A.getAttribute("aria-owns");if(i){let n=i.replace(this._liveElementId,"").trim();n.length>0?A.setAttribute("aria-owns",n):A.removeAttribute("aria-owns")}}),this._trackedModals.clear()}_assertNotAttached(){this._portalOutlet.hasAttached()}_screenReaderAnnounce(){this._announceTimeoutId||this._ngZone.runOutsideAngular(()=>{this._announceTimeoutId=setTimeout(()=>{let A=this._elementRef.nativeElement.querySelector("[aria-hidden]"),i=this._elementRef.nativeElement.querySelector("[aria-live]");if(A&&i){let n=null;this._platform.isBrowser&&document.activeElement instanceof HTMLElement&&A.contains(document.activeElement)&&(n=document.activeElement),A.removeAttribute("aria-hidden"),i.appendChild(A),n?.focus(),this._onAnnounce.next(),this._onAnnounce.complete()}},this._announceDelay)})}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-snack-bar-container"]],viewQuery:function(i,n){if(i&1&&(Te(fa,7),Te(E3A,7)),i&2){let o;XA(o=$A())&&(n._portalOutlet=o.first),XA(o=$A())&&(n._label=o.first)}},hostAttrs:[1,"mdc-snackbar","mat-mdc-snack-bar-container"],hostVars:1,hostBindings:function(i,n){i&1&&Hb("@state.done",function(r){return n.onAnimationEnd(r)}),i&2&&Tb("@state",n._animationState)},features:[lt],decls:6,vars:3,consts:[["label",""],[1,"mdc-snackbar__surface","mat-mdc-snackbar-surface"],[1,"mat-mdc-snack-bar-label"],["aria-hidden","true"],["cdkPortalOutlet",""]],template:function(i,n){i&1&&(S(0,"div",1)(1,"div",2,0)(3,"div",3),_A(4,Q3A,0,0,"ng-template",4),F(),JA(5,"div"),F()()),i&2&&(G(5),Ne("aria-live",n._live)("role",n._role)("id",n._liveElementId))},dependencies:[fa],styles:[".mat-mdc-snack-bar-container{display:flex;align-items:center;justify-content:center;box-sizing:border-box;-webkit-tap-highlight-color:rgba(0,0,0,0);margin:8px}.mat-mdc-snack-bar-handset .mat-mdc-snack-bar-container{width:100vw}.mat-mdc-snackbar-surface{box-shadow:0px 3px 5px -1px rgba(0, 0, 0, 0.2), 0px 6px 10px 0px rgba(0, 0, 0, 0.14), 0px 1px 18px 0px rgba(0, 0, 0, 0.12);display:flex;align-items:center;justify-content:flex-start;box-sizing:border-box;padding-left:0;padding-right:8px}[dir=rtl] .mat-mdc-snackbar-surface{padding-right:0;padding-left:8px}.mat-mdc-snack-bar-container .mat-mdc-snackbar-surface{min-width:344px;max-width:672px}.mat-mdc-snack-bar-handset .mat-mdc-snackbar-surface{width:100%;min-width:0}@media(forced-colors: active){.mat-mdc-snackbar-surface{outline:solid 1px}}.mat-mdc-snack-bar-container .mat-mdc-snackbar-surface{color:var(--mdc-snackbar-supporting-text-color, var(--mat-sys-inverse-on-surface));border-radius:var(--mdc-snackbar-container-shape, var(--mat-sys-corner-extra-small));background-color:var(--mdc-snackbar-container-color, var(--mat-sys-inverse-surface))}.mdc-snackbar__label{width:100%;flex-grow:1;box-sizing:border-box;margin:0;padding:14px 8px 14px 16px}[dir=rtl] .mdc-snackbar__label{padding-left:8px;padding-right:16px}.mat-mdc-snack-bar-container .mdc-snackbar__label{font-family:var(--mdc-snackbar-supporting-text-font, var(--mat-sys-body-medium-font));font-size:var(--mdc-snackbar-supporting-text-size, var(--mat-sys-body-medium-size));font-weight:var(--mdc-snackbar-supporting-text-weight, var(--mat-sys-body-medium-weight));line-height:var(--mdc-snackbar-supporting-text-line-height, var(--mat-sys-body-medium-line-height))}.mat-mdc-snack-bar-actions{display:flex;flex-shrink:0;align-items:center;box-sizing:border-box}.mat-mdc-snack-bar-handset,.mat-mdc-snack-bar-container,.mat-mdc-snack-bar-label{flex:1 1 auto}.mat-mdc-snack-bar-container .mat-mdc-button.mat-mdc-snack-bar-action:not(:disabled).mat-unthemed{color:var(--mat-snack-bar-button-color, var(--mat-sys-inverse-primary))}.mat-mdc-snack-bar-container .mat-mdc-button.mat-mdc-snack-bar-action:not(:disabled){--mat-text-button-state-layer-color:currentColor;--mat-text-button-ripple-color:currentColor}.mat-mdc-snack-bar-container .mat-mdc-button.mat-mdc-snack-bar-action:not(:disabled) .mat-ripple-element{opacity:.1}"],encapsulation:2,data:{animation:[w3A.snackBarState]}})}return t})();function y3A(){return new GB}var v3A=new dA("mat-snack-bar-default-options",{providedIn:"root",factory:y3A}),UP=(()=>{class t{_overlay=f(nr);_live=f(o8);_injector=f(Rt);_breakpointObserver=f(Z6);_parentSnackBar=f(t,{optional:!0,skipSelf:!0});_defaultConfig=f(v3A);_snackBarRefAtThisLevel=null;simpleSnackBarComponent=p3A;snackBarContainerComponent=D3A;handsetCssClass="mat-mdc-snack-bar-handset";get _openedSnackBarRef(){let A=this._parentSnackBar;return A?A._openedSnackBarRef:this._snackBarRefAtThisLevel}set _openedSnackBarRef(A){this._parentSnackBar?this._parentSnackBar._openedSnackBarRef=A:this._snackBarRefAtThisLevel=A}constructor(){}openFromComponent(A,i){return this._attach(A,i)}openFromTemplate(A,i){return this._attach(A,i)}open(A,i="",n){let o=rA(rA({},this._defaultConfig),n);return o.data={message:A,action:i},o.announcementMessage===A&&(o.announcementMessage=void 0),this.openFromComponent(this.simpleSnackBarComponent,o)}dismiss(){this._openedSnackBarRef&&this._openedSnackBarRef.dismiss()}ngOnDestroy(){this._snackBarRefAtThisLevel&&this._snackBarRefAtThisLevel.dismiss()}_attachSnackBarContainer(A,i){let n=i&&i.viewContainerRef&&i.viewContainerRef.injector,o=Rt.create({parent:n||this._injector,providers:[{provide:GB,useValue:i}]}),r=new dl(this.snackBarContainerComponent,i.viewContainerRef,o),s=A.attach(r);return s.instance.snackBarConfig=i,s.instance}_attach(A,i){let n=rA(rA(rA({},new GB),this._defaultConfig),i),o=this._createOverlay(n),r=this._attachSnackBarContainer(o,n),s=new Uu(r,o);if(A instanceof wn){let a=new ys(A,null,{$implicit:n.data,snackBarRef:s});s.instance=r.attachTemplatePortal(a)}else{let a=this._createInjector(n,s),c=new dl(A,void 0,a),l=r.attachComponentPortal(c);s.instance=l.instance}return this._breakpointObserver.observe(_O.HandsetPortrait).pipe(St(o.detachments())).subscribe(a=>{o.overlayElement.classList.toggle(this.handsetCssClass,a.matches)}),n.announcementMessage&&r._onAnnounce.subscribe(()=>{this._live.announce(n.announcementMessage,n.politeness)}),this._animateSnackBar(s,n),this._openedSnackBarRef=s,this._openedSnackBarRef}_animateSnackBar(A,i){A.afterDismissed().subscribe(()=>{this._openedSnackBarRef==A&&(this._openedSnackBarRef=null),i.announcementMessage&&this._live.clear()}),this._openedSnackBarRef?(this._openedSnackBarRef.afterDismissed().subscribe(()=>{A.containerInstance.enter()}),this._openedSnackBarRef.dismiss()):A.containerInstance.enter(),i.duration&&i.duration>0&&A.afterOpened().subscribe(()=>A._dismissAfter(i.duration))}_createOverlay(A){let i=new Bg;i.direction=A.direction;let n=this._overlay.position().global(),o=A.direction==="rtl",r=A.horizontalPosition==="left"||A.horizontalPosition==="start"&&!o||A.horizontalPosition==="end"&&o,s=!r&&A.horizontalPosition!=="center";return r?n.left("0"):s?n.right("0"):n.centerHorizontally(),A.verticalPosition==="top"?n.top("0"):n.bottom("0"),i.positionStrategy=n,this._overlay.create(i)}_createInjector(A,i){let n=A&&A.viewContainerRef&&A.viewContainerRef.injector;return Rt.create({parent:n||this._injector,providers:[{provide:Uu,useValue:i},{provide:GP,useValue:A.data}]})}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var b3A=(()=>{var t=import.meta.url;return function(e={}){var A,i=e,n,o,r=new Promise((Q,m)=>{n=Q,o=m});i.agerrMessages=[],i.stderrMessages=[],B=Q=>i.stderrMessages.push(Q);var s=Object.assign({},i),a="./this.program",c=(Q,m)=>{throw m},l="",I,C;typeof document<"u"&&document.currentScript&&(l=document.currentScript.src),t&&(l=t),l.startsWith("blob:")?l="":l=l.substr(0,l.replace(/[?#].*/,"").lastIndexOf("/")+1),I=Q=>fetch(Q,{credentials:"same-origin"}).then(m=>m.ok?m.arrayBuffer():Promise.reject(new Error(m.status+" : "+m.url)));var d=console.log.bind(console),B=console.error.bind(console);Object.assign(i,s),s=null;var E;function h(Q){for(var m=atob(Q),v=new Uint8Array(m.length),p=0;p<m.length;++p)v[p]=m.charCodeAt(p);return v}function u(Q){if(He(Q))return h(Q.slice(Ze.length))}var D,L=!1,R,w,_,K,z,U,H,q;function j(){var Q=D.buffer;i.HEAP8=R=new Int8Array(Q),i.HEAP16=_=new Int16Array(Q),i.HEAPU8=w=new Uint8Array(Q),i.HEAPU16=new Uint16Array(Q),i.HEAP32=K=new Int32Array(Q),i.HEAPU32=z=new Uint32Array(Q),i.HEAPF32=U=new Float32Array(Q),i.HEAPF64=q=new Float64Array(Q),i.HEAP64=H=new BigInt64Array(Q),i.HEAPU64=new BigUint64Array(Q)}var gA=[],QA=[],BA=[];function lA(){tt(gA)}function vA(){!i.noFSInit&&!M.initialized&&M.init(),M.ignorePermissions=!1,tt(QA)}function tA(){tt(BA)}function cA(Q){QA.unshift(Q)}var pA=0,VA=null;function oe(Q){return Q}function KA(Q){pA++}function CA(Q){if(pA--,pA==0&&VA){var m=VA;VA=null,m()}}function TA(Q){Q="Aborted("+Q+")",B(Q),L=!0,Q+=". Build with -sASSERTIONS for more info.";var m=new WebAssembly.RuntimeError(Q);throw o(m),m}var Ze="data:application/octet-stream;base64,",He=Q=>Q.startsWith(Ze);function uA(){var Q="data:application/octet-stream;base64, ";return Q}var eA;function UA(Q){if(Q==eA&&E)return new Uint8Array(E);var m=u(Q);if(m)return m;throw"both async and sync fetching of the wasm failed"}function aA(Q){return Promise.resolve().then(()=>UA(Q))}function le(Q,m,v){return aA(Q).then(p=>WebAssembly.instantiate(p,m)).then(v,p=>{B(`failed to asynchronously prepare wasm: ${p}`),TA(p)})}function SA(Q,m,v,p){return le(m,v,p)}function Ue(){return{a:Xn}}function mA(){var Q=Ue();function m(p,N){return se=p.exports,D=se.y,j(),cA(se.z),CA(),se}KA();function v(p){m(p.instance)}return eA??=uA(),SA(E,eA,Q,v).catch(o),{}}function sA(Q){return i.agerrMessages.push(Re(Q)),0}function xt(Q){this.name="ExitStatus",this.message=`Program terminated with exit(${Q})`,this.status=Q}var tt=Q=>{Q.forEach(m=>m(i))};function de(Q,m="i8"){switch(m.endsWith("*")&&(m="*"),m){case"i1":return R[Q];case"i8":return R[Q];case"i16":return _[Q>>1];case"i32":return K[Q>>2];case"i64":return H[Q>>3];case"float":return U[Q>>2];case"double":return q[Q>>3];case"*":return z[Q>>2];default:TA(`invalid type for getValue: ${m}`)}}var Dt=Q=>Yi(Q),_e=()=>Hr(),Le=typeof TextDecoder<"u"?new TextDecoder:void 0,bt=(Q,m=0,v=NaN)=>{for(var p=m+v,N=m;Q[N]&&!(N>=p);)++N;if(N-m>16&&Q.buffer&&Le)return Le.decode(Q.subarray(m,N));for(var J="";m<N;){var V=Q[m++];if(!(V&128)){J+=String.fromCharCode(V);continue}var Z=Q[m++]&63;if((V&224)==192){J+=String.fromCharCode((V&31)<<6|Z);continue}var FA=Q[m++]&63;if((V&240)==224?V=(V&15)<<12|Z<<6|FA:V=(V&7)<<18|Z<<12|FA<<6|Q[m++]&63,V<65536)J+=String.fromCharCode(V);else{var te=V-65536;J+=String.fromCharCode(55296|te>>10,56320|te&1023)}}return J},Re=(Q,m)=>Q?bt(w,Q,m):"",$t=(Q,m,v,p)=>{TA(`Assertion failed: ${Re(Q)}, at: `+[m?Re(m):"unknown filename",v,p?Re(p):"unknown function"])};class x{constructor(m){this.excPtr=m,this.ptr=m-24}set_type(m){z[this.ptr+4>>2]=m}get_type(){return z[this.ptr+4>>2]}set_destructor(m){z[this.ptr+8>>2]=m}get_destructor(){return z[this.ptr+8>>2]}set_caught(m){m=m?1:0,R[this.ptr+12]=m}get_caught(){return R[this.ptr+12]!=0}set_rethrown(m){m=m?1:0,R[this.ptr+13]=m}get_rethrown(){return R[this.ptr+13]!=0}init(m,v){this.set_adjusted_ptr(0),this.set_type(m),this.set_destructor(v)}set_adjusted_ptr(m){z[this.ptr+16>>2]=m}get_adjusted_ptr(){return z[this.ptr+16>>2]}}var Y=0,P=(Q,m,v)=>{var p=new x(Q);throw p.init(m,v),Y=Q,Y},X={isAbs:Q=>Q.charAt(0)==="/",splitPath:Q=>{var m=/^(\/?|)([\s\S]*?)((?:\.{1,2}|[^\/]+?|)(\.[^.\/]*|))(?:[\/]*)$/;return m.exec(Q).slice(1)},normalizeArray:(Q,m)=>{for(var v=0,p=Q.length-1;p>=0;p--){var N=Q[p];N==="."?Q.splice(p,1):N===".."?(Q.splice(p,1),v++):v&&(Q.splice(p,1),v--)}if(m)for(;v;v--)Q.unshift("..");return Q},normalize:Q=>{var m=X.isAbs(Q),v=Q.substr(-1)==="/";return Q=X.normalizeArray(Q.split("/").filter(p=>!!p),!m).join("/"),!Q&&!m&&(Q="."),Q&&v&&(Q+="/"),(m?"/":"")+Q},dirname:Q=>{var m=X.splitPath(Q),v=m[0],p=m[1];return!v&&!p?".":(p&&(p=p.substr(0,p.length-1)),v+p)},basename:Q=>{if(Q==="/")return"/";Q=X.normalize(Q),Q=Q.replace(/\/$/,"");var m=Q.lastIndexOf("/");return m===-1?Q:Q.substr(m+1)},join:(...Q)=>X.normalize(Q.join("/")),join2:(Q,m)=>X.normalize(Q+"/"+m)},bA=()=>{if(typeof crypto=="object"&&typeof crypto.getRandomValues=="function")return Q=>crypto.getRandomValues(Q);TA("initRandomDevice")},Be=Q=>(Be=bA())(Q),Ee={resolve:(...Q)=>{for(var m="",v=!1,p=Q.length-1;p>=-1&&!v;p--){var N=p>=0?Q[p]:M.cwd();if(typeof N!="string")throw new TypeError("Arguments to path.resolve must be strings");if(!N)return"";m=N+"/"+m,v=X.isAbs(N)}return m=X.normalizeArray(m.split("/").filter(J=>!!J),!v).join("/"),(v?"/":"")+m||"."},relative:(Q,m)=>{Q=Ee.resolve(Q).substr(1),m=Ee.resolve(m).substr(1);function v(te){for(var re=0;re<te.length&&te[re]==="";re++);for(var Pe=te.length-1;Pe>=0&&te[Pe]==="";Pe--);return re>Pe?[]:te.slice(re,Pe-re+1)}for(var p=v(Q.split("/")),N=v(m.split("/")),J=Math.min(p.length,N.length),V=J,Z=0;Z<J;Z++)if(p[Z]!==N[Z]){V=Z;break}for(var FA=[],Z=V;Z<p.length;Z++)FA.push("..");return FA=FA.concat(N.slice(V)),FA.join("/")}},kA=[],DA=Q=>{for(var m=0,v=0;v<Q.length;++v){var p=Q.charCodeAt(v);p<=127?m++:p<=2047?m+=2:p>=55296&&p<=57343?(m+=4,++v):m+=3}return m},gt=(Q,m,v,p)=>{if(!(p>0))return 0;for(var N=v,J=v+p-1,V=0;V<Q.length;++V){var Z=Q.charCodeAt(V);if(Z>=55296&&Z<=57343){var FA=Q.charCodeAt(++V);Z=65536+((Z&1023)<<10)|FA&1023}if(Z<=127){if(v>=J)break;m[v++]=Z}else if(Z<=2047){if(v+1>=J)break;m[v++]=192|Z>>6,m[v++]=128|Z&63}else if(Z<=65535){if(v+2>=J)break;m[v++]=224|Z>>12,m[v++]=128|Z>>6&63,m[v++]=128|Z&63}else{if(v+3>=J)break;m[v++]=240|Z>>18,m[v++]=128|Z>>12&63,m[v++]=128|Z>>6&63,m[v++]=128|Z&63}}return m[v]=0,v-N};function Ve(Q,m,v){var p=v>0?v:DA(Q)+1,N=new Array(p),J=gt(Q,N,0,N.length);return m&&(N.length=J),N}var ZA=()=>{if(!kA.length){var Q=null;if(typeof window<"u"&&typeof window.prompt=="function"&&(Q=window.prompt("Input: "),Q!==null&&(Q+=`
`)),!Q)return null;kA=Ve(Q,!0)}return kA.shift()},rt={ttys:[],init(){},shutdown(){},register(Q,m){rt.ttys[Q]={input:[],output:[],ops:m},M.registerDevice(Q,rt.stream_ops)},stream_ops:{open(Q){var m=rt.ttys[Q.node.rdev];if(!m)throw new M.ErrnoError(43);Q.tty=m,Q.seekable=!1},close(Q){Q.tty.ops.fsync(Q.tty)},fsync(Q){Q.tty.ops.fsync(Q.tty)},read(Q,m,v,p,N){if(!Q.tty||!Q.tty.ops.get_char)throw new M.ErrnoError(60);for(var J=0,V=0;V<p;V++){var Z;try{Z=Q.tty.ops.get_char(Q.tty)}catch{throw new M.ErrnoError(29)}if(Z===void 0&&J===0)throw new M.ErrnoError(6);if(Z==null)break;J++,m[v+V]=Z}return J&&(Q.node.timestamp=Date.now()),J},write(Q,m,v,p,N){if(!Q.tty||!Q.tty.ops.put_char)throw new M.ErrnoError(60);try{for(var J=0;J<p;J++)Q.tty.ops.put_char(Q.tty,m[v+J])}catch{throw new M.ErrnoError(29)}return p&&(Q.node.timestamp=Date.now()),J}},default_tty_ops:{get_char(Q){return ZA()},put_char(Q,m){m===null||m===10?(d(bt(Q.output)),Q.output=[]):m!=0&&Q.output.push(m)},fsync(Q){Q.output&&Q.output.length>0&&(d(bt(Q.output)),Q.output=[])},ioctl_tcgets(Q){return{c_iflag:25856,c_oflag:5,c_cflag:191,c_lflag:35387,c_cc:[3,28,127,21,4,0,1,0,17,19,26,0,18,15,23,22,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0]}},ioctl_tcsets(Q,m,v){return 0},ioctl_tiocgwinsz(Q){return[24,80]}},default_tty1_ops:{put_char(Q,m){m===null||m===10?(B(bt(Q.output)),Q.output=[]):m!=0&&Q.output.push(m)},fsync(Q){Q.output&&Q.output.length>0&&(B(bt(Q.output)),Q.output=[])}}},Ei=(Q,m)=>{w.fill(0,Q,Q+m)},tn=(Q,m)=>Math.ceil(Q/m)*m,qi=Q=>{Q=tn(Q,65536);var m=vi(65536,Q);return m&&Ei(m,Q),m},xe={ops_table:null,mount(Q){return xe.createNode(null,"/",16895,0)},createNode(Q,m,v,p){if(M.isBlkdev(v)||M.isFIFO(v))throw new M.ErrnoError(63);xe.ops_table||={dir:{node:{getattr:xe.node_ops.getattr,setattr:xe.node_ops.setattr,lookup:xe.node_ops.lookup,mknod:xe.node_ops.mknod,rename:xe.node_ops.rename,unlink:xe.node_ops.unlink,rmdir:xe.node_ops.rmdir,readdir:xe.node_ops.readdir,symlink:xe.node_ops.symlink},stream:{llseek:xe.stream_ops.llseek}},file:{node:{getattr:xe.node_ops.getattr,setattr:xe.node_ops.setattr},stream:{llseek:xe.stream_ops.llseek,read:xe.stream_ops.read,write:xe.stream_ops.write,allocate:xe.stream_ops.allocate,mmap:xe.stream_ops.mmap,msync:xe.stream_ops.msync}},link:{node:{getattr:xe.node_ops.getattr,setattr:xe.node_ops.setattr,readlink:xe.node_ops.readlink},stream:{}},chrdev:{node:{getattr:xe.node_ops.getattr,setattr:xe.node_ops.setattr},stream:M.chrdev_stream_ops}};var N=M.createNode(Q,m,v,p);return M.isDir(N.mode)?(N.node_ops=xe.ops_table.dir.node,N.stream_ops=xe.ops_table.dir.stream,N.contents={}):M.isFile(N.mode)?(N.node_ops=xe.ops_table.file.node,N.stream_ops=xe.ops_table.file.stream,N.usedBytes=0,N.contents=null):M.isLink(N.mode)?(N.node_ops=xe.ops_table.link.node,N.stream_ops=xe.ops_table.link.stream):M.isChrdev(N.mode)&&(N.node_ops=xe.ops_table.chrdev.node,N.stream_ops=xe.ops_table.chrdev.stream),N.timestamp=Date.now(),Q&&(Q.contents[m]=N,Q.timestamp=N.timestamp),N},getFileDataAsTypedArray(Q){return Q.contents?Q.contents.subarray?Q.contents.subarray(0,Q.usedBytes):new Uint8Array(Q.contents):new Uint8Array(0)},expandFileStorage(Q,m){var v=Q.contents?Q.contents.length:0;if(!(v>=m)){var p=1024*1024;m=Math.max(m,v*(v<p?2:1.125)>>>0),v!=0&&(m=Math.max(m,256));var N=Q.contents;Q.contents=new Uint8Array(m),Q.usedBytes>0&&Q.contents.set(N.subarray(0,Q.usedBytes),0)}},resizeFileStorage(Q,m){if(Q.usedBytes!=m)if(m==0)Q.contents=null,Q.usedBytes=0;else{var v=Q.contents;Q.contents=new Uint8Array(m),v&&Q.contents.set(v.subarray(0,Math.min(m,Q.usedBytes))),Q.usedBytes=m}},node_ops:{getattr(Q){var m={};return m.dev=M.isChrdev(Q.mode)?Q.id:1,m.ino=Q.id,m.mode=Q.mode,m.nlink=1,m.uid=0,m.gid=0,m.rdev=Q.rdev,M.isDir(Q.mode)?m.size=4096:M.isFile(Q.mode)?m.size=Q.usedBytes:M.isLink(Q.mode)?m.size=Q.link.length:m.size=0,m.atime=new Date(Q.timestamp),m.mtime=new Date(Q.timestamp),m.ctime=new Date(Q.timestamp),m.blksize=4096,m.blocks=Math.ceil(m.size/m.blksize),m},setattr(Q,m){m.mode!==void 0&&(Q.mode=m.mode),m.timestamp!==void 0&&(Q.timestamp=m.timestamp),m.size!==void 0&&xe.resizeFileStorage(Q,m.size)},lookup(Q,m){throw M.genericErrors[44]},mknod(Q,m,v,p){return xe.createNode(Q,m,v,p)},rename(Q,m,v){if(M.isDir(Q.mode)){var p;try{p=M.lookupNode(m,v)}catch{}if(p)for(var N in p.contents)throw new M.ErrnoError(55)}delete Q.parent.contents[Q.name],Q.parent.timestamp=Date.now(),Q.name=v,m.contents[v]=Q,m.timestamp=Q.parent.timestamp},unlink(Q,m){delete Q.contents[m],Q.timestamp=Date.now()},rmdir(Q,m){var v=M.lookupNode(Q,m);for(var p in v.contents)throw new M.ErrnoError(55);delete Q.contents[m],Q.timestamp=Date.now()},readdir(Q){var m=[".",".."];for(var v of Object.keys(Q.contents))m.push(v);return m},symlink(Q,m,v){var p=xe.createNode(Q,m,41471,0);return p.link=v,p},readlink(Q){if(!M.isLink(Q.mode))throw new M.ErrnoError(28);return Q.link}},stream_ops:{read(Q,m,v,p,N){var J=Q.node.contents;if(N>=Q.node.usedBytes)return 0;var V=Math.min(Q.node.usedBytes-N,p);if(V>8&&J.subarray)m.set(J.subarray(N,N+V),v);else for(var Z=0;Z<V;Z++)m[v+Z]=J[N+Z];return V},write(Q,m,v,p,N,J){if(m.buffer===R.buffer&&(J=!1),!p)return 0;var V=Q.node;if(V.timestamp=Date.now(),m.subarray&&(!V.contents||V.contents.subarray)){if(J)return V.contents=m.subarray(v,v+p),V.usedBytes=p,p;if(V.usedBytes===0&&N===0)return V.contents=m.slice(v,v+p),V.usedBytes=p,p;if(N+p<=V.usedBytes)return V.contents.set(m.subarray(v,v+p),N),p}if(xe.expandFileStorage(V,N+p),V.contents.subarray&&m.subarray)V.contents.set(m.subarray(v,v+p),N);else for(var Z=0;Z<p;Z++)V.contents[N+Z]=m[v+Z];return V.usedBytes=Math.max(V.usedBytes,N+p),p},llseek(Q,m,v){var p=m;if(v===1?p+=Q.position:v===2&&M.isFile(Q.node.mode)&&(p+=Q.node.usedBytes),p<0)throw new M.ErrnoError(28);return p},allocate(Q,m,v){xe.expandFileStorage(Q.node,m+v),Q.node.usedBytes=Math.max(Q.node.usedBytes,m+v)},mmap(Q,m,v,p,N){if(!M.isFile(Q.node.mode))throw new M.ErrnoError(43);var J,V,Z=Q.node.contents;if(!(N&2)&&Z&&Z.buffer===R.buffer)V=!1,J=Z.byteOffset;else{if(V=!0,J=qi(m),!J)throw new M.ErrnoError(48);Z&&((v>0||v+m<Z.length)&&(Z.subarray?Z=Z.subarray(v,v+m):Z=Array.prototype.slice.call(Z,v,v+m)),R.set(Z,J))}return{ptr:J,allocated:V}},msync(Q,m,v,p,N){return xe.stream_ops.write(Q,m,0,p,v,!1),0}}},nn=(Q,m,v,p)=>{var N=p?"":`al ${Q}`;I(Q).then(J=>{m(new Uint8Array(J)),N&&CA()},J=>{if(v)v();else throw`Loading data file "${Q}" failed.`}),N&&KA()},mi=(Q,m,v,p,N,J)=>{M.createDataFile(Q,m,v,p,N,J)},Ot=[],Lt=(Q,m,v,p)=>{typeof Browser<"u"&&Browser.init();var N=!1;return Ot.forEach(J=>{N||J.canHandle(m)&&(J.handle(Q,m,v,p),N=!0)}),N},ii=(Q,m,v,p,N,J,V,Z,FA,te)=>{var re=m?Ee.resolve(X.join2(Q,m)):Q;function Pe(ze){function ye(Ge){te?.(),Z||mi(Q,m,Ge,p,N,FA),J?.(),CA()}Lt(ze,re,ye,()=>{V?.(),CA()})||ye(ze)}KA(),typeof v=="string"?nn(v,Pe,V):Pe(v)},_i=Q=>{var m={r:0,"r+":2,w:577,"w+":578,a:1089,"a+":1090},v=m[Q];if(typeof v>"u")throw new Error(`Unknown file open mode: ${Q}`);return v},Tt=(Q,m)=>{var v=0;return Q&&(v|=365),m&&(v|=146),v},M={root:null,mounts:[],devices:{},streams:[],nextInode:1,nameTable:null,currentPath:"/",initialized:!1,ignorePermissions:!0,ErrnoError:class{constructor(Q){this.name="ErrnoError",this.errno=Q}},genericErrors:{},filesystems:null,syncFSRequests:0,FSStream:class{constructor(){this.shared={}}get object(){return this.node}set object(Q){this.node=Q}get isRead(){return(this.flags&2097155)!==1}get isWrite(){return(this.flags&2097155)!==0}get isAppend(){return this.flags&1024}get flags(){return this.shared.flags}set flags(Q){this.shared.flags=Q}get position(){return this.shared.position}set position(Q){this.shared.position=Q}},FSNode:class{constructor(Q,m,v,p){Q||(Q=this),this.parent=Q,this.mount=Q.mount,this.mounted=null,this.id=M.nextInode++,this.name=m,this.mode=v,this.node_ops={},this.stream_ops={},this.rdev=p,this.readMode=365,this.writeMode=146}get read(){return(this.mode&this.readMode)===this.readMode}set read(Q){Q?this.mode|=this.readMode:this.mode&=~this.readMode}get write(){return(this.mode&this.writeMode)===this.writeMode}set write(Q){Q?this.mode|=this.writeMode:this.mode&=~this.writeMode}get isFolder(){return M.isDir(this.mode)}get isDevice(){return M.isChrdev(this.mode)}},lookupPath(Q,m={}){if(Q=Ee.resolve(Q),!Q)return{path:"",node:null};var v={follow_mount:!0,recurse_count:0};if(m=Object.assign(v,m),m.recurse_count>8)throw new M.ErrnoError(32);for(var p=Q.split("/").filter(Pe=>!!Pe),N=M.root,J="/",V=0;V<p.length;V++){var Z=V===p.length-1;if(Z&&m.parent)break;if(N=M.lookupNode(N,p[V]),J=X.join2(J,p[V]),M.isMountpoint(N)&&(!Z||Z&&m.follow_mount)&&(N=N.mounted.root),!Z||m.follow)for(var FA=0;M.isLink(N.mode);){var te=M.readlink(J);J=Ee.resolve(X.dirname(J),te);var re=M.lookupPath(J,{recurse_count:m.recurse_count+1});if(N=re.node,FA++>40)throw new M.ErrnoError(32)}}return{path:J,node:N}},getPath(Q){for(var m;;){if(M.isRoot(Q)){var v=Q.mount.mountpoint;return m?v[v.length-1]!=="/"?`${v}/${m}`:v+m:v}m=m?`${Q.name}/${m}`:Q.name,Q=Q.parent}},hashName(Q,m){for(var v=0,p=0;p<m.length;p++)v=(v<<5)-v+m.charCodeAt(p)|0;return(Q+v>>>0)%M.nameTable.length},hashAddNode(Q){var m=M.hashName(Q.parent.id,Q.name);Q.name_next=M.nameTable[m],M.nameTable[m]=Q},hashRemoveNode(Q){var m=M.hashName(Q.parent.id,Q.name);if(M.nameTable[m]===Q)M.nameTable[m]=Q.name_next;else for(var v=M.nameTable[m];v;){if(v.name_next===Q){v.name_next=Q.name_next;break}v=v.name_next}},lookupNode(Q,m){var v=M.mayLookup(Q);if(v)throw new M.ErrnoError(v);for(var p=M.hashName(Q.id,m),N=M.nameTable[p];N;N=N.name_next){var J=N.name;if(N.parent.id===Q.id&&J===m)return N}return M.lookup(Q,m)},createNode(Q,m,v,p){var N=new M.FSNode(Q,m,v,p);return M.hashAddNode(N),N},destroyNode(Q){M.hashRemoveNode(Q)},isRoot(Q){return Q===Q.parent},isMountpoint(Q){return!!Q.mounted},isFile(Q){return(Q&61440)===32768},isDir(Q){return(Q&61440)===16384},isLink(Q){return(Q&61440)===40960},isChrdev(Q){return(Q&61440)===8192},isBlkdev(Q){return(Q&61440)===24576},isFIFO(Q){return(Q&61440)===4096},isSocket(Q){return(Q&49152)===49152},flagsToPermissionString(Q){var m=["r","w","rw"][Q&3];return Q&512&&(m+="w"),m},nodePermissions(Q,m){return M.ignorePermissions?0:m.includes("r")&&!(Q.mode&292)||m.includes("w")&&!(Q.mode&146)||m.includes("x")&&!(Q.mode&73)?2:0},mayLookup(Q){if(!M.isDir(Q.mode))return 54;var m=M.nodePermissions(Q,"x");return m||(Q.node_ops.lookup?0:2)},mayCreate(Q,m){try{var v=M.lookupNode(Q,m);return 20}catch{}return M.nodePermissions(Q,"wx")},mayDelete(Q,m,v){var p;try{p=M.lookupNode(Q,m)}catch(J){return J.errno}var N=M.nodePermissions(Q,"wx");if(N)return N;if(v){if(!M.isDir(p.mode))return 54;if(M.isRoot(p)||M.getPath(p)===M.cwd())return 10}else if(M.isDir(p.mode))return 31;return 0},mayOpen(Q,m){return Q?M.isLink(Q.mode)?32:M.isDir(Q.mode)&&(M.flagsToPermissionString(m)!=="r"||m&512)?31:M.nodePermissions(Q,M.flagsToPermissionString(m)):44},MAX_OPEN_FDS:4096,nextfd(){for(var Q=0;Q<=M.MAX_OPEN_FDS;Q++)if(!M.streams[Q])return Q;throw new M.ErrnoError(33)},getStreamChecked(Q){var m=M.getStream(Q);if(!m)throw new M.ErrnoError(8);return m},getStream:Q=>M.streams[Q],createStream(Q,m=-1){return Q=Object.assign(new M.FSStream,Q),m==-1&&(m=M.nextfd()),Q.fd=m,M.streams[m]=Q,Q},closeStream(Q){M.streams[Q]=null},dupStream(Q,m=-1){var v=M.createStream(Q,m);return v.stream_ops?.dup?.(v),v},chrdev_stream_ops:{open(Q){var m=M.getDevice(Q.node.rdev);Q.stream_ops=m.stream_ops,Q.stream_ops.open?.(Q)},llseek(){throw new M.ErrnoError(70)}},major:Q=>Q>>8,minor:Q=>Q&255,makedev:(Q,m)=>Q<<8|m,registerDevice(Q,m){M.devices[Q]={stream_ops:m}},getDevice:Q=>M.devices[Q],getMounts(Q){for(var m=[],v=[Q];v.length;){var p=v.pop();m.push(p),v.push(...p.mounts)}return m},syncfs(Q,m){typeof Q=="function"&&(m=Q,Q=!1),M.syncFSRequests++,M.syncFSRequests>1&&B(`warning: ${M.syncFSRequests} FS.syncfs operations in flight at once, probably just doing extra work`);var v=M.getMounts(M.root.mount),p=0;function N(V){return M.syncFSRequests--,m(V)}function J(V){if(V)return J.errored?void 0:(J.errored=!0,N(V));++p>=v.length&&N(null)}v.forEach(V=>{if(!V.type.syncfs)return J(null);V.type.syncfs(V,Q,J)})},mount(Q,m,v){var p=v==="/",N=!v,J;if(p&&M.root)throw new M.ErrnoError(10);if(!p&&!N){var V=M.lookupPath(v,{follow_mount:!1});if(v=V.path,J=V.node,M.isMountpoint(J))throw new M.ErrnoError(10);if(!M.isDir(J.mode))throw new M.ErrnoError(54)}var Z={type:Q,opts:m,mountpoint:v,mounts:[]},FA=Q.mount(Z);return FA.mount=Z,Z.root=FA,p?M.root=FA:J&&(J.mounted=Z,J.mount&&J.mount.mounts.push(Z)),FA},unmount(Q){var m=M.lookupPath(Q,{follow_mount:!1});if(!M.isMountpoint(m.node))throw new M.ErrnoError(28);var v=m.node,p=v.mounted,N=M.getMounts(p);Object.keys(M.nameTable).forEach(V=>{for(var Z=M.nameTable[V];Z;){var FA=Z.name_next;N.includes(Z.mount)&&M.destroyNode(Z),Z=FA}}),v.mounted=null;var J=v.mount.mounts.indexOf(p);v.mount.mounts.splice(J,1)},lookup(Q,m){return Q.node_ops.lookup(Q,m)},mknod(Q,m,v){var p=M.lookupPath(Q,{parent:!0}),N=p.node,J=X.basename(Q);if(!J||J==="."||J==="..")throw new M.ErrnoError(28);var V=M.mayCreate(N,J);if(V)throw new M.ErrnoError(V);if(!N.node_ops.mknod)throw new M.ErrnoError(63);return N.node_ops.mknod(N,J,m,v)},create(Q,m){return m=m!==void 0?m:438,m&=4095,m|=32768,M.mknod(Q,m,0)},mkdir(Q,m){return m=m!==void 0?m:511,m&=1023,m|=16384,M.mknod(Q,m,0)},mkdirTree(Q,m){for(var v=Q.split("/"),p="",N=0;N<v.length;++N)if(v[N]){p+="/"+v[N];try{M.mkdir(p,m)}catch(J){if(J.errno!=20)throw J}}},mkdev(Q,m,v){return typeof v>"u"&&(v=m,m=438),m|=8192,M.mknod(Q,m,v)},symlink(Q,m){if(!Ee.resolve(Q))throw new M.ErrnoError(44);var v=M.lookupPath(m,{parent:!0}),p=v.node;if(!p)throw new M.ErrnoError(44);var N=X.basename(m),J=M.mayCreate(p,N);if(J)throw new M.ErrnoError(J);if(!p.node_ops.symlink)throw new M.ErrnoError(63);return p.node_ops.symlink(p,N,Q)},rename(Q,m){var v=X.dirname(Q),p=X.dirname(m),N=X.basename(Q),J=X.basename(m),V,Z,FA;if(V=M.lookupPath(Q,{parent:!0}),Z=V.node,V=M.lookupPath(m,{parent:!0}),FA=V.node,!Z||!FA)throw new M.ErrnoError(44);if(Z.mount!==FA.mount)throw new M.ErrnoError(75);var te=M.lookupNode(Z,N),re=Ee.relative(Q,p);if(re.charAt(0)!==".")throw new M.ErrnoError(28);if(re=Ee.relative(m,v),re.charAt(0)!==".")throw new M.ErrnoError(55);var Pe;try{Pe=M.lookupNode(FA,J)}catch{}if(te!==Pe){var ze=M.isDir(te.mode),ye=M.mayDelete(Z,N,ze);if(ye)throw new M.ErrnoError(ye);if(ye=Pe?M.mayDelete(FA,J,ze):M.mayCreate(FA,J),ye)throw new M.ErrnoError(ye);if(!Z.node_ops.rename)throw new M.ErrnoError(63);if(M.isMountpoint(te)||Pe&&M.isMountpoint(Pe))throw new M.ErrnoError(10);if(FA!==Z&&(ye=M.nodePermissions(Z,"w"),ye))throw new M.ErrnoError(ye);M.hashRemoveNode(te);try{Z.node_ops.rename(te,FA,J),te.parent=FA}catch(Ge){throw Ge}finally{M.hashAddNode(te)}}},rmdir(Q){var m=M.lookupPath(Q,{parent:!0}),v=m.node,p=X.basename(Q),N=M.lookupNode(v,p),J=M.mayDelete(v,p,!0);if(J)throw new M.ErrnoError(J);if(!v.node_ops.rmdir)throw new M.ErrnoError(63);if(M.isMountpoint(N))throw new M.ErrnoError(10);v.node_ops.rmdir(v,p),M.destroyNode(N)},readdir(Q){var m=M.lookupPath(Q,{follow:!0}),v=m.node;if(!v.node_ops.readdir)throw new M.ErrnoError(54);return v.node_ops.readdir(v)},unlink(Q){var m=M.lookupPath(Q,{parent:!0}),v=m.node;if(!v)throw new M.ErrnoError(44);var p=X.basename(Q),N=M.lookupNode(v,p),J=M.mayDelete(v,p,!1);if(J)throw new M.ErrnoError(J);if(!v.node_ops.unlink)throw new M.ErrnoError(63);if(M.isMountpoint(N))throw new M.ErrnoError(10);v.node_ops.unlink(v,p),M.destroyNode(N)},readlink(Q){var m=M.lookupPath(Q),v=m.node;if(!v)throw new M.ErrnoError(44);if(!v.node_ops.readlink)throw new M.ErrnoError(28);return Ee.resolve(M.getPath(v.parent),v.node_ops.readlink(v))},stat(Q,m){var v=M.lookupPath(Q,{follow:!m}),p=v.node;if(!p)throw new M.ErrnoError(44);if(!p.node_ops.getattr)throw new M.ErrnoError(63);return p.node_ops.getattr(p)},lstat(Q){return M.stat(Q,!0)},chmod(Q,m,v){var p;if(typeof Q=="string"){var N=M.lookupPath(Q,{follow:!v});p=N.node}else p=Q;if(!p.node_ops.setattr)throw new M.ErrnoError(63);p.node_ops.setattr(p,{mode:m&4095|p.mode&-4096,timestamp:Date.now()})},lchmod(Q,m){M.chmod(Q,m,!0)},fchmod(Q,m){var v=M.getStreamChecked(Q);M.chmod(v.node,m)},chown(Q,m,v,p){var N;if(typeof Q=="string"){var J=M.lookupPath(Q,{follow:!p});N=J.node}else N=Q;if(!N.node_ops.setattr)throw new M.ErrnoError(63);N.node_ops.setattr(N,{timestamp:Date.now()})},lchown(Q,m,v){M.chown(Q,m,v,!0)},fchown(Q,m,v){var p=M.getStreamChecked(Q);M.chown(p.node,m,v)},truncate(Q,m){if(m<0)throw new M.ErrnoError(28);var v;if(typeof Q=="string"){var p=M.lookupPath(Q,{follow:!0});v=p.node}else v=Q;if(!v.node_ops.setattr)throw new M.ErrnoError(63);if(M.isDir(v.mode))throw new M.ErrnoError(31);if(!M.isFile(v.mode))throw new M.ErrnoError(28);var N=M.nodePermissions(v,"w");if(N)throw new M.ErrnoError(N);v.node_ops.setattr(v,{size:m,timestamp:Date.now()})},ftruncate(Q,m){var v=M.getStreamChecked(Q);if((v.flags&2097155)===0)throw new M.ErrnoError(28);M.truncate(v.node,m)},utime(Q,m,v){var p=M.lookupPath(Q,{follow:!0}),N=p.node;N.node_ops.setattr(N,{timestamp:Math.max(m,v)})},open(Q,m,v){if(Q==="")throw new M.ErrnoError(44);m=typeof m=="string"?_i(m):m,m&64?(v=typeof v>"u"?438:v,v=v&4095|32768):v=0;var p;if(typeof Q=="object")p=Q;else{Q=X.normalize(Q);try{var N=M.lookupPath(Q,{follow:!(m&131072)});p=N.node}catch{}}var J=!1;if(m&64)if(p){if(m&128)throw new M.ErrnoError(20)}else p=M.mknod(Q,v,0),J=!0;if(!p)throw new M.ErrnoError(44);if(M.isChrdev(p.mode)&&(m&=-513),m&65536&&!M.isDir(p.mode))throw new M.ErrnoError(54);if(!J){var V=M.mayOpen(p,m);if(V)throw new M.ErrnoError(V)}m&512&&!J&&M.truncate(p,0),m&=-131713;var Z=M.createStream({node:p,path:M.getPath(p),flags:m,seekable:!0,position:0,stream_ops:p.stream_ops,ungotten:[],error:!1});return Z.stream_ops.open&&Z.stream_ops.open(Z),Z},close(Q){if(M.isClosed(Q))throw new M.ErrnoError(8);Q.getdents&&(Q.getdents=null);try{Q.stream_ops.close&&Q.stream_ops.close(Q)}catch(m){throw m}finally{M.closeStream(Q.fd)}Q.fd=null},isClosed(Q){return Q.fd===null},llseek(Q,m,v){if(M.isClosed(Q))throw new M.ErrnoError(8);if(!Q.seekable||!Q.stream_ops.llseek)throw new M.ErrnoError(70);if(v!=0&&v!=1&&v!=2)throw new M.ErrnoError(28);return Q.position=Q.stream_ops.llseek(Q,m,v),Q.ungotten=[],Q.position},read(Q,m,v,p,N){if(p<0||N<0)throw new M.ErrnoError(28);if(M.isClosed(Q))throw new M.ErrnoError(8);if((Q.flags&2097155)===1)throw new M.ErrnoError(8);if(M.isDir(Q.node.mode))throw new M.ErrnoError(31);if(!Q.stream_ops.read)throw new M.ErrnoError(28);var J=typeof N<"u";if(!J)N=Q.position;else if(!Q.seekable)throw new M.ErrnoError(70);var V=Q.stream_ops.read(Q,m,v,p,N);return J||(Q.position+=V),V},write(Q,m,v,p,N,J){if(p<0||N<0)throw new M.ErrnoError(28);if(M.isClosed(Q))throw new M.ErrnoError(8);if((Q.flags&2097155)===0)throw new M.ErrnoError(8);if(M.isDir(Q.node.mode))throw new M.ErrnoError(31);if(!Q.stream_ops.write)throw new M.ErrnoError(28);Q.seekable&&Q.flags&1024&&M.llseek(Q,0,2);var V=typeof N<"u";if(!V)N=Q.position;else if(!Q.seekable)throw new M.ErrnoError(70);var Z=Q.stream_ops.write(Q,m,v,p,N,J);return V||(Q.position+=Z),Z},allocate(Q,m,v){if(M.isClosed(Q))throw new M.ErrnoError(8);if(m<0||v<=0)throw new M.ErrnoError(28);if((Q.flags&2097155)===0)throw new M.ErrnoError(8);if(!M.isFile(Q.node.mode)&&!M.isDir(Q.node.mode))throw new M.ErrnoError(43);if(!Q.stream_ops.allocate)throw new M.ErrnoError(138);Q.stream_ops.allocate(Q,m,v)},mmap(Q,m,v,p,N){if((p&2)!==0&&(N&2)===0&&(Q.flags&2097155)!==2)throw new M.ErrnoError(2);if((Q.flags&2097155)===1)throw new M.ErrnoError(2);if(!Q.stream_ops.mmap)throw new M.ErrnoError(43);if(!m)throw new M.ErrnoError(28);return Q.stream_ops.mmap(Q,m,v,p,N)},msync(Q,m,v,p,N){return Q.stream_ops.msync?Q.stream_ops.msync(Q,m,v,p,N):0},ioctl(Q,m,v){if(!Q.stream_ops.ioctl)throw new M.ErrnoError(59);return Q.stream_ops.ioctl(Q,m,v)},readFile(Q,m={}){if(m.flags=m.flags||0,m.encoding=m.encoding||"binary",m.encoding!=="utf8"&&m.encoding!=="binary")throw new Error(`Invalid encoding type "${m.encoding}"`);var v,p=M.open(Q,m.flags),N=M.stat(Q),J=N.size,V=new Uint8Array(J);return M.read(p,V,0,J,0),m.encoding==="utf8"?v=bt(V):m.encoding==="binary"&&(v=V),M.close(p),v},writeFile(Q,m,v={}){v.flags=v.flags||577;var p=M.open(Q,v.flags,v.mode);if(typeof m=="string"){var N=new Uint8Array(DA(m)+1),J=gt(m,N,0,N.length);M.write(p,N,0,J,void 0,v.canOwn)}else if(ArrayBuffer.isView(m))M.write(p,m,0,m.byteLength,void 0,v.canOwn);else throw new Error("Unsupported data type");M.close(p)},cwd:()=>M.currentPath,chdir(Q){var m=M.lookupPath(Q,{follow:!0});if(m.node===null)throw new M.ErrnoError(44);if(!M.isDir(m.node.mode))throw new M.ErrnoError(54);var v=M.nodePermissions(m.node,"x");if(v)throw new M.ErrnoError(v);M.currentPath=m.path},createDefaultDirectories(){M.mkdir("/tmp"),M.mkdir("/home"),M.mkdir("/home/web_user")},createDefaultDevices(){M.mkdir("/dev"),M.registerDevice(M.makedev(1,3),{read:()=>0,write:(p,N,J,V,Z)=>V}),M.mkdev("/dev/null",M.makedev(1,3)),rt.register(M.makedev(5,0),rt.default_tty_ops),rt.register(M.makedev(6,0),rt.default_tty1_ops),M.mkdev("/dev/tty",M.makedev(5,0)),M.mkdev("/dev/tty1",M.makedev(6,0));var Q=new Uint8Array(1024),m=0,v=()=>(m===0&&(m=Be(Q).byteLength),Q[--m]);M.createDevice("/dev","random",v),M.createDevice("/dev","urandom",v),M.mkdir("/dev/shm"),M.mkdir("/dev/shm/tmp")},createSpecialDirectories(){M.mkdir("/proc");var Q=M.mkdir("/proc/self");M.mkdir("/proc/self/fd"),M.mount({mount(){var m=M.createNode(Q,"fd",16895,73);return m.node_ops={lookup(v,p){var N=+p,J=M.getStreamChecked(N),V={parent:null,mount:{mountpoint:"fake"},node_ops:{readlink:()=>J.path}};return V.parent=V,V}},m}},{},"/proc/self/fd")},createStandardStreams(Q,m,v){Q?M.createDevice("/dev","stdin",Q):M.symlink("/dev/tty","/dev/stdin"),m?M.createDevice("/dev","stdout",null,m):M.symlink("/dev/tty","/dev/stdout"),v?M.createDevice("/dev","stderr",null,v):M.symlink("/dev/tty1","/dev/stderr"),M.open("/dev/stdin",0),M.open("/dev/stdout",1),M.open("/dev/stderr",1)},staticInit(){[44].forEach(Q=>{M.genericErrors[Q]=new M.ErrnoError(Q),M.genericErrors[Q].stack="<generic error, no stack>"}),M.nameTable=new Array(4096),M.mount(xe,{},"/"),M.createDefaultDirectories(),M.createDefaultDevices(),M.createSpecialDirectories(),M.filesystems={MEMFS:xe}},init(Q,m,v){M.initialized=!0,M.createStandardStreams(Q,m,v)},quit(){M.initialized=!1;for(var Q=0;Q<M.streams.length;Q++){var m=M.streams[Q];m&&M.close(m)}},findObject(Q,m){var v=M.analyzePath(Q,m);return v.exists?v.object:null},analyzePath(Q,m){try{var v=M.lookupPath(Q,{follow:!m});Q=v.path}catch{}var p={isRoot:!1,exists:!1,error:0,name:null,path:null,object:null,parentExists:!1,parentPath:null,parentObject:null};try{var v=M.lookupPath(Q,{parent:!0});p.parentExists=!0,p.parentPath=v.path,p.parentObject=v.node,p.name=X.basename(Q),v=M.lookupPath(Q,{follow:!m}),p.exists=!0,p.path=v.path,p.object=v.node,p.name=v.node.name,p.isRoot=v.path==="/"}catch(N){p.error=N.errno}return p},createPath(Q,m,v,p){Q=typeof Q=="string"?Q:M.getPath(Q);for(var N=m.split("/").reverse();N.length;){var J=N.pop();if(J){var V=X.join2(Q,J);try{M.mkdir(V)}catch{}Q=V}}return V},createFile(Q,m,v,p,N){var J=X.join2(typeof Q=="string"?Q:M.getPath(Q),m),V=Tt(p,N);return M.create(J,V)},createDataFile(Q,m,v,p,N,J){var V=m;Q&&(Q=typeof Q=="string"?Q:M.getPath(Q),V=m?X.join2(Q,m):Q);var Z=Tt(p,N),FA=M.create(V,Z);if(v){if(typeof v=="string"){for(var te=new Array(v.length),re=0,Pe=v.length;re<Pe;++re)te[re]=v.charCodeAt(re);v=te}M.chmod(FA,Z|146);var ze=M.open(FA,577);M.write(ze,v,0,v.length,0,J),M.close(ze),M.chmod(FA,Z)}},createDevice(Q,m,v,p){var N=X.join2(typeof Q=="string"?Q:M.getPath(Q),m),J=Tt(!!v,!!p);M.createDevice.major??=64;var V=M.makedev(M.createDevice.major++,0);return M.registerDevice(V,{open(Z){Z.seekable=!1},close(Z){p?.buffer?.length&&p(10)},read(Z,FA,te,re,Pe){for(var ze=0,ye=0;ye<re;ye++){var Ge;try{Ge=v()}catch{throw new M.ErrnoError(29)}if(Ge===void 0&&ze===0)throw new M.ErrnoError(6);if(Ge==null)break;ze++,FA[te+ye]=Ge}return ze&&(Z.node.timestamp=Date.now()),ze},write(Z,FA,te,re,Pe){for(var ze=0;ze<re;ze++)try{p(FA[te+ze])}catch{throw new M.ErrnoError(29)}return re&&(Z.node.timestamp=Date.now()),ze}}),M.mkdev(N,J,V)},forceLoadFile(Q){if(Q.isDevice||Q.isFolder||Q.link||Q.contents)return!0;if(typeof XMLHttpRequest<"u")throw new Error("Lazy loading should have been performed (contents set) in createLazyFile, but it was not. Lazy loading only works in web workers. Use --embed-file or --preload-file in emcc on the main thread.");try{Q.contents=C(Q.url),Q.usedBytes=Q.contents.length}catch{throw new M.ErrnoError(29)}},createLazyFile(Q,m,v,p,N){class J{constructor(){this.lengthKnown=!1,this.chunks=[]}get(ye){if(!(ye>this.length-1||ye<0)){var Ge=ye%this.chunkSize,Vi=ye/this.chunkSize|0;return this.getter(Vi)[Ge]}}setDataGetter(ye){this.getter=ye}cacheLength(){var ye=new XMLHttpRequest;if(ye.open("HEAD",v,!1),ye.send(null),!(ye.status>=200&&ye.status<300||ye.status===304))throw new Error("Couldn't load "+v+". Status: "+ye.status);var Ge=Number(ye.getResponseHeader("Content-length")),Vi,T=(Vi=ye.getResponseHeader("Accept-Ranges"))&&Vi==="bytes",oA=(Vi=ye.getResponseHeader("Content-Encoding"))&&Vi==="gzip",YA=1024*1024;T||(YA=Ge);var pe=(ge,Bt)=>{if(ge>Bt)throw new Error("invalid range ("+ge+", "+Bt+") or no bytes requested!");if(Bt>Ge-1)throw new Error("only "+Ge+" bytes available! programmer error!");var $e=new XMLHttpRequest;if($e.open("GET",v,!1),Ge!==YA&&$e.setRequestHeader("Range","bytes="+ge+"-"+Bt),$e.responseType="arraybuffer",$e.overrideMimeType&&$e.overrideMimeType("text/plain; charset=x-user-defined"),$e.send(null),!($e.status>=200&&$e.status<300||$e.status===304))throw new Error("Couldn't load "+v+". Status: "+$e.status);return $e.response!==void 0?new Uint8Array($e.response||[]):Ve($e.responseText||"",!0)},he=this;he.setDataGetter(ge=>{var Bt=ge*YA,$e=(ge+1)*YA-1;if($e=Math.min($e,Ge-1),typeof he.chunks[ge]>"u"&&(he.chunks[ge]=pe(Bt,$e)),typeof he.chunks[ge]>"u")throw new Error("doXHR failed!");return he.chunks[ge]}),(oA||!Ge)&&(YA=Ge=1,Ge=this.getter(0).length,YA=Ge,d("LazyFiles on gzip forces download of the whole file when length is accessed")),this._length=Ge,this._chunkSize=YA,this.lengthKnown=!0}get length(){return this.lengthKnown||this.cacheLength(),this._length}get chunkSize(){return this.lengthKnown||this.cacheLength(),this._chunkSize}}if(typeof XMLHttpRequest<"u"){throw"Cannot do synchronous binary XHRs outside webworkers in modern browsers. Use --embed-file or --preload-file in emcc";var V,Z}else var Z={isDevice:!1,url:v};var FA=M.createFile(Q,m,Z,p,N);Z.contents?FA.contents=Z.contents:Z.url&&(FA.contents=null,FA.url=Z.url),Object.defineProperties(FA,{usedBytes:{get:function(){return this.contents.length}}});var te={},re=Object.keys(FA.stream_ops);re.forEach(ze=>{var ye=FA.stream_ops[ze];te[ze]=(...Ge)=>(M.forceLoadFile(FA),ye(...Ge))});function Pe(ze,ye,Ge,Vi,T){var oA=ze.node.contents;if(T>=oA.length)return 0;var YA=Math.min(oA.length-T,Vi);if(oA.slice)for(var pe=0;pe<YA;pe++)ye[Ge+pe]=oA[T+pe];else for(var pe=0;pe<YA;pe++)ye[Ge+pe]=oA.get(T+pe);return YA}return te.read=(ze,ye,Ge,Vi,T)=>(M.forceLoadFile(FA),Pe(ze,ye,Ge,Vi,T)),te.mmap=(ze,ye,Ge,Vi,T)=>{M.forceLoadFile(FA);var oA=qi(ye);if(!oA)throw new M.ErrnoError(48);return Pe(ze,R,oA,ye,Ge),{ptr:oA,allocated:!0}},FA.stream_ops=te,FA}},We={DEFAULT_POLLMASK:5,calculateAt(Q,m,v){if(X.isAbs(m))return m;var p;if(Q===-100)p=M.cwd();else{var N=We.getStreamFromFD(Q);p=N.path}if(m.length==0){if(!v)throw new M.ErrnoError(44);return p}return X.join2(p,m)},doStat(Q,m,v){var p=Q(m);K[v>>2]=p.dev,K[v+4>>2]=p.mode,z[v+8>>2]=p.nlink,K[v+12>>2]=p.uid,K[v+16>>2]=p.gid,K[v+20>>2]=p.rdev,H[v+24>>3]=BigInt(p.size),K[v+32>>2]=4096,K[v+36>>2]=p.blocks;var N=p.atime.getTime(),J=p.mtime.getTime(),V=p.ctime.getTime();return H[v+40>>3]=BigInt(Math.floor(N/1e3)),z[v+48>>2]=N%1e3*1e3*1e3,H[v+56>>3]=BigInt(Math.floor(J/1e3)),z[v+64>>2]=J%1e3*1e3*1e3,H[v+72>>3]=BigInt(Math.floor(V/1e3)),z[v+80>>2]=V%1e3*1e3*1e3,H[v+88>>3]=BigInt(p.ino),0},doMsync(Q,m,v,p,N){if(!M.isFile(m.node.mode))throw new M.ErrnoError(43);if(p&2)return 0;var J=w.slice(Q,Q+v);M.msync(m,J,N,v,p)},getStreamFromFD(Q){var m=M.getStreamChecked(Q);return m},varargs:void 0,getStr(Q){var m=Re(Q);return m}};function ni(Q,m,v,p){try{if(m=We.getStr(m),m=We.calculateAt(Q,m),v&-8)return-28;var N=M.lookupPath(m,{follow:!0}),J=N.node;if(!J)return-44;var V="";return v&4&&(V+="r"),v&2&&(V+="w"),v&1&&(V+="x"),V&&M.nodePermissions(J,V)?-2:0}catch(Z){if(typeof M>"u"||Z.name!=="ErrnoError")throw Z;return-Z.errno}}function pi(){var Q=K[+We.varargs>>2];return We.varargs+=4,Q}var dn=pi;function mn(Q,m,v){We.varargs=v;try{var p=We.getStreamFromFD(Q);switch(m){case 0:{var N=pi();if(N<0)return-28;for(;M.streams[N];)N++;var J;return J=M.dupStream(p,N),J.fd}case 1:case 2:return 0;case 3:return p.flags;case 4:{var N=pi();return p.flags|=N,0}case 12:{var N=dn(),V=0;return _[N+V>>1]=2,0}case 13:case 14:return 0}return-28}catch(Z){if(typeof M>"u"||Z.name!=="ErrnoError")throw Z;return-Z.errno}}function Uo(Q,m){try{var v=We.getStreamFromFD(Q);return We.doStat(M.stat,v.path,m)}catch(p){if(typeof M>"u"||p.name!=="ErrnoError")throw p;return-p.errno}}function kn(Q,m,v){We.varargs=v;try{var p=We.getStreamFromFD(Q);switch(m){case 21509:return p.tty?0:-59;case 21505:{if(!p.tty)return-59;if(p.tty.ops.ioctl_tcgets){var N=p.tty.ops.ioctl_tcgets(p),J=dn();K[J>>2]=N.c_iflag||0,K[J+4>>2]=N.c_oflag||0,K[J+8>>2]=N.c_cflag||0,K[J+12>>2]=N.c_lflag||0;for(var V=0;V<32;V++)R[J+V+17]=N.c_cc[V]||0;return 0}return 0}case 21510:case 21511:case 21512:return p.tty?0:-59;case 21506:case 21507:case 21508:{if(!p.tty)return-59;if(p.tty.ops.ioctl_tcsets){for(var J=dn(),Z=K[J>>2],FA=K[J+4>>2],te=K[J+8>>2],re=K[J+12>>2],Pe=[],V=0;V<32;V++)Pe.push(R[J+V+17]);return p.tty.ops.ioctl_tcsets(p.tty,m,{c_iflag:Z,c_oflag:FA,c_cflag:te,c_lflag:re,c_cc:Pe})}return 0}case 21519:{if(!p.tty)return-59;var J=dn();return K[J>>2]=0,0}case 21520:return p.tty?-28:-59;case 21531:{var J=dn();return M.ioctl(p,m,J)}case 21523:{if(!p.tty)return-59;if(p.tty.ops.ioctl_tiocgwinsz){var ze=p.tty.ops.ioctl_tiocgwinsz(p.tty),J=dn();_[J>>1]=ze[0],_[J+2>>1]=ze[1]}return 0}case 21524:return p.tty?0:-59;case 21515:return p.tty?0:-59;default:return-28}}catch(ye){if(typeof M>"u"||ye.name!=="ErrnoError")throw ye;return-ye.errno}}function Wn(Q,m,v,p){try{m=We.getStr(m);var N=p&256,J=p&4096;return p=p&-6401,m=We.calculateAt(Q,m,J),We.doStat(N?M.lstat:M.stat,m,v)}catch(V){if(typeof M>"u"||V.name!=="ErrnoError")throw V;return-V.errno}}function Vo(Q,m,v,p){We.varargs=p;try{m=We.getStr(m),m=We.calculateAt(Q,m);var N=p?pi():0;return M.open(m,v,N).fd}catch(J){if(typeof M>"u"||J.name!=="ErrnoError")throw J;return-J.errno}}function vo(Q,m){try{return Q=We.getStr(Q),We.doStat(M.stat,Q,m)}catch(v){if(typeof M>"u"||v.name!=="ErrnoError")throw v;return-v.errno}}var bo=()=>{TA("")},Yn=Q=>Q%4===0&&(Q%100!==0||Q%400===0),Mo=[0,31,60,91,121,152,182,213,244,274,305,335],ne=[0,31,59,90,120,151,181,212,243,273,304,334],wi=Q=>{var m=Yn(Q.getFullYear()),v=m?Mo:ne,p=v[Q.getMonth()]+Q.getDate()-1;return p},MA=9007199254740992,me=-9007199254740992,nt=Q=>Q<me||Q>MA?NaN:Number(Q);function Wt(Q,m){Q=nt(Q);var v=new Date(Q*1e3);K[m>>2]=v.getSeconds(),K[m+4>>2]=v.getMinutes(),K[m+8>>2]=v.getHours(),K[m+12>>2]=v.getDate(),K[m+16>>2]=v.getMonth(),K[m+20>>2]=v.getFullYear()-1900,K[m+24>>2]=v.getDay();var p=wi(v)|0;K[m+28>>2]=p,K[m+36>>2]=-(v.getTimezoneOffset()*60);var N=new Date(v.getFullYear(),0,1),J=new Date(v.getFullYear(),6,1).getTimezoneOffset(),V=N.getTimezoneOffset(),Z=(J!=V&&v.getTimezoneOffset()==Math.min(V,J))|0;K[m+32>>2]=Z}function Xe(Q,m,v,p,N,J,V){N=nt(N);try{if(isNaN(N))return 61;var Z=We.getStreamFromFD(p),FA=M.mmap(Z,Q,N,m,v),te=FA.ptr;return K[J>>2]=FA.allocated,z[V>>2]=te,0}catch(re){if(typeof M>"u"||re.name!=="ErrnoError")throw re;return-re.errno}}function oi(Q,m,v,p,N,J){J=nt(J);try{var V=We.getStreamFromFD(N);v&2&&We.doMsync(Q,V,m,p,J)}catch(Z){if(typeof M>"u"||Z.name!=="ErrnoError")throw Z;return-Z.errno}}var Di=(Q,m,v)=>gt(Q,w,m,v),Ut=(Q,m,v,p)=>{var N=new Date().getFullYear(),J=new Date(N,0,1),V=new Date(N,6,1),Z=J.getTimezoneOffset(),FA=V.getTimezoneOffset(),te=Math.max(Z,FA);z[Q>>2]=te*60,K[m>>2]=+(Z!=FA);var re=ye=>{var Ge=ye>=0?"-":"+",Vi=Math.abs(ye),T=String(Math.floor(Vi/60)).padStart(2,"0"),oA=String(Vi%60).padStart(2,"0");return`UTC${Ge}${T}${oA}`},Pe=re(Z),ze=re(FA);FA<Z?(Di(Pe,v,17),Di(ze,p,17)):(Di(Pe,p,17),Di(ze,v,17))},cn=()=>Date.now(),ft=()=>2147483648,Qi=Q=>{var m=D.buffer,v=(Q-m.byteLength+65535)/65536|0;try{return D.grow(v),j(),1}catch{}},ot=Q=>{var m=w.length;Q>>>=0;var v=ft();if(Q>v)return!1;for(var p=1;p<=4;p*=2){var N=m*(1+.2/p);N=Math.min(N,Q+100663296);var J=Math.min(v,tn(Math.max(Q,N),65536)),V=Qi(J);if(V)return!0}return!1},Mt={},on=()=>a,hn=()=>{if(!hn.strings){var Q=(typeof navigator=="object"&&navigator.languages&&navigator.languages[0]||"C").replace("-","_")+".UTF-8",m={USER:"web_user",LOGNAME:"web_user",PATH:"/",PWD:"/",HOME:"/home/web_user",LANG:Q,_:on()};for(var v in Mt)Mt[v]===void 0?delete m[v]:m[v]=Mt[v];var p=[];for(var v in m)p.push(`${v}=${m[v]}`);hn.strings=p}return hn.strings},Ai=(Q,m)=>{for(var v=0;v<Q.length;++v)R[m++]=Q.charCodeAt(v);R[m]=0},Ki=(Q,m)=>{var v=0;return hn().forEach((p,N)=>{var J=m+v;z[Q+N*4>>2]=J,Ai(p,J),v+=p.length+1}),0},dt=(Q,m)=>{var v=hn();z[Q>>2]=v.length;var p=0;return v.forEach(N=>p+=N.length+1),z[m>>2]=p,0},EA=Q=>{c(Q,new xt(Q))},HA=(Q,m)=>{EA(Q)},ve=HA;function Qt(Q){try{var m=We.getStreamFromFD(Q);return M.close(m),0}catch(v){if(typeof M>"u"||v.name!=="ErrnoError")throw v;return v.errno}}var yi=(Q,m,v,p)=>{for(var N=0,J=0;J<v;J++){var V=z[m>>2],Z=z[m+4>>2];m+=8;var FA=M.read(Q,R,V,Z,p);if(FA<0)return-1;if(N+=FA,FA<Z)break;typeof p<"u"&&(p+=FA)}return N};function ri(Q,m,v,p){try{var N=We.getStreamFromFD(Q),J=yi(N,m,v);return z[p>>2]=J,0}catch(V){if(typeof M>"u"||V.name!=="ErrnoError")throw V;return V.errno}}function pn(Q,m,v,p){m=nt(m);try{if(isNaN(m))return 61;var N=We.getStreamFromFD(Q);return M.llseek(N,m,v),H[p>>3]=BigInt(N.position),N.getdents&&m===0&&v===0&&(N.getdents=null),0}catch(J){if(typeof M>"u"||J.name!=="ErrnoError")throw J;return J.errno}}var Fn=(Q,m,v,p)=>{for(var N=0,J=0;J<v;J++){var V=z[m>>2],Z=z[m+4>>2];m+=8;var FA=M.write(Q,R,V,Z,p);if(FA<0)return-1;if(N+=FA,FA<Z)break;typeof p<"u"&&(p+=FA)}return N};function Jn(Q,m,v,p){try{var N=We.getStreamFromFD(Q),J=Fn(N,m,v);return z[p>>2]=J,0}catch(V){if(typeof M>"u"||V.name!=="ErrnoError")throw V;return V.errno}}var ln=Q=>{var m=i["_"+Q];return m},Pt=(Q,m)=>{R.set(Q,m)},$i=Q=>rn(Q),Rr=Q=>{var m=DA(Q)+1,v=$i(m);return Di(Q,v,m),v},Ft=(Q,m,v,p,N)=>{var J={string:Ge=>{var Vi=0;return Ge!=null&&Ge!==0&&(Vi=Rr(Ge)),Vi},array:Ge=>{var Vi=$i(Ge.length);return Pt(Ge,Vi),Vi}};function V(Ge){return m==="string"?Re(Ge):m==="boolean"?!!Ge:Ge}var Z=ln(Q),FA=[],te=0;if(p)for(var re=0;re<p.length;re++){var Pe=J[v[re]];Pe?(te===0&&(te=_e()),FA[re]=Pe(p[re])):FA[re]=p[re]}var ze=Z(...FA);function ye(Ge){return te!==0&&Dt(te),V(Ge)}return ze=ye(ze),ze};M.createPreloadedFile=ii,M.staticInit();var Xn={a:$t,b:P,x:ni,f:mn,w:Uo,j:kn,u:Wn,l:Vo,v:vo,i:bo,p:Wt,m:Xe,n:oi,q:Ut,g:cn,k:ot,s:Ki,t:dt,h:ve,e:Qt,d:ri,r:pn,c:Jn,o:sA},se=mA();i._viz_set_y_invert=Q=>(i._viz_set_y_invert=se.A)(Q),i._viz_set_reduce=Q=>(i._viz_set_reduce=se.B)(Q),i._viz_get_graphviz_version=()=>(i._viz_get_graphviz_version=se.C)(),i._free=Q=>(i._free=se.D)(Q),i._malloc=Q=>(i._malloc=se.E)(Q),i._viz_get_plugin_list=Q=>(i._viz_get_plugin_list=se.G)(Q),i._viz_create_graph=(Q,m,v)=>(i._viz_create_graph=se.H)(Q,m,v),i._viz_read_one_graph=Q=>(i._viz_read_one_graph=se.I)(Q),i._viz_string_dup=(Q,m)=>(i._viz_string_dup=se.J)(Q,m),i._viz_string_dup_html=(Q,m)=>(i._viz_string_dup_html=se.K)(Q,m),i._viz_string_free=(Q,m)=>(i._viz_string_free=se.L)(Q,m),i._viz_string_free_html=(Q,m)=>(i._viz_string_free_html=se.M)(Q,m),i._viz_add_node=(Q,m)=>(i._viz_add_node=se.N)(Q,m),i._viz_add_edge=(Q,m,v)=>(i._viz_add_edge=se.O)(Q,m,v),i._viz_add_subgraph=(Q,m)=>(i._viz_add_subgraph=se.P)(Q,m),i._viz_set_default_graph_attribute=(Q,m,v)=>(i._viz_set_default_graph_attribute=se.Q)(Q,m,v),i._viz_set_default_node_attribute=(Q,m,v)=>(i._viz_set_default_node_attribute=se.R)(Q,m,v),i._viz_set_default_edge_attribute=(Q,m,v)=>(i._viz_set_default_edge_attribute=se.S)(Q,m,v),i._viz_set_attribute=(Q,m,v)=>(i._viz_set_attribute=se.T)(Q,m,v),i._viz_free_graph=Q=>(i._viz_free_graph=se.U)(Q),i._viz_create_context=()=>(i._viz_create_context=se.V)(),i._viz_free_context=Q=>(i._viz_free_context=se.W)(Q),i._viz_layout=(Q,m,v)=>(i._viz_layout=se.X)(Q,m,v),i._viz_free_layout=(Q,m)=>(i._viz_free_layout=se.Y)(Q,m),i._viz_reset_errors=()=>(i._viz_reset_errors=se.Z)(),i._viz_render=(Q,m,v)=>(i._viz_render=se._)(Q,m,v);var vi=(Q,m)=>(vi=se.$)(Q,m),Yi=Q=>(Yi=se.aa)(Q),rn=Q=>(rn=se.ba)(Q),Hr=()=>(Hr=se.ca)();i.ccall=Ft,i.getValue=de,i.PATH=X,i.UTF8ToString=Re,i.stringToUTF8=Di,i.lengthBytesUTF8=DA,i.FS=M;var Ri,fs;VA=function Q(){Ri||Bo(),Ri||(VA=Q)};function Bo(){if(pA>0||!fs&&(fs=1,lA(),pA>0))return;function Q(){Ri||(Ri=1,i.calledRun=1,!L&&(vA(),n(i),tA()))}Q()}return Bo(),A=r,A}})(),KP=[[/^Error: (.*)/,"error"],[/^Warning: (.*)/,"warning"]];function M3A(t){return t.map(e=>{for(let A=0;A<KP.length;A++){let[i,n]=KP[A],o;if((o=i.exec(e))!==null)return{message:o[1].trimEnd(),level:n}}return{message:e.trimEnd()}})}function k3A(t){let e=[],A;for(let i=0;i<t.length;i++)t[i]=="Error"&&t[i+1]==": "?(A="error",i+=1):t[i]=="Warning"&&t[i+1]==": "?(A="warning",i+=1):e.push({message:t[i].trimEnd(),level:A});return e}function S3A(t){let e=t.ccall("viz_get_graphviz_version","number",[],[]);return t.UTF8ToString(e)}function YP(t,e){let A=t.ccall("viz_get_plugin_list","number",["string"],[e]);if(A==0)throw new Error(`couldn't get plugin list: ${e}`);let i=[],n=A,o;for(;o=t.getValue(n,"*");)i.push(t.UTF8ToString(o)),t.ccall("free","number",["number"],[o]),n+=4;return t.ccall("free","number",["number"],[A]),i}function JP(t,e,A,i){let n,o,r,s;try{if(t.agerrMessages=[],t.stderrMessages=[],s=R3A(t,i.images),typeof e=="string")n=L3A(t,e,i);else if(typeof e=="object")n=F3A(t,e,i);else throw new Error("input must be a string or object");if(n===0)return{status:"failure",output:void 0,errors:Ku(t)};if(zP(t,n,i),t.ccall("viz_set_y_invert","number",["number"],[i.yInvert?1:0]),t.ccall("viz_set_reduce","number",["number"],[i.reduce?1:0]),o=t.ccall("viz_create_context"),t.ccall("viz_reset_errors"),t.ccall("viz_layout","number",["number","number","string"],[o,n,i.engine])!==0)return{status:"failure",output:void 0,errors:Ku(t)};let c={};for(let l of A){if(r=t.ccall("viz_render","number",["number","number","string"],[o,n,l]),r===0)return{status:"failure",output:void 0,errors:Ku(t)};c[l]=t.UTF8ToString(r),t.ccall("free","number",["number"],[r]),r=0}return{status:"success",output:c,errors:Ku(t)}}catch(a){if(/^exit\(\d+\)/.test(a))return{status:"failure",output:void 0,errors:Ku(t)};throw a}finally{o&&n&&t.ccall("viz_free_layout","number",["number"],[o,n]),n&&t.ccall("viz_free_graph","number",["number"],[n]),o&&t.ccall("viz_free_context","number",["number"],[o]),r&&t.ccall("free","number",["number"],[r]),s&&x3A(t,s)}}function Ku(t){return k3A(t.agerrMessages).concat(M3A(t.stderrMessages))}function R3A(t,e){return e?e.map(A=>{if(typeof A.name!="string")throw new Error("image name must be a string");if(typeof A.width!="number"&&typeof A.width!="string")throw new Error("image width must be a number or string");if(typeof A.height!="number"&&typeof A.height!="string")throw new Error("image height must be a number or string");let i=t.PATH.join("/",A.name),n=`<?xml version="1.0" encoding="UTF-8"?>
<svg xmlns="http://www.w3.org/2000/svg" width="${A.width}" height="${A.height}"></svg>
`;return t.FS.createPath("/",t.PATH.dirname(i)),t.FS.writeFile(i,n),i}):[]}function x3A(t,e){for(let A of e)t.FS.analyzePath(A).exists&&t.FS.unlink(A)}function L3A(t,e,A){let i;try{let n=t.lengthBytesUTF8(e);return i=t.ccall("malloc","number",["number"],[n+1]),t.stringToUTF8(e,i,n+1),t.ccall("viz_read_one_graph","number",["number"],[i])}finally{i&&t.ccall("free","number",["number"],[i])}}function F3A(t,e,A){let i=t.ccall("viz_create_graph","number",["string","number","number"],[e.name,typeof e.directed<"u"?e.directed:!0,typeof e.strict<"u"?e.strict:!1]);return HP(t,i,e),i}function HP(t,e,A){zP(t,e,A),A.nodes&&A.nodes.forEach(i=>{let n=t.ccall("viz_add_node","number",["number","string"],[e,String(i.name)]);i.attributes&&TP(t,e,n,i.attributes)}),A.edges&&A.edges.forEach(i=>{let n=t.ccall("viz_add_edge","number",["number","string","string"],[e,String(i.tail),String(i.head)]);i.attributes&&TP(t,e,n,i.attributes)}),A.subgraphs&&A.subgraphs.forEach(i=>{let n=t.ccall("viz_add_subgraph","number",["number","string"],[e,String(i.name)]);HP(t,n,i)})}function zP(t,e,A){if(A.graphAttributes)for(let[i,n]of Object.entries(A.graphAttributes))m8(t,e,n,o=>{t.ccall("viz_set_default_graph_attribute","number",["number","string","number"],[e,i,o])});if(A.nodeAttributes)for(let[i,n]of Object.entries(A.nodeAttributes))m8(t,e,n,o=>{t.ccall("viz_set_default_node_attribute","number",["number","string","number"],[e,i,o])});if(A.edgeAttributes)for(let[i,n]of Object.entries(A.edgeAttributes))m8(t,e,n,o=>{t.ccall("viz_set_default_edge_attribute","number",["number","string","number"],[e,i,o])})}function TP(t,e,A,i){for(let[n,o]of Object.entries(i))m8(t,e,o,r=>{t.ccall("viz_set_attribute","number",["number","string","number"],[A,n,r])})}function m8(t,e,A,i){let n;if(typeof A=="object"&&"html"in A?n=t.ccall("viz_string_dup_html","number",["number","string"],[e,String(A.html)]):n=t.ccall("viz_string_dup","number",["number","string"],[e,String(A)]),n==0)throw new Error("couldn't dup string");i(n),typeof A=="object"&&"html"in A?t.ccall("viz_string_free_html","number",["number","number"],[e,n]):t.ccall("viz_string_free","number",["number","number"],[e,n])}var Pk=class{constructor(e){this.module=e}get graphvizVersion(){return S3A(this.module)}get formats(){return YP(this.module,"device")}get engines(){return YP(this.module,"layout")}renderFormats(e,A,i={}){return JP(this.module,e,A,rA({engine:"dot"},i))}render(e,A={}){let i;A.format===void 0?i="dot":i=A.format;let n=JP(this.module,e,[i],rA({engine:"dot"},A));return n.status==="success"&&(n.output=n.output[i]),n}renderString(e,A={}){let i=this.render(e,A);if(i.status!=="success")throw new Error(i.errors.find(n=>n.level=="error")?.message||"render failed");return i.output}renderSVGElement(e,A={}){let i=this.renderString(e,Ye(rA({},A),{format:"svg"}));return new DOMParser().parseFromString(i,"image/svg+xml").documentElement}renderJSON(e,A={}){let i=this.renderString(e,Ye(rA({},A),{format:"json"}));return JSON.parse(i)}};function Yu(){return b3A().then(t=>new Pk(t))}var gU=jQ(rq());var vs=class{static getBaseUrlWithoutPath(){let e=window.location.href;return new URL(e).origin+"/dev-ui/"}static getApiServerBaseUrl(){return window.runtimeConfig?.backendUrl}static getWSServerUrl(){let e=this.getApiServerBaseUrl();return!e||e==""?window.location.host:e.startsWith("http://")?e.slice(7):e.startsWith("https://")?e.slice(8):e}};var H2=class t{constructor(e,A){this.http=e;this.zone=A}apiServerDomain=vs.getApiServerBaseUrl();_currentApp=new Mi("");currentApp=this._currentApp.asObservable();isLoading=new Mi(!1);getApp(){return this.currentApp}setApp(e){this._currentApp.next(e)}getLoadingState(){return this.isLoading}runSse(e){let A=this.apiServerDomain+"/run_sse";return this.isLoading.next(!0),new ct(i=>{let n=this;fetch(A,{method:"POST",headers:{"Content-Type":"application/json",Accept:"text/event-stream"},body:JSON.stringify(e)}).then(o=>{let r=o.body?.getReader(),s=new TextDecoder("utf-8"),a="",c=()=>{r?.read().then(({done:l,value:I})=>{if(this.isLoading.next(!0),l)return this.isLoading.next(!1),i.complete();let C=s.decode(I,{stream:!0});a+=C;try{a.split(/\r?\n/).filter(B=>B.startsWith("data:")).forEach(B=>{let E=B.replace(/^data:\s*/,"");JSON.parse(E),n.zone.run(()=>i.next(E))}),a=""}catch(d){d instanceof SyntaxError&&c()}c()}).catch(l=>{n.zone.run(()=>i.error(l))})};c()}).catch(o=>{n.zone.run(()=>i.error(o))})})}listApps(){if(this.apiServerDomain!=null){let e=this.apiServerDomain+"/list-apps?relative_path=./";return this.http.get(e)}return new ct}static \u0275fac=function(A){return new(A||t)(we(Ds),we(Qe))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var YmA="import_session",JmA="edit_function_args";var TmA="a2a_card",KB=class t{route=f(ha);constructor(){}isImportSessionEnabled(){return this.route.queryParams.pipe(je(e=>e[YmA]==="true"))}isEditFunctionArgsEnabled(){return this.route.queryParams.pipe(je(e=>e[JmA]==="true"))}isSessionUrlEnabled(){return Me(!0)}isA2ACardEnabled(){return this.route.queryParams.pipe(je(e=>e[TmA]==="true"))}static \u0275fac=function(A){return new(A||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};function HmA(t,e){}var z2=class{viewContainerRef;injector;id;role="dialog";panelClass="";hasBackdrop=!0;backdropClass="";disableClose=!1;width="";height="";minWidth;minHeight;maxWidth;maxHeight;positionStrategy;data=null;direction;ariaDescribedBy=null;ariaLabelledBy=null;ariaLabel=null;ariaModal=!1;autoFocus="first-tabbable";restoreFocus=!0;scrollStrategy;closeOnNavigation=!0;closeOnDestroy=!0;closeOnOverlayDetachments=!0;componentFactoryResolver;providers;container;templateContext};var rS=(()=>{class t extends J2{_elementRef=f(ee);_focusTrapFactory=f(n8);_config;_interactivityChecker=f(Mu);_ngZone=f(Qe);_overlayRef=f(LB);_focusMonitor=f(dr);_renderer=f(Wi);_platform=f(Ii);_document=f(at,{optional:!0});_portalOutlet;_focusTrap=null;_elementFocusedBeforeDialogWasOpened=null;_closeInteractionType=null;_ariaLabelledByQueue=[];_changeDetectorRef=f(It);_injector=f(Rt);_isDestroyed=!1;constructor(){super(),this._config=f(z2,{optional:!0})||new z2,this._config.ariaLabelledBy&&this._ariaLabelledByQueue.push(this._config.ariaLabelledBy)}_addAriaLabelledBy(A){this._ariaLabelledByQueue.push(A),this._changeDetectorRef.markForCheck()}_removeAriaLabelledBy(A){let i=this._ariaLabelledByQueue.indexOf(A);i>-1&&(this._ariaLabelledByQueue.splice(i,1),this._changeDetectorRef.markForCheck())}_contentAttached(){this._initializeFocusTrap(),this._handleBackdropClicks(),this._captureInitialFocus()}_captureInitialFocus(){this._trapFocus()}ngOnDestroy(){this._isDestroyed=!0,this._restoreFocus()}attachComponentPortal(A){this._portalOutlet.hasAttached();let i=this._portalOutlet.attachComponentPortal(A);return this._contentAttached(),i}attachTemplatePortal(A){this._portalOutlet.hasAttached();let i=this._portalOutlet.attachTemplatePortal(A);return this._contentAttached(),i}attachDomPortal=A=>{this._portalOutlet.hasAttached();let i=this._portalOutlet.attachDomPortal(A);return this._contentAttached(),i};_recaptureFocus(){this._containsFocus()||this._trapFocus()}_forceFocus(A,i){this._interactivityChecker.isFocusable(A)||(A.tabIndex=-1,this._ngZone.runOutsideAngular(()=>{let n=()=>{o(),r(),A.removeAttribute("tabindex")},o=this._renderer.listen(A,"blur",n),r=this._renderer.listen(A,"mousedown",n)})),A.focus(i)}_focusByCssSelector(A,i){let n=this._elementRef.nativeElement.querySelector(A);n&&this._forceFocus(n,i)}_trapFocus(){this._isDestroyed||To(()=>{let A=this._elementRef.nativeElement;switch(this._config.autoFocus){case!1:case"dialog":this._containsFocus()||A.focus();break;case!0:case"first-tabbable":this._focusTrap?.focusInitialElement()||this._focusDialogContainer();break;case"first-heading":this._focusByCssSelector('h1, h2, h3, h4, h5, h6, [role="heading"]');break;default:this._focusByCssSelector(this._config.autoFocus);break}},{injector:this._injector})}_restoreFocus(){let A=this._config.restoreFocus,i=null;if(typeof A=="string"?i=this._document.querySelector(A):typeof A=="boolean"?i=A?this._elementFocusedBeforeDialogWasOpened:null:A&&(i=A),this._config.restoreFocus&&i&&typeof i.focus=="function"){let n=pB(),o=this._elementRef.nativeElement;(!n||n===this._document.body||n===o||o.contains(n))&&(this._focusMonitor?(this._focusMonitor.focusVia(i,this._closeInteractionType),this._closeInteractionType=null):i.focus())}this._focusTrap&&this._focusTrap.destroy()}_focusDialogContainer(){this._elementRef.nativeElement.focus&&this._elementRef.nativeElement.focus()}_containsFocus(){let A=this._elementRef.nativeElement,i=pB();return A===i||A.contains(i)}_initializeFocusTrap(){this._platform.isBrowser&&(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement),this._document&&(this._elementFocusedBeforeDialogWasOpened=pB()))}_handleBackdropClicks(){this._overlayRef.backdropClick().subscribe(()=>{this._config.disableClose&&this._recaptureFocus()})}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["cdk-dialog-container"]],viewQuery:function(i,n){if(i&1&&Te(fa,7),i&2){let o;XA(o=$A())&&(n._portalOutlet=o.first)}},hostAttrs:["tabindex","-1",1,"cdk-dialog-container"],hostVars:6,hostBindings:function(i,n){i&2&&Ne("id",n._config.id||null)("role",n._config.role)("aria-modal",n._config.ariaModal)("aria-labelledby",n._config.ariaLabel?null:n._ariaLabelledByQueue[0])("aria-label",n._config.ariaLabel)("aria-describedby",n._config.ariaDescribedBy||null)},features:[lt],decls:1,vars:0,consts:[["cdkPortalOutlet",""]],template:function(i,n){i&1&&_A(0,HmA,0,0,"ng-template",0)},dependencies:[fa],styles:[".cdk-dialog-container{display:block;width:100%;height:100%;min-height:inherit;max-height:inherit}"],encapsulation:2})}return t})(),Tu=class{overlayRef;config;componentInstance;componentRef;containerInstance;disableClose;closed=new OA;backdropClick;keydownEvents;outsidePointerEvents;id;_detachSubscription;constructor(e,A){this.overlayRef=e,this.config=A,this.disableClose=A.disableClose,this.backdropClick=e.backdropClick(),this.keydownEvents=e.keydownEvents(),this.outsidePointerEvents=e.outsidePointerEvents(),this.id=A.id,this.keydownEvents.subscribe(i=>{i.keyCode===27&&!this.disableClose&&!ir(i)&&(i.preventDefault(),this.close(void 0,{focusOrigin:"keyboard"}))}),this.backdropClick.subscribe(()=>{this.disableClose||this.close(void 0,{focusOrigin:"mouse"})}),this._detachSubscription=e.detachments().subscribe(()=>{A.closeOnOverlayDetachments!==!1&&this.close()})}close(e,A){if(this.containerInstance){let i=this.closed;this.containerInstance._closeInteractionType=A?.focusOrigin||"program",this._detachSubscription.unsubscribe(),this.overlayRef.dispose(),i.next(e),i.complete(),this.componentInstance=this.containerInstance=null}}updatePosition(){return this.overlayRef.updatePosition(),this}updateSize(e="",A=""){return this.overlayRef.updateSize({width:e,height:A}),this}addPanelClass(e){return this.overlayRef.addPanelClass(e),this}removePanelClass(e){return this.overlayRef.removePanelClass(e),this}},zmA=new dA("DialogScrollStrategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.block()}}),OmA=new dA("DialogData"),PmA=new dA("DefaultDialogConfig");var sS=(()=>{class t{_overlay=f(nr);_injector=f(Rt);_defaultOptions=f(PmA,{optional:!0});_parentDialog=f(t,{optional:!0,skipSelf:!0});_overlayContainer=f(d8);_idGenerator=f(sn);_openDialogsAtThisLevel=[];_afterAllClosedAtThisLevel=new OA;_afterOpenedAtThisLevel=new OA;_ariaHiddenElements=new Map;_scrollStrategy=f(zmA);get openDialogs(){return this._parentDialog?this._parentDialog.openDialogs:this._openDialogsAtThisLevel}get afterOpened(){return this._parentDialog?this._parentDialog.afterOpened:this._afterOpenedAtThisLevel}afterAllClosed=jl(()=>this.openDialogs.length?this._getAfterAllClosed():this._getAfterAllClosed().pipe(Pn(void 0)));constructor(){}open(A,i){let n=this._defaultOptions||new z2;i=rA(rA({},n),i),i.id=i.id||this._idGenerator.getId("cdk-dialog-"),i.id&&this.getDialogById(i.id);let o=this._getOverlayConfig(i),r=this._overlay.create(o),s=new Tu(r,i),a=this._attachContainer(r,s,i);return s.containerInstance=a,this._attachDialogContent(A,s,a,i),this.openDialogs.length||this._hideNonDialogContentFromAssistiveTechnology(),this.openDialogs.push(s),s.closed.subscribe(()=>this._removeOpenDialog(s,!0)),this.afterOpened.next(s),s}closeAll(){oS(this.openDialogs,A=>A.close())}getDialogById(A){return this.openDialogs.find(i=>i.id===A)}ngOnDestroy(){oS(this._openDialogsAtThisLevel,A=>{A.config.closeOnDestroy===!1&&this._removeOpenDialog(A,!1)}),oS(this._openDialogsAtThisLevel,A=>A.close()),this._afterAllClosedAtThisLevel.complete(),this._afterOpenedAtThisLevel.complete(),this._openDialogsAtThisLevel=[]}_getOverlayConfig(A){let i=new Bg({positionStrategy:A.positionStrategy||this._overlay.position().global().centerHorizontally().centerVertically(),scrollStrategy:A.scrollStrategy||this._scrollStrategy(),panelClass:A.panelClass,hasBackdrop:A.hasBackdrop,direction:A.direction,minWidth:A.minWidth,minHeight:A.minHeight,maxWidth:A.maxWidth,maxHeight:A.maxHeight,width:A.width,height:A.height,disposeOnNavigation:A.closeOnNavigation});return A.backdropClass&&(i.backdropClass=A.backdropClass),i}_attachContainer(A,i,n){let o=n.injector||n.viewContainerRef?.injector,r=[{provide:z2,useValue:n},{provide:Tu,useValue:i},{provide:LB,useValue:A}],s;n.container?typeof n.container=="function"?s=n.container:(s=n.container.type,r.push(...n.container.providers(n))):s=rS;let a=new dl(s,n.viewContainerRef,Rt.create({parent:o||this._injector,providers:r}));return A.attach(a).instance}_attachDialogContent(A,i,n,o){if(A instanceof wn){let r=this._createInjector(o,i,n,void 0),s={$implicit:o.data,dialogRef:i};o.templateContext&&(s=rA(rA({},s),typeof o.templateContext=="function"?o.templateContext():o.templateContext)),n.attachTemplatePortal(new ys(A,null,s,r))}else{let r=this._createInjector(o,i,n,this._injector),s=n.attachComponentPortal(new dl(A,o.viewContainerRef,r));i.componentRef=s,i.componentInstance=s.instance}}_createInjector(A,i,n,o){let r=A.injector||A.viewContainerRef?.injector,s=[{provide:OmA,useValue:A.data},{provide:Tu,useValue:i}];return A.providers&&(typeof A.providers=="function"?s.push(...A.providers(i,A,n)):s.push(...A.providers)),A.direction&&(!r||!r.get(mo,null,{optional:!0}))&&s.push({provide:mo,useValue:{value:A.direction,change:Me()}}),Rt.create({parent:r||o,providers:s})}_removeOpenDialog(A,i){let n=this.openDialogs.indexOf(A);n>-1&&(this.openDialogs.splice(n,1),this.openDialogs.length||(this._ariaHiddenElements.forEach((o,r)=>{o?r.setAttribute("aria-hidden",o):r.removeAttribute("aria-hidden")}),this._ariaHiddenElements.clear(),i&&this._getAfterAllClosed().next()))}_hideNonDialogContentFromAssistiveTechnology(){let A=this._overlayContainer.getContainerElement();if(A.parentElement){let i=A.parentElement.children;for(let n=i.length-1;n>-1;n--){let o=i[n];o!==A&&o.nodeName!=="SCRIPT"&&o.nodeName!=="STYLE"&&!o.hasAttribute("aria-live")&&(this._ariaHiddenElements.set(o,o.getAttribute("aria-hidden")),o.setAttribute("aria-hidden","true"))}}}_getAfterAllClosed(){let A=this._parentDialog;return A?A._getAfterAllClosed():this._afterAllClosedAtThisLevel}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function oS(t,e){let A=t.length;for(;A--;)e(t[A])}var sq=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[sS],imports:[El,dg,r8,dg]})}return t})();function jmA(t,e){}var D8=class{viewContainerRef;injector;id;role="dialog";panelClass="";hasBackdrop=!0;backdropClass="";disableClose=!1;width="";height="";minWidth;minHeight;maxWidth;maxHeight;position;data=null;direction;ariaDescribedBy=null;ariaLabelledBy=null;ariaLabel=null;ariaModal=!1;autoFocus="first-tabbable";restoreFocus=!0;delayFocusTrap=!0;scrollStrategy;closeOnNavigation=!0;componentFactoryResolver;enterAnimationDuration;exitAnimationDuration},aS="mdc-dialog--open",aq="mdc-dialog--opening",cq="mdc-dialog--closing",qmA=150,VmA=75,ZmA=(()=>{class t extends rS{_animationMode=f(Si,{optional:!0});_animationStateChanged=new WA;_animationsEnabled=this._animationMode!=="NoopAnimations";_actionSectionCount=0;_hostElement=this._elementRef.nativeElement;_enterAnimationDuration=this._animationsEnabled?gq(this._config.enterAnimationDuration)??qmA:0;_exitAnimationDuration=this._animationsEnabled?gq(this._config.exitAnimationDuration)??VmA:0;_animationTimer=null;_contentAttached(){super._contentAttached(),this._startOpenAnimation()}_startOpenAnimation(){this._animationStateChanged.emit({state:"opening",totalTime:this._enterAnimationDuration}),this._animationsEnabled?(this._hostElement.style.setProperty(lq,`${this._enterAnimationDuration}ms`),this._requestAnimationFrame(()=>this._hostElement.classList.add(aq,aS)),this._waitForAnimationToComplete(this._enterAnimationDuration,this._finishDialogOpen)):(this._hostElement.classList.add(aS),Promise.resolve().then(()=>this._finishDialogOpen()))}_startExitAnimation(){this._animationStateChanged.emit({state:"closing",totalTime:this._exitAnimationDuration}),this._hostElement.classList.remove(aS),this._animationsEnabled?(this._hostElement.style.setProperty(lq,`${this._exitAnimationDuration}ms`),this._requestAnimationFrame(()=>this._hostElement.classList.add(cq)),this._waitForAnimationToComplete(this._exitAnimationDuration,this._finishDialogClose)):Promise.resolve().then(()=>this._finishDialogClose())}_updateActionSectionCount(A){this._actionSectionCount+=A,this._changeDetectorRef.markForCheck()}_finishDialogOpen=()=>{this._clearAnimationClasses(),this._openAnimationDone(this._enterAnimationDuration)};_finishDialogClose=()=>{this._clearAnimationClasses(),this._animationStateChanged.emit({state:"closed",totalTime:this._exitAnimationDuration})};_clearAnimationClasses(){this._hostElement.classList.remove(aq,cq)}_waitForAnimationToComplete(A,i){this._animationTimer!==null&&clearTimeout(this._animationTimer),this._animationTimer=setTimeout(i,A)}_requestAnimationFrame(A){this._ngZone.runOutsideAngular(()=>{typeof requestAnimationFrame=="function"?requestAnimationFrame(A):A()})}_captureInitialFocus(){this._config.delayFocusTrap||this._trapFocus()}_openAnimationDone(A){this._config.delayFocusTrap&&this._trapFocus(),this._animationStateChanged.next({state:"opened",totalTime:A})}ngOnDestroy(){super.ngOnDestroy(),this._animationTimer!==null&&clearTimeout(this._animationTimer)}attachComponentPortal(A){let i=super.attachComponentPortal(A);return i.location.nativeElement.classList.add("mat-mdc-dialog-component-host"),i}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-dialog-container"]],hostAttrs:["tabindex","-1",1,"mat-mdc-dialog-container","mdc-dialog"],hostVars:10,hostBindings:function(i,n){i&2&&(Hs("id",n._config.id),Ne("aria-modal",n._config.ariaModal)("role",n._config.role)("aria-labelledby",n._config.ariaLabel?null:n._ariaLabelledByQueue[0])("aria-label",n._config.ariaLabel)("aria-describedby",n._config.ariaDescribedBy||null),ue("_mat-animation-noopable",!n._animationsEnabled)("mat-mdc-dialog-container-with-actions",n._actionSectionCount>0))},features:[lt],decls:3,vars:0,consts:[[1,"mat-mdc-dialog-inner-container","mdc-dialog__container"],[1,"mat-mdc-dialog-surface","mdc-dialog__surface"],["cdkPortalOutlet",""]],template:function(i,n){i&1&&(S(0,"div",0)(1,"div",1),_A(2,jmA,0,0,"ng-template",2),F()())},dependencies:[fa],styles:['.mat-mdc-dialog-container{width:100%;height:100%;display:block;box-sizing:border-box;max-height:inherit;min-height:inherit;min-width:inherit;max-width:inherit;outline:0}.cdk-overlay-pane.mat-mdc-dialog-panel{max-width:var(--mat-dialog-container-max-width, 560px);min-width:var(--mat-dialog-container-min-width, 280px)}@media(max-width: 599px){.cdk-overlay-pane.mat-mdc-dialog-panel{max-width:var(--mat-dialog-container-small-max-width, calc(100vw - 32px))}}.mat-mdc-dialog-inner-container{display:flex;flex-direction:row;align-items:center;justify-content:space-around;box-sizing:border-box;height:100%;opacity:0;transition:opacity linear var(--mat-dialog-transition-duration, 0ms);max-height:inherit;min-height:inherit;min-width:inherit;max-width:inherit}.mdc-dialog--closing .mat-mdc-dialog-inner-container{transition:opacity 75ms linear;transform:none}.mdc-dialog--open .mat-mdc-dialog-inner-container{opacity:1}._mat-animation-noopable .mat-mdc-dialog-inner-container{transition:none}.mat-mdc-dialog-surface{display:flex;flex-direction:column;flex-grow:0;flex-shrink:0;box-sizing:border-box;width:100%;height:100%;position:relative;overflow-y:auto;outline:0;transform:scale(0.8);transition:transform var(--mat-dialog-transition-duration, 0ms) cubic-bezier(0, 0, 0.2, 1);max-height:inherit;min-height:inherit;min-width:inherit;max-width:inherit;box-shadow:var(--mat-dialog-container-elevation-shadow, none);border-radius:var(--mdc-dialog-container-shape, var(--mat-sys-corner-extra-large, 4px));background-color:var(--mdc-dialog-container-color, var(--mat-sys-surface, white))}[dir=rtl] .mat-mdc-dialog-surface{text-align:right}.mdc-dialog--open .mat-mdc-dialog-surface,.mdc-dialog--closing .mat-mdc-dialog-surface{transform:none}._mat-animation-noopable .mat-mdc-dialog-surface{transition:none}.mat-mdc-dialog-surface::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:2px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}.mat-mdc-dialog-title{display:block;position:relative;flex-shrink:0;box-sizing:border-box;margin:0 0 1px;padding:var(--mat-dialog-headline-padding, 6px 24px 13px)}.mat-mdc-dialog-title::before{display:inline-block;width:0;height:40px;content:"";vertical-align:0}[dir=rtl] .mat-mdc-dialog-title{text-align:right}.mat-mdc-dialog-container .mat-mdc-dialog-title{color:var(--mdc-dialog-subhead-color, var(--mat-sys-on-surface, rgba(0, 0, 0, 0.87)));font-family:var(--mdc-dialog-subhead-font, var(--mat-sys-headline-small-font, inherit));line-height:var(--mdc-dialog-subhead-line-height, var(--mat-sys-headline-small-line-height, 1.5rem));font-size:var(--mdc-dialog-subhead-size, var(--mat-sys-headline-small-size, 1rem));font-weight:var(--mdc-dialog-subhead-weight, var(--mat-sys-headline-small-weight, 400));letter-spacing:var(--mdc-dialog-subhead-tracking, var(--mat-sys-headline-small-tracking, 0.03125em))}.mat-mdc-dialog-content{display:block;flex-grow:1;box-sizing:border-box;margin:0;overflow:auto;max-height:65vh}.mat-mdc-dialog-content>:first-child{margin-top:0}.mat-mdc-dialog-content>:last-child{margin-bottom:0}.mat-mdc-dialog-container .mat-mdc-dialog-content{color:var(--mdc-dialog-supporting-text-color, var(--mat-sys-on-surface-variant, rgba(0, 0, 0, 0.6)));font-family:var(--mdc-dialog-supporting-text-font, var(--mat-sys-body-medium-font, inherit));line-height:var(--mdc-dialog-supporting-text-line-height, var(--mat-sys-body-medium-line-height, 1.5rem));font-size:var(--mdc-dialog-supporting-text-size, var(--mat-sys-body-medium-size, 1rem));font-weight:var(--mdc-dialog-supporting-text-weight, var(--mat-sys-body-medium-weight, 400));letter-spacing:var(--mdc-dialog-supporting-text-tracking, var(--mat-sys-body-medium-tracking, 0.03125em))}.mat-mdc-dialog-container .mat-mdc-dialog-content{padding:var(--mat-dialog-content-padding, 20px 24px)}.mat-mdc-dialog-container-with-actions .mat-mdc-dialog-content{padding:var(--mat-dialog-with-actions-content-padding, 20px 24px 0)}.mat-mdc-dialog-container .mat-mdc-dialog-title+.mat-mdc-dialog-content{padding-top:0}.mat-mdc-dialog-actions{display:flex;position:relative;flex-shrink:0;flex-wrap:wrap;align-items:center;justify-content:flex-end;box-sizing:border-box;min-height:52px;margin:0;padding:8px;border-top:1px solid rgba(0,0,0,0);padding:var(--mat-dialog-actions-padding, 16px 24px);justify-content:var(--mat-dialog-actions-alignment, flex-end)}@media(forced-colors: active){.mat-mdc-dialog-actions{border-top-color:CanvasText}}.mat-mdc-dialog-actions.mat-mdc-dialog-actions-align-start,.mat-mdc-dialog-actions[align=start]{justify-content:start}.mat-mdc-dialog-actions.mat-mdc-dialog-actions-align-center,.mat-mdc-dialog-actions[align=center]{justify-content:center}.mat-mdc-dialog-actions.mat-mdc-dialog-actions-align-end,.mat-mdc-dialog-actions[align=end]{justify-content:flex-end}.mat-mdc-dialog-actions .mat-button-base+.mat-button-base,.mat-mdc-dialog-actions .mat-mdc-button-base+.mat-mdc-button-base{margin-left:8px}[dir=rtl] .mat-mdc-dialog-actions .mat-button-base+.mat-button-base,[dir=rtl] .mat-mdc-dialog-actions .mat-mdc-button-base+.mat-mdc-button-base{margin-left:0;margin-right:8px}.mat-mdc-dialog-component-host{display:contents}'],encapsulation:2})}return t})(),lq="--mat-dialog-transition-duration";function gq(t){return t==null?null:typeof t=="number"?t:t.endsWith("ms")?zs(t.substring(0,t.length-2)):t.endsWith("s")?zs(t.substring(0,t.length-1))*1e3:t==="0"?0:null}var w8=function(t){return t[t.OPEN=0]="OPEN",t[t.CLOSING=1]="CLOSING",t[t.CLOSED=2]="CLOSED",t}(w8||{}),Br=class{_ref;_containerInstance;componentInstance;componentRef;disableClose;id;_afterOpened=new OA;_beforeClosed=new OA;_result;_closeFallbackTimeout;_state=w8.OPEN;_closeInteractionType;constructor(e,A,i){this._ref=e,this._containerInstance=i,this.disableClose=A.disableClose,this.id=e.id,e.addPanelClass("mat-mdc-dialog-panel"),i._animationStateChanged.pipe(kt(n=>n.state==="opened"),On(1)).subscribe(()=>{this._afterOpened.next(),this._afterOpened.complete()}),i._animationStateChanged.pipe(kt(n=>n.state==="closed"),On(1)).subscribe(()=>{clearTimeout(this._closeFallbackTimeout),this._finishDialogClose()}),e.overlayRef.detachments().subscribe(()=>{this._beforeClosed.next(this._result),this._beforeClosed.complete(),this._finishDialogClose()}),zn(this.backdropClick(),this.keydownEvents().pipe(kt(n=>n.keyCode===27&&!this.disableClose&&!ir(n)))).subscribe(n=>{this.disableClose||(n.preventDefault(),Iq(this,n.type==="keydown"?"keyboard":"mouse"))})}close(e){this._result=e,this._containerInstance._animationStateChanged.pipe(kt(A=>A.state==="closing"),On(1)).subscribe(A=>{this._beforeClosed.next(e),this._beforeClosed.complete(),this._ref.overlayRef.detachBackdrop(),this._closeFallbackTimeout=setTimeout(()=>this._finishDialogClose(),A.totalTime+100)}),this._state=w8.CLOSING,this._containerInstance._startExitAnimation()}afterOpened(){return this._afterOpened}afterClosed(){return this._ref.closed}beforeClosed(){return this._beforeClosed}backdropClick(){return this._ref.backdropClick}keydownEvents(){return this._ref.keydownEvents}updatePosition(e){let A=this._ref.config.positionStrategy;return e&&(e.left||e.right)?e.left?A.left(e.left):A.right(e.right):A.centerHorizontally(),e&&(e.top||e.bottom)?e.top?A.top(e.top):A.bottom(e.bottom):A.centerVertically(),this._ref.updatePosition(),this}updateSize(e="",A=""){return this._ref.updateSize(e,A),this}addPanelClass(e){return this._ref.addPanelClass(e),this}removePanelClass(e){return this._ref.removePanelClass(e),this}getState(){return this._state}_finishDialogClose(){this._state=w8.CLOSED,this._ref.close(this._result,{focusOrigin:this._closeInteractionType}),this.componentInstance=null}};function Iq(t,e,A){return t._closeInteractionType=e,t.close(A)}var as=new dA("MatMdcDialogData"),WmA=new dA("mat-mdc-dialog-default-options"),XmA=new dA("mat-mdc-dialog-scroll-strategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.block()}});var qs=(()=>{class t{_overlay=f(nr);_defaultOptions=f(WmA,{optional:!0});_scrollStrategy=f(XmA);_parentDialog=f(t,{optional:!0,skipSelf:!0});_idGenerator=f(sn);_dialog=f(sS);_openDialogsAtThisLevel=[];_afterAllClosedAtThisLevel=new OA;_afterOpenedAtThisLevel=new OA;dialogConfigClass=D8;_dialogRefConstructor;_dialogContainerType;_dialogDataToken;get openDialogs(){return this._parentDialog?this._parentDialog.openDialogs:this._openDialogsAtThisLevel}get afterOpened(){return this._parentDialog?this._parentDialog.afterOpened:this._afterOpenedAtThisLevel}_getAfterAllClosed(){let A=this._parentDialog;return A?A._getAfterAllClosed():this._afterAllClosedAtThisLevel}afterAllClosed=jl(()=>this.openDialogs.length?this._getAfterAllClosed():this._getAfterAllClosed().pipe(Pn(void 0)));constructor(){this._dialogRefConstructor=Br,this._dialogContainerType=ZmA,this._dialogDataToken=as}open(A,i){let n;i=rA(rA({},this._defaultOptions||new D8),i),i.id=i.id||this._idGenerator.getId("mat-mdc-dialog-"),i.scrollStrategy=i.scrollStrategy||this._scrollStrategy();let o=this._dialog.open(A,Ye(rA({},i),{positionStrategy:this._overlay.position().global().centerHorizontally().centerVertically(),disableClose:!0,closeOnDestroy:!1,closeOnOverlayDetachments:!1,container:{type:this._dialogContainerType,providers:()=>[{provide:this.dialogConfigClass,useValue:i},{provide:z2,useValue:i}]},templateContext:()=>({dialogRef:n}),providers:(r,s,a)=>(n=new this._dialogRefConstructor(r,i,a),n.updatePosition(i?.position),[{provide:this._dialogContainerType,useValue:a},{provide:this._dialogDataToken,useValue:s.data},{provide:this._dialogRefConstructor,useValue:n}])}));return n.componentRef=o.componentRef,n.componentInstance=o.componentInstance,this.openDialogs.push(n),this.afterOpened.next(n),n.afterClosed().subscribe(()=>{let r=this.openDialogs.indexOf(n);r>-1&&(this.openDialogs.splice(r,1),this.openDialogs.length||this._getAfterAllClosed().next())}),n}closeAll(){this._closeDialogs(this.openDialogs)}getDialogById(A){return this.openDialogs.find(i=>i.id===A)}ngOnDestroy(){this._closeDialogs(this._openDialogsAtThisLevel),this._afterAllClosedAtThisLevel.complete(),this._afterOpenedAtThisLevel.complete()}_closeDialogs(A){let i=A.length;for(;i--;)A[i].close()}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})(),hg=(()=>{class t{dialogRef=f(Br,{optional:!0});_elementRef=f(ee);_dialog=f(qs);ariaLabel;type="button";dialogResult;_matDialogClose;constructor(){}ngOnInit(){this.dialogRef||(this.dialogRef=dq(this._elementRef,this._dialog.openDialogs))}ngOnChanges(A){let i=A._matDialogClose||A._matDialogCloseResult;i&&(this.dialogResult=i.currentValue)}_onButtonClick(A){Iq(this.dialogRef,A.screenX===0&&A.screenY===0?"keyboard":"mouse",this.dialogResult)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","mat-dialog-close",""],["","matDialogClose",""]],hostVars:2,hostBindings:function(i,n){i&1&&hA("click",function(r){return n._onButtonClick(r)}),i&2&&Ne("aria-label",n.ariaLabel||null)("type",n.type)},inputs:{ariaLabel:[0,"aria-label","ariaLabel"],type:"type",dialogResult:[0,"mat-dialog-close","dialogResult"],_matDialogClose:[0,"matDialogClose","_matDialogClose"]},exportAs:["matDialogClose"],features:[ti]})}return t})(),Cq=(()=>{class t{_dialogRef=f(Br,{optional:!0});_elementRef=f(ee);_dialog=f(qs);constructor(){}ngOnInit(){this._dialogRef||(this._dialogRef=dq(this._elementRef,this._dialog.openDialogs)),this._dialogRef&&Promise.resolve().then(()=>{this._onAdd()})}ngOnDestroy(){this._dialogRef?._containerInstance&&Promise.resolve().then(()=>{this._onRemove()})}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t})}return t})(),bs=(()=>{class t extends Cq{id=f(sn).getId("mat-mdc-dialog-title-");_onAdd(){this._dialogRef._containerInstance?._addAriaLabelledBy?.(this.id)}_onRemove(){this._dialogRef?._containerInstance?._removeAriaLabelledBy?.(this.id)}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","mat-dialog-title",""],["","matDialogTitle",""]],hostAttrs:[1,"mat-mdc-dialog-title","mdc-dialog__title"],hostVars:1,hostBindings:function(i,n){i&2&&Hs("id",n.id)},inputs:{id:"id"},exportAs:["matDialogTitle"],features:[lt]})}return t})(),ma=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","mat-dialog-content",""],["mat-dialog-content"],["","matDialogContent",""]],hostAttrs:[1,"mat-mdc-dialog-content","mdc-dialog__content"],features:[YT([D0])]})}return t})(),pa=(()=>{class t extends Cq{align;_onAdd(){this._dialogRef._containerInstance?._updateActionSectionCount?.(1)}_onRemove(){this._dialogRef._containerInstance?._updateActionSectionCount?.(-1)}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","mat-dialog-actions",""],["mat-dialog-actions"],["","matDialogActions",""]],hostAttrs:[1,"mat-mdc-dialog-actions","mdc-dialog__actions"],hostVars:6,hostBindings:function(i,n){i&2&&ue("mat-mdc-dialog-actions-align-start",n.align==="start")("mat-mdc-dialog-actions-align-center",n.align==="center")("mat-mdc-dialog-actions-align-end",n.align==="end")},inputs:{align:"align"},features:[lt]})}return t})();function dq(t,e){let A=t.nativeElement.parentElement;for(;A&&!A.classList.contains("mat-mdc-dialog-container");)A=A.parentElement;return A?e.find(i=>i.id===A.id):null}var Bq=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[qs],imports:[sq,El,dg,it,it]})}return t})();function $mA(t,e){if(t&1&&JA(0,"img",5),t&2){let A=O(2);yA("src",A.displayContent,ja)}}function ApA(t,e){t&1&&(S(0,"div",6),AA(1," No image data provided. "),F())}function epA(t,e){if(t&1&&(S(0,"div",3),_A(1,$mA,1,1,"img",5)(2,ApA,2,0,"div",6),F()),t&2){let A=O();G(),GA(A.displayContent?1:-1),G(),GA(A.displayContent?-1:2)}}function tpA(t,e){if(t&1&&JA(0,"div",4),t&2){let A=O();yA("innerHTML",A.displayContent,RI)}}var v0=class t{constructor(e,A,i){this.dialogRef=e;this.data=A;this.sanitizer=i}displayContent=null;isSvgContent=!1;ngOnInit(){this.processImageData()}processImageData(){let e=this.data.imageData;if(!e){this.displayContent=null,this.isSvgContent=!1;return}if(e.trim().includes("<svg"))this.isSvgContent=!0,this.displayContent=this.sanitizer.bypassSecurityTrustHtml(e);else{let A=e.startsWith("data:image/")?"":"data:image/png;base64,";this.isSvgContent=!1,this.displayContent=this.sanitizer.bypassSecurityTrustUrl(A+e)}}close(){this.dialogRef.close()}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(as),PA(cl))};static \u0275cmp=zA({type:t,selectors:[["app-view-image-dialog"]],standalone:!1,decls:6,vars:2,consts:[[1,"close-button",3,"click"],["xmlns","[http://www.w3.org/2000/svg](http://www.w3.org/2000/svg)","viewBox","0 0 24 24","fill","currentColor","width","24px","height","24px"],["d","M19 6.41L17.59 5 12 10.59 6.41 5 5 6.41 10.59 12 5 17.59 6.41 19 12 13.41 17.59 19 19 17.59 13.41 12z"],[1,"image-wrapper"],[3,"innerHTML"],["alt","Viewed Image",3,"src"],[1,"no-image-placeholder"]],template:function(A,i){A&1&&(S(0,"div")(1,"button",0),hA("click",function(){return i.close()}),ar(),S(2,"svg",1),JA(3,"path",2),F()(),_A(4,epA,3,2,"div",3)(5,tpA,1,1,"div",4),F()),A&2&&(G(4),GA(i.isSvgContent?-1:4),G(),GA(i.isSvgContent?5:-1))},styles:[".close-button[_ngcontent-%COMP%]{position:absolute;top:5px;right:10px;background:none;border:none;cursor:pointer;padding:8px;border-radius:50%;transition:background-color .2s ease;color:#666;display:flex;align-items:center;justify-content:center;z-index:10;margin-bottom:15px}.close-button[_ngcontent-%COMP%]:hover{background-color:#f0f0f0;color:#333}.close-button[_ngcontent-%COMP%]   svg[_ngcontent-%COMP%]{width:24px;height:24px;fill:currentColor}.image-wrapper[_ngcontent-%COMP%]{flex-grow:1;display:flex;justify-content:center;align-items:center;overflow:auto}.image-wrapper[_ngcontent-%COMP%]   img[_ngcontent-%COMP%], .image-wrapper[_ngcontent-%COMP%]   .svg-container[_ngcontent-%COMP%]{max-width:90vw;max-height:90vh;object-fit:contain;border-radius:8px}.no-image-placeholder[_ngcontent-%COMP%]{color:#888;font-style:italic;text-align:center;padding:20px}@media (max-width: 1768px){.close-button[_ngcontent-%COMP%]{top:5px;right:5px;padding:5px}}"]})};var O2=class t{downloadBase64Data(e,A,i="image.png"){try{let n=document.createElement("a");n.href=e,n.download=i,document.body.appendChild(n),n.click(),document.body.removeChild(n)}catch(n){throw console.error("Error downloading base64 data:",n),n}}downloadObjectAsJson(e,A="session.json"){let i=JSON.stringify(e,null,2),n=new Blob([i],{type:"application/json"}),o=URL.createObjectURL(n),r=document.createElement("a");r.href=o,r.download=A,document.body.appendChild(r),r.click(),document.body.removeChild(r),URL.revokeObjectURL(o)}static \u0275fac=function(A){return new(A||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var ipA=["*"],y8;function npA(){if(y8===void 0&&(y8=null,typeof window<"u")){let t=window;t.trustedTypes!==void 0&&(y8=t.trustedTypes.createPolicy("angular#components",{createHTML:e=>e}))}return y8}function Hu(t){return npA()?.createHTML(t)||t}function Qq(t){return Error(`Unable to find icon with the name "${t}"`)}function opA(){return Error("Could not find HttpClient for use with Angular Material icons. Please add provideHttpClient() to your providers.")}function hq(t){return Error(`The URL provided to MatIconRegistry was not trusted as a resource URL via Angular's DomSanitizer. Attempted URL was "${t}".`)}function uq(t){return Error(`The literal provided to MatIconRegistry was not trusted as safe HTML by Angular's DomSanitizer. Attempted literal was "${t}".`)}var b0=class{url;svgText;options;svgElement;constructor(e,A,i){this.url=e,this.svgText=A,this.options=i}},rpA=(()=>{class t{_httpClient;_sanitizer;_errorHandler;_document;_svgIconConfigs=new Map;_iconSetConfigs=new Map;_cachedIconsByUrl=new Map;_inProgressUrlFetches=new Map;_fontCssClassesByAlias=new Map;_resolvers=[];_defaultFontSetClass=["material-icons","mat-ligature-font"];constructor(A,i,n,o){this._httpClient=A,this._sanitizer=i,this._errorHandler=o,this._document=n}addSvgIcon(A,i,n){return this.addSvgIconInNamespace("",A,i,n)}addSvgIconLiteral(A,i,n){return this.addSvgIconLiteralInNamespace("",A,i,n)}addSvgIconInNamespace(A,i,n,o){return this._addSvgIconConfig(A,i,new b0(n,null,o))}addSvgIconResolver(A){return this._resolvers.push(A),this}addSvgIconLiteralInNamespace(A,i,n,o){let r=this._sanitizer.sanitize(jr.HTML,n);if(!r)throw uq(n);let s=Hu(r);return this._addSvgIconConfig(A,i,new b0("",s,o))}addSvgIconSet(A,i){return this.addSvgIconSetInNamespace("",A,i)}addSvgIconSetLiteral(A,i){return this.addSvgIconSetLiteralInNamespace("",A,i)}addSvgIconSetInNamespace(A,i,n){return this._addSvgIconSetConfig(A,new b0(i,null,n))}addSvgIconSetLiteralInNamespace(A,i,n){let o=this._sanitizer.sanitize(jr.HTML,i);if(!o)throw uq(i);let r=Hu(o);return this._addSvgIconSetConfig(A,new b0("",r,n))}registerFontClassAlias(A,i=A){return this._fontCssClassesByAlias.set(A,i),this}classNameForFontAlias(A){return this._fontCssClassesByAlias.get(A)||A}setDefaultFontSetClass(...A){return this._defaultFontSetClass=A,this}getDefaultFontSetClass(){return this._defaultFontSetClass}getSvgIconFromUrl(A){let i=this._sanitizer.sanitize(jr.RESOURCE_URL,A);if(!i)throw hq(A);let n=this._cachedIconsByUrl.get(i);return n?Me(v8(n)):this._loadSvgIconFromConfig(new b0(A,null)).pipe(Qo(o=>this._cachedIconsByUrl.set(i,o)),je(o=>v8(o)))}getNamedSvgIcon(A,i=""){let n=fq(i,A),o=this._svgIconConfigs.get(n);if(o)return this._getSvgFromConfig(o);if(o=this._getIconConfigFromResolvers(i,A),o)return this._svgIconConfigs.set(n,o),this._getSvgFromConfig(o);let r=this._iconSetConfigs.get(i);return r?this._getSvgFromIconSetConfigs(A,r):g2(Qq(n))}ngOnDestroy(){this._resolvers=[],this._svgIconConfigs.clear(),this._iconSetConfigs.clear(),this._cachedIconsByUrl.clear()}_getSvgFromConfig(A){return A.svgText?Me(v8(this._svgElementFromConfig(A))):this._loadSvgIconFromConfig(A).pipe(je(i=>v8(i)))}_getSvgFromIconSetConfigs(A,i){let n=this._extractIconWithNameFromAnySet(A,i);if(n)return Me(n);let o=i.filter(r=>!r.svgText).map(r=>this._loadSvgIconSetFromConfig(r).pipe(mr(s=>{let c=`Loading icon set URL: ${this._sanitizer.sanitize(jr.RESOURCE_URL,r.url)} failed: ${s.message}`;return this._errorHandler.handleError(new Error(c)),Me(null)})));return nh(o).pipe(je(()=>{let r=this._extractIconWithNameFromAnySet(A,i);if(!r)throw Qq(A);return r}))}_extractIconWithNameFromAnySet(A,i){for(let n=i.length-1;n>=0;n--){let o=i[n];if(o.svgText&&o.svgText.toString().indexOf(A)>-1){let r=this._svgElementFromConfig(o),s=this._extractSvgIconFromSet(r,A,o.options);if(s)return s}}return null}_loadSvgIconFromConfig(A){return this._fetchIcon(A).pipe(Qo(i=>A.svgText=i),je(()=>this._svgElementFromConfig(A)))}_loadSvgIconSetFromConfig(A){return A.svgText?Me(null):this._fetchIcon(A).pipe(Qo(i=>A.svgText=i))}_extractSvgIconFromSet(A,i,n){let o=A.querySelector(`[id="${i}"]`);if(!o)return null;let r=o.cloneNode(!0);if(r.removeAttribute("id"),r.nodeName.toLowerCase()==="svg")return this._setSvgAttributes(r,n);if(r.nodeName.toLowerCase()==="symbol")return this._setSvgAttributes(this._toSvgElement(r),n);let s=this._svgElementFromString(Hu("<svg></svg>"));return s.appendChild(r),this._setSvgAttributes(s,n)}_svgElementFromString(A){let i=this._document.createElement("DIV");i.innerHTML=A;let n=i.querySelector("svg");if(!n)throw Error("<svg> tag not found");return n}_toSvgElement(A){let i=this._svgElementFromString(Hu("<svg></svg>")),n=A.attributes;for(let o=0;o<n.length;o++){let{name:r,value:s}=n[o];r!=="id"&&i.setAttribute(r,s)}for(let o=0;o<A.childNodes.length;o++)A.childNodes[o].nodeType===this._document.ELEMENT_NODE&&i.appendChild(A.childNodes[o].cloneNode(!0));return i}_setSvgAttributes(A,i){return A.setAttribute("fit",""),A.setAttribute("height","100%"),A.setAttribute("width","100%"),A.setAttribute("preserveAspectRatio","xMidYMid meet"),A.setAttribute("focusable","false"),i&&i.viewBox&&A.setAttribute("viewBox",i.viewBox),A}_fetchIcon(A){let{url:i,options:n}=A,o=n?.withCredentials??!1;if(!this._httpClient)throw opA();if(i==null)throw Error(`Cannot fetch icon from URL "${i}".`);let r=this._sanitizer.sanitize(jr.RESOURCE_URL,i);if(!r)throw hq(i);let s=this._inProgressUrlFetches.get(r);if(s)return s;let a=this._httpClient.get(r,{responseType:"text",withCredentials:o}).pipe(je(c=>Hu(c)),Vl(()=>this._inProgressUrlFetches.delete(r)),rh());return this._inProgressUrlFetches.set(r,a),a}_addSvgIconConfig(A,i,n){return this._svgIconConfigs.set(fq(A,i),n),this}_addSvgIconSetConfig(A,i){let n=this._iconSetConfigs.get(A);return n?n.push(i):this._iconSetConfigs.set(A,[i]),this}_svgElementFromConfig(A){if(!A.svgElement){let i=this._svgElementFromString(A.svgText);this._setSvgAttributes(i,A.options),A.svgElement=i}return A.svgElement}_getIconConfigFromResolvers(A,i){for(let n=0;n<this._resolvers.length;n++){let o=this._resolvers[n](i,A);if(o)return spA(o)?new b0(o.url,null,o.options):new b0(o,null)}}static \u0275fac=function(i){return new(i||t)(we(Ds,8),we(cl),we(at,8),we(aa))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();function v8(t){return t.cloneNode(!0)}function fq(t,e){return t+":"+e}function spA(t){return!!(t.url&&t.options)}var apA=new dA("MAT_ICON_DEFAULT_OPTIONS"),cpA=new dA("mat-icon-location",{providedIn:"root",factory:lpA});function lpA(){let t=f(at),e=t?t.location:null;return{getPathname:()=>e?e.pathname+e.search:""}}var mq=["clip-path","color-profile","src","cursor","fill","filter","marker","marker-start","marker-mid","marker-end","mask","stroke"],gpA=mq.map(t=>`[${t}]`).join(", "),IpA=/^url\(['"]?#(.*?)['"]?\)$/,P2=(()=>{class t{_elementRef=f(ee);_iconRegistry=f(rpA);_location=f(cpA);_errorHandler=f(aa);_defaultColor;get color(){return this._color||this._defaultColor}set color(A){this._color=A}_color;inline=!1;get svgIcon(){return this._svgIcon}set svgIcon(A){A!==this._svgIcon&&(A?this._updateSvgIcon(A):this._svgIcon&&this._clearSvgElement(),this._svgIcon=A)}_svgIcon;get fontSet(){return this._fontSet}set fontSet(A){let i=this._cleanupFontValue(A);i!==this._fontSet&&(this._fontSet=i,this._updateFontIconClasses())}_fontSet;get fontIcon(){return this._fontIcon}set fontIcon(A){let i=this._cleanupFontValue(A);i!==this._fontIcon&&(this._fontIcon=i,this._updateFontIconClasses())}_fontIcon;_previousFontSetClass=[];_previousFontIconClass;_svgName;_svgNamespace;_previousPath;_elementsWithExternalReferences;_currentIconFetch=Kt.EMPTY;constructor(){let A=f(new wr("aria-hidden"),{optional:!0}),i=f(apA,{optional:!0});i&&(i.color&&(this.color=this._defaultColor=i.color),i.fontSet&&(this.fontSet=i.fontSet)),A||this._elementRef.nativeElement.setAttribute("aria-hidden","true")}_splitIconName(A){if(!A)return["",""];let i=A.split(":");switch(i.length){case 1:return["",i[0]];case 2:return i;default:throw Error(`Invalid icon name: "${A}"`)}}ngOnInit(){this._updateFontIconClasses()}ngAfterViewChecked(){let A=this._elementsWithExternalReferences;if(A&&A.size){let i=this._location.getPathname();i!==this._previousPath&&(this._previousPath=i,this._prependPathToReferences(i))}}ngOnDestroy(){this._currentIconFetch.unsubscribe(),this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear()}_usingFontIcon(){return!this.svgIcon}_setSvgElement(A){this._clearSvgElement();let i=this._location.getPathname();this._previousPath=i,this._cacheChildrenWithExternalReferences(A),this._prependPathToReferences(i),this._elementRef.nativeElement.appendChild(A)}_clearSvgElement(){let A=this._elementRef.nativeElement,i=A.childNodes.length;for(this._elementsWithExternalReferences&&this._elementsWithExternalReferences.clear();i--;){let n=A.childNodes[i];(n.nodeType!==1||n.nodeName.toLowerCase()==="svg")&&n.remove()}}_updateFontIconClasses(){if(!this._usingFontIcon())return;let A=this._elementRef.nativeElement,i=(this.fontSet?this._iconRegistry.classNameForFontAlias(this.fontSet).split(/ +/):this._iconRegistry.getDefaultFontSetClass()).filter(n=>n.length>0);this._previousFontSetClass.forEach(n=>A.classList.remove(n)),i.forEach(n=>A.classList.add(n)),this._previousFontSetClass=i,this.fontIcon!==this._previousFontIconClass&&!i.includes("mat-ligature-font")&&(this._previousFontIconClass&&A.classList.remove(this._previousFontIconClass),this.fontIcon&&A.classList.add(this.fontIcon),this._previousFontIconClass=this.fontIcon)}_cleanupFontValue(A){return typeof A=="string"?A.trim().split(" ")[0]:A}_prependPathToReferences(A){let i=this._elementsWithExternalReferences;i&&i.forEach((n,o)=>{n.forEach(r=>{o.setAttribute(r.name,`url('${A}#${r.value}')`)})})}_cacheChildrenWithExternalReferences(A){let i=A.querySelectorAll(gpA),n=this._elementsWithExternalReferences=this._elementsWithExternalReferences||new Map;for(let o=0;o<i.length;o++)mq.forEach(r=>{let s=i[o],a=s.getAttribute(r),c=a?a.match(IpA):null;if(c){let l=n.get(s);l||(l=[],n.set(s,l)),l.push({name:r,value:c[1]})}})}_updateSvgIcon(A){if(this._svgNamespace=null,this._svgName=null,this._currentIconFetch.unsubscribe(),A){let[i,n]=this._splitIconName(A);i&&(this._svgNamespace=i),n&&(this._svgName=n),this._currentIconFetch=this._iconRegistry.getNamedSvgIcon(n,i).pipe(On(1)).subscribe(o=>this._setSvgElement(o),o=>{let r=`Error retrieving icon ${i}:${n}! ${o.message}`;this._errorHandler.handleError(new Error(r))})}}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-icon"]],hostAttrs:["role","img",1,"mat-icon","notranslate"],hostVars:10,hostBindings:function(i,n){i&2&&(Ne("data-mat-icon-type",n._usingFontIcon()?"font":"svg")("data-mat-icon-name",n._svgName||n.fontIcon)("data-mat-icon-namespace",n._svgNamespace||n.fontSet)("fontIcon",n._usingFontIcon()?n.fontIcon:null),fo(n.color?"mat-"+n.color:""),ue("mat-icon-inline",n.inline)("mat-icon-no-color",n.color!=="primary"&&n.color!=="accent"&&n.color!=="warn"))},inputs:{color:"color",inline:[2,"inline","inline",ie],svgIcon:"svgIcon",fontSet:"fontSet",fontIcon:"fontIcon"},exportAs:["matIcon"],ngContentSelectors:ipA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},styles:["mat-icon,mat-icon.mat-primary,mat-icon.mat-accent,mat-icon.mat-warn{color:var(--mat-icon-color, inherit)}.mat-icon{-webkit-user-select:none;user-select:none;background-repeat:no-repeat;display:inline-block;fill:currentColor;height:24px;width:24px;overflow:hidden}.mat-icon.mat-icon-inline{font-size:inherit;height:inherit;line-height:inherit;width:inherit}.mat-icon.mat-ligature-font[fontIcon]::before{content:attr(fontIcon)}[dir=rtl] .mat-icon-rtl-mirror{transform:scale(-1, 1)}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon{display:block}.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-prefix .mat-icon-button .mat-icon,.mat-form-field:not(.mat-form-field-appearance-legacy) .mat-form-field-suffix .mat-icon-button .mat-icon{margin:auto}"],encapsulation:2,changeDetection:0})}return t})(),pq=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,it]})}return t})();var CpA=["audioPlayer"],nC=class t{base64data="";audioPlayerRef;audioSrc="";constructor(){}ngOnChanges(e){e.base64data&&this.base64data&&this.setAudioSource(this.base64data)}setAudioSource(e){e.startsWith("data:")?this.audioSrc=e:this.audioSrc=`data:audio/mpeg;base64,${e}`,this.audioPlayerRef&&this.audioPlayerRef.nativeElement&&this.audioPlayerRef.nativeElement.load()}play(){this.audioPlayerRef&&this.audioPlayerRef.nativeElement&&this.audioPlayerRef.nativeElement.play()}pause(){this.audioPlayerRef&&this.audioPlayerRef.nativeElement&&this.audioPlayerRef.nativeElement.pause()}stop(){this.audioPlayerRef&&this.audioPlayerRef.nativeElement&&(this.audioPlayerRef.nativeElement.pause(),this.audioPlayerRef.nativeElement.currentTime=0)}static \u0275fac=function(A){return new(A||t)};static \u0275cmp=zA({type:t,selectors:[["app-audio-player"]],viewQuery:function(A,i){if(A&1&&Te(CpA,5),A&2){let n;XA(n=$A())&&(i.audioPlayerRef=n.first)}},inputs:{base64data:"base64data"},standalone:!1,features:[ti],decls:3,vars:1,consts:[["audioPlayer",""],["controls","",3,"src"]],template:function(A,i){A&1&&(S(0,"div"),JA(1,"audio",1,0),F()),A&2&&(G(),yA("src",i.audioSrc,ja))},styles:[".audio-player-container[_ngcontent-%COMP%]{display:flex;justify-content:center;align-items:center;padding:15px;background-color:#f0f0f0;border-radius:8px;box-shadow:0 2px 5px #0000001a;margin:20px auto;max-width:350px}audio[_ngcontent-%COMP%]{outline:none;border-radius:5px;width:350px}.custom-controls[_ngcontent-%COMP%]{margin-top:10px;display:flex;gap:10px}.custom-controls[_ngcontent-%COMP%]   button[_ngcontent-%COMP%]{padding:8px 15px;border:none;border-radius:5px;background-color:#007bff;color:#fff;cursor:pointer;font-size:14px;transition:background-color .2s ease}.custom-controls[_ngcontent-%COMP%]   button[_ngcontent-%COMP%]:hover{background-color:#0056b3}"]})};function dpA(t,e){t&1&&JA(0,"hr",2)}function BpA(t,e){if(t&1&&(S(0,"mat-option",7),AA(1),F()),t&2){let A=e.$implicit;yA("value",A),G(),Gt(A.versionId)}}function EpA(t,e){if(t&1){let A=be();S(0,"div")(1,"img",9),hA("click",function(){RA(A);let n=O().$index,o=O();return xA(o.openViewImageDialog(o.selectedArtifacts[n].data))}),F()()}if(t&2){let A,i=O().$index,n=O();G(),yA("src",(A=n.selectedArtifacts[i].data)!==null&&A!==void 0?A:"",ja)}}function QpA(t,e){if(t&1&&(S(0,"div"),JA(1,"app-audio-player",10),F()),t&2){let A=O().$index,i=O();G(),yA("base64data",i.selectedArtifacts[A].data)}}function hpA(t,e){if(t&1){let A=be();S(0,"div",1),_A(1,dpA,1,0,"hr",2),S(2,"div",3)(3,"button",4),hA("click",function(){let n=RA(A).$index,o=O();return xA(o.openArtifact(o.selectedArtifacts[n].data,o.selectedArtifacts[n].mimeType))}),AA(4),F()(),S(5,"div",3)(6,"span"),AA(7," Version: "),F(),S(8,"div",5)(9,"mat-select",6),da("ngModelChange",function(n){let o=RA(A).$index,r=O();return Va(r.selectedArtifacts[o],n)||(r.selectedArtifacts[o]=n),xA(n)}),hA("selectionChange",function(n){let o=RA(A).$index,r=O();return xA(r.onArtifactVersionChange(n,o))}),Dn(10,BpA,2,2,"mat-option",7,to),F()(),S(12,"button",8),hA("click",function(){let n=RA(A).$index,o=O();return xA(o.downloadArtifact(o.selectedArtifacts[n]))}),S(13,"mat-icon"),AA(14,"file_download"),F(),AA(15," Download "),F()(),S(16,"div"),_A(17,EpA,2,1,"div")(18,QpA,2,1,"div"),F()()}if(t&2){let A,i=e.$implicit,n=e.$index,o=O();G(),GA(n>0?1:-1),G(3),Et(" ",o.getArtifactName(i)," "),G(5),Ca("ngModel",o.selectedArtifacts[n]),G(),yn(o.getSortedArtifactsFromId(i)),G(7),GA((A=o.selectedArtifacts[n].mediaType)===o.MediaType.IMAGE?17:A===o.MediaType.AUDIO?18:-1)}}var upA="default_artifact_name",Ou=(n=>(n.IMAGE="image",n.AUDIO="audio",n.TEXT="text",n.UNSPECIFIED="unspecified",n))(Ou||{});function M8(t){let e=t.toLowerCase();for(let A of Object.values(Ou))if(A!=="unspecified"&&e.startsWith(A+"/"))return A;return"unspecified"}function fpA(t){return t?t.startsWith("image/"):!1}function mpA(t){return t?t.startsWith("audio/"):!1}function cS(t,e){try{if(!t)return;let A=t;if(t.startsWith("data:")&&t.includes(";base64,")&&(A=A.substring(A.indexOf(";base64,")+8)),!e||!A)return;let i=atob(A),n=new Array(i.length);for(let c=0;c<i.length;c++)n[c]=i.charCodeAt(c);let o=new Uint8Array(n),r=new Blob([o],{type:e}),s=URL.createObjectURL(r),a=window.open(s,"_blank");a?a.focus():alert("Pop-up blocked! Please allow pop-ups for this site to open the data in a new tab.")}catch{alert("Could not open the data. It might be invalid or too large. Check the browser console for errors.")}}var zu=class t{constructor(e,A){this.downloadService=e;this.dialog=A}artifacts=[];selectedArtifacts=[];isArtifactAudio=mpA;isArtifactImage=fpA;MediaType=Ou;openBase64InNewTab=cS;ngOnChanges(e){if(e.artifacts){this.selectedArtifacts=[];for(let A of this.getDistinctArtifactIds())this.selectedArtifacts.push(this.getSortedArtifactsFromId(A)[0])}}downloadArtifact(e){this.downloadService.downloadBase64Data(e.data,e.mimeType,e.id)}getArtifactName(e){return e??upA}getDistinctArtifactIds(){return[...new Set(this.artifacts.map(e=>e.id))]}getSortedArtifactsFromId(e){return this.artifacts.filter(A=>A.id===e).sort((A,i)=>i.versionId-A.versionId)}onArtifactVersionChange(e,A){this.selectedArtifacts[A]=e.value}openViewImageDialog(e){if(!e||!e.startsWith("data:")||e.indexOf(";base64,")===-1)return;let A=this.dialog.open(v0,{maxWidth:"90vw",maxHeight:"90vh",data:{imageData:e}})}openArtifact(e,A){if(this.isArtifactImage(A)){this.openViewImageDialog(e);return}this.openBase64InNewTab(e,A)}static \u0275fac=function(A){return new(A||t)(PA(O2),PA(qs))};static \u0275cmp=zA({type:t,selectors:[["app-artifact-tab"]],inputs:{artifacts:"artifacts"},standalone:!1,features:[ti],decls:3,vars:0,consts:[[1,"artifact-container"],[1,"artifact-box"],[1,"white-separator"],[1,"artifact-metadata"],[1,"link-style-button",3,"click"],[1,"version-select-container"],[3,"ngModelChange","selectionChange","ngModel"],[3,"value"],["mat-flat-button","",1,"download-button",3,"click"],["alt","artifact.id",1,"generated-image",3,"click","src"],[3,"base64data"]],template:function(A,i){A&1&&(S(0,"div",0),Dn(1,hpA,19,4,"div",1,to),F()),A&2&&(G(),yn(i.getDistinctArtifactIds()))},dependencies:[Ea,ec,P2,Dr,NB,U2,nC],styles:[".artifact-container[_ngcontent-%COMP%]{display:flex;flex-wrap:wrap}.artifact-box[_ngcontent-%COMP%]{padding:10px;max-width:100%;margin-left:26px;display:flex;flex-direction:column}.artifact-metadata[_ngcontent-%COMP%]{display:flex;align-items:center;margin-bottom:15px;flex-wrap:wrap;gap:5px}.download-button[_ngcontent-%COMP%]{background-color:#8ab4f8!important;margin-left:35px;width:130px;height:28px;font-size:14px}.generated-image[_ngcontent-%COMP%]{max-width:60%;border-radius:8px;cursor:pointer}hr.white-separator[_ngcontent-%COMP%]{border:none;border-top:1px solid white;margin-bottom:1.2em;margin-right:15px}.version-select-container[_ngcontent-%COMP%]{background-color:#212123;width:80px;margin-left:15px}.link-style-button[_ngcontent-%COMP%]{background:none;border:none;padding:0;font:inherit;color:#007bff!important;text-decoration:underline;cursor:pointer;outline:none}.link-style-button[_ngcontent-%COMP%]:hover{color:#0056b3;text-decoration:underline}.link-style-button[_ngcontent-%COMP%]:focus{outline:1px dotted #007bff}.link-style-button[_ngcontent-%COMP%]:active{color:#004085}.link-style-button[_ngcontent-%COMP%]:disabled{color:#6c757d;text-decoration:none;cursor:not-allowed}"]})};function wo(t){return Array.isArray(t)}function xo(t){return t!==null&&typeof t=="object"&&(t.constructor===void 0||t.constructor.name==="Object")}function lS(t){return t&&typeof t=="object"?t.op==="add":!1}function gS(t){return t&&typeof t=="object"?t.op==="remove":!1}function k8(t){return t&&typeof t=="object"?t.op==="replace":!1}function S8(t){return t&&typeof t=="object"?t.op==="copy":!1}function j2(t){return t&&typeof t=="object"?t.op==="move":!1}function Dq(t,e){return JSON.stringify(t)===JSON.stringify(e)}function wpA(t,e){return t===e}function IS(t){return t.slice(0,t.length-1)}function yq(t){return t[t.length-1]}function vq(t,e){let A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:wpA;if(t.length<e.length)return!1;for(let i=0;i<e.length;i++)if(!A(t[i],e[i]))return!1;return!0}function CS(t){return typeof t=="object"&&t!==null}function dS(t){if(wo(t)){let e=t.slice();return Object.getOwnPropertySymbols(t).forEach(A=>{e[A]=t[A]}),e}else if(xo(t)){let e=rA({},t);return Object.getOwnPropertySymbols(t).forEach(A=>{e[A]=t[A]}),e}else return t}function BS(t,e,A){if(t[e]===A)return t;{let i=dS(t);return i[e]=A,i}}function Ke(t,e){let A=t,i=0;for(;i<e.length;)xo(A)?A=A[e[i]]:wo(A)?A=A[parseInt(e[i])]:A=void 0,i++;return A}function cs(t,e,A){let i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:!1;if(e.length===0)return A;let n=e[0],o=cs(t?t[n]:void 0,e.slice(1),A,i);if(xo(t)||wo(t))return BS(t,n,o);if(i){let r=DpA.test(n)?[]:{};return r[n]=o,r}else throw new Error("Path does not exist")}var DpA=/^\d+$/;function Pu(t,e,A){if(e.length===0)return A(t);if(!CS(t))throw new Error("Path doesn't exist");let i=e[0],n=Pu(t[i],e.slice(1),A);return BS(t,i,n)}function oC(t,e){if(e.length===0)return t;if(!CS(t))throw new Error("Path does not exist");if(e.length===1){let n=e[0];if(n in t){let o=dS(t);return wo(o)&&o.splice(parseInt(n),1),xo(o)&&delete o[n],o}else return t}let A=e[0],i=oC(t[A],e.slice(1));return BS(t,A,i)}function ju(t,e,A){let i=e.slice(0,e.length-1),n=e[e.length-1];return Pu(t,i,o=>{if(!Array.isArray(o))throw new TypeError("Array expected at path "+JSON.stringify(i));let r=dS(o);return r.splice(parseInt(n),0,A),r})}function Ms(t,e){return t===void 0?!1:e.length===0?!0:t===null?!1:Ms(t[e[0]],e.slice(1))}function ks(t){let e=t.split("/");return e.shift(),e.map(A=>A.replace(/~1/g,"/").replace(/~0/g,"~"))}function Ct(t){return t.map(bq).join("")}function bq(t){return"/"+String(t).replace(/~/g,"~0").replace(/\//g,"~1")}function qu(t,e){return t+bq(e)}function Da(t,e,A){let i=t;for(let n=0;n<e.length;n++){xpA(e[n]);let o=e[n];if(A&&A.before){let a=A.before(i,o);if(a!==void 0){if(a.document!==void 0&&(i=a.document),a.json!==void 0)throw new Error('Deprecation warning: returned object property ".json" has been renamed to ".document"');a.operation!==void 0&&(o=a.operation)}}let r=i,s=ya(i,o.path);if(o.op==="add")i=bpA(i,s,o.value);else if(o.op==="remove")i=vpA(i,s);else if(o.op==="replace")i=ypA(i,s,o.value);else if(o.op==="copy")i=MpA(i,s,Vu(o.from));else if(o.op==="move")i=kpA(i,s,Vu(o.from));else if(o.op==="test")SpA(i,s,o.value);else throw new Error("Unknown JSONPatch operation "+JSON.stringify(o));if(A&&A.after){let a=A.after(i,o,r);a!==void 0&&(i=a)}}return i}function ypA(t,e,A){return cs(t,e,A)}function vpA(t,e){return oC(t,e)}function bpA(t,e,A){return YB(t,e)?ju(t,e,A):cs(t,e,A)}function MpA(t,e,A){let i=Ke(t,A);if(YB(t,e))return ju(t,e,i);{let n=Ke(t,A);return cs(t,e,n)}}function kpA(t,e,A){let i=Ke(t,A),n=oC(t,A);return YB(n,e)?ju(n,e,i):cs(n,e,i)}function SpA(t,e,A){if(A===void 0)throw new Error(`Test failed: no value provided (path: "${Ct(e)}")`);if(!Ms(t,e))throw new Error(`Test failed: path not found (path: "${Ct(e)}")`);let i=Ke(t,e);if(!Dq(i,A))throw new Error(`Test failed, value differs (path: "${Ct(e)}")`)}function YB(t,e){if(e.length===0)return!1;let A=Ke(t,IS(e));return Array.isArray(A)}function RpA(t,e){if(yq(e)!=="-")return e;let A=IS(e),i=Ke(t,A);return A.concat(i.length)}function xpA(t){if(!["add","remove","replace","copy","move","test"].includes(t.op))throw new Error("Unknown JSONPatch op "+JSON.stringify(t.op));if(typeof t.path!="string")throw new Error('Required property "path" missing or not a string in operation '+JSON.stringify(t));if((t.op==="copy"||t.op==="move")&&typeof t.from!="string")throw new Error('Required property "from" missing or not a string in operation '+JSON.stringify(t))}function ya(t,e){return RpA(t,ks(e))}function Vu(t){return ks(t)}function R8(t,e,A){let i=[];return Da(t,e,{before:(o,r)=>{let s,a=ya(o,r.path);if(r.op==="add")s=Sq(o,a);else if(r.op==="remove")s=kq(o,a);else if(r.op==="replace")s=Mq(o,a);else if(r.op==="copy")s=LpA(o,a);else if(r.op==="move")s=FpA(o,a,Vu(r.from));else if(r.op==="test")s=[];else throw new Error("Unknown JSONPatch operation "+JSON.stringify(r));let c;if(A&&A.before){let l=A.before(o,r,s);if(l&&l.revertOperations&&(s=l.revertOperations),l&&l.document&&(c=l.document),l&&l.json)throw new Error('Deprecation warning: returned object property ".json" has been renamed to ".document"')}if(i=s.concat(i),c!==void 0)return{document:c}}}),i}function Mq(t,e){return[{op:"replace",path:Ct(e),value:Ke(t,e)}]}function kq(t,e){return[{op:"add",path:Ct(e),value:Ke(t,e)}]}function Sq(t,e){return YB(t,e)||!Ms(t,e)?[{op:"remove",path:Ct(e)}]:Mq(t,e)}function LpA(t,e){return Sq(t,e)}function FpA(t,e,A){if(e.length<A.length&&vq(A,e))return[{op:"replace",path:Ct(e),value:t}];let i={op:"move",from:Ct(e),path:Ct(A)};return!YB(t,e)&&Ms(t,e)?[i,...kq(t,e)]:[i]}var s_=jQ(xq(),1);var M0=class extends Error{constructor(e,A){super(`${e} at position ${A}`),this.position=A}};function Lq(t){return/^[0-9A-Fa-f]$/.test(t)}function V2(t){return t>="0"&&t<="9"}function Fq(t){return t>=" "}function Zu(t){return`,:[]/{}()
+`.includes(t)}function hS(t){return t>="a"&&t<="z"||t>="A"&&t<="Z"||t==="_"||t==="$"}function uS(t){return t>="a"&&t<="z"||t>="A"&&t<="Z"||t==="_"||t==="$"||t>="0"&&t<="9"}var fS=/^(http|https|ftp|mailto|file|data|irc):\/\/$/,mS=/^[A-Za-z0-9-._~:/?#@!$&'()*+;=]$/;function pS(t){return`,[]/{}
+`.includes(t)}function wS(t){return Wu(t)||OpA.test(t)}var OpA=/^[[{\w-]$/;function Nq(t){return t===`
`||t==="\r"||t==="	"||t==="\b"||t==="\f"}function q2(t,e){let A=t.charCodeAt(e);return A===32||A===10||A===9||A===13}function _q(t,e){let A=t.charCodeAt(e);return A===32||A===9||A===13}function Gq(t,e){let A=t.charCodeAt(e);return A===160||A>=8192&&A<=8202||A===8239||A===8287||A===12288}function Wu(t){return DS(t)||F8(t)}function DS(t){return t==='"'||t==="\u201C"||t==="\u201D"}function yS(t){return t==='"'}function F8(t){return t==="'"||t==="\u2018"||t==="\u2019"||t==="`"||t==="\xB4"}function vS(t){return t==="'"}function JB(t,e){let A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:!1,i=t.lastIndexOf(e);return i!==-1?t.substring(0,i)+(A?"":t.substring(i+1)):t}function Sc(t,e){let A=t.length;if(!q2(t,A-1))return t+e;for(;q2(t,A-1);)A--;return t.substring(0,A)+e+t.substring(A)}function Uq(t,e,A){return t.substring(0,e)+t.substring(e+A)}function Kq(t){return/[,\n][ \t\r]*$/.test(t)}var PpA={"\b":"\\b","\f":"\\f","\n":"\\n","\r":"\\r","	":"\\t"},jpA={'"':'"',"\\":"\\","/":"/",b:"\b",f:"\f",n:`
`,r:"\r",t:"	"};function Rc(t){let e=0,A="";c(["```","[```","{```"]),o()||QA(),c(["```","```]","```}"]);let n=I(",");for(n&&r(),wS(t[e])&&Kq(A)?(n||(A=Sc(A,",")),u()):n&&(A=JB(A,","));t[e]==="}"||t[e]==="]";)e++,r();if(e>=t.length)return A;gA();function o(){r();let tA=E()||h()||D()||R()||w()||K(!1)||z();return r(),tA}function r(){let tA=arguments.length>0&&arguments[0]!==void 0?arguments[0]:!0,cA=e,pA=s(tA);do pA=a(),pA&&(pA=s(tA));while(pA);return e>cA}function s(tA){let cA=tA?q2:_q,pA="";for(;;)if(cA(t,e))pA+=t[e],e++;else if(Gq(t,e))pA+=" ",e++;else break;return pA.length>0?(A+=pA,!0):!1}function a(){if(t[e]==="/"&&t[e+1]==="*"){for(;e<t.length&&!qpA(t,e);)e++;return e+=2,!0}if(t[e]==="/"&&t[e+1]==="/"){for(;e<t.length&&t[e]!==`
`;)e++;return!0}return!1}function c(tA){if(l(tA)){if(hS(t[e]))for(;e<t.length&&uS(t[e]);)e++;return r(),!0}return!1}function l(tA){for(let cA of tA){let pA=e+cA.length;if(t.slice(e,pA)===cA)return e=pA,!0}return!1}function I(tA){return t[e]===tA?(A+=t[e],e++,!0):!1}function C(tA){return t[e]===tA?(e++,!0):!1}function d(){return C("\\")}function B(){return r(),t[e]==="."&&t[e+1]==="."&&t[e+2]==="."?(e+=3,r(),C(","),!0):!1}function E(){if(t[e]==="{"){A+="{",e++,r(),C(",")&&r();let tA=!0;for(;e<t.length&&t[e]!=="}";){let cA;if(tA?(cA=!0,tA=!1):(cA=I(","),cA||(A=Sc(A,",")),r()),B(),!(D()||K(!0))){t[e]==="}"||t[e]==="{"||t[e]==="]"||t[e]==="["||t[e]===void 0?A=JB(A,","):BA();break}r();let VA=I(":"),oe=e>=t.length;VA||(wS(t[e])||oe?A=Sc(A,":"):lA()),o()||(VA||oe?A+="null":lA())}return t[e]==="}"?(A+="}",e++):A=Sc(A,"}"),!0}return!1}function h(){if(t[e]==="["){A+="[",e++,r(),C(",")&&r();let tA=!0;for(;e<t.length&&t[e]!=="]";)if(tA?tA=!1:I(",")||(A=Sc(A,",")),B(),!o()){A=JB(A,",");break}return t[e]==="]"?(A+="]",e++):A=Sc(A,"]"),!0}return!1}function u(){let tA=!0,cA=!0;for(;cA;)tA?tA=!1:I(",")||(A=Sc(A,",")),cA=o();cA||(A=JB(A,",")),A=`[
${A}
]`}function D(){let tA=arguments.length>0&&arguments[0]!==void 0?arguments[0]:!1,cA=arguments.length>1&&arguments[1]!==void 0?arguments[1]:-1,pA=t[e]==="\\";if(pA&&(e++,pA=!0),Wu(t[e])){let VA=yS(t[e])?yS:vS(t[e])?vS:F8(t[e])?F8:DS,oe=e,KA=A.length,CA='"';for(e++;;){if(e>=t.length){let TA=U(e-1);return!tA&&Zu(t.charAt(TA))?(e=oe,A=A.substring(0,KA),D(!0)):(CA=Sc(CA,'"'),A+=CA,!0)}if(e===cA)return CA=Sc(CA,'"'),A+=CA,!0;if(VA(t[e])){let TA=e,Ze=CA.length;if(CA+='"',e++,A+=CA,r(!1),tA||e>=t.length||Zu(t[e])||Wu(t[e])||V2(t[e]))return L(),!0;let He=U(TA-1),uA=t.charAt(He);if(uA===",")return e=oe,A=A.substring(0,KA),D(!1,He);if(Zu(uA))return e=oe,A=A.substring(0,KA),D(!0);A=A.substring(0,KA),e=TA+1,CA=`${CA.substring(0,Ze)}\\${CA.substring(Ze)}`}else if(tA&&pS(t[e])){if(t[e-1]===":"&&fS.test(t.substring(oe+1,e+2)))for(;e<t.length&&mS.test(t[e]);)CA+=t[e],e++;return CA=Sc(CA,'"'),A+=CA,L(),!0}else if(t[e]==="\\"){let TA=t.charAt(e+1);if(jpA[TA]!==void 0)CA+=t.slice(e,e+2),e+=2;else if(TA==="u"){let He=2;for(;He<6&&Lq(t[e+He]);)He++;He===6?(CA+=t.slice(e,e+6),e+=6):e+He>=t.length?e=t.length:vA()}else CA+=TA,e+=2}else{let TA=t.charAt(e);TA==='"'&&t[e-1]!=="\\"?(CA+=`\\${TA}`,e++):Nq(TA)?(CA+=PpA[TA],e++):(Fq(TA)||j(TA),CA+=TA,e++)}pA&&d()}}return!1}function L(){let tA=!1;for(r();t[e]==="+";){tA=!0,e++,r(),A=JB(A,'"',!0);let cA=A.length;D()?A=Uq(A,cA,1):A=Sc(A,'"')}return tA}function R(){let tA=e;if(t[e]==="-"){if(e++,H())return q(tA),!0;if(!V2(t[e]))return e=tA,!1}for(;V2(t[e]);)e++;if(t[e]==="."){if(e++,H())return q(tA),!0;if(!V2(t[e]))return e=tA,!1;for(;V2(t[e]);)e++}if(t[e]==="e"||t[e]==="E"){if(e++,(t[e]==="-"||t[e]==="+")&&e++,H())return q(tA),!0;if(!V2(t[e]))return e=tA,!1;for(;V2(t[e]);)e++}if(!H())return e=tA,!1;if(e>tA){let cA=t.slice(tA,e),pA=/^0\d/.test(cA);return A+=pA?`"${cA}"`:cA,!0}return!1}function w(){return _("true","true")||_("false","false")||_("null","null")||_("True","true")||_("False","false")||_("None","null")}function _(tA,cA){return t.slice(e,e+tA.length)===tA?(A+=cA,e+=tA.length,!0):!1}function K(tA){let cA=e;if(hS(t[e])){for(;e<t.length&&uS(t[e]);)e++;let pA=e;for(;q2(t,pA);)pA++;if(t[pA]==="(")return e=pA+1,o(),t[e]===")"&&(e++,t[e]===";"&&e++),!0}for(;e<t.length&&!pS(t[e])&&!Wu(t[e])&&(!tA||t[e]!==":");)e++;if(t[e-1]===":"&&fS.test(t.substring(cA,e+2)))for(;e<t.length&&mS.test(t[e]);)e++;if(e>cA){for(;q2(t,e-1)&&e>0;)e--;let pA=t.slice(cA,e);return A+=pA==="undefined"?"null":JSON.stringify(pA),t[e]==='"'&&e++,!0}}function z(){if(t[e]==="/"){let tA=e;for(e++;e<t.length&&(t[e]!=="/"||t[e-1]==="\\");)e++;return e++,A+=`"${t.substring(tA,e)}"`,!0}}function U(tA){let cA=tA;for(;cA>0&&q2(t,cA);)cA--;return cA}function H(){return e>=t.length||Zu(t[e])||q2(t,e)}function q(tA){A+=`${t.slice(tA,e)}0`}function j(tA){throw new M0(`Invalid character ${JSON.stringify(tA)}`,e)}function gA(){throw new M0(`Unexpected character ${JSON.stringify(t[e])}`,e)}function QA(){throw new M0("Unexpected end of json string",t.length)}function BA(){throw new M0("Object key expected",e)}function lA(){throw new M0("Colon expected",e)}function vA(){let tA=t.slice(e,e+6);throw new M0(`Invalid unicode character "${tA}"`,e)}}function qpA(t,e){return t[e]==="*"&&t[e+1]==="/"}var VpA=typeof global=="object"&&global&&global.Object===Object&&global,N8=VpA;var ZpA=typeof self=="object"&&self&&self.Object===Object&&self,WpA=N8||ZpA||Function("return this")(),gr=WpA;var XpA=gr.Symbol,Zr=XpA;var Yq=Object.prototype,$pA=Yq.hasOwnProperty,A6A=Yq.toString,Xu=Zr?Zr.toStringTag:void 0;function e6A(t){var e=$pA.call(t,Xu),A=t[Xu];try{t[Xu]=void 0;var i=!0}catch{}var n=A6A.call(t);return i&&(e?t[Xu]=A:delete t[Xu]),n}var Jq=e6A;var t6A=Object.prototype,i6A=t6A.toString;function n6A(t){return i6A.call(t)}var Tq=n6A;var o6A="[object Null]",r6A="[object Undefined]",Hq=Zr?Zr.toStringTag:void 0;function s6A(t){return t==null?t===void 0?r6A:o6A:Hq&&Hq in Object(t)?Jq(t):Tq(t)}var Ql=s6A;function a6A(t){return t!=null&&typeof t=="object"}var Vs=a6A;var c6A="[object Symbol]";function l6A(t){return typeof t=="symbol"||Vs(t)&&Ql(t)==c6A}var ac=l6A;function g6A(t,e){for(var A=-1,i=t==null?0:t.length,n=Array(i);++A<i;)n[A]=e(t[A],A,t);return n}var Z2=g6A;var I6A=Array.isArray,Gn=I6A;var C6A=1/0,zq=Zr?Zr.prototype:void 0,Oq=zq?zq.toString:void 0;function Pq(t){if(typeof t=="string")return t;if(Gn(t))return Z2(t,Pq)+"";if(ac(t))return Oq?Oq.call(t):"";var e=t+"";return e=="0"&&1/t==-C6A?"-0":e}var jq=Pq;var d6A=/\s/;function B6A(t){for(var e=t.length;e--&&d6A.test(t.charAt(e)););return e}var qq=B6A;var E6A=/^\s+/;function Q6A(t){return t&&t.slice(0,qq(t)+1).replace(E6A,"")}var Vq=Q6A;function h6A(t){var e=typeof t;return t!=null&&(e=="object"||e=="function")}var Kr=h6A;var Zq=NaN,u6A=/^[-+]0x[0-9a-f]+$/i,f6A=/^0b[01]+$/i,m6A=/^0o[0-7]+$/i,p6A=parseInt;function w6A(t){if(typeof t=="number")return t;if(ac(t))return Zq;if(Kr(t)){var e=typeof t.valueOf=="function"?t.valueOf():t;t=Kr(e)?e+"":e}if(typeof t!="string")return t===0?t:+t;t=Vq(t);var A=f6A.test(t);return A||m6A.test(t)?p6A(t.slice(2),A?2:8):u6A.test(t)?Zq:+t}var $u=w6A;var Wq=1/0,D6A=17976931348623157e292;function y6A(t){if(!t)return t===0?t:0;if(t=$u(t),t===Wq||t===-Wq){var e=t<0?-1:1;return e*D6A}return t===t?t:0}var TB=y6A;function v6A(t){var e=TB(t),A=e%1;return e===e?A?e-A:e:0}var Xq=v6A;function b6A(t){return t}var ug=b6A;var M6A="[object AsyncFunction]",k6A="[object Function]",S6A="[object GeneratorFunction]",R6A="[object Proxy]";function x6A(t){if(!Kr(t))return!1;var e=Ql(t);return e==k6A||e==S6A||e==M6A||e==R6A}var _8=x6A;var L6A=gr["__core-js_shared__"],G8=L6A;var $q=function(){var t=/[^.]+$/.exec(G8&&G8.keys&&G8.keys.IE_PROTO||"");return t?"Symbol(src)_1."+t:""}();function F6A(t){return!!$q&&$q in t}var AV=F6A;var N6A=Function.prototype,_6A=N6A.toString;function G6A(t){if(t!=null){try{return _6A.call(t)}catch{}try{return t+""}catch{}}return""}var k0=G6A;var U6A=/[\\^$.*+?()[\]{}|]/g,K6A=/^\[object .+?Constructor\]$/,Y6A=Function.prototype,J6A=Object.prototype,T6A=Y6A.toString,H6A=J6A.hasOwnProperty,z6A=RegExp("^"+T6A.call(H6A).replace(U6A,"\\$&").replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g,"$1.*?")+"$");function O6A(t){if(!Kr(t)||AV(t))return!1;var e=_8(t)?z6A:K6A;return e.test(k0(t))}var eV=O6A;function P6A(t,e){return t?.[e]}var tV=P6A;function j6A(t,e){var A=tV(t,e);return eV(A)?A:void 0}var va=j6A;var q6A=va(gr,"WeakMap"),U8=q6A;var iV=Object.create,V6A=function(){function t(){}return function(e){if(!Kr(e))return{};if(iV)return iV(e);t.prototype=e;var A=new t;return t.prototype=void 0,A}}(),nV=V6A;function Z6A(t,e,A){switch(A.length){case 0:return t.call(e);case 1:return t.call(e,A[0]);case 2:return t.call(e,A[0],A[1]);case 3:return t.call(e,A[0],A[1],A[2])}return t.apply(e,A)}var oV=Z6A;function W6A(){}var $o=W6A;function X6A(t,e){var A=-1,i=t.length;for(e||(e=Array(i));++A<i;)e[A]=t[A];return e}var rV=X6A;var $6A=800,A8A=16,e8A=Date.now;function t8A(t){var e=0,A=0;return function(){var i=e8A(),n=A8A-(i-A);if(A=i,n>0){if(++e>=$6A)return arguments[0]}else e=0;return t.apply(void 0,arguments)}}var sV=t8A;function i8A(t){return function(){return t}}var aV=i8A;var n8A=function(){try{var t=va(Object,"defineProperty");return t({},"",{}),t}catch{}}(),HB=n8A;var o8A=HB?function(t,e){return HB(t,"toString",{configurable:!0,enumerable:!1,value:aV(e),writable:!0})}:ug,cV=o8A;var r8A=sV(cV),lV=r8A;function s8A(t,e){for(var A=-1,i=t==null?0:t.length;++A<i&&e(t[A],A,t)!==!1;);return t}var gV=s8A;var a8A=9007199254740991,c8A=/^(?:0|[1-9]\d*)$/;function l8A(t,e){var A=typeof t;return e=e??a8A,!!e&&(A=="number"||A!="symbol"&&c8A.test(t))&&t>-1&&t%1==0&&t<e}var zB=l8A;function g8A(t,e,A){e=="__proto__"&&HB?HB(t,e,{configurable:!0,enumerable:!0,value:A,writable:!0}):t[e]=A}var W2=g8A;function I8A(t,e){return t===e||t!==t&&e!==e}var X2=I8A;var C8A=Object.prototype,d8A=C8A.hasOwnProperty;function B8A(t,e,A){var i=t[e];(!(d8A.call(t,e)&&X2(i,A))||A===void 0&&!(e in t))&&W2(t,e,A)}var K8=B8A;function E8A(t,e,A,i){var n=!A;A||(A={});for(var o=-1,r=e.length;++o<r;){var s=e[o],a=i?i(A[s],t[s],s,A,t):void 0;a===void 0&&(a=t[s]),n?W2(A,s,a):K8(A,s,a)}return A}var $2=E8A;var IV=Math.max;function Q8A(t,e,A){return e=IV(e===void 0?t.length-1:e,0),function(){for(var i=arguments,n=-1,o=IV(i.length-e,0),r=Array(o);++n<o;)r[n]=i[e+n];n=-1;for(var s=Array(e+1);++n<e;)s[n]=i[n];return s[e]=A(r),oV(t,this,s)}}var CV=Q8A;function h8A(t,e){return lV(CV(t,e,ug),t+"")}var dV=h8A;var u8A=9007199254740991;function f8A(t){return typeof t=="number"&&t>-1&&t%1==0&&t<=u8A}var OB=f8A;function m8A(t){return t!=null&&OB(t.length)&&!_8(t)}var xc=m8A;function p8A(t,e,A){if(!Kr(A))return!1;var i=typeof e;return(i=="number"?xc(A)&&zB(e,A.length):i=="string"&&e in A)?X2(A[e],t):!1}var A4=p8A;var w8A=Object.prototype;function D8A(t){var e=t&&t.constructor,A=typeof e=="function"&&e.prototype||w8A;return t===A}var A1=D8A;function y8A(t,e){for(var A=-1,i=Array(t);++A<t;)i[A]=e(A);return i}var Y8=y8A;var v8A="[object Arguments]";function b8A(t){return Vs(t)&&Ql(t)==v8A}var bS=b8A;var BV=Object.prototype,M8A=BV.hasOwnProperty,k8A=BV.propertyIsEnumerable,S8A=bS(function(){return arguments}())?bS:function(t){return Vs(t)&&M8A.call(t,"callee")&&!k8A.call(t,"callee")},e1=S8A;function R8A(){return!1}var EV=R8A;var uV=typeof exports=="object"&&exports&&!exports.nodeType&&exports,QV=uV&&typeof module=="object"&&module&&!module.nodeType&&module,x8A=QV&&QV.exports===uV,hV=x8A?gr.Buffer:void 0,L8A=hV?hV.isBuffer:void 0,F8A=L8A||EV,S0=F8A;var N8A="[object Arguments]",_8A="[object Array]",G8A="[object Boolean]",U8A="[object Date]",K8A="[object Error]",Y8A="[object Function]",J8A="[object Map]",T8A="[object Number]",H8A="[object Object]",z8A="[object RegExp]",O8A="[object Set]",P8A="[object String]",j8A="[object WeakMap]",q8A="[object ArrayBuffer]",V8A="[object DataView]",Z8A="[object Float32Array]",W8A="[object Float64Array]",X8A="[object Int8Array]",$8A="[object Int16Array]",A5A="[object Int32Array]",e5A="[object Uint8Array]",t5A="[object Uint8ClampedArray]",i5A="[object Uint16Array]",n5A="[object Uint32Array]",Ho={};Ho[Z8A]=Ho[W8A]=Ho[X8A]=Ho[$8A]=Ho[A5A]=Ho[e5A]=Ho[t5A]=Ho[i5A]=Ho[n5A]=!0;Ho[N8A]=Ho[_8A]=Ho[q8A]=Ho[G8A]=Ho[V8A]=Ho[U8A]=Ho[K8A]=Ho[Y8A]=Ho[J8A]=Ho[T8A]=Ho[H8A]=Ho[z8A]=Ho[O8A]=Ho[P8A]=Ho[j8A]=!1;function o5A(t){return Vs(t)&&OB(t.length)&&!!Ho[Ql(t)]}var fV=o5A;function r5A(t){return function(e){return t(e)}}var t1=r5A;var mV=typeof exports=="object"&&exports&&!exports.nodeType&&exports,e4=mV&&typeof module=="object"&&module&&!module.nodeType&&module,s5A=e4&&e4.exports===mV,MS=s5A&&N8.process,a5A=function(){try{var t=e4&&e4.require&&e4.require("util").types;return t||MS&&MS.binding&&MS.binding("util")}catch{}}(),R0=a5A;var pV=R0&&R0.isTypedArray,c5A=pV?t1(pV):fV,PB=c5A;var l5A=Object.prototype,g5A=l5A.hasOwnProperty;function I5A(t,e){var A=Gn(t),i=!A&&e1(t),n=!A&&!i&&S0(t),o=!A&&!i&&!n&&PB(t),r=A||i||n||o,s=r?Y8(t.length,String):[],a=s.length;for(var c in t)(e||g5A.call(t,c))&&!(r&&(c=="length"||n&&(c=="offset"||c=="parent")||o&&(c=="buffer"||c=="byteLength"||c=="byteOffset")||zB(c,a)))&&s.push(c);return s}var J8=I5A;function C5A(t,e){return function(A){return t(e(A))}}var T8=C5A;var d5A=T8(Object.keys,Object),wV=d5A;var B5A=Object.prototype,E5A=B5A.hasOwnProperty;function Q5A(t){if(!A1(t))return wV(t);var e=[];for(var A in Object(t))E5A.call(t,A)&&A!="constructor"&&e.push(A);return e}var H8=Q5A;function h5A(t){return xc(t)?J8(t):H8(t)}var Lc=h5A;function u5A(t){var e=[];if(t!=null)for(var A in Object(t))e.push(A);return e}var DV=u5A;var f5A=Object.prototype,m5A=f5A.hasOwnProperty;function p5A(t){if(!Kr(t))return DV(t);var e=A1(t),A=[];for(var i in t)i=="constructor"&&(e||!m5A.call(t,i))||A.push(i);return A}var yV=p5A;function w5A(t){return xc(t)?J8(t,!0):yV(t)}var jB=w5A;var D5A=/\.|\[(?:[^[\]]*|(["'])(?:(?!\1)[^\\]|\\.)*?\1)\]/,y5A=/^\w*$/;function v5A(t,e){if(Gn(t))return!1;var A=typeof t;return A=="number"||A=="symbol"||A=="boolean"||t==null||ac(t)?!0:y5A.test(t)||!D5A.test(t)||e!=null&&t in Object(e)}var qB=v5A;var b5A=va(Object,"create"),x0=b5A;function M5A(){this.__data__=x0?x0(null):{},this.size=0}var vV=M5A;function k5A(t){var e=this.has(t)&&delete this.__data__[t];return this.size-=e?1:0,e}var bV=k5A;var S5A="__lodash_hash_undefined__",R5A=Object.prototype,x5A=R5A.hasOwnProperty;function L5A(t){var e=this.__data__;if(x0){var A=e[t];return A===S5A?void 0:A}return x5A.call(e,t)?e[t]:void 0}var MV=L5A;var F5A=Object.prototype,N5A=F5A.hasOwnProperty;function _5A(t){var e=this.__data__;return x0?e[t]!==void 0:N5A.call(e,t)}var kV=_5A;var G5A="__lodash_hash_undefined__";function U5A(t,e){var A=this.__data__;return this.size+=this.has(t)?0:1,A[t]=x0&&e===void 0?G5A:e,this}var SV=U5A;function VB(t){var e=-1,A=t==null?0:t.length;for(this.clear();++e<A;){var i=t[e];this.set(i[0],i[1])}}VB.prototype.clear=vV;VB.prototype.delete=bV;VB.prototype.get=MV;VB.prototype.has=kV;VB.prototype.set=SV;var kS=VB;function K5A(){this.__data__=[],this.size=0}var RV=K5A;function Y5A(t,e){for(var A=t.length;A--;)if(X2(t[A][0],e))return A;return-1}var i1=Y5A;var J5A=Array.prototype,T5A=J5A.splice;function H5A(t){var e=this.__data__,A=i1(e,t);if(A<0)return!1;var i=e.length-1;return A==i?e.pop():T5A.call(e,A,1),--this.size,!0}var xV=H5A;function z5A(t){var e=this.__data__,A=i1(e,t);return A<0?void 0:e[A][1]}var LV=z5A;function O5A(t){return i1(this.__data__,t)>-1}var FV=O5A;function P5A(t,e){var A=this.__data__,i=i1(A,t);return i<0?(++this.size,A.push([t,e])):A[i][1]=e,this}var NV=P5A;function ZB(t){var e=-1,A=t==null?0:t.length;for(this.clear();++e<A;){var i=t[e];this.set(i[0],i[1])}}ZB.prototype.clear=RV;ZB.prototype.delete=xV;ZB.prototype.get=LV;ZB.prototype.has=FV;ZB.prototype.set=NV;var n1=ZB;var j5A=va(gr,"Map"),o1=j5A;function q5A(){this.size=0,this.__data__={hash:new kS,map:new(o1||n1),string:new kS}}var _V=q5A;function V5A(t){var e=typeof t;return e=="string"||e=="number"||e=="symbol"||e=="boolean"?t!=="__proto__":t===null}var GV=V5A;function Z5A(t,e){var A=t.__data__;return GV(e)?A[typeof e=="string"?"string":"hash"]:A.map}var r1=Z5A;function W5A(t){var e=r1(this,t).delete(t);return this.size-=e?1:0,e}var UV=W5A;function X5A(t){return r1(this,t).get(t)}var KV=X5A;function $5A(t){return r1(this,t).has(t)}var YV=$5A;function AwA(t,e){var A=r1(this,t),i=A.size;return A.set(t,e),this.size+=A.size==i?0:1,this}var JV=AwA;function WB(t){var e=-1,A=t==null?0:t.length;for(this.clear();++e<A;){var i=t[e];this.set(i[0],i[1])}}WB.prototype.clear=_V;WB.prototype.delete=UV;WB.prototype.get=KV;WB.prototype.has=YV;WB.prototype.set=JV;var rC=WB;var ewA="Expected a function";function SS(t,e){if(typeof t!="function"||e!=null&&typeof e!="function")throw new TypeError(ewA);var A=function(){var i=arguments,n=e?e.apply(this,i):i[0],o=A.cache;if(o.has(n))return o.get(n);var r=t.apply(this,i);return A.cache=o.set(n,r)||o,r};return A.cache=new(SS.Cache||rC),A}SS.Cache=rC;var TV=SS;var twA=500;function iwA(t){var e=TV(t,function(i){return A.size===twA&&A.clear(),i}),A=e.cache;return e}var HV=iwA;var nwA=/[^.[\]]+|\[(?:(-?\d+(?:\.\d+)?)|(["'])((?:(?!\2)[^\\]|\\.)*?)\2)\]|(?=(?:\.|\[\])(?:\.|\[\]|$))/g,owA=/\\(\\)?/g,rwA=HV(function(t){var e=[];return t.charCodeAt(0)===46&&e.push(""),t.replace(nwA,function(A,i,n,o){e.push(n?o.replace(owA,"$1"):i||A)}),e}),zV=rwA;function swA(t){return t==null?"":jq(t)}var z8=swA;function awA(t,e){return Gn(t)?t:qB(t,e)?[t]:zV(z8(t))}var O8=awA;var cwA=1/0;function lwA(t){if(typeof t=="string"||ac(t))return t;var e=t+"";return e=="0"&&1/t==-cwA?"-0":e}var s1=lwA;function gwA(t,e){e=O8(e,t);for(var A=0,i=e.length;t!=null&&A<i;)t=t[s1(e[A++])];return A&&A==i?t:void 0}var XB=gwA;function IwA(t,e,A){var i=t==null?void 0:XB(t,e);return i===void 0?A:i}var OV=IwA;function CwA(t,e){for(var A=-1,i=e.length,n=t.length;++A<i;)t[n+A]=e[A];return t}var $B=CwA;var PV=Zr?Zr.isConcatSpreadable:void 0;function dwA(t){return Gn(t)||e1(t)||!!(PV&&t&&t[PV])}var jV=dwA;function qV(t,e,A,i,n){var o=-1,r=t.length;for(A||(A=jV),n||(n=[]);++o<r;){var s=t[o];e>0&&A(s)?e>1?qV(s,e-1,A,i,n):$B(n,s):i||(n[n.length]=s)}return n}var VV=qV;var BwA=T8(Object.getPrototypeOf,Object),P8=BwA;function EwA(t,e,A){var i=-1,n=t.length;e<0&&(e=-e>n?0:n+e),A=A>n?n:A,A<0&&(A+=n),n=e>A?0:A-e>>>0,e>>>=0;for(var o=Array(n);++i<n;)o[i]=t[i+e];return o}var ZV=EwA;function QwA(){this.__data__=new n1,this.size=0}var WV=QwA;function hwA(t){var e=this.__data__,A=e.delete(t);return this.size=e.size,A}var XV=hwA;function uwA(t){return this.__data__.get(t)}var $V=uwA;function fwA(t){return this.__data__.has(t)}var AZ=fwA;var mwA=200;function pwA(t,e){var A=this.__data__;if(A instanceof n1){var i=A.__data__;if(!o1||i.length<mwA-1)return i.push([t,e]),this.size=++A.size,this;A=this.__data__=new rC(i)}return A.set(t,e),this.size=A.size,this}var eZ=pwA;function AE(t){var e=this.__data__=new n1(t);this.size=e.size}AE.prototype.clear=WV;AE.prototype.delete=XV;AE.prototype.get=$V;AE.prototype.has=AZ;AE.prototype.set=eZ;var a1=AE;function wwA(t,e){return t&&$2(e,Lc(e),t)}var tZ=wwA;function DwA(t,e){return t&&$2(e,jB(e),t)}var iZ=DwA;var sZ=typeof exports=="object"&&exports&&!exports.nodeType&&exports,nZ=sZ&&typeof module=="object"&&module&&!module.nodeType&&module,ywA=nZ&&nZ.exports===sZ,oZ=ywA?gr.Buffer:void 0,rZ=oZ?oZ.allocUnsafe:void 0;function vwA(t,e){if(e)return t.slice();var A=t.length,i=rZ?rZ(A):new t.constructor(A);return t.copy(i),i}var aZ=vwA;function bwA(t,e){for(var A=-1,i=t==null?0:t.length,n=0,o=[];++A<i;){var r=t[A];e(r,A,t)&&(o[n++]=r)}return o}var cZ=bwA;function MwA(){return[]}var j8=MwA;var kwA=Object.prototype,SwA=kwA.propertyIsEnumerable,lZ=Object.getOwnPropertySymbols,RwA=lZ?function(t){return t==null?[]:(t=Object(t),cZ(lZ(t),function(e){return SwA.call(t,e)}))}:j8,eE=RwA;function xwA(t,e){return $2(t,eE(t),e)}var gZ=xwA;var LwA=Object.getOwnPropertySymbols,FwA=LwA?function(t){for(var e=[];t;)$B(e,eE(t)),t=P8(t);return e}:j8,q8=FwA;function NwA(t,e){return $2(t,q8(t),e)}var IZ=NwA;function _wA(t,e,A){var i=e(t);return Gn(t)?i:$B(i,A(t))}var V8=_wA;function GwA(t){return V8(t,Lc,eE)}var t4=GwA;function UwA(t){return V8(t,jB,q8)}var CZ=UwA;var KwA=va(gr,"DataView"),Z8=KwA;var YwA=va(gr,"Promise"),W8=YwA;var JwA=va(gr,"Set"),X8=JwA;var dZ="[object Map]",TwA="[object Object]",BZ="[object Promise]",EZ="[object Set]",QZ="[object WeakMap]",hZ="[object DataView]",HwA=k0(Z8),zwA=k0(o1),OwA=k0(W8),PwA=k0(X8),jwA=k0(U8),sC=Ql;(Z8&&sC(new Z8(new ArrayBuffer(1)))!=hZ||o1&&sC(new o1)!=dZ||W8&&sC(W8.resolve())!=BZ||X8&&sC(new X8)!=EZ||U8&&sC(new U8)!=QZ)&&(sC=function(t){var e=Ql(t),A=e==TwA?t.constructor:void 0,i=A?k0(A):"";if(i)switch(i){case HwA:return hZ;case zwA:return dZ;case OwA:return BZ;case PwA:return EZ;case jwA:return QZ}return e});var hl=sC;var qwA=Object.prototype,VwA=qwA.hasOwnProperty;function ZwA(t){var e=t.length,A=new t.constructor(e);return e&&typeof t[0]=="string"&&VwA.call(t,"index")&&(A.index=t.index,A.input=t.input),A}var uZ=ZwA;var WwA=gr.Uint8Array,tE=WwA;function XwA(t){var e=new t.constructor(t.byteLength);return new tE(e).set(new tE(t)),e}var iE=XwA;function $wA(t,e){var A=e?iE(t.buffer):t.buffer;return new t.constructor(A,t.byteOffset,t.byteLength)}var fZ=$wA;var ADA=/\w*$/;function eDA(t){var e=new t.constructor(t.source,ADA.exec(t));return e.lastIndex=t.lastIndex,e}var mZ=eDA;var pZ=Zr?Zr.prototype:void 0,wZ=pZ?pZ.valueOf:void 0;function tDA(t){return wZ?Object(wZ.call(t)):{}}var DZ=tDA;function iDA(t,e){var A=e?iE(t.buffer):t.buffer;return new t.constructor(A,t.byteOffset,t.length)}var yZ=iDA;var nDA="[object Boolean]",oDA="[object Date]",rDA="[object Map]",sDA="[object Number]",aDA="[object RegExp]",cDA="[object Set]",lDA="[object String]",gDA="[object Symbol]",IDA="[object ArrayBuffer]",CDA="[object DataView]",dDA="[object Float32Array]",BDA="[object Float64Array]",EDA="[object Int8Array]",QDA="[object Int16Array]",hDA="[object Int32Array]",uDA="[object Uint8Array]",fDA="[object Uint8ClampedArray]",mDA="[object Uint16Array]",pDA="[object Uint32Array]";function wDA(t,e,A){var i=t.constructor;switch(e){case IDA:return iE(t);case nDA:case oDA:return new i(+t);case CDA:return fZ(t,A);case dDA:case BDA:case EDA:case QDA:case hDA:case uDA:case fDA:case mDA:case pDA:return yZ(t,A);case rDA:return new i;case sDA:case lDA:return new i(t);case aDA:return mZ(t);case cDA:return new i;case gDA:return DZ(t)}}var vZ=wDA;function DDA(t){return typeof t.constructor=="function"&&!A1(t)?nV(P8(t)):{}}var bZ=DDA;var yDA="[object Map]";function vDA(t){return Vs(t)&&hl(t)==yDA}var MZ=vDA;var kZ=R0&&R0.isMap,bDA=kZ?t1(kZ):MZ,SZ=bDA;var MDA="[object Set]";function kDA(t){return Vs(t)&&hl(t)==MDA}var RZ=kDA;var xZ=R0&&R0.isSet,SDA=xZ?t1(xZ):RZ,LZ=SDA;var RDA=1,xDA=2,LDA=4,FZ="[object Arguments]",FDA="[object Array]",NDA="[object Boolean]",_DA="[object Date]",GDA="[object Error]",NZ="[object Function]",UDA="[object GeneratorFunction]",KDA="[object Map]",YDA="[object Number]",_Z="[object Object]",JDA="[object RegExp]",TDA="[object Set]",HDA="[object String]",zDA="[object Symbol]",ODA="[object WeakMap]",PDA="[object ArrayBuffer]",jDA="[object DataView]",qDA="[object Float32Array]",VDA="[object Float64Array]",ZDA="[object Int8Array]",WDA="[object Int16Array]",XDA="[object Int32Array]",$DA="[object Uint8Array]",AyA="[object Uint8ClampedArray]",eyA="[object Uint16Array]",tyA="[object Uint32Array]",Lo={};Lo[FZ]=Lo[FDA]=Lo[PDA]=Lo[jDA]=Lo[NDA]=Lo[_DA]=Lo[qDA]=Lo[VDA]=Lo[ZDA]=Lo[WDA]=Lo[XDA]=Lo[KDA]=Lo[YDA]=Lo[_Z]=Lo[JDA]=Lo[TDA]=Lo[HDA]=Lo[zDA]=Lo[$DA]=Lo[AyA]=Lo[eyA]=Lo[tyA]=!0;Lo[GDA]=Lo[NZ]=Lo[ODA]=!1;function $8(t,e,A,i,n,o){var r,s=e&RDA,a=e&xDA,c=e&LDA;if(A&&(r=n?A(t,i,n,o):A(t)),r!==void 0)return r;if(!Kr(t))return t;var l=Gn(t);if(l){if(r=uZ(t),!s)return rV(t,r)}else{var I=hl(t),C=I==NZ||I==UDA;if(S0(t))return aZ(t,s);if(I==_Z||I==FZ||C&&!n){if(r=a||C?{}:bZ(t),!s)return a?IZ(t,iZ(r,t)):gZ(t,tZ(r,t))}else{if(!Lo[I])return n?t:{};r=vZ(t,I,s)}}o||(o=new a1);var d=o.get(t);if(d)return d;o.set(t,r),LZ(t)?t.forEach(function(h){r.add($8(h,e,A,h,t,o))}):SZ(t)&&t.forEach(function(h,u){r.set(u,$8(h,e,A,u,t,o))});var B=c?a?CZ:t4:a?jB:Lc,E=l?void 0:B(t);return gV(E||t,function(h,u){E&&(u=h,h=t[u]),K8(r,u,$8(h,e,A,u,t,o))}),r}var A5=$8;var iyA=1,nyA=4;function oyA(t){return A5(t,iyA|nyA)}var i4=oyA;var ryA=1,syA=4;function ayA(t,e){return e=typeof e=="function"?e:void 0,A5(t,ryA|syA,e)}var RS=ayA;var cyA="__lodash_hash_undefined__";function lyA(t){return this.__data__.set(t,cyA),this}var GZ=lyA;function gyA(t){return this.__data__.has(t)}var UZ=gyA;function e5(t){var e=-1,A=t==null?0:t.length;for(this.__data__=new rC;++e<A;)this.add(t[e])}e5.prototype.add=e5.prototype.push=GZ;e5.prototype.has=UZ;var KZ=e5;function IyA(t,e){for(var A=-1,i=t==null?0:t.length;++A<i;)if(e(t[A],A,t))return!0;return!1}var YZ=IyA;function CyA(t,e){return t.has(e)}var JZ=CyA;var dyA=1,ByA=2;function EyA(t,e,A,i,n,o){var r=A&dyA,s=t.length,a=e.length;if(s!=a&&!(r&&a>s))return!1;var c=o.get(t),l=o.get(e);if(c&&l)return c==e&&l==t;var I=-1,C=!0,d=A&ByA?new KZ:void 0;for(o.set(t,e),o.set(e,t);++I<s;){var B=t[I],E=e[I];if(i)var h=r?i(E,B,I,e,t,o):i(B,E,I,t,e,o);if(h!==void 0){if(h)continue;C=!1;break}if(d){if(!YZ(e,function(u,D){if(!JZ(d,D)&&(B===u||n(B,u,A,i,o)))return d.push(D)})){C=!1;break}}else if(!(B===E||n(B,E,A,i,o))){C=!1;break}}return o.delete(t),o.delete(e),C}var t5=EyA;function QyA(t){var e=-1,A=Array(t.size);return t.forEach(function(i,n){A[++e]=[n,i]}),A}var TZ=QyA;function hyA(t){var e=-1,A=Array(t.size);return t.forEach(function(i){A[++e]=i}),A}var HZ=hyA;var uyA=1,fyA=2,myA="[object Boolean]",pyA="[object Date]",wyA="[object Error]",DyA="[object Map]",yyA="[object Number]",vyA="[object RegExp]",byA="[object Set]",MyA="[object String]",kyA="[object Symbol]",SyA="[object ArrayBuffer]",RyA="[object DataView]",zZ=Zr?Zr.prototype:void 0,xS=zZ?zZ.valueOf:void 0;function xyA(t,e,A,i,n,o,r){switch(A){case RyA:if(t.byteLength!=e.byteLength||t.byteOffset!=e.byteOffset)return!1;t=t.buffer,e=e.buffer;case SyA:return!(t.byteLength!=e.byteLength||!o(new tE(t),new tE(e)));case myA:case pyA:case yyA:return X2(+t,+e);case wyA:return t.name==e.name&&t.message==e.message;case vyA:case MyA:return t==e+"";case DyA:var s=TZ;case byA:var a=i&uyA;if(s||(s=HZ),t.size!=e.size&&!a)return!1;var c=r.get(t);if(c)return c==e;i|=fyA,r.set(t,e);var l=t5(s(t),s(e),i,n,o,r);return r.delete(t),l;case kyA:if(xS)return xS.call(t)==xS.call(e)}return!1}var OZ=xyA;var LyA=1,FyA=Object.prototype,NyA=FyA.hasOwnProperty;function _yA(t,e,A,i,n,o){var r=A&LyA,s=t4(t),a=s.length,c=t4(e),l=c.length;if(a!=l&&!r)return!1;for(var I=a;I--;){var C=s[I];if(!(r?C in e:NyA.call(e,C)))return!1}var d=o.get(t),B=o.get(e);if(d&&B)return d==e&&B==t;var E=!0;o.set(t,e),o.set(e,t);for(var h=r;++I<a;){C=s[I];var u=t[C],D=e[C];if(i)var L=r?i(D,u,C,e,t,o):i(u,D,C,t,e,o);if(!(L===void 0?u===D||n(u,D,A,i,o):L)){E=!1;break}h||(h=C=="constructor")}if(E&&!h){var R=t.constructor,w=e.constructor;R!=w&&"constructor"in t&&"constructor"in e&&!(typeof R=="function"&&R instanceof R&&typeof w=="function"&&w instanceof w)&&(E=!1)}return o.delete(t),o.delete(e),E}var PZ=_yA;var GyA=1,jZ="[object Arguments]",qZ="[object Array]",i5="[object Object]",UyA=Object.prototype,VZ=UyA.hasOwnProperty;function KyA(t,e,A,i,n,o){var r=Gn(t),s=Gn(e),a=r?qZ:hl(t),c=s?qZ:hl(e);a=a==jZ?i5:a,c=c==jZ?i5:c;var l=a==i5,I=c==i5,C=a==c;if(C&&S0(t)){if(!S0(e))return!1;r=!0,l=!1}if(C&&!l)return o||(o=new a1),r||PB(t)?t5(t,e,A,i,n,o):OZ(t,e,a,A,i,n,o);if(!(A&GyA)){var d=l&&VZ.call(t,"__wrapped__"),B=I&&VZ.call(e,"__wrapped__");if(d||B){var E=d?t.value():t,h=B?e.value():e;return o||(o=new a1),n(E,h,A,i,o)}}return C?(o||(o=new a1),PZ(t,e,A,i,n,o)):!1}var ZZ=KyA;function WZ(t,e,A,i,n){return t===e?!0:t==null||e==null||!Vs(t)&&!Vs(e)?t!==t&&e!==e:ZZ(t,e,A,i,WZ,n)}var nE=WZ;var YyA=1,JyA=2;function TyA(t,e,A,i){var n=A.length,o=n,r=!i;if(t==null)return!o;for(t=Object(t);n--;){var s=A[n];if(r&&s[2]?s[1]!==t[s[0]]:!(s[0]in t))return!1}for(;++n<o;){s=A[n];var a=s[0],c=t[a],l=s[1];if(r&&s[2]){if(c===void 0&&!(a in t))return!1}else{var I=new a1;if(i)var C=i(c,l,a,t,e,I);if(!(C===void 0?nE(l,c,YyA|JyA,i,I):C))return!1}}return!0}var XZ=TyA;function HyA(t){return t===t&&!Kr(t)}var n5=HyA;function zyA(t){for(var e=Lc(t),A=e.length;A--;){var i=e[A],n=t[i];e[A]=[i,n,n5(n)]}return e}var $Z=zyA;function OyA(t,e){return function(A){return A==null?!1:A[t]===e&&(e!==void 0||t in Object(A))}}var o5=OyA;function PyA(t){var e=$Z(t);return e.length==1&&e[0][2]?o5(e[0][0],e[0][1]):function(A){return A===t||XZ(A,t,e)}}var AW=PyA;function jyA(t,e){return t!=null&&e in Object(t)}var eW=jyA;function qyA(t,e,A){e=O8(e,t);for(var i=-1,n=e.length,o=!1;++i<n;){var r=s1(e[i]);if(!(o=t!=null&&A(t,r)))break;t=t[r]}return o||++i!=n?o:(n=t==null?0:t.length,!!n&&OB(n)&&zB(r,n)&&(Gn(t)||e1(t)))}var tW=qyA;function VyA(t,e){return t!=null&&tW(t,e,eW)}var iW=VyA;var ZyA=1,WyA=2;function XyA(t,e){return qB(t)&&n5(e)?o5(s1(t),e):function(A){var i=OV(A,t);return i===void 0&&i===e?iW(A,t):nE(e,i,ZyA|WyA)}}var nW=XyA;function $yA(t){return function(e){return e?.[t]}}var oW=$yA;function A7A(t){return function(e){return XB(e,t)}}var rW=A7A;function e7A(t){return qB(t)?oW(s1(t)):rW(t)}var sW=e7A;function t7A(t){return typeof t=="function"?t:t==null?ug:typeof t=="object"?Gn(t)?nW(t[0],t[1]):AW(t):sW(t)}var fg=t7A;function i7A(t,e,A,i){for(var n=-1,o=t==null?0:t.length;++n<o;){var r=t[n];e(i,r,A(r),t)}return i}var aW=i7A;function n7A(t){return function(e,A,i){for(var n=-1,o=Object(e),r=i(e),s=r.length;s--;){var a=r[t?s:++n];if(A(o[a],a,o)===!1)break}return e}}var r5=n7A;var o7A=r5(),cW=o7A;function r7A(t,e){return t&&cW(t,e,Lc)}var s5=r7A;function s7A(t,e){return function(A,i){if(A==null)return A;if(!xc(A))return t(A,i);for(var n=A.length,o=e?n:-1,r=Object(A);(e?o--:++o<n)&&i(r[o],o,r)!==!1;);return A}}var a5=s7A;var a7A=a5(s5),c5=a7A;function c7A(t,e,A,i){return c5(t,function(n,o,r){e(i,n,A(n),r)}),i}var lW=c7A;function l7A(t,e){return function(A,i){var n=Gn(A)?aW:lW,o=e?e():{};return n(A,t,fg(i,2),o)}}var l5=l7A;var g7A=function(){return gr.Date.now()},g5=g7A;var I7A="Expected a function",C7A=Math.max,d7A=Math.min;function B7A(t,e,A){var i,n,o,r,s,a,c=0,l=!1,I=!1,C=!0;if(typeof t!="function")throw new TypeError(I7A);e=$u(e)||0,Kr(A)&&(l=!!A.leading,I="maxWait"in A,o=I?C7A($u(A.maxWait)||0,e):o,C="trailing"in A?!!A.trailing:C);function d(_){var K=i,z=n;return i=n=void 0,c=_,r=t.apply(z,K),r}function B(_){return c=_,s=setTimeout(u,e),l?d(_):r}function E(_){var K=_-a,z=_-c,U=e-K;return I?d7A(U,o-z):U}function h(_){var K=_-a,z=_-c;return a===void 0||K>=e||K<0||I&&z>=o}function u(){var _=g5();if(h(_))return D(_);s=setTimeout(u,E(_))}function D(_){return s=void 0,C&&i?d(_):(i=n=void 0,r)}function L(){s!==void 0&&clearTimeout(s),c=0,i=a=n=s=void 0}function R(){return s===void 0?r:D(g5())}function w(){var _=g5(),K=h(_);if(i=arguments,n=this,a=_,K){if(s===void 0)return B(a);if(I)return clearTimeout(s),s=setTimeout(u,e),d(a)}return s===void 0&&(s=setTimeout(u,e)),r}return w.cancel=L,w.flush=R,w}var oE=B7A;function E7A(t){var e=t==null?0:t.length;return e?t[e-1]:void 0}var hi=E7A;function Q7A(t){return typeof t=="function"?t:ug}var I5=Q7A;function h7A(t,e){for(var A=t==null?0:t.length;A--&&e(t[A],A,t)!==!1;);return t}var gW=h7A;var u7A=r5(!0),IW=u7A;function f7A(t,e){return t&&IW(t,e,Lc)}var CW=f7A;var m7A=a5(CW,!0),dW=m7A;function p7A(t,e){var A=Gn(t)?gW:dW;return A(t,I5(e))}var LS=p7A;function w7A(t){return t&&t.length?t[0]:void 0}var Fc=w7A;function D7A(t,e){var A=-1,i=xc(t)?Array(t.length):[];return c5(t,function(n,o,r){i[++A]=e(n,o,r)}),i}var C5=D7A;function y7A(t,e){var A=Gn(t)?Z2:C5;return A(t,fg(e,3))}var FS=y7A;var v7A=Object.prototype,b7A=v7A.hasOwnProperty,M7A=l5(function(t,e,A){b7A.call(t,A)?t[A].push(e):W2(t,A,[e])}),NS=M7A;function k7A(t){var e=t==null?0:t.length;return e?ZV(t,0,-1):[]}var Fi=k7A;var S7A="[object Map]",R7A="[object Set]",x7A=Object.prototype,L7A=x7A.hasOwnProperty;function F7A(t){if(t==null)return!0;if(xc(t)&&(Gn(t)||typeof t=="string"||typeof t.splice=="function"||S0(t)||PB(t)||e1(t)))return!t.length;var e=hl(t);if(e==S7A||e==R7A)return!t.size;if(A1(t))return!H8(t).length;for(var A in t)if(L7A.call(t,A))return!1;return!0}var Oi=F7A;function N7A(t,e){return nE(t,e)}var di=N7A;function _7A(t,e){return t<e}var BW=_7A;function G7A(t,e){var A={};return e=fg(e,3),s5(t,function(i,n,o){W2(A,n,e(i,n,o))}),A}var _S=G7A;function U7A(t,e,A){for(var i=-1,n=t.length;++i<n;){var o=t[i],r=e(o);if(r!=null&&(s===void 0?r===r&&!ac(r):A(r,s)))var s=r,a=o}return a}var EW=U7A;function K7A(t,e){return t&&t.length?EW(t,fg(e,2),BW):void 0}var rE=K7A;function Y7A(t,e){var A=t.length;for(t.sort(e);A--;)t[A]=t[A].value;return t}var QW=Y7A;function J7A(t,e){if(t!==e){var A=t!==void 0,i=t===null,n=t===t,o=ac(t),r=e!==void 0,s=e===null,a=e===e,c=ac(e);if(!s&&!c&&!o&&t>e||o&&r&&a&&!s&&!c||i&&r&&a||!A&&a||!n)return 1;if(!i&&!o&&!c&&t<e||c&&A&&n&&!i&&!o||s&&A&&n||!r&&n||!a)return-1}return 0}var hW=J7A;function T7A(t,e,A){for(var i=-1,n=t.criteria,o=e.criteria,r=n.length,s=A.length;++i<r;){var a=hW(n[i],o[i]);if(a){if(i>=s)return a;var c=A[i];return a*(c=="desc"?-1:1)}}return t.index-e.index}var uW=T7A;function H7A(t,e,A){e.length?e=Z2(e,function(o){return Gn(o)?function(r){return XB(r,o.length===1?o[0]:o)}:o}):e=[ug];var i=-1;e=Z2(e,t1(fg));var n=C5(t,function(o,r,s){var a=Z2(e,function(c){return c(o)});return{criteria:a,index:++i,value:o}});return QW(n,function(o,r){return uW(o,r,A)})}var fW=H7A;var z7A=l5(function(t,e,A){t[A?0:1].push(e)},function(){return[[],[]]}),GS=z7A;var O7A=Math.ceil,P7A=Math.max;function j7A(t,e,A,i){for(var n=-1,o=P7A(O7A((e-t)/(A||1)),0),r=Array(o);o--;)r[i?o:++n]=t,t+=A;return r}var mW=j7A;function q7A(t){return function(e,A,i){return i&&typeof i!="number"&&A4(e,A,i)&&(A=i=void 0),e=TB(e),A===void 0?(A=e,e=0):A=TB(A),i=i===void 0?e<A?1:-1:TB(i),mW(e,A,i,t)}}var pW=q7A;var V7A=pW(),d5=V7A;var Z7A=dV(function(t,e){if(t==null)return[];var A=e.length;return A>1&&A4(t,e[0],e[1])?e=[]:A>2&&A4(e[0],e[1],e[2])&&(e=[e[0]]),fW(t,VV(e,1),[])}),US=Z7A;var W7A=9007199254740991,KS=4294967295,X7A=Math.min;function $7A(t,e){if(t=Xq(t),t<1||t>W7A)return[];var A=KS,i=X7A(t,KS);e=I5(e),t-=KS;for(var n=Y8(i,e);++A<t;)e(A);return n}var YS=$7A;var AvA=0;function evA(t){var e=++AvA;return z8(t)+e}var c1=evA;var l1=t=>Array.isArray(t),tvA=t=>t!==null&&typeof t=="object"&&!l1(t),ivA=t=>typeof t=="string",aC=(t,e)=>t===e?!0:t!==null&&e!==null&&typeof t=="object"&&typeof e=="object"&&Object.keys(t).length===Object.keys(e).length&&Object.entries(t).every(([A,i])=>aC(i,e[A]));function Er(t){return(...e)=>{let A=e.map(o=>Yr(o)),i=A[0],n=A[1];return A.length===1?o=>t(i(o)):A.length===2?o=>t(i(o),n(o)):o=>t(...A.map(r=>r(o)))}}var r4={boolean:0,number:1,string:2},wW=3,yW=(t,e)=>typeof t==typeof e&&typeof t in r4?t>e:!1,nvA=(t,e)=>aC(t,e)||yW(t,e),vW=(t,e)=>typeof t==typeof e&&typeof t in r4?t<e:!1,ovA=(t,e)=>aC(t,e)||vW(t,e),o4={pipe:(...t)=>{let e=t.map(A=>Yr(A));return A=>e.reduce((i,n)=>n(i),A)},object:t=>{let e=Object.keys(t).map(A=>[A,Yr(t[A])]);return A=>{let i={};for(let[n,o]of e)i[n]=o(A);return i}},array:(...t)=>{let e=t.map(A=>Yr(A));return A=>e.map(i=>i(A))},get:(...t)=>{if(t.length===0)return e=>e??null;if(t.length===1){let e=t[0];return A=>A?.[e]??null}return e=>{let A=e;for(let i of t)A=A?.[i];return A??null}},map:t=>{let e=Yr(t);return A=>A.map(e)},mapObject:t=>{let e=Yr(t);return A=>{let i={};for(let n of Object.keys(A)){let o=e({key:n,value:A[n]});i[o.key]=o.value}return i}},mapKeys:t=>{let e=Yr(t);return A=>{let i={};for(let n of Object.keys(A)){let o=e(n);i[o]=A[n]}return i}},mapValues:t=>{let e=Yr(t);return A=>{let i={};for(let n of Object.keys(A))i[n]=e(A[n]);return i}},filter:t=>{let e=Yr(t);return A=>A.filter(i=>DW(e(i)))},sort:(t=["get"],e)=>{let A=Yr(t),i=e==="desc"?-1:1;function n(o,r){let s=A(o),a=A(r);if(typeof s!=typeof a){let c=r4[typeof s]??wW,l=r4[typeof a]??wW;return c>l?i:c<l?-i:0}return typeof s in r4?s>a?i:s<a?-i:0:0}return o=>o.slice().sort(n)},reverse:()=>t=>t.toReversed(),pick:(...t)=>{let e=t.map(([i,...n])=>[n[n.length-1],o4.get(...n)]),A=(i,n)=>{let o={};for(let[r,s]of n)o[r]=s(i);return o};return i=>l1(i)?i.map(n=>A(n,e)):A(i,e)},groupBy:t=>{let e=Yr(t);return A=>{let i={};for(let n of A){let o=e(n);i[o]?i[o].push(n):i[o]=[n]}return i}},keyBy:t=>{let e=Yr(t);return A=>{let i={};for(let n of A){let o=e(n);o in i||(i[o]=n)}return i}},flatten:()=>t=>t.flat(),join:(t="")=>e=>e.join(t),split:Er((t,e)=>e!==void 0?t.split(e):t.trim().split(/\s+/)),substring:Er((t,e,A)=>t.slice(Math.max(e,0),A)),uniq:()=>t=>{let e=[];for(let A of t)e.findIndex(i=>aC(i,A))===-1&&e.push(A);return e},uniqBy:t=>e=>Object.values(o4.keyBy(t)(e)),limit:t=>e=>e.slice(0,Math.max(t,0)),size:()=>t=>t.length,keys:()=>Object.keys,values:()=>Object.values,prod:()=>t=>n4(t,(e,A)=>e*A),sum:()=>t=>l1(t)?t.reduce((e,A)=>e+A,0):JS(),average:()=>t=>l1(t)?t.length>0?t.reduce((e,A)=>e+A)/t.length:null:JS(),min:()=>t=>n4(t,(e,A)=>Math.min(e,A)),max:()=>t=>n4(t,(e,A)=>Math.max(e,A)),and:Er((...t)=>n4(t,(e,A)=>!!(e&&A))),or:Er((...t)=>n4(t,(e,A)=>!!(e||A))),not:Er(t=>!t),exists:t=>{let e=t.slice(1),A=e.pop(),i=o4.get(...e);return n=>{let o=i(n);return!!o&&Object.hasOwnProperty.call(o,A)}},if:(t,e,A)=>{let i=Yr(t),n=Yr(e),o=Yr(A);return r=>DW(i(r))?n(r):o(r)},in:(t,e)=>{let A=Yr(t),i=Yr(e);return n=>{let o=A(n);return i(n).findIndex(r=>aC(r,o))!==-1}},"not in":(t,e)=>{let A=o4.in(t,e);return i=>!A(i)},regex:(t,e,A)=>{let i=new RegExp(e,A),n=Yr(t);return o=>i.test(n(o))},eq:Er(aC),gt:Er(yW),gte:Er(nvA),lt:Er(vW),lte:Er(ovA),ne:Er((t,e)=>!aC(t,e)),add:Er((t,e)=>t+e),subtract:Er((t,e)=>t-e),multiply:Er((t,e)=>t*e),divide:Er((t,e)=>t/e),mod:Er((t,e)=>t%e),pow:Er((t,e)=>t**e),abs:Er(Math.abs),round:Er((t,e=0)=>+`${Math.round(+`${t}e${e}`)}e${-e}`),number:Er(t=>{let e=Number(t);return Number.isNaN(Number(t))?null:e}),string:Er(String)},DW=t=>t!==null&&t!==0&&t!==!1,n4=(t,e)=>(l1(t)||JS(),t.length===0?null:t.reduce(e)),JS=()=>{TS("Array expected")},TS=t=>{throw new TypeError(t)},B5=[];function Yr(t,e){B5.unshift(rA(rA(rA({},o4),B5[0]),e?.functions));try{let A=l1(t)?rvA(t,B5[0]):tvA(t)?TS(`Function notation ["object", {...}] expected but got ${JSON.stringify(t)}`):()=>t;return i=>{try{return A(i)}catch(n){throw n.jsonquery=[{data:i,query:t},...n.jsonquery??[]],n}}}finally{B5.shift()}}function rvA(t,e){let[A,...i]=t,n=e[A];return n||TS(`Unknown function '${A}'`),n(...i)}var bW=[{pow:"^"},{multiply:"*",divide:"/",mod:"%"},{add:"+",subtract:"-"},{gt:">",gte:">=",lt:"<",lte:"<=",in:"in","not in":"not in"},{eq:"==",ne:"!="},{and:"and"},{or:"or"},{pipe:"|"}],svA=["|","and","or"],MW=["|","and","or","*","/","%","+","-"];function kW(t,e){if(!l1(e))throw new Error("Invalid custom operators");return e.reduce(avA,t)}function avA(t,{name:e,op:A,at:i,after:n,before:o}){if(i)return t.map(a=>Object.values(a).includes(i)?Ye(rA({},a),{[e]:A}):a);let r=n??o,s=t.findIndex(a=>Object.values(a).includes(r));if(s!==-1)return t.toSpliced(s+(n?1:0),0,{[e]:A});throw new Error("Invalid custom operator")}var cvA=/^[a-zA-Z_$][a-zA-Z\d_$]*$/,lvA=/^[a-zA-Z_$][a-zA-Z\d_$]*/,gvA=/^"(?:[^"\\]|\\.)*"/,IvA=/^-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?/,CvA=/^(0|[1-9][0-9]*)/,dvA=/^(true|false|null)/,BvA=/^[ \n\t\r]+/;function HS(t,e){let A=e?.operators??[],i=kW(bW,A),n=Object.assign({},...i),o=svA.concat(A.filter(H=>H.vararg).map(H=>H.op)),r=MW.concat(A.filter(H=>H.leftAssociative).map(H=>H.op)),s=(H=i.length-1)=>{let q=i[H];if(!q)return c();let j=t[z]==="(",gA=s(H-1);for(;;){w();let QA=z,BA=a(q);if(!BA)break;let lA=s(H-1),vA=gA[0],tA=BA===vA&&!j;if(tA&&!r.includes(n[BA])){z=QA;break}gA=tA&&o.includes(n[BA])?[...gA,lA]:[BA,gA,lA]}return gA},a=H=>{let q=Object.keys(H).sort((j,gA)=>gA.length-j.length);for(let j of q){let gA=H[j];if(t.substring(z,z+gA.length)===gA)return z+=gA.length,w(),j}},c=()=>{if(w(),t[z]==="("){z++;let H=s();return _(")"),H}return l()},l=()=>{if(t[z]==="."){let H=[];for(;t[z]===".";)z++,H.push(B()??E()??u()??K("Property expected"));return["get",...H]}return I()},I=()=>{let H=z,q=E();if(w(),!q||t[z]!=="(")return z=H,C();z++,w();let j=t[z]!==")"?[s()]:[];for(;z<t.length&&t[z]!==")";)w(),_(","),j.push(s());return _(")"),[q,...j]},C=()=>{if(t[z]==="{"){z++,w();let H={},q=!0;for(;z<t.length&&t[z]!=="}";){q?q=!1:(_(","),w());let j=B()??E()??u()??K("Key expected");w(),_(":"),H[j]=s()}return _("}"),["object",H]}return d()},d=()=>{if(t[z]==="["){z++,w();let H=[],q=!0;for(;z<t.length&&t[z]!=="]";)q?q=!1:(_(","),w()),H.push(s());return _("]"),["array",...H]}return B()??h()??D()},B=()=>R(gvA,JSON.parse),E=()=>R(lvA,H=>H),h=()=>R(IvA,JSON.parse),u=()=>R(CvA,JSON.parse),D=()=>{let H=R(dvA,JSON.parse);if(H!==void 0)return H;K("Value expected")},L=()=>{w(),z<t.length&&K(`Unexpected part '${t.substring(z)}'`)},R=(H,q)=>{let j=t.substring(z).match(H);if(j)return z+=j[0].length,q(j[0])},w=()=>R(BvA,H=>H),_=H=>{t[z]!==H&&K(`Character '${H}' expected`),z++},K=(H,q=z)=>{throw new SyntaxError(`${H} (pos: ${q})`)},z=0,U=s();return L(),U}var EvA=40,QvA="  ",SW=(t,e)=>{let A=e?.indentation??QvA,i=e?.operators??[],n=kW(bW,i),o=Object.assign({},...n),r=MW.concat(i.filter(d=>d.leftAssociative).map(d=>d.op)),s=(d,B,E=!1)=>l1(d)?a(d,B,E):JSON.stringify(d),a=(d,B,E)=>{let[h,...u]=d;if(h==="get"&&u.length>0)return l(u);if(h==="object")return c(u[0],B);if(h==="array"){let w=u.map(_=>s(_,B));return C(w,["[",", ","]"],[`[
${B+A}`,`,
${B+A}`,`
${B}]`])}let D=o[h];if(D){let w=E?"(":"",_=E?")":"",K=u.map((z,U)=>{let H=z?.[0],q=n.findIndex(QA=>h in QA),j=n.findIndex(QA=>H in QA),gA=q<j||q===j&&U>0||h===H&&!r.includes(D);return s(z,B+A,gA)});return C(K,[w,` ${D} `,_],[w,`
${B+A}${D} `,_])}let L=u.length===1?B:B+A,R=u.map(w=>s(w,L));return C(R,[`${h}(`,", ",")"],u.length===1?[`${h}(`,`,
${B}`,")"]:[`${h}(
${L}`,`,
${L}`,`
${B})`])},c=(d,B)=>{let E=B+A,h=Object.entries(d).map(([u,D])=>`${I(u)}: ${s(D,E)}`);return C(h,["{ ",", "," }"],[`{
${E}`,`,
${E}`,`
${B}}`])},l=d=>d.map(B=>`.${I(B)}`).join(""),I=d=>cvA.test(d)?d:JSON.stringify(d),C=(d,[B,E,h],[u,D,L])=>B.length+d.reduce((R,w)=>R+w.length+E.length,0)-E.length+h.length<=(e?.maxLineLength??EvA)?B+d.join(E)+h:u+d.join(D)+L;return s(t,"")};function RW(t,e,A){return Yr(ivA(e)?HS(e,A):e,A)(t)}var xW={prefix:"far",iconName:"lightbulb",icon:[384,512,[128161],"f0eb","M297.2 248.9C311.6 228.3 320 203.2 320 176c0-70.7-57.3-128-128-128S64 105.3 64 176c0 27.2 8.4 52.3 22.8 72.9c3.7 5.3 8.1 11.3 12.8 17.7c0 0 0 0 0 0c12.9 17.7 28.3 38.9 39.8 59.8c10.4 19 15.7 38.8 18.3 57.5L109 384c-2.2-12-5.9-23.7-11.8-34.5c-9.9-18-22.2-34.9-34.5-51.8c0 0 0 0 0 0s0 0 0 0c-5.2-7.1-10.4-14.2-15.4-21.4C27.6 247.9 16 213.3 16 176C16 78.8 94.8 0 192 0s176 78.8 176 176c0 37.3-11.6 71.9-31.4 100.3c-5 7.2-10.2 14.3-15.4 21.4c0 0 0 0 0 0s0 0 0 0c-12.3 16.8-24.6 33.7-34.5 51.8c-5.9 10.8-9.6 22.5-11.8 34.5l-48.6 0c2.6-18.7 7.9-38.6 18.3-57.5c11.5-20.9 26.9-42.1 39.8-59.8c0 0 0 0 0 0s0 0 0 0s0 0 0 0c4.7-6.4 9-12.4 12.7-17.7zM192 128c-26.5 0-48 21.5-48 48c0 8.8-7.2 16-16 16s-16-7.2-16-16c0-44.2 35.8-80 80-80c8.8 0 16 7.2 16 16s-7.2 16-16 16zm0 384c-44.2 0-80-35.8-80-80l0-16 160 0 0 16c0 44.2-35.8 80-80 80z"]};var hvA={prefix:"far",iconName:"square-check",icon:[448,512,[9745,9989,61510,"check-square"],"f14a","M64 80c-8.8 0-16 7.2-16 16l0 320c0 8.8 7.2 16 16 16l320 0c8.8 0 16-7.2 16-16l0-320c0-8.8-7.2-16-16-16L64 80zM0 96C0 60.7 28.7 32 64 32l320 0c35.3 0 64 28.7 64 64l0 320c0 35.3-28.7 64-64 64L64 480c-35.3 0-64-28.7-64-64L0 96zM337 209L209 337c-9.4 9.4-24.6 9.4-33.9 0l-64-64c-9.4-9.4-9.4-24.6 0-33.9s24.6-9.4 33.9 0l47 47L303 175c9.4-9.4 24.6-9.4 33.9 0s9.4 24.6 0 33.9z"]},zS=hvA;var OS={prefix:"far",iconName:"square",icon:[448,512,[9632,9723,9724,61590],"f0c8","M384 80c8.8 0 16 7.2 16 16l0 320c0 8.8-7.2 16-16 16L64 432c-8.8 0-16-7.2-16-16L48 96c0-8.8 7.2-16 16-16l320 0zM64 32C28.7 32 0 60.7 0 96L0 416c0 35.3 28.7 64 64 64l320 0c35.3 0 64-28.7 64-64l0-320c0-35.3-28.7-64-64-64L64 32z"]};var LW={prefix:"far",iconName:"clock",icon:[512,512,[128339,"clock-four"],"f017","M464 256A208 208 0 1 1 48 256a208 208 0 1 1 416 0zM0 256a256 256 0 1 0 512 0A256 256 0 1 0 0 256zM232 120l0 136c0 8 4 15.5 10.7 20l96 64c11 7.4 25.9 4.4 33.3-6.7s4.4-25.9-6.7-33.3L280 243.2 280 120c0-13.3-10.7-24-24-24s-24 10.7-24 24z"]};var E5={prefix:"fas",iconName:"trash-can",icon:[448,512,[61460,"trash-alt"],"f2ed","M135.2 17.7C140.6 6.8 151.7 0 163.8 0L284.2 0c12.1 0 23.2 6.8 28.6 17.7L320 32l96 0c17.7 0 32 14.3 32 32s-14.3 32-32 32L32 96C14.3 96 0 81.7 0 64S14.3 32 32 32l96 0 7.2-14.3zM32 128l384 0 0 320c0 35.3-28.7 64-64 64L96 512c-35.3 0-64-28.7-64-64l0-320zm96 64c-8.8 0-16 7.2-16 16l0 224c0 8.8 7.2 16 16 16s16-7.2 16-16l0-224c0-8.8-7.2-16-16-16zm96 0c-8.8 0-16 7.2-16 16l0 224c0 8.8 7.2 16 16 16s16-7.2 16-16l0-224c0-8.8-7.2-16-16-16zm96 0c-8.8 0-16 7.2-16 16l0 224c0 8.8 7.2 16 16 16s16-7.2 16-16l0-224c0-8.8-7.2-16-16-16z"]};var FW={prefix:"fas",iconName:"down-left-and-up-right-to-center",icon:[512,512,["compress-alt"],"f422","M439 7c9.4-9.4 24.6-9.4 33.9 0l32 32c9.4 9.4 9.4 24.6 0 33.9l-87 87 39 39c6.9 6.9 8.9 17.2 5.2 26.2s-12.5 14.8-22.2 14.8l-144 0c-13.3 0-24-10.7-24-24l0-144c0-9.7 5.8-18.5 14.8-22.2s19.3-1.7 26.2 5.2l39 39L439 7zM72 272l144 0c13.3 0 24 10.7 24 24l0 144c0 9.7-5.8 18.5-14.8 22.2s-19.3 1.7-26.2-5.2l-39-39L73 505c-9.4 9.4-24.6 9.4-33.9 0L7 473c-9.4-9.4-9.4-24.6 0-33.9l87-87L55 313c-6.9-6.9-8.9-17.2-5.2-26.2s12.5-14.8 22.2-14.8z"]};var sE={prefix:"fas",iconName:"caret-right",icon:[256,512,[],"f0da","M246.6 278.6c12.5-12.5 12.5-32.8 0-45.3l-128-128c-9.2-9.2-22.9-11.9-34.9-6.9s-19.8 16.6-19.8 29.6l0 256c0 12.9 7.8 24.6 19.8 29.6s25.7 2.2 34.9-6.9l128-128z"]};var PS={prefix:"fas",iconName:"paste",icon:[512,512,["file-clipboard"],"f0ea","M160 0c-23.7 0-44.4 12.9-55.4 32L48 32C21.5 32 0 53.5 0 80L0 400c0 26.5 21.5 48 48 48l144 0 0-272c0-44.2 35.8-80 80-80l48 0 0-16c0-26.5-21.5-48-48-48l-56.6 0C204.4 12.9 183.7 0 160 0zM272 128c-26.5 0-48 21.5-48 48l0 272 0 16c0 26.5 21.5 48 48 48l192 0c26.5 0 48-21.5 48-48l0-220.1c0-12.7-5.1-24.9-14.1-33.9l-67.9-67.9c-9-9-21.2-14.1-33.9-14.1L320 128l-48 0zM160 40a24 24 0 1 1 0 48 24 24 0 1 1 0-48z"]};var NW={prefix:"fas",iconName:"circle-notch",icon:[512,512,[],"f1ce","M222.7 32.1c5 16.9-4.6 34.8-21.5 39.8C121.8 95.6 64 169.1 64 256c0 106 86 192 192 192s192-86 192-192c0-86.9-57.8-160.4-137.1-184.1c-16.9-5-26.6-22.9-21.5-39.8s22.9-26.6 39.8-21.5C434.9 42.1 512 140 512 256c0 141.4-114.6 256-256 256S0 397.4 0 256C0 140 77.1 42.1 182.9 10.6c16.9-5 34.8 4.6 39.8 21.5z"]};var uvA={prefix:"fas",iconName:"scissors",icon:[512,512,[9984,9986,9988,"cut"],"f0c4","M256 192l-39.5-39.5c4.9-12.6 7.5-26.2 7.5-40.5C224 50.1 173.9 0 112 0S0 50.1 0 112s50.1 112 112 112c14.3 0 27.9-2.7 40.5-7.5L192 256l-39.5 39.5c-12.6-4.9-26.2-7.5-40.5-7.5C50.1 288 0 338.1 0 400s50.1 112 112 112s112-50.1 112-112c0-14.3-2.7-27.9-7.5-40.5L499.2 76.8c7.1-7.1 7.1-18.5 0-25.6c-28.3-28.3-74.1-28.3-102.4 0L256 192zm22.6 150.6L396.8 460.8c28.3 28.3 74.1 28.3 102.4 0c7.1-7.1 7.1-18.5 0-25.6L342.6 278.6l-64 64zM64 112a48 48 0 1 1 96 0 48 48 0 1 1 -96 0zm48 240a48 48 0 1 1 0 96 48 48 0 1 1 0-96z"]},cC=uvA;var fvA={prefix:"fas",iconName:"square-caret-down",icon:[448,512,["caret-square-down"],"f150","M384 480c35.3 0 64-28.7 64-64l0-320c0-35.3-28.7-64-64-64L64 32C28.7 32 0 60.7 0 96L0 416c0 35.3 28.7 64 64 64l320 0zM224 352c-6.7 0-13-2.8-17.6-7.7l-104-112c-6.5-7-8.2-17.2-4.4-25.9s12.5-14.4 22-14.4l208 0c9.5 0 18.2 5.7 22 14.4s2.1 18.9-4.4 25.9l-104 112c-4.5 4.9-10.9 7.7-17.6 7.7z"]},_W=fvA;var GW={prefix:"fas",iconName:"caret-left",icon:[256,512,[],"f0d9","M9.4 278.6c-12.5-12.5-12.5-32.8 0-45.3l128-128c9.2-9.2 22.9-11.9 34.9-6.9s19.8 16.6 19.8 29.6l0 256c0 12.9-7.8 24.6-19.8 29.6s-25.7 2.2-34.9-6.9l-128-128z"]};var mvA={prefix:"fas",iconName:"square-check",icon:[448,512,[9745,9989,61510,"check-square"],"f14a","M64 32C28.7 32 0 60.7 0 96L0 416c0 35.3 28.7 64 64 64l320 0c35.3 0 64-28.7 64-64l0-320c0-35.3-28.7-64-64-64L64 32zM337 209L209 337c-9.4 9.4-24.6 9.4-33.9 0l-64-64c-9.4-9.4-9.4-24.6 0-33.9s24.6-9.4 33.9 0l47 47L303 175c9.4-9.4 24.6-9.4 33.9 0s9.4 24.6 0 33.9z"]},jS=mvA;var pvA={prefix:"fas",iconName:"pen-to-square",icon:[512,512,["edit"],"f044","M471.6 21.7c-21.9-21.9-57.3-21.9-79.2 0L362.3 51.7l97.9 97.9 30.1-30.1c21.9-21.9 21.9-57.3 0-79.2L471.6 21.7zm-299.2 220c-6.1 6.1-10.8 13.6-13.5 21.9l-29.6 88.8c-2.9 8.6-.6 18.1 5.8 24.6s15.9 8.7 24.6 5.8l88.8-29.6c8.2-2.7 15.7-7.4 21.9-13.5L437.7 172.3 339.7 74.3 172.4 241.7zM96 64C43 64 0 107 0 160L0 416c0 53 43 96 96 96l256 0c53 0 96-43 96-96l0-96c0-17.7-14.3-32-32-32s-32 14.3-32 32l0 96c0 17.7-14.3 32-32 32L96 448c-17.7 0-32-14.3-32-32l0-256c0-17.7 14.3-32 32-32l96 0c17.7 0 32-14.3 32-32s-14.3-32-32-32L96 64z"]},UW=pvA;var KW={prefix:"fas",iconName:"chevron-up",icon:[512,512,[],"f077","M233.4 105.4c12.5-12.5 32.8-12.5 45.3 0l192 192c12.5 12.5 12.5 32.8 0 45.3s-32.8 12.5-45.3 0L256 173.3 86.6 342.6c-12.5 12.5-32.8 12.5-45.3 0s-12.5-32.8 0-45.3l192-192z"]};var qS={prefix:"fas",iconName:"angle-right",icon:[320,512,[8250],"f105","M278.6 233.4c12.5 12.5 12.5 32.8 0 45.3l-160 160c-12.5 12.5-32.8 12.5-45.3 0s-12.5-32.8 0-45.3L210.7 256 73.4 118.6c-12.5-12.5-12.5-32.8 0-45.3s32.8-12.5 45.3 0l160 160z"]};var wvA={prefix:"fas",iconName:"square-caret-up",icon:[448,512,["caret-square-up"],"f151","M64 32C28.7 32 0 60.7 0 96L0 416c0 35.3 28.7 64 64 64l320 0c35.3 0 64-28.7 64-64l0-320c0-35.3-28.7-64-64-64L64 32zM224 160c6.7 0 13 2.8 17.6 7.7l104 112c6.5 7 8.2 17.2 4.4 25.9s-12.5 14.4-22 14.4l-208 0c-9.5 0-18.2-5.7-22-14.4s-2.1-18.9 4.4-25.9l104-112c4.5-4.9 10.9-7.7 17.6-7.7z"]},YW=wvA;var VS={prefix:"fas",iconName:"caret-up",icon:[320,512,[],"f0d8","M182.6 137.4c-12.5-12.5-32.8-12.5-45.3 0l-128 128c-9.2 9.2-11.9 22.9-6.9 34.9s16.6 19.8 29.6 19.8l256 0c12.9 0 24.6-7.8 29.6-19.8s2.2-25.7-6.9-34.9l-128-128z"]};var ZS={prefix:"fas",iconName:"square",icon:[448,512,[9632,9723,9724,61590],"f0c8","M0 96C0 60.7 28.7 32 64 32H384c35.3 0 64 28.7 64 64V416c0 35.3-28.7 64-64 64H64c-35.3 0-64-28.7-64-64V96z"]};var s4={prefix:"fas",iconName:"filter",icon:[512,512,[],"f0b0","M3.9 54.9C10.5 40.9 24.5 32 40 32l432 0c15.5 0 29.5 8.9 36.1 22.9s4.6 30.5-5.2 42.5L320 320.9 320 448c0 12.1-6.8 23.2-17.7 28.6s-23.8 4.3-33.5-3l-64-48c-8.1-6-12.8-15.5-12.8-25.6l0-79.1L9 97.3C-.7 85.4-2.8 68.8 3.9 54.9z"]};var a4={prefix:"fas",iconName:"code",icon:[640,512,[],"f121","M392.8 1.2c-17-4.9-34.7 5-39.6 22l-128 448c-4.9 17 5 34.7 22 39.6s34.7-5 39.6-22l128-448c4.9-17-5-34.7-22-39.6zm80.6 120.1c-12.5 12.5-12.5 32.8 0 45.3L562.7 256l-89.4 89.4c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0l112-112c12.5-12.5 12.5-32.8 0-45.3l-112-112c-12.5-12.5-32.8-12.5-45.3 0zm-306.7 0c-12.5-12.5-32.8-12.5-45.3 0l-112 112c-12.5 12.5-12.5 32.8 0 45.3l112 112c12.5 12.5 32.8 12.5 45.3 0s12.5-32.8 0-45.3L77.3 256l89.4-89.4c12.5-12.5 12.5-32.8 0-45.3z"]};var L0={prefix:"fas",iconName:"wrench",icon:[512,512,[128295],"f0ad","M352 320c88.4 0 160-71.6 160-160c0-15.3-2.2-30.1-6.2-44.2c-3.1-10.8-16.4-13.2-24.3-5.3l-76.8 76.8c-3 3-7.1 4.7-11.3 4.7L336 192c-8.8 0-16-7.2-16-16l0-57.4c0-4.2 1.7-8.3 4.7-11.3l76.8-76.8c7.9-7.9 5.4-21.2-5.3-24.3C382.1 2.2 367.3 0 352 0C263.6 0 192 71.6 192 160c0 19.1 3.4 37.5 9.5 54.5L19.9 396.1C7.2 408.8 0 426.1 0 444.1C0 481.6 30.4 512 67.9 512c18 0 35.3-7.2 48-19.9L297.5 310.5c17 6.2 35.4 9.5 54.5 9.5zM80 408a24 24 0 1 1 0 48 24 24 0 1 1 0-48z"]};var JW={prefix:"fas",iconName:"eye",icon:[576,512,[128065],"f06e","M288 32c-80.8 0-145.5 36.8-192.6 80.6C48.6 156 17.3 208 2.5 243.7c-3.3 7.9-3.3 16.7 0 24.6C17.3 304 48.6 356 95.4 399.4C142.5 443.2 207.2 480 288 480s145.5-36.8 192.6-80.6c46.8-43.5 78.1-95.4 93-131.1c3.3-7.9 3.3-16.7 0-24.6c-14.9-35.7-46.2-87.7-93-131.1C433.5 68.8 368.8 32 288 32zM144 256a144 144 0 1 1 288 0 144 144 0 1 1 -288 0zm144-64c0 35.3-28.7 64-64 64c-7.1 0-13.9-1.2-20.3-3.3c-5.5-1.8-11.9 1.6-11.7 7.4c.3 6.9 1.3 13.8 3.2 20.7c13.7 51.2 66.4 81.6 117.6 67.9s81.6-66.4 67.9-117.6c-11.1-41.5-47.8-69.4-88.6-71.1c-5.8-.2-9.2 6.1-7.4 11.7c2.1 6.4 3.3 13.2 3.3 20.3z"]};var lC={prefix:"fas",iconName:"pen",icon:[512,512,[128394],"f304","M362.7 19.3L314.3 67.7 444.3 197.7l48.4-48.4c25-25 25-65.5 0-90.5L453.3 19.3c-25-25-65.5-25-90.5 0zm-71 71L58.6 323.5c-10.4 10.4-18 23.3-22.2 37.4L1 481.2C-1.5 489.7 .8 498.8 7 505s15.3 8.5 23.7 6.1l120.3-35.4c14.1-4.2 27-11.8 37.4-22.2L421.7 220.3 291.7 90.3z"]};var DvA={prefix:"fas",iconName:"arrow-rotate-right",icon:[512,512,[8635,"arrow-right-rotate","arrow-rotate-forward","redo"],"f01e","M386.3 160L336 160c-17.7 0-32 14.3-32 32s14.3 32 32 32l128 0c17.7 0 32-14.3 32-32l0-128c0-17.7-14.3-32-32-32s-32 14.3-32 32l0 51.2L414.4 97.6c-87.5-87.5-229.3-87.5-316.8 0s-87.5 229.3 0 316.8s229.3 87.5 316.8 0c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0c-62.5 62.5-163.8 62.5-226.3 0s-62.5-163.8 0-226.3s163.8-62.5 226.3 0L386.3 160z"]};var Q5=DvA;var yvA={prefix:"fas",iconName:"arrow-rotate-left",icon:[512,512,[8634,"arrow-left-rotate","arrow-rotate-back","arrow-rotate-backward","undo"],"f0e2","M125.7 160l50.3 0c17.7 0 32 14.3 32 32s-14.3 32-32 32L48 224c-17.7 0-32-14.3-32-32L16 64c0-17.7 14.3-32 32-32s32 14.3 32 32l0 51.2L97.6 97.6c87.5-87.5 229.3-87.5 316.8 0s87.5 229.3 0 316.8s-229.3 87.5-316.8 0c-12.5-12.5-12.5-32.8 0-45.3s32.8-12.5 45.3 0c62.5 62.5 163.8 62.5 226.3 0s62.5-163.8 0-226.3s-163.8-62.5-226.3 0L125.7 160z"]};var h5=yvA;var vvA={prefix:"fas",iconName:"crop-simple",icon:[512,512,["crop-alt"],"f565","M128 32c0-17.7-14.3-32-32-32S64 14.3 64 32l0 32L32 64C14.3 64 0 78.3 0 96s14.3 32 32 32l32 0 0 256c0 35.3 28.7 64 64 64l224 0 0-64-224 0 0-352zM384 480c0 17.7 14.3 32 32 32s32-14.3 32-32l0-32 32 0c17.7 0 32-14.3 32-32s-14.3-32-32-32l-32 0 0-256c0-35.3-28.7-64-64-64L160 64l0 64 224 0 0 352z"]},TW=vvA;var bvA={prefix:"fas",iconName:"gear",icon:[512,512,[9881,"cog"],"f013","M495.9 166.6c3.2 8.7 .5 18.4-6.4 24.6l-43.3 39.4c1.1 8.3 1.7 16.8 1.7 25.4s-.6 17.1-1.7 25.4l43.3 39.4c6.9 6.2 9.6 15.9 6.4 24.6c-4.4 11.9-9.7 23.3-15.8 34.3l-4.7 8.1c-6.6 11-14 21.4-22.1 31.2c-5.9 7.2-15.7 9.6-24.5 6.8l-55.7-17.7c-13.4 10.3-28.2 18.9-44 25.4l-12.5 57.1c-2 9.1-9 16.3-18.2 17.8c-13.8 2.3-28 3.5-42.5 3.5s-28.7-1.2-42.5-3.5c-9.2-1.5-16.2-8.7-18.2-17.8l-12.5-57.1c-15.8-6.5-30.6-15.1-44-25.4L83.1 425.9c-8.8 2.8-18.6 .3-24.5-6.8c-8.1-9.8-15.5-20.2-22.1-31.2l-4.7-8.1c-6.1-11-11.4-22.4-15.8-34.3c-3.2-8.7-.5-18.4 6.4-24.6l43.3-39.4C64.6 273.1 64 264.6 64 256s.6-17.1 1.7-25.4L22.4 191.2c-6.9-6.2-9.6-15.9-6.4-24.6c4.4-11.9 9.7-23.3 15.8-34.3l4.7-8.1c6.6-11 14-21.4 22.1-31.2c5.9-7.2 15.7-9.6 24.5-6.8l55.7 17.7c13.4-10.3 28.2-18.9 44-25.4l12.5-57.1c2-9.1 9-16.3 18.2-17.8C227.3 1.2 241.5 0 256 0s28.7 1.2 42.5 3.5c9.2 1.5 16.2 8.7 18.2 17.8l12.5 57.1c15.8 6.5 30.6 15.1 44 25.4l55.7-17.7c8.8-2.8 18.6-.3 24.5 6.8c8.1 9.8 15.5 20.2 22.1 31.2l4.7 8.1c6.1 11 11.4 22.4 15.8 34.3zM256 336a80 80 0 1 0 0-160 80 80 0 1 0 0 160z"]},HW=bvA;var mg={prefix:"fas",iconName:"caret-down",icon:[320,512,[],"f0d7","M137.4 374.6c12.5 12.5 32.8 12.5 45.3 0l128-128c9.2-9.2 11.9-22.9 6.9-34.9s-16.6-19.8-29.6-19.8L32 192c-12.9 0-24.6 7.8-29.6 19.8s-2.2 25.7 6.9 34.9l128 128z"]};var MvA={prefix:"fas",iconName:"ellipsis-vertical",icon:[128,512,["ellipsis-v"],"f142","M64 360a56 56 0 1 0 0 112 56 56 0 1 0 0-112zm0-160a56 56 0 1 0 0 112 56 56 0 1 0 0-112zM120 96A56 56 0 1 0 8 96a56 56 0 1 0 112 0z"]},WS=MvA;var c4={prefix:"fas",iconName:"arrow-right-arrow-left",icon:[448,512,[8644,"exchange"],"f0ec","M438.6 150.6c12.5-12.5 12.5-32.8 0-45.3l-96-96c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3L338.7 96 32 96C14.3 96 0 110.3 0 128s14.3 32 32 32l306.7 0-41.4 41.4c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0l96-96zm-333.3 352c12.5 12.5 32.8 12.5 45.3 0s12.5-32.8 0-45.3L109.3 416 416 416c17.7 0 32-14.3 32-32s-14.3-32-32-32l-306.7 0 41.4-41.4c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0l-96 96c-12.5 12.5-12.5 32.8 0 45.3l96 96z"]};var kvA={prefix:"fas",iconName:"arrow-down-short-wide",icon:[576,512,["sort-amount-desc","sort-amount-down-alt"],"f884","M151.6 469.6C145.5 476.2 137 480 128 480s-17.5-3.8-23.6-10.4l-88-96c-11.9-13-11.1-33.3 2-45.2s33.3-11.1 45.2 2L96 365.7 96 64c0-17.7 14.3-32 32-32s32 14.3 32 32l0 301.7 32.4-35.4c11.9-13 32.2-13.9 45.2-2s13.9 32.2 2 45.2l-88 96zM320 32l32 0c17.7 0 32 14.3 32 32s-14.3 32-32 32l-32 0c-17.7 0-32-14.3-32-32s14.3-32 32-32zm0 128l96 0c17.7 0 32 14.3 32 32s-14.3 32-32 32l-96 0c-17.7 0-32-14.3-32-32s14.3-32 32-32zm0 128l160 0c17.7 0 32 14.3 32 32s-14.3 32-32 32l-160 0c-17.7 0-32-14.3-32-32s14.3-32 32-32zm0 128l224 0c17.7 0 32 14.3 32 32s-14.3 32-32 32l-224 0c-17.7 0-32-14.3-32-32s14.3-32 32-32z"]};var l4=kvA;var zW={prefix:"fas",iconName:"angle-down",icon:[448,512,[8964],"f107","M201.4 374.6c12.5 12.5 32.8 12.5 45.3 0l160-160c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L224 306.7 86.6 169.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3l160 160z"]};var XS={prefix:"fas",iconName:"arrow-down",icon:[384,512,[8595],"f063","M169.4 470.6c12.5 12.5 32.8 12.5 45.3 0l160-160c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L224 370.8 224 64c0-17.7-14.3-32-32-32s-32 14.3-32 32l0 306.7L54.6 265.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3l160 160z"]};var SvA={prefix:"fas",iconName:"magnifying-glass",icon:[512,512,[128269,"search"],"f002","M416 208c0 45.9-14.9 88.3-40 122.7L502.6 457.4c12.5 12.5 12.5 32.8 0 45.3s-32.8 12.5-45.3 0L330.7 376c-34.4 25.2-76.8 40-122.7 40C93.1 416 0 322.9 0 208S93.1 0 208 0S416 93.1 416 208zM208 352a144 144 0 1 0 0-288 144 144 0 1 0 0 288z"]},g4=SvA;var OW={prefix:"fas",iconName:"chevron-down",icon:[512,512,[],"f078","M233.4 406.6c12.5 12.5 32.8 12.5 45.3 0l192-192c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L256 338.7 86.6 169.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3l192 192z"]};var F0={prefix:"fas",iconName:"copy",icon:[448,512,[],"f0c5","M208 0L332.1 0c12.7 0 24.9 5.1 33.9 14.1l67.9 67.9c9 9 14.1 21.2 14.1 33.9L448 336c0 26.5-21.5 48-48 48l-192 0c-26.5 0-48-21.5-48-48l0-288c0-26.5 21.5-48 48-48zM48 128l80 0 0 64-64 0 0 256 192 0 0-32 64 0 0 48c0 26.5-21.5 48-48 48L48 512c-26.5 0-48-21.5-48-48L0 176c0-26.5 21.5-48 48-48z"]};var gC={prefix:"fas",iconName:"plus",icon:[448,512,[10133,61543,"add"],"2b","M256 80c0-17.7-14.3-32-32-32s-32 14.3-32 32l0 144L48 224c-17.7 0-32 14.3-32 32s14.3 32 32 32l144 0 0 144c0 17.7 14.3 32 32 32s32-14.3 32-32l0-144 144 0c17.7 0 32-14.3 32-32s-14.3-32-32-32l-144 0 0-144z"]};var PW={prefix:"fas",iconName:"xmark",icon:[384,512,[128473,10005,10006,10060,215,"close","multiply","remove","times"],"f00d","M342.6 150.6c12.5-12.5 12.5-32.8 0-45.3s-32.8-12.5-45.3 0L192 210.7 86.6 105.4c-12.5-12.5-32.8-12.5-45.3 0s-12.5 32.8 0 45.3L146.7 256 41.4 361.4c-12.5 12.5-12.5 32.8 0 45.3s32.8 12.5 45.3 0L192 301.3 297.4 406.6c12.5 12.5 32.8 12.5 45.3 0s12.5-32.8 0-45.3L237.3 256 342.6 150.6z"]},jW=PW;var I4=PW;var qW={prefix:"fas",iconName:"rotate",icon:[512,512,[128260,"sync-alt"],"f2f1","M142.9 142.9c-17.5 17.5-30.1 38-37.8 59.8c-5.9 16.7-24.2 25.4-40.8 19.5s-25.4-24.2-19.5-40.8C55.6 150.7 73.2 122 97.6 97.6c87.2-87.2 228.3-87.5 315.8-1L455 55c6.9-6.9 17.2-8.9 26.2-5.2s14.8 12.5 14.8 22.2l0 128c0 13.3-10.7 24-24 24l-8.4 0c0 0 0 0 0 0L344 224c-9.7 0-18.5-5.8-22.2-14.8s-1.7-19.3 5.2-26.2l41.1-41.1c-62.6-61.5-163.1-61.2-225.3 1zM16 312c0-13.3 10.7-24 24-24l7.6 0 .7 0L168 288c9.7 0 18.5 5.8 22.2 14.8s1.7 19.3-5.2 26.2l-41.1 41.1c62.6 61.5 163.1 61.2 225.3-1c17.5-17.5 30.1-38 37.8-59.8c5.9-16.7 24.2-25.4 40.8-19.5s25.4 24.2 19.5 40.8c-10.8 30.6-28.4 59.3-52.9 83.8c-87.2 87.2-228.3 87.5-315.8 1L57 457c-6.9 6.9-17.2 8.9-26.2 5.2S16 449.7 16 440l0-119.6 0-.7 0-7.6z"]};var VW={prefix:"fas",iconName:"up-right-and-down-left-from-center",icon:[512,512,["expand-alt"],"f424","M344 0L488 0c13.3 0 24 10.7 24 24l0 144c0 9.7-5.8 18.5-14.8 22.2s-19.3 1.7-26.2-5.2l-39-39-87 87c-9.4 9.4-24.6 9.4-33.9 0l-32-32c-9.4-9.4-9.4-24.6 0-33.9l87-87L327 41c-6.9-6.9-8.9-17.2-5.2-26.2S334.3 0 344 0zM168 512L24 512c-13.3 0-24-10.7-24-24L0 344c0-9.7 5.8-18.5 14.8-22.2s19.3-1.7 26.2 5.2l39 39 87-87c9.4-9.4 24.6-9.4 33.9 0l32 32c9.4 9.4 9.4 24.6 0 33.9l-87 87 39 39c6.9 6.9 8.9 17.2 5.2 26.2s-12.5 14.8-22.2 14.8z"]};var $S={prefix:"fas",iconName:"clone",icon:[512,512,[],"f24d","M288 448L64 448l0-224 64 0 0-64-64 0c-35.3 0-64 28.7-64 64L0 448c0 35.3 28.7 64 64 64l224 0c35.3 0 64-28.7 64-64l0-64-64 0 0 64zm-64-96l224 0c35.3 0 64-28.7 64-64l0-224c0-35.3-28.7-64-64-64L224 0c-35.3 0-64 28.7-64 64l0 224c0 35.3 28.7 64 64 64z"]};var u5={prefix:"fas",iconName:"check",icon:[448,512,[10003,10004],"f00c","M438.6 105.4c12.5 12.5 12.5 32.8 0 45.3l-256 256c-12.5 12.5-32.8 12.5-45.3 0l-128-128c-12.5-12.5-12.5-32.8 0-45.3s32.8-12.5 45.3 0L160 338.7 393.4 105.4c12.5-12.5 32.8-12.5 45.3 0z"]};var RvA={prefix:"fas",iconName:"triangle-exclamation",icon:[512,512,[9888,"exclamation-triangle","warning"],"f071","M256 32c14.2 0 27.3 7.5 34.5 19.8l216 368c7.3 12.4 7.3 27.7 .2 40.1S486.3 480 472 480L40 480c-14.3 0-27.6-7.7-34.7-20.1s-7-27.8 .2-40.1l216-368C228.7 39.5 241.8 32 256 32zm0 128c-13.3 0-24 10.7-24 24l0 112c0 13.3 10.7 24 24 24s24-10.7 24-24l0-112c0-13.3-10.7-24-24-24zm32 224a32 32 0 1 0 -64 0 32 32 0 1 0 64 0z"]},g1=RvA;var OrA=jQ(XW(),1);var $W=Number.isNaN||function(e){return typeof e=="number"&&e!==e};function xvA(t,e){return!!(t===e||$W(t)&&$W(e))}function LvA(t,e){if(t.length!==e.length)return!1;for(var A=0;A<t.length;A++)if(!xvA(t[A],e[A]))return!1;return!0}function aE(t,e){e===void 0&&(e=LvA);var A=null;function i(){for(var n=[],o=0;o<arguments.length;o++)n[o]=arguments[o];if(A&&A.lastThis===this&&e(n,A.lastArgs))return A.lastResult;var r=t.apply(this,n);return A={lastResult:r,lastArgs:n,lastThis:this},r}return i.clear=function(){A=null},i}var H_A=jQ(_AA(),1),z_A=jQ(GAA(),1);var Kx=class{add(e,A,i){if(typeof arguments[0]!="string")for(let n in arguments[0])this.add(n,arguments[0][n],arguments[1]);else(Array.isArray(e)?e:[e]).forEach(function(n){this[n]=this[n]||[],A&&this[n][i?"unshift":"push"](A)},this)}run(e,A){this[e]=this[e]||[],this[e].forEach(function(i){i.call(A&&A.context?A.context:A,A)})}},Yx=class{constructor(e){this.jsep=e,this.registered={}}register(){for(var e=arguments.length,A=new Array(e),i=0;i<e;i++)A[i]=arguments[i];A.forEach(n=>{if(typeof n!="object"||!n.name||!n.init)throw new Error("Invalid JSEP plugin format");this.registered[n.name]||(n.init(this.jsep),this.registered[n.name]=n)})}},Ra=class t{static get version(){return"1.4.0"}static toString(){return"JavaScript Expression Parser (JSEP) v"+t.version}static addUnaryOp(e){return t.max_unop_len=Math.max(e.length,t.max_unop_len),t.unary_ops[e]=1,t}static addBinaryOp(e,A,i){return t.max_binop_len=Math.max(e.length,t.max_binop_len),t.binary_ops[e]=A,i?t.right_associative.add(e):t.right_associative.delete(e),t}static addIdentifierChar(e){return t.additional_identifier_chars.add(e),t}static addLiteral(e,A){return t.literals[e]=A,t}static removeUnaryOp(e){return delete t.unary_ops[e],e.length===t.max_unop_len&&(t.max_unop_len=t.getMaxKeyLen(t.unary_ops)),t}static removeAllUnaryOps(){return t.unary_ops={},t.max_unop_len=0,t}static removeIdentifierChar(e){return t.additional_identifier_chars.delete(e),t}static removeBinaryOp(e){return delete t.binary_ops[e],e.length===t.max_binop_len&&(t.max_binop_len=t.getMaxKeyLen(t.binary_ops)),t.right_associative.delete(e),t}static removeAllBinaryOps(){return t.binary_ops={},t.max_binop_len=0,t}static removeLiteral(e){return delete t.literals[e],t}static removeAllLiterals(){return t.literals={},t}get char(){return this.expr.charAt(this.index)}get code(){return this.expr.charCodeAt(this.index)}constructor(e){this.expr=e,this.index=0}static parse(e){return new t(e).parse()}static getMaxKeyLen(e){return Math.max(0,...Object.keys(e).map(A=>A.length))}static isDecimalDigit(e){return e>=48&&e<=57}static binaryPrecedence(e){return t.binary_ops[e]||0}static isIdentifierStart(e){return e>=65&&e<=90||e>=97&&e<=122||e>=128&&!t.binary_ops[String.fromCharCode(e)]||t.additional_identifier_chars.has(String.fromCharCode(e))}static isIdentifierPart(e){return t.isIdentifierStart(e)||t.isDecimalDigit(e)}throwError(e){let A=new Error(e+" at character "+this.index);throw A.index=this.index,A.description=e,A}runHook(e,A){if(t.hooks[e]){let i={context:this,node:A};return t.hooks.run(e,i),i.node}return A}searchHook(e){if(t.hooks[e]){let A={context:this};return t.hooks[e].find(function(i){return i.call(A.context,A),A.node}),A.node}}gobbleSpaces(){let e=this.code;for(;e===t.SPACE_CODE||e===t.TAB_CODE||e===t.LF_CODE||e===t.CR_CODE;)e=this.expr.charCodeAt(++this.index);this.runHook("gobble-spaces")}parse(){this.runHook("before-all");let e=this.gobbleExpressions(),A=e.length===1?e[0]:{type:t.COMPOUND,body:e};return this.runHook("after-all",A)}gobbleExpressions(e){let A=[],i,n;for(;this.index<this.expr.length;)if(i=this.code,i===t.SEMCOL_CODE||i===t.COMMA_CODE)this.index++;else if(n=this.gobbleExpression())A.push(n);else if(this.index<this.expr.length){if(i===e)break;this.throwError('Unexpected "'+this.char+'"')}return A}gobbleExpression(){let e=this.searchHook("gobble-expression")||this.gobbleBinaryExpression();return this.gobbleSpaces(),this.runHook("after-expression",e)}gobbleBinaryOp(){this.gobbleSpaces();let e=this.expr.substr(this.index,t.max_binop_len),A=e.length;for(;A>0;){if(t.binary_ops.hasOwnProperty(e)&&(!t.isIdentifierStart(this.code)||this.index+e.length<this.expr.length&&!t.isIdentifierPart(this.expr.charCodeAt(this.index+e.length))))return this.index+=A,e;e=e.substr(0,--A)}return!1}gobbleBinaryExpression(){let e,A,i,n,o,r,s,a,c;if(r=this.gobbleToken(),!r||(A=this.gobbleBinaryOp(),!A))return r;for(o={value:A,prec:t.binaryPrecedence(A),right_a:t.right_associative.has(A)},s=this.gobbleToken(),s||this.throwError("Expected expression after "+A),n=[r,o,s];A=this.gobbleBinaryOp();){if(i=t.binaryPrecedence(A),i===0){this.index-=A.length;break}o={value:A,prec:i,right_a:t.right_associative.has(A)},c=A;let l=I=>o.right_a&&I.right_a?i>I.prec:i<=I.prec;for(;n.length>2&&l(n[n.length-2]);)s=n.pop(),A=n.pop().value,r=n.pop(),e={type:t.BINARY_EXP,operator:A,left:r,right:s},n.push(e);e=this.gobbleToken(),e||this.throwError("Expected expression after "+c),n.push(o,e)}for(a=n.length-1,e=n[a];a>1;)e={type:t.BINARY_EXP,operator:n[a-1].value,left:n[a-2],right:e},a-=2;return e}gobbleToken(){let e,A,i,n;if(this.gobbleSpaces(),n=this.searchHook("gobble-token"),n)return this.runHook("after-token",n);if(e=this.code,t.isDecimalDigit(e)||e===t.PERIOD_CODE)return this.gobbleNumericLiteral();if(e===t.SQUOTE_CODE||e===t.DQUOTE_CODE)n=this.gobbleStringLiteral();else if(e===t.OBRACK_CODE)n=this.gobbleArray();else{for(A=this.expr.substr(this.index,t.max_unop_len),i=A.length;i>0;){if(t.unary_ops.hasOwnProperty(A)&&(!t.isIdentifierStart(this.code)||this.index+A.length<this.expr.length&&!t.isIdentifierPart(this.expr.charCodeAt(this.index+A.length)))){this.index+=i;let o=this.gobbleToken();return o||this.throwError("missing unaryOp argument"),this.runHook("after-token",{type:t.UNARY_EXP,operator:A,argument:o,prefix:!0})}A=A.substr(0,--i)}t.isIdentifierStart(e)?(n=this.gobbleIdentifier(),t.literals.hasOwnProperty(n.name)?n={type:t.LITERAL,value:t.literals[n.name],raw:n.name}:n.name===t.this_str&&(n={type:t.THIS_EXP})):e===t.OPAREN_CODE&&(n=this.gobbleGroup())}return n?(n=this.gobbleTokenProperty(n),this.runHook("after-token",n)):this.runHook("after-token",!1)}gobbleTokenProperty(e){this.gobbleSpaces();let A=this.code;for(;A===t.PERIOD_CODE||A===t.OBRACK_CODE||A===t.OPAREN_CODE||A===t.QUMARK_CODE;){let i;if(A===t.QUMARK_CODE){if(this.expr.charCodeAt(this.index+1)!==t.PERIOD_CODE)break;i=!0,this.index+=2,this.gobbleSpaces(),A=this.code}this.index++,A===t.OBRACK_CODE?(e={type:t.MEMBER_EXP,computed:!0,object:e,property:this.gobbleExpression()},e.property||this.throwError('Unexpected "'+this.char+'"'),this.gobbleSpaces(),A=this.code,A!==t.CBRACK_CODE&&this.throwError("Unclosed ["),this.index++):A===t.OPAREN_CODE?e={type:t.CALL_EXP,arguments:this.gobbleArguments(t.CPAREN_CODE),callee:e}:(A===t.PERIOD_CODE||i)&&(i&&this.index--,this.gobbleSpaces(),e={type:t.MEMBER_EXP,computed:!1,object:e,property:this.gobbleIdentifier()}),i&&(e.optional=!0),this.gobbleSpaces(),A=this.code}return e}gobbleNumericLiteral(){let e="",A,i;for(;t.isDecimalDigit(this.code);)e+=this.expr.charAt(this.index++);if(this.code===t.PERIOD_CODE)for(e+=this.expr.charAt(this.index++);t.isDecimalDigit(this.code);)e+=this.expr.charAt(this.index++);if(A=this.char,A==="e"||A==="E"){for(e+=this.expr.charAt(this.index++),A=this.char,(A==="+"||A==="-")&&(e+=this.expr.charAt(this.index++));t.isDecimalDigit(this.code);)e+=this.expr.charAt(this.index++);t.isDecimalDigit(this.expr.charCodeAt(this.index-1))||this.throwError("Expected exponent ("+e+this.char+")")}return i=this.code,t.isIdentifierStart(i)?this.throwError("Variable names cannot start with a number ("+e+this.char+")"):(i===t.PERIOD_CODE||e.length===1&&e.charCodeAt(0)===t.PERIOD_CODE)&&this.throwError("Unexpected period"),{type:t.LITERAL,value:parseFloat(e),raw:e}}gobbleStringLiteral(){let e="",A=this.index,i=this.expr.charAt(this.index++),n=!1;for(;this.index<this.expr.length;){let o=this.expr.charAt(this.index++);if(o===i){n=!0;break}else if(o==="\\")switch(o=this.expr.charAt(this.index++),o){case"n":e+=`
`;break;case"r":e+="\r";break;case"t":e+="	";break;case"b":e+="\b";break;case"f":e+="\f";break;case"v":e+="\v";break;default:e+=o}else e+=o}return n||this.throwError('Unclosed quote after "'+e+'"'),{type:t.LITERAL,value:e,raw:this.expr.substring(A,this.index)}}gobbleIdentifier(){let e=this.code,A=this.index;for(t.isIdentifierStart(e)?this.index++:this.throwError("Unexpected "+this.char);this.index<this.expr.length&&(e=this.code,t.isIdentifierPart(e));)this.index++;return{type:t.IDENTIFIER,name:this.expr.slice(A,this.index)}}gobbleArguments(e){let A=[],i=!1,n=0;for(;this.index<this.expr.length;){this.gobbleSpaces();let o=this.code;if(o===e){i=!0,this.index++,e===t.CPAREN_CODE&&n&&n>=A.length&&this.throwError("Unexpected token "+String.fromCharCode(e));break}else if(o===t.COMMA_CODE){if(this.index++,n++,n!==A.length){if(e===t.CPAREN_CODE)this.throwError("Unexpected token ,");else if(e===t.CBRACK_CODE)for(let r=A.length;r<n;r++)A.push(null)}}else if(A.length!==n&&n!==0)this.throwError("Expected comma");else{let r=this.gobbleExpression();(!r||r.type===t.COMPOUND)&&this.throwError("Expected comma"),A.push(r)}}return i||this.throwError("Expected "+String.fromCharCode(e)),A}gobbleGroup(){this.index++;let e=this.gobbleExpressions(t.CPAREN_CODE);if(this.code===t.CPAREN_CODE)return this.index++,e.length===1?e[0]:e.length?{type:t.SEQUENCE_EXP,expressions:e}:!1;this.throwError("Unclosed (")}gobbleArray(){return this.index++,{type:t.ARRAY_EXP,elements:this.gobbleArguments(t.CBRACK_CODE)}}},nRA=new Kx;Object.assign(Ra,{hooks:nRA,plugins:new Yx(Ra),COMPOUND:"Compound",SEQUENCE_EXP:"SequenceExpression",IDENTIFIER:"Identifier",MEMBER_EXP:"MemberExpression",LITERAL:"Literal",THIS_EXP:"ThisExpression",CALL_EXP:"CallExpression",UNARY_EXP:"UnaryExpression",BINARY_EXP:"BinaryExpression",ARRAY_EXP:"ArrayExpression",TAB_CODE:9,LF_CODE:10,CR_CODE:13,SPACE_CODE:32,PERIOD_CODE:46,COMMA_CODE:44,SQUOTE_CODE:39,DQUOTE_CODE:34,OPAREN_CODE:40,CPAREN_CODE:41,OBRACK_CODE:91,CBRACK_CODE:93,QUMARK_CODE:63,SEMCOL_CODE:59,COLON_CODE:58,unary_ops:{"-":1,"!":1,"~":1,"+":1},binary_ops:{"||":1,"??":1,"&&":2,"|":3,"^":4,"&":5,"==":6,"!=":6,"===":6,"!==":6,"<":7,">":7,"<=":7,">=":7,"<<":8,">>":8,">>>":8,"+":9,"-":9,"*":10,"/":10,"%":10,"**":11},right_associative:new Set(["**"]),additional_identifier_chars:new Set(["$","_"]),literals:{true:!0,false:!1,null:null},this_str:"this"});Ra.max_unop_len=Ra.getMaxKeyLen(Ra.unary_ops);Ra.max_binop_len=Ra.getMaxKeyLen(Ra.binary_ops);var Y0=t=>new Ra(t).parse(),oRA=Object.getOwnPropertyNames(class{});Object.getOwnPropertyNames(Ra).filter(t=>!oRA.includes(t)&&Y0[t]===void 0).forEach(t=>{Y0[t]=Ra[t]});Y0.Jsep=Ra;var rRA="ConditionalExpression",sRA={name:"ternary",init(t){t.hooks.add("after-expression",function(A){if(A.node&&this.code===t.QUMARK_CODE){this.index++;let i=A.node,n=this.gobbleExpression();if(n||this.throwError("Expected expression"),this.gobbleSpaces(),this.code===t.COLON_CODE){this.index++;let o=this.gobbleExpression();if(o||this.throwError("Expected expression"),A.node={type:rRA,test:i,consequent:n,alternate:o},i.operator&&t.binary_ops[i.operator]<=.9){let r=i;for(;r.right.operator&&t.binary_ops[r.right.operator]<=.9;)r=r.right;A.node.test=r.right,r.right=A.node,A.node=i}}else this.throwError("Expected :")}})}};Y0.plugins.register(sRA);var UAA=47,aRA=92,cRA={name:"regex",init(t){t.hooks.add("gobble-token",function(A){if(this.code===UAA){let i=++this.index,n=!1;for(;this.index<this.expr.length;){if(this.code===UAA&&!n){let o=this.expr.slice(i,this.index),r="";for(;++this.index<this.expr.length;){let a=this.code;if(a>=97&&a<=122||a>=65&&a<=90||a>=48&&a<=57)r+=this.char;else break}let s;try{s=new RegExp(o,r)}catch(a){this.throwError(a.message)}return A.node={type:t.LITERAL,value:s,raw:this.expr.slice(i-1,this.index)},A.node=this.gobbleTokenProperty(A.node),A.node}this.code===t.OBRACK_CODE?n=!0:n&&this.code===t.CBRACK_CODE&&(n=!1),this.index+=this.code===aRA?2:1}this.throwError("Unclosed Regex")}})}},Ux=43,lRA=45,pE={name:"assignment",assignmentOperators:new Set(["=","*=","**=","/=","%=","+=","-=","<<=",">>=",">>>=","&=","^=","|=","||=","&&=","??="]),updateOperators:[Ux,lRA],assignmentPrecedence:.9,init(t){let e=[t.IDENTIFIER,t.MEMBER_EXP];pE.assignmentOperators.forEach(i=>t.addBinaryOp(i,pE.assignmentPrecedence,!0)),t.hooks.add("gobble-token",function(n){let o=this.code;pE.updateOperators.some(r=>r===o&&r===this.expr.charCodeAt(this.index+1))&&(this.index+=2,n.node={type:"UpdateExpression",operator:o===Ux?"++":"--",argument:this.gobbleTokenProperty(this.gobbleIdentifier()),prefix:!0},(!n.node.argument||!e.includes(n.node.argument.type))&&this.throwError(`Unexpected ${n.node.operator}`))}),t.hooks.add("after-token",function(n){if(n.node){let o=this.code;pE.updateOperators.some(r=>r===o&&r===this.expr.charCodeAt(this.index+1))&&(e.includes(n.node.type)||this.throwError(`Unexpected ${n.node.operator}`),this.index+=2,n.node={type:"UpdateExpression",operator:o===Ux?"++":"--",argument:n.node,prefix:!1})}}),t.hooks.add("after-expression",function(n){n.node&&A(n.node)});function A(i){pE.assignmentOperators.has(i.operator)?(i.type="AssignmentExpression",A(i.left),A(i.right)):i.operator||Object.values(i).forEach(n=>{n&&typeof n=="object"&&A(n)})}}};Y0.plugins.register(cRA,pE);Y0.addUnaryOp("typeof");Y0.addLiteral("null",null);Y0.addLiteral("undefined",void 0);var gRA=new Set(["constructor","__proto__","__defineGetter__","__defineSetter__"]),so={evalAst(t,e){switch(t.type){case"BinaryExpression":case"LogicalExpression":return so.evalBinaryExpression(t,e);case"Compound":return so.evalCompound(t,e);case"ConditionalExpression":return so.evalConditionalExpression(t,e);case"Identifier":return so.evalIdentifier(t,e);case"Literal":return so.evalLiteral(t,e);case"MemberExpression":return so.evalMemberExpression(t,e);case"UnaryExpression":return so.evalUnaryExpression(t,e);case"ArrayExpression":return so.evalArrayExpression(t,e);case"CallExpression":return so.evalCallExpression(t,e);case"AssignmentExpression":return so.evalAssignmentExpression(t,e);default:throw SyntaxError("Unexpected expression",t)}},evalBinaryExpression(t,e){return{"||":(i,n)=>i||n(),"&&":(i,n)=>i&&n(),"|":(i,n)=>i|n(),"^":(i,n)=>i^n(),"&":(i,n)=>i&n(),"==":(i,n)=>i==n(),"!=":(i,n)=>i!=n(),"===":(i,n)=>i===n(),"!==":(i,n)=>i!==n(),"<":(i,n)=>i<n(),">":(i,n)=>i>n(),"<=":(i,n)=>i<=n(),">=":(i,n)=>i>=n(),"<<":(i,n)=>i<<n(),">>":(i,n)=>i>>n(),">>>":(i,n)=>i>>>n(),"+":(i,n)=>i+n(),"-":(i,n)=>i-n(),"*":(i,n)=>i*n(),"/":(i,n)=>i/n(),"%":(i,n)=>i%n()}[t.operator](so.evalAst(t.left,e),()=>so.evalAst(t.right,e))},evalCompound(t,e){let A;for(let i=0;i<t.body.length;i++){t.body[i].type==="Identifier"&&["var","let","const"].includes(t.body[i].name)&&t.body[i+1]&&t.body[i+1].type==="AssignmentExpression"&&(i+=1);let n=t.body[i];A=so.evalAst(n,e)}return A},evalConditionalExpression(t,e){return so.evalAst(t.test,e)?so.evalAst(t.consequent,e):so.evalAst(t.alternate,e)},evalIdentifier(t,e){if(Object.hasOwn(e,t.name))return e[t.name];throw ReferenceError(`${t.name} is not defined`)},evalLiteral(t){return t.value},evalMemberExpression(t,e){let A=String(t.computed?so.evalAst(t.property):t.property.name),i=so.evalAst(t.object,e);if(i==null)throw TypeError(`Cannot read properties of ${i} (reading '${A}')`);if(!Object.hasOwn(i,A)&&gRA.has(A))throw TypeError(`Cannot read properties of ${i} (reading '${A}')`);let n=i[A];return typeof n=="function"?n.bind(i):n},evalUnaryExpression(t,e){return{"-":i=>-so.evalAst(i,e),"!":i=>!so.evalAst(i,e),"~":i=>~so.evalAst(i,e),"+":i=>+so.evalAst(i,e),typeof:i=>typeof so.evalAst(i,e)}[t.operator](t.argument)},evalArrayExpression(t,e){return t.elements.map(A=>so.evalAst(A,e))},evalCallExpression(t,e){let A=t.arguments.map(n=>so.evalAst(n,e));return so.evalAst(t.callee,e)(...A)},evalAssignmentExpression(t,e){if(t.left.type!=="Identifier")throw SyntaxError("Invalid left-hand side in assignment");let A=t.left.name,i=so.evalAst(t.right,e);return e[A]=i,e[A]}},Jx=class{constructor(e){this.code=e,this.ast=Y0(this.code)}runInNewContext(e){let A=Object.assign(Object.create(null),e);return so.evalAst(this.ast,A)}};function u1(t,e){return t=t.slice(),t.push(e),t}function Tx(t,e){return e=e.slice(),e.unshift(t),e}var Hx=class extends Error{constructor(e){super('JSONPath should not be called with "new" (it prevents return of (unwrapped) scalar values)'),this.avoidNew=!0,this.value=e,this.name="NewError"}};function xn(t,e,A,i,n){if(!(this instanceof xn))try{return new xn(t,e,A,i,n)}catch(r){if(!r.avoidNew)throw r;return r.value}typeof t=="string"&&(n=i,i=A,A=e,e=t,t=null);let o=t&&typeof t=="object";if(t=t||{},this.json=t.json||A,this.path=t.path||e,this.resultType=t.resultType||"value",this.flatten=t.flatten||!1,this.wrap=Object.hasOwn(t,"wrap")?t.wrap:!0,this.sandbox=t.sandbox||{},this.eval=t.eval===void 0?"safe":t.eval,this.ignoreEvalErrors=typeof t.ignoreEvalErrors>"u"?!1:t.ignoreEvalErrors,this.parent=t.parent||null,this.parentProperty=t.parentProperty||null,this.callback=t.callback||i||null,this.otherTypeCallback=t.otherTypeCallback||n||function(){throw new TypeError("You must supply an otherTypeCallback callback option with the @other() operator.")},t.autostart!==!1){let r={path:o?t.path:e};o?"json"in t&&(r.json=t.json):r.json=A;let s=this.evaluate(r);if(!s||typeof s!="object")throw new Hx(s);return s}}xn.prototype.evaluate=function(t,e,A,i){let n=this.parent,o=this.parentProperty,{flatten:r,wrap:s}=this;if(this.currResultType=this.resultType,this.currEval=this.eval,this.currSandbox=this.sandbox,A=A||this.callback,this.currOtherTypeCallback=i||this.otherTypeCallback,e=e||this.json,t=t||this.path,t&&typeof t=="object"&&!Array.isArray(t)){if(!t.path&&t.path!=="")throw new TypeError('You must supply a "path" property when providing an object argument to JSONPath.evaluate().');if(!Object.hasOwn(t,"json"))throw new TypeError('You must supply a "json" property when providing an object argument to JSONPath.evaluate().');({json:e}=t),r=Object.hasOwn(t,"flatten")?t.flatten:r,this.currResultType=Object.hasOwn(t,"resultType")?t.resultType:this.currResultType,this.currSandbox=Object.hasOwn(t,"sandbox")?t.sandbox:this.currSandbox,s=Object.hasOwn(t,"wrap")?t.wrap:s,this.currEval=Object.hasOwn(t,"eval")?t.eval:this.currEval,A=Object.hasOwn(t,"callback")?t.callback:A,this.currOtherTypeCallback=Object.hasOwn(t,"otherTypeCallback")?t.otherTypeCallback:this.currOtherTypeCallback,n=Object.hasOwn(t,"parent")?t.parent:n,o=Object.hasOwn(t,"parentProperty")?t.parentProperty:o,t=t.path}if(n=n||null,o=o||null,Array.isArray(t)&&(t=xn.toPathString(t)),!t&&t!==""||!e)return;let a=xn.toPathArray(t);a[0]==="$"&&a.length>1&&a.shift(),this._hasParentSelector=null;let c=this._trace(a,e,["$"],n,o,A).filter(function(l){return l&&!l.isParentSelector});return c.length?!s&&c.length===1&&!c[0].hasArrExpr?this._getPreferredOutput(c[0]):c.reduce((l,I)=>{let C=this._getPreferredOutput(I);return r&&Array.isArray(C)?l=l.concat(C):l.push(C),l},[]):s?[]:void 0};xn.prototype._getPreferredOutput=function(t){let e=this.currResultType;switch(e){case"all":{let A=Array.isArray(t.path)?t.path:xn.toPathArray(t.path);return t.pointer=xn.toPointer(A),t.path=typeof t.path=="string"?t.path:xn.toPathString(t.path),t}case"value":case"parent":case"parentProperty":return t[e];case"path":return xn.toPathString(t[e]);case"pointer":return xn.toPointer(t.path);default:throw new TypeError("Unknown result type")}};xn.prototype._handleCallback=function(t,e,A){if(e){let i=this._getPreferredOutput(t);t.path=typeof t.path=="string"?t.path:xn.toPathString(t.path),e(i,A,t)}};xn.prototype._trace=function(t,e,A,i,n,o,r,s){let a;if(!t.length)return a={path:A,value:e,parent:i,parentProperty:n,hasArrExpr:r},this._handleCallback(a,o,"value"),a;let c=t[0],l=t.slice(1),I=[];function C(d){Array.isArray(d)?d.forEach(B=>{I.push(B)}):I.push(d)}if((typeof c!="string"||s)&&e&&Object.hasOwn(e,c))C(this._trace(l,e[c],u1(A,c),e,c,o,r));else if(c==="*")this._walk(e,d=>{C(this._trace(l,e[d],u1(A,d),e,d,o,!0,!0))});else if(c==="..")C(this._trace(l,e,A,i,n,o,r)),this._walk(e,d=>{typeof e[d]=="object"&&C(this._trace(t.slice(),e[d],u1(A,d),e,d,o,!0))});else{if(c==="^")return this._hasParentSelector=!0,{path:A.slice(0,-1),expr:l,isParentSelector:!0};if(c==="~")return a={path:u1(A,c),value:n,parent:i,parentProperty:null},this._handleCallback(a,o,"property"),a;if(c==="$")C(this._trace(l,e,A,null,null,o,r));else if(/^(-?\d*):(-?\d*):?(\d*)$/u.test(c))C(this._slice(c,l,e,A,i,n,o));else if(c.indexOf("?(")===0){if(this.currEval===!1)throw new Error("Eval [?(expr)] prevented in JSONPath expression.");let d=c.replace(/^\?\((.*?)\)$/u,"$1"),B=/@.?([^?]*)[['](\??\(.*?\))(?!.\)\])[\]']/gu.exec(d);B?this._walk(e,E=>{let h=[B[2]],u=B[1]?e[E][B[1]]:e[E];this._trace(h,u,A,i,n,o,!0).length>0&&C(this._trace(l,e[E],u1(A,E),e,E,o,!0))}):this._walk(e,E=>{this._eval(d,e[E],E,A,i,n)&&C(this._trace(l,e[E],u1(A,E),e,E,o,!0))})}else if(c[0]==="("){if(this.currEval===!1)throw new Error("Eval [(expr)] prevented in JSONPath expression.");C(this._trace(Tx(this._eval(c,e,A.at(-1),A.slice(0,-1),i,n),l),e,A,i,n,o,r))}else if(c[0]==="@"){let d=!1,B=c.slice(1,-2);switch(B){case"scalar":(!e||!["object","function"].includes(typeof e))&&(d=!0);break;case"boolean":case"string":case"undefined":case"function":typeof e===B&&(d=!0);break;case"integer":Number.isFinite(e)&&!(e%1)&&(d=!0);break;case"number":Number.isFinite(e)&&(d=!0);break;case"nonFinite":typeof e=="number"&&!Number.isFinite(e)&&(d=!0);break;case"object":e&&typeof e===B&&(d=!0);break;case"array":Array.isArray(e)&&(d=!0);break;case"other":d=this.currOtherTypeCallback(e,A,i,n);break;case"null":e===null&&(d=!0);break;default:throw new TypeError("Unknown value type "+B)}if(d)return a={path:A,value:e,parent:i,parentProperty:n},this._handleCallback(a,o,"value"),a}else if(c[0]==="`"&&e&&Object.hasOwn(e,c.slice(1))){let d=c.slice(1);C(this._trace(l,e[d],u1(A,d),e,d,o,r,!0))}else if(c.includes(",")){let d=c.split(",");for(let B of d)C(this._trace(Tx(B,l),e,A,i,n,o,!0))}else!s&&e&&Object.hasOwn(e,c)&&C(this._trace(l,e[c],u1(A,c),e,c,o,r,!0))}if(this._hasParentSelector)for(let d=0;d<I.length;d++){let B=I[d];if(B&&B.isParentSelector){let E=this._trace(B.expr,e,B.path,i,n,o,r);if(Array.isArray(E)){I[d]=E[0];let h=E.length;for(let u=1;u<h;u++)d++,I.splice(d,0,E[u])}else I[d]=E}}return I};xn.prototype._walk=function(t,e){if(Array.isArray(t)){let A=t.length;for(let i=0;i<A;i++)e(i)}else t&&typeof t=="object"&&Object.keys(t).forEach(A=>{e(A)})};xn.prototype._slice=function(t,e,A,i,n,o,r){if(!Array.isArray(A))return;let s=A.length,a=t.split(":"),c=a[2]&&Number.parseInt(a[2])||1,l=a[0]&&Number.parseInt(a[0])||0,I=a[1]&&Number.parseInt(a[1])||s;l=l<0?Math.max(0,l+s):Math.min(s,l),I=I<0?Math.max(0,I+s):Math.min(s,I);let C=[];for(let d=l;d<I;d+=c)this._trace(Tx(d,e),A,i,n,o,r,!0).forEach(E=>{C.push(E)});return C};xn.prototype._eval=function(t,e,A,i,n,o){this.currSandbox._$_parentProperty=o,this.currSandbox._$_parent=n,this.currSandbox._$_property=A,this.currSandbox._$_root=this.json,this.currSandbox._$_v=e;let r=t.includes("@path");r&&(this.currSandbox._$_path=xn.toPathString(i.concat([A])));let s=this.currEval+"Script:"+t;if(!xn.cache[s]){let a=t.replaceAll("@parentProperty","_$_parentProperty").replaceAll("@parent","_$_parent").replaceAll("@property","_$_property").replaceAll("@root","_$_root").replaceAll(/@([.\s)[])/gu,"_$_v$1");if(r&&(a=a.replaceAll("@path","_$_path")),this.currEval==="safe"||this.currEval===!0||this.currEval===void 0)xn.cache[s]=new this.safeVm.Script(a);else if(this.currEval==="native")xn.cache[s]=new this.vm.Script(a);else if(typeof this.currEval=="function"&&this.currEval.prototype&&Object.hasOwn(this.currEval.prototype,"runInNewContext")){let c=this.currEval;xn.cache[s]=new c(a)}else if(typeof this.currEval=="function")xn.cache[s]={runInNewContext:c=>this.currEval(a,c)};else throw new TypeError(`Unknown "eval" property "${this.currEval}"`)}try{return xn.cache[s].runInNewContext(this.currSandbox)}catch(a){if(this.ignoreEvalErrors)return!1;throw new Error("jsonPath: "+a.message+": "+t)}};xn.cache={};xn.toPathString=function(t){let e=t,A=e.length,i="$";for(let n=1;n<A;n++)/^(~|\^|@.*?\(\))$/u.test(e[n])||(i+=/^[0-9*]+$/u.test(e[n])?"["+e[n]+"]":"['"+e[n]+"']");return i};xn.toPointer=function(t){let e=t,A=e.length,i="";for(let n=1;n<A;n++)/^(~|\^|@.*?\(\))$/u.test(e[n])||(i+="/"+e[n].toString().replaceAll("~","~0").replaceAll("/","~1"));return i};xn.toPathArray=function(t){let{cache:e}=xn;if(e[t])return e[t].concat();let A=[],n=t.replaceAll(/@(?:null|boolean|number|string|integer|undefined|nonFinite|scalar|array|object|function|other)\(\)/gu,";$&;").replaceAll(/[['](\??\(.*?\))[\]'](?!.\])/gu,function(o,r){return"[#"+(A.push(r)-1)+"]"}).replaceAll(/\[['"]([^'\]]*)['"]\]/gu,function(o,r){return"['"+r.replaceAll(".","%@%").replaceAll("~","%%@@%%")+"']"}).replaceAll("~",";~;").replaceAll(/['"]?\.['"]?(?![^[]*\])|\[['"]?/gu,";").replaceAll("%@%",".").replaceAll("%%@@%%","~").replaceAll(/(?:;)?(\^+)(?:;)?/gu,function(o,r){return";"+r.split("").join(";")+";"}).replaceAll(/;;;|;;/gu,";..;").replaceAll(/;$|'?\]|'$/gu,"").split(";").map(function(o){let r=o.match(/#(\d+)/u);return!r||!r[1]?o:A[r[1]]});return e[t]=n,e[t].concat()};xn.prototype.safeVm={Script:Jx};var IRA=function(t,e,A){let i=t.length;for(let n=0;n<i;n++){let o=t[n];A(o)&&e.push(t.splice(n--,1)[0])}},zx=class{constructor(e){this.code=e}runInNewContext(e){let A=this.code,i=Object.keys(e),n=[];IRA(i,n,c=>typeof e[c]=="function");let o=i.map(c=>e[c]);A=n.reduce((c,l)=>{let I=e[l].toString();return/function/u.test(I)||(I="function "+I),"var "+l+"="+I+";"+c},"")+A,!/(['"])use strict\1/u.test(A)&&!i.includes("arguments")&&(A="var arguments = undefined;"+A),A=A.replace(/;\s*$/u,"");let s=A.lastIndexOf(";"),a=s!==-1?A.slice(0,s+1)+" return "+A.slice(s+1):" return "+A;return new Function(...i,a)(...o)}};xn.prototype.vm={Script:zx};var Px=[],TAA=[];(()=>{let t="lc,34,7n,7,7b,19,,,,2,,2,,,20,b,1c,l,g,,2t,7,2,6,2,2,,4,z,,u,r,2j,b,1m,9,9,,o,4,,9,,3,,5,17,3,3b,f,,w,1j,,,,4,8,4,,3,7,a,2,t,,1m,,,,2,4,8,,9,,a,2,q,,2,2,1l,,4,2,4,2,2,3,3,,u,2,3,,b,2,1l,,4,5,,2,4,,k,2,m,6,,,1m,,,2,,4,8,,7,3,a,2,u,,1n,,,,c,,9,,14,,3,,1l,3,5,3,,4,7,2,b,2,t,,1m,,2,,2,,3,,5,2,7,2,b,2,s,2,1l,2,,,2,4,8,,9,,a,2,t,,20,,4,,2,3,,,8,,29,,2,7,c,8,2q,,2,9,b,6,22,2,r,,,,,,1j,e,,5,,2,5,b,,10,9,,2u,4,,6,,2,2,2,p,2,4,3,g,4,d,,2,2,6,,f,,jj,3,qa,3,t,3,t,2,u,2,1s,2,,7,8,,2,b,9,,19,3,3b,2,y,,3a,3,4,2,9,,6,3,63,2,2,,1m,,,7,,,,,2,8,6,a,2,,1c,h,1r,4,1c,7,,,5,,14,9,c,2,w,4,2,2,,3,1k,,,2,3,,,3,1m,8,2,2,48,3,,d,,7,4,,6,,3,2,5i,1m,,5,ek,,5f,x,2da,3,3x,,2o,w,fe,6,2x,2,n9w,4,,a,w,2,28,2,7k,,3,,4,,p,2,5,,47,2,q,i,d,,12,8,p,b,1a,3,1c,,2,4,2,2,13,,1v,6,2,2,2,2,c,,8,,1b,,1f,,,3,2,2,5,2,,,16,2,8,,6m,,2,,4,,fn4,,kh,g,g,g,a6,2,gt,,6a,,45,5,1ae,3,,2,5,4,14,3,4,,4l,2,fx,4,ar,2,49,b,4w,,1i,f,1k,3,1d,4,2,2,1x,3,10,5,,8,1q,,c,2,1g,9,a,4,2,,2n,3,2,,,2,6,,4g,,3,8,l,2,1l,2,,,,,m,,e,7,3,5,5f,8,2,3,,,n,,29,,2,6,,,2,,,2,,2,6j,,2,4,6,2,,2,r,2,2d,8,2,,,2,2y,,,,2,6,,,2t,3,2,4,,5,77,9,,2,6t,,a,2,,,4,,40,4,2,2,4,,w,a,14,6,2,4,8,,9,6,2,3,1a,d,,2,ba,7,,6,,,2a,m,2,7,,2,,2,3e,6,3,,,2,,7,,,20,2,3,,,,9n,2,f0b,5,1n,7,t4,,1r,4,29,,f5k,2,43q,,,3,4,5,8,8,2,7,u,4,44,3,1iz,1j,4,1e,8,,e,,m,5,,f,11s,7,,h,2,7,,2,,5,79,7,c5,4,15s,7,31,7,240,5,gx7k,2o,3k,6o".split(",").map(e=>e?parseInt(e,36):1);for(let e=0,A=0;e<t.length;e++)(e%2?TAA:Px).push(A=A+t[e])})();function CRA(t){if(t<768)return!1;for(let e=0,A=Px.length;;){let i=e+A>>1;if(t<Px[i])A=i;else if(t>=TAA[i])e=i+1;else return!0;if(e==A)return!1}}function KAA(t){return t>=127462&&t<=127487}var YAA=8205;function HAA(t,e,A=!0,i=!0){return(A?zAA:dRA)(t,e,i)}function zAA(t,e,A){if(e==t.length)return e;e&&OAA(t.charCodeAt(e))&&PAA(t.charCodeAt(e-1))&&e--;let i=Ox(t,e);for(e+=JAA(i);e<t.length;){let n=Ox(t,e);if(i==YAA||n==YAA||A&&CRA(n))e+=JAA(n),i=n;else if(KAA(n)){let o=0,r=e-2;for(;r>=0&&KAA(Ox(t,r));)o++,r-=2;if(o%2==0)break;e+=2}else break}return e}function dRA(t,e,A){for(;e>0;){let i=zAA(t,e-2,A);if(i<e)return i;e--}return 0}function Ox(t,e){let A=t.charCodeAt(e);if(!PAA(A)||e+1==t.length)return A;let i=t.charCodeAt(e+1);return OAA(i)?(A-55296<<10)+(i-56320)+65536:A}function OAA(t){return t>=56320&&t<57344}function PAA(t){return t>=55296&&t<56320}function JAA(t){return t<65536?1:2}var In=class t{lineAt(e){if(e<0||e>this.length)throw new RangeError(`Invalid position ${e} in document of length ${this.length}`);return this.lineInner(e,!1,1,0)}line(e){if(e<1||e>this.lines)throw new RangeError(`Invalid line number ${e} in ${this.lines}-line document`);return this.lineInner(e,!0,1,0)}replace(e,A,i){[e,A]=bE(this,e,A);let n=[];return this.decompose(0,e,n,2),i.length&&i.decompose(0,i.length,n,3),this.decompose(A,this.length,n,1),DE.from(n,this.length-(A-e)+i.length)}append(e){return this.replace(this.length,this.length,e)}slice(e,A=this.length){[e,A]=bE(this,e,A);let i=[];return this.decompose(e,A,i,0),DE.from(i,A-e)}eq(e){if(e==this)return!0;if(e.length!=this.length||e.lines!=this.lines)return!1;let A=this.scanIdentical(e,1),i=this.length-this.scanIdentical(e,-1),n=new pC(this),o=new pC(e);for(let r=A,s=A;;){if(n.next(r),o.next(r),r=0,n.lineBreak!=o.lineBreak||n.done!=o.done||n.value!=o.value)return!1;if(s+=n.value.length,n.done||s>=i)return!0}}iter(e=1){return new pC(this,e)}iterRange(e,A=this.length){return new sw(this,e,A)}iterLines(e,A){let i;if(e==null)i=this.iter();else{A==null&&(A=this.lines+1);let n=this.line(e).from;i=this.iterRange(n,Math.max(n,A==this.lines+1?this.length:A<=1?0:this.line(A-1).to))}return new aw(i)}toString(){return this.sliceString(0)}toJSON(){let e=[];return this.flatten(e),e}constructor(){}static of(e){if(e.length==0)throw new RangeError("A document must have at least one line");return e.length==1&&!e[0]?t.empty:e.length<=32?new cc(e):DE.from(cc.split(e,[]))}},cc=class t extends In{constructor(e,A=BRA(e)){super(),this.text=e,this.length=A}get lines(){return this.text.length}get children(){return null}lineInner(e,A,i,n){for(let o=0;;o++){let r=this.text[o],s=n+r.length;if((A?i:s)>=e)return new Vx(n,s,i,r);n=s+1,i++}}decompose(e,A,i,n){let o=e<=0&&A>=this.length?this:new t(jAA(this.text,e,A),Math.min(A,this.length)-Math.max(0,e));if(n&1){let r=i.pop(),s=rw(o.text,r.text.slice(),0,o.length);if(s.length<=32)i.push(new t(s,r.length+o.length));else{let a=s.length>>1;i.push(new t(s.slice(0,a)),new t(s.slice(a)))}}else i.push(o)}replace(e,A,i){if(!(i instanceof t))return super.replace(e,A,i);[e,A]=bE(this,e,A);let n=rw(this.text,rw(i.text,jAA(this.text,0,e)),A),o=this.length+i.length-(A-e);return n.length<=32?new t(n,o):DE.from(t.split(n,[]),o)}sliceString(e,A=this.length,i=`
`){[e,A]=bE(this,e,A);let n="";for(let o=0,r=0;o<=A&&r<this.text.length;r++){let s=this.text[r],a=o+s.length;o>e&&r&&(n+=i),e<a&&A>o&&(n+=s.slice(Math.max(0,e-o),A-o)),o=a+1}return n}flatten(e){for(let A of this.text)e.push(A)}scanIdentical(){return 0}static split(e,A){let i=[],n=-1;for(let o of e)i.push(o),n+=o.length+1,i.length==32&&(A.push(new t(i,n)),i=[],n=-1);return n>-1&&A.push(new t(i,n)),A}},DE=class t extends In{constructor(e,A){super(),this.children=e,this.length=A,this.lines=0;for(let i of e)this.lines+=i.lines}lineInner(e,A,i,n){for(let o=0;;o++){let r=this.children[o],s=n+r.length,a=i+r.lines-1;if((A?a:s)>=e)return r.lineInner(e,A,i,n);n=s+1,i=a+1}}decompose(e,A,i,n){for(let o=0,r=0;r<=A&&o<this.children.length;o++){let s=this.children[o],a=r+s.length;if(e<=a&&A>=r){let c=n&((r<=e?1:0)|(a>=A?2:0));r>=e&&a<=A&&!c?i.push(s):s.decompose(e-r,A-r,i,c)}r=a+1}}replace(e,A,i){if([e,A]=bE(this,e,A),i.lines<this.lines)for(let n=0,o=0;n<this.children.length;n++){let r=this.children[n],s=o+r.length;if(e>=o&&A<=s){let a=r.replace(e-o,A-o,i),c=this.lines-r.lines+a.lines;if(a.lines<c>>4&&a.lines>c>>6){let l=this.children.slice();return l[n]=a,new t(l,this.length-(A-e)+i.length)}return super.replace(o,s,a)}o=s+1}return super.replace(e,A,i)}sliceString(e,A=this.length,i=`
`){[e,A]=bE(this,e,A);let n="";for(let o=0,r=0;o<this.children.length&&r<=A;o++){let s=this.children[o],a=r+s.length;r>e&&o&&(n+=i),e<a&&A>r&&(n+=s.sliceString(e-r,A-r,i)),r=a+1}return n}flatten(e){for(let A of this.children)A.flatten(e)}scanIdentical(e,A){if(!(e instanceof t))return 0;let i=0,[n,o,r,s]=A>0?[0,0,this.children.length,e.children.length]:[this.children.length-1,e.children.length-1,-1,-1];for(;;n+=A,o+=A){if(n==r||o==s)return i;let a=this.children[n],c=e.children[o];if(a!=c)return i+a.scanIdentical(c,A);i+=a.length+1}}static from(e,A=e.reduce((i,n)=>i+n.length+1,-1)){let i=0;for(let d of e)i+=d.lines;if(i<32){let d=[];for(let B of e)B.flatten(d);return new cc(d,A)}let n=Math.max(32,i>>5),o=n<<1,r=n>>1,s=[],a=0,c=-1,l=[];function I(d){let B;if(d.lines>o&&d instanceof t)for(let E of d.children)I(E);else d.lines>r&&(a>r||!a)?(C(),s.push(d)):d instanceof cc&&a&&(B=l[l.length-1])instanceof cc&&d.lines+B.lines<=32?(a+=d.lines,c+=d.length+1,l[l.length-1]=new cc(B.text.concat(d.text),B.length+1+d.length)):(a+d.lines>n&&C(),a+=d.lines,c+=d.length+1,l.push(d))}function C(){a!=0&&(s.push(l.length==1?l[0]:t.from(l,c)),c=-1,a=l.length=0)}for(let d of e)I(d);return C(),s.length==1?s[0]:new t(s,A)}};In.empty=new cc([""],0);function BRA(t){let e=-1;for(let A of t)e+=A.length+1;return e}function rw(t,e,A=0,i=1e9){for(let n=0,o=0,r=!0;o<t.length&&n<=i;o++){let s=t[o],a=n+s.length;a>=A&&(a>i&&(s=s.slice(0,i-n)),n<A&&(s=s.slice(A-n)),r?(e[e.length-1]+=s,r=!1):e.push(s)),n=a+1}return e}function jAA(t,e,A){return rw(t,[""],e,A)}var pC=class{constructor(e,A=1){this.dir=A,this.done=!1,this.lineBreak=!1,this.value="",this.nodes=[e],this.offsets=[A>0?1:(e instanceof cc?e.text.length:e.children.length)<<1]}nextInner(e,A){for(this.done=this.lineBreak=!1;;){let i=this.nodes.length-1,n=this.nodes[i],o=this.offsets[i],r=o>>1,s=n instanceof cc?n.text.length:n.children.length;if(r==(A>0?s:0)){if(i==0)return this.done=!0,this.value="",this;A>0&&this.offsets[i-1]++,this.nodes.pop(),this.offsets.pop()}else if((o&1)==(A>0?0:1)){if(this.offsets[i]+=A,e==0)return this.lineBreak=!0,this.value=`
`,this;e--}else if(n instanceof cc){let a=n.text[r+(A<0?-1:0)];if(this.offsets[i]+=A,a.length>Math.max(0,e))return this.value=e==0?a:A>0?a.slice(e):a.slice(0,a.length-e),this;e-=a.length}else{let a=n.children[r+(A<0?-1:0)];e>a.length?(e-=a.length,this.offsets[i]+=A):(A<0&&this.offsets[i]--,this.nodes.push(a),this.offsets.push(A>0?1:(a instanceof cc?a.text.length:a.children.length)<<1))}}}next(e=0){return e<0&&(this.nextInner(-e,-this.dir),e=this.value.length),this.nextInner(e,this.dir)}},sw=class{constructor(e,A,i){this.value="",this.done=!1,this.cursor=new pC(e,A>i?-1:1),this.pos=A>i?e.length:0,this.from=Math.min(A,i),this.to=Math.max(A,i)}nextInner(e,A){if(A<0?this.pos<=this.from:this.pos>=this.to)return this.value="",this.done=!0,this;e+=Math.max(0,A<0?this.pos-this.to:this.from-this.pos);let i=A<0?this.pos-this.from:this.to-this.pos;e>i&&(e=i),i-=e;let{value:n}=this.cursor.next(e);return this.pos+=(n.length+e)*A,this.value=n.length<=i?n:A<0?n.slice(n.length-i):n.slice(0,i),this.done=!this.value,this}next(e=0){return e<0?e=Math.max(e,this.from-this.pos):e>0&&(e=Math.min(e,this.to-this.pos)),this.nextInner(e,this.cursor.dir)}get lineBreak(){return this.cursor.lineBreak&&this.value!=""}},aw=class{constructor(e){this.inner=e,this.afterBreak=!0,this.value="",this.done=!1}next(e=0){let{done:A,lineBreak:i,value:n}=this.inner.next(e);return A&&this.afterBreak?(this.value="",this.afterBreak=!1):A?(this.done=!0,this.value=""):i?this.afterBreak?this.value="":(this.afterBreak=!0,this.next()):(this.value=n,this.afterBreak=!1),this}get lineBreak(){return!1}};typeof Symbol<"u"&&(In.prototype[Symbol.iterator]=function(){return this.iter()},pC.prototype[Symbol.iterator]=sw.prototype[Symbol.iterator]=aw.prototype[Symbol.iterator]=function(){return this});var Vx=class{constructor(e,A,i,n){this.from=e,this.to=A,this.number=i,this.text=n}get length(){return this.to-this.from}};function bE(t,e,A){return e=Math.max(0,Math.min(t.length,e)),[e,Math.max(e,Math.min(t.length,A))]}function yr(t,e,A=!0,i=!0){return HAA(t,e,A,i)}function ERA(t){return t>=56320&&t<57344}function QRA(t){return t>=55296&&t<56320}function Bs(t,e){let A=t.charCodeAt(e);if(!QRA(A)||e+1==t.length)return A;let i=t.charCodeAt(e+1);return ERA(i)?(A-55296<<10)+(i-56320)+65536:A}function T4(t){return t<=65535?String.fromCharCode(t):(t-=65536,String.fromCharCode((t>>10)+55296,(t&1023)+56320))}function lc(t){return t<65536?1:2}var Zx=/\r\n?|\n/,Is=function(t){return t[t.Simple=0]="Simple",t[t.TrackDel=1]="TrackDel",t[t.TrackBefore=2]="TrackBefore",t[t.TrackAfter=3]="TrackAfter",t}(Is||(Is={})),m1=class t{constructor(e){this.sections=e}get length(){let e=0;for(let A=0;A<this.sections.length;A+=2)e+=this.sections[A];return e}get newLength(){let e=0;for(let A=0;A<this.sections.length;A+=2){let i=this.sections[A+1];e+=i<0?this.sections[A]:i}return e}get empty(){return this.sections.length==0||this.sections.length==2&&this.sections[1]<0}iterGaps(e){for(let A=0,i=0,n=0;A<this.sections.length;){let o=this.sections[A++],r=this.sections[A++];r<0?(e(i,n,o),n+=o):n+=r,i+=o}}iterChangedRanges(e,A=!1){Wx(this,e,A)}get invertedDesc(){let e=[];for(let A=0;A<this.sections.length;){let i=this.sections[A++],n=this.sections[A++];n<0?e.push(i,n):e.push(n,i)}return new t(e)}composeDesc(e){return this.empty?e:e.empty?this:AeA(this,e)}mapDesc(e,A=!1){return e.empty?this:Xx(this,e,A)}mapPos(e,A=-1,i=Is.Simple){let n=0,o=0;for(let r=0;r<this.sections.length;){let s=this.sections[r++],a=this.sections[r++],c=n+s;if(a<0){if(c>e)return o+(e-n);o+=s}else{if(i!=Is.Simple&&c>=e&&(i==Is.TrackDel&&n<e&&c>e||i==Is.TrackBefore&&n<e||i==Is.TrackAfter&&c>e))return null;if(c>e||c==e&&A<0&&!s)return e==n||A<0?o:o+a;o+=a}n=c}if(e>n)throw new RangeError(`Position ${e} is out of range for changeset of length ${n}`);return o}touchesRange(e,A=e){for(let i=0,n=0;i<this.sections.length&&n<=A;){let o=this.sections[i++],r=this.sections[i++],s=n+o;if(r>=0&&n<=A&&s>=e)return n<e&&s>A?"cover":!0;n=s}return!1}toString(){let e="";for(let A=0;A<this.sections.length;){let i=this.sections[A++],n=this.sections[A++];e+=(e?" ":"")+i+(n>=0?":"+n:"")}return e}toJSON(){return this.sections}static fromJSON(e){if(!Array.isArray(e)||e.length%2||e.some(A=>typeof A!="number"))throw new RangeError("Invalid JSON representation of ChangeDesc");return new t(e)}static create(e){return new t(e)}},Cs=class t extends m1{constructor(e,A){super(e),this.inserted=A}apply(e){if(this.length!=e.length)throw new RangeError("Applying change set to a document with the wrong length");return Wx(this,(A,i,n,o,r)=>e=e.replace(n,n+(i-A),r),!1),e}mapDesc(e,A=!1){return Xx(this,e,A,!0)}invert(e){let A=this.sections.slice(),i=[];for(let n=0,o=0;n<A.length;n+=2){let r=A[n],s=A[n+1];if(s>=0){A[n]=s,A[n+1]=r;let a=n>>1;for(;i.length<a;)i.push(In.empty);i.push(r?e.slice(o,o+r):In.empty)}o+=r}return new t(A,i)}compose(e){return this.empty?e:e.empty?this:AeA(this,e,!0)}map(e,A=!1){return e.empty?this:Xx(this,e,A,!0)}iterChanges(e,A=!1){Wx(this,e,A)}get desc(){return m1.create(this.sections)}filter(e){let A=[],i=[],n=[],o=new wC(this);A:for(let r=0,s=0;;){let a=r==e.length?1e9:e[r++];for(;s<a||s==a&&o.len==0;){if(o.done)break A;let l=Math.min(o.len,a-s);Ls(n,l,-1);let I=o.ins==-1?-1:o.off==0?o.ins:0;Ls(A,l,I),I>0&&f1(i,A,o.text),o.forward(l),s+=l}let c=e[r++];for(;s<c;){if(o.done)break A;let l=Math.min(o.len,c-s);Ls(A,l,-1),Ls(n,l,o.ins==-1?-1:o.off==0?o.ins:0),o.forward(l),s+=l}}return{changes:new t(A,i),filtered:m1.create(n)}}toJSON(){let e=[];for(let A=0;A<this.sections.length;A+=2){let i=this.sections[A],n=this.sections[A+1];n<0?e.push(i):n==0?e.push([i]):e.push([i].concat(this.inserted[A>>1].toJSON()))}return e}static of(e,A,i){let n=[],o=[],r=0,s=null;function a(l=!1){if(!l&&!n.length)return;r<A&&Ls(n,A-r,-1);let I=new t(n,o);s=s?s.compose(I.map(s)):I,n=[],o=[],r=0}function c(l){if(Array.isArray(l))for(let I of l)c(I);else if(l instanceof t){if(l.length!=A)throw new RangeError(`Mismatched change set length (got ${l.length}, expected ${A})`);a(),s=s?s.compose(l.map(s)):l}else{let{from:I,to:C=I,insert:d}=l;if(I>C||I<0||C>A)throw new RangeError(`Invalid change range ${I} to ${C} (in doc of length ${A})`);let B=d?typeof d=="string"?In.of(d.split(i||Zx)):d:In.empty,E=B.length;if(I==C&&E==0)return;I<r&&a(),I>r&&Ls(n,I-r,-1),Ls(n,C-I,E),f1(o,n,B),r=C}}return c(e),a(!s),s}static empty(e){return new t(e?[e,-1]:[],[])}static fromJSON(e){if(!Array.isArray(e))throw new RangeError("Invalid JSON representation of ChangeSet");let A=[],i=[];for(let n=0;n<e.length;n++){let o=e[n];if(typeof o=="number")A.push(o,-1);else{if(!Array.isArray(o)||typeof o[0]!="number"||o.some((r,s)=>s&&typeof r!="string"))throw new RangeError("Invalid JSON representation of ChangeSet");if(o.length==1)A.push(o[0],0);else{for(;i.length<n;)i.push(In.empty);i[n]=In.of(o.slice(1)),A.push(o[0],i[n].length)}}}return new t(A,i)}static createSet(e,A){return new t(e,A)}};function Ls(t,e,A,i=!1){if(e==0&&A<=0)return;let n=t.length-2;n>=0&&A<=0&&A==t[n+1]?t[n]+=e:n>=0&&e==0&&t[n]==0?t[n+1]+=A:i?(t[n]+=e,t[n+1]+=A):t.push(e,A)}function f1(t,e,A){if(A.length==0)return;let i=e.length-2>>1;if(i<t.length)t[t.length-1]=t[t.length-1].append(A);else{for(;t.length<i;)t.push(In.empty);t.push(A)}}function Wx(t,e,A){let i=t.inserted;for(let n=0,o=0,r=0;r<t.sections.length;){let s=t.sections[r++],a=t.sections[r++];if(a<0)n+=s,o+=s;else{let c=n,l=o,I=In.empty;for(;c+=s,l+=a,a&&i&&(I=I.append(i[r-2>>1])),!(A||r==t.sections.length||t.sections[r+1]<0);)s=t.sections[r++],a=t.sections[r++];e(n,c,o,l,I),n=c,o=l}}}function Xx(t,e,A,i=!1){let n=[],o=i?[]:null,r=new wC(t),s=new wC(e);for(let a=-1;;){if(r.done&&s.len||s.done&&r.len)throw new Error("Mismatched change set lengths");if(r.ins==-1&&s.ins==-1){let c=Math.min(r.len,s.len);Ls(n,c,-1),r.forward(c),s.forward(c)}else if(s.ins>=0&&(r.ins<0||a==r.i||r.off==0&&(s.len<r.len||s.len==r.len&&!A))){let c=s.len;for(Ls(n,s.ins,-1);c;){let l=Math.min(r.len,c);r.ins>=0&&a<r.i&&r.len<=l&&(Ls(n,0,r.ins),o&&f1(o,n,r.text),a=r.i),r.forward(l),c-=l}s.next()}else if(r.ins>=0){let c=0,l=r.len;for(;l;)if(s.ins==-1){let I=Math.min(l,s.len);c+=I,l-=I,s.forward(I)}else if(s.ins==0&&s.len<l)l-=s.len,s.next();else break;Ls(n,c,a<r.i?r.ins:0),o&&a<r.i&&f1(o,n,r.text),a=r.i,r.forward(r.len-l)}else{if(r.done&&s.done)return o?Cs.createSet(n,o):m1.create(n);throw new Error("Mismatched change set lengths")}}}function AeA(t,e,A=!1){let i=[],n=A?[]:null,o=new wC(t),r=new wC(e);for(let s=!1;;){if(o.done&&r.done)return n?Cs.createSet(i,n):m1.create(i);if(o.ins==0)Ls(i,o.len,0,s),o.next();else if(r.len==0&&!r.done)Ls(i,0,r.ins,s),n&&f1(n,i,r.text),r.next();else{if(o.done||r.done)throw new Error("Mismatched change set lengths");{let a=Math.min(o.len2,r.len),c=i.length;if(o.ins==-1){let l=r.ins==-1?-1:r.off?0:r.ins;Ls(i,a,l,s),n&&l&&f1(n,i,r.text)}else r.ins==-1?(Ls(i,o.off?0:o.len,a,s),n&&f1(n,i,o.textBit(a))):(Ls(i,o.off?0:o.len,r.off?0:r.ins,s),n&&!r.off&&f1(n,i,r.text));s=(o.ins>a||r.ins>=0&&r.len>a)&&(s||i.length>c),o.forward2(a),r.forward(a)}}}}var wC=class{constructor(e){this.set=e,this.i=0,this.next()}next(){let{sections:e}=this.set;this.i<e.length?(this.len=e[this.i++],this.ins=e[this.i++]):(this.len=0,this.ins=-2),this.off=0}get done(){return this.ins==-2}get len2(){return this.ins<0?this.len:this.ins}get text(){let{inserted:e}=this.set,A=this.i-2>>1;return A>=e.length?In.empty:e[A]}textBit(e){let{inserted:A}=this.set,i=this.i-2>>1;return i>=A.length&&!e?In.empty:A[i].slice(this.off,e==null?void 0:this.off+e)}forward(e){e==this.len?this.next():(this.len-=e,this.off+=e)}forward2(e){this.ins==-1?this.forward(e):e==this.ins?this.next():(this.ins-=e,this.off+=e)}},wE=class t{constructor(e,A,i){this.from=e,this.to=A,this.flags=i}get anchor(){return this.flags&32?this.to:this.from}get head(){return this.flags&32?this.from:this.to}get empty(){return this.from==this.to}get assoc(){return this.flags&8?-1:this.flags&16?1:0}get bidiLevel(){let e=this.flags&7;return e==7?null:e}get goalColumn(){let e=this.flags>>6;return e==16777215?void 0:e}map(e,A=-1){let i,n;return this.empty?i=n=e.mapPos(this.from,A):(i=e.mapPos(this.from,1),n=e.mapPos(this.to,-1)),i==this.from&&n==this.to?this:new t(i,n,this.flags)}extend(e,A=e){if(e<=this.anchor&&A>=this.anchor)return ae.range(e,A);let i=Math.abs(e-this.anchor)>Math.abs(A-this.anchor)?e:A;return ae.range(this.anchor,i)}eq(e,A=!1){return this.anchor==e.anchor&&this.head==e.head&&(!A||!this.empty||this.assoc==e.assoc)}toJSON(){return{anchor:this.anchor,head:this.head}}static fromJSON(e){if(!e||typeof e.anchor!="number"||typeof e.head!="number")throw new RangeError("Invalid JSON representation for SelectionRange");return ae.range(e.anchor,e.head)}static create(e,A,i){return new t(e,A,i)}},ae=class t{constructor(e,A){this.ranges=e,this.mainIndex=A}map(e,A=-1){return e.empty?this:t.create(this.ranges.map(i=>i.map(e,A)),this.mainIndex)}eq(e,A=!1){if(this.ranges.length!=e.ranges.length||this.mainIndex!=e.mainIndex)return!1;for(let i=0;i<this.ranges.length;i++)if(!this.ranges[i].eq(e.ranges[i],A))return!1;return!0}get main(){return this.ranges[this.mainIndex]}asSingle(){return this.ranges.length==1?this:new t([this.main],0)}addRange(e,A=!0){return t.create([e].concat(this.ranges),A?0:this.mainIndex+1)}replaceRange(e,A=this.mainIndex){let i=this.ranges.slice();return i[A]=e,t.create(i,this.mainIndex)}toJSON(){return{ranges:this.ranges.map(e=>e.toJSON()),main:this.mainIndex}}static fromJSON(e){if(!e||!Array.isArray(e.ranges)||typeof e.main!="number"||e.main>=e.ranges.length)throw new RangeError("Invalid JSON representation for EditorSelection");return new t(e.ranges.map(A=>wE.fromJSON(A)),e.main)}static single(e,A=e){return new t([t.range(e,A)],0)}static create(e,A=0){if(e.length==0)throw new RangeError("A selection needs at least one range");for(let i=0,n=0;n<e.length;n++){let o=e[n];if(o.empty?o.from<=i:o.from<i)return t.normalized(e.slice(),A);i=o.to}return new t(e,A)}static cursor(e,A=0,i,n){return wE.create(e,e,(A==0?0:A<0?8:16)|(i==null?7:Math.min(6,i))|(n??16777215)<<6)}static range(e,A,i,n){let o=(i??16777215)<<6|(n==null?7:Math.min(6,n));return A<e?wE.create(A,e,48|o):wE.create(e,A,(A>e?8:0)|o)}static normalized(e,A=0){let i=e[A];e.sort((n,o)=>n.from-o.from),A=e.indexOf(i);for(let n=1;n<e.length;n++){let o=e[n],r=e[n-1];if(o.empty?o.from<=r.to:o.from<r.to){let s=r.from,a=Math.max(o.to,r.to);n<=A&&A--,e.splice(--n,2,o.anchor>o.head?t.range(a,s):t.range(s,a))}}return new t(e,A)}};function eeA(t,e){for(let A of t.ranges)if(A.to>e)throw new RangeError("Selection points outside of document")}var sL=0,qe=class t{constructor(e,A,i,n,o){this.combine=e,this.compareInput=A,this.compare=i,this.isStatic=n,this.id=sL++,this.default=e([]),this.extensions=typeof o=="function"?o(this):o}get reader(){return this}static define(e={}){return new t(e.combine||(A=>A),e.compareInput||((A,i)=>A===i),e.compare||(e.combine?(A,i)=>A===i:aL),!!e.static,e.enables)}of(e){return new yE([],this,0,e)}compute(e,A){if(this.isStatic)throw new Error("Can't compute a static facet");return new yE(e,this,1,A)}computeN(e,A){if(this.isStatic)throw new Error("Can't compute a static facet");return new yE(e,this,2,A)}from(e,A){return A||(A=i=>i),this.compute([e],i=>A(i.field(e)))}};function aL(t,e){return t==e||t.length==e.length&&t.every((A,i)=>A===e[i])}var yE=class{constructor(e,A,i,n){this.dependencies=e,this.facet=A,this.type=i,this.value=n,this.id=sL++}dynamicSlot(e){var A;let i=this.value,n=this.facet.compareInput,o=this.id,r=e[o]>>1,s=this.type==2,a=!1,c=!1,l=[];for(let I of this.dependencies)I=="doc"?a=!0:I=="selection"?c=!0:(((A=e[I.id])!==null&&A!==void 0?A:1)&1)==0&&l.push(e[I.id]);return{create(I){return I.values[r]=i(I),1},update(I,C){if(a&&C.docChanged||c&&(C.docChanged||C.selection)||$x(I,l)){let d=i(I);if(s?!qAA(d,I.values[r],n):!n(d,I.values[r]))return I.values[r]=d,1}return 0},reconfigure:(I,C)=>{let d,B=C.config.address[o];if(B!=null){let E=gw(C,B);if(this.dependencies.every(h=>h instanceof qe?C.facet(h)===I.facet(h):h instanceof Ar?C.field(h,!1)==I.field(h,!1):!0)||(s?qAA(d=i(I),E,n):n(d=i(I),E)))return I.values[r]=E,0}else d=i(I);return I.values[r]=d,1}}}};function qAA(t,e,A){if(t.length!=e.length)return!1;for(let i=0;i<t.length;i++)if(!A(t[i],e[i]))return!1;return!0}function $x(t,e){let A=!1;for(let i of e)U4(t,i)&1&&(A=!0);return A}function hRA(t,e,A){let i=A.map(a=>t[a.id]),n=A.map(a=>a.type),o=i.filter(a=>!(a&1)),r=t[e.id]>>1;function s(a){let c=[];for(let l=0;l<i.length;l++){let I=gw(a,i[l]);if(n[l]==2)for(let C of I)c.push(C);else c.push(I)}return e.combine(c)}return{create(a){for(let c of i)U4(a,c);return a.values[r]=s(a),1},update(a,c){if(!$x(a,o))return 0;let l=s(a);return e.compare(l,a.values[r])?0:(a.values[r]=l,1)},reconfigure(a,c){let l=$x(a,i),I=c.config.facets[e.id],C=c.facet(e);if(I&&!l&&aL(A,I))return a.values[r]=C,0;let d=s(a);return e.compare(d,C)?(a.values[r]=C,0):(a.values[r]=d,1)}}}var iw=qe.define({static:!0}),Ar=class t{constructor(e,A,i,n,o){this.id=e,this.createF=A,this.updateF=i,this.compareF=n,this.spec=o,this.provides=void 0}static define(e){let A=new t(sL++,e.create,e.update,e.compare||((i,n)=>i===n),e);return e.provide&&(A.provides=e.provide(A)),A}create(e){let A=e.facet(iw).find(i=>i.field==this);return(A?.create||this.createF)(e)}slot(e){let A=e[this.id]>>1;return{create:i=>(i.values[A]=this.create(i),1),update:(i,n)=>{let o=i.values[A],r=this.updateF(o,n);return this.compareF(o,r)?0:(i.values[A]=r,1)},reconfigure:(i,n)=>{let o=i.facet(iw),r=n.facet(iw),s;return(s=o.find(a=>a.field==this))&&s!=r.find(a=>a.field==this)?(i.values[A]=s.create(i),1):n.config.address[this.id]!=null?(i.values[A]=n.field(this),0):(i.values[A]=this.create(i),1)}}}init(e){return[this,iw.of({field:this,create:e})]}get extension(){return this}},fC={lowest:4,low:3,default:2,high:1,highest:0};function G4(t){return e=>new cw(e,t)}var Dl={highest:G4(fC.highest),high:G4(fC.high),default:G4(fC.default),low:G4(fC.low),lowest:G4(fC.lowest)},cw=class{constructor(e,A){this.inner=e,this.prec=A}},bg=class t{of(e){return new K4(this,e)}reconfigure(e){return t.reconfigure.of({compartment:this,extension:e})}get(e){return e.config.compartments.get(this)}},K4=class{constructor(e,A){this.compartment=e,this.inner=A}},lw=class t{constructor(e,A,i,n,o,r){for(this.base=e,this.compartments=A,this.dynamicSlots=i,this.address=n,this.staticValues=o,this.facets=r,this.statusTemplate=[];this.statusTemplate.length<i.length;)this.statusTemplate.push(0)}staticFacet(e){let A=this.address[e.id];return A==null?e.default:this.staticValues[A>>1]}static resolve(e,A,i){let n=[],o=Object.create(null),r=new Map;for(let C of uRA(e,A,r))C instanceof Ar?n.push(C):(o[C.facet.id]||(o[C.facet.id]=[])).push(C);let s=Object.create(null),a=[],c=[];for(let C of n)s[C.id]=c.length<<1,c.push(d=>C.slot(d));let l=i?.config.facets;for(let C in o){let d=o[C],B=d[0].facet,E=l&&l[C]||[];if(d.every(h=>h.type==0))if(s[B.id]=a.length<<1|1,aL(E,d))a.push(i.facet(B));else{let h=B.combine(d.map(u=>u.value));a.push(i&&B.compare(h,i.facet(B))?i.facet(B):h)}else{for(let h of d)h.type==0?(s[h.id]=a.length<<1|1,a.push(h.value)):(s[h.id]=c.length<<1,c.push(u=>h.dynamicSlot(u)));s[B.id]=c.length<<1,c.push(h=>hRA(h,B,d))}}let I=c.map(C=>C(s));return new t(e,r,I,s,a,o)}};function uRA(t,e,A){let i=[[],[],[],[],[]],n=new Map;function o(r,s){let a=n.get(r);if(a!=null){if(a<=s)return;let c=i[a].indexOf(r);c>-1&&i[a].splice(c,1),r instanceof K4&&A.delete(r.compartment)}if(n.set(r,s),Array.isArray(r))for(let c of r)o(c,s);else if(r instanceof K4){if(A.has(r.compartment))throw new RangeError("Duplicate use of compartment in extensions");let c=e.get(r.compartment)||r.inner;A.set(r.compartment,c),o(c,s)}else if(r instanceof cw)o(r.inner,r.prec);else if(r instanceof Ar)i[s].push(r),r.provides&&o(r.provides,s);else if(r instanceof yE)i[s].push(r),r.facet.extensions&&o(r.facet.extensions,fC.default);else{let c=r.extension;if(!c)throw new Error(`Unrecognized extension value in extension set (${r}). This sometimes happens because multiple instances of @codemirror/state are loaded, breaking instanceof checks.`);o(c,s)}}return o(t,fC.default),i.reduce((r,s)=>r.concat(s))}function U4(t,e){if(e&1)return 2;let A=e>>1,i=t.status[A];if(i==4)throw new Error("Cyclic dependency between fields and/or facets");if(i&2)return i;t.status[A]=4;let n=t.computeSlot(t,t.config.dynamicSlots[A]);return t.status[A]=2|n}function gw(t,e){return e&1?t.config.staticValues[e>>1]:t.values[e>>1]}var VAA=qe.define(),jx=qe.define({combine:t=>t.some(e=>e),static:!0}),teA=qe.define({combine:t=>t.length?t[0]:void 0,static:!0}),ieA=qe.define(),neA=qe.define(),oeA=qe.define(),ZAA=qe.define({combine:t=>t.length?t[0]:!1}),xa=class{constructor(e,A){this.type=e,this.value=A}static define(){return new AL}},AL=class{of(e){return new xa(this,e)}},eL=class{constructor(e){this.map=e}of(e){return new Pi(this,e)}},Pi=(()=>{class t{constructor(A,i){this.type=A,this.value=i}map(A){let i=this.type.map(this.value,A);return i===void 0?void 0:i==this.value?this:new t(this.type,i)}is(A){return this.type==A}static define(A={}){return new eL(A.map||(i=>i))}static mapEffects(A,i){if(!A.length)return A;let n=[];for(let o of A){let r=o.map(i);r&&n.push(r)}return n}}return t.reconfigure=t.define(),t.appendConfig=t.define(),t})(),vg=(()=>{class t{constructor(A,i,n,o,r,s){this.startState=A,this.changes=i,this.selection=n,this.effects=o,this.annotations=r,this.scrollIntoView=s,this._doc=null,this._state=null,n&&eeA(n,i.newLength),r.some(a=>a.type==t.time)||(this.annotations=r.concat(t.time.of(Date.now())))}static create(A,i,n,o,r,s){return new t(A,i,n,o,r,s)}get newDoc(){return this._doc||(this._doc=this.changes.apply(this.startState.doc))}get newSelection(){return this.selection||this.startState.selection.map(this.changes)}get state(){return this._state||this.startState.applyTransaction(this),this._state}annotation(A){for(let i of this.annotations)if(i.type==A)return i.value}get docChanged(){return!this.changes.empty}get reconfigured(){return this.startState.config!=this.state.config}isUserEvent(A){let i=this.annotation(t.userEvent);return!!(i&&(i==A||i.length>A.length&&i.slice(0,A.length)==A&&i[A.length]=="."))}}return t.time=xa.define(),t.userEvent=xa.define(),t.addToHistory=xa.define(),t.remote=xa.define(),t})();function fRA(t,e){let A=[];for(let i=0,n=0;;){let o,r;if(i<t.length&&(n==e.length||e[n]>=t[i]))o=t[i++],r=t[i++];else if(n<e.length)o=e[n++],r=e[n++];else return A;!A.length||A[A.length-1]<o?A.push(o,r):A[A.length-1]<r&&(A[A.length-1]=r)}}function reA(t,e,A){var i;let n,o,r;return A?(n=e.changes,o=Cs.empty(e.changes.length),r=t.changes.compose(e.changes)):(n=e.changes.map(t.changes),o=t.changes.mapDesc(e.changes,!0),r=t.changes.compose(n)),{changes:r,selection:e.selection?e.selection.map(o):(i=t.selection)===null||i===void 0?void 0:i.map(n),effects:Pi.mapEffects(t.effects,n).concat(Pi.mapEffects(e.effects,o)),annotations:t.annotations.length?t.annotations.concat(e.annotations):e.annotations,scrollIntoView:t.scrollIntoView||e.scrollIntoView}}function tL(t,e,A){let i=e.selection,n=vE(e.annotations);return e.userEvent&&(n=n.concat(vg.userEvent.of(e.userEvent))),{changes:e.changes instanceof Cs?e.changes:Cs.of(e.changes||[],A,t.facet(teA)),selection:i&&(i instanceof ae?i:ae.single(i.anchor,i.head)),effects:vE(e.effects),annotations:n,scrollIntoView:!!e.scrollIntoView}}function seA(t,e,A){let i=tL(t,e.length?e[0]:{},t.doc.length);e.length&&e[0].filter===!1&&(A=!1);for(let o=1;o<e.length;o++){e[o].filter===!1&&(A=!1);let r=!!e[o].sequential;i=reA(i,tL(t,e[o],r?i.changes.newLength:t.doc.length),r)}let n=vg.create(t,i.changes,i.selection,i.effects,i.annotations,i.scrollIntoView);return pRA(A?mRA(n):n)}function mRA(t){let e=t.startState,A=!0;for(let n of e.facet(ieA)){let o=n(t);if(o===!1){A=!1;break}Array.isArray(o)&&(A=A===!0?o:fRA(A,o))}if(A!==!0){let n,o;if(A===!1)o=t.changes.invertedDesc,n=Cs.empty(e.doc.length);else{let r=t.changes.filter(A);n=r.changes,o=r.filtered.mapDesc(r.changes).invertedDesc}t=vg.create(e,n,t.selection&&t.selection.map(o),Pi.mapEffects(t.effects,o),t.annotations,t.scrollIntoView)}let i=e.facet(neA);for(let n=i.length-1;n>=0;n--){let o=i[n](t);o instanceof vg?t=o:Array.isArray(o)&&o.length==1&&o[0]instanceof vg?t=o[0]:t=seA(e,vE(o),!1)}return t}function pRA(t){let e=t.startState,A=e.facet(oeA),i=t;for(let n=A.length-1;n>=0;n--){let o=A[n](t);o&&Object.keys(o).length&&(i=reA(i,tL(e,o,t.changes.newLength),!0))}return i==t?t:vg.create(e,t.changes,t.selection,i.effects,i.annotations,i.scrollIntoView)}var wRA=[];function vE(t){return t==null?wRA:Array.isArray(t)?t:[t]}var ao=function(t){return t[t.Word=0]="Word",t[t.Space=1]="Space",t[t.Other=2]="Other",t}(ao||(ao={})),DRA=/[\u00df\u0587\u0590-\u05f4\u0600-\u06ff\u3040-\u309f\u30a0-\u30ff\u3400-\u4db5\u4e00-\u9fcc\uac00-\ud7af]/,iL;try{iL=new RegExp("[\\p{Alphabetic}\\p{Number}_]","u")}catch{}function yRA(t){if(iL)return iL.test(t);for(let e=0;e<t.length;e++){let A=t[e];if(/\w/.test(A)||A>"\x80"&&(A.toUpperCase()!=A.toLowerCase()||DRA.test(A)))return!0}return!1}function vRA(t){return e=>{if(!/\S/.test(e))return ao.Space;if(yRA(e))return ao.Word;for(let A=0;A<t.length;A++)if(e.indexOf(t[A])>-1)return ao.Word;return ao.Other}}var hr=(()=>{class t{constructor(A,i,n,o,r,s){this.config=A,this.doc=i,this.selection=n,this.values=o,this.status=A.statusTemplate.slice(),this.computeSlot=r,s&&(s._state=this);for(let a=0;a<this.config.dynamicSlots.length;a++)U4(this,a<<1);this.computeSlot=null}field(A,i=!0){let n=this.config.address[A.id];if(n==null){if(i)throw new RangeError("Field is not present in this state");return}return U4(this,n),gw(this,n)}update(...A){return seA(this,A,!0)}applyTransaction(A){let i=this.config,{base:n,compartments:o}=i;for(let a of A.effects)a.is(bg.reconfigure)?(i&&(o=new Map,i.compartments.forEach((c,l)=>o.set(l,c)),i=null),o.set(a.value.compartment,a.value.extension)):a.is(Pi.reconfigure)?(i=null,n=a.value):a.is(Pi.appendConfig)&&(i=null,n=vE(n).concat(a.value));let r;i?r=A.startState.values.slice():(i=lw.resolve(n,o,this),r=new t(i,this.doc,this.selection,i.dynamicSlots.map(()=>null),(c,l)=>l.reconfigure(c,this),null).values);let s=A.startState.facet(jx)?A.newSelection:A.newSelection.asSingle();new t(i,A.newDoc,s,r,(a,c)=>c.update(a,A),A)}replaceSelection(A){return typeof A=="string"&&(A=this.toText(A)),this.changeByRange(i=>({changes:{from:i.from,to:i.to,insert:A},range:ae.cursor(i.from+A.length)}))}changeByRange(A){let i=this.selection,n=A(i.ranges[0]),o=this.changes(n.changes),r=[n.range],s=vE(n.effects);for(let a=1;a<i.ranges.length;a++){let c=A(i.ranges[a]),l=this.changes(c.changes),I=l.map(o);for(let d=0;d<a;d++)r[d]=r[d].map(I);let C=o.mapDesc(l,!0);r.push(c.range.map(C)),o=o.compose(I),s=Pi.mapEffects(s,I).concat(Pi.mapEffects(vE(c.effects),C))}return{changes:o,selection:ae.create(r,i.mainIndex),effects:s}}changes(A=[]){return A instanceof Cs?A:Cs.of(A,this.doc.length,this.facet(t.lineSeparator))}toText(A){return In.of(A.split(this.facet(t.lineSeparator)||Zx))}sliceDoc(A=0,i=this.doc.length){return this.doc.sliceString(A,i,this.lineBreak)}facet(A){let i=this.config.address[A.id];return i==null?A.default:(U4(this,i),gw(this,i))}toJSON(A){let i={doc:this.sliceDoc(),selection:this.selection.toJSON()};if(A)for(let n in A){let o=A[n];o instanceof Ar&&this.config.address[o.id]!=null&&(i[n]=o.spec.toJSON(this.field(A[n]),this))}return i}static fromJSON(A,i={},n){if(!A||typeof A.doc!="string")throw new RangeError("Invalid JSON representation for EditorState");let o=[];if(n){for(let r in n)if(Object.prototype.hasOwnProperty.call(A,r)){let s=n[r],a=A[r];o.push(s.init(c=>s.spec.fromJSON(a,c)))}}return t.create({doc:A.doc,selection:ae.fromJSON(A.selection),extensions:i.extensions?o.concat([i.extensions]):o})}static create(A={}){let i=lw.resolve(A.extensions||[],new Map),n=A.doc instanceof In?A.doc:In.of((A.doc||"").split(i.staticFacet(t.lineSeparator)||Zx)),o=A.selection?A.selection instanceof ae?A.selection:ae.single(A.selection.anchor,A.selection.head):ae.single(0);return eeA(o,n.length),i.staticFacet(jx)||(o=o.asSingle()),new t(i,n,o,i.dynamicSlots.map(()=>null),(r,s)=>s.create(r),null)}get tabSize(){return this.facet(t.tabSize)}get lineBreak(){return this.facet(t.lineSeparator)||`
`}get readOnly(){return this.facet(ZAA)}phrase(A,...i){for(let n of this.facet(t.phrases))if(Object.prototype.hasOwnProperty.call(n,A)){A=n[A];break}return i.length&&(A=A.replace(/\$(\$|\d*)/g,(n,o)=>{if(o=="$")return"$";let r=+(o||1);return!r||r>i.length?n:i[r-1]})),A}languageDataAt(A,i,n=-1){let o=[];for(let r of this.facet(VAA))for(let s of r(this,i,n))Object.prototype.hasOwnProperty.call(s,A)&&o.push(s[A]);return o}charCategorizer(A){return vRA(this.languageDataAt("wordChars",A).join(""))}wordAt(A){let{text:i,from:n,length:o}=this.doc.lineAt(A),r=this.charCategorizer(A),s=A-n,a=A-n;for(;s>0;){let c=yr(i,s,!1);if(r(i.slice(c,s))!=ao.Word)break;s=c}for(;a<o;){let c=yr(i,a);if(r(i.slice(a,c))!=ao.Word)break;a=c}return s==a?null:ae.range(s+n,a+n)}}return t.allowMultipleSelections=jx,t.tabSize=qe.define({combine:e=>e.length?e[0]:4}),t.lineSeparator=teA,t.readOnly=ZAA,t.phrases=qe.define({compare(e,A){let i=Object.keys(e),n=Object.keys(A);return i.length==n.length&&i.every(o=>e[o]==A[o])}}),t.languageData=VAA,t.changeFilter=ieA,t.transactionFilter=neA,t.transactionExtender=oeA,t})();bg.reconfigure=Pi.define();function Wr(t,e,A={}){let i={};for(let n of t)for(let o of Object.keys(n)){let r=n[o],s=i[o];if(s===void 0)i[o]=r;else if(!(s===r||r===void 0))if(Object.hasOwnProperty.call(A,o))i[o]=A[o](s,r);else throw new Error("Config merge conflict for field "+o)}for(let n in e)i[n]===void 0&&(i[n]=e[n]);return i}var wl=class{eq(e){return this==e}range(e,A=e){return Y4.create(e,A,this)}};wl.prototype.startSide=wl.prototype.endSide=0;wl.prototype.point=!1;wl.prototype.mapMode=Is.TrackDel;var Y4=class t{constructor(e,A,i){this.from=e,this.to=A,this.value=i}static create(e,A,i){return new t(e,A,i)}};function nL(t,e){return t.from-e.from||t.value.startSide-e.value.startSide}var oL=class t{constructor(e,A,i,n){this.from=e,this.to=A,this.value=i,this.maxPoint=n}get length(){return this.to[this.to.length-1]}findIndex(e,A,i,n=0){let o=i?this.to:this.from;for(let r=n,s=o.length;;){if(r==s)return r;let a=r+s>>1,c=o[a]-e||(i?this.value[a].endSide:this.value[a].startSide)-A;if(a==r)return c>=0?r:s;c>=0?s=a:r=a+1}}between(e,A,i,n){for(let o=this.findIndex(A,-1e9,!0),r=this.findIndex(i,1e9,!1,o);o<r;o++)if(n(this.from[o]+e,this.to[o]+e,this.value[o])===!1)return!1}map(e,A){let i=[],n=[],o=[],r=-1,s=-1;for(let a=0;a<this.value.length;a++){let c=this.value[a],l=this.from[a]+e,I=this.to[a]+e,C,d;if(l==I){let B=A.mapPos(l,c.startSide,c.mapMode);if(B==null||(C=d=B,c.startSide!=c.endSide&&(d=A.mapPos(l,c.endSide),d<C)))continue}else if(C=A.mapPos(l,c.startSide),d=A.mapPos(I,c.endSide),C>d||C==d&&c.startSide>0&&c.endSide<=0)continue;(d-C||c.endSide-c.startSide)<0||(r<0&&(r=C),c.point&&(s=Math.max(s,d-C)),i.push(c),n.push(C-r),o.push(d-r))}return{mapped:i.length?new t(n,o,i,s):null,pos:r}}},co=(()=>{class t{constructor(A,i,n,o){this.chunkPos=A,this.chunk=i,this.nextLayer=n,this.maxPoint=o}static create(A,i,n,o){return new t(A,i,n,o)}get length(){let A=this.chunk.length-1;return A<0?0:Math.max(this.chunkEnd(A),this.nextLayer.length)}get size(){if(this.isEmpty)return 0;let A=this.nextLayer.size;for(let i of this.chunk)A+=i.value.length;return A}chunkEnd(A){return this.chunkPos[A]+this.chunk[A].length}update(A){let{add:i=[],sort:n=!1,filterFrom:o=0,filterTo:r=this.length}=A,s=A.filter;if(i.length==0&&!s)return this;if(n&&(i=i.slice().sort(nL)),this.isEmpty)return i.length?t.of(i):this;let a=new Iw(this,null,-1).goto(0),c=0,l=[],I=new ds;for(;a.value||c<i.length;)if(c<i.length&&(a.from-i[c].from||a.startSide-i[c].value.startSide)>=0){let C=i[c++];I.addInner(C.from,C.to,C.value)||l.push(C)}else a.rangeIndex==1&&a.chunkIndex<this.chunk.length&&(c==i.length||this.chunkEnd(a.chunkIndex)<i[c].from)&&(!s||o>this.chunkEnd(a.chunkIndex)||r<this.chunkPos[a.chunkIndex])&&I.addChunk(this.chunkPos[a.chunkIndex],this.chunk[a.chunkIndex])?a.nextChunk():((!s||o>a.to||r<a.from||s(a.from,a.to,a.value))&&(I.addInner(a.from,a.to,a.value)||l.push(Y4.create(a.from,a.to,a.value))),a.next());return I.finishInner(this.nextLayer.isEmpty&&!l.length?t.empty:this.nextLayer.update({add:l,filter:s,filterFrom:o,filterTo:r}))}map(A){if(A.empty||this.isEmpty)return this;let i=[],n=[],o=-1;for(let s=0;s<this.chunk.length;s++){let a=this.chunkPos[s],c=this.chunk[s],l=A.touchesRange(a,a+c.length);if(l===!1)o=Math.max(o,c.maxPoint),i.push(c),n.push(A.mapPos(a));else if(l===!0){let{mapped:I,pos:C}=c.map(a,A);I&&(o=Math.max(o,I.maxPoint),i.push(I),n.push(C))}}let r=this.nextLayer.map(A);return i.length==0?r:new t(n,i,r||t.empty,o)}between(A,i,n){if(!this.isEmpty){for(let o=0;o<this.chunk.length;o++){let r=this.chunkPos[o],s=this.chunk[o];if(i>=r&&A<=r+s.length&&s.between(r,A-r,i-r,n)===!1)return}this.nextLayer.between(A,i,n)}}iter(A=0){return J4.from([this]).goto(A)}get isEmpty(){return this.nextLayer==this}static iter(A,i=0){return J4.from(A).goto(i)}static compare(A,i,n,o,r=-1){let s=A.filter(C=>C.maxPoint>0||!C.isEmpty&&C.maxPoint>=r),a=i.filter(C=>C.maxPoint>0||!C.isEmpty&&C.maxPoint>=r),c=WAA(s,a,n),l=new mC(s,c,r),I=new mC(a,c,r);n.iterGaps((C,d,B)=>XAA(l,C,I,d,B,o)),n.empty&&n.length==0&&XAA(l,0,I,0,0,o)}static eq(A,i,n=0,o){o==null&&(o=999999999);let r=A.filter(I=>!I.isEmpty&&i.indexOf(I)<0),s=i.filter(I=>!I.isEmpty&&A.indexOf(I)<0);if(r.length!=s.length)return!1;if(!r.length)return!0;let a=WAA(r,s),c=new mC(r,a,0).goto(n),l=new mC(s,a,0).goto(n);for(;;){if(c.to!=l.to||!rL(c.active,l.active)||c.point&&(!l.point||!c.point.eq(l.point)))return!1;if(c.to>o)return!0;c.next(),l.next()}}static spans(A,i,n,o,r=-1){let s=new mC(A,null,r).goto(i),a=i,c=s.openStart;for(;;){let l=Math.min(s.to,n);if(s.point){let I=s.activeForPoint(s.to),C=s.pointFrom<i?I.length+1:s.point.startSide<0?I.length:Math.min(I.length,c);o.point(a,l,s.point,I,C,s.pointRank),c=Math.min(s.openEnd(l),I.length)}else l>a&&(o.span(a,l,s.active,c),c=s.openEnd(l));if(s.to>n)return c+(s.point&&s.to>n?1:0);a=s.to,s.next()}}static of(A,i=!1){let n=new ds;for(let o of A instanceof Y4?[A]:i?bRA(A):A)n.add(o.from,o.to,o.value);return n.finish()}static join(A){if(!A.length)return t.empty;let i=A[A.length-1];for(let n=A.length-2;n>=0;n--)for(let o=A[n];o!=t.empty;o=o.nextLayer)i=new t(o.chunkPos,o.chunk,i,Math.max(o.maxPoint,i.maxPoint));return i}}return t.empty=new t([],[],null,-1),t})();function bRA(t){if(t.length>1)for(let e=t[0],A=1;A<t.length;A++){let i=t[A];if(nL(e,i)>0)return t.slice().sort(nL);e=i}return t}co.empty.nextLayer=co.empty;var ds=class t{finishChunk(e){this.chunks.push(new oL(this.from,this.to,this.value,this.maxPoint)),this.chunkPos.push(this.chunkStart),this.chunkStart=-1,this.setMaxPoint=Math.max(this.setMaxPoint,this.maxPoint),this.maxPoint=-1,e&&(this.from=[],this.to=[],this.value=[])}constructor(){this.chunks=[],this.chunkPos=[],this.chunkStart=-1,this.last=null,this.lastFrom=-1e9,this.lastTo=-1e9,this.from=[],this.to=[],this.value=[],this.maxPoint=-1,this.setMaxPoint=-1,this.nextLayer=null}add(e,A,i){this.addInner(e,A,i)||(this.nextLayer||(this.nextLayer=new t)).add(e,A,i)}addInner(e,A,i){let n=e-this.lastTo||i.startSide-this.last.endSide;if(n<=0&&(e-this.lastFrom||i.startSide-this.last.startSide)<0)throw new Error("Ranges must be added sorted by `from` position and `startSide`");return n<0?!1:(this.from.length==250&&this.finishChunk(!0),this.chunkStart<0&&(this.chunkStart=e),this.from.push(e-this.chunkStart),this.to.push(A-this.chunkStart),this.last=i,this.lastFrom=e,this.lastTo=A,this.value.push(i),i.point&&(this.maxPoint=Math.max(this.maxPoint,A-e)),!0)}addChunk(e,A){if((e-this.lastTo||A.value[0].startSide-this.last.endSide)<0)return!1;this.from.length&&this.finishChunk(!0),this.setMaxPoint=Math.max(this.setMaxPoint,A.maxPoint),this.chunks.push(A),this.chunkPos.push(e);let i=A.value.length-1;return this.last=A.value[i],this.lastFrom=A.from[i]+e,this.lastTo=A.to[i]+e,!0}finish(){return this.finishInner(co.empty)}finishInner(e){if(this.from.length&&this.finishChunk(!1),this.chunks.length==0)return e;let A=co.create(this.chunkPos,this.chunks,this.nextLayer?this.nextLayer.finishInner(e):e,this.setMaxPoint);return this.from=null,A}};function WAA(t,e,A){let i=new Map;for(let o of t)for(let r=0;r<o.chunk.length;r++)o.chunk[r].maxPoint<=0&&i.set(o.chunk[r],o.chunkPos[r]);let n=new Set;for(let o of e)for(let r=0;r<o.chunk.length;r++){let s=i.get(o.chunk[r]);s!=null&&(A?A.mapPos(s):s)==o.chunkPos[r]&&!A?.touchesRange(s,s+o.chunk[r].length)&&n.add(o.chunk[r])}return n}var Iw=class{constructor(e,A,i,n=0){this.layer=e,this.skip=A,this.minPoint=i,this.rank=n}get startSide(){return this.value?this.value.startSide:0}get endSide(){return this.value?this.value.endSide:0}goto(e,A=-1e9){return this.chunkIndex=this.rangeIndex=0,this.gotoInner(e,A,!1),this}gotoInner(e,A,i){for(;this.chunkIndex<this.layer.chunk.length;){let n=this.layer.chunk[this.chunkIndex];if(!(this.skip&&this.skip.has(n)||this.layer.chunkEnd(this.chunkIndex)<e||n.maxPoint<this.minPoint))break;this.chunkIndex++,i=!1}if(this.chunkIndex<this.layer.chunk.length){let n=this.layer.chunk[this.chunkIndex].findIndex(e-this.layer.chunkPos[this.chunkIndex],A,!0);(!i||this.rangeIndex<n)&&this.setRangeIndex(n)}this.next()}forward(e,A){(this.to-e||this.endSide-A)<0&&this.gotoInner(e,A,!0)}next(){for(;;)if(this.chunkIndex==this.layer.chunk.length){this.from=this.to=1e9,this.value=null;break}else{let e=this.layer.chunkPos[this.chunkIndex],A=this.layer.chunk[this.chunkIndex],i=e+A.from[this.rangeIndex];if(this.from=i,this.to=e+A.to[this.rangeIndex],this.value=A.value[this.rangeIndex],this.setRangeIndex(this.rangeIndex+1),this.minPoint<0||this.value.point&&this.to-this.from>=this.minPoint)break}}setRangeIndex(e){if(e==this.layer.chunk[this.chunkIndex].value.length){if(this.chunkIndex++,this.skip)for(;this.chunkIndex<this.layer.chunk.length&&this.skip.has(this.layer.chunk[this.chunkIndex]);)this.chunkIndex++;this.rangeIndex=0}else this.rangeIndex=e}nextChunk(){this.chunkIndex++,this.rangeIndex=0,this.next()}compare(e){return this.from-e.from||this.startSide-e.startSide||this.rank-e.rank||this.to-e.to||this.endSide-e.endSide}},J4=class t{constructor(e){this.heap=e}static from(e,A=null,i=-1){let n=[];for(let o=0;o<e.length;o++)for(let r=e[o];!r.isEmpty;r=r.nextLayer)r.maxPoint>=i&&n.push(new Iw(r,A,i,o));return n.length==1?n[0]:new t(n)}get startSide(){return this.value?this.value.startSide:0}goto(e,A=-1e9){for(let i of this.heap)i.goto(e,A);for(let i=this.heap.length>>1;i>=0;i--)qx(this.heap,i);return this.next(),this}forward(e,A){for(let i of this.heap)i.forward(e,A);for(let i=this.heap.length>>1;i>=0;i--)qx(this.heap,i);(this.to-e||this.value.endSide-A)<0&&this.next()}next(){if(this.heap.length==0)this.from=this.to=1e9,this.value=null,this.rank=-1;else{let e=this.heap[0];this.from=e.from,this.to=e.to,this.value=e.value,this.rank=e.rank,e.value&&e.next(),qx(this.heap,0)}}};function qx(t,e){for(let A=t[e];;){let i=(e<<1)+1;if(i>=t.length)break;let n=t[i];if(i+1<t.length&&n.compare(t[i+1])>=0&&(n=t[i+1],i++),A.compare(n)<0)break;t[i]=A,t[e]=n,e=i}}var mC=class{constructor(e,A,i){this.minPoint=i,this.active=[],this.activeTo=[],this.activeRank=[],this.minActive=-1,this.point=null,this.pointFrom=0,this.pointRank=0,this.to=-1e9,this.endSide=0,this.openStart=-1,this.cursor=J4.from(e,A,i)}goto(e,A=-1e9){return this.cursor.goto(e,A),this.active.length=this.activeTo.length=this.activeRank.length=0,this.minActive=-1,this.to=e,this.endSide=A,this.openStart=-1,this.next(),this}forward(e,A){for(;this.minActive>-1&&(this.activeTo[this.minActive]-e||this.active[this.minActive].endSide-A)<0;)this.removeActive(this.minActive);this.cursor.forward(e,A)}removeActive(e){nw(this.active,e),nw(this.activeTo,e),nw(this.activeRank,e),this.minActive=$AA(this.active,this.activeTo)}addActive(e){let A=0,{value:i,to:n,rank:o}=this.cursor;for(;A<this.activeRank.length&&(o-this.activeRank[A]||n-this.activeTo[A])>0;)A++;ow(this.active,A,i),ow(this.activeTo,A,n),ow(this.activeRank,A,o),e&&ow(e,A,this.cursor.from),this.minActive=$AA(this.active,this.activeTo)}next(){let e=this.to,A=this.point;this.point=null;let i=this.openStart<0?[]:null;for(;;){let n=this.minActive;if(n>-1&&(this.activeTo[n]-this.cursor.from||this.active[n].endSide-this.cursor.startSide)<0){if(this.activeTo[n]>e){this.to=this.activeTo[n],this.endSide=this.active[n].endSide;break}this.removeActive(n),i&&nw(i,n)}else if(this.cursor.value)if(this.cursor.from>e){this.to=this.cursor.from,this.endSide=this.cursor.startSide;break}else{let o=this.cursor.value;if(!o.point)this.addActive(i),this.cursor.next();else if(A&&this.cursor.to==this.to&&this.cursor.from<this.cursor.to)this.cursor.next();else{this.point=o,this.pointFrom=this.cursor.from,this.pointRank=this.cursor.rank,this.to=this.cursor.to,this.endSide=o.endSide,this.cursor.next(),this.forward(this.to,this.endSide);break}}else{this.to=this.endSide=1e9;break}}if(i){this.openStart=0;for(let n=i.length-1;n>=0&&i[n]<e;n--)this.openStart++}}activeForPoint(e){if(!this.active.length)return this.active;let A=[];for(let i=this.active.length-1;i>=0&&!(this.activeRank[i]<this.pointRank);i--)(this.activeTo[i]>e||this.activeTo[i]==e&&this.active[i].endSide>=this.point.endSide)&&A.push(this.active[i]);return A.reverse()}openEnd(e){let A=0;for(let i=this.activeTo.length-1;i>=0&&this.activeTo[i]>e;i--)A++;return A}};function XAA(t,e,A,i,n,o){t.goto(e),A.goto(i);let r=i+n,s=i,a=i-e;for(;;){let c=t.to+a-A.to,l=c||t.endSide-A.endSide,I=l<0?t.to+a:A.to,C=Math.min(I,r);if(t.point||A.point?t.point&&A.point&&(t.point==A.point||t.point.eq(A.point))&&rL(t.activeForPoint(t.to),A.activeForPoint(A.to))||o.comparePoint(s,C,t.point,A.point):C>s&&!rL(t.active,A.active)&&o.compareRange(s,C,t.active,A.active),I>r)break;(c||t.openEnd!=A.openEnd)&&o.boundChange&&o.boundChange(I),s=I,l<=0&&t.next(),l>=0&&A.next()}}function rL(t,e){if(t.length!=e.length)return!1;for(let A=0;A<t.length;A++)if(t[A]!=e[A]&&!t[A].eq(e[A]))return!1;return!0}function nw(t,e){for(let A=e,i=t.length-1;A<i;A++)t[A]=t[A+1];t.pop()}function ow(t,e,A){for(let i=t.length-1;i>=e;i--)t[i+1]=t[i];t[e]=A}function $AA(t,e){let A=-1,i=1e9;for(let n=0;n<e.length;n++)(e[n]-i||t[n].endSide-t[A].endSide)<0&&(A=n,i=e[n]);return A}function J0(t,e,A=t.length){let i=0;for(let n=0;n<A&&n<t.length;)t.charCodeAt(n)==9?(i+=e-i%e,n++):(i++,n=yr(t,n));return i}function Cw(t,e,A,i){for(let n=0,o=0;;){if(o>=e)return n;if(n==t.length)break;o+=t.charCodeAt(n)==9?A-o%A:1,n=yr(t,n)}return i===!0?-1:t.length}var cL="\u037C",aeA=typeof Symbol>"u"?"__"+cL:Symbol.for(cL),lL=typeof Symbol>"u"?"__styleSet"+Math.floor(Math.random()*1e8):Symbol("styleSet"),ceA=typeof globalThis<"u"?globalThis:typeof window<"u"?window:{},Kc=class{constructor(e,A){this.rules=[];let{finish:i}=A||{};function n(r){return/^@/.test(r)?[r]:r.split(/,\s*/)}function o(r,s,a,c){let l=[],I=/^@(\w+)\b/.exec(r[0]),C=I&&I[1]=="keyframes";if(I&&s==null)return a.push(r[0]+";");for(let d in s){let B=s[d];if(/&/.test(d))o(d.split(/,\s*/).map(E=>r.map(h=>E.replace(/&/,h))).reduce((E,h)=>E.concat(h)),B,a);else if(B&&typeof B=="object"){if(!I)throw new RangeError("The value of a property ("+d+") should be a primitive value.");o(n(d),B,l,C)}else B!=null&&l.push(d.replace(/_.*/,"").replace(/[A-Z]/g,E=>"-"+E.toLowerCase())+": "+B+";")}(l.length||C)&&a.push((i&&!I&&!c?r.map(i):r).join(", ")+" {"+l.join(" ")+"}")}for(let r in e)o(n(r),e[r],this.rules)}getRules(){return this.rules.join(`
`)}static newName(){let e=ceA[aeA]||1;return ceA[aeA]=e+1,cL+e.toString(36)}static mount(e,A,i){let n=e[lL],o=i&&i.nonce;n?o&&n.setNonce(o):n=new gL(e,o),n.mount(Array.isArray(A)?A:[A],e)}},leA=new Map,gL=class{constructor(e,A){let i=e.ownerDocument||e,n=i.defaultView;if(!e.head&&e.adoptedStyleSheets&&n.CSSStyleSheet){let o=leA.get(i);if(o)return e[lL]=o;this.sheet=new n.CSSStyleSheet,leA.set(i,this)}else this.styleTag=i.createElement("style"),A&&this.styleTag.setAttribute("nonce",A);this.modules=[],e[lL]=this}mount(e,A){let i=this.sheet,n=0,o=0;for(let r=0;r<e.length;r++){let s=e[r],a=this.modules.indexOf(s);if(a<o&&a>-1&&(this.modules.splice(a,1),o--,a=-1),a==-1){if(this.modules.splice(o++,0,s),i)for(let c=0;c<s.rules.length;c++)i.insertRule(s.rules[c],n++)}else{for(;o<a;)n+=this.modules[o++].rules.length;n+=s.rules.length,o++}}if(i)A.adoptedStyleSheets.indexOf(this.sheet)<0&&(A.adoptedStyleSheets=[this.sheet,...A.adoptedStyleSheets]);else{let r="";for(let a=0;a<this.modules.length;a++)r+=this.modules[a].getRules()+`
`;this.styleTag.textContent=r;let s=A.head||A;this.styleTag.parentNode!=s&&s.insertBefore(this.styleTag,s.firstChild)}}setNonce(e){this.styleTag&&this.styleTag.getAttribute("nonce")!=e&&this.styleTag.setAttribute("nonce",e)}};var T0={8:"Backspace",9:"Tab",10:"Enter",12:"NumLock",13:"Enter",16:"Shift",17:"Control",18:"Alt",20:"CapsLock",27:"Escape",32:" ",33:"PageUp",34:"PageDown",35:"End",36:"Home",37:"ArrowLeft",38:"ArrowUp",39:"ArrowRight",40:"ArrowDown",44:"PrintScreen",45:"Insert",46:"Delete",59:";",61:"=",91:"Meta",92:"Meta",106:"*",107:"+",108:",",109:"-",110:".",111:"/",144:"NumLock",145:"ScrollLock",160:"Shift",161:"Shift",162:"Control",163:"Control",164:"Alt",165:"Alt",173:"-",186:";",187:"=",188:",",189:"-",190:".",191:"/",192:"`",219:"[",220:"\\",221:"]",222:"'"},ME={48:")",49:"!",50:"@",51:"#",52:"$",53:"%",54:"^",55:"&",56:"*",57:"(",59:":",61:"+",173:"_",186:":",187:"+",188:"<",189:"_",190:">",191:"?",192:"~",219:"{",220:"|",221:"}",222:'"'},MRA=typeof navigator<"u"&&/Mac/.test(navigator.platform),kRA=typeof navigator<"u"&&/MSIE \d|Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(navigator.userAgent);for(vr=0;vr<10;vr++)T0[48+vr]=T0[96+vr]=String(vr);var vr;for(vr=1;vr<=24;vr++)T0[vr+111]="F"+vr;var vr;for(vr=65;vr<=90;vr++)T0[vr]=String.fromCharCode(vr+32),ME[vr]=String.fromCharCode(vr);var vr;for(dw in T0)ME.hasOwnProperty(dw)||(ME[dw]=T0[dw]);var dw;function geA(t){var e=MRA&&t.metaKey&&t.shiftKey&&!t.ctrlKey&&!t.altKey||kRA&&t.shiftKey&&t.key&&t.key.length==1||t.key=="Unidentified",A=!e&&t.key||(t.shiftKey?ME:T0)[t.keyCode]||t.key||"Unidentified";return A=="Esc"&&(A="Escape"),A=="Del"&&(A="Delete"),A=="Left"&&(A="ArrowLeft"),A=="Up"&&(A="ArrowUp"),A=="Right"&&(A="ArrowRight"),A=="Down"&&(A="ArrowDown"),A}function Un(){var t=arguments[0];typeof t=="string"&&(t=document.createElement(t));var e=1,A=arguments[1];if(A&&typeof A=="object"&&A.nodeType==null&&!Array.isArray(A)){for(var i in A)if(Object.prototype.hasOwnProperty.call(A,i)){var n=A[i];typeof n=="string"?t.setAttribute(i,n):n!=null&&(t[i]=n)}e++}for(;e<arguments.length;e++)IeA(t,arguments[e]);return t}function IeA(t,e){if(typeof e=="string")t.appendChild(document.createTextNode(e));else if(e!=null)if(e.nodeType!=null)t.appendChild(e);else if(Array.isArray(e))for(var A=0;A<e.length;A++)IeA(t,e[A]);else throw new RangeError("Unsupported child node: "+e)}function e3(t){let e;return t.nodeType==11?e=t.getSelection?t:t.ownerDocument:e=t,e.getSelection()}function mL(t,e){return e?t==e||t.contains(e.nodeType!=1?e.parentNode:e):!1}function Dw(t,e){if(!e.anchorNode)return!1;try{return mL(t,e.anchorNode)}catch{return!1}}function t3(t){return t.nodeType==3?vC(t,0,t.nodeValue.length).getClientRects():t.nodeType==1?t.getClientRects():[]}function q4(t,e,A,i){return A?CeA(t,e,A,i,-1)||CeA(t,e,A,i,1):!1}function yC(t){for(var e=0;;e++)if(t=t.previousSibling,!t)return e}function Rw(t){return t.nodeType==1&&/^(DIV|P|LI|UL|OL|BLOCKQUOTE|DD|DT|H\d|SECTION|PRE)$/.test(t.nodeName)}function CeA(t,e,A,i,n){for(;;){if(t==A&&e==i)return!0;if(e==(n<0?0:xg(t))){if(t.nodeName=="DIV")return!1;let o=t.parentNode;if(!o||o.nodeType!=1)return!1;e=yC(t)+(n<0?0:1),t=o}else if(t.nodeType==1){if(t=t.childNodes[e+(n<0?-1:0)],t.nodeType==1&&t.contentEditable=="false")return!1;e=n<0?xg(t):0}else return!1}}function xg(t){return t.nodeType==3?t.nodeValue.length:t.childNodes.length}function Pw(t,e){let A=e?t.left:t.right;return{left:A,right:A,top:t.top,bottom:t.bottom}}function SRA(t){let e=t.visualViewport;return e?{left:0,right:e.width,top:0,bottom:e.height}:{left:0,right:t.innerWidth,top:0,bottom:t.innerHeight}}function ntA(t,e){let A=e.width/t.offsetWidth,i=e.height/t.offsetHeight;return(A>.995&&A<1.005||!isFinite(A)||Math.abs(e.width-t.offsetWidth)<1)&&(A=1),(i>.995&&i<1.005||!isFinite(i)||Math.abs(e.height-t.offsetHeight)<1)&&(i=1),{scaleX:A,scaleY:i}}function RRA(t,e,A,i,n,o,r,s){let a=t.ownerDocument,c=a.defaultView||window;for(let l=t,I=!1;l&&!I;)if(l.nodeType==1){let C,d=l==a.body,B=1,E=1;if(d)C=SRA(c);else{if(/^(fixed|sticky)$/.test(getComputedStyle(l).position)&&(I=!0),l.scrollHeight<=l.clientHeight&&l.scrollWidth<=l.clientWidth){l=l.assignedSlot||l.parentNode;continue}let D=l.getBoundingClientRect();({scaleX:B,scaleY:E}=ntA(l,D)),C={left:D.left,right:D.left+l.clientWidth*B,top:D.top,bottom:D.top+l.clientHeight*E}}let h=0,u=0;if(n=="nearest")e.top<C.top?(u=e.top-(C.top+r),A>0&&e.bottom>C.bottom+u&&(u=e.bottom-C.bottom+r)):e.bottom>C.bottom&&(u=e.bottom-C.bottom+r,A<0&&e.top-u<C.top&&(u=e.top-(C.top+r)));else{let D=e.bottom-e.top,L=C.bottom-C.top;u=(n=="center"&&D<=L?e.top+D/2-L/2:n=="start"||n=="center"&&A<0?e.top-r:e.bottom-L+r)-C.top}if(i=="nearest"?e.left<C.left?(h=e.left-(C.left+o),A>0&&e.right>C.right+h&&(h=e.right-C.right+o)):e.right>C.right&&(h=e.right-C.right+o,A<0&&e.left<C.left+h&&(h=e.left-(C.left+o))):h=(i=="center"?e.left+(e.right-e.left)/2-(C.right-C.left)/2:i=="start"==s?e.left-o:e.right-(C.right-C.left)+o)-C.left,h||u)if(d)c.scrollBy(h,u);else{let D=0,L=0;if(u){let R=l.scrollTop;l.scrollTop+=u/E,L=(l.scrollTop-R)*E}if(h){let R=l.scrollLeft;l.scrollLeft+=h/B,D=(l.scrollLeft-R)*B}e={left:e.left-D,top:e.top-L,right:e.right-D,bottom:e.bottom-L},D&&Math.abs(D-h)<1&&(i="nearest"),L&&Math.abs(L-u)<1&&(n="nearest")}if(d)break;(e.top<C.top||e.bottom>C.bottom||e.left<C.left||e.right>C.right)&&(e={left:Math.max(e.left,C.left),right:Math.min(e.right,C.right),top:Math.max(e.top,C.top),bottom:Math.min(e.bottom,C.bottom)}),l=l.assignedSlot||l.parentNode}else if(l.nodeType==11)l=l.host;else break}function xRA(t){let e=t.ownerDocument,A,i;for(let n=t.parentNode;n&&!(n==e.body||A&&i);)if(n.nodeType==1)!i&&n.scrollHeight>n.clientHeight&&(i=n),!A&&n.scrollWidth>n.clientWidth&&(A=n),n=n.assignedSlot||n.parentNode;else if(n.nodeType==11)n=n.host;else break;return{x:A,y:i}}var pL=class{constructor(){this.anchorNode=null,this.anchorOffset=0,this.focusNode=null,this.focusOffset=0}eq(e){return this.anchorNode==e.anchorNode&&this.anchorOffset==e.anchorOffset&&this.focusNode==e.focusNode&&this.focusOffset==e.focusOffset}setRange(e){let{anchorNode:A,focusNode:i}=e;this.set(A,Math.min(e.anchorOffset,A?xg(A):0),i,Math.min(e.focusOffset,i?xg(i):0))}set(e,A,i,n){this.anchorNode=e,this.anchorOffset=A,this.focusNode=i,this.focusOffset=n}},kE=null;function otA(t){if(t.setActive)return t.setActive();if(kE)return t.focus(kE);let e=[];for(let A=t;A&&(e.push(A,A.scrollTop,A.scrollLeft),A!=A.ownerDocument);A=A.parentNode);if(t.focus(kE==null?{get preventScroll(){return kE={preventScroll:!0},!0}}:void 0),!kE){kE=!1;for(let A=0;A<e.length;){let i=e[A++],n=e[A++],o=e[A++];i.scrollTop!=n&&(i.scrollTop=n),i.scrollLeft!=o&&(i.scrollLeft=o)}}}var deA;function vC(t,e,A=e){let i=deA||(deA=document.createRange());return i.setEnd(t,A),i.setStart(t,e),i}function FE(t,e,A,i){let n={key:e,code:e,keyCode:A,which:A,cancelable:!0};i&&({altKey:n.altKey,ctrlKey:n.ctrlKey,shiftKey:n.shiftKey,metaKey:n.metaKey}=i);let o=new KeyboardEvent("keydown",n);o.synthetic=!0,t.dispatchEvent(o);let r=new KeyboardEvent("keyup",n);return r.synthetic=!0,t.dispatchEvent(r),o.defaultPrevented||r.defaultPrevented}function LRA(t){for(;t;){if(t&&(t.nodeType==9||t.nodeType==11&&t.host))return t;t=t.assignedSlot||t.parentNode}return null}function rtA(t){for(;t.attributes.length;)t.removeAttributeNode(t.attributes[0])}function FRA(t,e){let A=e.focusNode,i=e.focusOffset;if(!A||e.anchorNode!=A||e.anchorOffset!=i)return!1;for(i=Math.min(i,xg(A));;)if(i){if(A.nodeType!=1)return!1;let n=A.childNodes[i-1];n.contentEditable=="false"?i--:(A=n,i=xg(A))}else{if(A==t)return!0;i=yC(A),A=A.parentNode}}function stA(t){return t.scrollTop>Math.max(1,t.scrollHeight-t.clientHeight-4)}function atA(t,e){for(let A=t,i=e;;){if(A.nodeType==3&&i>0)return{node:A,offset:i};if(A.nodeType==1&&i>0){if(A.contentEditable=="false")return null;A=A.childNodes[i-1],i=xg(A)}else if(A.parentNode&&!Rw(A))i=yC(A),A=A.parentNode;else return null}}function ctA(t,e){for(let A=t,i=e;;){if(A.nodeType==3&&i<A.nodeValue.length)return{node:A,offset:i};if(A.nodeType==1&&i<A.childNodes.length){if(A.contentEditable=="false")return null;A=A.childNodes[i],i=0}else if(A.parentNode&&!Rw(A))i=yC(A)+1,A=A.parentNode;else return null}}var Xs=class t{constructor(e,A,i=!0){this.node=e,this.offset=A,this.precise=i}static before(e,A){return new t(e.parentNode,yC(e),A)}static after(e,A){return new t(e.parentNode,yC(e)+1,A)}},sF=[],Fo=class t{constructor(){this.parent=null,this.dom=null,this.flags=2}get overrideDOMText(){return null}get posAtStart(){return this.parent?this.parent.posBefore(this):0}get posAtEnd(){return this.posAtStart+this.length}posBefore(e){let A=this.posAtStart;for(let i of this.children){if(i==e)return A;A+=i.length+i.breakAfter}throw new RangeError("Invalid child in posBefore")}posAfter(e){return this.posBefore(e)+e.length}sync(e,A){if(this.flags&2){let i=this.dom,n=null,o;for(let r of this.children){if(r.flags&7){if(!r.dom&&(o=n?n.nextSibling:i.firstChild)){let s=t.get(o);(!s||!s.parent&&s.canReuseDOM(r))&&r.reuseDOM(o)}r.sync(e,A),r.flags&=-8}if(o=n?n.nextSibling:i.firstChild,A&&!A.written&&A.node==i&&o!=r.dom&&(A.written=!0),r.dom.parentNode==i)for(;o&&o!=r.dom;)o=BeA(o);else i.insertBefore(r.dom,o);n=r.dom}for(o=n?n.nextSibling:i.firstChild,o&&A&&A.node==i&&(A.written=!0);o;)o=BeA(o)}else if(this.flags&1)for(let i of this.children)i.flags&7&&(i.sync(e,A),i.flags&=-8)}reuseDOM(e){}localPosFromDOM(e,A){let i;if(e==this.dom)i=this.dom.childNodes[A];else{let n=xg(e)==0?0:A==0?-1:1;for(;;){let o=e.parentNode;if(o==this.dom)break;n==0&&o.firstChild!=o.lastChild&&(e==o.firstChild?n=-1:n=1),e=o}n<0?i=e:i=e.nextSibling}if(i==this.dom.firstChild)return 0;for(;i&&!t.get(i);)i=i.nextSibling;if(!i)return this.length;for(let n=0,o=0;;n++){let r=this.children[n];if(r.dom==i)return o;o+=r.length+r.breakAfter}}domBoundsAround(e,A,i=0){let n=-1,o=-1,r=-1,s=-1;for(let a=0,c=i,l=i;a<this.children.length;a++){let I=this.children[a],C=c+I.length;if(c<e&&C>A)return I.domBoundsAround(e,A,c);if(C>=e&&n==-1&&(n=a,o=c),c>A&&I.dom.parentNode==this.dom){r=a,s=l;break}l=C,c=C+I.breakAfter}return{from:o,to:s<0?i+this.length:s,startDOM:(n?this.children[n-1].dom.nextSibling:null)||this.dom.firstChild,endDOM:r<this.children.length&&r>=0?this.children[r].dom:null}}markDirty(e=!1){this.flags|=2,this.markParentsDirty(e)}markParentsDirty(e){for(let A=this.parent;A;A=A.parent){if(e&&(A.flags|=2),A.flags&1)return;A.flags|=1,e=!1}}setParent(e){this.parent!=e&&(this.parent=e,this.flags&7&&this.markParentsDirty(!0))}setDOM(e){this.dom!=e&&(this.dom&&(this.dom.cmView=null),this.dom=e,e.cmView=this)}get rootView(){for(let e=this;;){let A=e.parent;if(!A)return e;e=A}}replaceChildren(e,A,i=sF){this.markDirty();for(let n=e;n<A;n++){let o=this.children[n];o.parent==this&&i.indexOf(o)<0&&o.destroy()}i.length<250?this.children.splice(e,A-e,...i):this.children=[].concat(this.children.slice(0,e),i,this.children.slice(A));for(let n=0;n<i.length;n++)i[n].setParent(this)}ignoreMutation(e){return!1}ignoreEvent(e){return!1}childCursor(e=this.length){return new xw(this.children,e,this.children.length)}childPos(e,A=1){return this.childCursor().findPos(e,A)}toString(){let e=this.constructor.name.replace("View","");return e+(this.children.length?"("+this.children.join()+")":this.length?"["+(e=="Text"?this.text:this.length)+"]":"")+(this.breakAfter?"#":"")}static get(e){return e.cmView}get isEditable(){return!0}get isWidget(){return!1}get isHidden(){return!1}merge(e,A,i,n,o,r){return!1}become(e){return!1}canReuseDOM(e){return e.constructor==this.constructor&&!((this.flags|e.flags)&8)}getSide(){return 0}destroy(){for(let e of this.children)e.parent==this&&e.destroy();this.parent=null}};Fo.prototype.breakAfter=0;function BeA(t){let e=t.nextSibling;return t.parentNode.removeChild(t),e}var xw=class{constructor(e,A,i){this.children=e,this.pos=A,this.i=i,this.off=0}findPos(e,A=1){for(;;){if(e>this.pos||e==this.pos&&(A>0||this.i==0||this.children[this.i-1].breakAfter))return this.off=e-this.pos,this;let i=this.children[--this.i];this.pos-=i.length+i.breakAfter}}};function ltA(t,e,A,i,n,o,r,s,a){let{children:c}=t,l=c.length?c[e]:null,I=o.length?o[o.length-1]:null,C=I?I.breakAfter:r;if(!(e==i&&l&&!r&&!C&&o.length<2&&l.merge(A,n,o.length?I:null,A==0,s,a))){if(i<c.length){let d=c[i];d&&(n<d.length||d.breakAfter&&I?.breakAfter)?(e==i&&(d=d.split(n),n=0),!C&&I&&d.merge(0,n,I,!0,0,a)?o[o.length-1]=d:((n||d.children.length&&!d.children[0].length)&&d.merge(0,n,null,!1,0,a),o.push(d))):d?.breakAfter&&(I?I.breakAfter=1:r=1),i++}for(l&&(l.breakAfter=r,A>0&&(!r&&o.length&&l.merge(A,l.length,o[0],!1,s,0)?l.breakAfter=o.shift().breakAfter:(A<l.length||l.children.length&&l.children[l.children.length-1].length==0)&&l.merge(A,l.length,null,!1,s,0),e++));e<i&&o.length;)if(c[i-1].become(o[o.length-1]))i--,o.pop(),a=o.length?0:s;else if(c[e].become(o[0]))e++,o.shift(),s=o.length?0:a;else break;!o.length&&e&&i<c.length&&!c[e-1].breakAfter&&c[i].merge(0,0,c[e-1],!1,s,a)&&e--,(e<i||o.length)&&t.replaceChildren(e,i,o)}}function gtA(t,e,A,i,n,o){let r=t.childCursor(),{i:s,off:a}=r.findPos(A,1),{i:c,off:l}=r.findPos(e,-1),I=e-A;for(let C of i)I+=C.length;t.length+=I,ltA(t,c,l,s,a,i,0,n,o)}var La=typeof navigator<"u"?navigator:{userAgent:"",vendor:"",platform:""},wL=typeof document<"u"?document:{documentElement:{style:{}}},DL=/Edge\/(\d+)/.exec(La.userAgent),ItA=/MSIE \d/.test(La.userAgent),yL=/Trident\/(?:[7-9]|\d{2,})\..*rv:(\d+)/.exec(La.userAgent),jw=!!(ItA||yL||DL),EeA=!jw&&/gecko\/(\d+)/i.test(La.userAgent),IL=!jw&&/Chrome\/(\d+)/.exec(La.userAgent),QeA="webkitFontSmoothing"in wL.documentElement.style,CtA=!jw&&/Apple Computer/.test(La.vendor),heA=CtA&&(/Mobile\/\w+/.test(La.userAgent)||La.maxTouchPoints>2),At={mac:heA||/Mac/.test(La.platform),windows:/Win/.test(La.platform),linux:/Linux|X11/.test(La.platform),ie:jw,ie_version:ItA?wL.documentMode||6:yL?+yL[1]:DL?+DL[1]:0,gecko:EeA,gecko_version:EeA?+(/Firefox\/(\d+)/.exec(La.userAgent)||[0,0])[1]:0,chrome:!!IL,chrome_version:IL?+IL[1]:0,ios:heA,android:/Android\b/.test(La.userAgent),webkit:QeA,safari:CtA,webkit_version:QeA?+(/\bAppleWebKit\/(\d+)/.exec(La.userAgent)||[0,0])[1]:0,tabSize:wL.documentElement.style.tabSize!=null?"tab-size":"-moz-tab-size"},NRA=256,Lg=class t extends Fo{constructor(e){super(),this.text=e}get length(){return this.text.length}createDOM(e){this.setDOM(e||document.createTextNode(this.text))}sync(e,A){this.dom||this.createDOM(),this.dom.nodeValue!=this.text&&(A&&A.node==this.dom&&(A.written=!0),this.dom.nodeValue=this.text)}reuseDOM(e){e.nodeType==3&&this.createDOM(e)}merge(e,A,i){return this.flags&8||i&&(!(i instanceof t)||this.length-(A-e)+i.length>NRA||i.flags&8)?!1:(this.text=this.text.slice(0,e)+(i?i.text:"")+this.text.slice(A),this.markDirty(),!0)}split(e){let A=new t(this.text.slice(e));return this.text=this.text.slice(0,e),this.markDirty(),A.flags|=this.flags&8,A}localPosFromDOM(e,A){return e==this.dom?A:A?this.text.length:0}domAtPos(e){return new Xs(this.dom,e)}domBoundsAround(e,A,i){return{from:i,to:i+this.length,startDOM:this.dom,endDOM:this.dom.nextSibling}}coordsAt(e,A){return _RA(this.dom,e,A)}},D1=class t extends Fo{constructor(e,A=[],i=0){super(),this.mark=e,this.children=A,this.length=i;for(let n of A)n.setParent(this)}setAttrs(e){if(rtA(e),this.mark.class&&(e.className=this.mark.class),this.mark.attrs)for(let A in this.mark.attrs)e.setAttribute(A,this.mark.attrs[A]);return e}canReuseDOM(e){return super.canReuseDOM(e)&&!((this.flags|e.flags)&8)}reuseDOM(e){e.nodeName==this.mark.tagName.toUpperCase()&&(this.setDOM(e),this.flags|=6)}sync(e,A){this.dom?this.flags&4&&this.setAttrs(this.dom):this.setDOM(this.setAttrs(document.createElement(this.mark.tagName))),super.sync(e,A)}merge(e,A,i,n,o,r){return i&&(!(i instanceof t&&i.mark.eq(this.mark))||e&&o<=0||A<this.length&&r<=0)?!1:(gtA(this,e,A,i?i.children.slice():[],o-1,r-1),this.markDirty(),!0)}split(e){let A=[],i=0,n=-1,o=0;for(let s of this.children){let a=i+s.length;a>e&&A.push(i<e?s.split(e-i):s),n<0&&i>=e&&(n=o),i=a,o++}let r=this.length-e;return this.length=e,n>-1&&(this.children.length=n,this.markDirty()),new t(this.mark,A,r)}domAtPos(e){return dtA(this,e)}coordsAt(e,A){return EtA(this,e,A)}};function _RA(t,e,A){let i=t.nodeValue.length;e>i&&(e=i);let n=e,o=e,r=0;e==0&&A<0||e==i&&A>=0?At.chrome||At.gecko||(e?(n--,r=1):o<i&&(o++,r=-1)):A<0?n--:o<i&&o++;let s=vC(t,n,o).getClientRects();if(!s.length)return null;let a=s[(r?r<0:A>=0)?0:s.length-1];return At.safari&&!r&&a.width==0&&(a=Array.prototype.find.call(s,c=>c.width)||a),r?Pw(a,r<0):a||null}var i3=class t extends Fo{static create(e,A,i){return new t(e,A,i)}constructor(e,A,i){super(),this.widget=e,this.length=A,this.side=i,this.prevWidget=null}split(e){let A=t.create(this.widget,this.length-e,this.side);return this.length-=e,A}sync(e){(!this.dom||!this.widget.updateDOM(this.dom,e))&&(this.dom&&this.prevWidget&&this.prevWidget.destroy(this.dom),this.prevWidget=null,this.setDOM(this.widget.toDOM(e)),this.widget.editable||(this.dom.contentEditable="false"))}getSide(){return this.side}merge(e,A,i,n,o,r){return i&&(!(i instanceof t)||!this.widget.compare(i.widget)||e>0&&o<=0||A<this.length&&r<=0)?!1:(this.length=e+(i?i.length:0)+(this.length-A),!0)}become(e){return e instanceof t&&e.side==this.side&&this.widget.constructor==e.widget.constructor?(this.widget.compare(e.widget)||this.markDirty(!0),this.dom&&!this.prevWidget&&(this.prevWidget=this.widget),this.widget=e.widget,this.length=e.length,!0):!1}ignoreMutation(){return!0}ignoreEvent(e){return this.widget.ignoreEvent(e)}get overrideDOMText(){if(this.length==0)return In.empty;let e=this;for(;e.parent;)e=e.parent;let{view:A}=e,i=A&&A.state.doc,n=this.posAtStart;return i?i.slice(n,n+this.length):In.empty}domAtPos(e){return(this.length?e==0:this.side>0)?Xs.before(this.dom):Xs.after(this.dom,e==this.length)}domBoundsAround(){return null}coordsAt(e,A){let i=this.widget.coordsAt(this.dom,e,A);if(i)return i;let n=this.dom.getClientRects(),o=null;if(!n.length)return null;let r=this.side?this.side<0:e>0;for(let s=r?n.length-1:0;o=n[s],!(e>0?s==0:s==n.length-1||o.top<o.bottom);s+=r?-1:1);return Pw(o,!r)}get isEditable(){return!1}get isWidget(){return!0}get isHidden(){return this.widget.isHidden}destroy(){super.destroy(),this.dom&&this.widget.destroy(this.dom)}},n3=class t extends Fo{constructor(e){super(),this.side=e}get length(){return 0}merge(){return!1}become(e){return e instanceof t&&e.side==this.side}split(){return new t(this.side)}sync(){if(!this.dom){let e=document.createElement("img");e.className="cm-widgetBuffer",e.setAttribute("aria-hidden","true"),this.setDOM(e)}}getSide(){return this.side}domAtPos(e){return this.side>0?Xs.before(this.dom):Xs.after(this.dom)}localPosFromDOM(){return 0}domBoundsAround(){return null}coordsAt(e){return this.dom.getBoundingClientRect()}get overrideDOMText(){return In.empty}get isHidden(){return!0}};Lg.prototype.children=i3.prototype.children=n3.prototype.children=sF;function dtA(t,e){let A=t.dom,{children:i}=t,n=0;for(let o=0;n<i.length;n++){let r=i[n],s=o+r.length;if(!(s==o&&r.getSide()<=0)){if(e>o&&e<s&&r.dom.parentNode==A)return r.domAtPos(e-o);if(e<=o)break;o=s}}for(let o=n;o>0;o--){let r=i[o-1];if(r.dom.parentNode==A)return r.domAtPos(r.length)}for(let o=n;o<i.length;o++){let r=i[o];if(r.dom.parentNode==A)return r.domAtPos(0)}return new Xs(A,0)}function BtA(t,e,A){let i,{children:n}=t;A>0&&e instanceof D1&&n.length&&(i=n[n.length-1])instanceof D1&&i.mark.eq(e.mark)?BtA(i,e.children[0],A-1):(n.push(e),e.setParent(t)),t.length+=e.length}function EtA(t,e,A){let i=null,n=-1,o=null,r=-1;function s(c,l){for(let I=0,C=0;I<c.children.length&&C<=l;I++){let d=c.children[I],B=C+d.length;B>=l&&(d.children.length?s(d,l-C):(!o||o.isHidden&&(A>0||URA(o,d)))&&(B>l||C==B&&d.getSide()>0)?(o=d,r=l-C):(C<l||C==B&&d.getSide()<0&&!d.isHidden)&&(i=d,n=l-C)),C=B}}s(t,e);let a=(A<0?i:o)||i||o;return a?a.coordsAt(Math.max(0,a==i?n:r),A):GRA(t)}function GRA(t){let e=t.dom.lastChild;if(!e)return t.dom.getBoundingClientRect();let A=t3(e);return A[A.length-1]||null}function URA(t,e){let A=t.coordsAt(0,1),i=e.coordsAt(0,1);return A&&i&&i.top<A.bottom}function vL(t,e){for(let A in t)A=="class"&&e.class?e.class+=" "+t.class:A=="style"&&e.style?e.style+=";"+t.style:e[A]=t[A];return e}var ueA=Object.create(null);function Lw(t,e,A){if(t==e)return!0;t||(t=ueA),e||(e=ueA);let i=Object.keys(t),n=Object.keys(e);if(i.length-(A&&i.indexOf(A)>-1?1:0)!=n.length-(A&&n.indexOf(A)>-1?1:0))return!1;for(let o of i)if(o!=A&&(n.indexOf(o)==-1||t[o]!==e[o]))return!1;return!0}function bL(t,e,A){let i=!1;if(e)for(let n in e)A&&n in A||(i=!0,n=="style"?t.style.cssText="":t.removeAttribute(n));if(A)for(let n in A)e&&e[n]==A[n]||(i=!0,n=="style"?t.style.cssText=A[n]:t.setAttribute(n,A[n]));return i}function KRA(t){let e=Object.create(null);for(let A=0;A<t.attributes.length;A++){let i=t.attributes[A];e[i.name]=i.value}return e}var Ic=class{eq(e){return!1}updateDOM(e,A){return!1}compare(e){return this==e||this.constructor==e.constructor&&this.eq(e)}get estimatedHeight(){return-1}get lineBreaks(){return 0}ignoreEvent(e){return!0}coordsAt(e,A,i){return null}get isHidden(){return!1}get editable(){return!1}destroy(e){}},$s=function(t){return t[t.Text=0]="Text",t[t.WidgetBefore=1]="WidgetBefore",t[t.WidgetAfter=2]="WidgetAfter",t[t.WidgetRange=3]="WidgetRange",t}($s||($s={})),ut=class extends wl{constructor(e,A,i,n){super(),this.startSide=e,this.endSide=A,this.widget=i,this.spec=n}get heightRelevant(){return!1}static mark(e){return new o3(e)}static widget(e){let A=Math.max(-1e4,Math.min(1e4,e.side||0)),i=!!e.block;return A+=i&&!e.inlineOrder?A>0?3e8:-4e8:A>0?1e8:-1e8,new y1(e,A,A,i,e.widget||null,!1)}static replace(e){let A=!!e.block,i,n;if(e.isBlockGap)i=-5e8,n=4e8;else{let{start:o,end:r}=QtA(e,A);i=(o?A?-3e8:-1:5e8)-1,n=(r?A?2e8:1:-6e8)+1}return new y1(e,i,n,A,e.widget||null,!0)}static line(e){return new r3(e)}static set(e,A=!1){return co.of(e,A)}hasHeight(){return this.widget?this.widget.estimatedHeight>-1:!1}};ut.none=co.empty;var o3=class t extends ut{constructor(e){let{start:A,end:i}=QtA(e);super(A?-1:5e8,i?1:-6e8,null,e),this.tagName=e.tagName||"span",this.class=e.class||"",this.attrs=e.attributes||null}eq(e){var A,i;return this==e||e instanceof t&&this.tagName==e.tagName&&(this.class||((A=this.attrs)===null||A===void 0?void 0:A.class))==(e.class||((i=e.attrs)===null||i===void 0?void 0:i.class))&&Lw(this.attrs,e.attrs,"class")}range(e,A=e){if(e>=A)throw new RangeError("Mark decorations may not be empty");return super.range(e,A)}};o3.prototype.point=!1;var r3=class t extends ut{constructor(e){super(-2e8,-2e8,null,e)}eq(e){return e instanceof t&&this.spec.class==e.spec.class&&Lw(this.spec.attributes,e.spec.attributes)}range(e,A=e){if(A!=e)throw new RangeError("Line decoration ranges must be zero-length");return super.range(e,A)}};r3.prototype.mapMode=Is.TrackBefore;r3.prototype.point=!0;var y1=class t extends ut{constructor(e,A,i,n,o,r){super(A,i,o,e),this.block=n,this.isReplace=r,this.mapMode=n?A<=0?Is.TrackBefore:Is.TrackAfter:Is.TrackDel}get type(){return this.startSide!=this.endSide?$s.WidgetRange:this.startSide<=0?$s.WidgetBefore:$s.WidgetAfter}get heightRelevant(){return this.block||!!this.widget&&(this.widget.estimatedHeight>=5||this.widget.lineBreaks>0)}eq(e){return e instanceof t&&YRA(this.widget,e.widget)&&this.block==e.block&&this.startSide==e.startSide&&this.endSide==e.endSide}range(e,A=e){if(this.isReplace&&(e>A||e==A&&this.startSide>0&&this.endSide<=0))throw new RangeError("Invalid range for replacement decoration");if(!this.isReplace&&A!=e)throw new RangeError("Widget decorations can only have zero-length ranges");return super.range(e,A)}};y1.prototype.point=!0;function QtA(t,e=!1){let{inclusiveStart:A,inclusiveEnd:i}=t;return A==null&&(A=t.inclusive),i==null&&(i=t.inclusive),{start:A??e,end:i??e}}function YRA(t,e){return t==e||!!(t&&e&&t.compare(e))}function yw(t,e,A,i=0){let n=A.length-1;n>=0&&A[n]+i>=t?A[n]=Math.max(A[n],e):A.push(t,e)}var Es=class t extends Fo{constructor(){super(...arguments),this.children=[],this.length=0,this.prevAttrs=void 0,this.attrs=null,this.breakAfter=0}merge(e,A,i,n,o,r){if(i){if(!(i instanceof t))return!1;this.dom||i.transferDOM(this)}return n&&this.setDeco(i?i.attrs:null),gtA(this,e,A,i?i.children.slice():[],o,r),!0}split(e){let A=new t;if(A.breakAfter=this.breakAfter,this.length==0)return A;let{i,off:n}=this.childPos(e);n&&(A.append(this.children[i].split(n),0),this.children[i].merge(n,this.children[i].length,null,!1,0,0),i++);for(let o=i;o<this.children.length;o++)A.append(this.children[o],0);for(;i>0&&this.children[i-1].length==0;)this.children[--i].destroy();return this.children.length=i,this.markDirty(),this.length=e,A}transferDOM(e){this.dom&&(this.markDirty(),e.setDOM(this.dom),e.prevAttrs=this.prevAttrs===void 0?this.attrs:this.prevAttrs,this.prevAttrs=void 0,this.dom=null)}setDeco(e){Lw(this.attrs,e)||(this.dom&&(this.prevAttrs=this.attrs,this.markDirty()),this.attrs=e)}append(e,A){BtA(this,e,A)}addLineDeco(e){let A=e.spec.attributes,i=e.spec.class;A&&(this.attrs=vL(A,this.attrs||{})),i&&(this.attrs=vL({class:i},this.attrs||{}))}domAtPos(e){return dtA(this,e)}reuseDOM(e){e.nodeName=="DIV"&&(this.setDOM(e),this.flags|=6)}sync(e,A){var i;this.dom?this.flags&4&&(rtA(this.dom),this.dom.className="cm-line",this.prevAttrs=this.attrs?null:void 0):(this.setDOM(document.createElement("div")),this.dom.className="cm-line",this.prevAttrs=this.attrs?null:void 0),this.prevAttrs!==void 0&&(bL(this.dom,this.prevAttrs,this.attrs),this.dom.classList.add("cm-line"),this.prevAttrs=void 0),super.sync(e,A);let n=this.dom.lastChild;for(;n&&Fo.get(n)instanceof D1;)n=n.lastChild;if(!n||!this.length||n.nodeName!="BR"&&((i=Fo.get(n))===null||i===void 0?void 0:i.isEditable)==!1&&(!At.ios||!this.children.some(o=>o instanceof Lg))){let o=document.createElement("BR");o.cmIgnore=!0,this.dom.appendChild(o)}}measureTextSize(){if(this.children.length==0||this.length>20)return null;let e=0,A;for(let i of this.children){if(!(i instanceof Lg)||/[^ -~]/.test(i.text))return null;let n=t3(i.dom);if(n.length!=1)return null;e+=n[0].width,A=n[0].height}return e?{lineHeight:this.dom.getBoundingClientRect().height,charWidth:e/this.length,textHeight:A}:null}coordsAt(e,A){let i=EtA(this,e,A);if(!this.children.length&&i&&this.parent){let{heightOracle:n}=this.parent.view.viewState,o=i.bottom-i.top;if(Math.abs(o-n.lineHeight)<2&&n.textHeight<o){let r=(o-n.textHeight)/2;return{top:i.top+r,bottom:i.bottom-r,left:i.left,right:i.left}}}return i}become(e){return e instanceof t&&this.children.length==0&&e.children.length==0&&Lw(this.attrs,e.attrs)&&this.breakAfter==e.breakAfter}covers(){return!0}static find(e,A){for(let i=0,n=0;i<e.children.length;i++){let o=e.children[i],r=n+o.length;if(r>=A){if(o instanceof t)return o;if(r>A)break}n=r+o.breakAfter}return null}},DC=class t extends Fo{constructor(e,A,i){super(),this.widget=e,this.length=A,this.deco=i,this.breakAfter=0,this.prevWidget=null}merge(e,A,i,n,o,r){return i&&(!(i instanceof t)||!this.widget.compare(i.widget)||e>0&&o<=0||A<this.length&&r<=0)?!1:(this.length=e+(i?i.length:0)+(this.length-A),!0)}domAtPos(e){return e==0?Xs.before(this.dom):Xs.after(this.dom,e==this.length)}split(e){let A=this.length-e;this.length=e;let i=new t(this.widget,A,this.deco);return i.breakAfter=this.breakAfter,i}get children(){return sF}sync(e){(!this.dom||!this.widget.updateDOM(this.dom,e))&&(this.dom&&this.prevWidget&&this.prevWidget.destroy(this.dom),this.prevWidget=null,this.setDOM(this.widget.toDOM(e)),this.widget.editable||(this.dom.contentEditable="false"))}get overrideDOMText(){return this.parent?this.parent.view.state.doc.slice(this.posAtStart,this.posAtEnd):In.empty}domBoundsAround(){return null}become(e){return e instanceof t&&e.widget.constructor==this.widget.constructor?(e.widget.compare(this.widget)||this.markDirty(!0),this.dom&&!this.prevWidget&&(this.prevWidget=this.widget),this.widget=e.widget,this.length=e.length,this.deco=e.deco,this.breakAfter=e.breakAfter,!0):!1}ignoreMutation(){return!0}ignoreEvent(e){return this.widget.ignoreEvent(e)}get isEditable(){return!1}get isWidget(){return!0}coordsAt(e,A){let i=this.widget.coordsAt(this.dom,e,A);return i||(this.widget instanceof s3?null:Pw(this.dom.getBoundingClientRect(),this.length?e==0:A<=0))}destroy(){super.destroy(),this.dom&&this.widget.destroy(this.dom)}covers(e){let{startSide:A,endSide:i}=this.deco;return A==i?!1:e<0?A<0:i>0}},s3=class extends Ic{constructor(e){super(),this.height=e}toDOM(){let e=document.createElement("div");return e.className="cm-gap",this.updateDOM(e),e}eq(e){return e.height==this.height}updateDOM(e){return e.style.height=this.height+"px",!0}get editable(){return!0}get estimatedHeight(){return this.height}ignoreEvent(){return!1}},V4=class t{constructor(e,A,i,n){this.doc=e,this.pos=A,this.end=i,this.disallowBlockEffectsFor=n,this.content=[],this.curLine=null,this.breakAtStart=0,this.pendingBuffer=0,this.bufferMarks=[],this.atCursorPos=!0,this.openStart=-1,this.openEnd=-1,this.text="",this.textOff=0,this.cursor=e.iter(),this.skip=A}posCovered(){if(this.content.length==0)return!this.breakAtStart&&this.doc.lineAt(this.pos).from!=this.pos;let e=this.content[this.content.length-1];return!(e.breakAfter||e instanceof DC&&e.deco.endSide<0)}getLine(){return this.curLine||(this.content.push(this.curLine=new Es),this.atCursorPos=!0),this.curLine}flushBuffer(e=this.bufferMarks){this.pendingBuffer&&(this.curLine.append(Bw(new n3(-1),e),e.length),this.pendingBuffer=0)}addBlockWidget(e){this.flushBuffer(),this.curLine=null,this.content.push(e)}finish(e){this.pendingBuffer&&e<=this.bufferMarks.length?this.flushBuffer():this.pendingBuffer=0,!this.posCovered()&&!(e&&this.content.length&&this.content[this.content.length-1]instanceof DC)&&this.getLine()}buildText(e,A,i){for(;e>0;){if(this.textOff==this.text.length){let{value:o,lineBreak:r,done:s}=this.cursor.next(this.skip);if(this.skip=0,s)throw new Error("Ran out of text content when drawing inline views");if(r){this.posCovered()||this.getLine(),this.content.length?this.content[this.content.length-1].breakAfter=1:this.breakAtStart=1,this.flushBuffer(),this.curLine=null,this.atCursorPos=!0,e--;continue}else this.text=o,this.textOff=0}let n=Math.min(this.text.length-this.textOff,e,512);this.flushBuffer(A.slice(A.length-i)),this.getLine().append(Bw(new Lg(this.text.slice(this.textOff,this.textOff+n)),A),i),this.atCursorPos=!0,this.textOff+=n,e-=n,i=0}}span(e,A,i,n){this.buildText(A-e,i,n),this.pos=A,this.openStart<0&&(this.openStart=n)}point(e,A,i,n,o,r){if(this.disallowBlockEffectsFor[r]&&i instanceof y1){if(i.block)throw new RangeError("Block decorations may not be specified via plugins");if(A>this.doc.lineAt(this.pos).to)throw new RangeError("Decorations that replace line breaks may not be specified via plugins")}let s=A-e;if(i instanceof y1)if(i.block)i.startSide>0&&!this.posCovered()&&this.getLine(),this.addBlockWidget(new DC(i.widget||feA.block,s,i));else{let a=i3.create(i.widget||feA.inline,s,s?0:i.startSide),c=this.atCursorPos&&!a.isEditable&&o<=n.length&&(e<A||i.startSide>0),l=!a.isEditable&&(e<A||o>n.length||i.startSide<=0),I=this.getLine();this.pendingBuffer==2&&!c&&!a.isEditable&&(this.pendingBuffer=0),this.flushBuffer(n),c&&(I.append(Bw(new n3(1),n),o),o=n.length+Math.max(0,o-n.length)),I.append(Bw(a,n),o),this.atCursorPos=l,this.pendingBuffer=l?e<A||o>n.length?1:2:0,this.pendingBuffer&&(this.bufferMarks=n.slice())}else this.doc.lineAt(this.pos).from==this.pos&&this.getLine().addLineDeco(i);s&&(this.textOff+s<=this.text.length?this.textOff+=s:(this.skip+=s-(this.text.length-this.textOff),this.text="",this.textOff=0),this.pos=A),this.openStart<0&&(this.openStart=o)}static build(e,A,i,n,o){let r=new t(e,A,i,o);return r.openEnd=co.spans(n,A,i,r),r.openStart<0&&(r.openStart=r.openEnd),r.finish(r.openEnd),r}};function Bw(t,e){for(let A of e)t=new D1(A,[t],t.length);return t}var feA=(()=>{class t extends Ic{constructor(A){super(),this.tag=A}eq(A){return A.tag==this.tag}toDOM(){return document.createElement(this.tag)}updateDOM(A){return A.nodeName.toLowerCase()==this.tag}get isHidden(){return!0}}return t.inline=new t("span"),t.block=new t("div"),t})(),lo=function(t){return t[t.LTR=0]="LTR",t[t.RTL=1]="RTL",t}(lo||(lo={})),bC=lo.LTR,aF=lo.RTL;function htA(t){let e=[];for(let A=0;A<t.length;A++)e.push(1<<+t[A]);return e}var JRA=htA("88888888888888888888888888888888888666888888787833333333337888888000000000000000000000000008888880000000000000000000000000088888888888888888888888888888888888887866668888088888663380888308888800000000000000000000000800000000000000000000000000000008"),TRA=htA("4444448826627288999999999992222222222222222222222222222222222222222222222229999999999999999999994444444444644222822222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222222999999949999999229989999223333333333"),ML=Object.create(null),Mg=[];for(let t of["()","[]","{}"]){let e=t.charCodeAt(0),A=t.charCodeAt(1);ML[e]=A,ML[A]=-e}function utA(t){return t<=247?JRA[t]:1424<=t&&t<=1524?2:1536<=t&&t<=1785?TRA[t-1536]:1774<=t&&t<=2220?4:8192<=t&&t<=8204?256:64336<=t&&t<=65023?4:1}var HRA=/[\u0590-\u05f4\u0600-\u06ff\u0700-\u08ac\ufb50-\ufdff]/,Sg=class{get dir(){return this.level%2?aF:bC}constructor(e,A,i){this.from=e,this.to=A,this.level=i}side(e,A){return this.dir==A==e?this.to:this.from}forward(e,A){return e==(this.dir==A)}static find(e,A,i,n){let o=-1;for(let r=0;r<e.length;r++){let s=e[r];if(s.from<=A&&s.to>=A){if(s.level==i)return r;(o<0||(n!=0?n<0?s.from<A:s.to>A:e[o].level>s.level))&&(o=r)}}if(o<0)throw new RangeError("Index out of range");return o}};function ftA(t,e){if(t.length!=e.length)return!1;for(let A=0;A<t.length;A++){let i=t[A],n=e[A];if(i.from!=n.from||i.to!=n.to||i.direction!=n.direction||!ftA(i.inner,n.inner))return!1}return!0}var yo=[];function zRA(t,e,A,i,n){for(let o=0;o<=i.length;o++){let r=o?i[o-1].to:e,s=o<i.length?i[o].from:A,a=o?256:n;for(let c=r,l=a,I=a;c<s;c++){let C=utA(t.charCodeAt(c));C==512?C=l:C==8&&I==4&&(C=16),yo[c]=C==4?2:C,C&7&&(I=C),l=C}for(let c=r,l=a,I=a;c<s;c++){let C=yo[c];if(C==128)c<s-1&&l==yo[c+1]&&l&24?C=yo[c]=l:yo[c]=256;else if(C==64){let d=c+1;for(;d<s&&yo[d]==64;)d++;let B=c&&l==8||d<A&&yo[d]==8?I==1?1:8:256;for(let E=c;E<d;E++)yo[E]=B;c=d-1}else C==8&&I==1&&(yo[c]=1);l=C,C&7&&(I=C)}}}function ORA(t,e,A,i,n){let o=n==1?2:1;for(let r=0,s=0,a=0;r<=i.length;r++){let c=r?i[r-1].to:e,l=r<i.length?i[r].from:A;for(let I=c,C,d,B;I<l;I++)if(d=ML[C=t.charCodeAt(I)])if(d<0){for(let E=s-3;E>=0;E-=3)if(Mg[E+1]==-d){let h=Mg[E+2],u=h&2?n:h&4?h&1?o:n:0;u&&(yo[I]=yo[Mg[E]]=u),s=E;break}}else{if(Mg.length==189)break;Mg[s++]=I,Mg[s++]=C,Mg[s++]=a}else if((B=yo[I])==2||B==1){let E=B==n;a=E?0:1;for(let h=s-3;h>=0;h-=3){let u=Mg[h+2];if(u&2)break;if(E)Mg[h+2]|=2;else{if(u&4)break;Mg[h+2]|=4}}}}}function PRA(t,e,A,i){for(let n=0,o=i;n<=A.length;n++){let r=n?A[n-1].to:t,s=n<A.length?A[n].from:e;for(let a=r;a<s;){let c=yo[a];if(c==256){let l=a+1;for(;;)if(l==s){if(n==A.length)break;l=A[n++].to,s=n<A.length?A[n].from:e}else if(yo[l]==256)l++;else break;let I=o==1,C=(l<e?yo[l]:i)==1,d=I==C?I?1:2:i;for(let B=l,E=n,h=E?A[E-1].to:t;B>a;)B==h&&(B=A[--E].from,h=E?A[E-1].to:t),yo[--B]=d;a=l}else o=c,a++}}}function kL(t,e,A,i,n,o,r){let s=i%2?2:1;if(i%2==n%2)for(let a=e,c=0;a<A;){let l=!0,I=!1;if(c==o.length||a<o[c].from){let E=yo[a];E!=s&&(l=!1,I=E==16)}let C=!l&&s==1?[]:null,d=l?i:i+1,B=a;A:for(;;)if(c<o.length&&B==o[c].from){if(I)break A;let E=o[c];if(!l)for(let h=E.to,u=c+1;;){if(h==A)break A;if(u<o.length&&o[u].from==h)h=o[u++].to;else{if(yo[h]==s)break A;break}}if(c++,C)C.push(E);else{E.from>a&&r.push(new Sg(a,E.from,d));let h=E.direction==bC!=!(d%2);SL(t,h?i+1:i,n,E.inner,E.from,E.to,r),a=E.to}B=E.to}else{if(B==A||(l?yo[B]!=s:yo[B]==s))break;B++}C?kL(t,a,B,i+1,n,C,r):a<B&&r.push(new Sg(a,B,d)),a=B}else for(let a=A,c=o.length;a>e;){let l=!0,I=!1;if(!c||a>o[c-1].to){let E=yo[a-1];E!=s&&(l=!1,I=E==16)}let C=!l&&s==1?[]:null,d=l?i:i+1,B=a;A:for(;;)if(c&&B==o[c-1].to){if(I)break A;let E=o[--c];if(!l)for(let h=E.from,u=c;;){if(h==e)break A;if(u&&o[u-1].to==h)h=o[--u].from;else{if(yo[h-1]==s)break A;break}}if(C)C.push(E);else{E.to<a&&r.push(new Sg(E.to,a,d));let h=E.direction==bC!=!(d%2);SL(t,h?i+1:i,n,E.inner,E.from,E.to,r),a=E.from}B=E.from}else{if(B==e||(l?yo[B-1]!=s:yo[B-1]==s))break;B--}C?kL(t,B,a,i+1,n,C,r):B<a&&r.push(new Sg(B,a,d)),a=B}}function SL(t,e,A,i,n,o,r){let s=e%2?2:1;zRA(t,n,o,i,s),ORA(t,n,o,i,s),PRA(n,o,i,s),kL(t,n,o,e,A,i,r)}function jRA(t,e,A){if(!t)return[new Sg(0,0,e==aF?1:0)];if(e==bC&&!A.length&&!HRA.test(t))return mtA(t.length);if(A.length)for(;t.length>yo.length;)yo[yo.length]=256;let i=[],n=e==bC?0:1;return SL(t,n,n,A,0,t.length,i),i}function mtA(t){return[new Sg(0,t,0)]}var ptA="";function qRA(t,e,A,i,n){var o;let r=i.head-t.from,s=Sg.find(e,r,(o=i.bidiLevel)!==null&&o!==void 0?o:-1,i.assoc),a=e[s],c=a.side(n,A);if(r==c){let C=s+=n?1:-1;if(C<0||C>=e.length)return null;a=e[s=C],r=a.side(!n,A),c=a.side(n,A)}let l=yr(t.text,r,a.forward(n,A));(l<a.from||l>a.to)&&(l=c),ptA=t.text.slice(Math.min(r,l),Math.max(r,l));let I=s==(n?e.length-1:0)?null:e[s+(n?1:-1)];return I&&l==c&&I.level+(n?0:1)<a.level?ae.cursor(I.side(!n,A)+t.from,I.forward(n,A)?1:-1,I.level):ae.cursor(l+t.from,a.forward(n,A)?-1:1,a.level)}function VRA(t,e,A){for(let i=e;i<A;i++){let n=utA(t.charCodeAt(i));if(n==1)return bC;if(n==2||n==4)return aF}return bC}var wtA=qe.define(),DtA=qe.define(),ytA=qe.define(),vtA=qe.define(),CL=qe.define(),btA=qe.define(),MtA=qe.define(),cF=qe.define(),lF=qe.define(),meA=qe.define({combine:t=>t.some(e=>e)}),ktA=qe.define({combine:t=>t.some(e=>e)}),StA=qe.define(),Z4=class t{constructor(e,A="nearest",i="nearest",n=5,o=5,r=!1){this.range=e,this.y=A,this.x=i,this.yMargin=n,this.xMargin=o,this.isSnapshot=r}map(e){return e.empty?this:new t(this.range.map(e),this.y,this.x,this.yMargin,this.xMargin,this.isSnapshot)}clip(e){return this.range.to<=e.doc.length?this:new t(ae.cursor(e.doc.length),this.y,this.x,this.yMargin,this.xMargin,this.isSnapshot)}},Ew=Pi.define({map:(t,e)=>t.map(e)}),RtA=Pi.define();function Xr(t,e,A){let i=t.facet(vtA);i.length?i[0](e):window.onerror&&window.onerror(String(e),A,void 0,void 0,e)||(A?console.error(A+":",e):console.error(e))}var H0=qe.define({combine:t=>t.length?t[0]:!0}),ZRA=0,SE=qe.define({combine(t){return t.filter((e,A)=>{for(let i=0;i<A;i++)if(t[i].plugin==e.plugin)return!1;return!0})}}),go=class t{constructor(e,A,i,n,o){this.id=e,this.create=A,this.domEventHandlers=i,this.domEventObservers=n,this.baseExtensions=o(this),this.extension=this.baseExtensions.concat(SE.of({plugin:this,arg:void 0}))}of(e){return this.baseExtensions.concat(SE.of({plugin:this,arg:e}))}static define(e,A){let{eventHandlers:i,eventObservers:n,provide:o,decorations:r}=A||{};return new t(ZRA++,e,i,n,s=>{let a=[];return r&&a.push(a3.of(c=>{let l=c.plugin(s);return l?r(l):ut.none})),o&&a.push(o(s)),a})}static fromClass(e,A){return t.define((i,n)=>new e(i,n),A)}},W4=class{constructor(e){this.spec=e,this.mustUpdate=null,this.value=null}get plugin(){return this.spec&&this.spec.plugin}update(e){if(this.value){if(this.mustUpdate){let A=this.mustUpdate;if(this.mustUpdate=null,this.value.update)try{this.value.update(A)}catch(i){if(Xr(A.state,i,"CodeMirror plugin crashed"),this.value.destroy)try{this.value.destroy()}catch{}this.deactivate()}}}else if(this.spec)try{this.value=this.spec.plugin.create(e,this.spec.arg)}catch(A){Xr(e.state,A,"CodeMirror plugin crashed"),this.deactivate()}return this}destroy(e){var A;if(!((A=this.value)===null||A===void 0)&&A.destroy)try{this.value.destroy()}catch(i){Xr(e.state,i,"CodeMirror plugin crashed")}}deactivate(){this.spec=this.value=null}},peA=qe.define(),RL=qe.define(),a3=qe.define(),xtA=qe.define(),gF=qe.define(),LtA=qe.define();function weA(t,e){let A=t.state.facet(LtA);if(!A.length)return A;let i=A.map(o=>o instanceof Function?o(t):o),n=[];return co.spans(i,e.from,e.to,{point(){},span(o,r,s,a){let c=o-e.from,l=r-e.from,I=n;for(let C=s.length-1;C>=0;C--,a--){let d=s[C].spec.bidiIsolate,B;if(d==null&&(d=VRA(e.text,c,l)),a>0&&I.length&&(B=I[I.length-1]).to==c&&B.direction==d)B.to=l,I=B.inner;else{let E={from:c,to:l,direction:d,inner:[]};I.push(E),I=E.inner}}}}),n}var FtA=qe.define();function IF(t){let e=0,A=0,i=0,n=0;for(let o of t.state.facet(FtA)){let r=o(t);r&&(r.left!=null&&(e=Math.max(e,r.left)),r.right!=null&&(A=Math.max(A,r.right)),r.top!=null&&(i=Math.max(i,r.top)),r.bottom!=null&&(n=Math.max(n,r.bottom)))}return{left:e,right:A,top:i,bottom:n}}var H4=qe.define(),Rg=class t{constructor(e,A,i,n){this.fromA=e,this.toA=A,this.fromB=i,this.toB=n}join(e){return new t(Math.min(this.fromA,e.fromA),Math.max(this.toA,e.toA),Math.min(this.fromB,e.fromB),Math.max(this.toB,e.toB))}addToSet(e){let A=e.length,i=this;for(;A>0;A--){let n=e[A-1];if(!(n.fromA>i.toA)){if(n.toA<i.fromA)break;i=i.join(n),e.splice(A-1,1)}}return e.splice(A,0,i),e}static extendWithRanges(e,A){if(A.length==0)return e;let i=[];for(let n=0,o=0,r=0,s=0;;n++){let a=n==e.length?null:e[n],c=r-s,l=a?a.fromB:1e9;for(;o<A.length&&A[o]<l;){let I=A[o],C=A[o+1],d=Math.max(s,I),B=Math.min(l,C);if(d<=B&&new t(d+c,B+c,d,B).addToSet(i),C>l)break;o+=2}if(!a)return i;new t(a.fromA,a.toA,a.fromB,a.toB).addToSet(i),r=a.toA,s=a.toB}}},Fw=class t{constructor(e,A,i){this.view=e,this.state=A,this.transactions=i,this.flags=0,this.startState=e.state,this.changes=Cs.empty(this.startState.doc.length);for(let o of i)this.changes=this.changes.compose(o.changes);let n=[];this.changes.iterChangedRanges((o,r,s,a)=>n.push(new Rg(o,r,s,a))),this.changedRanges=n}static create(e,A,i){return new t(e,A,i)}get viewportChanged(){return(this.flags&4)>0}get viewportMoved(){return(this.flags&8)>0}get heightChanged(){return(this.flags&2)>0}get geometryChanged(){return this.docChanged||(this.flags&18)>0}get focusChanged(){return(this.flags&1)>0}get docChanged(){return!this.changes.empty}get selectionSet(){return this.transactions.some(e=>e.selection)}get empty(){return this.flags==0&&this.transactions.length==0}},Nw=class extends Fo{get length(){return this.view.state.doc.length}constructor(e){super(),this.view=e,this.decorations=[],this.dynamicDecorationMap=[!1],this.domChanged=null,this.hasComposition=null,this.markedForComposition=new Set,this.editContextFormatting=ut.none,this.lastCompositionAfterCursor=!1,this.minWidth=0,this.minWidthFrom=0,this.minWidthTo=0,this.impreciseAnchor=null,this.impreciseHead=null,this.forceSelection=!1,this.lastUpdate=Date.now(),this.setDOM(e.contentDOM),this.children=[new Es],this.children[0].setParent(this),this.updateDeco(),this.updateInner([new Rg(0,0,0,e.state.doc.length)],0,null)}update(e){var A;let i=e.changedRanges;this.minWidth>0&&i.length&&(i.every(({fromA:c,toA:l})=>l<this.minWidthFrom||c>this.minWidthTo)?(this.minWidthFrom=e.changes.mapPos(this.minWidthFrom,1),this.minWidthTo=e.changes.mapPos(this.minWidthTo,1)):this.minWidth=this.minWidthFrom=this.minWidthTo=0),this.updateEditContextFormatting(e);let n=-1;this.view.inputState.composing>=0&&!this.view.observer.editContext&&(!((A=this.domChanged)===null||A===void 0)&&A.newSel?n=this.domChanged.newSel.head:!ixA(e.changes,this.hasComposition)&&!e.selectionSet&&(n=e.state.selection.main.head));let o=n>-1?XRA(this.view,e.changes,n):null;if(this.domChanged=null,this.hasComposition){this.markedForComposition.clear();let{from:c,to:l}=this.hasComposition;i=new Rg(c,l,e.changes.mapPos(c,-1),e.changes.mapPos(l,1)).addToSet(i.slice())}this.hasComposition=o?{from:o.range.fromB,to:o.range.toB}:null,(At.ie||At.chrome)&&!o&&e&&e.state.doc.lines!=e.startState.doc.lines&&(this.forceSelection=!0);let r=this.decorations,s=this.updateDeco(),a=exA(r,s,e.changes);return i=Rg.extendWithRanges(i,a),!(this.flags&7)&&i.length==0?!1:(this.updateInner(i,e.startState.doc.length,o),e.transactions.length&&(this.lastUpdate=Date.now()),!0)}updateInner(e,A,i){this.view.viewState.mustMeasureContent=!0,this.updateChildren(e,A,i);let{observer:n}=this.view;n.ignore(()=>{this.dom.style.height=this.view.viewState.contentHeight/this.view.scaleY+"px",this.dom.style.flexBasis=this.minWidth?this.minWidth+"px":"";let r=At.chrome||At.ios?{node:n.selectionRange.focusNode,written:!1}:void 0;this.sync(this.view,r),this.flags&=-8,r&&(r.written||n.selectionRange.focusNode!=r.node)&&(this.forceSelection=!0),this.dom.style.height=""}),this.markedForComposition.forEach(r=>r.flags&=-9);let o=[];if(this.view.viewport.from||this.view.viewport.to<this.view.state.doc.length)for(let r of this.children)r instanceof DC&&r.widget instanceof s3&&o.push(r.dom);n.updateGaps(o)}updateChildren(e,A,i){let n=i?i.range.addToSet(e.slice()):e,o=this.childCursor(A);for(let r=n.length-1;;r--){let s=r>=0?n[r]:null;if(!s)break;let{fromA:a,toA:c,fromB:l,toB:I}=s,C,d,B,E;if(i&&i.range.fromB<I&&i.range.toB>l){let R=V4.build(this.view.state.doc,l,i.range.fromB,this.decorations,this.dynamicDecorationMap),w=V4.build(this.view.state.doc,i.range.toB,I,this.decorations,this.dynamicDecorationMap);d=R.breakAtStart,B=R.openStart,E=w.openEnd;let _=this.compositionView(i);w.breakAtStart?_.breakAfter=1:w.content.length&&_.merge(_.length,_.length,w.content[0],!1,w.openStart,0)&&(_.breakAfter=w.content[0].breakAfter,w.content.shift()),R.content.length&&_.merge(0,0,R.content[R.content.length-1],!0,0,R.openEnd)&&R.content.pop(),C=R.content.concat(_).concat(w.content)}else({content:C,breakAtStart:d,openStart:B,openEnd:E}=V4.build(this.view.state.doc,l,I,this.decorations,this.dynamicDecorationMap));let{i:h,off:u}=o.findPos(c,1),{i:D,off:L}=o.findPos(a,-1);ltA(this,D,L,h,u,C,d,B,E)}i&&this.fixCompositionDOM(i)}updateEditContextFormatting(e){this.editContextFormatting=this.editContextFormatting.map(e.changes);for(let A of e.transactions)for(let i of A.effects)i.is(RtA)&&(this.editContextFormatting=i.value)}compositionView(e){let A=new Lg(e.text.nodeValue);A.flags|=8;for(let{deco:n}of e.marks)A=new D1(n,[A],A.length);let i=new Es;return i.append(A,0),i}fixCompositionDOM(e){let A=(o,r)=>{r.flags|=8|(r.children.some(a=>a.flags&7)?1:0),this.markedForComposition.add(r);let s=Fo.get(o);s&&s!=r&&(s.dom=null),r.setDOM(o)},i=this.childPos(e.range.fromB,1),n=this.children[i.i];A(e.line,n);for(let o=e.marks.length-1;o>=-1;o--)i=n.childPos(i.off,1),n=n.children[i.i],A(o>=0?e.marks[o].node:e.text,n)}updateSelection(e=!1,A=!1){(e||!this.view.observer.selectionRange.focusNode)&&this.view.observer.readSelectionRange();let i=this.view.root.activeElement,n=i==this.dom,o=!n&&!(this.view.state.facet(H0)||this.dom.tabIndex>-1)&&Dw(this.dom,this.view.observer.selectionRange)&&!(i&&this.dom.contains(i));if(!(n||A||o))return;let r=this.forceSelection;this.forceSelection=!1;let s=this.view.state.selection.main,a=this.moveToLine(this.domAtPos(s.anchor)),c=s.empty?a:this.moveToLine(this.domAtPos(s.head));if(At.gecko&&s.empty&&!this.hasComposition&&WRA(a)){let I=document.createTextNode("");this.view.observer.ignore(()=>a.node.insertBefore(I,a.node.childNodes[a.offset]||null)),a=c=new Xs(I,0),r=!0}let l=this.view.observer.selectionRange;(r||!l.focusNode||(!q4(a.node,a.offset,l.anchorNode,l.anchorOffset)||!q4(c.node,c.offset,l.focusNode,l.focusOffset))&&!this.suppressWidgetCursorChange(l,s))&&(this.view.observer.ignore(()=>{At.android&&At.chrome&&this.dom.contains(l.focusNode)&&txA(l.focusNode,this.dom)&&(this.dom.blur(),this.dom.focus({preventScroll:!0}));let I=e3(this.view.root);if(I)if(s.empty){if(At.gecko){let C=$RA(a.node,a.offset);if(C&&C!=3){let d=(C==1?atA:ctA)(a.node,a.offset);d&&(a=new Xs(d.node,d.offset))}}I.collapse(a.node,a.offset),s.bidiLevel!=null&&I.caretBidiLevel!==void 0&&(I.caretBidiLevel=s.bidiLevel)}else if(I.extend){I.collapse(a.node,a.offset);try{I.extend(c.node,c.offset)}catch{}}else{let C=document.createRange();s.anchor>s.head&&([a,c]=[c,a]),C.setEnd(c.node,c.offset),C.setStart(a.node,a.offset),I.removeAllRanges(),I.addRange(C)}o&&this.view.root.activeElement==this.dom&&(this.dom.blur(),i&&i.focus())}),this.view.observer.setSelectionRange(a,c)),this.impreciseAnchor=a.precise?null:new Xs(l.anchorNode,l.anchorOffset),this.impreciseHead=c.precise?null:new Xs(l.focusNode,l.focusOffset)}suppressWidgetCursorChange(e,A){return this.hasComposition&&A.empty&&q4(e.focusNode,e.focusOffset,e.anchorNode,e.anchorOffset)&&this.posFromDOM(e.focusNode,e.focusOffset)==A.head}enforceCursorAssoc(){if(this.hasComposition)return;let{view:e}=this,A=e.state.selection.main,i=e3(e.root),{anchorNode:n,anchorOffset:o}=e.observer.selectionRange;if(!i||!A.empty||!A.assoc||!i.modify)return;let r=Es.find(this,A.head);if(!r)return;let s=r.posAtStart;if(A.head==s||A.head==s+r.length)return;let a=this.coordsAt(A.head,-1),c=this.coordsAt(A.head,1);if(!a||!c||a.bottom>c.top)return;let l=this.domAtPos(A.head+A.assoc);i.collapse(l.node,l.offset),i.modify("move",A.assoc<0?"forward":"backward","lineboundary"),e.observer.readSelectionRange();let I=e.observer.selectionRange;e.docView.posFromDOM(I.anchorNode,I.anchorOffset)!=A.from&&i.collapse(n,o)}moveToLine(e){let A=this.dom,i;if(e.node!=A)return e;for(let n=e.offset;!i&&n<A.childNodes.length;n++){let o=Fo.get(A.childNodes[n]);o instanceof Es&&(i=o.domAtPos(0))}for(let n=e.offset-1;!i&&n>=0;n--){let o=Fo.get(A.childNodes[n]);o instanceof Es&&(i=o.domAtPos(o.length))}return i?new Xs(i.node,i.offset,!0):e}nearest(e){for(let A=e;A;){let i=Fo.get(A);if(i&&i.rootView==this)return i;A=A.parentNode}return null}posFromDOM(e,A){let i=this.nearest(e);if(!i)throw new RangeError("Trying to find position for a DOM position outside of the document");return i.localPosFromDOM(e,A)+i.posAtStart}domAtPos(e){let{i:A,off:i}=this.childCursor().findPos(e,-1);for(;A<this.children.length-1;){let n=this.children[A];if(i<n.length||n instanceof Es)break;A++,i=0}return this.children[A].domAtPos(i)}coordsAt(e,A){let i=null,n=0;for(let o=this.length,r=this.children.length-1;r>=0;r--){let s=this.children[r],a=o-s.breakAfter,c=a-s.length;if(a<e)break;if(c<=e&&(c<e||s.covers(-1))&&(a>e||s.covers(1))&&(!i||s instanceof Es&&!(i instanceof Es&&A>=0)))i=s,n=c;else if(i&&c==e&&a==e&&s instanceof DC&&Math.abs(A)<2){if(s.deco.startSide<0)break;r&&(i=null)}o=c}return i?i.coordsAt(e-n,A):null}coordsForChar(e){let{i:A,off:i}=this.childPos(e,1),n=this.children[A];if(!(n instanceof Es))return null;for(;n.children.length;){let{i:s,off:a}=n.childPos(i,1);for(;;s++){if(s==n.children.length)return null;if((n=n.children[s]).length)break}i=a}if(!(n instanceof Lg))return null;let o=yr(n.text,i);if(o==i)return null;let r=vC(n.dom,i,o).getClientRects();for(let s=0;s<r.length;s++){let a=r[s];if(s==r.length-1||a.top<a.bottom&&a.left<a.right)return a}return null}measureVisibleLineHeights(e){let A=[],{from:i,to:n}=e,o=this.view.contentDOM.clientWidth,r=o>Math.max(this.view.scrollDOM.clientWidth,this.minWidth)+1,s=-1,a=this.view.textDirection==lo.LTR;for(let c=0,l=0;l<this.children.length;l++){let I=this.children[l],C=c+I.length;if(C>n)break;if(c>=i){let d=I.dom.getBoundingClientRect();if(A.push(d.height),r){let B=I.dom.lastChild,E=B?t3(B):[];if(E.length){let h=E[E.length-1],u=a?h.right-d.left:d.right-h.left;u>s&&(s=u,this.minWidth=o,this.minWidthFrom=c,this.minWidthTo=C)}}}c=C+I.breakAfter}return A}textDirectionAt(e){let{i:A}=this.childPos(e,1);return getComputedStyle(this.children[A].dom).direction=="rtl"?lo.RTL:lo.LTR}measureTextSize(){for(let o of this.children)if(o instanceof Es){let r=o.measureTextSize();if(r)return r}let e=document.createElement("div"),A,i,n;return e.className="cm-line",e.style.width="99999px",e.style.position="absolute",e.textContent="abc def ghi jkl mno pqr stu",this.view.observer.ignore(()=>{this.dom.appendChild(e);let o=t3(e.firstChild)[0];A=e.getBoundingClientRect().height,i=o?o.width/27:7,n=o?o.height:A,e.remove()}),{lineHeight:A,charWidth:i,textHeight:n}}childCursor(e=this.length){let A=this.children.length;return A&&(e-=this.children[--A].length),new xw(this.children,e,A)}computeBlockGapDeco(){let e=[],A=this.view.viewState;for(let i=0,n=0;;n++){let o=n==A.viewports.length?null:A.viewports[n],r=o?o.from-1:this.length;if(r>i){let s=(A.lineBlockAt(r).bottom-A.lineBlockAt(i).top)/this.view.scaleY;e.push(ut.replace({widget:new s3(s),block:!0,inclusive:!0,isBlockGap:!0}).range(i,r))}if(!o)break;i=o.to+1}return ut.set(e)}updateDeco(){let e=1,A=this.view.state.facet(a3).map(o=>(this.dynamicDecorationMap[e++]=typeof o=="function")?o(this.view):o),i=!1,n=this.view.state.facet(xtA).map((o,r)=>{let s=typeof o=="function";return s&&(i=!0),s?o(this.view):o});for(n.length&&(this.dynamicDecorationMap[e++]=i,A.push(co.join(n))),this.decorations=[this.editContextFormatting,...A,this.computeBlockGapDeco(),this.view.viewState.lineGapDeco];e<this.decorations.length;)this.dynamicDecorationMap[e++]=!1;return this.decorations}scrollIntoView(e){if(e.isSnapshot){let c=this.view.viewState.lineBlockAt(e.range.head);this.view.scrollDOM.scrollTop=c.top-e.yMargin,this.view.scrollDOM.scrollLeft=e.xMargin;return}for(let c of this.view.state.facet(StA))try{if(c(this.view,e.range,e))return!0}catch(l){Xr(this.view.state,l,"scroll handler")}let{range:A}=e,i=this.coordsAt(A.head,A.empty?A.assoc:A.head>A.anchor?-1:1),n;if(!i)return;!A.empty&&(n=this.coordsAt(A.anchor,A.anchor>A.head?-1:1))&&(i={left:Math.min(i.left,n.left),top:Math.min(i.top,n.top),right:Math.max(i.right,n.right),bottom:Math.max(i.bottom,n.bottom)});let o=IF(this.view),r={left:i.left-o.left,top:i.top-o.top,right:i.right+o.right,bottom:i.bottom+o.bottom},{offsetWidth:s,offsetHeight:a}=this.view.scrollDOM;RRA(this.view.scrollDOM,r,A.head<A.anchor?-1:1,e.x,e.y,Math.max(Math.min(e.xMargin,s),-s),Math.max(Math.min(e.yMargin,a),-a),this.view.textDirection==lo.LTR)}};function WRA(t){return t.node.nodeType==1&&t.node.firstChild&&(t.offset==0||t.node.childNodes[t.offset-1].contentEditable=="false")&&(t.offset==t.node.childNodes.length||t.node.childNodes[t.offset].contentEditable=="false")}function NtA(t,e){let A=t.observer.selectionRange;if(!A.focusNode)return null;let i=atA(A.focusNode,A.focusOffset),n=ctA(A.focusNode,A.focusOffset),o=i||n;if(n&&i&&n.node!=i.node){let s=Fo.get(n.node);if(!s||s instanceof Lg&&s.text!=n.node.nodeValue)o=n;else if(t.docView.lastCompositionAfterCursor){let a=Fo.get(i.node);!a||a instanceof Lg&&a.text!=i.node.nodeValue||(o=n)}}if(t.docView.lastCompositionAfterCursor=o!=i,!o)return null;let r=e-o.offset;return{from:r,to:r+o.node.nodeValue.length,node:o.node}}function XRA(t,e,A){let i=NtA(t,A);if(!i)return null;let{node:n,from:o,to:r}=i,s=n.nodeValue;if(/[\n\r]/.test(s)||t.state.doc.sliceString(i.from,i.to)!=s)return null;let a=e.invertedDesc,c=new Rg(a.mapPos(o),a.mapPos(r),o,r),l=[];for(let I=n.parentNode;;I=I.parentNode){let C=Fo.get(I);if(C instanceof D1)l.push({node:I,deco:C.mark});else{if(C instanceof Es||I.nodeName=="DIV"&&I.parentNode==t.contentDOM)return{range:c,text:n,marks:l,line:I};if(I!=t.contentDOM)l.push({node:I,deco:new o3({inclusive:!0,attributes:KRA(I),tagName:I.tagName.toLowerCase()})});else return null}}}function $RA(t,e){return t.nodeType!=1?0:(e&&t.childNodes[e-1].contentEditable=="false"?1:0)|(e<t.childNodes.length&&t.childNodes[e].contentEditable=="false"?2:0)}var AxA=class{constructor(){this.changes=[]}compareRange(e,A){yw(e,A,this.changes)}comparePoint(e,A){yw(e,A,this.changes)}boundChange(e){yw(e,e,this.changes)}};function exA(t,e,A){let i=new AxA;return co.compare(t,e,A,i),i.changes}function txA(t,e){for(let A=t;A&&A!=e;A=A.assignedSlot||A.parentNode)if(A.nodeType==1&&A.contentEditable=="false")return!0;return!1}function ixA(t,e){let A=!1;return e&&t.iterChangedRanges((i,n)=>{i<e.to&&n>e.from&&(A=!0)}),A}function nxA(t,e,A=1){let i=t.charCategorizer(e),n=t.doc.lineAt(e),o=e-n.from;if(n.length==0)return ae.cursor(e);o==0?A=1:o==n.length&&(A=-1);let r=o,s=o;A<0?r=yr(n.text,o,!1):s=yr(n.text,o);let a=i(n.text.slice(r,s));for(;r>0;){let c=yr(n.text,r,!1);if(i(n.text.slice(c,r))!=a)break;r=c}for(;s<n.length;){let c=yr(n.text,s);if(i(n.text.slice(s,c))!=a)break;s=c}return ae.range(r+n.from,s+n.from)}function oxA(t,e){return e.left>t?e.left-t:Math.max(0,t-e.right)}function rxA(t,e){return e.top>t?e.top-t:Math.max(0,t-e.bottom)}function dL(t,e){return t.top<e.bottom-1&&t.bottom>e.top+1}function DeA(t,e){return e<t.top?{top:e,left:t.left,right:t.right,bottom:t.bottom}:t}function yeA(t,e){return e>t.bottom?{top:t.top,left:t.left,right:t.right,bottom:e}:t}function xL(t,e,A){let i,n,o,r,s=!1,a,c,l,I;for(let B=t.firstChild;B;B=B.nextSibling){let E=t3(B);for(let h=0;h<E.length;h++){let u=E[h];n&&dL(n,u)&&(u=DeA(yeA(u,n.bottom),n.top));let D=oxA(e,u),L=rxA(A,u);if(D==0&&L==0)return B.nodeType==3?veA(B,e,A):xL(B,e,A);(!i||r>L||r==L&&o>D)&&(i=B,n=u,o=D,r=L,s=D?e<u.left?h>0:h<E.length-1:!0),D==0?A>u.bottom&&(!l||l.bottom<u.bottom)?(a=B,l=u):A<u.top&&(!I||I.top>u.top)&&(c=B,I=u):l&&dL(l,u)?l=yeA(l,u.bottom):I&&dL(I,u)&&(I=DeA(I,u.top))}}if(l&&l.bottom>=A?(i=a,n=l):I&&I.top<=A&&(i=c,n=I),!i)return{node:t,offset:0};let C=Math.max(n.left,Math.min(n.right,e));if(i.nodeType==3)return veA(i,C,A);if(s&&i.contentEditable!="false")return xL(i,C,A);let d=Array.prototype.indexOf.call(t.childNodes,i)+(e>=(n.left+n.right)/2?1:0);return{node:t,offset:d}}function veA(t,e,A){let i=t.nodeValue.length,n=-1,o=1e9,r=0;for(let s=0;s<i;s++){let a=vC(t,s,s+1).getClientRects();for(let c=0;c<a.length;c++){let l=a[c];if(l.top==l.bottom)continue;r||(r=e-l.left);let I=(l.top>A?l.top-A:A-l.bottom)-1;if(l.left-1<=e&&l.right+1>=e&&I<o){let C=e>=(l.left+l.right)/2,d=C;if((At.chrome||At.gecko)&&vC(t,s).getBoundingClientRect().left==l.right&&(d=!C),I<=0)return{node:t,offset:s+(d?1:0)};n=s+(d?1:0),o=I}}}return{node:t,offset:n>-1?n:r>0?t.nodeValue.length:0}}function _tA(t,e,A,i=-1){var n,o;let r=t.contentDOM.getBoundingClientRect(),s=r.top+t.viewState.paddingTop,a,{docHeight:c}=t.viewState,{x:l,y:I}=e,C=I-s;if(C<0)return 0;if(C>c)return t.state.doc.length;for(let R=t.viewState.heightOracle.textHeight/2,w=!1;a=t.elementAtHeight(C),a.type!=$s.Text;)for(;C=i>0?a.bottom+R:a.top-R,!(C>=0&&C<=c);){if(w)return A?null:0;w=!0,i=-i}I=s+C;let d=a.from;if(d<t.viewport.from)return t.viewport.from==0?0:A?null:beA(t,r,a,l,I);if(d>t.viewport.to)return t.viewport.to==t.state.doc.length?t.state.doc.length:A?null:beA(t,r,a,l,I);let B=t.dom.ownerDocument,E=t.root.elementFromPoint?t.root:B,h=E.elementFromPoint(l,I);h&&!t.contentDOM.contains(h)&&(h=null),h||(l=Math.max(r.left+1,Math.min(r.right-1,l)),h=E.elementFromPoint(l,I),h&&!t.contentDOM.contains(h)&&(h=null));let u,D=-1;if(h&&((n=t.docView.nearest(h))===null||n===void 0?void 0:n.isEditable)!=!1){if(B.caretPositionFromPoint){let R=B.caretPositionFromPoint(l,I);R&&({offsetNode:u,offset:D}=R)}else if(B.caretRangeFromPoint){let R=B.caretRangeFromPoint(l,I);R&&({startContainer:u,startOffset:D}=R,(!t.contentDOM.contains(u)||At.safari&&sxA(u,D,l)||At.chrome&&axA(u,D,l))&&(u=void 0))}u&&(D=Math.min(xg(u),D))}if(!u||!t.docView.dom.contains(u)){let R=Es.find(t.docView,d);if(!R)return C>a.top+a.height/2?a.to:a.from;({node:u,offset:D}=xL(R.dom,l,I))}let L=t.docView.nearest(u);if(!L)return null;if(L.isWidget&&((o=L.dom)===null||o===void 0?void 0:o.nodeType)==1){let R=L.dom.getBoundingClientRect();return e.y<R.top||e.y<=R.bottom&&e.x<=(R.left+R.right)/2?L.posAtStart:L.posAtEnd}else return L.localPosFromDOM(u,D)+L.posAtStart}function beA(t,e,A,i,n){let o=Math.round((i-e.left)*t.defaultCharacterWidth);if(t.lineWrapping&&A.height>t.defaultLineHeight*1.5){let s=t.viewState.heightOracle.textHeight,a=Math.floor((n-A.top-(t.defaultLineHeight-s)*.5)/s);o+=a*t.viewState.heightOracle.lineLength}let r=t.state.sliceDoc(A.from,A.to);return A.from+Cw(r,o,t.state.tabSize)}function sxA(t,e,A){let i,n=t;if(t.nodeType!=3||e!=(i=t.nodeValue.length))return!1;for(;;){let o=n.nextSibling;if(o){if(o.nodeName=="BR")break;return!1}else{let r=n.parentNode;if(!r||r.nodeName=="DIV")break;n=r}}return vC(t,i-1,i).getBoundingClientRect().right>A}function axA(t,e,A){if(e!=0)return!1;for(let n=t;;){let o=n.parentNode;if(!o||o.nodeType!=1||o.firstChild!=n)return!1;if(o.classList.contains("cm-line"))break;n=o}let i=t.nodeType==1?t.getBoundingClientRect():vC(t,0,Math.max(t.nodeValue.length,1)).getBoundingClientRect();return A-i.left>5}function LL(t,e,A){let i=t.lineBlockAt(e);if(Array.isArray(i.type)){let n;for(let o of i.type){if(o.from>e)break;if(!(o.to<e)){if(o.from<e&&o.to>e)return o;(!n||o.type==$s.Text&&(n.type!=o.type||(A<0?o.from<e:o.to>e)))&&(n=o)}}return n||i}return i}function cxA(t,e,A,i){let n=LL(t,e.head,e.assoc||-1),o=!i||n.type!=$s.Text||!(t.lineWrapping||n.widgetLineBreaks)?null:t.coordsAtPos(e.assoc<0&&e.head>n.from?e.head-1:e.head);if(o){let r=t.dom.getBoundingClientRect(),s=t.textDirectionAt(n.from),a=t.posAtCoords({x:A==(s==lo.LTR)?r.right-1:r.left+1,y:(o.top+o.bottom)/2});if(a!=null)return ae.cursor(a,A?-1:1)}return ae.cursor(A?n.to:n.from,A?-1:1)}function MeA(t,e,A,i){let n=t.state.doc.lineAt(e.head),o=t.bidiSpans(n),r=t.textDirectionAt(n.from);for(let s=e,a=null;;){let c=qRA(n,o,r,s,A),l=ptA;if(!c){if(n.number==(A?t.state.doc.lines:1))return s;l=`
`,n=t.state.doc.line(n.number+(A?1:-1)),o=t.bidiSpans(n),c=t.visualLineSide(n,!A)}if(a){if(!a(l))return s}else{if(!i)return c;a=i(l)}s=c}}function lxA(t,e,A){let i=t.state.charCategorizer(e),n=i(A);return o=>{let r=i(o);return n==ao.Space&&(n=r),n==r}}function gxA(t,e,A,i){let n=e.head,o=A?1:-1;if(n==(A?t.state.doc.length:0))return ae.cursor(n,e.assoc);let r=e.goalColumn,s,a=t.contentDOM.getBoundingClientRect(),c=t.coordsAtPos(n,e.assoc||-1),l=t.documentTop;if(c)r==null&&(r=c.left-a.left),s=o<0?c.top:c.bottom;else{let d=t.viewState.lineBlockAt(n);r==null&&(r=Math.min(a.right-a.left,t.defaultCharacterWidth*(n-d.from))),s=(o<0?d.top:d.bottom)+l}let I=a.left+r,C=i??t.viewState.heightOracle.textHeight>>1;for(let d=0;;d+=10){let B=s+(C+d)*o,E=_tA(t,{x:I,y:B},!1,o);if(B<a.top||B>a.bottom||(o<0?E<n:E>n)){let h=t.docView.coordsForChar(E),u=!h||B<h.top?-1:1;return ae.cursor(E,u,void 0,r)}}}function vw(t,e,A){for(;;){let i=0;for(let n of t)n.between(e-1,e+1,(o,r,s)=>{if(e>o&&e<r){let a=i||A||(e-o<r-e?-1:1);e=a<0?o:r,i=a}});if(!i)return e}}function BL(t,e,A){let i=vw(t.state.facet(gF).map(n=>n(t)),A.from,e.head>A.from?-1:1);return i==A.from?A:ae.cursor(i,i<A.from?1:-1)}var O4="\uFFFF",FL=class{constructor(e,A){this.points=e,this.text="",this.lineSeparator=A.facet(hr.lineSeparator)}append(e){this.text+=e}lineBreak(){this.text+=O4}readRange(e,A){if(!e)return this;let i=e.parentNode;for(let n=e;;){this.findPointBefore(i,n);let o=this.text.length;this.readNode(n);let r=n.nextSibling;if(r==A)break;let s=Fo.get(n),a=Fo.get(r);(s&&a?s.breakAfter:(s?s.breakAfter:Rw(n))||Rw(r)&&(n.nodeName!="BR"||n.cmIgnore)&&this.text.length>o)&&this.lineBreak(),n=r}return this.findPointBefore(i,A),this}readTextNode(e){let A=e.nodeValue;for(let i of this.points)i.node==e&&(i.pos=this.text.length+Math.min(i.offset,A.length));for(let i=0,n=this.lineSeparator?null:/\r\n?|\n/g;;){let o=-1,r=1,s;if(this.lineSeparator?(o=A.indexOf(this.lineSeparator,i),r=this.lineSeparator.length):(s=n.exec(A))&&(o=s.index,r=s[0].length),this.append(A.slice(i,o<0?A.length:o)),o<0)break;if(this.lineBreak(),r>1)for(let a of this.points)a.node==e&&a.pos>this.text.length&&(a.pos-=r-1);i=o+r}}readNode(e){if(e.cmIgnore)return;let A=Fo.get(e),i=A&&A.overrideDOMText;if(i!=null){this.findPointInside(e,i.length);for(let n=i.iter();!n.next().done;)n.lineBreak?this.lineBreak():this.append(n.value)}else e.nodeType==3?this.readTextNode(e):e.nodeName=="BR"?e.nextSibling&&this.lineBreak():e.nodeType==1&&this.readRange(e.firstChild,null)}findPointBefore(e,A){for(let i of this.points)i.node==e&&e.childNodes[i.offset]==A&&(i.pos=this.text.length)}findPointInside(e,A){for(let i of this.points)(e.nodeType==3?i.node==e:e.contains(i.node))&&(i.pos=this.text.length+(IxA(e,i.node,i.offset)?A:0))}};function IxA(t,e,A){for(;;){if(!e||A<xg(e))return!1;if(e==t)return!0;A=yC(e)+1,e=e.parentNode}}var _w=class{constructor(e,A){this.node=e,this.offset=A,this.pos=-1}},NL=class{constructor(e,A,i,n){this.typeOver=n,this.bounds=null,this.text="",this.domChanged=A>-1;let{impreciseHead:o,impreciseAnchor:r}=e.docView;if(e.state.readOnly&&A>-1)this.newSel=null;else if(A>-1&&(this.bounds=e.docView.domBoundsAround(A,i,0))){let s=o||r?[]:BxA(e),a=new FL(s,e.state);a.readRange(this.bounds.startDOM,this.bounds.endDOM),this.text=a.text,this.newSel=ExA(s,this.bounds.from)}else{let s=e.observer.selectionRange,a=o&&o.node==s.focusNode&&o.offset==s.focusOffset||!mL(e.contentDOM,s.focusNode)?e.state.selection.main.head:e.docView.posFromDOM(s.focusNode,s.focusOffset),c=r&&r.node==s.anchorNode&&r.offset==s.anchorOffset||!mL(e.contentDOM,s.anchorNode)?e.state.selection.main.anchor:e.docView.posFromDOM(s.anchorNode,s.anchorOffset),l=e.viewport;if((At.ios||At.chrome)&&e.state.selection.main.empty&&a!=c&&(l.from>0||l.to<e.state.doc.length)){let I=Math.min(a,c),C=Math.max(a,c),d=l.from-I,B=l.to-C;(d==0||d==1||I==0)&&(B==0||B==-1||C==e.state.doc.length)&&(a=0,c=e.state.doc.length)}this.newSel=ae.single(c,a)}}};function GtA(t,e){let A,{newSel:i}=e,n=t.state.selection.main,o=t.inputState.lastKeyTime>Date.now()-100?t.inputState.lastKeyCode:-1;if(e.bounds){let{from:r,to:s}=e.bounds,a=n.from,c=null;(o===8||At.android&&e.text.length<s-r)&&(a=n.to,c="end");let l=dxA(t.state.doc.sliceString(r,s,O4),e.text,a-r,c);l&&(At.chrome&&o==13&&l.toB==l.from+2&&e.text.slice(l.from,l.toB)==O4+O4&&l.toB--,A={from:r+l.from,to:r+l.toA,insert:In.of(e.text.slice(l.from,l.toB).split(O4))})}else i&&(!t.hasFocus&&t.state.facet(H0)||i.main.eq(n))&&(i=null);if(!A&&!i)return!1;if(!A&&e.typeOver&&!n.empty&&i&&i.main.empty?A={from:n.from,to:n.to,insert:t.state.doc.slice(n.from,n.to)}:(At.mac||At.android)&&A&&A.from==A.to&&A.from==n.head-1&&/^\. ?$/.test(A.insert.toString())&&t.contentDOM.getAttribute("autocorrect")=="off"?(i&&A.insert.length==2&&(i=ae.single(i.main.anchor-1,i.main.head-1)),A={from:A.from,to:A.to,insert:In.of([A.insert.toString().replace("."," ")])}):A&&A.from>=n.from&&A.to<=n.to&&(A.from!=n.from||A.to!=n.to)&&n.to-n.from-(A.to-A.from)<=4?A={from:n.from,to:n.to,insert:t.state.doc.slice(n.from,A.from).append(A.insert).append(t.state.doc.slice(A.to,n.to))}:At.chrome&&A&&A.from==A.to&&A.from==n.head&&A.insert.toString()==`
 `&&t.lineWrapping&&(i&&(i=ae.single(i.main.anchor-1,i.main.head-1)),A={from:n.from,to:n.to,insert:In.of([" "])}),A)return CF(t,A,i,o);if(i&&!i.main.eq(n)){let r=!1,s="select";return t.inputState.lastSelectionTime>Date.now()-50&&(t.inputState.lastSelectionOrigin=="select"&&(r=!0),s=t.inputState.lastSelectionOrigin),t.dispatch({selection:i,scrollIntoView:r,userEvent:s}),!0}else return!1}function CF(t,e,A,i=-1){if(At.ios&&t.inputState.flushIOSKey(e))return!0;let n=t.state.selection.main;if(At.android&&(e.to==n.to&&(e.from==n.from||e.from==n.from-1&&t.state.sliceDoc(e.from,n.from)==" ")&&e.insert.length==1&&e.insert.lines==2&&FE(t.contentDOM,"Enter",13)||(e.from==n.from-1&&e.to==n.to&&e.insert.length==0||i==8&&e.insert.length<e.to-e.from&&e.to>n.head)&&FE(t.contentDOM,"Backspace",8)||e.from==n.from&&e.to==n.to+1&&e.insert.length==0&&FE(t.contentDOM,"Delete",46)))return!0;let o=e.insert.toString();t.inputState.composing>=0&&t.inputState.composing++;let r,s=()=>r||(r=CxA(t,e,A));return t.state.facet(btA).some(a=>a(t,e.from,e.to,o,s))||t.dispatch(s()),!0}function CxA(t,e,A){let i,n=t.state,o=n.selection.main;if(e.from>=o.from&&e.to<=o.to&&e.to-e.from>=(o.to-o.from)/3&&(!A||A.main.empty&&A.main.from==e.from+e.insert.length)&&t.inputState.composing<0){let s=o.from<e.from?n.sliceDoc(o.from,e.from):"",a=o.to>e.to?n.sliceDoc(e.to,o.to):"";i=n.replaceSelection(t.state.toText(s+e.insert.sliceString(0,void 0,t.state.lineBreak)+a))}else{let s=n.changes(e),a=A&&A.main.to<=s.newLength?A.main:void 0;if(n.selection.ranges.length>1&&t.inputState.composing>=0&&e.to<=o.to&&e.to>=o.to-10){let c=t.state.sliceDoc(e.from,e.to),l,I=A&&NtA(t,A.main.head);if(I){let B=e.insert.length-(e.to-e.from);l={from:I.from,to:I.to-B}}else l=t.state.doc.lineAt(o.head);let C=o.to-e.to,d=o.to-o.from;i=n.changeByRange(B=>{if(B.from==o.from&&B.to==o.to)return{changes:s,range:a||B.map(s)};let E=B.to-C,h=E-c.length;if(B.to-B.from!=d||t.state.sliceDoc(h,E)!=c||B.to>=l.from&&B.from<=l.to)return{range:B};let u=n.changes({from:h,to:E,insert:e.insert}),D=B.to-o.to;return{changes:u,range:a?ae.range(Math.max(0,a.anchor+D),Math.max(0,a.head+D)):B.map(u)}})}else i={changes:s,selection:a&&n.selection.replaceRange(a)}}let r="input.type";return(t.composing||t.inputState.compositionPendingChange&&t.inputState.compositionEndedAt>Date.now()-50)&&(t.inputState.compositionPendingChange=!1,r+=".compose",t.inputState.compositionFirstChange&&(r+=".start",t.inputState.compositionFirstChange=!1)),n.update(i,{userEvent:r,scrollIntoView:!0})}function dxA(t,e,A,i){let n=Math.min(t.length,e.length),o=0;for(;o<n&&t.charCodeAt(o)==e.charCodeAt(o);)o++;if(o==n&&t.length==e.length)return null;let r=t.length,s=e.length;for(;r>0&&s>0&&t.charCodeAt(r-1)==e.charCodeAt(s-1);)r--,s--;if(i=="end"){let a=Math.max(0,o-Math.min(r,s));A-=r+a-o}if(r<o&&t.length<e.length){let a=A<=o&&A>=r?o-A:0;o-=a,s=o+(s-r),r=o}else if(s<o){let a=A<=o&&A>=s?o-A:0;o-=a,r=o+(r-s),s=o}return{from:o,toA:r,toB:s}}function BxA(t){let e=[];if(t.root.activeElement!=t.contentDOM)return e;let{anchorNode:A,anchorOffset:i,focusNode:n,focusOffset:o}=t.observer.selectionRange;return A&&(e.push(new _w(A,i)),(n!=A||o!=i)&&e.push(new _w(n,o))),e}function ExA(t,e){if(t.length==0)return null;let A=t[0].pos,i=t.length==2?t[1].pos:A;return A>-1&&i>-1?ae.single(A+e,i+e):null}var _L=class{setSelectionOrigin(e){this.lastSelectionOrigin=e,this.lastSelectionTime=Date.now()}constructor(e){this.view=e,this.lastKeyCode=0,this.lastKeyTime=0,this.lastTouchTime=0,this.lastFocusTime=0,this.lastScrollTop=0,this.lastScrollLeft=0,this.pendingIOSKey=void 0,this.tabFocusMode=-1,this.lastSelectionOrigin=null,this.lastSelectionTime=0,this.lastContextMenu=0,this.scrollHandlers=[],this.handlers=Object.create(null),this.composing=-1,this.compositionFirstChange=null,this.compositionEndedAt=0,this.compositionPendingKey=!1,this.compositionPendingChange=!1,this.mouseSelection=null,this.draggedContent=null,this.handleEvent=this.handleEvent.bind(this),this.notifiedFocused=e.hasFocus,At.safari&&e.contentDOM.addEventListener("input",()=>null),At.gecko&&RxA(e.contentDOM.ownerDocument)}handleEvent(e){!wxA(this.view,e)||this.ignoreDuringComposition(e)||e.type=="keydown"&&this.keydown(e)||(this.view.updateState!=0?Promise.resolve().then(()=>this.runHandlers(e.type,e)):this.runHandlers(e.type,e))}runHandlers(e,A){let i=this.handlers[e];if(i){for(let n of i.observers)n(this.view,A);for(let n of i.handlers){if(A.defaultPrevented)break;if(n(this.view,A)){A.preventDefault();break}}}}ensureHandlers(e){let A=QxA(e),i=this.handlers,n=this.view.contentDOM;for(let o in A)if(o!="scroll"){let r=!A[o].handlers.length,s=i[o];s&&r!=!s.handlers.length&&(n.removeEventListener(o,this.handleEvent),s=null),s||n.addEventListener(o,this.handleEvent,{passive:r})}for(let o in i)o!="scroll"&&!A[o]&&n.removeEventListener(o,this.handleEvent);this.handlers=A}keydown(e){if(this.lastKeyCode=e.keyCode,this.lastKeyTime=Date.now(),e.keyCode==9&&this.tabFocusMode>-1&&(!this.tabFocusMode||Date.now()<=this.tabFocusMode))return!0;if(this.tabFocusMode>0&&e.keyCode!=27&&KtA.indexOf(e.keyCode)<0&&(this.tabFocusMode=-1),At.android&&At.chrome&&!e.synthetic&&(e.keyCode==13||e.keyCode==8))return this.view.observer.delayAndroidKey(e.key,e.keyCode),!0;let A;return At.ios&&!e.synthetic&&!e.altKey&&!e.metaKey&&((A=UtA.find(i=>i.keyCode==e.keyCode))&&!e.ctrlKey||hxA.indexOf(e.key)>-1&&e.ctrlKey&&!e.shiftKey)?(this.pendingIOSKey=A||e,setTimeout(()=>this.flushIOSKey(),250),!0):(e.keyCode!=229&&this.view.observer.forceFlush(),!1)}flushIOSKey(e){let A=this.pendingIOSKey;return!A||A.key=="Enter"&&e&&e.from<e.to&&/^\S+$/.test(e.insert.toString())?!1:(this.pendingIOSKey=void 0,FE(this.view.contentDOM,A.key,A.keyCode,A instanceof KeyboardEvent?A:void 0))}ignoreDuringComposition(e){return/^key/.test(e.type)?this.composing>0?!0:At.safari&&!At.ios&&this.compositionPendingKey&&Date.now()-this.compositionEndedAt<100?(this.compositionPendingKey=!1,!0):!1:!1}startMouseSelection(e){this.mouseSelection&&this.mouseSelection.destroy(),this.mouseSelection=e}update(e){this.view.observer.update(e),this.mouseSelection&&this.mouseSelection.update(e),this.draggedContent&&e.docChanged&&(this.draggedContent=this.draggedContent.map(e.changes)),e.transactions.length&&(this.lastKeyCode=this.lastSelectionTime=0)}destroy(){this.mouseSelection&&this.mouseSelection.destroy()}};function keA(t,e){return(A,i)=>{try{return e.call(t,i,A)}catch(n){Xr(A.state,n)}}}function QxA(t){let e=Object.create(null);function A(i){return e[i]||(e[i]={observers:[],handlers:[]})}for(let i of t){let n=i.spec,o=n&&n.plugin.domEventHandlers,r=n&&n.plugin.domEventObservers;if(o)for(let s in o){let a=o[s];a&&A(s).handlers.push(keA(i.value,a))}if(r)for(let s in r){let a=r[s];a&&A(s).observers.push(keA(i.value,a))}}for(let i in yl)A(i).handlers.push(yl[i]);for(let i in Jc)A(i).observers.push(Jc[i]);return e}var UtA=[{key:"Backspace",keyCode:8,inputType:"deleteContentBackward"},{key:"Enter",keyCode:13,inputType:"insertParagraph"},{key:"Enter",keyCode:13,inputType:"insertLineBreak"},{key:"Delete",keyCode:46,inputType:"deleteContentForward"}],hxA="dthko",KtA=[16,17,18,20,91,92,224,225],Qw=6;function hw(t){return Math.max(0,t)*.7+8}function uxA(t,e){return Math.max(Math.abs(t.clientX-e.clientX),Math.abs(t.clientY-e.clientY))}var GL=class{constructor(e,A,i,n){this.view=e,this.startEvent=A,this.style=i,this.mustSelect=n,this.scrollSpeed={x:0,y:0},this.scrolling=-1,this.lastEvent=A,this.scrollParents=xRA(e.contentDOM),this.atoms=e.state.facet(gF).map(r=>r(e));let o=e.contentDOM.ownerDocument;o.addEventListener("mousemove",this.move=this.move.bind(this)),o.addEventListener("mouseup",this.up=this.up.bind(this)),this.extend=A.shiftKey,this.multiple=e.state.facet(hr.allowMultipleSelections)&&fxA(e,A),this.dragging=pxA(e,A)&&TtA(A)==1?null:!1}start(e){this.dragging===!1&&this.select(e)}move(e){if(e.buttons==0)return this.destroy();if(this.dragging||this.dragging==null&&uxA(this.startEvent,e)<10)return;this.select(this.lastEvent=e);let A=0,i=0,n=0,o=0,r=this.view.win.innerWidth,s=this.view.win.innerHeight;this.scrollParents.x&&({left:n,right:r}=this.scrollParents.x.getBoundingClientRect()),this.scrollParents.y&&({top:o,bottom:s}=this.scrollParents.y.getBoundingClientRect());let a=IF(this.view);e.clientX-a.left<=n+Qw?A=-hw(n-e.clientX):e.clientX+a.right>=r-Qw&&(A=hw(e.clientX-r)),e.clientY-a.top<=o+Qw?i=-hw(o-e.clientY):e.clientY+a.bottom>=s-Qw&&(i=hw(e.clientY-s)),this.setScrollSpeed(A,i)}up(e){this.dragging==null&&this.select(this.lastEvent),this.dragging||e.preventDefault(),this.destroy()}destroy(){this.setScrollSpeed(0,0);let e=this.view.contentDOM.ownerDocument;e.removeEventListener("mousemove",this.move),e.removeEventListener("mouseup",this.up),this.view.inputState.mouseSelection=this.view.inputState.draggedContent=null}setScrollSpeed(e,A){this.scrollSpeed={x:e,y:A},e||A?this.scrolling<0&&(this.scrolling=setInterval(()=>this.scroll(),50)):this.scrolling>-1&&(clearInterval(this.scrolling),this.scrolling=-1)}scroll(){let{x:e,y:A}=this.scrollSpeed;e&&this.scrollParents.x&&(this.scrollParents.x.scrollLeft+=e,e=0),A&&this.scrollParents.y&&(this.scrollParents.y.scrollTop+=A,A=0),(e||A)&&this.view.win.scrollBy(e,A),this.dragging===!1&&this.select(this.lastEvent)}skipAtoms(e){let A=null;for(let i=0;i<e.ranges.length;i++){let n=e.ranges[i],o=null;if(n.empty){let r=vw(this.atoms,n.from,0);r!=n.from&&(o=ae.cursor(r,-1))}else{let r=vw(this.atoms,n.from,-1),s=vw(this.atoms,n.to,1);(r!=n.from||s!=n.to)&&(o=ae.range(n.from==n.anchor?r:s,n.from==n.head?r:s))}o&&(A||(A=e.ranges.slice()),A[i]=o)}return A?ae.create(A,e.mainIndex):e}select(e){let{view:A}=this,i=this.skipAtoms(this.style.get(e,this.extend,this.multiple));(this.mustSelect||!i.eq(A.state.selection,this.dragging===!1))&&this.view.dispatch({selection:i,userEvent:"select.pointer"}),this.mustSelect=!1}update(e){e.transactions.some(A=>A.isUserEvent("input.type"))?this.destroy():this.style.update(e)&&setTimeout(()=>this.select(this.lastEvent),20)}};function fxA(t,e){let A=t.state.facet(wtA);return A.length?A[0](e):At.mac?e.metaKey:e.ctrlKey}function mxA(t,e){let A=t.state.facet(DtA);return A.length?A[0](e):At.mac?!e.altKey:!e.ctrlKey}function pxA(t,e){let{main:A}=t.state.selection;if(A.empty)return!1;let i=e3(t.root);if(!i||i.rangeCount==0)return!0;let n=i.getRangeAt(0).getClientRects();for(let o=0;o<n.length;o++){let r=n[o];if(r.left<=e.clientX&&r.right>=e.clientX&&r.top<=e.clientY&&r.bottom>=e.clientY)return!0}return!1}function wxA(t,e){if(!e.bubbles)return!0;if(e.defaultPrevented)return!1;for(let A=e.target,i;A!=t.contentDOM;A=A.parentNode)if(!A||A.nodeType==11||(i=Fo.get(A))&&i.ignoreEvent(e))return!1;return!0}var yl=Object.create(null),Jc=Object.create(null),YtA=At.ie&&At.ie_version<15||At.ios&&At.webkit_version<604;function DxA(t){let e=t.dom.parentNode;if(!e)return;let A=e.appendChild(document.createElement("textarea"));A.style.cssText="position: fixed; left: -10000px; top: 10px",A.focus(),setTimeout(()=>{t.focus(),A.remove(),JtA(t,A.value)},50)}function qw(t,e,A){for(let i of t.facet(e))A=i(A,t);return A}function JtA(t,e){e=qw(t.state,cF,e);let{state:A}=t,i,n=1,o=A.toText(e),r=o.lines==A.selection.ranges.length;if(UL!=null&&A.selection.ranges.every(a=>a.empty)&&UL==o.toString()){let a=-1;i=A.changeByRange(c=>{let l=A.doc.lineAt(c.from);if(l.from==a)return{range:c};a=l.from;let I=A.toText((r?o.line(n++).text:e)+A.lineBreak);return{changes:{from:l.from,insert:I},range:ae.cursor(c.from+I.length)}})}else r?i=A.changeByRange(a=>{let c=o.line(n++);return{changes:{from:a.from,to:a.to,insert:c.text},range:ae.cursor(a.from+c.length)}}):i=A.replaceSelection(o);t.dispatch(i,{userEvent:"input.paste",scrollIntoView:!0})}Jc.scroll=t=>{t.inputState.lastScrollTop=t.scrollDOM.scrollTop,t.inputState.lastScrollLeft=t.scrollDOM.scrollLeft};yl.keydown=(t,e)=>(t.inputState.setSelectionOrigin("select"),e.keyCode==27&&t.inputState.tabFocusMode!=0&&(t.inputState.tabFocusMode=Date.now()+2e3),!1);Jc.touchstart=(t,e)=>{t.inputState.lastTouchTime=Date.now(),t.inputState.setSelectionOrigin("select.pointer")};Jc.touchmove=t=>{t.inputState.setSelectionOrigin("select.pointer")};yl.mousedown=(t,e)=>{if(t.observer.flush(),t.inputState.lastTouchTime>Date.now()-2e3)return!1;let A=null;for(let i of t.state.facet(ytA))if(A=i(t,e),A)break;if(!A&&e.button==0&&(A=bxA(t,e)),A){let i=!t.hasFocus;t.inputState.startMouseSelection(new GL(t,e,A,i)),i&&t.observer.ignore(()=>{otA(t.contentDOM);let o=t.root.activeElement;o&&!o.contains(t.contentDOM)&&o.blur()});let n=t.inputState.mouseSelection;if(n)return n.start(e),n.dragging===!1}return!1};function SeA(t,e,A,i){if(i==1)return ae.cursor(e,A);if(i==2)return nxA(t.state,e,A);{let n=Es.find(t.docView,e),o=t.state.doc.lineAt(n?n.posAtEnd:e),r=n?n.posAtStart:o.from,s=n?n.posAtEnd:o.to;return s<t.state.doc.length&&s==o.to&&s++,ae.range(r,s)}}var ReA=(t,e,A)=>e>=A.top&&e<=A.bottom&&t>=A.left&&t<=A.right;function yxA(t,e,A,i){let n=Es.find(t.docView,e);if(!n)return 1;let o=e-n.posAtStart;if(o==0)return 1;if(o==n.length)return-1;let r=n.coordsAt(o,-1);if(r&&ReA(A,i,r))return-1;let s=n.coordsAt(o,1);return s&&ReA(A,i,s)?1:r&&r.bottom>=i?-1:1}function xeA(t,e){let A=t.posAtCoords({x:e.clientX,y:e.clientY},!1);return{pos:A,bias:yxA(t,A,e.clientX,e.clientY)}}var vxA=At.ie&&At.ie_version<=11,LeA=null,FeA=0,NeA=0;function TtA(t){if(!vxA)return t.detail;let e=LeA,A=NeA;return LeA=t,NeA=Date.now(),FeA=!e||A>Date.now()-400&&Math.abs(e.clientX-t.clientX)<2&&Math.abs(e.clientY-t.clientY)<2?(FeA+1)%3:1}function bxA(t,e){let A=xeA(t,e),i=TtA(e),n=t.state.selection;return{update(o){o.docChanged&&(A.pos=o.changes.mapPos(A.pos),n=n.map(o.changes))},get(o,r,s){let a=xeA(t,o),c,l=SeA(t,a.pos,a.bias,i);if(A.pos!=a.pos&&!r){let I=SeA(t,A.pos,A.bias,i),C=Math.min(I.from,l.from),d=Math.max(I.to,l.to);l=C<l.from?ae.range(C,d):ae.range(d,C)}return r?n.replaceRange(n.main.extend(l.from,l.to)):s&&i==1&&n.ranges.length>1&&(c=MxA(n,a.pos))?c:s?n.addRange(l):ae.create([l])}}}function MxA(t,e){for(let A=0;A<t.ranges.length;A++){let{from:i,to:n}=t.ranges[A];if(i<=e&&n>=e)return ae.create(t.ranges.slice(0,A).concat(t.ranges.slice(A+1)),t.mainIndex==A?0:t.mainIndex-(t.mainIndex>A?1:0))}return null}yl.dragstart=(t,e)=>{let{selection:{main:A}}=t.state;if(e.target.draggable){let n=t.docView.nearest(e.target);if(n&&n.isWidget){let o=n.posAtStart,r=o+n.length;(o>=A.to||r<=A.from)&&(A=ae.range(o,r))}}let{inputState:i}=t;return i.mouseSelection&&(i.mouseSelection.dragging=!0),i.draggedContent=A,e.dataTransfer&&(e.dataTransfer.setData("Text",qw(t.state,lF,t.state.sliceDoc(A.from,A.to))),e.dataTransfer.effectAllowed="copyMove"),!1};yl.dragend=t=>(t.inputState.draggedContent=null,!1);function _eA(t,e,A,i){if(A=qw(t.state,cF,A),!A)return;let n=t.posAtCoords({x:e.clientX,y:e.clientY},!1),{draggedContent:o}=t.inputState,r=i&&o&&mxA(t,e)?{from:o.from,to:o.to}:null,s={from:n,insert:A},a=t.state.changes(r?[r,s]:s);t.focus(),t.dispatch({changes:a,selection:{anchor:a.mapPos(n,-1),head:a.mapPos(n,1)},userEvent:r?"move.drop":"input.drop"}),t.inputState.draggedContent=null}yl.drop=(t,e)=>{if(!e.dataTransfer)return!1;if(t.state.readOnly)return!0;let A=e.dataTransfer.files;if(A&&A.length){let i=Array(A.length),n=0,o=()=>{++n==A.length&&_eA(t,e,i.filter(r=>r!=null).join(t.state.lineBreak),!1)};for(let r=0;r<A.length;r++){let s=new FileReader;s.onerror=o,s.onload=()=>{/[\x00-\x08\x0e-\x1f]{2}/.test(s.result)||(i[r]=s.result),o()},s.readAsText(A[r])}return!0}else{let i=e.dataTransfer.getData("Text");if(i)return _eA(t,e,i,!0),!0}return!1};yl.paste=(t,e)=>{if(t.state.readOnly)return!0;t.observer.flush();let A=YtA?null:e.clipboardData;return A?(JtA(t,A.getData("text/plain")||A.getData("text/uri-list")),!0):(DxA(t),!1)};function kxA(t,e){let A=t.dom.parentNode;if(!A)return;let i=A.appendChild(document.createElement("textarea"));i.style.cssText="position: fixed; left: -10000px; top: 10px",i.value=e,i.focus(),i.selectionEnd=e.length,i.selectionStart=0,setTimeout(()=>{i.remove(),t.focus()},50)}function SxA(t){let e=[],A=[],i=!1;for(let n of t.selection.ranges)n.empty||(e.push(t.sliceDoc(n.from,n.to)),A.push(n));if(!e.length){let n=-1;for(let{from:o}of t.selection.ranges){let r=t.doc.lineAt(o);r.number>n&&(e.push(r.text),A.push({from:r.from,to:Math.min(t.doc.length,r.to+1)})),n=r.number}i=!0}return{text:qw(t,lF,e.join(t.lineBreak)),ranges:A,linewise:i}}var UL=null;yl.copy=yl.cut=(t,e)=>{let{text:A,ranges:i,linewise:n}=SxA(t.state);if(!A&&!n)return!1;UL=n?A:null,e.type=="cut"&&!t.state.readOnly&&t.dispatch({changes:i,scrollIntoView:!0,userEvent:"delete.cut"});let o=YtA?null:e.clipboardData;return o?(o.clearData(),o.setData("text/plain",A),!0):(kxA(t,A),!1)};var HtA=xa.define();function ztA(t,e){let A=[];for(let i of t.facet(MtA)){let n=i(t,e);n&&A.push(n)}return A.length?t.update({effects:A,annotations:HtA.of(!0)}):null}function OtA(t){setTimeout(()=>{let e=t.hasFocus;if(e!=t.inputState.notifiedFocused){let A=ztA(t.state,e);A?t.dispatch(A):t.update([])}},10)}Jc.focus=t=>{t.inputState.lastFocusTime=Date.now(),!t.scrollDOM.scrollTop&&(t.inputState.lastScrollTop||t.inputState.lastScrollLeft)&&(t.scrollDOM.scrollTop=t.inputState.lastScrollTop,t.scrollDOM.scrollLeft=t.inputState.lastScrollLeft),OtA(t)};Jc.blur=t=>{t.observer.clearSelectionRange(),OtA(t)};Jc.compositionstart=Jc.compositionupdate=t=>{t.observer.editContext||(t.inputState.compositionFirstChange==null&&(t.inputState.compositionFirstChange=!0),t.inputState.composing<0&&(t.inputState.composing=0))};Jc.compositionend=t=>{t.observer.editContext||(t.inputState.composing=-1,t.inputState.compositionEndedAt=Date.now(),t.inputState.compositionPendingKey=!0,t.inputState.compositionPendingChange=t.observer.pendingRecords().length>0,t.inputState.compositionFirstChange=null,At.chrome&&At.android?t.observer.flushSoon():t.inputState.compositionPendingChange?Promise.resolve().then(()=>t.observer.flush()):setTimeout(()=>{t.inputState.composing<0&&t.docView.hasComposition&&t.update([])},50))};Jc.contextmenu=t=>{t.inputState.lastContextMenu=Date.now()};yl.beforeinput=(t,e)=>{var A,i;if(e.inputType=="insertReplacementText"&&t.observer.editContext){let o=(A=e.dataTransfer)===null||A===void 0?void 0:A.getData("text/plain"),r=e.getTargetRanges();if(o&&r.length){let s=r[0],a=t.posAtDOM(s.startContainer,s.startOffset),c=t.posAtDOM(s.endContainer,s.endOffset);return CF(t,{from:a,to:c,insert:t.state.toText(o)},null),!0}}let n;if(At.chrome&&At.android&&(n=UtA.find(o=>o.inputType==e.inputType))&&(t.observer.delayAndroidKey(n.key,n.keyCode),n.key=="Backspace"||n.key=="Delete")){let o=((i=window.visualViewport)===null||i===void 0?void 0:i.height)||0;setTimeout(()=>{var r;(((r=window.visualViewport)===null||r===void 0?void 0:r.height)||0)>o+10&&t.hasFocus&&(t.contentDOM.blur(),t.focus())},100)}return At.ios&&e.inputType=="deleteContentForward"&&t.observer.flushSoon(),At.safari&&e.inputType=="insertText"&&t.inputState.composing>=0&&setTimeout(()=>Jc.compositionend(t,e),20),!1};var GeA=new Set;function RxA(t){GeA.has(t)||(GeA.add(t),t.addEventListener("copy",()=>{}),t.addEventListener("cut",()=>{}))}var UeA=["pre-wrap","normal","pre-line","break-spaces"],NE=!1;function KeA(){NE=!1}var KL=class{constructor(e){this.lineWrapping=e,this.doc=In.empty,this.heightSamples={},this.lineHeight=14,this.charWidth=7,this.textHeight=14,this.lineLength=30}heightForGap(e,A){let i=this.doc.lineAt(A).number-this.doc.lineAt(e).number+1;return this.lineWrapping&&(i+=Math.max(0,Math.ceil((A-e-i*this.lineLength*.5)/this.lineLength))),this.lineHeight*i}heightForLine(e){return this.lineWrapping?(1+Math.max(0,Math.ceil((e-this.lineLength)/Math.max(1,this.lineLength-5))))*this.lineHeight:this.lineHeight}setDoc(e){return this.doc=e,this}mustRefreshForWrapping(e){return UeA.indexOf(e)>-1!=this.lineWrapping}mustRefreshForHeights(e){let A=!1;for(let i=0;i<e.length;i++){let n=e[i];n<0?i++:this.heightSamples[Math.floor(n*10)]||(A=!0,this.heightSamples[Math.floor(n*10)]=!0)}return A}refresh(e,A,i,n,o,r){let s=UeA.indexOf(e)>-1,a=Math.round(A)!=Math.round(this.lineHeight)||this.lineWrapping!=s;if(this.lineWrapping=s,this.lineHeight=A,this.charWidth=i,this.textHeight=n,this.lineLength=o,a){this.heightSamples={};for(let c=0;c<r.length;c++){let l=r[c];l<0?c++:this.heightSamples[Math.floor(l*10)]=!0}}return a}},YL=class{constructor(e,A){this.from=e,this.heights=A,this.index=0}get more(){return this.index<this.heights.length}},kg=class t{constructor(e,A,i,n,o){this.from=e,this.length=A,this.top=i,this.height=n,this._content=o}get type(){return typeof this._content=="number"?$s.Text:Array.isArray(this._content)?this._content:this._content.type}get to(){return this.from+this.length}get bottom(){return this.top+this.height}get widget(){return this._content instanceof y1?this._content.widget:null}get widgetLineBreaks(){return typeof this._content=="number"?this._content:0}join(e){let A=(Array.isArray(this._content)?this._content:[this]).concat(Array.isArray(e._content)?e._content:[e]);return new t(this.from,this.length+e.length,this.top,this.height+e.height,A)}},Po=function(t){return t[t.ByPos=0]="ByPos",t[t.ByHeight=1]="ByHeight",t[t.ByPosNoHeight=2]="ByPosNoHeight",t}(Po||(Po={})),bw=.001,gc=class t{constructor(e,A,i=2){this.length=e,this.height=A,this.flags=i}get outdated(){return(this.flags&2)>0}set outdated(e){this.flags=(e?2:0)|this.flags&-3}setHeight(e){this.height!=e&&(Math.abs(this.height-e)>bw&&(NE=!0),this.height=e)}replace(e,A,i){return t.of(i)}decomposeLeft(e,A){A.push(this)}decomposeRight(e,A){A.push(this)}applyChanges(e,A,i,n){let o=this,r=i.doc;for(let s=n.length-1;s>=0;s--){let{fromA:a,toA:c,fromB:l,toB:I}=n[s],C=o.lineAt(a,Po.ByPosNoHeight,i.setDoc(A),0,0),d=C.to>=c?C:o.lineAt(c,Po.ByPosNoHeight,i,0,0);for(I+=d.to-c,c=d.to;s>0&&C.from<=n[s-1].toA;)a=n[s-1].fromA,l=n[s-1].fromB,s--,a<C.from&&(C=o.lineAt(a,Po.ByPosNoHeight,i,0,0));l+=C.from-a,a=C.from;let B=TL.build(i.setDoc(r),e,l,I);o=Gw(o,o.replace(a,c,B))}return o.updateHeight(i,0)}static empty(){return new Yc(0,0)}static of(e){if(e.length==1)return e[0];let A=0,i=e.length,n=0,o=0;for(;;)if(A==i)if(n>o*2){let s=e[A-1];s.break?e.splice(--A,1,s.left,null,s.right):e.splice(--A,1,s.left,s.right),i+=1+s.break,n-=s.size}else if(o>n*2){let s=e[i];s.break?e.splice(i,1,s.left,null,s.right):e.splice(i,1,s.left,s.right),i+=2+s.break,o-=s.size}else break;else if(n<o){let s=e[A++];s&&(n+=s.size)}else{let s=e[--i];s&&(o+=s.size)}let r=0;return e[A-1]==null?(r=1,A--):e[A]==null&&(r=1,i++),new JL(t.of(e.slice(0,A)),r,t.of(e.slice(i)))}};function Gw(t,e){return t==e?t:(t.constructor!=e.constructor&&(NE=!0),e)}gc.prototype.size=1;var Uw=class extends gc{constructor(e,A,i){super(e,A),this.deco=i}blockAt(e,A,i,n){return new kg(n,this.length,i,this.height,this.deco||0)}lineAt(e,A,i,n,o){return this.blockAt(0,i,n,o)}forEachLine(e,A,i,n,o,r){e<=o+this.length&&A>=o&&r(this.blockAt(0,i,n,o))}updateHeight(e,A=0,i=!1,n){return n&&n.from<=A&&n.more&&this.setHeight(n.heights[n.index++]),this.outdated=!1,this}toString(){return`block(${this.length})`}},Yc=class t extends Uw{constructor(e,A){super(e,A,null),this.collapsed=0,this.widgetHeight=0,this.breaks=0}blockAt(e,A,i,n){return new kg(n,this.length,i,this.height,this.breaks)}replace(e,A,i){let n=i[0];return i.length==1&&(n instanceof t||n instanceof w1&&n.flags&4)&&Math.abs(this.length-n.length)<10?(n instanceof w1?n=new t(n.length,this.height):n.height=this.height,this.outdated||(n.outdated=!1),n):gc.of(i)}updateHeight(e,A=0,i=!1,n){return n&&n.from<=A&&n.more?this.setHeight(n.heights[n.index++]):(i||this.outdated)&&this.setHeight(Math.max(this.widgetHeight,e.heightForLine(this.length-this.collapsed))+this.breaks*e.lineHeight),this.outdated=!1,this}toString(){return`line(${this.length}${this.collapsed?-this.collapsed:""}${this.widgetHeight?":"+this.widgetHeight:""})`}},w1=class t extends gc{constructor(e){super(e,0)}heightMetrics(e,A){let i=e.doc.lineAt(A).number,n=e.doc.lineAt(A+this.length).number,o=n-i+1,r,s=0;if(e.lineWrapping){let a=Math.min(this.height,e.lineHeight*o);r=a/o,this.length>o+1&&(s=(this.height-a)/(this.length-o-1))}else r=this.height/o;return{firstLine:i,lastLine:n,perLine:r,perChar:s}}blockAt(e,A,i,n){let{firstLine:o,lastLine:r,perLine:s,perChar:a}=this.heightMetrics(A,n);if(A.lineWrapping){let c=n+(e<A.lineHeight?0:Math.round(Math.max(0,Math.min(1,(e-i)/this.height))*this.length)),l=A.doc.lineAt(c),I=s+l.length*a,C=Math.max(i,e-I/2);return new kg(l.from,l.length,C,I,0)}else{let c=Math.max(0,Math.min(r-o,Math.floor((e-i)/s))),{from:l,length:I}=A.doc.line(o+c);return new kg(l,I,i+s*c,s,0)}}lineAt(e,A,i,n,o){if(A==Po.ByHeight)return this.blockAt(e,i,n,o);if(A==Po.ByPosNoHeight){let{from:d,to:B}=i.doc.lineAt(e);return new kg(d,B-d,0,0,0)}let{firstLine:r,perLine:s,perChar:a}=this.heightMetrics(i,o),c=i.doc.lineAt(e),l=s+c.length*a,I=c.number-r,C=n+s*I+a*(c.from-o-I);return new kg(c.from,c.length,Math.max(n,Math.min(C,n+this.height-l)),l,0)}forEachLine(e,A,i,n,o,r){e=Math.max(e,o),A=Math.min(A,o+this.length);let{firstLine:s,perLine:a,perChar:c}=this.heightMetrics(i,o);for(let l=e,I=n;l<=A;){let C=i.doc.lineAt(l);if(l==e){let B=C.number-s;I+=a*B+c*(e-o-B)}let d=a+c*C.length;r(new kg(C.from,C.length,I,d,0)),I+=d,l=C.to+1}}replace(e,A,i){let n=this.length-A;if(n>0){let o=i[i.length-1];o instanceof t?i[i.length-1]=new t(o.length+n):i.push(null,new t(n-1))}if(e>0){let o=i[0];o instanceof t?i[0]=new t(e+o.length):i.unshift(new t(e-1),null)}return gc.of(i)}decomposeLeft(e,A){A.push(new t(e-1),null)}decomposeRight(e,A){A.push(null,new t(this.length-e-1))}updateHeight(e,A=0,i=!1,n){let o=A+this.length;if(n&&n.from<=A+this.length&&n.more){let r=[],s=Math.max(A,n.from),a=-1;for(n.from>A&&r.push(new t(n.from-A-1).updateHeight(e,A));s<=o&&n.more;){let l=e.doc.lineAt(s).length;r.length&&r.push(null);let I=n.heights[n.index++];a==-1?a=I:Math.abs(I-a)>=bw&&(a=-2);let C=new Yc(l,I);C.outdated=!1,r.push(C),s+=l+1}s<=o&&r.push(null,new t(o-s).updateHeight(e,s));let c=gc.of(r);return(a<0||Math.abs(c.height-this.height)>=bw||Math.abs(a-this.heightMetrics(e,A).perLine)>=bw)&&(NE=!0),Gw(this,c)}else(i||this.outdated)&&(this.setHeight(e.heightForGap(A,A+this.length)),this.outdated=!1);return this}toString(){return`gap(${this.length})`}},JL=class extends gc{constructor(e,A,i){super(e.length+A+i.length,e.height+i.height,A|(e.outdated||i.outdated?2:0)),this.left=e,this.right=i,this.size=e.size+i.size}get break(){return this.flags&1}blockAt(e,A,i,n){let o=i+this.left.height;return e<o?this.left.blockAt(e,A,i,n):this.right.blockAt(e,A,o,n+this.left.length+this.break)}lineAt(e,A,i,n,o){let r=n+this.left.height,s=o+this.left.length+this.break,a=A==Po.ByHeight?e<r:e<s,c=a?this.left.lineAt(e,A,i,n,o):this.right.lineAt(e,A,i,r,s);if(this.break||(a?c.to<s:c.from>s))return c;let l=A==Po.ByPosNoHeight?Po.ByPosNoHeight:Po.ByPos;return a?c.join(this.right.lineAt(s,l,i,r,s)):this.left.lineAt(s,l,i,n,o).join(c)}forEachLine(e,A,i,n,o,r){let s=n+this.left.height,a=o+this.left.length+this.break;if(this.break)e<a&&this.left.forEachLine(e,A,i,n,o,r),A>=a&&this.right.forEachLine(e,A,i,s,a,r);else{let c=this.lineAt(a,Po.ByPos,i,n,o);e<c.from&&this.left.forEachLine(e,c.from-1,i,n,o,r),c.to>=e&&c.from<=A&&r(c),A>c.to&&this.right.forEachLine(c.to+1,A,i,s,a,r)}}replace(e,A,i){let n=this.left.length+this.break;if(A<n)return this.balanced(this.left.replace(e,A,i),this.right);if(e>this.left.length)return this.balanced(this.left,this.right.replace(e-n,A-n,i));let o=[];e>0&&this.decomposeLeft(e,o);let r=o.length;for(let s of i)o.push(s);if(e>0&&YeA(o,r-1),A<this.length){let s=o.length;this.decomposeRight(A,o),YeA(o,s)}return gc.of(o)}decomposeLeft(e,A){let i=this.left.length;if(e<=i)return this.left.decomposeLeft(e,A);A.push(this.left),this.break&&(i++,e>=i&&A.push(null)),e>i&&this.right.decomposeLeft(e-i,A)}decomposeRight(e,A){let i=this.left.length,n=i+this.break;if(e>=n)return this.right.decomposeRight(e-n,A);e<i&&this.left.decomposeRight(e,A),this.break&&e<n&&A.push(null),A.push(this.right)}balanced(e,A){return e.size>2*A.size||A.size>2*e.size?gc.of(this.break?[e,null,A]:[e,A]):(this.left=Gw(this.left,e),this.right=Gw(this.right,A),this.setHeight(e.height+A.height),this.outdated=e.outdated||A.outdated,this.size=e.size+A.size,this.length=e.length+this.break+A.length,this)}updateHeight(e,A=0,i=!1,n){let{left:o,right:r}=this,s=A+o.length+this.break,a=null;return n&&n.from<=A+o.length&&n.more?a=o=o.updateHeight(e,A,i,n):o.updateHeight(e,A,i),n&&n.from<=s+r.length&&n.more?a=r=r.updateHeight(e,s,i,n):r.updateHeight(e,s,i),a?this.balanced(o,r):(this.height=this.left.height+this.right.height,this.outdated=!1,this)}toString(){return this.left+(this.break?" ":"-")+this.right}};function YeA(t,e){let A,i;t[e]==null&&(A=t[e-1])instanceof w1&&(i=t[e+1])instanceof w1&&t.splice(e-1,3,new w1(A.length+1+i.length))}var xxA=5,TL=class t{constructor(e,A){this.pos=e,this.oracle=A,this.nodes=[],this.lineStart=-1,this.lineEnd=-1,this.covering=null,this.writtenTo=e}get isCovered(){return this.covering&&this.nodes[this.nodes.length-1]==this.covering}span(e,A){if(this.lineStart>-1){let i=Math.min(A,this.lineEnd),n=this.nodes[this.nodes.length-1];n instanceof Yc?n.length+=i-this.pos:(i>this.pos||!this.isCovered)&&this.nodes.push(new Yc(i-this.pos,-1)),this.writtenTo=i,A>i&&(this.nodes.push(null),this.writtenTo++,this.lineStart=-1)}this.pos=A}point(e,A,i){if(e<A||i.heightRelevant){let n=i.widget?i.widget.estimatedHeight:0,o=i.widget?i.widget.lineBreaks:0;n<0&&(n=this.oracle.lineHeight);let r=A-e;i.block?this.addBlock(new Uw(r,n,i)):(r||o||n>=xxA)&&this.addLineDeco(n,o,r)}else A>e&&this.span(e,A);this.lineEnd>-1&&this.lineEnd<this.pos&&(this.lineEnd=this.oracle.doc.lineAt(this.pos).to)}enterLine(){if(this.lineStart>-1)return;let{from:e,to:A}=this.oracle.doc.lineAt(this.pos);this.lineStart=e,this.lineEnd=A,this.writtenTo<e&&((this.writtenTo<e-1||this.nodes[this.nodes.length-1]==null)&&this.nodes.push(this.blankContent(this.writtenTo,e-1)),this.nodes.push(null)),this.pos>e&&this.nodes.push(new Yc(this.pos-e,-1)),this.writtenTo=this.pos}blankContent(e,A){let i=new w1(A-e);return this.oracle.doc.lineAt(e).to==A&&(i.flags|=4),i}ensureLine(){this.enterLine();let e=this.nodes.length?this.nodes[this.nodes.length-1]:null;if(e instanceof Yc)return e;let A=new Yc(0,-1);return this.nodes.push(A),A}addBlock(e){this.enterLine();let A=e.deco;A&&A.startSide>0&&!this.isCovered&&this.ensureLine(),this.nodes.push(e),this.writtenTo=this.pos=this.pos+e.length,A&&A.endSide>0&&(this.covering=e)}addLineDeco(e,A,i){let n=this.ensureLine();n.length+=i,n.collapsed+=i,n.widgetHeight=Math.max(n.widgetHeight,e),n.breaks+=A,this.writtenTo=this.pos=this.pos+i}finish(e){let A=this.nodes.length==0?null:this.nodes[this.nodes.length-1];this.lineStart>-1&&!(A instanceof Yc)&&!this.isCovered?this.nodes.push(new Yc(0,-1)):(this.writtenTo<this.pos||A==null)&&this.nodes.push(this.blankContent(this.writtenTo,this.pos));let i=e;for(let n of this.nodes)n instanceof Yc&&n.updateHeight(this.oracle,i),i+=n?n.length:1;return this.nodes}static build(e,A,i,n){let o=new t(i,e);return co.spans(A,i,n,o,0),o.finish(i)}};function LxA(t,e,A){let i=new HL;return co.compare(t,e,A,i,0),i.changes}var HL=class{constructor(){this.changes=[]}compareRange(){}comparePoint(e,A,i,n){(e<A||i&&i.heightRelevant||n&&n.heightRelevant)&&yw(e,A,this.changes,5)}};function FxA(t,e){let A=t.getBoundingClientRect(),i=t.ownerDocument,n=i.defaultView||window,o=Math.max(0,A.left),r=Math.min(n.innerWidth,A.right),s=Math.max(0,A.top),a=Math.min(n.innerHeight,A.bottom);for(let c=t.parentNode;c&&c!=i.body;)if(c.nodeType==1){let l=c,I=window.getComputedStyle(l);if((l.scrollHeight>l.clientHeight||l.scrollWidth>l.clientWidth)&&I.overflow!="visible"){let C=l.getBoundingClientRect();o=Math.max(o,C.left),r=Math.min(r,C.right),s=Math.max(s,C.top),a=Math.min(c==t.parentNode?n.innerHeight:a,C.bottom)}c=I.position=="absolute"||I.position=="fixed"?l.offsetParent:l.parentNode}else if(c.nodeType==11)c=c.host;else break;return{left:o-A.left,right:Math.max(o,r)-A.left,top:s-(A.top+e),bottom:Math.max(s,a)-(A.top+e)}}function NxA(t){let e=t.getBoundingClientRect(),A=t.ownerDocument.defaultView||window;return e.left<A.innerWidth&&e.right>0&&e.top<A.innerHeight&&e.bottom>0}function _xA(t,e){let A=t.getBoundingClientRect();return{left:0,right:A.right-A.left,top:e,bottom:A.bottom-(A.top+e)}}var X4=class{constructor(e,A,i,n){this.from=e,this.to=A,this.size=i,this.displaySize=n}static same(e,A){if(e.length!=A.length)return!1;for(let i=0;i<e.length;i++){let n=e[i],o=A[i];if(n.from!=o.from||n.to!=o.to||n.size!=o.size)return!1}return!0}draw(e,A){return ut.replace({widget:new zL(this.displaySize*(A?e.scaleY:e.scaleX),A)}).range(this.from,this.to)}},zL=class extends Ic{constructor(e,A){super(),this.size=e,this.vertical=A}eq(e){return e.size==this.size&&e.vertical==this.vertical}toDOM(){let e=document.createElement("div");return this.vertical?e.style.height=this.size+"px":(e.style.width=this.size+"px",e.style.height="2px",e.style.display="inline-block"),e}get estimatedHeight(){return this.vertical?this.size:-1}},Kw=class{constructor(e){this.state=e,this.pixelViewport={left:0,right:window.innerWidth,top:0,bottom:0},this.inView=!0,this.paddingTop=0,this.paddingBottom=0,this.contentDOMWidth=0,this.contentDOMHeight=0,this.editorHeight=0,this.editorWidth=0,this.scrollTop=0,this.scrolledToBottom=!1,this.scaleX=1,this.scaleY=1,this.scrollAnchorPos=0,this.scrollAnchorHeight=-1,this.scaler=JeA,this.scrollTarget=null,this.printing=!1,this.mustMeasureContent=!0,this.defaultTextDirection=lo.LTR,this.visibleRanges=[],this.mustEnforceCursorAssoc=!1;let A=e.facet(RL).some(i=>typeof i!="function"&&i.class=="cm-lineWrapping");this.heightOracle=new KL(A),this.stateDeco=e.facet(a3).filter(i=>typeof i!="function"),this.heightMap=gc.empty().applyChanges(this.stateDeco,In.empty,this.heightOracle.setDoc(e.doc),[new Rg(0,0,0,e.doc.length)]);for(let i=0;i<2&&(this.viewport=this.getViewport(0,null),!!this.updateForViewport());i++);this.updateViewportLines(),this.lineGaps=this.ensureLineGaps([]),this.lineGapDeco=ut.set(this.lineGaps.map(i=>i.draw(this,!1))),this.computeVisibleRanges()}updateForViewport(){let e=[this.viewport],{main:A}=this.state.selection;for(let i=0;i<=1;i++){let n=i?A.head:A.anchor;if(!e.some(({from:o,to:r})=>n>=o&&n<=r)){let{from:o,to:r}=this.lineBlockAt(n);e.push(new RE(o,r))}}return this.viewports=e.sort((i,n)=>i.from-n.from),this.updateScaler()}updateScaler(){let e=this.scaler;return this.scaler=this.heightMap.height<=7e6?JeA:new OL(this.heightOracle,this.heightMap,this.viewports),e.eq(this.scaler)?0:2}updateViewportLines(){this.viewportLines=[],this.heightMap.forEachLine(this.viewport.from,this.viewport.to,this.heightOracle.setDoc(this.state.doc),0,0,e=>{this.viewportLines.push(P4(e,this.scaler))})}update(e,A=null){this.state=e.state;let i=this.stateDeco;this.stateDeco=this.state.facet(a3).filter(l=>typeof l!="function");let n=e.changedRanges,o=Rg.extendWithRanges(n,LxA(i,this.stateDeco,e?e.changes:Cs.empty(this.state.doc.length))),r=this.heightMap.height,s=this.scrolledToBottom?null:this.scrollAnchorAt(this.scrollTop);KeA(),this.heightMap=this.heightMap.applyChanges(this.stateDeco,e.startState.doc,this.heightOracle.setDoc(this.state.doc),o),(this.heightMap.height!=r||NE)&&(e.flags|=2),s?(this.scrollAnchorPos=e.changes.mapPos(s.from,-1),this.scrollAnchorHeight=s.top):(this.scrollAnchorPos=-1,this.scrollAnchorHeight=r);let a=o.length?this.mapViewport(this.viewport,e.changes):this.viewport;(A&&(A.range.head<a.from||A.range.head>a.to)||!this.viewportIsAppropriate(a))&&(a=this.getViewport(0,A));let c=a.from!=this.viewport.from||a.to!=this.viewport.to;this.viewport=a,e.flags|=this.updateForViewport(),(c||!e.changes.empty||e.flags&2)&&this.updateViewportLines(),(this.lineGaps.length||this.viewport.to-this.viewport.from>4e3)&&this.updateLineGaps(this.ensureLineGaps(this.mapLineGaps(this.lineGaps,e.changes))),e.flags|=this.computeVisibleRanges(e.changes),A&&(this.scrollTarget=A),!this.mustEnforceCursorAssoc&&e.selectionSet&&e.view.lineWrapping&&e.state.selection.main.empty&&e.state.selection.main.assoc&&!e.state.facet(ktA)&&(this.mustEnforceCursorAssoc=!0)}measure(e){let A=e.contentDOM,i=window.getComputedStyle(A),n=this.heightOracle,o=i.whiteSpace;this.defaultTextDirection=i.direction=="rtl"?lo.RTL:lo.LTR;let r=this.heightOracle.mustRefreshForWrapping(o),s=A.getBoundingClientRect(),a=r||this.mustMeasureContent||this.contentDOMHeight!=s.height;this.contentDOMHeight=s.height,this.mustMeasureContent=!1;let c=0,l=0;if(s.width&&s.height){let{scaleX:R,scaleY:w}=ntA(A,s);(R>.005&&Math.abs(this.scaleX-R)>.005||w>.005&&Math.abs(this.scaleY-w)>.005)&&(this.scaleX=R,this.scaleY=w,c|=16,r=a=!0)}let I=(parseInt(i.paddingTop)||0)*this.scaleY,C=(parseInt(i.paddingBottom)||0)*this.scaleY;(this.paddingTop!=I||this.paddingBottom!=C)&&(this.paddingTop=I,this.paddingBottom=C,c|=18),this.editorWidth!=e.scrollDOM.clientWidth&&(n.lineWrapping&&(a=!0),this.editorWidth=e.scrollDOM.clientWidth,c|=16);let d=e.scrollDOM.scrollTop*this.scaleY;this.scrollTop!=d&&(this.scrollAnchorHeight=-1,this.scrollTop=d),this.scrolledToBottom=stA(e.scrollDOM);let B=(this.printing?_xA:FxA)(A,this.paddingTop),E=B.top-this.pixelViewport.top,h=B.bottom-this.pixelViewport.bottom;this.pixelViewport=B;let u=this.pixelViewport.bottom>this.pixelViewport.top&&this.pixelViewport.right>this.pixelViewport.left;if(u!=this.inView&&(this.inView=u,u&&(a=!0)),!this.inView&&!this.scrollTarget&&!NxA(e.dom))return 0;let D=s.width;if((this.contentDOMWidth!=D||this.editorHeight!=e.scrollDOM.clientHeight)&&(this.contentDOMWidth=s.width,this.editorHeight=e.scrollDOM.clientHeight,c|=16),a){let R=e.docView.measureVisibleLineHeights(this.viewport);if(n.mustRefreshForHeights(R)&&(r=!0),r||n.lineWrapping&&Math.abs(D-this.contentDOMWidth)>n.charWidth){let{lineHeight:w,charWidth:_,textHeight:K}=e.docView.measureTextSize();r=w>0&&n.refresh(o,w,_,K,Math.max(5,D/_),R),r&&(e.docView.minWidth=0,c|=16)}E>0&&h>0?l=Math.max(E,h):E<0&&h<0&&(l=Math.min(E,h)),KeA();for(let w of this.viewports){let _=w.from==this.viewport.from?R:e.docView.measureVisibleLineHeights(w);this.heightMap=(r?gc.empty().applyChanges(this.stateDeco,In.empty,this.heightOracle,[new Rg(0,0,0,e.state.doc.length)]):this.heightMap).updateHeight(n,0,r,new YL(w.from,_))}NE&&(c|=2)}let L=!this.viewportIsAppropriate(this.viewport,l)||this.scrollTarget&&(this.scrollTarget.range.head<this.viewport.from||this.scrollTarget.range.head>this.viewport.to);return L&&(c&2&&(c|=this.updateScaler()),this.viewport=this.getViewport(l,this.scrollTarget),c|=this.updateForViewport()),(c&2||L)&&this.updateViewportLines(),(this.lineGaps.length||this.viewport.to-this.viewport.from>4e3)&&this.updateLineGaps(this.ensureLineGaps(r?[]:this.lineGaps,e)),c|=this.computeVisibleRanges(),this.mustEnforceCursorAssoc&&(this.mustEnforceCursorAssoc=!1,e.docView.enforceCursorAssoc()),c}get visibleTop(){return this.scaler.fromDOM(this.pixelViewport.top)}get visibleBottom(){return this.scaler.fromDOM(this.pixelViewport.bottom)}getViewport(e,A){let i=.5-Math.max(-.5,Math.min(.5,e/1e3/2)),n=this.heightMap,o=this.heightOracle,{visibleTop:r,visibleBottom:s}=this,a=new RE(n.lineAt(r-i*1e3,Po.ByHeight,o,0,0).from,n.lineAt(s+(1-i)*1e3,Po.ByHeight,o,0,0).to);if(A){let{head:c}=A.range;if(c<a.from||c>a.to){let l=Math.min(this.editorHeight,this.pixelViewport.bottom-this.pixelViewport.top),I=n.lineAt(c,Po.ByPos,o,0,0),C;A.y=="center"?C=(I.top+I.bottom)/2-l/2:A.y=="start"||A.y=="nearest"&&c<a.from?C=I.top:C=I.bottom-l,a=new RE(n.lineAt(C-1e3/2,Po.ByHeight,o,0,0).from,n.lineAt(C+l+1e3/2,Po.ByHeight,o,0,0).to)}}return a}mapViewport(e,A){let i=A.mapPos(e.from,-1),n=A.mapPos(e.to,1);return new RE(this.heightMap.lineAt(i,Po.ByPos,this.heightOracle,0,0).from,this.heightMap.lineAt(n,Po.ByPos,this.heightOracle,0,0).to)}viewportIsAppropriate({from:e,to:A},i=0){if(!this.inView)return!0;let{top:n}=this.heightMap.lineAt(e,Po.ByPos,this.heightOracle,0,0),{bottom:o}=this.heightMap.lineAt(A,Po.ByPos,this.heightOracle,0,0),{visibleTop:r,visibleBottom:s}=this;return(e==0||n<=r-Math.max(10,Math.min(-i,250)))&&(A==this.state.doc.length||o>=s+Math.max(10,Math.min(i,250)))&&n>r-2*1e3&&o<s+2*1e3}mapLineGaps(e,A){if(!e.length||A.empty)return e;let i=[];for(let n of e)A.touchesRange(n.from,n.to)||i.push(new X4(A.mapPos(n.from),A.mapPos(n.to),n.size,n.displaySize));return i}ensureLineGaps(e,A){let i=this.heightOracle.lineWrapping,n=i?1e4:2e3,o=n>>1,r=n<<1;if(this.defaultTextDirection!=lo.LTR&&!i)return[];let s=[],a=(l,I,C,d)=>{if(I-l<o)return;let B=this.state.selection.main,E=[B.from];B.empty||E.push(B.to);for(let u of E)if(u>l&&u<I){a(l,u-10,C,d),a(u+10,I,C,d);return}let h=UxA(e,u=>u.from>=C.from&&u.to<=C.to&&Math.abs(u.from-l)<o&&Math.abs(u.to-I)<o&&!E.some(D=>u.from<D&&u.to>D));if(!h){if(I<C.to&&A&&i&&A.visibleRanges.some(L=>L.from<=I&&L.to>=I)){let L=A.moveToLineBoundary(ae.cursor(I),!1,!0).head;L>l&&(I=L)}let u=this.gapSize(C,l,I,d),D=i||u<2e6?u:2e6;h=new X4(l,I,u,D)}s.push(h)},c=l=>{if(l.length<r||l.type!=$s.Text)return;let I=GxA(l.from,l.to,this.stateDeco);if(I.total<r)return;let C=this.scrollTarget?this.scrollTarget.range.head:null,d,B;if(i){let E=n/this.heightOracle.lineLength*this.heightOracle.lineHeight,h,u;if(C!=null){let D=fw(I,C),L=((this.visibleBottom-this.visibleTop)/2+E)/l.height;h=D-L,u=D+L}else h=(this.visibleTop-l.top-E)/l.height,u=(this.visibleBottom-l.top+E)/l.height;d=uw(I,h),B=uw(I,u)}else{let E=I.total*this.heightOracle.charWidth,h=n*this.heightOracle.charWidth,u=0;if(E>2e6)for(let _ of e)_.from>=l.from&&_.from<l.to&&_.size!=_.displaySize&&_.from*this.heightOracle.charWidth+u<this.pixelViewport.left&&(u=_.size-_.displaySize);let D=this.pixelViewport.left+u,L=this.pixelViewport.right+u,R,w;if(C!=null){let _=fw(I,C),K=((L-D)/2+h)/E;R=_-K,w=_+K}else R=(D-h)/E,w=(L+h)/E;d=uw(I,R),B=uw(I,w)}d>l.from&&a(l.from,d,l,I),B<l.to&&a(B,l.to,l,I)};for(let l of this.viewportLines)Array.isArray(l.type)?l.type.forEach(c):c(l);return s}gapSize(e,A,i,n){let o=fw(n,i)-fw(n,A);return this.heightOracle.lineWrapping?e.height*o:n.total*this.heightOracle.charWidth*o}updateLineGaps(e){X4.same(e,this.lineGaps)||(this.lineGaps=e,this.lineGapDeco=ut.set(e.map(A=>A.draw(this,this.heightOracle.lineWrapping))))}computeVisibleRanges(e){let A=this.stateDeco;this.lineGaps.length&&(A=A.concat(this.lineGapDeco));let i=[];co.spans(A,this.viewport.from,this.viewport.to,{span(o,r){i.push({from:o,to:r})},point(){}},20);let n=0;if(i.length!=this.visibleRanges.length)n=12;else for(let o=0;o<i.length&&!(n&8);o++){let r=this.visibleRanges[o],s=i[o];(r.from!=s.from||r.to!=s.to)&&(n|=4,e&&e.mapPos(r.from,-1)==s.from&&e.mapPos(r.to,1)==s.to||(n|=8))}return this.visibleRanges=i,n}lineBlockAt(e){return e>=this.viewport.from&&e<=this.viewport.to&&this.viewportLines.find(A=>A.from<=e&&A.to>=e)||P4(this.heightMap.lineAt(e,Po.ByPos,this.heightOracle,0,0),this.scaler)}lineBlockAtHeight(e){return e>=this.viewportLines[0].top&&e<=this.viewportLines[this.viewportLines.length-1].bottom&&this.viewportLines.find(A=>A.top<=e&&A.bottom>=e)||P4(this.heightMap.lineAt(this.scaler.fromDOM(e),Po.ByHeight,this.heightOracle,0,0),this.scaler)}scrollAnchorAt(e){let A=this.lineBlockAtHeight(e+8);return A.from>=this.viewport.from||this.viewportLines[0].top-e>200?A:this.viewportLines[0]}elementAtHeight(e){return P4(this.heightMap.blockAt(this.scaler.fromDOM(e),this.heightOracle,0,0),this.scaler)}get docHeight(){return this.scaler.toDOM(this.heightMap.height)}get contentHeight(){return this.docHeight+this.paddingTop+this.paddingBottom}},RE=class{constructor(e,A){this.from=e,this.to=A}};function GxA(t,e,A){let i=[],n=t,o=0;return co.spans(A,t,e,{span(){},point(r,s){r>n&&(i.push({from:n,to:r}),o+=r-n),n=s}},20),n<e&&(i.push({from:n,to:e}),o+=e-n),{total:o,ranges:i}}function uw({total:t,ranges:e},A){if(A<=0)return e[0].from;if(A>=1)return e[e.length-1].to;let i=Math.floor(t*A);for(let n=0;;n++){let{from:o,to:r}=e[n],s=r-o;if(i<=s)return o+i;i-=s}}function fw(t,e){let A=0;for(let{from:i,to:n}of t.ranges){if(e<=n){A+=e-i;break}A+=n-i}return A/t.total}function UxA(t,e){for(let A of t)if(e(A))return A}var JeA={toDOM(t){return t},fromDOM(t){return t},scale:1,eq(t){return t==this}},OL=class t{constructor(e,A,i){let n=0,o=0,r=0;this.viewports=i.map(({from:s,to:a})=>{let c=A.lineAt(s,Po.ByPos,e,0,0).top,l=A.lineAt(a,Po.ByPos,e,0,0).bottom;return n+=l-c,{from:s,to:a,top:c,bottom:l,domTop:0,domBottom:0}}),this.scale=(7e6-n)/(A.height-n);for(let s of this.viewports)s.domTop=r+(s.top-o)*this.scale,r=s.domBottom=s.domTop+(s.bottom-s.top),o=s.bottom}toDOM(e){for(let A=0,i=0,n=0;;A++){let o=A<this.viewports.length?this.viewports[A]:null;if(!o||e<o.top)return n+(e-i)*this.scale;if(e<=o.bottom)return o.domTop+(e-o.top);i=o.bottom,n=o.domBottom}}fromDOM(e){for(let A=0,i=0,n=0;;A++){let o=A<this.viewports.length?this.viewports[A]:null;if(!o||e<o.domTop)return i+(e-n)/this.scale;if(e<=o.domBottom)return o.top+(e-o.domTop);i=o.bottom,n=o.domBottom}}eq(e){return e instanceof t?this.scale==e.scale&&this.viewports.length==e.viewports.length&&this.viewports.every((A,i)=>A.from==e.viewports[i].from&&A.to==e.viewports[i].to):!1}};function P4(t,e){if(e.scale==1)return t;let A=e.toDOM(t.top),i=e.toDOM(t.bottom);return new kg(t.from,t.length,A,i-A,Array.isArray(t._content)?t._content.map(n=>P4(n,e)):t._content)}var mw=qe.define({combine:t=>t.join(" ")}),EL=qe.define({combine:t=>t.indexOf(!0)>-1}),PL=Kc.newName(),PtA=Kc.newName(),jtA=Kc.newName(),qtA={"&light":"."+PtA,"&dark":"."+jtA};function jL(t,e,A){return new Kc(e,{finish(i){return/&/.test(i)?i.replace(/&\w*/,n=>{if(n=="&")return t;if(!A||!A[n])throw new RangeError(`Unsupported selector: ${n}`);return A[n]}):t+" "+i}})}var KxA=jL("."+PL,{"&":{position:"relative !important",boxSizing:"border-box","&.cm-focused":{outline:"1px dotted #212121"},display:"flex !important",flexDirection:"column"},".cm-scroller":{display:"flex !important",alignItems:"flex-start !important",fontFamily:"monospace",lineHeight:1.4,height:"100%",overflowX:"auto",position:"relative",zIndex:0,overflowAnchor:"none"},".cm-content":{margin:0,flexGrow:2,flexShrink:0,display:"block",whiteSpace:"pre",wordWrap:"normal",boxSizing:"border-box",minHeight:"100%",padding:"4px 0",outline:"none","&[contenteditable=true]":{WebkitUserModify:"read-write-plaintext-only"}},".cm-lineWrapping":{whiteSpace_fallback:"pre-wrap",whiteSpace:"break-spaces",wordBreak:"break-word",overflowWrap:"anywhere",flexShrink:1},"&light .cm-content":{caretColor:"black"},"&dark .cm-content":{caretColor:"white"},".cm-line":{display:"block",padding:"0 2px 0 6px"},".cm-layer":{position:"absolute",left:0,top:0,contain:"size style","& > *":{position:"absolute"}},"&light .cm-selectionBackground":{background:"#d9d9d9"},"&dark .cm-selectionBackground":{background:"#222"},"&light.cm-focused > .cm-scroller > .cm-selectionLayer .cm-selectionBackground":{background:"#d7d4f0"},"&dark.cm-focused > .cm-scroller > .cm-selectionLayer .cm-selectionBackground":{background:"#233"},".cm-cursorLayer":{pointerEvents:"none"},"&.cm-focused > .cm-scroller > .cm-cursorLayer":{animation:"steps(1) cm-blink 1.2s infinite"},"@keyframes cm-blink":{"0%":{},"50%":{opacity:0},"100%":{}},"@keyframes cm-blink2":{"0%":{},"50%":{opacity:0},"100%":{}},".cm-cursor, .cm-dropCursor":{borderLeft:"1.2px solid black",marginLeft:"-0.6px",pointerEvents:"none"},".cm-cursor":{display:"none"},"&dark .cm-cursor":{borderLeftColor:"#ddd"},".cm-dropCursor":{position:"absolute"},"&.cm-focused > .cm-scroller > .cm-cursorLayer .cm-cursor":{display:"block"},".cm-iso":{unicodeBidi:"isolate"},".cm-announced":{position:"fixed",top:"-10000px"},"@media print":{".cm-announced":{display:"none"}},"&light .cm-activeLine":{backgroundColor:"#cceeff44"},"&dark .cm-activeLine":{backgroundColor:"#99eeff33"},"&light .cm-specialChar":{color:"red"},"&dark .cm-specialChar":{color:"#f78"},".cm-gutters":{flexShrink:0,display:"flex",height:"100%",boxSizing:"border-box",zIndex:200},".cm-gutters-before":{insetInlineStart:0},".cm-gutters-after":{insetInlineEnd:0},"&light .cm-gutters":{backgroundColor:"#f5f5f5",color:"#6c6c6c",border:"0px solid #ddd","&.cm-gutters-before":{borderRightWidth:"1px"},"&.cm-gutters-after":{borderLeftWidth:"1px"}},"&dark .cm-gutters":{backgroundColor:"#333338",color:"#ccc"},".cm-gutter":{display:"flex !important",flexDirection:"column",flexShrink:0,boxSizing:"border-box",minHeight:"100%",overflow:"hidden"},".cm-gutterElement":{boxSizing:"border-box"},".cm-lineNumbers .cm-gutterElement":{padding:"0 3px 0 5px",minWidth:"20px",textAlign:"right",whiteSpace:"nowrap"},"&light .cm-activeLineGutter":{backgroundColor:"#e2f2ff"},"&dark .cm-activeLineGutter":{backgroundColor:"#222227"},".cm-panels":{boxSizing:"border-box",position:"sticky",left:0,right:0,zIndex:300},"&light .cm-panels":{backgroundColor:"#f5f5f5",color:"black"},"&light .cm-panels-top":{borderBottom:"1px solid #ddd"},"&light .cm-panels-bottom":{borderTop:"1px solid #ddd"},"&dark .cm-panels":{backgroundColor:"#333338",color:"white"},".cm-dialog":{padding:"2px 19px 4px 6px",position:"relative","& label":{fontSize:"80%"}},".cm-dialog-close":{position:"absolute",top:"3px",right:"4px",backgroundColor:"inherit",border:"none",font:"inherit",fontSize:"14px",padding:"0"},".cm-tab":{display:"inline-block",overflow:"hidden",verticalAlign:"bottom"},".cm-widgetBuffer":{verticalAlign:"text-top",height:"1em",width:0,display:"inline"},".cm-placeholder":{color:"#888",display:"inline-block",verticalAlign:"top",userSelect:"none"},".cm-highlightSpace":{backgroundImage:"radial-gradient(circle at 50% 55%, #aaa 20%, transparent 5%)",backgroundPosition:"center"},".cm-highlightTab":{backgroundImage:`url('data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" width="200" height="20"><path stroke="%23888" stroke-width="1" fill="none" d="M1 10H196L190 5M190 15L196 10M197 4L197 16"/></svg>')`,backgroundSize:"auto 100%",backgroundPosition:"right 90%",backgroundRepeat:"no-repeat"},".cm-trailingSpace":{backgroundColor:"#ff332255"},".cm-button":{verticalAlign:"middle",color:"inherit",fontSize:"70%",padding:".2em 1em",borderRadius:"1px"},"&light .cm-button":{backgroundImage:"linear-gradient(#eff1f5, #d9d9df)",border:"1px solid #888","&:active":{backgroundImage:"linear-gradient(#b4b4b4, #d0d3d6)"}},"&dark .cm-button":{backgroundImage:"linear-gradient(#393939, #111)",border:"1px solid #888","&:active":{backgroundImage:"linear-gradient(#111, #333)"}},".cm-textfield":{verticalAlign:"middle",color:"inherit",fontSize:"70%",border:"1px solid silver",padding:".2em .5em"},"&light .cm-textfield":{backgroundColor:"white"},"&dark .cm-textfield":{border:"1px solid #555",backgroundColor:"inherit"}},qtA),YxA={childList:!0,characterData:!0,subtree:!0,attributes:!0,characterDataOldValue:!0},QL=At.ie&&At.ie_version<=11,qL=class{constructor(e){this.view=e,this.active=!1,this.editContext=null,this.selectionRange=new pL,this.selectionChanged=!1,this.delayedFlush=-1,this.resizeTimeout=-1,this.queue=[],this.delayedAndroidKey=null,this.flushingAndroidKey=-1,this.lastChange=0,this.scrollTargets=[],this.intersection=null,this.resizeScroll=null,this.intersecting=!1,this.gapIntersection=null,this.gaps=[],this.printQuery=null,this.parentCheck=-1,this.dom=e.contentDOM,this.observer=new MutationObserver(A=>{for(let i of A)this.queue.push(i);(At.ie&&At.ie_version<=11||At.ios&&e.composing)&&A.some(i=>i.type=="childList"&&i.removedNodes.length||i.type=="characterData"&&i.oldValue.length>i.target.nodeValue.length)?this.flushSoon():this.flush()}),window.EditContext&&At.android&&e.constructor.EDIT_CONTEXT!==!1&&!(At.chrome&&At.chrome_version<126)&&(this.editContext=new VL(e),e.state.facet(H0)&&(e.contentDOM.editContext=this.editContext.editContext)),QL&&(this.onCharData=A=>{this.queue.push({target:A.target,type:"characterData",oldValue:A.prevValue}),this.flushSoon()}),this.onSelectionChange=this.onSelectionChange.bind(this),this.onResize=this.onResize.bind(this),this.onPrint=this.onPrint.bind(this),this.onScroll=this.onScroll.bind(this),window.matchMedia&&(this.printQuery=window.matchMedia("print")),typeof ResizeObserver=="function"&&(this.resizeScroll=new ResizeObserver(()=>{var A;((A=this.view.docView)===null||A===void 0?void 0:A.lastUpdate)<Date.now()-75&&this.onResize()}),this.resizeScroll.observe(e.scrollDOM)),this.addWindowListeners(this.win=e.win),this.start(),typeof IntersectionObserver=="function"&&(this.intersection=new IntersectionObserver(A=>{this.parentCheck<0&&(this.parentCheck=setTimeout(this.listenForScroll.bind(this),1e3)),A.length>0&&A[A.length-1].intersectionRatio>0!=this.intersecting&&(this.intersecting=!this.intersecting,this.intersecting!=this.view.inView&&this.onScrollChanged(document.createEvent("Event")))},{threshold:[0,.001]}),this.intersection.observe(this.dom),this.gapIntersection=new IntersectionObserver(A=>{A.length>0&&A[A.length-1].intersectionRatio>0&&this.onScrollChanged(document.createEvent("Event"))},{})),this.listenForScroll(),this.readSelectionRange()}onScrollChanged(e){this.view.inputState.runHandlers("scroll",e),this.intersecting&&this.view.measure()}onScroll(e){this.intersecting&&this.flush(!1),this.editContext&&this.view.requestMeasure(this.editContext.measureReq),this.onScrollChanged(e)}onResize(){this.resizeTimeout<0&&(this.resizeTimeout=setTimeout(()=>{this.resizeTimeout=-1,this.view.requestMeasure()},50))}onPrint(e){(e.type=="change"||!e.type)&&!e.matches||(this.view.viewState.printing=!0,this.view.measure(),setTimeout(()=>{this.view.viewState.printing=!1,this.view.requestMeasure()},500))}updateGaps(e){if(this.gapIntersection&&(e.length!=this.gaps.length||this.gaps.some((A,i)=>A!=e[i]))){this.gapIntersection.disconnect();for(let A of e)this.gapIntersection.observe(A);this.gaps=e}}onSelectionChange(e){let A=this.selectionChanged;if(!this.readSelectionRange()||this.delayedAndroidKey)return;let{view:i}=this,n=this.selectionRange;if(i.state.facet(H0)?i.root.activeElement!=this.dom:!Dw(this.dom,n))return;let o=n.anchorNode&&i.docView.nearest(n.anchorNode);if(o&&o.ignoreEvent(e)){A||(this.selectionChanged=!1);return}(At.ie&&At.ie_version<=11||At.android&&At.chrome)&&!i.state.selection.main.empty&&n.focusNode&&q4(n.focusNode,n.focusOffset,n.anchorNode,n.anchorOffset)?this.flushSoon():this.flush(!1)}readSelectionRange(){let{view:e}=this,A=e3(e.root);if(!A)return!1;let i=At.safari&&e.root.nodeType==11&&e.root.activeElement==this.dom&&JxA(this.view,A)||A;if(!i||this.selectionRange.eq(i))return!1;let n=Dw(this.dom,i);return n&&!this.selectionChanged&&e.inputState.lastFocusTime>Date.now()-200&&e.inputState.lastTouchTime<Date.now()-300&&FRA(this.dom,i)?(this.view.inputState.lastFocusTime=0,e.docView.updateSelection(),!1):(this.selectionRange.setRange(i),n&&(this.selectionChanged=!0),!0)}setSelectionRange(e,A){this.selectionRange.set(e.node,e.offset,A.node,A.offset),this.selectionChanged=!1}clearSelectionRange(){this.selectionRange.set(null,0,null,0)}listenForScroll(){this.parentCheck=-1;let e=0,A=null;for(let i=this.dom;i;)if(i.nodeType==1)!A&&e<this.scrollTargets.length&&this.scrollTargets[e]==i?e++:A||(A=this.scrollTargets.slice(0,e)),A&&A.push(i),i=i.assignedSlot||i.parentNode;else if(i.nodeType==11)i=i.host;else break;if(e<this.scrollTargets.length&&!A&&(A=this.scrollTargets.slice(0,e)),A){for(let i of this.scrollTargets)i.removeEventListener("scroll",this.onScroll);for(let i of this.scrollTargets=A)i.addEventListener("scroll",this.onScroll)}}ignore(e){if(!this.active)return e();try{return this.stop(),e()}finally{this.start(),this.clear()}}start(){this.active||(this.observer.observe(this.dom,YxA),QL&&this.dom.addEventListener("DOMCharacterDataModified",this.onCharData),this.active=!0)}stop(){this.active&&(this.active=!1,this.observer.disconnect(),QL&&this.dom.removeEventListener("DOMCharacterDataModified",this.onCharData))}clear(){this.processRecords(),this.queue.length=0,this.selectionChanged=!1}delayAndroidKey(e,A){var i;if(!this.delayedAndroidKey){let n=()=>{let o=this.delayedAndroidKey;o&&(this.clearDelayedAndroidKey(),this.view.inputState.lastKeyCode=o.keyCode,this.view.inputState.lastKeyTime=Date.now(),!this.flush()&&o.force&&FE(this.dom,o.key,o.keyCode))};this.flushingAndroidKey=this.view.win.requestAnimationFrame(n)}(!this.delayedAndroidKey||e=="Enter")&&(this.delayedAndroidKey={key:e,keyCode:A,force:this.lastChange<Date.now()-50||!!(!((i=this.delayedAndroidKey)===null||i===void 0)&&i.force)})}clearDelayedAndroidKey(){this.win.cancelAnimationFrame(this.flushingAndroidKey),this.delayedAndroidKey=null,this.flushingAndroidKey=-1}flushSoon(){this.delayedFlush<0&&(this.delayedFlush=this.view.win.requestAnimationFrame(()=>{this.delayedFlush=-1,this.flush()}))}forceFlush(){this.delayedFlush>=0&&(this.view.win.cancelAnimationFrame(this.delayedFlush),this.delayedFlush=-1),this.flush()}pendingRecords(){for(let e of this.observer.takeRecords())this.queue.push(e);return this.queue}processRecords(){let e=this.pendingRecords();e.length&&(this.queue=[]);let A=-1,i=-1,n=!1;for(let o of e){let r=this.readMutation(o);r&&(r.typeOver&&(n=!0),A==-1?{from:A,to:i}=r:(A=Math.min(r.from,A),i=Math.max(r.to,i)))}return{from:A,to:i,typeOver:n}}readChange(){let{from:e,to:A,typeOver:i}=this.processRecords(),n=this.selectionChanged&&Dw(this.dom,this.selectionRange);if(e<0&&!n)return null;e>-1&&(this.lastChange=Date.now()),this.view.inputState.lastFocusTime=0,this.selectionChanged=!1;let o=new NL(this.view,e,A,i);return this.view.docView.domChanged={newSel:o.newSel?o.newSel.main:null},o}flush(e=!0){if(this.delayedFlush>=0||this.delayedAndroidKey)return!1;e&&this.readSelectionRange();let A=this.readChange();if(!A)return this.view.requestMeasure(),!1;let i=this.view.state,n=GtA(this.view,A);return this.view.state==i&&(A.domChanged||A.newSel&&!A.newSel.main.eq(this.view.state.selection.main))&&this.view.update([]),n}readMutation(e){let A=this.view.docView.nearest(e.target);if(!A||A.ignoreMutation(e))return null;if(A.markDirty(e.type=="attributes"),e.type=="attributes"&&(A.flags|=4),e.type=="childList"){let i=TeA(A,e.previousSibling||e.target.previousSibling,-1),n=TeA(A,e.nextSibling||e.target.nextSibling,1);return{from:i?A.posAfter(i):A.posAtStart,to:n?A.posBefore(n):A.posAtEnd,typeOver:!1}}else return e.type=="characterData"?{from:A.posAtStart,to:A.posAtEnd,typeOver:e.target.nodeValue==e.oldValue}:null}setWindow(e){e!=this.win&&(this.removeWindowListeners(this.win),this.win=e,this.addWindowListeners(this.win))}addWindowListeners(e){e.addEventListener("resize",this.onResize),this.printQuery?this.printQuery.addEventListener?this.printQuery.addEventListener("change",this.onPrint):this.printQuery.addListener(this.onPrint):e.addEventListener("beforeprint",this.onPrint),e.addEventListener("scroll",this.onScroll),e.document.addEventListener("selectionchange",this.onSelectionChange)}removeWindowListeners(e){e.removeEventListener("scroll",this.onScroll),e.removeEventListener("resize",this.onResize),this.printQuery?this.printQuery.removeEventListener?this.printQuery.removeEventListener("change",this.onPrint):this.printQuery.removeListener(this.onPrint):e.removeEventListener("beforeprint",this.onPrint),e.document.removeEventListener("selectionchange",this.onSelectionChange)}update(e){this.editContext&&(this.editContext.update(e),e.startState.facet(H0)!=e.state.facet(H0)&&(e.view.contentDOM.editContext=e.state.facet(H0)?this.editContext.editContext:null))}destroy(){var e,A,i;this.stop(),(e=this.intersection)===null||e===void 0||e.disconnect(),(A=this.gapIntersection)===null||A===void 0||A.disconnect(),(i=this.resizeScroll)===null||i===void 0||i.disconnect();for(let n of this.scrollTargets)n.removeEventListener("scroll",this.onScroll);this.removeWindowListeners(this.win),clearTimeout(this.parentCheck),clearTimeout(this.resizeTimeout),this.win.cancelAnimationFrame(this.delayedFlush),this.win.cancelAnimationFrame(this.flushingAndroidKey),this.editContext&&(this.view.contentDOM.editContext=null,this.editContext.destroy())}};function TeA(t,e,A){for(;e;){let i=Fo.get(e);if(i&&i.parent==t)return i;let n=e.parentNode;e=n!=t.dom?n:A>0?e.nextSibling:e.previousSibling}return null}function HeA(t,e){let A=e.startContainer,i=e.startOffset,n=e.endContainer,o=e.endOffset,r=t.docView.domAtPos(t.state.selection.main.anchor);return q4(r.node,r.offset,n,o)&&([A,i,n,o]=[n,o,A,i]),{anchorNode:A,anchorOffset:i,focusNode:n,focusOffset:o}}function JxA(t,e){if(e.getComposedRanges){let n=e.getComposedRanges(t.root)[0];if(n)return HeA(t,n)}let A=null;function i(n){n.preventDefault(),n.stopImmediatePropagation(),A=n.getTargetRanges()[0]}return t.contentDOM.addEventListener("beforeinput",i,!0),t.dom.ownerDocument.execCommand("indent"),t.contentDOM.removeEventListener("beforeinput",i,!0),A?HeA(t,A):null}var VL=class{constructor(e){this.from=0,this.to=0,this.pendingContextChange=null,this.handlers=Object.create(null),this.composing=null,this.resetRange(e.state);let A=this.editContext=new window.EditContext({text:e.state.doc.sliceString(this.from,this.to),selectionStart:this.toContextPos(Math.max(this.from,Math.min(this.to,e.state.selection.main.anchor))),selectionEnd:this.toContextPos(e.state.selection.main.head)});this.handlers.textupdate=i=>{let n=e.state.selection.main,{anchor:o,head:r}=n,s=this.toEditorPos(i.updateRangeStart),a=this.toEditorPos(i.updateRangeEnd);e.inputState.composing>=0&&!this.composing&&(this.composing={contextBase:i.updateRangeStart,editorBase:s,drifted:!1});let c={from:s,to:a,insert:In.of(i.text.split(`
`))};if(c.from==this.from&&o<this.from?c.from=o:c.to==this.to&&o>this.to&&(c.to=o),c.from==c.to&&!c.insert.length){let l=ae.single(this.toEditorPos(i.selectionStart),this.toEditorPos(i.selectionEnd));l.main.eq(n)||e.dispatch({selection:l,userEvent:"select"});return}if((At.mac||At.android)&&c.from==r-1&&/^\. ?$/.test(i.text)&&e.contentDOM.getAttribute("autocorrect")=="off"&&(c={from:s,to:a,insert:In.of([i.text.replace("."," ")])}),this.pendingContextChange=c,!e.state.readOnly){let l=this.to-this.from+(c.to-c.from+c.insert.length);CF(e,c,ae.single(this.toEditorPos(i.selectionStart,l),this.toEditorPos(i.selectionEnd,l)))}this.pendingContextChange&&(this.revertPending(e.state),this.setSelection(e.state))},this.handlers.characterboundsupdate=i=>{let n=[],o=null;for(let r=this.toEditorPos(i.rangeStart),s=this.toEditorPos(i.rangeEnd);r<s;r++){let a=e.coordsForChar(r);o=a&&new DOMRect(a.left,a.top,a.right-a.left,a.bottom-a.top)||o||new DOMRect,n.push(o)}A.updateCharacterBounds(i.rangeStart,n)},this.handlers.textformatupdate=i=>{let n=[];for(let o of i.getTextFormats()){let r=o.underlineStyle,s=o.underlineThickness;if(r!="None"&&s!="None"){let a=this.toEditorPos(o.rangeStart),c=this.toEditorPos(o.rangeEnd);if(a<c){let l=`text-decoration: underline ${r=="Dashed"?"dashed ":r=="Squiggle"?"wavy ":""}${s=="Thin"?1:2}px`;n.push(ut.mark({attributes:{style:l}}).range(a,c))}}}e.dispatch({effects:RtA.of(ut.set(n))})},this.handlers.compositionstart=()=>{e.inputState.composing<0&&(e.inputState.composing=0,e.inputState.compositionFirstChange=!0)},this.handlers.compositionend=()=>{if(e.inputState.composing=-1,e.inputState.compositionFirstChange=null,this.composing){let{drifted:i}=this.composing;this.composing=null,i&&this.reset(e.state)}};for(let i in this.handlers)A.addEventListener(i,this.handlers[i]);this.measureReq={read:i=>{this.editContext.updateControlBounds(i.contentDOM.getBoundingClientRect());let n=e3(i.root);n&&n.rangeCount&&this.editContext.updateSelectionBounds(n.getRangeAt(0).getBoundingClientRect())}}}applyEdits(e){let A=0,i=!1,n=this.pendingContextChange;return e.changes.iterChanges((o,r,s,a,c)=>{if(i)return;let l=c.length-(r-o);if(n&&r>=n.to)if(n.from==o&&n.to==r&&n.insert.eq(c)){n=this.pendingContextChange=null,A+=l,this.to+=l;return}else n=null,this.revertPending(e.state);if(o+=A,r+=A,r<=this.from)this.from+=l,this.to+=l;else if(o<this.to){if(o<this.from||r>this.to||this.to-this.from+c.length>3e4){i=!0;return}this.editContext.updateText(this.toContextPos(o),this.toContextPos(r),c.toString()),this.to+=l}A+=l}),n&&!i&&this.revertPending(e.state),!i}update(e){let A=this.pendingContextChange,i=e.startState.selection.main;this.composing&&(this.composing.drifted||!e.changes.touchesRange(i.from,i.to)&&e.transactions.some(n=>!n.isUserEvent("input.type")&&n.changes.touchesRange(this.from,this.to)))?(this.composing.drifted=!0,this.composing.editorBase=e.changes.mapPos(this.composing.editorBase)):!this.applyEdits(e)||!this.rangeIsValid(e.state)?(this.pendingContextChange=null,this.reset(e.state)):(e.docChanged||e.selectionSet||A)&&this.setSelection(e.state),(e.geometryChanged||e.docChanged||e.selectionSet)&&e.view.requestMeasure(this.measureReq)}resetRange(e){let{head:A}=e.selection.main;this.from=Math.max(0,A-1e4),this.to=Math.min(e.doc.length,A+1e4)}reset(e){this.resetRange(e),this.editContext.updateText(0,this.editContext.text.length,e.doc.sliceString(this.from,this.to)),this.setSelection(e)}revertPending(e){let A=this.pendingContextChange;this.pendingContextChange=null,this.editContext.updateText(this.toContextPos(A.from),this.toContextPos(A.from+A.insert.length),e.doc.sliceString(A.from,A.to))}setSelection(e){let{main:A}=e.selection,i=this.toContextPos(Math.max(this.from,Math.min(this.to,A.anchor))),n=this.toContextPos(A.head);(this.editContext.selectionStart!=i||this.editContext.selectionEnd!=n)&&this.editContext.updateSelection(i,n)}rangeIsValid(e){let{head:A}=e.selection.main;return!(this.from>0&&A-this.from<500||this.to<e.doc.length&&this.to-A<500||this.to-this.from>1e4*3)}toEditorPos(e,A=this.to-this.from){e=Math.min(e,A);let i=this.composing;return i&&i.drifted?i.editorBase+(e-i.contextBase):e+this.from}toContextPos(e){let A=this.composing;return A&&A.drifted?A.contextBase+(e-A.editorBase):e-this.from}destroy(){for(let e in this.handlers)this.editContext.removeEventListener(e,this.handlers[e])}},qt=(()=>{class t{get state(){return this.viewState.state}get viewport(){return this.viewState.viewport}get visibleRanges(){return this.viewState.visibleRanges}get inView(){return this.viewState.inView}get composing(){return!!this.inputState&&this.inputState.composing>0}get compositionStarted(){return!!this.inputState&&this.inputState.composing>=0}get root(){return this._root}get win(){return this.dom.ownerDocument.defaultView||window}constructor(A={}){var i;this.plugins=[],this.pluginMap=new Map,this.editorAttrs={},this.contentAttrs={},this.bidiCache=[],this.destroyed=!1,this.updateState=2,this.measureScheduled=-1,this.measureRequests=[],this.contentDOM=document.createElement("div"),this.scrollDOM=document.createElement("div"),this.scrollDOM.tabIndex=-1,this.scrollDOM.className="cm-scroller",this.scrollDOM.appendChild(this.contentDOM),this.announceDOM=document.createElement("div"),this.announceDOM.className="cm-announced",this.announceDOM.setAttribute("aria-live","polite"),this.dom=document.createElement("div"),this.dom.appendChild(this.announceDOM),this.dom.appendChild(this.scrollDOM),A.parent&&A.parent.appendChild(this.dom);let{dispatch:n}=A;this.dispatchTransactions=A.dispatchTransactions||n&&(o=>o.forEach(r=>n(r,this)))||(o=>this.update(o)),this.dispatch=this.dispatch.bind(this),this._root=A.root||LRA(A.parent)||document,this.viewState=new Kw(A.state||hr.create(A)),A.scrollTo&&A.scrollTo.is(Ew)&&(this.viewState.scrollTarget=A.scrollTo.value.clip(this.viewState.state)),this.plugins=this.state.facet(SE).map(o=>new W4(o));for(let o of this.plugins)o.update(this);this.observer=new qL(this),this.inputState=new _L(this),this.inputState.ensureHandlers(this.plugins),this.docView=new Nw(this),this.mountStyles(),this.updateAttrs(),this.updateState=0,this.requestMeasure(),!((i=document.fonts)===null||i===void 0)&&i.ready&&document.fonts.ready.then(()=>this.requestMeasure())}dispatch(...A){let i=A.length==1&&A[0]instanceof vg?A:A.length==1&&Array.isArray(A[0])?A[0]:[this.state.update(...A)];this.dispatchTransactions(i,this)}update(A){if(this.updateState!=0)throw new Error("Calls to EditorView.update are not allowed while an update is in progress");let i=!1,n=!1,o,r=this.state;for(let d of A){if(d.startState!=r)throw new RangeError("Trying to update state with a transaction that doesn't start from the previous state.");r=d.state}if(this.destroyed){this.viewState.state=r;return}let s=this.hasFocus,a=0,c=null;A.some(d=>d.annotation(HtA))?(this.inputState.notifiedFocused=s,a=1):s!=this.inputState.notifiedFocused&&(this.inputState.notifiedFocused=s,c=ztA(r,s),c||(a=1));let l=this.observer.delayedAndroidKey,I=null;if(l?(this.observer.clearDelayedAndroidKey(),I=this.observer.readChange(),(I&&!this.state.doc.eq(r.doc)||!this.state.selection.eq(r.selection))&&(I=null)):this.observer.clear(),r.facet(hr.phrases)!=this.state.facet(hr.phrases))return this.setState(r);o=Fw.create(this,r,A),o.flags|=a;let C=this.viewState.scrollTarget;try{this.updateState=2;for(let d of A){if(C&&(C=C.map(d.changes)),d.scrollIntoView){let{main:B}=d.state.selection;C=new Z4(B.empty?B:ae.cursor(B.head,B.head>B.anchor?-1:1))}for(let B of d.effects)B.is(Ew)&&(C=B.value.clip(this.state))}this.viewState.update(o,C),this.bidiCache=Yw.update(this.bidiCache,o.changes),o.empty||(this.updatePlugins(o),this.inputState.update(o)),i=this.docView.update(o),this.state.facet(H4)!=this.styleModules&&this.mountStyles(),n=this.updateAttrs(),this.showAnnouncements(A),this.docView.updateSelection(i,A.some(d=>d.isUserEvent("select.pointer")))}finally{this.updateState=0}if(o.startState.facet(mw)!=o.state.facet(mw)&&(this.viewState.mustMeasureContent=!0),(i||n||C||this.viewState.mustEnforceCursorAssoc||this.viewState.mustMeasureContent)&&this.requestMeasure(),i&&this.docViewUpdate(),!o.empty)for(let d of this.state.facet(CL))try{d(o)}catch(B){Xr(this.state,B,"update listener")}(c||I)&&Promise.resolve().then(()=>{c&&this.state==c.startState&&this.dispatch(c),I&&!GtA(this,I)&&l.force&&FE(this.contentDOM,l.key,l.keyCode)})}setState(A){if(this.updateState!=0)throw new Error("Calls to EditorView.setState are not allowed while an update is in progress");if(this.destroyed){this.viewState.state=A;return}this.updateState=2;let i=this.hasFocus;try{for(let n of this.plugins)n.destroy(this);this.viewState=new Kw(A),this.plugins=A.facet(SE).map(n=>new W4(n)),this.pluginMap.clear();for(let n of this.plugins)n.update(this);this.docView.destroy(),this.docView=new Nw(this),this.inputState.ensureHandlers(this.plugins),this.mountStyles(),this.updateAttrs(),this.bidiCache=[]}finally{this.updateState=0}i&&this.focus(),this.requestMeasure()}updatePlugins(A){let i=A.startState.facet(SE),n=A.state.facet(SE);if(i!=n){let o=[];for(let r of n){let s=i.indexOf(r);if(s<0)o.push(new W4(r));else{let a=this.plugins[s];a.mustUpdate=A,o.push(a)}}for(let r of this.plugins)r.mustUpdate!=A&&r.destroy(this);this.plugins=o,this.pluginMap.clear()}else for(let o of this.plugins)o.mustUpdate=A;for(let o=0;o<this.plugins.length;o++)this.plugins[o].update(this);i!=n&&this.inputState.ensureHandlers(this.plugins)}docViewUpdate(){for(let A of this.plugins){let i=A.value;if(i&&i.docViewUpdate)try{i.docViewUpdate(this)}catch(n){Xr(this.state,n,"doc view update listener")}}}measure(A=!0){if(this.destroyed)return;if(this.measureScheduled>-1&&this.win.cancelAnimationFrame(this.measureScheduled),this.observer.delayedAndroidKey){this.measureScheduled=-1,this.requestMeasure();return}this.measureScheduled=0,A&&this.observer.forceFlush();let i=null,n=this.scrollDOM,o=n.scrollTop*this.scaleY,{scrollAnchorPos:r,scrollAnchorHeight:s}=this.viewState;Math.abs(o-this.viewState.scrollTop)>1&&(s=-1),this.viewState.scrollAnchorHeight=-1;try{for(let a=0;;a++){if(s<0)if(stA(n))r=-1,s=this.viewState.heightMap.height;else{let B=this.viewState.scrollAnchorAt(o);r=B.from,s=B.top}this.updateState=1;let c=this.viewState.measure(this);if(!c&&!this.measureRequests.length&&this.viewState.scrollTarget==null)break;if(a>5){console.warn(this.measureRequests.length?"Measure loop restarted more than 5 times":"Viewport failed to stabilize");break}let l=[];c&4||([this.measureRequests,l]=[l,this.measureRequests]);let I=l.map(B=>{try{return B.read(this)}catch(E){return Xr(this.state,E),zeA}}),C=Fw.create(this,this.state,[]),d=!1;C.flags|=c,i?i.flags|=c:i=C,this.updateState=2,C.empty||(this.updatePlugins(C),this.inputState.update(C),this.updateAttrs(),d=this.docView.update(C),d&&this.docViewUpdate());for(let B=0;B<l.length;B++)if(I[B]!=zeA)try{let E=l[B];E.write&&E.write(I[B],this)}catch(E){Xr(this.state,E)}if(d&&this.docView.updateSelection(!0),!C.viewportChanged&&this.measureRequests.length==0){if(this.viewState.editorHeight)if(this.viewState.scrollTarget){this.docView.scrollIntoView(this.viewState.scrollTarget),this.viewState.scrollTarget=null,s=-1;continue}else{let E=(r<0?this.viewState.heightMap.height:this.viewState.lineBlockAt(r).top)-s;if(E>1||E<-1){o=o+E,n.scrollTop=o/this.scaleY,s=-1;continue}}break}}}finally{this.updateState=0,this.measureScheduled=-1}if(i&&!i.empty)for(let a of this.state.facet(CL))a(i)}get themeClasses(){return PL+" "+(this.state.facet(EL)?jtA:PtA)+" "+this.state.facet(mw)}updateAttrs(){let A=OeA(this,peA,{class:"cm-editor"+(this.hasFocus?" cm-focused ":" ")+this.themeClasses}),i={spellcheck:"false",autocorrect:"off",autocapitalize:"off",writingsuggestions:"false",translate:"no",contenteditable:this.state.facet(H0)?"true":"false",class:"cm-content",style:`${At.tabSize}: ${this.state.tabSize}`,role:"textbox","aria-multiline":"true"};this.state.readOnly&&(i["aria-readonly"]="true"),OeA(this,RL,i);let n=this.observer.ignore(()=>{let o=bL(this.contentDOM,this.contentAttrs,i),r=bL(this.dom,this.editorAttrs,A);return o||r});return this.editorAttrs=A,this.contentAttrs=i,n}showAnnouncements(A){let i=!0;for(let n of A)for(let o of n.effects)if(o.is(t.announce)){i&&(this.announceDOM.textContent=""),i=!1;let r=this.announceDOM.appendChild(document.createElement("div"));r.textContent=o.value}}mountStyles(){this.styleModules=this.state.facet(H4);let A=this.state.facet(t.cspNonce);Kc.mount(this.root,this.styleModules.concat(KxA).reverse(),A?{nonce:A}:void 0)}readMeasured(){if(this.updateState==2)throw new Error("Reading the editor layout isn't allowed during an update");this.updateState==0&&this.measureScheduled>-1&&this.measure(!1)}requestMeasure(A){if(this.measureScheduled<0&&(this.measureScheduled=this.win.requestAnimationFrame(()=>this.measure())),A){if(this.measureRequests.indexOf(A)>-1)return;if(A.key!=null){for(let i=0;i<this.measureRequests.length;i++)if(this.measureRequests[i].key===A.key){this.measureRequests[i]=A;return}}this.measureRequests.push(A)}}plugin(A){let i=this.pluginMap.get(A);return(i===void 0||i&&i.plugin!=A)&&this.pluginMap.set(A,i=this.plugins.find(n=>n.plugin==A)||null),i&&i.update(this).value}get documentTop(){return this.contentDOM.getBoundingClientRect().top+this.viewState.paddingTop}get documentPadding(){return{top:this.viewState.paddingTop,bottom:this.viewState.paddingBottom}}get scaleX(){return this.viewState.scaleX}get scaleY(){return this.viewState.scaleY}elementAtHeight(A){return this.readMeasured(),this.viewState.elementAtHeight(A)}lineBlockAtHeight(A){return this.readMeasured(),this.viewState.lineBlockAtHeight(A)}get viewportLineBlocks(){return this.viewState.viewportLines}lineBlockAt(A){return this.viewState.lineBlockAt(A)}get contentHeight(){return this.viewState.contentHeight}moveByChar(A,i,n){return BL(this,A,MeA(this,A,i,n))}moveByGroup(A,i){return BL(this,A,MeA(this,A,i,n=>lxA(this,A.head,n)))}visualLineSide(A,i){let n=this.bidiSpans(A),o=this.textDirectionAt(A.from),r=n[i?n.length-1:0];return ae.cursor(r.side(i,o)+A.from,r.forward(!i,o)?1:-1)}moveToLineBoundary(A,i,n=!0){return cxA(this,A,i,n)}moveVertically(A,i,n){return BL(this,A,gxA(this,A,i,n))}domAtPos(A){return this.docView.domAtPos(A)}posAtDOM(A,i=0){return this.docView.posFromDOM(A,i)}posAtCoords(A,i=!0){return this.readMeasured(),_tA(this,A,i)}coordsAtPos(A,i=1){this.readMeasured();let n=this.docView.coordsAt(A,i);if(!n||n.left==n.right)return n;let o=this.state.doc.lineAt(A),r=this.bidiSpans(o),s=r[Sg.find(r,A-o.from,-1,i)];return Pw(n,s.dir==lo.LTR==i>0)}coordsForChar(A){return this.readMeasured(),this.docView.coordsForChar(A)}get defaultCharacterWidth(){return this.viewState.heightOracle.charWidth}get defaultLineHeight(){return this.viewState.heightOracle.lineHeight}get textDirection(){return this.viewState.defaultTextDirection}textDirectionAt(A){return!this.state.facet(meA)||A<this.viewport.from||A>this.viewport.to?this.textDirection:(this.readMeasured(),this.docView.textDirectionAt(A))}get lineWrapping(){return this.viewState.heightOracle.lineWrapping}bidiSpans(A){if(A.length>TxA)return mtA(A.length);let i=this.textDirectionAt(A.from),n;for(let r of this.bidiCache)if(r.from==A.from&&r.dir==i&&(r.fresh||ftA(r.isolates,n=weA(this,A))))return r.order;n||(n=weA(this,A));let o=jRA(A.text,i,n);return this.bidiCache.push(new Yw(A.from,A.to,i,n,!0,o)),o}get hasFocus(){var A;return(this.dom.ownerDocument.hasFocus()||At.safari&&((A=this.inputState)===null||A===void 0?void 0:A.lastContextMenu)>Date.now()-3e4)&&this.root.activeElement==this.contentDOM}focus(){this.observer.ignore(()=>{otA(this.contentDOM),this.docView.updateSelection()})}setRoot(A){this._root!=A&&(this._root=A,this.observer.setWindow((A.nodeType==9?A:A.ownerDocument).defaultView||window),this.mountStyles())}destroy(){this.root.activeElement==this.contentDOM&&this.contentDOM.blur();for(let A of this.plugins)A.destroy(this);this.plugins=[],this.inputState.destroy(),this.docView.destroy(),this.dom.remove(),this.observer.destroy(),this.measureScheduled>-1&&this.win.cancelAnimationFrame(this.measureScheduled),this.destroyed=!0}static scrollIntoView(A,i={}){return Ew.of(new Z4(typeof A=="number"?ae.cursor(A):A,i.y,i.x,i.yMargin,i.xMargin))}scrollSnapshot(){let{scrollTop:A,scrollLeft:i}=this.scrollDOM,n=this.viewState.scrollAnchorAt(A);return Ew.of(new Z4(ae.cursor(n.from),"start","start",n.top-A,i,!0))}setTabFocusMode(A){A==null?this.inputState.tabFocusMode=this.inputState.tabFocusMode<0?0:-1:typeof A=="boolean"?this.inputState.tabFocusMode=A?0:-1:this.inputState.tabFocusMode!=0&&(this.inputState.tabFocusMode=Date.now()+A)}static domEventHandlers(A){return go.define(()=>({}),{eventHandlers:A})}static domEventObservers(A){return go.define(()=>({}),{eventObservers:A})}static theme(A,i){let n=Kc.newName(),o=[mw.of(n),H4.of(jL(`.${n}`,A))];return i&&i.dark&&o.push(EL.of(!0)),o}static baseTheme(A){return Dl.lowest(H4.of(jL("."+PL,A,qtA)))}static findFromDOM(A){var i;let n=A.querySelector(".cm-content"),o=n&&Fo.get(n)||Fo.get(A);return((i=o?.rootView)===null||i===void 0?void 0:i.view)||null}}return t.styleModule=H4,t.inputHandler=btA,t.clipboardInputFilter=cF,t.clipboardOutputFilter=lF,t.scrollHandler=StA,t.focusChangeEffect=MtA,t.perLineTextDirection=meA,t.exceptionSink=vtA,t.updateListener=CL,t.editable=H0,t.mouseSelectionStyle=ytA,t.dragMovesSelection=DtA,t.clickAddsSelectionRange=wtA,t.decorations=a3,t.outerDecorations=xtA,t.atomicRanges=gF,t.bidiIsolatedRanges=LtA,t.scrollMargins=FtA,t.darkTheme=EL,t.cspNonce=qe.define({combine:e=>e.length?e[0]:""}),t.contentAttributes=RL,t.editorAttributes=peA,t.lineWrapping=t.contentAttributes.of({class:"cm-lineWrapping"}),t.announce=Pi.define(),t})(),TxA=4096,zeA={},Yw=class t{constructor(e,A,i,n,o,r){this.from=e,this.to=A,this.dir=i,this.isolates=n,this.fresh=o,this.order=r}static update(e,A){if(A.empty&&!e.some(o=>o.fresh))return e;let i=[],n=e.length?e[e.length-1].dir:lo.LTR;for(let o=Math.max(0,e.length-10);o<e.length;o++){let r=e[o];r.dir==n&&!A.touchesRange(r.from,r.to)&&i.push(new t(A.mapPos(r.from,1),A.mapPos(r.to,-1),r.dir,r.isolates,!1,r.order))}return i}};function OeA(t,e,A){for(let i=t.state.facet(e),n=i.length-1;n>=0;n--){let o=i[n],r=typeof o=="function"?o(t):o;r&&vL(r,A)}return A}var HxA=At.mac?"mac":At.windows?"win":At.linux?"linux":"key";function zxA(t,e){let A=t.split(/-(?!$)/),i=A[A.length-1];i=="Space"&&(i=" ");let n,o,r,s;for(let a=0;a<A.length-1;++a){let c=A[a];if(/^(cmd|meta|m)$/i.test(c))s=!0;else if(/^a(lt)?$/i.test(c))n=!0;else if(/^(c|ctrl|control)$/i.test(c))o=!0;else if(/^s(hift)?$/i.test(c))r=!0;else if(/^mod$/i.test(c))e=="mac"?s=!0:o=!0;else throw new Error("Unrecognized modifier name: "+c)}return n&&(i="Alt-"+i),o&&(i="Ctrl-"+i),s&&(i="Meta-"+i),r&&(i="Shift-"+i),i}function pw(t,e,A){return e.altKey&&(t="Alt-"+t),e.ctrlKey&&(t="Ctrl-"+t),e.metaKey&&(t="Meta-"+t),A!==!1&&e.shiftKey&&(t="Shift-"+t),t}var OxA=Dl.default(qt.domEventHandlers({keydown(t,e){return WtA(VtA(e.state),t,e,"editor")}})),_E=qe.define({enables:OxA}),PeA=new WeakMap;function VtA(t){let e=t.facet(_E),A=PeA.get(e);return A||PeA.set(e,A=jxA(e.reduce((i,n)=>i.concat(n),[]))),A}function ZtA(t,e,A){return WtA(VtA(t.state),e,t,A)}var p1=null,PxA=4e3;function jxA(t,e=HxA){let A=Object.create(null),i=Object.create(null),n=(r,s)=>{let a=i[r];if(a==null)i[r]=s;else if(a!=s)throw new Error("Key binding "+r+" is used both as a regular binding and as a multi-stroke prefix")},o=(r,s,a,c,l)=>{var I,C;let d=A[r]||(A[r]=Object.create(null)),B=s.split(/ (?!$)/).map(u=>zxA(u,e));for(let u=1;u<B.length;u++){let D=B.slice(0,u).join(" ");n(D,!0),d[D]||(d[D]={preventDefault:!0,stopPropagation:!1,run:[L=>{let R=p1={view:L,prefix:D,scope:r};return setTimeout(()=>{p1==R&&(p1=null)},PxA),!0}]})}let E=B.join(" ");n(E,!1);let h=d[E]||(d[E]={preventDefault:!1,stopPropagation:!1,run:((C=(I=d._any)===null||I===void 0?void 0:I.run)===null||C===void 0?void 0:C.slice())||[]});a&&h.run.push(a),c&&(h.preventDefault=!0),l&&(h.stopPropagation=!0)};for(let r of t){let s=r.scope?r.scope.split(" "):["editor"];if(r.any)for(let c of s){let l=A[c]||(A[c]=Object.create(null));l._any||(l._any={preventDefault:!1,stopPropagation:!1,run:[]});let{any:I}=r;for(let C in l)l[C].run.push(d=>I(d,ZL))}let a=r[e]||r.key;if(a)for(let c of s)o(c,a,r.run,r.preventDefault,r.stopPropagation),r.shift&&o(c,"Shift-"+a,r.shift,r.preventDefault,r.stopPropagation)}return A}var ZL=null;function WtA(t,e,A,i){ZL=e;let n=geA(e),o=Bs(n,0),r=lc(o)==n.length&&n!=" ",s="",a=!1,c=!1,l=!1;p1&&p1.view==A&&p1.scope==i&&(s=p1.prefix+" ",KtA.indexOf(e.keyCode)<0&&(c=!0,p1=null));let I=new Set,C=h=>{if(h){for(let u of h.run)if(!I.has(u)&&(I.add(u),u(A)))return h.stopPropagation&&(l=!0),!0;h.preventDefault&&(h.stopPropagation&&(l=!0),c=!0)}return!1},d=t[i],B,E;return d&&(C(d[s+pw(n,e,!r)])?a=!0:r&&(e.altKey||e.metaKey||e.ctrlKey)&&!(At.windows&&e.ctrlKey&&e.altKey)&&!(At.mac&&e.altKey&&!e.ctrlKey)&&(B=T0[e.keyCode])&&B!=n?(C(d[s+pw(B,e,!0)])||e.shiftKey&&(E=ME[e.keyCode])!=n&&E!=B&&C(d[s+pw(E,e,!1)]))&&(a=!0):r&&e.shiftKey&&C(d[s+pw(n,e,!0)])&&(a=!0),!a&&C(d._any)&&(a=!0)),c&&(a=!0),a&&l&&e.stopPropagation(),ZL=null,a}var c3=class t{constructor(e,A,i,n,o){this.className=e,this.left=A,this.top=i,this.width=n,this.height=o}draw(){let e=document.createElement("div");return e.className=this.className,this.adjust(e),e}update(e,A){return A.className!=this.className?!1:(this.adjust(e),!0)}adjust(e){e.style.left=this.left+"px",e.style.top=this.top+"px",this.width!=null&&(e.style.width=this.width+"px"),e.style.height=this.height+"px"}eq(e){return this.left==e.left&&this.top==e.top&&this.width==e.width&&this.height==e.height&&this.className==e.className}static forRange(e,A,i){if(i.empty){let n=e.coordsAtPos(i.head,i.assoc||1);if(!n)return[];let o=XtA(e);return[new t(A,n.left-o.left,n.top-o.top,null,n.bottom-n.top)]}else return qxA(e,A,i)}};function XtA(t){let e=t.scrollDOM.getBoundingClientRect();return{left:(t.textDirection==lo.LTR?e.left:e.right-t.scrollDOM.clientWidth*t.scaleX)-t.scrollDOM.scrollLeft*t.scaleX,top:e.top-t.scrollDOM.scrollTop*t.scaleY}}function jeA(t,e,A,i){let n=t.coordsAtPos(e,A*2);if(!n)return i;let o=t.dom.getBoundingClientRect(),r=(n.top+n.bottom)/2,s=t.posAtCoords({x:o.left+1,y:r}),a=t.posAtCoords({x:o.right-1,y:r});return s==null||a==null?i:{from:Math.max(i.from,Math.min(s,a)),to:Math.min(i.to,Math.max(s,a))}}function qxA(t,e,A){if(A.to<=t.viewport.from||A.from>=t.viewport.to)return[];let i=Math.max(A.from,t.viewport.from),n=Math.min(A.to,t.viewport.to),o=t.textDirection==lo.LTR,r=t.contentDOM,s=r.getBoundingClientRect(),a=XtA(t),c=r.querySelector(".cm-line"),l=c&&window.getComputedStyle(c),I=s.left+(l?parseInt(l.paddingLeft)+Math.min(0,parseInt(l.textIndent)):0),C=s.right-(l?parseInt(l.paddingRight):0),d=LL(t,i,1),B=LL(t,n,-1),E=d.type==$s.Text?d:null,h=B.type==$s.Text?B:null;if(E&&(t.lineWrapping||d.widgetLineBreaks)&&(E=jeA(t,i,1,E)),h&&(t.lineWrapping||B.widgetLineBreaks)&&(h=jeA(t,n,-1,h)),E&&h&&E.from==h.from&&E.to==h.to)return D(L(A.from,A.to,E));{let w=E?L(A.from,null,E):R(d,!1),_=h?L(null,A.to,h):R(B,!0),K=[];return(E||d).to<(h||B).from-(E&&h?1:0)||d.widgetLineBreaks>1&&w.bottom+t.defaultLineHeight/2<_.top?K.push(u(I,w.bottom,C,_.top)):w.bottom<_.top&&t.elementAtHeight((w.bottom+_.top)/2).type==$s.Text&&(w.bottom=_.top=(w.bottom+_.top)/2),D(w).concat(K).concat(D(_))}function u(w,_,K,z){return new c3(e,w-a.left,_-a.top,K-w,z-_)}function D({top:w,bottom:_,horizontal:K}){let z=[];for(let U=0;U<K.length;U+=2)z.push(u(K[U],w,K[U+1],_));return z}function L(w,_,K){let z=1e9,U=-1e9,H=[];function q(QA,BA,lA,vA,tA){let cA=t.coordsAtPos(QA,QA==K.to?-2:2),pA=t.coordsAtPos(lA,lA==K.from?2:-2);!cA||!pA||(z=Math.min(cA.top,pA.top,z),U=Math.max(cA.bottom,pA.bottom,U),tA==lo.LTR?H.push(o&&BA?I:cA.left,o&&vA?C:pA.right):H.push(!o&&vA?I:pA.left,!o&&BA?C:cA.right))}let j=w??K.from,gA=_??K.to;for(let QA of t.visibleRanges)if(QA.to>j&&QA.from<gA)for(let BA=Math.max(QA.from,j),lA=Math.min(QA.to,gA);;){let vA=t.state.doc.lineAt(BA);for(let tA of t.bidiSpans(vA)){let cA=tA.from+vA.from,pA=tA.to+vA.from;if(cA>=lA)break;pA>BA&&q(Math.max(cA,BA),w==null&&cA<=j,Math.min(pA,lA),_==null&&pA>=gA,tA.dir)}if(BA=vA.to+1,BA>=lA)break}return H.length==0&&q(j,w==null,gA,_==null,t.textDirection),{top:z,bottom:U,horizontal:H}}function R(w,_){let K=s.top+(_?w.top:w.bottom);return{top:K,bottom:K,horizontal:[]}}}function VxA(t,e){return t.constructor==e.constructor&&t.eq(e)}var WL=class{constructor(e,A){this.view=e,this.layer=A,this.drawn=[],this.scaleX=1,this.scaleY=1,this.measureReq={read:this.measure.bind(this),write:this.draw.bind(this)},this.dom=e.scrollDOM.appendChild(document.createElement("div")),this.dom.classList.add("cm-layer"),A.above&&this.dom.classList.add("cm-layer-above"),A.class&&this.dom.classList.add(A.class),this.scale(),this.dom.setAttribute("aria-hidden","true"),this.setOrder(e.state),e.requestMeasure(this.measureReq),A.mount&&A.mount(this.dom,e)}update(e){e.startState.facet(Mw)!=e.state.facet(Mw)&&this.setOrder(e.state),(this.layer.update(e,this.dom)||e.geometryChanged)&&(this.scale(),e.view.requestMeasure(this.measureReq))}docViewUpdate(e){this.layer.updateOnDocViewUpdate!==!1&&e.requestMeasure(this.measureReq)}setOrder(e){let A=0,i=e.facet(Mw);for(;A<i.length&&i[A]!=this.layer;)A++;this.dom.style.zIndex=String((this.layer.above?150:-1)-A)}measure(){return this.layer.markers(this.view)}scale(){let{scaleX:e,scaleY:A}=this.view;(e!=this.scaleX||A!=this.scaleY)&&(this.scaleX=e,this.scaleY=A,this.dom.style.transform=`scale(${1/e}, ${1/A})`)}draw(e){if(e.length!=this.drawn.length||e.some((A,i)=>!VxA(A,this.drawn[i]))){let A=this.dom.firstChild,i=0;for(let n of e)n.update&&A&&n.constructor&&this.drawn[i].constructor&&n.update(A,this.drawn[i])?(A=A.nextSibling,i++):this.dom.insertBefore(n.draw(),A);for(;A;){let n=A.nextSibling;A.remove(),A=n}this.drawn=e}}destroy(){this.layer.destroy&&this.layer.destroy(this.dom,this.view),this.dom.remove()}},Mw=qe.define();function $tA(t){return[go.define(e=>new WL(e,t)),Mw.of(t)]}var l3=qe.define({combine(t){return Wr(t,{cursorBlinkRate:1200,drawRangeCursor:!0},{cursorBlinkRate:(e,A)=>Math.min(e,A),drawRangeCursor:(e,A)=>e||A})}});function AiA(t={}){return[l3.of(t),ZxA,WxA,XxA,ktA.of(!0)]}function eiA(t){return t.startState.facet(l3)!=t.state.facet(l3)}var ZxA=$tA({above:!0,markers(t){let{state:e}=t,A=e.facet(l3),i=[];for(let n of e.selection.ranges){let o=n==e.selection.main;if(n.empty||A.drawRangeCursor){let r=o?"cm-cursor cm-cursor-primary":"cm-cursor cm-cursor-secondary",s=n.empty?n:ae.cursor(n.head,n.head>n.anchor?-1:1);for(let a of c3.forRange(t,r,s))i.push(a)}}return i},update(t,e){t.transactions.some(i=>i.selection)&&(e.style.animationName=e.style.animationName=="cm-blink"?"cm-blink2":"cm-blink");let A=eiA(t);return A&&qeA(t.state,e),t.docChanged||t.selectionSet||A},mount(t,e){qeA(e.state,t)},class:"cm-cursorLayer"});function qeA(t,e){e.style.animationDuration=t.facet(l3).cursorBlinkRate+"ms"}var WxA=$tA({above:!1,markers(t){return t.state.selection.ranges.map(e=>e.empty?[]:c3.forRange(t,"cm-selectionBackground",e)).reduce((e,A)=>e.concat(A))},update(t,e){return t.docChanged||t.selectionSet||t.viewportChanged||eiA(t)},class:"cm-selectionLayer"}),XxA=Dl.highest(qt.theme({".cm-line":{"& ::selection, &::selection":{backgroundColor:"transparent !important"},caretColor:"transparent !important"},".cm-content":{caretColor:"transparent !important","& :focus":{caretColor:"initial !important","&::selection, & ::selection":{backgroundColor:"Highlight !important"}}}})),tiA=Pi.define({map(t,e){return t==null?null:e.mapPos(t)}}),j4=Ar.define({create(){return null},update(t,e){return t!=null&&(t=e.changes.mapPos(t)),e.effects.reduce((A,i)=>i.is(tiA)?i.value:A,t)}}),$xA=go.fromClass(class{constructor(t){this.view=t,this.cursor=null,this.measureReq={read:this.readPos.bind(this),write:this.drawCursor.bind(this)}}update(t){var e;let A=t.state.field(j4);A==null?this.cursor!=null&&((e=this.cursor)===null||e===void 0||e.remove(),this.cursor=null):(this.cursor||(this.cursor=this.view.scrollDOM.appendChild(document.createElement("div")),this.cursor.className="cm-dropCursor"),(t.startState.field(j4)!=A||t.docChanged||t.geometryChanged)&&this.view.requestMeasure(this.measureReq))}readPos(){let{view:t}=this,e=t.state.field(j4),A=e!=null&&t.coordsAtPos(e);if(!A)return null;let i=t.scrollDOM.getBoundingClientRect();return{left:A.left-i.left+t.scrollDOM.scrollLeft*t.scaleX,top:A.top-i.top+t.scrollDOM.scrollTop*t.scaleY,height:A.bottom-A.top}}drawCursor(t){if(this.cursor){let{scaleX:e,scaleY:A}=this.view;t?(this.cursor.style.left=t.left/e+"px",this.cursor.style.top=t.top/A+"px",this.cursor.style.height=t.height/A+"px"):this.cursor.style.left="-100000px"}}destroy(){this.cursor&&this.cursor.remove()}setDropPos(t){this.view.state.field(j4)!=t&&this.view.dispatch({effects:tiA.of(t)})}},{eventObservers:{dragover(t){this.setDropPos(this.view.posAtCoords({x:t.clientX,y:t.clientY}))},dragleave(t){(t.target==this.view.contentDOM||!this.view.contentDOM.contains(t.relatedTarget))&&this.setDropPos(null)},dragend(){this.setDropPos(null)},drop(){this.setDropPos(null)}}});function iiA(){return[j4,$xA]}function VeA(t,e,A,i,n){e.lastIndex=0;for(let o=t.iterRange(A,i),r=A,s;!o.next().done;r+=o.value.length)if(!o.lineBreak)for(;s=e.exec(o.value);)n(r+s.index,s)}function ALA(t,e){let A=t.visibleRanges;if(A.length==1&&A[0].from==t.viewport.from&&A[0].to==t.viewport.to)return A;let i=[];for(let{from:n,to:o}of A)n=Math.max(t.state.doc.lineAt(n).from,n-e),o=Math.min(t.state.doc.lineAt(o).to,o+e),i.length&&i[i.length-1].to>=n?i[i.length-1].to=o:i.push({from:n,to:o});return i}var XL=class{constructor(e){let{regexp:A,decoration:i,decorate:n,boundary:o,maxLength:r=1e3}=e;if(!A.global)throw new RangeError("The regular expression given to MatchDecorator should have its 'g' flag set");if(this.regexp=A,n)this.addMatch=(s,a,c,l)=>n(l,c,c+s[0].length,s,a);else if(typeof i=="function")this.addMatch=(s,a,c,l)=>{let I=i(s,a,c);I&&l(c,c+s[0].length,I)};else if(i)this.addMatch=(s,a,c,l)=>l(c,c+s[0].length,i);else throw new RangeError("Either 'decorate' or 'decoration' should be provided to MatchDecorator");this.boundary=o,this.maxLength=r}createDeco(e){let A=new ds,i=A.add.bind(A);for(let{from:n,to:o}of ALA(e,this.maxLength))VeA(e.state.doc,this.regexp,n,o,(r,s)=>this.addMatch(s,e,r,i));return A.finish()}updateDeco(e,A){let i=1e9,n=-1;return e.docChanged&&e.changes.iterChanges((o,r,s,a)=>{a>=e.view.viewport.from&&s<=e.view.viewport.to&&(i=Math.min(s,i),n=Math.max(a,n))}),e.viewportMoved||n-i>1e3?this.createDeco(e.view):n>-1?this.updateRange(e.view,A.map(e.changes),i,n):A}updateRange(e,A,i,n){for(let o of e.visibleRanges){let r=Math.max(o.from,i),s=Math.min(o.to,n);if(s>=r){let a=e.state.doc.lineAt(r),c=a.to<s?e.state.doc.lineAt(s):a,l=Math.max(o.from,a.from),I=Math.min(o.to,c.to);if(this.boundary){for(;r>a.from;r--)if(this.boundary.test(a.text[r-1-a.from])){l=r;break}for(;s<c.to;s++)if(this.boundary.test(c.text[s-c.from])){I=s;break}}let C=[],d,B=(E,h,u)=>C.push(u.range(E,h));if(a==c)for(this.regexp.lastIndex=l-a.from;(d=this.regexp.exec(a.text))&&d.index<I-a.from;)this.addMatch(d,e,d.index+a.from,B);else VeA(e.state.doc,this.regexp,l,I,(E,h)=>this.addMatch(h,e,E,B));A=A.update({filterFrom:l,filterTo:I,filter:(E,h)=>E<l||h>I,add:C})}}return A}},$L=/x/.unicode!=null?"gu":"g",eLA=new RegExp(`[\0-\b
-\x7F-\x9F\xAD\u061C\u200B\u200E\u200F\u2028\u2029\u202D\u202E\u2066\u2067\u2069\uFEFF\uFFF9-\uFFFC]`,$L),tLA={0:"null",7:"bell",8:"backspace",10:"newline",11:"vertical tab",13:"carriage return",27:"escape",8203:"zero width space",8204:"zero width non-joiner",8205:"zero width joiner",8206:"left-to-right mark",8207:"right-to-left mark",8232:"line separator",8237:"left-to-right override",8238:"right-to-left override",8294:"left-to-right isolate",8295:"right-to-left isolate",8297:"pop directional isolate",8233:"paragraph separator",65279:"zero width no-break space",65532:"object replacement"},hL=null;function iLA(){var t;if(hL==null&&typeof document<"u"&&document.body){let e=document.body.style;hL=((t=e.tabSize)!==null&&t!==void 0?t:e.MozTabSize)!=null}return hL||!1}var kw=qe.define({combine(t){let e=Wr(t,{render:null,specialChars:eLA,addSpecialChars:null});return(e.replaceTabs=!iLA())&&(e.specialChars=new RegExp("	|"+e.specialChars.source,$L)),e.addSpecialChars&&(e.specialChars=new RegExp(e.specialChars.source+"|"+e.addSpecialChars.source,$L)),e}});function niA(t={}){return[kw.of(t),nLA()]}var ZeA=null;function nLA(){return ZeA||(ZeA=go.fromClass(class{constructor(t){this.view=t,this.decorations=ut.none,this.decorationCache=Object.create(null),this.decorator=this.makeDecorator(t.state.facet(kw)),this.decorations=this.decorator.createDeco(t)}makeDecorator(t){return new XL({regexp:t.specialChars,decoration:(e,A,i)=>{let{doc:n}=A.state,o=Bs(e[0],0);if(o==9){let r=n.lineAt(i),s=A.state.tabSize,a=J0(r.text,s,i-r.from);return ut.replace({widget:new eF((s-a%s)*this.view.defaultCharacterWidth/this.view.scaleX)})}return this.decorationCache[o]||(this.decorationCache[o]=ut.replace({widget:new AF(t,o)}))},boundary:t.replaceTabs?void 0:/[^]/})}update(t){let e=t.state.facet(kw);t.startState.facet(kw)!=e?(this.decorator=this.makeDecorator(e),this.decorations=this.decorator.createDeco(t.view)):this.decorations=this.decorator.updateDeco(t,this.decorations)}},{decorations:t=>t.decorations}))}var oLA="\u2022";function rLA(t){return t>=32?oLA:t==10?"\u2424":String.fromCharCode(9216+t)}var AF=class extends Ic{constructor(e,A){super(),this.options=e,this.code=A}eq(e){return e.code==this.code}toDOM(e){let A=rLA(this.code),i=e.state.phrase("Control character")+" "+(tLA[this.code]||"0x"+this.code.toString(16)),n=this.options.render&&this.options.render(this.code,i,A);if(n)return n;let o=document.createElement("span");return o.textContent=A,o.title=i,o.setAttribute("aria-label",i),o.className="cm-specialChar",o}ignoreEvent(){return!1}},eF=class extends Ic{constructor(e){super(),this.width=e}eq(e){return e.width==this.width}toDOM(){let e=document.createElement("span");return e.textContent="	",e.className="cm-tab",e.style.width=this.width+"px",e}ignoreEvent(){return!1}};function oiA(){return aLA}var sLA=ut.line({class:"cm-activeLine"}),aLA=go.fromClass(class{constructor(t){this.decorations=this.getDeco(t)}update(t){(t.docChanged||t.selectionSet)&&(this.decorations=this.getDeco(t.view))}getDeco(t){let e=-1,A=[];for(let i of t.state.selection.ranges){let n=t.lineBlockAt(i.head);n.from>e&&(A.push(sLA.range(n.from)),e=n.from)}return ut.set(A)}},{decorations:t=>t.decorations});var tF=2e3;function cLA(t,e,A){let i=Math.min(e.line,A.line),n=Math.max(e.line,A.line),o=[];if(e.off>tF||A.off>tF||e.col<0||A.col<0){let r=Math.min(e.off,A.off),s=Math.max(e.off,A.off);for(let a=i;a<=n;a++){let c=t.doc.line(a);c.length<=s&&o.push(ae.range(c.from+r,c.to+s))}}else{let r=Math.min(e.col,A.col),s=Math.max(e.col,A.col);for(let a=i;a<=n;a++){let c=t.doc.line(a),l=Cw(c.text,r,t.tabSize,!0);if(l<0)o.push(ae.cursor(c.to));else{let I=Cw(c.text,s,t.tabSize);o.push(ae.range(c.from+l,c.from+I))}}}return o}function lLA(t,e){let A=t.coordsAtPos(t.viewport.from);return A?Math.round(Math.abs((A.left-e)/t.defaultCharacterWidth)):-1}function WeA(t,e){let A=t.posAtCoords({x:e.clientX,y:e.clientY},!1),i=t.state.doc.lineAt(A),n=A-i.from,o=n>tF?-1:n==i.length?lLA(t,e.clientX):J0(i.text,t.state.tabSize,A-i.from);return{line:i.number,col:o,off:n}}function gLA(t,e){let A=WeA(t,e),i=t.state.selection;return A?{update(n){if(n.docChanged){let o=n.changes.mapPos(n.startState.doc.line(A.line).from),r=n.state.doc.lineAt(o);A={line:r.number,col:A.col,off:Math.min(A.off,r.length)},i=i.map(n.changes)}},get(n,o,r){let s=WeA(t,n);if(!s)return i;let a=cLA(t.state,A,s);return a.length?r?ae.create(a.concat(i.ranges)):ae.create(a):i}}:null}function riA(t){let e=t?.eventFilter||(A=>A.altKey&&A.button==0);return qt.mouseSelectionStyle.of((A,i)=>e(i)?gLA(A,i):null)}var ILA={Alt:[18,t=>!!t.altKey],Control:[17,t=>!!t.ctrlKey],Shift:[16,t=>!!t.shiftKey],Meta:[91,t=>!!t.metaKey]},CLA={style:"cursor: crosshair"};function siA(t={}){let[e,A]=ILA[t.key||"Alt"],i=go.fromClass(class{constructor(n){this.view=n,this.isDown=!1}set(n){this.isDown!=n&&(this.isDown=n,this.view.update([]))}},{eventObservers:{keydown(n){this.set(n.keyCode==e||A(n))},keyup(n){(n.keyCode==e||!A(n))&&this.set(!1)},mousemove(n){this.set(A(n))}}});return[i,qt.contentAttributes.of(n=>{var o;return!((o=n.plugin(i))===null||o===void 0)&&o.isDown?CLA:null})]}var z4="-10000px",Jw=class{constructor(e,A,i,n){this.facet=A,this.createTooltipView=i,this.removeTooltipView=n,this.input=e.state.facet(A),this.tooltips=this.input.filter(r=>r);let o=null;this.tooltipViews=this.tooltips.map(r=>o=i(r,o))}update(e,A){var i;let n=e.state.facet(this.facet),o=n.filter(a=>a);if(n===this.input){for(let a of this.tooltipViews)a.update&&a.update(e);return!1}let r=[],s=A?[]:null;for(let a=0;a<o.length;a++){let c=o[a],l=-1;if(c){for(let I=0;I<this.tooltips.length;I++){let C=this.tooltips[I];C&&C.create==c.create&&(l=I)}if(l<0)r[a]=this.createTooltipView(c,a?r[a-1]:null),s&&(s[a]=!!c.above);else{let I=r[a]=this.tooltipViews[l];s&&(s[a]=A[l]),I.update&&I.update(e)}}}for(let a of this.tooltipViews)r.indexOf(a)<0&&(this.removeTooltipView(a),(i=a.destroy)===null||i===void 0||i.call(a));return A&&(s.forEach((a,c)=>A[c]=a),A.length=s.length),this.input=n,this.tooltips=o,this.tooltipViews=r,!0}};function dLA(t){let e=t.dom.ownerDocument.documentElement;return{top:0,left:0,bottom:e.clientHeight,right:e.clientWidth}}var uL=qe.define({combine:t=>{var e,A,i;return{position:At.ios?"absolute":((e=t.find(n=>n.position))===null||e===void 0?void 0:e.position)||"fixed",parent:((A=t.find(n=>n.parent))===null||A===void 0?void 0:A.parent)||null,tooltipSpace:((i=t.find(n=>n.tooltipSpace))===null||i===void 0?void 0:i.tooltipSpace)||dLA}}}),XeA=new WeakMap,dF=go.fromClass(class{constructor(t){this.view=t,this.above=[],this.inView=!0,this.madeAbsolute=!1,this.lastTransaction=0,this.measureTimeout=-1;let e=t.state.facet(uL);this.position=e.position,this.parent=e.parent,this.classes=t.themeClasses,this.createContainer(),this.measureReq={read:this.readMeasure.bind(this),write:this.writeMeasure.bind(this),key:this},this.resizeObserver=typeof ResizeObserver=="function"?new ResizeObserver(()=>this.measureSoon()):null,this.manager=new Jw(t,GE,(A,i)=>this.createTooltip(A,i),A=>{this.resizeObserver&&this.resizeObserver.unobserve(A.dom),A.dom.remove()}),this.above=this.manager.tooltips.map(A=>!!A.above),this.intersectionObserver=typeof IntersectionObserver=="function"?new IntersectionObserver(A=>{Date.now()>this.lastTransaction-50&&A.length>0&&A[A.length-1].intersectionRatio<1&&this.measureSoon()},{threshold:[1]}):null,this.observeIntersection(),t.win.addEventListener("resize",this.measureSoon=this.measureSoon.bind(this)),this.maybeMeasure()}createContainer(){this.parent?(this.container=document.createElement("div"),this.container.style.position="relative",this.container.className=this.view.themeClasses,this.parent.appendChild(this.container)):this.container=this.view.dom}observeIntersection(){if(this.intersectionObserver){this.intersectionObserver.disconnect();for(let t of this.manager.tooltipViews)this.intersectionObserver.observe(t.dom)}}measureSoon(){this.measureTimeout<0&&(this.measureTimeout=setTimeout(()=>{this.measureTimeout=-1,this.maybeMeasure()},50))}update(t){t.transactions.length&&(this.lastTransaction=Date.now());let e=this.manager.update(t,this.above);e&&this.observeIntersection();let A=e||t.geometryChanged,i=t.state.facet(uL);if(i.position!=this.position&&!this.madeAbsolute){this.position=i.position;for(let n of this.manager.tooltipViews)n.dom.style.position=this.position;A=!0}if(i.parent!=this.parent){this.parent&&this.container.remove(),this.parent=i.parent,this.createContainer();for(let n of this.manager.tooltipViews)this.container.appendChild(n.dom);A=!0}else this.parent&&this.view.themeClasses!=this.classes&&(this.classes=this.container.className=this.view.themeClasses);A&&this.maybeMeasure()}createTooltip(t,e){let A=t.create(this.view),i=e?e.dom:null;if(A.dom.classList.add("cm-tooltip"),t.arrow&&!A.dom.querySelector(".cm-tooltip > .cm-tooltip-arrow")){let n=document.createElement("div");n.className="cm-tooltip-arrow",A.dom.appendChild(n)}return A.dom.style.position=this.position,A.dom.style.top=z4,A.dom.style.left="0px",this.container.insertBefore(A.dom,i),A.mount&&A.mount(this.view),this.resizeObserver&&this.resizeObserver.observe(A.dom),A}destroy(){var t,e,A;this.view.win.removeEventListener("resize",this.measureSoon);for(let i of this.manager.tooltipViews)i.dom.remove(),(t=i.destroy)===null||t===void 0||t.call(i);this.parent&&this.container.remove(),(e=this.resizeObserver)===null||e===void 0||e.disconnect(),(A=this.intersectionObserver)===null||A===void 0||A.disconnect(),clearTimeout(this.measureTimeout)}readMeasure(){let t=1,e=1,A=!1;if(this.position=="fixed"&&this.manager.tooltipViews.length){let{dom:o}=this.manager.tooltipViews[0];if(At.gecko)A=o.offsetParent!=this.container.ownerDocument.body;else if(o.style.top==z4&&o.style.left=="0px"){let r=o.getBoundingClientRect();A=Math.abs(r.top+1e4)>1||Math.abs(r.left)>1}}if(A||this.position=="absolute")if(this.parent){let o=this.parent.getBoundingClientRect();o.width&&o.height&&(t=o.width/this.parent.offsetWidth,e=o.height/this.parent.offsetHeight)}else({scaleX:t,scaleY:e}=this.view.viewState);let i=this.view.scrollDOM.getBoundingClientRect(),n=IF(this.view);return{visible:{left:i.left+n.left,top:i.top+n.top,right:i.right-n.right,bottom:i.bottom-n.bottom},parent:this.parent?this.container.getBoundingClientRect():this.view.dom.getBoundingClientRect(),pos:this.manager.tooltips.map((o,r)=>{let s=this.manager.tooltipViews[r];return s.getCoords?s.getCoords(o.pos):this.view.coordsAtPos(o.pos)}),size:this.manager.tooltipViews.map(({dom:o})=>o.getBoundingClientRect()),space:this.view.state.facet(uL).tooltipSpace(this.view),scaleX:t,scaleY:e,makeAbsolute:A}}writeMeasure(t){var e;if(t.makeAbsolute){this.madeAbsolute=!0,this.position="absolute";for(let s of this.manager.tooltipViews)s.dom.style.position="absolute"}let{visible:A,space:i,scaleX:n,scaleY:o}=t,r=[];for(let s=0;s<this.manager.tooltips.length;s++){let a=this.manager.tooltips[s],c=this.manager.tooltipViews[s],{dom:l}=c,I=t.pos[s],C=t.size[s];if(!I||a.clip!==!1&&(I.bottom<=Math.max(A.top,i.top)||I.top>=Math.min(A.bottom,i.bottom)||I.right<Math.max(A.left,i.left)-.1||I.left>Math.min(A.right,i.right)+.1)){l.style.top=z4;continue}let d=a.arrow?c.dom.querySelector(".cm-tooltip-arrow"):null,B=d?7:0,E=C.right-C.left,h=(e=XeA.get(c))!==null&&e!==void 0?e:C.bottom-C.top,u=c.offset||ELA,D=this.view.textDirection==lo.LTR,L=C.width>i.right-i.left?D?i.left:i.right-C.width:D?Math.max(i.left,Math.min(I.left-(d?14:0)+u.x,i.right-E)):Math.min(Math.max(i.left,I.left-E+(d?14:0)-u.x),i.right-E),R=this.above[s];!a.strictSide&&(R?I.top-h-B-u.y<i.top:I.bottom+h+B+u.y>i.bottom)&&R==i.bottom-I.bottom>I.top-i.top&&(R=this.above[s]=!R);let w=(R?I.top-i.top:i.bottom-I.bottom)-B;if(w<h&&c.resize!==!1){if(w<this.view.defaultLineHeight){l.style.top=z4;continue}XeA.set(c,h),l.style.height=(h=w)/o+"px"}else l.style.height&&(l.style.height="");let _=R?I.top-h-B-u.y:I.bottom+B+u.y,K=L+E;if(c.overlap!==!0)for(let z of r)z.left<K&&z.right>L&&z.top<_+h&&z.bottom>_&&(_=R?z.top-h-2-B:z.bottom+B+2);if(this.position=="absolute"?(l.style.top=(_-t.parent.top)/o+"px",$eA(l,(L-t.parent.left)/n)):(l.style.top=_/o+"px",$eA(l,L/n)),d){let z=I.left+(D?u.x:-u.x)-(L+14-7);d.style.left=z/n+"px"}c.overlap!==!0&&r.push({left:L,top:_,right:K,bottom:_+h}),l.classList.toggle("cm-tooltip-above",R),l.classList.toggle("cm-tooltip-below",!R),c.positioned&&c.positioned(t.space)}}maybeMeasure(){if(this.manager.tooltips.length&&(this.view.inView&&this.view.requestMeasure(this.measureReq),this.inView!=this.view.inView&&(this.inView=this.view.inView,!this.inView)))for(let t of this.manager.tooltipViews)t.dom.style.top=z4}},{eventObservers:{scroll(){this.maybeMeasure()}}});function $eA(t,e){let A=parseInt(t.style.left,10);(isNaN(A)||Math.abs(e-A)>1)&&(t.style.left=e+"px")}var BLA=qt.baseTheme({".cm-tooltip":{zIndex:500,boxSizing:"border-box"},"&light .cm-tooltip":{border:"1px solid #bbb",backgroundColor:"#f5f5f5"},"&light .cm-tooltip-section:not(:first-child)":{borderTop:"1px solid #bbb"},"&dark .cm-tooltip":{backgroundColor:"#333338",color:"white"},".cm-tooltip-arrow":{height:"7px",width:`${7*2}px`,position:"absolute",zIndex:-1,overflow:"hidden","&:before, &:after":{content:"''",position:"absolute",width:0,height:0,borderLeft:"7px solid transparent",borderRight:"7px solid transparent"},".cm-tooltip-above &":{bottom:"-7px","&:before":{borderTop:"7px solid #bbb"},"&:after":{borderTop:"7px solid #f5f5f5",bottom:"1px"}},".cm-tooltip-below &":{top:"-7px","&:before":{borderBottom:"7px solid #bbb"},"&:after":{borderBottom:"7px solid #f5f5f5",top:"1px"}}},"&dark .cm-tooltip .cm-tooltip-arrow":{"&:before":{borderTopColor:"#333338",borderBottomColor:"#333338"},"&:after":{borderTopColor:"transparent",borderBottomColor:"transparent"}}}),ELA={x:0,y:0},GE=qe.define({enables:[dF,BLA]}),Tw=qe.define({combine:t=>t.reduce((e,A)=>e.concat(A),[])}),Hw=class t{static create(e){return new t(e)}constructor(e){this.view=e,this.mounted=!1,this.dom=document.createElement("div"),this.dom.classList.add("cm-tooltip-hover"),this.manager=new Jw(e,Tw,(A,i)=>this.createHostedView(A,i),A=>A.dom.remove())}createHostedView(e,A){let i=e.create(this.view);return i.dom.classList.add("cm-tooltip-section"),this.dom.insertBefore(i.dom,A?A.dom.nextSibling:this.dom.firstChild),this.mounted&&i.mount&&i.mount(this.view),i}mount(e){for(let A of this.manager.tooltipViews)A.mount&&A.mount(e);this.mounted=!0}positioned(e){for(let A of this.manager.tooltipViews)A.positioned&&A.positioned(e)}update(e){this.manager.update(e)}destroy(){var e;for(let A of this.manager.tooltipViews)(e=A.destroy)===null||e===void 0||e.call(A)}passProp(e){let A;for(let i of this.manager.tooltipViews){let n=i[e];if(n!==void 0){if(A===void 0)A=n;else if(A!==n)return}}return A}get offset(){return this.passProp("offset")}get getCoords(){return this.passProp("getCoords")}get overlap(){return this.passProp("overlap")}get resize(){return this.passProp("resize")}},QLA=GE.compute([Tw],t=>{let e=t.facet(Tw);return e.length===0?null:{pos:Math.min(...e.map(A=>A.pos)),end:Math.max(...e.map(A=>{var i;return(i=A.end)!==null&&i!==void 0?i:A.pos})),create:Hw.create,above:e[0].above,arrow:e.some(A=>A.arrow)}}),iF=class{constructor(e,A,i,n,o){this.view=e,this.source=A,this.field=i,this.setHover=n,this.hoverTime=o,this.hoverTimeout=-1,this.restartTimeout=-1,this.pending=null,this.lastMove={x:0,y:0,target:e.dom,time:0},this.checkHover=this.checkHover.bind(this),e.dom.addEventListener("mouseleave",this.mouseleave=this.mouseleave.bind(this)),e.dom.addEventListener("mousemove",this.mousemove=this.mousemove.bind(this))}update(){this.pending&&(this.pending=null,clearTimeout(this.restartTimeout),this.restartTimeout=setTimeout(()=>this.startHover(),20))}get active(){return this.view.state.field(this.field)}checkHover(){if(this.hoverTimeout=-1,this.active.length)return;let e=Date.now()-this.lastMove.time;e<this.hoverTime?this.hoverTimeout=setTimeout(this.checkHover,this.hoverTime-e):this.startHover()}startHover(){clearTimeout(this.restartTimeout);let{view:e,lastMove:A}=this,i=e.docView.nearest(A.target);if(!i)return;let n,o=1;if(i instanceof i3)n=i.posAtStart;else{if(n=e.posAtCoords(A),n==null)return;let s=e.coordsAtPos(n);if(!s||A.y<s.top||A.y>s.bottom||A.x<s.left-e.defaultCharacterWidth||A.x>s.right+e.defaultCharacterWidth)return;let a=e.bidiSpans(e.state.doc.lineAt(n)).find(l=>l.from<=n&&l.to>=n),c=a&&a.dir==lo.RTL?-1:1;o=A.x<s.left?-c:c}let r=this.source(e,n,o);if(r?.then){let s=this.pending={pos:n};r.then(a=>{this.pending==s&&(this.pending=null,a&&!(Array.isArray(a)&&!a.length)&&e.dispatch({effects:this.setHover.of(Array.isArray(a)?a:[a])}))},a=>Xr(e.state,a,"hover tooltip"))}else r&&!(Array.isArray(r)&&!r.length)&&e.dispatch({effects:this.setHover.of(Array.isArray(r)?r:[r])})}get tooltip(){let e=this.view.plugin(dF),A=e?e.manager.tooltips.findIndex(i=>i.create==Hw.create):-1;return A>-1?e.manager.tooltipViews[A]:null}mousemove(e){var A,i;this.lastMove={x:e.clientX,y:e.clientY,target:e.target,time:Date.now()},this.hoverTimeout<0&&(this.hoverTimeout=setTimeout(this.checkHover,this.hoverTime));let{active:n,tooltip:o}=this;if(n.length&&o&&!hLA(o.dom,e)||this.pending){let{pos:r}=n[0]||this.pending,s=(i=(A=n[0])===null||A===void 0?void 0:A.end)!==null&&i!==void 0?i:r;(r==s?this.view.posAtCoords(this.lastMove)!=r:!uLA(this.view,r,s,e.clientX,e.clientY))&&(this.view.dispatch({effects:this.setHover.of([])}),this.pending=null)}}mouseleave(e){clearTimeout(this.hoverTimeout),this.hoverTimeout=-1;let{active:A}=this;if(A.length){let{tooltip:i}=this;i&&i.dom.contains(e.relatedTarget)?this.watchTooltipLeave(i.dom):this.view.dispatch({effects:this.setHover.of([])})}}watchTooltipLeave(e){let A=i=>{e.removeEventListener("mouseleave",A),this.active.length&&!this.view.dom.contains(i.relatedTarget)&&this.view.dispatch({effects:this.setHover.of([])})};e.addEventListener("mouseleave",A)}destroy(){clearTimeout(this.hoverTimeout),this.view.dom.removeEventListener("mouseleave",this.mouseleave),this.view.dom.removeEventListener("mousemove",this.mousemove)}},ww=4;function hLA(t,e){let{left:A,right:i,top:n,bottom:o}=t.getBoundingClientRect(),r;if(r=t.querySelector(".cm-tooltip-arrow")){let s=r.getBoundingClientRect();n=Math.min(s.top,n),o=Math.max(s.bottom,o)}return e.clientX>=A-ww&&e.clientX<=i+ww&&e.clientY>=n-ww&&e.clientY<=o+ww}function uLA(t,e,A,i,n,o){let r=t.scrollDOM.getBoundingClientRect(),s=t.documentTop+t.documentPadding.top+t.contentHeight;if(r.left>i||r.right<i||r.top>n||Math.min(r.bottom,s)<n)return!1;let a=t.posAtCoords({x:i,y:n},!1);return a>=e&&a<=A}function aiA(t,e={}){let A=Pi.define(),i=Ar.define({create(){return[]},update(n,o){if(n.length&&(e.hideOnChange&&(o.docChanged||o.selection)?n=[]:e.hideOn&&(n=n.filter(r=>!e.hideOn(o,r))),o.docChanged)){let r=[];for(let s of n){let a=o.changes.mapPos(s.pos,-1,Is.TrackDel);if(a!=null){let c=Object.assign(Object.create(null),s);c.pos=a,c.end!=null&&(c.end=o.changes.mapPos(c.end)),r.push(c)}}n=r}for(let r of o.effects)r.is(A)&&(n=r.value),r.is(fLA)&&(n=[]);return n},provide:n=>Tw.from(n)});return{active:i,extension:[i,go.define(n=>new iF(n,t,i,A,e.hoverTime||300)),QLA]}}function BF(t,e){let A=t.plugin(dF);if(!A)return null;let i=A.manager.tooltips.indexOf(e);return i<0?null:A.manager.tooltipViews[i]}var fLA=Pi.define();var AtA=qe.define({combine(t){let e,A;for(let i of t)e=e||i.topContainer,A=A||i.bottomContainer;return{topContainer:e,bottomContainer:A}}});function kC(t,e){let A=t.plugin(ciA),i=A?A.specs.indexOf(e):-1;return i>-1?A.panels[i]:null}var ciA=go.fromClass(class{constructor(t){this.input=t.state.facet(MC),this.specs=this.input.filter(A=>A),this.panels=this.specs.map(A=>A(t));let e=t.state.facet(AtA);this.top=new xE(t,!0,e.topContainer),this.bottom=new xE(t,!1,e.bottomContainer),this.top.sync(this.panels.filter(A=>A.top)),this.bottom.sync(this.panels.filter(A=>!A.top));for(let A of this.panels)A.dom.classList.add("cm-panel"),A.mount&&A.mount()}update(t){let e=t.state.facet(AtA);this.top.container!=e.topContainer&&(this.top.sync([]),this.top=new xE(t.view,!0,e.topContainer)),this.bottom.container!=e.bottomContainer&&(this.bottom.sync([]),this.bottom=new xE(t.view,!1,e.bottomContainer)),this.top.syncClasses(),this.bottom.syncClasses();let A=t.state.facet(MC);if(A!=this.input){let i=A.filter(a=>a),n=[],o=[],r=[],s=[];for(let a of i){let c=this.specs.indexOf(a),l;c<0?(l=a(t.view),s.push(l)):(l=this.panels[c],l.update&&l.update(t)),n.push(l),(l.top?o:r).push(l)}this.specs=i,this.panels=n,this.top.sync(o),this.bottom.sync(r);for(let a of s)a.dom.classList.add("cm-panel"),a.mount&&a.mount()}else for(let i of this.panels)i.update&&i.update(t)}destroy(){this.top.sync([]),this.bottom.sync([])}},{provide:t=>qt.scrollMargins.of(e=>{let A=e.plugin(t);return A&&{top:A.top.scrollMargin(),bottom:A.bottom.scrollMargin()}})}),xE=class{constructor(e,A,i){this.view=e,this.top=A,this.container=i,this.dom=void 0,this.classes="",this.panels=[],this.syncClasses()}sync(e){for(let A of this.panels)A.destroy&&e.indexOf(A)<0&&A.destroy();this.panels=e,this.syncDOM()}syncDOM(){if(this.panels.length==0){this.dom&&(this.dom.remove(),this.dom=void 0);return}if(!this.dom){this.dom=document.createElement("div"),this.dom.className=this.top?"cm-panels cm-panels-top":"cm-panels cm-panels-bottom",this.dom.style[this.top?"top":"bottom"]="0";let A=this.container||this.view.dom;A.insertBefore(this.dom,this.top?A.firstChild:null)}let e=this.dom.firstChild;for(let A of this.panels)if(A.dom.parentNode==this.dom){for(;e!=A.dom;)e=etA(e);e=e.nextSibling}else this.dom.insertBefore(A.dom,e);for(;e;)e=etA(e)}scrollMargin(){return!this.dom||this.container?0:Math.max(0,this.top?this.dom.getBoundingClientRect().bottom-Math.max(0,this.view.scrollDOM.getBoundingClientRect().top):Math.min(innerHeight,this.view.scrollDOM.getBoundingClientRect().bottom)-this.dom.getBoundingClientRect().top)}syncClasses(){if(!(!this.container||this.classes==this.view.themeClasses)){for(let e of this.classes.split(" "))e&&this.container.classList.remove(e);for(let e of(this.classes=this.view.themeClasses).split(" "))e&&this.container.classList.add(e)}}};function etA(t){let e=t.nextSibling;return t.remove(),e}var MC=qe.define({enables:ciA});var Fa=class extends wl{compare(e){return this==e||this.constructor==e.constructor&&this.eq(e)}eq(e){return!1}destroy(e){}};Fa.prototype.elementClass="";Fa.prototype.toDOM=void 0;Fa.prototype.mapMode=Is.TrackBefore;Fa.prototype.startSide=Fa.prototype.endSide=-1;Fa.prototype.point=!0;var Sw=qe.define(),mLA=qe.define(),pLA={class:"",renderEmptyElements:!1,elementStyle:"",markers:()=>co.empty,lineMarker:()=>null,widgetMarker:()=>null,lineMarkerChange:null,initialSpacer:null,updateSpacer:null,domEventHandlers:{},side:"before"},$4=qe.define();function Vw(t){return[liA(),$4.of(rA(rA({},pLA),t))]}var nF=qe.define({combine:t=>t.some(e=>e)});function liA(t){let e=[wLA];return t&&t.fixed===!1&&e.push(nF.of(!0)),e}var wLA=go.fromClass(class{constructor(t){this.view=t,this.domAfter=null,this.prevViewport=t.viewport,this.dom=document.createElement("div"),this.dom.className="cm-gutters cm-gutters-before",this.dom.setAttribute("aria-hidden","true"),this.dom.style.minHeight=this.view.contentHeight/this.view.scaleY+"px",this.gutters=t.state.facet($4).map(e=>new zw(t,e)),this.fixed=!t.state.facet(nF);for(let e of this.gutters)e.config.side=="after"?this.getDOMAfter().appendChild(e.dom):this.dom.appendChild(e.dom);this.fixed&&(this.dom.style.position="sticky"),this.syncGutters(!1),t.scrollDOM.insertBefore(this.dom,t.contentDOM)}getDOMAfter(){return this.domAfter||(this.domAfter=document.createElement("div"),this.domAfter.className="cm-gutters cm-gutters-after",this.domAfter.setAttribute("aria-hidden","true"),this.domAfter.style.minHeight=this.view.contentHeight/this.view.scaleY+"px",this.domAfter.style.position=this.fixed?"sticky":"",this.view.scrollDOM.appendChild(this.domAfter)),this.domAfter}update(t){if(this.updateGutters(t)){let e=this.prevViewport,A=t.view.viewport,i=Math.min(e.to,A.to)-Math.max(e.from,A.from);this.syncGutters(i<(A.to-A.from)*.8)}if(t.geometryChanged){let e=this.view.contentHeight/this.view.scaleY+"px";this.dom.style.minHeight=e,this.domAfter&&(this.domAfter.style.minHeight=e)}this.view.state.facet(nF)!=!this.fixed&&(this.fixed=!this.fixed,this.dom.style.position=this.fixed?"sticky":"",this.domAfter&&(this.domAfter.style.position=this.fixed?"sticky":"")),this.prevViewport=t.view.viewport}syncGutters(t){let e=this.dom.nextSibling;t&&(this.dom.remove(),this.domAfter&&this.domAfter.remove());let A=co.iter(this.view.state.facet(Sw),this.view.viewport.from),i=[],n=this.gutters.map(o=>new rF(o,this.view.viewport,-this.view.documentPadding.top));for(let o of this.view.viewportLineBlocks)if(i.length&&(i=[]),Array.isArray(o.type)){let r=!0;for(let s of o.type)if(s.type==$s.Text&&r){oF(A,i,s.from);for(let a of n)a.line(this.view,s,i);r=!1}else if(s.widget)for(let a of n)a.widget(this.view,s)}else if(o.type==$s.Text){oF(A,i,o.from);for(let r of n)r.line(this.view,o,i)}else if(o.widget)for(let r of n)r.widget(this.view,o);for(let o of n)o.finish();t&&(this.view.scrollDOM.insertBefore(this.dom,e),this.domAfter&&this.view.scrollDOM.appendChild(this.domAfter))}updateGutters(t){let e=t.startState.facet($4),A=t.state.facet($4),i=t.docChanged||t.heightChanged||t.viewportChanged||!co.eq(t.startState.facet(Sw),t.state.facet(Sw),t.view.viewport.from,t.view.viewport.to);if(e==A)for(let n of this.gutters)n.update(t)&&(i=!0);else{i=!0;let n=[];for(let o of A){let r=e.indexOf(o);r<0?n.push(new zw(this.view,o)):(this.gutters[r].update(t),n.push(this.gutters[r]))}for(let o of this.gutters)o.dom.remove(),n.indexOf(o)<0&&o.destroy();for(let o of n)o.config.side=="after"?this.getDOMAfter().appendChild(o.dom):this.dom.appendChild(o.dom);this.gutters=n}return i}destroy(){for(let t of this.gutters)t.destroy();this.dom.remove(),this.domAfter&&this.domAfter.remove()}},{provide:t=>qt.scrollMargins.of(e=>{let A=e.plugin(t);if(!A||A.gutters.length==0||!A.fixed)return null;let i=A.dom.offsetWidth*e.scaleX,n=A.domAfter?A.domAfter.offsetWidth*e.scaleX:0;return e.textDirection==lo.LTR?{left:i,right:n}:{right:i,left:n}})});function ttA(t){return Array.isArray(t)?t:[t]}function oF(t,e,A){for(;t.value&&t.from<=A;)t.from==A&&e.push(t.value),t.next()}var rF=class{constructor(e,A,i){this.gutter=e,this.height=i,this.i=0,this.cursor=co.iter(e.markers,A.from)}addElement(e,A,i){let{gutter:n}=this,o=(A.top-this.height)/e.scaleY,r=A.height/e.scaleY;if(this.i==n.elements.length){let s=new Ow(e,r,o,i);n.elements.push(s),n.dom.appendChild(s.dom)}else n.elements[this.i].update(e,r,o,i);this.height=A.bottom,this.i++}line(e,A,i){let n=[];oF(this.cursor,n,A.from),i.length&&(n=n.concat(i));let o=this.gutter.config.lineMarker(e,A,n);o&&n.unshift(o);let r=this.gutter;n.length==0&&!r.config.renderEmptyElements||this.addElement(e,A,n)}widget(e,A){let i=this.gutter.config.widgetMarker(e,A.widget,A),n=i?[i]:null;for(let o of e.state.facet(mLA)){let r=o(e,A.widget,A);r&&(n||(n=[])).push(r)}n&&this.addElement(e,A,n)}finish(){let e=this.gutter;for(;e.elements.length>this.i;){let A=e.elements.pop();e.dom.removeChild(A.dom),A.destroy()}}},zw=class{constructor(e,A){this.view=e,this.config=A,this.elements=[],this.spacer=null,this.dom=document.createElement("div"),this.dom.className="cm-gutter"+(this.config.class?" "+this.config.class:"");for(let i in A.domEventHandlers)this.dom.addEventListener(i,n=>{let o=n.target,r;if(o!=this.dom&&this.dom.contains(o)){for(;o.parentNode!=this.dom;)o=o.parentNode;let a=o.getBoundingClientRect();r=(a.top+a.bottom)/2}else r=n.clientY;let s=e.lineBlockAtHeight(r-e.documentTop);A.domEventHandlers[i](e,s,n)&&n.preventDefault()});this.markers=ttA(A.markers(e)),A.initialSpacer&&(this.spacer=new Ow(e,0,0,[A.initialSpacer(e)]),this.dom.appendChild(this.spacer.dom),this.spacer.dom.style.cssText+="visibility: hidden; pointer-events: none")}update(e){let A=this.markers;if(this.markers=ttA(this.config.markers(e.view)),this.spacer&&this.config.updateSpacer){let n=this.config.updateSpacer(this.spacer.markers[0],e);n!=this.spacer.markers[0]&&this.spacer.update(e.view,0,0,[n])}let i=e.view.viewport;return!co.eq(this.markers,A,i.from,i.to)||(this.config.lineMarkerChange?this.config.lineMarkerChange(e):!1)}destroy(){for(let e of this.elements)e.destroy()}},Ow=class{constructor(e,A,i,n){this.height=-1,this.above=0,this.markers=[],this.dom=document.createElement("div"),this.dom.className="cm-gutterElement",this.update(e,A,i,n)}update(e,A,i,n){this.height!=A&&(this.height=A,this.dom.style.height=A+"px"),this.above!=i&&(this.dom.style.marginTop=(this.above=i)?i+"px":""),DLA(this.markers,n)||this.setMarkers(e,n)}setMarkers(e,A){let i="cm-gutterElement",n=this.dom.firstChild;for(let o=0,r=0;;){let s=r,a=o<A.length?A[o++]:null,c=!1;if(a){let l=a.elementClass;l&&(i+=" "+l);for(let I=r;I<this.markers.length;I++)if(this.markers[I].compare(a)){s=I,c=!0;break}}else s=this.markers.length;for(;r<s;){let l=this.markers[r++];if(l.toDOM){l.destroy(n);let I=n.nextSibling;n.remove(),n=I}}if(!a)break;a.toDOM&&(c?n=n.nextSibling:this.dom.insertBefore(a.toDOM(e),n)),c&&r++}this.dom.className=i,this.markers=A}destroy(){this.setMarkers(null,[])}};function DLA(t,e){if(t.length!=e.length)return!1;for(let A=0;A<t.length;A++)if(!t[A].compare(e[A]))return!1;return!0}var yLA=qe.define(),vLA=qe.define(),LE=qe.define({combine(t){return Wr(t,{formatNumber:String,domEventHandlers:{}},{domEventHandlers(e,A){let i=Object.assign({},e);for(let n in A){let o=i[n],r=A[n];i[n]=o?(s,a,c)=>o(s,a,c)||r(s,a,c):r}return i}})}}),A3=class extends Fa{constructor(e){super(),this.number=e}eq(e){return this.number==e.number}toDOM(){return document.createTextNode(this.number)}};function fL(t,e){return t.state.facet(LE).formatNumber(e,t.state)}var bLA=$4.compute([LE],t=>({class:"cm-lineNumbers",renderEmptyElements:!1,markers(e){return e.state.facet(yLA)},lineMarker(e,A,i){return i.some(n=>n.toDOM)?null:new A3(fL(e,e.state.doc.lineAt(A.from).number))},widgetMarker:(e,A,i)=>{for(let n of e.state.facet(vLA)){let o=n(e,A,i);if(o)return o}return null},lineMarkerChange:e=>e.startState.facet(LE)!=e.state.facet(LE),initialSpacer(e){return new A3(fL(e,itA(e.state.doc.lines)))},updateSpacer(e,A){let i=fL(A.view,itA(A.view.state.doc.lines));return i==e.number?e:new A3(i)},domEventHandlers:t.facet(LE).domEventHandlers,side:"before"}));function giA(t={}){return[LE.of(t),liA(),bLA]}function itA(t){let e=9;for(;e<t;)e=e*10+9;return e}var MLA=new class extends Fa{constructor(){super(...arguments),this.elementClass="cm-activeLineGutter"}},kLA=Sw.compute(["selection"],t=>{let e=[],A=-1;for(let i of t.selection.ranges){let n=t.doc.lineAt(i.head).from;n>A&&(A=n,e.push(MLA.range(n)))}return co.of(e)});function IiA(){return kLA}var SLA=0,g3=class{constructor(e,A){this.from=e,this.to=A}},fi=class{constructor(e={}){this.id=SLA++,this.perNode=!!e.perNode,this.deserialize=e.deserialize||(()=>{throw new Error("This node type doesn't define a deserialize function")})}add(e){if(this.perNode)throw new RangeError("Can't add per-node props to node types");return typeof e!="function"&&(e=Fs.match(e)),A=>{let i=e(A);return i===void 0?null:[this,i]}}};fi.closedBy=new fi({deserialize:t=>t.split(" ")});fi.openedBy=new fi({deserialize:t=>t.split(" ")});fi.group=new fi({deserialize:t=>t.split(" ")});fi.isolate=new fi({deserialize:t=>{if(t&&t!="rtl"&&t!="ltr"&&t!="auto")throw new RangeError("Invalid value for isolate: "+t);return t||"auto"}});fi.contextHash=new fi({perNode:!0});fi.lookAhead=new fi({perNode:!0});fi.mounted=new fi({perNode:!0});var UE=class{constructor(e,A,i){this.tree=e,this.overlay=A,this.parser=i}static get(e){return e&&e.props&&e.props[fi.mounted.id]}},RLA=Object.create(null),Fs=class t{constructor(e,A,i,n=0){this.name=e,this.props=A,this.id=i,this.flags=n}static define(e){let A=e.props&&e.props.length?Object.create(null):RLA,i=(e.top?1:0)|(e.skipped?2:0)|(e.error?4:0)|(e.name==null?8:0),n=new t(e.name||"",A,e.id,i);if(e.props){for(let o of e.props)if(Array.isArray(o)||(o=o(n)),o){if(o[0].perNode)throw new RangeError("Can't store a per-node prop on a node type");A[o[0].id]=o[1]}}return n}prop(e){return this.props[e.id]}get isTop(){return(this.flags&1)>0}get isSkipped(){return(this.flags&2)>0}get isError(){return(this.flags&4)>0}get isAnonymous(){return(this.flags&8)>0}is(e){if(typeof e=="string"){if(this.name==e)return!0;let A=this.prop(fi.group);return A?A.indexOf(e)>-1:!1}return this.id==e}static match(e){let A=Object.create(null);for(let i in e)for(let n of i.split(" "))A[n]=e[i];return i=>{for(let n=i.prop(fi.group),o=-1;o<(n?n.length:0);o++){let r=A[o<0?i.name:n[o]];if(r)return r}}}};Fs.none=new Fs("",Object.create(null),0,8);var I3=class t{constructor(e){this.types=e;for(let A=0;A<e.length;A++)if(e[A].id!=A)throw new RangeError("Node type ids should correspond to array positions when creating a node set")}extend(...e){let A=[];for(let i of this.types){let n=null;for(let o of e){let r=o(i);r&&(n||(n=Object.assign({},i.props)),n[r[0].id]=r[1])}A.push(n?new Fs(i.name,n,i.id,i.flags):i)}return new t(A)}},Zw=new WeakMap,CiA=new WeakMap,Jr=function(t){return t[t.ExcludeBuffers=1]="ExcludeBuffers",t[t.IncludeAnonymous=2]="IncludeAnonymous",t[t.IgnoreMounts=4]="IgnoreMounts",t[t.IgnoreOverlays=8]="IgnoreOverlays",t}(Jr||{}),ur=class t{constructor(e,A,i,n,o){if(this.type=e,this.children=A,this.positions=i,this.length=n,this.props=null,o&&o.length){this.props=Object.create(null);for(let[r,s]of o)this.props[typeof r=="number"?r:r.id]=s}}toString(){let e=UE.get(this);if(e&&!e.overlay)return e.tree.toString();let A="";for(let i of this.children){let n=i.toString();n&&(A&&(A+=","),A+=n)}return this.type.name?(/\W/.test(this.type.name)&&!this.type.isError?JSON.stringify(this.type.name):this.type.name)+(A.length?"("+A+")":""):A}cursor(e=0){return new B3(this.topNode,e)}cursorAt(e,A=0,i=0){let n=Zw.get(this)||this.topNode,o=new B3(n);return o.moveTo(e,A),Zw.set(this,o._tree),o}get topNode(){return new Fg(this,0,0,null)}resolve(e,A=0){let i=C3(Zw.get(this)||this.topNode,e,A,!1);return Zw.set(this,i),i}resolveInner(e,A=0){let i=C3(CiA.get(this)||this.topNode,e,A,!0);return CiA.set(this,i),i}resolveStack(e,A=0){return xLA(this,e,A)}iterate(e){let{enter:A,leave:i,from:n=0,to:o=this.length}=e,r=e.mode||0,s=(r&Jr.IncludeAnonymous)>0;for(let a=this.cursor(r|Jr.IncludeAnonymous);;){let c=!1;if(a.from<=o&&a.to>=n&&(!s&&a.type.isAnonymous||A(a)!==!1)){if(a.firstChild())continue;c=!0}for(;c&&i&&(s||!a.type.isAnonymous)&&i(a),!a.nextSibling();){if(!a.parent())return;c=!0}}}prop(e){return e.perNode?this.props?this.props[e.id]:void 0:this.type.prop(e)}get propValues(){let e=[];if(this.props)for(let A in this.props)e.push([+A,this.props[A]]);return e}balance(e={}){return this.children.length<=8?this:pF(Fs.none,this.children,this.positions,0,this.children.length,0,this.length,(A,i,n)=>new t(this.type,A,i,n,this.propValues),e.makeTree||((A,i,n)=>new t(Fs.none,A,i,n)))}static build(e){return LLA(e)}};ur.empty=new ur(Fs.none,[],[],0);var EF=class t{constructor(e,A){this.buffer=e,this.index=A}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}get pos(){return this.index}next(){this.index-=4}fork(){return new t(this.buffer,this.index)}},v1=class t{constructor(e,A,i){this.buffer=e,this.length=A,this.set=i}get type(){return Fs.none}toString(){let e=[];for(let A=0;A<this.buffer.length;)e.push(this.childString(A)),A=this.buffer[A+3];return e.join(",")}childString(e){let A=this.buffer[e],i=this.buffer[e+3],n=this.set.types[A],o=n.name;if(/\W/.test(o)&&!n.isError&&(o=JSON.stringify(o)),e+=4,i==e)return o;let r=[];for(;e<i;)r.push(this.childString(e)),e=this.buffer[e+3];return o+"("+r.join(",")+")"}findChild(e,A,i,n,o){let{buffer:r}=this,s=-1;for(let a=e;a!=A&&!(EiA(o,n,r[a+1],r[a+2])&&(s=a,i>0));a=r[a+3]);return s}slice(e,A,i){let n=this.buffer,o=new Uint16Array(A-e),r=0;for(let s=e,a=0;s<A;){o[a++]=n[s++],o[a++]=n[s++]-i;let c=o[a++]=n[s++]-i;o[a++]=n[s++]-e,r=Math.max(r,c)}return new t(o,r,this.set)}};function EiA(t,e,A,i){switch(t){case-2:return A<e;case-1:return i>=e&&A<e;case 0:return A<e&&i>e;case 1:return A<=e&&i>e;case 2:return i>e;case 4:return!0}}function C3(t,e,A,i){for(var n;t.from==t.to||(A<1?t.from>=e:t.from>e)||(A>-1?t.to<=e:t.to<e);){let r=!i&&t instanceof Fg&&t.index<0?null:t.parent;if(!r)return t;t=r}let o=i?0:Jr.IgnoreOverlays;if(i)for(let r=t,s=r.parent;s;r=s,s=r.parent)r instanceof Fg&&r.index<0&&((n=s.enter(e,A,o))===null||n===void 0?void 0:n.from)!=r.from&&(t=s);for(;;){let r=t.enter(e,A,o);if(!r)return t;t=r}}var Xw=class{cursor(e=0){return new B3(this,e)}getChild(e,A=null,i=null){let n=diA(this,e,A,i);return n.length?n[0]:null}getChildren(e,A=null,i=null){return diA(this,e,A,i)}resolve(e,A=0){return C3(this,e,A,!1)}resolveInner(e,A=0){return C3(this,e,A,!0)}matchContext(e){return QF(this.parent,e)}enterUnfinishedNodesBefore(e){let A=this.childBefore(e),i=this;for(;A;){let n=A.lastChild;if(!n||n.to!=A.to)break;n.type.isError&&n.from==n.to?(i=A,A=n.prevSibling):A=n}return i}get node(){return this}get next(){return this.parent}},Fg=class t extends Xw{constructor(e,A,i,n){super(),this._tree=e,this.from=A,this.index=i,this._parent=n}get type(){return this._tree.type}get name(){return this._tree.type.name}get to(){return this.from+this._tree.length}nextChild(e,A,i,n,o=0){for(let r=this;;){for(let{children:s,positions:a}=r._tree,c=A>0?s.length:-1;e!=c;e+=A){let l=s[e],I=a[e]+r.from;if(EiA(n,i,I,I+l.length)){if(l instanceof v1){if(o&Jr.ExcludeBuffers)continue;let C=l.findChild(0,l.buffer.length,A,i-I,n);if(C>-1)return new d3(new hF(r,l,e,I),null,C)}else if(o&Jr.IncludeAnonymous||!l.type.isAnonymous||mF(l)){let C;if(!(o&Jr.IgnoreMounts)&&(C=UE.get(l))&&!C.overlay)return new t(C.tree,I,e,r);let d=new t(l,I,e,r);return o&Jr.IncludeAnonymous||!d.type.isAnonymous?d:d.nextChild(A<0?l.children.length-1:0,A,i,n)}}}if(o&Jr.IncludeAnonymous||!r.type.isAnonymous||(r.index>=0?e=r.index+A:e=A<0?-1:r._parent._tree.children.length,r=r._parent,!r))return null}}get firstChild(){return this.nextChild(0,1,0,4)}get lastChild(){return this.nextChild(this._tree.children.length-1,-1,0,4)}childAfter(e){return this.nextChild(0,1,e,2)}childBefore(e){return this.nextChild(this._tree.children.length-1,-1,e,-2)}enter(e,A,i=0){let n;if(!(i&Jr.IgnoreOverlays)&&(n=UE.get(this._tree))&&n.overlay){let o=e-this.from;for(let{from:r,to:s}of n.overlay)if((A>0?r<=o:r<o)&&(A<0?s>=o:s>o))return new t(n.tree,n.overlay[0].from+this.from,-1,this)}return this.nextChild(0,1,e,A,i)}nextSignificantParent(){let e=this;for(;e.type.isAnonymous&&e._parent;)e=e._parent;return e}get parent(){return this._parent?this._parent.nextSignificantParent():null}get nextSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index+1,1,0,4):null}get prevSibling(){return this._parent&&this.index>=0?this._parent.nextChild(this.index-1,-1,0,4):null}get tree(){return this._tree}toTree(){return this._tree}toString(){return this._tree.toString()}};function diA(t,e,A,i){let n=t.cursor(),o=[];if(!n.firstChild())return o;if(A!=null){for(let r=!1;!r;)if(r=n.type.is(A),!n.nextSibling())return o}for(;;){if(i!=null&&n.type.is(i))return o;if(n.type.is(e)&&o.push(n.node),!n.nextSibling())return i==null?o:[]}}function QF(t,e,A=e.length-1){for(let i=t;A>=0;i=i.parent){if(!i)return!1;if(!i.type.isAnonymous){if(e[A]&&e[A]!=i.name)return!1;A--}}return!0}var hF=class{constructor(e,A,i,n){this.parent=e,this.buffer=A,this.index=i,this.start=n}},d3=class t extends Xw{get name(){return this.type.name}get from(){return this.context.start+this.context.buffer.buffer[this.index+1]}get to(){return this.context.start+this.context.buffer.buffer[this.index+2]}constructor(e,A,i){super(),this.context=e,this._parent=A,this.index=i,this.type=e.buffer.set.types[e.buffer.buffer[i]]}child(e,A,i){let{buffer:n}=this.context,o=n.findChild(this.index+4,n.buffer[this.index+3],e,A-this.context.start,i);return o<0?null:new t(this.context,this,o)}get firstChild(){return this.child(1,0,4)}get lastChild(){return this.child(-1,0,4)}childAfter(e){return this.child(1,e,2)}childBefore(e){return this.child(-1,e,-2)}enter(e,A,i=0){if(i&Jr.ExcludeBuffers)return null;let{buffer:n}=this.context,o=n.findChild(this.index+4,n.buffer[this.index+3],A>0?1:-1,e-this.context.start,A);return o<0?null:new t(this.context,this,o)}get parent(){return this._parent||this.context.parent.nextSignificantParent()}externalSibling(e){return this._parent?null:this.context.parent.nextChild(this.context.index+e,e,0,4)}get nextSibling(){let{buffer:e}=this.context,A=e.buffer[this.index+3];return A<(this._parent?e.buffer[this._parent.index+3]:e.buffer.length)?new t(this.context,this._parent,A):this.externalSibling(1)}get prevSibling(){let{buffer:e}=this.context,A=this._parent?this._parent.index+4:0;return this.index==A?this.externalSibling(-1):new t(this.context,this._parent,e.findChild(A,this.index,-1,0,4))}get tree(){return null}toTree(){let e=[],A=[],{buffer:i}=this.context,n=this.index+4,o=i.buffer[this.index+3];if(o>n){let r=i.buffer[this.index+1];e.push(i.slice(n,o,r)),A.push(0)}return new ur(this.type,e,A,this.to-this.from)}toString(){return this.context.buffer.childString(this.index)}};function QiA(t){if(!t.length)return null;let e=0,A=t[0];for(let o=1;o<t.length;o++){let r=t[o];(r.from>A.from||r.to<A.to)&&(A=r,e=o)}let i=A instanceof Fg&&A.index<0?null:A.parent,n=t.slice();return i?n[e]=i:n.splice(e,1),new uF(n,A)}var uF=class{constructor(e,A){this.heads=e,this.node=A}get next(){return QiA(this.heads)}};function xLA(t,e,A){let i=t.resolveInner(e,A),n=null;for(let o=i instanceof Fg?i:i.context.parent;o;o=o.parent)if(o.index<0){let r=o.parent;(n||(n=[i])).push(r.resolve(e,A)),o=r}else{let r=UE.get(o.tree);if(r&&r.overlay&&r.overlay[0].from<=e&&r.overlay[r.overlay.length-1].to>=e){let s=new Fg(r.tree,r.overlay[0].from+o.from,-1,o);(n||(n=[i])).push(C3(s,e,A,!1))}}return n?QiA(n):i}var B3=class{get name(){return this.type.name}constructor(e,A=0){if(this.mode=A,this.buffer=null,this.stack=[],this.index=0,this.bufferNode=null,e instanceof Fg)this.yieldNode(e);else{this._tree=e.context.parent,this.buffer=e.context;for(let i=e._parent;i;i=i._parent)this.stack.unshift(i.index);this.bufferNode=e,this.yieldBuf(e.index)}}yieldNode(e){return e?(this._tree=e,this.type=e.type,this.from=e.from,this.to=e.to,!0):!1}yieldBuf(e,A){this.index=e;let{start:i,buffer:n}=this.buffer;return this.type=A||n.set.types[n.buffer[e]],this.from=i+n.buffer[e+1],this.to=i+n.buffer[e+2],!0}yield(e){return e?e instanceof Fg?(this.buffer=null,this.yieldNode(e)):(this.buffer=e.context,this.yieldBuf(e.index,e.type)):!1}toString(){return this.buffer?this.buffer.buffer.childString(this.index):this._tree.toString()}enterChild(e,A,i){if(!this.buffer)return this.yield(this._tree.nextChild(e<0?this._tree._tree.children.length-1:0,e,A,i,this.mode));let{buffer:n}=this.buffer,o=n.findChild(this.index+4,n.buffer[this.index+3],e,A-this.buffer.start,i);return o<0?!1:(this.stack.push(this.index),this.yieldBuf(o))}firstChild(){return this.enterChild(1,0,4)}lastChild(){return this.enterChild(-1,0,4)}childAfter(e){return this.enterChild(1,e,2)}childBefore(e){return this.enterChild(-1,e,-2)}enter(e,A,i=this.mode){return this.buffer?i&Jr.ExcludeBuffers?!1:this.enterChild(1,e,A):this.yield(this._tree.enter(e,A,i))}parent(){if(!this.buffer)return this.yieldNode(this.mode&Jr.IncludeAnonymous?this._tree._parent:this._tree.parent);if(this.stack.length)return this.yieldBuf(this.stack.pop());let e=this.mode&Jr.IncludeAnonymous?this.buffer.parent:this.buffer.parent.nextSignificantParent();return this.buffer=null,this.yieldNode(e)}sibling(e){if(!this.buffer)return this._tree._parent?this.yield(this._tree.index<0?null:this._tree._parent.nextChild(this._tree.index+e,e,0,4,this.mode)):!1;let{buffer:A}=this.buffer,i=this.stack.length-1;if(e<0){let n=i<0?0:this.stack[i]+4;if(this.index!=n)return this.yieldBuf(A.findChild(n,this.index,-1,0,4))}else{let n=A.buffer[this.index+3];if(n<(i<0?A.buffer.length:A.buffer[this.stack[i]+3]))return this.yieldBuf(n)}return i<0?this.yield(this.buffer.parent.nextChild(this.buffer.index+e,e,0,4,this.mode)):!1}nextSibling(){return this.sibling(1)}prevSibling(){return this.sibling(-1)}atLastNode(e){let A,i,{buffer:n}=this;if(n){if(e>0){if(this.index<n.buffer.buffer.length)return!1}else for(let o=0;o<this.index;o++)if(n.buffer.buffer[o+3]<this.index)return!1;({index:A,parent:i}=n)}else({index:A,_parent:i}=this._tree);for(;i;{index:A,_parent:i}=i)if(A>-1)for(let o=A+e,r=e<0?-1:i._tree.children.length;o!=r;o+=e){let s=i._tree.children[o];if(this.mode&Jr.IncludeAnonymous||s instanceof v1||!s.type.isAnonymous||mF(s))return!1}return!0}move(e,A){if(A&&this.enterChild(e,0,4))return!0;for(;;){if(this.sibling(e))return!0;if(this.atLastNode(e)||!this.parent())return!1}}next(e=!0){return this.move(1,e)}prev(e=!0){return this.move(-1,e)}moveTo(e,A=0){for(;(this.from==this.to||(A<1?this.from>=e:this.from>e)||(A>-1?this.to<=e:this.to<e))&&this.parent(););for(;this.enterChild(1,e,A););return this}get node(){if(!this.buffer)return this._tree;let e=this.bufferNode,A=null,i=0;if(e&&e.context==this.buffer)A:for(let n=this.index,o=this.stack.length;o>=0;){for(let r=e;r;r=r._parent)if(r.index==n){if(n==this.index)return r;A=r,i=o+1;break A}n=this.stack[--o]}for(let n=i;n<this.stack.length;n++)A=new d3(this.buffer,A,this.stack[n]);return this.bufferNode=new d3(this.buffer,A,this.index)}get tree(){return this.buffer?null:this._tree._tree}iterate(e,A){for(let i=0;;){let n=!1;if(this.type.isAnonymous||e(this)!==!1){if(this.firstChild()){i++;continue}this.type.isAnonymous||(n=!0)}for(;;){if(n&&A&&A(this),n=this.type.isAnonymous,!i)return;if(this.nextSibling())break;this.parent(),i--,n=!0}}}matchContext(e){if(!this.buffer)return QF(this.node.parent,e);let{buffer:A}=this.buffer,{types:i}=A.set;for(let n=e.length-1,o=this.stack.length-1;n>=0;o--){if(o<0)return QF(this._tree,e,n);let r=i[A.buffer[this.stack[o]]];if(!r.isAnonymous){if(e[n]&&e[n]!=r.name)return!1;n--}}return!0}};function mF(t){return t.children.some(e=>e instanceof v1||!e.type.isAnonymous||mF(e))}function LLA(t){var e;let{buffer:A,nodeSet:i,maxBufferLength:n=1024,reused:o=[],minRepeatType:r=i.types.length}=t,s=Array.isArray(A)?new EF(A,A.length):A,a=i.types,c=0,l=0;function I(w,_,K,z,U,H){let{id:q,start:j,end:gA,size:QA}=s,BA=l,lA=c;for(;QA<0;)if(s.next(),QA==-1){let VA=o[q];K.push(VA),z.push(j-w);return}else if(QA==-3){c=q;return}else if(QA==-4){l=q;return}else throw new RangeError(`Unrecognized record size: ${QA}`);let vA=a[q],tA,cA,pA=j-w;if(gA-j<=n&&(cA=h(s.pos-_,U))){let VA=new Uint16Array(cA.size-cA.skip),oe=s.pos-cA.size,KA=VA.length;for(;s.pos>oe;)KA=u(cA.start,VA,KA);tA=new v1(VA,gA-cA.start,i),pA=cA.start-w}else{let VA=s.pos-QA;s.next();let oe=[],KA=[],CA=q>=r?q:-1,TA=0,Ze=gA;for(;s.pos>VA;)CA>=0&&s.id==CA&&s.size>=0?(s.end<=Ze-n&&(B(oe,KA,j,TA,s.end,Ze,CA,BA,lA),TA=oe.length,Ze=s.end),s.next()):H>2500?C(j,VA,oe,KA):I(j,VA,oe,KA,CA,H+1);if(CA>=0&&TA>0&&TA<oe.length&&B(oe,KA,j,TA,j,Ze,CA,BA,lA),oe.reverse(),KA.reverse(),CA>-1&&TA>0){let He=d(vA,lA);tA=pF(vA,oe,KA,0,oe.length,0,gA-j,He,He)}else tA=E(vA,oe,KA,gA-j,BA-gA,lA)}K.push(tA),z.push(pA)}function C(w,_,K,z){let U=[],H=0,q=-1;for(;s.pos>_;){let{id:j,start:gA,end:QA,size:BA}=s;if(BA>4)s.next();else{if(q>-1&&gA<q)break;q<0&&(q=QA-n),U.push(j,gA,QA),H++,s.next()}}if(H){let j=new Uint16Array(H*4),gA=U[U.length-2];for(let QA=U.length-3,BA=0;QA>=0;QA-=3)j[BA++]=U[QA],j[BA++]=U[QA+1]-gA,j[BA++]=U[QA+2]-gA,j[BA++]=BA;K.push(new v1(j,U[2]-gA,i)),z.push(gA-w)}}function d(w,_){return(K,z,U)=>{let H=0,q=K.length-1,j,gA;if(q>=0&&(j=K[q])instanceof ur){if(!q&&j.type==w&&j.length==U)return j;(gA=j.prop(fi.lookAhead))&&(H=z[q]+j.length+gA)}return E(w,K,z,U,H,_)}}function B(w,_,K,z,U,H,q,j,gA){let QA=[],BA=[];for(;w.length>z;)QA.push(w.pop()),BA.push(_.pop()+K-U);w.push(E(i.types[q],QA,BA,H-U,j-H,gA)),_.push(U-K)}function E(w,_,K,z,U,H,q){if(H){let j=[fi.contextHash,H];q=q?[j].concat(q):[j]}if(U>25){let j=[fi.lookAhead,U];q=q?[j].concat(q):[j]}return new ur(w,_,K,z,q)}function h(w,_){let K=s.fork(),z=0,U=0,H=0,q=K.end-n,j={size:0,start:0,skip:0};A:for(let gA=K.pos-w;K.pos>gA;){let QA=K.size;if(K.id==_&&QA>=0){j.size=z,j.start=U,j.skip=H,H+=4,z+=4,K.next();continue}let BA=K.pos-QA;if(QA<0||BA<gA||K.start<q)break;let lA=K.id>=r?4:0,vA=K.start;for(K.next();K.pos>BA;){if(K.size<0)if(K.size==-3)lA+=4;else break A;else K.id>=r&&(lA+=4);K.next()}U=vA,z+=QA,H+=lA}return(_<0||z==w)&&(j.size=z,j.start=U,j.skip=H),j.size>4?j:void 0}function u(w,_,K){let{id:z,start:U,end:H,size:q}=s;if(s.next(),q>=0&&z<r){let j=K;if(q>4){let gA=s.pos-(q-4);for(;s.pos>gA;)K=u(w,_,K)}_[--K]=j,_[--K]=H-w,_[--K]=U-w,_[--K]=z}else q==-3?c=z:q==-4&&(l=z);return K}let D=[],L=[];for(;s.pos>0;)I(t.start||0,t.bufferStart||0,D,L,-1,0);let R=(e=t.length)!==null&&e!==void 0?e:D.length?L[0]+D[0].length:0;return new ur(a[t.topID],D.reverse(),L.reverse(),R)}var BiA=new WeakMap;function Ww(t,e){if(!t.isAnonymous||e instanceof v1||e.type!=t)return 1;let A=BiA.get(e);if(A==null){A=1;for(let i of e.children){if(i.type!=t||!(i instanceof ur)){A=1;break}A+=Ww(t,i)}BiA.set(e,A)}return A}function pF(t,e,A,i,n,o,r,s,a){let c=0;for(let B=i;B<n;B++)c+=Ww(t,e[B]);let l=Math.ceil(c*1.5/8),I=[],C=[];function d(B,E,h,u,D){for(let L=h;L<u;){let R=L,w=E[L],_=Ww(t,B[L]);for(L++;L<u;L++){let K=Ww(t,B[L]);if(_+K>=l)break;_+=K}if(L==R+1){if(_>l){let K=B[R];d(K.children,K.positions,0,K.children.length,E[R]+D);continue}I.push(B[R])}else{let K=E[L-1]+B[L-1].length-w;I.push(pF(t,B,E,R,L,w,K,null,a))}C.push(w+D-o)}}return d(e,A,i,n,0),(s||a)(I,C,r)}var SC=class t{constructor(e,A,i,n,o=!1,r=!1){this.from=e,this.to=A,this.tree=i,this.offset=n,this.open=(o?1:0)|(r?2:0)}get openStart(){return(this.open&1)>0}get openEnd(){return(this.open&2)>0}static addTree(e,A=[],i=!1){let n=[new t(0,e.length,e,0,!1,i)];for(let o of A)o.to>e.length&&n.push(o);return n}static applyChanges(e,A,i=128){if(!A.length)return e;let n=[],o=1,r=e.length?e[0]:null;for(let s=0,a=0,c=0;;s++){let l=s<A.length?A[s]:null,I=l?l.fromA:1e9;if(I-a>=i)for(;r&&r.from<I;){let C=r;if(a>=C.from||I<=C.to||c){let d=Math.max(C.from,a)-c,B=Math.min(C.to,I)-c;C=d>=B?null:new t(d,B,C.tree,C.offset+c,s>0,!!l)}if(C&&n.push(C),r.to>I)break;r=o<e.length?e[o++]:null}if(!l)break;a=l.toA,c=l.toA-l.toB}return n}},KE=class{startParse(e,A,i){return typeof e=="string"&&(e=new fF(e)),i=i?i.length?i.map(n=>new g3(n.from,n.to)):[new g3(0,0)]:[new g3(0,e.length)],this.createParse(e,A||[],i)}parse(e,A,i){let n=this.startParse(e,A,i);for(;;){let o=n.advance();if(o)return o}}},fF=class{constructor(e){this.string=e}get length(){return this.string.length}chunk(e){return this.string.slice(e)}get lineChunks(){return!1}read(e,A){return this.string.slice(e,A)}};var Gpe=new fi({perNode:!0});var FLA=0,vl=class t{constructor(e,A,i,n){this.name=e,this.set=A,this.base=i,this.modified=n,this.id=FLA++}toString(){let{name:e}=this;for(let A of this.modified)A.name&&(e=`${A.name}(${e})`);return e}static define(e,A){let i=typeof e=="string"?e:"?";if(e instanceof t&&(A=e),A?.base)throw new Error("Can not derive from a modified tag");let n=new t(i,[],null,[]);if(n.set.push(n),A)for(let o of A.set)n.set.push(o);return n}static defineModifier(e){let A=new tD(e);return i=>i.modified.indexOf(A)>-1?i:tD.get(i.base||i,i.modified.concat(A).sort((n,o)=>n.id-o.id))}},NLA=0,tD=class t{constructor(e){this.name=e,this.instances=[],this.id=NLA++}static get(e,A){if(!A.length)return e;let i=A[0].instances.find(s=>s.base==e&&_LA(A,s.modified));if(i)return i;let n=[],o=new vl(e.name,n,e,A);for(let s of A)s.instances.push(o);let r=GLA(A);for(let s of e.set)if(!s.modified.length)for(let a of r)n.push(t.get(s,a));return o}};function _LA(t,e){return t.length==e.length&&t.every((A,i)=>A==e[i])}function GLA(t){let e=[[]];for(let A=0;A<t.length;A++)for(let i=0,n=e.length;i<n;i++)e.push(e[i].concat(t[A]));return e.sort((A,i)=>i.length-A.length)}function iD(t){let e=Object.create(null);for(let A in t){let i=t[A];Array.isArray(i)||(i=[i]);for(let n of A.split(" "))if(n){let o=[],r=2,s=n;for(let I=0;;){if(s=="..."&&I>0&&I+3==n.length){r=1;break}let C=/^"(?:[^"\\]|\\.)*?"|[^\/!]+/.exec(s);if(!C)throw new RangeError("Invalid path: "+n);if(o.push(C[0]=="*"?"":C[0][0]=='"'?JSON.parse(C[0]):C[0]),I+=C[0].length,I==n.length)break;let d=n[I++];if(I==n.length&&d=="!"){r=0;break}if(d!="/")throw new RangeError("Invalid path: "+n);s=n.slice(I)}let a=o.length-1,c=o[a];if(!c)throw new RangeError("Invalid path: "+n);let l=new YE(i,r,a>0?o.slice(0,a):null);e[c]=l.sort(e[c])}}return fiA.add(e)}var fiA=new fi,YE=class{constructor(e,A,i,n){this.tags=e,this.mode=A,this.context=i,this.next=n}get opaque(){return this.mode==0}get inherit(){return this.mode==1}sort(e){return!e||e.depth<this.depth?(this.next=e,this):(e.next=this.sort(e.next),e)}get depth(){return this.context?this.context.length:0}};YE.empty=new YE([],2,null);function vF(t,e){let A=Object.create(null);for(let o of t)if(!Array.isArray(o.tag))A[o.tag.id]=o.class;else for(let r of o.tag)A[r.id]=o.class;let{scope:i,all:n=null}=e||{};return{style:o=>{let r=n;for(let s of o)for(let a of s.set){let c=A[a.id];if(c){r=r?r+" "+c:c;break}}return r},scope:i}}function ULA(t,e){let A=null;for(let i of t){let n=i.style(e);n&&(A=A?A+" "+n:n)}return A}function miA(t,e,A,i=0,n=t.length){let o=new DF(i,Array.isArray(e)?e:[e],A);o.highlightRange(t.cursor(),i,n,"",o.highlighters),o.flush(n)}var DF=class{constructor(e,A,i){this.at=e,this.highlighters=A,this.span=i,this.class=""}startSpan(e,A){A!=this.class&&(this.flush(e),e>this.at&&(this.at=e),this.class=A)}flush(e){e>this.at&&this.class&&this.span(this.at,e,this.class)}highlightRange(e,A,i,n,o){let{type:r,from:s,to:a}=e;if(s>=i||a<=A)return;r.isTop&&(o=this.highlighters.filter(d=>!d.scope||d.scope(r)));let c=n,l=KLA(e)||YE.empty,I=ULA(o,l.tags);if(I&&(c&&(c+=" "),c+=I,l.mode==1&&(n+=(n?" ":"")+I)),this.startSpan(Math.max(A,s),c),l.opaque)return;let C=e.tree&&e.tree.prop(fi.mounted);if(C&&C.overlay){let d=e.node.enter(C.overlay[0].from+s,1),B=this.highlighters.filter(h=>!h.scope||h.scope(C.tree.type)),E=e.firstChild();for(let h=0,u=s;;h++){let D=h<C.overlay.length?C.overlay[h]:null,L=D?D.from+s:a,R=Math.max(A,u),w=Math.min(i,L);if(R<w&&E)for(;e.from<w&&(this.highlightRange(e,R,w,n,o),this.startSpan(Math.min(w,e.to),c),!(e.to>=L||!e.nextSibling())););if(!D||L>i)break;u=D.to+s,u>A&&(this.highlightRange(d.cursor(),Math.max(A,D.from+s),Math.min(i,u),"",B),this.startSpan(Math.min(i,u),c))}E&&e.parent()}else if(e.firstChild()){C&&(n="");do if(!(e.to<=A)){if(e.from>=i)break;this.highlightRange(e,A,i,n,o),this.startSpan(Math.min(i,e.to),c)}while(e.nextSibling());e.parent()}}};function KLA(t){let e=t.type.prop(fiA);for(;e&&e.context&&!t.matchContext(e.context);)e=e.next;return e||null}var Oe=vl.define,$w=Oe(),b1=Oe(),hiA=Oe(b1),uiA=Oe(b1),M1=Oe(),AD=Oe(M1),wF=Oe(M1),Gg=Oe(),RC=Oe(Gg),Ng=Oe(),_g=Oe(),yF=Oe(),E3=Oe(yF),eD=Oe(),Se={comment:$w,lineComment:Oe($w),blockComment:Oe($w),docComment:Oe($w),name:b1,variableName:Oe(b1),typeName:hiA,tagName:Oe(hiA),propertyName:uiA,attributeName:Oe(uiA),className:Oe(b1),labelName:Oe(b1),namespace:Oe(b1),macroName:Oe(b1),literal:M1,string:AD,docString:Oe(AD),character:Oe(AD),attributeValue:Oe(AD),number:wF,integer:Oe(wF),float:Oe(wF),bool:Oe(M1),regexp:Oe(M1),escape:Oe(M1),color:Oe(M1),url:Oe(M1),keyword:Ng,self:Oe(Ng),null:Oe(Ng),atom:Oe(Ng),unit:Oe(Ng),modifier:Oe(Ng),operatorKeyword:Oe(Ng),controlKeyword:Oe(Ng),definitionKeyword:Oe(Ng),moduleKeyword:Oe(Ng),operator:_g,derefOperator:Oe(_g),arithmeticOperator:Oe(_g),logicOperator:Oe(_g),bitwiseOperator:Oe(_g),compareOperator:Oe(_g),updateOperator:Oe(_g),definitionOperator:Oe(_g),typeOperator:Oe(_g),controlOperator:Oe(_g),punctuation:yF,separator:Oe(yF),bracket:E3,angleBracket:Oe(E3),squareBracket:Oe(E3),paren:Oe(E3),brace:Oe(E3),content:Gg,heading:RC,heading1:Oe(RC),heading2:Oe(RC),heading3:Oe(RC),heading4:Oe(RC),heading5:Oe(RC),heading6:Oe(RC),contentSeparator:Oe(Gg),list:Oe(Gg),quote:Oe(Gg),emphasis:Oe(Gg),strong:Oe(Gg),link:Oe(Gg),monospace:Oe(Gg),strikethrough:Oe(Gg),inserted:Oe(),deleted:Oe(),changed:Oe(),invalid:Oe(),meta:eD,documentMeta:Oe(eD),annotation:Oe(eD),processingInstruction:Oe(eD),definition:vl.defineModifier("definition"),constant:vl.defineModifier("constant"),function:vl.defineModifier("function"),standard:vl.defineModifier("standard"),local:vl.defineModifier("local"),special:vl.defineModifier("special")};for(let t in Se){let e=Se[t];e instanceof vl&&(e.name=t)}var Ype=vF([{tag:Se.link,class:"tok-link"},{tag:Se.heading,class:"tok-heading"},{tag:Se.emphasis,class:"tok-emphasis"},{tag:Se.strong,class:"tok-strong"},{tag:Se.keyword,class:"tok-keyword"},{tag:Se.atom,class:"tok-atom"},{tag:Se.bool,class:"tok-bool"},{tag:Se.url,class:"tok-url"},{tag:Se.labelName,class:"tok-labelName"},{tag:Se.inserted,class:"tok-inserted"},{tag:Se.deleted,class:"tok-deleted"},{tag:Se.literal,class:"tok-literal"},{tag:Se.string,class:"tok-string"},{tag:Se.number,class:"tok-number"},{tag:[Se.regexp,Se.escape,Se.special(Se.string)],class:"tok-string2"},{tag:Se.variableName,class:"tok-variableName"},{tag:Se.local(Se.variableName),class:"tok-variableName tok-local"},{tag:Se.definition(Se.variableName),class:"tok-variableName tok-definition"},{tag:Se.special(Se.variableName),class:"tok-variableName2"},{tag:Se.definition(Se.propertyName),class:"tok-propertyName tok-definition"},{tag:Se.typeName,class:"tok-typeName"},{tag:Se.namespace,class:"tok-namespace"},{tag:Se.className,class:"tok-className"},{tag:Se.macroName,class:"tok-macroName"},{tag:Se.propertyName,class:"tok-propertyName"},{tag:Se.operator,class:"tok-operator"},{tag:Se.comment,class:"tok-comment"},{tag:Se.meta,class:"tok-meta"},{tag:Se.invalid,class:"tok-invalid"},{tag:Se.punctuation,class:"tok-punctuation"}]);var bF,JE=new fi;function YLA(t){return qe.define({combine:t?e=>e.concat(t):void 0})}var JLA=new fi,Ug=(()=>{class t{constructor(A,i,n=[],o=""){this.data=A,this.name=o,hr.prototype.hasOwnProperty("tree")||Object.defineProperty(hr.prototype,"tree",{get(){return $r(this)}}),this.parser=i,this.extension=[k1.of(this),hr.languageData.of((r,s,a)=>{let c=piA(r,s,a),l=c.type.prop(JE);if(!l)return[];let I=r.facet(l),C=c.type.prop(JLA);if(C){let d=c.resolve(s-c.from,a);for(let B of C)if(B.test(d,r)){let E=r.facet(B.facet);return B.type=="replace"?E:E.concat(I)}}return I})].concat(n)}isActiveAt(A,i,n=-1){return piA(A,i,n).type.prop(JE)==this.data}findRegions(A){let i=A.facet(k1);if(i?.data==this.data)return[{from:0,to:A.doc.length}];if(!i||!i.allowsNesting)return[];let n=[],o=(r,s)=>{if(r.prop(JE)==this.data){n.push({from:s,to:s+r.length});return}let a=r.prop(fi.mounted);if(a){if(a.tree.prop(JE)==this.data){if(a.overlay)for(let c of a.overlay)n.push({from:c.from+s,to:c.to+s});else n.push({from:s,to:s+r.length});return}else if(a.overlay){let c=n.length;if(o(a.tree,a.overlay[0].from+s),n.length>c)return}}for(let c=0;c<r.children.length;c++){let l=r.children[c];l instanceof ur&&o(l,r.positions[c]+s)}};return o($r(A),0),n}get allowsNesting(){return!0}}return t.setState=Pi.define(),t})();function piA(t,e,A){let i=t.facet(k1),n=$r(t).topNode;if(!i||i.allowsNesting)for(let o=n;o;o=o.enter(e,A,Jr.ExcludeBuffers))o.type.isTop&&(n=o);return n}var nD=class t extends Ug{constructor(e,A,i){super(e,A,[],i),this.parser=A}static define(e){let A=YLA(e.languageData);return new t(A,e.parser.configure({props:[JE.add(i=>i.isTop?A:void 0)]}),e.name)}configure(e,A){return new t(this.data,this.parser.configure(e),A||this.name)}get allowsNesting(){return this.parser.hasWrappers()}};function $r(t){let e=t.field(Ug.state,!1);return e?e.tree:ur.empty}var RF=class{constructor(e){this.doc=e,this.cursorPos=0,this.string="",this.cursor=e.iter()}get length(){return this.doc.length}syncTo(e){return this.string=this.cursor.next(e-this.cursorPos).value,this.cursorPos=e+this.string.length,this.cursorPos-this.string.length}chunk(e){return this.syncTo(e),this.string}get lineChunks(){return!0}read(e,A){let i=this.cursorPos-this.string.length;return e<i||A>=this.cursorPos?this.doc.sliceString(e,A):this.string.slice(e-i,A-i)}},Q3=null,xF=class t{constructor(e,A,i=[],n,o,r,s,a){this.parser=e,this.state=A,this.fragments=i,this.tree=n,this.treeLen=o,this.viewport=r,this.skipped=s,this.scheduleOn=a,this.parse=null,this.tempSkipped=[]}static create(e,A,i){return new t(e,A,[],ur.empty,0,i,[],null)}startParse(){return this.parser.startParse(new RF(this.state.doc),this.fragments)}work(e,A){return A!=null&&A>=this.state.doc.length&&(A=void 0),this.tree!=ur.empty&&this.isDone(A??this.state.doc.length)?(this.takeTree(),!0):this.withContext(()=>{var i;if(typeof e=="number"){let n=Date.now()+e;e=()=>Date.now()>n}for(this.parse||(this.parse=this.startParse()),A!=null&&(this.parse.stoppedAt==null||this.parse.stoppedAt>A)&&A<this.state.doc.length&&this.parse.stopAt(A);;){let n=this.parse.advance();if(n)if(this.fragments=this.withoutTempSkipped(SC.addTree(n,this.fragments,this.parse.stoppedAt!=null)),this.treeLen=(i=this.parse.stoppedAt)!==null&&i!==void 0?i:this.state.doc.length,this.tree=n,this.parse=null,this.treeLen<(A??this.state.doc.length))this.parse=this.startParse();else return!0;if(e())return!1}})}takeTree(){let e,A;this.parse&&(e=this.parse.parsedPos)>=this.treeLen&&((this.parse.stoppedAt==null||this.parse.stoppedAt>e)&&this.parse.stopAt(e),this.withContext(()=>{for(;!(A=this.parse.advance()););}),this.treeLen=e,this.tree=A,this.fragments=this.withoutTempSkipped(SC.addTree(this.tree,this.fragments,!0)),this.parse=null)}withContext(e){let A=Q3;Q3=this;try{return e()}finally{Q3=A}}withoutTempSkipped(e){for(let A;A=this.tempSkipped.pop();)e=wiA(e,A.from,A.to);return e}changes(e,A){let{fragments:i,tree:n,treeLen:o,viewport:r,skipped:s}=this;if(this.takeTree(),!e.empty){let a=[];if(e.iterChangedRanges((c,l,I,C)=>a.push({fromA:c,toA:l,fromB:I,toB:C})),i=SC.applyChanges(i,a),n=ur.empty,o=0,r={from:e.mapPos(r.from,-1),to:e.mapPos(r.to,1)},this.skipped.length){s=[];for(let c of this.skipped){let l=e.mapPos(c.from,1),I=e.mapPos(c.to,-1);l<I&&s.push({from:l,to:I})}}}return new t(this.parser,A,i,n,o,r,s,this.scheduleOn)}updateViewport(e){if(this.viewport.from==e.from&&this.viewport.to==e.to)return!1;this.viewport=e;let A=this.skipped.length;for(let i=0;i<this.skipped.length;i++){let{from:n,to:o}=this.skipped[i];n<e.to&&o>e.from&&(this.fragments=wiA(this.fragments,n,o),this.skipped.splice(i--,1))}return this.skipped.length>=A?!1:(this.reset(),!0)}reset(){this.parse&&(this.takeTree(),this.parse=null)}skipUntilInView(e,A){this.skipped.push({from:e,to:A})}static getSkippingParser(e){return new class extends KE{createParse(A,i,n){let o=n[0].from,r=n[n.length-1].to;return{parsedPos:o,advance(){let a=Q3;if(a){for(let c of n)a.tempSkipped.push(c);e&&(a.scheduleOn=a.scheduleOn?Promise.all([a.scheduleOn,e]):e)}return this.parsedPos=r,new ur(Fs.none,[],[],r-o)},stoppedAt:null,stopAt(){}}}}}isDone(e){e=Math.min(e,this.state.doc.length);let A=this.fragments;return this.treeLen>=e&&A.length&&A[0].from==0&&A[0].to>=e}static get(){return Q3}};function wiA(t,e,A){return SC.applyChanges(t,[{fromA:e,toA:A,fromB:e,toB:A}])}var u3=class t{constructor(e){this.context=e,this.tree=e.tree}apply(e){if(!e.docChanged&&this.tree==this.context.tree)return this;let A=this.context.changes(e.changes,e.state),i=this.context.treeLen==e.startState.doc.length?void 0:Math.max(e.changes.mapPos(this.context.treeLen),A.viewport.to);return A.work(20,i)||A.takeTree(),new t(A)}static init(e){let A=Math.min(3e3,e.doc.length),i=xF.create(e.facet(k1).parser,e,{from:0,to:A});return i.work(20,A)||i.takeTree(),new t(i)}};Ug.state=Ar.define({create:u3.init,update(t,e){for(let A of e.effects)if(A.is(Ug.setState))return A.value;return e.startState.facet(k1)!=e.state.facet(k1)?u3.init(e.state):t.apply(e)}});var kiA=t=>{let e=setTimeout(()=>t(),500);return()=>clearTimeout(e)};typeof requestIdleCallback<"u"&&(kiA=t=>{let e=-1,A=setTimeout(()=>{e=requestIdleCallback(t,{timeout:400})},100);return()=>e<0?clearTimeout(A):cancelIdleCallback(e)});var MF=typeof navigator<"u"&&(!((bF=navigator.scheduling)===null||bF===void 0)&&bF.isInputPending)?()=>navigator.scheduling.isInputPending():null,TLA=go.fromClass(class{constructor(e){this.view=e,this.working=null,this.workScheduled=0,this.chunkEnd=-1,this.chunkBudget=-1,this.work=this.work.bind(this),this.scheduleWork()}update(e){let A=this.view.state.field(Ug.state).context;(A.updateViewport(e.view.viewport)||this.view.viewport.to>A.treeLen)&&this.scheduleWork(),(e.docChanged||e.selectionSet)&&(this.view.hasFocus&&(this.chunkBudget+=50),this.scheduleWork()),this.checkAsyncSchedule(A)}scheduleWork(){if(this.working)return;let{state:e}=this.view,A=e.field(Ug.state);(A.tree!=A.context.tree||!A.context.isDone(e.doc.length))&&(this.working=kiA(this.work))}work(e){this.working=null;let A=Date.now();if(this.chunkEnd<A&&(this.chunkEnd<0||this.view.hasFocus)&&(this.chunkEnd=A+3e4,this.chunkBudget=3e3),this.chunkBudget<=0)return;let{state:i,viewport:{to:n}}=this.view,o=i.field(Ug.state);if(o.tree==o.context.tree&&o.context.isDone(n+1e5))return;let r=Date.now()+Math.min(this.chunkBudget,100,e&&!MF?Math.max(25,e.timeRemaining()-5):1e9),s=o.context.treeLen<n&&i.doc.length>n+1e3,a=o.context.work(()=>MF&&MF()||Date.now()>r,n+(s?0:1e5));this.chunkBudget-=Date.now()-A,(a||this.chunkBudget<=0)&&(o.context.takeTree(),this.view.dispatch({effects:Ug.setState.of(new u3(o.context))})),this.chunkBudget>0&&!(a&&!s)&&this.scheduleWork(),this.checkAsyncSchedule(o.context)}checkAsyncSchedule(e){e.scheduleOn&&(this.workScheduled++,e.scheduleOn.then(()=>this.scheduleWork()).catch(A=>Xr(this.view.state,A)).then(()=>this.workScheduled--),e.scheduleOn=null)}destroy(){this.working&&this.working()}isWorking(){return!!(this.working||this.workScheduled>0)}},{eventHandlers:{focus(){this.scheduleWork()}}}),k1=qe.define({combine(t){return t.length?t[0]:null},enables:t=>[Ug.state,TLA,qt.contentAttributes.compute([t],e=>{let A=e.facet(t);return A&&A.name?{"data-language":A.name}:{}})]}),oD=class{constructor(e,A=[]){this.language=e,this.support=A,this.extension=[e,A]}};var HLA=qe.define(),FC=qe.define({combine:t=>{if(!t.length)return"  ";let e=t[0];if(!e||/\S/.test(e)||Array.from(e).some(A=>A!=e[0]))throw new Error("Invalid indent unit: "+JSON.stringify(t[0]));return e}});function Ml(t){let e=t.facet(FC);return e.charCodeAt(0)==9?t.tabSize*e.length:e.length}function HE(t,e){let A="",i=t.tabSize,n=t.facet(FC)[0];if(n=="	"){for(;e>=i;)A+="	",e-=i;n=" "}for(let o=0;o<e;o++)A+=n;return A}function aD(t,e){t instanceof hr&&(t=new xC(t));for(let i of t.state.facet(HLA)){let n=i(t,e);if(n!==void 0)return n}let A=$r(t.state);return A.length>=e?zLA(t,A,e):null}var xC=class{constructor(e,A={}){this.state=e,this.options=A,this.unit=Ml(e)}lineAt(e,A=1){let i=this.state.doc.lineAt(e),{simulateBreak:n,simulateDoubleBreak:o}=this.options;return n!=null&&n>=i.from&&n<=i.to?o&&n==e?{text:"",from:e}:(A<0?n<e:n<=e)?{text:i.text.slice(n-i.from),from:n}:{text:i.text.slice(0,n-i.from),from:i.from}:i}textAfterPos(e,A=1){if(this.options.simulateDoubleBreak&&e==this.options.simulateBreak)return"";let{text:i,from:n}=this.lineAt(e,A);return i.slice(e-n,Math.min(i.length,e+100-n))}column(e,A=1){let{text:i,from:n}=this.lineAt(e,A),o=this.countColumn(i,e-n),r=this.options.overrideIndentation?this.options.overrideIndentation(n):-1;return r>-1&&(o+=r-this.countColumn(i,i.search(/\S|$/))),o}countColumn(e,A=e.length){return J0(e,this.state.tabSize,A)}lineIndent(e,A=1){let{text:i,from:n}=this.lineAt(e,A),o=this.options.overrideIndentation;if(o){let r=o(n);if(r>-1)return r}return this.countColumn(i,i.search(/\S|$/))}get simulatedBreak(){return this.options.simulateBreak||null}},KF=new fi;function zLA(t,e,A){let i=e.resolveStack(A),n=e.resolveInner(A,-1).resolve(A,0).enterUnfinishedNodesBefore(A);if(n!=i.node){let o=[];for(let r=n;r&&!(r.from<i.node.from||r.to>i.node.to||r.from==i.node.from&&r.type==i.node.type);r=r.parent)o.push(r);for(let r=o.length-1;r>=0;r--)i={node:o[r],next:i}}return SiA(i,t,A)}function SiA(t,e,A){for(let i=t;i;i=i.next){let n=PLA(i.node);if(n)return n(LF.create(e,A,i))}return 0}function OLA(t){return t.pos==t.options.simulateBreak&&t.options.simulateDoubleBreak}function PLA(t){let e=t.type.prop(KF);if(e)return e;let A=t.firstChild,i;if(A&&(i=A.type.prop(fi.closedBy))){let n=t.lastChild,o=n&&i.indexOf(n.name)>-1;return r=>ZLA(r,!0,1,void 0,o&&!OLA(r)?n.from:void 0)}return t.parent==null?jLA:null}function jLA(){return 0}var LF=class t extends xC{constructor(e,A,i){super(e.state,e.options),this.base=e,this.pos=A,this.context=i}get node(){return this.context.node}static create(e,A,i){return new t(e,A,i)}get textAfter(){return this.textAfterPos(this.pos)}get baseIndent(){return this.baseIndentFor(this.node)}baseIndentFor(e){let A=this.state.doc.lineAt(e.from);for(;;){let i=e.resolve(A.from);for(;i.parent&&i.parent.from==i.from;)i=i.parent;if(qLA(i,e))break;A=this.state.doc.lineAt(i.from)}return this.lineIndent(A.from)}continue(){return SiA(this.context.next,this.base,this.pos)}};function qLA(t,e){for(let A=e;A;A=A.parent)if(t==A)return!0;return!1}function VLA(t){let e=t.node,A=e.childAfter(e.from),i=e.lastChild;if(!A)return null;let n=t.options.simulateBreak,o=t.state.doc.lineAt(A.from),r=n==null||n<=o.from?o.to:Math.min(o.to,n);for(let s=A.to;;){let a=e.childAfter(s);if(!a||a==i)return null;if(!a.type.isSkipped){if(a.from>=r)return null;let c=/^ */.exec(o.text.slice(A.to-o.from))[0].length;return{from:A.from,to:A.to+c}}s=a.to}}function ZLA(t,e,A,i,n){let o=t.textAfter,r=o.match(/^\s*/)[0].length,s=i&&o.slice(r,r+i.length)==i||n==t.pos+r,a=e?VLA(t):null;return a?s?t.column(a.from):t.column(a.to):t.baseIndent+(s?0:t.unit*A)}function YF({except:t,units:e=1}={}){return A=>{let i=t&&t.test(A.textAfter);return A.baseIndent+(i?0:e*A.unit)}}var WLA=200;function RiA(){return hr.transactionFilter.of(t=>{if(!t.docChanged||!t.isUserEvent("input.type")&&!t.isUserEvent("input.complete"))return t;let e=t.startState.languageDataAt("indentOnInput",t.startState.selection.main.head);if(!e.length)return t;let A=t.newDoc,{head:i}=t.newSelection.main,n=A.lineAt(i);if(i>n.from+WLA)return t;let o=A.sliceString(n.from,i);if(!e.some(c=>c.test(o)))return t;let{state:r}=t,s=-1,a=[];for(let{head:c}of r.selection.ranges){let l=r.doc.lineAt(c);if(l.from==s)continue;s=l.from;let I=aD(r,l.from);if(I==null)continue;let C=/^\s*/.exec(l.text)[0],d=HE(r,I);C!=d&&a.push({from:l.from,to:l.from+C.length,insert:d})}return a.length?[t,{changes:a,sequential:!0}]:t})}var XLA=qe.define(),JF=new fi;function xiA(t){let e=t.firstChild,A=t.lastChild;return e&&e.to<A.from?{from:e.to,to:A.type.isError?t.to:A.from}:null}function $LA(t,e,A){let i=$r(t);if(i.length<A)return null;let n=i.resolveStack(A,1),o=null;for(let r=n;r;r=r.next){let s=r.node;if(s.to<=A||s.from>A)continue;if(o&&s.from<e)break;let a=s.type.prop(JF);if(a&&(s.to<i.length-50||i.length==t.doc.length||!AFA(s))){let c=a(s,t);c&&c.from<=A&&c.from>=e&&c.to>A&&(o=c)}}return o}function AFA(t){let e=t.lastChild;return e&&e.to==t.to&&e.type.isError}function rD(t,e,A){for(let i of t.facet(XLA)){let n=i(t,e,A);if(n)return n}return $LA(t,e,A)}function LiA(t,e){let A=e.mapPos(t.from,1),i=e.mapPos(t.to,-1);return A>=i?void 0:{from:A,to:i}}var cD=Pi.define({map:LiA}),f3=Pi.define({map:LiA});function FiA(t){let e=[];for(let{head:A}of t.state.selection.ranges)e.some(i=>i.from<=A&&i.to>=A)||e.push(t.lineBlockAt(A));return e}var LC=Ar.define({create(){return ut.none},update(t,e){e.isUserEvent("delete")&&e.changes.iterChangedRanges((A,i)=>t=DiA(t,A,i)),t=t.map(e.changes);for(let A of e.effects)if(A.is(cD)&&!eFA(t,A.value.from,A.value.to)){let{preparePlaceholder:i}=e.state.facet(TF),n=i?ut.replace({widget:new FF(i(e.state,A.value))}):yiA;t=t.update({add:[n.range(A.value.from,A.value.to)]})}else A.is(f3)&&(t=t.update({filter:(i,n)=>A.value.from!=i||A.value.to!=n,filterFrom:A.value.from,filterTo:A.value.to}));return e.selection&&(t=DiA(t,e.selection.main.head)),t},provide:t=>qt.decorations.from(t),toJSON(t,e){let A=[];return t.between(0,e.doc.length,(i,n)=>{A.push(i,n)}),A},fromJSON(t){if(!Array.isArray(t)||t.length%2)throw new RangeError("Invalid JSON for fold state");let e=[];for(let A=0;A<t.length;){let i=t[A++],n=t[A++];if(typeof i!="number"||typeof n!="number")throw new RangeError("Invalid JSON for fold state");e.push(yiA.range(i,n))}return ut.set(e,!0)}});function DiA(t,e,A=e){let i=!1;return t.between(e,A,(n,o)=>{n<A&&o>e&&(i=!0)}),i?t.update({filterFrom:e,filterTo:A,filter:(n,o)=>n>=A||o<=e}):t}function sD(t,e,A){var i;let n=null;return(i=t.field(LC,!1))===null||i===void 0||i.between(e,A,(o,r)=>{(!n||n.from>o)&&(n={from:o,to:r})}),n}function eFA(t,e,A){let i=!1;return t.between(e,e,(n,o)=>{n==e&&o==A&&(i=!0)}),i}function NiA(t,e){return t.field(LC,!1)?e:e.concat(Pi.appendConfig.of(UiA()))}var tFA=t=>{for(let e of FiA(t)){let A=rD(t.state,e.from,e.to);if(A)return t.dispatch({effects:NiA(t.state,[cD.of(A),_iA(t,A)])}),!0}return!1},iFA=t=>{if(!t.state.field(LC,!1))return!1;let e=[];for(let A of FiA(t)){let i=sD(t.state,A.from,A.to);i&&e.push(f3.of(i),_iA(t,i,!1))}return e.length&&t.dispatch({effects:e}),e.length>0};function _iA(t,e,A=!0){let i=t.state.doc.lineAt(e.from).number,n=t.state.doc.lineAt(e.to).number;return qt.announce.of(`${t.state.phrase(A?"Folded lines":"Unfolded lines")} ${i} ${t.state.phrase("to")} ${n}.`)}var nFA=t=>{let{state:e}=t,A=[];for(let i=0;i<e.doc.length;){let n=t.lineBlockAt(i),o=rD(e,n.from,n.to);o&&A.push(cD.of(o)),i=(o?t.lineBlockAt(o.to):n).to+1}return A.length&&t.dispatch({effects:NiA(t.state,A)}),!!A.length},oFA=t=>{let e=t.state.field(LC,!1);if(!e||!e.size)return!1;let A=[];return e.between(0,t.state.doc.length,(i,n)=>{A.push(f3.of({from:i,to:n}))}),t.dispatch({effects:A}),!0};var GiA=[{key:"Ctrl-Shift-[",mac:"Cmd-Alt-[",run:tFA},{key:"Ctrl-Shift-]",mac:"Cmd-Alt-]",run:iFA},{key:"Ctrl-Alt-[",run:nFA},{key:"Ctrl-Alt-]",run:oFA}],rFA={placeholderDOM:null,preparePlaceholder:null,placeholderText:"\u2026"},TF=qe.define({combine(t){return Wr(t,rFA)}});function UiA(t){let e=[LC,aFA];return t&&e.push(TF.of(t)),e}function KiA(t,e){let{state:A}=t,i=A.facet(TF),n=r=>{let s=t.lineBlockAt(t.posAtDOM(r.target)),a=sD(t.state,s.from,s.to);a&&t.dispatch({effects:f3.of(a)}),r.preventDefault()};if(i.placeholderDOM)return i.placeholderDOM(t,n,e);let o=document.createElement("span");return o.textContent=i.placeholderText,o.setAttribute("aria-label",A.phrase("folded code")),o.title=A.phrase("unfold"),o.className="cm-foldPlaceholder",o.onclick=n,o}var yiA=ut.replace({widget:new class extends Ic{toDOM(t){return KiA(t,null)}}}),FF=class extends Ic{constructor(e){super(),this.value=e}eq(e){return this.value==e.value}toDOM(e){return KiA(e,this.value)}},sFA={openText:"\u2304",closedText:"\u203A",markerDOM:null,domEventHandlers:{},foldingChanged:()=>!1},h3=class extends Fa{constructor(e,A){super(),this.config=e,this.open=A}eq(e){return this.config==e.config&&this.open==e.open}toDOM(e){if(this.config.markerDOM)return this.config.markerDOM(this.open);let A=document.createElement("span");return A.textContent=this.open?this.config.openText:this.config.closedText,A.title=e.state.phrase(this.open?"Fold line":"Unfold line"),A}};function YiA(t={}){let e=rA(rA({},sFA),t),A=new h3(e,!0),i=new h3(e,!1),n=go.fromClass(class{constructor(r){this.from=r.viewport.from,this.markers=this.buildMarkers(r)}update(r){(r.docChanged||r.viewportChanged||r.startState.facet(k1)!=r.state.facet(k1)||r.startState.field(LC,!1)!=r.state.field(LC,!1)||$r(r.startState)!=$r(r.state)||e.foldingChanged(r))&&(this.markers=this.buildMarkers(r.view))}buildMarkers(r){let s=new ds;for(let a of r.viewportLineBlocks){let c=sD(r.state,a.from,a.to)?i:rD(r.state,a.from,a.to)?A:null;c&&s.add(a.from,a.from,c)}return s.finish()}}),{domEventHandlers:o}=e;return[n,Vw({class:"cm-foldGutter",markers(r){var s;return((s=r.plugin(n))===null||s===void 0?void 0:s.markers)||co.empty},initialSpacer(){return new h3(e,!1)},domEventHandlers:Ye(rA({},o),{click:(r,s,a)=>{if(o.click&&o.click(r,s,a))return!0;let c=sD(r.state,s.from,s.to);if(c)return r.dispatch({effects:f3.of(c)}),!0;let l=rD(r.state,s.from,s.to);return l?(r.dispatch({effects:cD.of(l)}),!0):!1}})}),UiA()]}var aFA=qt.baseTheme({".cm-foldPlaceholder":{backgroundColor:"#eee",border:"1px solid #ddd",color:"#888",borderRadius:".2em",margin:"0 1px",padding:"0 1px",cursor:"pointer"},".cm-foldGutter span":{padding:"0 1px",cursor:"pointer"}}),TE=class t{constructor(e,A){this.specs=e;let i;function n(s){let a=Kc.newName();return(i||(i=Object.create(null)))["."+a]=s,a}let o=typeof A.all=="string"?A.all:A.all?n(A.all):void 0,r=A.scope;this.scope=r instanceof Ug?s=>s.prop(JE)==r.data:r?s=>s==r:void 0,this.style=vF(e.map(s=>({tag:s.tag,class:s.class||n(Object.assign({},s,{tag:null}))})),{all:o}).style,this.module=i?new Kc(i):null,this.themeType=A.themeType}static define(e,A){return new t(e,A||{})}},NF=qe.define(),JiA=qe.define({combine(t){return t.length?[t[0]]:null}});function kF(t){let e=t.facet(NF);return e.length?e:t.facet(JiA)}function HF(t,e){let A=[cFA],i;return t instanceof TE&&(t.module&&A.push(qt.styleModule.of(t.module)),i=t.themeType),e?.fallback?A.push(JiA.of(t)):i?A.push(NF.computeN([qt.darkTheme],n=>n.facet(qt.darkTheme)==(i=="dark")?[t]:[])):A.push(NF.of(t)),A}var _F=class{constructor(e){this.markCache=Object.create(null),this.tree=$r(e.state),this.decorations=this.buildDeco(e,kF(e.state)),this.decoratedTo=e.viewport.to}update(e){let A=$r(e.state),i=kF(e.state),n=i!=kF(e.startState),{viewport:o}=e.view,r=e.changes.mapPos(this.decoratedTo,1);A.length<o.to&&!n&&A.type==this.tree.type&&r>=o.to?(this.decorations=this.decorations.map(e.changes),this.decoratedTo=r):(A!=this.tree||e.viewportChanged||n)&&(this.tree=A,this.decorations=this.buildDeco(e.view,i),this.decoratedTo=o.to)}buildDeco(e,A){if(!A||!this.tree.length)return ut.none;let i=new ds;for(let{from:n,to:o}of e.visibleRanges)miA(this.tree,A,(r,s,a)=>{i.add(r,s,this.markCache[a]||(this.markCache[a]=ut.mark({class:a})))},n,o);return i.finish()}},cFA=Dl.high(go.fromClass(_F,{decorations:t=>t.decorations})),TiA=TE.define([{tag:Se.meta,color:"#404740"},{tag:Se.link,textDecoration:"underline"},{tag:Se.heading,textDecoration:"underline",fontWeight:"bold"},{tag:Se.emphasis,fontStyle:"italic"},{tag:Se.strong,fontWeight:"bold"},{tag:Se.strikethrough,textDecoration:"line-through"},{tag:Se.keyword,color:"#708"},{tag:[Se.atom,Se.bool,Se.url,Se.contentSeparator,Se.labelName],color:"#219"},{tag:[Se.literal,Se.inserted],color:"#164"},{tag:[Se.string,Se.deleted],color:"#a11"},{tag:[Se.regexp,Se.escape,Se.special(Se.string)],color:"#e40"},{tag:Se.definition(Se.variableName),color:"#00f"},{tag:Se.local(Se.variableName),color:"#30a"},{tag:[Se.typeName,Se.namespace],color:"#085"},{tag:Se.className,color:"#167"},{tag:[Se.special(Se.variableName),Se.macroName],color:"#256"},{tag:Se.definition(Se.propertyName),color:"#00c"},{tag:Se.comment,color:"#940"},{tag:Se.invalid,color:"#f00"}]),lFA=qt.baseTheme({"&.cm-focused .cm-matchingBracket":{backgroundColor:"#328c8252"},"&.cm-focused .cm-nonmatchingBracket":{backgroundColor:"#bb555544"}}),HiA=1e4,ziA="()[]{}",OiA=qe.define({combine(t){return Wr(t,{afterCursor:!0,brackets:ziA,maxScanDistance:HiA,renderMatch:CFA})}}),gFA=ut.mark({class:"cm-matchingBracket"}),IFA=ut.mark({class:"cm-nonmatchingBracket"});function CFA(t){let e=[],A=t.matched?gFA:IFA;return e.push(A.range(t.start.from,t.start.to)),t.end&&e.push(A.range(t.end.from,t.end.to)),e}var dFA=Ar.define({create(){return ut.none},update(t,e){if(!e.docChanged&&!e.selection)return t;let A=[],i=e.state.facet(OiA);for(let n of e.state.selection.ranges){if(!n.empty)continue;let o=bl(e.state,n.head,-1,i)||n.head>0&&bl(e.state,n.head-1,1,i)||i.afterCursor&&(bl(e.state,n.head,1,i)||n.head<e.state.doc.length&&bl(e.state,n.head+1,-1,i));o&&(A=A.concat(i.renderMatch(o,e.state)))}return ut.set(A,!0)},provide:t=>qt.decorations.from(t)}),BFA=[dFA,lFA];function PiA(t={}){return[OiA.of(t),BFA]}var EFA=new fi;function GF(t,e,A){let i=t.prop(e<0?fi.openedBy:fi.closedBy);if(i)return i;if(t.name.length==1){let n=A.indexOf(t.name);if(n>-1&&n%2==(e<0?1:0))return[A[n+e]]}return null}function UF(t){let e=t.type.prop(EFA);return e?e(t.node):t}function bl(t,e,A,i={}){let n=i.maxScanDistance||HiA,o=i.brackets||ziA,r=$r(t),s=r.resolveInner(e,A);for(let a=s;a;a=a.parent){let c=GF(a.type,A,o);if(c&&a.from<a.to){let l=UF(a);if(l&&(A>0?e>=l.from&&e<l.to:e>l.from&&e<=l.to))return QFA(t,e,A,a,l,c,o)}}return hFA(t,e,A,r,s.type,n,o)}function QFA(t,e,A,i,n,o,r){let s=i.parent,a={from:n.from,to:n.to},c=0,l=s?.cursor();if(l&&(A<0?l.childBefore(i.from):l.childAfter(i.to)))do if(A<0?l.to<=i.from:l.from>=i.to){if(c==0&&o.indexOf(l.type.name)>-1&&l.from<l.to){let I=UF(l);return{start:a,end:I?{from:I.from,to:I.to}:void 0,matched:!0}}else if(GF(l.type,A,r))c++;else if(GF(l.type,-A,r)){if(c==0){let I=UF(l);return{start:a,end:I&&I.from<I.to?{from:I.from,to:I.to}:void 0,matched:!1}}c--}}while(A<0?l.prevSibling():l.nextSibling());return{start:a,matched:!1}}function hFA(t,e,A,i,n,o,r){let s=A<0?t.sliceDoc(e-1,e):t.sliceDoc(e,e+1),a=r.indexOf(s);if(a<0||a%2==0!=A>0)return null;let c={from:A<0?e-1:e,to:A>0?e+1:e},l=t.doc.iterRange(e,A>0?t.doc.length:0),I=0;for(let C=0;!l.next().done&&C<=o;){let d=l.value;A<0&&(C+=d.length);let B=e+C*A;for(let E=A>0?0:d.length-1,h=A>0?d.length:-1;E!=h;E+=A){let u=r.indexOf(d[E]);if(!(u<0||i.resolveInner(B+E,1).type!=n))if(u%2==0==A>0)I++;else{if(I==1)return{start:c,end:{from:B+E,to:B+E+1},matched:u>>1==a>>1};I--}}A>0&&(C+=d.length)}return l.done?{start:c,matched:!1}:null}var uFA=Object.create(null),viA=[Fs.none];var biA=[],MiA=Object.create(null),fFA=Object.create(null);for(let[t,e]of[["variable","variableName"],["variable-2","variableName.special"],["string-2","string.special"],["def","variableName.definition"],["tag","tagName"],["attribute","attributeName"],["type","typeName"],["builtin","variableName.standard"],["qualifier","modifier"],["error","invalid"],["header","heading"],["property","propertyName"]])fFA[t]=mFA(uFA,e);function SF(t,e){biA.indexOf(t)>-1||(biA.push(t),console.warn(e))}function mFA(t,e){let A=[];for(let s of e.split(" ")){let a=[];for(let c of s.split(".")){let l=t[c]||Se[c];l?typeof l=="function"?a.length?a=a.map(l):SF(c,`Modifier ${c} used at start of tag`):a.length?SF(c,`Tag ${c} used as modifier`):a=Array.isArray(l)?l:[l]:SF(c,`Unknown highlighting tag ${c}`)}for(let c of a)A.push(c)}if(!A.length)return 0;let i=e.replace(/ /g,"_"),n=i+" "+A.map(s=>s.id),o=MiA[n];if(o)return o.id;let r=MiA[n]=Fs.define({id:viA.length,name:i,props:[iD({[i]:A})]});return viA.push(r),r.id}var qpe={rtl:ut.mark({class:"cm-iso",inclusive:!0,attributes:{dir:"rtl"},bidiIsolate:lo.RTL}),ltr:ut.mark({class:"cm-iso",inclusive:!0,attributes:{dir:"ltr"},bidiIsolate:lo.LTR}),auto:ut.mark({class:"cm-iso",inclusive:!0,attributes:{dir:"auto"},bidiIsolate:null})};var pFA=t=>{let{state:e}=t,A=e.doc.lineAt(e.selection.main.from),i=jF(t.state,A.from);return i.line?wFA(t):i.block?yFA(t):!1};function PF(t,e){return({state:A,dispatch:i})=>{if(A.readOnly)return!1;let n=t(e,A);return n?(i(A.update(n)),!0):!1}}var wFA=PF(MFA,0);var DFA=PF(enA,0);var yFA=PF((t,e)=>enA(t,e,bFA(e)),0);function jF(t,e){let A=t.languageDataAt("commentTokens",e,1);return A.length?A[0]:{}}var m3=50;function vFA(t,{open:e,close:A},i,n){let o=t.sliceDoc(i-m3,i),r=t.sliceDoc(n,n+m3),s=/\s*$/.exec(o)[0].length,a=/^\s*/.exec(r)[0].length,c=o.length-s;if(o.slice(c-e.length,c)==e&&r.slice(a,a+A.length)==A)return{open:{pos:i-s,margin:s&&1},close:{pos:n+a,margin:a&&1}};let l,I;n-i<=2*m3?l=I=t.sliceDoc(i,n):(l=t.sliceDoc(i,i+m3),I=t.sliceDoc(n-m3,n));let C=/^\s*/.exec(l)[0].length,d=/\s*$/.exec(I)[0].length,B=I.length-d-A.length;return l.slice(C,C+e.length)==e&&I.slice(B,B+A.length)==A?{open:{pos:i+C+e.length,margin:/\s/.test(l.charAt(C+e.length))?1:0},close:{pos:n-d-A.length,margin:/\s/.test(I.charAt(B-1))?1:0}}:null}function bFA(t){let e=[];for(let A of t.selection.ranges){let i=t.doc.lineAt(A.from),n=A.to<=i.to?i:t.doc.lineAt(A.to);n.from>i.from&&n.from==A.to&&(n=A.to==i.to+1?i:t.doc.lineAt(A.to-1));let o=e.length-1;o>=0&&e[o].to>i.from?e[o].to=n.to:e.push({from:i.from+/^\s*/.exec(i.text)[0].length,to:n.to})}return e}function enA(t,e,A=e.selection.ranges){let i=A.map(o=>jF(e,o.from).block);if(!i.every(o=>o))return null;let n=A.map((o,r)=>vFA(e,i[r],o.from,o.to));if(t!=2&&!n.every(o=>o))return{changes:e.changes(A.map((o,r)=>n[r]?[]:[{from:o.from,insert:i[r].open+" "},{from:o.to,insert:" "+i[r].close}]))};if(t!=1&&n.some(o=>o)){let o=[];for(let r=0,s;r<n.length;r++)if(s=n[r]){let a=i[r],{open:c,close:l}=s;o.push({from:c.pos-a.open.length,to:c.pos+c.margin},{from:l.pos-l.margin,to:l.pos+a.close.length})}return{changes:o}}return null}function MFA(t,e,A=e.selection.ranges){let i=[],n=-1;for(let{from:o,to:r}of A){let s=i.length,a=1e9,c=jF(e,o).line;if(c){for(let l=o;l<=r;){let I=e.doc.lineAt(l);if(I.from>n&&(o==r||r>I.from)){n=I.from;let C=/^\s*/.exec(I.text)[0].length,d=C==I.length,B=I.text.slice(C,C+c.length)==c?C:-1;C<I.text.length&&C<a&&(a=C),i.push({line:I,comment:B,token:c,indent:C,empty:d,single:!1})}l=I.to+1}if(a<1e9)for(let l=s;l<i.length;l++)i[l].indent<i[l].line.text.length&&(i[l].indent=a);i.length==s+1&&(i[s].single=!0)}}if(t!=2&&i.some(o=>o.comment<0&&(!o.empty||o.single))){let o=[];for(let{line:s,token:a,indent:c,empty:l,single:I}of i)(I||!l)&&o.push({from:s.from+c,insert:a+" "});let r=e.changes(o);return{changes:r,selection:e.selection.map(r,1)}}else if(t!=1&&i.some(o=>o.comment>=0)){let o=[];for(let{line:r,comment:s,token:a}of i)if(s>=0){let c=r.from+s,l=c+a.length;r.text[l-r.from]==" "&&l++,o.push({from:c,to:l})}return{changes:o}}return null}function zE(t,e){return ae.create(t.ranges.map(e),t.mainIndex)}function Kg(t,e){return t.update({selection:e,scrollIntoView:!0,userEvent:"select"})}function kl({state:t,dispatch:e},A){let i=zE(t.selection,A);return i.eq(t.selection,!0)?!1:(e(Kg(t,i)),!0)}function gD(t,e){return ae.cursor(e?t.to:t.from)}function tnA(t,e){return kl(t,A=>A.empty?t.moveByChar(A,e):gD(A,e))}function Ns(t){return t.textDirectionAt(t.state.selection.main.head)==lo.LTR}var inA=t=>tnA(t,!Ns(t)),nnA=t=>tnA(t,Ns(t));function onA(t,e){return kl(t,A=>A.empty?t.moveByGroup(A,e):gD(A,e))}var kFA=t=>onA(t,!Ns(t)),SFA=t=>onA(t,Ns(t));var n6e=typeof Intl<"u"&&Intl.Segmenter?new Intl.Segmenter(void 0,{granularity:"word"}):null;function RFA(t,e,A){if(e.type.prop(A))return!0;let i=e.to-e.from;return i&&(i>2||/[^\s,.;:]/.test(t.sliceDoc(e.from,e.to)))||e.firstChild}function ID(t,e,A){let i=$r(t).resolveInner(e.head),n=A?fi.closedBy:fi.openedBy;for(let a=e.head;;){let c=A?i.childAfter(a):i.childBefore(a);if(!c)break;RFA(t,c,n)?i=c:a=A?c.to:c.from}let o=i.type.prop(n),r,s;return o&&(r=A?bl(t,i.from,1):bl(t,i.to,-1))&&r.matched?s=A?r.end.to:r.end.from:s=A?i.to:i.from,ae.cursor(s,A?-1:1)}var xFA=t=>kl(t,e=>ID(t.state,e,!Ns(t))),LFA=t=>kl(t,e=>ID(t.state,e,Ns(t)));function rnA(t,e){return kl(t,A=>{if(!A.empty)return gD(A,e);let i=t.moveVertically(A,e);return i.head!=A.head?i:t.moveToLineBoundary(A,e)})}var snA=t=>rnA(t,!1),anA=t=>rnA(t,!0);function cnA(t){let e=t.scrollDOM.clientHeight<t.scrollDOM.scrollHeight-2,A=0,i=0,n;if(e){for(let o of t.state.facet(qt.scrollMargins)){let r=o(t);r?.top&&(A=Math.max(r?.top,A)),r?.bottom&&(i=Math.max(r?.bottom,i))}n=t.scrollDOM.clientHeight-A-i}else n=(t.dom.ownerDocument.defaultView||window).innerHeight;return{marginTop:A,marginBottom:i,selfScroll:e,height:Math.max(t.defaultLineHeight,n-5)}}function lnA(t,e){let A=cnA(t),{state:i}=t,n=zE(i.selection,r=>r.empty?t.moveVertically(r,e,A.height):gD(r,e));if(n.eq(i.selection))return!1;let o;if(A.selfScroll){let r=t.coordsAtPos(i.selection.main.head),s=t.scrollDOM.getBoundingClientRect(),a=s.top+A.marginTop,c=s.bottom-A.marginBottom;r&&r.top>a&&r.bottom<c&&(o=qt.scrollIntoView(n.main.head,{y:"start",yMargin:r.top-a}))}return t.dispatch(Kg(i,n),{effects:o}),!0}var jiA=t=>lnA(t,!1),zF=t=>lnA(t,!0);function S1(t,e,A){let i=t.lineBlockAt(e.head),n=t.moveToLineBoundary(e,A);if(n.head==e.head&&n.head!=(A?i.to:i.from)&&(n=t.moveToLineBoundary(e,A,!1)),!A&&n.head==i.from&&i.length){let o=/^\s*/.exec(t.state.sliceDoc(i.from,Math.min(i.from+100,i.to)))[0].length;o&&e.head!=i.from+o&&(n=ae.cursor(i.from+o))}return n}var FFA=t=>kl(t,e=>S1(t,e,!0)),NFA=t=>kl(t,e=>S1(t,e,!1)),_FA=t=>kl(t,e=>S1(t,e,!Ns(t))),GFA=t=>kl(t,e=>S1(t,e,Ns(t))),UFA=t=>kl(t,e=>ae.cursor(t.lineBlockAt(e.head).from,1)),KFA=t=>kl(t,e=>ae.cursor(t.lineBlockAt(e.head).to,-1));function YFA(t,e,A){let i=!1,n=zE(t.selection,o=>{let r=bl(t,o.head,-1)||bl(t,o.head,1)||o.head>0&&bl(t,o.head-1,1)||o.head<t.doc.length&&bl(t,o.head+1,-1);if(!r||!r.end)return o;i=!0;let s=r.start.from==o.head?r.end.to:r.end.from;return A?ae.range(o.anchor,s):ae.cursor(s)});return i?(e(Kg(t,n)),!0):!1}var JFA=({state:t,dispatch:e})=>YFA(t,e,!1);function Tc(t,e){let A=zE(t.state.selection,i=>{let n=e(i);return ae.range(i.anchor,n.head,n.goalColumn,n.bidiLevel||void 0)});return A.eq(t.state.selection)?!1:(t.dispatch(Kg(t.state,A)),!0)}function gnA(t,e){return Tc(t,A=>t.moveByChar(A,e))}var InA=t=>gnA(t,!Ns(t)),CnA=t=>gnA(t,Ns(t));function dnA(t,e){return Tc(t,A=>t.moveByGroup(A,e))}var TFA=t=>dnA(t,!Ns(t)),HFA=t=>dnA(t,Ns(t));var zFA=t=>Tc(t,e=>ID(t.state,e,!Ns(t))),OFA=t=>Tc(t,e=>ID(t.state,e,Ns(t)));function BnA(t,e){return Tc(t,A=>t.moveVertically(A,e))}var EnA=t=>BnA(t,!1),QnA=t=>BnA(t,!0);function hnA(t,e){return Tc(t,A=>t.moveVertically(A,e,cnA(t).height))}var qiA=t=>hnA(t,!1),ViA=t=>hnA(t,!0),PFA=t=>Tc(t,e=>S1(t,e,!0)),jFA=t=>Tc(t,e=>S1(t,e,!1)),qFA=t=>Tc(t,e=>S1(t,e,!Ns(t))),VFA=t=>Tc(t,e=>S1(t,e,Ns(t))),ZFA=t=>Tc(t,e=>ae.cursor(t.lineBlockAt(e.head).from)),WFA=t=>Tc(t,e=>ae.cursor(t.lineBlockAt(e.head).to)),ZiA=({state:t,dispatch:e})=>(e(Kg(t,{anchor:0})),!0),WiA=({state:t,dispatch:e})=>(e(Kg(t,{anchor:t.doc.length})),!0),XiA=({state:t,dispatch:e})=>(e(Kg(t,{anchor:t.selection.main.anchor,head:0})),!0),$iA=({state:t,dispatch:e})=>(e(Kg(t,{anchor:t.selection.main.anchor,head:t.doc.length})),!0),XFA=({state:t,dispatch:e})=>(e(t.update({selection:{anchor:0,head:t.doc.length},userEvent:"select"})),!0),$FA=({state:t,dispatch:e})=>{let A=CD(t).map(({from:i,to:n})=>ae.range(i,Math.min(n+1,t.doc.length)));return e(t.update({selection:ae.create(A),userEvent:"select"})),!0},ANA=({state:t,dispatch:e})=>{let A=zE(t.selection,i=>{let n=$r(t),o=n.resolveStack(i.from,1);if(i.empty){let r=n.resolveStack(i.from,-1);r.node.from>=o.node.from&&r.node.to<=o.node.to&&(o=r)}for(let r=o;r;r=r.next){let{node:s}=r;if((s.from<i.from&&s.to>=i.to||s.to>i.to&&s.from<=i.from)&&r.next)return ae.range(s.to,s.from)}return i});return A.eq(t.selection)?!1:(e(Kg(t,A)),!0)},eNA=({state:t,dispatch:e})=>{let A=t.selection,i=null;return A.ranges.length>1?i=ae.create([A.main]):A.main.empty||(i=ae.create([ae.cursor(A.main.head)])),i?(e(Kg(t,i)),!0):!1};function p3(t,e){if(t.state.readOnly)return!1;let A="delete.selection",{state:i}=t,n=i.changeByRange(o=>{let{from:r,to:s}=o;if(r==s){let a=e(o);a<r?(A="delete.backward",a=lD(t,a,!1)):a>r&&(A="delete.forward",a=lD(t,a,!0)),r=Math.min(r,a),s=Math.max(s,a)}else r=lD(t,r,!1),s=lD(t,s,!0);return r==s?{range:o}:{changes:{from:r,to:s},range:ae.cursor(r,r<o.head?-1:1)}});return n.changes.empty?!1:(t.dispatch(i.update(n,{scrollIntoView:!0,userEvent:A,effects:A=="delete.selection"?qt.announce.of(i.phrase("Selection deleted")):void 0})),!0)}function lD(t,e,A){if(t instanceof qt)for(let i of t.state.facet(qt.atomicRanges).map(n=>n(t)))i.between(e,e,(n,o)=>{n<e&&o>e&&(e=A?o:n)});return e}var unA=(t,e,A)=>p3(t,i=>{let n=i.from,{state:o}=t,r=o.doc.lineAt(n),s,a;if(A&&!e&&n>r.from&&n<r.from+200&&!/[^ \t]/.test(s=r.text.slice(0,n-r.from))){if(s[s.length-1]=="	")return n-1;let c=J0(s,o.tabSize),l=c%Ml(o)||Ml(o);for(let I=0;I<l&&s[s.length-1-I]==" ";I++)n--;a=n}else a=yr(r.text,n-r.from,e,e)+r.from,a==n&&r.number!=(e?o.doc.lines:1)?a+=e?1:-1:!e&&/[\ufe00-\ufe0f]/.test(r.text.slice(a-r.from,n-r.from))&&(a=yr(r.text,a-r.from,!1,!1)+r.from);return a}),OF=t=>unA(t,!1,!0);var fnA=t=>unA(t,!0,!1),mnA=(t,e)=>p3(t,A=>{let i=A.head,{state:n}=t,o=n.doc.lineAt(i),r=n.charCategorizer(i);for(let s=null;;){if(i==(e?o.to:o.from)){i==A.head&&o.number!=(e?n.doc.lines:1)&&(i+=e?1:-1);break}let a=yr(o.text,i-o.from,e)+o.from,c=o.text.slice(Math.min(i,a)-o.from,Math.max(i,a)-o.from),l=r(c);if(s!=null&&l!=s)break;(c!=" "||i!=A.head)&&(s=l),i=a}return i}),pnA=t=>mnA(t,!1),tNA=t=>mnA(t,!0),iNA=t=>p3(t,e=>{let A=t.lineBlockAt(e.head).to;return e.head<A?A:Math.min(t.state.doc.length,e.head+1)});var nNA=t=>p3(t,e=>{let A=t.moveToLineBoundary(e,!1).head;return e.head>A?A:Math.max(0,e.head-1)}),oNA=t=>p3(t,e=>{let A=t.moveToLineBoundary(e,!0).head;return e.head<A?A:Math.min(t.state.doc.length,e.head+1)});var rNA=({state:t,dispatch:e})=>{if(t.readOnly)return!1;let A=t.changeByRange(i=>({changes:{from:i.from,to:i.to,insert:In.of(["",""])},range:ae.cursor(i.from)}));return e(t.update(A,{scrollIntoView:!0,userEvent:"input"})),!0},sNA=({state:t,dispatch:e})=>{if(t.readOnly)return!1;let A=t.changeByRange(i=>{if(!i.empty||i.from==0||i.from==t.doc.length)return{range:i};let n=i.from,o=t.doc.lineAt(n),r=n==o.from?n-1:yr(o.text,n-o.from,!1)+o.from,s=n==o.to?n+1:yr(o.text,n-o.from,!0)+o.from;return{changes:{from:r,to:s,insert:t.doc.slice(n,s).append(t.doc.slice(r,n))},range:ae.cursor(s)}});return A.changes.empty?!1:(e(t.update(A,{scrollIntoView:!0,userEvent:"move.character"})),!0)};function CD(t){let e=[],A=-1;for(let i of t.selection.ranges){let n=t.doc.lineAt(i.from),o=t.doc.lineAt(i.to);if(!i.empty&&i.to==o.from&&(o=t.doc.lineAt(i.to-1)),A>=n.number){let r=e[e.length-1];r.to=o.to,r.ranges.push(i)}else e.push({from:n.from,to:o.to,ranges:[i]});A=o.number+1}return e}function wnA(t,e,A){if(t.readOnly)return!1;let i=[],n=[];for(let o of CD(t)){if(A?o.to==t.doc.length:o.from==0)continue;let r=t.doc.lineAt(A?o.to+1:o.from-1),s=r.length+1;if(A){i.push({from:o.to,to:r.to},{from:o.from,insert:r.text+t.lineBreak});for(let a of o.ranges)n.push(ae.range(Math.min(t.doc.length,a.anchor+s),Math.min(t.doc.length,a.head+s)))}else{i.push({from:r.from,to:o.from},{from:o.to,insert:t.lineBreak+r.text});for(let a of o.ranges)n.push(ae.range(a.anchor-s,a.head-s))}}return i.length?(e(t.update({changes:i,scrollIntoView:!0,selection:ae.create(n,t.selection.mainIndex),userEvent:"move.line"})),!0):!1}var aNA=({state:t,dispatch:e})=>wnA(t,e,!1),cNA=({state:t,dispatch:e})=>wnA(t,e,!0);function DnA(t,e,A){if(t.readOnly)return!1;let i=[];for(let n of CD(t))A?i.push({from:n.from,insert:t.doc.slice(n.from,n.to)+t.lineBreak}):i.push({from:n.to,insert:t.lineBreak+t.doc.slice(n.from,n.to)});return e(t.update({changes:i,scrollIntoView:!0,userEvent:"input.copyline"})),!0}var lNA=({state:t,dispatch:e})=>DnA(t,e,!1),gNA=({state:t,dispatch:e})=>DnA(t,e,!0),INA=t=>{if(t.state.readOnly)return!1;let{state:e}=t,A=e.changes(CD(e).map(({from:n,to:o})=>(n>0?n--:o<e.doc.length&&o++,{from:n,to:o}))),i=zE(e.selection,n=>{let o;if(t.lineWrapping){let r=t.lineBlockAt(n.head),s=t.coordsAtPos(n.head,n.assoc||1);s&&(o=r.bottom+t.documentTop-s.bottom+t.defaultLineHeight/2)}return t.moveVertically(n,!0,o)}).map(A);return t.dispatch({changes:A,selection:i,scrollIntoView:!0,userEvent:"delete.line"}),!0};function CNA(t,e){if(/\(\)|\[\]|\{\}/.test(t.sliceDoc(e-1,e+1)))return{from:e,to:e};let A=$r(t).resolveInner(e),i=A.childBefore(e),n=A.childAfter(e),o;return i&&n&&i.to<=e&&n.from>=e&&(o=i.type.prop(fi.closedBy))&&o.indexOf(n.name)>-1&&t.doc.lineAt(i.to).from==t.doc.lineAt(n.from).from&&!/\S/.test(t.sliceDoc(i.to,n.from))?{from:i.to,to:n.from}:null}var AnA=ynA(!1),dNA=ynA(!0);function ynA(t){return({state:e,dispatch:A})=>{if(e.readOnly)return!1;let i=e.changeByRange(n=>{let{from:o,to:r}=n,s=e.doc.lineAt(o),a=!t&&o==r&&CNA(e,o);t&&(o=r=(r<=s.to?s:e.doc.lineAt(r)).to);let c=new xC(e,{simulateBreak:o,simulateDoubleBreak:!!a}),l=aD(c,o);for(l==null&&(l=J0(/^\s*/.exec(e.doc.lineAt(o).text)[0],e.tabSize));r<s.to&&/\s/.test(s.text[r-s.from]);)r++;a?{from:o,to:r}=a:o>s.from&&o<s.from+100&&!/\S/.test(s.text.slice(0,o))&&(o=s.from);let I=["",HE(e,l)];return a&&I.push(HE(e,c.lineIndent(s.from,-1))),{changes:{from:o,to:r,insert:In.of(I)},range:ae.cursor(o+1+I[1].length)}});return A(e.update(i,{scrollIntoView:!0,userEvent:"input"})),!0}}function qF(t,e){let A=-1;return t.changeByRange(i=>{let n=[];for(let r=i.from;r<=i.to;){let s=t.doc.lineAt(r);s.number>A&&(i.empty||i.to>s.from)&&(e(s,n,i),A=s.number),r=s.to+1}let o=t.changes(n);return{changes:n,range:ae.range(o.mapPos(i.anchor,1),o.mapPos(i.head,1))}})}var BNA=({state:t,dispatch:e})=>{if(t.readOnly)return!1;let A=Object.create(null),i=new xC(t,{overrideIndentation:o=>{let r=A[o];return r??-1}}),n=qF(t,(o,r,s)=>{let a=aD(i,o.from);if(a==null)return;/\S/.test(o.text)||(a=0);let c=/^\s*/.exec(o.text)[0],l=HE(t,a);(c!=l||s.from<o.from+c.length)&&(A[o.from]=a,r.push({from:o.from,to:o.from+c.length,insert:l}))});return n.changes.empty||e(t.update(n,{userEvent:"indent"})),!0},vnA=({state:t,dispatch:e})=>t.readOnly?!1:(e(t.update(qF(t,(A,i)=>{i.push({from:A.from,insert:t.facet(FC)})}),{userEvent:"input.indent"})),!0),bnA=({state:t,dispatch:e})=>t.readOnly?!1:(e(t.update(qF(t,(A,i)=>{let n=/^\s*/.exec(A.text)[0];if(!n)return;let o=J0(n,t.tabSize),r=0,s=HE(t,Math.max(0,o-Ml(t)));for(;r<n.length&&r<s.length&&n.charCodeAt(r)==s.charCodeAt(r);)r++;i.push({from:A.from+r,to:A.from+n.length,insert:s.slice(r)})}),{userEvent:"delete.dedent"})),!0),ENA=t=>(t.setTabFocusMode(),!0);var QNA=[{key:"Ctrl-b",run:inA,shift:InA,preventDefault:!0},{key:"Ctrl-f",run:nnA,shift:CnA},{key:"Ctrl-p",run:snA,shift:EnA},{key:"Ctrl-n",run:anA,shift:QnA},{key:"Ctrl-a",run:UFA,shift:ZFA},{key:"Ctrl-e",run:KFA,shift:WFA},{key:"Ctrl-d",run:fnA},{key:"Ctrl-h",run:OF},{key:"Ctrl-k",run:iNA},{key:"Ctrl-Alt-h",run:pnA},{key:"Ctrl-o",run:rNA},{key:"Ctrl-t",run:sNA},{key:"Ctrl-v",run:zF}],hNA=[{key:"ArrowLeft",run:inA,shift:InA,preventDefault:!0},{key:"Mod-ArrowLeft",mac:"Alt-ArrowLeft",run:kFA,shift:TFA,preventDefault:!0},{mac:"Cmd-ArrowLeft",run:_FA,shift:qFA,preventDefault:!0},{key:"ArrowRight",run:nnA,shift:CnA,preventDefault:!0},{key:"Mod-ArrowRight",mac:"Alt-ArrowRight",run:SFA,shift:HFA,preventDefault:!0},{mac:"Cmd-ArrowRight",run:GFA,shift:VFA,preventDefault:!0},{key:"ArrowUp",run:snA,shift:EnA,preventDefault:!0},{mac:"Cmd-ArrowUp",run:ZiA,shift:XiA},{mac:"Ctrl-ArrowUp",run:jiA,shift:qiA},{key:"ArrowDown",run:anA,shift:QnA,preventDefault:!0},{mac:"Cmd-ArrowDown",run:WiA,shift:$iA},{mac:"Ctrl-ArrowDown",run:zF,shift:ViA},{key:"PageUp",run:jiA,shift:qiA},{key:"PageDown",run:zF,shift:ViA},{key:"Home",run:NFA,shift:jFA,preventDefault:!0},{key:"Mod-Home",run:ZiA,shift:XiA},{key:"End",run:FFA,shift:PFA,preventDefault:!0},{key:"Mod-End",run:WiA,shift:$iA},{key:"Enter",run:AnA,shift:AnA},{key:"Mod-a",run:XFA},{key:"Backspace",run:OF,shift:OF},{key:"Delete",run:fnA},{key:"Mod-Backspace",mac:"Alt-Backspace",run:pnA},{key:"Mod-Delete",mac:"Alt-Delete",run:tNA},{mac:"Mod-Backspace",run:nNA},{mac:"Mod-Delete",run:oNA}].concat(QNA.map(t=>({mac:t.key,run:t.run,shift:t.shift}))),MnA=[{key:"Alt-ArrowLeft",mac:"Ctrl-ArrowLeft",run:xFA,shift:zFA},{key:"Alt-ArrowRight",mac:"Ctrl-ArrowRight",run:LFA,shift:OFA},{key:"Alt-ArrowUp",run:aNA},{key:"Shift-Alt-ArrowUp",run:lNA},{key:"Alt-ArrowDown",run:cNA},{key:"Shift-Alt-ArrowDown",run:gNA},{key:"Escape",run:eNA},{key:"Mod-Enter",run:dNA},{key:"Alt-l",mac:"Ctrl-l",run:$FA},{key:"Mod-i",run:ANA,preventDefault:!0},{key:"Mod-[",run:bnA},{key:"Mod-]",run:vnA},{key:"Mod-Alt-\\",run:BNA},{key:"Shift-Mod-k",run:INA},{key:"Shift-Mod-\\",run:JFA},{key:"Mod-/",run:pFA},{key:"Alt-A",run:DFA},{key:"Ctrl-m",mac:"Shift-Alt-m",run:ENA}].concat(hNA),knA={key:"Tab",run:vnA,shift:bnA};var ED=class{constructor(e,A,i){this.from=e,this.to=A,this.diagnostic=i}},NC=class t{constructor(e,A,i){this.diagnostics=e,this.panel=A,this.selected=i}static init(e,A,i){let n=i.facet(Yg).markerFilter;n&&(e=n(e,i));let o=e.slice().sort((l,I)=>l.from-I.from||l.to-I.to),r=new ds,s=[],a=0;for(let l=0;;){let I=l==o.length?null:o[l];if(!I&&!s.length)break;let C,d;for(s.length?(C=a,d=s.reduce((E,h)=>Math.min(E,h.to),I&&I.from>C?I.from:1e8)):(C=I.from,d=I.to,s.push(I),l++);l<o.length;){let E=o[l];if(E.from==C&&(E.to>E.from||E.to==C))s.push(E),l++,d=Math.min(E.to,d);else{d=Math.min(E.from,d);break}}let B=KnA(s);if(s.some(E=>E.from==E.to||E.from==E.to-1&&i.doc.lineAt(E.from).to==E.from))r.add(C,C,ut.widget({widget:new VF(B),diagnostics:s.slice()}));else{let E=s.reduce((h,u)=>u.markClass?h+" "+u.markClass:h,"");r.add(C,d,ut.mark({class:"cm-lintRange cm-lintRange-"+B+E,diagnostics:s.slice(),inclusiveEnd:s.some(h=>h.to>d)}))}a=d;for(let E=0;E<s.length;E++)s[E].to<=a&&s.splice(E--,1)}let c=r.finish();return new t(c,A,OE(c))}};function OE(t,e=null,A=0){let i=null;return t.between(A,1e9,(n,o,{spec:r})=>{if(!(e&&r.diagnostics.indexOf(e)<0))if(!i)i=new ED(n,o,e||r.diagnostics[0]);else{if(r.diagnostics.indexOf(i.diagnostic)<0)return!1;i=new ED(i.from,o,i.diagnostic)}}),i}function RnA(t,e){let A=e.pos,i=e.end||A,n=t.state.facet(Yg).hideOn(t,A,i);if(n!=null)return n;let o=t.startState.doc.lineAt(e.pos);return!!(t.effects.some(r=>r.is(uD))||t.changes.touchesRange(o.from,Math.max(o.to,i)))}function xnA(t,e){return t.field(Cc,!1)?e:e.concat(Pi.appendConfig.of(JnA))}function uNA(t,e){return{effects:xnA(t,[uD.of(e)])}}var uD=Pi.define(),WF=Pi.define(),LnA=Pi.define(),Cc=Ar.define({create(){return new NC(ut.none,null,null)},update(t,e){if(e.docChanged&&t.diagnostics.size){let A=t.diagnostics.map(e.changes),i=null,n=t.panel;if(t.selected){let o=e.changes.mapPos(t.selected.from,1);i=OE(A,t.selected.diagnostic,o)||OE(A,null,o)}!A.size&&n&&e.state.facet(Yg).autoPanel&&(n=null),t=new NC(A,n,i)}for(let A of e.effects)if(A.is(uD)){let i=e.state.facet(Yg).autoPanel?A.value.length?w3.open:null:t.panel;t=NC.init(A.value,i,e.state)}else A.is(WF)?t=new NC(t.diagnostics,A.value?w3.open:null,t.selected):A.is(LnA)&&(t=new NC(t.diagnostics,t.panel,A.value));return t},provide:t=>[MC.from(t,e=>e.panel),qt.decorations.from(t,e=>e.diagnostics)]});var fNA=ut.mark({class:"cm-lintRange cm-lintRange-active"});function mNA(t,e,A){let{diagnostics:i}=t.state.field(Cc),n,o=-1,r=-1;i.between(e-(A<0?1:0),e+(A>0?1:0),(a,c,{spec:l})=>{if(e>=a&&e<=c&&(a==c||(e>a||A>0)&&(e<c||A<0)))return n=l.diagnostics,o=a,r=c,!1});let s=t.state.facet(Yg).tooltipFilter;return n&&s&&(n=s(n,t.state)),n?{pos:o,end:r,above:t.state.doc.lineAt(o).to<r,create(){return{dom:FnA(t,n)}}}:null}function FnA(t,e){return Un("ul",{class:"cm-tooltip-lint"},e.map(A=>UnA(t,A,!1)))}var pNA=t=>{let e=t.state.field(Cc,!1);(!e||!e.panel)&&t.dispatch({effects:xnA(t.state,[WF.of(!0)])});let A=kC(t,w3.open);return A&&A.dom.querySelector(".cm-panel-lint ul").focus(),!0},SnA=t=>{let e=t.state.field(Cc,!1);return!e||!e.panel?!1:(t.dispatch({effects:WF.of(!1)}),!0)},wNA=t=>{let e=t.state.field(Cc,!1);if(!e)return!1;let A=t.state.selection.main,i=e.diagnostics.iter(A.to+1);return!i.value&&(i=e.diagnostics.iter(0),!i.value||i.from==A.from&&i.to==A.to)?!1:(t.dispatch({selection:{anchor:i.from,head:i.to},scrollIntoView:!0}),!0)};var NnA=[{key:"Mod-Shift-m",run:pNA,preventDefault:!0},{key:"F8",run:wNA}],DNA=go.fromClass(class{constructor(t){this.view=t,this.timeout=-1,this.set=!0;let{delay:e}=t.state.facet(Yg);this.lintTime=Date.now()+e,this.run=this.run.bind(this),this.timeout=setTimeout(this.run,e)}run(){clearTimeout(this.timeout);let t=Date.now();if(t<this.lintTime-10)this.timeout=setTimeout(this.run,this.lintTime-t);else{this.set=!1;let{state:e}=this.view,{sources:A}=e.facet(Yg);A.length&&yNA(A.map(i=>Promise.resolve(i(this.view))),i=>{this.view.state.doc==e.doc&&this.view.dispatch(uNA(this.view.state,i.reduce((n,o)=>n.concat(o))))},i=>{Xr(this.view.state,i)})}}update(t){let e=t.state.facet(Yg);(t.docChanged||e!=t.startState.facet(Yg)||e.needsRefresh&&e.needsRefresh(t))&&(this.lintTime=Date.now()+e.delay,this.set||(this.set=!0,this.timeout=setTimeout(this.run,e.delay)))}force(){this.set&&(this.lintTime=Date.now(),this.run())}destroy(){clearTimeout(this.timeout)}});function yNA(t,e,A){let i=[],n=-1;for(let o of t)o.then(r=>{i.push(r),clearTimeout(n),i.length==t.length?e(i):n=setTimeout(()=>e(i),200)},A)}var Yg=qe.define({combine(t){return Object.assign({sources:t.map(e=>e.source).filter(e=>e!=null)},Wr(t.map(e=>e.config),{delay:750,markerFilter:null,tooltipFilter:null,needsRefresh:null,hideOn:()=>null},{needsRefresh:(e,A)=>e?A?i=>e(i)||A(i):e:A}))}});function _nA(t,e={}){return[Yg.of({source:t,config:e}),DNA,JnA]}function GnA(t){let e=[];if(t)A:for(let{name:A}of t){for(let i=0;i<A.length;i++){let n=A[i];if(/[a-zA-Z]/.test(n)&&!e.some(o=>o.toLowerCase()==n.toLowerCase())){e.push(n);continue A}}e.push("")}return e}function UnA(t,e,A){var i;let n=A?GnA(e.actions):[];return Un("li",{class:"cm-diagnostic cm-diagnostic-"+e.severity},Un("span",{class:"cm-diagnosticText"},e.renderMessage?e.renderMessage(t):e.message),(i=e.actions)===null||i===void 0?void 0:i.map((o,r)=>{let s=!1,a=C=>{if(C.preventDefault(),s)return;s=!0;let d=OE(t.state.field(Cc).diagnostics,e);d&&o.apply(t,d.from,d.to)},{name:c}=o,l=n[r]?c.indexOf(n[r]):-1,I=l<0?c:[c.slice(0,l),Un("u",c.slice(l,l+1)),c.slice(l+1)];return Un("button",{type:"button",class:"cm-diagnosticAction",onclick:a,onmousedown:a,"aria-label":` Action: ${c}${l<0?"":` (access key "${n[r]})"`}.`},I)}),e.source&&Un("div",{class:"cm-diagnosticSource"},e.source))}var VF=class extends Ic{constructor(e){super(),this.sev=e}eq(e){return e.sev==this.sev}toDOM(){return Un("span",{class:"cm-lintPoint cm-lintPoint-"+this.sev})}},QD=class{constructor(e,A){this.diagnostic=A,this.id="item_"+Math.floor(Math.random()*4294967295).toString(16),this.dom=UnA(e,A,!0),this.dom.id=this.id,this.dom.setAttribute("role","option")}},w3=class t{constructor(e){this.view=e,this.items=[];let A=n=>{if(n.keyCode==27)SnA(this.view),this.view.focus();else if(n.keyCode==38||n.keyCode==33)this.moveSelection((this.selectedIndex-1+this.items.length)%this.items.length);else if(n.keyCode==40||n.keyCode==34)this.moveSelection((this.selectedIndex+1)%this.items.length);else if(n.keyCode==36)this.moveSelection(0);else if(n.keyCode==35)this.moveSelection(this.items.length-1);else if(n.keyCode==13)this.view.focus();else if(n.keyCode>=65&&n.keyCode<=90&&this.selectedIndex>=0){let{diagnostic:o}=this.items[this.selectedIndex],r=GnA(o.actions);for(let s=0;s<r.length;s++)if(r[s].toUpperCase().charCodeAt(0)==n.keyCode){let a=OE(this.view.state.field(Cc).diagnostics,o);a&&o.actions[s].apply(e,a.from,a.to)}}else return;n.preventDefault()},i=n=>{for(let o=0;o<this.items.length;o++)this.items[o].dom.contains(n.target)&&this.moveSelection(o)};this.list=Un("ul",{tabIndex:0,role:"listbox","aria-label":this.view.state.phrase("Diagnostics"),onkeydown:A,onclick:i}),this.dom=Un("div",{class:"cm-panel-lint"},this.list,Un("button",{type:"button",name:"close","aria-label":this.view.state.phrase("close"),onclick:()=>SnA(this.view)},"\xD7")),this.update()}get selectedIndex(){let e=this.view.state.field(Cc).selected;if(!e)return-1;for(let A=0;A<this.items.length;A++)if(this.items[A].diagnostic==e.diagnostic)return A;return-1}update(){let{diagnostics:e,selected:A}=this.view.state.field(Cc),i=0,n=!1,o=null,r=new Set;for(e.between(0,this.view.state.doc.length,(s,a,{spec:c})=>{for(let l of c.diagnostics){if(r.has(l))continue;r.add(l);let I=-1,C;for(let d=i;d<this.items.length;d++)if(this.items[d].diagnostic==l){I=d;break}I<0?(C=new QD(this.view,l),this.items.splice(i,0,C),n=!0):(C=this.items[I],I>i&&(this.items.splice(i,I-i),n=!0)),A&&C.diagnostic==A.diagnostic?C.dom.hasAttribute("aria-selected")||(C.dom.setAttribute("aria-selected","true"),o=C):C.dom.hasAttribute("aria-selected")&&C.dom.removeAttribute("aria-selected"),i++}});i<this.items.length&&!(this.items.length==1&&this.items[0].diagnostic.from<0);)n=!0,this.items.pop();this.items.length==0&&(this.items.push(new QD(this.view,{from:-1,to:-1,severity:"info",message:this.view.state.phrase("No diagnostics")})),n=!0),o?(this.list.setAttribute("aria-activedescendant",o.id),this.view.requestMeasure({key:this,read:()=>({sel:o.dom.getBoundingClientRect(),panel:this.list.getBoundingClientRect()}),write:({sel:s,panel:a})=>{let c=a.height/this.list.offsetHeight;s.top<a.top?this.list.scrollTop-=(a.top-s.top)/c:s.bottom>a.bottom&&(this.list.scrollTop+=(s.bottom-a.bottom)/c)}})):this.selectedIndex<0&&this.list.removeAttribute("aria-activedescendant"),n&&this.sync()}sync(){let e=this.list.firstChild;function A(){let i=e;e=i.nextSibling,i.remove()}for(let i of this.items)if(i.dom.parentNode==this.list){for(;e!=i.dom;)A();e=i.dom.nextSibling}else this.list.insertBefore(i.dom,e);for(;e;)A()}moveSelection(e){if(this.selectedIndex<0)return;let A=this.view.state.field(Cc),i=OE(A.diagnostics,this.items[e].diagnostic);i&&this.view.dispatch({selection:{anchor:i.from,head:i.to},scrollIntoView:!0,effects:LnA.of(i)})}static open(e){return new t(e)}};function BD(t,e='viewBox="0 0 40 40"'){return`url('data:image/svg+xml,<svg xmlns="http://www.w3.org/2000/svg" ${e}>${encodeURIComponent(t)}</svg>')`}function dD(t){return BD(`<path d="m0 2.5 l2 -1.5 l1 0 l2 1.5 l1 0" stroke="${t}" fill="none" stroke-width=".7"/>`,'width="6" height="3"')}var vNA=qt.baseTheme({".cm-diagnostic":{padding:"3px 6px 3px 8px",marginLeft:"-1px",display:"block",whiteSpace:"pre-wrap"},".cm-diagnostic-error":{borderLeft:"5px solid #d11"},".cm-diagnostic-warning":{borderLeft:"5px solid orange"},".cm-diagnostic-info":{borderLeft:"5px solid #999"},".cm-diagnostic-hint":{borderLeft:"5px solid #66d"},".cm-diagnosticAction":{font:"inherit",border:"none",padding:"2px 4px",backgroundColor:"#444",color:"white",borderRadius:"3px",marginLeft:"8px",cursor:"pointer"},".cm-diagnosticSource":{fontSize:"70%",opacity:.7},".cm-lintRange":{backgroundPosition:"left bottom",backgroundRepeat:"repeat-x",paddingBottom:"0.7px"},".cm-lintRange-error":{backgroundImage:dD("#d11")},".cm-lintRange-warning":{backgroundImage:dD("orange")},".cm-lintRange-info":{backgroundImage:dD("#999")},".cm-lintRange-hint":{backgroundImage:dD("#66d")},".cm-lintRange-active":{backgroundColor:"#ffdd9980"},".cm-tooltip-lint":{padding:0,margin:0},".cm-lintPoint":{position:"relative","&:after":{content:'""',position:"absolute",bottom:0,left:"-2px",borderLeft:"3px solid transparent",borderRight:"3px solid transparent",borderBottom:"4px solid #d11"}},".cm-lintPoint-warning":{"&:after":{borderBottomColor:"orange"}},".cm-lintPoint-info":{"&:after":{borderBottomColor:"#999"}},".cm-lintPoint-hint":{"&:after":{borderBottomColor:"#66d"}},".cm-panel.cm-panel-lint":{position:"relative","& ul":{maxHeight:"100px",overflowY:"auto","& [aria-selected]":{backgroundColor:"#ddd","& u":{textDecoration:"underline"}},"&:focus [aria-selected]":{background_fallback:"#bdf",backgroundColor:"Highlight",color_fallback:"white",color:"HighlightText"},"& u":{textDecoration:"none"},padding:0,margin:0},"& [name=close]":{position:"absolute",top:"0",right:"2px",background:"inherit",border:"none",font:"inherit",padding:0,margin:0}}});function bNA(t){return t=="error"?4:t=="warning"?3:t=="info"?2:1}function KnA(t){let e="hint",A=1;for(let i of t){let n=bNA(i.severity);n>A&&(A=n,e=i.severity)}return e}var hD=class extends Fa{constructor(e){super(),this.diagnostics=e,this.severity=KnA(e)}toDOM(e){let A=document.createElement("div");A.className="cm-lint-marker cm-lint-marker-"+this.severity;let i=this.diagnostics,n=e.state.facet(fD).tooltipFilter;return n&&(i=n(i,e.state)),i.length&&(A.onmouseover=()=>kNA(e,A,i)),A}};function MNA(t,e){let A=i=>{let n=e.getBoundingClientRect();if(!(i.clientX>n.left-10&&i.clientX<n.right+10&&i.clientY>n.top-10&&i.clientY<n.bottom+10)){for(let o=i.target;o;o=o.parentNode)if(o.nodeType==1&&o.classList.contains("cm-tooltip-lint"))return;window.removeEventListener("mousemove",A),t.state.field(YnA)&&t.dispatch({effects:XF.of(null)})}};window.addEventListener("mousemove",A)}function kNA(t,e,A){function i(){let r=t.elementAtHeight(e.getBoundingClientRect().top+5-t.documentTop);t.coordsAtPos(r.from)&&t.dispatch({effects:XF.of({pos:r.from,above:!1,clip:!1,create(){return{dom:FnA(t,A),getCoords:()=>e.getBoundingClientRect()}}})}),e.onmouseout=e.onmousemove=null,MNA(t,e)}let{hoverTime:n}=t.state.facet(fD),o=setTimeout(i,n);e.onmouseout=()=>{clearTimeout(o),e.onmouseout=e.onmousemove=null},e.onmousemove=()=>{clearTimeout(o),o=setTimeout(i,n)}}function SNA(t,e){let A=Object.create(null);for(let n of e){let o=t.lineAt(n.from);(A[o.from]||(A[o.from]=[])).push(n)}let i=[];for(let n in A)i.push(new hD(A[n]).range(+n));return co.of(i,!0)}var RNA=Vw({class:"cm-gutter-lint",markers:t=>t.state.field(ZF),widgetMarker:(t,e,A)=>{let i=[];return t.state.field(ZF).between(A.from,A.to,(n,o,r)=>{n>A.from&&n<A.to&&i.push(...r.diagnostics)}),i.length?new hD(i):null}}),ZF=Ar.define({create(){return co.empty},update(t,e){t=t.map(e.changes);let A=e.state.facet(fD).markerFilter;for(let i of e.effects)if(i.is(uD)){let n=i.value;A&&(n=A(n||[],e.state)),t=SNA(e.state.doc,n.slice(0))}return t}}),XF=Pi.define(),YnA=Ar.define({create(){return null},update(t,e){return t&&e.docChanged&&(t=RnA(e,t)?null:Object.assign(Object.assign({},t),{pos:e.changes.mapPos(t.pos)})),e.effects.reduce((A,i)=>i.is(XF)?i.value:A,t)},provide:t=>GE.from(t)}),xNA=qt.baseTheme({".cm-gutter-lint":{width:"1.4em","& .cm-gutterElement":{padding:".2em"}},".cm-lint-marker":{width:"1em",height:"1em"},".cm-lint-marker-info":{content:BD('<path fill="#aaf" stroke="#77e" stroke-width="6" stroke-linejoin="round" d="M5 5L35 5L35 35L5 35Z"/>')},".cm-lint-marker-warning":{content:BD('<path fill="#fe8" stroke="#fd7" stroke-width="6" stroke-linejoin="round" d="M20 6L37 35L3 35Z"/>')},".cm-lint-marker-error":{content:BD('<circle cx="20" cy="20" r="15" fill="#f87" stroke="#f43" stroke-width="6"/>')}}),JnA=[Cc,qt.decorations.compute([Cc],t=>{let{selected:e,panel:A}=t.field(Cc);return!e||!A||e.from==e.to?ut.none:ut.set([fNA.range(e.from,e.to)])}),aiA(mNA,{hideOn:RnA}),vNA],fD=qe.define({combine(t){return Wr(t,{hoverTime:300,markerFilter:null,tooltipFilter:null})}});function TnA(t={}){return[fD.of(t),ZF,RNA,xNA,YnA]}var AN=class t{constructor(e,A,i,n,o,r,s,a,c,l=0,I){this.p=e,this.stack=A,this.state=i,this.reducePos=n,this.pos=o,this.score=r,this.buffer=s,this.bufferBase=a,this.curContext=c,this.lookAhead=l,this.parent=I}toString(){return`[${this.stack.filter((e,A)=>A%3==0).concat(this.state)}]@${this.pos}${this.score?"!"+this.score:""}`}static start(e,A,i=0){let n=e.parser.context;return new t(e,[],A,i,i,0,[],0,n?new mD(n,n.start):null,0,null)}get context(){return this.curContext?this.curContext.context:null}pushState(e,A){this.stack.push(this.state,A,this.bufferBase+this.buffer.length),this.state=e}reduce(e){var A;let i=e>>19,n=e&65535,{parser:o}=this.p,r=this.reducePos<this.pos-25;r&&this.setLookAhead(this.pos);let s=o.dynamicPrecedence(n);if(s&&(this.score+=s),i==0){this.pushState(o.getGoto(this.state,n,!0),this.reducePos),n<o.minRepeatTerm&&this.storeNode(n,this.reducePos,this.reducePos,r?8:4,!0),this.reduceContext(n,this.reducePos);return}let a=this.stack.length-(i-1)*3-(e&262144?6:0),c=a?this.stack[a-2]:this.p.ranges[0].from,l=this.reducePos-c;l>=2e3&&!(!((A=this.p.parser.nodeSet.types[n])===null||A===void 0)&&A.isAnonymous)&&(c==this.p.lastBigReductionStart?(this.p.bigReductionCount++,this.p.lastBigReductionSize=l):this.p.lastBigReductionSize<l&&(this.p.bigReductionCount=1,this.p.lastBigReductionStart=c,this.p.lastBigReductionSize=l));let I=a?this.stack[a-1]:0,C=this.bufferBase+this.buffer.length-I;if(n<o.minRepeatTerm||e&131072){let d=o.stateFlag(this.state,1)?this.pos:this.reducePos;this.storeNode(n,c,d,C+4,!0)}if(e&262144)this.state=this.stack[a];else{let d=this.stack[a-3];this.state=o.getGoto(d,n,!0)}for(;this.stack.length>a;)this.stack.pop();this.reduceContext(n,c)}storeNode(e,A,i,n=4,o=!1){if(e==0&&(!this.stack.length||this.stack[this.stack.length-1]<this.buffer.length+this.bufferBase)){let r=this,s=this.buffer.length;if(s==0&&r.parent&&(s=r.bufferBase-r.parent.bufferBase,r=r.parent),s>0&&r.buffer[s-4]==0&&r.buffer[s-1]>-1){if(A==i)return;if(r.buffer[s-2]>=A){r.buffer[s-2]=i;return}}}if(!o||this.pos==i)this.buffer.push(e,A,i,n);else{let r=this.buffer.length;if(r>0&&this.buffer[r-4]!=0){let s=!1;for(let a=r;a>0&&this.buffer[a-2]>i;a-=4)if(this.buffer[a-1]>=0){s=!0;break}if(s)for(;r>0&&this.buffer[r-2]>i;)this.buffer[r]=this.buffer[r-4],this.buffer[r+1]=this.buffer[r-3],this.buffer[r+2]=this.buffer[r-2],this.buffer[r+3]=this.buffer[r-1],r-=4,n>4&&(n-=4)}this.buffer[r]=e,this.buffer[r+1]=A,this.buffer[r+2]=i,this.buffer[r+3]=n}}shift(e,A,i,n){if(e&131072)this.pushState(e&65535,this.pos);else if((e&262144)==0){let o=e,{parser:r}=this.p;(n>this.pos||A<=r.maxNode)&&(this.pos=n,r.stateFlag(o,1)||(this.reducePos=n)),this.pushState(o,i),this.shiftContext(A,i),A<=r.maxNode&&this.buffer.push(A,i,n,4)}else this.pos=n,this.shiftContext(A,i),A<=this.p.parser.maxNode&&this.buffer.push(A,i,n,4)}apply(e,A,i,n){e&65536?this.reduce(e):this.shift(e,A,i,n)}useNode(e,A){let i=this.p.reused.length-1;(i<0||this.p.reused[i]!=e)&&(this.p.reused.push(e),i++);let n=this.pos;this.reducePos=this.pos=n+e.length,this.pushState(A,n),this.buffer.push(i,n,this.reducePos,-1),this.curContext&&this.updateContext(this.curContext.tracker.reuse(this.curContext.context,e,this,this.p.stream.reset(this.pos-e.length)))}split(){let e=this,A=e.buffer.length;for(;A>0&&e.buffer[A-2]>e.reducePos;)A-=4;let i=e.buffer.slice(A),n=e.bufferBase+A;for(;e&&n==e.bufferBase;)e=e.parent;return new t(this.p,this.stack.slice(),this.state,this.reducePos,this.pos,this.score,i,n,this.curContext,this.lookAhead,e)}recoverByDelete(e,A){let i=e<=this.p.parser.maxNode;i&&this.storeNode(e,this.pos,A,4),this.storeNode(0,this.pos,A,i?8:4),this.pos=this.reducePos=A,this.score-=190}canShift(e){for(let A=new eN(this);;){let i=this.p.parser.stateSlot(A.state,4)||this.p.parser.hasAction(A.state,e);if(i==0)return!1;if((i&65536)==0)return!0;A.reduce(i)}}recoverByInsert(e){if(this.stack.length>=300)return[];let A=this.p.parser.nextStates(this.state);if(A.length>8||this.stack.length>=120){let n=[];for(let o=0,r;o<A.length;o+=2)(r=A[o+1])!=this.state&&this.p.parser.hasAction(r,e)&&n.push(A[o],r);if(this.stack.length<120)for(let o=0;n.length<8&&o<A.length;o+=2){let r=A[o+1];n.some((s,a)=>a&1&&s==r)||n.push(A[o],r)}A=n}let i=[];for(let n=0;n<A.length&&i.length<4;n+=2){let o=A[n+1];if(o==this.state)continue;let r=this.split();r.pushState(o,this.pos),r.storeNode(0,r.pos,r.pos,4,!0),r.shiftContext(A[n],this.pos),r.reducePos=this.pos,r.score-=200,i.push(r)}return i}forceReduce(){let{parser:e}=this.p,A=e.stateSlot(this.state,5);if((A&65536)==0)return!1;if(!e.validAction(this.state,A)){let i=A>>19,n=A&65535,o=this.stack.length-i*3;if(o<0||e.getGoto(this.stack[o],n,!1)<0){let r=this.findForcedReduction();if(r==null)return!1;A=r}this.storeNode(0,this.pos,this.pos,4,!0),this.score-=100}return this.reducePos=this.pos,this.reduce(A),!0}findForcedReduction(){let{parser:e}=this.p,A=[],i=(n,o)=>{if(!A.includes(n))return A.push(n),e.allActions(n,r=>{if(!(r&393216))if(r&65536){let s=(r>>19)-o;if(s>1){let a=r&65535,c=this.stack.length-s*3;if(c>=0&&e.getGoto(this.stack[c],a,!1)>=0)return s<<19|65536|a}}else{let s=i(r,o+1);if(s!=null)return s}})};return i(this.state,0)}forceAll(){for(;!this.p.parser.stateFlag(this.state,2);)if(!this.forceReduce()){this.storeNode(0,this.pos,this.pos,4,!0);break}return this}get deadEnd(){if(this.stack.length!=3)return!1;let{parser:e}=this.p;return e.data[e.stateSlot(this.state,1)]==65535&&!e.stateSlot(this.state,4)}restart(){this.storeNode(0,this.pos,this.pos,4,!0),this.state=this.stack[0],this.stack.length=0}sameState(e){if(this.state!=e.state||this.stack.length!=e.stack.length)return!1;for(let A=0;A<this.stack.length;A+=3)if(this.stack[A]!=e.stack[A])return!1;return!0}get parser(){return this.p.parser}dialectEnabled(e){return this.p.parser.dialect.flags[e]}shiftContext(e,A){this.curContext&&this.updateContext(this.curContext.tracker.shift(this.curContext.context,e,this,this.p.stream.reset(A)))}reduceContext(e,A){this.curContext&&this.updateContext(this.curContext.tracker.reduce(this.curContext.context,e,this,this.p.stream.reset(A)))}emitContext(){let e=this.buffer.length-1;(e<0||this.buffer[e]!=-3)&&this.buffer.push(this.curContext.hash,this.pos,this.pos,-3)}emitLookAhead(){let e=this.buffer.length-1;(e<0||this.buffer[e]!=-4)&&this.buffer.push(this.lookAhead,this.pos,this.pos,-4)}updateContext(e){if(e!=this.curContext.context){let A=new mD(this.curContext.tracker,e);A.hash!=this.curContext.hash&&this.emitContext(),this.curContext=A}}setLookAhead(e){e>this.lookAhead&&(this.emitLookAhead(),this.lookAhead=e)}close(){this.curContext&&this.curContext.tracker.strict&&this.emitContext(),this.lookAhead>0&&this.emitLookAhead()}},mD=class{constructor(e,A){this.tracker=e,this.context=A,this.hash=e.strict?e.hash(A):0}},eN=class{constructor(e){this.start=e,this.state=e.state,this.stack=e.stack,this.base=this.stack.length}reduce(e){let A=e&65535,i=e>>19;i==0?(this.stack==this.start.stack&&(this.stack=this.stack.slice()),this.stack.push(this.state,0,0),this.base+=3):this.base-=(i-1)*3;let n=this.start.p.parser.getGoto(this.stack[this.base-3],A,!0);this.state=n}},tN=class t{constructor(e,A,i){this.stack=e,this.pos=A,this.index=i,this.buffer=e.buffer,this.index==0&&this.maybeNext()}static create(e,A=e.bufferBase+e.buffer.length){return new t(e,A,A-e.bufferBase)}maybeNext(){let e=this.stack.parent;e!=null&&(this.index=this.stack.bufferBase-e.bufferBase,this.stack=e,this.buffer=e.buffer)}get id(){return this.buffer[this.index-4]}get start(){return this.buffer[this.index-3]}get end(){return this.buffer[this.index-2]}get size(){return this.buffer[this.index-1]}next(){this.index-=4,this.pos-=4,this.index==0&&this.maybeNext()}fork(){return new t(this.stack,this.pos,this.index)}};function D3(t,e=Uint16Array){if(typeof t!="string")return t;let A=null;for(let i=0,n=0;i<t.length;){let o=0;for(;;){let r=t.charCodeAt(i++),s=!1;if(r==126){o=65535;break}r>=92&&r--,r>=34&&r--;let a=r-32;if(a>=46&&(a-=46,s=!0),o+=a,s)break;o*=46}A?A[n++]=o:A=new e(o)}return A}var PE=class{constructor(){this.start=-1,this.value=-1,this.end=-1,this.extended=-1,this.lookAhead=0,this.mask=0,this.context=0}},HnA=new PE,iN=class{constructor(e,A){this.input=e,this.ranges=A,this.chunk="",this.chunkOff=0,this.chunk2="",this.chunk2Pos=0,this.next=-1,this.token=HnA,this.rangeIndex=0,this.pos=this.chunkPos=A[0].from,this.range=A[0],this.end=A[A.length-1].to,this.readNext()}resolveOffset(e,A){let i=this.range,n=this.rangeIndex,o=this.pos+e;for(;o<i.from;){if(!n)return null;let r=this.ranges[--n];o-=i.from-r.to,i=r}for(;A<0?o>i.to:o>=i.to;){if(n==this.ranges.length-1)return null;let r=this.ranges[++n];o+=r.from-i.to,i=r}return o}clipPos(e){if(e>=this.range.from&&e<this.range.to)return e;for(let A of this.ranges)if(A.to>e)return Math.max(e,A.from);return this.end}peek(e){let A=this.chunkOff+e,i,n;if(A>=0&&A<this.chunk.length)i=this.pos+e,n=this.chunk.charCodeAt(A);else{let o=this.resolveOffset(e,1);if(o==null)return-1;if(i=o,i>=this.chunk2Pos&&i<this.chunk2Pos+this.chunk2.length)n=this.chunk2.charCodeAt(i-this.chunk2Pos);else{let r=this.rangeIndex,s=this.range;for(;s.to<=i;)s=this.ranges[++r];this.chunk2=this.input.chunk(this.chunk2Pos=i),i+this.chunk2.length>s.to&&(this.chunk2=this.chunk2.slice(0,s.to-i)),n=this.chunk2.charCodeAt(0)}}return i>=this.token.lookAhead&&(this.token.lookAhead=i+1),n}acceptToken(e,A=0){let i=A?this.resolveOffset(A,-1):this.pos;if(i==null||i<this.token.start)throw new RangeError("Token end out of bounds");this.token.value=e,this.token.end=i}acceptTokenTo(e,A){this.token.value=e,this.token.end=A}getChunk(){if(this.pos>=this.chunk2Pos&&this.pos<this.chunk2Pos+this.chunk2.length){let{chunk:e,chunkPos:A}=this;this.chunk=this.chunk2,this.chunkPos=this.chunk2Pos,this.chunk2=e,this.chunk2Pos=A,this.chunkOff=this.pos-this.chunkPos}else{this.chunk2=this.chunk,this.chunk2Pos=this.chunkPos;let e=this.input.chunk(this.pos),A=this.pos+e.length;this.chunk=A>this.range.to?e.slice(0,this.range.to-this.pos):e,this.chunkPos=this.pos,this.chunkOff=0}}readNext(){return this.chunkOff>=this.chunk.length&&(this.getChunk(),this.chunkOff==this.chunk.length)?this.next=-1:this.next=this.chunk.charCodeAt(this.chunkOff)}advance(e=1){for(this.chunkOff+=e;this.pos+e>=this.range.to;){if(this.rangeIndex==this.ranges.length-1)return this.setDone();e-=this.range.to-this.pos,this.range=this.ranges[++this.rangeIndex],this.pos=this.range.from}return this.pos+=e,this.pos>=this.token.lookAhead&&(this.token.lookAhead=this.pos+1),this.readNext()}setDone(){return this.pos=this.chunkPos=this.end,this.range=this.ranges[this.rangeIndex=this.ranges.length-1],this.chunk="",this.next=-1}reset(e,A){if(A?(this.token=A,A.start=e,A.lookAhead=e+1,A.value=A.extended=-1):this.token=HnA,this.pos!=e){if(this.pos=e,e==this.end)return this.setDone(),this;for(;e<this.range.from;)this.range=this.ranges[--this.rangeIndex];for(;e>=this.range.to;)this.range=this.ranges[++this.rangeIndex];e>=this.chunkPos&&e<this.chunkPos+this.chunk.length?this.chunkOff=e-this.chunkPos:(this.chunk="",this.chunkOff=0),this.readNext()}return this}read(e,A){if(e>=this.chunkPos&&A<=this.chunkPos+this.chunk.length)return this.chunk.slice(e-this.chunkPos,A-this.chunkPos);if(e>=this.chunk2Pos&&A<=this.chunk2Pos+this.chunk2.length)return this.chunk2.slice(e-this.chunk2Pos,A-this.chunk2Pos);if(e>=this.range.from&&A<=this.range.to)return this.input.read(e,A);let i="";for(let n of this.ranges){if(n.from>=A)break;n.to>e&&(i+=this.input.read(Math.max(n.from,e),Math.min(n.to,A)))}return i}},R1=class{constructor(e,A){this.data=e,this.id=A}token(e,A){let{parser:i}=A.p;qnA(this.data,e,A,this.id,i.data,i.tokenPrecTable)}};R1.prototype.contextual=R1.prototype.fallback=R1.prototype.extend=!1;var nN=class{constructor(e,A,i){this.precTable=A,this.elseToken=i,this.data=typeof e=="string"?D3(e):e}token(e,A){let i=e.pos,n=0;for(;;){let o=e.next<0,r=e.resolveOffset(1,1);if(qnA(this.data,e,A,0,this.data,this.precTable),e.token.value>-1)break;if(this.elseToken==null)return;if(o||n++,r==null)break;e.reset(r,e.token)}n&&(e.reset(i,e.token),e.acceptToken(this.elseToken,n))}};nN.prototype.contextual=R1.prototype.fallback=R1.prototype.extend=!1;function qnA(t,e,A,i,n,o){let r=0,s=1<<i,{dialect:a}=A.p.parser;A:for(;(s&t[r])!=0;){let c=t[r+1];for(let d=r+3;d<c;d+=2)if((t[d+1]&s)>0){let B=t[d];if(a.allows(B)&&(e.token.value==-1||e.token.value==B||FNA(B,e.token.value,n,o))){e.acceptToken(B);break}}let l=e.next,I=0,C=t[r+2];if(e.next<0&&C>I&&t[c+C*3-3]==65535){r=t[c+C*3-1];continue A}for(;I<C;){let d=I+C>>1,B=c+d+(d<<1),E=t[B],h=t[B+1]||65536;if(l<E)C=d;else if(l>=h)I=d+1;else{r=t[B+2],e.advance();continue A}}break}}function znA(t,e,A){for(let i=e,n;(n=t[i])!=65535;i++)if(n==A)return i-e;return-1}function FNA(t,e,A,i){let n=znA(A,i,e);return n<0||znA(A,i,t)<n}var dc=typeof process<"u"&&process.env&&/\bparse\b/.test(process.env.LOG),$F=null;function OnA(t,e,A){let i=t.cursor(Jr.IncludeAnonymous);for(i.moveTo(e);;)if(!(A<0?i.childBefore(e):i.childAfter(e)))for(;;){if((A<0?i.to<e:i.from>e)&&!i.type.isError)return A<0?Math.max(0,Math.min(i.to-1,e-25)):Math.min(t.length,Math.max(i.from+1,e+25));if(A<0?i.prevSibling():i.nextSibling())break;if(!i.parent())return A<0?0:t.length}}var oN=class{constructor(e,A){this.fragments=e,this.nodeSet=A,this.i=0,this.fragment=null,this.safeFrom=-1,this.safeTo=-1,this.trees=[],this.start=[],this.index=[],this.nextFragment()}nextFragment(){let e=this.fragment=this.i==this.fragments.length?null:this.fragments[this.i++];if(e){for(this.safeFrom=e.openStart?OnA(e.tree,e.from+e.offset,1)-e.offset:e.from,this.safeTo=e.openEnd?OnA(e.tree,e.to+e.offset,-1)-e.offset:e.to;this.trees.length;)this.trees.pop(),this.start.pop(),this.index.pop();this.trees.push(e.tree),this.start.push(-e.offset),this.index.push(0),this.nextStart=this.safeFrom}else this.nextStart=1e9}nodeAt(e){if(e<this.nextStart)return null;for(;this.fragment&&this.safeTo<=e;)this.nextFragment();if(!this.fragment)return null;for(;;){let A=this.trees.length-1;if(A<0)return this.nextFragment(),null;let i=this.trees[A],n=this.index[A];if(n==i.children.length){this.trees.pop(),this.start.pop(),this.index.pop();continue}let o=i.children[n],r=this.start[A]+i.positions[n];if(r>e)return this.nextStart=r,null;if(o instanceof ur){if(r==e){if(r<this.safeFrom)return null;let s=r+o.length;if(s<=this.safeTo){let a=o.prop(fi.lookAhead);if(!a||s+a<this.fragment.to)return o}}this.index[A]++,r+o.length>=Math.max(this.safeFrom,e)&&(this.trees.push(o),this.start.push(r),this.index.push(0))}else this.index[A]++,this.nextStart=r+o.length}}},rN=class{constructor(e,A){this.stream=A,this.tokens=[],this.mainToken=null,this.actions=[],this.tokens=e.tokenizers.map(i=>new PE)}getActions(e){let A=0,i=null,{parser:n}=e.p,{tokenizers:o}=n,r=n.stateSlot(e.state,3),s=e.curContext?e.curContext.hash:0,a=0;for(let c=0;c<o.length;c++){if((1<<c&r)==0)continue;let l=o[c],I=this.tokens[c];if(!(i&&!l.fallback)&&((l.contextual||I.start!=e.pos||I.mask!=r||I.context!=s)&&(this.updateCachedToken(I,l,e),I.mask=r,I.context=s),I.lookAhead>I.end+25&&(a=Math.max(I.lookAhead,a)),I.value!=0)){let C=A;if(I.extended>-1&&(A=this.addActions(e,I.extended,I.end,A)),A=this.addActions(e,I.value,I.end,A),!l.extend&&(i=I,A>C))break}}for(;this.actions.length>A;)this.actions.pop();return a&&e.setLookAhead(a),!i&&e.pos==this.stream.end&&(i=new PE,i.value=e.p.parser.eofTerm,i.start=i.end=e.pos,A=this.addActions(e,i.value,i.end,A)),this.mainToken=i,this.actions}getMainToken(e){if(this.mainToken)return this.mainToken;let A=new PE,{pos:i,p:n}=e;return A.start=i,A.end=Math.min(i+1,n.stream.end),A.value=i==n.stream.end?n.parser.eofTerm:0,A}updateCachedToken(e,A,i){let n=this.stream.clipPos(i.pos);if(A.token(this.stream.reset(n,e),i),e.value>-1){let{parser:o}=i.p;for(let r=0;r<o.specialized.length;r++)if(o.specialized[r]==e.value){let s=o.specializers[r](this.stream.read(e.start,e.end),i);if(s>=0&&i.p.parser.dialect.allows(s>>1)){(s&1)==0?e.value=s>>1:e.extended=s>>1;break}}}else e.value=0,e.end=this.stream.clipPos(n+1)}putAction(e,A,i,n){for(let o=0;o<n;o+=3)if(this.actions[o]==e)return n;return this.actions[n++]=e,this.actions[n++]=A,this.actions[n++]=i,n}addActions(e,A,i,n){let{state:o}=e,{parser:r}=e.p,{data:s}=r;for(let a=0;a<2;a++)for(let c=r.stateSlot(o,a?2:1);;c+=3){if(s[c]==65535)if(s[c+1]==1)c=z0(s,c+2);else{n==0&&s[c+1]==2&&(n=this.putAction(z0(s,c+2),A,i,n));break}s[c]==A&&(n=this.putAction(z0(s,c+1),A,i,n))}return n}},sN=class{constructor(e,A,i,n){this.parser=e,this.input=A,this.ranges=n,this.recovering=0,this.nextStackID=9812,this.minStackPos=0,this.reused=[],this.stoppedAt=null,this.lastBigReductionStart=-1,this.lastBigReductionSize=0,this.bigReductionCount=0,this.stream=new iN(A,n),this.tokens=new rN(e,this.stream),this.topTerm=e.top[1];let{from:o}=n[0];this.stacks=[AN.start(this,e.top[0],o)],this.fragments=i.length&&this.stream.end-o>e.bufferLength*4?new oN(i,e.nodeSet):null}get parsedPos(){return this.minStackPos}advance(){let e=this.stacks,A=this.minStackPos,i=this.stacks=[],n,o;if(this.bigReductionCount>300&&e.length==1){let[r]=e;for(;r.forceReduce()&&r.stack.length&&r.stack[r.stack.length-2]>=this.lastBigReductionStart;);this.bigReductionCount=this.lastBigReductionSize=0}for(let r=0;r<e.length;r++){let s=e[r];for(;;){if(this.tokens.mainToken=null,s.pos>A)i.push(s);else{if(this.advanceStack(s,i,e))continue;{n||(n=[],o=[]),n.push(s);let a=this.tokens.getMainToken(s);o.push(a.value,a.end)}}break}}if(!i.length){let r=n&&NNA(n);if(r)return dc&&console.log("Finish with "+this.stackID(r)),this.stackToTree(r);if(this.parser.strict)throw dc&&n&&console.log("Stuck with token "+(this.tokens.mainToken?this.parser.getName(this.tokens.mainToken.value):"none")),new SyntaxError("No parse at "+A);this.recovering||(this.recovering=5)}if(this.recovering&&n){let r=this.stoppedAt!=null&&n[0].pos>this.stoppedAt?n[0]:this.runRecovery(n,o,i);if(r)return dc&&console.log("Force-finish "+this.stackID(r)),this.stackToTree(r.forceAll())}if(this.recovering){let r=this.recovering==1?1:this.recovering*3;if(i.length>r)for(i.sort((s,a)=>a.score-s.score);i.length>r;)i.pop();i.some(s=>s.reducePos>A)&&this.recovering--}else if(i.length>1){A:for(let r=0;r<i.length-1;r++){let s=i[r];for(let a=r+1;a<i.length;a++){let c=i[a];if(s.sameState(c)||s.buffer.length>500&&c.buffer.length>500)if((s.score-c.score||s.buffer.length-c.buffer.length)>0)i.splice(a--,1);else{i.splice(r--,1);continue A}}}i.length>12&&i.splice(12,i.length-12)}this.minStackPos=i[0].pos;for(let r=1;r<i.length;r++)i[r].pos<this.minStackPos&&(this.minStackPos=i[r].pos);return null}stopAt(e){if(this.stoppedAt!=null&&this.stoppedAt<e)throw new RangeError("Can't move stoppedAt forward");this.stoppedAt=e}advanceStack(e,A,i){let n=e.pos,{parser:o}=this,r=dc?this.stackID(e)+" -> ":"";if(this.stoppedAt!=null&&n>this.stoppedAt)return e.forceReduce()?e:null;if(this.fragments){let c=e.curContext&&e.curContext.tracker.strict,l=c?e.curContext.hash:0;for(let I=this.fragments.nodeAt(n);I;){let C=this.parser.nodeSet.types[I.type.id]==I.type?o.getGoto(e.state,I.type.id):-1;if(C>-1&&I.length&&(!c||(I.prop(fi.contextHash)||0)==l))return e.useNode(I,C),dc&&console.log(r+this.stackID(e)+` (via reuse of ${o.getName(I.type.id)})`),!0;if(!(I instanceof ur)||I.children.length==0||I.positions[0]>0)break;let d=I.children[0];if(d instanceof ur&&I.positions[0]==0)I=d;else break}}let s=o.stateSlot(e.state,4);if(s>0)return e.reduce(s),dc&&console.log(r+this.stackID(e)+` (via always-reduce ${o.getName(s&65535)})`),!0;if(e.stack.length>=8400)for(;e.stack.length>6e3&&e.forceReduce(););let a=this.tokens.getActions(e);for(let c=0;c<a.length;){let l=a[c++],I=a[c++],C=a[c++],d=c==a.length||!i,B=d?e:e.split(),E=this.tokens.mainToken;if(B.apply(l,I,E?E.start:B.pos,C),dc&&console.log(r+this.stackID(B)+` (via ${(l&65536)==0?"shift":`reduce of ${o.getName(l&65535)}`} for ${o.getName(I)} @ ${n}${B==e?"":", split"})`),d)return!0;B.pos>n?A.push(B):i.push(B)}return!1}advanceFully(e,A){let i=e.pos;for(;;){if(!this.advanceStack(e,null,null))return!1;if(e.pos>i)return PnA(e,A),!0}}runRecovery(e,A,i){let n=null,o=!1;for(let r=0;r<e.length;r++){let s=e[r],a=A[r<<1],c=A[(r<<1)+1],l=dc?this.stackID(s)+" -> ":"";if(s.deadEnd&&(o||(o=!0,s.restart(),dc&&console.log(l+this.stackID(s)+" (restarted)"),this.advanceFully(s,i))))continue;let I=s.split(),C=l;for(let d=0;I.forceReduce()&&d<10&&(dc&&console.log(C+this.stackID(I)+" (via force-reduce)"),!this.advanceFully(I,i));d++)dc&&(C=this.stackID(I)+" -> ");for(let d of s.recoverByInsert(a))dc&&console.log(l+this.stackID(d)+" (via recover-insert)"),this.advanceFully(d,i);this.stream.end>s.pos?(c==s.pos&&(c++,a=0),s.recoverByDelete(a,c),dc&&console.log(l+this.stackID(s)+` (via recover-delete ${this.parser.getName(a)})`),PnA(s,i)):(!n||n.score<s.score)&&(n=s)}return n}stackToTree(e){return e.close(),ur.build({buffer:tN.create(e),nodeSet:this.parser.nodeSet,topID:this.topTerm,maxBufferLength:this.parser.bufferLength,reused:this.reused,start:this.ranges[0].from,length:e.pos-this.ranges[0].from,minRepeatType:this.parser.minRepeatTerm})}stackID(e){let A=($F||($F=new WeakMap)).get(e);return A||$F.set(e,A=String.fromCodePoint(this.nextStackID++)),A+e}};function PnA(t,e){for(let A=0;A<e.length;A++){let i=e[A];if(i.pos==t.pos&&i.sameState(t)){e[A].score<t.score&&(e[A]=t);return}}e.push(t)}var aN=class{constructor(e,A,i){this.source=e,this.flags=A,this.disabled=i}allows(e){return!this.disabled||this.disabled[e]==0}};var pD=class t extends KE{constructor(e){if(super(),this.wrappers=[],e.version!=14)throw new RangeError(`Parser version (${e.version}) doesn't match runtime version (14)`);let A=e.nodeNames.split(" ");this.minRepeatTerm=A.length;for(let s=0;s<e.repeatNodeCount;s++)A.push("");let i=Object.keys(e.topRules).map(s=>e.topRules[s][1]),n=[];for(let s=0;s<A.length;s++)n.push([]);function o(s,a,c){n[s].push([a,a.deserialize(String(c))])}if(e.nodeProps)for(let s of e.nodeProps){let a=s[0];typeof a=="string"&&(a=fi[a]);for(let c=1;c<s.length;){let l=s[c++];if(l>=0)o(l,a,s[c++]);else{let I=s[c+-l];for(let C=-l;C>0;C--)o(s[c++],a,I);c++}}}this.nodeSet=new I3(A.map((s,a)=>Fs.define({name:a>=this.minRepeatTerm?void 0:s,id:a,props:n[a],top:i.indexOf(a)>-1,error:a==0,skipped:e.skippedNodes&&e.skippedNodes.indexOf(a)>-1}))),e.propSources&&(this.nodeSet=this.nodeSet.extend(...e.propSources)),this.strict=!1,this.bufferLength=1024;let r=D3(e.tokenData);this.context=e.context,this.specializerSpecs=e.specialized||[],this.specialized=new Uint16Array(this.specializerSpecs.length);for(let s=0;s<this.specializerSpecs.length;s++)this.specialized[s]=this.specializerSpecs[s].term;this.specializers=this.specializerSpecs.map(jnA),this.states=D3(e.states,Uint32Array),this.data=D3(e.stateData),this.goto=D3(e.goto),this.maxTerm=e.maxTerm,this.tokenizers=e.tokenizers.map(s=>typeof s=="number"?new R1(r,s):s),this.topRules=e.topRules,this.dialects=e.dialects||{},this.dynamicPrecedences=e.dynamicPrecedences||null,this.tokenPrecTable=e.tokenPrec,this.termNames=e.termNames||null,this.maxNode=this.nodeSet.types.length-1,this.dialect=this.parseDialect(),this.top=this.topRules[Object.keys(this.topRules)[0]]}createParse(e,A,i){let n=new sN(this,e,A,i);for(let o of this.wrappers)n=o(n,e,A,i);return n}getGoto(e,A,i=!1){let n=this.goto;if(A>=n[0])return-1;for(let o=n[A+1];;){let r=n[o++],s=r&1,a=n[o++];if(s&&i)return a;for(let c=o+(r>>1);o<c;o++)if(n[o]==e)return a;if(s)return-1}}hasAction(e,A){let i=this.data;for(let n=0;n<2;n++)for(let o=this.stateSlot(e,n?2:1),r;;o+=3){if((r=i[o])==65535)if(i[o+1]==1)r=i[o=z0(i,o+2)];else{if(i[o+1]==2)return z0(i,o+2);break}if(r==A||r==0)return z0(i,o+1)}return 0}stateSlot(e,A){return this.states[e*6+A]}stateFlag(e,A){return(this.stateSlot(e,0)&A)>0}validAction(e,A){return!!this.allActions(e,i=>i==A?!0:null)}allActions(e,A){let i=this.stateSlot(e,4),n=i?A(i):void 0;for(let o=this.stateSlot(e,1);n==null;o+=3){if(this.data[o]==65535)if(this.data[o+1]==1)o=z0(this.data,o+2);else break;n=A(z0(this.data,o+1))}return n}nextStates(e){let A=[];for(let i=this.stateSlot(e,1);;i+=3){if(this.data[i]==65535)if(this.data[i+1]==1)i=z0(this.data,i+2);else break;if((this.data[i+2]&1)==0){let n=this.data[i+1];A.some((o,r)=>r&1&&o==n)||A.push(this.data[i],n)}}return A}configure(e){let A=Object.assign(Object.create(t.prototype),this);if(e.props&&(A.nodeSet=this.nodeSet.extend(...e.props)),e.top){let i=this.topRules[e.top];if(!i)throw new RangeError(`Invalid top rule name ${e.top}`);A.top=i}return e.tokenizers&&(A.tokenizers=this.tokenizers.map(i=>{let n=e.tokenizers.find(o=>o.from==i);return n?n.to:i})),e.specializers&&(A.specializers=this.specializers.slice(),A.specializerSpecs=this.specializerSpecs.map((i,n)=>{let o=e.specializers.find(s=>s.from==i.external);if(!o)return i;let r=Object.assign(Object.assign({},i),{external:o.to});return A.specializers[n]=jnA(r),r})),e.contextTracker&&(A.context=e.contextTracker),e.dialect&&(A.dialect=this.parseDialect(e.dialect)),e.strict!=null&&(A.strict=e.strict),e.wrap&&(A.wrappers=A.wrappers.concat(e.wrap)),e.bufferLength!=null&&(A.bufferLength=e.bufferLength),A}hasWrappers(){return this.wrappers.length>0}getName(e){return this.termNames?this.termNames[e]:String(e<=this.maxNode&&this.nodeSet.types[e].name||e)}get eofTerm(){return this.maxNode+1}get topNode(){return this.nodeSet.types[this.top[1]]}dynamicPrecedence(e){let A=this.dynamicPrecedences;return A==null?0:A[e]||0}parseDialect(e){let A=Object.keys(this.dialects),i=A.map(()=>!1);if(e)for(let o of e.split(" ")){let r=A.indexOf(o);r>=0&&(i[r]=!0)}let n=null;for(let o=0;o<A.length;o++)if(!i[o])for(let r=this.dialects[A[o]],s;(s=this.data[r++])!=65535;)(n||(n=new Uint8Array(this.maxTerm+1)))[s]=1;return new aN(e,i,n)}static deserialize(e){return new t(e)}};function z0(t,e){return t[e]|t[e+1]<<16}function NNA(t){let e=null;for(let A of t){let i=A.p.stoppedAt;(A.pos==A.p.stream.end||i!=null&&A.pos>i)&&A.p.parser.stateFlag(A.state,2)&&(!e||e.score<A.score)&&(e=A)}return e}function jnA(t){if(t.external){let e=t.extend?1:0;return(A,i)=>t.external(A,i)<<1|e}return t.get}var _NA=iD({String:Se.string,Number:Se.number,"True False":Se.bool,PropertyName:Se.propertyName,Null:Se.null,", :":Se.separator,"[ ]":Se.squareBracket,"{ }":Se.brace}),VnA=pD.deserialize({version:14,states:"$bOVQPOOOOQO'#Cb'#CbOnQPO'#CeOvQPO'#ClOOQO'#Cr'#CrQOQPOOOOQO'#Cg'#CgO}QPO'#CfO!SQPO'#CtOOQO,59P,59PO![QPO,59PO!aQPO'#CuOOQO,59W,59WO!iQPO,59WOVQPO,59QOqQPO'#CmO!nQPO,59`OOQO1G.k1G.kOVQPO'#CnO!vQPO,59aOOQO1G.r1G.rOOQO1G.l1G.lOOQO,59X,59XOOQO-E6k-E6kOOQO,59Y,59YOOQO-E6l-E6l",stateData:"#O~OeOS~OQSORSOSSOTSOWQO_ROgPO~OVXOgUO~O^[O~PVO[^O~O]_OVhX~OVaO~O]bO^iX~O^dO~O]_OVha~O]bO^ia~O",goto:"!kjPPPPPPkPPkqwPPPPk{!RPPP!XP!e!hXSOR^bQWQRf_TVQ_Q`WRg`QcZRicQTOQZRQe^RhbRYQR]R",nodeNames:"\u26A0 JsonText True False Null Number String } { Object Property PropertyName : , ] [ Array",maxTerm:25,nodeProps:[["isolate",-2,6,11,""],["openedBy",7,"{",14,"["],["closedBy",8,"}",15,"]"]],propSources:[_NA],skippedNodes:[0],repeatNodeCount:2,tokenData:"(|~RaXY!WYZ!W]^!Wpq!Wrs!]|}$u}!O$z!Q!R%T!R![&c![!]&t!}#O&y#P#Q'O#Y#Z'T#b#c'r#h#i(Z#o#p(r#q#r(w~!]Oe~~!`Wpq!]qr!]rs!xs#O!]#O#P!}#P;'S!];'S;=`$o<%lO!]~!}Og~~#QXrs!]!P!Q!]#O#P!]#U#V!]#Y#Z!]#b#c!]#f#g!]#h#i!]#i#j#m~#pR!Q![#y!c!i#y#T#Z#y~#|R!Q![$V!c!i$V#T#Z$V~$YR!Q![$c!c!i$c#T#Z$c~$fR!Q![!]!c!i!]#T#Z!]~$rP;=`<%l!]~$zO]~~$}Q!Q!R%T!R![&c~%YRT~!O!P%c!g!h%w#X#Y%w~%fP!Q![%i~%nRT~!Q![%i!g!h%w#X#Y%w~%zR{|&T}!O&T!Q![&Z~&WP!Q![&Z~&`PT~!Q![&Z~&hST~!O!P%c!Q![&c!g!h%w#X#Y%w~&yO[~~'OO_~~'TO^~~'WP#T#U'Z~'^P#`#a'a~'dP#g#h'g~'jP#X#Y'm~'rOR~~'uP#i#j'x~'{P#`#a(O~(RP#`#a(U~(ZOS~~(^P#f#g(a~(dP#i#j(g~(jP#X#Y(m~(rOQ~~(wOW~~(|OV~",tokenizers:[0],topRules:{JsonText:[0,1]},tokenPrec:0});var GNA=nD.define({name:"json",parser:VnA.configure({props:[KF.add({Object:YF({except:/^\s*\}/}),Array:YF({except:/^\s*\]/})}),JF.add({"Object Array":xiA})]}),languageData:{closeBrackets:{brackets:["[","{",'"']},indentOnInput:/^\s*[\}\]]$/}});function ZnA(){return new oD(GNA)}var WnA=typeof String.prototype.normalize=="function"?t=>t.normalize("NFKD"):t=>t,L1=class{constructor(e,A,i=0,n=e.length,o,r){this.test=r,this.value={from:0,to:0},this.done=!1,this.matches=[],this.buffer="",this.bufferPos=0,this.iter=e.iterRange(i,n),this.bufferStart=i,this.normalize=o?s=>o(WnA(s)):WnA,this.query=this.normalize(A)}peek(){if(this.bufferPos==this.buffer.length){if(this.bufferStart+=this.buffer.length,this.iter.next(),this.iter.done)return-1;this.bufferPos=0,this.buffer=this.iter.value}return Bs(this.buffer,this.bufferPos)}next(){for(;this.matches.length;)this.matches.pop();return this.nextOverlapping()}nextOverlapping(){for(;;){let e=this.peek();if(e<0)return this.done=!0,this;let A=T4(e),i=this.bufferStart+this.bufferPos;this.bufferPos+=lc(e);let n=this.normalize(A);if(n.length)for(let o=0,r=i;;o++){let s=n.charCodeAt(o),a=this.match(s,r,this.bufferPos+this.bufferStart);if(o==n.length-1){if(a)return this.value=a,this;break}r==i&&o<A.length&&A.charCodeAt(o)==s&&r++}}}match(e,A,i){let n=null;for(let o=0;o<this.matches.length;o+=2){let r=this.matches[o],s=!1;this.query.charCodeAt(r)==e&&(r==this.query.length-1?n={from:this.matches[o+1],to:i}:(this.matches[o]++,s=!0)),s||(this.matches.splice(o,2),o-=2)}return this.query.charCodeAt(0)==e&&(this.query.length==1?n={from:A,to:i}:this.matches.push(1,A)),n&&this.test&&!this.test(n.from,n.to,this.buffer,this.bufferStart)&&(n=null),n}};typeof Symbol<"u"&&(L1.prototype[Symbol.iterator]=function(){return this});var eoA={from:-1,to:-1,match:/.*/.exec("")},EN="gm"+(/x/.unicode==null?"":"u"),yD=class{constructor(e,A,i,n=0,o=e.length){if(this.text=e,this.to=o,this.curLine="",this.done=!1,this.value=eoA,/\\[sWDnr]|\n|\r|\[\^/.test(A))return new bD(e,A,i,n,o);this.re=new RegExp(A,EN+(i?.ignoreCase?"i":"")),this.test=i?.test,this.iter=e.iter();let r=e.lineAt(n);this.curLineStart=r.from,this.matchPos=MD(e,n),this.getLine(this.curLineStart)}getLine(e){this.iter.next(e),this.iter.lineBreak?this.curLine="":(this.curLine=this.iter.value,this.curLineStart+this.curLine.length>this.to&&(this.curLine=this.curLine.slice(0,this.to-this.curLineStart)),this.iter.next())}nextLine(){this.curLineStart=this.curLineStart+this.curLine.length+1,this.curLineStart>this.to?this.curLine="":this.getLine(0)}next(){for(let e=this.matchPos-this.curLineStart;;){this.re.lastIndex=e;let A=this.matchPos<=this.to&&this.re.exec(this.curLine);if(A){let i=this.curLineStart+A.index,n=i+A[0].length;if(this.matchPos=MD(this.text,n+(i==n?1:0)),i==this.curLineStart+this.curLine.length&&this.nextLine(),(i<n||i>this.value.to)&&(!this.test||this.test(i,n,A)))return this.value={from:i,to:n,match:A},this;e=this.matchPos-this.curLineStart}else if(this.curLineStart+this.curLine.length<this.to)this.nextLine(),e=0;else return this.done=!0,this}}},cN=new WeakMap,vD=class t{constructor(e,A){this.from=e,this.text=A}get to(){return this.from+this.text.length}static get(e,A,i){let n=cN.get(e);if(!n||n.from>=i||n.to<=A){let s=new t(A,e.sliceString(A,i));return cN.set(e,s),s}if(n.from==A&&n.to==i)return n;let{text:o,from:r}=n;return r>A&&(o=e.sliceString(A,r)+o,r=A),n.to<i&&(o+=e.sliceString(n.to,i)),cN.set(e,new t(r,o)),new t(A,o.slice(A-r,i-r))}},bD=class{constructor(e,A,i,n,o){this.text=e,this.to=o,this.done=!1,this.value=eoA,this.matchPos=MD(e,n),this.re=new RegExp(A,EN+(i?.ignoreCase?"i":"")),this.test=i?.test,this.flat=vD.get(e,n,this.chunkEnd(n+5e3))}chunkEnd(e){return e>=this.to?this.to:this.text.lineAt(e).to}next(){for(;;){let e=this.re.lastIndex=this.matchPos-this.flat.from,A=this.re.exec(this.flat.text);if(A&&!A[0]&&A.index==e&&(this.re.lastIndex=e+1,A=this.re.exec(this.flat.text)),A){let i=this.flat.from+A.index,n=i+A[0].length;if((this.flat.to>=this.to||A.index+A[0].length<=this.flat.text.length-10)&&(!this.test||this.test(i,n,A)))return this.value={from:i,to:n,match:A},this.matchPos=MD(this.text,n+(i==n?1:0)),this}if(this.flat.to==this.to)return this.done=!0,this;this.flat=vD.get(this.text,this.flat.from,this.chunkEnd(this.flat.from+this.flat.text.length*2))}}};typeof Symbol<"u"&&(yD.prototype[Symbol.iterator]=bD.prototype[Symbol.iterator]=function(){return this});function UNA(t){try{return new RegExp(t,EN),!0}catch{return!1}}function MD(t,e){if(e>=t.length)return e;let A=t.lineAt(e),i;for(;e<A.to&&(i=A.text.charCodeAt(e-A.from))>=56320&&i<57344;)e++;return e}function lN(t){let e=String(t.state.doc.lineAt(t.state.selection.main.head).number),A=Un("input",{class:"cm-textfield",name:"line",value:e}),i=Un("form",{class:"cm-gotoLine",onkeydown:o=>{o.keyCode==27?(o.preventDefault(),t.dispatch({effects:y3.of(!1)}),t.focus()):o.keyCode==13&&(o.preventDefault(),n())},onsubmit:o=>{o.preventDefault(),n()}},Un("label",t.state.phrase("Go to line"),": ",A)," ",Un("button",{class:"cm-button",type:"submit"},t.state.phrase("go")),Un("button",{name:"close",onclick:()=>{t.dispatch({effects:y3.of(!1)}),t.focus()},"aria-label":t.state.phrase("close"),type:"button"},["\xD7"]));function n(){let o=/^([+-])?(\d+)?(:\d+)?(%)?$/.exec(A.value);if(!o)return;let{state:r}=t,s=r.doc.lineAt(r.selection.main.head),[,a,c,l,I]=o,C=l?+l.slice(1):0,d=c?+c:s.number;if(c&&I){let h=d/100;a&&(h=h*(a=="-"?-1:1)+s.number/r.doc.lines),d=Math.round(r.doc.lines*h)}else c&&a&&(d=d*(a=="-"?-1:1)+s.number);let B=r.doc.line(Math.max(1,Math.min(r.doc.lines,d))),E=ae.cursor(B.from+Math.max(0,Math.min(C,B.length)));t.dispatch({effects:[y3.of(!1),qt.scrollIntoView(E.from,{y:"center"})],selection:E}),t.focus()}return{dom:i}}var y3=Pi.define(),XnA=Ar.define({create(){return!0},update(t,e){for(let A of e.effects)A.is(y3)&&(t=A.value);return t},provide:t=>MC.from(t,e=>e?lN:null)}),KNA=t=>{let e=kC(t,lN);if(!e){let A=[y3.of(!0)];t.state.field(XnA,!1)==null&&A.push(Pi.appendConfig.of([XnA,YNA])),t.dispatch({effects:A}),e=kC(t,lN)}return e&&e.dom.querySelector("input").select(),!0},YNA=qt.baseTheme({".cm-panel.cm-gotoLine":{padding:"2px 6px 4px",position:"relative","& label":{fontSize:"80%"},"& [name=close]":{position:"absolute",top:"0",bottom:"0",right:"4px",backgroundColor:"inherit",border:"none",font:"inherit",padding:"0"}}}),JNA={highlightWordAroundCursor:!1,minSelectionLength:1,maxMatches:100,wholeWords:!1},toA=qe.define({combine(t){return Wr(t,JNA,{highlightWordAroundCursor:(e,A)=>e||A,minSelectionLength:Math.min,maxMatches:Math.min})}});function ioA(t){let e=[PNA,ONA];return t&&e.push(toA.of(t)),e}var TNA=ut.mark({class:"cm-selectionMatch"}),HNA=ut.mark({class:"cm-selectionMatch cm-selectionMatch-main"});function $nA(t,e,A,i){return(A==0||t(e.sliceDoc(A-1,A))!=ao.Word)&&(i==e.doc.length||t(e.sliceDoc(i,i+1))!=ao.Word)}function zNA(t,e,A,i){return t(e.sliceDoc(A,A+1))==ao.Word&&t(e.sliceDoc(i-1,i))==ao.Word}var ONA=go.fromClass(class{constructor(t){this.decorations=this.getDeco(t)}update(t){(t.selectionSet||t.docChanged||t.viewportChanged)&&(this.decorations=this.getDeco(t.view))}getDeco(t){let e=t.state.facet(toA),{state:A}=t,i=A.selection;if(i.ranges.length>1)return ut.none;let n=i.main,o,r=null;if(n.empty){if(!e.highlightWordAroundCursor)return ut.none;let a=A.wordAt(n.head);if(!a)return ut.none;r=A.charCategorizer(n.head),o=A.sliceDoc(a.from,a.to)}else{let a=n.to-n.from;if(a<e.minSelectionLength||a>200)return ut.none;if(e.wholeWords){if(o=A.sliceDoc(n.from,n.to),r=A.charCategorizer(n.head),!($nA(r,A,n.from,n.to)&&zNA(r,A,n.from,n.to)))return ut.none}else if(o=A.sliceDoc(n.from,n.to),!o)return ut.none}let s=[];for(let a of t.visibleRanges){let c=new L1(A.doc,o,a.from,a.to);for(;!c.next().done;){let{from:l,to:I}=c.value;if((!r||$nA(r,A,l,I))&&(n.empty&&l<=n.from&&I>=n.to?s.push(HNA.range(l,I)):(l>=n.to||I<=n.from)&&s.push(TNA.range(l,I)),s.length>e.maxMatches))return ut.none}}return ut.set(s)}},{decorations:t=>t.decorations}),PNA=qt.baseTheme({".cm-selectionMatch":{backgroundColor:"#99ff7780"},".cm-searchMatch .cm-selectionMatch":{backgroundColor:"transparent"}}),jNA=({state:t,dispatch:e})=>{let{selection:A}=t,i=ae.create(A.ranges.map(n=>t.wordAt(n.head)||ae.cursor(n.head)),A.mainIndex);return i.eq(A)?!1:(e(t.update({selection:i})),!0)};function qNA(t,e){let{main:A,ranges:i}=t.selection,n=t.wordAt(A.head),o=n&&n.from==A.from&&n.to==A.to;for(let r=!1,s=new L1(t.doc,e,i[i.length-1].to);;)if(s.next(),s.done){if(r)return null;s=new L1(t.doc,e,0,Math.max(0,i[i.length-1].from-1)),r=!0}else{if(r&&i.some(a=>a.from==s.value.from))continue;if(o){let a=t.wordAt(s.value.from);if(!a||a.from!=s.value.from||a.to!=s.value.to)continue}return s.value}}var VNA=({state:t,dispatch:e})=>{let{ranges:A}=t.selection;if(A.some(o=>o.from===o.to))return jNA({state:t,dispatch:e});let i=t.sliceDoc(A[0].from,A[0].to);if(t.selection.ranges.some(o=>t.sliceDoc(o.from,o.to)!=i))return!1;let n=qNA(t,i);return n?(e(t.update({selection:t.selection.addRange(ae.range(n.from,n.to),!1),effects:qt.scrollIntoView(n.to)})),!0):!1},_C=qe.define({combine(t){return Wr(t,{top:!1,caseSensitive:!1,literal:!1,regexp:!1,wholeWord:!1,createPanel:e=>new dN(e),scrollToMatch:e=>qt.scrollIntoView(e)})}});function noA(t){return t?[_C.of(t),BN]:BN}var kD=class{constructor(e){this.search=e.search,this.caseSensitive=!!e.caseSensitive,this.literal=!!e.literal,this.regexp=!!e.regexp,this.replace=e.replace||"",this.valid=!!this.search&&(!this.regexp||UNA(this.search)),this.unquoted=this.unquote(this.search),this.wholeWord=!!e.wholeWord}unquote(e){return this.literal?e:e.replace(/\\([nrt\\])/g,(A,i)=>i=="n"?`
`:i=="r"?"\r":i=="t"?"	":"\\")}eq(e){return this.search==e.search&&this.replace==e.replace&&this.caseSensitive==e.caseSensitive&&this.regexp==e.regexp&&this.wholeWord==e.wholeWord}create(){return this.regexp?new IN(this):new gN(this)}getCursor(e,A=0,i){let n=e.doc?e:hr.create({doc:e});return i==null&&(i=n.doc.length),this.regexp?qE(this,n,A,i):jE(this,n,A,i)}},SD=class{constructor(e){this.spec=e}};function jE(t,e,A,i){return new L1(e.doc,t.unquoted,A,i,t.caseSensitive?void 0:n=>n.toLowerCase(),t.wholeWord?ZNA(e.doc,e.charCategorizer(e.selection.main.head)):void 0)}function ZNA(t,e){return(A,i,n,o)=>((o>A||o+n.length<i)&&(o=Math.max(0,A-2),n=t.sliceString(o,Math.min(t.length,i+2))),(e(RD(n,A-o))!=ao.Word||e(xD(n,A-o))!=ao.Word)&&(e(xD(n,i-o))!=ao.Word||e(RD(n,i-o))!=ao.Word))}var gN=class extends SD{constructor(e){super(e)}nextMatch(e,A,i){let n=jE(this.spec,e,i,e.doc.length).nextOverlapping();if(n.done){let o=Math.min(e.doc.length,A+this.spec.unquoted.length);n=jE(this.spec,e,0,o).nextOverlapping()}return n.done||n.value.from==A&&n.value.to==i?null:n.value}prevMatchInRange(e,A,i){for(let n=i;;){let o=Math.max(A,n-1e4-this.spec.unquoted.length),r=jE(this.spec,e,o,n),s=null;for(;!r.nextOverlapping().done;)s=r.value;if(s)return s;if(o==A)return null;n-=1e4}}prevMatch(e,A,i){let n=this.prevMatchInRange(e,0,A);return n||(n=this.prevMatchInRange(e,Math.max(0,i-this.spec.unquoted.length),e.doc.length)),n&&(n.from!=A||n.to!=i)?n:null}getReplacement(e){return this.spec.unquote(this.spec.replace)}matchAll(e,A){let i=jE(this.spec,e,0,e.doc.length),n=[];for(;!i.next().done;){if(n.length>=A)return null;n.push(i.value)}return n}highlight(e,A,i,n){let o=jE(this.spec,e,Math.max(0,A-this.spec.unquoted.length),Math.min(i+this.spec.unquoted.length,e.doc.length));for(;!o.next().done;)n(o.value.from,o.value.to)}};function qE(t,e,A,i){return new yD(e.doc,t.search,{ignoreCase:!t.caseSensitive,test:t.wholeWord?WNA(e.charCategorizer(e.selection.main.head)):void 0},A,i)}function RD(t,e){return t.slice(yr(t,e,!1),e)}function xD(t,e){return t.slice(e,yr(t,e))}function WNA(t){return(e,A,i)=>!i[0].length||(t(RD(i.input,i.index))!=ao.Word||t(xD(i.input,i.index))!=ao.Word)&&(t(xD(i.input,i.index+i[0].length))!=ao.Word||t(RD(i.input,i.index+i[0].length))!=ao.Word)}var IN=class extends SD{nextMatch(e,A,i){let n=qE(this.spec,e,i,e.doc.length).next();return n.done&&(n=qE(this.spec,e,0,A).next()),n.done?null:n.value}prevMatchInRange(e,A,i){for(let n=1;;n++){let o=Math.max(A,i-n*1e4),r=qE(this.spec,e,o,i),s=null;for(;!r.next().done;)s=r.value;if(s&&(o==A||s.from>o+10))return s;if(o==A)return null}}prevMatch(e,A,i){return this.prevMatchInRange(e,0,A)||this.prevMatchInRange(e,i,e.doc.length)}getReplacement(e){return this.spec.unquote(this.spec.replace).replace(/\$([$&]|\d+)/g,(A,i)=>{if(i=="&")return e.match[0];if(i=="$")return"$";for(let n=i.length;n>0;n--){let o=+i.slice(0,n);if(o>0&&o<e.match.length)return e.match[o]+i.slice(n)}return A})}matchAll(e,A){let i=qE(this.spec,e,0,e.doc.length),n=[];for(;!i.next().done;){if(n.length>=A)return null;n.push(i.value)}return n}highlight(e,A,i,n){let o=qE(this.spec,e,Math.max(0,A-250),Math.min(i+250,e.doc.length));for(;!o.next().done;)n(o.value.from,o.value.to)}},b3=Pi.define(),QN=Pi.define(),x1=Ar.define({create(t){return new v3(CN(t).create(),null)},update(t,e){for(let A of e.effects)A.is(b3)?t=new v3(A.value.create(),t.panel):A.is(QN)&&(t=new v3(t.query,A.value?hN:null));return t},provide:t=>MC.from(t,e=>e.panel)});var v3=class{constructor(e,A){this.query=e,this.panel=A}},XNA=ut.mark({class:"cm-searchMatch"}),$NA=ut.mark({class:"cm-searchMatch cm-searchMatch-selected"}),A_A=go.fromClass(class{constructor(t){this.view=t,this.decorations=this.highlight(t.state.field(x1))}update(t){let e=t.state.field(x1);(e!=t.startState.field(x1)||t.docChanged||t.selectionSet||t.viewportChanged)&&(this.decorations=this.highlight(e))}highlight({query:t,panel:e}){if(!e||!t.spec.valid)return ut.none;let{view:A}=this,i=new ds;for(let n=0,o=A.visibleRanges,r=o.length;n<r;n++){let{from:s,to:a}=o[n];for(;n<r-1&&a>o[n+1].from-2*250;)a=o[++n].to;t.highlight(A.state,s,a,(c,l)=>{let I=A.state.selection.ranges.some(C=>C.from==c&&C.to==l);i.add(c,l,I?$NA:XNA)})}return i.finish()}},{decorations:t=>t.decorations});function M3(t){return e=>{let A=e.state.field(x1,!1);return A&&A.query.spec.valid?t(e,A):ND(e)}}var LD=M3((t,{query:e})=>{let{to:A}=t.state.selection.main,i=e.nextMatch(t.state,A,A);if(!i)return!1;let n=ae.single(i.from,i.to),o=t.state.facet(_C);return t.dispatch({selection:n,effects:[uN(t,i),o.scrollToMatch(n.main,t)],userEvent:"select.search"}),roA(t),!0}),FD=M3((t,{query:e})=>{let{state:A}=t,{from:i}=A.selection.main,n=e.prevMatch(A,i,i);if(!n)return!1;let o=ae.single(n.from,n.to),r=t.state.facet(_C);return t.dispatch({selection:o,effects:[uN(t,n),r.scrollToMatch(o.main,t)],userEvent:"select.search"}),roA(t),!0}),e_A=M3((t,{query:e})=>{let A=e.matchAll(t.state,1e3);return!A||!A.length?!1:(t.dispatch({selection:ae.create(A.map(i=>ae.range(i.from,i.to))),userEvent:"select.search.matches"}),!0)}),t_A=({state:t,dispatch:e})=>{let A=t.selection;if(A.ranges.length>1||A.main.empty)return!1;let{from:i,to:n}=A.main,o=[],r=0;for(let s=new L1(t.doc,t.sliceDoc(i,n));!s.next().done;){if(o.length>1e3)return!1;s.value.from==i&&(r=o.length),o.push(ae.range(s.value.from,s.value.to))}return e(t.update({selection:ae.create(o,r),userEvent:"select.search.matches"})),!0},AoA=M3((t,{query:e})=>{let{state:A}=t,{from:i,to:n}=A.selection.main;if(A.readOnly)return!1;let o=e.nextMatch(A,i,i);if(!o)return!1;let r=o,s=[],a,c,l=[];r.from==i&&r.to==n&&(c=A.toText(e.getReplacement(r)),s.push({from:r.from,to:r.to,insert:c}),r=e.nextMatch(A,r.from,r.to),l.push(qt.announce.of(A.phrase("replaced match on line $",A.doc.lineAt(i).number)+".")));let I=t.state.changes(s);return r&&(a=ae.single(r.from,r.to).map(I),l.push(uN(t,r)),l.push(A.facet(_C).scrollToMatch(a.main,t))),t.dispatch({changes:I,selection:a,effects:l,userEvent:"input.replace"}),!0}),i_A=M3((t,{query:e})=>{if(t.state.readOnly)return!1;let A=e.matchAll(t.state,1e9).map(n=>{let{from:o,to:r}=n;return{from:o,to:r,insert:e.getReplacement(n)}});if(!A.length)return!1;let i=t.state.phrase("replaced $ matches",A.length)+".";return t.dispatch({changes:A,effects:qt.announce.of(i),userEvent:"input.replace.all"}),!0});function hN(t){return t.state.facet(_C).createPanel(t)}function CN(t,e){var A,i,n,o,r;let s=t.selection.main,a=s.empty||s.to>s.from+100?"":t.sliceDoc(s.from,s.to);if(e&&!a)return e;let c=t.facet(_C);return new kD({search:((A=e?.literal)!==null&&A!==void 0?A:c.literal)?a:a.replace(/\n/g,"\\n"),caseSensitive:(i=e?.caseSensitive)!==null&&i!==void 0?i:c.caseSensitive,literal:(n=e?.literal)!==null&&n!==void 0?n:c.literal,regexp:(o=e?.regexp)!==null&&o!==void 0?o:c.regexp,wholeWord:(r=e?.wholeWord)!==null&&r!==void 0?r:c.wholeWord})}function ooA(t){let e=kC(t,hN);return e&&e.dom.querySelector("[main-field]")}function roA(t){let e=ooA(t);e&&e==t.root.activeElement&&e.select()}var ND=t=>{let e=t.state.field(x1,!1);if(e&&e.panel){let A=ooA(t);if(A&&A!=t.root.activeElement){let i=CN(t.state,e.query.spec);i.valid&&t.dispatch({effects:b3.of(i)}),A.focus(),A.select()}}else t.dispatch({effects:[QN.of(!0),e?b3.of(CN(t.state,e.query.spec)):Pi.appendConfig.of(BN)]});return!0},_D=t=>{let e=t.state.field(x1,!1);if(!e||!e.panel)return!1;let A=kC(t,hN);return A&&A.dom.contains(t.root.activeElement)&&t.focus(),t.dispatch({effects:QN.of(!1)}),!0},soA=[{key:"Mod-f",run:ND,scope:"editor search-panel"},{key:"F3",run:LD,shift:FD,scope:"editor search-panel",preventDefault:!0},{key:"Mod-g",run:LD,shift:FD,scope:"editor search-panel",preventDefault:!0},{key:"Escape",run:_D,scope:"editor search-panel"},{key:"Mod-Shift-l",run:t_A},{key:"Mod-Alt-g",run:KNA},{key:"Mod-d",run:VNA,preventDefault:!0}],dN=class{constructor(e){this.view=e;let A=this.query=e.state.field(x1).query.spec;this.commit=this.commit.bind(this),this.searchField=Un("input",{value:A.search,placeholder:Bc(e,"Find"),"aria-label":Bc(e,"Find"),class:"cm-textfield",name:"search",form:"","main-field":"true",onchange:this.commit,onkeyup:this.commit}),this.replaceField=Un("input",{value:A.replace,placeholder:Bc(e,"Replace"),"aria-label":Bc(e,"Replace"),class:"cm-textfield",name:"replace",form:"",onchange:this.commit,onkeyup:this.commit}),this.caseField=Un("input",{type:"checkbox",name:"case",form:"",checked:A.caseSensitive,onchange:this.commit}),this.reField=Un("input",{type:"checkbox",name:"re",form:"",checked:A.regexp,onchange:this.commit}),this.wordField=Un("input",{type:"checkbox",name:"word",form:"",checked:A.wholeWord,onchange:this.commit});function i(n,o,r){return Un("button",{class:"cm-button",name:n,onclick:o,type:"button"},r)}this.dom=Un("div",{onkeydown:n=>this.keydown(n),class:"cm-search"},[this.searchField,i("next",()=>LD(e),[Bc(e,"next")]),i("prev",()=>FD(e),[Bc(e,"previous")]),i("select",()=>e_A(e),[Bc(e,"all")]),Un("label",null,[this.caseField,Bc(e,"match case")]),Un("label",null,[this.reField,Bc(e,"regexp")]),Un("label",null,[this.wordField,Bc(e,"by word")]),...e.state.readOnly?[]:[Un("br"),this.replaceField,i("replace",()=>AoA(e),[Bc(e,"replace")]),i("replaceAll",()=>i_A(e),[Bc(e,"replace all")])],Un("button",{name:"close",onclick:()=>_D(e),"aria-label":Bc(e,"close"),type:"button"},["\xD7"])])}commit(){let e=new kD({search:this.searchField.value,caseSensitive:this.caseField.checked,regexp:this.reField.checked,wholeWord:this.wordField.checked,replace:this.replaceField.value});e.eq(this.query)||(this.query=e,this.view.dispatch({effects:b3.of(e)}))}keydown(e){ZtA(this.view,e,"search-panel")?e.preventDefault():e.keyCode==13&&e.target==this.searchField?(e.preventDefault(),(e.shiftKey?FD:LD)(this.view)):e.keyCode==13&&e.target==this.replaceField&&(e.preventDefault(),AoA(this.view))}update(e){for(let A of e.transactions)for(let i of A.effects)i.is(b3)&&!i.value.eq(this.query)&&this.setQuery(i.value)}setQuery(e){this.query=e,this.searchField.value=e.search,this.replaceField.value=e.replace,this.caseField.checked=e.caseSensitive,this.reField.checked=e.regexp,this.wordField.checked=e.wholeWord}mount(){this.searchField.select()}get pos(){return 80}get top(){return this.view.state.facet(_C).top}};function Bc(t,e){return t.state.phrase(e)}var wD=30,DD=/[\s\.,:;?!]/;function uN(t,{from:e,to:A}){let i=t.state.doc.lineAt(e),n=t.state.doc.lineAt(A).to,o=Math.max(i.from,e-wD),r=Math.min(n,A+wD),s=t.state.sliceDoc(o,r);if(o!=i.from){for(let a=0;a<wD;a++)if(!DD.test(s[a+1])&&DD.test(s[a])){s=s.slice(a);break}}if(r!=n){for(let a=s.length-1;a>s.length-wD;a--)if(!DD.test(s[a-1])&&DD.test(s[a])){s=s.slice(0,a);break}}return qt.announce.of(`${t.state.phrase("current match")}. ${s} ${t.state.phrase("on line")} ${i.number}.`)}var n_A=qt.baseTheme({".cm-panel.cm-search":{padding:"2px 6px 4px",position:"relative","& [name=close]":{position:"absolute",top:"0",right:"4px",backgroundColor:"inherit",border:"none",font:"inherit",padding:0,margin:0},"& input, & button, & label":{margin:".2em .6em .2em 0"},"& input[type=checkbox]":{marginRight:".2em"},"& label":{fontSize:"80%",whiteSpace:"pre"}},"&light .cm-searchMatch":{backgroundColor:"#ffff0054"},"&dark .cm-searchMatch":{backgroundColor:"#00ffff8a"},"&light .cm-searchMatch-selected":{backgroundColor:"#ff6a0054"},"&dark .cm-searchMatch-selected":{backgroundColor:"#ff00ff8a"}}),BN=[x1,Dl.low(A_A),n_A];var UD=class{constructor(e,A,i,n){this.state=e,this.pos=A,this.explicit=i,this.view=n,this.abortListeners=[],this.abortOnDocChange=!1}tokenBefore(e){let A=$r(this.state).resolveInner(this.pos,-1);for(;A&&e.indexOf(A.name)<0;)A=A.parent;return A?{from:A.from,to:this.pos,text:this.state.sliceDoc(A.from,this.pos),type:A.type}:null}matchBefore(e){let A=this.state.doc.lineAt(this.pos),i=Math.max(A.from,this.pos-250),n=A.text.slice(i-A.from,this.pos-A.from),o=n.search(EoA(e,!1));return o<0?null:{from:i+o,to:this.pos,text:n.slice(o)}}get aborted(){return this.abortListeners==null}addEventListener(e,A,i){e=="abort"&&this.abortListeners&&(this.abortListeners.push(A),i&&i.onDocChange&&(this.abortOnDocChange=!0))}};function aoA(t){let e=Object.keys(t).join(""),A=/\w/.test(e);return A&&(e=e.replace(/\w/g,"")),`[${A?"\\w":""}${e.replace(/[^\w\s]/g,"\\$&")}]`}function o_A(t){let e=Object.create(null),A=Object.create(null);for(let{label:n}of t){e[n[0]]=!0;for(let o=1;o<n.length;o++)A[n[o]]=!0}let i=aoA(e)+aoA(A)+"*$";return[new RegExp("^"+i),new RegExp(i)]}function r_A(t){let e=t.map(n=>typeof n=="string"?{label:n}:n),[A,i]=e.every(n=>/^\w+$/.test(n.label))?[/\w*$/,/\w+$/]:o_A(e);return n=>{let o=n.matchBefore(i);return o||n.explicit?{from:o?o.from:n.pos,options:e,validFor:A}:null}}var KD=class{constructor(e,A,i,n){this.completion=e,this.source=A,this.match=i,this.score=n}};function UC(t){return t.selection.main.from}function EoA(t,e){var A;let{source:i}=t,n=e&&i[0]!="^",o=i[i.length-1]!="$";return!n&&!o?t:new RegExp(`${n?"^":""}(?:${i})${o?"$":""}`,(A=t.flags)!==null&&A!==void 0?A:t.ignoreCase?"i":"")}var QoA=xa.define();function s_A(t,e,A,i){let{main:n}=t.selection,o=A-n.from,r=i-n.from;return Object.assign(Object.assign({},t.changeByRange(s=>{if(s!=n&&A!=i&&t.sliceDoc(s.from+o,s.from+r)!=t.sliceDoc(A,i))return{range:s};let a=t.toText(e);return{changes:{from:s.from+o,to:i==n.from?s.to:s.from+r,insert:a},range:ae.cursor(s.from+o+a.length)}})),{scrollIntoView:!0,userEvent:"input.complete"})}var coA=new WeakMap;function a_A(t){if(!Array.isArray(t))return t;let e=coA.get(t);return e||coA.set(t,e=r_A(t)),e}var YD=Pi.define(),k3=Pi.define(),pN=class{constructor(e){this.pattern=e,this.chars=[],this.folded=[],this.any=[],this.precise=[],this.byWord=[],this.score=0,this.matched=[];for(let A=0;A<e.length;){let i=Bs(e,A),n=lc(i);this.chars.push(i);let o=e.slice(A,A+n),r=o.toUpperCase();this.folded.push(Bs(r==o?o.toLowerCase():r,0)),A+=n}this.astral=e.length!=this.chars.length}ret(e,A){return this.score=e,this.matched=A,this}match(e){if(this.pattern.length==0)return this.ret(-100,[]);if(e.length<this.pattern.length)return null;let{chars:A,folded:i,any:n,precise:o,byWord:r}=this;if(A.length==1){let D=Bs(e,0),L=lc(D),R=L==e.length?0:-100;if(D!=A[0])if(D==i[0])R+=-200;else return null;return this.ret(R,[0,L])}let s=e.indexOf(this.pattern);if(s==0)return this.ret(e.length==this.pattern.length?0:-100,[0,this.pattern.length]);let a=A.length,c=0;if(s<0){for(let D=0,L=Math.min(e.length,200);D<L&&c<a;){let R=Bs(e,D);(R==A[c]||R==i[c])&&(n[c++]=D),D+=lc(R)}if(c<a)return null}let l=0,I=0,C=!1,d=0,B=-1,E=-1,h=/[a-z]/.test(e),u=!0;for(let D=0,L=Math.min(e.length,200),R=0;D<L&&I<a;){let w=Bs(e,D);s<0&&(l<a&&w==A[l]&&(o[l++]=D),d<a&&(w==A[d]||w==i[d]?(d==0&&(B=D),E=D+1,d++):d=0));let _,K=w<255?w>=48&&w<=57||w>=97&&w<=122?2:w>=65&&w<=90?1:0:(_=T4(w))!=_.toLowerCase()?1:_!=_.toUpperCase()?2:0;(!D||K==1&&h||R==0&&K!=0)&&(A[I]==w||i[I]==w&&(C=!0)?r[I++]=D:r.length&&(u=!1)),R=K,D+=lc(w)}return I==a&&r[0]==0&&u?this.result(-100+(C?-200:0),r,e):d==a&&B==0?this.ret(-200-e.length+(E==e.length?0:-100),[0,E]):s>-1?this.ret(-700-e.length,[s,s+this.pattern.length]):d==a?this.ret(-900-e.length,[B,E]):I==a?this.result(-100+(C?-200:0)+-700+(u?0:-1100),r,e):A.length==2?null:this.result((n[0]?-700:0)+-200+-1100,n,e)}result(e,A,i){let n=[],o=0;for(let r of A){let s=r+(this.astral?lc(Bs(i,r)):1);o&&n[o-1]==r?n[o-1]=s:(n[o++]=r,n[o++]=s)}return this.ret(e-i.length,n)}},wN=class{constructor(e){this.pattern=e,this.matched=[],this.score=0,this.folded=e.toLowerCase()}match(e){if(e.length<this.pattern.length)return null;let A=e.slice(0,this.pattern.length),i=A==this.pattern?0:A.toLowerCase()==this.folded?-200:null;return i==null?null:(this.matched=[0,A.length],this.score=i+(e.length==this.pattern.length?0:-100),this)}},As=qe.define({combine(t){return Wr(t,{activateOnTyping:!0,activateOnCompletion:()=>!1,activateOnTypingDelay:100,selectOnOpen:!0,override:null,closeOnBlur:!0,maxRenderedOptions:100,defaultKeymap:!0,tooltipClass:()=>"",optionClass:()=>"",aboveCursor:!1,icons:!0,addToOptions:[],positionInfo:c_A,filterStrict:!1,compareCompletions:(e,A)=>e.label.localeCompare(A.label),interactionDelay:75,updateSyncTime:100},{defaultKeymap:(e,A)=>e&&A,closeOnBlur:(e,A)=>e&&A,icons:(e,A)=>e&&A,tooltipClass:(e,A)=>i=>loA(e(i),A(i)),optionClass:(e,A)=>i=>loA(e(i),A(i)),addToOptions:(e,A)=>e.concat(A),filterStrict:(e,A)=>e||A})}});function loA(t,e){return t?e?t+" "+e:t:e}function c_A(t,e,A,i,n,o){let r=t.textDirection==lo.RTL,s=r,a=!1,c="top",l,I,C=e.left-n.left,d=n.right-e.right,B=i.right-i.left,E=i.bottom-i.top;if(s&&C<Math.min(B,d)?s=!1:!s&&d<Math.min(B,C)&&(s=!0),B<=(s?C:d))l=Math.max(n.top,Math.min(A.top,n.bottom-E))-e.top,I=Math.min(400,s?C:d);else{a=!0,I=Math.min(400,(r?e.right:n.right-e.left)-30);let D=n.bottom-e.bottom;D>=E||D>e.top?l=A.bottom-e.top:(c="bottom",l=e.bottom-A.top)}let h=(e.bottom-e.top)/o.offsetHeight,u=(e.right-e.left)/o.offsetWidth;return{style:`${c}: ${l/h}px; max-width: ${I/u}px`,class:"cm-completionInfo-"+(a?r?"left-narrow":"right-narrow":s?"left":"right")}}function l_A(t){let e=t.addToOptions.slice();return t.icons&&e.push({render(A){let i=document.createElement("div");return i.classList.add("cm-completionIcon"),A.type&&i.classList.add(...A.type.split(/\s+/g).map(n=>"cm-completionIcon-"+n)),i.setAttribute("aria-hidden","true"),i},position:20}),e.push({render(A,i,n,o){let r=document.createElement("span");r.className="cm-completionLabel";let s=A.displayLabel||A.label,a=0;for(let c=0;c<o.length;){let l=o[c++],I=o[c++];l>a&&r.appendChild(document.createTextNode(s.slice(a,l)));let C=r.appendChild(document.createElement("span"));C.appendChild(document.createTextNode(s.slice(l,I))),C.className="cm-completionMatchedText",a=I}return a<s.length&&r.appendChild(document.createTextNode(s.slice(a))),r},position:50},{render(A){if(!A.detail)return null;let i=document.createElement("span");return i.className="cm-completionDetail",i.textContent=A.detail,i},position:80}),e.sort((A,i)=>A.position-i.position).map(A=>A.render)}function fN(t,e,A){if(t<=A)return{from:0,to:t};if(e<0&&(e=0),e<=t>>1){let n=Math.floor(e/A);return{from:n*A,to:(n+1)*A}}let i=Math.floor((t-e)/A);return{from:t-(i+1)*A,to:t-i*A}}var DN=class{constructor(e,A,i){this.view=e,this.stateField=A,this.applyCompletion=i,this.info=null,this.infoDestroy=null,this.placeInfoReq={read:()=>this.measureInfo(),write:a=>this.placeInfo(a),key:this},this.space=null,this.currentClass="";let n=e.state.field(A),{options:o,selected:r}=n.open,s=e.state.facet(As);this.optionContent=l_A(s),this.optionClass=s.optionClass,this.tooltipClass=s.tooltipClass,this.range=fN(o.length,r,s.maxRenderedOptions),this.dom=document.createElement("div"),this.dom.className="cm-tooltip-autocomplete",this.updateTooltipClass(e.state),this.dom.addEventListener("mousedown",a=>{let{options:c}=e.state.field(A).open;for(let l=a.target,I;l&&l!=this.dom;l=l.parentNode)if(l.nodeName=="LI"&&(I=/-(\d+)$/.exec(l.id))&&+I[1]<c.length){this.applyCompletion(e,c[+I[1]]),a.preventDefault();return}}),this.dom.addEventListener("focusout",a=>{let c=e.state.field(this.stateField,!1);c&&c.tooltip&&e.state.facet(As).closeOnBlur&&a.relatedTarget!=e.contentDOM&&e.dispatch({effects:k3.of(null)})}),this.showOptions(o,n.id)}mount(){this.updateSel()}showOptions(e,A){this.list&&this.list.remove(),this.list=this.dom.appendChild(this.createListBox(e,A,this.range)),this.list.addEventListener("scroll",()=>{this.info&&this.view.requestMeasure(this.placeInfoReq)})}update(e){var A;let i=e.state.field(this.stateField),n=e.startState.field(this.stateField);if(this.updateTooltipClass(e.state),i!=n){let{options:o,selected:r,disabled:s}=i.open;(!n.open||n.open.options!=o)&&(this.range=fN(o.length,r,e.state.facet(As).maxRenderedOptions),this.showOptions(o,i.id)),this.updateSel(),s!=((A=n.open)===null||A===void 0?void 0:A.disabled)&&this.dom.classList.toggle("cm-tooltip-autocomplete-disabled",!!s)}}updateTooltipClass(e){let A=this.tooltipClass(e);if(A!=this.currentClass){for(let i of this.currentClass.split(" "))i&&this.dom.classList.remove(i);for(let i of A.split(" "))i&&this.dom.classList.add(i);this.currentClass=A}}positioned(e){this.space=e,this.info&&this.view.requestMeasure(this.placeInfoReq)}updateSel(){let e=this.view.state.field(this.stateField),A=e.open;if((A.selected>-1&&A.selected<this.range.from||A.selected>=this.range.to)&&(this.range=fN(A.options.length,A.selected,this.view.state.facet(As).maxRenderedOptions),this.showOptions(A.options,e.id)),this.updateSelectedOption(A.selected)){this.destroyInfo();let{completion:i}=A.options[A.selected],{info:n}=i;if(!n)return;let o=typeof n=="string"?document.createTextNode(n):n(i);if(!o)return;"then"in o?o.then(r=>{r&&this.view.state.field(this.stateField,!1)==e&&this.addInfoPane(r,i)}).catch(r=>Xr(this.view.state,r,"completion info")):this.addInfoPane(o,i)}}addInfoPane(e,A){this.destroyInfo();let i=this.info=document.createElement("div");if(i.className="cm-tooltip cm-completionInfo",e.nodeType!=null)i.appendChild(e),this.infoDestroy=null;else{let{dom:n,destroy:o}=e;i.appendChild(n),this.infoDestroy=o||null}this.dom.appendChild(i),this.view.requestMeasure(this.placeInfoReq)}updateSelectedOption(e){let A=null;for(let i=this.list.firstChild,n=this.range.from;i;i=i.nextSibling,n++)i.nodeName!="LI"||!i.id?n--:n==e?i.hasAttribute("aria-selected")||(i.setAttribute("aria-selected","true"),A=i):i.hasAttribute("aria-selected")&&i.removeAttribute("aria-selected");return A&&I_A(this.list,A),A}measureInfo(){let e=this.dom.querySelector("[aria-selected]");if(!e||!this.info)return null;let A=this.dom.getBoundingClientRect(),i=this.info.getBoundingClientRect(),n=e.getBoundingClientRect(),o=this.space;if(!o){let r=this.dom.ownerDocument.documentElement;o={left:0,top:0,right:r.clientWidth,bottom:r.clientHeight}}return n.top>Math.min(o.bottom,A.bottom)-10||n.bottom<Math.max(o.top,A.top)+10?null:this.view.state.facet(As).positionInfo(this.view,A,n,i,o,this.dom)}placeInfo(e){this.info&&(e?(e.style&&(this.info.style.cssText=e.style),this.info.className="cm-tooltip cm-completionInfo "+(e.class||"")):this.info.style.cssText="top: -1e6px")}createListBox(e,A,i){let n=document.createElement("ul");n.id=A,n.setAttribute("role","listbox"),n.setAttribute("aria-expanded","true"),n.setAttribute("aria-label",this.view.state.phrase("Completions")),n.addEventListener("mousedown",r=>{r.target==n&&r.preventDefault()});let o=null;for(let r=i.from;r<i.to;r++){let{completion:s,match:a}=e[r],{section:c}=s;if(c){let C=typeof c=="string"?c:c.name;if(C!=o&&(r>i.from||i.from==0))if(o=C,typeof c!="string"&&c.header)n.appendChild(c.header(c));else{let d=n.appendChild(document.createElement("completion-section"));d.textContent=C}}let l=n.appendChild(document.createElement("li"));l.id=A+"-"+r,l.setAttribute("role","option");let I=this.optionClass(s);I&&(l.className=I);for(let C of this.optionContent){let d=C(s,this.view.state,this.view,a);d&&l.appendChild(d)}}return i.from&&n.classList.add("cm-completionListIncompleteTop"),i.to<e.length&&n.classList.add("cm-completionListIncompleteBottom"),n}destroyInfo(){this.info&&(this.infoDestroy&&this.infoDestroy(),this.info.remove(),this.info=null)}destroy(){this.destroyInfo()}};function g_A(t,e){return A=>new DN(A,t,e)}function I_A(t,e){let A=t.getBoundingClientRect(),i=e.getBoundingClientRect(),n=A.height/t.offsetHeight;i.top<A.top?t.scrollTop-=(A.top-i.top)/n:i.bottom>A.bottom&&(t.scrollTop+=(i.bottom-A.bottom)/n)}function goA(t){return(t.boost||0)*100+(t.apply?10:0)+(t.info?5:0)+(t.type?1:0)}function C_A(t,e){let A=[],i=null,n=c=>{A.push(c);let{section:l}=c.completion;if(l){i||(i=[]);let I=typeof l=="string"?l:l.name;i.some(C=>C.name==I)||i.push(typeof l=="string"?{name:I}:l)}},o=e.facet(As);for(let c of t)if(c.hasResult()){let l=c.result.getMatch;if(c.result.filter===!1)for(let I of c.result.options)n(new KD(I,c.source,l?l(I):[],1e9-A.length));else{let I=e.sliceDoc(c.from,c.to),C,d=o.filterStrict?new wN(I):new pN(I);for(let B of c.result.options)if(C=d.match(B.label)){let E=B.displayLabel?l?l(B,C.matched):[]:C.matched;n(new KD(B,c.source,E,C.score+(B.boost||0)))}}}if(i){let c=Object.create(null),l=0,I=(C,d)=>{var B,E;return((B=C.rank)!==null&&B!==void 0?B:1e9)-((E=d.rank)!==null&&E!==void 0?E:1e9)||(C.name<d.name?-1:1)};for(let C of i.sort(I))l-=1e5,c[C.name]=l;for(let C of A){let{section:d}=C.completion;d&&(C.score+=c[typeof d=="string"?d:d.name])}}let r=[],s=null,a=o.compareCompletions;for(let c of A.sort((l,I)=>I.score-l.score||a(l.completion,I.completion))){let l=c.completion;!s||s.label!=l.label||s.detail!=l.detail||s.type!=null&&l.type!=null&&s.type!=l.type||s.apply!=l.apply||s.boost!=l.boost?r.push(c):goA(c.completion)>goA(s)&&(r[r.length-1]=c),s=c.completion}return r}var yN=class t{constructor(e,A,i,n,o,r){this.options=e,this.attrs=A,this.tooltip=i,this.timestamp=n,this.selected=o,this.disabled=r}setSelected(e,A){return e==this.selected||e>=this.options.length?this:new t(this.options,IoA(A,e),this.tooltip,this.timestamp,e,this.disabled)}static build(e,A,i,n,o,r){if(n&&!r&&e.some(c=>c.isPending))return n.setDisabled();let s=C_A(e,A);if(!s.length)return n&&e.some(c=>c.isPending)?n.setDisabled():null;let a=A.facet(As).selectOnOpen?0:-1;if(n&&n.selected!=a&&n.selected!=-1){let c=n.options[n.selected].completion;for(let l=0;l<s.length;l++)if(s[l].completion==c){a=l;break}}return new t(s,IoA(i,a),{pos:e.reduce((c,l)=>l.hasResult()?Math.min(c,l.from):c,1e8),create:u_A,above:o.aboveCursor},n?n.timestamp:Date.now(),a,!1)}map(e){return new t(this.options,this.attrs,Object.assign(Object.assign({},this.tooltip),{pos:e.mapPos(this.tooltip.pos)}),this.timestamp,this.selected,this.disabled)}setDisabled(){return new t(this.options,this.attrs,this.tooltip,this.timestamp,this.selected,!0)}},vN=class t{constructor(e,A,i){this.active=e,this.id=A,this.open=i}static start(){return new t(Q_A,"cm-ac-"+Math.floor(Math.random()*2e6).toString(36),null)}update(e){let{state:A}=e,i=A.facet(As),o=(i.override||A.languageDataAt("autocomplete",UC(A)).map(a_A)).map(a=>(this.active.find(l=>l.source==a)||new O0(a,this.active.some(l=>l.state!=0)?1:0)).update(e,i));o.length==this.active.length&&o.every((a,c)=>a==this.active[c])&&(o=this.active);let r=this.open,s=e.effects.some(a=>a.is(MN));r&&e.docChanged&&(r=r.map(e.changes)),e.selection||o.some(a=>a.hasResult()&&e.changes.touchesRange(a.from,a.to))||!d_A(o,this.active)||s?r=yN.build(o,A,this.id,r,i,s):r&&r.disabled&&!o.some(a=>a.isPending)&&(r=null),!r&&o.every(a=>!a.isPending)&&o.some(a=>a.hasResult())&&(o=o.map(a=>a.hasResult()?new O0(a.source,0):a));for(let a of e.effects)a.is(uoA)&&(r=r&&r.setSelected(a.value,this.id));return o==this.active&&r==this.open?this:new t(o,this.id,r)}get tooltip(){return this.open?this.open.tooltip:null}get attrs(){return this.open?this.open.attrs:this.active.length?B_A:E_A}};function d_A(t,e){if(t==e)return!0;for(let A=0,i=0;;){for(;A<t.length&&!t[A].hasResult();)A++;for(;i<e.length&&!e[i].hasResult();)i++;let n=A==t.length,o=i==e.length;if(n||o)return n==o;if(t[A++].result!=e[i++].result)return!1}}var B_A={"aria-autocomplete":"list"},E_A={};function IoA(t,e){let A={"aria-autocomplete":"list","aria-haspopup":"listbox","aria-controls":t};return e>-1&&(A["aria-activedescendant"]=t+"-"+e),A}var Q_A=[];function hoA(t,e){if(t.isUserEvent("input.complete")){let i=t.annotation(QoA);if(i&&e.activateOnCompletion(i))return 12}let A=t.isUserEvent("input.type");return A&&e.activateOnTyping?5:A?1:t.isUserEvent("delete.backward")?2:t.selection?8:t.docChanged?16:0}var O0=class t{constructor(e,A,i=!1){this.source=e,this.state=A,this.explicit=i}hasResult(){return!1}get isPending(){return this.state==1}update(e,A){let i=hoA(e,A),n=this;(i&8||i&16&&this.touches(e))&&(n=new t(n.source,0)),i&4&&n.state==0&&(n=new t(this.source,1)),n=n.updateFor(e,i);for(let o of e.effects)if(o.is(YD))n=new t(n.source,1,o.value);else if(o.is(k3))n=new t(n.source,0);else if(o.is(MN))for(let r of o.value)r.source==n.source&&(n=r);return n}updateFor(e,A){return this.map(e.changes)}map(e){return this}touches(e){return e.changes.touchesRange(UC(e.state))}},JD=class t extends O0{constructor(e,A,i,n,o,r){super(e,3,A),this.limit=i,this.result=n,this.from=o,this.to=r}hasResult(){return!0}updateFor(e,A){var i;if(!(A&3))return this.map(e.changes);let n=this.result;n.map&&!e.changes.empty&&(n=n.map(n,e.changes));let o=e.changes.mapPos(this.from),r=e.changes.mapPos(this.to,1),s=UC(e.state);if(s>r||!n||A&2&&(UC(e.startState)==this.from||s<this.limit))return new O0(this.source,A&4?1:0);let a=e.changes.mapPos(this.limit);return h_A(n.validFor,e.state,o,r)?new t(this.source,this.explicit,a,n,o,r):n.update&&(n=n.update(n,o,r,new UD(e.state,s,!1)))?new t(this.source,this.explicit,a,n,n.from,(i=n.to)!==null&&i!==void 0?i:UC(e.state)):new O0(this.source,1,this.explicit)}map(e){return e.empty?this:(this.result.map?this.result.map(this.result,e):this.result)?new t(this.source,this.explicit,e.mapPos(this.limit),this.result,e.mapPos(this.from),e.mapPos(this.to,1)):new O0(this.source,0)}touches(e){return e.changes.touchesRange(this.from,this.to)}};function h_A(t,e,A,i){if(!t)return!1;let n=e.sliceDoc(A,i);return typeof t=="function"?t(n,A,i,e):EoA(t,!0).test(n)}var MN=Pi.define({map(t,e){return t.map(A=>A.map(e))}}),uoA=Pi.define(),Na=Ar.define({create(){return vN.start()},update(t,e){return t.update(e)},provide:t=>[GE.from(t,e=>e.tooltip),qt.contentAttributes.from(t,e=>e.attrs)]});function kN(t,e){let A=e.completion.apply||e.completion.label,i=t.state.field(Na).active.find(n=>n.source==e.source);return i instanceof JD?(typeof A=="string"?t.dispatch(Object.assign(Object.assign({},s_A(t.state,A,i.from,i.to)),{annotations:QoA.of(e.completion)})):A(t,e.completion,i.from,i.to),!0):!1}var u_A=g_A(Na,kN);function GD(t,e="option"){return A=>{let i=A.state.field(Na,!1);if(!i||!i.open||i.open.disabled||Date.now()-i.open.timestamp<A.state.facet(As).interactionDelay)return!1;let n=1,o;e=="page"&&(o=BF(A,i.open.tooltip))&&(n=Math.max(2,Math.floor(o.dom.offsetHeight/o.dom.querySelector("li").offsetHeight)-1));let{length:r}=i.open.options,s=i.open.selected>-1?i.open.selected+n*(t?1:-1):t?0:r-1;return s<0?s=e=="page"?0:r-1:s>=r&&(s=e=="page"?r-1:0),A.dispatch({effects:uoA.of(s)}),!0}}var f_A=t=>{let e=t.state.field(Na,!1);return t.state.readOnly||!e||!e.open||e.open.selected<0||e.open.disabled||Date.now()-e.open.timestamp<t.state.facet(As).interactionDelay?!1:kN(t,e.open.options[e.open.selected])},CoA=t=>t.state.field(Na,!1)?(t.dispatch({effects:YD.of(!0)}),!0):!1,m_A=t=>{let e=t.state.field(Na,!1);return!e||!e.active.some(A=>A.state!=0)?!1:(t.dispatch({effects:k3.of(null)}),!0)},bN=class{constructor(e,A){this.active=e,this.context=A,this.time=Date.now(),this.updates=[],this.done=void 0}},p_A=50,w_A=1e3,D_A=go.fromClass(class{constructor(t){this.view=t,this.debounceUpdate=-1,this.running=[],this.debounceAccept=-1,this.pendingStart=!1,this.composing=0;for(let e of t.state.field(Na).active)e.isPending&&this.startQuery(e)}update(t){let e=t.state.field(Na),A=t.state.facet(As);if(!t.selectionSet&&!t.docChanged&&t.startState.field(Na)==e)return;let i=t.transactions.some(o=>{let r=hoA(o,A);return r&8||(o.selection||o.docChanged)&&!(r&3)});for(let o=0;o<this.running.length;o++){let r=this.running[o];if(i||r.context.abortOnDocChange&&t.docChanged||r.updates.length+t.transactions.length>p_A&&Date.now()-r.time>w_A){for(let s of r.context.abortListeners)try{s()}catch(a){Xr(this.view.state,a)}r.context.abortListeners=null,this.running.splice(o--,1)}else r.updates.push(...t.transactions)}this.debounceUpdate>-1&&clearTimeout(this.debounceUpdate),t.transactions.some(o=>o.effects.some(r=>r.is(YD)))&&(this.pendingStart=!0);let n=this.pendingStart?50:A.activateOnTypingDelay;if(this.debounceUpdate=e.active.some(o=>o.isPending&&!this.running.some(r=>r.active.source==o.source))?setTimeout(()=>this.startUpdate(),n):-1,this.composing!=0)for(let o of t.transactions)o.isUserEvent("input.type")?this.composing=2:this.composing==2&&o.selection&&(this.composing=3)}startUpdate(){this.debounceUpdate=-1,this.pendingStart=!1;let{state:t}=this.view,e=t.field(Na);for(let A of e.active)A.isPending&&!this.running.some(i=>i.active.source==A.source)&&this.startQuery(A);this.running.length&&e.open&&e.open.disabled&&(this.debounceAccept=setTimeout(()=>this.accept(),this.view.state.facet(As).updateSyncTime))}startQuery(t){let{state:e}=this.view,A=UC(e),i=new UD(e,A,t.explicit,this.view),n=new bN(t,i);this.running.push(n),Promise.resolve(t.source(i)).then(o=>{n.context.aborted||(n.done=o||null,this.scheduleAccept())},o=>{this.view.dispatch({effects:k3.of(null)}),Xr(this.view.state,o)})}scheduleAccept(){this.running.every(t=>t.done!==void 0)?this.accept():this.debounceAccept<0&&(this.debounceAccept=setTimeout(()=>this.accept(),this.view.state.facet(As).updateSyncTime))}accept(){var t;this.debounceAccept>-1&&clearTimeout(this.debounceAccept),this.debounceAccept=-1;let e=[],A=this.view.state.facet(As),i=this.view.state.field(Na);for(let n=0;n<this.running.length;n++){let o=this.running[n];if(o.done===void 0)continue;if(this.running.splice(n--,1),o.done){let s=UC(o.updates.length?o.updates[0].startState:this.view.state),a=Math.min(s,o.done.from+(o.active.explicit?0:1)),c=new JD(o.active.source,o.active.explicit,a,o.done,o.done.from,(t=o.done.to)!==null&&t!==void 0?t:s);for(let l of o.updates)c=c.update(l,A);if(c.hasResult()){e.push(c);continue}}let r=i.active.find(s=>s.source==o.active.source);if(r&&r.isPending)if(o.done==null){let s=new O0(o.active.source,0);for(let a of o.updates)s=s.update(a,A);s.isPending||e.push(s)}else this.startQuery(r)}(e.length||i.open&&i.open.disabled)&&this.view.dispatch({effects:MN.of(e)})}},{eventHandlers:{blur(t){let e=this.view.state.field(Na,!1);if(e&&e.tooltip&&this.view.state.facet(As).closeOnBlur){let A=e.open&&BF(this.view,e.open.tooltip);(!A||!A.dom.contains(t.relatedTarget))&&setTimeout(()=>this.view.dispatch({effects:k3.of(null)}),10)}},compositionstart(){this.composing=1},compositionend(){this.composing==3&&setTimeout(()=>this.view.dispatch({effects:YD.of(!1)}),20),this.composing=0}}}),y_A=typeof navigator=="object"&&/Win/.test(navigator.platform),v_A=Dl.highest(qt.domEventHandlers({keydown(t,e){let A=e.state.field(Na,!1);if(!A||!A.open||A.open.disabled||A.open.selected<0||t.key.length>1||t.ctrlKey&&!(y_A&&t.altKey)||t.metaKey)return!1;let i=A.open.options[A.open.selected],n=A.active.find(r=>r.source==i.source),o=i.completion.commitCharacters||n.result.commitCharacters;return o&&o.indexOf(t.key)>-1&&kN(e,i),!1}})),b_A=qt.baseTheme({".cm-tooltip.cm-tooltip-autocomplete":{"& > ul":{fontFamily:"monospace",whiteSpace:"nowrap",overflow:"hidden auto",maxWidth_fallback:"700px",maxWidth:"min(700px, 95vw)",minWidth:"250px",maxHeight:"10em",height:"100%",listStyle:"none",margin:0,padding:0,"& > li, & > completion-section":{padding:"1px 3px",lineHeight:1.2},"& > li":{overflowX:"hidden",textOverflow:"ellipsis",cursor:"pointer"},"& > completion-section":{display:"list-item",borderBottom:"1px solid silver",paddingLeft:"0.5em",opacity:.7}}},"&light .cm-tooltip-autocomplete ul li[aria-selected]":{background:"#17c",color:"white"},"&light .cm-tooltip-autocomplete-disabled ul li[aria-selected]":{background:"#777"},"&dark .cm-tooltip-autocomplete ul li[aria-selected]":{background:"#347",color:"white"},"&dark .cm-tooltip-autocomplete-disabled ul li[aria-selected]":{background:"#444"},".cm-completionListIncompleteTop:before, .cm-completionListIncompleteBottom:after":{content:'"\xB7\xB7\xB7"',opacity:.5,display:"block",textAlign:"center"},".cm-tooltip.cm-completionInfo":{position:"absolute",padding:"3px 9px",width:"max-content",maxWidth:"400px",boxSizing:"border-box",whiteSpace:"pre-line"},".cm-completionInfo.cm-completionInfo-left":{right:"100%"},".cm-completionInfo.cm-completionInfo-right":{left:"100%"},".cm-completionInfo.cm-completionInfo-left-narrow":{right:"30px"},".cm-completionInfo.cm-completionInfo-right-narrow":{left:"30px"},"&light .cm-snippetField":{backgroundColor:"#00000022"},"&dark .cm-snippetField":{backgroundColor:"#ffffff22"},".cm-snippetFieldPosition":{verticalAlign:"text-top",width:0,height:"1.15em",display:"inline-block",margin:"0 -0.7px -.7em",borderLeft:"1.4px dotted #888"},".cm-completionMatchedText":{textDecoration:"underline"},".cm-completionDetail":{marginLeft:"0.5em",fontStyle:"italic"},".cm-completionIcon":{fontSize:"90%",width:".8em",display:"inline-block",textAlign:"center",paddingRight:".6em",opacity:"0.6",boxSizing:"content-box"},".cm-completionIcon-function, .cm-completionIcon-method":{"&:after":{content:"'\u0192'"}},".cm-completionIcon-class":{"&:after":{content:"'\u25CB'"}},".cm-completionIcon-interface":{"&:after":{content:"'\u25CC'"}},".cm-completionIcon-variable":{"&:after":{content:"'\u{1D465}'"}},".cm-completionIcon-constant":{"&:after":{content:"'\u{1D436}'"}},".cm-completionIcon-type":{"&:after":{content:"'\u{1D461}'"}},".cm-completionIcon-enum":{"&:after":{content:"'\u222A'"}},".cm-completionIcon-property":{"&:after":{content:"'\u25A1'"}},".cm-completionIcon-keyword":{"&:after":{content:"'\u{1F511}\uFE0E'"}},".cm-completionIcon-namespace":{"&:after":{content:"'\u25A2'"}},".cm-completionIcon-text":{"&:after":{content:"'abc'",fontSize:"50%",verticalAlign:"middle"}}});var S3={brackets:["(","[","{","'",'"'],before:")]}:;>",stringPrefixes:[]},GC=Pi.define({map(t,e){let A=e.mapPos(t,-1,Is.TrackAfter);return A??void 0}}),SN=new class extends wl{};SN.startSide=1;SN.endSide=-1;var foA=Ar.define({create(){return co.empty},update(t,e){if(t=t.map(e.changes),e.selection){let A=e.state.doc.lineAt(e.selection.main.head);t=t.update({filter:i=>i>=A.from&&i<=A.to})}for(let A of e.effects)A.is(GC)&&(t=t.update({add:[SN.range(A.value,A.value+1)]}));return t}});function moA(){return[k_A,foA]}var mN="()[]{}<>\xAB\xBB\xBB\xAB\uFF3B\uFF3D\uFF5B\uFF5D";function poA(t){for(let e=0;e<mN.length;e+=2)if(mN.charCodeAt(e)==t)return mN.charAt(e+1);return T4(t<128?t:t+1)}function woA(t,e){return t.languageDataAt("closeBrackets",e)[0]||S3}var M_A=typeof navigator=="object"&&/Android\b/.test(navigator.userAgent),k_A=qt.inputHandler.of((t,e,A,i)=>{if((M_A?t.composing:t.compositionStarted)||t.state.readOnly)return!1;let n=t.state.selection.main;if(i.length>2||i.length==2&&lc(Bs(i,0))==1||e!=n.from||A!=n.to)return!1;let o=R_A(t.state,i);return o?(t.dispatch(o),!0):!1}),S_A=({state:t,dispatch:e})=>{if(t.readOnly)return!1;let i=woA(t,t.selection.main.head).brackets||S3.brackets,n=null,o=t.changeByRange(r=>{if(r.empty){let s=x_A(t.doc,r.head);for(let a of i)if(a==s&&TD(t.doc,r.head)==poA(Bs(a,0)))return{changes:{from:r.head-a.length,to:r.head+a.length},range:ae.cursor(r.head-a.length)}}return{range:n=r}});return n||e(t.update(o,{scrollIntoView:!0,userEvent:"delete.backward"})),!n},DoA=[{key:"Backspace",run:S_A}];function R_A(t,e){let A=woA(t,t.selection.main.head),i=A.brackets||S3.brackets;for(let n of i){let o=poA(Bs(n,0));if(e==n)return o==n?N_A(t,n,i.indexOf(n+n+n)>-1,A):L_A(t,n,o,A.before||S3.before);if(e==o&&yoA(t,t.selection.main.from))return F_A(t,n,o)}return null}function yoA(t,e){let A=!1;return t.field(foA).between(0,t.doc.length,i=>{i==e&&(A=!0)}),A}function TD(t,e){let A=t.sliceString(e,e+2);return A.slice(0,lc(Bs(A,0)))}function x_A(t,e){let A=t.sliceString(e-2,e);return lc(Bs(A,0))==A.length?A:A.slice(1)}function L_A(t,e,A,i){let n=null,o=t.changeByRange(r=>{if(!r.empty)return{changes:[{insert:e,from:r.from},{insert:A,from:r.to}],effects:GC.of(r.to+e.length),range:ae.range(r.anchor+e.length,r.head+e.length)};let s=TD(t.doc,r.head);return!s||/\s/.test(s)||i.indexOf(s)>-1?{changes:{insert:e+A,from:r.head},effects:GC.of(r.head+e.length),range:ae.cursor(r.head+e.length)}:{range:n=r}});return n?null:t.update(o,{scrollIntoView:!0,userEvent:"input.type"})}function F_A(t,e,A){let i=null,n=t.changeByRange(o=>o.empty&&TD(t.doc,o.head)==A?{changes:{from:o.head,to:o.head+A.length,insert:A},range:ae.cursor(o.head+A.length)}:i={range:o});return i?null:t.update(n,{scrollIntoView:!0,userEvent:"input.type"})}function N_A(t,e,A,i){let n=i.stringPrefixes||S3.stringPrefixes,o=null,r=t.changeByRange(s=>{if(!s.empty)return{changes:[{insert:e,from:s.from},{insert:e,from:s.to}],effects:GC.of(s.to+e.length),range:ae.range(s.anchor+e.length,s.head+e.length)};let a=s.head,c=TD(t.doc,a),l;if(c==e){if(doA(t,a))return{changes:{insert:e+e,from:a},effects:GC.of(a+e.length),range:ae.cursor(a+e.length)};if(yoA(t,a)){let C=A&&t.sliceDoc(a,a+e.length*3)==e+e+e?e+e+e:e;return{changes:{from:a,to:a+C.length,insert:C},range:ae.cursor(a+C.length)}}}else{if(A&&t.sliceDoc(a-2*e.length,a)==e+e&&(l=BoA(t,a-2*e.length,n))>-1&&doA(t,l))return{changes:{insert:e+e+e+e,from:a},effects:GC.of(a+e.length),range:ae.cursor(a+e.length)};if(t.charCategorizer(a)(c)!=ao.Word&&BoA(t,a,n)>-1&&!__A(t,a,e,n))return{changes:{insert:e+e,from:a},effects:GC.of(a+e.length),range:ae.cursor(a+e.length)}}return{range:o=s}});return o?null:t.update(r,{scrollIntoView:!0,userEvent:"input.type"})}function doA(t,e){let A=$r(t).resolveInner(e+1);return A.parent&&A.from==e}function __A(t,e,A,i){let n=$r(t).resolveInner(e,-1),o=i.reduce((r,s)=>Math.max(r,s.length),0);for(let r=0;r<5;r++){let s=t.sliceDoc(n.from,Math.min(n.to,n.from+A.length+o)),a=s.indexOf(A);if(!a||a>-1&&i.indexOf(s.slice(0,a))>-1){let l=n.firstChild;for(;l&&l.from==n.from&&l.to-l.from>A.length+a;){if(t.sliceDoc(l.to-A.length,l.to)==A)return!1;l=l.firstChild}return!0}let c=n.to==e&&n.parent;if(!c)break;n=c}return!1}function BoA(t,e,A){let i=t.charCategorizer(e);if(i(t.sliceDoc(e-1,e))!=ao.Word)return e;for(let n of A){let o=e-n.length;if(t.sliceDoc(o,e)==n&&i(t.sliceDoc(o-1,o))!=ao.Word)return o}return-1}function voA(t={}){return[v_A,Na,As.of(t),D_A,G_A,b_A]}var RN=[{key:"Ctrl-Space",run:CoA},{mac:"Alt-`",run:CoA},{key:"Escape",run:m_A},{key:"ArrowDown",run:GD(!0)},{key:"ArrowUp",run:GD(!1)},{key:"PageDown",run:GD(!0,"page")},{key:"PageUp",run:GD(!1,"page")},{key:"Enter",run:f_A}],G_A=Dl.highest(_E.computeN([As],t=>t.facet(As).defaultKeymap?[RN]:[]));function U_A(t,e=t.state){let A=new Set;for(let{from:i,to:n}of t.visibleRanges){let o=i;for(;o<=n;){let r=e.doc.lineAt(o);A.has(r)||A.add(r),o=r.to+1}}return A}function xN(t){let e=t.selection.main.head;return t.doc.lineAt(e)}function boA(t,e){let A=0;A:for(let i=0;i<t.length;i++)switch(t[i]){case" ":case"\xA0":{A+=1;continue A}case"	":{A+=e-A%e;continue A}case"\r":continue A;default:break A}return A}var zD=qe.define({combine(t){return Wr(t,{highlightActiveBlock:!0,hideFirstIndent:!1,markerType:"fullScope",thickness:1})}}),LN=class{constructor(e,A,i,n){this.lines=e,this.state=A,this.map=new Map,this.unitWidth=i,this.markerType=n;for(let o of this.lines)this.add(o);this.state.facet(zD).highlightActiveBlock&&this.findAndSetActiveLines()}has(e){return this.map.has(typeof e=="number"?e:e.number)}get(e){let A=this.map.get(typeof e=="number"?e:e.number);if(!A)throw new Error("Line not found in indentation map");return A}set(e,A,i){let n=!e.text.trim().length,o={line:e,col:A,level:i,empty:n};return this.map.set(o.line.number,o),o}add(e){if(this.has(e))return this.get(e);if(!e.length||!e.text.trim().length){if(e.number===1)return this.set(e,0,0);if(e.number===this.state.doc.lines){let r=this.closestNonEmpty(e,-1);return this.set(e,0,r.level)}let n=this.closestNonEmpty(e,-1),o=this.closestNonEmpty(e,1);return n.level>=o.level&&this.markerType!=="codeOnly"?this.set(e,0,n.level):n.empty&&n.level===0&&o.level!==0?this.set(e,0,0):o.level>n.level?this.set(e,0,n.level+1):this.set(e,0,o.level)}let A=boA(e.text,this.state.tabSize),i=Math.floor(A/this.unitWidth);return this.set(e,A,i)}closestNonEmpty(e,A){let i=e.number+A;for(;A===-1?i>=1:i<=this.state.doc.lines;){if(this.has(i)){let r=this.get(i);if(!r.empty)return r}let o=this.state.doc.line(i);if(o.text.trim().length){let r=boA(o.text,this.state.tabSize),s=Math.floor(r/this.unitWidth);return this.set(o,r,s)}i+=A}let n=this.state.doc.line(A===-1?1:this.state.doc.lines);return this.set(n,0,0)}findAndSetActiveLines(){let e=xN(this.state);if(!this.has(e))return;let A=this.get(e);if(this.has(A.line.number+1)){let o=this.get(A.line.number+1);o.level>A.level&&(A=o)}if(this.has(A.line.number-1)){let o=this.get(A.line.number-1);o.level>A.level&&(A=o)}if(A.level===0)return;A.active=A.level;let i,n;for(i=A.line.number;i>1;i--){if(!this.has(i-1))continue;let o=this.get(i-1);if(o.level<A.level)break;o.active=A.level}for(n=A.line.number;n<this.state.doc.lines;n++){if(!this.has(n+1))continue;let o=this.get(n+1);if(o.level<A.level)break;o.active=A.level}}};function K_A(t){let e={light:"#F0F1F2",dark:"#2B3245",activeLight:"#E4E5E6",activeDark:"#3C445C"},A=e;return t&&(A=Object.assign(Object.assign({},e),t)),qt.baseTheme({"&light":{"--indent-marker-bg-color":A.light,"--indent-marker-active-bg-color":A.activeLight},"&dark":{"--indent-marker-bg-color":A.dark,"--indent-marker-active-bg-color":A.activeDark},".cm-line":{position:"relative"},".cm-indent-markers::before":{content:'""',position:"absolute",top:0,left:"2px",right:0,bottom:0,background:"var(--indent-markers)",pointerEvents:"none",zIndex:"-1"}})}function HD(t,e,A,i,n){return`${`repeating-linear-gradient(to right, var(${t}) 0 ${e}px, transparent ${e}px ${A}ch)`} ${i*A}.5ch/calc(${A*n}ch - 1px) no-repeat`}function Y_A(t,e,A,i,n){let{level:o,active:r}=t;if(n=n??i,A&&o===0)return[];let s=A?1:0,a=[];if(r!==void 0){let c=r-s-1;c>0&&a.push(HD("--indent-marker-bg-color",i,e,s,c)),a.push(HD("--indent-marker-active-bg-color",n,e,r-1,1)),r!==o&&a.push(HD("--indent-marker-bg-color",i,e,r,o-r))}else a.push(HD("--indent-marker-bg-color",i,e,s,o-s));return a.join(",")}var FN=class{constructor(e){this.view=e,this.unitWidth=Ml(e.state),this.currentLineNumber=xN(e.state).number,this.generate(e.state)}update(e){let A=Ml(e.state),i=A!==this.unitWidth;i&&(this.unitWidth=A);let n=xN(e.state).number,o=n!==this.currentLineNumber;this.currentLineNumber=n;let r=e.state.facet(zD).highlightActiveBlock&&o;(e.docChanged||e.viewportChanged||i||r)&&this.generate(e.state)}generate(e){let A=new ds,i=U_A(this.view,e),{hideFirstIndent:n,markerType:o,thickness:r,activeThickness:s}=e.facet(zD),a=new LN(i,e,this.unitWidth,o);for(let c of i){let l=a.get(c.number);if(!l?.level)continue;let I=Y_A(l,this.unitWidth,n,r,s);A.add(c.from,c.from,ut.line({class:"cm-indent-markers",attributes:{style:`--indent-markers: ${I}`}}))}this.decorations=A.finish()}};function MoA(t={}){return[zD.of(t),K_A(t.colors),go.fromClass(FN,{decorations:e=>e.decorations})]}var J_A=["mainAxis","crossAxis","fallbackPlacements","fallbackStrategy","fallbackAxisSideDirection","flipAlignment"],T_A=["mainAxis","crossAxis","limiter"];function zrA(t,e){if(t==null)return{};var A,i,n=function(r,s){if(r==null)return{};var a={};for(var c in r)if({}.hasOwnProperty.call(r,c)){if(s.indexOf(c)!==-1)continue;a[c]=r[c]}return a}(t,e);if(Object.getOwnPropertySymbols){var o=Object.getOwnPropertySymbols(t);for(i=0;i<o.length;i++)A=o[i],e.indexOf(A)===-1&&{}.propertyIsEnumerable.call(t,A)&&(n[A]=t[A])}return n}function koA(t,e){var A=Object.keys(t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(t);e&&(i=i.filter(function(n){return Object.getOwnPropertyDescriptor(t,n).enumerable})),A.push.apply(A,i)}return A}function fe(t){for(var e=1;e<arguments.length;e++){var A=arguments[e]!=null?arguments[e]:{};e%2?koA(Object(A),!0).forEach(function(i){r_(t,i,A[i])}):Object.getOwnPropertyDescriptors?Object.defineProperties(t,Object.getOwnPropertyDescriptors(A)):koA(Object(A)).forEach(function(i){Object.defineProperty(t,i,Object.getOwnPropertyDescriptor(A,i))})}return t}function SoA(t,e,A,i,n,o,r){try{var s=t[o](r),a=s.value}catch(c){return void A(c)}s.done?e(a):Promise.resolve(a).then(i,n)}function Yt(t){return function(){var e=this,A=arguments;return new Promise(function(i,n){var o=t.apply(e,A);function r(a){SoA(o,i,n,r,s,"next",a)}function s(a){SoA(o,i,n,r,s,"throw",a)}r(void 0)})}}function r_(t,e,A){return(e=function(i){var n=function(o,r){if(typeof o!="object"||!o)return o;var s=o[Symbol.toPrimitive];if(s!==void 0){var a=s.call(o,r||"default");if(typeof a!="object")return a;throw new TypeError("@@toPrimitive must return a primitive value.")}return(r==="string"?String:Number)(o)}(i,"string");return typeof n=="symbol"?n:n+""}(e))in t?Object.defineProperty(t,e,{value:A,enumerable:!0,configurable:!0,writable:!0}):t[e]=A,t}var RoA,NN,xoA,_N;typeof window<"u"&&((NN=(RoA=(_N=(xoA=window).__svelte)!==null&&_N!==void 0?_N:xoA.__svelte={}).v)!==null&&NN!==void 0?NN:RoA.v=new Set).add("5");var fQ=!1;fQ=!0;var Qs=Symbol(),mQ=Array.isArray,O_A=Array.prototype.indexOf,a_=Array.from,P_A=Object.defineProperty,$0=Object.getOwnPropertyDescriptor,PrA=Object.getOwnPropertyDescriptors,j_A=Object.prototype,q_A=Array.prototype,K_=Object.getPrototypeOf,LoA=Object.isExtensible;function R3(t){return typeof t=="function"}var FoA=()=>{};function V_A(t){return t()}function gy(t){for(var e=0;e<t.length;e++)t[e]()}var Ry=32,jrA=64,xl=256,c_=512,pc=1024,IQ=2048,Ad=4096,H1=8192,Y_=16384,Af=65536,Z_A=1<<17,W_A=1<<19,l_=1<<20,X_A=1<<21,Og=Symbol("$state"),qrA=Symbol("legacy props"),$_A=Symbol(""),VrA=new class extends Error{constructor(){super(...arguments),r_(this,"name","StaleReactionError"),r_(this,"message","The reaction that called `getAbortSignal()` was re-run or destroyed")}};function ZrA(t){return t===this.v}function J_(t,e){return t!=t?e==e:t!==e||t!==null&&typeof t=="object"||typeof t=="function"}function AGA(t,e){return t!==e}function WrA(t){return!J_(t,this.v)}function ef(t){throw new Error("https://svelte.dev/e/lifecycle_outside_component")}var qn=null;function NoA(t){qn=t}function $1(t){return XrA().get(t)}function mt(t){qn={p:qn,c:null,e:null,s:t,x:null,l:null},fQ&&!(arguments.length>1&&arguments[1]!==void 0&&arguments[1])&&(qn.l={s:null,u:null,r1:[],r2:jC(!1)})}function pt(t){var e=qn,A=e.e;if(A!==null)for(var i of(e.e=null,A))rsA(i);return t!==void 0&&(e.x=t),qn=e.p,t??{}}function pQ(){return!fQ||qn!==null&&qn.l===null}function XrA(t){var e,A;return qn===null&&ef(),(A=(e=qn).c)!==null&&A!==void 0?A:e.c=new Map(function(i){for(var n=i.p;n!==null;){var o=n.c;if(o!==null)return o;n=n.p}return null}(qn)||void 0)}function iQ(t){if(typeof t!="object"||t===null||Og in t)return t;var e=K_(t);if(e!==j_A&&e!==q_A)return t;var A=new Map,i=mQ(t),n=P0(0),o=zC,r=s=>{if(zC===o)return s();var a=Go,c=zC;q1(null),KoA(o);var l=s();return q1(a),KoA(c),l};return i&&A.set("length",P0(t.length)),new Proxy(t,{defineProperty(s,a,c){"value"in c&&c.configurable!==!1&&c.enumerable!==!1&&c.writable!==!1||function(){throw new Error("https://svelte.dev/e/state_descriptors_fixed")}();var l=A.get(a);return l===void 0?l=r(()=>{var I=P0(c.value);return A.set(a,I),I}):y(l,c.value,!0),!0},deleteProperty(s,a){var c=A.get(a);if(c===void 0){if(a in s){var l=r(()=>P0(Qs));A.set(a,l),GN(n)}}else{if(i&&typeof a=="string"){var I=A.get("length"),C=Number(a);Number.isInteger(C)&&C<I.v&&y(I,C)}y(c,Qs),GN(n)}return!0},get(s,a,c){var l;if(a===Og)return t;var I=A.get(a),C=a in s;if(I===void 0&&(!C||(l=$0(s,a))!==null&&l!==void 0&&l.writable)&&(I=r(()=>P0(iQ(C?s[a]:Qs))),A.set(a,I)),I!==void 0){var d=g(I);return d===Qs?void 0:d}return Reflect.get(s,a,c)},getOwnPropertyDescriptor(s,a){var c=Reflect.getOwnPropertyDescriptor(s,a);if(c&&"value"in c){var l=A.get(a);l&&(c.value=g(l))}else if(c===void 0){var I=A.get(a),C=I?.v;if(I!==void 0&&C!==Qs)return{enumerable:!0,configurable:!0,value:C,writable:!0}}return c},has(s,a){var c;if(a===Og)return!0;var l=A.get(a),I=l!==void 0&&l.v!==Qs||Reflect.has(s,a);return(l!==void 0||Vn!==null&&(!I||(c=$0(s,a))!==null&&c!==void 0&&c.writable))&&(l===void 0&&(l=r(()=>P0(I?iQ(s[a]):Qs)),A.set(a,l)),g(l)===Qs)?!1:I},set(s,a,c,l){var I,C=A.get(a),d=a in s;if(i&&a==="length")for(var B=c;B<C.v;B+=1){var E=A.get(B+"");E!==void 0?y(E,Qs):B in s&&(E=r(()=>P0(Qs)),A.set(B+"",E))}C===void 0?(!d||(I=$0(s,a))!==null&&I!==void 0&&I.writable)&&(y(C=r(()=>P0(void 0)),iQ(c)),A.set(a,C)):(d=C.v!==Qs,y(C,r(()=>iQ(c))));var h=Reflect.getOwnPropertyDescriptor(s,a);if(h!=null&&h.set&&h.set.call(l,c),!d){if(i&&typeof a=="string"){var u=A.get("length"),D=Number(a);Number.isInteger(D)&&D>=u.v&&y(u,D+1)}GN(n)}return!0},ownKeys(s){g(n);var a=Reflect.ownKeys(s).filter(I=>{var C=A.get(I);return C===void 0||C.v!==Qs});for(var[c,l]of A)l.v===Qs||c in s||a.push(c);return a},setPrototypeOf(){(function(){throw new Error("https://svelte.dev/e/state_prototype_fixed")})()}})}function _oA(t){try{if(t!==null&&typeof t=="object"&&Og in t)return t[Og]}catch{}return t}function eGA(t,e){return Object.is(_oA(t),_oA(e))}function wQ(t){var e=2050,A=Go!==null&&2&Go.f?Go:null;return Vn===null||A!==null&&(A.f&xl)!==0?e|=xl:Vn.f|=W_A,{ctx:qn,deps:null,effects:null,equals:ZrA,f:e,fn:t,reactions:null,rv:0,v:Qs,wv:0,parent:A??Vn,ac:null}}function Ga(t){var e=wQ(t);return BsA(e),e}function qA(t){var e=wQ(t);return e.equals=WrA,e}function $rA(t){var e=t.effects;if(e!==null){t.effects=null;for(var A=0;A<e.length;A+=1)jc(e[A])}}function T_(t){var e,A=Vn;V1(function(i){for(var n=i.parent;n!==null;){if(!(2&n.f))return n;n=n.parent}return null}(t));try{$rA(t),e=hsA(t)}finally{V1(A)}return e}function AsA(t){var e=T_(t);t.equals(e)||(t.v=e,t.wv=EsA()),j1||Ul(t,!K1&&(t.f&xl)===0||t.deps===null?pc:Ad)}var TC=new Map;function jC(t,e){return{f:0,v:t,reactions:null,equals:ZrA,rv:0,wv:0}}function P0(t,e){var A=jC(t);return BsA(A),A}function $(t){var e,A,i=arguments.length>1&&arguments[1]!==void 0&&arguments[1],n=!(arguments.length>2&&arguments[2]!==void 0)||arguments[2],o=jC(t);return i||(o.equals=WrA),fQ&&n&&qn!==null&&qn.l!==null&&((A=(e=qn.l).s)!==null&&A!==void 0?A:e.s=[]).push(o),o}function hc(t,e){return y(t,nA(()=>g(t))),e}function y(t,e){var A,i=arguments.length>2&&arguments[2]!==void 0&&arguments[2];return Go===null||Tg&&(Go.f&Z_A)===0||!pQ()||!(131090&Go.f)||(A=e2)!==null&&A!==void 0&&A.includes(t)||function(){throw new Error("https://svelte.dev/e/state_unsafe_mutation")}(),g_(t,i?iQ(e):e)}function g_(t,e){if(!t.equals(e)){var A=t.v;j1?TC.set(t,e):TC.set(t,A),t.v=e,2&t.f&&((t.f&IQ)!==0&&T_(t),Ul(t,(t.f&xl)===0?pc:Ad)),t.wv=EsA(),esA(t,IQ),!pQ()||Vn===null||(Vn.f&pc)===0||96&Vn.f||(Hc===null?function(i){Hc=i}([t]):Hc.push(t))}return e}function GoA(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:1,A=g(t),i=e===1?A++:A--;return y(t,A),i}function GN(t){y(t,t.v+1)}function esA(t,e){var A=t.reactions;if(A!==null)for(var i=pQ(),n=A.length,o=0;o<n;o++){var r=A[o],s=r.f;(s&IQ)===0&&(i||r!==Vn)&&(Ul(r,e),1280&s&&(2&s?esA(r,Ad):_y(r)))}}var A2,tsA,isA,nsA;function xy(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:"";return document.createTextNode(t)}function uc(t){return isA.call(t)}function Ly(t){return nsA.call(t)}function W(t,e){return uc(t)}function vt(t,e){var A=uc(t);return A instanceof Comment&&A.data===""?Ly(A):A}function IA(t){for(var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:1,A=t;e--;)A=Ly(A);return A}function osA(t){Vn===null&&Go===null&&function(){throw new Error("https://svelte.dev/e/effect_orphan")}(),Go!==null&&(Go.f&xl)!==0&&Vn===null&&function(){throw new Error("https://svelte.dev/e/effect_in_unowned_derived")}(),j1&&function(){throw new Error("https://svelte.dev/e/effect_in_teardown")}()}function AI(t,e,A){var i=!(arguments.length>3&&arguments[3]!==void 0)||arguments[3],n=Vn,o={ctx:qn,deps:null,nodes_start:null,nodes_end:null,f:t|IQ,first:null,fn:e,last:null,next:null,parent:n,b:n&&n.b,prev:null,teardown:null,transitions:null,wv:0,ac:null};if(A)try{Ny(o),o.f|=32768}catch(a){throw jc(o),a}else e!==null&&_y(o);if(!(A&&o.deps===null&&o.first===null&&o.nodes_start===null&&o.teardown===null&&!(524416&o.f))&&i&&(n!==null&&function(a,c){var l=c.last;l===null?c.last=c.first=a:(l.next=a,a.prev=l,c.last=a)}(o,n),Go!==null&&2&Go.f)){var r,s=Go;((r=s.effects)!==null&&r!==void 0?r:s.effects=[]).push(o)}return o}function H_(t){var e=AI(8,null,!1);return Ul(e,pc),e.teardown=t,e}function I_(t){if(osA(),Go||!Vn||(Vn.f&Ry)===0)return rsA(t);var e,A=qn;((e=A.e)!==null&&e!==void 0?e:A.e=[]).push(t)}function rsA(t){return AI(2097156,t,!1)}function es(t){return AI(4,t,!1)}function fA(t,e){var A=qn,i={effect:null,ran:!1};A.l.r1.push(i),i.effect=tf(()=>{t(),i.ran||(i.ran=!0,y(A.l.r2,!0),nA(e))})}function Qn(){var t=qn;tf(()=>{if(g(t.l.r2)){for(var e of t.l.r1){var A=e.effect;(A.f&pc)!==0&&Ul(A,Ad),nf(A)&&Ny(A),e.ran=!1}t.l.r2.v=!1}})}function tf(t){return AI(8,t,!0)}function De(t){var e=arguments.length>2&&arguments[2]!==void 0?arguments[2]:wQ,A=(arguments.length>1&&arguments[1]!==void 0?arguments[1]:[]).map(e);return eI(()=>t(...A.map(g)))}function eI(t){return AI(24|(arguments.length>1&&arguments[1]!==void 0?arguments[1]:0),t,!0)}function Vg(t){return AI(40,t,!0,!(arguments.length>1&&arguments[1]!==void 0)||arguments[1])}function ssA(t){var e=t.teardown;if(e!==null){var A=j1,i=Go;UoA(!0),q1(null);try{e.call(null)}finally{UoA(A),q1(i)}}}function asA(t){var e=arguments.length>1&&arguments[1]!==void 0&&arguments[1],A=t.first;for(t.first=t.last=null;A!==null;){var i;(i=A.ac)===null||i===void 0||i.abort(VrA);var n=A.next;(A.f&jrA)!==0?A.parent=null:jc(A,e),A=n}}function jc(t){var e=!(arguments.length>1&&arguments[1]!==void 0)||arguments[1],A=!1;(e||262144&t.f)&&t.nodes_start!==null&&t.nodes_end!==null&&(csA(t.nodes_start,t.nodes_end),A=!0),asA(t,e&&!A),dy(t,0),Ul(t,Y_);var i=t.transitions;if(i!==null)for(var n of i)n.stop();ssA(t);var o=t.parent;o!==null&&o.first!==null&&lsA(t),t.next=t.prev=t.teardown=t.ctx=t.deps=t.fn=t.nodes_start=t.nodes_end=t.ac=null}function csA(t,e){for(;t!==null;){var A=t===e?null:Ly(t);t.remove(),t=A}}function lsA(t){var e=t.parent,A=t.prev,i=t.next;A!==null&&(A.next=i),i!==null&&(i.prev=A),e!==null&&(e.first===t&&(e.first=i),e.last===t&&(e.last=A))}function CQ(t,e){var A=[];z_(t,A,!0),gsA(A,()=>{jc(t),e&&e()})}function gsA(t,e){var A=t.length;if(A>0){var i=()=>--A||e();for(var n of t)n.out(i)}else e()}function z_(t,e,A){if((t.f&H1)===0){if(t.f^=H1,t.transitions!==null)for(var i of t.transitions)(i.is_global||A)&&e.push(i);for(var n=t.first;n!==null;){var o=n.next;z_(n,e,((n.f&Af)!==0||(n.f&Ry)!==0)&&A),n=o}}}function Iy(t){IsA(t,!0)}function IsA(t,e){if((t.f&H1)!==0){t.f^=H1;for(var A=t.first;A!==null;){var i=A.next;IsA(A,((A.f&Af)!==0||(A.f&Ry)!==0)&&e),A=i}if(t.transitions!==null)for(var n of t.transitions)(n.is_global||e)&&n.in()}}var K3=[],UN=[];function CsA(){var t=K3;K3=[],gy(t)}function Fy(t){K3.length===0&&queueMicrotask(CsA),K3.push(t)}function tGA(){var t;K3.length>0&&CsA(),UN.length>0&&(t=UN,UN=[],gy(t))}function dsA(t,e){for(;e!==null;){if(128&e.f)try{return void e.b.error(t)}catch{}e=e.parent}throw t}var Y3=!1,J3=null,HC=!1,j1=!1;function UoA(t){j1=t}var U3=[],Go=null,Tg=!1;function q1(t){Go=t}var Vn=null;function V1(t){Vn=t}var e2=null;function BsA(t){Go!==null&&Go.f&l_&&(e2===null?e2=[t]:e2.push(t))}var Aa=null,Ec=0,Hc=null,Cy=1,T3=0,zC=T3;function KoA(t){zC=t}var K1=!1,YoA=null;function EsA(){return++Cy}function nf(t){var e=t.f;if((e&IQ)!==0)return!0;if((e&Ad)!==0){var A=t.deps,i=(e&xl)!==0;if(A!==null){var n,o,r=(e&c_)!==0,s=i&&Vn!==null&&!K1,a=A.length;if(r||s){var c=t,l=c.parent;for(n=0;n<a;n++){var I,C,d;o=A[n],(r||(I=o)===null||I===void 0||(I=I.reactions)===null||I===void 0||!I.includes(c))&&((d=(C=o).reactions)!==null&&d!==void 0?d:C.reactions=[]).push(c)}r&&(c.f^=c_),s&&l!==null&&(l.f&xl)===0&&(c.f^=xl)}for(n=0;n<a;n++)if(nf(o=A[n])&&AsA(o),o.wv>t.wv)return!0}i&&(Vn===null||K1)||Ul(t,pc)}return!1}function QsA(t,e){var A,i=!(arguments.length>2&&arguments[2]!==void 0)||arguments[2],n=t.reactions;if(n!==null&&((A=e2)===null||A===void 0||!A.includes(t)))for(var o=0;o<n.length;o++){var r=n[o];2&r.f?QsA(r,e,!1):e===r&&(i?Ul(r,IQ):(r.f&pc)!==0&&Ul(r,Ad),_y(r))}}function hsA(t){var e=Aa,A=Ec,i=Hc,n=Go,o=K1,r=e2,s=qn,a=Tg,c=zC,l=t.f;Aa=null,Ec=0,Hc=null,K1=(l&xl)!==0&&(Tg||!HC||Go===null),Go=96&l?null:t,e2=null,NoA(t.ctx),Tg=!1,zC=++T3,t.f|=l_,t.ac!==null&&(t.ac.abort(VrA),t.ac=null);try{var I=(0,t.fn)(),C=t.deps;if(Aa!==null){var d;if(dy(t,Ec),C!==null&&Ec>0)for(C.length=Ec+Aa.length,d=0;d<Aa.length;d++)C[Ec+d]=Aa[d];else t.deps=C=Aa;if(!K1||2&l&&t.reactions!==null)for(d=Ec;d<C.length;d++){var B,E;((E=(B=C[d]).reactions)!==null&&E!==void 0?E:B.reactions=[]).push(t)}}else C!==null&&Ec<C.length&&(dy(t,Ec),C.length=Ec);if(pQ()&&Hc!==null&&!Tg&&C!==null&&!(6146&t.f))for(d=0;d<Hc.length;d++)QsA(Hc[d],t);return n!==null&&n!==t&&(T3++,Hc!==null&&(i===null?i=Hc:i.push(...Hc))),I}catch(h){(function(u){var D=Vn;if(32768&D.f)dsA(u,D);else{if(!(128&D.f))throw u;D.fn(u)}})(h)}finally{Aa=e,Ec=A,Hc=i,Go=n,K1=o,e2=r,NoA(s),Tg=a,zC=c,t.f^=l_}}function iGA(t,e){var A=e.reactions;if(A!==null){var i=O_A.call(A,t);if(i!==-1){var n=A.length-1;n===0?A=e.reactions=null:(A[i]=A[n],A.pop())}}A===null&&2&e.f&&(Aa===null||!Aa.includes(e))&&(Ul(e,Ad),768&e.f||(e.f^=c_),$rA(e),dy(e,0))}function dy(t,e){var A=t.deps;if(A!==null)for(var i=e;i<A.length;i++)iGA(t,A[i])}function Ny(t){var e=t.f;if((e&Y_)===0){Ul(t,pc);var A=Vn,i=HC;Vn=t,HC=!0;try{16&e?function(o){for(var r=o.first;r!==null;){var s=r.next;(r.f&Ry)===0&&jc(r),r=s}}(t):asA(t),ssA(t);var n=hsA(t);t.teardown=typeof n=="function"?n:null,t.wv=Cy}finally{HC=i,Vn=A}}}function nGA(){try{(function(){throw new Error("https://svelte.dev/e/effect_update_depth_exceeded")})()}catch(t){if(J3===null)throw t;dsA(t,J3)}}function usA(){var t=HC;try{var e=0;for(HC=!0;U3.length>0;){e++>1e3&&nGA();var A=U3,i=A.length;U3=[];for(var n=0;n<i;n++)oGA(rGA(A[n]));TC.clear()}}finally{Y3=!1,HC=t,J3=null}}function oGA(t){var e=t.length;if(e!==0){for(var A=0;A<e;A++){var i=t[A];if(!(24576&i.f)&&nf(i)){var n=Cy;if(Ny(i),i.deps===null&&i.first===null&&i.nodes_start===null&&(i.teardown===null?lsA(i):i.fn=null),Cy>n&&(i.f&X_A)!==0)break}}for(;A<e;A+=1)_y(t[A])}}function _y(t){Y3||(Y3=!0,queueMicrotask(usA));for(var e=J3=t;e.parent!==null;){var A=(e=e.parent).f;if(96&A){if((A&pc)===0)return;e.f^=pc}}U3.push(e)}function rGA(t){for(var e=[],A=t;A!==null;){var i=A.f,n=!!(96&i);if(!(n&&(i&pc)!==0)&&(i&H1)===0){4&i?e.push(A):n?A.f^=pc:nf(A)&&Ny(A);var o=A.first;if(o!==null){A=o;continue}}var r=A.parent;for(A=A.next;A===null&&r!==null;)A=r.next,r=r.parent}return e}function io(t){for(;;){if(tGA(),U3.length===0)return Y3=!1,void(J3=null);Y3=!0,usA()}}function fsA(){return(fsA=Yt(function*(){yield Promise.resolve(),io()})).apply(this,arguments)}function g(t){var e,A=!!(2&t.f);if(YoA!==null&&YoA.add(t),Go===null||Tg){if(A&&t.deps===null&&t.effects===null){var i=t,n=i.parent;n!==null&&(n.f&xl)===0&&(i.f^=xl)}}else if((e=e2)===null||e===void 0||!e.includes(t)){var o=Go.deps;t.rv<T3&&(t.rv=T3,Aa===null&&o!==null&&o[Ec]===t?Ec++:Aa===null?Aa=[t]:K1&&Aa.includes(t)||Aa.push(t))}if(A&&!j1&&nf(i=t)&&AsA(i),j1){if(TC.has(t))return TC.get(t);if(A){var r=(i=t).v;return((i.f&pc)!==0||msA(i))&&(r=T_(i)),TC.set(i,r),r}}return t.v}function msA(t){if(t.v===Qs)return!0;if(t.deps===null)return!1;for(var e of t.deps)if(TC.has(e)||2&e.f&&msA(e))return!0;return!1}function nA(t){var e=Tg;try{return Tg=!0,t()}finally{Tg=e}}var sGA=-7169;function Ul(t,e){t.f=t.f&sGA|e}function k(t){if(typeof t=="object"&&t&&!(t instanceof EventTarget)){if(Og in t)C_(t);else if(!Array.isArray(t))for(var e in t){var A=t[e];typeof A=="object"&&A&&Og in A&&C_(A)}}}function C_(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:new Set;if(!(typeof t!="object"||t===null||t instanceof EventTarget||e.has(t))){for(var A in e.add(t),t instanceof Date&&t.getTime(),t)try{C_(t[A],e)}catch{}var i=K_(t);if(i!==Object.prototype&&i!==Array.prototype&&i!==Map.prototype&&i!==Set.prototype&&i!==Date.prototype){var n=PrA(i);for(var o in n){var r=n[o].get;if(r)try{r.call(t)}catch{}}}}}var JoA=!1;function psA(t){var e=Go,A=Vn;q1(null),V1(null);try{return t()}finally{q1(e),V1(A)}}function aGA(t,e,A){var i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:A;t.addEventListener(e,()=>psA(A));var n=t.__on_r;t.__on_r=n?()=>{n(),i(!0)}:()=>i(!0),JoA||(JoA=!0,document.addEventListener("reset",o=>{Promise.resolve().then(()=>{if(!o.defaultPrevented)for(var r of o.target.elements){var s;(s=r.__on_r)===null||s===void 0||s.call(r)}})},{capture:!0}))}var wsA=new Set,d_=new Set;function DsA(t,e,A){var i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:{};function n(o){if(i.capture||F3.call(e,o),!o.cancelBubble)return psA(()=>A?.call(this,o))}return t.startsWith("pointer")||t.startsWith("touch")||t==="wheel"?Fy(()=>{e.addEventListener(t,n,i)}):e.addEventListener(t,n,i),n}function ce(t,e,A,i,n){var o={capture:i,passive:n},r=DsA(t,e,A,o);(e===document.body||e===window||e===document||e instanceof HTMLMediaElement)&&H_(()=>{e.removeEventListener(t,r,o)})}function of(t){for(var e=0;e<t.length;e++)wsA.add(t[e]);for(var A of d_)A(t)}function F3(t){var e,A=this,i=A.ownerDocument,n=t.type,o=((e=t.composedPath)===null||e===void 0?void 0:e.call(t))||[],r=o[0]||t.target,s=0,a=t.__root;if(a){var c=o.indexOf(a);if(c!==-1&&(A===document||A===window))return void(t.__root=A);var l=o.indexOf(A);if(l===-1)return;c<=l&&(s=c)}if((r=o[s]||t.target)!==A){P_A(t,"currentTarget",{configurable:!0,get:()=>r||i});var I=Go,C=Vn;q1(null),V1(null);try{for(var d,B=[];r!==null;){var E=r.assignedSlot||r.parentNode||r.host||null;try{var h=r["__"+n];if(h!=null&&(!r.disabled||t.target===r))if(mQ(h)){var[u,...D]=h;u.apply(r,[t,...D])}else h.call(r,t)}catch(w){d?B.push(w):d=w}if(t.cancelBubble||E===A||E===null)break;r=E}if(d){var L=function(w){queueMicrotask(()=>{throw w})};for(var R of B)L(R);throw d}}finally{t.__root=A,delete t.currentTarget,q1(I),V1(C)}}}function O_(t){var e=document.createElement("template");return e.innerHTML=t.replaceAll("<!>","<!---->"),e.content}function qC(t,e){var A=Vn;A.nodes_start===null&&(A.nodes_start=t,A.nodes_end=e)}function wA(t,e){var A,i=!!(1&e),n=!!(2&e),o=!t.startsWith("<!>");return()=>{A===void 0&&(A=O_(o?t:"<!>"+t),i||(A=uc(A)));var r=n||tsA?document.importNode(A,!0):A.cloneNode(!0);return i?qC(uc(r),r.lastChild):qC(r,r),r}}function tI(t,e){return function(A,i){var n,o=arguments.length>2&&arguments[2]!==void 0?arguments[2]:"svg",r=!A.startsWith("<!>"),s=!!(1&i),a="<".concat(o,">").concat(r?A:"<!>"+A,"</").concat(o,">");return()=>{if(!n){var c=uc(O_(a));if(s)for(n=document.createDocumentFragment();uc(c);)n.appendChild(uc(c));else n=uc(c)}var l=n.cloneNode(!0);return s?qC(uc(l),l.lastChild):qC(l,l),l}}(t,e,"svg")}function Tr(){var t=xy((arguments.length>0&&arguments[0]!==void 0?arguments[0]:"")+"");return qC(t,t),t}function _o(){var t=document.createDocumentFragment(),e=document.createComment(""),A=xy();return t.append(e,A),qC(e,A),t}function iA(t,e){t!==null&&t.before(e)}var cGA=["beforeinput","click","change","dblclick","contextmenu","focusin","focusout","input","keydown","keyup","mousedown","mousemove","mouseout","mouseover","mouseup","pointerdown","pointermove","pointerout","pointerover","pointerup","touchend","touchmove","touchstart"],lGA={formnovalidate:"formNoValidate",ismap:"isMap",nomodule:"noModule",playsinline:"playsInline",readonly:"readOnly",defaultvalue:"defaultValue",defaultchecked:"defaultChecked",srcobject:"srcObject",novalidate:"noValidate",allowfullscreen:"allowFullscreen",disablepictureinpicture:"disablePictureInPicture",disableremoteplayback:"disableRemotePlayback"},gGA=["touchstart","touchmove"];function IGA(t){return gGA.includes(t)}function wt(t,e){var A,i=e==null?"":typeof e=="object"?e+"":e;i!==((A=t.__t)!==null&&A!==void 0?A:t.__t=t.nodeValue)&&(t.__t=i,t.nodeValue=i+"")}function CGA(t,e){return function(A,i){var{target:n,anchor:o,props:r={},events:s,context:a,intro:c=!0}=i;(function(){if(A2===void 0){A2=window,tsA=/Firefox/.test(navigator.userAgent);var B=Element.prototype,E=Node.prototype,h=Text.prototype;isA=$0(E,"firstChild").get,nsA=$0(E,"nextSibling").get,LoA(B)&&(B.__click=void 0,B.__className=void 0,B.__attributes=null,B.__style=void 0,B.__e=void 0),LoA(h)&&(h.__t=void 0)}})();var l=new Set,I=B=>{for(var E=0;E<B.length;E++){var h=B[E];if(!l.has(h)){l.add(h);var u=IGA(h);n.addEventListener(h,F3,{passive:u});var D=VE.get(h);D===void 0?(document.addEventListener(h,F3,{passive:u}),VE.set(h,1)):VE.set(h,D+1)}}};I(a_(wsA)),d_.add(I);var C=void 0,d=function(B){var E=AI(jrA,B,!0);return function(){var h=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};return new Promise(u=>{h.outro?CQ(E,()=>{jc(E),u(void 0)}):(jc(E),u(void 0))})}}(()=>{var B=o??n.appendChild(xy());return Vg(()=>{a&&(mt({}),qn.c=a),s&&(r.$$events=s),C=A(B,r)||{},a&&pt()}),()=>{for(var E of l){n.removeEventListener(E,F3);var h=VE.get(E);--h===0?(document.removeEventListener(E,F3),VE.delete(E)):VE.set(E,h)}var u;d_.delete(I),B!==o&&((u=B.parentNode)===null||u===void 0||u.removeChild(B))}});return B_.set(C,d),C}(t,e)}var VE=new Map,B_=new WeakMap;function hs(t){qn===null&&ef(),fQ&&qn.l!==null?ysA(qn).m.push(t):I_(()=>{var e=nA(t);if(typeof e=="function")return e})}function qc(t){qn===null&&ef(),hs(()=>()=>nA(t))}function dGA(){var t=qn;return t===null&&ef(),(e,A,i)=>{var n,o=(n=t.s.$$events)===null||n===void 0?void 0:n[e];if(o){var r=mQ(o)?o.slice():[o],s=function(c,l){var{bubbles:I=!1,cancelable:C=!1}=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{};return new CustomEvent(c,{detail:l,bubbles:I,cancelable:C})}(e,A,i);for(var a of r)a.call(t.x,s);return!s.defaultPrevented}return!0}}function BGA(t){qn===null&&ef(),qn.l===null&&function(){throw new Error("https://svelte.dev/e/lifecycle_legacy_only")}(),ysA(qn).b.push(t)}function ysA(t){var e,A=t.l;return(e=A.u)!==null&&e!==void 0?e:A.u={a:[],b:[],m:[]}}function LA(t,e){var[A,i]=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[0,0],n=t,o=null,r=null,s=Qs,a=!1,c=function(I){a=!0,l(!(arguments.length>1&&arguments[1]!==void 0)||arguments[1],I)},l=(I,C)=>{s!==(s=I)&&(s?(o?Iy(o):C&&(o=Vg(()=>C(n))),r&&CQ(r,()=>{r=null})):(r?Iy(r):C&&(r=Vg(()=>C(n,[A+1,i]))),o&&CQ(o,()=>{o=null})))};eI(()=>{a=!1,e(c),a||l(null,null)},A>0?Af:0)}function vsA(t,e,A){var i,n=t,o=Qs,r=pQ()?AGA:J_;eI(()=>{r(o,o=e())&&(i&&CQ(i),i=Vg(()=>A(n)))})}function or(t,e){return e}function qo(t,e,A,i,n){var o=arguments.length>5&&arguments[5]!==void 0?arguments[5]:null,r=t,s={flags:e,items:new Map,first:null};!(4&e)||(r=t.appendChild(xy()));var a=null,c=!1,l=qA(()=>{var I=A();return mQ(I)?I:I==null?[]:a_(I)});eI(()=>{var I=g(l),C=I.length;c&&C===0||(c=C===0,function(d,B,E,h,u,D,L){var R,w,_,K,z,U,H=!!(8&u),q=!!(3&u),j=d.length,gA=B.items,QA=B.first,BA=QA,lA=null,vA=[],tA=[];if(H)for(U=0;U<j;U+=1){var cA;K=D(_=d[U],U),(z=gA.get(K))!==void 0&&((cA=z.a)===null||cA===void 0||cA.measure(),(w??(w=new Set)).add(z))}for(U=0;U<j;U+=1)if(K=D(_=d[U],U),(z=gA.get(K))!==void 0){var pA;if(q&&EGA(z,_,U,u),(z.e.f&H1)!==0&&(Iy(z.e),H&&((pA=z.a)===null||pA===void 0||pA.unfix(),(w??(w=new Set)).delete(z))),z!==BA){if(R!==void 0&&R.has(z)){if(vA.length<tA.length){var VA,oe=tA[0];lA=oe.prev;var KA=vA[0],CA=vA[vA.length-1];for(VA=0;VA<vA.length;VA+=1)ToA(vA[VA],oe,E);for(VA=0;VA<tA.length;VA+=1)R.delete(tA[VA]);F1(B,KA.prev,CA.next),F1(B,lA,KA),F1(B,CA,oe),BA=oe,lA=CA,U-=1,vA=[],tA=[]}else R.delete(z),ToA(z,BA,E),F1(B,z.prev,z.next),F1(B,z,lA===null?B.first:lA.next),F1(B,lA,z),lA=z;continue}for(vA=[],tA=[];BA!==null&&BA.k!==K;)(BA.e.f&H1)===0&&(R??(R=new Set)).add(BA),tA.push(BA),BA=BA.next;if(BA===null)continue;z=BA}vA.push(z),lA=z,BA=z.next}else lA=QGA(BA?BA.e.nodes_start:E,B,lA,lA===null?B.first:lA.next,_,K,U,h,u,L),gA.set(K,lA),vA=[],tA=[],BA=lA.next;if(BA!==null||R!==void 0){for(var TA=R===void 0?[]:a_(R);BA!==null;)(BA.e.f&H1)===0&&TA.push(BA),BA=BA.next;var Ze=TA.length;if(Ze>0){var He=4&u&&j===0?E:null;if(H){for(U=0;U<Ze;U+=1){var uA;(uA=TA[U].a)===null||uA===void 0||uA.measure()}for(U=0;U<Ze;U+=1){var eA;(eA=TA[U].a)===null||eA===void 0||eA.fix()}}(function(UA,aA,le,SA){for(var Ue=[],mA=aA.length,sA=0;sA<mA;sA++)z_(aA[sA].e,Ue,!0);var xt=mA>0&&Ue.length===0&&le!==null;if(xt){var tt=le.parentNode;tt.textContent="",tt.append(le),SA.clear(),F1(UA,aA[0].prev,aA[mA-1].next)}gsA(Ue,()=>{for(var de=0;de<mA;de++){var Dt=aA[de];xt||(SA.delete(Dt.k),F1(UA,Dt.prev,Dt.next)),jc(Dt.e,!xt)}})})(B,TA,He,gA)}}H&&Fy(()=>{if(w!==void 0)for(z of w){var UA;(UA=z.a)===null||UA===void 0||UA.apply()}}),Vn.first=B.first&&B.first.e,Vn.last=lA&&lA.e}(I,s,r,n,e,i,A),o!==null&&(C===0?a?Iy(a):a=Vg(()=>o(r)):a!==null&&CQ(a,()=>{a=null})),g(l))})}function EGA(t,e,A,i){1&i&&g_(t.v,e),2&i?g_(t.i,A):t.i=A}function QGA(t,e,A,i,n,o,r,s,a,c){var l=1&a?16&a?jC(n):$(n,!1,!1):n,I=2&a?jC(r):r,C={i:I,v:l,k:o,a:null,e:null,prev:A,next:i};try{return C.e=Vg(()=>s(t,l,I,c),!1),C.e.prev=A&&A.e,C.e.next=i&&i.e,A===null?e.first=C:(A.next=C,A.e.next=C.e),i!==null&&(i.prev=C,i.e.prev=C.e),C}finally{}}function ToA(t,e,A){for(var i=t.next?t.next.e.nodes_start:A,n=e?e.e.nodes_start:A,o=t.e.nodes_start;o!==i;){var r=Ly(o);n.before(o),o=r}}function F1(t,e,A){e===null?t.first=A:(e.next=A,e.e.next=A&&A.e),A!==null&&(A.prev=e,A.e.prev=e&&e.e)}function bsA(t,e){var A=arguments.length>2&&arguments[2]!==void 0&&arguments[2],i=arguments.length>3&&arguments[3]!==void 0&&arguments[3],n=t,o="";De(()=>{var r,s=Vn;if(o!==(o=(r=e())!==null&&r!==void 0?r:"")&&(s.nodes_start!==null&&(csA(s.nodes_start,s.nodes_end),s.nodes_start=s.nodes_end=null),o!=="")){var a=o+"";A?a="<svg>".concat(a,"</svg>"):i&&(a="<math>".concat(a,"</math>"));var c=O_(a);if((A||i)&&(c=uc(c)),qC(uc(c),c.lastChild),A||i)for(;uc(c);)n.before(uc(c));else n.before(c)}})}function jo(t,e,A,i,n){var o,r=(o=e.$$slots)===null||o===void 0?void 0:o[A],s=!1;r===!0&&(r=e[A==="default"?"children":A],s=!0),r===void 0?n!==null&&n(t):r(t,s?()=>i:i)}function MsA(t,e,A){var i,n,o=t;eI(()=>{i!==(i=e())&&(n&&(CQ(n),n=null),i&&(n=Vg(()=>A(o,i))))},Af)}function Us(t,e,A){es(()=>{var i=nA(()=>e(t,A?.())||{});if(A&&i!=null&&i.update){var n=!1,o={};tf(()=>{var r=A();k(r),n&&J_(o,r)&&(o=r,i.update(r))}),n=!0}if(i!=null&&i.destroy)return()=>i.destroy()})}function hGA(t,e){var A,i=void 0;eI(()=>{i!==(i=e())&&(A&&(jc(A),A=null),i&&(A=Vg(()=>{es(()=>i(t))})))})}function ksA(t){var e,A,i="";if(typeof t=="string"||typeof t=="number")i+=t;else if(typeof t=="object")if(Array.isArray(t)){var n=t.length;for(e=0;e<n;e++)t[e]&&(A=ksA(t[e]))&&(i&&(i+=" "),i+=A)}else for(A in t)t[A]&&(i&&(i+=" "),i+=A);return i}function Z1(t){return typeof t=="object"?function(){for(var e,A,i=0,n="",o=arguments.length;i<o;i++)(e=arguments[i])&&(A=ksA(e))&&(n&&(n+=" "),n+=A);return n}(t):t??""}var HoA=[...` 	
\r\f\xA0\v\uFEFF`];function zoA(t){var e=arguments.length>1&&arguments[1]!==void 0&&arguments[1]?" !important;":";",A="";for(var i in t){var n=t[i];n!=null&&n!==""&&(A+=" "+i+": "+n+e)}return A}function KN(t){return t[0]!=="-"||t[1]!=="-"?t.toLowerCase():t}function Vt(t,e,A,i,n,o){var r=t.__className;if(r!==A||r===void 0){var s=function(l,I,C){var d=l==null?"":""+l;if(I&&(d=d?d+" "+I:I),C){for(var B in C)if(C[B])d=d?d+" "+B:B;else if(d.length)for(var E=B.length,h=0;(h=d.indexOf(B,h))>=0;){var u=h+E;h!==0&&!HoA.includes(d[h-1])||u!==d.length&&!HoA.includes(d[u])?h=u:d=(h===0?"":d.substring(0,h))+d.substring(u+1)}}return d===""?null:d}(A,i,o);s==null?t.removeAttribute("class"):e?t.className=s:t.setAttribute("class",s),t.__className=A}else if(o&&n!==o)for(var a in o){var c=!!o[a];n!=null&&c===!!n[a]||t.classList.toggle(a,c)}return o}function YN(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},A=arguments.length>2?arguments[2]:void 0,i=arguments.length>3?arguments[3]:void 0;for(var n in A){var o=A[n];e[n]!==o&&(A[n]==null?t.style.removeProperty(n):t.style.setProperty(n,o,i))}}function Ll(t,e,A,i){if(t.__style!==e){var n=function(o,r){if(r){var s,a,c="";if(Array.isArray(r)?(s=r[0],a=r[1]):s=r,o){o=String(o).replaceAll(/\s*\/\*.*?\*\/\s*/g,"").trim();var l=!1,I=0,C=!1,d=[];s&&d.push(...Object.keys(s).map(KN)),a&&d.push(...Object.keys(a).map(KN));for(var B=0,E=-1,h=o.length,u=0;u<h;u++){var D=o[u];if(C?D==="/"&&o[u-1]==="*"&&(C=!1):l?l===D&&(l=!1):D==="/"&&o[u+1]==="*"?C=!0:D==='"'||D==="'"?l=D:D==="("?I++:D===")"&&I--,!C&&l===!1&&I===0){if(D===":"&&E===-1)E=u;else if(D===";"||u===h-1){if(E!==-1){var L=KN(o.substring(B,E).trim());d.includes(L)||(D!==";"&&u++,c+=" "+o.substring(B,u).trim()+";")}B=u+1,E=-1}}}}return s&&(c+=zoA(s)),a&&(c+=zoA(a,!0)),(c=c.trim())===""?null:c}return o==null?null:String(o)}(e,i);n==null?t.removeAttribute("style"):t.style.cssText=n,t.__style=e}else i&&(Array.isArray(i)?(YN(t,A?.[0],i[0]),YN(t,A?.[1],i[1],"important")):YN(t,A,i));return i}function E_(t,e){var A=arguments.length>2&&arguments[2]!==void 0&&arguments[2];if(t.multiple){if(e==null)return;if(!mQ(e))return void console.warn("https://svelte.dev/e/select_multiple_invalid_value");for(var i of t.options)i.selected=e.includes(OoA(i))}else{for(i of t.options)if(eGA(OoA(i),e))return void(i.selected=!0);A&&e===void 0||(t.selectedIndex=-1)}}function uGA(t){var e=new MutationObserver(()=>{E_(t,t.__value)});e.observe(t,{childList:!0,subtree:!0,attributes:!0,attributeFilter:["value"]}),H_(()=>{e.disconnect()})}function OoA(t){return"__value"in t?t.__value:t.value}var eQ=Symbol("class"),x3=Symbol("style"),SsA=Symbol("is custom element"),RsA=Symbol("is html");function VC(t,e){var A=P_(t);A.value!==(A.value=e??void 0)&&(t.value!==e||e===0&&t.nodeName==="PROGRESS")&&(t.value=e??"")}function En(t,e,A,i){var n=P_(t);n[e]!==(n[e]=A)&&(e==="loading"&&(t[$_A]=A),A==null?t.removeAttribute(e):typeof A!="string"&&xsA(t).includes(e)?t[e]=A:t.setAttribute(e,A))}function fGA(t,e,A,i){var n,o=P_(t),r=o[SsA],s=!o[RsA],a=e||{},c=t.tagName==="OPTION";for(var l in e)l in A||(A[l]=null);A.class?A.class=Z1(A.class):(i||A[eQ])&&(A.class=null),A[x3]&&((n=A.style)!==null&&n!==void 0||(A.style=null));var I,C,d,B,E,h,u=xsA(t),D=function(R){var w=A[R];if(c&&R==="value"&&w==null)return t.value=t.__value="",a[R]=w,0;if(R==="class")return I=t.namespaceURI==="http://www.w3.org/1999/xhtml",Vt(t,I,w,i,e?.[eQ],A[eQ]),a[R]=w,a[eQ]=A[eQ],0;if(R==="style")return Ll(t,w,e?.[x3],A[x3]),a[R]=w,a[x3]=A[x3],0;if(w===(C=a[R])&&(w!==void 0||!t.hasAttribute(R))||(a[R]=w,(d=R[0]+R[1])==="$$"))return 0;if(d==="on"){var _={},K="$$"+R,z=R.slice(2);if(B=function(QA){return cGA.includes(QA)}(z),function(QA){return QA.endsWith("capture")&&QA!=="gotpointercapture"&&QA!=="lostpointercapture"}(z)&&(z=z.slice(0,-7),_.capture=!0),!B&&C){if(w!=null)return 0;t.removeEventListener(z,a[K],_),a[K]=null}if(w!=null)if(B)t["__".concat(z)]=w,of([z]);else{let QA=function(BA){a[R].call(this,BA)};var gA=QA;a[K]=DsA(z,t,QA,_)}else B&&(t["__".concat(z)]=void 0)}else if(R==="style")En(t,R,w);else if(R==="autofocus")(function(QA,BA){if(BA){var lA=document.body;QA.autofocus=!0,Fy(()=>{document.activeElement===lA&&QA.focus()})}})(t,!!w);else if(r||R!=="__value"&&(R!=="value"||w==null))if(R==="selected"&&c)(function(QA,BA){BA?QA.hasAttribute("selected")||QA.setAttribute("selected",""):QA.removeAttribute("selected")})(t,w);else if(E=R,s||(E=function(QA){var BA;return QA=QA.toLowerCase(),(BA=lGA[QA])!==null&&BA!==void 0?BA:QA}(E)),h=E==="defaultValue"||E==="defaultChecked",w!=null||r||h)h||u.includes(E)&&(r||typeof w!="string")?t[E]=w:typeof w!="function"&&En(t,E,w);else if(o[R]=null,E==="value"||E==="checked"){var U=t,H=e===void 0;if(E==="value"){var q=U.defaultValue;U.removeAttribute(E),U.defaultValue=q,U.value=U.__value=H?q:null}else{var j=U.defaultChecked;U.removeAttribute(E),U.defaultChecked=j,U.checked=!!H&&j}}else t.removeAttribute(R);else t.value=t.__value=w};for(var L in A)D(L);return a}function ry(t,e){var A=arguments.length>3?arguments[3]:void 0,i=arguments.length>4&&arguments[4]!==void 0&&arguments[4],n=arguments.length>5&&arguments[5]!==void 0?arguments[5]:wQ,o=(arguments.length>2&&arguments[2]!==void 0?arguments[2]:[]).map(n),r=void 0,s={},a=t.nodeName==="SELECT",c=!1;if(eI(()=>{var I=e(...o.map(g)),C=fGA(t,r,I,A,i);for(var d of(c&&a&&"value"in I&&E_(t,I.value),Object.getOwnPropertySymbols(s)))I[d]||jc(s[d]);for(var B of Object.getOwnPropertySymbols(I)){var E=I[B];B.description!=="@attach"||r&&E===r[B]||(s[B]&&jc(s[B]),s[B]=Vg(()=>hGA(t,()=>E))),C[B]=E}r=C}),a){var l=t;es(()=>{E_(l,r.value,!0),uGA(l)})}c=!0}function P_(t){var e;return(e=t.__attributes)!==null&&e!==void 0?e:t.__attributes={[SsA]:t.nodeName.includes("-"),[RsA]:t.namespaceURI==="http://www.w3.org/1999/xhtml"}}var PoA=new Map;function xsA(t){var e,A=PoA.get(t.nodeName);if(A)return A;PoA.set(t.nodeName,A=[]);for(var i=t,n=Element.prototype;n!==i;){for(var o in e=PrA(i))e[o].set&&A.push(o);i=K_(i)}return A}function By(t,e){var A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:e,i=pQ();aGA(t,"input",n=>{var o=n?t.defaultValue:t.value;if(o=JN(t)?TN(o):o,A(o),i&&o!==(o=e())){var r=t.selectionStart,s=t.selectionEnd;t.value=o??"",s!==null&&(t.selectionStart=r,t.selectionEnd=Math.min(s,t.value.length))}}),nA(e)==null&&t.value&&A(JN(t)?TN(t.value):t.value),tf(()=>{var n=e();JN(t)&&n===TN(t.value)||(t.type!=="date"||n||t.value)&&n!==t.value&&(t.value=n??"")})}function JN(t){var e=t.type;return e==="number"||e==="range"}function TN(t){return t===""?null:+t}function zt(t,e,A){var i=$0(t,e);i&&i.set&&(t[e]=A,H_(()=>{t[e]=null}))}function joA(t,e){return t===e||t?.[Og]===e}function Co(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},e=arguments.length>1?arguments[1]:void 0,A=arguments.length>2?arguments[2]:void 0;return es(()=>{var i,n;return tf(()=>{i=n,n=[],nA(()=>{t!==A(...n)&&(e(t,...n),i&&joA(A(...i),t)&&e(null,...i))})}),()=>{Fy(()=>{n&&joA(A(...n),t)&&e(null,...n)})}}),t}function j0(t){return function(){for(var e=arguments.length,A=new Array(e),i=0;i<e;i++)A[i]=arguments[i];return A[0].stopPropagation(),t?.apply(this,A)}}function N1(t){return function(){for(var e=arguments.length,A=new Array(e),i=0;i<e;i++)A[i]=arguments[i];return A[0].preventDefault(),t?.apply(this,A)}}function Zt(){var t=arguments.length>0&&arguments[0]!==void 0&&arguments[0],e=qn,A=e.l.u;if(A){var i,n=()=>k(e.s);if(t){var o=0,r={},s=wQ(()=>{var a=!1,c=e.s;for(var l in c)c[l]!==r[l]&&(r[l]=c[l],a=!0);return a&&o++,o});n=()=>g(s)}A.b.length&&(i=()=>{qoA(e,n),gy(A.b)},osA(),AI(2097160,i,!0)),I_(()=>{var a=nA(()=>A.m.map(V_A));return()=>{for(var c of a)typeof c=="function"&&c()}}),A.a.length&&I_(()=>{qoA(e,n),gy(A.a)})}}function qoA(t,e){if(t.l.s)for(var A of t.l.s)g(A);e()}function Gy(t){var e=jC(0);return function(){return arguments.length===1?(y(e,g(e)+1),arguments[0]):(g(e),t())}}function N3(t,e){var A,i=(A=t.$$events)===null||A===void 0?void 0:A[e.type],n=mQ(i)?i.slice():i==null?[]:[i];for(var o of n)o.call(this,e)}var OD=!1,mGA={get(t,e){if(!t.exclude.includes(e))return g(t.version),e in t.special?t.special[e]():t.props[e]},set(t,e,A){if(!(e in t.special)){var i=Vn;try{V1(t.parent_effect),t.special[e]=b({get[e](){return t.props[e]}},e,4)}finally{V1(i)}}return t.special[e](A),GoA(t.version),!0},getOwnPropertyDescriptor(t,e){if(!t.exclude.includes(e))return e in t.props?{enumerable:!0,configurable:!0,value:t.props[e]}:void 0},deleteProperty:(t,e)=>(t.exclude.includes(e)||(t.exclude.push(e),GoA(t.version)),!0),has:(t,e)=>!t.exclude.includes(e)&&e in t.props,ownKeys:t=>Reflect.ownKeys(t.props).filter(e=>!t.exclude.includes(e))};function PD(t,e){return new Proxy({props:t,exclude:e,special:{},version:jC(0),parent_effect:Vn},mGA)}var pGA={get(t,e){for(var A=t.props.length;A--;){var i=t.props[A];if(R3(i)&&(i=i()),typeof i=="object"&&i!==null&&e in i)return i[e]}},set(t,e,A){for(var i=t.props.length;i--;){var n=t.props[i];R3(n)&&(n=n());var o=$0(n,e);if(o&&o.set)return o.set(A),!0}return!1},getOwnPropertyDescriptor(t,e){for(var A=t.props.length;A--;){var i=t.props[A];if(R3(i)&&(i=i()),typeof i=="object"&&i!==null&&e in i){var n=$0(i,e);return n&&!n.configurable&&(n.configurable=!0),n}}},has(t,e){if(e===Og||e===qrA)return!1;for(var A of t.props)if(R3(A)&&(A=A()),A!=null&&e in A)return!0;return!1},ownKeys(t){var e=[];for(var A of t.props)if(R3(A)&&(A=A()),A){for(var i in A)e.includes(i)||e.push(i);for(var n of Object.getOwnPropertySymbols(A))e.includes(n)||e.push(n)}return e}};function z1(){for(var t=arguments.length,e=new Array(t),A=0;A<t;A++)e[A]=arguments[A];return new Proxy({props:e},pGA)}function b(t,e,A,i){var n,o,r=!fQ||!!(2&A),s=!!(8&A),a=!!(16&A),c=i,l=!0,I=()=>(l&&(l=!1,c=a?nA(i):i),c);if(s){var C,d,B=Og in t||qrA in t;n=(C=(d=$0(t,e))===null||d===void 0?void 0:d.set)!==null&&C!==void 0?C:B&&e in t?w=>t[e]=w:void 0}var E,h=!1;if(s?[o,h]=function(w){var _=OD;try{return OD=!1,[w(),OD]}finally{OD=_}}(()=>t[e]):o=t[e],o===void 0&&i!==void 0&&(o=I(),n&&(r&&function(){throw new Error("https://svelte.dev/e/props_invalid_value")}(),n(o))),E=r?()=>{var w=t[e];return w===void 0?I():(l=!0,w)}:()=>{var w=t[e];return w!==void 0&&(c=void 0),w===void 0?c:w},r&&!(4&A))return E;if(n){var u=t.$$legacy;return function(w,_){return arguments.length>0?(r&&_&&!u&&!h||n(_?E():w),w):E()}}var D=!1,L=(1&A?wQ:qA)(()=>(D=!1,E()));s&&g(L);var R=Vn;return function(w,_){if(arguments.length>0){var K=_?g(L):r&&s?iQ(w):w;return y(L,K),D=!0,c!==void 0&&(c=K),w}return j1&&D||(R.f&Y_)!==0?L.v:g(L)}}function Sr(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:function(i){var n=function(o){try{if(typeof window<"u"&&window.localStorage!==void 0)return window.localStorage[o]}catch{}}("debug");return n!=null&&n.endsWith("*")?i.startsWith(n.slice(0,-1)):i===n}(t);if(!e)return wGA;var A=function(i){for(var n=0,o=0;o<i.length;o++)n=(n<<5)-n+i.charCodeAt(o),n|=0;return VoA[Math.abs(n)%VoA.length]}(t);return function(){for(var i=arguments.length,n=new Array(i),o=0;o<i;o++)n[o]=arguments[o];console.log("%c".concat(t),"color:".concat(A),...n)}}function wGA(){}var VoA=["#0000CC","#0099FF","#009400","#8dd200","#CCCC00","#CC9933","#ae04e7","#ff35d7","#FF3333","#FF6600","#FF9933","#FFCC33"],DGA=0;function nQ(){return++DGA}function ts(t){return parseInt(t,10)}function j_(t){return yGA.test(t)}var yGA=/^-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?$/;function Cn(t){return typeof t=="object"&&t!==null&&(t.constructor===void 0||t.constructor.name==="Object")}function No(t){return typeof t=="object"&&t!==null&&(t.constructor===void 0||t.constructor.name==="Object"||t.constructor.name==="Array")}function vGA(t){return t===!0||t===!1}function Q_(t){if(typeof t=="number")return t>9466848e5&&isFinite(t)&&Math.floor(t)===t&&!isNaN(new Date(t).valueOf());if(typeof t=="bigint")return Q_(Number(t));try{var e=t&&t.valueOf();if(e!==t)return Q_(e)}catch{return!1}return!1}function LsA(t){(jD=jD||window.document.createElement("div")).style.color="",jD.style.color=t;var e=jD.style.color;return e!==""?e.replace(/\s+/g,"").toLowerCase():void 0}var jD=void 0;function bGA(t){return typeof t=="string"&&t.length<99&&!!LsA(t)}function q_(t,e){if(typeof t=="number"||typeof t=="string"||typeof t=="boolean"||t===void 0)return typeof t;if(typeof t=="bigint")return"number";if(t===null)return"null";if(Array.isArray(t))return"array";if(Cn(t))return"object";var A=e.stringify(t);return A&&j_(A)?"number":A==="true"||A==="false"?"boolean":A==="null"?"null":"unknown"}var MGA=/^https?:\/\/\S+$/;function Uy(t){return typeof t=="string"&&MGA.test(t)}function DQ(t,e){if(t==="")return"";var A=t.trim();return A==="null"?null:A==="true"||A!=="false"&&(j_(A)?e.parse(A):t)}var kGA=[];function ZoA(t,e){if(t.length!==e.length)return!1;for(var A=0;A<t.length;A++)if(t[A]!==e[A])return!1;return!0}function h_(t){var e=arguments.length>1&&arguments[1]!==void 0&&arguments[1],A={};if(!Array.isArray(t))throw new TypeError("Array expected");function i(r,s){(!Array.isArray(r)&&!Cn(r)||e&&s.length>0)&&(A[Ct(s)]=!0),Cn(r)&&Object.keys(r).forEach(a=>{i(r[a],s.concat(a))})}for(var n=Math.min(t.length,1e4),o=0;o<n;o++)i(t[o],kGA);return Object.keys(A).sort().map(ks)}function FsA(t,e,A){if(!(e<=t))for(var i=t;i<e;i++)A(i)}function Ey(t,e){return t.length>e?t.slice(0,e):t}function WoA(t){return fe({},t)}function XoA(t){return Object.values(t)}function $oA(t,e,A,i){var n=t.slice(0),o=n.splice(e,A);return n.splice.apply(n,[e+i,0,...o]),n}function SGA(t,e,A){return t.slice(0,e).concat(A).concat(t.slice(e))}function rf(t,e){try{return e.parse(t)}catch{return e.parse(Rc(t))}}function NsA(t,e){try{return rf(t,e)}catch{return}}function sf(t,e){t=t.replace(GsA,"");try{return e(t)}catch{}try{return e("{"+t+"}")}catch{}try{return e("["+t+"]")}catch{}throw new Error("Failed to parse partial JSON")}function _sA(t){t=t.replace(GsA,"");try{return Rc(t)}catch{}try{var e=Rc("["+t+"]");return e.substring(1,e.length-1)}catch{}try{var A=Rc("{"+t+"}");return A.substring(1,A.length-1)}catch{}throw new Error("Failed to repair partial JSON")}var GsA=/,\s*$/;function dQ(t,e){var A=erA.exec(e);if(A){var i=ts(A[2]),n=function(d,B){for(var E=arguments.length>2&&arguments[2]!==void 0?arguments[2]:0,h=arguments.length>3&&arguments[3]!==void 0?arguments[3]:d.length,u=0,D=E;D<h;D++)d.charAt(D)===B&&u++;return u}(t,`
`,0,i),o=i-t.lastIndexOf(`
`,i)-1;return{position:i,line:n,column:o,message:e.replace(erA,()=>"line ".concat(n+1," column ").concat(o+1))}}var r=FGA.exec(e),s=r?ts(r[1]):void 0,a=s!==void 0?s-1:void 0,c=NGA.exec(e),l=c?ts(c[1]):void 0,I=l!==void 0?l-1:void 0,C=a!==void 0&&I!==void 0?function(d,B,E){for(var h=d.indexOf(`
`),u=1;u<B&&h!==-1;)h=d.indexOf(`
`,h+1),u++;return h!==-1?h+E+1:void 0}(t,a,I):void 0;return{position:C,line:a,column:I,message:e.replace(/^JSON.parse: /,"").replace(/ of the JSON data$/,"")}}function HN(t){return Cn(t)?t.json!==void 0?t.text!==void 0?'Content must contain either a property "json" or a property "text" but not both':void 0:t.text===void 0?'Content must contain either a property "json" or a property "text"':typeof t.text!="string"?'Content "text" property must be a string containing a JSON document. Did you mean to use the "json" property instead?':void 0:"Content must be an object"}function H3(t){return Cn(t)&&typeof t.text=="string"}function z3(t){return Cn(t)&&t.json!==void 0}function RGA(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:void 0,A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:JSON;return H3(t)?t:{text:A.stringify(t.json,null,e)}}function ArA(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:JSON;return z3(t)?t:{json:e.parse(t.text)}}function u_(t,e,A){return RGA(t,e,A).text}function xGA(t,e){return LGA(t,e)>e}function LGA(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:1/0;if(H3(t))return t.text.length;var A=t.json,i=0;return function n(o){if(Array.isArray(o)){if((i+=o.length-1+2)>e)return;for(var r=0;r<o.length;r++)if(n(o[r]),i>e)return}else if(Cn(o)){var s=Object.keys(o);i+=2+s.length+(s.length-1);for(var a=0;a<s.length;a++){var c=s[a],l=o[c];i+=c.length+2,n(l)}}else i+=typeof o=="string"?o.length+2:String(o).length}(A),i}var erA=/(position|char) (\d+)/,FGA=/line (\d+)/,NGA=/column (\d+)/;function UsA(t,e){return t.parse===e.parse&&t.stringify===e.stringify}function trA(t){var e=t.substring(0,999).trim();return!e.includes(`
`)&&_GA.test(e)}var er,Ln,Rl,Fl,Nl,Pc,O1,_GA=/[,:]\S/;function V_(t){var{escapeControlCharacters:e,escapeUnicodeCharacters:A}=t;return e?A?GGA:UGA:A?KGA:YGA}(function(t){t.text="text",t.tree="tree",t.table="table"})(er||(er={})),function(t){t.after="after",t.inside="inside",t.key="key",t.value="value",t.multi="multi",t.text="text"}(Ln||(Ln={})),function(t){t.after="after",t.key="key",t.value="value",t.inside="inside"}(Rl||(Rl={})),function(t){t.info="info",t.warning="warning",t.error="error"}(Fl||(Fl={})),function(t){t.key="key",t.value="value"}(Nl||(Nl={})),function(t){t.asc="asc",t.desc="desc"}(Pc||(Pc={})),function(t){t.no="no",t.self="self",t.nextInside="nextInside"}(O1||(O1={}));var GGA={escapeValue:t=>KsA(TsA(String(t))),unescapeValue:t=>HsA(YsA(t))},UGA={escapeValue:t=>TsA(String(t)),unescapeValue:t=>HsA(t)},KGA={escapeValue:t=>KsA(String(t)),unescapeValue:t=>YsA(t)},YGA={escapeValue:t=>String(t),unescapeValue:t=>t};function KsA(t){return t.replace(/[^\x20-\x7F]/g,e=>{var A;return e==="\b"||e==="\f"||e===`
`||e==="\r"||e==="	"?e:"\\u"+("000"+((A=e.codePointAt(0))===null||A===void 0?void 0:A.toString(16))).slice(-4)})}function YsA(t){return t.replace(/\\u[a-fA-F0-9]{4}/g,e=>{try{var A=JSON.parse('"'+e+'"');return JsA[A]||A}catch{return e}})}var JsA={'"':'\\"',"\\":"\\\\","\b":"\\b","\f":"\\f","\n":"\\n","\r":"\\r","	":"\\t"},JGA={'\\"':'"',"\\\\":"\\","\\/":"/","\\b":"\b","\\f":"\f","\\n":`
`,"\\r":"\r","\\t":"	"};function TsA(t){return t.replace(/["\b\f\n\r\t\\]/g,e=>JsA[e]||e)}function HsA(t){return t.replace(/\\["bfnrt\\]/g,e=>JGA[e]||e)}function BQ(t){return typeof t!="string"?String(t):t.endsWith(`
`)?t+`
`:t}function zsA(t,e){return yQ(t,A=>A.nodeName.toUpperCase()===e.toUpperCase())}function Y1(t,e,A){return yQ(t,i=>function(n,o,r){return typeof n.getAttribute=="function"&&n.getAttribute(o)===r}(i,e,A))}function yQ(t,e){return!!Z_(t,e)}function Z_(t,e){for(var A=t;A&&!e(A);)A=A.parentNode;return A}function af(t){var e,A;return(e=t==null||(A=t.ownerDocument)===null||A===void 0?void 0:A.defaultView)!==null&&e!==void 0?e:void 0}function W_(t){var e=af(t),A=e?.document.activeElement;return!!A&&yQ(A,i=>i===t)}function OsA(t,e){return Z_(t,A=>A.nodeName===e)}function zN(t){return Y1(t,"data-type","selectable-key")?Ln.key:Y1(t,"data-type","selectable-value")?Ln.value:Y1(t,"data-type","insert-selection-area-inside")?Ln.inside:Y1(t,"data-type","insert-selection-area-after")?Ln.after:Ln.multi}function sy(t){return encodeURIComponent(Ct(t))}function PsA(t){var e,A=Z_(t,n=>!(n==null||!n.hasAttribute)&&n.hasAttribute("data-path")),i=(e=A?.getAttribute("data-path"))!==null&&e!==void 0?e:void 0;return i?ks(decodeURIComponent(i)):void 0}function TGA(t){var{allElements:e,currentElement:A,direction:i,hasPrio:n=()=>!0,margin:o=10}=t,r=FS(e.filter(function(u){var D=u.getBoundingClientRect();return D.width>0&&D.height>0}),a),s=a(A);function a(u){var D=u.getBoundingClientRect();return{x:D.left+D.width/2,y:D.top+D.height/2,rect:D,element:u}}function c(u,D){var L=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1,R=u.x-D.x,w=(u.y-D.y)*L;return Math.sqrt(R*R+w*w)}var l=u=>c(u,s);if(i==="Left"||i==="Right"){var I=i==="Left"?r.filter(u=>{return D=s,u.rect.left+o<D.rect.left;var D}):r.filter(u=>{return D=s,u.rect.right>D.rect.right+o;var D}),C=I.filter(u=>{return D=u,L=s,Math.abs(D.y-L.y)<o;var D,L}),d=rE(C,l)||rE(I,u=>c(u,s,10));return d?.element}if(i==="Up"||i==="Down"){var B=i==="Up"?r.filter(u=>{return D=s,u.y+o<D.y;var D}):r.filter(u=>{return D=s,u.y>D.y+o;var D}),E=B.filter(u=>n(u.element)),h=rE(E,l)||rE(B,l);return h?.element}}function X_(){var t,e,A,i;return typeof navigator<"u"&&(t=(e=(A=navigator)===null||A===void 0||(A=A.platform)===null||A===void 0?void 0:A.toUpperCase().includes("MAC"))!==null&&e!==void 0?e:(i=navigator)===null||i===void 0||(i=i.userAgentData)===null||i===void 0||(i=i.platform)===null||i===void 0?void 0:i.toUpperCase().includes("MAC"))!==null&&t!==void 0&&t}function n2(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:"+",A=[];$_(t,arguments.length>2&&arguments[2]!==void 0?arguments[2]:X_)&&A.push("Ctrl"),t.altKey&&A.push("Alt"),t.shiftKey&&A.push("Shift");var i=t.key.length===1?t.key.toUpperCase():t.key;return i in HGA||A.push(i),A.join(e)}function $_(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:X_;return t.ctrlKey||t.metaKey&&e()}var HGA={Ctrl:!0,Command:!0,Control:!0,Alt:!0,Option:!0,Shift:!0};function Jt(t,e){e===void 0&&(e={});var A=e.insertAt;if(t&&typeof document<"u"){var i=document.head||document.getElementsByTagName("head")[0],n=document.createElement("style");n.type="text/css",A==="top"&&i.firstChild?i.insertBefore(n,i.firstChild):i.appendChild(n),n.styleSheet?n.styleSheet.cssText=t:n.appendChild(document.createTextNode(t))}}Jt(`.jse-absolute-popup.svelte-1r8q3m8 {
  position: relative;
  left: 0;
  top: 0;
  width: 0;
  height: 0;
  z-index: 1001;
}
.jse-absolute-popup.svelte-1r8q3m8 .jse-hidden-input:where(.svelte-1r8q3m8) {
  position: fixed;
  left: 0;
  top: 0;
  width: 0;
  height: 0;
  padding: 0;
  margin: 0;
  border: none;
  outline: none;
  overflow: hidden;
}
.jse-absolute-popup.svelte-1r8q3m8 .jse-absolute-popup-content:where(.svelte-1r8q3m8) {
  position: absolute;
}`);var zGA=wA('<div class="jse-absolute-popup-content svelte-1r8q3m8"><input type="text" readonly="" tabindex="-1" class="jse-hidden-input svelte-1r8q3m8"/> <!></div>'),OGA=wA('<div role="none" class="jse-absolute-popup svelte-1r8q3m8"><!></div>');function PGA(t,e){mt(e,!1);var A=b(e,"popup",8),i=b(e,"closeAbsolutePopup",8),n=$(),o=$();function r(I){A().options&&A().options.closeOnOuterClick&&!yQ(I.target,C=>C===g(n))&&i()(A().id)}function s(I){n2(I)==="Escape"&&(I.preventDefault(),I.stopPropagation(),i()(A().id))}hs(function(){g(o)&&g(o).focus()}),Zt();var a=OGA();ce("mousedown",A2,function(I){r(I)},!0),ce("keydown",A2,s,!0),ce("wheel",A2,function(I){r(I)},!0);var c=W(a),l=I=>{var C=zGA(),d=W(C);Co(d,B=>y(o,B),()=>g(o)),MsA(IA(d,2),()=>A().component,(B,E)=>{E(B,z1(()=>A().props))}),De(B=>Ll(C,B),[()=>(g(n),k(A()),nA(()=>function(B,E){var h=B.getBoundingClientRect(),{left:u,top:D,positionAbove:L,positionLeft:R}=function(){if(E.anchor){var{anchor:w,width:_=0,height:K=0,offsetTop:z=0,offsetLeft:U=0,position:H}=E,{left:q,top:j,bottom:gA,right:QA}=w.getBoundingClientRect(),BA=H==="top"||j+K>window.innerHeight&&j>K,lA=H==="left"||q+_>window.innerWidth&&q>_;return{left:lA?QA-U:q+U,top:BA?j-z:gA+z,positionAbove:BA,positionLeft:lA}}if(typeof E.left=="number"&&typeof E.top=="number"){var{left:vA,top:tA,width:cA=0,height:pA=0}=E;return{left:vA,top:tA,positionAbove:tA+pA>window.innerHeight&&tA>pA,positionLeft:vA+cA>window.innerWidth&&vA>cA}}throw new Error('Invalid config: pass either "left" and "top", or pass "anchor"')}();return(L?"bottom: ".concat(h.top-D,"px;"):"top: ".concat(D-h.top,"px;"))+(R?"right: ".concat(h.left-u,"px;"):"left: ".concat(u-h.left,"px;"))}(g(n),A().options)))],qA),iA(I,C)};LA(c,I=>{g(n)&&I(l)}),Co(a,I=>y(n,I),()=>g(n)),ce("mousedown",a,function(I){I.stopPropagation()}),ce("keydown",a,s),iA(t,a),pt()}var jGA=wA("<!> <!>",1);function f_(t,e){mt(e,!1);var A,i,n=Sr("jsoneditor:AbsolutePopup"),o=$([],!0);function r(c){var l=g(o).findIndex(C=>C.id===c);if(l!==-1){var I=g(o)[l];I.options.onClose&&I.options.onClose(),y(o,g(o).filter(C=>C.id!==c))}}A="absolute-popup",i={openAbsolutePopup:function(c,l,I){n("open...",l,I);var C={id:nQ(),component:c,props:l||{},options:I||{}};return y(o,[...g(o),C]),C.id},closeAbsolutePopup:r},XrA().set(A,i),fA(()=>g(o),()=>{n("popups",g(o))}),Qn(),Zt(!0);var s=jGA(),a=vt(s);qo(a,1,()=>g(o),or,(c,l)=>{PGA(c,{get popup(){return g(l)},closeAbsolutePopup:r})}),jo(IA(a,2),e,"default",{},null),iA(t,s),pt()}function cf(t,e){for(var A=new Set(e),i=t.replace(/ \(copy( \d+)?\)$/,""),n=t,o=1;A.has(n);){var r="copy"+(o>1?" "+o:"");n="".concat(i," (").concat(r,")"),o++}return n}function V0(t,e){var A=e-3;return t.length>e?t.substring(0,A)+"...":t}function qGA(t){if(t==="")return"";var e=t.toLowerCase();if(e==="null")return null;if(e==="true")return!0;if(e==="false")return!1;if(e!=="undefined"){var A=Number(t),i=parseFloat(t);return isNaN(A)||isNaN(i)?t:A}}var VGA={id:"jsonquery",name:"JSONQuery",description:`
<p>
  Enter a <a href="https://jsonquerylang.org" target="_blank" 
  rel="noopener noreferrer">JSON Query</a> function to filter, sort, or transform the data.
  You can use functions like <code>get</code>, <code>filter</code>,
  <code>sort</code>, <code>pick</code>, <code>groupBy</code>, <code>uniq</code>, etcetera. 
  Example query: <code>filter(.age >= 18)</code>
</p>
`,createQuery:function(t,e){var{filter:A,sort:i,projection:n}=e,o=[];A&&A.path&&A.relation&&A.value&&o.push(["filter",[(r=A.relation,HS("1 ".concat(r," 1"))[0]),qD(A.path),qGA(A.value)]]);var r;return i&&i.path&&i.direction&&o.push(["sort",qD(i.path),i.direction==="desc"?"desc":"asc"]),n&&n.paths&&(n.paths.length>1?o.push(["pick",...n.paths.map(qD)]):o.push(["map",qD(n.paths[0])])),SW(["pipe",...o])},executeQuery:function(t,e,A){var i=UsA(A,JSON)?t:function(n){var o=A.stringify(n);return o!==void 0?JSON.parse(o):void 0}(t);return e.trim()!==""?RW(i,e):i}};function qD(t){return["get",...t]}var ZGA=tI("<g><!></g>");function WGA(t,e){mt(e,!1);var A=870711,i=$(""),n=b(e,"data",8);function o(s){if(!s||!s.raw)return"";var a=s.raw,c={};return a=a.replace(/\s(?:xml:)?id=["']?([^"')\s]+)/g,(l,I)=>{var C="fa-".concat((A+=1).toString(16));return c[I]=C,' id="'.concat(C,'"')}),a=a.replace(/#(?:([^'")\s]+)|xpointer\(id\((['"]?)([^')]+)\2\)\))/g,(l,I,C,d)=>{var B=I||d;return B&&c[B]?"#".concat(c[B]):l}),a}fA(()=>k(n()),()=>{y(i,o(n()))}),Qn();var r=ZGA();bsA(W(r),()=>g(i),!0),iA(t,r),pt()}Jt(`
  .fa-icon.svelte-1mc5hvj {
    display: inline-block;
    fill: currentColor;
  }
  .fa-flip-horizontal.svelte-1mc5hvj {
    transform: scale(-1, 1);
  }
  .fa-flip-vertical.svelte-1mc5hvj {
    transform: scale(1, -1);
  }
  .fa-spin.svelte-1mc5hvj {
    animation: svelte-1mc5hvj-fa-spin 1s 0s infinite linear;
  }
  .fa-inverse.svelte-1mc5hvj {
    color: #fff;
  }
  .fa-pulse.svelte-1mc5hvj {
    animation: svelte-1mc5hvj-fa-spin 1s infinite steps(8);
  }
  @keyframes svelte-1mc5hvj-fa-spin {
    0% {
      transform: rotate(0deg);
    }
    100% {
      transform: rotate(360deg);
    }
  }
`);var XGA=tI("<svg><!></svg>"),$GA=tI("<path></path>"),AUA=tI("<polygon></polygon>"),eUA=tI("<!><!><!>",1);function ji(t,e){var A=PD(e,["children","$$slots","$$events","$$legacy"]),i=PD(A,["class","data","scale","spin","inverse","pulse","flip","label","style"]);mt(e,!1);var n=b(e,"class",8,""),o=b(e,"data",8),r=$(),s=b(e,"scale",8,1),a=b(e,"spin",8,!1),c=b(e,"inverse",8,!1),l=b(e,"pulse",8,!1),I=b(e,"flip",8,void 0),C=b(e,"label",8,""),d=b(e,"style",8,""),B=$(10),E=$(10),h=$(),u=$();function D(){var R=1;return s()!==void 0&&(R=Number(s())),isNaN(R)||R<=0?(console.warn('Invalid prop: prop "scale" should be a number over 0.'),1):1*R}function L(){return g(r)?Math.max(g(r).width,g(r).height)/16:1}fA(()=>(k(o()),k(d()),k(s())),()=>{y(r,function(R){var w;if(R){if(!("definition"in R)){if("iconName"in R&&"icon"in R){R.iconName;var[_,K,,,z]=R.icon;w={width:_,height:K,paths:(Array.isArray(z)?z:[z]).map(U=>({d:U}))}}else w=R[Object.keys(R)[0]];return w}console.error("`import faIconName from '@fortawesome/package-name/faIconName` not supported - Please use `import { faIconName } from '@fortawesome/package-name/faIconName'` instead")}}(o())),d(),s(),y(B,g(r)?g(r).width/L()*D():0),y(E,g(r)?g(r).height/L()*D():0),y(h,function(){var R="";d()!==null&&(R+=d());var w=D();return w===1?R.length===0?"":R:(R===""||R.endsWith(";")||(R+="; "),"".concat(R,"font-size: ").concat(w,"em"))}()),y(u,g(r)?"0 0 ".concat(g(r).width," ").concat(g(r).height):"0 0 ".concat(g(B)," ").concat(g(E)))}),Qn(),Zt(),function(R,w){var _=PD(w,["children","$$slots","$$events","$$legacy"]),K=PD(_,["class","width","height","box","spin","inverse","pulse","flip","style","label"]),z=b(w,"class",8,""),U=b(w,"width",8),H=b(w,"height",8),q=b(w,"box",8,"0 0 0 0"),j=b(w,"spin",8,!1),gA=b(w,"inverse",8,!1),QA=b(w,"pulse",8,!1),BA=b(w,"flip",8,"none"),lA=b(w,"style",8,""),vA=b(w,"label",8,""),tA=XGA();ry(tA,cA=>{var pA;return fe(fe({version:"1.1",class:"fa-icon ".concat((pA=z())!==null&&pA!==void 0?pA:""),width:U(),height:H(),"aria-label":vA(),role:vA()?"img":"presentation",viewBox:q(),style:lA()},K),{},{[eQ]:cA})},[()=>({"fa-spin":j(),"fa-pulse":QA(),"fa-inverse":gA(),"fa-flip-horizontal":BA()==="horizontal","fa-flip-vertical":BA()==="vertical"})],"svelte-1mc5hvj"),jo(W(tA),w,"default",{},null),iA(R,tA)}(t,z1({get label(){return C()},get width(){return g(B)},get height(){return g(E)},get box(){return g(u)},get style(){return g(h)},get spin(){return a()},get flip(){return I()},get inverse(){return c()},get pulse(){return l()},get class(){return n()}},()=>i,{children:(R,w)=>{var _=_o();jo(vt(_),e,"default",{},K=>{var z=eUA(),U=vt(z);qo(U,1,()=>(g(r),nA(()=>{var gA;return((gA=g(r))===null||gA===void 0?void 0:gA.paths)||[]})),or,(gA,QA)=>{var BA=$GA();ry(BA,()=>fe({},g(QA))),iA(gA,BA)});var H=IA(U);qo(H,1,()=>(g(r),nA(()=>{var gA;return((gA=g(r))===null||gA===void 0?void 0:gA.polygons)||[]})),or,(gA,QA)=>{var BA=AUA();ry(BA,()=>fe({},g(QA))),iA(gA,BA)});var q=IA(H),j=gA=>{WGA(gA,{get data(){return g(r)},set data(QA){y(r,QA)},$$legacy:!0})};LA(q,gA=>{g(r),nA(()=>{var QA;return(QA=g(r))===null||QA===void 0?void 0:QA.raw})&&gA(j)}),iA(K,z)}),iA(R,_)},$$slots:{default:!0}})),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-boolean-toggle.svelte-1ryp01u {
  padding: 0;
  margin: 1px 0 0;
  vertical-align: top;
  display: inline-flex;
  color: var(--jse-value-color-boolean, #ff8c00);
}

.jse-boolean-toggle.svelte-1ryp01u:not(.jse-readonly) {
  cursor: pointer;
}`);var tUA=wA('<div role="checkbox" tabindex="-1"><!></div>');function iUA(t,e){mt(e,!1);var A=b(e,"path",9),i=b(e,"value",9),n=b(e,"readOnly",9),o=b(e,"onPatch",9),r=b(e,"focus",9);Zt(!0);var s,a=tUA(),c=W(a),l=qA(()=>i()===!0?zS:OS);ji(c,{get data(){return g(l)}}),De(I=>{En(a,"aria-checked",i()===!0),s=Vt(a,1,"jse-boolean-toggle svelte-1ryp01u",null,s,I),En(a,"title",n()?"Boolean value ".concat(i()):"Click to toggle this boolean value")},[()=>({"jse-readonly":n()})],qA),ce("mousedown",a,function(I){I.stopPropagation(),n()||(o()([{op:"replace",path:Ct(A()),value:!i()}]),r()())}),iA(t,a),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-color-picker-popup.svelte-s1wu8v .picker_wrapper.popup,
.jse-color-picker-popup.svelte-s1wu8v .picker_wrapper.popup .picker_arrow::before,
.jse-color-picker-popup.svelte-s1wu8v .picker_wrapper.popup .picker_arrow::after {
  background: var(--jse-color-picker-background, var(--jse-panel-background, #ebebeb));
  line-height: normal;
}
.jse-color-picker-popup.svelte-s1wu8v .picker_slider,
.jse-color-picker-popup.svelte-s1wu8v .picker_sl,
.jse-color-picker-popup.svelte-s1wu8v .picker_editor input,
.jse-color-picker-popup.svelte-s1wu8v .picker_sample,
.jse-color-picker-popup.svelte-s1wu8v .picker_done button {
  box-shadow: var(--jse-color-picker-border-box-shadow, #cbcbcb 0 0 0 1px);
}
.jse-color-picker-popup.svelte-s1wu8v .picker_editor input {
  background: var(--jse-background-color, #fff);
  color: var(--jse-text-color, #4d4d4d);
}
.jse-color-picker-popup.svelte-s1wu8v .picker_done button {
  background: var(--jse-button-background, #e0e0e0);
  color: var(--jse-button-color, var(--jse-text-color, #4d4d4d));
}
.jse-color-picker-popup.svelte-s1wu8v .picker_done button:hover {
  background: var(--jse-button-background-highlight, #e7e7e7);
}`);var nUA=wA('<div class="jse-color-picker-popup svelte-s1wu8v"></div>');function oUA(t,e){mt(e,!1);var A=b(e,"color",8),i=b(e,"onChange",8),n=b(e,"showOnTop",8),o=$(),r=()=>{};hs(Yt(function*(){var a,c=new((a=yield import("./chunk-TXJFAAIW.js"))===null||a===void 0?void 0:a.default)({parent:g(o),color:A(),popup:n()?"top":"bottom",onDone(l){var I=l.rgba[3]===1?l.hex.substring(0,7):l.hex;i()(I)}});c.show(),r=()=>{c.destroy()}})),qc(()=>{r()}),Zt();var s=nUA();Co(s,a=>y(o,a),()=>g(o)),iA(t,s),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-color-picker-button.svelte-xeg9n6 {
  font-size: var(--jse-font-size-mono, 14px);
  width: var(--jse-color-picker-button-size, 1em);
  height: var(--jse-color-picker-button-size, 1em);
  box-sizing: border-box;
  padding: 0;
  margin: 2px 0 0 calc(0.5 * var(--jse-padding, 10px));
  display: inline-flex;
  vertical-align: top;
  border: 1px solid var(--jse-text-color, #4d4d4d);
  border-radius: 2px;
  background: inherit;
  outline: none;
}

.jse-color-picker-button.svelte-xeg9n6:not(.jse-readonly) {
  cursor: pointer;
}`);var rUA=wA('<button type="button"></button>');function sUA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),{openAbsolutePopup:n}=$1("absolute-popup"),o=b(e,"path",9),r=b(e,"value",9),s=b(e,"readOnly",9),a=b(e,"onPatch",9),c=b(e,"focus",9);function l(B){a()([{op:"replace",path:Ct(o()),value:B}]),I()}function I(){c()()}fA(()=>k(r()),()=>{y(A,LsA(r()))}),fA(()=>(k(s()),k(r())),()=>{y(i,s()?"Color ".concat(r()):"Click to open a color picker")}),Qn(),Zt(!0);var C,d=rUA();De(B=>{var E;C=Vt(d,1,"jse-color-picker-button svelte-xeg9n6",null,C,B),Ll(d,"background: ".concat((E=g(A))!==null&&E!==void 0?E:"")),En(d,"title",g(i)),En(d,"aria-label",g(i))},[()=>({"jse-readonly":s()})],qA),ce("click",d,function(B){var E,h;if(!s()){var u=B.target,D=u.getBoundingClientRect().top,L=((E=(h=af(u))===null||h===void 0?void 0:h.innerHeight)!==null&&E!==void 0?E:0)-D<300&&D>300,R={color:r(),onChange:l,showOnTop:L};n(oUA,R,{anchor:u,closeOnOuterClick:!0,onClose:I,offsetTop:18,offsetLeft:-8,height:300})}}),iA(t,d),pt()}var ON=1e3,O3=100,VD=100,Qy=2e4,aQ=[{start:0,end:O3}],aUA=1048576,cUA=1048576,PN=10485760,jN="Insert or paste contents, enter [ insert a new array, enter { to insert a new object, or start typing to insert a new value",AG="Open context menu (Click here, right click on the selection, or use the context menu button or Ctrl+Q)",KC="hover-insert-inside",ZD="hover-insert-after",irA="hover-collection",qN="valid",nrA="repairable",Z0=336,W0=260,_3=100,orA={[Pc.asc]:"ascending",[Pc.desc]:"descending"};function jsA(t){for(var e=US(t,s=>s.start),A=[e[0]],i=0;i<e.length;i++){var n=A.length-1,o=A[n],r=e[i];r.start<=o.end?A[n]={start:Math.min(o.start,r.start),end:Math.max(o.end,r.end)}:A.push(r)}return A}function m_(t){return hy(t)+O3}function hy(t){return Math.floor(t/O3)*O3}function p_(t){return!!t&&(t.type==="space"||t.space===!0)}function G1(t){return!!t&&(t.type==="separator"||t.separator===!0)}function rrA(t){return!!t&&t.type==="label"&&typeof t.text=="string"}function q0(t){return!!t&&typeof t.onClick=="function"}function ZE(t){return!!t&&t.type==="dropdown-button"&&q0(t.main)&&Array.isArray(t.items)}function srA(t){return!!t&&t.type==="row"&&Array.isArray(t.items)}function arA(t){return!!t&&t.type==="column"&&Array.isArray(t.items)}function crA(t){return Cn(t)&&Cn(t.parseError)}function lUA(t){return Cn(t)&&Array.isArray(t.validationErrors)}function gUA(t){return Cn(t)&&Array.isArray(t.path)&&typeof t.message=="string"&&"severity"in t}function IUA(t){return Cn(t)&&gUA(t)&&typeof t.isChildError=="boolean"}function CUA(t){return Cn(t)&&typeof t.action=="function"&&Cn(t.props)}function _a(t){return t!==void 0&&t.type==="object"}function Mr(t){return t!==void 0&&t.type==="array"}function eG(t){return t!==void 0&&t.type==="value"}function ZC(t){return _a(t)||Mr(t)}function qsA(t){return t!==void 0&&Array.isArray(t.searchResults)}function uy(t){return!!t&&t.type==="tree"}function lrA(t){return!!t&&t.type==="text"}function grA(t){return!!t&&t.type==="mode"}function w_(t){var{json:e,expand:A}=t,i=function(n){var{json:o,factory:r}=n;return Array.isArray(o)?r.createArrayDocumentState():Cn(o)?r.createObjectDocumentState():o!==void 0?r.createValueDocumentState():void 0}({json:e,factory:nG});return A&&i?Sl(e,i,[],A):i}function tG(){var{expanded:t}=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{expanded:!1};return{type:"array",expanded:t,visibleSections:aQ,items:[]}}function iG(){var{expanded:t}=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{expanded:!1};return{type:"object",expanded:t,properties:{}}}var nG={createObjectDocumentState:iG,createArrayDocumentState:tG,createValueDocumentState:function(){return{type:"value"}}};function VsA(t,e,A,i){var{createObjectDocumentState:n,createArrayDocumentState:o,createValueDocumentState:r}=i;return function s(a,c,l){if(Array.isArray(a)){var I=Mr(c)?c:o();if(l.length===0)return I;var C=ts(l[0]),d=s(a[C],I.items[C],l.slice(1));return cs(I,["items",l[0]],d)}if(Cn(a)){var B=_a(c)?c:n();if(l.length===0)return B;var E=l[0],h=s(a[E],B.properties[E],l.slice(1));return cs(B,["properties",E],h)}return eG(c)?c:r()}(t,e,A)}function Qc(t,e){return P3(t,e,arguments.length>2&&arguments[2]!==void 0?arguments[2]:[],(A,i)=>{if(A!==void 0&&i!==void 0)return Array.isArray(A)?Mr(i)?i:tG({expanded:!!ZC(i)&&i.expanded}):Cn(A)?_a(i)?i:iG({expanded:!!ZC(i)&&i.expanded}):eG(i)?i:void 0},()=>!0)}function P3(t,e,A,i,n){var o=i(t,e,A);if(Array.isArray(t)&&Mr(o)&&n(o)){var r=[];return oG(t,o.visibleSections,a=>{var c=A.concat(String(a)),l=P3(t[a],o.items[a],c,i,n);l!==void 0&&(r[a]=l)}),ZoA(r,o.items)?o:fe(fe({},o),{},{items:r})}if(Cn(t)&&_a(o)&&n(o)){var s={};return Object.keys(t).forEach(a=>{var c=A.concat(a),l=P3(t[a],o.properties[a],c,i,n);l!==void 0&&(s[a]=l)}),ZoA(Object.values(s),Object.values(o.properties))?o:fe(fe({},o),{},{properties:s})}return o}function oG(t,e,A){e.forEach(i=>{var{start:n,end:o}=i;FsA(n,Math.min(t.length,o),A)})}function j3(t,e){for(var A=t,i=[],n=0;n<e.length;){if(Array.isArray(A)){var o=e[n];i.push("items",o),A=A[ts(o)]}else{if(!Cn(A))throw new Error("Cannot convert path: Object or Array expected at index ".concat(n));var r=e[n];i.push("properties",r),A=A[r]}n++}return i}function Sl(t,e,A,i){for(var n=e,o=function(s){var a=A.slice(0,s);n=EQ(t,n,a,(c,l)=>{var I=ZC(l)&&!l.expanded?fe(fe({},l),{},{expanded:!0}):l;return Mr(I)?function(C,d){if(function(h,u){return h.some(D=>u>=D.start&&u<D.end)}(C.visibleSections,d))return C;var B=hy(d),E={start:B,end:m_(B)};return fe(fe({},C),{},{visibleSections:jsA(C.visibleSections.concat(E))})}(I,ts(A[s])):I})},r=0;r<A.length;r++)o(r);return EQ(t,n,A,(s,a)=>function(c,l,I,C){return P3(c,l,I,(d,B,E)=>Array.isArray(d)&&C(E)?Mr(B)?B.expanded?B:fe(fe({},B),{},{expanded:!0}):tG({expanded:!0}):Cn(d)&&C(E)?_a(B)?B.expanded?B:fe(fe({},B),{},{expanded:!0}):iG({expanded:!0}):B,d=>ZC(d)&&d.expanded)}(s,a,[],i))}function IrA(t,e,A,i){return EQ(t,e,A,(n,o)=>i?function(r,s,a){return P3(r,s,a,(c,l)=>CrA(l),()=>!0)}(n,o,A):CrA(o))}function CrA(t){return Mr(t)&&t.expanded?fe(fe({},t),{},{expanded:!1,visibleSections:aQ}):_a(t)&&t.expanded?fe(fe({},t),{},{expanded:!1}):t}function ZsA(t,e,A){var i={json:t,documentState:e},n=A.reduce((o,r)=>({json:Da(o.json,[r]),documentState:dUA(o.json,o.documentState,r)}),i);return{json:n.json,documentState:Qc(n.json,n.documentState)}}function dUA(t,e,A){if(lS(A))return drA(t,e,A,void 0);if(gS(A))return BrA(t,e,A);if(k8(A)){var i=ya(t,A.path),n=zg(t,e,i);return n?Ky(t,e,i,{type:"value",enforceString:n}):e}return S8(A)||j2(A)?function(o,r,s){if(j2(s)&&s.from===s.path)return r;var a=r,c=ya(o,s.from),l=Jg(o,a,c);return j2(s)&&(a=BrA(o,a,{path:s.from})),a=drA(o,a,{path:s.path},l),a}(t,e,A):e}function Jg(t,e,A){try{return Ke(e,j3(t,A))}catch{return}}function rG(t,e,A,i,n){var o=VsA(t,e,A,n);return Pu(o,j3(t,A),r=>{var s=Ke(t,A);return i(s,r)})}function Ky(t,e,A,i){return function(n,o,r,s,a){var c=VsA(n,o,r,a);return cs(c,j3(n,r),s)}(t,e,A,i,nG)}function EQ(t,e,A,i){return rG(t,e,A,i,nG)}function drA(t,e,A,i){var n=ya(t,A.path),o=e;return o=EQ(t,o,Fi(n),(r,s)=>{if(!Mr(s))return s;var a=ts(hi(n)),{items:c,visibleSections:l}=s;return fe(fe({},s),{},{items:a<c.length?SGA(c,a,i!==void 0?[i]:Array(1)):c,visibleSections:WsA(l,a,1)})}),Ky(t,o,n,i)}function BrA(t,e,A){var i=ya(t,A.path),n=Fi(i),o=Ke(t,n);return Array.isArray(o)?EQ(t,e,n,(r,s)=>{if(!Mr(s))return s;var a=ts(hi(i)),{items:c,visibleSections:l}=s;return fe(fe({},s),{},{items:c.slice(0,a).concat(c.slice(a+1)),visibleSections:WsA(l,a,-1)})}):function(r,s,a){var c=j3(r,a);return Ms(s,c)?oC(s,j3(r,a)):s}(t,e,i)}function WsA(t,e,A){return function(i){for(var n=i.slice(0),o=1;o<n.length;)n[o-1].end===n[o].start&&(n[o-1]={start:n[o-1].start,end:n[o].end},n.splice(o)),o++;return n}(t.map(i=>({start:i.start>e?i.start+A:i.start,end:i.end>e?i.end+A:i.end})))}function zg(t,e,A){var i,n=Ke(t,A),o=Jg(t,e,A),r=eG(o)?o.enforceString:void 0;return typeof r=="boolean"?r:typeof(i=n)=="string"&&typeof DQ(i,JSON)!="string"}function lf(t,e){var A=arguments.length>2&&arguments[2]!==void 0&&arguments[2],i=t.indexOf(e);return i!==-1?A?t.slice(i):t.slice(i+1):[]}function sG(t,e){var A=[];return function i(n,o,r){A.push(r),wo(n)&&Mr(o)&&o.expanded&&oG(n,o.visibleSections,s=>{i(n[s],o.items[s],r.concat(String(s)))}),xo(n)&&_a(o)&&o.expanded&&Object.keys(n).forEach(s=>{i(n[s],o.properties[s],r.concat(s))})}(t,e,[]),A}function XsA(t,e){var A=!(arguments.length>2&&arguments[2]!==void 0)||arguments[2],i=[];return function n(o,r){i.push({path:r,type:Rl.value});var s=Jg(t,e,r);if(o&&ZC(s)&&s.expanded){if(A&&i.push({path:r,type:Rl.inside}),wo(o)){var a=Mr(s)?s.visibleSections:aQ;oG(o,a,c=>{var l=r.concat(String(c));n(o[c],l),A&&i.push({path:l,type:Rl.after})})}xo(o)&&Object.keys(o).forEach(c=>{var l=r.concat(c);i.push({path:l,type:Rl.key}),n(o[c],l),A&&i.push({path:l,type:Rl.after})})}}(t,[]),i}function VN(t,e,A){var i=sG(t,e),n=i.map(Ct).indexOf(Ct(A));if(n!==-1&&n<i.length-1)return i[n+1]}function WC(t,e,A){var i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:10240;return Sl(t,e,A,xGA({json:Ke(t,A)},i)?G3:aG)}function ZN(t,e,A){var i=Jg(t,e,A);return ZC(i)&&i.expanded?e:WC(t,e,A)}function G3(t){return t.length===0||t.length===1&&t[0]==="0"}function ErA(t){return t.length===0}function aG(){return!0}function ay(){return!1}function Ua(t){return t&&t.type===Ln.after||!1}function fr(t){return t&&t.type===Ln.inside||!1}function kr(t){return t&&t.type===Ln.key||!1}function an(t){return t&&t.type===Ln.value||!1}function Kn(t){return t&&t.type===Ln.multi||!1}function Yy(t){return Kn(t)&&di(t.focusPath,t.anchorPath)}function q3(t){return Kn(t)||Ua(t)||fr(t)||kr(t)||an(t)}function WN(t){return t&&t.type===Ln.text||!1}function W1(t,e){var A=[];return function(i,n,o){if(n){var r=OC(n),s=et(n);if(di(r,s))return o(r);if(i!==void 0){var a=AaA(r,s);if(r.length===a.length||s.length===a.length)return o(a);var c=_s(r,s),l=X0(i,c),I=P1(i,c),C=i2(i,c,l),d=i2(i,c,I);if(!(C===-1||d===-1)){var B=Ke(i,a);if(xo(B)){for(var E=Object.keys(B),h=C;h<=d;h++){var u=o(a.concat(E[h]));if(u!==void 0)return u}return}if(wo(B)){for(var D=C;D<=d;D++){var L=o(a.concat(String(D)));if(L!==void 0)return L}return}throw new Error("Failed to create selection")}}}}(t,e,i=>{A.push(i)}),A}function $sA(t){return fr(t)?t.path:Fi(et(t))}function X0(t,e){if(!Kn(e))return e.path;var A=i2(t,e,e.anchorPath);return i2(t,e,e.focusPath)<A?e.focusPath:e.anchorPath}function P1(t,e){if(!Kn(e))return e.path;var A=i2(t,e,e.anchorPath);return i2(t,e,e.focusPath)>A?e.focusPath:e.anchorPath}function QrA(t,e,A){var i=arguments.length>3&&arguments[3]!==void 0&&arguments[3];if(A){var n=i?et(A):X0(t,A),o=function(a,c,l){var I=sG(a,c),C=I.map(Ct),d=Ct(l),B=C.indexOf(d);if(B!==-1&&B>0)return I[B-1]}(t,e,n);if(i)return fr(A)||Ua(A)?o!==void 0?_s(n,n):void 0:o!==void 0?_s(OC(A),o):void 0;if(Ua(A)||fr(A))return Ni(n);if(kr(A)){if(o===void 0||o.length===0)return;var r=Fi(o),s=Ke(t,r);return Array.isArray(s)||Oi(o)?Ni(o):o2(o)}return an(A),o!==void 0?Ni(o):void 0}}function hrA(t,e,A,i){if(!A)return{caret:void 0,previous:void 0,next:void 0};var n=XsA(t,e,i),o=n.findIndex(r=>di(r.path,et(A))&&String(r.type)===String(A.type));return{caret:o!==-1?n[o]:void 0,previous:o!==-1&&o>0?n[o-1]:void 0,next:o!==-1&&o<n.length-1?n[o+1]:void 0}}function WE(t,e){for(var A=sG(t,e),i=0;i<A.length-1&&A[i+1].length>A[i].length;)i++;var n=A[i];return n===void 0||n.length===0||Array.isArray(Ke(t,Fi(n)))?Ni(n):o2(n)}function QQ(t,e){if(e.length===1){var A=Fc(e);if(A.op==="replace")return Ni(ya(t,A.path))}if(!Oi(e)&&e.every(r=>r.op==="move")){var i=Fc(e),n=e.slice(1);if((S8(i)||j2(i))&&i.from!==i.path&&n.every(r=>(S8(r)||j2(r))&&r.from===r.path))return o2(ya(t,i.path))}var o=e.filter(r=>r.op!=="test"&&r.op!=="remove"&&(r.op!=="move"||r.from!==r.path)&&typeof r.path=="string").map(r=>ya(t,r.path));if(!Oi(o))return{type:Ln.multi,anchorPath:Fc(o),focusPath:hi(o)}}function AaA(t,e){for(var A=0;A<t.length&&A<e.length&&t[A]===e[A];)A++;return t.slice(0,A)}function fy(t){return kr(t)||an(t)||Yy(t)}function urA(t,e){return fy(e)&&No(Ke(t,et(e)))?et(e):Fi(et(e))}function Pg(t,e){if(t.length<e.length)return!1;for(var A=0;A<e.length;A++)if(t[A]!==e[A])return!1;return!0}function Hg(t){if(br(t)){var{type:e,path:A}=t;return{type:e,path:A}}return t}function o2(t){return{type:Ln.key,path:t}}function cG(t,e){return{type:Ln.key,path:t,edit:!0,initialValue:e}}function Ni(t){return{type:Ln.value,path:t}}function my(t,e){return{type:Ln.value,path:t,edit:!0,initialValue:e}}function r2(t){return{type:Ln.inside,path:t}}function t2(t){return{type:Ln.after,path:t}}function _s(t,e){var A=AaA(t,e),i=t.length>A.length&&e.length>A.length;return{type:Ln.multi,anchorPath:i?A.concat(t[A.length]):A,focusPath:i?A.concat(e[A.length]):A}}function eaA(t,e,A,i){if(kr(e))return String(hi(e.path));if(an(e)){var n=Ke(t,e.path);return typeof n=="string"?n:i.stringify(n,null,A)}if(Kn(e)){if(Oi(e.focusPath))return i.stringify(t,null,A);var o=$sA(e),r=Ke(t,o);if(Array.isArray(r)){if(Yy(e)){var s=Ke(t,e.focusPath);return i.stringify(s,null,A)}return W1(t,e).map(a=>{var c=Ke(t,a);return"".concat(i.stringify(c,null,A),",")}).join(`
`)}return W1(t,e).map(a=>{var c=hi(a),l=Ke(t,a);return"".concat(i.stringify(c),": ").concat(i.stringify(l,null,A),",")}).join(`
`)}}function br(t){return(kr(t)||an(t))&&t.edit===!0}function oQ(t){return kr(t)||an(t)||Kn(t)}function WD(t){return kr(t)||an(t)||Yy(t)}function D_(t){switch(t.type){case Rl.key:return o2(t.path);case Rl.value:return Ni(t.path);case Rl.after:return t2(t.path);case Rl.inside:return r2(t.path)}}function frA(t,e){switch(t){case Ln.key:return o2(e);case Ln.value:return Ni(e);case Ln.after:return t2(e);case Ln.inside:return r2(e);case Ln.multi:case Ln.text:return _s(e,e)}}function XD(t,e,A){if(e)return V3(t,e,A)||Pg(Kn(e)?Fi(e.focusPath):e.path,A)?e:void 0}function V3(t,e,A){if(t===void 0||!e)return!1;if(kr(e)||fr(e)||Ua(e))return di(e.path,A);if(an(e))return Pg(A,e.path);if(Kn(e)){var i=X0(t,e),n=P1(t,e),o=Fi(e.focusPath);if(!Pg(A,o)||A.length<=o.length)return!1;var r=i2(t,e,i),s=i2(t,e,n),a=i2(t,e,A);return a!==-1&&a>=r&&a<=s}return!1}function i2(t,e,A){var i=Fi(e.focusPath);if(!Pg(A,i)||A.length<=i.length)return-1;var n=A[i.length],o=Ke(t,i);if(xo(o))return Object.keys(o).indexOf(n);if(wo(o)){var r=ts(n);if(r<o.length)return r}return-1}function et(t){return Kn(t)?t.focusPath:t.path}function OC(t){return Kn(t)?t.anchorPath:t.path}function Kl(){for(var t=[],e=arguments.length,A=new Array(e),i=0;i<e;i++)A[i]=arguments[i];for(var n of A)if(typeof n=="string"&&t.push(n),n&&typeof n=="object")for(var o in n)Object.hasOwnProperty.call(n,o)&&n[o]&&t.push(o);return t.join(" ")}function taA(t,e,A){return Kl("jse-value","jse-"+q_(t,A),{"jse-url":Uy(t),"jse-empty":typeof t=="string"&&t.length===0,"jse-table-cell":e===er.table})}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-value.jse-string.svelte-f9kmxj {
  color: var(--jse-value-color-string, #008000);
}
.jse-value.jse-object.svelte-f9kmxj, .jse-value.jse-array.svelte-f9kmxj {
  min-width: 16px;
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
}
.jse-value.jse-number.svelte-f9kmxj {
  color: var(--jse-value-color-number, #ee422e);
}
.jse-value.jse-boolean.svelte-f9kmxj {
  color: var(--jse-value-color-boolean, #ff8c00);
}
.jse-value.jse-null.svelte-f9kmxj {
  color: var(--jse-value-color-null, #004ed0);
}
.jse-value.jse-invalid.svelte-f9kmxj {
  color: var(--jse-text-color, #4d4d4d);
}
.jse-value.jse-url.svelte-f9kmxj {
  color: var(--jse-value-color-url, #008000);
  text-decoration: underline;
}

div.jse-editable-div.svelte-f9kmxj {
  min-width: 2em;
  padding: 0 5px;
  box-sizing: border-box;
  outline: none;
  border-radius: 1px;
  vertical-align: top;
  cursor: text !important;
  word-break: normal;
  white-space: pre-wrap;
  overflow-wrap: anywhere;
}
div.jse-editable-div.jse-short-text.svelte-f9kmxj {
  overflow-wrap: normal;
}
div.jse-editable-div.jse-table-cell.svelte-f9kmxj {
  overflow-wrap: normal;
  white-space: nowrap;
}
div.jse-editable-div[contenteditable=true].svelte-f9kmxj {
  outline: var(--jse-edit-outline, 2px solid #656565);
  background: var(--jse-background-color, #fff);
  position: relative;
  display: inline-block;
  border-radius: 0;
  z-index: 3;
}
div.jse-editable-div.jse-empty.svelte-f9kmxj:not(:focus) {
  outline: 1px dotted var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  -moz-outline-radius: 2px;
}
div.jse-editable-div.jse-empty.svelte-f9kmxj::after {
  pointer-events: none;
  color: var(--jse-tag-background, rgba(0, 0, 0, 0.2));
}`);var BUA=wA('<div role="textbox" tabindex="0" contenteditable="true" spellcheck="false"></div>');function iaA(t,e){mt(e,!1);var A=Sr("jsoneditor:EditableDiv"),i=b(e,"value",9),n=b(e,"initialValue",9),o=b(e,"shortText",9,!1),r=b(e,"label",9),s=b(e,"onChange",9),a=b(e,"onCancel",9),c=b(e,"onFind",9),l=b(e,"onPaste",9,$o),I=b(e,"onValueClass",9,()=>""),C=$(void 0,!0),d=$(void 0,!0),B=!1;function E(){return g(C)?function(D){return D.replace(/\n$/,"")}(g(C).innerText):""}function h(D){g(C)&&hc(C,g(C).innerText=BQ(D))}hs(()=>{A("onMount",{value:i(),initialValue:n()}),h(n()!==void 0?n():i()),g(C)&&function(D){if(D.firstChild!=null){var L=document.createRange(),R=window.getSelection();L.setStart(D,1),L.collapse(!0),R?.removeAllRanges(),R?.addRange(L)}else D.focus()}(g(C))}),qc(()=>{var D=E();A("onDestroy",{closed:B,value:i(),newValue:D}),B||D===i()||s()(D,O1.no)}),fA(()=>(k(I()),k(i())),()=>{y(d,I()(i()))}),Qn(),Zt(!0);var u=BUA();Co(u,D=>y(C,D),()=>g(C)),De(D=>{En(u,"aria-label",r()),Vt(u,1,D,"svelte-f9kmxj")},[()=>Z1((k(Kl),g(d),k(o()),nA(()=>Kl("jse-editable-div",g(d),{"jse-short-text":o()}))))],qA),ce("input",u,function(){var D=E();D===""&&h(""),y(d,I()(D))}),ce("keydown",u,function(D){D.stopPropagation();var L=n2(D);if(L==="Escape"&&(D.preventDefault(),B=!0,a()()),L==="Enter"||L==="Tab"){D.preventDefault(),B=!0;var R=E();s()(R,O1.nextInside)}L==="Ctrl+F"&&(D.preventDefault(),c()(!1)),L==="Ctrl+H"&&(D.preventDefault(),c()(!0))}),ce("paste",u,function(D){if(D.stopPropagation(),l()&&D.clipboardData){var L=D.clipboardData.getData("text/plain");l()(L)}}),ce("blur",u,function(){var D=document.hasFocus(),L=E();A("handleBlur",{hasFocus:D,closed:B,value:i(),newValue:L}),document.hasFocus()&&!B&&(B=!0,L!==i()&&s()(L,O1.self))}),iA(t,u),pt()}function EUA(t,e){mt(e,!1);var A=b(e,"path",9),i=b(e,"value",9),n=b(e,"selection",9),o=b(e,"mode",9),r=b(e,"parser",9),s=b(e,"normalization",9),a=b(e,"enforceString",9),c=b(e,"onPatch",9),l=b(e,"onPasteJson",9),I=b(e,"onSelect",9),C=b(e,"onFind",9),d=b(e,"focus",9),B=b(e,"findNextInside",9);function E(L){return a()?L:DQ(L,r())}function h(){I()(Ni(A())),d()()}Zt(!0);var u=qA(()=>(k(s()),k(i()),nA(()=>s().escapeValue(i())))),D=qA(()=>(k(br),k(n()),nA(()=>br(n())?n().initialValue:void 0)));iaA(t,{get value(){return g(u)},get initialValue(){return g(D)},label:"Edit value",onChange:function(L,R){c()([{op:"replace",path:Ct(A()),value:E(s().unescapeValue(L))}],(w,_,K)=>{if(!K||di(A(),et(K)))return{state:_,selection:R===O1.nextInside?B()(A()):Ni(A())}}),d()()},onCancel:h,onPaste:function(L){try{var R=r().parse(L);No(R)&&l()({path:A(),contents:R,onPasteAsJson:()=>{h();var w=[{op:"replace",path:Ct(A()),value:R}];c()(w,(_,K)=>({state:WC(_,K,A())}))}})}catch{}},get onFind(){return C()},onValueClass:function(L){return taA(E(s().unescapeValue(L)),o(),r())}}),pt()}function rQ(t,e,A){var i=Fi(e),n=Ke(t,i);if(wo(n)){var o=ts(hi(e));return A.map((c,l)=>({op:"add",path:Ct(i.concat(String(o+l))),value:c.value}))}if(xo(n)){var r=hi(e),s=Object.keys(n),a=r!==void 0?lf(s,r,!0):[];return[...A.map(c=>{var l=cf(c.key,s);return{op:"add",path:Ct(i.concat(l)),value:c.value}}),...a.map(c=>X1(i,c))]}throw new Error("Cannot create insert operations: parent must be an Object or Array")}function y_(t,e,A){var i=Ke(t,e);if(Array.isArray(i)){var n=i.length;return A.map((o,r)=>({op:"add",path:Ct(e.concat(String(n+r))),value:o.value}))}return A.map(o=>{var r=cf(o.key,Object.keys(i));return{op:"add",path:Ct(e.concat(r)),value:o.value}})}function gf(t,e,A,i){var n=cf(i,e.filter(r=>r!==A)),o=lf(e,A,!1);return[{op:"move",from:Ct(t.concat(A)),path:Ct(t.concat(n))},...o.map(r=>X1(t,r))]}function naA(t,e){var A=hi(e);if(Oi(A))throw new Error("Cannot duplicate root object");var i=Fi(A),n=hi(A),o=Ke(t,i);if(wo(o)){var r=hi(e),s=r?ts(hi(r))+1:0;return[...e.map((l,I)=>({op:"copy",from:Ct(l),path:Ct(i.concat(String(I+s)))}))]}if(xo(o)){var a=Object.keys(o),c=n!==void 0?lf(a,n,!1):[];return[...e.map(l=>{var I=cf(hi(l),a);return{op:"copy",from:Ct(l),path:Ct(i.concat(I))}}),...c.map(l=>X1(i,l))]}throw new Error("Cannot create duplicate operations: parent must be an Object or Array")}function oaA(t,e){if(an(e))return[{op:"move",from:Ct(e.path),path:""}];if(!Kn(e))throw new Error("Cannot create extract operations: parent must be an Object or Array");var A=Fi(e.focusPath),i=Ke(t,A);if(wo(i)){var n=W1(t,e).map(r=>{var s=ts(hi(r));return i[s]});return[{op:"replace",path:"",value:n}]}if(xo(i)){var o={};return W1(t,e).forEach(r=>{var s=String(hi(r));o[s]=i[s]}),[{op:"replace",path:"",value:o}]}throw new Error("Cannot extract: unsupported type of selection "+JSON.stringify(e))}function raA(t,e,A,i){if(kr(e)){var n=NsA(A,i),o=Fi(e.path),r=Ke(t,o);return gf(o,Object.keys(r),hi(e.path),typeof n=="string"?n:A)}if(an(e)||Kn(e)&&Oi(e.focusPath))try{return[{op:"replace",path:Ct(et(e)),value:sf(A,_=>rf(_,i))}]}catch{return[{op:"replace",path:Ct(et(e)),value:A}]}if(Kn(e)){var s=XN(A,i);return function(_,K,z){var U=Fc(K),H=Fi(U),q=Ke(_,H);if(wo(q)){var j=Fc(K),gA=j?ts(hi(j)):0;return[...py(K),...z.map((VA,oe)=>({op:"add",path:Ct(H.concat(String(oe+gA))),value:VA.value}))]}if(xo(q)){var QA=hi(K),BA=Fi(QA),lA=hi(QA),vA=Object.keys(q),tA=lA!==void 0?lf(vA,lA,!1):[],cA=new Set(K.map(VA=>hi(VA))),pA=vA.filter(VA=>!cA.has(VA));return[...py(K),...z.map(VA=>{var oe=cf(VA.key,pA);return{op:"add",path:Ct(BA.concat(oe)),value:VA.value}}),...tA.map(VA=>X1(BA,VA))]}throw new Error("Cannot create replace operations: parent must be an Object or Array")}(t,W1(t,e),s)}if(Ua(e)){var a=XN(A,i),c=e.path,l=Fi(c),I=Ke(t,l);if(wo(I)){var C=ts(hi(c));return rQ(t,l.concat(String(C+1)),a)}if(xo(I)){var d=String(hi(c)),B=Object.keys(I);if(Oi(B)||hi(B)===d)return y_(t,l,a);var E=B.indexOf(d),h=B[E+1];return rQ(t,l.concat(h),a)}throw new Error("Cannot create insert operations: parent must be an Object or Array")}if(fr(e)){var u=XN(A,i),D=e.path,L=Ke(t,D);if(wo(L))return rQ(t,D.concat("0"),u);if(xo(L)){var R=Object.keys(L);if(Oi(R))return y_(t,D,u);var w=Fc(R);return rQ(t,D.concat(w),u)}throw new Error("Cannot create insert operations: parent must be an Object or Array")}throw new Error("Cannot insert: unsupported type of selection "+JSON.stringify(e))}function py(t){return t.map(e=>({op:"remove",path:Ct(e)})).reverse()}function X1(t,e){return{op:"move",from:Ct(t.concat(e)),path:Ct(t.concat(e))}}function XN(t,e){var A=/^\s*{/.test(t),i=/^\s*\[/.test(t),n=NsA(t,e),o=n!==void 0?n:sf(t,r=>rf(r,e));return A&&Cn(o)||i&&Array.isArray(o)?[{key:"New item",value:o}]:Array.isArray(o)?o.map((r,s)=>({key:"New item "+s,value:r})):Cn(o)?Object.keys(o).map(r=>({key:r,value:o[r]})):[{key:"New item",value:o}]}function saA(t,e){if(kr(e)){var A=Fi(e.path),i=Ke(t,A),n=gf(A,Object.keys(i),hi(e.path),"");return{operations:n,newSelection:QQ(t,n)}}if(an(e))return{operations:[{op:"replace",path:Ct(e.path),value:""}],newSelection:e};if(Kn(e)){var o=W1(t,e),r=py(o),s=hi(o);if(Oi(s))return{operations:[{op:"replace",path:"",value:""}],newSelection:Ni([])};var a=Fi(s),c=Ke(t,a);if(wo(c)){var l=Fc(o),I=ts(hi(l));return{operations:r,newSelection:I===0?r2(a):t2(a.concat(String(I-1)))}}if(xo(c)){var C=Object.keys(c),d=Fc(o),B=hi(d),E=C.indexOf(B),h=C[E-1];return{operations:r,newSelection:E===0?r2(a):t2(a.concat(h))}}throw new Error("Cannot create remove operations: parent must be an Object or Array")}throw new Error("Cannot remove: unsupported type of selection "+JSON.stringify(e))}function aaA(t,e){var A=function(i,n){if(Oi(n)||!n.every(j2))return n;var o=[];for(var r of n){var s=mrA(ks(r.from)),a=mrA(ks(r.path));if(!s||!a)return n;o.push({from:s,path:a,operation:r})}var c=o[0].path.parent,l=Ke(i,c);if(!xo(l)||!o.every(B=>function(E,h){return di(E.from.parent,h)&&di(E.path.parent,h)}(B,c)))return n;var I=function(B,E){var h=Object.keys(E),u=h.slice();for(var D of B){var L=u.indexOf(D.from.key);L!==-1&&(u.splice(L,1),u.push(D.path.key))}for(var R=0;R<h.length&&h[R]===u[R];)R++;return u[R]}(o,i),C=B=>B.operation,d=o.filter(B=>B.operation.from!==B.operation.path);return d.some(B=>B.path.key===I)?d.map(C):[X1(c,I),...d.map(C)]}(t,e);return R8(t,A,{before:(i,n,o)=>{if(gS(n)){var r=ks(n.path);return{revertOperations:[...o,...$N(i,r)]}}if(j2(n)){var s=ks(n.from);return{revertOperations:n.from===n.path?[n,...$N(i,s)]:[...o,...$N(i,s)]}}return{document:i}}})}function mrA(t){return t.length>0?{parent:Fi(t),key:hi(t)}:void 0}function $N(t,e){var A=Fi(e),i=hi(e),n=Ke(t,A);return xo(n)?lf(Object.keys(n),i,!1).map(o=>X1(A,o)):[]}function prA(t){var e=t.activeIndex<t.items.length-1?t.activeIndex+1:t.items.length>0?0:-1,A=t.items[e],i=t.items.map((n,o)=>fe(fe({},n),{},{active:o===e}));return fe(fe({},t),{},{items:i,activeItem:A,activeIndex:e})}function wrA(t,e){var A,i=arguments.length>2&&arguments[2]!==void 0?arguments[2]:{},n=t.toLowerCase(),o=(A=i?.maxResults)!==null&&A!==void 0?A:1/0,r=i?.columns,s=[],a=[];function c(h){s.length>=o||s.push(h)}function l(h,u){if(wo(u)){var D=a.length;a.push("0");for(var L=0;L<u.length;L++)if(a[D]=String(L),l(h,u[L]),s.length>=o)return;a.pop()}else if(xo(u)){var R=Object.keys(u),w=a.length;for(var _ of(a.push(""),R))if(a[w]=_,DrA(_,h,a,Nl.key,c),l(h,u[_]),s.length>=o)return;a.pop()}else DrA(String(u),h,a,Nl.value,c)}if(t==="")return[];if(r){if(!Array.isArray(e))throw new Error("json must be an Array when option columns is defined");for(var I=0;I<e.length;I++){a[0]=String(I);for(var C=e[I],d=0;d<r.length;d++){var B=r[d];if(B.length===1)a[1]=B[0];else for(var E=0;E<B.length;E++)a[E+1]=B[E];for(;a.length>B.length+1;)a.pop();l(n,Ke(C,B))}if(s.length>=o)break}return s}return l(n,e),s}function DrA(t,e,A,i,n){var o=t.toLowerCase(),r=0,s=-1,a=-1;do(a=o.indexOf(e,s))!==-1&&(s=a+e.length,n({path:A.slice(0),field:i,fieldIndex:r,start:a,end:s}),r++);while(a!==-1)}function v_(t,e,A,i){return t.substring(0,A)+e+t.substring(i)}function yrA(t,e,A){var i=t;return LS(A,n=>{i=v_(i,e,n.start,n.end)}),i}function QUA(t,e,A,i,n){var{field:o,path:r,start:s,end:a}=i;if(o===Nl.key){var c=Fi(r),l=Ke(t,c),I=hi(r),C=gf(c,Object.keys(l),I,v_(I,A,s,a));return{newSelection:QQ(t,C),operations:C}}if(o===Nl.value){var d=Ke(t,r);if(d===void 0)throw new Error("Cannot replace: path not found ".concat(Ct(r)));var B=typeof d=="string"?d:String(d),E=zg(t,e,r),h=v_(B,A,s,a),u=[{op:"replace",path:Ct(r),value:E?h:DQ(h,n)}];return{newSelection:QQ(t,u),operations:u}}throw new Error("Cannot replace: unknown type of search result field ".concat(o))}function vrA(t){return t.path.concat(t.field,String(t.fieldIndex))}function brA(t){var e=qsA(t)?t.searchResults.filter(A=>A.field===Nl.key):void 0;return e&&e.length>0?e:void 0}function MrA(t){var e=qsA(t)?t.searchResults.filter(A=>A.field===Nl.value):void 0;return e&&e.length>0?e:void 0}var hUA={createObjectDocumentState:()=>({type:"object",properties:{}}),createArrayDocumentState:()=>({type:"array",items:[]}),createValueDocumentState:()=>({type:"value"})};function caA(t,e){return e.reduce((A,i)=>function(n,o,r,s){return rG(n,o,r,s,hUA)}(t,A,i.path,(n,o)=>fe(fe({},o),{},{searchResults:o.searchResults?o.searchResults.concat(i):[i]})),void 0)}function wy(t){var e,A=(e=t?.searchResults)!==null&&e!==void 0?e:[],i=_a(t)?Object.values(t.properties).flatMap(wy):Mr(t)?t.items.flatMap(wy):[];return A.concat(i)}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-highlight.svelte-5fb7bl {
  background-color: var(--jse-search-match-color, #ffe665);
  outline: var(--jse-search-match-outline, none);
}
.jse-highlight.jse-active.svelte-5fb7bl {
  background-color: var(--jse-search-match-active-color, var(--jse-search-match-color, #ffe665));
  outline: var(--jse-search-match-outline, 2px solid #e0be00);
}`);var uUA=wA("<span> </span>");function laA(t,e){mt(e,!1);var A=$(),i=b(e,"text",8),n=b(e,"searchResultItems",8);fA(()=>(k(i()),k(n())),()=>{y(A,function(r,s){var a=[],c=0;for(var l of s){var I=r.slice(c,l.start);I!==""&&a.push({resultIndex:void 0,type:"normal",text:I,active:!1});var C=r.slice(l.start,l.end);a.push({resultIndex:l.resultIndex,type:"highlight",text:C,active:l.active}),c=l.end}var d=hi(s);return d&&d.end<r.length&&a.push({type:"normal",text:r.slice(d.end),resultIndex:void 0,active:!1}),a}(String(i()),n()))}),Qn(),Zt();var o=_o();qo(vt(o),1,()=>g(A),or,(r,s)=>{var a=_o(),c=vt(a),l=C=>{var d=Tr();De(()=>wt(d,(g(s),nA(()=>g(s).text)))),iA(C,d)},I=C=>{var d,B=uUA(),E=W(B);De((h,u,D)=>{d=Vt(B,1,"jse-highlight svelte-5fb7bl",null,d,h),En(B,"data-search-result-index",u),wt(E,D)},[()=>({"jse-active":g(s).active}),()=>(g(s),nA(()=>String(g(s).resultIndex))),()=>(k(BQ),g(s),nA(()=>BQ(g(s).text)))],qA),iA(C,B)};LA(c,C=>{g(s),nA(()=>g(s).type==="normal")?C(l):C(I,!1)}),iA(r,a)}),iA(t,o),pt()}function cy(t){var e=1e3;if(t<900)return t.toFixed()+" B";var A=t/e;if(A<900)return A.toFixed(1)+" KB";var i=A/e;if(i<900)return i.toFixed(1)+" MB";var n=i/e;return n<900?n.toFixed(1)+" GB":(n/e).toFixed(1)+" TB"}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-tag.svelte-jlw0fj {
  border: none;
  font-size: 80%;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  color: var(--jse-tag-color, var(--jse-text-color-inverse, #fff));
  background: var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  border-radius: 2px;
  cursor: pointer;
  display: inline-block;
  padding: 0 4px;
  line-height: normal;
  margin: 1px 0;
}
.jse-tag.svelte-jlw0fj:hover {
  opacity: 0.8;
}
.jse-tag.disabled.svelte-jlw0fj {
  opacity: 0.7;
  cursor: inherit;
}`);var fUA=wA('<button type="button"><!></button>');function ly(t,e){mt(e,!0);var A,i=Ga(()=>e.onclick?o=>{o.preventDefault(),o.stopPropagation(),e.onclick()}:void 0),n=fUA();n.__click=function(){for(var o,r=arguments.length,s=new Array(r),a=0;a<r;a++)s[a]=arguments[a];(o=g(i))===null||o===void 0||o.apply(this,s)},function(o,r){for(var s=arguments.length,a=new Array(s>2?s-2:0),c=2;c<s;c++)a[c-2]=arguments[c];var l,I=o,C=FoA;eI(()=>{C!==(C=r())&&(l&&(jc(l),l=null),l=Vg(()=>C(I,...a)))},Af)}(W(n),()=>{var o;return(o=e.children)!==null&&o!==void 0?o:FoA}),De(o=>A=Vt(n,1,"jse-tag svelte-jlw0fj",null,A,o),[()=>({disabled:!e.onclick})]),iA(t,n),pt()}of(["click"]);function mUA(t,e,A){typeof e.value=="string"&&g(A)&&$_(t)&&(t.preventDefault(),t.stopPropagation(),window.open(e.value,"_blank"))}function pUA(t,e){e.readOnly||(t.preventDefault(),e.onSelect(my(e.path)))}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-value.jse-string.svelte-c0g9qz {
  color: var(--jse-value-color-string, #008000);
}
.jse-value.jse-object.svelte-c0g9qz, .jse-value.jse-array.svelte-c0g9qz {
  min-width: 16px;
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
}
.jse-value.jse-number.svelte-c0g9qz {
  color: var(--jse-value-color-number, #ee422e);
}
.jse-value.jse-boolean.svelte-c0g9qz {
  color: var(--jse-value-color-boolean, #ff8c00);
}
.jse-value.jse-null.svelte-c0g9qz {
  color: var(--jse-value-color-null, #004ed0);
}
.jse-value.jse-invalid.svelte-c0g9qz {
  color: var(--jse-text-color, #4d4d4d);
}
.jse-value.jse-url.svelte-c0g9qz {
  color: var(--jse-value-color-url, #008000);
  text-decoration: underline;
}

.jse-value.svelte-c0g9qz {
  display: inline-block;
  min-width: 2em;
  padding: 0 5px;
  box-sizing: border-box;
  outline: none;
  border-radius: 1px;
  vertical-align: top;
  word-break: normal;
  overflow-wrap: anywhere;
  white-space: pre-wrap;
}
.jse-value.jse-table-cell.svelte-c0g9qz {
  overflow-wrap: normal;
  white-space: nowrap;
}
.jse-value.jse-empty.svelte-c0g9qz {
  min-width: 4em;
  outline: 1px dotted var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  -moz-outline-radius: 2px;
}
.jse-value.jse-empty.svelte-c0g9qz::after {
  pointer-events: none;
  color: var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  content: "value";
}`);var wUA=wA('<div role="button" tabindex="-1" data-type="selectable-value"><!> <!></div>');function DUA(t,e){mt(e,!0);var A=P0(!0),i=Ga(()=>g(A)&&typeof e.value=="string"&&e.value.length>e.truncateTextSize&&(!e.searchResultItems||!e.searchResultItems.some(d=>d.active&&d.end>e.truncateTextSize))),n=Ga(()=>g(i)&&typeof e.value=="string"?e.value.substring(0,e.truncateTextSize).trim():e.value),o=Ga(()=>Uy(e.value));function r(){y(A,!1)}var s=wUA();s.__click=[mUA,e,o],s.__dblclick=[pUA,e];var a=W(s),c=d=>{var B=Ga(()=>e.normalization.escapeValue(g(n)));laA(d,{get text(){return g(B)},get searchResultItems(){return e.searchResultItems}})},l=d=>{var B=Tr();De(E=>wt(B,E),[()=>BQ(e.normalization.escapeValue(g(n)))]),iA(d,B)};LA(a,d=>{e.searchResultItems?d(c):d(l,!1)});var I=IA(a,2),C=d=>{ly(d,{onclick:r,children:(B,E)=>{var h=Tr();De(u=>wt(h,"Show more (".concat(u??"",")")),[()=>cy(e.value.length)]),iA(B,h)},$$slots:{default:!0}})};LA(I,d=>{g(i)&&typeof e.value=="string"&&d(C)}),De(d=>{Vt(s,1,d,"svelte-c0g9qz"),En(s,"title",g(o)?"Ctrl+Click or Ctrl+Enter to open url in new window":void 0)},[()=>Z1(taA(e.value,e.mode,e.parser))]),iA(t,s),pt()}of(["click","dblclick"]);Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-tooltip.svelte-14y3y8t {
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  line-height: normal;
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
  border-radius: 3px;
  background: var(--jse-context-menu-background, #656565);
  color: var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff));
  white-space: nowrap;
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
}`);var yUA=wA('<div class="jse-tooltip svelte-14y3y8t"> </div>');function vUA(t,e){var A=b(e,"text",8),i=yUA(),n=W(i);De(()=>wt(n,A())),iA(t,i)}function hQ(t,e){var A,{text:i,openAbsolutePopup:n,closeAbsolutePopup:o}=e;function r(){A=n(vUA,{text:i},{position:"top",width:10*i.length,offsetTop:3,anchor:t,closeOnOuterClick:!0})}function s(){o(A)}return t.addEventListener("mouseenter",r),t.addEventListener("mouseleave",s),{destroy(){t.removeEventListener("mouseenter",r),t.removeEventListener("mouseleave",s)}}}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-timestamp.svelte-1jla5ec {
  padding: 0;
  margin: 0;
  vertical-align: middle;
  display: inline-flex;
  color: var(--jse-value-color-number, #ee422e);
}`);var bUA=wA('<div class="jse-timestamp svelte-1jla5ec"><!></div>');function MUA(t,e){mt(e,!1);var A=$(void 0,!0),i=$1("absolute-popup"),n=b(e,"value",9);fA(()=>k(n()),()=>{y(A,"Time: ".concat(new Date(n()).toString()))}),Qn(),Zt(!0);var o=bUA();ji(W(o),{get data(){return LW}}),Us(o,(r,s)=>hQ?.(r,s),()=>fe({text:g(A)},i)),iA(t,o),pt()}function kUA(t){var e=[];return!t.isEditing&&vGA(t.value)&&e.push({component:iUA,props:t}),!t.isEditing&&bGA(t.value)&&e.push({component:sUA,props:t}),t.isEditing&&e.push({component:EUA,props:t}),t.isEditing||e.push({component:DUA,props:t}),!t.isEditing&&Q_(t.value)&&e.push({component:MUA,props:t}),e}function Ka(t){return t.map((e,A)=>RUA.test(e)?"["+e+"]":/[.[\]]/.test(e)||e===""?'["'+function(i){return i.replace(/"/g,'\\"')}(e)+'"]':(A>0?".":"")+e).join("")}function SUA(t){for(var e=[],A=0;A<t.length;)t[A]==="."&&A++,t[A]==="["?(A++,t[A]==='"'?(A++,e.push(i(o=>o==='"',!0)),n('"')):e.push(i(o=>o==="]")),n("]")):e.push(i(o=>o==="."||o==="["));function i(o){for(var r=arguments.length>1&&arguments[1]!==void 0&&arguments[1],s="";A<t.length&&!o(t[A]);)r&&t[A]==="\\"&&t[A+1]==='"'?(s+='"',A+=2):(s+=t[A],A++);return s}function n(o){if(t[A]!==o)throw new SyntaxError("Invalid JSON path: ".concat(o," expected at position ").concat(A));A++}return e}function U1(t){return{value:t,label:Oi(t)?"(item root)":Ka(t)}}var RUA=/^\d+$/,xUA={},LUA={showWizard:!0,showOriginal:!0},Dy=Math.min,PC=Math.max,yy=Math.round,$D=Math.floor,jg=t=>({x:t,y:t}),FUA={left:"right",right:"left",bottom:"top",top:"bottom"},NUA={start:"end",end:"start"};function krA(t,e,A){return PC(t,Dy(e,A))}function Jy(t,e){return typeof t=="function"?t(e):t}function XC(t){return t.split("-")[0]}function Ty(t){return t.split("-")[1]}function gaA(t){return t==="x"?"y":"x"}function IaA(t){return t==="y"?"height":"width"}var _UA=new Set(["top","bottom"]);function J1(t){return _UA.has(XC(t))?"y":"x"}function CaA(t){return gaA(J1(t))}function b_(t){return t.replace(/start|end/g,e=>NUA[e])}var SrA=["left","right"],RrA=["right","left"],GUA=["top","bottom"],UUA=["bottom","top"];function KUA(t,e,A,i){var n=Ty(t),o=function(r,s,a){switch(r){case"top":case"bottom":return a?s?RrA:SrA:s?SrA:RrA;case"left":case"right":return s?GUA:UUA;default:return[]}}(XC(t),A==="start",i);return n&&(o=o.map(r=>r+"-"+n),e&&(o=o.concat(o.map(b_)))),o}function Ay(t){return t.replace(/left|right|bottom|top/g,e=>FUA[e])}function YUA(t){return typeof t!="number"?function(e){return fe({top:0,right:0,bottom:0,left:0},e)}(t):{top:t,right:t,bottom:t,left:t}}function vy(t){var{x:e,y:A,width:i,height:n}=t;return{width:i,height:n,top:A,left:e,right:e+i,bottom:A+n,x:e,y:A}}function xrA(t,e,A){var i,{reference:n,floating:o}=t,r=J1(e),s=CaA(e),a=IaA(s),c=XC(e),l=r==="y",I=n.x+n.width/2-o.width/2,C=n.y+n.height/2-o.height/2,d=n[a]/2-o[a]/2;switch(c){case"top":i={x:I,y:n.y-o.height};break;case"bottom":i={x:I,y:n.y+n.height};break;case"right":i={x:n.x+n.width,y:C};break;case"left":i={x:n.x-o.width,y:C};break;default:i={x:n.x,y:n.y}}switch(Ty(e)){case"start":i[s]-=d*(A&&l?-1:1);break;case"end":i[s]+=d*(A&&l?-1:1)}return i}var JUA=function(){var t=Yt(function*(e,A,i){for(var{placement:n="bottom",strategy:o="absolute",middleware:r=[],platform:s}=i,a=r.filter(Boolean),c=yield s.isRTL==null?void 0:s.isRTL(A),l=yield s.getElementRects({reference:e,floating:A,strategy:o}),{x:I,y:C}=xrA(l,n,c),d=n,B={},E=0,h=0;h<a.length;h++){var{name:u,fn:D}=a[h],{x:L,y:R,data:w,reset:_}=yield D({x:I,y:C,initialPlacement:n,placement:d,strategy:o,middlewareData:B,rects:l,platform:s,elements:{reference:e,floating:A}});I=L??I,C=R??C,B=fe(fe({},B),{},{[u]:fe(fe({},B[u]),w)}),_&&E<=50&&(E++,typeof _=="object"&&(_.placement&&(d=_.placement),_.rects&&(l=_.rects===!0?yield s.getElementRects({reference:e,floating:A,strategy:o}):_.rects),{x:I,y:C}=xrA(l,d,c)),h=-1)}return{x:I,y:C,placement:d,strategy:o,middlewareData:B}});return function(e,A,i){return t.apply(this,arguments)}}();function daA(t,e){return M_.apply(this,arguments)}function M_(){return M_=Yt(function*(t,e){var A;e===void 0&&(e={});var{x:i,y:n,platform:o,rects:r,elements:s,strategy:a}=t,{boundary:c="clippingAncestors",rootBoundary:l="viewport",elementContext:I="floating",altBoundary:C=!1,padding:d=0}=Jy(e,t),B=YUA(d),E=s[C?I==="floating"?"reference":"floating":I],h=vy(yield o.getClippingRect({element:(A=yield o.isElement==null?void 0:o.isElement(E))==null||A?E:E.contextElement||(yield o.getDocumentElement==null?void 0:o.getDocumentElement(s.floating)),boundary:c,rootBoundary:l,strategy:a})),u=I==="floating"?{x:i,y:n,width:r.floating.width,height:r.floating.height}:r.reference,D=yield o.getOffsetParent==null?void 0:o.getOffsetParent(s.floating),L=(yield o.isElement==null?void 0:o.isElement(D))&&(yield o.getScale==null?void 0:o.getScale(D))||{x:1,y:1},R=vy(o.convertOffsetParentRelativeRectToViewportRelativeRect?yield o.convertOffsetParentRelativeRectToViewportRelativeRect({elements:s,rect:u,offsetParent:D,strategy:a}):u);return{top:(h.top-R.top+B.top)/L.y,bottom:(R.bottom-h.bottom+B.bottom)/L.y,left:(h.left-R.left+B.left)/L.x,right:(R.right-h.right+B.right)/L.x}}),M_.apply(this,arguments)}var TUA=new Set(["left","top"]);function k_(){return k_=Yt(function*(t,e){var{placement:A,platform:i,elements:n}=t,o=yield i.isRTL==null?void 0:i.isRTL(n.floating),r=XC(A),s=Ty(A),a=J1(A)==="y",c=TUA.has(r)?-1:1,l=o&&a?-1:1,I=Jy(e,t),{mainAxis:C,crossAxis:d,alignmentAxis:B}=typeof I=="number"?{mainAxis:I,crossAxis:0,alignmentAxis:null}:{mainAxis:I.mainAxis||0,crossAxis:I.crossAxis||0,alignmentAxis:I.alignmentAxis};return s&&typeof B=="number"&&(d=s==="end"?-1*B:B),a?{x:d*l,y:C*c}:{x:C*c,y:d*l}}),k_.apply(this,arguments)}function Hy(){return typeof window<"u"}function uQ(t){return BaA(t)?(t.nodeName||"").toLowerCase():"#document"}function fc(t){var e;return(t==null||(e=t.ownerDocument)==null?void 0:e.defaultView)||window}function qg(t){var e;return(e=(BaA(t)?t.ownerDocument:t.document)||window.document)==null?void 0:e.documentElement}function BaA(t){return!!Hy()&&(t instanceof Node||t instanceof fc(t).Node)}function _l(t){return!!Hy()&&(t instanceof Element||t instanceof fc(t).Element)}function Zg(t){return!!Hy()&&(t instanceof HTMLElement||t instanceof fc(t).HTMLElement)}function LrA(t){return!(!Hy()||typeof ShadowRoot>"u")&&(t instanceof ShadowRoot||t instanceof fc(t).ShadowRoot)}var HUA=new Set(["inline","contents"]);function Z3(t){var{overflow:e,overflowX:A,overflowY:i,display:n}=Gl(t);return/auto|scroll|overlay|hidden|clip/.test(e+i+A)&&!HUA.has(n)}var zUA=new Set(["table","td","th"]);function OUA(t){return zUA.has(uQ(t))}var PUA=[":popover-open",":modal"];function by(t){return PUA.some(e=>{try{return t.matches(e)}catch{return!1}})}var jUA=["transform","translate","scale","rotate","perspective"],qUA=["transform","translate","scale","rotate","perspective","filter"],VUA=["paint","layout","strict","content"];function S_(t){var e=lG(),A=_l(t)?Gl(t):t;return jUA.some(i=>!!A[i]&&A[i]!=="none")||!!A.containerType&&A.containerType!=="normal"||!e&&!!A.backdropFilter&&A.backdropFilter!=="none"||!e&&!!A.filter&&A.filter!=="none"||qUA.some(i=>(A.willChange||"").includes(i))||VUA.some(i=>(A.contain||"").includes(i))}function lG(){return!(typeof CSS>"u"||!CSS.supports)&&CSS.supports("-webkit-backdrop-filter","none")}var ZUA=new Set(["html","body","#document"]);function cQ(t){return ZUA.has(uQ(t))}function Gl(t){return fc(t).getComputedStyle(t)}function zy(t){return _l(t)?{scrollLeft:t.scrollLeft,scrollTop:t.scrollTop}:{scrollLeft:t.scrollX,scrollTop:t.scrollY}}function T1(t){if(uQ(t)==="html")return t;var e=t.assignedSlot||t.parentNode||LrA(t)&&t.host||qg(t);return LrA(e)?e.host:e}function EaA(t){var e=T1(t);return cQ(e)?t.ownerDocument?t.ownerDocument.body:t.body:Zg(e)&&Z3(e)?e:EaA(e)}function W3(t,e,A){var i;e===void 0&&(e=[]),A===void 0&&(A=!0);var n=EaA(t),o=n===((i=t.ownerDocument)==null?void 0:i.body),r=fc(n);if(o){var s=R_(r);return e.concat(r,r.visualViewport||[],Z3(n)?n:[],s&&A?W3(s):[])}return e.concat(n,W3(n,[],A))}function R_(t){return t.parent&&Object.getPrototypeOf(t.parent)?t.frameElement:null}function QaA(t){var e=Gl(t),A=parseFloat(e.width)||0,i=parseFloat(e.height)||0,n=Zg(t),o=n?t.offsetWidth:A,r=n?t.offsetHeight:i,s=yy(A)!==o||yy(i)!==r;return s&&(A=o,i=r),{width:A,height:i,$:s}}function gG(t){return _l(t)?t:t.contextElement}function lQ(t){var e=gG(t);if(!Zg(e))return jg(1);var A=e.getBoundingClientRect(),{width:i,height:n,$:o}=QaA(e),r=(o?yy(A.width):A.width)/i,s=(o?yy(A.height):A.height)/n;return r&&Number.isFinite(r)||(r=1),s&&Number.isFinite(s)||(s=1),{x:r,y:s}}var WUA=jg(0);function haA(t){var e=fc(t);return lG()&&e.visualViewport?{x:e.visualViewport.offsetLeft,y:e.visualViewport.offsetTop}:WUA}function $C(t,e,A,i){e===void 0&&(e=!1),A===void 0&&(A=!1);var n=t.getBoundingClientRect(),o=gG(t),r=jg(1);e&&(i?_l(i)&&(r=lQ(i)):r=lQ(t));var s=function(w,_,K){return _===void 0&&(_=!1),!(!K||_&&K!==fc(w))&&_}(o,A,i)?haA(o):jg(0),a=(n.left+s.x)/r.x,c=(n.top+s.y)/r.y,l=n.width/r.x,I=n.height/r.y;if(o)for(var C=fc(o),d=i&&_l(i)?fc(i):i,B=C,E=R_(B);E&&i&&d!==B;){var h=lQ(E),u=E.getBoundingClientRect(),D=Gl(E),L=u.left+(E.clientLeft+parseFloat(D.paddingLeft))*h.x,R=u.top+(E.clientTop+parseFloat(D.paddingTop))*h.y;a*=h.x,c*=h.y,l*=h.x,I*=h.y,a+=L,c+=R,E=R_(B=fc(E))}return vy({width:l,height:I,x:a,y:c})}function IG(t,e){var A=zy(t).scrollLeft;return e?e.left+A:$C(qg(t)).left+A}function uaA(t,e,A){A===void 0&&(A=!1);var i=t.getBoundingClientRect();return{x:i.left+e.scrollLeft-(A?0:IG(t,i)),y:i.top+e.scrollTop}}var XUA=new Set(["absolute","fixed"]);function FrA(t,e,A){var i;if(e==="viewport")i=function(o,r){var s=fc(o),a=qg(o),c=s.visualViewport,l=a.clientWidth,I=a.clientHeight,C=0,d=0;if(c){l=c.width,I=c.height;var B=lG();(!B||B&&r==="fixed")&&(C=c.offsetLeft,d=c.offsetTop)}return{width:l,height:I,x:C,y:d}}(t,A);else if(e==="document")i=function(o){var r=qg(o),s=zy(o),a=o.ownerDocument.body,c=PC(r.scrollWidth,r.clientWidth,a.scrollWidth,a.clientWidth),l=PC(r.scrollHeight,r.clientHeight,a.scrollHeight,a.clientHeight),I=-s.scrollLeft+IG(o),C=-s.scrollTop;return Gl(a).direction==="rtl"&&(I+=PC(r.clientWidth,a.clientWidth)-c),{width:c,height:l,x:I,y:C}}(qg(t));else if(_l(e))i=function(o,r){var s=$C(o,!0,r==="fixed"),a=s.top+o.clientTop,c=s.left+o.clientLeft,l=Zg(o)?lQ(o):jg(1);return{width:o.clientWidth*l.x,height:o.clientHeight*l.y,x:c*l.x,y:a*l.y}}(e,A);else{var n=haA(t);i={x:e.x-n.x,y:e.y-n.y,width:e.width,height:e.height}}return vy(i)}function faA(t,e){var A=T1(t);return!(A===e||!_l(A)||cQ(A))&&(Gl(A).position==="fixed"||faA(A,e))}function $UA(t,e,A){var i=Zg(e),n=qg(e),o=A==="fixed",r=$C(t,!0,o,e),s={scrollLeft:0,scrollTop:0},a=jg(0);function c(){a.x=IG(n)}if(i||!i&&!o)if((uQ(e)!=="body"||Z3(n))&&(s=zy(e)),i){var l=$C(e,!0,o,e);a.x=l.x+e.clientLeft,a.y=l.y+e.clientTop}else n&&c();o&&!i&&n&&c();var I=!n||i||o?jg(0):uaA(n,s);return{x:r.left+s.scrollLeft-a.x-I.x,y:r.top+s.scrollTop-a.y-I.y,width:r.width,height:r.height}}function A_(t){return Gl(t).position==="static"}function NrA(t,e){if(!Zg(t)||Gl(t).position==="fixed")return null;if(e)return e(t);var A=t.offsetParent;return qg(t)===A&&(A=A.ownerDocument.body),A}function _rA(t,e){var A=fc(t);if(by(t))return A;if(!Zg(t)){for(var i=T1(t);i&&!cQ(i);){if(_l(i)&&!A_(i))return i;i=T1(i)}return A}for(var n=NrA(t,e);n&&OUA(n)&&A_(n);)n=NrA(n,e);return n&&cQ(n)&&A_(n)&&!S_(n)?A:n||function(o){for(var r=T1(o);Zg(r)&&!cQ(r);){if(S_(r))return r;if(by(r))return null;r=T1(r)}return null}(t)||A}var AKA={convertOffsetParentRelativeRectToViewportRelativeRect:function(t){var{elements:e,rect:A,offsetParent:i,strategy:n}=t,o=n==="fixed",r=qg(i),s=!!e&&by(e.floating);if(i===r||s&&o)return A;var a={scrollLeft:0,scrollTop:0},c=jg(1),l=jg(0),I=Zg(i);if((I||!I&&!o)&&((uQ(i)!=="body"||Z3(r))&&(a=zy(i)),Zg(i))){var C=$C(i);c=lQ(i),l.x=C.x+i.clientLeft,l.y=C.y+i.clientTop}var d=!r||I||o?jg(0):uaA(r,a,!0);return{width:A.width*c.x,height:A.height*c.y,x:A.x*c.x-a.scrollLeft*c.x+l.x+d.x,y:A.y*c.y-a.scrollTop*c.y+l.y+d.y}},getDocumentElement:qg,getClippingRect:function(t){var{element:e,boundary:A,rootBoundary:i,strategy:n}=t,o=[...A==="clippingAncestors"?by(e)?[]:function(a,c){var l=c.get(a);if(l)return l;for(var I=W3(a,[],!1).filter(u=>_l(u)&&uQ(u)!=="body"),C=null,d=Gl(a).position==="fixed",B=d?T1(a):a;_l(B)&&!cQ(B);){var E=Gl(B),h=S_(B);h||E.position!=="fixed"||(C=null),(d?!h&&!C:!h&&E.position==="static"&&C&&XUA.has(C.position)||Z3(B)&&!h&&faA(a,B))?I=I.filter(u=>u!==B):C=E,B=T1(B)}return c.set(a,I),I}(e,this._c):[].concat(A),i],r=o[0],s=o.reduce((a,c)=>{var l=FrA(e,c,n);return a.top=PC(l.top,a.top),a.right=Dy(l.right,a.right),a.bottom=Dy(l.bottom,a.bottom),a.left=PC(l.left,a.left),a},FrA(e,r,n));return{width:s.right-s.left,height:s.bottom-s.top,x:s.left,y:s.top}},getOffsetParent:_rA,getElementRects:function(){var t=Yt(function*(e){var A=this.getOffsetParent||_rA,i=this.getDimensions,n=yield i(e.floating);return{reference:$UA(e.reference,yield A(e.floating),e.strategy),floating:{x:0,y:0,width:n.width,height:n.height}}});return function(e){return t.apply(this,arguments)}}(),getClientRects:function(t){return Array.from(t.getClientRects())},getDimensions:function(t){var{width:e,height:A}=QaA(t);return{width:e,height:A}},getScale:lQ,isElement:_l,isRTL:function(t){return Gl(t).direction==="rtl"}};function GrA(t,e){return t.x===e.x&&t.y===e.y&&t.width===e.width&&t.height===e.height}function eKA(t,e,A,i){i===void 0&&(i={});var{ancestorScroll:n=!0,ancestorResize:o=!0,elementResize:r=typeof ResizeObserver=="function",layoutShift:s=typeof IntersectionObserver=="function",animationFrame:a=!1}=i,c=gG(t),l=n||o?[...c?W3(c):[],...W3(e)]:[];l.forEach(h=>{n&&h.addEventListener("scroll",A,{passive:!0}),o&&h.addEventListener("resize",A)});var I,C=c&&s?function(h,u){var D,L=null,R=qg(h);function w(){var _;clearTimeout(D),(_=L)==null||_.disconnect(),L=null}return function _(K,z){K===void 0&&(K=!1),z===void 0&&(z=1),w();var U=h.getBoundingClientRect(),{left:H,top:q,width:j,height:gA}=U;if(K||u(),j&&gA){var QA={rootMargin:-$D(q)+"px "+-$D(R.clientWidth-(H+j))+"px "+-$D(R.clientHeight-(q+gA))+"px "+-$D(H)+"px",threshold:PC(0,Dy(1,z))||1},BA=!0;try{L=new IntersectionObserver(lA,fe(fe({},QA),{},{root:R.ownerDocument}))}catch{L=new IntersectionObserver(lA,QA)}L.observe(h)}function lA(vA){var tA=vA[0].intersectionRatio;if(tA!==z){if(!BA)return _();tA?_(!1,tA):D=setTimeout(()=>{_(!1,1e-7)},1e3)}tA!==1||GrA(U,h.getBoundingClientRect())||_(),BA=!1}}(!0),w}(c,A):null,d=-1,B=null;r&&(B=new ResizeObserver(h=>{var[u]=h;u&&u.target===c&&B&&(B.unobserve(e),cancelAnimationFrame(d),d=requestAnimationFrame(()=>{var D;(D=B)==null||D.observe(e)})),A()}),c&&!a&&B.observe(c),B.observe(e));var E=a?$C(t):null;return a&&function h(){var u=$C(t);E&&!GrA(E,u)&&A(),E=u,I=requestAnimationFrame(h)}(),A(),()=>{var h;l.forEach(u=>{n&&u.removeEventListener("scroll",A),o&&u.removeEventListener("resize",A)}),C?.(),(h=B)==null||h.disconnect(),B=null,a&&cancelAnimationFrame(I)}}var tKA=function(t){return t===void 0&&(t=0),{name:"offset",options:t,fn:e=>Yt(function*(){var A,i,{x:n,y:o,placement:r,middlewareData:s}=e,a=yield function(c,l){return k_.apply(this,arguments)}(e,t);return r===((A=s.offset)==null?void 0:A.placement)&&(i=s.arrow)!=null&&i.alignmentOffset?{}:{x:n+a.x,y:o+a.y,data:fe(fe({},a),{},{placement:r})}})()}},iKA=function(t){return t===void 0&&(t={}),{name:"shift",options:t,fn:e=>Yt(function*(){var{x:A,y:i,placement:n}=e,o=Jy(t,e),{mainAxis:r=!0,crossAxis:s=!1,limiter:a={fn:L=>{var{x:R,y:w}=L;return{x:R,y:w}}}}=o,c=zrA(o,T_A),l={x:A,y:i},I=yield daA(e,c),C=J1(XC(n)),d=gaA(C),B=l[d],E=l[C];if(r){var h=d==="y"?"bottom":"right";B=krA(B+I[d==="y"?"top":"left"],B,B-I[h])}if(s){var u=C==="y"?"bottom":"right";E=krA(E+I[C==="y"?"top":"left"],E,E-I[u])}var D=a.fn(fe(fe({},e),{},{[d]:B,[C]:E}));return fe(fe({},D),{},{data:{x:D.x-A,y:D.y-i,enabled:{[d]:r,[C]:s}}})})()}},nKA=function(t){return t===void 0&&(t={}),{name:"flip",options:t,fn:e=>Yt(function*(){var A,i,{placement:n,middlewareData:o,rects:r,initialPlacement:s,platform:a,elements:c}=e,l=Jy(t,e),{mainAxis:I=!0,crossAxis:C=!0,fallbackPlacements:d,fallbackStrategy:B="bestFit",fallbackAxisSideDirection:E="none",flipAlignment:h=!0}=l,u=zrA(l,J_A);if((A=o.arrow)!=null&&A.alignmentOffset)return{};var D=XC(n),L=J1(s),R=XC(s)===s,w=yield a.isRTL==null?void 0:a.isRTL(c.floating),_=d||(R||!h?[Ay(s)]:function(pA){var VA=Ay(pA);return[b_(pA),VA,b_(VA)]}(s)),K=E!=="none";!d&&K&&_.push(...KUA(s,h,E,w));var z=[s,..._],U=yield daA(e,u),H=[],q=((i=o.flip)==null?void 0:i.overflows)||[];if(I&&H.push(U[D]),C){var j=function(pA,VA,oe){oe===void 0&&(oe=!1);var KA=Ty(pA),CA=CaA(pA),TA=IaA(CA),Ze=CA==="x"?KA===(oe?"end":"start")?"right":"left":KA==="start"?"bottom":"top";return VA.reference[TA]>VA.floating[TA]&&(Ze=Ay(Ze)),[Ze,Ay(Ze)]}(n,r,w);H.push(U[j[0]],U[j[1]])}if(q=[...q,{placement:n,overflows:H}],!H.every(pA=>pA<=0)){var gA,QA,BA=(((gA=o.flip)==null?void 0:gA.index)||0)+1,lA=z[BA];if(lA&&(!(C==="alignment"&&L!==J1(lA))||q.every(pA=>pA.overflows[0]>0&&J1(pA.placement)===L)))return{data:{index:BA,overflows:q},reset:{placement:lA}};var vA=(QA=q.filter(pA=>pA.overflows[0]<=0).sort((pA,VA)=>pA.overflows[1]-VA.overflows[1])[0])==null?void 0:QA.placement;if(!vA)switch(B){case"bestFit":var tA,cA=(tA=q.filter(pA=>{if(K){var VA=J1(pA.placement);return VA===L||VA==="y"}return!0}).map(pA=>[pA.placement,pA.overflows.filter(VA=>VA>0).reduce((VA,oe)=>VA+oe,0)]).sort((pA,VA)=>pA[1]-VA[1])[0])==null?void 0:tA[0];cA&&(vA=cA);break;case"initialPlacement":vA=s}if(n!==vA)return{reset:{placement:vA}}}return{}})()}};function oKA(t){var e,A,i={autoUpdate:!0},n=t,o=a=>fe(fe(fe({},i),t||{}),a||{}),r=a=>{e&&A&&(n=o(a),((c,l,I)=>{var C=new Map,d=fe({platform:AKA},I),B=fe(fe({},d.platform),{},{_c:C});return JUA(c,l,fe(fe({},d),{},{platform:B}))})(e,A,n).then(c=>{var l;Object.assign(A.style,{position:c.strategy,left:"".concat(c.x,"px"),top:"".concat(c.y,"px")}),!((l=n)===null||l===void 0)&&l.onComputed&&n.onComputed(c)}))},s=a=>{qc(a.subscribe(c=>{e===void 0?(e=c,r()):(Object.assign(e,c),r())}))};return[a=>{if("subscribe"in a)return s(a),{};e=a,r()},(a,c)=>{var l;A=a,n=o(c),setTimeout(()=>r(c),0),r(c);var I=()=>{l&&(l(),l=void 0)},C=function(){var{autoUpdate:d}=arguments.length>0&&arguments[0]!==void 0?arguments[0]:n||{};I(),d!==!1&&function(){return fsA.apply(this,arguments)}().then(()=>eKA(e,A,()=>r(n),d===!0?{}:d))};return l=C(),{update(d){r(d),l=C(d)},destroy(){I()}}},r]}function rKA(t){var{loadOptions:e,filterText:A,items:i,multiple:n,value:o,itemId:r,groupBy:s,filterSelectedItems:a,itemFilter:c,convertStringItemsToObjects:l,filterGroupedItems:I,label:C}=t;if(i&&e)return i;if(!i)return[];i&&i.length>0&&typeof i[0]!="object"&&(i=l(i));var d=i.filter(B=>{var E=c(B[C],A,B);return E&&n&&o!=null&&o.length&&(E=!o.some(h=>!!a&&h[r]===B[r])),E});return s&&(d=I(d)),d}function sKA(t){return maA.apply(this,arguments)}function maA(){return(maA=Yt(function*(t){var{dispatch:e,loadOptions:A,convertStringItemsToObjects:i,filterText:n}=t,o=yield A(n).catch(r=>{console.warn("svelte-select loadOptions error :>> ",r),e("error",{type:"loadOptions",details:r})});if(o&&!o.cancelled)return o?(o&&o.length>0&&typeof o[0]!="object"&&(o=i(o)),e("loaded",{items:o})):o=[],{filteredItems:o,loading:!1,focused:!0,listOpen:!0}})).apply(this,arguments)}Jt(`
  svg.svelte-qbd276 {
      width: var(--chevron-icon-width, 20px);
      height: var(--chevron-icon-width, 20px);
      color: var(--chevron-icon-colour, currentColor);
  }
`);var aKA=tI(`<svg width="100%" height="100%" viewBox="0 0 20 20" focusable="false" aria-hidden="true" class="svelte-qbd276"><path fill="currentColor" d="M4.516 7.548c0.436-0.446 1.043-0.481 1.576 0l3.908 3.747
          3.908-3.747c0.533-0.481 1.141-0.446 1.574 0 0.436 0.445 0.408 1.197 0
          1.615-0.406 0.418-4.695 4.502-4.695 4.502-0.217 0.223-0.502
          0.335-0.787 0.335s-0.57-0.112-0.789-0.335c0
          0-4.287-4.084-4.695-4.502s-0.436-1.17 0-1.615z"></path></svg>`);Jt(`
    svg.svelte-whdbu1 {
        width: var(--clear-icon-width, 20px);
        height: var(--clear-icon-width, 20px);
        color: var(--clear-icon-color, currentColor);
    }
`);var cKA=tI(`<svg width="100%" height="100%" viewBox="-2 -2 50 50" focusable="false" aria-hidden="true" role="presentation" class="svelte-whdbu1"><path fill="currentColor" d="M34.923,37.251L24,26.328L13.077,37.251L9.436,33.61l10.923-10.923L9.436,11.765l3.641-3.641L24,19.047L34.923,8.124
    l3.641,3.641L27.641,22.688L38.564,33.61L34.923,37.251z"></path></svg>`);function e_(t){iA(t,cKA())}Jt(`
    .loading.svelte-1p3nqvd {
        width: var(--spinner-width, 20px);
        height: var(--spinner-height, 20px);
        color: var(--spinner-color, var(--icons-color));
        animation: svelte-1p3nqvd-rotate 0.75s linear infinite;
        transform-origin: center center;
        transform: none;
    }

    .circle_path.svelte-1p3nqvd {
        stroke-dasharray: 90;
        stroke-linecap: round;
    }

    @keyframes svelte-1p3nqvd-rotate {
        100% {
            transform: rotate(360deg);
        }
    }
`);var lKA=tI('<svg class="loading svelte-1p3nqvd" viewBox="25 25 50 50"><circle class="circle_path svelte-1p3nqvd" cx="50" cy="50" r="20" fill="none" stroke="currentColor" stroke-width="5" stroke-miterlimit="10"></circle></svg>');Jt(`
    .svelte-select.svelte-82qwg8 {
        /* deprecating camelCase custom props in favour of kebab-case for v5 */
        --borderRadius: var(--border-radius);
        --clearSelectColor: var(--clear-select-color);
        --clearSelectWidth: var(--clear-select-width);
        --disabledBackground: var(--disabled-background);
        --disabledBorderColor: var(--disabled-border-color);
        --disabledColor: var(--disabled-color);
        --disabledPlaceholderColor: var(--disabled-placeholder-color);
        --disabledPlaceholderOpacity: var(--disabled-placeholder-opacity);
        --errorBackground: var(--error-background);
        --errorBorder: var(--error-border);
        --groupItemPaddingLeft: var(--group-item-padding-left);
        --groupTitleColor: var(--group-title-color);
        --groupTitleFontSize: var(--group-title-font-size);
        --groupTitleFontWeight: var(--group-title-font-weight);
        --groupTitlePadding: var(--group-title-padding);
        --groupTitleTextTransform: var(--group-title-text-transform);
        --groupTitleBorderColor: var(--group-title-border-color);
        --groupTitleBorderWidth: var(--group-title-border-width);
        --groupTitleBorderStyle: var(--group-title-border-style);
        --indicatorColor: var(--chevron-color);
        --indicatorHeight: var(--chevron-height);
        --indicatorWidth: var(--chevron-width);
        --inputColor: var(--input-color);
        --inputLeft: var(--input-left);
        --inputLetterSpacing: var(--input-letter-spacing);
        --inputMargin: var(--input-margin);
        --inputPadding: var(--input-padding);
        --itemActiveBackground: var(--item-active-background);
        --itemColor: var(--item-color);
        --itemFirstBorderRadius: var(--item-first-border-radius);
        --itemHoverBG: var(--item-hover-bg);
        --itemHoverColor: var(--item-hover-color);
        --itemIsActiveBG: var(--item-is-active-bg);
        --itemIsActiveColor: var(--item-is-active-color);
        --itemIsNotSelectableColor: var(--item-is-not-selectable-color);
        --itemPadding: var(--item-padding);
        --listBackground: var(--list-background);
        --listBorder: var(--list-border);
        --listBorderRadius: var(--list-border-radius);
        --listEmptyColor: var(--list-empty-color);
        --listEmptyPadding: var(--list-empty-padding);
        --listEmptyTextAlign: var(--list-empty-text-align);
        --listMaxHeight: var(--list-max-height);
        --listPosition: var(--list-position);
        --listShadow: var(--list-shadow);
        --listZIndex: var(--list-z-index);
        --multiItemBG: var(--multi-item-bg);
        --multiItemBorderRadius: var(--multi-item-border-radius);
        --multiItemDisabledHoverBg: var(--multi-item-disabled-hover-bg);
        --multiItemDisabledHoverColor: var(--multi-item-disabled-hover-color);
        --multiItemHeight: var(--multi-item-height);
        --multiItemMargin: var(--multi-item-margin);
        --multiItemPadding: var(--multi-item-padding);
        --multiSelectInputMargin: var(--multi-select-input-margin);
        --multiSelectInputPadding: var(--multi-select-input-padding);
        --multiSelectPadding: var(--multi-select-padding);
        --placeholderColor: var(--placeholder-color);
        --placeholderOpacity: var(--placeholder-opacity);
        --selectedItemPadding: var(--selected-item-padding);
        --spinnerColor: var(--spinner-color);
        --spinnerHeight: var(--spinner-height);
        --spinnerWidth: var(--spinner-width);

        --internal-padding: 0 0 0 16px;

        border: var(--border, 1px solid #d8dbdf);
        border-radius: var(--border-radius, 6px);
        min-height: var(--height, 42px);
        position: relative;
        display: flex;
        align-items: stretch;
        padding: var(--padding, var(--internal-padding));
        background: var(--background, #fff);
        margin: var(--margin, 0);
        width: var(--width, 100%);
        font-size: var(--font-size, 16px);
        max-height: var(--max-height);
    }

    .svelte-82qwg8 {
        box-sizing: var(--box-sizing, border-box);
    }

    .svelte-select.svelte-82qwg8:hover {
        border: var(--border-hover, 1px solid #b2b8bf);
    }

    .value-container.svelte-82qwg8 {
        display: flex;
        flex: 1 1 0%;
        flex-wrap: wrap;
        align-items: center;
        gap: 5px 10px;
        padding: var(--value-container-padding, 5px 0);
        position: relative;
        overflow: var(--value-container-overflow, hidden);
        align-self: stretch;
    }

    .prepend.svelte-82qwg8,
    .indicators.svelte-82qwg8 {
        display: flex;
        flex-shrink: 0;
        align-items: center;
    }

    .indicators.svelte-82qwg8 {
        position: var(--indicators-position);
        top: var(--indicators-top);
        right: var(--indicators-right);
        bottom: var(--indicators-bottom);
    }

    input.svelte-82qwg8 {
        position: absolute;
        cursor: default;
        border: none;
        color: var(--input-color, var(--item-color));
        padding: var(--input-padding, 0);
        letter-spacing: var(--input-letter-spacing, inherit);
        margin: var(--input-margin, 0);
        min-width: 10px;
        top: 0;
        right: 0;
        bottom: 0;
        left: 0;
        background: transparent;
        font-size: var(--font-size, 16px);
    }

    .svelte-82qwg8:not(.multi) > .value-container:where(.svelte-82qwg8) > input:where(.svelte-82qwg8) {
        width: 100%;
        height: 100%;
    }

    input.svelte-82qwg8::placeholder {
        color: var(--placeholder-color, #78848f);
        opacity: var(--placeholder-opacity, 1);
    }

    input.svelte-82qwg8:focus {
        outline: none;
    }

    .svelte-select.focused.svelte-82qwg8 {
        border: var(--border-focused, 1px solid #006fe8);
        border-radius: var(--border-radius-focused, var(--border-radius, 6px));
    }

    .disabled.svelte-82qwg8 {
        background: var(--disabled-background, #ebedef);
        border-color: var(--disabled-border-color, #ebedef);
        color: var(--disabled-color, #c1c6cc);
    }

    .disabled.svelte-82qwg8 input:where(.svelte-82qwg8)::placeholder {
        color: var(--disabled-placeholder-color, #c1c6cc);
        opacity: var(--disabled-placeholder-opacity, 1);
    }

    .selected-item.svelte-82qwg8 {
        position: relative;
        overflow: var(--selected-item-overflow, hidden);
        padding: var(--selected-item-padding, 0 20px 0 0);
        text-overflow: ellipsis;
        white-space: nowrap;
        color: var(--selected-item-color, inherit);
        font-size: var(--font-size, 16px);
    }

    .multi.svelte-82qwg8 .selected-item:where(.svelte-82qwg8) {
        position: absolute;
        line-height: var(--height, 42px);
        height: var(--height, 42px);
    }

    .selected-item.svelte-82qwg8:focus {
        outline: none;
    }

    .hide-selected-item.svelte-82qwg8 {
        opacity: 0;
    }

    .icon.svelte-82qwg8 {
        display: flex;
        align-items: center;
        justify-content: center;
    }

    .clear-select.svelte-82qwg8 {
        all: unset;
        display: flex;
        align-items: center;
        justify-content: center;
        width: var(--clear-select-width, 40px);
        height: var(--clear-select-height, 100%);
        color: var(--clear-select-color, var(--icons-color));
        margin: var(--clear-select-margin, 0);
        pointer-events: all;
        flex-shrink: 0;
    }

    .clear-select.svelte-82qwg8:focus {
        outline: var(--clear-select-focus-outline, 1px solid #006fe8);
    }

    .loading.svelte-82qwg8 {
        width: var(--loading-width, 40px);
        height: var(--loading-height);
        color: var(--loading-color, var(--icons-color));
        margin: var(--loading--margin, 0);
        flex-shrink: 0;
    }

    .chevron.svelte-82qwg8 {
        width: var(--chevron-width, 40px);
        height: var(--chevron-height, 40px);
        background: var(--chevron-background, transparent);
        pointer-events: var(--chevron-pointer-events, none);
        color: var(--chevron-color, var(--icons-color));
        border: var(--chevron-border, 0 0 0 1px solid #d8dbdf);
        flex-shrink: 0;
    }

    .multi.svelte-82qwg8 {
        padding: var(--multi-select-padding, var(--internal-padding));
    }

    .multi.svelte-82qwg8 input:where(.svelte-82qwg8) {
        padding: var(--multi-select-input-padding, 0);
        position: relative;
        margin: var(--multi-select-input-margin, 5px 0);
        flex: 1 1 40px;
    }

    .svelte-select.error.svelte-82qwg8 {
        border: var(--error-border, 1px solid #ff2d55);
        background: var(--error-background, #fff);
    }

    .a11y-text.svelte-82qwg8 {
        z-index: 9999;
        border: 0px;
        clip: rect(1px, 1px, 1px, 1px);
        height: 1px;
        width: 1px;
        position: absolute;
        overflow: hidden;
        padding: 0px;
        white-space: nowrap;
    }

    .multi-item.svelte-82qwg8 {
        background: var(--multi-item-bg, #ebedef);
        margin: var(--multi-item-margin, 0);
        outline: var(--multi-item-outline, 1px solid #ddd);
        border-radius: var(--multi-item-border-radius, 4px);
        height: var(--multi-item-height, 25px);
        line-height: var(--multi-item-height, 25px);
        display: flex;
        cursor: default;
        padding: var(--multi-item-padding, 0 5px);
        overflow: hidden;
        gap: var(--multi-item-gap, 4px);
        outline-offset: -1px;
        max-width: var(--multi-max-width, none);
        color: var(--multi-item-color, var(--item-color));
    }

    .multi-item.disabled.svelte-82qwg8:hover {
        background: var(--multi-item-disabled-hover-bg, #ebedef);
        color: var(--multi-item-disabled-hover-color, #c1c6cc);
    }

    .multi-item-text.svelte-82qwg8 {
        overflow: hidden;
        text-overflow: ellipsis;
        white-space: nowrap;
    }

    .multi-item-clear.svelte-82qwg8 {
        display: flex;
        align-items: center;
        justify-content: center;
        --clear-icon-color: var(--multi-item-clear-icon-color, #000);
    }

    .multi-item.active.svelte-82qwg8 {
        outline: var(--multi-item-active-outline, 1px solid #006fe8);
    }

    .svelte-select-list.svelte-82qwg8 {
        box-shadow: var(--list-shadow, 0 2px 3px 0 rgba(44, 62, 80, 0.24));
        border-radius: var(--list-border-radius, 4px);
        max-height: var(--list-max-height, 252px);
        overflow-y: auto;
        background: var(--list-background, #fff);
        position: var(--list-position, absolute);
        z-index: var(--list-z-index, 2);
        border: var(--list-border);
    }

    .prefloat.svelte-82qwg8 {
        opacity: 0;
        pointer-events: none;
    }

    .list-group-title.svelte-82qwg8 {
        color: var(--group-title-color, #8f8f8f);
        cursor: default;
        font-size: var(--group-title-font-size, 16px);
        font-weight: var(--group-title-font-weight, 600);
        height: var(--height, 42px);
        line-height: var(--height, 42px);
        padding: var(--group-title-padding, 0 20px);
        text-overflow: ellipsis;
        overflow-x: hidden;
        white-space: nowrap;
        text-transform: var(--group-title-text-transform, uppercase);
        border-width: var(--group-title-border-width, medium);
        border-style: var(--group-title-border-style, none);
        border-color: var(--group-title-border-color, color);
    }

    .empty.svelte-82qwg8 {
        text-align: var(--list-empty-text-align, center);
        padding: var(--list-empty-padding, 20px 0);
        color: var(--list-empty-color, #78848f);
    }

    .item.svelte-82qwg8 {
        cursor: default;
        height: var(--item-height, var(--height, 42px));
        line-height: var(--item-line-height, var(--height, 42px));
        padding: var(--item-padding, 0 20px);
        color: var(--item-color, inherit);
        text-overflow: ellipsis;
        overflow: hidden;
        white-space: nowrap;
        transition: var(--item-transition, all 0.2s);
        align-items: center;
        width: 100%;
    }

    .item.group-item.svelte-82qwg8 {
        padding-left: var(--group-item-padding-left, 40px);
    }

    .item.svelte-82qwg8:active {
        background: var(--item-active-background, #b9daff);
    }

    .item.active.svelte-82qwg8 {
        background: var(--item-is-active-bg, #007aff);
        color: var(--item-is-active-color, #fff);
    }

    .item.first.svelte-82qwg8 {
        border-radius: var(--item-first-border-radius, 4px 4px 0 0);
    }

    .item.hover.svelte-82qwg8:not(.active) {
        background: var(--item-hover-bg, #e7f2ff);
        color: var(--item-hover-color, inherit);
    }

    .item.not-selectable.svelte-82qwg8,
    .item.hover.item.not-selectable.svelte-82qwg8,
    .item.active.item.not-selectable.svelte-82qwg8,
    .item.not-selectable.svelte-82qwg8:active {
        color: var(--item-is-not-selectable-color, #999);
        background: transparent;
    }

    .required.svelte-82qwg8 {
        opacity: 0;
        z-index: -1;
        position: absolute;
        top: 0;
        left: 0;
        bottom: 0;
        right: 0;
    }
`);var gKA=wA('<div class="list-item svelte-82qwg8" tabindex="-1" role="none"><div><!></div></div>'),IKA=wA('<div class="empty svelte-82qwg8">No options</div>'),CKA=wA('<div role="none"><!> <!> <!></div>'),dKA=wA('<span id="aria-selection" class="svelte-82qwg8"> </span> <span id="aria-context" class="svelte-82qwg8"> </span>',1),BKA=wA('<div class="multi-item-clear svelte-82qwg8"><!></div>'),EKA=wA('<div role="none"><span class="multi-item-text svelte-82qwg8"><!></span> <!></div>'),QKA=wA("<div><!></div>"),hKA=wA('<div class="icon loading svelte-82qwg8" aria-hidden="true"><!></div>'),uKA=wA('<button type="button" class="icon clear-select svelte-82qwg8"><!></button>'),fKA=wA('<div class="icon chevron svelte-82qwg8" aria-hidden="true"><!></div>'),mKA=wA('<input type="hidden" class="svelte-82qwg8"/>'),pKA=wA('<select class="required svelte-82qwg8" required tabindex="-1" aria-hidden="true"></select>'),wKA=wA('<div role="none"><!> <span aria-live="polite" aria-atomic="false" aria-relevant="additions text" class="a11y-text svelte-82qwg8"><!></span> <div class="prepend svelte-82qwg8"><!></div> <div class="value-container svelte-82qwg8"><!> <input/></div> <div class="indicators svelte-82qwg8"><!> <!> <!></div> <!> <!></div>');function JC(t,e){var A=function(EA){var HA={};for(var ve in EA.children&&(HA.default=!0),EA.$$slots)HA[ve]=!0;return HA}(e);mt(e,!1);var i,n=$(),o=$(),r=$(),s=$(),a=$(),c=$(),l=$(),I=$(),C=$(),d=dGA(),B=b(e,"justValue",12,null),E=b(e,"filter",8,rKA),h=b(e,"getItems",8,sKA),u=b(e,"id",8,null),D=b(e,"name",8,null),L=b(e,"container",12,void 0),R=b(e,"input",12,void 0),w=b(e,"multiple",8,!1),_=b(e,"multiFullItemClearable",8,!1),K=b(e,"disabled",8,!1),z=b(e,"focused",12,!1),U=b(e,"value",12,null),H=b(e,"filterText",12,""),q=b(e,"placeholder",8,"Please select"),j=b(e,"placeholderAlwaysShow",8,!1),gA=b(e,"items",12,null),QA=b(e,"label",8,"label"),BA=b(e,"itemFilter",8,(EA,HA,ve)=>"".concat(EA).toLowerCase().includes(HA.toLowerCase())),lA=b(e,"groupBy",8,void 0),vA=b(e,"groupFilter",8,EA=>EA),tA=b(e,"groupHeaderSelectable",8,!1),cA=b(e,"itemId",8,"value"),pA=b(e,"loadOptions",8,void 0),VA=b(e,"containerStyles",8,""),oe=b(e,"hasError",8,!1),KA=b(e,"filterSelectedItems",8,!0),CA=b(e,"required",8,!1),TA=b(e,"closeListOnChange",8,!0),Ze=b(e,"clearFilterTextOnBlur",8,!0),He=b(e,"createGroupHeaderItem",8,(EA,HA)=>({value:EA,[QA()]:EA})),uA=()=>g(l),eA=b(e,"searchable",8,!0),UA=b(e,"inputStyles",8,""),aA=b(e,"clearable",8,!0),le=b(e,"loading",12,!1),SA=b(e,"listOpen",12,!1),Ue=b(e,"debounce",8,function(EA){var HA=arguments.length>1&&arguments[1]!==void 0?arguments[1]:1;clearTimeout(i),i=setTimeout(EA,HA)}),mA=b(e,"debounceWait",8,300),sA=b(e,"hideEmptyState",8,!1),xt=b(e,"inputAttributes",24,()=>({})),tt=b(e,"listAutoWidth",8,!0),de=b(e,"showChevron",8,!1),Dt=b(e,"listOffset",8,5),_e=b(e,"hoverItemIndex",12,0),Le=b(e,"floatingConfig",24,()=>({})),bt=b(e,"class",8,""),Re=$(),$t=$(),x=$(),Y=$(),P=$();function X(EA){return EA.map((HA,ve)=>({index:ve,value:HA,label:"".concat(HA)}))}function bA(EA){var HA=[],ve={};EA.forEach(yi=>{var ri=lA()(yi);HA.includes(ri)||(HA.push(ri),ve[ri]=[],ri&&ve[ri].push(Object.assign(He()(ri,yi),{id:ri,groupHeader:!0,selectable:tA()}))),ve[ri].push(Object.assign({groupItem:!!ri},yi))});var Qt=[];return vA()(HA).forEach(yi=>{ve[yi]&&Qt.push(...ve[yi])}),Qt}function Be(){var EA=arguments.length>0&&arguments[0]!==void 0?arguments[0]:0,HA=arguments.length>1?arguments[1]:void 0;_e(EA<0?0:EA),!HA&&lA()&&g(l)[_e()]&&!g(l)[_e()].selectable&&ni(1)}function Ee(){var EA=!0;if(U()){var HA=[],ve=[];U().forEach(Qt=>{HA.includes(Qt[cA()])?EA=!1:(HA.push(Qt[cA()]),ve.push(Qt))}),EA||U(ve)}return EA}function kA(EA){var HA=EA?EA[cA()]:U()[cA()];return gA().find(ve=>ve[cA()]===HA)}function DA(EA){return gt.apply(this,arguments)}function gt(){return(gt=Yt(function*(EA){var HA=U()[EA];U().length===1?U(void 0):U(U().filter(ve=>ve!==HA)),d("clear",HA)})).apply(this,arguments)}function Ve(EA){if(z())switch(EA.stopPropagation(),EA.key){case"Escape":EA.preventDefault(),xe();break;case"Enter":if(EA.preventDefault(),SA()){if(g(l).length===0)break;var HA=g(l)[_e()];if(U()&&!w()&&U()[cA()]===HA[cA()]){xe();break}M(g(l)[_e()])}break;case"ArrowDown":EA.preventDefault(),SA()?ni(1):(SA(!0),y(Re,void 0));break;case"ArrowUp":EA.preventDefault(),SA()?ni(-1):(SA(!0),y(Re,void 0));break;case"Tab":if(SA()&&z()){if(g(l).length===0||U()&&U()[cA()]===g(l)[_e()][cA()])return xe();EA.preventDefault(),M(g(l)[_e()]),xe()}break;case"Backspace":if(!w()||H().length>0)return;if(w()&&U()&&U().length>0){if(DA(g(Re)!==void 0?g(Re):U().length-1),g(Re)===0||g(Re)===void 0)break;y(Re,U().length>g(Re)?g(Re)-1:void 0)}break;case"ArrowLeft":if(!U()||!w()||H().length>0)return;g(Re)===void 0?y(Re,U().length-1):U().length>g(Re)&&g(Re)!==0&&y(Re,g(Re)-1);break;case"ArrowRight":if(!U()||!w()||H().length>0||g(Re)===void 0)return;g(Re)===U().length-1?y(Re,void 0):g(Re)<U().length-1&&y(Re,g(Re)+1)}}function ZA(EA){var HA,ve;z()&&R()===((HA=document)===null||HA===void 0?void 0:HA.activeElement)||(EA&&d("focus",EA),(ve=R())===null||ve===void 0||ve.focus(),z(!0))}function rt(EA){return Ei.apply(this,arguments)}function Ei(){return(Ei=Yt(function*(EA){var HA;Tt||(SA()||z())&&(d("blur",EA),xe(),z(!1),y(Re,void 0),(HA=R())===null||HA===void 0||HA.blur())})).apply(this,arguments)}function tn(){if(!K())return H().length>0?SA(!0):void SA(!SA())}function qi(){d("clear",U()),U(void 0),xe(),ZA()}function xe(){Ze()&&H(""),SA(!1)}BGA(Yt(function*(){y($t,U()),y(x,H()),y(Y,w())})),hs(()=>{SA()&&z(!0),z()&&R()&&R().focus()});var nn=b(e,"ariaValues",8,EA=>"Option ".concat(EA,", selected.")),mi=b(e,"ariaListOpen",8,(EA,HA)=>"You are currently focused on option ".concat(EA,". There are ").concat(HA," results available.")),Ot=b(e,"ariaFocused",8,()=>"Select is focused, type to refine list, press down to open the menu."),Lt,ii=$(null);function _i(){clearTimeout(Lt),Lt=setTimeout(()=>{Tt=!1},100)}qc(()=>{var EA;(EA=g(ii))===null||EA===void 0||EA.remove()});var Tt=!1;function M(EA){EA&&EA.selectable!==!1&&function(HA){if(HA){H("");var ve=Object.assign({},HA);if(ve.groupHeader&&!ve.selectable)return;U(w()?U()?U().concat([ve]):[ve]:U(ve)),setTimeout(()=>{TA()&&xe(),y(Re,void 0),d("change",U()),d("select",HA)})}}(EA)}function We(EA){Tt||_e(EA)}function ni(EA){if(g(l).filter(ve=>!Object.hasOwn(ve,"selectable")||ve.selectable===!0).length===0)return _e(0);EA>0&&_e()===g(l).length-1?_e(0):EA<0&&_e()===0?_e(g(l).length-1):_e(_e()+EA);var HA=g(l)[_e()];HA&&HA.selectable===!1&&(EA!==1&&EA!==-1||ni(EA))}function pi(EA,HA,ve){if(!w())return HA&&HA[ve]===EA[ve]}var dn=Uo,mn=Uo;function Uo(EA){return{update(HA){HA.scroll&&(_i(),EA.scrollIntoView({behavior:"auto",block:"nearest"}))}}}var kn=$({strategy:"absolute",placement:"bottom-start",middleware:[tKA(Dt()),nKA(),iKA()],autoUpdate:!1}),[Wn,Vo,vo]=oKA(g(kn)),bo=$(!0);fA(()=>(k(gA()),k(U())),()=>{gA(),U()&&function(){if(typeof U()=="string"){var EA=(gA()||[]).find(HA=>HA[cA()]===U());U(EA||{[cA()]:U(),label:U()})}else w()&&Array.isArray(U())&&U().length>0&&U(U().map(HA=>typeof HA=="string"?{value:HA,label:HA}:HA))}()}),fA(()=>(k(xt()),k(eA())),()=>{!xt()&&eA()||(y(P,Object.assign({autocapitalize:"none",autocomplete:"off",autocorrect:"off",spellcheck:!1,tabindex:0,type:"text","aria-autocomplete":"list"},xt())),u()&&hc(P,g(P).id=u()),eA()||hc(P,g(P).readonly=!0))}),fA(()=>k(w()),()=>{w()&&U()&&(Array.isArray(U())?U([...U()]):U([U()]))}),fA(()=>(g(Y),k(w())),()=>{g(Y)&&!w()&&U()&&U(null)}),fA(()=>(k(w()),k(U())),()=>{w()&&U()&&U().length>1&&Ee()}),fA(()=>k(U()),()=>{U()&&(w()?JSON.stringify(U())!==JSON.stringify(g($t))&&Ee()&&d("input",U()):g($t)&&JSON.stringify(U()[cA()])===JSON.stringify(g($t)[cA()])||d("input",U()))}),fA(()=>(k(U()),k(w()),g($t)),()=>{!U()&&w()&&g($t)&&d("input",U())}),fA(()=>(k(z()),k(R())),()=>{!z()&&R()&&xe()}),fA(()=>(k(H()),g(x)),()=>{H()!==g(x)&&(pA()||H().length!==0)&&(pA()?Ue()(Yt(function*(){le(!0);var EA=yield h()({dispatch:d,loadOptions:pA(),convertStringItemsToObjects:X,filterText:H()});EA?(le(EA.loading),SA(SA()?EA.listOpen:H().length>0),z(SA()&&EA.focused),gA(lA()?bA(EA.filteredItems):EA.filteredItems)):(le(!1),z(!0),SA(!0))}),mA()):(SA(!0),w()&&y(Re,void 0)))}),fA(()=>(k(E()),k(pA()),k(H()),k(gA()),k(w()),k(U()),k(cA()),k(lA()),k(QA()),k(KA()),k(BA())),()=>{y(l,E()({loadOptions:pA(),filterText:H(),items:gA(),multiple:w(),value:U(),itemId:cA(),groupBy:lA(),label:QA(),filterSelectedItems:KA(),itemFilter:BA(),convertStringItemsToObjects:X,filterGroupedItems:bA}))}),fA(()=>(k(w()),k(SA()),k(U()),g(l)),()=>{!w()&&SA()&&U()&&g(l)&&Be(g(l).findIndex(EA=>EA[cA()]===U()[cA()]),!0)}),fA(()=>(k(SA()),k(w())),()=>{SA()&&w()&&_e(0)}),fA(()=>k(H()),()=>{H()&&_e(0)}),fA(()=>k(_e()),()=>{var EA;EA=_e(),d("hoverItem",EA)}),fA(()=>(k(w()),k(U())),()=>{y(n,w()?U()&&U().length>0:U())}),fA(()=>(g(n),k(H())),()=>{y(o,g(n)&&H().length>0)}),fA(()=>(g(n),k(aA()),k(K()),k(le())),()=>{y(r,g(n)&&aA()&&!K()&&!le())}),fA(()=>(k(j()),k(w()),k(q()),k(U())),()=>{var EA;y(s,j()&&w()||w()&&((EA=U())===null||EA===void 0?void 0:EA.length)===0?q():U()?"":q())}),fA(()=>(k(U()),k(w())),()=>{var EA,HA;y(a,U()?(EA=w(),HA=void 0,HA=EA&&U().length>0?U().map(ve=>ve[QA()]).join(", "):U()[QA()],nn()(HA)):"")}),fA(()=>(g(l),k(_e()),k(z()),k(SA())),()=>{y(c,function(){if(!g(l)||g(l).length===0)return"";var EA=g(l)[_e()];if(SA()&&EA){var HA=g(l)?g(l).length:0;return mi()(EA[QA()],HA)}return Ot()()}((g(l),_e(),z(),SA())))}),fA(()=>k(gA()),()=>{(function(EA){EA&&EA.length!==0&&!EA.some(HA=>typeof HA!="object")&&U()&&(w()?!U().some(HA=>!HA||!HA[cA()]):U()[cA()])&&(Array.isArray(U())?U(U().map(HA=>kA(HA)||HA)):U(kA()||U()))})(gA())}),fA(()=>(k(w()),k(U()),k(cA())),()=>{B((w(),U(),cA(),w()?U()?U().map(EA=>EA[cA()]):null:U()?U()[cA()]:U()))}),fA(()=>(k(w()),g($t),k(U())),()=>{w()||!g($t)||U()||d("input",U())}),fA(()=>(k(SA()),g(l),k(w()),k(U())),()=>{SA()&&g(l)&&!w()&&!U()&&Be()}),fA(()=>g(l),()=>{(function(EA){SA()&&d("filter",EA)})(g(l))}),fA(()=>(k(L()),k(Le()),g(kn)),()=>{L()&&Le()&&vo(Object.assign(g(kn),Le()))}),fA(()=>g(ii),()=>{y(I,!!g(ii))}),fA(()=>(g(ii),k(SA())),()=>{(function(EA,HA){if(!EA||!HA)return y(bo,!0);setTimeout(()=>{y(bo,!1)},0)})(g(ii),SA())}),fA(()=>(k(SA()),k(L()),g(ii)),()=>{SA()&&L()&&g(ii)&&function(){var{width:EA}=L().getBoundingClientRect();hc(ii,g(ii).style.width=tt()?EA+"px":"auto")}()}),fA(()=>k(_e()),()=>{y(C,_e())}),fA(()=>(k(R()),k(SA()),k(z())),()=>{R()&&SA()&&!z()&&ZA()}),fA(()=>(k(L()),k(Le())),()=>{var EA;L()&&((EA=Le())===null||EA===void 0?void 0:EA.autoUpdate)===void 0&&hc(kn,g(kn).autoUpdate=!0)}),Qn(),Zt();var Yn,Mo=wKA();ce("click",A2,function(EA){var HA;SA()||z()||!L()||L().contains(EA.target)||(HA=g(ii))!==null&&HA!==void 0&&HA.contains(EA.target)||rt()}),ce("keydown",A2,Ve);var ne=W(Mo),wi=EA=>{var HA,ve=CKA(),Qt=W(ve),yi=Pt=>{var $i=_o();jo(vt($i),e,"list-prepend",{},null),iA(Pt,$i)};LA(Qt,Pt=>{nA(()=>A["list-prepend"])&&Pt(yi)});var ri=IA(Qt,2),pn=Pt=>{var $i=_o();jo(vt($i),e,"list",{get filteredItems(){return g(l)}},null),iA(Pt,$i)},Fn=(Pt,$i)=>{var Rr=Xn=>{var se=_o();qo(vt(se),1,()=>g(l),or,(vi,Yi,rn)=>{var Hr,Ri=gKA(),fs=W(Ri);jo(W(fs),e,"item",{get item(){return g(Yi)},index:rn},Bo=>{var Q=Tr();De(()=>wt(Q,(g(Yi),k(QA()),nA(()=>{var m;return(m=g(Yi))===null||m===void 0?void 0:m[QA()]})))),iA(Bo,Q)}),Us(fs,(Bo,Q)=>dn?.(Bo),()=>({scroll:pi(g(Yi),U(),cA()),listDom:g(I)})),Us(fs,(Bo,Q)=>mn?.(Bo),()=>({scroll:g(C)===rn,listDom:g(I)})),De(Bo=>Hr=Vt(fs,1,"item svelte-82qwg8",null,Hr,Bo),[()=>{var Bo,Q;return{"list-group-title":g(Yi).groupHeader,active:pi(g(Yi),U(),cA()),first:(Q=rn,Q===0),hover:_e()===rn,"group-item":g(Yi).groupItem,"not-selectable":((Bo=g(Yi))===null||Bo===void 0?void 0:Bo.selectable)===!1}}],qA),ce("mouseover",Ri,()=>We(rn)),ce("focus",Ri,()=>We(rn)),ce("click",Ri,j0(()=>function(Bo){var{item:Q,i:m}=Bo;if(Q?.selectable!==!1)return U()&&!w()&&U()[cA()]===Q[cA()]?xe():void(function(v){return v.groupHeader&&v.selectable||v.selectable||!v.hasOwnProperty("selectable")}(Q)&&(_e(m),M(Q)))}({item:g(Yi),i:rn}))),ce("keydown",Ri,N1(j0(function(Bo){N3.call(this,e,Bo)}))),iA(vi,Ri)}),iA(Xn,se)},Ft=(Xn,se)=>{var vi=Yi=>{var rn=_o();jo(vt(rn),e,"empty",{},Hr=>{iA(Hr,IKA())}),iA(Yi,rn)};LA(Xn,Yi=>{sA()||Yi(vi)},se)};LA(Pt,Xn=>{g(l),nA(()=>g(l).length>0)?Xn(Rr):Xn(Ft,!1)},$i)};LA(ri,Pt=>{nA(()=>A.list)?Pt(pn):Pt(Fn,!1)});var Jn=IA(ri,2),ln=Pt=>{var $i=_o();jo(vt($i),e,"list-append",{},null),iA(Pt,$i)};LA(Jn,Pt=>{nA(()=>A["list-append"])&&Pt(ln)}),Us(ve,Pt=>Vo?.(Pt)),Co(ve,Pt=>y(ii,Pt),()=>g(ii)),es(()=>ce("scroll",ve,_i)),es(()=>ce("pointerup",ve,N1(j0(function(Pt){N3.call(this,e,Pt)})))),es(()=>ce("mousedown",ve,N1(j0(function(Pt){N3.call(this,e,Pt)})))),De(Pt=>HA=Vt(ve,1,"svelte-select-list svelte-82qwg8",null,HA,Pt),[()=>({prefloat:g(bo)})],qA),iA(EA,ve)};LA(ne,EA=>{SA()&&EA(wi)});var MA=IA(ne,2),me=W(MA),nt=EA=>{var HA=dKA(),ve=vt(HA),Qt=W(ve),yi=W(IA(ve,2));De(()=>{wt(Qt,g(a)),wt(yi,g(c))}),iA(EA,HA)};LA(me,EA=>{z()&&EA(nt)});var Wt=IA(MA,2);jo(W(Wt),e,"prepend",{},null);var Xe=IA(Wt,2),oi=W(Xe),Di=EA=>{var HA=_o(),ve=vt(HA),Qt=ri=>{var pn=_o();qo(vt(pn),1,U,or,(Fn,Jn,ln)=>{var Pt,$i=EKA(),Rr=W($i);jo(W(Rr),e,"selection",{get selection(){return g(Jn)},index:ln},se=>{var vi=Tr();De(()=>wt(vi,(g(Jn),k(QA()),nA(()=>g(Jn)[QA()])))),iA(se,vi)});var Ft=IA(Rr,2),Xn=se=>{var vi=BKA();jo(W(vi),e,"multi-clear-icon",{},Yi=>{e_(Yi)}),ce("pointerup",vi,N1(j0(()=>DA(ln)))),iA(se,vi)};LA(Ft,se=>{K()||_()||!e_||se(Xn)}),De(se=>Pt=Vt($i,1,"multi-item svelte-82qwg8",null,Pt,se),[()=>({active:g(Re)===ln,disabled:K()})],qA),ce("click",$i,N1(()=>_()?DA(ln):{})),ce("keydown",$i,N1(j0(function(se){N3.call(this,e,se)}))),iA(Fn,$i)}),iA(ri,pn)},yi=ri=>{var pn,Fn=QKA();jo(W(Fn),e,"selection",{get selection(){return U()}},Jn=>{var ln=Tr();De(()=>wt(ln,(k(U()),k(QA()),nA(()=>U()[QA()])))),iA(Jn,ln)}),De(Jn=>pn=Vt(Fn,1,"selected-item svelte-82qwg8",null,pn,Jn),[()=>({"hide-selected-item":g(o)})],qA),iA(ri,Fn)};LA(ve,ri=>{w()?ri(Qt):ri(yi,!1)}),iA(EA,HA)};LA(oi,EA=>{g(n)&&EA(Di)});var Ut=IA(oi,2);ry(Ut,()=>fe(fe({readOnly:!eA()},g(P)),{},{placeholder:g(s),style:UA(),disabled:K()}),void 0,"svelte-82qwg8"),Co(Ut,EA=>R(EA),()=>R());var cn=IA(Xe,2),ft=W(cn),Qi=EA=>{var HA=hKA();jo(W(HA),e,"loading-icon",{},ve=>{(function(Qt){iA(Qt,lKA())})(ve)}),iA(EA,HA)};LA(ft,EA=>{le()&&EA(Qi)});var ot=IA(ft,2),Mt=EA=>{var HA=uKA();jo(W(HA),e,"clear-icon",{},ve=>{e_(ve)}),ce("click",HA,qi),iA(EA,HA)};LA(ot,EA=>{g(r)&&EA(Mt)});var on=IA(ot,2),hn=EA=>{var HA=fKA();jo(W(HA),e,"chevron-icon",{get listOpen(){return SA()}},ve=>{(function(Qt){iA(Qt,aKA())})(ve)}),iA(EA,HA)};LA(on,EA=>{de()&&EA(hn)});var Ai=IA(cn,2);jo(Ai,e,"input-hidden",{get value(){return U()}},EA=>{var HA=mKA();De(ve=>{En(HA,"name",D()),VC(HA,ve)},[()=>(k(U()),nA(()=>U()?JSON.stringify(U()):null))],qA),iA(EA,HA)});var Ki=IA(Ai,2),dt=EA=>{var HA=_o();jo(vt(HA),e,"required",{get value(){return U()}},ve=>{iA(ve,pKA())}),iA(EA,HA)};return LA(Ki,EA=>{k(CA()),k(U()),nA(()=>CA()&&(!U()||U().length===0))&&EA(dt)}),es(()=>ce("pointerup",Mo,N1(tn))),Co(Mo,EA=>L(EA),()=>L()),Us(Mo,EA=>Wn?.(EA)),De(EA=>{var HA;Yn=Vt(Mo,1,"svelte-select ".concat((HA=bt())!==null&&HA!==void 0?HA:""),"svelte-82qwg8",Yn,EA),Ll(Mo,VA())},[()=>({multi:w(),disabled:K(),focused:z(),"list-open":SA(),"show-chevron":de(),error:oe()})],qA),ce("keydown",Ut,Ve),ce("blur",Ut,rt),ce("focus",Ut,ZA),By(Ut,H),iA(t,Mo),zt(e,"getFilteredItems",uA),zt(e,"handleClear",qi),pt({getFilteredItems:uA,handleClear:qi})}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
table.jse-transform-wizard.svelte-qbze6z {
  border-collapse: collapse;
  border-spacing: 0;
  width: 100%;
}
table.jse-transform-wizard.svelte-qbze6z input:where(.svelte-qbze6z) {
  font-family: inherit;
  font-size: inherit;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) th:where(.svelte-qbze6z) {
  font-weight: normal;
  text-align: left;
  width: 60px;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) {
  width: 100%;
  display: flex;
  flex-direction: row;
  margin-bottom: calc(0.5 * var(--jse-padding, 10px));
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select .multi-item {
  align-items: center;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select .value-container {
  gap: 0 !important;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select.jse-filter-path {
  flex: 4;
  margin-right: calc(0.5 * var(--jse-padding, 10px));
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select.jse-filter-relation {
  flex: 1.5;
  margin-right: calc(0.5 * var(--jse-padding, 10px));
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select.jse-sort-path {
  flex: 3;
  margin-right: calc(0.5 * var(--jse-padding, 10px));
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select.jse-sort-direction {
  flex: 1;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select.jse-projection-paths {
  flex: 1;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .svelte-select input {
  box-sizing: border-box;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .jse-filter-value:where(.svelte-qbze6z) {
  flex: 4;
  padding: 4px 8px;
  border: var(--jse-input-border, 1px solid #d8dbdf);
  border-radius: var(--jse-input-radius, 3px);
  outline: none;
  background: var(--jse-input-background, var(--jse-background-color, #fff));
  color: inherit;
}
table.jse-transform-wizard.svelte-qbze6z tr:where(.svelte-qbze6z) td:where(.svelte-qbze6z) .jse-horizontal:where(.svelte-qbze6z) .jse-filter-value:where(.svelte-qbze6z):focus {
  border: var(--jse-input-border-focus, 1px solid var(--jse-input-border-focus, var(--jse-theme-color, #3883fa)));
}`);var DKA=wA('<table class="jse-transform-wizard svelte-qbze6z"><tbody><tr class="svelte-qbze6z"><th class="svelte-qbze6z">Filter</th><td class="svelte-qbze6z"><div class="jse-horizontal svelte-qbze6z"><!> <!> <input class="jse-filter-value svelte-qbze6z"/></div></td></tr><tr class="svelte-qbze6z"><th class="svelte-qbze6z">Sort</th><td class="svelte-qbze6z"><div class="jse-horizontal svelte-qbze6z"><!> <!></div></td></tr><tr class="svelte-qbze6z"><th class="svelte-qbze6z">Pick</th><td class="svelte-qbze6z"><div class="jse-horizontal svelte-qbze6z"><!></div></td></tr></tbody></table>');function yKA(t,e){var A,i,n,o,r;mt(e,!1);var s=$(void 0,!0),a=$(void 0,!0),c=$(void 0,!0),l=$(void 0,!0),I=$(void 0,!0),C=$(void 0,!0),d=Sr("jsoneditor:TransformWizard"),B=b(e,"json",9),E=b(e,"queryOptions",29,()=>({})),h=b(e,"onChange",9),u=["==","!=","<","<=",">",">="].map(KA=>({value:KA,label:KA})),D=[{value:"asc",label:"ascending"},{value:"desc",label:"descending"}],L=$((A=E())!==null&&A!==void 0&&(A=A.filter)!==null&&A!==void 0&&A.path?U1(E().filter.path):void 0,!0),R=$((i=u.find(KA=>{var CA;return KA.value===((CA=E().filter)===null||CA===void 0?void 0:CA.relation)}))!==null&&i!==void 0?i:u[0],!0),w=$(((n=E())===null||n===void 0||(n=n.filter)===null||n===void 0?void 0:n.value)||"",!0),_=$((o=E())!==null&&o!==void 0&&(o=o.sort)!==null&&o!==void 0&&o.path?U1(E().sort.path):void 0,!0),K=$((r=D.find(KA=>{var CA;return KA.value===((CA=E().sort)===null||CA===void 0?void 0:CA.direction)}))!==null&&r!==void 0?r:D[0],!0);fA(()=>k(B()),()=>{y(s,Array.isArray(B()))}),fA(()=>(g(s),k(B())),()=>{y(a,g(s)?h_(B()):[])}),fA(()=>(g(s),k(B())),()=>{y(c,g(s)?h_(B(),!0):[])}),fA(()=>(g(a),U1),()=>{y(l,g(a).map(U1))}),fA(()=>(g(c),U1),()=>{y(I,g(c)?g(c).map(U1):[])}),fA(()=>(k(E()),g(I),di),()=>{var KA;y(C,(KA=E())!==null&&KA!==void 0&&(KA=KA.projection)!==null&&KA!==void 0&&KA.paths&&g(I)?E().projection.paths.map(CA=>g(I).find(TA=>di(TA.value,CA))).filter(CA=>!!CA):void 0)}),fA(()=>g(L),()=>{var KA,CA,TA;CA=(KA=g(L))===null||KA===void 0?void 0:KA.value,di((TA=E())===null||TA===void 0||(TA=TA.filter)===null||TA===void 0?void 0:TA.path,CA)||(d("changeFilterPath",CA),E(cs(E(),["filter","path"],CA,!0)),h()(E()))}),fA(()=>g(R),()=>{var KA,CA,TA;CA=(KA=g(R))===null||KA===void 0?void 0:KA.value,di((TA=E())===null||TA===void 0||(TA=TA.filter)===null||TA===void 0?void 0:TA.relation,CA)||(d("changeFilterRelation",CA),E(cs(E(),["filter","relation"],CA,!0)),h()(E()))}),fA(()=>g(w),()=>{var KA,CA;KA=g(w),di((CA=E())===null||CA===void 0||(CA=CA.filter)===null||CA===void 0?void 0:CA.value,KA)||(d("changeFilterValue",KA),E(cs(E(),["filter","value"],KA,!0)),h()(E()))}),fA(()=>g(_),()=>{var KA,CA,TA;CA=(KA=g(_))===null||KA===void 0?void 0:KA.value,di((TA=E())===null||TA===void 0||(TA=TA.sort)===null||TA===void 0?void 0:TA.path,CA)||(d("changeSortPath",CA),E(cs(E(),["sort","path"],CA,!0)),h()(E()))}),fA(()=>g(K),()=>{var KA,CA,TA;CA=(KA=g(K))===null||KA===void 0?void 0:KA.value,di((TA=E())===null||TA===void 0||(TA=TA.sort)===null||TA===void 0?void 0:TA.direction,CA)||(d("changeSortDirection",CA),E(cs(E(),["sort","direction"],CA,!0)),h()(E()))}),fA(()=>g(C),()=>{(function(KA){var CA;di((CA=E())===null||CA===void 0||(CA=CA.projection)===null||CA===void 0?void 0:CA.paths,KA)||(d("changeProjectionPaths",KA),E(cs(E(),["projection","paths"],KA,!0)),h()(E()))})(g(C)?g(C).map(KA=>KA.value):void 0)}),Qn(),Zt(!0);var z=DKA(),U=W(z),H=W(U),q=IA(W(H)),j=W(q),gA=W(j);JC(gA,{class:"jse-filter-path",showChevron:!0,get items(){return g(l)},get value(){return g(L)},set value(KA){y(L,KA)},$$legacy:!0});var QA=IA(gA,2);JC(QA,{class:"jse-filter-relation",showChevron:!0,clearable:!1,get items(){return u},get value(){return g(R)},set value(KA){y(R,KA)},$$legacy:!0});var BA=IA(QA,2),lA=IA(H),vA=IA(W(lA)),tA=W(vA),cA=W(tA);JC(cA,{class:"jse-sort-path",showChevron:!0,get items(){return g(l)},get value(){return g(_)},set value(KA){y(_,KA)},$$legacy:!0}),JC(IA(cA,2),{class:"jse-sort-direction",showChevron:!0,clearable:!1,get items(){return D},get value(){return g(K)},set value(KA){y(K,KA)},$$legacy:!0});var pA=IA(lA),VA=IA(W(pA)),oe=W(VA);JC(W(oe),{class:"jse-projection-paths",multiple:!0,showChevron:!0,get items(){return g(I)},get value(){return g(C)},set value(KA){y(C,KA)},$$legacy:!0}),By(BA,()=>g(w),KA=>y(w,KA)),iA(t,z),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-select-query-language.svelte-atm4um {
  position: relative;
  width: 32px;
}
.jse-select-query-language.svelte-atm4um .jse-select-query-language-container:where(.svelte-atm4um) {
  position: absolute;
  top: 0;
  right: 0;
  display: flex;
  flex-direction: column;
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
}
.jse-select-query-language.svelte-atm4um .jse-select-query-language-container:where(.svelte-atm4um) .jse-query-language:where(.svelte-atm4um) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  text-align: left;
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  white-space: nowrap;
  color: var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff));
  background: var(--jse-context-menu-background, #656565);
}
.jse-select-query-language.svelte-atm4um .jse-select-query-language-container:where(.svelte-atm4um) .jse-query-language:where(.svelte-atm4um):hover {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
}`);var vKA=wA('<button type="button"><!> </button>'),bKA=wA('<div class="jse-select-query-language svelte-atm4um"><div class="jse-select-query-language-container svelte-atm4um"></div></div>');function MKA(t,e){mt(e,!1);var A=b(e,"queryLanguages",8),i=b(e,"queryLanguageId",12),n=b(e,"onChangeQueryLanguage",8);Zt();var o=bKA();qo(W(o),5,A,or,(r,s)=>{var a,c=vKA(),l=W(c),I=B=>{ji(B,{get data(){return zS}})},C=B=>{ji(B,{get data(){return OS}})};LA(l,B=>{g(s),k(i()),nA(()=>g(s).id===i())?B(I):B(C,!1)});var d=IA(l);De(B=>{var E;a=Vt(c,1,"jse-query-language svelte-atm4um",null,a,B),En(c,"title",(g(s),nA(()=>"Select ".concat(g(s).name," as query language")))),wt(d," ".concat((g(s),(E=nA(()=>g(s).name))!==null&&E!==void 0?E:"")))},[()=>({selected:g(s).id===i()})],qA),ce("click",c,()=>{return B=g(s).id,i(B),void n()(B);var B}),iA(r,c)}),iA(t,o),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-header.svelte-1y24war {
  display: flex;
  background: var(--jse-theme-color, #3883fa);
  color: var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
}
.jse-header.svelte-1y24war .jse-title:where(.svelte-1y24war) {
  flex: 1;
  padding: 5px;
  vertical-align: middle;
}
.jse-header.svelte-1y24war button:where(.svelte-1y24war) {
  border: none;
  background: transparent;
  min-width: 32px;
  color: inherit;
  cursor: pointer;
}
.jse-header.svelte-1y24war button:where(.svelte-1y24war):hover {
  background: rgba(255, 255, 255, 0.1);
}`);var kKA=wA('<button type="button" class="jse-fullscreen svelte-1y24war" title="Toggle full screen"><!></button>'),SKA=wA('<div class="jse-header svelte-1y24war"><div class="jse-title svelte-1y24war"> </div> <!> <!> <button type="button" class="jse-close svelte-1y24war"><!></button></div>');function My(t,e){mt(e,!1);var A=b(e,"title",9,"Modal"),i=b(e,"fullScreenButton",9,!1),n=b(e,"fullscreen",13,!1),o=b(e,"onClose",9,void 0);Zt(!0);var r=SKA(),s=W(r),a=W(s),c=IA(s,2);jo(c,e,"actions",{},null);var l=IA(c,2),I=d=>{var B=kKA(),E=W(B),h=qA(()=>n()?FW:VW);ji(E,{get data(){return g(h)}}),ce("click",B,()=>n(!n())),iA(d,B)};LA(l,d=>{i()&&d(I)});var C=IA(l,2);ji(W(C),{get data(){return I4}}),De(()=>wt(a,A())),ce("click",C,()=>{var d;return(d=o())===null||d===void 0?void 0:d()}),iA(t,r),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-config.svelte-1kpylsp {
  border: none;
  background: transparent;
  min-width: 32px;
  color: inherit;
  cursor: pointer;
}
.jse-config.svelte-1kpylsp:hover {
  background: rgba(255, 255, 255, 0.1);
}
.jse-config.hide.svelte-1kpylsp {
  display: none;
}`);var RKA=wA('<button slot="actions" type="button" title="Select a query language"><!></button>'),t_=Sr("jsoneditor:AutoScrollHandler");function UrA(t){var e,A;function i(s){return s<20?200:s<50?400:1200}function n(){if(t){var s=.05*(e||0);t.scrollTop+=s}}function o(s){A&&s===e||(r(),t_("startAutoScroll",s),e=s,A=setInterval(n,50))}function r(){A&&(t_("stopAutoScroll"),clearInterval(A),A=void 0,e=void 0)}return t_("createAutoScrollHandler",t),{onDrag:function(s){if(t){var a=s.clientY,{top:c,bottom:l}=t.getBoundingClientRect();a<c?o(-i(c-a)):a>l?o(i(a-l)):r()}},onDragEnd:function(){r()}}}var xKA=(t,e,A,i)=>(t/=i/2)<1?A/2*t*t+e:-A/2*(--t*(t-2)-1)+e,paA=()=>{var t,e,A,i,n,o,r,s,a,c,l,I,C;function d(h){return h.getBoundingClientRect().top-(t.getBoundingClientRect?t.getBoundingClientRect().top:0)+A}function B(h){t.scrollTo?t.scrollTo(t.scrollLeft,h):t.scrollTop=h}function E(h){c||(c=h),B(o(l=h-c,A,s,a)),C=!0,l<a?requestAnimationFrame(E):function(){B(A+s),e&&r&&(e.setAttribute("tabindex","-1"),e.focus()),typeof I=="function"&&I(),c=0,C=!1}()}return function(h){var u=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};switch(a=1e3,n=u.offset||0,I=u.callback,o=u.easing||xKA,r=u.a11y||!1,typeof u.container){case"object":t=u.container;break;case"string":t=document.querySelector(u.container);break;default:t=window.document.documentElement}switch(A=t.scrollTop,typeof h){case"number":e=void 0,r=!1,i=A+h;break;case"object":i=d(e=h);break;case"string":e=document.querySelector(h),i=d(e)}switch(s=i-A+n,typeof u.duration){case"number":a=u.duration;break;case"function":a=u.duration(s)}C?c=0:requestAnimationFrame(E)}};function sQ(t,e){var A=Date.now(),i=t();return e(Date.now()-A),i}var tQ=Sr("validation"),LKA={createObjectDocumentState:()=>({type:"object",properties:{}}),createArrayDocumentState:()=>({type:"array",items:[]}),createValueDocumentState:()=>({type:"value"})};function KrA(t,e,A,i){return rG(t,e,A,i,LKA)}function waA(t,e,A,i){if(tQ("validateJSON"),!e)return[];if(A!==i){var n=A.stringify(t);return e(n!==void 0?i.parse(n):void 0)}return e(t)}function FKA(t,e,A,i){if(tQ("validateText"),t.length>104857600)return{validationErrors:[{path:[],message:"Validation turned off: the document is too large",severity:Fl.info}]};if(t.length!==0)try{var n=sQ(()=>A.parse(t),a=>tQ("validate: parsed json in ".concat(a," ms")));if(!e)return;var o=A===i?n:sQ(()=>i.parse(t),a=>tQ("validate: parsed json with the validationParser in ".concat(a," ms"))),r=sQ(()=>e(o),a=>tQ("validate: validated json in ".concat(a," ms")));return Oi(r)?void 0:{validationErrors:r}}catch(a){var s=sQ(()=>function(c,l){if(c.length>aUA)return!1;try{return l.parse(Rc(c)),!0}catch{return!1}}(t,A),c=>tQ("validate: checked whether repairable in ".concat(c," ms")));return{parseError:dQ(t,a.message||a.toString()),isRepairable:s}}}var ey=Sr("jsoneditor:FocusTracker");function CG(t){var e,{onMount:A,onDestroy:i,getWindow:n,hasFocus:o,onFocus:r,onBlur:s}=t,a=!1;function c(){var I=o();I&&(clearTimeout(e),a||(ey("focus"),r(),a=I))}function l(){a&&(clearTimeout(e),e=setTimeout(()=>{o()||(ey("blur"),a=!1,s())}))}A(()=>{ey("mount FocusTracker");var I=n();I&&(I.addEventListener("focusin",c,!0),I.addEventListener("focusout",l,!0))}),i(()=>{ey("destroy FocusTracker");var I=n();I&&(I.removeEventListener("focusin",c,!0),I.removeEventListener("focusout",l,!0))})}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-message.svelte-czprfx {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  padding: var(--jse-padding, 10px);
  display: flex;
  gap: var(--jse-padding, 10px);
  flex-wrap: wrap;
  align-items: stretch;
}
.jse-message.jse-success.svelte-czprfx {
  background: var(--message-success-background, #9ac45d);
  color: var(--jse-message-success-color, #fff);
}
.jse-message.svelte-czprfx .jse-text:where(.svelte-czprfx) {
  display: flex;
  flex: 1;
  min-width: 60%;
  align-items: center;
}
.jse-message.svelte-czprfx .jse-text.jse-clickable:where(.svelte-czprfx) {
  cursor: pointer;
}
.jse-message.svelte-czprfx .jse-text.jse-clickable:where(.svelte-czprfx):hover {
  background-color: rgba(255, 255, 255, 0.1);
}
.jse-message.jse-error.svelte-czprfx {
  background: var(--jse-message-error-background, var(--jse-error-color, #ee5341));
  color: var(--jse-message-error-color, #fff);
}
.jse-message.jse-warning.svelte-czprfx {
  background: var(--jse-message-warning-background, #ffde5c);
  color: var(--jse-message-warning-color, #4d4d4d);
}
.jse-message.jse-info.svelte-czprfx {
  background: var(--jse-message-info-background, #4f91ff);
  color: var(--jse-message-info-color, #fff);
}
.jse-message.svelte-czprfx .jse-actions:where(.svelte-czprfx) {
  display: flex;
  gap: var(--jse-padding, 10px);
}
.jse-message.svelte-czprfx .jse-actions:where(.svelte-czprfx) button.jse-action:where(.svelte-czprfx) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-message-action-background, rgba(255, 255, 255, 0.2));
  color: inherit;
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
}
.jse-message.svelte-czprfx .jse-actions:where(.svelte-czprfx) button.jse-action:where(.svelte-czprfx):hover {
  background: var(--jse-message-action-background-highlight, rgba(255, 255, 255, 0.3));
}`);var NKA=wA('<button type="button" class="jse-button jse-action jse-primary svelte-czprfx"><!> </button>'),_KA=wA('<div><div role="button" tabindex="-1"><div class="jse-text-centered"><!> </div></div> <div class="jse-actions svelte-czprfx"></div></div>');function mc(t,e){mt(e,!1);var A=b(e,"type",9,"success"),i=b(e,"icon",9,void 0),n=b(e,"message",9,void 0),o=b(e,"actions",25,()=>[]),r=b(e,"onClick",9,void 0),s=b(e,"onClose",9,void 0);s()&&qc(s()),Zt(!0);var a,c=_KA(),l=W(c),I=W(l),C=W(I),d=E=>{ji(E,{get data(){return i()}})};LA(C,E=>{i()&&E(d)});var B=IA(C);qo(IA(l,2),5,o,or,(E,h)=>{var u=NKA(),D=W(u),L=w=>{ji(w,{get data(){return g(h),nA(()=>g(h).icon)}})};LA(D,w=>{g(h),nA(()=>g(h).icon)&&w(L)});var R=IA(D);De(()=>{var w;En(u,"title",(g(h),nA(()=>g(h).title))),u.disabled=(g(h),nA(()=>g(h).disabled)),wt(R," ".concat((g(h),(w=nA(()=>g(h).text))!==null&&w!==void 0?w:"")))}),ce("click",u,()=>{g(h).onClick&&g(h).onClick()}),ce("mousedown",u,()=>{g(h).onMouseDown&&g(h).onMouseDown()}),iA(E,u)}),De(E=>{var h,u;Vt(c,1,"jse-message jse-".concat((h=A())!==null&&h!==void 0?h:""),"svelte-czprfx"),a=Vt(l,1,"jse-text svelte-czprfx",null,a,E),wt(B," ".concat((u=n())!==null&&u!==void 0?u:""))},[()=>({"jse-clickable":!!r()})],qA),ce("click",l,function(){r()&&r()()}),iA(t,c),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-validation-errors-overview.svelte-1uindol {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  overflow: auto;
  max-height: 25%;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) {
  border-collapse: collapse;
  width: 100%;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) {
  cursor: pointer;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr.jse-validation-error:where(.svelte-1uindol) {
  background: var(--jse-message-error-background, var(--jse-error-color, #ee5341));
  color: var(--jse-message-error-color, #fff);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr.jse-validation-warning:where(.svelte-1uindol) {
  background: var(--jse-message-warning-background, #ffde5c);
  color: var(--jse-message-warning-color, #4d4d4d);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr.jse-validation-warning:where(.svelte-1uindol):hover {
  filter: brightness(105%);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr.jse-validation-info:where(.svelte-1uindol) {
  background: var(--jse-message-info-background, #4f91ff);
  color: var(--jse-message-info-color, #fff);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol):hover {
  filter: brightness(110%);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td:where(.svelte-1uindol) {
  padding: 4px var(--jse-padding, 10px);
  vertical-align: middle;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td.jse-validation-error-icon:where(.svelte-1uindol) {
  width: 36px;
  box-sizing: border-box;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td.jse-validation-error-action:where(.svelte-1uindol) {
  width: 36px;
  box-sizing: border-box;
  padding: 0;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td.jse-validation-error-action:where(.svelte-1uindol) button.jse-validation-errors-collapse:where(.svelte-1uindol) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  width: 36px;
  height: 26px;
  cursor: pointer;
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td.jse-validation-error-action:where(.svelte-1uindol) button.jse-validation-errors-collapse:where(.svelte-1uindol):hover {
  background-color: rgba(255, 255, 255, 0.2);
}
.jse-validation-errors-overview.svelte-1uindol table:where(.svelte-1uindol) tr:where(.svelte-1uindol) td:where(.svelte-1uindol) div.jse-validation-errors-expand:where(.svelte-1uindol) {
  display: inline-block;
  position: relative;
  top: 3px;
}`);var GKA=wA('<button type="button" class="jse-validation-errors-collapse svelte-1uindol" title="Collapse validation errors"><!></button>'),UKA=wA('<tr tabindex="0"><td class="jse-validation-error-icon svelte-1uindol"><!></td><td class="jse-validation-error-path svelte-1uindol"> </td><td class="jse-validation-error-message svelte-1uindol"> </td><td class="jse-validation-error-action svelte-1uindol"><!></td></tr>'),KKA=wA('<tr class="jse-validation-error svelte-1uindol"><td class="svelte-1uindol"></td><td class="svelte-1uindol"></td><td class="svelte-1uindol"> </td><td class="svelte-1uindol"></td></tr>'),YKA=wA('<table class="jse-validation-errors-overview-expanded svelte-1uindol"><tbody><!><!></tbody></table>'),JKA=wA('<table class="jse-validation-errors-overview-collapsed svelte-1uindol"><tbody><tr><td class="jse-validation-error-icon svelte-1uindol"><!></td><td class="jse-validation-error-count svelte-1uindol"> <div class="jse-validation-errors-expand svelte-1uindol"><!></div></td></tr></tbody></table>'),TKA=wA('<div class="jse-validation-errors-overview svelte-1uindol"><!></div>');function dG(t,e){mt(e,!1);var A=$(void 0,!0),i=b(e,"validationErrors",9),n=b(e,"selectError",9),o=$(!0,!0);function r(){y(o,!1)}function s(){y(o,!0)}fA(()=>k(i()),()=>{y(A,i().length)}),Qn(),Zt(!0);var a=_o(),c=vt(a),l=I=>{var C=TKA(),d=W(C),B=h=>{var u=YKA(),D=W(u),L=W(D);qo(L,1,()=>(k(Ey),k(i()),k(VD),nA(()=>Ey(i(),VD))),or,(_,K,z)=>{var U=UKA(),H=W(U);ji(W(H),{get data(){return g1}});var q=IA(H),j=W(q),gA=IA(q),QA=W(gA),BA=W(IA(gA)),lA=vA=>{var tA=GKA();ji(W(tA),{get data(){return zW}}),ce("click",tA,j0(r)),iA(vA,tA)};LA(BA,vA=>{k(i()),nA(()=>z===0&&i().length>1)&&vA(lA)}),De(vA=>{var tA;Vt(U,1,"jse-validation-".concat((g(K),(tA=nA(()=>g(K).severity))!==null&&tA!==void 0?tA:"")),"svelte-1uindol"),wt(j,vA),wt(QA,(g(K),nA(()=>g(K).message)))},[()=>(k(Ka),g(K),nA(()=>Ka(g(K).path)))],qA),ce("click",U,()=>{setTimeout(()=>n()(g(K)))}),iA(_,U)});var R=IA(L),w=_=>{var K=KKA(),z=IA(W(K),2),U=W(z);De(()=>wt(U,"(and ".concat(g(A)-VD," more errors)"))),iA(_,K)};LA(R,_=>{g(A)>VD&&_(w)}),iA(h,u)},E=h=>{var u=JKA(),D=W(u),L=W(D),R=W(L);ji(W(R),{get data(){return g1}});var w=W(IA(R));ji(W(IA(w)),{get data(){return qS}}),De(_=>{var K;Vt(L,1,"jse-validation-".concat(_??""),"svelte-1uindol"),wt(w,"".concat((K=g(A))!==null&&K!==void 0?K:""," validation errors "))},[()=>(k(i()),nA(()=>{return _=i(),[Fl.error,Fl.warning,Fl.info].find(K=>_.some(z=>z.severity===K));var _}))],qA),ce("click",L,s),iA(h,u)};LA(d,h=>{g(o)||g(A)===1?h(B):h(E,!1)}),iA(I,C)};LA(c,I=>{k(Oi),k(i()),nA(()=>!Oi(i()))&&I(l)}),iA(t,a),pt()}function ky(t,e){if(t)return t.addEventListener("keydown",A),{destroy(){t.removeEventListener("keydown",A)}};function A(i){i.key==="Escape"&&(i.preventDefault(),i.stopPropagation(),e())}}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
dialog.jse-modal.svelte-1s9c2ql {
  border-radius: 3px;
  font-size: var(--jse-padding, 10px);
  border: none;
  padding: 0;
  display: flex;
  min-width: 0;
  margin: auto;
  overflow: visible;
  transition: width 0.1s ease-in-out, height 0.1s ease-in-out;
}
dialog.jse-modal.jse-sort-modal.svelte-1s9c2ql {
  width: 400px;
}
dialog.jse-modal.jse-repair-modal.svelte-1s9c2ql {
  width: 600px;
  height: 500px;
}
dialog.jse-modal.jse-jsoneditor-modal.svelte-1s9c2ql {
  width: 800px;
  height: 600px;
}
dialog.jse-modal.jse-transform-modal.svelte-1s9c2ql {
  width: 1200px;
  height: 800px;
}
dialog.jse-modal.jse-fullscreen.svelte-1s9c2ql {
  width: 100%;
  height: 100%;
}
dialog.jse-modal.svelte-1s9c2ql::backdrop {
  background: var(--jse-overlay-background, rgba(0, 0, 0, 0.3));
}
dialog.jse-modal[open].svelte-1s9c2ql {
  animation: svelte-1s9c2ql-zoom 0.3s cubic-bezier(0.34, 1.56, 0.64, 1);
}
dialog.jse-modal[open].svelte-1s9c2ql::backdrop {
  animation: svelte-1s9c2ql-fade 0.2s ease-out;
}
dialog.jse-modal.svelte-1s9c2ql .jse-modal-inner:where(.svelte-1s9c2ql) {
  flex: 1;
  display: flex;
  flex-direction: column;
  min-width: 0;
  min-height: 0;
  padding: 0;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  line-height: normal;
  background: var(--jse-modal-background, #f5f5f5);
  color: var(--jse-text-color, #4d4d4d);
}
@keyframes svelte-1s9c2ql-zoom {
  from {
    transform: scale(0.95);
  }
  to {
    transform: scale(1);
  }
}
@keyframes svelte-1s9c2ql-fade {
  from {
    opacity: 0;
  }
  to {
    opacity: 1;
  }
}
dialog.jse-modal.svelte-1s9c2ql .svelte-select {
  --border: var(--jse-svelte-select-border, 1px solid #d8dbdf);
  --item-is-active-bg: var(--jse-item-is-active-bg, #3883fa);
  --border-radius: var(--jse-svelte-select-border-radius, 3px);
  --background: var(--jse-svelte-select-background, #fff);
  --padding: var(--jse-svelte-select-padding, 0 10px);
  --multi-select-padding: var(--jse-svelte-select-multi-select-padding, 0 10px);
  --font-size: var(--jse-svelte-select-font-size, var(--jse-font-size, 16px));
  --height: 36px;
  --multi-item-height: 28px;
  --multi-item-margin: 2px;
  --multi-item-padding: 2px 8px;
  --multi-item-border-radius: 6px;
  --indicator-top: 8px;
}`);var HKA=wA('<dialog><div class="jse-modal-inner svelte-1s9c2ql"><!></div></dialog>');function X3(t,e){mt(e,!1);var A=b(e,"className",8,void 0),i=b(e,"fullscreen",8,!1),n=b(e,"onClose",8),o=$();function r(){n()()}hs(()=>g(o).showModal()),qc(()=>g(o).close()),Zt();var s,a=HKA(),c=W(a);jo(W(c),e,"default",{},null),Co(a,l=>y(o,l),()=>g(o)),es(()=>ce("close",a,r)),es(()=>{return ce("pointerdown",a,(l=r,function(){for(var I=arguments.length,C=new Array(I),d=0;d<I;d++)C[d]=arguments[d];C[0].target===this&&l?.apply(this,C)}));var l}),es(()=>ce("cancel",a,N1(function(l){N3.call(this,e,l)}))),Us(a,(l,I)=>ky?.(l,I),()=>r),De((l,I)=>s=Vt(a,1,l,"svelte-1s9c2ql",s,I),[()=>Z1((k(Kl),k(A()),nA(()=>Kl("jse-modal",A())))),()=>({"jse-fullscreen":i()})],qA),iA(t,a),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-modal-contents.svelte-189qksl {
  flex: 1;
  display: flex;
  flex-direction: column;
  padding: 20px;
  overflow: auto;
  min-width: 0;
  min-height: 0;
}
.jse-modal-contents.svelte-189qksl .jse-actions:where(.svelte-189qksl) {
  display: flex;
  flex-direction: row;
  justify-content: flex-end;
  padding-top: var(--jse-padding, 10px);
}
.jse-modal-contents.svelte-189qksl .jse-actions:where(.svelte-189qksl) button.jse-primary:where(.svelte-189qksl) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-modal-contents.svelte-189qksl .jse-actions:where(.svelte-189qksl) button.jse-primary:where(.svelte-189qksl):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-modal-contents.svelte-189qksl .jse-actions:where(.svelte-189qksl) button.jse-primary:where(.svelte-189qksl):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}

.jse-shortcuts.svelte-189qksl {
  display: flex;
  flex-wrap: wrap;
  justify-content: space-around;
  margin: calc(2 * var(--jse-padding, 10px)) 0;
}
.jse-shortcuts.svelte-189qksl .jse-shortcut:where(.svelte-189qksl) .jse-key:where(.svelte-189qksl) {
  font-size: 200%;
  color: var(--jse-theme-color, #3883fa);
}`);var zKA=wA('<!> <div class="jse-modal-contents svelte-189qksl"><div>Clipboard permission is disabled by your browser. You can use:</div> <div class="jse-shortcuts svelte-189qksl"><div class="jse-shortcut svelte-189qksl"><div class="jse-key svelte-189qksl"> </div> for copy</div> <div class="jse-shortcut svelte-189qksl"><div class="jse-key svelte-189qksl"> </div> for cut</div> <div class="jse-shortcut svelte-189qksl"><div class="jse-key svelte-189qksl"> </div> for paste</div></div> <div class="jse-actions svelte-189qksl"><button type="button" class="jse-primary svelte-189qksl">Close</button></div></div>',1);function DaA(t,e){mt(e,!1);var A=b(e,"onClose",9),i=X_()?"\u2318":"Ctrl";Zt(!0),X3(t,{get onClose(){return A()},className:"jse-copy-paste",children:(n,o)=>{var r=zKA(),s=vt(r);My(s,{title:"Copying and pasting",get onClose(){return A()}});var a=IA(s,2),c=IA(W(a),2),l=W(c),I=W(l),C=W(I),d=IA(l,2),B=W(d),E=W(B),h=W(IA(d,2)),u=W(h),D=W(IA(c,2));De(()=>{wt(C,"".concat(i,"+C")),wt(E,"".concat(i,"+X")),wt(u,"".concat(i,"+V"))}),ce("click",D,function(){for(var L,R=arguments.length,w=new Array(R),_=0;_<R;_++)w[_]=arguments[_];(L=A())===null||L===void 0||L.apply(this,w)}),iA(n,r)},$$slots:{default:!0}}),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-menu.svelte-pf7s2l {
  background: var(--jse-theme-color, #3883fa);
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size-main-menu, 14px);
  color: var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
  display: flex;
  flex-wrap: wrap;
  align-items: stretch;
  position: relative;
}
.jse-menu.svelte-pf7s2l .jse-button:where(.svelte-pf7s2l) {
  font-family: inherit;
  font-size: inherit;
  line-height: 1.5em;
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  width: var(--jse-menu-button-size, 32px);
  height: var(--jse-menu-button-size, 32px);
  padding: calc(0.5 * var(--jse-padding, 10px));
  margin: 0;
  border-radius: 0;
  display: inline-flex;
  align-items: center;
  text-align: center;
  justify-content: center;
}
.jse-menu.svelte-pf7s2l .jse-button:where(.svelte-pf7s2l):hover, .jse-menu.svelte-pf7s2l .jse-button:where(.svelte-pf7s2l):focus {
  background: var(--jse-theme-color-highlight, #5f9dff);
}
.jse-menu.svelte-pf7s2l .jse-button:where(.svelte-pf7s2l):disabled {
  color: var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
  opacity: 0.5;
  background: transparent;
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button:where(.svelte-pf7s2l) {
  width: auto;
  height: calc(var(--jse-menu-button-size, 32px) - var(--jse-padding, 10px));
  margin: calc(0.5 * var(--jse-padding, 10px)) 0;
  padding: 0 calc(0.5 * var(--jse-padding, 10px)) 1px;
  border: 1px solid var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button:where(.svelte-pf7s2l):not(.jse-last) {
  border-right: none;
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button.jse-first:where(.svelte-pf7s2l) {
  margin-left: calc(0.5 * var(--jse-padding, 10px));
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button.jse-last:where(.svelte-pf7s2l) {
  margin-right: calc(0.5 * var(--jse-padding, 10px));
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button:where(.svelte-pf7s2l):hover, .jse-menu.svelte-pf7s2l .jse-button.jse-group-button:where(.svelte-pf7s2l):focus {
  background: var(--jse-theme-color-highlight, #5f9dff);
}
.jse-menu.svelte-pf7s2l .jse-button.jse-group-button.jse-selected:where(.svelte-pf7s2l) {
  background: var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
  color: var(--jse-theme-color, #3883fa);
}
.jse-menu.svelte-pf7s2l .jse-space:where(.svelte-pf7s2l) {
  flex: 1;
}
.jse-menu.svelte-pf7s2l .jse-separator:where(.svelte-pf7s2l) {
  background: var(--jse-menu-color, var(--jse-text-color-inverse, #fff));
  opacity: 0.3;
  width: 1px;
  margin: 3px;
}`);var OKA=wA('<div class="jse-separator svelte-pf7s2l"></div>'),PKA=wA('<div class="jse-space svelte-pf7s2l"></div>'),jKA=wA('<button type="button"><!> <!></button>'),qKA=wA('<div class="jse-menu svelte-pf7s2l"><!> <!> <!></div>');function Oy(t,e){mt(e,!1);var A=b(e,"items",25,()=>[]);Zt(!0);var i=qKA(),n=W(i);jo(n,e,"left",{},null);var o=IA(n,2);qo(o,1,A,or,(r,s)=>{var a=_o(),c=vt(a),l=C=>{iA(C,OKA())},I=(C,d)=>{var B=h=>{iA(h,PKA())},E=(h,u)=>{var D=R=>{var w=jKA(),_=W(w),K=H=>{ji(H,{get data(){return g(s),nA(()=>g(s).icon)}})};LA(_,H=>{g(s),nA(()=>g(s).icon)&&H(K)});var z=IA(_,2),U=H=>{var q=Tr();De(()=>wt(q,(g(s),nA(()=>g(s).text)))),iA(H,q)};LA(z,H=>{g(s),nA(()=>g(s).text)&&H(U)}),De(()=>{var H;Vt(w,1,"jse-button ".concat((g(s),(H=nA(()=>g(s).className))!==null&&H!==void 0?H:"")),"svelte-pf7s2l"),En(w,"title",(g(s),nA(()=>g(s).title))),w.disabled=(g(s),nA(()=>g(s).disabled||!1))}),ce("click",w,function(){for(var H,q=arguments.length,j=new Array(q),gA=0;gA<q;gA++)j[gA]=arguments[gA];(H=g(s).onClick)===null||H===void 0||H.apply(this,j)}),iA(R,w)},L=R=>{var w=Tr();De(_=>wt(w,_),[()=>(g(s),nA(()=>function(_){return console.error("Unknown type of menu item",_),"???"}(g(s))))],qA),iA(R,w)};LA(h,R=>{k(q0),g(s),nA(()=>q0(g(s)))?R(D):R(L,!1)},u)};LA(C,h=>{k(p_),g(s),nA(()=>p_(g(s)))?h(B):h(E,!1)},d)};LA(c,C=>{k(G1),g(s),nA(()=>G1(g(s)))?C(l):C(I,!1)}),iA(r,a)}),jo(IA(o,2),e,"right",{},null),iA(t,i),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-json-repair-component.svelte-3golau {
  flex: 1;
  display: flex;
  flex-direction: column;
  background: var(--jse-background-color, #fff);
  color: var(--jse-text-color, #4d4d4d);
}
.jse-json-repair-component.svelte-3golau .jse-info:where(.svelte-3golau) {
  padding: calc(0.5 * var(--jse-padding, 10px));
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  vertical-align: center;
}
.jse-json-repair-component.svelte-3golau .jse-json-text:where(.svelte-3golau) {
  flex: 1;
  border: none;
  padding: 2px;
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  background: var(--jse-input-background, var(--jse-background-color, #fff));
  color: var(--jse-text-color, #4d4d4d);
  resize: none;
  outline: none;
}`);var VKA=wA('<div slot="left" class="jse-info svelte-3golau">Repair invalid JSON, then click apply</div>'),ZKA=wA('<div class="jse-json-repair-component svelte-3golau"><!> <!> <textarea class="jse-json-text svelte-3golau" autocomplete="off" autocapitalize="off" spellcheck="false"></textarea></div>');function WKA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=$(void 0,!0),o=$(void 0,!0),r=$(void 0,!0),s=$(void 0,!0),a=b(e,"text",13,""),c=b(e,"readOnly",9,!1),l=b(e,"onParse",9),I=b(e,"onRepair",9),C=b(e,"onChange",9,void 0),d=b(e,"onApply",9),B=b(e,"onCancel",9),E=Sr("jsoneditor:JSONRepair"),h=$(void 0,!0);function u(){if(g(h)&&g(A)){var q=g(A).position!==void 0?g(A).position:0;g(h).setSelectionRange(q,q),g(h).focus()}}function D(){d()(a())}function L(){try{a(I()(a())),C()&&C()(a())}catch{}}var R=$(void 0,!0);fA(()=>k(a()),()=>{y(A,function(q){try{return void l()(q)}catch(j){return dQ(q,j.message)}}(a()))}),fA(()=>k(a()),()=>{y(i,function(q){try{return I()(q),!0}catch{return!1}}(a()))}),fA(()=>g(A),()=>{E("error",g(A))}),fA(()=>k(B()),()=>{y(R,[{type:"space"},{type:"button",icon:I4,title:"Cancel repair",className:"jse-cancel",onClick:B()}])}),fA(()=>XS,()=>{y(n,{icon:XS,text:"Show me",title:"Scroll to the error location",onClick:u})}),fA(()=>L0,()=>{y(o,{icon:L0,text:"Auto repair",title:"Automatically repair JSON",onClick:L})}),fA(()=>(g(i),g(n),g(o)),()=>{y(r,g(i)?[g(n),g(o)]:[g(n)])}),fA(()=>k(c()),()=>{y(s,[{icon:u5,text:"Apply",title:"Apply fixed JSON",disabled:c(),onClick:D}])}),Qn(),Zt(!0);var w=ZKA(),_=W(w);Oy(_,{get items(){return g(R)},$$slots:{left:(q,j)=>{iA(q,VKA())}}});var K=IA(_,2),z=q=>{var j=qA(()=>(g(A),nA(()=>"Cannot parse JSON: ".concat(g(A).message))));mc(q,{type:"error",get icon(){return g1},get message(){return g(j)},get actions(){return g(r)}})},U=q=>{mc(q,{type:"success",message:"JSON is valid now and can be parsed.",get actions(){return g(s)}})};LA(K,q=>{g(A)?q(z):q(U,!1)});var H=IA(K,2);Co(H,q=>y(h,q),()=>g(h)),De(()=>{H.readOnly=c(),VC(H,a())}),ce("input",H,function(q){E("handleChange");var j=q.target.value;a()!==j&&(a(j),C()&&C()(a()))}),iA(t,w),pt()}function yaA(t,e){mt(e,!1);var A=b(e,"text",13),i=b(e,"onParse",9),n=b(e,"onRepair",9),o=b(e,"onApply",9),r=b(e,"onClose",9);function s(c){o()(c),r()()}function a(){r()()}Zt(!0),X3(t,{get onClose(){return r()},className:"jse-repair-modal",children:(c,l)=>{WKA(c,{get onParse(){return i()},get onRepair(){return n()},onApply:s,onCancel:a,get text(){return A()},set text(I){A(I)},$$legacy:!0})},$$slots:{default:!0}}),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
div.jse-collapsed-items.svelte-1h6hzoq {
  margin-left: calc(var(--level) * var(--jse-indent-size, calc(1em + 4px)));
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  color: var(--jse-collapsed-items-link-color, rgba(0, 0, 0, 0.38));
  padding: calc(0.5 * var(--jse-padding, 10px));
  border: 8px solid transparent;
  border-width: 8px 0;
  background-color: var(--jse-contents-background-color, transparent);
  background-image: linear-gradient(var(--jse-collapsed-items-background-color, #f5f5f5), var(--jse-collapsed-items-background-color, #f5f5f5)), linear-gradient(to bottom right, transparent 50.5%, var(--jse-collapsed-items-background-color, #f5f5f5) 50.5%), linear-gradient(to bottom left, transparent 50.5%, var(--jse-collapsed-items-background-color, #f5f5f5) 50.5%), linear-gradient(to top right, transparent 50.5%, var(--jse-collapsed-items-background-color, #f5f5f5) 50.5%), linear-gradient(to top left, transparent 50.5%, var(--jse-collapsed-items-background-color, #f5f5f5) 50.5%);
  background-repeat: repeat, repeat-x, repeat-x, repeat-x, repeat-x;
  background-position: 0 0, 8px 0, 8px 0, 8px 100%, 8px 100%;
  background-size: auto auto, 16px 16px, 16px 16px, 16px 16px, 16px 16px;
  background-clip: padding-box, border-box, border-box, border-box, border-box;
  background-origin: padding-box, border-box, border-box, border-box, border-box;
  display: flex;
}
div.jse-collapsed-items.jse-selected.svelte-1h6hzoq {
  background-color: var(--jse-selection-background-color, #d3d3d3);
  --jse-collapsed-items-background-color: var(--jse-collapsed-items-selected-background-color, #c2c2c2);
}
div.jse-collapsed-items.svelte-1h6hzoq div.jse-text:where(.svelte-1h6hzoq),
div.jse-collapsed-items.svelte-1h6hzoq button.jse-expand-items:where(.svelte-1h6hzoq) {
  margin: 0 calc(0.5 * var(--jse-padding, 10px));
}
div.jse-collapsed-items.svelte-1h6hzoq div.jse-text:where(.svelte-1h6hzoq) {
  display: inline;
}
div.jse-collapsed-items.svelte-1h6hzoq button.jse-expand-items:where(.svelte-1h6hzoq) {
  font-family: inherit;
  font-size: inherit;
  color: var(--jse-collapsed-items-link-color, rgba(0, 0, 0, 0.38));
  background: none;
  border: none;
  padding: 0;
  text-decoration: underline;
  cursor: pointer;
}
div.jse-collapsed-items.svelte-1h6hzoq button.jse-expand-items:where(.svelte-1h6hzoq):hover, div.jse-collapsed-items.svelte-1h6hzoq button.jse-expand-items:where(.svelte-1h6hzoq):focus {
  color: var(--jse-collapsed-items-link-color-highlight, #ee5341);
}`);var XKA=wA('<button type="button" class="jse-expand-items svelte-1h6hzoq"> </button>'),$KA=wA('<div role="none"><div><div class="jse-text svelte-1h6hzoq"> </div> <!></div></div>');function AYA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=$(void 0,!0),o=$(void 0,!0),r=$(void 0,!0),s=b(e,"visibleSections",9),a=b(e,"sectionIndex",9),c=b(e,"total",9),l=b(e,"path",9),I=b(e,"selection",9),C=b(e,"onExpandSection",9),d=b(e,"context",9);fA(()=>(k(s()),k(a())),()=>{y(A,s()[a()])}),fA(()=>g(A),()=>{y(i,g(A).end)}),fA(()=>(k(s()),k(a()),k(c())),()=>{y(n,s()[a()+1]?s()[a()+1].start:c())}),fA(()=>(k(d()),k(I()),k(l()),g(i)),()=>{y(o,V3(d().getJson(),I(),l().concat(String(g(i)))))}),fA(()=>(g(i),g(n)),()=>{y(r,function(R,w){var _={start:R,end:Math.min(m_(R),w)},K=Math.max(hy((R+w)/2),R),z={start:K,end:Math.min(m_(K),w)},U=hy(w),H=U===w?U-O3:U,q={start:Math.max(H,R),end:w},j=[_],gA=z.start>=_.end&&z.end<=q.start;return gA&&j.push(z),q.start>=(gA?z.end:_.end)&&j.push(q),j}(g(i),g(n)))}),Qn(),Zt(!0);var B,E,h=$KA(),u=W(h),D=W(u),L=W(D);qo(IA(D,2),1,()=>g(r),or,(R,w)=>{var _=XKA(),K=W(_);De(()=>{var z,U;return wt(K,"show ".concat((g(w),(z=nA(()=>g(w).start))!==null&&z!==void 0?z:""),"-").concat((g(w),(U=nA(()=>g(w).end))!==null&&U!==void 0?U:"")))}),ce("click",_,()=>C()(l(),g(w))),iA(R,_)}),De((R,w)=>{var _,K;B=Vt(h,1,"jse-collapsed-items svelte-1h6hzoq",null,B,R),E=Ll(h,"",E,w),wt(L,"Items ".concat((_=g(i))!==null&&_!==void 0?_:"","-").concat((K=g(n))!==null&&K!==void 0?K:""))},[()=>({"jse-selected":g(o)}),()=>({"--level":(k(l()),nA(()=>l().length+2))})],qA),ce("mousemove",h,function(R){R.stopPropagation()}),iA(t,h),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-context-menu-pointer.svelte-137iwnw {
  position: absolute;
  top: calc(-0.5 * var(--jse-context-menu-pointer-size, calc(1em + 4px)));
  right: calc(-0.5 * var(--jse-context-menu-pointer-size, calc(1em + 4px)));
  width: var(--jse-context-menu-pointer-size, calc(1em + 4px));
  height: var(--jse-context-menu-pointer-size, calc(1em + 4px));
  padding: 0;
  margin: 0;
  cursor: pointer;
  background: transparent;
  border-radius: 2px;
  background: var(--jse-context-menu-pointer-hover-background, #b2b2b2);
  color: var(--jse-context-menu-pointer-color, var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff)));
  border: none;
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
}
.jse-context-menu-pointer.jse-root.svelte-137iwnw {
  top: 0;
  right: calc(-2px - var(--jse-context-menu-pointer-size, calc(1em + 4px)));
}
.jse-context-menu-pointer.jse-insert.svelte-137iwnw {
  right: -1px;
}
.jse-context-menu-pointer.svelte-137iwnw:hover {
  background: var(--jse-context-menu-pointer-background-highlight, var(--jse-context-menu-background-highlight, #7a7a7a));
}
.jse-context-menu-pointer.jse-selected.svelte-137iwnw {
  background: var(--jse-context-menu-pointer-background, var(--jse-context-menu-background, #656565));
}
.jse-context-menu-pointer.jse-selected.svelte-137iwnw:hover {
  background: var(--jse-context-menu-pointer-background-highlight, var(--jse-context-menu-background-highlight, #7a7a7a));
}`);var eYA=wA('<button type="button"><!></button>');function _1(t,e){mt(e,!1);var A=b(e,"root",9,!1),i=b(e,"insert",9,!1),n=b(e,"selected",9),o=b(e,"onContextMenu",9);Zt(!0);var r,s=eYA();ji(W(s),{get data(){return mg}}),De(a=>{r=Vt(s,1,"jse-context-menu-pointer svelte-137iwnw",null,r,a),En(s,"title",AG)},[()=>({"jse-root":A(),"jse-insert":i(),"jse-selected":n()})],qA),ce("click",s,function(a){for(var c=a.target;c&&c.nodeName!=="BUTTON";)c=c.parentNode;c&&o()({anchor:c,left:0,top:0,width:W0,height:Z0,offsetTop:2,offsetLeft:0,showTip:!0})}),iA(t,s),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-key.svelte-2iqnqn {
  display: inline-block;
  min-width: 2em;
  padding: 0 5px;
  box-sizing: border-box;
  outline: none;
  border-radius: 1px;
  vertical-align: top;
  color: var(--jse-key-color, #1a1a1a);
  word-break: normal;
  overflow-wrap: normal;
  white-space: pre-wrap;
}
.jse-key.jse-empty.svelte-2iqnqn {
  min-width: 3em;
  outline: 1px dotted var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  -moz-outline-radius: 2px;
}
.jse-key.jse-empty.svelte-2iqnqn::after {
  pointer-events: none;
  color: var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  content: "key";
}`);var tYA=wA('<div role="none" data-type="selectable-key"><!></div>'),iYA=wA("<!> <!>",1),nYA=wA('<div role="button" tabindex="-1" class="jse-value" data-type="selectable-value"></div>');function vaA(t,e){mt(e,!0);var A=Ga(()=>an(e.selection)&&br(e.selection)),i=Ga(()=>e.context.onRenderValue({path:e.path,value:e.value,mode:e.context.mode,truncateTextSize:e.context.truncateTextSize,readOnly:e.context.readOnly,enforceString:e.enforceString,isEditing:g(A),parser:e.context.parser,normalization:e.context.normalization,selection:e.selection,searchResultItems:e.searchResultItems,onPatch:e.context.onPatch,onPasteJson:e.context.onPasteJson,onSelect:e.context.onSelect,onFind:e.context.onFind,findNextInside:e.context.findNextInside,focus:e.context.focus})),n=_o();qo(vt(n),17,()=>g(i),or,(o,r)=>{var s=_o(),a=vt(s),c=I=>{var C=nYA(),d=Ga(()=>g(r).action);Us(C,(B,E)=>{var h;return(h=g(d))===null||h===void 0?void 0:h(B,E)},()=>g(r).props),iA(I,C)},l=I=>{var C=_o(),d=Ga(()=>g(r).component);MsA(vt(C),()=>g(d),(B,E)=>{E(B,z1(()=>g(r).props))}),iA(I,C)};LA(a,I=>{CUA(g(r))?I(c):I(l,!1)}),iA(o,s)}),iA(t,n),pt()}var oYA={selecting:!1,selectionAnchor:void 0,selectionAnchorType:void 0,selectionFocus:void 0,dragging:!1};function i_(t){var{json:e,selection:A,deltaY:i,items:n}=t;if(!A)return{operations:void 0,updatedSelection:void 0,offset:0};var o=i<0?function(l){for(var{json:I,items:C,selection:d,deltaY:B}=l,E=X0(I,d),h=C.findIndex(_=>di(_.path,E)),u=()=>{var _;return(_=C[D-1])===null||_===void 0?void 0:_.height},D=h,L=0;u()!==void 0&&Math.abs(B)>L+u()/2;)L+=u(),D-=1;var R=C[D].path,w=D-h;return D!==h&&C[D]!==void 0?{beforePath:R,offset:w}:void 0}({json:e,selection:A,deltaY:i,items:n}):function(l){for(var I,{json:C,items:d,selection:B,deltaY:E}=l,h=P1(C,B),u=d.findIndex(H=>di(H.path,h)),D=0,L=u,R=()=>{var H;return(H=d[L+1])===null||H===void 0?void 0:H.height};R()!==void 0&&Math.abs(E)>D+R()/2;)D+=R(),L+=1;var w=Fi(h),_=Ke(C,w),K=Array.isArray(_)?L:L+1,z=(I=d[K])===null||I===void 0?void 0:I.path,U=L-u;return z?{beforePath:z,offset:U}:{append:!0,offset:U}}({json:e,selection:A,deltaY:i,items:n});if(!o||o.offset===0)return{operations:void 0,updatedSelection:void 0,offset:0};var r=function(l,I,C){if(!I)return[];var d="beforePath"in C?C.beforePath:void 0,B="append"in C?C.append:void 0,E=Fi(et(I)),h=Ke(l,E);if(!(B||d&&Pg(d,E)&&d.length>E.length))return[];var u=X0(l,I),D=P1(l,I),L=hi(u),R=hi(D),w=d?d[E.length]:void 0;if(!xo(h)){if(wo(h)){var _=ts(L),K=ts(R),z=w!==void 0?ts(w):h.length;return YS(K-_+1,z<_?gA=>({op:"move",from:Ct(E.concat(String(_+gA))),path:Ct(E.concat(String(z+gA)))}):()=>({op:"move",from:Ct(E.concat(String(_))),path:Ct(E.concat(String(z)))}))}throw new Error("Cannot create move operations: parent must be an Object or Array")}var U=Object.keys(h),H=U.indexOf(L),q=U.indexOf(R),j=B?U.length:w!==void 0?U.indexOf(w):-1;return H!==-1&&q!==-1&&j!==-1?j>H?[...U.slice(H,q+1),...U.slice(j,U.length)].map(gA=>X1(E,gA)):[...U.slice(j,H),...U.slice(q+1,U.length)].map(gA=>X1(E,gA)):[]}(e,A,o),s=Fi(X0(e,A)),a=Ke(e,s);if(Array.isArray(a)){var c=function(l){var I,C,{items:d,json:B,selection:E,offset:h}=l,u=X0(B,E),D=P1(B,E),L=d.findIndex(K=>di(K.path,u)),R=d.findIndex(K=>di(K.path,D)),w=(I=d[L+h])===null||I===void 0?void 0:I.path,_=(C=d[R+h])===null||C===void 0?void 0:C.path;return _s(w,_)}({items:n,json:e,selection:A,offset:o.offset});return{operations:r,updatedSelection:c,offset:o.offset}}return{operations:r,updatedSelection:void 0,offset:o.offset}}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
button.jse-validation-error.svelte-1a8aobl {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  padding: 0;
  margin: 0;
  vertical-align: top;
  display: inline-flex;
  color: var(--jse-error-color, #ee5341);
}

button.jse-validation-info.svelte-1a8aobl {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  padding: 0;
  margin: 0;
  vertical-align: top;
  display: inline-flex;
  color: var(--jse-info-color, #4f91ff);
}

button.jse-validation-warning.svelte-1a8aobl {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  padding: 0;
  margin: 0;
  vertical-align: top;
  display: inline-flex;
  color: var(--jse-warning-color, #fdc539);
}`);var rYA=wA('<button type="button"><!></button>');function gQ(t,e){mt(e,!1);var A=$(),i=$1("absolute-popup"),n=b(e,"validationError",8),o=b(e,"onExpand",8);fA(()=>k(n()),()=>{y(A,IUA(n())&&n().isChildError?"Contains invalid data":n().message)}),Qn(),Zt();var r=rYA();ji(W(r),{get data(){return g1}}),es(()=>ce("click",r,function(){for(var s,a=arguments.length,c=new Array(a),l=0;l<a;l++)c[l]=arguments[l];(s=o())===null||s===void 0||s.apply(this,c)})),Us(r,(s,a)=>hQ?.(s,a),()=>fe({text:g(A)},i)),De(()=>{var s;return Vt(r,1,"jse-validation-".concat((k(n()),(s=nA(()=>n().severity))!==null&&s!==void 0?s:"")),"svelte-1a8aobl")}),iA(t,r),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-expand.svelte-oawf7x {
  width: var(--jse-indent-size, calc(1em + 4px));
  padding: 0;
  margin: 0;
  border: none;
  cursor: pointer;
  background: transparent;
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
  font-size: var(--jse-font-size-mono, 14px);
  height: var(--jse-line-height, calc(1em + 4px));
}
.jse-expand.svelte-oawf7x:hover {
  opacity: 0.8;
}

.jse-meta.svelte-oawf7x,
.jse-separator.svelte-oawf7x,
.jse-index.svelte-oawf7x,
.jse-bracket.svelte-oawf7x {
  vertical-align: top;
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
}

.jse-index.svelte-oawf7x {
  padding: 0 calc(0.5 * var(--jse-padding, 10px));
}

.jse-bracket.svelte-oawf7x {
  padding: 0 2px;
}
.jse-bracket.jse-expanded.svelte-oawf7x {
  padding-right: var(--jse-padding, 10px);
}

.jse-identifier.svelte-oawf7x {
  vertical-align: top;
  position: relative;
}

.jse-json-node.svelte-oawf7x {
  position: relative;
  color: var(--jse-text-color, #4d4d4d);
}
.jse-json-node.jse-root.svelte-oawf7x {
  min-height: 100%;
  padding-bottom: 2px;
  box-sizing: border-box;
}
.jse-json-node.jse-root.svelte-oawf7x > .jse-contents-outer:where(.svelte-oawf7x) > .jse-contents:where(.svelte-oawf7x) {
  padding-left: 0;
}
.jse-json-node.svelte-oawf7x .jse-props:where(.svelte-oawf7x),
.jse-json-node.svelte-oawf7x .jse-items:where(.svelte-oawf7x) {
  position: relative;
}
.jse-json-node.svelte-oawf7x .jse-header-outer:where(.svelte-oawf7x),
.jse-json-node.svelte-oawf7x .jse-footer-outer:where(.svelte-oawf7x) {
  display: flex;
  margin-left: calc(var(--level) * var(--jse-indent-size, calc(1em + 4px)));
}
.jse-json-node.svelte-oawf7x .jse-header:where(.svelte-oawf7x) {
  position: relative;
}
.jse-json-node.svelte-oawf7x .jse-header:where(.svelte-oawf7x) .jse-meta:where(.svelte-oawf7x) > .jse-meta-inner:where(.svelte-oawf7x) {
  display: flex;
  justify-content: center;
}
.jse-json-node.svelte-oawf7x .jse-contents-outer:where(.svelte-oawf7x) {
  display: flex;
  margin-left: calc(var(--level) * var(--jse-indent-size, calc(1em + 4px)));
}
.jse-json-node.svelte-oawf7x .jse-header:where(.svelte-oawf7x),
.jse-json-node.svelte-oawf7x .jse-contents:where(.svelte-oawf7x) {
  display: flex;
  flex-direction: row;
  align-items: flex-start;
}
.jse-json-node.svelte-oawf7x .jse-contents:where(.svelte-oawf7x) {
  padding-left: var(--jse-indent-size, calc(1em + 4px));
  cursor: var(--jse-contents-cursor, pointer);
}
.jse-json-node.svelte-oawf7x .jse-contents:where(.svelte-oawf7x) .jse-value-outer:where(.svelte-oawf7x) {
  display: inline-flex;
}
.jse-json-node.svelte-oawf7x .jse-footer:where(.svelte-oawf7x) {
  display: inline-flex;
  padding-left: calc(var(--jse-indent-size, calc(1em + 4px)) + 5px);
}
.jse-json-node.svelte-oawf7x .jse-header:where(.svelte-oawf7x),
.jse-json-node.svelte-oawf7x .jse-contents:where(.svelte-oawf7x),
.jse-json-node.svelte-oawf7x .jse-footer:where(.svelte-oawf7x) {
  background: var(--jse-contents-background-color, transparent);
}
.jse-json-node.svelte-oawf7x .jse-insert-selection-area:where(.svelte-oawf7x) {
  padding: 0 calc(0.5 * var(--jse-padding, 10px));
  flex: 1;
}
.jse-json-node.svelte-oawf7x .jse-insert-selection-area.jse-inside:where(.svelte-oawf7x) {
  display: inline-flex;
  align-items: center;
}
.jse-json-node.svelte-oawf7x .jse-insert-selection-area.jse-after:where(.svelte-oawf7x) {
  display: flex;
  align-items: flex-end;
}
.jse-json-node.svelte-oawf7x .jse-context-menu-pointer-anchor:where(.svelte-oawf7x) {
  position: relative;
}
.jse-json-node.svelte-oawf7x .jse-insert-area:where(.svelte-oawf7x) {
  display: flex;
  position: relative;
  z-index: 1;
  margin-left: calc(var(--level) * var(--jse-indent-size, calc(1em + 4px)));
  max-width: 250px;
  min-width: 100px;
  height: 0;
  margin-right: calc(0.5 * var(--jse-padding, 10px));
  outline: 1px solid;
}
.jse-json-node.svelte-oawf7x .jse-insert-area.jse-hovered:where(.svelte-oawf7x) {
  outline-color: var(--jse-context-menu-pointer-hover-background, #b2b2b2);
}
.jse-json-node.svelte-oawf7x .jse-key-outer:where(.svelte-oawf7x) {
  position: relative;
}
.jse-json-node.svelte-oawf7x .jse-key-outer:where(.svelte-oawf7x):hover,
.jse-json-node.svelte-oawf7x .jse-value-outer:where(.svelte-oawf7x):hover,
.jse-json-node.svelte-oawf7x .jse-meta:where(.svelte-oawf7x):hover,
.jse-json-node.svelte-oawf7x .jse-footer:where(.svelte-oawf7x):hover {
  background: var(--jse-hover-background-color, rgba(0, 0, 0, 0.06));
  cursor: var(--jse-contents-cursor, pointer);
}
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-header,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-contents,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-header,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-contents,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-footer {
  background: var(--jse-hover-background-color, rgba(0, 0, 0, 0.06));
  cursor: var(--jse-contents-cursor, pointer);
}
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-value-outer .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-value-outer .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-meta .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-meta .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-header .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-header .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-contents .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-items .jse-contents .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-header .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-header .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-contents .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-props .jse-contents .jse-meta,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-footer .jse-value-outer,
.jse-json-node.jse-hovered.svelte-oawf7x:not(.jse-selected):not(.jse-selected-value) .jse-footer .jse-meta {
  background: none;
}
.jse-json-node.jse-selected.svelte-oawf7x .jse-header:where(.svelte-oawf7x),
.jse-json-node.jse-selected.svelte-oawf7x .jse-contents:where(.svelte-oawf7x),
.jse-json-node.jse-selected.svelte-oawf7x .jse-footer:where(.svelte-oawf7x) {
  background: var(--jse-selection-background-color, #d3d3d3);
  cursor: var(--jse-contents-selected-cursor, grab);
}
.jse-json-node.jse-selected.svelte-oawf7x .jse-key-outer:where(.svelte-oawf7x):hover,
.jse-json-node.jse-selected.svelte-oawf7x .jse-value-outer:where(.svelte-oawf7x):hover,
.jse-json-node.jse-selected.svelte-oawf7x .jse-meta:where(.svelte-oawf7x):hover,
.jse-json-node.jse-selected.svelte-oawf7x .jse-footer:where(.svelte-oawf7x):hover {
  background: inherit;
  cursor: inherit;
}
.jse-json-node.svelte-oawf7x .jse-key-outer.jse-selected-key:where(.svelte-oawf7x) {
  background: var(--jse-selection-background-color, #d3d3d3);
  cursor: var(--jse-contents-selected-cursor, grab);
}
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-value-outer,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-meta,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-items .jse-header,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-items .jse-contents,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-props .jse-header,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-props .jse-contents,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-footer {
  background: var(--jse-selection-background-color, #d3d3d3);
  cursor: var(--jse-contents-selected-cursor, grab);
}
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-value-outer .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-meta .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-items .jse-header .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-items .jse-contents .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-props .jse-header .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-props .jse-contents .jse-key-outer:hover,
.jse-json-node.jse-selected-value.svelte-oawf7x .jse-footer .jse-key-outer:hover {
  background: inherit;
  cursor: inherit;
}
.jse-json-node.jse-readonly.svelte-oawf7x {
  --jse-contents-selected-cursor: pointer;
}
.jse-json-node.svelte-oawf7x .jse-insert-area.jse-selected:where(.svelte-oawf7x) {
  outline-color: var(--jse-context-menu-pointer-background, var(--jse-context-menu-background, #656565));
}`);var Io=Gy(()=>oYA),sYA=wA('<div class="jse-separator svelte-oawf7x">:</div>'),aYA=wA('<div class="jse-bracket svelte-oawf7x">[</div> <!> &nbsp;',1),cYA=wA('<div class="jse-bracket svelte-oawf7x">[</div> <!> <div class="jse-bracket svelte-oawf7x">]</div>',1),lYA=wA('<div class="jse-context-menu-pointer-anchor svelte-oawf7x"><!></div>'),gYA=wA('<div role="none" class="jse-insert-selection-area jse-inside svelte-oawf7x" data-type="insert-selection-area-inside"></div>'),IYA=wA('<div role="none" class="jse-insert-selection-area jse-after svelte-oawf7x" data-type="insert-selection-area-after"></div>'),CYA=wA('<div data-type="insert-selection-area-inside"><!></div>'),dYA=wA('<div slot="identifier" class="jse-identifier svelte-oawf7x"><div class="jse-index svelte-oawf7x"> </div></div>'),BYA=wA("<!> <!>",1),EYA=wA('<div role="none" class="jse-insert-selection-area jse-after svelte-oawf7x" data-type="insert-selection-area-after"></div>'),QYA=wA('<div class="jse-items svelte-oawf7x"><!> <!></div> <div class="jse-footer-outer svelte-oawf7x"><div data-type="selectable-value" class="jse-footer svelte-oawf7x"><span class="jse-bracket svelte-oawf7x">]</span></div> <!></div>',1),hYA=wA('<div class="jse-header-outer svelte-oawf7x"><div class="jse-header svelte-oawf7x"><button type="button" class="jse-expand svelte-oawf7x" title="Expand or collapse this array (Ctrl+Click to expand/collapse recursively)"><!></button> <!> <!> <div class="jse-meta svelte-oawf7x"><div class="jse-meta-inner svelte-oawf7x" data-type="selectable-value"><!></div></div> <!></div> <!> <!></div> <!>',1),uYA=wA('<div class="jse-separator svelte-oawf7x">:</div>'),fYA=wA('<div class="jse-bracket jse-expanded svelte-oawf7x">&lbrace;</div>'),mYA=wA('<div class="jse-bracket svelte-oawf7x">&lbrace;</div> <!> <div class="jse-bracket svelte-oawf7x">&rbrace;</div>',1),pYA=wA('<div class="jse-context-menu-pointer-anchor svelte-oawf7x"><!></div>'),wYA=wA('<div role="none" class="jse-insert-selection-area jse-inside svelte-oawf7x" data-type="insert-selection-area-inside"></div>'),DYA=wA('<div role="none" class="jse-insert-selection-area jse-after svelte-oawf7x" data-type="insert-selection-area-after"></div>'),yYA=wA('<div data-type="insert-selection-area-inside"><!></div>'),vYA=wA('<div slot="identifier"><!></div>'),bYA=wA('<div role="none" class="jse-insert-selection-area jse-after svelte-oawf7x" data-type="insert-selection-area-after"></div>'),MYA=wA('<div class="jse-props svelte-oawf7x"><!> <!></div> <div class="jse-footer-outer svelte-oawf7x"><div data-type="selectable-value" class="jse-footer svelte-oawf7x"><div class="jse-bracket svelte-oawf7x">&rbrace;</div></div> <!></div>',1),kYA=wA('<div class="jse-header-outer svelte-oawf7x"><div class="jse-header svelte-oawf7x"><button type="button" class="jse-expand svelte-oawf7x" title="Expand or collapse this object (Ctrl+Click to expand/collapse recursively)"><!></button> <!> <!> <div class="jse-meta svelte-oawf7x" data-type="selectable-value"><div class="jse-meta-inner svelte-oawf7x"><!></div></div> <!></div> <!> <!></div> <!>',1),SYA=wA('<div class="jse-separator svelte-oawf7x">:</div>'),RYA=wA('<div class="jse-context-menu-pointer-anchor svelte-oawf7x"><!></div>'),xYA=wA('<div role="none" class="jse-insert-selection-area jse-after svelte-oawf7x" data-type="insert-selection-area-after"></div>'),LYA=wA('<div class="jse-contents-outer svelte-oawf7x"><div class="jse-contents svelte-oawf7x"><!> <!> <div class="jse-value-outer svelte-oawf7x"><!></div> <!></div> <!> <!></div>'),FYA=wA('<div data-type="insert-selection-area-after"><!></div>'),NYA=wA('<div role="treeitem" tabindex="-1"><!> <!></div>');function x_(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=b(e,"pointer",9),o=b(e,"value",9),r=b(e,"state",9),s=b(e,"validationErrors",9),a=b(e,"searchResults",9),c=b(e,"selection",9),l=b(e,"context",9),I=b(e,"onDragSelectionStart",9),C=Sr("jsoneditor:JSONNode"),d=$(void 0,!0),B=void 0,E=$(void 0,!0),h=$(void 0,!0),u=$(void 0,!0),D=$(void 0,!0),L=$(void 0,!0),R=$(void 0,!0),w=$(void 0,!0);function _(uA){uA.stopPropagation();var eA=$_(uA);l().onExpand(g(h),!g(u),eA)}function K(){l().onExpand(g(h),!0)}function z(uA,eA){var UA=gf(g(h),Object.keys(o()),uA,eA);return l().onPatch(UA),hi(ks(UA[0].path))}function U(uA){l().onDrag(uA)}function H(uA){Io().selecting&&(Io(Io().selecting=!1),uA.stopPropagation()),l().onDragEnd(),document.removeEventListener("mousemove",U,!0),document.removeEventListener("mouseup",H)}function q(){var uA;return((uA=l().findElement([]))===null||uA===void 0||(uA=uA.getBoundingClientRect())===null||uA===void 0?void 0:uA.top)||0}function j(uA,eA){var UA=q()-uA.initialContentTop;return eA.clientY-uA.initialClientY-UA}function gA(uA){if(!l().readOnly&&c()){var eA=Fi(et(c()));if(di(g(h),eA)){var UA=function(mA,sA){var xt=[];function tt(Y){var P=g(h).concat(Y),X=l().findElement(P);X!==void 0&&xt.push({path:P,height:X.clientHeight})}if(Array.isArray(o())){var de=l().getJson();if(de===void 0)return;var Dt=X0(de,mA),_e=P1(de,mA),Le=parseInt(hi(Dt),10),bt=parseInt(hi(_e),10),Re=sA.find(Y=>Le>=Y.start&&bt<=Y.end);if(!Re)return;var{start:$t,end:x}=Re;FsA($t,Math.min(o().length,x),Y=>tt(String(Y)))}else Object.keys(o()).forEach(tt);return xt}(c(),g(L)||aQ);if(C("dragSelectionStart",{selection:c(),items:UA}),UA){var aA=l().getJson();if(aA!==void 0){var le=X0(aA,c()),SA=UA.findIndex(mA=>di(mA.path,le)),{offset:Ue}=i_({json:aA,selection:l().getSelection(),deltaY:0,items:UA});y(E,{initialTarget:uA.target,initialClientY:uA.clientY,initialContentTop:q(),selectionStartIndex:SA,selectionItemsCount:W1(aA,c()).length,items:UA,offset:Ue,didMoveItems:!1}),Io(Io().dragging=!0),document.addEventListener("mousemove",QA,!0),document.addEventListener("mouseup",BA)}}else C("Cannot drag the current selection (probably spread over multiple sections)")}else I()(uA)}}function QA(uA){if(g(E)){var eA=l().getJson();if(eA===void 0)return;var UA=j(g(E),uA),{offset:aA}=i_({json:eA,selection:l().getSelection(),deltaY:UA,items:g(E).items});aA!==g(E).offset&&(C("drag selection",aA,UA),y(E,fe(fe({},g(E)),{},{offset:aA,didMoveItems:!0})))}}function BA(uA){if(g(E)){var eA=l().getJson();if(eA===void 0)return;var UA=j(g(E),uA),{operations:aA,updatedSelection:le}=i_({json:eA,selection:l().getSelection(),deltaY:UA,items:g(E).items});if(aA)l().onPatch(aA,(mA,sA)=>({state:sA,selection:le??c()}));else if(uA.target===g(E).initialTarget&&!g(E).didMoveItems){var SA=zN(uA.target),Ue=PsA(uA.target);Ue&&l().onSelect(frA(SA,Ue))}y(E,void 0),Io(Io().dragging=!1),document.removeEventListener("mousemove",QA,!0),document.removeEventListener("mouseup",BA)}}function lA(uA){uA.shiftKey||(uA.stopPropagation(),uA.preventDefault(),l().onSelect(r2(g(h))))}function vA(uA){uA.shiftKey||(uA.stopPropagation(),uA.preventDefault(),l().onSelect(t2(g(h))))}function tA(uA){l().onSelect(r2(g(h))),io(),l().onContextMenu(uA)}function cA(uA){l().onSelect(t2(g(h))),io(),l().onContextMenu(uA)}fA(()=>k(n()),()=>{y(h,ks(n()))}),fA(()=>k(n()),()=>{y(A,encodeURIComponent(n()))}),fA(()=>k(r()),()=>{y(u,!!ZC(r())&&r().expanded)}),fA(()=>(k(o()),k(r())),()=>{y(D,zg(o(),r(),[]))}),fA(()=>k(r()),()=>{y(L,Mr(r())?r().visibleSections:void 0)}),fA(()=>k(s()),()=>{var uA;y(R,(uA=s())===null||uA===void 0?void 0:uA.validationError)}),fA(()=>(k(l()),k(c()),g(h)),()=>{y(w,V3(l().getJson(),c(),g(h)))}),fA(()=>g(h),()=>{y(i,g(h).length===0)}),Qn(),Zt(!0);var pA,VA,oe=NYA(),KA=W(oe),CA=uA=>{var eA=hYA(),UA=vt(eA),aA=W(UA),le=W(aA),SA=W(le),Ue=kA=>{ji(kA,{get data(){return mg}})},mA=kA=>{ji(kA,{get data(){return sE}})};LA(SA,kA=>{g(u)?kA(Ue):kA(mA,!1)});var sA=IA(le,2);jo(sA,e,"identifier",{},null);var xt=IA(sA,2),tt=kA=>{iA(kA,sYA())};LA(xt,kA=>{g(i)||kA(tt)});var de=IA(xt,2),Dt=W(de),_e=W(Dt),Le=kA=>{var DA=aYA();ly(IA(vt(DA),2),{children:(gt,Ve)=>{var ZA=Tr();De(()=>{var rt,Ei;return wt(ZA,"".concat((k(o()),(rt=nA(()=>o().length))!==null&&rt!==void 0?rt:""),`
                `).concat((k(o()),(Ei=nA(()=>o().length===1?"item":"items"))!==null&&Ei!==void 0?Ei:"")))}),iA(gt,ZA)},$$slots:{default:!0}}),iA(kA,DA)},bt=kA=>{var DA=cYA();ly(IA(vt(DA),2),{onclick:K,children:(gt,Ve)=>{var ZA=Tr();De(()=>{var rt,Ei;return wt(ZA,"".concat((k(o()),(rt=nA(()=>o().length))!==null&&rt!==void 0?rt:""),`
                `).concat((k(o()),(Ei=nA(()=>o().length===1?"item":"items"))!==null&&Ei!==void 0?Ei:"")))}),iA(gt,ZA)},$$slots:{default:!0}}),iA(kA,DA)};LA(_e,kA=>{g(u)?kA(Le):kA(bt,!1)});var Re=IA(de,2),$t=kA=>{var DA=lYA();_1(W(DA),{get root(){return g(i)},selected:!0,get onContextMenu(){return k(l()),nA(()=>l().onContextMenu)}}),iA(kA,DA)};LA(Re,kA=>{k(l()),g(w),k(c()),k(an),k(Kn),k(br),k(di),k(et),g(h),nA(()=>!l().readOnly&&g(w)&&c()&&(an(c())||Kn(c()))&&!br(c())&&di(et(c()),g(h)))&&kA($t)});var x=IA(aA,2),Y=kA=>{gQ(kA,{get validationError(){return g(R)},onExpand:K})};LA(x,kA=>{g(R),g(u),nA(()=>g(R)&&(!g(u)||!g(R).isChildError))&&kA(Y)});var P=IA(x,2),X=kA=>{var DA=gYA();ce("click",DA,lA),iA(kA,DA)},bA=kA=>{var DA=IYA();ce("click",DA,vA),iA(kA,DA)};LA(P,kA=>{g(u)?kA(X):kA(bA,!1)});var Be=IA(UA,2),Ee=kA=>{var DA=QYA(),gt=vt(DA),Ve=W(gt),ZA=qi=>{var xe,nn,mi=CYA(),Ot=W(mi),Lt=qA(()=>(g(w),k(fr),k(c()),nA(()=>g(w)&&fr(c()))));_1(Ot,{insert:!0,get selected(){return g(Lt)},onContextMenu:tA}),De((ii,_i)=>{xe=Vt(mi,1,"jse-insert-area jse-inside svelte-oawf7x",null,xe,ii),En(mi,"title",jN),nn=Ll(mi,"",nn,_i)},[()=>({"jse-hovered":g(d)===KC,"jse-selected":g(w)&&fr(c())}),()=>({"--level":(g(h),nA(()=>g(h).length+1))})],qA),iA(qi,mi)};LA(Ve,qi=>{k(l()),g(d),k(KC),g(w),k(fr),k(c()),nA(()=>!l().readOnly&&(g(d)===KC||g(w)&&fr(c())))&&qi(ZA)}),qo(IA(Ve,2),1,()=>g(L)||aQ,or,(qi,xe,nn)=>{var mi=BYA(),Ot=vt(mi);qo(Ot,1,()=>(k(o()),g(xe),g(E),nA(()=>function(_i,Tt,M){var We=Tt.start,ni=Math.min(Tt.end,_i.length),pi=d5(We,ni);return M&&M.offset!==0?$oA(pi,M.selectionStartIndex,M.selectionItemsCount,M.offset).map((dn,mn)=>({index:dn,gutterIndex:mn})):pi.map(dn=>({index:dn,gutterIndex:dn}))}(o(),g(xe),g(E)))),_i=>_i.index,(_i,Tt)=>{var M=_o(),We=qA(()=>(k(Mr),k(s()),g(Tt),nA(()=>Mr(s())?s().items[g(Tt).index]:void 0))),ni=qA(()=>(k(XD),k(l()),k(c()),g(h),g(Tt),nA(()=>XD(l().getJson(),c(),g(h).concat(String(g(Tt).index)))))),pi=vt(M),dn=qA(()=>(k(qu),k(n()),g(Tt),nA(()=>qu(n(),g(Tt).index)))),mn=qA(()=>(k(Mr),k(r()),g(Tt),nA(()=>Mr(r())?r().items[g(Tt).index]:void 0))),Uo=qA(()=>(k(Mr),k(a()),g(Tt),nA(()=>Mr(a())?a().items[g(Tt).index]:void 0)));x_(pi,{get value(){return k(o()),g(Tt),nA(()=>o()[g(Tt).index])},get pointer(){return g(dn)},get state(){return g(mn)},get validationErrors(){return g(We)},get searchResults(){return g(Uo)},get selection(){return g(ni)},get context(){return l()},onDragSelectionStart:gA,$$slots:{identifier:(kn,Wn)=>{var Vo=dYA(),vo=W(Vo),bo=W(vo);De(()=>wt(bo,(g(Tt),nA(()=>g(Tt).gutterIndex)))),iA(kn,Vo)}}}),iA(_i,M)});var Lt=IA(Ot,2),ii=_i=>{var Tt=qA(()=>g(L)||aQ);AYA(_i,{get visibleSections(){return g(Tt)},sectionIndex:nn,get total(){return k(o()),nA(()=>o().length)},get path(){return g(h)},get onExpandSection(){return k(l()),nA(()=>l().onExpandSection)},get selection(){return c()},get context(){return l()}})};LA(Lt,_i=>{g(xe),k(o()),nA(()=>g(xe).end<o().length)&&_i(ii)}),iA(qi,mi)});var rt=IA(gt,2),Ei=IA(W(rt),2),tn=qi=>{var xe=EYA();ce("click",xe,vA),iA(qi,xe)};LA(Ei,qi=>{g(i)||qi(tn)}),iA(kA,DA)};LA(Be,kA=>{g(u)&&kA(Ee)}),ce("click",le,_),iA(uA,eA)},TA=(uA,eA)=>{var UA=le=>{var SA=kYA(),Ue=vt(SA),mA=W(Ue),sA=W(mA),xt=W(sA),tt=ZA=>{ji(ZA,{get data(){return mg}})},de=ZA=>{ji(ZA,{get data(){return sE}})};LA(xt,ZA=>{g(u)?ZA(tt):ZA(de,!1)});var Dt=IA(sA,2);jo(Dt,e,"identifier",{},null);var _e=IA(Dt,2),Le=ZA=>{iA(ZA,uYA())};LA(_e,ZA=>{g(i)||ZA(Le)});var bt=IA(_e,2),Re=W(bt),$t=W(Re),x=ZA=>{iA(ZA,fYA())},Y=ZA=>{var rt=mYA();ly(IA(vt(rt),2),{onclick:K,children:(Ei,tn)=>{var qi=Tr();De((xe,nn)=>wt(qi,"".concat(xe??"",`
                `).concat(nn??"")),[()=>(k(o()),nA(()=>Object.keys(o()).length)),()=>(k(o()),nA(()=>Object.keys(o()).length===1?"prop":"props"))],qA),iA(Ei,qi)},$$slots:{default:!0}}),iA(ZA,rt)};LA($t,ZA=>{g(u)?ZA(x):ZA(Y,!1)});var P=IA(bt,2),X=ZA=>{var rt=pYA();_1(W(rt),{get root(){return g(i)},selected:!0,get onContextMenu(){return k(l()),nA(()=>l().onContextMenu)}}),iA(ZA,rt)};LA(P,ZA=>{k(l()),g(w),k(c()),k(an),k(Kn),k(br),k(di),k(et),g(h),nA(()=>!l().readOnly&&g(w)&&c()&&(an(c())||Kn(c()))&&!br(c())&&di(et(c()),g(h)))&&ZA(X)});var bA=IA(mA,2),Be=ZA=>{gQ(ZA,{get validationError(){return g(R)},onExpand:K})};LA(bA,ZA=>{g(R),g(u),nA(()=>g(R)&&(!g(u)||!g(R).isChildError))&&ZA(Be)});var Ee=IA(bA,2),kA=ZA=>{var rt=wYA();ce("click",rt,lA),iA(ZA,rt)},DA=(ZA,rt)=>{var Ei=tn=>{var qi=DYA();ce("click",qi,vA),iA(tn,qi)};LA(ZA,tn=>{g(i)||tn(Ei)},rt)};LA(Ee,ZA=>{g(u)?ZA(kA):ZA(DA,!1)});var gt=IA(Ue,2),Ve=ZA=>{var rt=MYA(),Ei=vt(rt),tn=W(Ei),qi=Ot=>{var Lt,ii,_i=yYA(),Tt=W(_i),M=qA(()=>(g(w),k(fr),k(c()),nA(()=>g(w)&&fr(c()))));_1(Tt,{insert:!0,get selected(){return g(M)},onContextMenu:tA}),De((We,ni)=>{Lt=Vt(_i,1,"jse-insert-area jse-inside svelte-oawf7x",null,Lt,We),En(_i,"title",jN),ii=Ll(_i,"",ii,ni)},[()=>({"jse-hovered":g(d)===KC,"jse-selected":g(w)&&fr(c())}),()=>({"--level":(g(h),nA(()=>g(h).length+1))})],qA),iA(Ot,_i)};LA(tn,Ot=>{k(l()),g(d),k(KC),g(w),k(fr),k(c()),nA(()=>!l().readOnly&&(g(d)===KC||g(w)&&fr(c())))&&Ot(qi)}),qo(IA(tn,2),1,()=>(k(o()),g(E),nA(()=>function(Ot,Lt){var ii=Object.keys(Ot);return Lt&&Lt.offset!==0?$oA(ii,Lt.selectionStartIndex,Lt.selectionItemsCount,Lt.offset):ii}(o(),g(E)))),or,(Ot,Lt)=>{var ii=_o(),_i=qA(()=>(k(qu),k(n()),g(Lt),nA(()=>qu(n(),g(Lt))))),Tt=qA(()=>(k(_a),k(a()),g(Lt),nA(()=>_a(a())?a().properties[g(Lt)]:void 0))),M=qA(()=>(k(_a),k(s()),g(Lt),nA(()=>_a(s())?s().properties[g(Lt)]:void 0))),We=qA(()=>(g(h),g(Lt),nA(()=>g(h).concat(g(Lt))))),ni=qA(()=>(k(XD),k(l()),k(c()),k(g(We)),nA(()=>XD(l().getJson(),c(),g(We))))),pi=vt(ii),dn=qA(()=>(k(_a),k(r()),g(Lt),nA(()=>_a(r())?r().properties[g(Lt)]:void 0)));x_(pi,{get value(){return k(o()),g(Lt),nA(()=>o()[g(Lt)])},get pointer(){return g(_i)},get state(){return g(dn)},get validationErrors(){return g(M)},get searchResults(){return g(Tt)},get selection(){return g(ni)},get context(){return l()},onDragSelectionStart:gA,$$slots:{identifier:(mn,Uo)=>{var kn,Wn=vYA(),Vo=W(Wn),vo=qA(()=>(k(brA),k(g(Tt)),nA(()=>brA(g(Tt)))));(function(bo,Yn){mt(Yn,!1);var Mo=$(void 0,!0),ne=$(void 0,!0),wi=b(Yn,"pointer",9),MA=b(Yn,"key",9),me=b(Yn,"selection",9),nt=b(Yn,"searchResultItems",9),Wt=b(Yn,"onUpdateKey",9),Xe=b(Yn,"context",9),oi=$(void 0,!0);function Di(Ai){g(ne)||Xe().readOnly||(Ai.preventDefault(),Xe().onSelect(cG(g(oi))))}function Ut(Ai,Ki){var dt=Wt()(MA(),Xe().normalization.unescapeValue(Ai)),EA=Fi(g(oi)).concat(dt);Xe().onSelect(Ki===O1.nextInside?Ni(EA):o2(EA)),Ki!==O1.self&&Xe().focus()}function cn(){Xe().onSelect(o2(g(oi))),Xe().focus()}fA(()=>k(wi()),()=>{y(oi,ks(wi()))}),fA(()=>(k(me()),g(oi)),()=>{y(Mo,kr(me())&&di(me().path,g(oi)))}),fA(()=>(g(Mo),k(me())),()=>{y(ne,g(Mo)&&br(me()))}),Qn(),Zt(!0);var ft=iYA(),Qi=vt(ft),ot=Ai=>{var Ki=qA(()=>(k(Xe()),k(MA()),nA(()=>Xe().normalization.escapeValue(MA())))),dt=qA(()=>(k(br),k(me()),nA(()=>br(me())?me().initialValue:void 0)));iaA(Ai,{get value(){return g(Ki)},get initialValue(){return g(dt)},label:"Edit key",shortText:!0,onChange:Ut,onCancel:cn,get onFind(){return k(Xe()),nA(()=>Xe().onFind)}})},Mt=Ai=>{var Ki,dt=tYA(),EA=W(dt),HA=Qt=>{var yi=qA(()=>(k(Xe()),k(MA()),nA(()=>Xe().normalization.escapeValue(MA()))));laA(Qt,{get text(){return g(yi)},get searchResultItems(){return nt()}})},ve=Qt=>{var yi=Tr();De(ri=>wt(yi,ri),[()=>(k(BQ),k(Xe()),k(MA()),nA(()=>BQ(Xe().normalization.escapeValue(MA()))))],qA),iA(Qt,yi)};LA(EA,Qt=>{nt()?Qt(HA):Qt(ve,!1)}),De(Qt=>Ki=Vt(dt,1,"jse-key svelte-2iqnqn",null,Ki,Qt),[()=>({"jse-empty":MA()===""})],qA),ce("dblclick",dt,Di),iA(Ai,dt)};LA(Qi,Ai=>{k(Xe()),g(ne),nA(()=>!Xe().readOnly&&g(ne))?Ai(ot):Ai(Mt,!1)});var on=IA(Qi,2),hn=Ai=>{_1(Ai,{selected:!0,get onContextMenu(){return k(Xe()),nA(()=>Xe().onContextMenu)}})};LA(on,Ai=>{k(Xe()),g(Mo),g(ne),nA(()=>!Xe().readOnly&&g(Mo)&&!g(ne))&&Ai(hn)}),iA(bo,ft),pt()})(Vo,{get pointer(){return g(_i)},get key(){return g(Lt)},get selection(){return g(ni)},get searchResultItems(){return g(vo)},get context(){return l()},onUpdateKey:z}),De(bo=>kn=Vt(Wn,1,"jse-key-outer svelte-oawf7x",null,kn,bo),[()=>({"jse-selected-key":kr(g(ni))&&di(g(ni).path,g(We))})],qA),iA(mn,Wn)}}}),iA(Ot,ii)});var xe=IA(Ei,2),nn=IA(W(xe),2),mi=Ot=>{var Lt=bYA();ce("click",Lt,vA),iA(Ot,Lt)};LA(nn,Ot=>{g(i)||Ot(mi)}),iA(ZA,rt)};LA(gt,ZA=>{g(u)&&ZA(Ve)}),ce("click",sA,_),iA(le,SA)},aA=le=>{var SA=LYA(),Ue=W(SA),mA=W(Ue);jo(mA,e,"identifier",{},null);var sA=IA(mA,2),xt=P=>{iA(P,SYA())};LA(sA,P=>{g(i)||P(xt)});var tt=IA(sA,2),de=W(tt),Dt=qA(()=>g(w)?c():void 0),_e=qA(()=>(k(MrA),k(a()),nA(()=>MrA(a()))));vaA(de,{get path(){return g(h)},get value(){return o()},get enforceString(){return g(D)},get selection(){return g(Dt)},get searchResultItems(){return g(_e)},get context(){return l()}});var Le=IA(tt,2),bt=P=>{var X=RYA();_1(W(X),{get root(){return g(i)},selected:!0,get onContextMenu(){return k(l()),nA(()=>l().onContextMenu)}}),iA(P,X)};LA(Le,P=>{k(l()),g(w),k(c()),k(an),k(Kn),k(br),k(di),k(et),g(h),nA(()=>!l().readOnly&&g(w)&&c()&&(an(c())||Kn(c()))&&!br(c())&&di(et(c()),g(h)))&&P(bt)});var Re=IA(Ue,2),$t=P=>{gQ(P,{get validationError(){return g(R)},onExpand:K})};LA(Re,P=>{g(R)&&P($t)});var x=IA(Re,2),Y=P=>{var X=xYA();ce("click",X,vA),iA(P,X)};LA(x,P=>{g(i)||P(Y)}),iA(le,SA)};LA(uA,le=>{k(Cn),k(o()),nA(()=>Cn(o()))?le(UA):le(aA,!1)},eA)};LA(KA,uA=>{k(o()),nA(()=>Array.isArray(o()))?uA(CA):uA(TA,!1)});var Ze=IA(KA,2),He=uA=>{var eA,UA=FYA(),aA=W(UA),le=qA(()=>(g(w),k(Ua),k(c()),nA(()=>g(w)&&Ua(c()))));_1(aA,{insert:!0,get selected(){return g(le)},onContextMenu:cA}),De(SA=>{eA=Vt(UA,1,"jse-insert-area jse-after svelte-oawf7x",null,eA,SA),En(UA,"title",jN)},[()=>({"jse-hovered":g(d)===ZD,"jse-selected":g(w)&&Ua(c())})],qA),iA(uA,UA)};LA(Ze,uA=>{k(l()),g(d),k(ZD),g(w),k(Ua),k(c()),nA(()=>!l().readOnly&&(g(d)===ZD||g(w)&&Ua(c())))&&uA(He)}),De((uA,eA,UA)=>{pA=Vt(oe,1,uA,"svelte-oawf7x",pA,eA),En(oe,"data-path",g(A)),En(oe,"aria-selected",g(w)),VA=Ll(oe,"",VA,UA)},[()=>Z1((k(Kl),g(u),k(l()),g(h),k(o()),nA(()=>Kl("jse-json-node",{"jse-expanded":g(u)},l().onClassName(g(h),o()))))),()=>({"jse-root":g(i),"jse-selected":g(w)&&Kn(c()),"jse-selected-value":g(w)&&an(c()),"jse-readonly":l().readOnly,"jse-hovered":g(d)===irA}),()=>({"--level":(g(h),nA(()=>g(h).length))})],qA),ce("mousedown",oe,function(uA){if((uA.buttons===1||uA.buttons===2)&&!((eA=uA.target).nodeName==="DIV"&&eA.contentEditable==="true"||uA.buttons===1&&zsA(uA.target,"BUTTON"))){var eA;uA.stopPropagation(),uA.preventDefault(),l().focus(),document.addEventListener("mousemove",U,!0),document.addEventListener("mouseup",H);var UA=zN(uA.target),aA=l().getJson(),le=l().getDocumentState();if(!c()||UA===Ln.after||UA===Ln.inside||c().type!==UA&&c().type!==Ln.multi||!V3(aA,c(),g(h)))if(Io(Io().selecting=!0),Io(Io().selectionAnchor=g(h)),Io(Io().selectionAnchorType=UA),Io(Io().selectionFocus=g(h)),uA.shiftKey){var SA=l().getSelection();SA&&l().onSelect(_s(OC(SA),g(h)))}else if(UA===Ln.multi)if(g(i)&&uA.target.hasAttribute("data-path")){var Ue=hi(XsA(o(),le));l().onSelect(D_(Ue))}else l().onSelect(_s(g(h),g(h)));else aA!==void 0&&l().onSelect(frA(UA,g(h)));else uA.button===0&&I()(uA)}}),ce("mousemove",oe,function(uA){if(Io().selecting){uA.preventDefault(),uA.stopPropagation(),Io().selectionFocus===void 0&&window.getSelection&&window.getSelection().empty();var eA=zN(uA.target);di(g(h),Io().selectionFocus)&&eA===Io().selectionAnchorType||(Io(Io().selectionFocus=g(h)),Io(Io().selectionAnchorType=eA),l().onSelect(_s(Io().selectionAnchor||Io().selectionFocus,Io().selectionFocus)))}}),ce("mouseover",oe,function(uA){Io().selecting||Io().dragging||(uA.stopPropagation(),Y1(uA.target,"data-type","selectable-value")?y(d,irA):Y1(uA.target,"data-type","selectable-key")?y(d,void 0):Y1(uA.target,"data-type","insert-selection-area-inside")?y(d,KC):Y1(uA.target,"data-type","insert-selection-area-after")&&y(d,ZD),clearTimeout(B))}),ce("mouseout",oe,function(uA){uA.stopPropagation(),B=window.setTimeout(()=>y(d,void 0))}),iA(t,oe),pt()}var _YA={prefix:"fas",iconName:"jsoneditor-expand",icon:[512,512,[],"","M 0,448 V 512 h 512 v -64 z M 0,0 V 64 H 512 V 0 Z M 256,96 128,224 h 256 z M 256,416 384,288 H 128 Z"]},GYA={prefix:"fas",iconName:"jsoneditor-collapse",icon:[512,512,[],"","m 0,224 v 64 h 512 v -64 z M 256,192 384,64 H 128 Z M 256,320 128,448 h 256 z"]},YrA={prefix:"fas",iconName:"jsoneditor-format",icon:[512,512,[],"","M 0,32 v 64 h 416 v -64 z M 160,160 v 64 h 352 v -64 z M 160,288 v 64 h 288 v -64 z M 0,416 v 64 h 320 v -64 z"]},UYA={prefix:"fas",iconName:"jsoneditor-compact",icon:[512,512,[],"","M 0,32 v 64 h 512 v -64 z M 0,160 v 64 h 512 v -64 z M 0,288 v 64 h 352 v -64 z"]};function KYA(t,e){t.stopPropagation(),e.onCreateObject()}function YYA(t,e){t.stopPropagation(),e.onCreateArray()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-welcome.svelte-1eamlhk {
  flex: 1;
  overflow: auto;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  display: flex;
  flex-direction: column;
  align-items: center;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-welcome.svelte-1eamlhk:last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-welcome.svelte-1eamlhk .jse-space.jse-before:where(.svelte-1eamlhk) {
  flex: 1;
}
.jse-welcome.svelte-1eamlhk .jse-space.jse-after:where(.svelte-1eamlhk) {
  flex: 2;
}
.jse-welcome.svelte-1eamlhk .jse-contents:where(.svelte-1eamlhk) {
  display: flex;
  flex-direction: column;
  max-width: 300px;
  margin: 2em var(--jse-padding, 10px);
  gap: var(--jse-padding, 10px);
}
.jse-welcome.svelte-1eamlhk .jse-contents:where(.svelte-1eamlhk) .jse-welcome-info:where(.svelte-1eamlhk) {
  color: var(--jse-panel-color-readonly, #b2b2b2);
}
.jse-welcome.svelte-1eamlhk .jse-contents:where(.svelte-1eamlhk) button:where(.svelte-1eamlhk) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-welcome.svelte-1eamlhk .jse-contents:where(.svelte-1eamlhk) button:where(.svelte-1eamlhk):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-welcome.svelte-1eamlhk .jse-contents:where(.svelte-1eamlhk) button:where(.svelte-1eamlhk):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}`);var JYA=(t,e)=>e.onClick(),TYA=wA('<div class="jse-welcome-info svelte-1eamlhk">You can paste clipboard data using <b>Ctrl+V</b>, or use the following options:</div> <button class="svelte-1eamlhk">Create object</button> <button class="svelte-1eamlhk">Create array</button>',1),HYA=wA('<div class="jse-welcome svelte-1eamlhk" role="none"><div class="jse-space jse-before svelte-1eamlhk"></div> <div class="jse-contents svelte-1eamlhk"><div class="jse-welcome-title">Empty document</div> <!></div> <div class="jse-space jse-after svelte-1eamlhk"></div></div>');function L_(t,e){var A=typeof t=="string"?t.toLowerCase():t,i=typeof e=="string"?e.toLowerCase():e;return(0,OrA.default)(A,i)}function baA(t){var e=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[],i=arguments.length>3&&arguments[3]!==void 0?arguments[3]:1,n=Ke(t,e);if(wo(n)){if(A===void 0)throw new Error("Cannot sort: no property selected by which to sort the array");return function(o){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],s=arguments.length>2&&arguments[2]!==void 0?arguments[2]:[],a=arguments.length>3&&arguments[3]!==void 0?arguments[3]:1,c=function(I,C){var d={boolean:0,number:1,string:2,undefined:4},B=3;return function(E,h){var u=Ke(E,I),D=Ke(h,I);if(typeof u!=typeof D){var L,R,w=(L=d[typeof u])!==null&&L!==void 0?L:B,_=(R=d[typeof D])!==null&&R!==void 0?R:B;return w>_?C:w<_?-C:0}return typeof u=="number"||typeof u=="boolean"?u>D?C:u<D?-C:0:No(u)?0:C*L_(u,D)}}(s,a),l=Ke(o,r);return[{op:"replace",path:Ct(r),value:l.slice(0).sort(c)}]}(t,e,A,i)}if(Cn(n))return function(o){var r=arguments.length>1&&arguments[1]!==void 0?arguments[1]:[],s=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1,a=Ke(o,r),c=Object.keys(a).slice();c.sort((I,C)=>s*L_(I,C));var l={};return c.forEach(I=>l[I]=a[I]),[{op:"replace",path:Ct(r),value:l}]}(t,e,i);throw new Error("Cannot sort: no array or object")}of(["click"]);Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-navigation-bar-dropdown.svelte-2nnd2m {
  position: absolute;
  top: 100%;
  left: 0;
  z-index: 3;
  background: var(--jse-navigation-bar-background, var(--jse-background-color, #fff));
  color: var(--jse-navigation-bar-dropdown-color, #656565);
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
  display: flex;
  flex-direction: column;
  max-height: 300px;
  overflow: auto;
  min-width: 80px;
}
.jse-navigation-bar-dropdown.svelte-2nnd2m button.jse-navigation-bar-dropdown-item:where(.svelte-2nnd2m) {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  outline: none;
  text-align: left;
  white-space: nowrap;
  box-sizing: border-box;
  padding: calc(0.5 * var(--jse-padding, 10px)) 36px;
}
.jse-navigation-bar-dropdown.svelte-2nnd2m button.jse-navigation-bar-dropdown-item:where(.svelte-2nnd2m):focus, .jse-navigation-bar-dropdown.svelte-2nnd2m button.jse-navigation-bar-dropdown-item:where(.svelte-2nnd2m):hover {
  background: var(--jse-navigation-bar-background-highlight, #e5e5e5);
}
.jse-navigation-bar-dropdown.svelte-2nnd2m button.jse-navigation-bar-dropdown-item.jse-selected:where(.svelte-2nnd2m) {
  background: var(--jse-navigation-bar-dropdown-color, #656565);
  color: var(--jse-navigation-bar-background, var(--jse-background-color, #fff));
}`);var zYA=wA('<button type="button"> </button>'),OYA=wA('<button type="button" class="jse-navigation-bar-dropdown-item svelte-2nnd2m">...</button>'),PYA=wA('<div class="jse-navigation-bar-dropdown svelte-2nnd2m"><!> <!></div>');function jYA(t,e){mt(e,!1);var A=b(e,"items",9),i=b(e,"selectedItem",9),n=b(e,"onSelect",9);Zt(!0);var o=PYA(),r=W(o);qo(r,1,()=>(k(Ey),k(A()),nA(()=>Ey(A(),100))),c=>c,(c,l)=>{var I,C=zYA(),d=W(C);De((B,E,h)=>{I=Vt(C,1,"jse-navigation-bar-dropdown-item svelte-2nnd2m",null,I,B),En(C,"title",E),wt(d,h)},[()=>({"jse-selected":g(l)===i()}),()=>(g(l),nA(()=>g(l).toString())),()=>(k(V0),g(l),nA(()=>V0(g(l).toString(),30)))],qA),ce("click",C,j0(()=>n()(g(l)))),iA(c,C)});var s=IA(r,2),a=c=>{var l=OYA();En(l,"title","Limited to 100 items"),iA(c,l)};LA(s,c=>{k(A()),nA(()=>A().length>100)&&c(a)}),iA(t,o),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-navigation-bar-item.svelte-752ro1 {
  position: relative;
  display: flex;
}
.jse-navigation-bar-item.svelte-752ro1 button.jse-navigation-bar-button:where(.svelte-752ro1) {
  font-family: inherit;
  font-size: inherit;
  padding: calc(0.5 * var(--jse-padding, 10px)) 2px;
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  outline: none;
  min-width: 2em;
  white-space: nowrap;
}
.jse-navigation-bar-item.svelte-752ro1 button.jse-navigation-bar-button:where(.svelte-752ro1):focus, .jse-navigation-bar-item.svelte-752ro1 button.jse-navigation-bar-button:where(.svelte-752ro1):hover {
  background: var(--jse-panel-button-background-highlight, #e0e0e0);
  color: var(--panel-button-color-highlight, var(--jse-text-color, #4d4d4d));
}
.jse-navigation-bar-item.svelte-752ro1 button.jse-navigation-bar-button.jse-navigation-bar-arrow:where(.svelte-752ro1) {
  padding: 2px var(--jse-padding, 10px) 0;
}
.jse-navigation-bar-item.svelte-752ro1 button.jse-navigation-bar-button.jse-navigation-bar-arrow.jse-open:where(.svelte-752ro1) {
  background: var(--jse-navigation-bar-background, var(--jse-background-color, #fff));
  color: var(--jse-navigation-bar-dropdown-color, #656565);
}
.jse-navigation-bar-item.svelte-752ro1:last-child {
  padding-right: var(--jse-padding, 10px);
}`);var qYA=wA('<button type="button" class="jse-navigation-bar-button svelte-752ro1"> </button>'),VYA=wA('<div class="jse-navigation-bar-item svelte-752ro1"><button type="button"><!></button> <!></div>');function JrA(t,e){mt(e,!1);var A,i=$(void 0,!0),n=$(void 0,!0),{openAbsolutePopup:o,closeAbsolutePopup:r}=$1("absolute-popup"),s=b(e,"path",9),a=b(e,"index",9),c=b(e,"onSelect",9),l=b(e,"getItems",9),I=$(void 0,!0),C=$(!1,!0);function d(L){r(A),c()(g(i).concat(L))}fA(()=>(k(s()),k(a())),()=>{y(i,s().slice(0,a()))}),fA(()=>(k(s()),k(a())),()=>{y(n,s()[a()])}),Qn(),Zt(!0);var B,E=VYA(),h=W(E);ji(W(h),{get data(){return qS}});var u=IA(h,2),D=L=>{var R=qYA(),w=W(R);De(()=>wt(w,g(n))),ce("click",R,()=>d(g(n))),iA(L,R)};LA(u,L=>{g(n)!==void 0&&L(D)}),Co(E,L=>y(I,L),()=>g(I)),De(L=>B=Vt(h,1,"jse-navigation-bar-button jse-navigation-bar-arrow svelte-752ro1",null,B,L),[()=>({"jse-open":g(C)})],qA),ce("click",h,function(){if(g(I)){y(C,!0);var L={items:l()(g(i)),selectedItem:g(n),onSelect:d};A=o(jYA,L,{anchor:g(I),closeOnOuterClick:!0,onClose:()=>{y(C,!1)}})}}),iA(t,E),pt()}function BG(t){var e,A;if(navigator.clipboard)return navigator.clipboard.writeText(t);if((e=(A=document).queryCommandSupported)!==null&&e!==void 0&&e.call(A,"copy")){var i=document.createElement("textarea");i.value=t,i.style.position="fixed",i.style.opacity="0",document.body.appendChild(i),i.select();try{document.execCommand("copy")}catch(n){console.error(n)}finally{document.body.removeChild(i)}return Promise.resolve()}return console.error("Copy failed."),Promise.resolve()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-navigation-bar-path-editor.svelte-zc2wx7 {
  flex: 1;
  display: flex;
  border: var(--jse-edit-outline, 2px solid #656565);
  background: var(--jse-background-color, #fff);
}
.jse-navigation-bar-path-editor.svelte-zc2wx7 input.jse-navigation-bar-text:where(.svelte-zc2wx7) {
  flex: 1;
  font-family: inherit;
  font-size: inherit;
  padding: 0 5px 1px;
  background: var(--jse-background-color, #fff);
  color: var(--jse-text-color, #4d4d4d);
  border: none;
  outline: none;
}
.jse-navigation-bar-path-editor.svelte-zc2wx7 button:where(.svelte-zc2wx7) {
  border: none;
  background: var(--jse-background-color, #fff);
  cursor: pointer;
  font-family: inherit;
  font-size: 80%;
  color: inherit;
}
.jse-navigation-bar-path-editor.svelte-zc2wx7 button.jse-navigation-bar-copy.copied:where(.svelte-zc2wx7) {
  color: var(--message-success-background, #9ac45d);
}
.jse-navigation-bar-path-editor.svelte-zc2wx7 button.jse-navigation-bar-validation-error:where(.svelte-zc2wx7) {
  color: var(--jse-error-color, #ee5341);
}
.jse-navigation-bar-path-editor.error.svelte-zc2wx7 {
  border-color: var(--jse-error-color, #ee5341);
}
.jse-navigation-bar-path-editor.error.svelte-zc2wx7 input.jse-navigation-bar-text:where(.svelte-zc2wx7) {
  color: var(--jse-error-color, #ee5341);
}
.jse-navigation-bar-path-editor.svelte-zc2wx7 .jse-copied-text:where(.svelte-zc2wx7) {
  background: var(--message-success-background, #9ac45d);
  color: var(--jse-message-success-color, #fff);
  position: relative;
  margin: 2px;
  padding: 0 5px;
  border-radius: 3px;
}`);var ZYA=wA('<button type="button" class="jse-navigation-bar-validation-error svelte-zc2wx7"><!></button>'),WYA=wA('<div class="jse-copied-text svelte-zc2wx7">Copied!</div>'),XYA=wA('<div><input type="text" class="jse-navigation-bar-text svelte-zc2wx7"/> <!> <!> <button type="button" title="Copy selected path to the clipboard"><!></button></div>');function $YA(t,e){mt(e,!1);var A=$(),i=$1("absolute-popup"),n=b(e,"path",8),o=b(e,"pathParser",8),r=b(e,"onChange",8),s=b(e,"onClose",8),a=b(e,"onError",8),c=b(e,"pathExists",8),l=$(),I=$(),C=$(!1),d=void 0,B=$(!1);function E(){g(l).focus()}function h(H){try{var q=o().parse(H);return function(j){if(!c()(j))throw new Error("Path does not exist in current document")}(q),{path:q,error:void 0}}catch(j){return{path:void 0,error:j}}}hs(()=>{E()}),qc(()=>{clearTimeout(d)}),fA(()=>(k(o()),k(n())),()=>{y(I,o().stringify(n()))}),fA(()=>(g(C),g(I)),()=>{y(A,g(C)?h(g(I)).error:void 0)}),Qn(),Zt();var u,D=XYA(),L=W(D);Co(L,H=>y(l,H),()=>g(l));var R=IA(L,2),w=H=>{var q=ZYA();ji(W(q),{get data(){return g1}}),Us(q,(j,gA)=>hQ?.(j,gA),()=>fe({text:String(g(A)||"")},i)),iA(H,q)};LA(R,H=>{g(A)&&H(w)});var _=IA(R,2),K=H=>{iA(H,WYA())};LA(_,H=>{g(B)&&H(K)});var z,U=IA(_,2);ji(W(U),{get data(){return F0}}),De((H,q)=>{u=Vt(D,1,"jse-navigation-bar-path-editor svelte-zc2wx7",null,u,H),VC(L,g(I)),z=Vt(U,1,"jse-navigation-bar-copy svelte-zc2wx7",null,z,q)},[()=>({error:g(A)}),()=>({copied:g(B)})],qA),ce("keydown",L,j0(function(H){var q=n2(H);if(q==="Escape"&&(H.preventDefault(),s()()),q==="Enter"){H.preventDefault(),y(C,!0);var j=h(g(I));j.path!==void 0?r()(j.path):a()(j.error)}})),ce("input",L,function(H){y(I,H.currentTarget.value)}),ce("click",U,function(){BG(g(I)),y(B,!0),d=window.setTimeout(()=>y(B,!1),1e3),E()}),iA(t,D),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-navigation-bar.svelte-xs03gj {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  background: var(--jse-panel-background, #ebebeb);
  color: var(--jse-panel-button-color, inherit);
  padding: 0;
  margin: 0;
  display: flex;
  overflow: auto;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit:where(.svelte-xs03gj) {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  background: transparent;
  border: none;
  display: flex;
  cursor: pointer;
  outline: none;
  align-items: center;
}
.jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit.flex:where(.svelte-xs03gj) {
  flex: 1;
}
.jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit:where(.svelte-xs03gj):focus, .jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit:where(.svelte-xs03gj):hover, .jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit.editing:where(.svelte-xs03gj) {
  background: var(--jse-panel-button-background-highlight, #e0e0e0);
  color: var(--panel-button-color-highlight, var(--jse-text-color, #4d4d4d));
  transition: color 0.2s ease-in, background 0.2s ease-in;
}
.jse-navigation-bar.svelte-xs03gj .jse-navigation-bar-edit:where(.svelte-xs03gj) .jse-navigation-bar-space:where(.svelte-xs03gj) {
  flex: 1;
  text-align: left;
}`);var AJA=wA("<!> <!>",1),eJA=wA('<div class="jse-navigation-bar svelte-xs03gj"><!> <button type="button"><span class="jse-navigation-bar-space svelte-xs03gj"> </span> <!></button></div>');function tJA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=Sr("jsoneditor:NavigationBar"),o=b(e,"json",9),r=b(e,"selection",9),s=b(e,"onSelect",9),a=b(e,"onError",9),c=b(e,"pathParser",9),l=$(void 0,!0),I=$(!1,!0);function C(q){n("get items for path",q);var j=Ke(o(),q);if(Array.isArray(j))return d5(0,j.length).map(String);if(Cn(j)){var gA=Object.keys(j).slice(0);return gA.sort(L_),gA}return[]}function d(q){return Ms(o(),q)}function B(q){n("select path",JSON.stringify(q)),s()(_s(q,q))}function E(){y(I,!1)}function h(q){E(),B(q)}fA(()=>(k(r()),et),()=>{y(A,r()?et(r()):[])}),fA(()=>(k(o()),g(A)),()=>{y(i,No(Ke(o(),g(A))))}),fA(()=>g(A),()=>{g(A),setTimeout(()=>{if(g(l)&&g(l).scrollTo){var q=g(l).scrollWidth-g(l).clientWidth;q>0&&(n("scrollTo ",q),g(l).scrollTo({left:q,behavior:"smooth"}))}})}),Qn(),Zt(!0);var u=eJA(),D=W(u),L=q=>{var j=AJA(),gA=vt(j);qo(gA,1,()=>g(A),or,(lA,vA,tA)=>{JrA(lA,{getItems:C,get path(){return g(A)},index:tA,onSelect:B})});var QA=IA(gA,2),BA=lA=>{JrA(lA,{getItems:C,get path(){return g(A)},get index(){return g(A),nA(()=>g(A).length)},onSelect:B})};LA(QA,lA=>{g(i)&&lA(BA)}),iA(q,j)},R=q=>{$YA(q,{get path(){return g(A)},onClose:E,onChange:h,get onError(){return a()},pathExists:d,get pathParser(){return c()}})};LA(D,q=>{g(I)?q(R,!1):q(L)});var w,_=IA(D,2),K=W(_),z=W(K),U=IA(K,2),H=qA(()=>g(I)?jW:UW);ji(U,{get data(){return g(H)}}),Co(u,q=>y(l,q),()=>g(l)),De((q,j)=>{w=Vt(_,1,"jse-navigation-bar-edit svelte-xs03gj",null,w,q),En(_,"title",g(I)?"Cancel editing the selected path":"Edit the selected path"),wt(z,j)},[()=>({flex:!g(I),editing:g(I)}),()=>(k(No),k(o()),g(I),nA(()=>No(o())||g(I)?"\xA0":"Navigation bar"))],qA),ce("click",_,function(){y(I,!g(I))}),iA(t,u),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-search-box.svelte-1mxl2uo {
  border: var(--jse-panel-border, var(--jse-main-border, 1px solid #d7d7d7));
  border-radius: 3px;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  background: var(--jse-panel-background, #ebebeb);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
  display: inline-block;
  width: 400px;
  max-width: 100%;
  overflow: auto;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) {
  display: flex;
  align-items: stretch;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) button:where(.svelte-1mxl2uo),
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) input:where(.svelte-1mxl2uo) {
  font-family: inherit;
  font-size: inherit;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) button:where(.svelte-1mxl2uo) {
  display: block;
  text-align: center;
  border: none;
  padding: 0 5px;
  margin: 0;
  cursor: pointer;
  color: var(--jse-panel-button-color, inherit);
  background: var(--jse-panel-button-background, transparent);
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) button:where(.svelte-1mxl2uo):hover {
  color: var(--panel-button-color-highlight, var(--jse-text-color, #4d4d4d));
  background: var(--jse-panel-button-background-highlight, #e0e0e0);
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) input:where(.svelte-1mxl2uo) {
  color: var(--jse-panel-color, var(--jse-text-color, #4d4d4d));
  border: var(--jse-input-border, 1px solid #d8dbdf);
  border-radius: 3px;
  background: var(--jse-input-background, var(--jse-background-color, #fff));
  height: 28px;
  padding: 0 5px;
  margin: 0;
  flex: 1;
  width: 0;
  min-width: 50px;
  outline: none;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-replace-toggle:where(.svelte-1mxl2uo) {
  padding: var(--jse-padding, 10px) calc(0.5 * var(--jse-padding, 10px));
  min-width: 20px;
  background: var(--jse-panel-button-background-highlight, #e0e0e0);
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) {
  flex: 1;
  display: flex;
  flex-direction: column;
  padding: calc(0.5 * var(--jse-padding, 10px));
  gap: calc(0.5 * var(--jse-padding, 10px));
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-search-section:where(.svelte-1mxl2uo) {
  flex: 1;
  display: flex;
  align-items: center;
  position: relative;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-search-section:where(.svelte-1mxl2uo) .jse-search-icon:where(.svelte-1mxl2uo) {
  color: inherit;
  cursor: inherit;
  background: inherit;
  width: 32px;
  text-align: center;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-search-section:where(.svelte-1mxl2uo) label.jse-search-input-label:where(.svelte-1mxl2uo) {
  flex: 1;
  display: flex;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-search-section:where(.svelte-1mxl2uo) .jse-search-count:where(.svelte-1mxl2uo) {
  color: inherit;
  font-size: 80%;
  visibility: hidden;
  padding: 0 5px;
  min-width: 36px;
  text-align: center;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-search-section:where(.svelte-1mxl2uo) .jse-search-count.jse-visible:where(.svelte-1mxl2uo) {
  visibility: visible;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-replace-section:where(.svelte-1mxl2uo) {
  flex: 1;
  display: flex;
  padding-left: 32px;
}
.jse-search-box.svelte-1mxl2uo .jse-search-form:where(.svelte-1mxl2uo) .jse-search-contents:where(.svelte-1mxl2uo) .jse-replace-section:where(.svelte-1mxl2uo) button:where(.svelte-1mxl2uo) {
  width: auto;
}`);var iJA=wA('<button type="button" class="jse-replace-toggle svelte-1mxl2uo" title="Toggle visibility of replace options (Ctrl+H)"><!></button>'),nJA=wA('<div class="jse-replace-section svelte-1mxl2uo"><input class="jse-replace-input svelte-1mxl2uo" title="Enter replacement text" type="text" placeholder="Replace"/> <button type="button" title="Replace current occurrence (Ctrl+Enter)" class="svelte-1mxl2uo">Replace</button> <button type="button" title="Replace all occurrences" class="svelte-1mxl2uo">All</button></div>'),oJA=wA('<div class="jse-search-box svelte-1mxl2uo"><form class="jse-search-form svelte-1mxl2uo"><!> <div class="jse-search-contents svelte-1mxl2uo"><div class="jse-search-section svelte-1mxl2uo"><div class="jse-search-icon svelte-1mxl2uo"><!></div> <label class="jse-search-input-label svelte-1mxl2uo" about="jse-search input"><input class="jse-search-input svelte-1mxl2uo" title="Enter text to search" type="text" placeholder="Find"/></label> <div> </div> <button type="button" class="jse-search-next svelte-1mxl2uo" title="Go to next search result (Enter)"><!></button> <button type="button" class="jse-search-previous svelte-1mxl2uo" title="Go to previous search result (Shift+Enter)"><!></button> <button type="button" class="jse-search-clear svelte-1mxl2uo" title="Close search box (Esc)"><!></button></div> <!></div></form></div>');function MaA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=$(void 0,!0),o=Sr("jsoneditor:SearchBox"),r=b(e,"json",9),s=b(e,"documentState",9),a=b(e,"parser",9),c=b(e,"showSearch",9),l=b(e,"showReplace",13),I=b(e,"readOnly",9),C=b(e,"columns",9),d=b(e,"onSearch",9),B=b(e,"onFocus",9),E=b(e,"onPatch",9),h=b(e,"onClose",9),u=$("",!0),D="",L=$("",!0),R=$(!1,!0),w=$(void 0,!0),_=oE(function(SA){return TA.apply(this,arguments)},300),K=oE(function(SA){return Ze.apply(this,arguments)},300);function z(){l(!l()&&!I())}function U(SA){SA.stopPropagation();var Ue=n2(SA);Ue==="Enter"&&(SA.preventDefault(),g(u)!==D?_.flush():tA()),Ue==="Shift+Enter"&&(SA.preventDefault(),pA()),Ue==="Ctrl+Enter"&&(SA.preventDefault(),l()?gA():tA()),Ue==="Ctrl+H"&&(SA.preventDefault(),z()),Ue==="Escape"&&(SA.preventDefault(),eA())}function H(SA){n2(SA)==="Enter"&&(SA.preventDefault(),SA.stopPropagation(),gA())}function q(){return j.apply(this,arguments)}function j(){return(j=Yt(function*(){io(),yield _.flush()})).apply(this,arguments)}function gA(){return QA.apply(this,arguments)}function QA(){return(QA=Yt(function*(){var SA;if(!I()){var Ue=(SA=g(w))===null||SA===void 0?void 0:SA.activeItem;if(o("handleReplace",{replaceText:g(L),activeItem:Ue}),g(w)&&Ue&&r()!==void 0){y(w,fe(fe({},prA(g(w))),{},{activeIndex:g(i)}));var{operations:mA,newSelection:sA}=QUA(r(),s(),g(L),Ue,a());E()(mA,(xt,tt)=>({state:tt,selection:sA})),io(),yield K.flush(),yield oe()}}})).apply(this,arguments)}function BA(){return lA.apply(this,arguments)}function lA(){return(lA=Yt(function*(){if(!I()){o("handleReplaceAll",{text:g(u),replaceText:g(L)});var{operations:SA,newSelection:Ue}=function(mA,sA,xt,tt,de){for(var Dt=wrA(xt,mA,{maxResults:1/0}),_e=[],Le=0;Le<Dt.length;Le++){var bt=Dt[Le-1],Re=Dt[Le];Le!==0&&Re.field===bt.field&&di(Re.path,bt.path)?hi(_e).items.push(Re):_e.push({path:Re.path,field:Re.field,items:[Re]})}_e.sort((Y,P)=>Y.field!==P.field?Y.field===Nl.key?1:-1:P.path.length-Y.path.length);var $t,x=[];return _e.forEach(Y=>{var{field:P,path:X,items:bA}=Y;if(P===Nl.key){var Be=Fi(X),Ee=Ke(mA,Be),kA=hi(X),DA=gf(Be,Object.keys(Ee),kA,yrA(kA,tt,bA));x=x.concat(DA),$t=QQ(mA,DA)}else{if(P!==Nl.value)throw new Error("Cannot replace: unknown type of search result field ".concat(P));var gt=Ke(mA,X);if(gt===void 0)throw new Error("Cannot replace: path not found ".concat(Ct(X)));var Ve=typeof gt=="string"?gt:String(gt),ZA=zg(mA,sA,X),rt=yrA(Ve,tt,bA),Ei=[{op:"replace",path:Ct(X),value:ZA?rt:DQ(rt,de)}];x=x.concat(Ei),$t=QQ(mA,Ei)}}),{operations:x,newSelection:$t}}(r(),s(),g(u),g(L),a());E()(SA,(mA,sA)=>({state:sA,selection:Ue})),yield oe()}})).apply(this,arguments)}function vA(SA){SA.select()}function tA(){return cA.apply(this,arguments)}function cA(){return(cA=Yt(function*(){y(w,g(w)?prA(g(w)):void 0),yield oe()})).apply(this,arguments)}function pA(){return VA.apply(this,arguments)}function VA(){return VA=Yt(function*(){y(w,g(w)?function(SA){var Ue=SA.activeIndex>0?SA.activeIndex-1:SA.items.length-1,mA=SA.items[Ue],sA=SA.items.map((xt,tt)=>fe(fe({},xt),{},{active:tt===Ue}));return fe(fe({},SA),{},{items:sA,activeItem:mA,activeIndex:Ue})}(g(w)):void 0),yield oe()}),VA.apply(this,arguments)}function oe(){return KA.apply(this,arguments)}function KA(){return(KA=Yt(function*(){var SA;o("handleFocus",g(w));var Ue=(SA=g(w))===null||SA===void 0?void 0:SA.activeItem;Ue&&r()!==void 0&&(yield B()(Ue.path,Ue.resultIndex))})).apply(this,arguments)}function CA(){return CA=Yt(function*(SA){yield He(SA,g(u),r())}),CA.apply(this,arguments)}function TA(){return TA=Yt(function*(SA){yield He(c(),SA,r()),yield oe()}),TA.apply(this,arguments)}function Ze(){return Ze=Yt(function*(SA){yield He(c(),g(u),SA)}),Ze.apply(this,arguments)}function He(SA,Ue,mA){return uA.apply(this,arguments)}function uA(){return uA=Yt(function*(SA,Ue,mA){return SA?(o("applySearch",{showSearch:SA,text:Ue}),Ue===""?(o("clearing search result"),g(w)!==void 0&&y(w,void 0),Promise.resolve()):(D=Ue,y(R,!0),new Promise(sA=>{setTimeout(()=>{var xt=wrA(Ue,mA,{maxResults:ON,columns:C()});y(w,function(tt,de){var Dt=de!=null&&de.activeItem?vrA(de.activeItem):void 0,_e=tt.findIndex(Re=>di(Dt,vrA(Re))),Le=_e!==-1?_e:de?.activeIndex!==void 0&&de?.activeIndex<tt.length?de?.activeIndex:tt.length>0?0:-1,bt=tt.map((Re,$t)=>fe(fe({resultIndex:$t},Re),{},{active:$t===Le}));return{items:bt,activeItem:bt[Le],activeIndex:Le}}(xt,g(w))),y(R,!1),sA()})}))):(g(w)&&y(w,void 0),Promise.resolve())}),uA.apply(this,arguments)}function eA(){o("handleClose"),_.cancel(),K.cancel(),He(!1,g(u),r()),h()()}fA(()=>g(w),()=>{var SA;y(A,((SA=g(w))===null||SA===void 0||(SA=SA.items)===null||SA===void 0?void 0:SA.length)||0)}),fA(()=>g(w),()=>{var SA;y(i,((SA=g(w))===null||SA===void 0?void 0:SA.activeIndex)||0)}),fA(()=>(g(A),ON),()=>{y(n,g(A)>=ON?"".concat(999,"+"):String(g(A)))}),fA(()=>(k(d()),g(w)),()=>{d()(g(w))}),fA(()=>k(c()),()=>{(function(SA){CA.apply(this,arguments)})(c())}),fA(()=>g(u),()=>{_(g(u))}),fA(()=>k(r()),()=>{K(r())}),Qn(),Zt(!0);var UA=_o(),aA=vt(UA),le=SA=>{var Ue=oJA(),mA=W(Ue),sA=W(mA),xt=kA=>{var DA=iJA(),gt=W(DA),Ve=qA(()=>l()?mg:sE);ji(gt,{get data(){return g(Ve)}}),ce("click",DA,z),iA(kA,DA)};LA(sA,kA=>{I()||kA(xt)});var tt=W(IA(sA,2)),de=W(tt),Dt=W(de),_e=kA=>{ji(kA,{get data(){return NW},spin:!0})},Le=kA=>{ji(kA,{get data(){return g4}})};LA(Dt,kA=>{g(R)?kA(_e):kA(Le,!1)});var bt=IA(de,2),Re=W(bt);es(()=>By(Re,()=>g(u),kA=>y(u,kA))),Us(Re,kA=>vA?.(kA)),es(()=>ce("paste",Re,q));var $t,x=IA(bt,2),Y=W(x),P=IA(x,2);ji(W(P),{get data(){return OW}});var X=IA(P,2);ji(W(X),{get data(){return KW}});var bA=IA(X,2);ji(W(bA),{get data(){return I4}});var Be=IA(tt,2),Ee=kA=>{var DA=nJA(),gt=W(DA),Ve=IA(gt,2),ZA=IA(Ve,2);By(gt,()=>g(L),rt=>y(L,rt)),ce("keydown",gt,H),ce("click",Ve,gA),ce("click",ZA,BA),iA(kA,DA)};LA(Be,kA=>{l()&&!I()&&kA(Ee)}),De(kA=>{var DA;$t=Vt(x,1,"jse-search-count svelte-1mxl2uo",null,$t,kA),wt(Y,"".concat(g(i)!==-1&&g(i)<g(A)?"".concat(g(i)+1,"/"):"").concat((DA=g(n))!==null&&DA!==void 0?DA:""))},[()=>({"jse-visible":g(u)!==""})],qA),ce("click",P,tA),ce("click",X,pA),ce("click",bA,eA),ce("keydown",mA,U),iA(SA,Ue)};LA(aA,SA=>{c()&&SA(le)}),iA(t,UA),pt()}var $3=Symbol("path");function rJA(t,e){var A=arguments.length>2&&arguments[2]!==void 0?arguments[2]:1/0,i={};Array.isArray(t)&&function(o,r,s){if(o.length<r)o.forEach(s);else for(var a=r>1?(o.length-1)/(r-1):o.length,c=0;c<r;c++){var l=Math.floor(c*a);s(o[l],l,o)}}(t,A,o=>{Cn(o)?kaA(o,i,e):i[$3]=!0});var n=[];return $3 in i&&n.push([]),SaA(i,[],n,e),n}function kaA(t,e,A){for(var i in t){var n=t[i],o=e[i]||(e[i]={});Cn(n)&&A?kaA(n,o,A):o[$3]===void 0&&(o[$3]=!0)}}function SaA(t,e,A,i){for(var n in t){var o=e.concat(n),r=t[n];r&&r[$3]===!0&&A.push(o),xo(r)&&i&&SaA(r,o,A,i)}}function sJA(t,e,A,i,n,o){for(var r=arguments.length>6&&arguments[6]!==void 0?arguments[6]:80,s=wo(A)?A.length:0,a=function(D,L){var R=Object.values(D);if(Oi(R))return L;var w=(_,K)=>_+K;return R.reduce(w)/R.length}(i,n),c=t-r,l=e+2*r,I=D=>i[D]||n,C=0,d=o;d<c&&C<s;)d+=I(C),C++;C>0&&(d-=I(--C));for(var B=C,E=0;E<l&&B<s;)E+=I(B),B++;for(var h=0,u=B;u<s;u++)h+=I(u);return{startIndex:C,endIndex:B,startHeight:d,endHeight:h,averageItemHeight:a,visibleHeight:E,visibleItems:wo(A)?A.slice(C,B):[]}}function TrA(t,e,A,i){for(var{rowIndex:n}=zc(t,e),o=0,r=0;r<n;r++)o+=A[r]||i;return o}function zc(t,e){var[A,...i]=t,n=parseInt(A,10);return{rowIndex:isNaN(n)?-1:n,columnIndex:e.findIndex(o=>Pg(i,o))}}function YC(t,e){var{rowIndex:A,columnIndex:i}=t;return[String(A),...e[i]]}function aJA(t,e){var[A,i]=GS(t,r=>j_(r.path[0])),n=NS(A,cJA),o=_S(n,r=>{var s={row:[],columns:{}};return r.forEach(a=>{var c=function(l,I){var C=zc(l.path,I);return C.columnIndex!==-1?C.columnIndex:-1}(a,e);c!==-1?(s.columns[c]===void 0&&(s.columns[c]=[]),s.columns[c].push(a)):s.row.push(a)}),s});return{root:i,rows:o}}function XE(t,e){if(e&&e.length!==0)return e.length===1?e[0]:{path:t,message:"Multiple validation issues: "+e.map(A=>Ka(A.path)+" "+A.message).join(", "),severity:Fl.warning}}function cJA(t){return parseInt(t.path[0],10)}function lJA(t,e,A){var i=e.some(n=>function(o,r,s){if(!o)return!1;if(r.op==="replace"){var a=ks(r.path),{rowIndex:c,columnIndex:l}=zc(a,s),I=s.findIndex(C=>di(C,o.path));if(c!==-1&&l!==-1&&l!==I)return!1}return!0}(t,n,A));return i?void 0:t}var Gs=Sr("jsoneditor:actions");function RaA(t){return F_.apply(this,arguments)}function F_(){return F_=Yt(function*(t){var{json:e,selection:A,indentation:i,readOnly:n,parser:o,onPatch:r}=t;if(!n&&e!==void 0&&A&&oQ(A)){var s=eaA(e,A,i,o);if(s!==void 0){Gs("cut",{selection:A,clipboard:s,indentation:i}),yield BG(s);var{operations:a,newSelection:c}=saA(e,A);r(a,(l,I)=>({state:I,selection:c}))}}}),F_.apply(this,arguments)}function xaA(t){return N_.apply(this,arguments)}function N_(){return N_=Yt(function*(t){var{json:e,selection:A,indentation:i,parser:n}=t,o=eaA(e,A,i,n);o!==void 0&&(Gs("copy",{clipboard:o,indentation:i}),yield BG(o))}),N_.apply(this,arguments)}function LaA(t){var{clipboardText:e,json:A,selection:i,readOnly:n,parser:o,onPatch:r,onChangeText:s,onPasteMultilineText:a,openRepairModal:c}=t;if(!n)try{l(e)}catch{c(e,C=>{Gs("repaired pasted text: ",C),l(C)})}function l(I){if(A!==void 0){var C=i||Ni([]),d=raA(A,C,I,o),B=function(E,h,u){var D=arguments.length>3&&arguments[3]!==void 0?arguments[3]:cUA;if(E.length>D)return!1;var L=/\n/.test(E);if(!L)return!1;var R=h.some(_=>_.op==="replace"&&Array.isArray(_.value)),w=h.filter(_=>_.op==="add").length>1;if(!R&&!w)return!1;try{return sf(E,u.parse),!1}catch{return!0}}(e,d,o);Gs("paste",{pastedText:I,operations:d,ensureSelection:C,pasteMultilineText:B}),r(d,(E,h)=>{var u=h;return d.filter(D=>(lS(D)||k8(D))&&No(D.value)).forEach(D=>{var L=ya(A,D.path);u=WC(E,u,L)}),{state:u}}),B&&a(I)}else Gs("paste text",{pastedText:I}),s(e,(E,h)=>{if(E)return{state:WC(E,h,[])}})}}function FaA(t){var{json:e,text:A,selection:i,keepSelection:n,readOnly:o,onChange:r,onPatch:s}=t;if(!o&&i){var a=e!==void 0&&(kr(i)||an(i))?_s(i.path,i.path):i;if(Oi(et(i)))Gs("remove root",{selection:i}),r&&r({text:"",json:void 0},e!==void 0?{text:void 0,json:e}:{text:A||"",json:e},{contentErrors:void 0,patchResult:void 0});else if(e!==void 0){var{operations:c,newSelection:l}=saA(e,a);Gs("remove",{operations:c,selection:i,newSelection:l}),s(c,(I,C)=>({state:C,selection:n?i:l}))}}}function Sy(t){var{insertType:e,selectInside:A,initialValue:i,json:n,selection:o,readOnly:r,parser:s,onPatch:a,onReplaceJson:c}=t;if(!r){var l=function(E,h,u){if(u==="object")return{};if(u==="array")return[];if(u==="structure"&&E!==void 0){var D=h?$sA(h):[],L=Ke(E,D);if(Array.isArray(L)&&!Oi(L)){var R=Fc(L);return No(R)?RS(R,w=>Array.isArray(w)?[]:Cn(w)?void 0:""):""}}return""}(n,o,e);if(n!==void 0){var I=s.stringify(l),C=raA(n,o,I,s);Gs("onInsert",{insertType:e,operations:C,newValue:l,data:I});var d=hi(C.filter(E=>E.op==="add"||E.op==="replace"));a(C,(E,h,u)=>{if(d){var D=ya(E,d.path);if(No(l))return{state:Sl(E,h,D,aG),selection:A?r2(D):u};if(l===""){var L=Oi(D)?void 0:Ke(E,Fi(D));return{state:Sl(E,h,D,ay),selection:Cn(L)?cG(D,i):my(D,i)}}}}),Gs("after patch")}else{Gs("onInsert",{insertType:e,newValue:l});var B=[];c(l,(E,h)=>({state:WC(E,h,B),selection:No(l)?r2(B):my(B)}))}}}function NaA(t){return __.apply(this,arguments)}function __(){return __=Yt(function*(t){var{char:e,selectInside:A,json:i,selection:n,readOnly:o,parser:r,onPatch:s,onReplaceJson:a,onSelect:c}=t;o||(kr(n)?c(fe(fe({},n),{},{edit:!0,initialValue:e})):e==="{"?Sy({insertType:"object",selectInside:A,initialValue:void 0,json:i,selection:n,readOnly:o,parser:r,onPatch:s,onReplaceJson:a}):e==="["?Sy({insertType:"array",selectInside:A,initialValue:void 0,json:i,selection:n,readOnly:o,parser:r,onPatch:s,onReplaceJson:a}):an(n)&&i!==void 0?No(Ke(i,n.path))||c(fe(fe({},n),{},{edit:!0,initialValue:e})):(Gs("onInsertValueWithCharacter",{char:e}),yield function(l){return G_.apply(this,arguments)}({char:e,json:i,selection:n,readOnly:o,parser:r,onPatch:s,onReplaceJson:a})))}),__.apply(this,arguments)}function G_(){return G_=Yt(function*(t){var{char:e,json:A,selection:i,readOnly:n,parser:o,onPatch:r,onReplaceJson:s}=t;n||Sy({insertType:"value",selectInside:!1,initialValue:e,json:A,selection:i,readOnly:n,parser:o,onPatch:r,onReplaceJson:s})}),G_.apply(this,arguments)}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-json-preview.svelte-1vjn89h {
  flex: 1;
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  overflow: auto;
  white-space: pre-wrap;
  padding: 2px;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}`);var gJA=wA('<div class="jse-json-preview svelte-1vjn89h"> </div>');function _aA(t,e){mt(e,!1);var A=$(),i=$(),n=b(e,"text",8),o=b(e,"json",8),r=b(e,"indentation",8),s=b(e,"parser",8);fA(()=>(k(o()),k(n())),()=>{y(A,o()!==void 0?{json:o()}:{text:n()||""})}),fA(()=>(g(A),k(r()),k(s()),Qy),()=>{y(i,V0(u_(g(A),r(),s()),Qy))}),Qn(),Zt();var a=gJA(),c=W(a);De(()=>wt(c,g(i))),iA(t,a),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
button.jse-context-menu-button.svelte-1idfykj {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  flex: 1;
  white-space: nowrap;
  padding: var(--jse-padding, 10px);
  color: inherit;
}
button.jse-context-menu-button.svelte-1idfykj:hover {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
}
button.jse-context-menu-button.svelte-1idfykj:focus {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
  z-index: 1;
}
button.jse-context-menu-button.svelte-1idfykj:disabled {
  color: var(--jse-context-menu-color-disabled, #9d9d9d);
  background: unset;
}
button.jse-context-menu-button.left.svelte-1idfykj {
  text-align: left;
}
button.jse-context-menu-button.svelte-1idfykj svg {
  width: 16px;
}`);var IJA=wA('<button type="button"><!> <!></button>');function n_(t,e){mt(e,!1);var A=b(e,"item",8),i=b(e,"className",8,void 0),n=b(e,"onRequestClose",8);Zt();var o=IJA(),r=W(o),s=l=>{ji(l,{get data(){return k(A()),nA(()=>A().icon)}})};LA(r,l=>{k(A()),nA(()=>A().icon)&&l(s)});var a=IA(r,2),c=l=>{var I=Tr();De(()=>wt(I,(k(A()),nA(()=>A().text)))),iA(l,I)};LA(a,l=>{k(A()),nA(()=>A().text)&&l(c)}),De(l=>{Vt(o,1,l,"svelte-1idfykj"),En(o,"title",(k(A()),nA(()=>A().title))),o.disabled=(k(A()),nA(()=>A().disabled||!1))},[()=>Z1((k(Kl),k(i()),k(A()),nA(()=>Kl("jse-context-menu-button",i(),A().className))))],qA),ce("click",o,l=>{n()(),A().onClick(l)}),iA(t,o),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-dropdown-button.svelte-11rxb2m {
  flex: 1;
  line-height: normal;
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  position: relative;
  padding: 0;
  display: flex;
}
.jse-dropdown-button.svelte-11rxb2m ul:where(.svelte-11rxb2m) {
  margin: 0;
  padding: 0;
}
.jse-dropdown-button.svelte-11rxb2m ul:where(.svelte-11rxb2m) li:where(.svelte-11rxb2m) {
  margin: 0;
  padding: 0;
  list-style-type: none;
}
.jse-dropdown-button.svelte-11rxb2m button.jse-open-dropdown:where(.svelte-11rxb2m) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  width: 2em;
  background: var(--jse-context-menu-background, #656565);
  color: var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff));
  border-radius: 0;
}
.jse-dropdown-button.svelte-11rxb2m button.jse-open-dropdown.jse-visible:where(.svelte-11rxb2m) {
  background: var(--jse-context-menu-background, #656565);
}
.jse-dropdown-button.svelte-11rxb2m button.jse-open-dropdown:where(.svelte-11rxb2m):hover {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
}
.jse-dropdown-button.svelte-11rxb2m button.jse-open-dropdown:where(.svelte-11rxb2m):focus {
  z-index: 1;
}
.jse-dropdown-button.svelte-11rxb2m button.jse-open-dropdown:where(.svelte-11rxb2m):disabled {
  color: var(--jse-context-menu-color-disabled, #9d9d9d);
  background: unset;
}
.jse-dropdown-button.svelte-11rxb2m .jse-dropdown-items:where(.svelte-11rxb2m) {
  display: none;
  position: absolute;
  top: 100%;
  left: 0;
  z-index: 1;
  background: var(--jse-context-menu-background, #656565);
  color: var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff));
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
}
.jse-dropdown-button.svelte-11rxb2m .jse-dropdown-items.jse-visible:where(.svelte-11rxb2m) {
  display: block;
}
.jse-dropdown-button.svelte-11rxb2m .jse-dropdown-items:where(.svelte-11rxb2m) button:where(.svelte-11rxb2m) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  width: 100%;
  text-align: left;
  padding: var(--jse-padding, 10px);
  margin: 0;
}
.jse-dropdown-button.svelte-11rxb2m .jse-dropdown-items:where(.svelte-11rxb2m) button:where(.svelte-11rxb2m):hover {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
}
.jse-dropdown-button.svelte-11rxb2m .jse-dropdown-items:where(.svelte-11rxb2m) button:where(.svelte-11rxb2m):disabled {
  color: var(--jse-context-menu-color-disabled, #9d9d9d);
  background: unset;
}`);var CJA=wA('<li class="svelte-11rxb2m"><button type="button"><!> </button></li>'),dJA=wA('<div role="button" tabindex="0" class="jse-dropdown-button svelte-11rxb2m"><!> <button type="button" data-type="jse-open-dropdown"><!></button> <div><ul class="svelte-11rxb2m"></ul></div></div>');Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
button.jse-context-menu-button.svelte-1idfykj {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  flex: 1;
  white-space: nowrap;
  padding: var(--jse-padding, 10px);
  color: inherit;
}
button.jse-context-menu-button.svelte-1idfykj:hover {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
}
button.jse-context-menu-button.svelte-1idfykj:focus {
  background: var(--jse-context-menu-background-highlight, #7a7a7a);
  z-index: 1;
}
button.jse-context-menu-button.svelte-1idfykj:disabled {
  color: var(--jse-context-menu-color-disabled, #9d9d9d);
  background: unset;
}
button.jse-context-menu-button.left.svelte-1idfykj {
  text-align: left;
}
button.jse-context-menu-button.svelte-1idfykj svg {
  width: 16px;
}`);var BJA=wA('<button type="button" slot="defaultItem"><!> </button>');function o_(t,e){mt(e,!1);var A=$(),i=b(e,"item",8),n=b(e,"className",8,void 0),o=b(e,"onRequestClose",8);fA(()=>(k(i()),k(o())),()=>{y(A,i().items.map(r=>fe(fe({},r),{},{onClick:s=>{o()(),r.onClick(s)}})))}),Qn(),Zt(),function(r,s){mt(s,!1);var a=$(void 0,!0),c=b(s,"items",25,()=>[]),l=b(s,"title",9,void 0),I=b(s,"width",9,"120px"),C=$(!1,!0);function d(){y(C,!1)}function B(w){n2(w)==="Escape"&&(w.preventDefault(),y(C,!1))}hs(()=>{document.addEventListener("click",d),document.addEventListener("keydown",B)}),qc(()=>{document.removeEventListener("click",d),document.removeEventListener("keydown",B)}),fA(()=>k(c()),()=>{y(a,c().every(w=>w.disabled===!0))}),Qn(),Zt(!0);var E=dJA(),h=W(E);jo(h,s,"defaultItem",{},null);var u,D=IA(h,2);ji(W(D),{get data(){return mg}});var L,R=IA(D,2);qo(W(R),5,c,or,(w,_)=>{var K=CJA(),z=W(K),U=W(z),H=j=>{ji(j,{get data(){return g(_),nA(()=>g(_).icon)}})};LA(U,j=>{g(_),nA(()=>g(_).icon)&&j(H)});var q=IA(U);De(()=>{var j;En(z,"title",(g(_),nA(()=>g(_).title))),z.disabled=(g(_),nA(()=>g(_).disabled)),Vt(z,1,Z1((g(_),nA(()=>g(_).className))),"svelte-11rxb2m"),wt(q," ".concat((g(_),(j=nA(()=>g(_).text))!==null&&j!==void 0?j:"")))}),ce("click",z,j=>g(_).onClick(j)),iA(w,K)}),De((w,_)=>{var K;En(E,"title",l()),u=Vt(D,1,"jse-open-dropdown svelte-11rxb2m",null,u,w),D.disabled=g(a),L=Vt(R,1,"jse-dropdown-items svelte-11rxb2m",null,L,_),Ll(R,"width: ".concat((K=I())!==null&&K!==void 0?K:"",";"))},[()=>({"jse-visible":g(C)}),()=>({"jse-visible":g(C)})],qA),ce("click",D,function(){var w=g(C);setTimeout(()=>y(C,!w))}),ce("click",E,d),iA(r,E),pt()}(t,{get width(){return k(i()),nA(()=>i().width)},get items(){return g(A)},$$slots:{defaultItem:(r,s)=>{var a=BJA(),c=W(a),l=C=>{ji(C,{get data(){return k(i()),nA(()=>i().main.icon)}})};LA(c,C=>{k(i()),nA(()=>i().main.icon)&&C(l)});var I=IA(c);De(C=>{var d;Vt(a,1,C,"svelte-1idfykj"),En(a,"title",(k(i()),nA(()=>i().main.title))),a.disabled=(k(i()),nA(()=>i().main.disabled||!1)),wt(I," ".concat((k(i()),(d=nA(()=>i().main.text))!==null&&d!==void 0?d:"")))},[()=>Z1((k(Kl),k(n()),k(i()),nA(()=>Kl("jse-context-menu-button",n(),i().main.className))))],qA),ce("click",a,C=>{o()(),i().main.onClick(C)}),iA(r,a)}}}),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-contextmenu.svelte-12z7bz1 {
  box-shadow: var(--jse-controls-box-shadow, 0 2px 6px 0 rgba(0, 0, 0, 0.24));
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  background: var(--jse-context-menu-background, #656565);
  color: var(--jse-context-menu-color, var(--jse-text-color-inverse, #fff));
}
.jse-contextmenu.svelte-12z7bz1 .jse-row:where(.svelte-12z7bz1) {
  display: flex;
  flex-direction: row;
  align-items: flex-start;
  justify-content: stretch;
}
.jse-contextmenu.svelte-12z7bz1 .jse-row:where(.svelte-12z7bz1) div.jse-label:where(.svelte-12z7bz1) {
  flex: 1;
  white-space: nowrap;
  padding: var(--jse-padding, 10px);
  color: var(--jse-context-menu-color-disabled, #9d9d9d);
  line-height: normal;
}
.jse-contextmenu.svelte-12z7bz1 .jse-row:where(.svelte-12z7bz1) div.jse-tip:where(.svelte-12z7bz1) {
  flex: 1;
  background: var(--jse-context-menu-tip-background, rgba(255, 255, 255, 0.2));
  color: var(--context-menu-tip-color, inherit);
  margin: calc(0.5 * var(--jse-padding, 10px));
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
  font-size: 80%;
  line-height: 1.3em;
  display: flex;
  flex-direction: row;
  align-items: flex-start;
  gap: var(--jse-padding, 10px);
  border-radius: 3px;
}
.jse-contextmenu.svelte-12z7bz1 .jse-row:where(.svelte-12z7bz1) div.jse-tip:where(.svelte-12z7bz1) div.jse-tip-icon:where(.svelte-12z7bz1) {
  padding-top: calc(0.5 * var(--jse-padding, 10px));
}
.jse-contextmenu.svelte-12z7bz1 .jse-column:where(.svelte-12z7bz1) {
  flex: 1;
  display: flex;
  flex-direction: column;
  align-items: stretch;
}
.jse-contextmenu.svelte-12z7bz1 .jse-column:where(.svelte-12z7bz1):not(:last-child) {
  border-right: 1px solid var(--jse-context-menu-separator-color, #7a7a7a);
}
.jse-contextmenu.svelte-12z7bz1 .jse-separator:where(.svelte-12z7bz1) {
  width: 100%;
  height: 1px;
  background: var(--jse-context-menu-separator-color, #7a7a7a);
}`);var EJA=wA('<div class="jse-separator svelte-12z7bz1"></div>'),QJA=wA('<div class="jse-label svelte-12z7bz1"> </div>'),hJA=wA('<div class="jse-column svelte-12z7bz1"></div>'),uJA=wA('<div class="jse-separator svelte-12z7bz1"></div>'),fJA=wA('<div class="jse-row svelte-12z7bz1"></div>'),mJA=wA('<div class="jse-separator svelte-12z7bz1"></div>'),pJA=wA('<div class="jse-row svelte-12z7bz1"><div class="jse-tip svelte-12z7bz1"><div class="jse-tip-icon svelte-12z7bz1"><!></div> <div class="jse-tip-text"> </div></div></div>'),wJA=wA('<div role="menu" tabindex="-1" class="jse-contextmenu svelte-12z7bz1"><!> <!></div>');function GaA(t,e){mt(e,!1);var A=b(e,"items",9),i=b(e,"onRequestClose",9),n=b(e,"tip",9),o=$(void 0,!0);hs(()=>{var C=Array.from(g(o).querySelectorAll("button")).find(d=>!d.disabled);C&&C.focus()});var r={ArrowUp:"Up",ArrowDown:"Down",ArrowLeft:"Left",ArrowRight:"Right"};function s(C){return console.error("Unknown type of context menu item",C),"???"}Zt(!0);var a=wJA(),c=W(a);qo(c,1,A,or,(C,d)=>{var B=_o(),E=vt(B),h=D=>{n_(D,{get item(){return g(d)},get onRequestClose(){return i()}})},u=(D,L)=>{var R=_=>{o_(_,{get item(){return g(d)},get onRequestClose(){return i()}})},w=(_,K)=>{var z=H=>{var q=fJA();qo(q,5,()=>(g(d),nA(()=>g(d).items)),or,(j,gA)=>{var QA=_o(),BA=vt(QA),lA=tA=>{n_(tA,{get item(){return g(gA)},get onRequestClose(){return i()}})},vA=(tA,cA)=>{var pA=oe=>{o_(oe,{get item(){return g(gA)},get onRequestClose(){return i()}})},VA=(oe,KA)=>{var CA=Ze=>{var He=hJA();qo(He,5,()=>(g(gA),nA(()=>g(gA).items)),or,(uA,eA)=>{var UA=_o(),aA=vt(UA),le=Ue=>{n_(Ue,{className:"left",get item(){return g(eA)},get onRequestClose(){return i()}})},SA=(Ue,mA)=>{var sA=tt=>{o_(tt,{className:"left",get item(){return g(eA)},get onRequestClose(){return i()}})},xt=(tt,de)=>{var Dt=Le=>{iA(Le,EJA())},_e=(Le,bt)=>{var Re=x=>{var Y=QJA(),P=W(Y);De(()=>wt(P,(g(eA),nA(()=>g(eA).text)))),iA(x,Y)},$t=x=>{var Y=Tr();De(P=>wt(Y,P),[()=>(g(eA),nA(()=>s(g(eA))))],qA),iA(x,Y)};LA(Le,x=>{k(rrA),g(eA),nA(()=>rrA(g(eA)))?x(Re):x($t,!1)},bt)};LA(tt,Le=>{k(G1),g(eA),nA(()=>G1(g(eA)))?Le(Dt):Le(_e,!1)},de)};LA(Ue,tt=>{k(ZE),g(eA),nA(()=>ZE(g(eA)))?tt(sA):tt(xt,!1)},mA)};LA(aA,Ue=>{k(q0),g(eA),nA(()=>q0(g(eA)))?Ue(le):Ue(SA,!1)}),iA(uA,UA)}),iA(Ze,He)},TA=(Ze,He)=>{var uA=UA=>{iA(UA,uJA())},eA=UA=>{var aA=Tr();De(le=>wt(aA,le),[()=>(g(gA),nA(()=>s(g(gA))))],qA),iA(UA,aA)};LA(Ze,UA=>{k(G1),g(gA),nA(()=>G1(g(gA)))?UA(uA):UA(eA,!1)},He)};LA(oe,Ze=>{k(arA),g(gA),nA(()=>arA(g(gA)))?Ze(CA):Ze(TA,!1)},KA)};LA(tA,oe=>{k(ZE),g(gA),nA(()=>ZE(g(gA)))?oe(pA):oe(VA,!1)},cA)};LA(BA,tA=>{k(q0),g(gA),nA(()=>q0(g(gA)))?tA(lA):tA(vA,!1)}),iA(j,QA)}),iA(H,q)},U=(H,q)=>{var j=QA=>{iA(QA,mJA())},gA=QA=>{var BA=Tr();De(lA=>wt(BA,lA),[()=>(g(d),nA(()=>s(g(d))))],qA),iA(QA,BA)};LA(H,QA=>{k(G1),g(d),nA(()=>G1(g(d)))?QA(j):QA(gA,!1)},q)};LA(_,H=>{k(srA),g(d),nA(()=>srA(g(d)))?H(z):H(U,!1)},K)};LA(D,_=>{k(ZE),g(d),nA(()=>ZE(g(d)))?_(R):_(w,!1)},L)};LA(E,D=>{k(q0),g(d),nA(()=>q0(g(d)))?D(h):D(u,!1)}),iA(C,B)});var l=IA(c,2),I=C=>{var d=pJA(),B=W(d),E=W(B);ji(W(E),{get data(){return xW}});var h=W(IA(E,2));De(()=>wt(h,n())),iA(C,d)};LA(l,C=>{n()&&C(I)}),Co(a,C=>y(o,C),()=>g(o)),ce("keydown",a,function(C){var d=n2(C),B=r[d];if(B&&C.target){C.preventDefault();var E=TGA({allElements:Array.from(g(o).querySelectorAll("button:not([disabled])")),currentElement:C.target,direction:B,hasPrio:h=>h.getAttribute("data-type")!=="jse-open-dropdown"});E&&E.focus()}}),iA(t,a),pt()}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-value.jse-string.svelte-6ttr41 {
  color: var(--jse-value-color-string, #008000);
}
.jse-value.jse-object.svelte-6ttr41, .jse-value.jse-array.svelte-6ttr41 {
  min-width: 16px;
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
}
.jse-value.jse-number.svelte-6ttr41 {
  color: var(--jse-value-color-number, #ee422e);
}
.jse-value.jse-boolean.svelte-6ttr41 {
  color: var(--jse-value-color-boolean, #ff8c00);
}
.jse-value.jse-null.svelte-6ttr41 {
  color: var(--jse-value-color-null, #004ed0);
}
.jse-value.jse-invalid.svelte-6ttr41 {
  color: var(--jse-text-color, #4d4d4d);
}
.jse-value.jse-url.svelte-6ttr41 {
  color: var(--jse-value-color-url, #008000);
  text-decoration: underline;
}

.jse-enum-value.svelte-6ttr41 {
  background: var(--jse-hover-background-color, rgba(0, 0, 0, 0.06));
  border: none;
  padding: 0;
  font-family: inherit;
  font-size: inherit;
  cursor: pointer;
  outline: none;
}
.jse-enum-value.jse-selected.svelte-6ttr41 {
  background: var(--jse-selection-background-color, #d3d3d3);
  color: inherit;
}
.jse-enum-value.jse-value.svelte-6ttr41:focus {
  color: var(--jse-text-color, #4d4d4d);
}`);var $6e=wA("<option> </option>"),A8e=wA("<select></select>");var ty,iy;function ny(t,e){return ty||(iy=new WeakMap,ty=new ResizeObserver(A=>{for(var i of A){var n=iy.get(i.target);n&&n(i.target)}})),iy.set(t,e),ty.observe(t),{destroy:()=>{iy.delete(t),ty.unobserve(t)}}}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-tree-mode.svelte-vrx1dr {
  flex: 1;
  display: flex;
  flex-direction: column;
  position: relative;
  background: var(--jse-background-color, #fff);
  min-width: 0;
  min-height: 0;
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: var(--jse-text-color, #4d4d4d);
  line-height: var(--jse-line-height, calc(1em + 4px));
}
.jse-tree-mode.svelte-vrx1dr .jse-hidden-input-label:where(.svelte-vrx1dr) .jse-hidden-input:where(.svelte-vrx1dr) {
  position: fixed;
  top: -10px;
  left: -10px;
  width: 1px;
  height: 1px;
  padding: 0;
  border: 0;
  outline: none;
}
.jse-tree-mode.no-main-menu.svelte-vrx1dr {
  border-top: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-tree-mode.svelte-vrx1dr .jse-search-box-container:where(.svelte-vrx1dr) {
  position: relative;
  height: 0;
  top: var(--jse-padding, 10px);
  margin-right: calc(var(--jse-padding, 10px) + 20px);
  margin-left: var(--jse-padding, 10px);
  text-align: right;
  z-index: 3;
}
.jse-tree-mode.svelte-vrx1dr .jse-contents:where(.svelte-vrx1dr) {
  flex: 1;
  overflow: auto;
  position: relative;
  padding: 2px;
  display: flex;
  flex-direction: column;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-tree-mode.svelte-vrx1dr .jse-contents:where(.svelte-vrx1dr):last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-tree-mode.svelte-vrx1dr .jse-contents:where(.svelte-vrx1dr) .jse-loading-space:where(.svelte-vrx1dr) {
  flex: 1;
}
.jse-tree-mode.svelte-vrx1dr .jse-contents:where(.svelte-vrx1dr) .jse-loading:where(.svelte-vrx1dr) {
  flex: 2;
  text-align: center;
  color: var(--jse-panel-color-readonly, #b2b2b2);
  box-sizing: border-box;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
}
.jse-tree-mode.svelte-vrx1dr .jse-contents:where(.svelte-vrx1dr) .jse-search-box-background:where(.svelte-vrx1dr) {
  border: 50px solid var(--jse-modal-background, #f5f5f5);
  margin: -2px;
  margin-bottom: 2px;
  display: inline-block;
}`);var DJA=wA("<!> <!>",1),yJA=wA('<div class="jse-search-box-background svelte-vrx1dr"></div>'),vJA=wA('<div class="jse-search-box-container svelte-vrx1dr"><!></div> <div class="jse-contents svelte-vrx1dr"><!> <!></div> <!> <!> <!> <!>',1),bJA=wA('<label class="jse-hidden-input-label svelte-vrx1dr"><input type="text" tabindex="-1" class="jse-hidden-input svelte-vrx1dr"/></label> <!>',1),MJA=wA('<div class="jse-contents svelte-vrx1dr"><div class="jse-loading-space svelte-vrx1dr"></div> <div class="jse-loading svelte-vrx1dr">loading...</div></div>'),kJA=wA('<div role="tree" tabindex="-1"><!> <!> <!></div> <!> <!>',1);function U_(t,e){mt(e,!1);var A=$(void 0,!0),i=Sr("jsoneditor:TreeMode"),n=typeof window>"u";i("isSSR:",n);var o=c1(),r=c1(),{openAbsolutePopup:s,closeAbsolutePopup:a}=$1("absolute-popup"),c=$(void 0,!0),l=$(void 0,!0),I=$(void 0,!0),C=!1,d=paA(),B=b(e,"readOnly",9),E=b(e,"externalContent",9),h=b(e,"externalSelection",9),u=b(e,"history",9),D=b(e,"truncateTextSize",9),L=b(e,"mainMenuBar",9),R=b(e,"navigationBar",9),w=b(e,"escapeControlCharacters",9),_=b(e,"escapeUnicodeCharacters",9),K=b(e,"parser",9),z=b(e,"parseMemoizeOne",9),U=b(e,"validator",9),H=b(e,"validationParser",9),q=b(e,"pathParser",9),j=b(e,"indentation",9),gA=b(e,"onError",9),QA=b(e,"onChange",9),BA=b(e,"onChangeMode",9),lA=b(e,"onSelect",9),vA=b(e,"onUndo",9),tA=b(e,"onRedo",9),cA=b(e,"onRenderValue",9),pA=b(e,"onRenderMenu",9),VA=b(e,"onRenderContextMenu",9),oe=b(e,"onClassName",9),KA=b(e,"onFocus",9),CA=b(e,"onBlur",9),TA=b(e,"onSortModal",9),Ze=b(e,"onTransformModal",9),He=b(e,"onJSONEditorModal",9),uA=!1,eA=$(!1,!0),UA=$(void 0,!0);CG({onMount:hs,onDestroy:qc,getWindow:()=>af(g(I)),hasFocus:()=>uA&&document.hasFocus()||W_(g(I)),onFocus:()=>{C=!0,KA()&&KA()()},onBlur:()=>{C=!1,CA()&&CA()()}});var aA=$(void 0,!0),le=$(void 0,!0),SA=void 0,Ue=!1,mA=$(w_({json:g(aA)}),!0),sA=$(q3(h())?h():void 0,!0);function xt(T){y(sA,T)}hs(()=>{if(g(sA)){var T=et(g(sA));y(mA,Sl(g(aA),g(mA),T,ay)),setTimeout(()=>Ai(T))}});var tt,de=$(void 0,!0),Dt=$(void 0,!0),_e=$(void 0,!0),Le=$(void 0,!0),bt=$(!1,!0),Re=$(!1,!0);function $t(T){y(Le,(tt=T)?caA(g(aA),tt.items):void 0)}function x(T,oA){return Y.apply(this,arguments)}function Y(){return(Y=Yt(function*(T,oA){y(mA,Sl(g(aA),g(mA),T,ay));var YA=hn(oA);yield ot(T,{element:YA})})).apply(this,arguments)}function P(){y(bt,!1),y(Re,!1),Ri()}function X(T){i("select validation error",T),y(sA,Ni(T.path)),ot(T.path)}function bA(T){var oA=arguments.length>1&&arguments[1]!==void 0?arguments[1]:ErA;i("expand"),y(mA,Sl(g(aA),g(mA),T,oA))}function Be(T,oA){y(mA,IrA(g(aA),g(mA),T,oA)),g(sA)&&function(YA,pe){return Pg(et(YA),pe)&&(et(YA).length>pe.length||fr(YA))}(g(sA),T)&&y(sA,void 0)}var Ee=$(!1,!0),kA=$([],!0),DA=$(void 0,!0),gt=aE(waA);function Ve(T,oA,YA,pe){sQ(()=>{var he;try{he=gt(T,oA,YA,pe)}catch(ge){he=[{path:[],message:"Failed to validate: "+ge.message,severity:Fl.warning}]}di(he,g(kA))||(i("validationErrors changed:",he),y(kA,he),y(DA,function(ge,Bt){var $e;return Bt.forEach(bi=>{$e=KrA(ge,$e,bi.path,(un,Ji)=>fe(fe({},Ji),{},{validationError:bi}))}),Bt.forEach(bi=>{for(var un=bi.path;un.length>0;)un=Fi(un),$e=KrA(ge,$e,un,(Ji,Tn)=>Tn.validationError?Tn:fe(fe({},Tn),{},{validationError:{isChildError:!0,path:un,message:"Contains invalid data",severity:Fl.warning}}))}),$e}(T,g(kA))))},he=>i("validationErrors updated in ".concat(he," ms")))}function ZA(){return i("validate"),SA?{parseError:SA,isRepairable:!1}:(Ve(g(aA),U(),K(),H()),Oi(g(kA))?void 0:{validationErrors:g(kA)})}function rt(){return g(aA)}function Ei(){return g(mA)}function tn(){return g(sA)}function qi(T){i("applyExternalContent",{updatedContent:T}),z3(T)?function(oA){if(oA!==void 0){var YA=!di(g(aA),oA);if(i("update external json",{isChanged:YA,currentlyText:g(aA)===void 0}),!!YA){var pe={documentState:g(mA),selection:g(sA),json:g(aA),text:g(le),textIsRepaired:g(Ee)};y(aA,oA),y(mA,Qc(oA,g(mA))),xe(g(aA)),y(le,void 0),y(Ee,!1),SA=void 0,nn(g(aA)),mi(pe)}}}(T.json):H3(T)&&function(oA){if(!(oA===void 0||z3(E()))){var YA=oA!==g(le);if(i("update external text",{isChanged:YA}),!!YA){var pe={documentState:g(mA),selection:g(sA),json:g(aA),text:g(le),textIsRepaired:g(Ee)};try{y(aA,z()(oA)),y(mA,Qc(g(aA),g(mA))),xe(g(aA)),y(le,oA),y(Ee,!1),SA=void 0}catch(he){try{y(aA,z()(Rc(oA))),y(mA,Qc(g(aA),g(mA))),xe(g(aA)),y(le,oA),y(Ee,!0),SA=void 0,nn(g(aA))}catch{y(aA,void 0),y(mA,void 0),y(le,E().text),y(Ee,!1),SA=g(le)!==void 0&&g(le)!==""?dQ(g(le),he.message||String(he)):void 0}}nn(g(aA)),mi(pe)}}}(T.text)}function xe(T){Ue||(Ue=!0,y(mA,WC(T,g(mA),[])))}function nn(T){g(sA)&&(Ms(T,OC(g(sA)))&&Ms(T,et(g(sA)))||(i("clearing selection: path does not exist anymore",g(sA)),y(sA,WE(T,g(mA)))))}function mi(T){if(T.json!==void 0||T.text!==void 0){var oA=g(aA)!==void 0&&T.json!==void 0;u().add({type:"tree",undo:{patch:oA?[{op:"replace",path:"",value:T.json}]:void 0,json:T.json,text:T.text,documentState:T.documentState,textIsRepaired:T.textIsRepaired,selection:Hg(T.selection),sortedColumn:void 0},redo:{patch:oA?[{op:"replace",path:"",value:g(aA)}]:void 0,json:g(aA),text:g(le),documentState:g(mA),textIsRepaired:g(Ee),selection:Hg(g(sA)),sortedColumn:void 0}})}}function Ot(T,oA){var YA;if(i("patch",T,oA),g(aA)===void 0)throw new Error("Cannot apply patch: no JSON");var pe=g(aA),he={json:void 0,text:g(le),documentState:g(mA),selection:Hg(g(sA)),textIsRepaired:g(Ee),sortedColumn:void 0},ge=aaA(g(aA),T),Bt=ZsA(g(aA),g(mA),T),$e=(YA=QQ(g(aA),T))!==null&&YA!==void 0?YA:g(sA),bi=typeof oA=="function"?oA(Bt.json,Bt.documentState,$e):void 0;return y(aA,bi?.json!==void 0?bi.json:Bt.json),y(mA,bi?.state!==void 0?bi.state:Bt.documentState),y(sA,bi?.selection!==void 0?bi.selection:$e),y(le,void 0),y(Ee,!1),y(Dt,void 0),y(_e,void 0),SA=void 0,nn(g(aA)),u().add({type:"tree",undo:fe({patch:ge},he),redo:{patch:T,json:void 0,text:g(le),documentState:g(mA),selection:Hg(g(sA)),sortedColumn:void 0,textIsRepaired:g(Ee)}}),{json:g(aA),previousJson:pe,undo:ge,redo:T}}function Lt(){!B()&&g(sA)&&y(sA,cG(et(g(sA))))}function ii(){if(!B()&&g(sA)){var T=et(g(sA)),oA=Ke(g(aA),T);No(oA)?function(YA,pe){i("openJSONEditorModal",{path:YA,value:pe}),uA=!0,He()({content:{json:pe},path:YA,onPatch:g(m).onPatch,onClose:()=>{uA=!1,setTimeout(Ri)}})}(T,oA):y(sA,my(T))}}function _i(){if(!B()&&an(g(sA))){var T=et(g(sA)),oA=Ct(T),YA=Ke(g(aA),T),pe=!zg(g(aA),g(mA),T),he=pe?String(YA):DQ(String(YA),K());i("handleToggleEnforceString",{enforceString:pe,value:YA,updatedValue:he}),dt([{op:"replace",path:oA,value:he}],(ge,Bt)=>({state:Ky(g(aA),Bt,T,{type:"value",enforceString:pe})}))}}function Tt(){return g(Ee)&&g(aA)!==void 0&&EA(g(aA)),g(aA)!==void 0?{json:g(aA)}:{text:g(le)||""}}function M(){return We.apply(this,arguments)}function We(){return We=Yt(function*(){var T=!(arguments.length>0&&arguments[0]!==void 0)||arguments[0];yield RaA({json:g(aA),selection:g(sA),indentation:T?j():void 0,readOnly:B(),parser:K(),onPatch:dt})}),We.apply(this,arguments)}function ni(){return pi.apply(this,arguments)}function pi(){return pi=Yt(function*(){var T=!(arguments.length>0&&arguments[0]!==void 0)||arguments[0];g(aA)!==void 0&&(yield xaA({json:g(aA),selection:g(sA),indentation:T?j():void 0,parser:K()}))}),pi.apply(this,arguments)}function dn(T){var oA;T.preventDefault(),kn((oA=T.clipboardData)===null||oA===void 0?void 0:oA.getData("text/plain"))}function mn(){return Uo.apply(this,arguments)}function Uo(){return(Uo=Yt(function*(){try{kn(yield navigator.clipboard.readText())}catch(T){console.error(T),y(eA,!0)}})).apply(this,arguments)}function kn(T){T!==void 0&&LaA({clipboardText:T,json:g(aA),selection:g(sA),readOnly:B(),parser:K(),onPatch:dt,onChangeText:HA,onPasteMultilineText:Jn,openRepairModal:Wn})}function Wn(T,oA){y(UA,{text:T,onParse:YA=>sf(YA,pe=>rf(pe,K())),onRepair:_sA,onApply:oA,onClose:Ri})}function Vo(){FaA({json:g(aA),text:g(le),selection:g(sA),keepSelection:!1,readOnly:B(),onChange:QA(),onPatch:dt})}function vo(){!B()&&g(aA)!==void 0&&g(sA)&&oQ&&!Oi(et(g(sA)))&&(i("duplicate",{selection:g(sA)}),dt(naA(g(aA),W1(g(aA),g(sA)))))}function bo(){B()||!g(sA)||!Kn(g(sA))&&!an(g(sA))||Oi(et(g(sA)))||(i("extract",{selection:g(sA)}),dt(oaA(g(aA),g(sA)),(T,oA)=>{if(No(T))return{state:ZN(T,oA,[])}}))}function Yn(T){Sy({insertType:T,selectInside:!0,initialValue:void 0,json:g(aA),selection:g(sA),readOnly:B(),parser:K(),onPatch:dt,onReplaceJson:EA})}function Mo(T){kr(g(sA))&&y(sA,Ni(g(sA).path)),g(sA)||y(sA,WE(g(aA),g(mA))),Yn(T)}function ne(T){if(!B()&&g(sA))if(WD(g(sA)))try{var oA=OC(g(sA)),YA=Ke(g(aA),oA),pe=function(ge,Bt,$e){if(Bt==="array"){if(Array.isArray(ge))return ge;if(Cn(ge))return XoA(ge);if(typeof ge=="string")try{var bi=$e.parse(ge);if(Array.isArray(bi))return bi;if(Cn(bi))return XoA(bi)}catch{return[ge]}return[ge]}if(Bt==="object"){if(Array.isArray(ge))return WoA(ge);if(Cn(ge))return ge;if(typeof ge=="string")try{var un=$e.parse(ge);if(Cn(un))return un;if(Array.isArray(un))return WoA(un)}catch{return{value:ge}}return{value:ge}}if(Bt==="value")return No(ge)?$e.stringify(ge):ge;throw new Error("Cannot convert ".concat(q_(ge,$e)," to ").concat(Bt))}(YA,T,K());if(pe===YA)return;var he=[{op:"replace",path:Ct(oA),value:pe}];i("handleConvert",{selection:g(sA),path:oA,type:T,operations:he}),dt(he,(ge,Bt)=>({state:g(sA)?WC(ge,Bt,et(g(sA))):g(mA)}))}catch(ge){gA()(ge)}else gA()(new Error("Cannot convert current selection to ".concat(T)))}function wi(){if(g(sA)){var T=QrA(g(aA),g(mA),g(sA),!1),oA=Fi(et(g(sA)));T&&!Oi(et(T))&&di(oA,Fi(et(T)))?y(sA,t2(et(T))):y(sA,r2(oA)),i("insert before",{selection:g(sA),selectionBefore:T,parentPath:oA}),io(),Pt()}}function MA(){if(g(sA)){var T=P1(g(aA),g(sA));i("insert after",T),y(sA,t2(T)),io(),Pt()}}function me(T){return nt.apply(this,arguments)}function nt(){return(nt=Yt(function*(T){yield NaA({char:T,selectInside:!0,json:g(aA),selection:g(sA),readOnly:B(),parser:K(),onPatch:dt,onReplaceJson:EA,onSelect:xt})})).apply(this,arguments)}function Wt(){if(!B()&&u().canUndo){var T=u().undo();if(uy(T)){var oA={json:g(aA),text:g(le)};y(aA,T.undo.patch?Da(g(aA),T.undo.patch):T.undo.json),y(mA,T.undo.documentState),y(sA,T.undo.selection),y(le,T.undo.text),y(Ee,T.undo.textIsRepaired),SA=void 0,i("undo",{item:T,json:g(aA),documentState:g(mA),selection:g(sA)}),Ki(oA,T.undo.patch&&T.redo.patch?{json:g(aA),previousJson:oA.json,redo:T.undo.patch,undo:T.redo.patch}:void 0),Ri(),g(sA)&&ot(et(g(sA)),{scrollToWhenVisible:!1})}else vA()(T)}}function Xe(){if(!B()&&u().canRedo){var T=u().redo();if(uy(T)){var oA={json:g(aA),text:g(le)};y(aA,T.redo.patch?Da(g(aA),T.redo.patch):T.redo.json),y(mA,T.redo.documentState),y(sA,T.redo.selection),y(le,T.redo.text),y(Ee,T.redo.textIsRepaired),SA=void 0,i("redo",{item:T,json:g(aA),documentState:g(mA),selection:g(sA)}),Ki(oA,T.undo.patch&&T.redo.patch?{json:g(aA),previousJson:oA.json,redo:T.redo.patch,undo:T.undo.patch}:void 0),Ri(),g(sA)&&ot(et(g(sA)),{scrollToWhenVisible:!1})}else tA()(T)}}function oi(T){var oA;B()||g(aA)===void 0||(uA=!0,TA()({id:o,json:g(aA),rootPath:T,onSort:(oA=Yt(function*(YA){var{operations:pe}=YA;i("onSort",T,pe),dt(pe,(he,ge)=>({state:ZN(he,ge,T),selection:Ni(T)}))}),function(YA){return oA.apply(this,arguments)}),onClose:()=>{uA=!1,setTimeout(Ri)}}))}function Di(){g(sA)&&oi(urA(g(aA),g(sA)))}function Ut(){oi([])}function cn(T){if(g(aA)!==void 0){var{id:oA,onTransform:YA,onClose:pe}=T,he=T.rootPath||[];uA=!0,Ze()({id:oA||r,json:g(aA),rootPath:he,onTransform:ge=>{YA?YA({operations:ge,json:g(aA),transformedJson:Da(g(aA),ge)}):(i("onTransform",he,ge),dt(ge,(Bt,$e)=>({state:ZN(Bt,$e,he),selection:Ni(he)})))},onClose:()=>{uA=!1,setTimeout(Ri),pe&&pe()}})}}function ft(){g(sA)&&cn({rootPath:urA(g(aA),g(sA))})}function Qi(){cn({rootPath:[]})}function ot(T){return Mt.apply(this,arguments)}function Mt(){return Mt=Yt(function*(T){var{scrollToWhenVisible:oA=!0,element:YA}=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{};y(mA,Sl(g(aA),g(mA),T,ay));var pe=YA??on(T);if(i("scrollTo",{path:T,elem:pe,refContents:g(c)}),!pe||!g(c))return Promise.resolve();var he=g(c).getBoundingClientRect(),ge=pe.getBoundingClientRect();if(!oA&&ge.bottom>he.top&&ge.top<he.bottom)return Promise.resolve();var Bt=-he.height/4;return new Promise($e=>{d(pe,{container:g(c),offset:Bt,duration:300,callback:()=>$e()})})}),Mt.apply(this,arguments)}function on(T){var oA,YA;return io(),(oA=(YA=g(c))===null||YA===void 0?void 0:YA.querySelector('div[data-path="'.concat(sy(T),'"]')))!==null&&oA!==void 0?oA:void 0}function hn(T){var oA,YA;return io(),(oA=(YA=g(c))===null||YA===void 0?void 0:YA.querySelector('span[data-search-result-index="'.concat(T,'"]')))!==null&&oA!==void 0?oA:void 0}function Ai(T){var oA=on(T);if(oA&&g(c)){var YA=g(c).getBoundingClientRect(),pe=oA.getBoundingClientRect(),he=No(Ke(g(aA),T))?20:pe.height;pe.top<YA.top+20?d(oA,{container:g(c),offset:-20,duration:0}):pe.top+he>YA.bottom-20&&d(oA,{container:g(c),offset:-(YA.height-he-20),duration:0})}}function Ki(T,oA){if(T.json!==void 0||T?.text!==void 0){if(g(le)!==void 0){var YA,pe={text:g(le),json:void 0};(YA=QA())===null||YA===void 0||YA(pe,T,{contentErrors:ZA(),patchResult:oA})}else if(g(aA)!==void 0){var he,ge={text:void 0,json:g(aA)};(he=QA())===null||he===void 0||he(ge,T,{contentErrors:ZA(),patchResult:oA})}}}function dt(T,oA){i("handlePatch",T,oA);var YA={json:g(aA),text:g(le)},pe=Ot(T,oA);return Ki(YA,pe),pe}function EA(T,oA){var YA={json:g(aA),text:g(le)},pe={documentState:g(mA),selection:g(sA),json:g(aA),text:g(le),textIsRepaired:g(Ee)},he=Sl(g(aA),Qc(T,g(mA)),[],G3),ge=typeof oA=="function"?oA(T,he,g(sA)):void 0;y(aA,ge?.json!==void 0?ge.json:T),y(mA,ge?.state!==void 0?ge.state:he),y(sA,ge?.selection!==void 0?ge.selection:g(sA)),y(le,void 0),y(Ee,!1),SA=void 0,nn(g(aA)),mi(pe),Ki(YA,void 0)}function HA(T,oA){i("handleChangeText");var YA={json:g(aA),text:g(le)},pe={documentState:g(mA),selection:g(sA),json:g(aA),text:g(le),textIsRepaired:g(Ee)};try{y(aA,z()(T)),y(mA,Sl(g(aA),Qc(g(aA),g(mA)),[],G3)),y(le,void 0),y(Ee,!1),SA=void 0}catch(ge){try{y(aA,z()(Rc(T))),y(mA,Sl(g(aA),Qc(g(aA),g(mA)),[],G3)),y(le,T),y(Ee,!0),SA=void 0}catch{y(aA,void 0),y(mA,w_({json:g(aA),expand:G3})),y(le,T),y(Ee,!1),SA=g(le)!==""?dQ(g(le),ge.message||String(ge)):void 0}}if(typeof oA=="function"){var he=oA(g(aA),g(mA),g(sA));y(aA,he?.json!==void 0?he.json:g(aA)),y(mA,he?.state!==void 0?he.state:g(mA)),y(sA,he?.selection!==void 0?he.selection:g(sA))}nn(g(aA)),mi(pe),Ki(YA,void 0)}function ve(T,oA){var YA=arguments.length>2&&arguments[2]!==void 0&&arguments[2];i("handleExpand",{path:T,expanded:oA,recursive:YA}),oA?bA(T,YA?aG:ErA):Be(T,YA),Ri()}function Qt(){ve([],!0,!0)}function yi(){ve([],!1,!0)}function ri(T){i("openFind",{findAndReplace:T}),y(bt,!1),y(Re,!1),io(),y(bt,!0),y(Re,T)}function pn(T,oA){i("handleExpandSection",T,oA),y(mA,function(YA,pe,he,ge){return EQ(YA,pe,he,(Bt,$e)=>{if(!Mr($e))return $e;var bi=jsA($e.visibleSections.concat(ge));return fe(fe({},$e),{},{visibleSections:bi})})}(g(aA),g(mA),T,oA))}function Fn(T){i("pasted json as text",T),y(Dt,T)}function Jn(T){i("pasted multiline text",{pastedText:T}),y(_e,T)}function ln(T){var oA,{anchor:YA,left:pe,top:he,width:ge,height:Bt,offsetTop:$e,offsetLeft:bi,showTip:un}=T,Ji=function(Sn){var{json:no,documentState:gn,selection:Nt,readOnly:Zi,onEditKey:_t,onEditValue:st,onToggleEnforceString:si,onCut:Eo,onCopy:xr,onPaste:Hn,onRemove:So,onDuplicate:zr,onExtract:t0,onInsertBefore:Ta,onInsert:Wc,onConvert:Hl,onInsertAfter:Xc,onSort:Or,onTransform:Pr}=Sn,Ha=no!==void 0,i0=!!Nt,za=!!Nt&&Oi(et(Nt)),Ko=Nt?Ke(no,et(Nt)):void 0,$n=Array.isArray(Ko)?"Edit array":Cn(Ko)?"Edit object":"Edit value",Zo=Ha&&(Kn(Nt)||kr(Nt)||an(Nt)),zl=Nt&&!za?Ke(no,Fi(et(Nt))):void 0,Id=!Zi&&Ha&&fy(Nt)&&!za&&!Array.isArray(zl),Cd=!Zi&&Ha&&Nt!==void 0&&fy(Nt),JQ=Cd&&!No(Ko),dd=!Zi&&Zo,TQ=Zo,U7=!Zi&&i0,K7=!Zi&&Ha&&Zo&&!za,Y7=!Zi&&Ha&&Nt!==void 0&&(Kn(Nt)||an(Nt))&&!za,Ol=Zo,sI=Ol?"Convert to:":"Insert:",rr=!Zi&&(fr(Nt)&&Array.isArray(Ko)||Ua(Nt)&&Array.isArray(zl)),ta=!Zi&&(Ol?WD(Nt)&&!Cn(Ko):i0),HQ=!Zi&&(Ol?WD(Nt)&&!Array.isArray(Ko):i0),zQ=!Zi&&(Ol?WD(Nt)&&No(Ko):i0),aI=Nt!==void 0&&zg(no,gn,et(Nt));function is(OQ){Zo?OQ!=="structure"&&Hl(OQ):Wc(OQ)}return[{type:"row",items:[{type:"button",onClick:()=>_t(),icon:lC,text:"Edit key",title:"Edit the key (Double-click on the key)",disabled:!Id},{type:"dropdown-button",main:{type:"button",onClick:()=>st(),icon:lC,text:$n,title:"Edit the value (Double-click on the value)",disabled:!Cd},width:"11em",items:[{type:"button",icon:lC,text:$n,title:"Edit the value (Double-click on the value)",onClick:()=>st(),disabled:!Cd},{type:"button",icon:aI?jS:ZS,text:"Enforce string",title:"Enforce keeping the value as string when it contains a numeric value",onClick:()=>si(),disabled:!JQ}]}]},{type:"separator"},{type:"row",items:[{type:"dropdown-button",main:{type:"button",onClick:()=>Eo(!0),icon:cC,text:"Cut",title:"Cut selected contents, formatted with indentation (Ctrl+X)",disabled:!dd},width:"10em",items:[{type:"button",icon:cC,text:"Cut formatted",title:"Cut selected contents, formatted with indentation (Ctrl+X)",onClick:()=>Eo(!0),disabled:!dd},{type:"button",icon:cC,text:"Cut compacted",title:"Cut selected contents, without indentation (Ctrl+Shift+X)",onClick:()=>Eo(!1),disabled:!dd}]},{type:"dropdown-button",main:{type:"button",onClick:()=>xr(!0),icon:F0,text:"Copy",title:"Copy selected contents, formatted with indentation (Ctrl+C)",disabled:!TQ},width:"12em",items:[{type:"button",icon:F0,text:"Copy formatted",title:"Copy selected contents, formatted with indentation (Ctrl+C)",onClick:()=>xr(!0),disabled:!TQ},{type:"button",icon:F0,text:"Copy compacted",title:"Copy selected contents, without indentation (Ctrl+Shift+C)",onClick:()=>xr(!1),disabled:!TQ}]},{type:"button",onClick:()=>Hn(),icon:PS,text:"Paste",title:"Paste clipboard contents (Ctrl+V)",disabled:!U7}]},{type:"separator"},{type:"row",items:[{type:"column",items:[{type:"button",onClick:()=>zr(),icon:$S,text:"Duplicate",title:"Duplicate selected contents (Ctrl+D)",disabled:!K7},{type:"button",onClick:()=>t0(),icon:TW,text:"Extract",title:"Extract selected contents",disabled:!Y7},{type:"button",onClick:()=>Or(),icon:l4,text:"Sort",title:"Sort array or object contents",disabled:Zi||!Zo},{type:"button",onClick:()=>Pr(),icon:s4,text:"Transform",title:"Transform array or object contents (filter, sort, project)",disabled:Zi||!Zo},{type:"button",onClick:()=>So(),icon:E5,text:"Remove",title:"Remove selected contents (Delete)",disabled:Zi||!Zo}]},{type:"column",items:[{type:"label",text:sI},{type:"button",onClick:()=>is("structure"),icon:Ol?c4:gC,text:"Structure",title:sI+" structure like the first item in the array",disabled:!rr},{type:"button",onClick:()=>is("object"),icon:Ol?c4:gC,text:"Object",title:sI+" object",disabled:!ta},{type:"button",onClick:()=>is("array"),icon:Ol?c4:gC,text:"Array",title:sI+" array",disabled:!HQ},{type:"button",onClick:()=>is("value"),icon:Ol?c4:gC,text:"Value",title:sI+" value",disabled:!zQ}]}]},{type:"separator"},{type:"row",items:[{type:"button",onClick:()=>Ta(),icon:YW,text:"Insert before",title:"Select area before current entry to insert or paste contents",disabled:Zi||!Zo||za},{type:"button",onClick:()=>Xc(),icon:_W,text:"Insert after",title:"Select area after current entry to insert or paste contents",disabled:Zi||!Zo||za}]}]}({json:g(aA),documentState:g(mA),selection:g(sA),readOnly:B(),onEditKey:Lt,onEditValue:ii,onToggleEnforceString:_i,onCut:M,onCopy:ni,onPaste:mn,onRemove:Vo,onDuplicate:vo,onExtract:bo,onInsertBefore:wi,onInsert:Mo,onInsertAfter:MA,onConvert:ne,onSort:Di,onTransform:ft}),Tn=(oA=VA()(Ji))!==null&&oA!==void 0?oA:Ji;if(Tn!==!1){var Ht={left:pe,top:he,offsetTop:$e,offsetLeft:bi,width:ge,height:Bt,anchor:YA,closeOnOuterClick:!0,onClose:()=>{uA=!1,Ri()}};uA=!0;var ko=s(GaA,{tip:un?"Tip: you can open this context menu via right-click or with Ctrl+Q":void 0,items:Tn,onRequestClose:()=>a(ko)},Ht)}}function Pt(T){if(!br(g(sA)))if(T&&(T.stopPropagation(),T.preventDefault()),T&&T.type==="contextmenu"&&T.target!==g(l))ln({left:T.clientX,top:T.clientY,width:W0,height:Z0,showTip:!1});else{var oA,YA=(oA=g(c))===null||oA===void 0?void 0:oA.querySelector(".jse-context-menu-pointer.jse-selected");if(YA)ln({anchor:YA,offsetTop:2,width:W0,height:Z0,showTip:!1});else{var pe,he=(pe=g(c))===null||pe===void 0?void 0:pe.getBoundingClientRect();he&&ln({top:he.top+2,left:he.left+2,width:W0,height:Z0,showTip:!1})}}}function $i(T){ln({anchor:OsA(T.target,"BUTTON"),offsetTop:0,width:W0,height:Z0,showTip:!0})}function Rr(){return Ft.apply(this,arguments)}function Ft(){return(Ft=Yt(function*(){if(i("apply pasted json",g(Dt)),g(Dt)){var{onPasteAsJson:T}=g(Dt);y(Dt,void 0),T(),setTimeout(Ri)}})).apply(this,arguments)}function Xn(){return se.apply(this,arguments)}function se(){return(se=Yt(function*(){i("apply pasted multiline text",g(_e)),g(_e)&&(kn(JSON.stringify(g(_e))),setTimeout(Ri))})).apply(this,arguments)}function vi(){i("clear pasted json"),y(Dt,void 0),Ri()}function Yi(){i("clear pasted multiline text"),y(_e,void 0),Ri()}function rn(){BA()(er.text)}function Hr(T){y(sA,T),Ri(),ot(et(T))}function Ri(){i("focus"),g(l)&&(g(l).focus(),g(l).select())}function fs(T){return function(oA,YA,pe){var he=Fi(pe),ge=[hi(pe)],Bt=Ke(oA,he),$e=Bt?VN(Bt,YA,ge):void 0;return $e?Ni(he.concat($e)):t2(pe)}(g(aA),g(mA),T)}function Bo(T){g(A)&&g(A).onDrag(T)}function Q(){g(A)&&g(A).onDragEnd()}var m=$(void 0,!0);fA(()=>g(sA),()=>{var T;T=g(sA),di(T,h())||(i("onSelect",T),lA()(T))}),fA(()=>(k(w()),k(_())),()=>{y(de,V_({escapeControlCharacters:w(),escapeUnicodeCharacters:_()}))}),fA(()=>g(bt),()=>{(function(T){g(c)&&T&&g(c).scrollTop===0&&(hc(c,g(c).style.overflowAnchor="none"),hc(c,g(c).scrollTop+=_3),setTimeout(()=>{g(c)&&hc(c,g(c).style.overflowAnchor="")}))})(g(bt))}),fA(()=>k(E()),()=>{qi(E())}),fA(()=>k(h()),()=>{(function(T){di(g(sA),T)||(i("applyExternalSelection",{selection:g(sA),externalSelection:T}),q3(T)&&y(sA,T))})(h())}),fA(()=>(g(aA),k(U()),k(K()),k(H())),()=>{Ve(g(aA),U(),K(),H())}),fA(()=>(g(c),UrA),()=>{y(A,g(c)?UrA(g(c)):void 0)}),fA(()=>(k(B()),k(D()),k(K()),g(de),k(cA()),k(oe())),()=>{y(m,{mode:er.tree,readOnly:B(),truncateTextSize:D(),parser:K(),normalization:g(de),getJson:rt,getDocumentState:Ei,getSelection:tn,findElement:on,findNextInside:fs,focus:Ri,onPatch:dt,onInsert:Yn,onExpand:ve,onSelect:xt,onFind:ri,onExpandSection:pn,onPasteJson:Fn,onRenderValue:cA(),onContextMenu:ln,onClassName:oe()||(()=>{}),onDrag:Bo,onDragEnd:Q})}),fA(()=>g(m),()=>{i("context changed",g(m))}),Qn(),Zt(!0);var v=kJA();ce("mousedown",A2,function(T){!yQ(T.target,oA=>oA===g(I))&&br(g(sA))&&(i("click outside the editor, exit edit mode"),y(sA,Hg(g(sA))),C&&g(l)&&(g(l).focus(),g(l).blur()),i("blur (outside editor)"),g(l)&&g(l).blur())});var p,N=vt(v),J=W(N),V=T=>{(function(oA,YA){mt(YA,!1);var pe=$(void 0,!0),he=$(void 0,!0),ge=$(void 0,!0),Bt=b(YA,"json",9),$e=b(YA,"selection",9),bi=b(YA,"readOnly",9),un=b(YA,"showSearch",13,!1),Ji=b(YA,"history",9),Tn=b(YA,"onExpandAll",9),Ht=b(YA,"onCollapseAll",9),ko=b(YA,"onUndo",9),Sn=b(YA,"onRedo",9),no=b(YA,"onSort",9),gn=b(YA,"onTransform",9),Nt=b(YA,"onContextMenu",9),Zi=b(YA,"onCopy",9),_t=b(YA,"onRenderMenu",9);function st(){un(!un())}var si=$(void 0,!0),Eo=$(void 0,!0),xr=$(void 0,!0),Hn=$(void 0,!0);fA(()=>k(Bt()),()=>{y(pe,Bt()!==void 0)}),fA(()=>(g(pe),k($e()),an),()=>{y(he,g(pe)&&(Kn($e())||kr($e())||an($e())))}),fA(()=>(k(Tn()),k(Bt())),()=>{y(si,{type:"button",icon:_YA,title:"Expand all",className:"jse-expand-all",onClick:Tn(),disabled:!No(Bt())})}),fA(()=>(k(Ht()),k(Bt())),()=>{y(Eo,{type:"button",icon:GYA,title:"Collapse all",className:"jse-collapse-all",onClick:Ht(),disabled:!No(Bt())})}),fA(()=>k(Bt()),()=>{y(xr,{type:"button",icon:g4,title:"Search (Ctrl+F)",className:"jse-search",onClick:st,disabled:Bt()===void 0})}),fA(()=>(k(bi()),g(si),g(Eo),k(no()),k(Bt()),k(gn()),g(xr),k(Nt()),k(ko()),k(Ji()),k(Sn()),k(Zi()),g(he)),()=>{y(Hn,bi()?[g(si),g(Eo),{type:"separator"},{type:"button",icon:F0,title:"Copy (Ctrl+C)",className:"jse-copy",onClick:Zi(),disabled:!g(he)},{type:"separator"},g(xr),{type:"space"}]:[g(si),g(Eo),{type:"separator"},{type:"button",icon:l4,title:"Sort",className:"jse-sort",onClick:no(),disabled:bi()||Bt()===void 0},{type:"button",icon:s4,title:"Transform contents (filter, sort, project)",className:"jse-transform",onClick:gn(),disabled:bi()||Bt()===void 0},g(xr),{type:"button",icon:WS,title:AG,className:"jse-contextmenu",onClick:Nt()},{type:"separator"},{type:"button",icon:h5,title:"Undo (Ctrl+Z)",className:"jse-undo",onClick:ko(),disabled:!Ji().canUndo},{type:"button",icon:Q5,title:"Redo (Ctrl+Shift+Z)",className:"jse-redo",onClick:Sn(),disabled:!Ji().canRedo},{type:"space"}])}),fA(()=>(k(_t()),g(Hn)),()=>{y(ge,_t()(g(Hn))||g(Hn))}),Qn(),Zt(!0),Oy(oA,{get items(){return g(ge)}}),pt()})(T,{get json(){return g(aA)},get selection(){return g(sA)},get readOnly(){return B()},get history(){return u()},onExpandAll:Qt,onCollapseAll:yi,onUndo:Wt,onRedo:Xe,onSort:Ut,onTransform:Qi,onContextMenu:$i,onCopy:ni,get onRenderMenu(){return pA()},get showSearch(){return g(bt)},set showSearch(oA){y(bt,oA)},$$legacy:!0})};LA(J,T=>{L()&&T(V)});var Z=IA(J,2),FA=T=>{tJA(T,{get json(){return g(aA)},get selection(){return g(sA)},onSelect:Hr,get onError(){return gA()},get pathParser(){return q()}})};LA(Z,T=>{R()&&T(FA)});var te=IA(Z,2),re=T=>{var oA=bJA(),YA=vt(oA),pe=W(YA);pe.readOnly=!0,Co(pe,$e=>y(l,$e),()=>g(l));var he=IA(YA,2),ge=$e=>{var bi=_o(),un=vt(bi),Ji=Ht=>{(function(ko,Sn){mt(Sn,!0);var no=HYA();no.__click=[JYA,Sn];var gn=IA(W(no),2),Nt=IA(W(gn),2),Zi=_t=>{var st=TYA(),si=IA(vt(st),2);En(si,"title","Create an empty JSON object (press '{')"),si.__click=[KYA,Sn];var Eo=IA(si,2);En(Eo,"title","Create an empty JSON array (press '[')"),Eo.__click=[YYA,Sn],iA(_t,st)};LA(Nt,_t=>{Sn.readOnly||_t(Zi)}),iA(ko,no),pt()})(Ht,{get readOnly(){return B()},onCreateObject:()=>{Ri(),me("{")},onCreateArray:()=>{Ri(),me("[")},onClick:()=>{Ri()}})},Tn=Ht=>{var ko=DJA(),Sn=vt(ko),no=qA(()=>B()?[]:[{icon:a4,text:"Repair manually",title:'Open the document in "code" mode and repair it manually',onClick:rn}]);mc(Sn,{type:"error",message:"The loaded JSON document is invalid and could not be repaired automatically.",get actions(){return g(no)}}),_aA(IA(Sn,2),{get text(){return g(le)},get json(){return g(aA)},get indentation(){return j()},get parser(){return K()}}),iA(Ht,ko)};LA(un,Ht=>{g(le)===""||g(le)===void 0?Ht(Ji):Ht(Tn,!1)}),iA($e,bi)},Bt=$e=>{var bi=vJA(),un=vt(bi);MaA(W(un),{get json(){return g(aA)},get documentState(){return g(mA)},get parser(){return K()},get showSearch(){return g(bt)},get showReplace(){return g(Re)},get readOnly(){return B()},columns:void 0,onSearch:$t,onFocus:x,onPatch:dt,onClose:P});var Ji=IA(un,2);En(Ji,"data-jsoneditor-scrollable-contents",!0);var Tn=W(Ji),Ht=_t=>{iA(_t,yJA())};LA(Tn,_t=>{g(bt)&&_t(Ht)}),x_(IA(Tn,2),{get value(){return g(aA)},pointer:"",get state(){return g(mA)},get validationErrors(){return g(DA)},get searchResults(){return g(Le)},get selection(){return g(sA)},get context(){return g(m)},get onDragSelectionStart(){return $o}}),Co(Ji,_t=>y(c,_t),()=>g(c));var ko=IA(Ji,2),Sn=_t=>{var st=qA(()=>(g(Dt),nA(()=>"You pasted a JSON ".concat(Array.isArray(g(Dt).contents)?"array":"object"," as text")))),si=qA(()=>[{icon:L0,text:"Paste as JSON instead",title:"Replace the value with the pasted JSON",onMouseDown:Rr},{text:"Leave as is",title:"Keep the JSON embedded in the value",onClick:vi}]);mc(_t,{type:"info",get message(){return g(st)},get actions(){return g(si)}})};LA(ko,_t=>{g(Dt)&&_t(Sn)});var no=IA(ko,2),gn=_t=>{var st=qA(()=>[{icon:L0,text:"Paste as string instead",title:"Paste the clipboard data as a single string value instead of an array",onClick:Xn},{text:"Leave as is",title:"Keep the pasted array",onClick:Yi}]);mc(_t,{type:"info",message:"Multiline text was pasted as array",get actions(){return g(st)}})};LA(no,_t=>{g(_e)&&_t(gn)});var Nt=IA(no,2),Zi=_t=>{var st=qA(()=>B()?[]:[{icon:u5,text:"Ok",title:"Accept the repaired document",onClick:Tt},{icon:a4,text:"Repair manually instead",title:"Leave the document unchanged and repair it manually instead",onClick:rn}]);mc(_t,{type:"success",message:"The loaded JSON document was invalid but is successfully repaired.",get actions(){return g(st)},onClose:Ri})};LA(Nt,_t=>{g(Ee)&&_t(Zi)}),dG(IA(Nt,2),{get validationErrors(){return g(kA)},selectError:X}),iA($e,bi)};LA(he,$e=>{g(aA)===void 0?$e(ge):$e(Bt,!1)}),ce("paste",pe,dn),iA(T,oA)},Pe=T=>{iA(T,MJA())};LA(te,T=>{n?T(Pe,!1):T(re)}),Co(N,T=>y(I,T),()=>g(I));var ze=IA(N,2),ye=T=>{DaA(T,{onClose:()=>y(eA,!1)})};LA(ze,T=>{g(eA)&&T(ye)});var Ge=IA(ze,2),Vi=T=>{yaA(T,z1(()=>g(UA),{onClose:()=>{var oA;(oA=g(UA))===null||oA===void 0||oA.onClose(),y(UA,void 0)}}))};return LA(Ge,T=>{g(UA)&&T(Vi)}),De(T=>p=Vt(N,1,"jse-tree-mode svelte-vrx1dr",null,p,T),[()=>({"no-main-menu":!L()})],qA),ce("keydown",N,function(T){var oA=n2(T),YA=T.shiftKey;if(i("keydown",{combo:oA,key:T.key}),oA==="Ctrl+X"&&(T.preventDefault(),M(!0)),oA==="Ctrl+Shift+X"&&(T.preventDefault(),M(!1)),oA==="Ctrl+C"&&(T.preventDefault(),ni(!0)),oA==="Ctrl+Shift+C"&&(T.preventDefault(),ni(!1)),oA==="Ctrl+D"&&(T.preventDefault(),vo()),oA!=="Delete"&&oA!=="Backspace"||(T.preventDefault(),Vo()),oA==="Insert"&&(T.preventDefault(),Yn("structure")),oA==="Ctrl+A"&&(T.preventDefault(),y(sA,Ni([]))),oA==="Ctrl+Q"&&Pt(T),oA==="ArrowUp"||oA==="Shift+ArrowUp"){T.preventDefault();var pe=g(sA)?QrA(g(aA),g(mA),g(sA),YA)||g(sA):WE(g(aA),g(mA));y(sA,pe),Ai(et(pe))}if(oA==="ArrowDown"||oA==="Shift+ArrowDown"){T.preventDefault();var he=g(sA)?function(Ji,Tn,Ht){var ko=arguments.length>3&&arguments[3]!==void 0&&arguments[3];if(Ht){var Sn=ko?et(Ht):P1(Ji,Ht),no=No(Ke(Ji,Sn))?IrA(Ji,Tn,Sn,!0):Tn,gn=VN(Ji,Tn,Sn),Nt=VN(Ji,no,Sn);if(ko)return fr(Ht)?gn!==void 0?_s(gn,gn):void 0:Ua(Ht)?Nt!==void 0?_s(Nt,Nt):void 0:Nt!==void 0?_s(OC(Ht),Nt):void 0;if(Ua(Ht))return Nt!==void 0?Ni(Nt):void 0;if(fr(Ht)||an(Ht))return gn!==void 0?Ni(gn):void 0;if(kr(Ht)){if(gn===void 0||gn.length===0)return;var Zi=Fi(gn),_t=Ke(Ji,Zi);return Array.isArray(_t)?Ni(gn):o2(gn)}return Kn(Ht)?Nt!==void 0?Ni(Nt):gn!==void 0?Ni(gn):void 0:void 0}}(g(aA),g(mA),g(sA),YA)||g(sA):WE(g(aA),g(mA));y(sA,he),Ai(et(he))}if(oA==="ArrowLeft"||oA==="Shift+ArrowLeft"){T.preventDefault();var ge=g(sA)?function(Ji,Tn,Ht){var ko=arguments.length>3&&arguments[3]!==void 0&&arguments[3],Sn=!(arguments.length>4&&arguments[4]!==void 0)||arguments[4];if(Ht){var{caret:no,previous:gn}=hrA(Ji,Tn,Ht,Sn);if(ko)return Kn(Ht)?void 0:_s(Ht.path,Ht.path);if(no&&gn)return D_(gn);var Nt=Fi(et(Ht)),Zi=Ke(Ji,Nt);return an(Ht)&&Array.isArray(Zi)?_s(Ht.path,Ht.path):Kn(Ht)&&!Array.isArray(Zi)?o2(Ht.focusPath):void 0}}(g(aA),g(mA),g(sA),YA,!B())||g(sA):WE(g(aA),g(mA));y(sA,ge),Ai(et(ge))}if(oA==="ArrowRight"||oA==="Shift+ArrowRight"){T.preventDefault();var Bt=g(sA)&&g(aA)!==void 0?function(Ji,Tn,Ht){var ko=arguments.length>3&&arguments[3]!==void 0&&arguments[3],Sn=!(arguments.length>4&&arguments[4]!==void 0)||arguments[4];if(Ht){var{caret:no,next:gn}=hrA(Ji,Tn,Ht,Sn);return ko?Kn(Ht)?void 0:_s(Ht.path,Ht.path):no&&gn?D_(gn):Kn(Ht)?Ni(Ht.focusPath):void 0}}(g(aA),g(mA),g(sA),YA,!B())||g(sA):WE(g(aA),g(mA));y(sA,Bt),Ai(et(Bt))}if(oA==="Enter"&&g(sA)){if(Yy(g(sA))){var $e=g(sA).focusPath,bi=Ke(g(aA),Fi($e));Array.isArray(bi)&&(T.preventDefault(),y(sA,Ni($e)))}kr(g(sA))&&(T.preventDefault(),y(sA,fe(fe({},g(sA)),{},{edit:!0}))),an(g(sA))&&(T.preventDefault(),No(Ke(g(aA),g(sA).path))?ve(g(sA).path,!0):y(sA,fe(fe({},g(sA)),{},{edit:!0})))}if(oA.replace(/^Shift\+/,"").length===1&&g(sA))return T.preventDefault(),void me(T.key);if(oA==="Enter"&&(Ua(g(sA))||fr(g(sA))))return T.preventDefault(),void me("");if(oA==="Ctrl+Enter"&&an(g(sA))){var un=Ke(g(aA),g(sA).path);Uy(un)&&window.open(String(un),"_blank")}oA==="Escape"&&g(sA)&&(T.preventDefault(),y(sA,void 0)),oA==="Ctrl+F"&&(T.preventDefault(),ri(!1)),oA==="Ctrl+H"&&(T.preventDefault(),ri(!0)),oA==="Ctrl+Z"&&(T.preventDefault(),Wt()),oA==="Ctrl+Shift+Z"&&(T.preventDefault(),Xe())}),ce("mousedown",N,function(T){i("handleMouseDown",T);var oA=T.target;zsA(oA,"BUTTON")||oA.isContentEditable||(Ri(),g(sA)||g(aA)!==void 0||g(le)!==""&&g(le)!==void 0||(i("createDefaultSelection"),y(sA,Ni([]))))}),ce("contextmenu",N,Pt),iA(t,v),zt(e,"expand",bA),zt(e,"collapse",Be),zt(e,"validate",ZA),zt(e,"getJson",rt),zt(e,"patch",Ot),zt(e,"acceptAutoRepair",Tt),zt(e,"openTransformModal",cn),zt(e,"scrollTo",ot),zt(e,"findElement",on),zt(e,"findSearchResult",hn),zt(e,"focus",Ri),pt({expand:bA,collapse:Be,validate:ZA,getJson:rt,patch:Ot,acceptAutoRepair:Tt,openTransformModal:cn,scrollTo:ot,findElement:on,findSearchResult:hn,focus:Ri})}function UaA(t){return typeof(e=t)!="object"||e===null?t:new Proxy(t,{get:(A,i,n)=>UaA(Reflect.get(A,i,n)),set:()=>!1,deleteProperty:()=>!1});var e}var oy=Sr("jsoneditor:History");function KaA(){var t=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{},e=t.maxItems||1e3,A=[],i=0;function n(){return i<A.length}function o(){return i>0}function r(){return{canUndo:n(),canRedo:o(),items:()=>A.slice().reverse(),add:a,undo:l,redo:I,clear:c}}function s(){t.onChange&&t.onChange(r())}function a(C){oy("add",C),A=[C].concat(A.slice(i)).slice(0,e),i=0,s()}function c(){oy("clear"),A=[],i=0,s()}function l(){if(n()){var C=A[i];return i+=1,oy("undo",C),s(),C}}function I(){if(o())return oy("redo",A[i-=1]),s(),A[i]}return{get:r}}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-transform-modal-inner.svelte-rrrjnb {
  flex: 1;
  display: flex;
  flex-direction: column;
  min-width: 0;
  min-height: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) {
  color: inherit;
  flex: 1;
  display: flex;
  flex-direction: column;
  padding: 0;
  overflow: auto;
  min-width: 0;
  min-height: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-actions:where(.svelte-rrrjnb) {
  display: flex;
  flex-direction: row;
  justify-content: flex-end;
  padding-top: var(--jse-padding, 10px);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-actions:where(.svelte-rrrjnb) button.jse-primary:where(.svelte-rrrjnb) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-actions:where(.svelte-rrrjnb) button.jse-primary:where(.svelte-rrrjnb):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-actions:where(.svelte-rrrjnb) button.jse-primary:where(.svelte-rrrjnb):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) {
  flex: 1;
  display: flex;
  gap: calc(2 * var(--jse-padding, 10px));
  min-height: 0;
  box-sizing: border-box;
  padding: 0 calc(2 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) {
  flex: 1;
  display: flex;
  flex-direction: column;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) .jse-description:where(.svelte-rrrjnb) p {
  margin: var(--jse-padding, 10px) 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) .jse-description:where(.svelte-rrrjnb) p:first-child {
  margin-top: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) .jse-description:where(.svelte-rrrjnb) p:last-child {
  margin-bottom: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) .jse-description:where(.svelte-rrrjnb) code {
  background: var(--jse-modal-code-background, rgba(0, 0, 0, 0.05));
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) .query-error:where(.svelte-rrrjnb) {
  color: var(--jse-error-color, #ee5341);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) textarea.jse-query:where(.svelte-rrrjnb) {
  flex: 1;
  outline: none;
  resize: vertical;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents:where(.svelte-rrrjnb) {
  flex: 1;
  display: flex;
  flex-direction: column;
  gap: calc(2 * var(--jse-padding, 10px));
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents:where(.svelte-rrrjnb) .jse-original-data:where(.svelte-rrrjnb) {
  flex: 1;
  display: flex;
  flex-direction: column;
  min-height: 0;
  box-sizing: border-box;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents:where(.svelte-rrrjnb) .jse-original-data.jse-hide:where(.svelte-rrrjnb) {
  flex: none;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents:where(.svelte-rrrjnb) .jse-preview-data:where(.svelte-rrrjnb) {
  flex: 1;
  display: flex;
  flex-direction: column;
  min-height: 0;
  box-sizing: border-box;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents.jse-hide-original-data:where(.svelte-rrrjnb) {
  flex-direction: column;
  gap: 0;
  margin-bottom: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-actions:where(.svelte-rrrjnb) {
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px)) calc(2 * var(--jse-padding, 10px));
}
@media screen and (max-width: 1200px) {
  .jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) {
    flex-direction: column;
    overflow: auto;
  }
  .jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-query-contents:where(.svelte-rrrjnb) textarea.jse-query:where(.svelte-rrrjnb) {
    min-height: 150px;
    flex: none;
  }
  .jse-transform-modal-inner.svelte-rrrjnb .jse-modal-contents:where(.svelte-rrrjnb) .jse-main-contents:where(.svelte-rrrjnb) .jse-data-contents:where(.svelte-rrrjnb) .jse-tree-mode {
    height: 300px;
    flex: none;
  }
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-label:where(.svelte-rrrjnb) {
  font-weight: bold;
  display: block;
  box-sizing: border-box;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-label:where(.svelte-rrrjnb) .jse-label-inner:where(.svelte-rrrjnb) {
  margin-top: calc(2 * var(--jse-padding, 10px));
  margin-bottom: calc(0.5 * var(--jse-padding, 10px));
  box-sizing: border-box;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-label:where(.svelte-rrrjnb) .jse-label-inner:where(.svelte-rrrjnb) button:where(.svelte-rrrjnb) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  font-weight: bold;
  padding: 0;
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-tree-mode {
  flex: 1;
  background: var(--jse-input-background-readonly, transparent);
  box-shadow: none;
  box-sizing: border-box;
  --jse-main-border: var(--jse-input-border, 1px solid #d8dbdf);
}
.jse-transform-modal-inner.svelte-rrrjnb input:where(.svelte-rrrjnb),
.jse-transform-modal-inner.svelte-rrrjnb textarea:where(.svelte-rrrjnb) {
  border: var(--jse-input-border, 1px solid #d8dbdf);
  outline: none;
  box-sizing: border-box;
  padding: calc(0.5 * var(--jse-padding, 10px));
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: inherit;
  background: var(--jse-input-background, var(--jse-background-color, #fff));
}
.jse-transform-modal-inner.svelte-rrrjnb input:where(.svelte-rrrjnb):focus,
.jse-transform-modal-inner.svelte-rrrjnb textarea:where(.svelte-rrrjnb):focus {
  border: var(--jse-input-border-focus, 1px solid var(--jse-input-border-focus, var(--jse-theme-color, #3883fa)));
}
.jse-transform-modal-inner.svelte-rrrjnb input:where(.svelte-rrrjnb):read-only,
.jse-transform-modal-inner.svelte-rrrjnb textarea:where(.svelte-rrrjnb):read-only {
  background: var(--jse-input-background-readonly, transparent);
}
.jse-transform-modal-inner.svelte-rrrjnb .jse-preview.jse-error:where(.svelte-rrrjnb) {
  flex: 1;
  background: var(--jse-input-background-readonly, transparent);
  border: var(--jse-input-border, 1px solid #d8dbdf);
  color: var(--jse-error-color, #ee5341);
  padding: calc(0.5 * var(--jse-padding, 10px));
}
.jse-transform-modal-inner.svelte-rrrjnb a {
  color: var(--jse-a-color, #156fc5);
}
.jse-transform-modal-inner.svelte-rrrjnb a:hover {
  color: var(--jse-a-color-highlight, #0f508d);
}`);var L3=Gy(()=>xUA),$E=Gy(()=>LUA),SJA=wA('<div class="query-error svelte-rrrjnb"> </div>'),RJA=wA("<!> <!>",1),xJA=wA('<div class="jse-preview jse-error svelte-rrrjnb"> </div>'),LJA=wA('<!> <div class="jse-modal-contents svelte-rrrjnb"><div class="jse-main-contents svelte-rrrjnb"><div class="jse-query-contents svelte-rrrjnb"><div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb">Language</div></div> <div class="jse-description svelte-rrrjnb"><!></div> <div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb">Path</div></div> <input class="jse-path svelte-rrrjnb" type="text" readonly="" title="Selected path"/> <div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb"><button type="button" class="svelte-rrrjnb"><!> Wizard</button></div></div> <!> <div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb">Query</div></div> <textarea class="jse-query svelte-rrrjnb" spellcheck="false"></textarea></div> <div><div><div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb"><button type="button" class="svelte-rrrjnb"><!> Original</button></div></div> <!></div> <div class="jse-preview-data svelte-rrrjnb"><div class="jse-label svelte-rrrjnb"><div class="jse-label-inner svelte-rrrjnb">Preview</div></div> <!></div></div></div> <div class="jse-actions svelte-rrrjnb"><button type="button" class="jse-primary svelte-rrrjnb">Transform</button></div></div>',1),FJA=wA('<div class="jse-transform-modal-inner svelte-rrrjnb"><!></div>');function NJA(t,e){var A,i,n;mt(e,!1);var o=Sr("jsoneditor:TransformModal"),r=b(e,"id",25,()=>"transform-modal-"+nQ()),s=b(e,"json",9),a=b(e,"rootPath",25,()=>[]),c=b(e,"indentation",9),l=b(e,"truncateTextSize",9),I=b(e,"escapeControlCharacters",9),C=b(e,"escapeUnicodeCharacters",9),d=b(e,"parser",9),B=b(e,"parseMemoizeOne",9),E=b(e,"validationParser",9),h=b(e,"pathParser",9),u=b(e,"queryLanguages",9),D=b(e,"queryLanguageId",13),L=b(e,"onChangeQueryLanguage",9),R=b(e,"onRenderValue",9),w=b(e,"onRenderMenu",9),_=b(e,"onRenderContextMenu",9),K=b(e,"onClassName",9),z=b(e,"onTransform",9),U=b(e,"onClose",9),H=$(void 0,!0),q=$(KaA({onChange:mA=>y(q,mA)}).get(),!0),j=$(void 0,!0),gA=$(void 0,!0),QA=$(!1,!0),BA="".concat(r(),":").concat(Ct(a())),lA=(A=L3()[BA])!==null&&A!==void 0?A:{},vA=$($E().showWizard!==!1,!0),tA=$($E().showOriginal!==!1,!0),cA=$((i=lA.queryOptions)!==null&&i!==void 0?i:{},!0),pA=$(D()===lA.queryLanguageId&&lA.query?lA.query:"",!0),VA=$((n=lA.isManual)!==null&&n!==void 0&&n,!0),oe=$(void 0,!0),KA=$(void 0,!0),CA=$({text:""},!0);function TA(mA){var sA;return(sA=u().find(xt=>xt.id===mA))!==null&&sA!==void 0?sA:u()[0]}function Ze(mA){try{y(cA,mA),y(pA,TA(D()).createQuery(g(j),mA)),y(oe,void 0),y(VA,!1),o("updateQueryByWizard",{queryOptions:g(cA),query:g(pA),isManual:g(VA)})}catch(sA){y(oe,String(sA))}}function He(mA){y(pA,mA.target.value),y(VA,!0),o("handleChangeQuery",{query:g(pA),isManual:g(VA)})}g(VA)||Ze(g(cA)),hs(()=>{var mA;(mA=g(H))===null||mA===void 0||mA.focus()});var uA=oE(function(mA,sA){if(mA===void 0)return y(CA,{text:""}),void y(KA,"Error: No JSON");if(sA.trim()!=="")try{o("previewTransform",{query:sA});var xt=TA(D()).executeQuery(mA,sA,d());y(CA,{json:xt}),y(KA,void 0)}catch(tt){y(CA,{text:""}),y(KA,String(tt))}else y(CA,{json:mA})},300);function eA(){if(g(j)===void 0)return y(CA,{text:""}),void y(KA,"Error: No JSON");try{o("handleTransform",{query:g(pA)});var mA=TA(D()).executeQuery(g(j),g(pA),d());z()([{op:"replace",path:Ct(a()),value:mA}]),U()()}catch(sA){console.error(sA),y(CA,{text:""}),y(KA,String(sA))}}function UA(){y(vA,!g(vA)),$E($E().showWizard=g(vA))}function aA(){y(tA,!g(tA)),$E($E().showOriginal=g(tA))}function le(mA){mA.focus()}function SA(mA){o("handleChangeQueryLanguage",mA),D(mA),L()(mA),Ze(g(cA))}function Ue(){g(QA)?y(QA,!g(QA)):U()()}fA(()=>(k(s()),k(a())),()=>{y(j,UaA(Ke(s(),a())))}),fA(()=>g(j),()=>{y(gA,g(j)?{json:g(j)}:{text:""})}),fA(()=>(g(j),g(pA)),()=>{uA(g(j),g(pA))}),fA(()=>(L3(),g(cA),g(pA),k(D()),g(VA)),()=>{L3(L3()[BA]={queryOptions:g(cA),query:g(pA),queryLanguageId:D(),isManual:g(VA)}),o("store state in memory",BA,L3()[BA])}),Qn(),Zt(!0),X3(t,{get onClose(){return U()},className:"jse-transform-modal",get fullscreen(){return g(QA)},children:(mA,sA)=>{var xt=FJA();f_(W(xt),{children:(tt,de)=>{var Dt=LJA(),_e=vt(Dt);(function(M,We){mt(We,!1);var ni,pi=b(We,"queryLanguages",9),dn=b(We,"queryLanguageId",9),mn=b(We,"fullscreen",13),Uo=b(We,"onChangeQueryLanguage",9),kn=b(We,"onClose",9),Wn=$(void 0,!0),{openAbsolutePopup:Vo,closeAbsolutePopup:vo}=$1("absolute-popup");function bo(){var Yn={queryLanguages:pi(),queryLanguageId:dn(),onChangeQueryLanguage:Mo=>{vo(ni),Uo()(Mo)}};ni=Vo(MKA,Yn,{offsetTop:-2,offsetLeft:0,anchor:g(Wn),closeOnOuterClick:!0})}Zt(!0),My(M,{title:"Transform",fullScreenButton:!0,get onClose(){return kn()},get fullscreen(){return mn()},set fullscreen(Yn){mn(Yn)},$$slots:{actions:(Yn,Mo)=>{var ne,wi=RKA();ji(W(wi),{get data(){return HW}}),Co(wi,MA=>y(Wn,MA),()=>g(Wn)),De(MA=>ne=Vt(wi,1,"jse-config svelte-1kpylsp",null,ne,MA),[()=>({hide:pi().length<=1})],qA),ce("click",wi,bo),iA(Yn,wi)}},$$legacy:!0}),pt()})(_e,{get queryLanguages(){return u()},get queryLanguageId(){return D()},onChangeQueryLanguage:SA,get onClose(){return U()},get fullscreen(){return g(QA)},set fullscreen(M){y(QA,M)},$$legacy:!0});var Le=W(IA(_e,2)),bt=W(Le),Re=IA(W(bt),2);bsA(W(Re),()=>(k(D()),nA(()=>TA(D()).description)));var $t=IA(Re,4),x=IA($t,2),Y=W(x),P=W(Y),X=W(P),bA=qA(()=>g(vA)?mg:sE);ji(X,{get data(){return g(bA)}});var Be=IA(x,2),Ee=M=>{var We=_o(),ni=vt(We),pi=mn=>{var Uo=RJA(),kn=vt(Uo);yKA(kn,{get queryOptions(){return g(cA)},get json(){return g(j)},onChange:Ze});var Wn=IA(kn,2),Vo=vo=>{var bo=SJA(),Yn=W(bo);De(()=>wt(Yn,g(oe))),iA(vo,bo)};LA(Wn,vo=>{g(oe)&&vo(Vo)}),iA(mn,Uo)},dn=mn=>{iA(mn,Tr("(Only available for arrays, not for objects)"))};LA(ni,mn=>{g(j),nA(()=>Array.isArray(g(j)))?mn(pi):mn(dn,!1)}),iA(M,We)};LA(Be,M=>{g(vA)&&M(Ee)});var kA=IA(Be,4);Co(kA,M=>y(H,M),()=>g(H));var DA,gt,Ve=IA(bt,2),ZA=W(Ve),rt=W(ZA),Ei=W(rt),tn=W(Ei),qi=W(tn),xe=qA(()=>g(tA)?mg:sE);ji(qi,{get data(){return g(xe)}});var nn=IA(rt,2),mi=M=>{U_(M,{get externalContent(){return g(gA)},externalSelection:void 0,get history(){return g(q)},readOnly:!0,get truncateTextSize(){return l()},mainMenuBar:!1,navigationBar:!1,get indentation(){return c()},get escapeControlCharacters(){return I()},get escapeUnicodeCharacters(){return C()},get parser(){return d()},get parseMemoizeOne(){return B()},get onRenderValue(){return R()},get onRenderMenu(){return w()},get onRenderContextMenu(){return _()},onError:nA(()=>console.error),get onChange(){return $o},get onChangeMode(){return $o},get onSelect(){return $o},get onUndo(){return $o},get onRedo(){return $o},get onFocus(){return $o},get onBlur(){return $o},get onSortModal(){return $o},get onTransformModal(){return $o},get onJSONEditorModal(){return $o},get onClassName(){return K()},validator:void 0,get validationParser(){return E()},get pathParser(){return h()}})};LA(nn,M=>{g(tA)&&M(mi)});var Ot=IA(ZA,2),Lt=IA(W(Ot),2),ii=M=>{U_(M,{get externalContent(){return g(CA)},externalSelection:void 0,get history(){return g(q)},readOnly:!0,get truncateTextSize(){return l()},mainMenuBar:!1,navigationBar:!1,get indentation(){return c()},get escapeControlCharacters(){return I()},get escapeUnicodeCharacters(){return C()},get parser(){return d()},get parseMemoizeOne(){return B()},get onRenderValue(){return R()},get onRenderMenu(){return w()},get onRenderContextMenu(){return _()},onError:nA(()=>console.error),get onChange(){return $o},get onChangeMode(){return $o},get onSelect(){return $o},get onUndo(){return $o},get onRedo(){return $o},get onFocus(){return $o},get onBlur(){return $o},get onSortModal(){return $o},get onTransformModal(){return $o},get onJSONEditorModal(){return $o},get onClassName(){return K()},validator:void 0,get validationParser(){return E()},get pathParser(){return h()}})},_i=M=>{var We=xJA(),ni=W(We);De(()=>wt(ni,g(KA))),iA(M,We)};LA(Lt,M=>{g(KA)?M(_i,!1):M(ii)});var Tt=W(IA(Le,2));es(()=>ce("click",Tt,eA)),Us(Tt,M=>le?.(M)),De((M,We,ni)=>{VC($t,M),VC(kA,g(pA)),DA=Vt(Ve,1,"jse-data-contents svelte-rrrjnb",null,DA,We),gt=Vt(ZA,1,"jse-original-data svelte-rrrjnb",null,gt,ni),Tt.disabled=!!g(KA)},[()=>(k(Oi),k(a()),k(Ka),nA(()=>Oi(a())?"(document root)":Ka(a()))),()=>({"jse-hide-original-data":!g(tA)}),()=>({"jse-hide":!g(tA)})],qA),ce("click",P,UA),ce("input",kA,He),ce("click",tn,aA),iA(tt,Dt)},$$slots:{default:!0}}),Us(xt,(tt,de)=>ky?.(tt,de),()=>Ue),iA(mA,xt)},$$slots:{default:!0}}),pt()}function Oc(){}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-status-bar.svelte-1ulj7zd {
  background: var(--jse-panel-background, #ebebeb);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  margin: 0;
  border-top: var(--jse-panel-border, var(--jse-main-border, 1px solid #d7d7d7));
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
  display: flex;
  gap: var(--jse-padding, 10px);
}
.jse-status-bar.svelte-1ulj7zd:last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-status-bar.svelte-1ulj7zd .jse-status-bar-info:where(.svelte-1ulj7zd) {
  padding: 2px;
}`);var _JA=wA('<div class="jse-status-bar-info svelte-1ulj7zd"> </div>'),GJA=wA('<div class="jse-status-bar-info svelte-1ulj7zd"> </div>'),UJA=wA('<div class="jse-status-bar-info svelte-1ulj7zd"> </div>'),KJA=wA('<div class="jse-status-bar svelte-1ulj7zd"><!> <!> <!></div>'),EG=TE.define([{tag:Se.propertyName,color:"var(--internal-key-color)"},{tag:Se.number,color:"var(--internal-value-color-number)"},{tag:Se.bool,color:"var(--internal-value-color-boolean)"},{tag:Se.string,color:"var(--internal-value-color-string)"},{tag:Se.keyword,color:"var(--internal-value-color-null)"}]),YJA=HF(EG),JJA=EG.style;EG.style=t=>JJA(t||[]);var TJA=[go.fromClass(class{constructor(t){this.view=t,this.indentUnit=Ml(t.state),this.initialPaddingLeft=null,this.isChrome=window?.navigator.userAgent.includes("Chrome"),this.generate(t.state)}update(t){var e=Ml(t.state);(e!==this.indentUnit||t.docChanged||t.viewportChanged)&&(this.indentUnit=e,this.generate(t.state))}generate(t){var e=new ds;this.initialPaddingLeft?this.addStyleToBuilder(e,t,this.initialPaddingLeft):this.view.requestMeasure({read:A=>{var i=A.contentDOM.querySelector(".cm-line");i&&(this.initialPaddingLeft=window.getComputedStyle(i).getPropertyValue("padding-left"),this.addStyleToBuilder(e,A.state,this.initialPaddingLeft)),this.decorations=e.finish()}}),this.decorations=e.finish()}addStyleToBuilder(t,e,A){var i=this.getVisibleLines(e);for(var n of i){var{numColumns:o,containsTab:r}=this.numColumns(n.text,e.tabSize),s="calc(".concat(o+this.indentUnit,"ch + ").concat(A,")"),a=this.isChrome?"calc(-".concat(o+this.indentUnit,"ch - ").concat(r?1:0,"px)"):"-".concat(o+this.indentUnit,"ch");t.add(n.from,n.from,ut.line({attributes:{style:"padding-left: ".concat(s,"; text-indent: ").concat(a,";")}}))}}getVisibleLines(t){var e=new Set,A=null;for(var{from:i,to:n}of this.view.visibleRanges)for(var o=i;o<=n;){var r=t.doc.lineAt(o);A!==r&&(e.add(r),A=r),o=r.to+1}return e}numColumns(t,e){var A=0,i=!1;A:for(var n=0;n<t.length;n++)switch(t[n]){case" ":A+=1;continue A;case"	":A+=e-A%e,i=!0;continue A;case"\r":continue A;default:break A}return{numColumns:A,containsTab:i}}},{decorations:t=>t.decorations})];Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-text-mode.svelte-xt61xw {
  --internal-key-color: var(--jse-key-color, #1a1a1a);
  --internal-value-color-number: var(--jse-value-color-number, #ee422e);
  --internal-value-color-boolean: var(--jse-value-color-boolean, #ff8c00);
  --internal-value-color-string: var(--jse-value-color-string, #008000);
  --internal-value-color-null: var(--jse-value-color-null, #004ed0);
  flex: 1;
  box-sizing: border-box;
  display: flex;
  flex-direction: column;
  background: var(--jse-background-color, #fff);
}
.jse-text-mode.no-main-menu.svelte-xt61xw {
  border-top: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) {
  flex: 1;
  display: flex;
  position: relative;
  flex-direction: column;
  overflow: hidden;
  min-width: 0;
  min-height: 0;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw):last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-text-mode.svelte-xt61xw .jse-contents.jse-hidden:where(.svelte-xt61xw) {
  visibility: hidden;
  position: absolute;
  top: 0;
  left: 0;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor {
  flex: 1;
  overflow: hidden;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-scroller {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  line-height: var(--jse-line-height, calc(1em + 4px));
  color: var(--jse-delimiter-color, rgba(0, 0, 0, 0.38));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-gutters {
  background: var(--jse-panel-background, #ebebeb);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  border-right: var(--jse-panel-border, var(--jse-main-border, 1px solid #d7d7d7));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-activeLine,
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-activeLineGutter {
  background: var(--jse-active-line-background-color, rgba(0, 0, 0, 0.06));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-selectionBackground {
  background: var(--jse-selection-background-color, #d3d3d3);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-searchMatch {
  background-color: var(--jse-search-match-color, #ffe665);
  outline: var(--jse-search-match-outline, none);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-searchMatch.cm-searchMatch-selected {
  background-color: var(--jse-search-match-active-color, var(--jse-search-match-color, #ffe665));
  outline: var(--jse-search-match-outline, 2px solid #e0be00);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-selectionMatch {
  background-color: var(--jse-search-match-background-color, rgba(153, 255, 119, 0.5019607843));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-foldPlaceholder {
  background: var(--jse-tag-background, rgba(0, 0, 0, 0.2));
  color: var(--jse-tag-color, var(--jse-text-color-inverse, #fff));
  border: none;
  padding: 0 var(--jse-padding, 10px);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-tooltip {
  font-size: var(--jse-font-size, 16px);
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  color: var(--jse-tooltip-color, var(--jse-text-color, #4d4d4d));
  background: var(--jse-tooltip-background, var(--jse-modal-background, #f5f5f5));
  border: var(--jse-tooltip-border, var(--jse-main-border, 1px solid #d7d7d7));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-diagnosticAction {
  background: var(--jse-tooltip-action-button-color, var(--jse-text-color-inverse, #fff));
  background: var(--jse-tooltip-action-button-background, #4d4d4d);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-panels {
  border-bottom: var(--jse-panel-border, var(--jse-main-border, 1px solid #d7d7d7));
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search {
  background: var(--jse-panel-background, #ebebeb);
  color: var(--jse-panel-color, var(--jse-text-color, #4d4d4d));
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search input {
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size-text-mode-search, 80%);
  color: var(--jse-input-color, var(--jse-text-color, #4d4d4d));
  border: var(--jse-input-border, 1px solid #d8dbdf);
  background: var(--jse-input-background, var(--jse-background-color, #fff));
  margin-right: 2px;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search button {
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size-text-mode-search, 80%);
  color: var(--jse-panel-button-color, inherit);
  background: var(--jse-panel-button-background, transparent);
  border: none;
  cursor: pointer;
  text-transform: capitalize;
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px);
  margin: 0;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search button:hover {
  color: var(--panel-button-color-highlight, var(--jse-text-color, #4d4d4d));
  background: var(--jse-panel-button-background-highlight, #e0e0e0);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search label {
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size-text-mode-search, 80%);
  padding-left: var(--jse-padding, 10px);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search label input {
  margin-right: 2px;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-search button[name="close"] {
  width: 32px;
  height: 32px;
  font-size: 24px;
  line-height: 24px;
  padding: 0;
  right: 0;
  top: -4px;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .cm-editor .cm-cursor-primary {
  border-color: var(--jse-text-color, #4d4d4d);
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .jse-loading-space:where(.svelte-xt61xw) {
  flex: 1;
}
.jse-text-mode.svelte-xt61xw .jse-contents:where(.svelte-xt61xw) .jse-loading:where(.svelte-xt61xw) {
  flex: 2;
  text-align: center;
  color: var(--jse-panel-color-readonly, #b2b2b2);
  box-sizing: border-box;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
}
.jse-text-mode.svelte-xt61xw .jse-contents.jse-preview:where(.svelte-xt61xw) {
  flex: 1;
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: var(--jse-panel-color-readonly, #b2b2b2);
  overflow: auto;
  white-space: pre-wrap;
  word-break: break-word;
  padding: 2px;
}`);var HJA=wA('<!> <div class="jse-contents jse-preview svelte-xt61xw"> </div>',1),zJA=wA("<!> <!> <!> <!>",1),OJA=wA("<div></div> <!> <!>",1),PJA=wA('<div class="jse-contents svelte-xt61xw"><div class="jse-loading-space svelte-xt61xw"></div> <div class="jse-loading svelte-xt61xw">loading...</div></div>'),jJA=wA("<div><!> <!></div>");function qJA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=b(e,"readOnly",9),o=b(e,"mainMenuBar",9),r=b(e,"statusBar",9),s=b(e,"askToFormat",9),a=b(e,"externalContent",9),c=b(e,"externalSelection",9),l=b(e,"history",9),I=b(e,"indentation",9),C=b(e,"tabSize",9),d=b(e,"escapeUnicodeCharacters",9),B=b(e,"parser",9),E=b(e,"validator",9),h=b(e,"validationParser",9),u=b(e,"onChange",9),D=b(e,"onChangeMode",9),L=b(e,"onSelect",9),R=b(e,"onUndo",9),w=b(e,"onRedo",9),_=b(e,"onError",9),K=b(e,"onFocus",9),z=b(e,"onBlur",9),U=b(e,"onRenderMenu",9),H=b(e,"onSortModal",9),q=b(e,"onTransformModal",9),j=Sr("jsoneditor:TextMode"),gA={key:"Mod-i",run:de,shift:Dt,preventDefault:!0},QA=typeof window>"u";j("isSSR:",QA);var BA,lA=$(void 0,!0),vA=$(void 0,!0),tA=$(void 0,!0),cA=$(!1,!0),pA=$(s(),!0),VA=$([],!0),oe=new bg,KA=new bg,CA=new bg,TA=new bg,Ze=new bg,He=a(),uA=$(u_(He,I(),B()),!0),eA=xa.define(),UA=null;function aA(){if(!UA||UA.length===0)return!1;var MA=UA[0].startState,me=UA[UA.length-1].state,nt=UA.map(Xe=>Xe.changes).reduce((Xe,oi)=>Xe.compose(oi)),Wt={type:"text",undo:{changes:nt.invert(MA.doc).toJSON(),selection:M(MA.selection)},redo:{changes:nt.toJSON(),selection:M(me.selection)}};return j("add history item",Wt),l().add(Wt),UA=null,!0}var le=$(d(),!0);hs(Yt(function*(){if(!QA)try{BA=function(MA){var{target:me,initialText:nt,readOnly:Wt,indentation:Xe}=MA;j("Create CodeMirror editor",{readOnly:Wt,indentation:Xe});var oi=function(Ut,cn){return WN(Ut)?Ut.ranges.every(ft=>ft.anchor<cn.length&&ft.head<cn.length):!1}(c(),nt)?tn(c()):void 0,Di=hr.create({doc:nt,selection:oi,extensions:[_E.of([knA,gA]),oe.of(DA()),TnA(),giA(),IiA(),niA(),YiA(),AiA(),iiA(),hr.allowMultipleSelections.of(!0),RiA(),HF(TiA,{fallback:!0}),PiA(),moA(),voA(),riA(),siA(),oiA(),ioA(),_E.of([...DoA,...MnA,...soA,{key:"Mod-z",run:x,preventDefault:!0},{key:"Mod-y",mac:"Mod-Shift-z",run:Y,preventDefault:!0},{key:"Ctrl-Shift-z",run:Y,preventDefault:!0},...GiA,...RN,...NnA]),YJA,MoA({hideFirstIndent:!0}),qt.domEventHandlers({dblclick:kA}),qt.updateListener.of(Ut=>{y(tA,Ut.state),Ut.docChanged&&(Ut.transactions.some(cn=>!!cn.annotation(eA))||(UA=[...UA??[],Ut]),Lt()),Ut.selectionSet&&Tt()}),ZnA(),noA({top:!0}),qt.lineWrapping,KA.of(hr.readOnly.of(Wt)),TA.of(hr.tabSize.of(C())),CA.of(Ot(Xe)),Ze.of(qt.theme({},{dark:gt()}))]});return BA=new qt({state:Di,parent:me}),oi&&BA.dispatch(BA.state.update({selection:oi.main,scrollIntoView:!0})),BA}({target:g(lA),initialText:We(g(uA),g(cA))?"":g(A).escapeValue(g(uA)),readOnly:n(),indentation:I()})}catch(MA){console.error(MA)}})),qc(()=>{ii(),BA&&(j("Destroy CodeMirror editor"),BA.destroy())});var SA=c1(),Ue=c1();function mA(){BA&&(j("focus"),BA.focus())}var sA=!1;function xt(MA){return tt(MA,!1)}function tt(MA,me){j("handlePatch",MA,me);var nt=B().parse(g(uA)),Wt=Da(nt,MA),Xe=R8(nt,MA);return Ei({text:B().stringify(Wt,null,I())},me,!1),{json:Wt,previousJson:nt,undo:Xe,redo:MA}}function de(){if(j("format"),n())return!1;try{var MA=B().parse(g(uA));return Ei({text:B().stringify(MA,null,I())},!0,!1),y(pA,s()),!0}catch(me){_()(me)}return!1}function Dt(){if(j("compact"),n())return!1;try{var MA=B().parse(g(uA));return Ei({text:B().stringify(MA)},!0,!1),y(pA,!1),!0}catch(me){_()(me)}return!1}function _e(){if(j("repair"),!n())try{Ei({text:Rc(g(uA))},!0,!1),y(ni,qN),y(pi,void 0)}catch(MA){_()(MA)}}function Le(){var MA;if(!n())try{var me=B().parse(g(uA));sA=!0,H()({id:SA,json:me,rootPath:[],onSort:(MA=Yt(function*(nt){var{operations:Wt}=nt;j("onSort",Wt),tt(Wt,!0)}),function(nt){return MA.apply(this,arguments)}),onClose:()=>{sA=!1,mA()}})}catch(nt){_()(nt)}}function bt(MA){var{id:me,rootPath:nt,onTransform:Wt,onClose:Xe}=MA;try{var oi=B().parse(g(uA));sA=!0,q()({id:me||Ue,json:oi,rootPath:nt||[],onTransform:Di=>{Wt?Wt({operations:Di,json:oi,transformedJson:Da(oi,Di)}):(j("onTransform",Di),tt(Di,!0))},onClose:()=>{sA=!1,mA(),Xe&&Xe()}})}catch(Di){_()(Di)}}function Re(){n()||bt({rootPath:[]})}function $t(){BA&&(g(lA)&&g(lA).querySelector(".cm-search")?_D(BA):ND(BA))}function x(){if(n())return!1;ii();var MA=l().undo();return j("undo",MA),lrA(MA)?(BA.dispatch({annotations:eA.of("undo"),changes:Cs.fromJSON(MA.undo.changes),selection:ae.fromJSON(MA.undo.selection),scrollIntoView:!0}),!0):(R()(MA),!1)}function Y(){if(n())return!1;ii();var MA=l().redo();return j("redo",MA),lrA(MA)?(BA.dispatch({annotations:eA.of("redo"),changes:Cs.fromJSON(MA.redo.changes),selection:ae.fromJSON(MA.redo.selection),scrollIntoView:!0}),!0):(w()(MA),!1)}function P(){y(cA,!0),Ei(a(),!0,!0)}function X(){D()(er.tree)}function bA(){nn()}function Be(MA){j("select validation error",MA);var{from:me,to:nt}=Ve(MA);me!==void 0&&nt!==void 0&&(Ee(me,nt),mA())}function Ee(MA,me){j("setSelection",{anchor:MA,head:me}),BA&&BA.dispatch(BA.state.update({selection:{anchor:MA,head:me},scrollIntoView:!0}))}function kA(MA,me){if(me.state.selection.ranges.length===1){var nt=me.state.selection.ranges[0],Wt=g(uA).slice(nt.from,nt.to);if(Wt==="{"||Wt==="["){var Xe=s_.default.parse(g(uA)),oi=Object.keys(Xe.pointers).find(Ut=>{var cn;return((cn=Xe.pointers[Ut].value)===null||cn===void 0?void 0:cn.pos)===nt.from}),Di=Xe.pointers[oi];oi&&Di&&Di.value&&Di.valueEnd&&(j("pointer found, selecting inner contents of path:",oi,Di),Ee(Di.value.pos+1,Di.valueEnd.pos-1))}}}function DA(){return _nA(dn,{delay:300})}function gt(){return!!g(lA)&&getComputedStyle(g(lA)).getPropertyValue("--jse-theme").includes("dark")}function Ve(MA){var{path:me,message:nt,severity:Wt}=MA,{line:Xe,column:oi,from:Di,to:Ut}=function(cn,ft){try{var Qi=s_.default.parse(cn),ot=Ct(ft),Mt=Qi.pointers[ot];if(Mt)return{path:ft,line:Mt.key?Mt.key.line:Mt.value?Mt.value.line:0,column:Mt.key?Mt.key.column:Mt.value?Mt.value.column:0,from:Mt.key?Mt.key.pos:Mt.value?Mt.value.pos:0,to:Mt.keyEnd?Mt.keyEnd.pos:Mt.valueEnd?Mt.valueEnd.pos:0}}catch(on){console.error(on)}return{path:ft,line:0,column:0,from:0,to:0}}(g(A).escapeValue(g(uA)),me);return{path:me,line:Xe,column:oi,from:Di,to:Ut,message:nt,severity:Wt,actions:[]}}function ZA(MA,me){var{line:nt,column:Wt,position:Xe,message:oi}=MA;return{path:[],line:nt,column:Wt,from:Xe,to:Xe,severity:Fl.error,message:oi,actions:me&&!n()?[{name:"Auto repair",apply:()=>_e()}]:void 0}}function rt(MA){return{from:MA.from||0,to:MA.to||0,message:MA.message||"",actions:MA.actions,severity:MA.severity}}function Ei(MA,me,nt){var Wt=u_(MA,I(),B()),Xe=!di(MA,He),oi=He;j("setCodeMirrorContent",{isChanged:Xe,emitChange:me,forceUpdate:nt}),BA&&(Xe||nt)&&(He=MA,y(uA,Wt),We(g(uA),g(cA))||BA.dispatch({changes:{from:0,to:BA.state.doc.length,insert:g(A).escapeValue(g(uA))}}),aA(),Xe&&me&&_i(He,oi))}function tn(MA){return WN(MA)?ae.fromJSON(MA):void 0}function qi(){return xe.apply(this,arguments)}function xe(){return xe=Yt(function*(){j("refresh"),yield function(){return mi.apply(this,arguments)}()}),xe.apply(this,arguments)}function nn(){if(BA){var MA=BA?g(A).unescapeValue(BA.state.doc.toString()):"",me=MA!==g(uA);if(j("onChangeCodeMirrorValue",{isChanged:me}),me){var nt=He;y(uA,MA),He={text:g(uA)},aA(),_i(He,nt),io(),Tt()}}}function mi(){return(mi=Yt(function*(){if(io(),BA){var MA=gt();return j("updateTheme",{dark:MA}),BA.dispatch({effects:[Ze.reconfigure(qt.theme({},{dark:MA}))]}),new Promise(me=>setTimeout(me))}return Promise.resolve()})).apply(this,arguments)}function Ot(MA){var me=FC.of(typeof MA=="number"?" ".repeat(MA):MA);return MA==="	"?[me]:[me,TJA]}CG({onMount:hs,onDestroy:qc,getWindow:()=>af(g(vA)),hasFocus:()=>sA&&document.hasFocus()||W_(g(vA)),onFocus:K(),onBlur:()=>{ii(),z()()}});var Lt=oE(nn,300);function ii(){Lt.flush()}function _i(MA,me){u()&&u()(MA,me,{contentErrors:mn(),patchResult:void 0})}function Tt(){L()(M(g(tA).selection))}function M(MA){return fe({type:Ln.text},MA.toJSON())}function We(MA,me){return!!MA&&MA.length>PN&&!me}var ni=$(qN,!0),pi=$(void 0,!0);function dn(){if(We(g(uA),g(cA)))return[];var MA=mn();if(crA(MA)){var{parseError:me,isRepairable:nt}=MA;return[rt(ZA(me,nt))]}return lUA(MA)?MA.validationErrors.map(Ve).map(rt):[]}function mn(){j("validate:start"),ii();var MA=Uo(g(A).escapeValue(g(uA)),E(),B(),h());return crA(MA)?(y(ni,MA.isRepairable?nrA:"invalid"),y(pi,MA.parseError),y(VA,[])):(y(ni,qN),y(pi,void 0),y(VA,MA?.validationErrors||[])),j("validate:end"),MA}var Uo=aE(FKA);function kn(){g(pi)&&function(MA){j("select parse error",MA);var me=ZA(MA,!1);Ee(me.from!=null?me.from:0,me.to!=null?me.to:0),mA()}(g(pi))}var Wn={icon:JW,text:"Show me",title:"Move to the parse error location",onClick:kn};fA(()=>k(d()),()=>{y(A,V_({escapeControlCharacters:!1,escapeUnicodeCharacters:d()}))}),fA(()=>k(a()),()=>{Ei(a(),!1,!1)}),fA(()=>k(c()),()=>{(function(MA){if(WN(MA)){var me=tn(MA);!BA||!me||g(tA)&&g(tA).selection.eq(me)||(j("applyExternalSelection",me),BA.dispatch({selection:me}))}})(c())}),fA(()=>k(E()),()=>{(function(MA){j("updateLinter",MA),BA&&BA.dispatch({effects:oe.reconfigure(DA())})})(E())}),fA(()=>k(I()),()=>{(function(MA){BA&&(j("updateIndentation",MA),BA.dispatch({effects:CA.reconfigure(Ot(MA))}))})(I())}),fA(()=>k(C()),()=>{(function(MA){BA&&(j("updateTabSize",MA),BA.dispatch({effects:TA.reconfigure(hr.tabSize.of(MA))}))})(C())}),fA(()=>k(n()),()=>{(function(MA){BA&&(j("updateReadOnly",MA),BA.dispatch({effects:[KA.reconfigure(hr.readOnly.of(MA))]}))})(n())}),fA(()=>(g(le),k(d())),()=>{g(le)!==d()&&(y(le,d()),j("forceUpdateText",{escapeUnicodeCharacters:d()}),BA&&BA.dispatch({changes:{from:0,to:BA.state.doc.length,insert:g(A).escapeValue(g(uA))}}))}),fA(()=>(g(ni),k(n()),L0),()=>{y(i,g(ni)!==nrA||n()?[Wn]:[{icon:L0,text:"Auto repair",title:"Automatically repair JSON",onClick:_e},Wn])}),Qn(),Zt(!0);var Vo,vo=jJA(),bo=W(vo),Yn=MA=>{var me=qA(()=>(g(uA),nA(()=>g(uA).length===0))),nt=qA(()=>!g(me)),Wt=qA(()=>!g(me)),Xe=qA(()=>!g(me)),oi=qA(()=>!g(me));(function(Di,Ut){mt(Ut,!1);var cn=$(void 0,!0),ft=b(Ut,"readOnly",9,!1),Qi=b(Ut,"onFormat",9),ot=b(Ut,"onCompact",9),Mt=b(Ut,"onSort",9),on=b(Ut,"onTransform",9),hn=b(Ut,"onToggleSearch",9),Ai=b(Ut,"onUndo",9),Ki=b(Ut,"onRedo",9),dt=b(Ut,"canUndo",9),EA=b(Ut,"canRedo",9),HA=b(Ut,"canFormat",9),ve=b(Ut,"canCompact",9),Qt=b(Ut,"canSort",9),yi=b(Ut,"canTransform",9),ri=b(Ut,"onRenderMenu",9),pn={type:"button",icon:g4,title:"Search (Ctrl+F)",className:"jse-search",onClick:hn()},Fn=$(void 0,!0);fA(()=>(k(ft()),k(Qi()),k(HA()),k(ot()),k(ve()),k(Mt()),k(Qt()),k(on()),k(yi()),k(Ai()),k(dt()),k(Ki()),k(EA())),()=>{y(Fn,ft()?[pn,{type:"space"}]:[{type:"button",icon:YrA,title:"Format JSON: add proper indentation and new lines (Ctrl+I)",className:"jse-format",onClick:Qi(),disabled:ft()||!HA()},{type:"button",icon:UYA,title:"Compact JSON: remove all white spacing and new lines (Ctrl+Shift+I)",className:"jse-compact",onClick:ot(),disabled:ft()||!ve()},{type:"separator"},{type:"button",icon:l4,title:"Sort",className:"jse-sort",onClick:Mt(),disabled:ft()||!Qt()},{type:"button",icon:s4,title:"Transform contents (filter, sort, project)",className:"jse-transform",onClick:on(),disabled:ft()||!yi()},pn,{type:"separator"},{type:"button",icon:h5,title:"Undo (Ctrl+Z)",className:"jse-undo",onClick:Ai(),disabled:!dt()},{type:"button",icon:Q5,title:"Redo (Ctrl+Shift+Z)",className:"jse-redo",onClick:Ki(),disabled:!EA()},{type:"space"}])}),fA(()=>(k(ri()),g(Fn)),()=>{y(cn,ri()(g(Fn))||g(Fn))}),Qn(),Zt(!0),Oy(Di,{get items(){return g(cn)}}),pt()})(MA,{get readOnly(){return n()},onFormat:de,onCompact:Dt,onSort:Le,onTransform:Re,onToggleSearch:$t,onUndo:x,onRedo:Y,get canFormat(){return g(nt)},get canCompact(){return g(Wt)},get canSort(){return g(Xe)},get canTransform(){return g(oi)},get canUndo(){return k(l()),nA(()=>l().canUndo)},get canRedo(){return k(l()),nA(()=>l().canRedo)},get onRenderMenu(){return U()}})};LA(bo,MA=>{o()&&MA(Yn)});var Mo=IA(bo,2),ne=MA=>{var me,nt=OJA(),Wt=qA(()=>(g(uA),g(cA),nA(()=>We(g(uA),g(cA))))),Xe=vt(nt);Co(Xe,ft=>y(lA,ft),()=>g(lA));var oi=IA(Xe,2),Di=ft=>{var Qi=HJA(),ot=vt(Qi),Mt=qA(()=>(k(cy),k(PN),g(uA),nA(()=>"The JSON document is larger than ".concat(cy(PN),", ")+"and may crash your browser when loading it in text mode. Actual size: ".concat(cy(g(uA).length),"."))));mc(ot,{get icon(){return g1},type:"error",get message(){return g(Mt)},actions:[{text:"Open anyway",title:"Open the document in text mode. This may freeze or crash your browser.",onClick:P},{text:"Open in tree mode",title:"Open the document in tree mode. Tree mode can handle large documents.",onClick:X},{text:"Cancel",title:"Cancel opening this large document.",onClick:bA}],onClose:mA});var on=W(IA(ot,2));De(hn=>wt(on,hn),[()=>(k(V0),g(uA),k(Qy),nA(()=>V0(g(uA)||"",Qy)))],qA),iA(ft,Qi)};LA(oi,ft=>{g(Wt)&&ft(Di)});var Ut=IA(oi,2),cn=ft=>{var Qi=zJA(),ot=vt(Qi),Mt=dt=>{(function(EA,HA){mt(HA,!1);var ve=b(HA,"editorState",8),Qt=$(),yi=$(),ri=$(),pn=$(),Fn=$();fA(()=>k(ve()),()=>{var se;y(Qt,(se=ve())===null||se===void 0||(se=se.selection)===null||se===void 0||(se=se.main)===null||se===void 0?void 0:se.head)}),fA(()=>(g(Qt),k(ve())),()=>{var se;y(yi,g(Qt)!==void 0?(se=ve())===null||se===void 0||(se=se.doc)===null||se===void 0?void 0:se.lineAt(g(Qt)):void 0)}),fA(()=>g(yi),()=>{y(ri,g(yi)!==void 0?g(yi).number:void 0)}),fA(()=>(g(yi),g(Qt)),()=>{y(pn,g(yi)!==void 0&&g(Qt)!==void 0?g(Qt)-g(yi).from+1:void 0)}),fA(()=>k(ve()),()=>{var se;y(Fn,(se=ve())===null||se===void 0||(se=se.selection)===null||se===void 0||(se=se.ranges)===null||se===void 0?void 0:se.reduce((vi,Yi)=>vi+Yi.to-Yi.from,0))}),Qn(),Zt();var Jn=KJA(),ln=W(Jn),Pt=se=>{var vi=_JA(),Yi=W(vi);De(()=>{var rn;return wt(Yi,"Line: ".concat((rn=g(ri))!==null&&rn!==void 0?rn:""))}),iA(se,vi)};LA(ln,se=>{g(ri)!==void 0&&se(Pt)});var $i=IA(ln,2),Rr=se=>{var vi=GJA(),Yi=W(vi);De(()=>{var rn;return wt(Yi,"Column: ".concat((rn=g(pn))!==null&&rn!==void 0?rn:""))}),iA(se,vi)};LA($i,se=>{g(pn)!==void 0&&se(Rr)});var Ft=IA($i,2),Xn=se=>{var vi=UJA(),Yi=W(vi);De(()=>{var rn;return wt(Yi,"Selection: ".concat((rn=g(Fn))!==null&&rn!==void 0?rn:""," characters"))}),iA(se,vi)};LA(Ft,se=>{g(Fn)!==void 0&&g(Fn)>0&&se(Xn)}),iA(EA,Jn),pt()})(dt,{get editorState(){return g(tA)}})};LA(ot,dt=>{r()&&dt(Mt)});var on=IA(ot,2),hn=dt=>{mc(dt,{type:"error",get icon(){return g1},get message(){return g(pi),nA(()=>g(pi).message)},get actions(){return g(i)},onClick:kn,onClose:mA})};LA(on,dt=>{g(pi)&&dt(hn)});var Ai=IA(on,2),Ki=dt=>{var EA=qA(()=>[{icon:YrA,text:"Format",title:"Format JSON: add proper indentation and new lines (Ctrl+I)",onClick:de},{icon:I4,text:"No thanks",title:"Close this message",onClick:()=>y(pA,!1)}]);mc(dt,{type:"success",message:"Do you want to format the JSON?",get actions(){return g(EA)},onClose:mA})};LA(Ai,dt=>{g(pi),g(pA),k(trA),g(uA),nA(()=>!g(pi)&&g(pA)&&trA(g(uA)))&&dt(Ki)}),dG(IA(Ai,2),{get validationErrors(){return g(VA)},selectError:Be}),iA(ft,Qi)};LA(Ut,ft=>{g(Wt)||ft(cn)}),De(ft=>me=Vt(Xe,1,"jse-contents svelte-xt61xw",null,me,ft),[()=>({"jse-hidden":g(Wt)})],qA),iA(MA,nt)},wi=MA=>{iA(MA,PJA())};return LA(Mo,MA=>{QA?MA(wi,!1):MA(ne)}),Co(vo,MA=>y(vA,MA),()=>g(vA)),De(MA=>Vo=Vt(vo,1,"jse-text-mode svelte-xt61xw",null,Vo,MA),[()=>({"no-main-menu":!o()})],qA),iA(t,vo),zt(e,"focus",mA),zt(e,"patch",xt),zt(e,"handlePatch",tt),zt(e,"openTransformModal",bt),zt(e,"refresh",qi),zt(e,"flush",ii),zt(e,"validate",mn),pt({focus:mA,patch:xt,handlePatch:tt,openTransformModal:bt,refresh:qi,flush:ii,validate:mn})}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-inline-value.svelte-h57m0p {
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  line-height: var(--jse-line-height, calc(1em + 4px));
  border: none;
  padding: 0 calc(0.5 * var(--jse-padding, 10px));
  background: transparent;
  color: inherit;
  cursor: inherit;
}
.jse-inline-value.jse-highlight.svelte-h57m0p {
  background-color: var(--jse-search-match-color, #ffe665);
  outline: var(--jse-search-match-outline, none);
}
.jse-inline-value.jse-highlight.jse-active.svelte-h57m0p {
  background-color: var(--jse-search-match-active-color, var(--jse-search-match-color, #ffe665));
  outline: var(--jse-search-match-outline, 2px solid #e0be00);
}`);var VJA=wA('<button type="button"> </button>');Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-column-header.svelte-2i3vdx {
  background: none;
  border: none;
  font-family: inherit;
  font-size: inherit;
  color: inherit;
  display: flex;
  gap: var(--jse-padding, 10px);
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px) calc(0.5 * var(--jse-padding, 10px)) calc(0.5 * var(--jse-padding, 10px));
  width: 100%;
}
.jse-column-header.svelte-2i3vdx:hover {
  background: var(--jse-table-header-background-highlight, #e8e8e8);
}
.jse-column-header.svelte-2i3vdx:not(.jse-column-header.jse-readonly) {
  cursor: pointer;
}
.jse-column-header.svelte-2i3vdx span.jse-column-sort-icon:where(.svelte-2i3vdx) {
  height: 1em;
}`);var ZJA=wA('<span class="jse-column-sort-icon svelte-2i3vdx"><!></span>'),WJA=wA('<button type="button"><span class="jse-column-name"> </span> <!></button>');Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-table-mode-welcome.svelte-17xl1jx {
  flex: 1;
  display: flex;
  flex-direction: column;
  overflow: auto;
  align-items: center;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-table-mode-welcome.svelte-17xl1jx:last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-space.jse-before:where(.svelte-17xl1jx) {
  flex: 1;
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) {
  display: flex;
  flex-direction: column;
  gap: var(--jse-padding, 10px);
  max-width: 400px;
  margin: 2em var(--jse-padding, 10px);
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) .jse-nested-arrays-info:where(.svelte-17xl1jx) {
  color: var(--jse-panel-color-readonly, #b2b2b2);
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) .jse-nested-property:where(.svelte-17xl1jx) {
  display: flex;
  align-items: center;
  gap: var(--jse-padding, 10px);
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) .jse-nested-property:where(.svelte-17xl1jx) .jse-nested-property-path:where(.svelte-17xl1jx) {
  flex: 1;
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) .jse-nested-property:where(.svelte-17xl1jx) .jse-nested-property-path:where(.svelte-17xl1jx) .jse-nested-property-count:where(.svelte-17xl1jx) {
  opacity: 0.5;
  white-space: nowrap;
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) button.jse-nested-array-action:where(.svelte-17xl1jx) {
  text-align: left;
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) button.jse-nested-array-action:where(.svelte-17xl1jx):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-nested-arrays:where(.svelte-17xl1jx) button.jse-nested-array-action:where(.svelte-17xl1jx):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}
.jse-table-mode-welcome.svelte-17xl1jx .jse-space.jse-after:where(.svelte-17xl1jx) {
  flex: 2;
}`);var XJA=(t,e)=>e.onClick(),$JA=wA(`An empty document cannot be opened in table mode. You can go to tree mode instead, or paste
        a JSON Array using <b>Ctrl+V</b>.`,1),ATA=(t,e,A)=>e.openJSONEditorModal(g(A)),eTA=(t,e,A)=>e.extractPath(g(A)),tTA=wA('<button type="button" class="jse-nested-array-action svelte-17xl1jx">Extract</button>'),iTA=wA('<div class="jse-nested-property svelte-17xl1jx"><div class="jse-nested-property-path svelte-17xl1jx"> <span class="jse-nested-property-count svelte-17xl1jx"> </span></div> <button type="button" class="jse-nested-array-action svelte-17xl1jx"> </button> <!></div>'),nTA=(t,e)=>e.onChangeMode(er.tree),oTA=wA('<div class="jse-table-mode-welcome svelte-17xl1jx" role="none"><div class="jse-space jse-before svelte-17xl1jx"></div> <div class="jse-nested-arrays svelte-17xl1jx"><div class="jse-nested-arrays-title"> </div> <div class="jse-nested-arrays-info svelte-17xl1jx"><!></div> <!> <button type="button" class="jse-nested-array-action svelte-17xl1jx">Switch to tree mode</button></div> <div class="jse-space jse-after svelte-17xl1jx"></div></div>');function rTA(t,e){mt(e,!0);var A=Ga(()=>e.json?function(E){var h=arguments.length>1&&arguments[1]!==void 0?arguments[1]:2,u=[];return function D(L,R){xo(L)&&R.length<h&&Object.keys(L).forEach(w=>{D(L[w],R.concat(w))}),wo(L)&&u.push(R)}(E,[]),u}(e.json).slice(0,99).filter(E=>E.length>0):[]),i=Ga(()=>!Oi(g(A))),n=Ga(()=>e.json===void 0&&(e.text===""||e.text===void 0)),o=Ga(()=>g(i)?"Object with nested arrays":g(n)?"An empty document":xo(e.json)?"An object":wo(e.json)?"An empty array":"A ".concat(q_(e.json,e.parser))),r=oTA();r.__click=[XJA,e];var s=IA(W(r),2),a=W(s),c=W(a),l=IA(a,2),I=W(l),C=E=>{iA(E,Tr(`An object cannot be opened in table mode. You can open a nested array instead, or open the
        document in tree mode.`))},d=(E,h)=>{var u=L=>{iA(L,$JA())},D=L=>{var R=Tr();De(()=>{var w;return wt(R,"".concat((w=g(o))!==null&&w!==void 0?w:""," cannot be opened in table mode. You can open the document in tree mode instead."))}),iA(L,R)};LA(E,L=>{g(n)&&!e.readOnly?L(u):L(D,!1)},h)};LA(I,E=>{g(i)?E(C):E(d,!1)});var B=IA(l,2);qo(B,17,()=>g(A),or,(E,h)=>{var u=iTA(),D=Ga(()=>function(H){return Ke(e.json,H).length}(g(h))),L=W(u),R=W(L),w=W(IA(R)),_=IA(L,2);_.__click=[ATA,e,h];var K=W(_),z=IA(_,2),U=H=>{var q=tTA();q.__click=[eTA,e,h],iA(H,q)};LA(z,H=>{e.readOnly||H(U)}),De(H=>{var q;wt(R,'"'.concat(H??"",'" ')),wt(w,"(".concat((q=g(D))!==null&&q!==void 0?q:""," ").concat(g(D)!==1?"items":"item",")")),wt(K,e.readOnly?"View":"Edit")},[()=>Ka(g(h))]),iA(E,u)}),IA(B,2).__click=[nTA,e],De(()=>wt(c,g(o))),iA(t,r),pt()}of(["click"]);Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-column-header.svelte-fzj761 {
  background: none;
  border: none;
  font-family: inherit;
  font-size: inherit;
  color: inherit;
  display: flex;
  gap: var(--jse-padding, 10px);
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px) calc(0.5 * var(--jse-padding, 10px)) calc(0.5 * var(--jse-padding, 10px));
  width: 100%;
}
.jse-column-header.svelte-fzj761:hover {
  background: var(--jse-table-header-background-highlight, #e8e8e8);
}
.jse-column-header.svelte-fzj761:not(.jse-column-header.jse-readonly) {
  cursor: pointer;
}`);var sTA=wA('<button type="button"><!></button>');Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-table-mode.svelte-u14cgx {
  flex: 1;
  display: flex;
  flex-direction: column;
  position: relative;
  background: var(--jse-background-color, #fff);
  min-width: 0;
  min-height: 0;
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: var(--jse-text-color, #4d4d4d);
  line-height: var(--jse-line-height, calc(1em + 4px));
}
.jse-table-mode.no-main-menu.svelte-u14cgx {
  border-top: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-table-mode.svelte-u14cgx .jse-search-box-container:where(.svelte-u14cgx) {
  position: relative;
  height: 0;
  top: calc(var(--jse-line-height, calc(1em + 4px)) + 2 * var(--jse-padding, 10px));
  margin-right: calc(var(--jse-padding, 10px) + 20px);
  margin-left: var(--jse-padding, 10px);
  text-align: right;
  z-index: 3;
}
.jse-table-mode.svelte-u14cgx .jse-hidden-input-label:where(.svelte-u14cgx) {
  position: fixed;
  right: 0;
  top: 0;
  width: 0;
  height: 0;
}
.jse-table-mode.svelte-u14cgx .jse-hidden-input-label:where(.svelte-u14cgx) .jse-hidden-input:where(.svelte-u14cgx) {
  width: 0;
  height: 0;
  padding: 0;
  border: 0;
  outline: none;
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) {
  flex: 1;
  align-items: flex-start;
  flex-direction: column;
  display: flex;
  overflow: auto;
  overflow-anchor: none;
  scrollbar-gutter: stable;
  border-left: var(--jse-main-border, 1px solid #d7d7d7);
  border-right: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx):last-child {
  border-bottom: var(--jse-main-border, 1px solid #d7d7d7);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) {
  border-collapse: collapse;
  border-spacing: 0;
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-invisible-start-section:where(.svelte-u14cgx) td:where(.svelte-u14cgx),
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-invisible-end-section:where(.svelte-u14cgx) td:where(.svelte-u14cgx) {
  margin: 0;
  padding: 0;
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-search-box-background:where(.svelte-u14cgx) {
  background: var(--jse-table-header-background, #f5f5f5);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-invisible-end-section:where(.svelte-u14cgx) td:where(.svelte-u14cgx) {
  padding-bottom: var(--jse-padding, 10px);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx):hover {
  background-color: var(--jse-table-row-odd-background, rgba(0, 0, 0, 0.05));
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell:where(.svelte-u14cgx) {
  padding: 0 var(--jse-padding, 10px) 0 0;
  vertical-align: top;
  white-space: nowrap;
  height: var(--jse-line-height, calc(1em + 4px));
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell.jse-table-cell-header:where(.svelte-u14cgx), .jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell.jse-table-cell-gutter:where(.svelte-u14cgx) {
  font-weight: normal;
  text-align: left;
  color: var(--jse-text-readonly, #8d8d8d);
  background: var(--jse-table-header-background, #f5f5f5);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell.jse-table-cell-header:where(.svelte-u14cgx) {
  padding: 0;
  position: sticky;
  top: 0;
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell.jse-table-cell-header:where(.svelte-u14cgx) .jse-table-root-error:where(.svelte-u14cgx) {
  padding: calc(0.5 * var(--jse-padding, 10px)) var(--jse-padding, 10px) calc(0.5 * var(--jse-padding, 10px)) calc(0.5 * var(--jse-padding, 10px));
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell.jse-table-cell-gutter:where(.svelte-u14cgx) {
  padding: 0 var(--jse-padding, 10px) 0 calc(0.5 * var(--jse-padding, 10px));
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell:where(.svelte-u14cgx) .jse-value-outer:where(.svelte-u14cgx) {
  display: inline-block;
  cursor: var(--jse-contents-cursor, pointer);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell:where(.svelte-u14cgx) .jse-value-outer:where(.svelte-u14cgx):hover {
  background: var(--jse-hover-background-color, rgba(0, 0, 0, 0.06));
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell:where(.svelte-u14cgx) .jse-value-outer.jse-selected-value:where(.svelte-u14cgx) {
  background: var(--jse-selection-background-color, #d3d3d3);
}
.jse-table-mode.svelte-u14cgx .jse-contents:where(.svelte-u14cgx) table.jse-table-main:where(.svelte-u14cgx) .jse-table-row:where(.svelte-u14cgx) .jse-table-cell:where(.svelte-u14cgx) .jse-context-menu-anchor:where(.svelte-u14cgx) {
  display: inline-flex;
  position: relative;
  vertical-align: top;
}
.jse-table-mode.svelte-u14cgx .jse-contents.jse-contents-loading:where(.svelte-u14cgx) {
  align-items: unset;
}
.jse-table-mode.svelte-u14cgx .jse-contents.jse-contents-loading:where(.svelte-u14cgx) .jse-loading-space:where(.svelte-u14cgx) {
  flex: 1;
}
.jse-table-mode.svelte-u14cgx .jse-contents.jse-contents-loading:where(.svelte-u14cgx) .jse-loading:where(.svelte-u14cgx) {
  flex: 2;
  text-align: center;
  color: var(--jse-panel-color-readonly, #b2b2b2);
  box-sizing: border-box;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
}`);var aTA=wA('<div class="jse-table-root-error svelte-u14cgx"><!></div>'),cTA=wA('<th class="jse-table-cell jse-table-cell-header svelte-u14cgx"><!></th>'),lTA=wA('<th class="jse-table-cell jse-table-cell-header svelte-u14cgx"><!></th>'),gTA=wA('<th class="jse-table-cell jse-table-cell-gutter svelte-u14cgx"> <!></th>'),ITA=wA('<div class="jse-context-menu-anchor svelte-u14cgx"><!></div>'),CTA=wA('<td class="jse-table-cell svelte-u14cgx"><div><!><!></div> <!></td>'),dTA=wA('<td class="jse-table-cell svelte-u14cgx"></td>'),BTA=wA('<tr class="jse-table-row svelte-u14cgx"><!><!><!></tr>'),ETA=wA('<div class="jse-search-box-container svelte-u14cgx"><!></div> <div class="jse-contents svelte-u14cgx"><table class="jse-table-main svelte-u14cgx"><tbody><tr class="jse-table-row jse-table-row-header svelte-u14cgx"><th class="jse-table-cell jse-table-cell-header svelte-u14cgx"><!></th><!><!></tr><tr><td class="svelte-u14cgx"></td></tr><!><tr class="jse-table-invisible-end-section svelte-u14cgx"><td class="svelte-u14cgx"></td></tr></tbody></table></div> <!> <!> <!> <!>',1),QTA=wA("<!> <!>",1),hTA=wA('<label class="jse-hidden-input-label svelte-u14cgx"><input type="text" tabindex="-1" class="jse-hidden-input svelte-u14cgx"/></label> <!>',1),uTA=wA('<div class="jse-contents jse-contents-loading svelte-u14cgx"><div class="jse-loading-space svelte-u14cgx"></div> <div class="jse-loading svelte-u14cgx">loading...</div></div>'),fTA=wA('<div role="table"><!> <!></div> <!> <!>',1);function mTA(t,e){mt(e,!1);var A=$(void 0,!0),i=$(void 0,!0),n=$(void 0,!0),o=Sr("jsoneditor:TableMode"),{openAbsolutePopup:r,closeAbsolutePopup:s}=$1("absolute-popup"),a=paA(),c=c1(),l=c1(),I=typeof window>"u";o("isSSR:",I);var C=b(e,"readOnly",9),d=b(e,"externalContent",9),B=b(e,"externalSelection",9),E=b(e,"history",9),h=b(e,"truncateTextSize",9),u=b(e,"mainMenuBar",9),D=b(e,"escapeControlCharacters",9),L=b(e,"escapeUnicodeCharacters",9),R=b(e,"flattenColumns",9),w=b(e,"parser",9),_=b(e,"parseMemoizeOne",9),K=b(e,"validator",9),z=b(e,"validationParser",9),U=b(e,"indentation",9),H=b(e,"onChange",9),q=b(e,"onChangeMode",9),j=b(e,"onSelect",9),gA=b(e,"onUndo",9),QA=b(e,"onRedo",9),BA=b(e,"onRenderValue",9),lA=b(e,"onRenderMenu",9),vA=b(e,"onRenderContextMenu",9),tA=b(e,"onFocus",9),cA=b(e,"onBlur",9),pA=b(e,"onSortModal",9),VA=b(e,"onTransformModal",9),oe=b(e,"onJSONEditorModal",9),KA=$(void 0,!0),CA=$(void 0,!0),TA=$(void 0,!0),Ze=$(void 0,!0),He=$(void 0,!0);CG({onMount:hs,onDestroy:qc,getWindow:()=>af(g(CA)),hasFocus:()=>Re&&document.hasFocus()||W_(g(CA)),onFocus:()=>{$t=!0,tA()&&tA()()},onBlur:()=>{$t=!1,cA()&&cA()()}});var uA,eA=$(void 0,!0),UA=$(void 0,!0),aA=$(void 0,!0),le=$(void 0,!0),SA=$(void 0,!0),Ue=$(void 0,!0),mA=$(!1,!0),sA=$(!1,!0);function xt(p){y(Ue,(uA=p)?caA(g(eA),uA.items):void 0)}function tt(p){return de.apply(this,arguments)}function de(){return(de=Yt(function*(p){y(DA,void 0),yield dn(p)})).apply(this,arguments)}function Dt(){y(mA,!1),y(sA,!1),M()}var _e=$(1e4,!0),Le=$([],!0),bt=$(void 0,!0),Re=!1,$t=!1,x=$(!1,!0),Y=$({},!0),P=$(600,!0),X=$(0,!0),bA=18;function Be(p){y(DA,p)}function Ee(p){g(DA)&&p!==void 0&&(Ms(p,OC(g(DA)))&&Ms(p,et(g(DA)))||(o("clearing selection: path does not exist anymore",g(DA)),y(DA,void 0)))}var kA=$(g(eA)!==void 0?w_({json:g(eA)}):void 0,!0),DA=$(q3(B())?B():void 0,!0),gt=$(void 0,!0),Ve=$(!1,!0);function ZA(p){if(!C()){o("onSortByHeader",p);var N=p.sortDirection===Pc.desc?-1:1;Ot(baA(g(eA),[],p.path,N),(J,V)=>({state:V,sortedColumn:p}))}}hs(()=>{g(DA)&&Uo(et(g(DA)))});var rt=$(void 0,!0);function Ei(p){if(p.json!==void 0||p.text!==void 0){var N=g(eA)!==void 0&&p.json!==void 0;E().add({type:"tree",undo:{patch:N?[{op:"replace",path:"",value:p.json}]:void 0,json:p.json,text:p.text,documentState:p.documentState,textIsRepaired:p.textIsRepaired,selection:Hg(p.selection),sortedColumn:p.sortedColumn},redo:{patch:N?[{op:"replace",path:"",value:g(eA)}]:void 0,json:g(eA),text:g(UA),documentState:g(kA),textIsRepaired:g(Ve),selection:Hg(g(DA)),sortedColumn:g(gt)}})}}var tn=$([],!0),qi=aE(waA);function xe(p,N,J,V){sQ(()=>{var Z;try{Z=qi(p,N,J,V)}catch(FA){Z=[{path:[],message:"Failed to validate: "+FA.message,severity:Fl.warning}]}di(Z,g(tn))||(o("validationErrors changed:",Z),y(tn,Z))},Z=>o("validationErrors updated in ".concat(Z," ms")))}function nn(){return o("validate"),g(aA)?{parseError:g(aA),isRepairable:!1}:(xe(g(eA),K(),w(),z()),Oi(g(tn))?void 0:{validationErrors:g(tn)})}function mi(p,N){if(o("patch",p,N),g(eA)===void 0)throw new Error("Cannot apply patch: no JSON");var J=g(eA),V={json:void 0,text:g(UA),documentState:g(kA),selection:Hg(g(DA)),sortedColumn:g(gt),textIsRepaired:g(Ve)},Z=aaA(g(eA),p),FA=ZsA(g(eA),g(kA),p),te=lJA(g(gt),p,g(Le)),re=typeof N=="function"?N(FA.json,FA.documentState,g(DA)):void 0;return y(eA,re?.json!==void 0?re.json:FA.json),y(kA,re?.state!==void 0?re.state:FA.documentState),y(DA,re?.selection!==void 0?re.selection:g(DA)),y(gt,re?.sortedColumn!==void 0?re.sortedColumn:te),y(UA,void 0),y(Ve,!1),y(le,void 0),y(SA,void 0),y(aA,void 0),E().add({type:"tree",undo:fe({patch:Z},V),redo:{patch:p,json:void 0,text:void 0,documentState:g(kA),selection:Hg(g(DA)),sortedColumn:g(gt),textIsRepaired:g(Ve)}}),{json:g(eA),previousJson:J,undo:Z,redo:p}}function Ot(p,N){o("handlePatch",p,N);var J={json:g(eA),text:g(UA)},V=mi(p,N);return Lt(J,V),V}function Lt(p,N){if((p.json!==void 0||p?.text!==void 0)&&H()){if(g(UA)!==void 0){var J={text:g(UA),json:void 0};H()(J,p,{contentErrors:nn(),patchResult:N})}else if(g(eA)!==void 0){var V={text:void 0,json:g(eA)};H()(V,p,{contentErrors:nn(),patchResult:N})}}}function ii(p){o("pasted json as text",p),y(le,p)}function _i(p){o("pasted multiline text",{pastedText:p}),y(SA,p)}function Tt(p){var N=parseInt(p[0],10),J=[String(N+1),...p.slice(1)];return Ms(g(eA),J)?Ni(J):Ni(p)}function M(){o("focus"),g(Ze)&&(g(Ze).focus(),g(Ze).select())}function We(p){y(X,p.target.scrollTop)}function ni(){g(DA)||y(DA,function(){if(wo(g(eA))&&!Oi(g(eA))&&!Oi(g(Le)))return Ni(["0",...g(Le)[0]])}())}function pi(){if(g(Ve)&&g(eA)!==void 0){var p={json:g(eA),text:g(UA)},N={json:g(eA),documentState:g(kA),selection:g(DA),sortedColumn:g(gt),text:g(UA),textIsRepaired:g(Ve)};y(UA,void 0),y(Ve,!1),Ee(g(eA)),Ei(N),Lt(p,void 0)}return{json:g(eA),text:g(UA)}}function dn(p){var{scrollToWhenVisible:N=!0}=arguments.length>1&&arguments[1]!==void 0?arguments[1]:{},J=g(mA)?_3:0,V=TrA(p,g(Le),Y,bA),Z=V-g(X)+J+bA,FA=kn(p);if(o("scrollTo",{path:p,top:V,scrollTop:g(X),elem:FA}),!g(TA))return Promise.resolve();var te=g(TA).getBoundingClientRect();if(FA&&!N){var re=FA.getBoundingClientRect();if(re.bottom>te.top&&re.top<te.bottom)return Promise.resolve()}var Pe=-Math.max(J+2*bA,te.height/4);return new Promise(FA?ze=>{a(FA,{container:g(TA),offset:Pe,duration:300,callback:()=>{mn(p),ze()}})}:ze=>{a(Z,{container:g(TA),offset:Pe,duration:300,callback:()=>{io(),mn(p),ze()}})})}function mn(p){var N=kn(p);if(N&&g(TA)){var J=g(TA).getBoundingClientRect(),V=N.getBoundingClientRect();if(V.right>J.right){var Z=V.right-J.right;hc(TA,g(TA).scrollLeft+=Z)}if(V.left<J.left){var FA=J.left-V.left;hc(TA,g(TA).scrollLeft-=FA)}}}function Uo(p){(function(N){if(g(TA)){var{rowIndex:J}=zc(N,g(Le)),V=TrA(N,g(Le),Y,bA),Z=V+(Y[J]||bA),FA=bA,te=g(TA).getBoundingClientRect(),re=g(X),Pe=g(X)+te.height-FA;if(Z>Pe){var ze=Z-Pe;hc(TA,g(TA).scrollTop+=ze)}if(V<re){var ye=re-V;hc(TA,g(TA).scrollTop-=ye)}}})(p),mn(p)}function kn(p){var N,J,V=g(Le).find(FA=>Pg(p.slice(1),FA)),Z=V?p.slice(0,1).concat(V):p;return(N=(J=g(TA))===null||J===void 0?void 0:J.querySelector('td[data-path="'.concat(sy(Z),'"]')))!==null&&N!==void 0?N:void 0}function Wn(p){var N,{anchor:J,left:V,top:Z,width:FA,height:te,offsetTop:re,offsetLeft:Pe,showTip:ze}=p,ye=function(oA){var{json:YA,documentState:pe,selection:he,readOnly:ge,onEditValue:Bt,onEditRow:$e,onToggleEnforceString:bi,onCut:un,onCopy:Ji,onPaste:Tn,onRemove:Ht,onDuplicateRow:ko,onInsertBeforeRow:Sn,onInsertAfterRow:no,onRemoveRow:gn}=oA,Nt=YA!==void 0,Zi=!!he,_t=YA!==void 0&&he?Ke(YA,et(he)):void 0,st=Nt&&(Kn(he)||kr(he)||an(he)),si=!ge&&Nt&&he!==void 0&&fy(he),Eo=si&&!No(_t),xr=!ge&&st,Hn=he!==void 0&&zg(YA,pe,et(he));return[{type:"separator"},{type:"row",items:[{type:"column",items:[{type:"label",text:"Table cell:"},{type:"dropdown-button",main:{type:"button",onClick:()=>Bt(),icon:lC,text:"Edit",title:"Edit the value (Double-click on the value)",disabled:!si},width:"11em",items:[{type:"button",icon:lC,text:"Edit",title:"Edit the value (Double-click on the value)",onClick:()=>Bt(),disabled:!si},{type:"button",icon:Hn?jS:ZS,text:"Enforce string",title:"Enforce keeping the value as string when it contains a numeric value",onClick:()=>bi(),disabled:!Eo}]},{type:"dropdown-button",main:{type:"button",onClick:()=>un(!0),icon:cC,text:"Cut",title:"Cut selected contents, formatted with indentation (Ctrl+X)",disabled:!xr},width:"10em",items:[{type:"button",icon:cC,text:"Cut formatted",title:"Cut selected contents, formatted with indentation (Ctrl+X)",onClick:()=>un(!0),disabled:ge||!st},{type:"button",icon:cC,text:"Cut compacted",title:"Cut selected contents, without indentation (Ctrl+Shift+X)",onClick:()=>un(!1),disabled:ge||!st}]},{type:"dropdown-button",main:{type:"button",onClick:()=>Ji(!0),icon:F0,text:"Copy",title:"Copy selected contents, formatted with indentation (Ctrl+C)",disabled:!st},width:"12em",items:[{type:"button",icon:F0,text:"Copy formatted",title:"Copy selected contents, formatted with indentation (Ctrl+C)",onClick:()=>Ji(!1),disabled:!st},{type:"button",icon:F0,text:"Copy compacted",title:"Copy selected contents, without indentation (Ctrl+Shift+C)",onClick:()=>Ji(!1),disabled:!st}]},{type:"button",onClick:()=>Tn(),icon:PS,text:"Paste",title:"Paste clipboard contents (Ctrl+V)",disabled:ge||!Zi},{type:"button",onClick:()=>Ht(),icon:E5,text:"Remove",title:"Remove selected contents (Delete)",disabled:ge||!st}]},{type:"column",items:[{type:"label",text:"Table row:"},{type:"button",onClick:()=>$e(),icon:lC,text:"Edit row",title:"Edit the current row",disabled:ge||!Zi||!Nt},{type:"button",onClick:()=>ko(),icon:$S,text:"Duplicate row",title:"Duplicate the current row (Ctrl+D)",disabled:ge||!Zi||!Nt},{type:"button",onClick:()=>Sn(),icon:gC,text:"Insert before",title:"Insert a row before the current row",disabled:ge||!Zi||!Nt},{type:"button",onClick:()=>no(),icon:gC,text:"Insert after",title:"Insert a row after the current row",disabled:ge||!Zi||!Nt},{type:"button",onClick:()=>gn(),icon:E5,text:"Remove row",title:"Remove current row",disabled:ge||!Zi||!Nt}]}]}]}({json:g(eA),documentState:g(kA),selection:g(DA),readOnly:C(),onEditValue:bo,onEditRow:Yn,onToggleEnforceString:Mo,onCut:Ut,onCopy:ft,onPaste:MA,onRemove:ot,onDuplicateRow:on,onInsertBeforeRow:hn,onInsertAfterRow:Ai,onRemoveRow:Ki}),Ge=(N=vA()(ye))!==null&&N!==void 0?N:ye;if(Ge!==!1){var Vi={left:V,top:Z,offsetTop:re,offsetLeft:Pe,width:FA,height:te,anchor:J,closeOnOuterClick:!0,onClose:()=>{Re=!1,M()}};Re=!0;var T=r(GaA,{tip:ze?"Tip: you can open this context menu via right-click or with Ctrl+Q":void 0,items:Ge,onRequestClose(){s(T),M()}},Vi)}}function Vo(p){if(!br(g(DA)))if(p&&(p.stopPropagation(),p.preventDefault()),p&&p.type==="contextmenu"&&p.target!==g(Ze))Wn({left:p.clientX,top:p.clientY,width:W0,height:Z0,showTip:!1});else{var N,J=(N=g(TA))===null||N===void 0?void 0:N.querySelector(".jse-table-cell.jse-selected-value");if(J)Wn({anchor:J,offsetTop:2,width:W0,height:Z0,showTip:!1});else{var V,Z=(V=g(TA))===null||V===void 0?void 0:V.getBoundingClientRect();Z&&Wn({top:Z.top+2,left:Z.left+2,width:W0,height:Z0,showTip:!1})}}}function vo(p){Wn({anchor:OsA(p.target,"BUTTON"),offsetTop:0,width:W0,height:Z0,showTip:!0})}function bo(){if(!C()&&g(DA)){var p=et(g(DA));No(Ke(g(eA),p))?pn(p):y(DA,Ni(p))}}function Yn(){!C()&&g(DA)&&pn(et(g(DA)).slice(0,1))}function Mo(){if(!C()&&an(g(DA))){var p=g(DA).path,N=Ct(p),J=Ke(g(eA),p),V=!zg(g(eA),g(kA),p),Z=V?String(J):DQ(String(J),w());o("handleToggleEnforceString",{enforceString:V,value:J,updatedValue:Z}),Ot([{op:"replace",path:N,value:Z}],(FA,te)=>({state:Ky(g(eA),te,p,{type:"value",enforceString:V})}))}}function ne(){return wi.apply(this,arguments)}function wi(){return(wi=Yt(function*(){if(o("apply pasted json",g(le)),g(le)){var{onPasteAsJson:p}=g(le);p(),setTimeout(M)}})).apply(this,arguments)}function MA(){return me.apply(this,arguments)}function me(){return(me=Yt(function*(){try{HA(yield navigator.clipboard.readText())}catch(p){console.error(p),y(x,!0)}})).apply(this,arguments)}function nt(){return Wt.apply(this,arguments)}function Wt(){return(Wt=Yt(function*(){o("apply pasted multiline text",g(SA)),g(SA)&&(HA(JSON.stringify(g(SA))),setTimeout(M))})).apply(this,arguments)}function Xe(){o("clear pasted json"),y(le,void 0),M()}function oi(){o("clear pasted multiline text"),y(SA,void 0),M()}function Di(){q()(er.text)}function Ut(p){return cn.apply(this,arguments)}function cn(){return(cn=Yt(function*(p){yield RaA({json:g(eA),selection:g(DA),indentation:p?U():void 0,readOnly:C(),parser:w(),onPatch:Ot})})).apply(this,arguments)}function ft(){return Qi.apply(this,arguments)}function Qi(){return Qi=Yt(function*(){var p=!(arguments.length>0&&arguments[0]!==void 0)||arguments[0];g(eA)!==void 0&&(yield xaA({json:g(eA),selection:g(DA),indentation:p?U():void 0,parser:w()}))}),Qi.apply(this,arguments)}function ot(){FaA({json:g(eA),text:g(UA),selection:g(DA),keepSelection:!0,readOnly:C(),onChange:H(),onPatch:Ot})}function Mt(p){C()||(o("extract",{path:p}),Ot(oaA(g(eA),Ni(p))))}function on(){(function(p){var{json:N,selection:J,columns:V,readOnly:Z,onPatch:FA}=p;if(!Z&&N!==void 0&&J&&oQ(J)){var{rowIndex:te,columnIndex:re}=zc(et(J),V);Gs("duplicate row",{rowIndex:te});var Pe=[String(te)];FA(naA(N,[Pe]),(ze,ye)=>({state:ye,selection:Ni(YC({rowIndex:te<N.length?te+1:te,columnIndex:re},V))}))}})({json:g(eA),selection:g(DA),columns:g(Le),readOnly:C(),onPatch:Ot})}function hn(){(function(p){var{json:N,selection:J,columns:V,readOnly:Z,onPatch:FA}=p;if(!Z&&N!==void 0&&J&&oQ(J)){var{rowIndex:te}=zc(et(J),V);Gs("insert before row",{rowIndex:te}),FA(rQ(N,[String(te)],[{key:"",value:xo(N[0])?{}:""}]))}})({json:g(eA),selection:g(DA),columns:g(Le),readOnly:C(),onPatch:Ot})}function Ai(){(function(p){var{json:N,selection:J,columns:V,readOnly:Z,onPatch:FA}=p;if(!Z&&N!==void 0&&J&&oQ(J)){var{rowIndex:te,columnIndex:re}=zc(et(J),V);Gs("insert after row",{rowIndex:te});var Pe=te+1,ze=[String(Pe)],ye=[{key:"",value:xo(N[0])?{}:""}];FA(Pe<N.length?rQ(N,ze,ye):y_(N,[],ye),(Ge,Vi)=>({state:Vi,selection:Ni(YC({rowIndex:Pe,columnIndex:re},V))}))}})({json:g(eA),selection:g(DA),columns:g(Le),readOnly:C(),onPatch:Ot})}function Ki(){(function(p){var{json:N,selection:J,columns:V,readOnly:Z,onPatch:FA}=p;if(!Z&&N!==void 0&&J&&oQ(J)){var{rowIndex:te,columnIndex:re}=zc(et(J),V);Gs("remove row",{rowIndex:te}),FA(py([[String(te)]]),(Pe,ze)=>{var ye=te<Pe.length?te:te>0?te-1:void 0,Ge=ye!==void 0?Ni(YC({rowIndex:ye,columnIndex:re},V)):void 0;return Gs("remove row new selection",{rowIndex:te,newRowIndex:ye,newSelection:Ge}),{state:ze,selection:Ge}})}})({json:g(eA),selection:g(DA),columns:g(Le),readOnly:C(),onPatch:Ot})}function dt(){return(dt=Yt(function*(p){yield NaA({char:p,selectInside:!1,json:g(eA),selection:g(DA),readOnly:C(),parser:w(),onPatch:Ot,onReplaceJson:ve,onSelect:Be})})).apply(this,arguments)}function EA(p){var N;p.preventDefault(),HA((N=p.clipboardData)===null||N===void 0?void 0:N.getData("text/plain"))}function HA(p){p!==void 0&&LaA({clipboardText:p,json:g(eA),selection:g(DA),readOnly:C(),parser:w(),onPatch:Ot,onChangeText:Qt,onPasteMultilineText:_i,openRepairModal:Fn})}function ve(p,N){var J={json:g(eA),text:g(UA)},V={json:g(eA),documentState:g(kA),selection:g(DA),sortedColumn:g(gt),text:g(UA),textIsRepaired:g(Ve)},Z=Qc(p,g(kA)),FA=typeof N=="function"?N(p,Z,g(DA)):void 0;y(eA,FA?.json!==void 0?FA.json:p),y(kA,FA?.state!==void 0?FA.state:Z),y(DA,FA?.selection!==void 0?FA.selection:g(DA)),y(gt,void 0),y(UA,void 0),y(Ve,!1),y(aA,void 0),Ee(g(eA)),Ei(V),Lt(J,void 0)}function Qt(p,N){o("handleChangeText");var J={json:g(eA),text:g(UA)},V={json:g(eA),documentState:g(kA),selection:g(DA),sortedColumn:g(gt),text:g(UA),textIsRepaired:g(Ve)};try{y(eA,_()(p)),y(kA,Qc(g(eA),g(kA))),y(UA,void 0),y(Ve,!1),y(aA,void 0)}catch(FA){try{y(eA,_()(Rc(p))),y(kA,Qc(g(eA),g(kA))),y(UA,p),y(Ve,!0),y(aA,void 0)}catch{y(eA,void 0),y(kA,void 0),y(UA,p),y(Ve,!1),y(aA,g(UA)!==""?dQ(g(UA),FA.message||String(FA)):void 0)}}if(typeof N=="function"){var Z=N(g(eA),g(kA),g(DA));y(eA,Z?.json!==void 0?Z.json:g(eA)),y(kA,Z?.state!==void 0?Z.state:g(kA)),y(DA,Z?.selection!==void 0?Z.selection:g(DA))}Ee(g(eA)),Ei(V),Lt(J,void 0)}function yi(p){o("select validation error",p),y(DA,Ni(p.path)),dn(p.path)}function ri(p){if(g(eA)!==void 0){var{id:N,onTransform:J,onClose:V}=p,Z=p.rootPath||[];Re=!0,VA()({id:N||l,json:g(eA),rootPath:Z||[],onTransform:FA=>{J?J({operations:FA,json:g(eA),transformedJson:Da(g(eA),FA)}):(o("onTransform",Z,FA),Ot(FA))},onClose:()=>{Re=!1,setTimeout(M),V&&V()}})}}function pn(p){o("openJSONEditorModal",{path:p}),Re=!0,oe()({content:{json:Ke(g(eA),p)},path:p,onPatch:Ot,onClose:()=>{Re=!1,setTimeout(M)}})}function Fn(p,N){y(He,{text:p,onParse:J=>sf(J,V=>rf(V,w())),onRepair:_sA,onApply:N,onClose:M})}function Jn(){(function(p){C()||g(eA)===void 0||(Re=!0,pA()({id:c,json:g(eA),rootPath:p,onSort:N=>{var{operations:J,itemPath:V,direction:Z}=N;o("onSort",J,p,V,Z),Ot(J,(FA,te)=>({state:te,sortedColumn:{path:V,sortDirection:Z===-1?Pc.desc:Pc.asc}}))},onClose:()=>{Re=!1,setTimeout(M)}}))})([])}function ln(){ri({rootPath:[]})}function Pt(p){o("openFind",{findAndReplace:p}),y(mA,!1),y(sA,!1),io(),y(mA,!0),y(sA,p)}function $i(){if(!C()&&E().canUndo){var p=E().undo();if(uy(p)){var N={json:g(eA),text:g(UA)};y(eA,p.undo.patch?Da(g(eA),p.undo.patch):p.undo.json),y(kA,p.undo.documentState),y(DA,p.undo.selection),y(gt,p.undo.sortedColumn),y(UA,p.undo.text),y(Ve,p.undo.textIsRepaired),y(aA,void 0),o("undo",{item:p,json:g(eA)}),Lt(N,p.undo.patch&&p.redo.patch?{json:g(eA),previousJson:N.json,redo:p.undo.patch,undo:p.redo.patch}:void 0),M(),g(DA)&&dn(et(g(DA)),{scrollToWhenVisible:!1})}else gA()(p)}}function Rr(){if(!C()&&E().canRedo){var p=E().redo();if(uy(p)){var N={json:g(eA),text:g(UA)};y(eA,p.redo.patch?Da(g(eA),p.redo.patch):p.redo.json),y(kA,p.redo.documentState),y(DA,p.redo.selection),y(gt,p.redo.sortedColumn),y(UA,p.redo.text),y(Ve,p.redo.textIsRepaired),y(aA,void 0),o("redo",{item:p,json:g(eA)}),Lt(N,p.undo.patch&&p.redo.patch?{json:g(eA),previousJson:N.json,redo:p.redo.patch,undo:p.undo.patch}:void 0),M(),g(DA)&&dn(et(g(DA)),{scrollToWhenVisible:!1})}else QA()(p)}}function Ft(p){y(P,p.getBoundingClientRect().height)}fA(()=>(k(D()),k(L())),()=>{y(KA,V_({escapeControlCharacters:D(),escapeUnicodeCharacters:L()}))}),fA(()=>g(mA),()=>{(function(p){if(g(TA)){var N=p?_3:-100;g(TA).scrollTo({top:hc(TA,g(TA).scrollTop+=N),left:g(TA).scrollLeft})}})(g(mA))}),fA(()=>k(d()),()=>{(function(p){var N={json:g(eA)},J=H3(p)?p.text!==g(UA):!di(N.json,p.json);if(o("update external content",{isChanged:J}),J){var V={json:g(eA),documentState:g(kA),selection:g(DA),sortedColumn:g(gt),text:g(UA),textIsRepaired:g(Ve)};if(H3(p))try{y(eA,_()(p.text)),y(kA,Qc(g(eA),g(kA))),y(UA,p.text),y(Ve,!1),y(aA,void 0)}catch(Z){try{y(eA,_()(Rc(p.text))),y(kA,Qc(g(eA),g(kA))),y(UA,p.text),y(Ve,!0),y(aA,void 0)}catch{y(eA,void 0),y(kA,void 0),y(UA,p.text),y(Ve,!1),y(aA,g(UA)!==""?dQ(g(UA),Z.message||String(Z)):void 0)}}else y(eA,p.json),y(kA,Qc(g(eA),g(kA))),y(UA,void 0),y(Ve,!1),y(aA,void 0);Ee(g(eA)),y(gt,void 0),Ei(V)}})(d())}),fA(()=>k(B()),()=>{(function(p){di(g(DA),p)||(o("applyExternalSelection",{selection:g(DA),externalSelection:p}),q3(p)&&y(DA,p))})(B())}),fA(()=>(g(Le),g(eA),k(R()),g(_e)),()=>{y(Le,wo(g(eA))?function(p,N){var J=new Set(N.map(Ct)),V=new Set(p.map(Ct));for(var Z of J)V.has(Z)||J.delete(Z);for(var FA of V)J.has(FA)||J.add(FA);return[...J].map(ks)}(rJA(g(eA),R(),g(_e)),g(Le)):[])}),fA(()=>(g(eA),g(Le)),()=>{y(bt,!(!g(eA)||Oi(g(Le))))}),fA(()=>(g(eA),g(_e)),()=>{y(A,Array.isArray(g(eA))&&g(eA).length>g(_e))}),fA(()=>(g(X),g(P),g(eA),g(mA),_3),()=>{y(i,sJA(g(X),g(P),g(eA),Y,bA,g(mA)?_3:0))}),fA(()=>g(eA),()=>{g(eA),g(TA)&&g(TA).scrollTo({top:g(TA).scrollTop,left:g(TA).scrollLeft})}),fA(()=>g(DA),()=>{var p;p=g(DA),di(p,B())||(o("onSelect",p),j()(p))}),fA(()=>(k(C()),k(h()),k(w()),g(KA),g(eA),g(kA),k(BA())),()=>{y(rt,{mode:er.table,readOnly:C(),truncateTextSize:h(),parser:w(),normalization:g(KA),getJson:()=>g(eA),getDocumentState:()=>g(kA),findElement:kn,findNextInside:Tt,focus:M,onPatch:(p,N)=>Ot(function(J,V){return J.flatMap(Z=>{if(k8(Z)){var FA=ks(Z.path);if(FA.length>0){for(var te=[Z],re=Fi(FA);re.length>0&&!Ms(V,re);)te.unshift({op:"add",path:Ct(re),value:{}}),re=Fi(re);return te}}return Z})}(p,g(eA)),N),onSelect:Be,onFind:Pt,onPasteJson:ii,onRenderValue:BA()})}),fA(()=>(g(eA),k(K()),k(w()),k(z())),()=>{xe(g(eA),K(),w(),z())}),fA(()=>(g(tn),g(Le)),()=>{y(n,aJA(g(tn),g(Le)))}),Qn(),Zt(!0);var Xn=fTA();ce("mousedown",A2,function(p){!yQ(p.target,N=>N===g(CA))&&br(g(DA))&&(o("click outside the editor, exit edit mode"),y(DA,Hg(g(DA))),$t&&g(Ze)&&(g(Ze).focus(),g(Ze).blur()),o("blur (outside editor)"),g(Ze)&&g(Ze).blur())});var se,vi=vt(Xn),Yi=W(vi),rn=p=>{(function(N,J){mt(J,!1);var V=b(J,"containsValidArray",9),Z=b(J,"readOnly",9),FA=b(J,"showSearch",13,!1),te=b(J,"history",9),re=b(J,"onSort",9),Pe=b(J,"onTransform",9),ze=b(J,"onContextMenu",9),ye=b(J,"onUndo",9),Ge=b(J,"onRedo",9),Vi=b(J,"onRenderMenu",9);function T(){FA(!FA())}var oA=$(void 0,!0),YA=$(void 0,!0);fA(()=>(k(Z()),k(re()),k(V()),k(Pe()),k(ze()),k(ye()),k(te()),k(Ge())),()=>{y(oA,Z()?[{type:"space"}]:[{type:"button",icon:l4,title:"Sort",className:"jse-sort",onClick:re(),disabled:Z()||!V()},{type:"button",icon:s4,title:"Transform contents (filter, sort, project)",className:"jse-transform",onClick:Pe(),disabled:Z()||!V()},{type:"button",icon:g4,title:"Search (Ctrl+F)",className:"jse-search",onClick:T,disabled:!V()},{type:"button",icon:WS,title:AG,className:"jse-contextmenu",onClick:ze()},{type:"separator"},{type:"button",icon:h5,title:"Undo (Ctrl+Z)",className:"jse-undo",onClick:ye(),disabled:!te().canUndo},{type:"button",icon:Q5,title:"Redo (Ctrl+Shift+Z)",className:"jse-redo",onClick:Ge(),disabled:!te().canRedo},{type:"space"}])}),fA(()=>(k(Vi()),g(oA)),()=>{y(YA,Vi()(g(oA))||g(oA))}),Qn(),Zt(!0),Oy(N,{get items(){return g(YA)}}),pt()})(p,{get containsValidArray(){return g(bt)},get readOnly(){return C()},get history(){return E()},onSort:Jn,onTransform:ln,onUndo:$i,onRedo:Rr,onContextMenu:vo,get onRenderMenu(){return lA()},get showSearch(){return g(mA)},set showSearch(N){y(mA,N)},$$legacy:!0})};LA(Yi,p=>{u()&&p(rn)});var Hr=IA(Yi,2),Ri=p=>{var N=hTA(),J=vt(N),V=W(J);V.readOnly=!0,Co(V,re=>y(Ze,re),()=>g(Ze));var Z=IA(J,2),FA=re=>{var Pe=ETA(),ze=vt(Pe);MaA(W(ze),{get json(){return g(eA)},get documentState(){return g(kA)},get parser(){return w()},get showSearch(){return g(mA)},get showReplace(){return g(sA)},get readOnly(){return C()},get columns(){return g(Le)},onSearch:xt,onFocus:tt,onPatch:Ot,onClose:Dt});var ye=IA(ze,2),Ge=W(ye),Vi=W(Ge),T=W(Vi),oA=W(T),YA=W(oA),pe=st=>{var si=_o(),Eo=qA(()=>(k(XE),g(n),nA(()=>{var So;return XE([],(So=g(n))===null||So===void 0?void 0:So.root)}))),xr=vt(si),Hn=So=>{var zr=aTA();gQ(W(zr),{get validationError(){return g(Eo)},get onExpand(){return Oc}}),iA(So,zr)};LA(xr,So=>{g(Eo)&&So(Hn)}),iA(st,si)};LA(YA,st=>{k(Oi),g(n),nA(()=>{var si;return!Oi((si=g(n))===null||si===void 0?void 0:si.root)})&&st(pe)});var he=IA(oA);qo(he,1,()=>g(Le),or,(st,si)=>{var Eo=cTA();(function(xr,Hn){mt(Hn,!1);var So=$(void 0,!0),zr=$(void 0,!0),t0=$(void 0,!0),Ta=b(Hn,"path",9),Wc=b(Hn,"sortedColumn",9),Hl=b(Hn,"readOnly",9),Xc=b(Hn,"onSort",9);fA(()=>(k(Ta()),Ka),()=>{y(So,Oi(Ta())?"values":Ka(Ta()))}),fA(()=>(k(Wc()),k(Ta())),()=>{var $n;y(zr,Wc()&&di(Ta(),($n=Wc())===null||$n===void 0?void 0:$n.path)?Wc().sortDirection:void 0)}),fA(()=>(g(zr),orA),()=>{y(t0,g(zr)?orA[g(zr)]:void 0)}),Qn(),Zt(!0);var Or,Pr=WJA(),Ha=W(Pr),i0=W(Ha),za=IA(Ha,2),Ko=$n=>{var Zo=ZJA(),zl=W(Zo),Id=qA(()=>(g(zr),k(Pc),k(mg),k(VS),nA(()=>g(zr)===Pc.asc?mg:VS)));ji(zl,{get data(){return g(Id)}}),De(()=>En(Zo,"title","Currently sorted in ".concat(g(t0)," order"))),iA($n,Zo)};LA(za,$n=>{g(zr)!==void 0&&$n(Ko)}),De(($n,Zo)=>{Or=Vt(Pr,1,"jse-column-header svelte-2i3vdx",null,Or,$n),En(Pr,"title",Hl()?g(So):g(So)+" (Click to sort the data by this column)"),wt(i0,Zo)},[()=>({"jse-readonly":Hl()}),()=>(k(V0),g(So),k(50),nA(()=>V0(g(So),50)))],qA),ce("click",Pr,function(){Hl()||Xc()({path:Ta(),sortDirection:g(zr)===Pc.asc?Pc.desc:Pc.asc})}),iA(xr,Pr),pt()})(W(Eo),{get path(){return g(si)},get sortedColumn(){return g(gt)},get readOnly(){return C()},onSort:ZA}),iA(st,Eo)});var ge=IA(he),Bt=st=>{var si=lTA(),Eo=W(si),xr=qA(()=>(g(eA),nA(()=>Array.isArray(g(eA))?g(eA).length:0)));(function(Hn,So){mt(So,!1);var zr=b(So,"count",9),t0=b(So,"maxSampleCount",9),Ta=b(So,"readOnly",9),Wc=b(So,"onRefresh",9);Zt(!0);var Hl,Xc=sTA();ji(W(Xc),{get data(){return qW}}),De(Or=>{Hl=Vt(Xc,1,"jse-column-header svelte-fzj761",null,Hl,Or),En(Xc,"title","The Columns are created by sampling ".concat(t0()," items out of ").concat(zr(),". ")+"If you're missing a column, click here to sample all of the items instead of a subset. This is slower.")},[()=>({"jse-readonly":Ta()})],qA),ce("click",Xc,()=>Wc()()),iA(Hn,Xc),pt()})(Eo,{get count(){return g(xr)},get maxSampleCount(){return g(_e)},get readOnly(){return C()},onRefresh:()=>y(_e,1/0)}),iA(st,si)};LA(ge,st=>{g(A)&&st(Bt)});var $e,bi,un=IA(T),Ji=W(un),Tn=IA(un);qo(Tn,1,()=>(g(i),nA(()=>g(i).visibleItems)),or,(st,si,Eo)=>{var xr=BTA(),Hn=qA(()=>(g(i),nA(()=>g(i).startIndex+Eo))),So=qA(()=>(g(n),k(g(Hn)),nA(()=>g(n).rows[g(Hn)]))),zr=qA(()=>(k(XE),k(g(Hn)),k(g(So)),nA(()=>{var Or;return XE([String(g(Hn))],(Or=g(So))===null||Or===void 0?void 0:Or.row)}))),t0=qA(()=>(k(Jg),g(eA),g(Ue),k(g(Hn)),nA(()=>Jg(g(eA),g(Ue),[String(g(Hn))])))),Ta=W(xr);vsA(Ta,()=>g(Hn),Or=>{var Pr=gTA(),Ha=W(Pr),i0=IA(Ha),za=Ko=>{gQ(Ko,{get validationError(){return g(zr)},get onExpand(){return Oc}})};LA(i0,Ko=>{g(zr)&&Ko(za)}),Us(Pr,(Ko,$n)=>ny?.(Ko,$n),()=>Ko=>function($n,Zo){Y[Zo]=$n.getBoundingClientRect().height}(Ko,g(Hn))),De(()=>{var Ko;return wt(Ha,"".concat((Ko=g(Hn))!==null&&Ko!==void 0?Ko:""," "))}),iA(Or,Pr)});var Wc=IA(Ta);qo(Wc,1,()=>g(Le),or,(Or,Pr,Ha,i0)=>{var za,Ko=CTA(),$n=qA(()=>(k(g(Hn)),g(Pr),nA(()=>[String(g(Hn))].concat(g(Pr))))),Zo=qA(()=>(k(Ke),g(si),g(Pr),nA(()=>Ke(g(si),g(Pr))))),zl=qA(()=>(k(an),g(DA),k(Pg),k(g($n)),nA(()=>an(g(DA))&&Pg(g(DA).path,g($n))))),Id=qA(()=>(k(g(So)),nA(()=>{var rr;return(rr=g(So))===null||rr===void 0?void 0:rr.columns[Ha]}))),Cd=qA(()=>(k(XE),k(g($n)),k(g(Id)),nA(()=>XE(g($n),g(Id))))),JQ=W(Ko),dd=W(JQ),TQ=rr=>{var ta=qA(()=>(k(wy),k(Jg),g(si),k(g(t0)),g(Pr),nA(()=>wy(Jg(g(si),g(t0),g(Pr)))))),HQ=qA(()=>(k(g(ta)),nA(()=>!!g(ta)&&g(ta).some(aI=>aI.active)))),zQ=qA(()=>(k(Oi),k(g(ta)),nA(()=>!Oi(g(ta)))));(function(aI,is){mt(is,!1);var OQ=b(is,"path",9),OU=b(is,"value",9),PU=b(is,"parser",9),LgA=b(is,"isSelected",9),FgA=b(is,"containsSearchResult",9),NgA=b(is,"containsActiveSearchResult",9),_gA=b(is,"onEdit",9);Zt(!0);var jU,Hf=VJA(),GgA=W(Hf);De((PQ,UgA)=>{jU=Vt(Hf,1,"jse-inline-value svelte-h57m0p",null,jU,PQ),wt(GgA,UgA)},[()=>({"jse-selected":LgA(),"jse-highlight":FgA(),"jse-active":NgA()}),()=>(k(V0),k(PU()),k(OU()),k(50),nA(()=>{var PQ;return V0((PQ=PU().stringify(OU()))!==null&&PQ!==void 0?PQ:"",50)}))],qA),ce("dblclick",Hf,()=>_gA()(OQ())),iA(aI,Hf),pt()})(rr,{get path(){return g($n)},get value(){return g(Zo)},get parser(){return w()},get isSelected(){return g(zl)},get containsSearchResult(){return g(zQ)},get containsActiveSearchResult(){return g(HQ)},onEdit:pn})},U7=rr=>{var ta=qA(()=>(k(Jg),g(eA),g(Ue),k(g($n)),nA(()=>{var is;return(is=Jg(g(eA),g(Ue),g($n)))===null||is===void 0?void 0:is.searchResults}))),HQ=qA(()=>g(Zo)!==void 0?g(Zo):""),zQ=qA(()=>(k(zg),g(eA),g(kA),k(g($n)),nA(()=>zg(g(eA),g(kA),g($n))))),aI=qA(()=>g(zl)?g(DA):void 0);vaA(rr,{get path(){return g($n)},get value(){return g(HQ)},get enforceString(){return g(zQ)},get selection(){return g(aI)},get searchResultItems(){return g(ta)},get context(){return g(rt)}})};LA(dd,rr=>{k(No),k(g(Zo)),nA(()=>No(g(Zo)))?rr(TQ):rr(U7,!1)});var K7=IA(dd),Y7=rr=>{var ta=ITA();_1(W(ta),{selected:!0,onContextMenu:Wn}),iA(rr,ta)};LA(K7,rr=>{k(C()),k(g(zl)),k(br),g(DA),nA(()=>!C()&&g(zl)&&!br(g(DA)))&&rr(Y7)});var Ol=IA(JQ,2),sI=rr=>{gQ(rr,{get validationError(){return g(Cd)},get onExpand(){return Oc}})};LA(Ol,rr=>{g(Cd)&&rr(sI)}),De((rr,ta)=>{En(Ko,"data-path",rr),za=Vt(JQ,1,"jse-value-outer svelte-u14cgx",null,za,ta)},[()=>(k(sy),k(g($n)),nA(()=>sy(g($n)))),()=>({"jse-selected-value":g(zl)})],qA),iA(Or,Ko)});var Hl=IA(Wc),Xc=Or=>{iA(Or,dTA())};LA(Hl,Or=>{g(A)&&Or(Xc)}),iA(st,xr)});var Ht,ko=W(IA(Tn));Co(ye,st=>y(TA,st),()=>g(TA)),Us(ye,(st,si)=>ny?.(st,si),()=>Ft),es(()=>ce("scroll",ye,We));var Sn=IA(ye,2),no=st=>{var si=qA(()=>(g(le),nA(()=>"You pasted a JSON ".concat(Array.isArray(g(le).contents)?"array":"object"," as text")))),Eo=qA(()=>[{icon:L0,text:"Paste as JSON instead",title:"Paste the text as JSON instead of a single value",onMouseDown:ne},{text:"Leave as is",title:"Keep the pasted content as a single value",onClick:Xe}]);mc(st,{type:"info",get message(){return g(si)},get actions(){return g(Eo)}})};LA(Sn,st=>{g(le)&&st(no)});var gn=IA(Sn,2),Nt=st=>{var si=qA(()=>[{icon:L0,text:"Paste as string instead",title:"Paste the clipboard data as a single string value instead of an array",onClick:nt},{text:"Leave as is",title:"Keep the pasted array",onClick:oi}]);mc(st,{type:"info",message:"Multiline text was pasted as array",get actions(){return g(si)}})};LA(gn,st=>{g(SA)&&st(Nt)});var Zi=IA(gn,2),_t=st=>{var si=qA(()=>C()?[]:[{icon:u5,text:"Ok",title:"Accept the repaired document",onClick:pi},{icon:a4,text:"Repair manually instead",title:"Leave the document unchanged and repair it manually instead",onClick:Di}]);mc(st,{type:"success",message:"The loaded JSON document was invalid but is successfully repaired.",get actions(){return g(si)},onClose:M})};LA(Zi,st=>{g(Ve)&&st(_t)}),dG(IA(Zi,2),{get validationErrors(){return g(tn)},selectError:yi}),De((st,si,Eo)=>{$e=Vt(un,1,"jse-table-invisible-start-section svelte-u14cgx",null,$e,st),En(Ji,"colspan",(g(Le),nA(()=>g(Le).length))),bi=Ll(Ji,"",bi,si),En(ko,"colspan",(g(Le),nA(()=>g(Le).length))),Ht=Ll(ko,"",Ht,Eo)},[()=>({"jse-search-box-background":g(mA)}),()=>({height:(g(i),nA(()=>g(i).startHeight+"px"))}),()=>({height:(g(i),nA(()=>g(i).endHeight+"px"))})],qA),iA(re,Pe)},te=(re,Pe)=>{var ze=Ge=>{var Vi=QTA(),T=vt(Vi),oA=qA(()=>C()?[]:[{icon:a4,text:"Repair manually",title:'Open the document in "code" mode and repair it manually',onClick:Di}]);mc(T,{type:"error",message:"The loaded JSON document is invalid and could not be repaired automatically.",get actions(){return g(oA)}}),_aA(IA(T,2),{get text(){return g(UA)},get json(){return g(eA)},get indentation(){return U()},get parser(){return w()}}),iA(Ge,Vi)},ye=Ge=>{rTA(Ge,{get text(){return g(UA)},get json(){return g(eA)},get readOnly(){return C()},get parser(){return w()},openJSONEditorModal:pn,extractPath:Mt,get onChangeMode(){return q()},onClick:()=>{M()}})};LA(re,Ge=>{g(aA)&&g(UA)!==void 0&&g(UA)!==""?Ge(ze):Ge(ye,!1)},Pe)};LA(Z,re=>{g(bt)?re(FA):re(te,!1)}),ce("paste",V,EA),iA(p,N)},fs=p=>{iA(p,uTA())};LA(Hr,p=>{I?p(fs,!1):p(Ri)}),Co(vi,p=>y(CA,p),()=>g(CA));var Bo=IA(vi,2),Q=p=>{DaA(p,{onClose:()=>y(x,!1)})};LA(Bo,p=>{g(x)&&p(Q)});var m=IA(Bo,2),v=p=>{yaA(p,z1(()=>g(He),{onClose:()=>{var N;(N=g(He))===null||N===void 0||N.onClose(),y(He,void 0)}}))};return LA(m,p=>{g(He)&&p(v)}),De(p=>se=Vt(vi,1,"jse-table-mode svelte-u14cgx",null,se,p),[()=>({"no-main-menu":!u()})],qA),ce("mousedown",vi,function(p){if(p.buttons===1||p.buttons===2){var N=p.target;N.isContentEditable||M();var J=PsA(N);if(J){if(br(g(DA))&&V3(g(eA),g(DA),J))return;y(DA,Ni(J)),p.preventDefault()}}}),ce("keydown",vi,function(p){var N=n2(p);if(o("keydown",{combo:N,key:p.key}),N==="Ctrl+X"&&(p.preventDefault(),Ut(!0)),N==="Ctrl+Shift+X"&&(p.preventDefault(),Ut(!1)),N==="Ctrl+C"&&(p.preventDefault(),ft(!0)),N==="Ctrl+Shift+C"&&(p.preventDefault(),ft(!1)),N==="Ctrl+D"&&(p.preventDefault(),on()),N!=="Delete"&&N!=="Backspace"||(p.preventDefault(),ot()),N==="Insert"&&p.preventDefault(),N==="Ctrl+A"&&p.preventDefault(),N==="Ctrl+Q"&&Vo(p),N==="ArrowLeft"&&(p.preventDefault(),ni(),g(DA))){var J=function(Pe,ze){var{rowIndex:ye,columnIndex:Ge}=zc(et(ze),Pe);return Ge>0?Ni(YC({rowIndex:ye,columnIndex:Ge-1},Pe)):ze}(g(Le),g(DA));y(DA,J),Uo(et(J))}if(N==="ArrowRight"&&(p.preventDefault(),ni(),g(DA))){var V=function(Pe,ze){var{rowIndex:ye,columnIndex:Ge}=zc(et(ze),Pe);return Ge<Pe.length-1?Ni(YC({rowIndex:ye,columnIndex:Ge+1},Pe)):ze}(g(Le),g(DA));y(DA,V),Uo(et(V))}if(N==="ArrowUp"&&(p.preventDefault(),ni(),g(DA))){var Z=function(Pe,ze){var{rowIndex:ye,columnIndex:Ge}=zc(et(ze),Pe);return ye>0?Ni(YC({rowIndex:ye-1,columnIndex:Ge},Pe)):ze}(g(Le),g(DA));y(DA,Z),Uo(et(Z))}if(N==="ArrowDown"&&(p.preventDefault(),ni(),g(DA))){var FA=function(Pe,ze,ye){var{rowIndex:Ge,columnIndex:Vi}=zc(et(ye),ze);return Ge<Pe.length-1?Ni(YC({rowIndex:Ge+1,columnIndex:Vi},ze)):ye}(g(eA),g(Le),g(DA));y(DA,FA),Uo(et(FA))}if(N==="Enter"&&g(DA)&&an(g(DA))){p.preventDefault();var te=g(DA).path;No(Ke(g(eA),te))?pn(te):C()||y(DA,fe(fe({},g(DA)),{},{edit:!0}))}if(N.replace(/^Shift\+/,"").length===1&&g(DA))return p.preventDefault(),void function(Pe){dt.apply(this,arguments)}(p.key);if(N==="Ctrl+Enter"&&an(g(DA))){p.preventDefault();var re=Ke(g(eA),g(DA).path);Uy(re)&&window.open(String(re),"_blank")}N==="Escape"&&g(DA)&&(p.preventDefault(),y(DA,void 0)),N==="Ctrl+F"&&(p.preventDefault(),Pt(!1)),N==="Ctrl+H"&&(p.preventDefault(),Pt(!0)),N==="Ctrl+Z"&&(p.preventDefault(),$i()),N==="Ctrl+Shift+Z"&&(p.preventDefault(),Rr())}),ce("contextmenu",vi,Vo),iA(t,Xn),zt(e,"validate",nn),zt(e,"patch",mi),zt(e,"focus",M),zt(e,"acceptAutoRepair",pi),zt(e,"scrollTo",dn),zt(e,"findElement",kn),zt(e,"openTransformModal",ri),pt({validate:nn,patch:mi,focus:M,acceptAutoRepair:pi,scrollTo:dn,findElement:kn,openTransformModal:ri})}function HrA(t,e){mt(e,!1);var A=b(e,"content",8),i=b(e,"selection",12),n=b(e,"readOnly",8),o=b(e,"indentation",8),r=b(e,"tabSize",8),s=b(e,"truncateTextSize",8),a=b(e,"externalMode",8),c=b(e,"mainMenuBar",8),l=b(e,"navigationBar",8),I=b(e,"statusBar",8),C=b(e,"askToFormat",8),d=b(e,"escapeControlCharacters",8),B=b(e,"escapeUnicodeCharacters",8),E=b(e,"flattenColumns",8),h=b(e,"parser",8),u=b(e,"parseMemoizeOne",8),D=b(e,"validator",8),L=b(e,"validationParser",8),R=b(e,"pathParser",8),w=b(e,"insideModal",8),_=b(e,"onChange",8),K=b(e,"onChangeMode",8),z=b(e,"onSelect",8),U=b(e,"onRenderValue",8),H=b(e,"onClassName",8),q=b(e,"onRenderMenu",8),j=b(e,"onRenderContextMenu",8),gA=b(e,"onError",8),QA=b(e,"onFocus",8),BA=b(e,"onBlur",8),lA=b(e,"onSortModal",8),vA=b(e,"onTransformModal",8),tA=b(e,"onJSONEditorModal",8),cA=$(),pA=$(),VA=$(),oe=Sr("jsoneditor:JSONEditorRoot"),KA=$(KaA({onChange:x=>y(KA,x)}).get()),CA=$(a());function TA(x){if(grA(x)){y(CA,x.undo.mode);var Y=g(KA).items(),P=Y.findIndex(bA=>bA===x),X=P!==-1?Y[P-1]:void 0;oe("handleUndo",{index:P,item:x,items:Y,prevItem:X}),X&&i(X.redo.selection),K()(g(CA))}}function Ze(x){if(grA(x)){y(CA,x.redo.mode);var Y=g(KA).items(),P=Y.findIndex(bA=>bA===x),X=P!==-1?Y[P+1]:void 0;oe("handleRedo",{index:P,item:x,items:Y,nextItem:X}),X&&i(X.undo.selection),K()(g(CA))}}var He=$(),uA={type:"separator"},eA=$(),UA=$();function aA(x){if(g(cA))return g(cA).patch(x);if(g(pA))return g(pA).patch(x);if(g(VA))return g(VA).patch(x);throw new Error('Method patch is not available in mode "'.concat(g(CA),'"'))}function le(x,Y){if(g(cA))return g(cA).expand(x,Y);throw new Error('Method expand is not available in mode "'.concat(g(CA),'"'))}function SA(x,Y){if(g(cA))return g(cA).collapse(x,Y);throw new Error('Method collapse is not available in mode "'.concat(g(CA),'"'))}function Ue(x){if(g(VA))g(VA).openTransformModal(x);else if(g(cA))g(cA).openTransformModal(x);else{if(!g(pA))throw new Error('Method transform is not available in mode "'.concat(g(CA),'"'));g(pA).openTransformModal(x)}}function mA(){if(g(VA))return g(VA).validate();if(g(cA))return g(cA).validate();if(g(pA))return g(pA).validate();throw new Error('Method validate is not available in mode "'.concat(g(CA),'"'))}function sA(){return g(cA)?g(cA).acceptAutoRepair():A()}function xt(x){if(g(cA))return g(cA).scrollTo(x);if(g(pA))return g(pA).scrollTo(x);throw new Error('Method scrollTo is not available in mode "'.concat(g(CA),'"'))}function tt(x){if(g(cA))return g(cA).findElement(x);if(g(pA))return g(pA).findElement(x);throw new Error('Method findElement is not available in mode "'.concat(g(CA),'"'))}function de(){g(VA)?g(VA).focus():g(cA)?g(cA).focus():g(pA)&&g(pA).focus()}function Dt(){return _e.apply(this,arguments)}function _e(){return(_e=Yt(function*(){g(VA)&&(yield g(VA).refresh())})).apply(this,arguments)}fA(()=>k(a()),()=>{(function(x){if(x!==g(CA)){var Y={type:"mode",undo:{mode:g(CA),selection:void 0},redo:{mode:x,selection:void 0}};g(CA)==="text"&&g(VA)&&g(VA).flush(),oe("add history item",Y),g(KA).add(Y),y(CA,x)}})(a())}),fA(()=>(g(CA),k(K())),()=>{y(He,[{type:"button",text:"text",title:"Switch to text mode (current mode: ".concat(g(CA),")"),className:"jse-group-button jse-first"+(g(CA)===er.text?" jse-selected":""),onClick:()=>K()(er.text)},{type:"button",text:"tree",title:"Switch to tree mode (current mode: ".concat(g(CA),")"),className:"jse-group-button "+(g(CA)===er.tree?" jse-selected":""),onClick:()=>K()(er.tree)},{type:"button",text:"table",title:"Switch to table mode (current mode: ".concat(g(CA),")"),className:"jse-group-button jse-last"+(g(CA)===er.table?" jse-selected":""),onClick:()=>K()(er.table)}])}),fA(()=>(g(He),k(q()),g(CA),k(w()),k(n())),()=>{y(eA,x=>{var Y=p_(x[0])?g(He).concat(x):g(He).concat(uA,x),P=i4(Y);return q()(Y,{mode:g(CA),modal:w(),readOnly:n()})||P})}),fA(()=>(k(j()),g(CA),k(w()),k(n()),k(i())),()=>{y(UA,x=>{var Y,P=i4(x);return(Y=j()(x,{mode:g(CA),modal:w(),readOnly:n(),selection:i()}))!==null&&Y!==void 0?Y:!n()&&P})}),Qn(),Zt();var Le=_o(),bt=vt(Le),Re=x=>{Co(qJA(x,{get externalContent(){return A()},get externalSelection(){return i()},get history(){return g(KA)},get readOnly(){return n()},get indentation(){return o()},get tabSize(){return r()},get mainMenuBar(){return c()},get statusBar(){return I()},get askToFormat(){return C()},get escapeUnicodeCharacters(){return B()},get parser(){return h()},get validator(){return D()},get validationParser(){return L()},get onChange(){return _()},get onChangeMode(){return K()},get onSelect(){return z()},onUndo:TA,onRedo:Ze,get onError(){return gA()},get onFocus(){return QA()},get onBlur(){return BA()},get onRenderMenu(){return g(eA)},get onSortModal(){return lA()},get onTransformModal(){return vA()},$$legacy:!0}),Y=>y(VA,Y),()=>g(VA))},$t=(x,Y)=>{var P=bA=>{Co(mTA(bA,{get externalContent(){return A()},get externalSelection(){return i()},get history(){return g(KA)},get readOnly(){return n()},get truncateTextSize(){return s()},get mainMenuBar(){return c()},get escapeControlCharacters(){return d()},get escapeUnicodeCharacters(){return B()},get flattenColumns(){return E()},get parser(){return h()},get parseMemoizeOne(){return u()},get validator(){return D()},get validationParser(){return L()},get indentation(){return o()},get onChange(){return _()},get onChangeMode(){return K()},get onSelect(){return z()},onUndo:TA,onRedo:Ze,get onRenderValue(){return U()},get onFocus(){return QA()},get onBlur(){return BA()},get onRenderMenu(){return g(eA)},get onRenderContextMenu(){return g(UA)},get onSortModal(){return lA()},get onTransformModal(){return vA()},get onJSONEditorModal(){return tA()},$$legacy:!0}),Be=>y(pA,Be),()=>g(pA))},X=bA=>{Co(U_(bA,{get externalContent(){return A()},get externalSelection(){return i()},get history(){return g(KA)},get readOnly(){return n()},get indentation(){return o()},get truncateTextSize(){return s()},get mainMenuBar(){return c()},get navigationBar(){return l()},get escapeControlCharacters(){return d()},get escapeUnicodeCharacters(){return B()},get parser(){return h()},get parseMemoizeOne(){return u()},get validator(){return D()},get validationParser(){return L()},get pathParser(){return R()},get onError(){return gA()},get onChange(){return _()},get onChangeMode(){return K()},get onSelect(){return z()},onUndo:TA,onRedo:Ze,get onRenderValue(){return U()},get onClassName(){return H()},get onFocus(){return QA()},get onBlur(){return BA()},get onRenderMenu(){return g(eA)},get onRenderContextMenu(){return g(UA)},get onSortModal(){return lA()},get onTransformModal(){return vA()},get onJSONEditorModal(){return tA()},$$legacy:!0}),Be=>y(cA,Be),()=>g(cA))};LA(x,bA=>{g(CA),k(er),nA(()=>g(CA)===er.table)?bA(P):bA(X,!1)},Y)};return LA(bt,x=>{g(CA),k(er),nA(()=>g(CA)===er.text||String(g(CA))==="code")?x(Re):x($t,!1)}),iA(t,Le),zt(e,"patch",aA),zt(e,"expand",le),zt(e,"collapse",SA),zt(e,"transform",Ue),zt(e,"validate",mA),zt(e,"acceptAutoRepair",sA),zt(e,"scrollTo",xt),zt(e,"findElement",tt),zt(e,"focus",de),zt(e,"refresh",Dt),pt({patch:aA,expand:le,collapse:SA,transform:Ue,validate:mA,acceptAutoRepair:sA,scrollTo:xt,findElement:tt,focus:de,refresh:Dt})}Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-modal-wrapper.svelte-v0el4e {
  flex: 1;
  display: flex;
  min-width: 0;
  min-height: 0;
  flex-direction: column;
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) {
  flex: 1;
  display: flex;
  flex-direction: column;
  padding: 20px;
  overflow: auto;
  min-width: 0;
  min-height: 0;
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-actions:where(.svelte-v0el4e) {
  display: flex;
  flex-direction: row;
  justify-content: flex-end;
  padding-top: var(--jse-padding, 10px);
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-actions:where(.svelte-v0el4e) button.jse-primary:where(.svelte-v0el4e) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-actions:where(.svelte-v0el4e) button.jse-primary:where(.svelte-v0el4e):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-actions:where(.svelte-v0el4e) button.jse-primary:where(.svelte-v0el4e):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-label:where(.svelte-v0el4e) {
  font-weight: bold;
  display: block;
  box-sizing: border-box;
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-label:where(.svelte-v0el4e) .jse-label-inner:where(.svelte-v0el4e) {
  margin-top: calc(2 * var(--jse-padding, 10px));
  margin-bottom: calc(0.5 * var(--jse-padding, 10px));
  box-sizing: border-box;
}
.jse-modal-wrapper.svelte-v0el4e .jse-modal-contents:where(.svelte-v0el4e) .jse-modal-inline-editor:where(.svelte-v0el4e) {
  flex: 1;
  min-height: 150px;
  min-width: 0;
  max-width: 100%;
  display: flex;
  --jse-theme-color: var(--jse-modal-editor-theme-color, #707070);
  --jse-theme-color-highlight: var(--jse-modal-editor-theme-color-highlight, #646464);
}
.jse-modal-wrapper.svelte-v0el4e .jse-actions:where(.svelte-v0el4e) {
  gap: var(--jse-padding, 10px);
  align-items: center;
}
.jse-modal-wrapper.svelte-v0el4e .jse-actions:where(.svelte-v0el4e) .jse-error:where(.svelte-v0el4e) {
  flex: 1;
  color: var(--jse-error-color, #ee5341);
}
.jse-modal-wrapper.svelte-v0el4e .jse-actions:where(.svelte-v0el4e) button.jse-secondary:where(.svelte-v0el4e) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-secondary-background, #d3d3d3);
  color: var(--jse-button-secondary-color, var(--jse-text-color, #4d4d4d));
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-modal-wrapper.svelte-v0el4e .jse-actions:where(.svelte-v0el4e) button.jse-secondary:where(.svelte-v0el4e):hover {
  background: var(--jse-button-secondary-background-highlight, #e1e1e1);
}
.jse-modal-wrapper.svelte-v0el4e .jse-actions:where(.svelte-v0el4e) button.jse-secondary:where(.svelte-v0el4e):disabled {
  background: var(--jse-button-secondary-background-disabled, #9d9d9d);
}
.jse-modal-wrapper.svelte-v0el4e input:where(.svelte-v0el4e) {
  border: var(--jse-input-border, 1px solid #d8dbdf);
  outline: none;
  box-sizing: border-box;
  padding: calc(0.5 * var(--jse-padding, 10px));
  font-family: var(--jse-font-family-mono, consolas, menlo, monaco, "Ubuntu Mono", "source-code-pro", monospace);
  font-size: var(--jse-font-size-mono, 14px);
  color: inherit;
  background: var(--jse-input-background, var(--jse-background-color, #fff));
}
.jse-modal-wrapper.svelte-v0el4e input:where(.svelte-v0el4e):focus {
  border: var(--jse-input-border-focus, 1px solid var(--jse-input-border-focus, var(--jse-theme-color, #3883fa)));
}
.jse-modal-wrapper.svelte-v0el4e input:where(.svelte-v0el4e):read-only {
  background: var(--jse-input-background-readonly, transparent);
}`);var pTA=wA('<div class="jse-error svelte-v0el4e"> </div>'),wTA=wA('<button type="button" class="jse-secondary svelte-v0el4e"><!> Back</button>'),DTA=wA('<button type="button" class="jse-primary svelte-v0el4e">Apply</button>'),yTA=wA('<button type="button" class="jse-primary svelte-v0el4e">Close</button>'),vTA=wA('<!> <div class="jse-modal-contents svelte-v0el4e"><div class="jse-label svelte-v0el4e"><div class="jse-label-inner svelte-v0el4e">Path</div></div> <input class="jse-path svelte-v0el4e" type="text" readonly="" title="Selected path"/> <div class="jse-label svelte-v0el4e"><div class="jse-label-inner svelte-v0el4e">Contents</div></div> <div class="jse-modal-inline-editor svelte-v0el4e"><!></div> <div class="jse-actions svelte-v0el4e"><!> <!> <!></div></div>',1),bTA=wA('<div class="jse-modal-wrapper svelte-v0el4e"><!></div>'),MTA={};Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-modal-contents.svelte-1v9c92j {
  flex: 1;
  display: flex;
  flex-direction: column;
  padding: 20px;
  overflow: auto;
  min-width: 0;
  min-height: 0;
}
.jse-modal-contents.svelte-1v9c92j .jse-actions:where(.svelte-1v9c92j) {
  display: flex;
  flex-direction: row;
  justify-content: flex-end;
  padding-top: var(--jse-padding, 10px);
}
.jse-modal-contents.svelte-1v9c92j .jse-actions:where(.svelte-1v9c92j) button.jse-primary:where(.svelte-1v9c92j) {
  border: none;
  background: transparent;
  color: inherit;
  cursor: pointer;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  padding: 5px;
  margin: 0;
  background: var(--jse-button-primary-background, var(--jse-theme-color, #3883fa));
  color: var(--jse-button-primary-color, #fff);
  padding: var(--jse-padding, 10px) calc(2 * var(--jse-padding, 10px));
  border-radius: 3px;
}
.jse-modal-contents.svelte-1v9c92j .jse-actions:where(.svelte-1v9c92j) button.jse-primary:where(.svelte-1v9c92j):hover {
  background: var(--jse-button-primary-background-highlight, var(--jse-theme-color-highlight, #5f9dff));
}
.jse-modal-contents.svelte-1v9c92j .jse-actions:where(.svelte-1v9c92j) button.jse-primary:where(.svelte-1v9c92j):disabled {
  background: var(--jse-button-primary-background-disabled, #9d9d9d);
}
.jse-modal-contents.svelte-1v9c92j table:where(.svelte-1v9c92j) {
  width: 100%;
  border-collapse: collapse;
  border-spacing: 0;
}
.jse-modal-contents.svelte-1v9c92j table:where(.svelte-1v9c92j) th:where(.svelte-1v9c92j),
.jse-modal-contents.svelte-1v9c92j table:where(.svelte-1v9c92j) td:where(.svelte-1v9c92j) {
  text-align: left;
  vertical-align: middle;
  font-weight: normal;
  padding-bottom: var(--jse-padding, 10px);
}
.jse-modal-contents.svelte-1v9c92j input.jse-path:where(.svelte-1v9c92j) {
  width: 100%;
  box-sizing: border-box;
  padding: 5px 10px;
  border: var(--jse-input-border, 1px solid #d8dbdf);
  border-radius: var(--jse-input-radius, 3px);
  font-family: inherit;
  font-size: inherit;
  background: inherit;
  background: var(--jse-input-background-readonly, transparent);
  color: inherit;
  outline: none;
}
.jse-modal-contents.svelte-1v9c92j .svelte-select input {
  box-sizing: border-box;
}
.jse-modal-contents.svelte-1v9c92j .jse-space:where(.svelte-1v9c92j) {
  height: 200px;
}
.jse-modal-contents.svelte-1v9c92j .jse-space:where(.svelte-1v9c92j) .jse-error:where(.svelte-1v9c92j) {
  color: var(--jse-error-color, #ee5341);
}`);var AQ=Gy(()=>MTA),kTA=wA('<tr><th class="svelte-1v9c92j">Property</th><td class="svelte-1v9c92j"><!></td></tr>'),STA=wA('<div class="jse-error svelte-1v9c92j"> </div>'),RTA=wA('<!> <div class="jse-modal-contents svelte-1v9c92j"><table class="svelte-1v9c92j"><colgroup><col width="25%"/><col width="75%"/></colgroup><tbody><tr><th class="svelte-1v9c92j">Path</th><td class="svelte-1v9c92j"><input class="jse-path svelte-1v9c92j" type="text" readonly="" title="Selected path"/></td></tr><!><tr><th class="svelte-1v9c92j">Direction</th><td class="svelte-1v9c92j"><!></td></tr></tbody></table> <div class="jse-space svelte-1v9c92j"><!></div> <div class="jse-actions svelte-1v9c92j"><button type="button" class="jse-primary svelte-1v9c92j">Sort</button></div></div>',1);Jt(`/* over all fonts, sizes, and colors */
/* "consolas" for Windows, "menlo" for Mac with fallback to "monaco", 'Ubuntu Mono' for Ubuntu */
/* (at Mac this font looks too large at 14px, but 13px is too small for the font on Windows) */
/* main, menu, modal */
/* jsoneditor modal */
/* tooltip in text mode */
/* panels: navigation bar, gutter, search box */
/* navigation-bar */
/* context menu */
/* contents: json key and values */
/* contents: selected or hovered */
/* contents: section of collapsed items in an array */
/* contents: highlighting of search matches */
/* contents: inline tags inside the JSON document */
/* contents: table */
/* controls in modals: inputs, buttons, and \`a\` */
/* messages */
/* svelte-select */
/* color picker */
.jse-main.svelte-57bmz4 {
  width: 100%;
  height: 100%;
  min-width: 0;
  min-height: 150px;
  font-family: var(--jse-font-family, -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, Oxygen-Sans, Ubuntu, Cantarell, "Helvetica Neue", sans-serif);
  font-size: var(--jse-font-size, 16px);
  line-height: normal;
  position: relative;
  display: flex;
  flex-direction: row;
}
.jse-main.svelte-57bmz4:not(.jse-focus) {
  --jse-selection-background-color: var(--jse-selection-background-inactive-color, #e8e8e8);
  --jse-context-menu-pointer-background: var(--jse-context-menu-pointer-hover-background, #b2b2b2);
}`);var xTA=wA('<div role="none"><!></div> <!> <!> <!>',1);function LTA(t,e){mt(e,!1);var A=$(void 0,!0),i=Sr("jsoneditor:JSONEditor"),n={text:""},o=void 0,r=!1,s=er.tree,a=!0,c=!0,l=!0,I=!0,C=!1,d=!1,B=!0,E=JSON,h=void 0,u=JSON,D={parse:SUA,stringify:Ka},L=[VGA],R=L[0].id,w=Oc,_=void 0,K=void 0,z=kUA,U=Oc,H=Oc,q=Oc,j=Oc,gA=ne=>{console.error(ne),alert(ne.toString())},QA=Oc,BA=Oc,lA=b(e,"content",13,n),vA=b(e,"selection",13,o),tA=b(e,"readOnly",13,r),cA=b(e,"indentation",13,2),pA=b(e,"tabSize",13,4),VA=b(e,"truncateTextSize",13,1e3),oe=b(e,"mode",13,s),KA=b(e,"mainMenuBar",13,a),CA=b(e,"navigationBar",13,c),TA=b(e,"statusBar",13,l),Ze=b(e,"askToFormat",13,I),He=b(e,"escapeControlCharacters",13,C),uA=b(e,"escapeUnicodeCharacters",13,d),eA=b(e,"flattenColumns",13,B),UA=b(e,"parser",13,E),aA=b(e,"validator",13,h),le=b(e,"validationParser",13,u),SA=b(e,"pathParser",13,D),Ue=b(e,"queryLanguages",13,L),mA=b(e,"queryLanguageId",13,R),sA=b(e,"onChangeQueryLanguage",13,w),xt=b(e,"onChange",13,_),tt=b(e,"onSelect",13,K),de=b(e,"onRenderValue",13,z),Dt=b(e,"onClassName",13,U),_e=b(e,"onRenderMenu",13,H),Le=b(e,"onRenderContextMenu",13,q),bt=b(e,"onChangeMode",13,j),Re=b(e,"onError",13,gA),$t=b(e,"onFocus",13,QA),x=b(e,"onBlur",13,BA),Y=$(nQ(),!0),P=$(!1,!0),X=$(void 0,!0),bA=$(void 0,!0),Be=$(void 0,!0),Ee=$(void 0,!0),kA=$(UA(),!0);function DA(){return lA()}function gt(ne){i("set");var wi=HN(ne);if(wi)throw new Error(wi);y(Y,nQ()),lA(ne),io()}function Ve(ne){i("update");var wi=HN(ne);if(wi)throw new Error(wi);lA(ne),io()}function ZA(ne){var wi=g(X).patch(ne);return io(),wi}function rt(ne){vA(ne),io()}function Ei(ne,wi){g(X).expand(ne,wi),io()}function tn(ne){var wi=arguments.length>1&&arguments[1]!==void 0&&arguments[1];g(X).collapse(ne,wi),io()}function qi(){var ne=arguments.length>0&&arguments[0]!==void 0?arguments[0]:{};g(X).transform(ne),io()}function xe(){return g(X).validate()}function nn(){var ne=g(X).acceptAutoRepair();return io(),ne}function mi(ne){return Ot.apply(this,arguments)}function Ot(){return(Ot=Yt(function*(ne){yield g(X).scrollTo(ne)})).apply(this,arguments)}function Lt(ne){return g(X).findElement(ne)}function ii(){g(X).focus(),io()}function _i(){return Tt.apply(this,arguments)}function Tt(){return(Tt=Yt(function*(){yield g(X).refresh()})).apply(this,arguments)}function M(ne){var wi,MA,me,nt,Wt,Xe,oi,Di,Ut,cn,ft,Qi,ot,Mt,on,hn,Ai,Ki,dt,EA,HA,ve,Qt,yi,ri,pn,Fn,Jn,ln,Pt,$i,Rr=Object.keys(ne);for(var Ft of Rr)switch(Ft){case"content":lA((wi=ne[Ft])!==null&&wi!==void 0?wi:n);break;case"selection":vA((MA=ne[Ft])!==null&&MA!==void 0?MA:o);break;case"readOnly":tA((me=ne[Ft])!==null&&me!==void 0?me:r);break;case"indentation":cA((nt=ne[Ft])!==null&&nt!==void 0?nt:2);break;case"tabSize":pA((Wt=ne[Ft])!==null&&Wt!==void 0?Wt:4);break;case"truncateTextSize":VA((Xe=ne[Ft])!==null&&Xe!==void 0?Xe:1e3);break;case"mode":oe((oi=ne[Ft])!==null&&oi!==void 0?oi:s);break;case"mainMenuBar":KA((Di=ne[Ft])!==null&&Di!==void 0?Di:a);break;case"navigationBar":CA((Ut=ne[Ft])!==null&&Ut!==void 0?Ut:c);break;case"statusBar":TA((cn=ne[Ft])!==null&&cn!==void 0?cn:l);break;case"askToFormat":Ze((ft=ne[Ft])!==null&&ft!==void 0?ft:I);break;case"escapeControlCharacters":He((Qi=ne[Ft])!==null&&Qi!==void 0?Qi:C);break;case"escapeUnicodeCharacters":uA((ot=ne[Ft])!==null&&ot!==void 0?ot:d);break;case"flattenColumns":eA((Mt=ne[Ft])!==null&&Mt!==void 0?Mt:B);break;case"parser":UA((on=ne[Ft])!==null&&on!==void 0?on:E);break;case"validator":aA((hn=ne[Ft])!==null&&hn!==void 0?hn:h);break;case"validationParser":le((Ai=ne[Ft])!==null&&Ai!==void 0?Ai:u);break;case"pathParser":SA((Ki=ne[Ft])!==null&&Ki!==void 0?Ki:D);break;case"queryLanguages":Ue((dt=ne[Ft])!==null&&dt!==void 0?dt:L);break;case"queryLanguageId":mA((EA=ne[Ft])!==null&&EA!==void 0?EA:R);break;case"onChangeQueryLanguage":sA((HA=ne[Ft])!==null&&HA!==void 0?HA:w);break;case"onChange":xt((ve=ne[Ft])!==null&&ve!==void 0?ve:_);break;case"onRenderValue":de((Qt=ne[Ft])!==null&&Qt!==void 0?Qt:z);break;case"onClassName":Dt((yi=ne[Ft])!==null&&yi!==void 0?yi:U);break;case"onRenderMenu":_e((ri=ne[Ft])!==null&&ri!==void 0?ri:H);break;case"onRenderContextMenu":Le((pn=ne[Ft])!==null&&pn!==void 0?pn:q);break;case"onChangeMode":bt((Fn=ne[Ft])!==null&&Fn!==void 0?Fn:j);break;case"onSelect":tt((Jn=ne[Ft])!==null&&Jn!==void 0?Jn:K);break;case"onError":Re((ln=ne[Ft])!==null&&ln!==void 0?ln:gA);break;case"onFocus":$t((Pt=ne[Ft])!==null&&Pt!==void 0?Pt:QA);break;case"onBlur":x(($i=ne[Ft])!==null&&$i!==void 0?$i:BA);break;default:Xn(Ft)}function Xn(se){i('Unknown property "'.concat(se,'"'))}Ue().some(se=>se.id===mA())||mA(Ue()[0].id),io()}function We(){return ni.apply(this,arguments)}function ni(){return(ni=Yt(function*(){throw new Error("class method destroy() is deprecated. It is replaced with a method destroy() in the vanilla library.")})).apply(this,arguments)}function pi(ne,wi,MA){lA(ne),xt()&&xt()(ne,wi,MA)}function dn(ne){vA(ne),tt()&&tt()(i4(ne))}function mn(){y(P,!0),$t()&&$t()()}function Uo(){y(P,!1),x()&&x()()}function kn(ne){return Wn.apply(this,arguments)}function Wn(){return(Wn=Yt(function*(ne){oe()!==ne&&(oe(ne),io(),ii(),bt()(ne))})).apply(this,arguments)}function Vo(ne){i("handleChangeQueryLanguage",ne),mA(ne),sA()(ne)}function vo(ne){var{id:wi,json:MA,rootPath:me,onTransform:nt,onClose:Wt}=ne;tA()||y(Ee,{id:wi,json:MA,rootPath:me,indentation:cA(),truncateTextSize:VA(),escapeControlCharacters:He(),escapeUnicodeCharacters:uA(),parser:UA(),parseMemoizeOne:g(A),validationParser:le(),pathParser:SA(),queryLanguages:Ue(),queryLanguageId:mA(),onChangeQueryLanguage:Vo,onRenderValue:de(),onRenderMenu:Xe=>_e()(Xe,{mode:oe(),modal:!0,readOnly:tA()}),onRenderContextMenu:Xe=>Le()(Xe,{mode:oe(),modal:!0,readOnly:tA(),selection:vA()}),onClassName:Dt(),onTransform:nt,onClose:Wt})}function bo(ne){tA()||y(Be,ne)}function Yn(ne){var{content:wi,path:MA,onPatch:me,onClose:nt}=ne;i("onJSONEditorModal",{content:wi,path:MA}),y(bA,{content:wi,path:MA,onPatch:me,readOnly:tA(),indentation:cA(),tabSize:pA(),truncateTextSize:VA(),mainMenuBar:KA(),navigationBar:CA(),statusBar:TA(),askToFormat:Ze(),escapeControlCharacters:He(),escapeUnicodeCharacters:uA(),flattenColumns:eA(),parser:UA(),validator:void 0,validationParser:le(),pathParser:SA(),onRenderValue:de(),onClassName:Dt(),onRenderMenu:_e(),onRenderContextMenu:Le(),onSortModal:bo,onTransformModal:vo,onClose:nt})}function Mo(ne){ne.stopPropagation()}return fA(()=>(k(UA()),g(kA),k(lA()),nQ),()=>{if(!UsA(UA(),g(kA))){if(i("parser changed, recreate editor"),z3(lA())){var ne=g(kA).stringify(lA().json);lA({json:ne!==void 0?UA().parse(ne):void 0})}y(kA,UA()),y(Y,nQ())}}),fA(()=>k(lA()),()=>{var ne=HN(lA());ne&&console.error("Error: "+ne)}),fA(()=>k(vA()),()=>{vA()===null&&console.warn("selection is invalid: it is null but should be undefined")}),fA(()=>k(UA()),()=>{y(A,aE(UA().parse))}),fA(()=>k(oe()),()=>{i("mode changed to",oe())}),Qn(),Zt(!0),f_(t,{children:(ne,wi)=>{var MA,me=xTA(),nt=vt(me);vsA(W(nt),()=>g(Y),ft=>{Co(HrA(ft,{get externalMode(){return oe()},get content(){return lA()},get selection(){return vA()},get readOnly(){return tA()},get indentation(){return cA()},get tabSize(){return pA()},get truncateTextSize(){return VA()},get statusBar(){return TA()},get askToFormat(){return Ze()},get mainMenuBar(){return KA()},get navigationBar(){return CA()},get escapeControlCharacters(){return He()},get escapeUnicodeCharacters(){return uA()},get flattenColumns(){return eA()},get parser(){return UA()},get parseMemoizeOne(){return g(A)},get validator(){return aA()},get validationParser(){return le()},get pathParser(){return SA()},insideModal:!1,get onError(){return Re()},onChange:pi,onChangeMode:kn,onSelect:dn,get onRenderValue(){return de()},get onClassName(){return Dt()},onFocus:mn,onBlur:Uo,get onRenderMenu(){return _e()},get onRenderContextMenu(){return Le()},onSortModal:bo,onTransformModal:vo,onJSONEditorModal:Yn,$$legacy:!0}),Qi=>y(X,Qi),()=>g(X))});var Wt=IA(nt,2),Xe=ft=>{(function(Qi,ot){var Mt,on;mt(ot,!1);var hn=$(void 0,!0),Ai=$(void 0,!0),Ki=$(void 0,!0),dt=$(void 0,!0),EA=Sr("jsoneditor:SortModal"),HA=b(ot,"id",9),ve=b(ot,"json",9),Qt=b(ot,"rootPath",9),yi=b(ot,"onSort",9),ri=b(ot,"onClose",9),pn={value:1,label:"ascending"},Fn=[pn,{value:-1,label:"descending"}],Jn="".concat(HA(),":").concat(Ct(Qt())),ln=$((Mt=AQ()[Jn])===null||Mt===void 0?void 0:Mt.selectedProperty,!0),Pt=$(((on=AQ()[Jn])===null||on===void 0?void 0:on.selectedDirection)||pn,!0),$i=$(void 0,!0);function Rr(){try{var Xn,se,vi;y($i,void 0);var Yi=((Xn=g(ln))===null||Xn===void 0?void 0:Xn.value)||((se=g(dt))===null||se===void 0||(se=se[0])===null||se===void 0?void 0:se.value)||[],rn=(vi=g(Pt))===null||vi===void 0?void 0:vi.value,Hr=baA(ve(),Qt(),Yi,rn);yi()!==void 0&&Qt()!==void 0&&yi()({operations:Hr,rootPath:Qt(),itemPath:Yi,direction:rn}),ri()()}catch(Ri){y($i,String(Ri))}}function Ft(Xn){Xn.focus()}fA(()=>(k(ve()),k(Qt())),()=>{y(hn,Ke(ve(),Qt()))}),fA(()=>g(hn),()=>{y(Ai,Array.isArray(g(hn)))}),fA(()=>(g(Ai),g(hn)),()=>{y(Ki,g(Ai)?h_(g(hn)):void 0)}),fA(()=>(g(Ki),U1),()=>{y(dt,g(Ki)?g(Ki).map(U1):void 0)}),fA(()=>(AQ(),g(ln),g(Pt)),()=>{AQ(AQ()[Jn]={selectedProperty:g(ln),selectedDirection:g(Pt)}),EA("store state in memory",Jn,AQ()[Jn])}),Qn(),Zt(!0),X3(Qi,{get onClose(){return ri()},className:"jse-sort-modal",children:(Xn,se)=>{var vi=RTA(),Yi=vt(vi),rn=qA(()=>g(Ai)?"Sort array items":"Sort object keys");My(Yi,{get title(){return g(rn)},get onClose(){return ri()}});var Hr=W(IA(Yi,2)),Ri=IA(W(Hr)),fs=W(Ri),Bo=IA(W(fs)),Q=W(Bo),m=IA(fs),v=te=>{var re=kTA(),Pe=IA(W(re));JC(W(Pe),{showChevron:!0,get items(){return g(dt)},get value(){return g(ln)},set value(ze){y(ln,ze)},$$legacy:!0}),iA(te,re)};LA(m,te=>{g(Ai),g(dt),nA(()=>{var re;return g(Ai)&&g(dt)&&((re=g(dt))===null||re===void 0?void 0:re.length)>1})&&te(v)});var p=IA(m),N=IA(W(p));JC(W(N),{showChevron:!0,clearable:!1,get items(){return Fn},get value(){return g(Pt)},set value(te){y(Pt,te)},$$legacy:!0});var J=IA(Hr,2),V=W(J),Z=te=>{var re=STA(),Pe=W(re);De(()=>wt(Pe,g($i))),iA(te,re)};LA(V,te=>{g($i)&&te(Z)});var FA=W(IA(J,2));es(()=>ce("click",FA,Rr)),Us(FA,te=>Ft?.(te)),De(te=>{VC(Q,te),FA.disabled=(g(Ai),g(dt),g(ln),nA(()=>{var re;return!!(g(Ai)&&g(dt)&&((re=g(dt))===null||re===void 0?void 0:re.length)>1)&&!g(ln)}))},[()=>(k(Qt()),k(Oi),k(Ka),nA(()=>Qt()&&!Oi(Qt())?Ka(Qt()):"(document root)"))],qA),iA(Xn,vi)},$$slots:{default:!0}}),pt()})(ft,z1(()=>g(Be),{onClose:()=>{var Qi;(Qi=g(Be))===null||Qi===void 0||Qi.onClose(),y(Be,void 0)}}))};LA(Wt,ft=>{g(Be)&&ft(Xe)});var oi=IA(Wt,2),Di=ft=>{NJA(ft,z1(()=>g(Ee),{onClose:()=>{var Qi;(Qi=g(Ee))===null||Qi===void 0||Qi.onClose(),y(Ee,void 0)}}))};LA(oi,ft=>{g(Ee)&&ft(Di)});var Ut=IA(oi,2),cn=ft=>{(function(Qi,ot){mt(ot,!1);var Mt=$(void 0,!0),on=$(void 0,!0),hn=$(void 0,!0),Ai=$(void 0,!0),Ki=Sr("jsoneditor:JSONEditorModal"),dt=b(ot,"content",9),EA=b(ot,"path",9),HA=b(ot,"onPatch",9),ve=b(ot,"readOnly",9),Qt=b(ot,"indentation",9),yi=b(ot,"tabSize",9),ri=b(ot,"truncateTextSize",9),pn=b(ot,"mainMenuBar",9),Fn=b(ot,"navigationBar",9),Jn=b(ot,"statusBar",9),ln=b(ot,"askToFormat",9),Pt=b(ot,"escapeControlCharacters",9),$i=b(ot,"escapeUnicodeCharacters",9),Rr=b(ot,"flattenColumns",9),Ft=b(ot,"parser",9),Xn=b(ot,"validator",9),se=b(ot,"validationParser",9),vi=b(ot,"pathParser",9),Yi=b(ot,"onRenderValue",9),rn=b(ot,"onClassName",9),Hr=b(ot,"onRenderMenu",9),Ri=b(ot,"onRenderContextMenu",9),fs=b(ot,"onSortModal",9),Bo=b(ot,"onTransformModal",9),Q=b(ot,"onClose",9),m=$(void 0,!0),v=$(void 0,!0),p={mode:V(dt()),content:dt(),selection:void 0,relativePath:EA()},N=$([p],!0),J=$(void 0,!0);function V(oA){return z3(oA)&&wo(oA.json)?er.table:er.tree}function Z(){var oA,YA=(oA=hi(g(N)))===null||oA===void 0?void 0:oA.selection;q3(YA)&&g(m).scrollTo(et(YA))}function FA(){if(Ki("handleApply"),!ve())try{y(J,void 0);var oA=g(Mt).relativePath,YA=g(Mt).content,pe=[{op:"replace",path:Ct(oA),value:ArA(YA,Ft()).json}];if(g(N).length>1){var he=ArA(g(N)[g(N).length-2].content,Ft()).json,ge={json:Da(he,pe)},Bt=fe(fe({},g(N)[g(N).length-2]||p),{},{content:ge});y(N,[...g(N).slice(0,g(N).length-2),Bt]),io(),Z()}else HA()(pe),Q()()}catch($e){y(J,String($e))}}function te(){if(Ki("handleClose"),g(v))y(v,!1);else if(g(N).length>1){var oA;y(N,Fi(g(N))),io(),(oA=g(m))===null||oA===void 0||oA.focus(),Z(),y(J,void 0)}else Q()()}function re(oA){Ki("handleChange",oA),ye(YA=>fe(fe({},YA),{},{content:oA}))}function Pe(oA){Ki("handleChangeSelection",oA),ye(YA=>fe(fe({},YA),{},{selection:oA}))}function ze(oA){Ki("handleChangeMode",oA),ye(YA=>fe(fe({},YA),{},{mode:oA}))}function ye(oA){var YA=oA(hi(g(N)));y(N,[...Fi(g(N)),YA])}function Ge(oA){y(J,oA.toString()),console.error(oA)}function Vi(oA){var YA,{content:pe,path:he}=oA;Ki("handleJSONEditorModal",{content:pe,path:he});var ge={mode:V(pe),content:pe,selection:void 0,relativePath:he};y(N,[...g(N),ge]),io(),(YA=g(m))===null||YA===void 0||YA.focus()}function T(oA){oA.focus()}hs(()=>{var oA;(oA=g(m))===null||oA===void 0||oA.focus()}),fA(()=>g(N),()=>{y(Mt,hi(g(N))||p)}),fA(()=>g(N),()=>{y(on,g(N).flatMap(oA=>oA.relativePath))}),fA(()=>(g(on),Ka),()=>{y(hn,Oi(g(on))?"(document root)":Ka(g(on)))}),fA(()=>k(Ft()),()=>{y(Ai,aE(Ft().parse))}),Qn(),Zt(!0),X3(Qi,{onClose:te,className:"jse-jsoneditor-modal",get fullscreen(){return g(v)},children:(oA,YA)=>{var pe=bTA();f_(W(pe),{children:(he,ge)=>{var Bt=vTA(),$e=vt(Bt),bi=qA(()=>(g(N),nA(()=>g(N).length>1?" (".concat(g(N).length,")"):"")));My($e,{get title(){var _t;return"Edit nested content ".concat((_t=g(bi))!==null&&_t!==void 0?_t:"")},fullScreenButton:!0,onClose:te,get fullscreen(){return g(v)},set fullscreen(_t){y(v,_t)},$$legacy:!0});var un=IA($e,2),Ji=IA(W(un),2),Tn=IA(Ji,4);Co(HrA(W(Tn),{get externalMode(){return g(Mt),nA(()=>g(Mt).mode)},get content(){return g(Mt),nA(()=>g(Mt).content)},get selection(){return g(Mt),nA(()=>g(Mt).selection)},get readOnly(){return ve()},get indentation(){return Qt()},get tabSize(){return yi()},get truncateTextSize(){return ri()},get statusBar(){return Jn()},get askToFormat(){return ln()},get mainMenuBar(){return pn()},get navigationBar(){return Fn()},get escapeControlCharacters(){return Pt()},get escapeUnicodeCharacters(){return $i()},get flattenColumns(){return Rr()},get parser(){return Ft()},get parseMemoizeOne(){return g(Ai)},get validator(){return Xn()},get validationParser(){return se()},get pathParser(){return vi()},insideModal:!0,onError:Ge,onChange:re,onChangeMode:ze,onSelect:Pe,get onRenderValue(){return Yi()},get onClassName(){return rn()},get onFocus(){return Oc},get onBlur(){return Oc},get onRenderMenu(){return Hr()},get onRenderContextMenu(){return Ri()},get onSortModal(){return fs()},get onTransformModal(){return Bo()},onJSONEditorModal:Vi,$$legacy:!0}),_t=>y(m,_t),()=>g(m));var Ht=W(IA(Tn,2)),ko=_t=>{var st=pTA(),si=W(st);De(()=>wt(si,g(J))),iA(_t,st)};LA(Ht,_t=>{g(J)&&_t(ko)});var Sn=IA(Ht,2),no=_t=>{var st=wTA();ji(W(st),{get data(){return GW}}),ce("click",st,te),iA(_t,st)};LA(Sn,_t=>{g(N),nA(()=>g(N).length>1)&&_t(no)});var gn=IA(Sn,2),Nt=_t=>{var st=DTA();es(()=>ce("click",st,FA)),Us(st,si=>T?.(si)),iA(_t,st)},Zi=_t=>{var st=yTA();ce("click",st,te),iA(_t,st)};LA(gn,_t=>{ve()?_t(Zi,!1):_t(Nt)}),De(()=>VC(Ji,g(hn))),iA(he,Bt)},$$slots:{default:!0}}),iA(oA,pe)},$$slots:{default:!0}}),pt()})(ft,z1(()=>g(bA),{onClose:()=>{var Qi;(Qi=g(bA))===null||Qi===void 0||Qi.onClose(),y(bA,void 0)}}))};LA(Ut,ft=>{g(bA)&&ft(cn)}),De(ft=>MA=Vt(nt,1,"jse-main svelte-57bmz4",null,MA,ft),[()=>({"jse-focus":g(P)})],qA),ce("keydown",nt,Mo),iA(ne,me)},$$slots:{default:!0}}),zt(e,"get",DA),zt(e,"set",gt),zt(e,"update",Ve),zt(e,"patch",ZA),zt(e,"select",rt),zt(e,"expand",Ei),zt(e,"collapse",tn),zt(e,"transform",qi),zt(e,"validate",xe),zt(e,"acceptAutoRepair",nn),zt(e,"scrollTo",mi),zt(e,"findElement",Lt),zt(e,"focus",ii),zt(e,"refresh",_i),zt(e,"updateProps",M),zt(e,"destroy",We),pt({get:DA,set:gt,update:Ve,patch:ZA,select:rt,expand:Ei,collapse:tn,transform:qi,validate:xe,acceptAutoRepair:nn,scrollTo:mi,findElement:Lt,focus:ii,refresh:_i,updateProps:M,destroy:We})}function YaA(t){var{target:e,props:A}=t,i=CGA(LTA,{target:e,props:A});return i.destroy=Yt(function*(){return function(n,o){var r=B_.get(n);return r?(B_.delete(n),r(o)):Promise.resolve()}(i)}),io(),i}var ed=class t{constructor(e){this.el=e}jsonString;editor=null;ngAfterViewInit(){let e={text:this.jsonString};this.editor=YaA({target:document.getElementById("json-editor"),props:{content:e,mode:er.text,mainMenuBar:!1}})}getJsonString(){return this.editor?.get().text}static \u0275fac=function(A){return new(A||t)(PA(ee))};static \u0275cmp=zA({type:t,selectors:[["app-json-editor"]],inputs:{jsonString:"jsonString"},standalone:!1,decls:1,vars:0,consts:[["id","json-editor",1,"json-editor-container","jse-theme-dark"]],template:function(A,i){A&1&&JA(0,"div",0)},styles:[".jse-theme-dark[_ngcontent-%COMP%]{--jse-theme: dark;--jse-theme-color: #2f6dd0;--jse-theme-color-highlight: #467cd2;--jse-background-color: #1e1e1e;--jse-text-color: #d4d4d4;--jse-text-color-inverse: #4d4d4d;--jse-main-border: 1px solid #4f4f4f;--jse-menu-color: #fff;--jse-modal-background: #2f2f2f;--jse-modal-overlay-background: rgba(0, 0, 0, .5);--jse-modal-code-background: #2f2f2f;--jse-tooltip-color: var(--jse-text-color);--jse-tooltip-background: #4b4b4b;--jse-tooltip-border: 1px solid #737373;--jse-tooltip-action-button-color: inherit;--jse-tooltip-action-button-background: #737373;--jse-panel-background: #333333;--jse-panel-background-border: 1px solid #464646;--jse-panel-color: var(--jse-text-color);--jse-panel-color-readonly: #737373;--jse-panel-border: 1px solid #3c3c3c;--jse-panel-button-color-highlight: #e5e5e5;--jse-panel-button-background-highlight: #464646;--jse-navigation-bar-background: #656565;--jse-navigation-bar-background-highlight: #7e7e7e;--jse-navigation-bar-dropdown-color: var(--jse-text-color);--jse-context-menu-background: #4b4b4b;--jse-context-menu-background-highlight: #595959;--jse-context-menu-separator-color: #595959;--jse-context-menu-color: var(--jse-text-color);--jse-context-menu-pointer-background: #737373;--jse-context-menu-pointer-background-highlight: #818181;--jse-context-menu-pointer-color: var(--jse-context-menu-color);--jse-key-color: #9cdcfe;--jse-value-color: var(--jse-text-color);--jse-value-color-number: #b5cea8;--jse-value-color-boolean: #569cd6;--jse-value-color-null: #569cd6;--jse-value-color-string: #ce9178;--jse-value-color-url: #ce9178;--jse-delimiter-color: #949494;--jse-edit-outline: 2px solid var(--jse-text-color);--jse-selection-background-color: #464646;--jse-selection-background-inactive-color: #333333;--jse-hover-background-color: #343434;--jse-active-line-background-color: rgba(255, 255, 255, .06);--jse-search-match-background-color: #343434;--jse-collapsed-items-background-color: #333333;--jse-collapsed-items-selected-background-color: #565656;--jse-collapsed-items-link-color: #b2b2b2;--jse-collapsed-items-link-color-highlight: #ec8477;--jse-search-match-color: #724c27;--jse-search-match-outline: 1px solid #966535;--jse-search-match-active-color: #9f6c39;--jse-search-match-active-outline: 1px solid #bb7f43;--jse-tag-background: #444444;--jse-tag-color: #bdbdbd;--jse-table-header-background: #333333;--jse-table-header-background-highlight: #424242;--jse-table-row-odd-background: rgba(255, 255, 255, .1);--jse-input-background: #3d3d3d;--jse-input-border: var(--jse-main-border);--jse-button-background: #808080;--jse-button-background-highlight: #7a7a7a;--jse-button-color: #e0e0e0;--jse-button-secondary-background: #494949;--jse-button-secondary-background-highlight: #5d5d5d;--jse-button-secondary-background-disabled: #9d9d9d;--jse-button-secondary-color: var(--jse-text-color);--jse-a-color: #55abff;--jse-a-color-highlight: #4387c9;--jse-svelte-select-background: #3d3d3d;--jse-svelte-select-border: 1px solid #4f4f4f;--list-background: #3d3d3d;--item-hover-bg: #505050;--multi-item-bg: #5b5b5b;--input-color: #d4d4d4;--multi-clear-bg: #8a8a8a;--multi-item-clear-icon-color: #d4d4d4;--multi-item-outline: 1px solid #696969;--list-shadow: 0 2px 8px 0 rgba(0, 0, 0, .4);--jse-color-picker-background: #656565;--jse-color-picker-border-box-shadow: #8c8c8c 0 0 0 1px}.json-editor-container[_ngcontent-%COMP%]{height:300px;max-height:300px}"]})};var vQ=class t{constructor(e,A){this.dialogRef=e;this.data=A;this.jsonString=JSON.stringify(A.jsonContent,null,2),this.functionName=A.functionName||""}jsonEditorComponent;jsonString="";functionName="";ngOnInit(){}onSave(){try{this.jsonString=this.jsonEditorComponent.getJsonString();let e=JSON.parse(this.jsonString);this.dialogRef.close(e)}catch(e){alert("Invalid JSON: "+e)}}onCancel(){this.dialogRef.close(null)}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(as))};static \u0275cmp=zA({type:t,selectors:[["app-edit-json-dialog"]],viewQuery:function(A,i){if(A&1&&Te(ed,5),A&2){let n;XA(n=$A())&&(i.jsonEditorComponent=n.first)}},standalone:!1,decls:11,vars:3,consts:[[1,"dialog-container"],["mat-dialog-title",""],[3,"jsonString"],["align","end"],["mat-button","","mat-dialog-close",""],["mat-button","","cdkFocusInitial","",3,"click"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"h2",1),AA(2),F(),S(3,"mat-dialog-content"),AA(4),JA(5,"app-json-editor",2),F(),S(6,"mat-dialog-actions",3)(7,"button",4),AA(8,"Cancel"),F(),S(9,"button",5),hA("click",function(){return i.onSave()}),AA(10,"Save"),F()()()),A&2&&(G(2),Gt(i.data.dialogHeader),G(2),Et(" ",i.functionName," "),G(),yA("jsonString",i.jsonString))},dependencies:[Dr,bs,ma,pa,hg,ed],styles:[".dialog-container[_ngcontent-%COMP%]{border-radius:12px;padding:18px;width:500px;box-shadow:0 8px 16px #0006}.editor[_ngcontent-%COMP%]{padding-top:12px}"]})};var _TA=["input"],GTA=["label"],UTA=["*"],KTA=new dA("mat-checkbox-default-options",{providedIn:"root",factory:TaA});function TaA(){return{color:"accent",clickAction:"check-indeterminate",disabledInteractive:!1}}var Ks=function(t){return t[t.Init=0]="Init",t[t.Checked=1]="Checked",t[t.Unchecked=2]="Unchecked",t[t.Indeterminate=3]="Indeterminate",t}(Ks||{}),YTA={provide:yc,useExisting:Ir(()=>bQ),multi:!0},QG=class{source;checked},JaA=TaA(),bQ=(()=>{class t{_elementRef=f(ee);_changeDetectorRef=f(It);_ngZone=f(Qe);_animationMode=f(Si,{optional:!0});_options=f(KTA,{optional:!0});focus(){this._inputElement.nativeElement.focus()}_createChangeEvent(A){let i=new QG;return i.source=this,i.checked=A,i}_getAnimationTargetElement(){return this._inputElement?.nativeElement}_animationClasses={uncheckedToChecked:"mdc-checkbox--anim-unchecked-checked",uncheckedToIndeterminate:"mdc-checkbox--anim-unchecked-indeterminate",checkedToUnchecked:"mdc-checkbox--anim-checked-unchecked",checkedToIndeterminate:"mdc-checkbox--anim-checked-indeterminate",indeterminateToChecked:"mdc-checkbox--anim-indeterminate-checked",indeterminateToUnchecked:"mdc-checkbox--anim-indeterminate-unchecked"};ariaLabel="";ariaLabelledby=null;ariaDescribedby;ariaExpanded;ariaControls;ariaOwns;_uniqueId;id;get inputId(){return`${this.id||this._uniqueId}-input`}required;labelPosition="after";name=null;change=new WA;indeterminateChange=new WA;value;disableRipple;_inputElement;_labelElement;tabIndex;color;disabledInteractive;_onTouched=()=>{};_currentAnimationClass="";_currentCheckState=Ks.Init;_controlValueAccessorChangeFn=()=>{};_validatorChangeFn=()=>{};constructor(){f(Rn).load(lr);let A=f(new wr("tabindex"),{optional:!0});this._options=this._options||JaA,this.color=this._options.color||JaA.color,this.tabIndex=A==null?0:parseInt(A)||0,this.id=this._uniqueId=f(sn).getId("mat-mdc-checkbox-"),this.disabledInteractive=this._options?.disabledInteractive??!1}ngOnChanges(A){A.required&&this._validatorChangeFn()}ngAfterViewInit(){this._syncIndeterminate(this._indeterminate)}get checked(){return this._checked}set checked(A){A!=this.checked&&(this._checked=A,this._changeDetectorRef.markForCheck())}_checked=!1;get disabled(){return this._disabled}set disabled(A){A!==this.disabled&&(this._disabled=A,this._changeDetectorRef.markForCheck())}_disabled=!1;get indeterminate(){return this._indeterminate}set indeterminate(A){let i=A!=this._indeterminate;this._indeterminate=A,i&&(this._indeterminate?this._transitionCheckState(Ks.Indeterminate):this._transitionCheckState(this.checked?Ks.Checked:Ks.Unchecked),this.indeterminateChange.emit(this._indeterminate)),this._syncIndeterminate(this._indeterminate)}_indeterminate=!1;_isRippleDisabled(){return this.disableRipple||this.disabled}_onLabelTextChange(){this._changeDetectorRef.detectChanges()}writeValue(A){this.checked=!!A}registerOnChange(A){this._controlValueAccessorChangeFn=A}registerOnTouched(A){this._onTouched=A}setDisabledState(A){this.disabled=A}validate(A){return this.required&&A.value!==!0?{required:!0}:null}registerOnValidatorChange(A){this._validatorChangeFn=A}_transitionCheckState(A){let i=this._currentCheckState,n=this._getAnimationTargetElement();if(!(i===A||!n)&&(this._currentAnimationClass&&n.classList.remove(this._currentAnimationClass),this._currentAnimationClass=this._getAnimationClassForCheckStateTransition(i,A),this._currentCheckState=A,this._currentAnimationClass.length>0)){n.classList.add(this._currentAnimationClass);let o=this._currentAnimationClass;this._ngZone.runOutsideAngular(()=>{setTimeout(()=>{n.classList.remove(o)},1e3)})}}_emitChangeEvent(){this._controlValueAccessorChangeFn(this.checked),this.change.emit(this._createChangeEvent(this.checked)),this._inputElement&&(this._inputElement.nativeElement.checked=this.checked)}toggle(){this.checked=!this.checked,this._controlValueAccessorChangeFn(this.checked)}_handleInputClick(){let A=this._options?.clickAction;!this.disabled&&A!=="noop"?(this.indeterminate&&A!=="check"&&Promise.resolve().then(()=>{this._indeterminate=!1,this.indeterminateChange.emit(this._indeterminate)}),this._checked=!this._checked,this._transitionCheckState(this._checked?Ks.Checked:Ks.Unchecked),this._emitChangeEvent()):(this.disabled&&this.disabledInteractive||!this.disabled&&A==="noop")&&(this._inputElement.nativeElement.checked=this.checked,this._inputElement.nativeElement.indeterminate=this.indeterminate)}_onInteractionEvent(A){A.stopPropagation()}_onBlur(){Promise.resolve().then(()=>{this._onTouched(),this._changeDetectorRef.markForCheck()})}_getAnimationClassForCheckStateTransition(A,i){if(this._animationMode==="NoopAnimations")return"";switch(A){case Ks.Init:if(i===Ks.Checked)return this._animationClasses.uncheckedToChecked;if(i==Ks.Indeterminate)return this._checked?this._animationClasses.checkedToIndeterminate:this._animationClasses.uncheckedToIndeterminate;break;case Ks.Unchecked:return i===Ks.Checked?this._animationClasses.uncheckedToChecked:this._animationClasses.uncheckedToIndeterminate;case Ks.Checked:return i===Ks.Unchecked?this._animationClasses.checkedToUnchecked:this._animationClasses.checkedToIndeterminate;case Ks.Indeterminate:return i===Ks.Checked?this._animationClasses.indeterminateToChecked:this._animationClasses.indeterminateToUnchecked}return""}_syncIndeterminate(A){let i=this._inputElement;i&&(i.nativeElement.indeterminate=A)}_onInputClick(){this._handleInputClick()}_onTouchTargetClick(){this._handleInputClick(),this.disabled||this._inputElement.nativeElement.focus()}_preventBubblingFromLabel(A){A.target&&this._labelElement.nativeElement.contains(A.target)&&A.stopPropagation()}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-checkbox"]],viewQuery:function(i,n){if(i&1&&(Te(_TA,5),Te(GTA,5)),i&2){let o;XA(o=$A())&&(n._inputElement=o.first),XA(o=$A())&&(n._labelElement=o.first)}},hostAttrs:[1,"mat-mdc-checkbox"],hostVars:16,hostBindings:function(i,n){i&2&&(Hs("id",n.id),Ne("tabindex",null)("aria-label",null)("aria-labelledby",null),fo(n.color?"mat-"+n.color:"mat-accent"),ue("_mat-animation-noopable",n._animationMode==="NoopAnimations")("mdc-checkbox--disabled",n.disabled)("mat-mdc-checkbox-disabled",n.disabled)("mat-mdc-checkbox-checked",n.checked)("mat-mdc-checkbox-disabled-interactive",n.disabledInteractive))},inputs:{ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],ariaDescribedby:[0,"aria-describedby","ariaDescribedby"],ariaExpanded:[2,"aria-expanded","ariaExpanded",ie],ariaControls:[0,"aria-controls","ariaControls"],ariaOwns:[0,"aria-owns","ariaOwns"],id:"id",required:[2,"required","required",ie],labelPosition:"labelPosition",name:"name",value:"value",disableRipple:[2,"disableRipple","disableRipple",ie],tabIndex:[2,"tabIndex","tabIndex",A=>A==null?void 0:zi(A)],color:"color",disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie],checked:[2,"checked","checked",ie],disabled:[2,"disabled","disabled",ie],indeterminate:[2,"indeterminate","indeterminate",ie]},outputs:{change:"change",indeterminateChange:"indeterminateChange"},exportAs:["matCheckbox"],features:[ht([YTA,{provide:w0,useExisting:t,multi:!0}]),ti],ngContentSelectors:UTA,decls:15,vars:23,consts:[["checkbox",""],["input",""],["label",""],["mat-internal-form-field","",3,"click","labelPosition"],[1,"mdc-checkbox"],[1,"mat-mdc-checkbox-touch-target",3,"click"],["type","checkbox",1,"mdc-checkbox__native-control",3,"blur","click","change","checked","indeterminate","disabled","id","required","tabIndex"],[1,"mdc-checkbox__ripple"],[1,"mdc-checkbox__background"],["focusable","false","viewBox","0 0 24 24","aria-hidden","true",1,"mdc-checkbox__checkmark"],["fill","none","d","M1.73,12.91 8.1,19.28 22.79,4.59",1,"mdc-checkbox__checkmark-path"],[1,"mdc-checkbox__mixedmark"],["mat-ripple","",1,"mat-mdc-checkbox-ripple","mat-focus-indicator",3,"matRippleTrigger","matRippleDisabled","matRippleCentered"],[1,"mdc-label",3,"for"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"div",3),hA("click",function(s){return RA(o),xA(n._preventBubblingFromLabel(s))}),S(1,"div",4,0)(3,"div",5),hA("click",function(){return RA(o),xA(n._onTouchTargetClick())}),F(),S(4,"input",6,1),hA("blur",function(){return RA(o),xA(n._onBlur())})("click",function(){return RA(o),xA(n._onInputClick())})("change",function(s){return RA(o),xA(n._onInteractionEvent(s))}),F(),JA(6,"div",7),S(7,"div",8),ar(),S(8,"svg",9),JA(9,"path",10),F(),SI(),JA(10,"div",11),F(),JA(11,"div",12),F(),S(12,"label",13,2),Fe(14),F()()}if(i&2){let o=cr(2);yA("labelPosition",n.labelPosition),G(4),ue("mdc-checkbox--selected",n.checked),yA("checked",n.checked)("indeterminate",n.indeterminate)("disabled",n.disabled&&!n.disabledInteractive)("id",n.inputId)("required",n.required)("tabIndex",n.disabled&&!n.disabledInteractive?-1:n.tabIndex),Ne("aria-label",n.ariaLabel||null)("aria-labelledby",n.ariaLabelledby)("aria-describedby",n.ariaDescribedby)("aria-checked",n.indeterminate?"mixed":null)("aria-controls",n.ariaControls)("aria-disabled",n.disabled&&n.disabledInteractive?!0:null)("aria-expanded",n.ariaExpanded)("aria-owns",n.ariaOwns)("name",n.name)("value",n.value),G(7),yA("matRippleTrigger",o)("matRippleDisabled",n.disableRipple||n.disabled)("matRippleCentered",!0),G(),yA("for",n.inputId)}},dependencies:[rs,MB],styles:['.mdc-checkbox{display:inline-block;position:relative;flex:0 0 18px;box-sizing:content-box;width:18px;height:18px;line-height:0;white-space:nowrap;cursor:pointer;vertical-align:bottom;padding:calc((var(--mdc-checkbox-state-layer-size, 40px) - 18px)/2);margin:calc((var(--mdc-checkbox-state-layer-size, 40px) - var(--mdc-checkbox-state-layer-size, 40px))/2)}.mdc-checkbox:hover>.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-unselected-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity));background-color:var(--mdc-checkbox-unselected-hover-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox:hover>.mat-mdc-checkbox-ripple>.mat-ripple-element{background-color:var(--mdc-checkbox-unselected-hover-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox .mdc-checkbox__native-control:focus+.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-unselected-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity));background-color:var(--mdc-checkbox-unselected-focus-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox .mdc-checkbox__native-control:focus~.mat-mdc-checkbox-ripple .mat-ripple-element{background-color:var(--mdc-checkbox-unselected-focus-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox:active>.mdc-checkbox__native-control+.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-unselected-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity));background-color:var(--mdc-checkbox-unselected-pressed-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox:active>.mdc-checkbox__native-control~.mat-mdc-checkbox-ripple .mat-ripple-element{background-color:var(--mdc-checkbox-unselected-pressed-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox:hover .mdc-checkbox__native-control:checked+.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-selected-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity));background-color:var(--mdc-checkbox-selected-hover-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox:hover .mdc-checkbox__native-control:checked~.mat-mdc-checkbox-ripple .mat-ripple-element{background-color:var(--mdc-checkbox-selected-hover-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox .mdc-checkbox__native-control:focus:checked+.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-selected-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity));background-color:var(--mdc-checkbox-selected-focus-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox .mdc-checkbox__native-control:focus:checked~.mat-mdc-checkbox-ripple .mat-ripple-element{background-color:var(--mdc-checkbox-selected-focus-state-layer-color, var(--mat-sys-primary))}.mdc-checkbox:active>.mdc-checkbox__native-control:checked+.mdc-checkbox__ripple{opacity:var(--mdc-checkbox-selected-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity));background-color:var(--mdc-checkbox-selected-pressed-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox:active>.mdc-checkbox__native-control:checked~.mat-mdc-checkbox-ripple .mat-ripple-element{background-color:var(--mdc-checkbox-selected-pressed-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox .mdc-checkbox__native-control~.mat-mdc-checkbox-ripple .mat-ripple-element,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox .mdc-checkbox__native-control+.mdc-checkbox__ripple{background-color:var(--mdc-checkbox-unselected-hover-state-layer-color, var(--mat-sys-on-surface))}.mdc-checkbox .mdc-checkbox__native-control{position:absolute;margin:0;padding:0;opacity:0;cursor:inherit;width:var(--mdc-checkbox-state-layer-size, 40px);height:var(--mdc-checkbox-state-layer-size, 40px);top:calc((var(--mdc-checkbox-state-layer-size, 40px) - var(--mdc-checkbox-state-layer-size, 40px))/2);right:calc((var(--mdc-checkbox-state-layer-size, 40px) - var(--mdc-checkbox-state-layer-size, 40px))/2);left:calc((var(--mdc-checkbox-state-layer-size, 40px) - var(--mdc-checkbox-state-layer-size, 40px))/2)}.mdc-checkbox--disabled{cursor:default;pointer-events:none}@media(forced-colors: active){.mdc-checkbox--disabled{opacity:.5}}.mdc-checkbox__background{display:inline-flex;position:absolute;align-items:center;justify-content:center;box-sizing:border-box;width:18px;height:18px;border:2px solid currentColor;border-radius:2px;background-color:rgba(0,0,0,0);pointer-events:none;will-change:background-color,border-color;transition:background-color 90ms cubic-bezier(0.4, 0, 0.6, 1),border-color 90ms cubic-bezier(0.4, 0, 0.6, 1);-webkit-print-color-adjust:exact;color-adjust:exact;border-color:var(--mdc-checkbox-unselected-icon-color, var(--mat-sys-on-surface-variant));top:calc((var(--mdc-checkbox-state-layer-size, 40px) - 18px)/2);left:calc((var(--mdc-checkbox-state-layer-size, 40px) - 18px)/2)}.mdc-checkbox__native-control:enabled:checked~.mdc-checkbox__background,.mdc-checkbox__native-control:enabled:indeterminate~.mdc-checkbox__background{border-color:var(--mdc-checkbox-selected-icon-color, var(--mat-sys-primary));background-color:var(--mdc-checkbox-selected-icon-color, var(--mat-sys-primary))}.mdc-checkbox--disabled .mdc-checkbox__background{border-color:var(--mdc-checkbox-disabled-unselected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-checkbox__native-control:disabled:checked~.mdc-checkbox__background,.mdc-checkbox__native-control:disabled:indeterminate~.mdc-checkbox__background{background-color:var(--mdc-checkbox-disabled-selected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));border-color:rgba(0,0,0,0)}.mdc-checkbox:hover>.mdc-checkbox__native-control:not(:checked)~.mdc-checkbox__background,.mdc-checkbox:hover>.mdc-checkbox__native-control:not(:indeterminate)~.mdc-checkbox__background{border-color:var(--mdc-checkbox-unselected-hover-icon-color, var(--mat-sys-on-surface));background-color:rgba(0,0,0,0)}.mdc-checkbox:hover>.mdc-checkbox__native-control:checked~.mdc-checkbox__background,.mdc-checkbox:hover>.mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background{border-color:var(--mdc-checkbox-selected-hover-icon-color, var(--mat-sys-primary));background-color:var(--mdc-checkbox-selected-hover-icon-color, var(--mat-sys-primary))}.mdc-checkbox__native-control:focus:focus:not(:checked)~.mdc-checkbox__background,.mdc-checkbox__native-control:focus:focus:not(:indeterminate)~.mdc-checkbox__background{border-color:var(--mdc-checkbox-unselected-focus-icon-color, var(--mat-sys-on-surface))}.mdc-checkbox__native-control:focus:focus:checked~.mdc-checkbox__background,.mdc-checkbox__native-control:focus:focus:indeterminate~.mdc-checkbox__background{border-color:var(--mdc-checkbox-selected-focus-icon-color, var(--mat-sys-primary));background-color:var(--mdc-checkbox-selected-focus-icon-color, var(--mat-sys-primary))}.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox:hover>.mdc-checkbox__native-control~.mdc-checkbox__background,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox .mdc-checkbox__native-control:focus~.mdc-checkbox__background,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__background{border-color:var(--mdc-checkbox-disabled-unselected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__native-control:checked~.mdc-checkbox__background,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background{background-color:var(--mdc-checkbox-disabled-selected-icon-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));border-color:rgba(0,0,0,0)}.mdc-checkbox__checkmark{position:absolute;top:0;right:0;bottom:0;left:0;width:100%;opacity:0;transition:opacity 180ms cubic-bezier(0.4, 0, 0.6, 1);color:var(--mdc-checkbox-selected-checkmark-color, var(--mat-sys-on-primary))}@media(forced-colors: active){.mdc-checkbox__checkmark{color:CanvasText}}.mdc-checkbox--disabled .mdc-checkbox__checkmark,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__checkmark{color:var(--mdc-checkbox-disabled-selected-checkmark-color, var(--mat-sys-surface))}@media(forced-colors: active){.mdc-checkbox--disabled .mdc-checkbox__checkmark,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__checkmark{color:CanvasText}}.mdc-checkbox__checkmark-path{transition:stroke-dashoffset 180ms cubic-bezier(0.4, 0, 0.6, 1);stroke:currentColor;stroke-width:3.12px;stroke-dashoffset:29.7833385;stroke-dasharray:29.7833385}.mdc-checkbox__mixedmark{width:100%;height:0;transform:scaleX(0) rotate(0deg);border-width:1px;border-style:solid;opacity:0;transition:opacity 90ms cubic-bezier(0.4, 0, 0.6, 1),transform 90ms cubic-bezier(0.4, 0, 0.6, 1);border-color:var(--mdc-checkbox-selected-checkmark-color, var(--mat-sys-on-primary))}@media(forced-colors: active){.mdc-checkbox__mixedmark{margin:0 1px}}.mdc-checkbox--disabled .mdc-checkbox__mixedmark,.mdc-checkbox--disabled.mat-mdc-checkbox-disabled-interactive .mdc-checkbox__mixedmark{border-color:var(--mdc-checkbox-disabled-selected-checkmark-color, var(--mat-sys-surface))}.mdc-checkbox--anim-unchecked-checked .mdc-checkbox__background,.mdc-checkbox--anim-unchecked-indeterminate .mdc-checkbox__background,.mdc-checkbox--anim-checked-unchecked .mdc-checkbox__background,.mdc-checkbox--anim-indeterminate-unchecked .mdc-checkbox__background{animation-duration:180ms;animation-timing-function:linear}.mdc-checkbox--anim-unchecked-checked .mdc-checkbox__checkmark-path{animation:mdc-checkbox-unchecked-checked-checkmark-path 180ms linear;transition:none}.mdc-checkbox--anim-unchecked-indeterminate .mdc-checkbox__mixedmark{animation:mdc-checkbox-unchecked-indeterminate-mixedmark 90ms linear;transition:none}.mdc-checkbox--anim-checked-unchecked .mdc-checkbox__checkmark-path{animation:mdc-checkbox-checked-unchecked-checkmark-path 90ms linear;transition:none}.mdc-checkbox--anim-checked-indeterminate .mdc-checkbox__checkmark{animation:mdc-checkbox-checked-indeterminate-checkmark 90ms linear;transition:none}.mdc-checkbox--anim-checked-indeterminate .mdc-checkbox__mixedmark{animation:mdc-checkbox-checked-indeterminate-mixedmark 90ms linear;transition:none}.mdc-checkbox--anim-indeterminate-checked .mdc-checkbox__checkmark{animation:mdc-checkbox-indeterminate-checked-checkmark 500ms linear;transition:none}.mdc-checkbox--anim-indeterminate-checked .mdc-checkbox__mixedmark{animation:mdc-checkbox-indeterminate-checked-mixedmark 500ms linear;transition:none}.mdc-checkbox--anim-indeterminate-unchecked .mdc-checkbox__mixedmark{animation:mdc-checkbox-indeterminate-unchecked-mixedmark 300ms linear;transition:none}.mdc-checkbox__native-control:checked~.mdc-checkbox__background,.mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background{transition:border-color 90ms cubic-bezier(0, 0, 0.2, 1),background-color 90ms cubic-bezier(0, 0, 0.2, 1)}.mdc-checkbox__native-control:checked~.mdc-checkbox__background>.mdc-checkbox__checkmark>.mdc-checkbox__checkmark-path,.mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background>.mdc-checkbox__checkmark>.mdc-checkbox__checkmark-path{stroke-dashoffset:0}.mdc-checkbox__native-control:checked~.mdc-checkbox__background>.mdc-checkbox__checkmark{transition:opacity 180ms cubic-bezier(0, 0, 0.2, 1),transform 180ms cubic-bezier(0, 0, 0.2, 1);opacity:1}.mdc-checkbox__native-control:checked~.mdc-checkbox__background>.mdc-checkbox__mixedmark{transform:scaleX(1) rotate(-45deg)}.mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background>.mdc-checkbox__checkmark{transform:rotate(45deg);opacity:0;transition:opacity 90ms cubic-bezier(0.4, 0, 0.6, 1),transform 90ms cubic-bezier(0.4, 0, 0.6, 1)}.mdc-checkbox__native-control:indeterminate~.mdc-checkbox__background>.mdc-checkbox__mixedmark{transform:scaleX(1) rotate(0deg);opacity:1}@keyframes mdc-checkbox-unchecked-checked-checkmark-path{0%,50%{stroke-dashoffset:29.7833385}50%{animation-timing-function:cubic-bezier(0, 0, 0.2, 1)}100%{stroke-dashoffset:0}}@keyframes mdc-checkbox-unchecked-indeterminate-mixedmark{0%,68.2%{transform:scaleX(0)}68.2%{animation-timing-function:cubic-bezier(0, 0, 0, 1)}100%{transform:scaleX(1)}}@keyframes mdc-checkbox-checked-unchecked-checkmark-path{from{animation-timing-function:cubic-bezier(0.4, 0, 1, 1);opacity:1;stroke-dashoffset:0}to{opacity:0;stroke-dashoffset:-29.7833385}}@keyframes mdc-checkbox-checked-indeterminate-checkmark{from{animation-timing-function:cubic-bezier(0, 0, 0.2, 1);transform:rotate(0deg);opacity:1}to{transform:rotate(45deg);opacity:0}}@keyframes mdc-checkbox-indeterminate-checked-checkmark{from{animation-timing-function:cubic-bezier(0.14, 0, 0, 1);transform:rotate(45deg);opacity:0}to{transform:rotate(360deg);opacity:1}}@keyframes mdc-checkbox-checked-indeterminate-mixedmark{from{animation-timing-function:cubic-bezier(0, 0, 0.2, 1);transform:rotate(-45deg);opacity:0}to{transform:rotate(0deg);opacity:1}}@keyframes mdc-checkbox-indeterminate-checked-mixedmark{from{animation-timing-function:cubic-bezier(0.14, 0, 0, 1);transform:rotate(0deg);opacity:1}to{transform:rotate(315deg);opacity:0}}@keyframes mdc-checkbox-indeterminate-unchecked-mixedmark{0%{animation-timing-function:linear;transform:scaleX(1);opacity:1}32.8%,100%{transform:scaleX(0);opacity:0}}.mat-mdc-checkbox{display:inline-block;position:relative;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mat-mdc-checkbox-touch-target,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__native-control,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__ripple,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mat-mdc-checkbox-ripple::before,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__background,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__background>.mdc-checkbox__checkmark,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__background>.mdc-checkbox__checkmark>.mdc-checkbox__checkmark-path,.mat-mdc-checkbox._mat-animation-noopable>.mat-internal-form-field>.mdc-checkbox>.mdc-checkbox__background>.mdc-checkbox__mixedmark{transition:none !important;animation:none !important}.mat-mdc-checkbox label{cursor:pointer}.mat-mdc-checkbox .mat-internal-form-field{color:var(--mat-checkbox-label-text-color, var(--mat-sys-on-surface));font-family:var(--mat-checkbox-label-text-font, var(--mat-sys-body-medium-font));line-height:var(--mat-checkbox-label-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mat-checkbox-label-text-size, var(--mat-sys-body-medium-size));letter-spacing:var(--mat-checkbox-label-text-tracking, var(--mat-sys-body-medium-tracking));font-weight:var(--mat-checkbox-label-text-weight, var(--mat-sys-body-medium-weight))}.mat-mdc-checkbox.mat-mdc-checkbox-disabled.mat-mdc-checkbox-disabled-interactive{pointer-events:auto}.mat-mdc-checkbox.mat-mdc-checkbox-disabled.mat-mdc-checkbox-disabled-interactive input{cursor:default}.mat-mdc-checkbox.mat-mdc-checkbox-disabled label{cursor:default;color:var(--mat-checkbox-disabled-label-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-checkbox label:empty{display:none}.mat-mdc-checkbox .mdc-checkbox__ripple{opacity:0}.mat-mdc-checkbox .mat-mdc-checkbox-ripple,.mdc-checkbox__ripple{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:50%;pointer-events:none}.mat-mdc-checkbox .mat-mdc-checkbox-ripple:not(:empty),.mdc-checkbox__ripple:not(:empty){transform:translateZ(0)}.mat-mdc-checkbox-ripple .mat-ripple-element{opacity:.1}.mat-mdc-checkbox-touch-target{position:absolute;top:50%;left:50%;height:48px;width:48px;transform:translate(-50%, -50%);display:var(--mat-checkbox-touch-target-display, block)}.mat-mdc-checkbox .mat-mdc-checkbox-ripple::before{border-radius:50%}.mdc-checkbox__native-control:focus~.mat-focus-indicator::before{content:""}'],encapsulation:2,changeDetection:0})}return t})();var HaA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[bQ,it,it]})}return t})();var HTA=[[["caption"]],[["colgroup"],["col"]],"*"],zTA=["caption","colgroup, col","*"];function OTA(t,e){t&1&&Fe(0,2)}function PTA(t,e){t&1&&(S(0,"thead",0),_r(1,1),F(),S(2,"tbody",0),_r(3,2)(4,3),F(),S(5,"tfoot",0),_r(6,4),F())}function jTA(t,e){t&1&&_r(0,1)(1,2)(2,3)(3,4)}var Yl=new dA("CDK_TABLE");var Wy=(()=>{class t{template=f(wn);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkCellDef",""]]})}return t})(),Xy=(()=>{class t{template=f(wn);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkHeaderCellDef",""]]})}return t})(),PaA=(()=>{class t{template=f(wn);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkFooterCellDef",""]]})}return t})(),MQ=(()=>{class t{_table=f(Yl,{optional:!0});_hasStickyChanged=!1;get name(){return this._name}set name(A){this._setNameInput(A)}_name;get sticky(){return this._sticky}set sticky(A){A!==this._sticky&&(this._sticky=A,this._hasStickyChanged=!0)}_sticky=!1;get stickyEnd(){return this._stickyEnd}set stickyEnd(A){A!==this._stickyEnd&&(this._stickyEnd=A,this._hasStickyChanged=!0)}_stickyEnd=!1;cell;headerCell;footerCell;cssClassFriendlyName;_columnCssClassName;constructor(){}hasStickyChanged(){let A=this._hasStickyChanged;return this.resetStickyChanged(),A}resetStickyChanged(){this._hasStickyChanged=!1}_updateColumnCssClassName(){this._columnCssClassName=[`cdk-column-${this.cssClassFriendlyName}`]}_setNameInput(A){A&&(this._name=A,this.cssClassFriendlyName=A.replace(/[^a-z0-9_-]/gi,"-"),this._updateColumnCssClassName())}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkColumnDef",""]],contentQueries:function(i,n,o){if(i&1&&(ci(o,Wy,5),ci(o,Xy,5),ci(o,PaA,5)),i&2){let r;XA(r=$A())&&(n.cell=r.first),XA(r=$A())&&(n.headerCell=r.first),XA(r=$A())&&(n.footerCell=r.first)}},inputs:{name:[0,"cdkColumnDef","name"],sticky:[2,"sticky","sticky",ie],stickyEnd:[2,"stickyEnd","stickyEnd",ie]},features:[ht([{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:t}])]})}return t})(),jy=class{constructor(e,A){A.nativeElement.classList.add(...e._columnCssClassName)}},jaA=(()=>{class t extends jy{constructor(){super(f(MQ),f(ee))}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["cdk-header-cell"],["th","cdk-header-cell",""]],hostAttrs:["role","columnheader",1,"cdk-header-cell"],features:[lt]})}return t})();var qaA=(()=>{class t extends jy{constructor(){let A=f(MQ),i=f(ee);super(A,i);let n=A._table?._getCellRole();n&&i.nativeElement.setAttribute("role",n)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["cdk-cell"],["td","cdk-cell",""]],hostAttrs:[1,"cdk-cell"],features:[lt]})}return t})(),qy=class{tasks=[];endTasks=[]},Vy=new dA("_COALESCED_STYLE_SCHEDULER"),uG=(()=>{class t{_currentSchedule=null;_ngZone=f(Qe);constructor(){}schedule(A){this._createScheduleIfNeeded(),this._currentSchedule.tasks.push(A)}scheduleEnd(A){this._createScheduleIfNeeded(),this._currentSchedule.endTasks.push(A)}_createScheduleIfNeeded(){this._currentSchedule||(this._currentSchedule=new qy,this._ngZone.runOutsideAngular(()=>queueMicrotask(()=>{for(;this._currentSchedule.tasks.length||this._currentSchedule.endTasks.length;){let A=this._currentSchedule;this._currentSchedule=new qy;for(let i of A.tasks)i();for(let i of A.endTasks)i()}this._currentSchedule=null})))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();var fG=(()=>{class t{template=f(wn);_differs=f(rg);columns;_columnsDiffer;constructor(){}ngOnChanges(A){if(!this._columnsDiffer){let i=A.columns&&A.columns.currentValue||[];this._columnsDiffer=this._differs.find(i).create(),this._columnsDiffer.diff(i)}}getColumnsDiff(){return this._columnsDiffer.diff(this.columns)}extractCellTemplate(A){return this instanceof If?A.headerCell.template:this instanceof mG?A.footerCell.template:A.cell.template}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,features:[ti]})}return t})(),If=(()=>{class t extends fG{_table=f(Yl,{optional:!0});_hasStickyChanged=!1;get sticky(){return this._sticky}set sticky(A){A!==this._sticky&&(this._sticky=A,this._hasStickyChanged=!0)}_sticky=!1;constructor(){super(f(wn),f(rg))}ngOnChanges(A){super.ngOnChanges(A)}hasStickyChanged(){let A=this._hasStickyChanged;return this.resetStickyChanged(),A}resetStickyChanged(){this._hasStickyChanged=!1}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkHeaderRowDef",""]],inputs:{columns:[0,"cdkHeaderRowDef","columns"],sticky:[2,"cdkHeaderRowDefSticky","sticky",ie]},features:[lt,ti]})}return t})(),mG=(()=>{class t extends fG{_table=f(Yl,{optional:!0});_hasStickyChanged=!1;get sticky(){return this._sticky}set sticky(A){A!==this._sticky&&(this._sticky=A,this._hasStickyChanged=!0)}_sticky=!1;constructor(){super(f(wn),f(rg))}ngOnChanges(A){super.ngOnChanges(A)}hasStickyChanged(){let A=this._hasStickyChanged;return this.resetStickyChanged(),A}resetStickyChanged(){this._hasStickyChanged=!1}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkFooterRowDef",""]],inputs:{columns:[0,"cdkFooterRowDef","columns"],sticky:[2,"cdkFooterRowDefSticky","sticky",ie]},features:[lt,ti]})}return t})(),$y=(()=>{class t extends fG{_table=f(Yl,{optional:!0});when;constructor(){super(f(wn),f(rg))}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkRowDef",""]],inputs:{columns:[0,"cdkRowDefColumns","columns"],when:[0,"cdkRowDefWhen","when"]},features:[lt]})}return t})(),td=(()=>{class t{_viewContainer=f(Nn);cells;context;static mostRecentCellOutlet=null;constructor(){t.mostRecentCellOutlet=this}ngOnDestroy(){t.mostRecentCellOutlet===this&&(t.mostRecentCellOutlet=null)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","cdkCellOutlet",""]]})}return t})(),pG=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["cdk-header-row"],["tr","cdk-header-row",""]],hostAttrs:["role","row",1,"cdk-header-row"],decls:1,vars:0,consts:[["cdkCellOutlet",""]],template:function(i,n){i&1&&_r(0,0)},dependencies:[td],encapsulation:2})}return t})();var wG=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["cdk-row"],["tr","cdk-row",""]],hostAttrs:["role","row",1,"cdk-row"],decls:1,vars:0,consts:[["cdkCellOutlet",""]],template:function(i,n){i&1&&_r(0,0)},dependencies:[td],encapsulation:2})}return t})(),VaA=(()=>{class t{templateRef=f(wn);_contentClassName="cdk-no-data-row";constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["ng-template","cdkNoDataRow",""]]})}return t})(),zaA=["top","bottom","left","right"],hG=class{_isNativeHtmlTable;_stickCellCss;direction;_coalescedStyleScheduler;_isBrowser;_needsPositionStickyOnElement;_positionListener;_tableInjector;_elemSizeCache=new WeakMap;_resizeObserver=globalThis?.ResizeObserver?new globalThis.ResizeObserver(e=>this._updateCachedSizes(e)):null;_updatedStickyColumnsParamsToReplay=[];_stickyColumnsReplayTimeout=null;_cachedCellWidths=[];_borderCellCss;_destroyed=!1;constructor(e,A,i,n,o=!0,r=!0,s,a){this._isNativeHtmlTable=e,this._stickCellCss=A,this.direction=i,this._coalescedStyleScheduler=n,this._isBrowser=o,this._needsPositionStickyOnElement=r,this._positionListener=s,this._tableInjector=a,this._borderCellCss={top:`${A}-border-elem-top`,bottom:`${A}-border-elem-bottom`,left:`${A}-border-elem-left`,right:`${A}-border-elem-right`}}clearStickyPositioning(e,A){(A.includes("left")||A.includes("right"))&&this._removeFromStickyColumnReplayQueue(e);let i=[];for(let n of e)n.nodeType===n.ELEMENT_NODE&&i.push(n,...Array.from(n.children));this._afterNextRender({write:()=>{for(let n of i)this._removeStickyStyle(n,A)}})}updateStickyColumns(e,A,i,n=!0,o=!0){if(!e.length||!this._isBrowser||!(A.some(h=>h)||i.some(h=>h))){this._positionListener?.stickyColumnsUpdated({sizes:[]}),this._positionListener?.stickyEndColumnsUpdated({sizes:[]});return}let r=e[0],s=r.children.length,a=this.direction==="rtl",c=a?"right":"left",l=a?"left":"right",I=A.lastIndexOf(!0),C=i.indexOf(!0),d,B,E;o&&this._updateStickyColumnReplayQueue({rows:[...e],stickyStartStates:[...A],stickyEndStates:[...i]}),this._afterNextRender({earlyRead:()=>{d=this._getCellWidths(r,n),B=this._getStickyStartColumnPositions(d,A),E=this._getStickyEndColumnPositions(d,i)},write:()=>{for(let h of e)for(let u=0;u<s;u++){let D=h.children[u];A[u]&&this._addStickyStyle(D,c,B[u],u===I),i[u]&&this._addStickyStyle(D,l,E[u],u===C)}this._positionListener&&d.some(h=>!!h)&&(this._positionListener.stickyColumnsUpdated({sizes:I===-1?[]:d.slice(0,I+1).map((h,u)=>A[u]?h:null)}),this._positionListener.stickyEndColumnsUpdated({sizes:C===-1?[]:d.slice(C).map((h,u)=>i[u+C]?h:null).reverse()}))}})}stickRows(e,A,i){if(!this._isBrowser)return;let n=i==="bottom"?e.slice().reverse():e,o=i==="bottom"?A.slice().reverse():A,r=[],s=[],a=[];this._afterNextRender({earlyRead:()=>{for(let c=0,l=0;c<n.length;c++){if(!o[c])continue;r[c]=l;let I=n[c];a[c]=this._isNativeHtmlTable?Array.from(I.children):[I];let C=this._retrieveElementSize(I).height;l+=C,s[c]=C}},write:()=>{let c=o.lastIndexOf(!0);for(let l=0;l<n.length;l++){if(!o[l])continue;let I=r[l],C=l===c;for(let d of a[l])this._addStickyStyle(d,i,I,C)}i==="top"?this._positionListener?.stickyHeaderRowsUpdated({sizes:s,offsets:r,elements:a}):this._positionListener?.stickyFooterRowsUpdated({sizes:s,offsets:r,elements:a})}})}updateStickyFooterContainer(e,A){this._isNativeHtmlTable&&this._afterNextRender({write:()=>{let i=e.querySelector("tfoot");i&&(A.some(n=>!n)?this._removeStickyStyle(i,["bottom"]):this._addStickyStyle(i,"bottom",0,!1))}})}destroy(){this._stickyColumnsReplayTimeout&&clearTimeout(this._stickyColumnsReplayTimeout),this._resizeObserver?.disconnect(),this._destroyed=!0}_removeStickyStyle(e,A){for(let n of A)e.style[n]="",e.classList.remove(this._borderCellCss[n]);zaA.some(n=>A.indexOf(n)===-1&&e.style[n])?e.style.zIndex=this._getCalculatedZIndex(e):(e.style.zIndex="",this._needsPositionStickyOnElement&&(e.style.position=""),e.classList.remove(this._stickCellCss))}_addStickyStyle(e,A,i,n){e.classList.add(this._stickCellCss),n&&e.classList.add(this._borderCellCss[A]),e.style[A]=`${i}px`,e.style.zIndex=this._getCalculatedZIndex(e),this._needsPositionStickyOnElement&&(e.style.cssText+="position: -webkit-sticky; position: sticky; ")}_getCalculatedZIndex(e){let A={top:100,bottom:10,left:1,right:1},i=0;for(let n of zaA)e.style[n]&&(i+=A[n]);return i?`${i}`:""}_getCellWidths(e,A=!0){if(!A&&this._cachedCellWidths.length)return this._cachedCellWidths;let i=[],n=e.children;for(let o=0;o<n.length;o++){let r=n[o];i.push(this._retrieveElementSize(r).width)}return this._cachedCellWidths=i,i}_getStickyStartColumnPositions(e,A){let i=[],n=0;for(let o=0;o<e.length;o++)A[o]&&(i[o]=n,n+=e[o]);return i}_getStickyEndColumnPositions(e,A){let i=[],n=0;for(let o=e.length;o>0;o--)A[o]&&(i[o]=n,n+=e[o]);return i}_retrieveElementSize(e){let A=this._elemSizeCache.get(e);if(A)return A;let i=e.getBoundingClientRect(),n={width:i.width,height:i.height};return this._resizeObserver&&(this._elemSizeCache.set(e,n),this._resizeObserver.observe(e,{box:"border-box"})),n}_updateStickyColumnReplayQueue(e){this._removeFromStickyColumnReplayQueue(e.rows),this._stickyColumnsReplayTimeout||this._updatedStickyColumnsParamsToReplay.push(e)}_removeFromStickyColumnReplayQueue(e){let A=new Set(e);for(let i of this._updatedStickyColumnsParamsToReplay)i.rows=i.rows.filter(n=>!A.has(n));this._updatedStickyColumnsParamsToReplay=this._updatedStickyColumnsParamsToReplay.filter(i=>!!i.rows.length)}_updateCachedSizes(e){let A=!1;for(let i of e){let n=i.borderBoxSize?.length?{width:i.borderBoxSize[0].inlineSize,height:i.borderBoxSize[0].blockSize}:{width:i.contentRect.width,height:i.contentRect.height};n.width!==this._elemSizeCache.get(i.target)?.width&&qTA(i.target)&&(A=!0),this._elemSizeCache.set(i.target,n)}A&&this._updatedStickyColumnsParamsToReplay.length&&(this._stickyColumnsReplayTimeout&&clearTimeout(this._stickyColumnsReplayTimeout),this._stickyColumnsReplayTimeout=setTimeout(()=>{if(!this._destroyed){for(let i of this._updatedStickyColumnsParamsToReplay)this.updateStickyColumns(i.rows,i.stickyStartStates,i.stickyEndStates,!0,!1);this._updatedStickyColumnsParamsToReplay=[],this._stickyColumnsReplayTimeout=null}},0))}_afterNextRender(e){this._tableInjector?To(e,{injector:this._tableInjector}):this._coalescedStyleScheduler.schedule(()=>{e.earlyRead?.(),e.write()})}};function qTA(t){return["cdk-cell","cdk-header-cell","cdk-footer-cell"].some(e=>t.classList.contains(e))}var Zy=new dA("CDK_SPL");var DG=(()=>{class t{viewContainer=f(Nn);elementRef=f(ee);constructor(){let A=f(Yl);A._rowOutlet=this,A._outletAssigned()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","rowOutlet",""]]})}return t})(),yG=(()=>{class t{viewContainer=f(Nn);elementRef=f(ee);constructor(){let A=f(Yl);A._headerRowOutlet=this,A._outletAssigned()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","headerRowOutlet",""]]})}return t})(),vG=(()=>{class t{viewContainer=f(Nn);elementRef=f(ee);constructor(){let A=f(Yl);A._footerRowOutlet=this,A._outletAssigned()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","footerRowOutlet",""]]})}return t})(),bG=(()=>{class t{viewContainer=f(Nn);elementRef=f(ee);constructor(){let A=f(Yl);A._noDataRowOutlet=this,A._outletAssigned()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","noDataRowOutlet",""]]})}return t})();var MG=(()=>{class t{_differs=f(rg);_changeDetectorRef=f(It);_elementRef=f(ee);_dir=f(mo,{optional:!0});_platform=f(Ii);_viewRepeater=f(Ru);_coalescedStyleScheduler=f(Vy);_viewportRuler=f(Mc);_stickyPositioningListener=f(Zy,{optional:!0,skipSelf:!0});_document=f(at);_data;_onDestroy=new OA;_renderRows;_renderChangeSubscription;_columnDefsByName=new Map;_rowDefs;_headerRowDefs;_footerRowDefs;_dataDiffer;_defaultRowDef;_customColumnDefs=new Set;_customRowDefs=new Set;_customHeaderRowDefs=new Set;_customFooterRowDefs=new Set;_customNoDataRow;_headerRowDefChanged=!0;_footerRowDefChanged=!0;_stickyColumnStylesNeedReset=!0;_forceRecalculateCellWidths=!0;_cachedRenderRowsMap=new Map;_isNativeHtmlTable;_stickyStyler;stickyCssClass="cdk-table-sticky";needsPositionStickyOnElement=!0;_isServer;_isShowingNoDataRow=!1;_hasAllOutlets=!1;_hasInitialized=!1;_getCellRole(){if(this._cellRoleInternal===void 0){let A=this._elementRef.nativeElement.getAttribute("role");return A==="grid"||A==="treegrid"?"gridcell":"cell"}return this._cellRoleInternal}_cellRoleInternal=void 0;get trackBy(){return this._trackByFn}set trackBy(A){this._trackByFn=A}_trackByFn;get dataSource(){return this._dataSource}set dataSource(A){this._dataSource!==A&&this._switchDataSource(A)}_dataSource;get multiTemplateDataRows(){return this._multiTemplateDataRows}set multiTemplateDataRows(A){this._multiTemplateDataRows=A,this._rowOutlet&&this._rowOutlet.viewContainer.length&&(this._forceRenderDataRows(),this.updateStickyColumnStyles())}_multiTemplateDataRows=!1;get fixedLayout(){return this._fixedLayout}set fixedLayout(A){this._fixedLayout=A,this._forceRecalculateCellWidths=!0,this._stickyColumnStylesNeedReset=!0}_fixedLayout=!1;contentChanged=new WA;viewChange=new Mi({start:0,end:Number.MAX_VALUE});_rowOutlet;_headerRowOutlet;_footerRowOutlet;_noDataRowOutlet;_contentColumnDefs;_contentRowDefs;_contentHeaderRowDefs;_contentFooterRowDefs;_noDataRow;_injector=f(Rt);constructor(){f(new wr("role"),{optional:!0})||this._elementRef.nativeElement.setAttribute("role","table"),this._isServer=!this._platform.isBrowser,this._isNativeHtmlTable=this._elementRef.nativeElement.nodeName==="TABLE"}ngOnInit(){this._setupStickyStyler(),this._dataDiffer=this._differs.find([]).create((A,i)=>this.trackBy?this.trackBy(i.dataIndex,i.data):i),this._viewportRuler.change().pipe(St(this._onDestroy)).subscribe(()=>{this._forceRecalculateCellWidths=!0})}ngAfterContentInit(){this._hasInitialized=!0}ngAfterContentChecked(){this._canRender()&&this._render()}ngOnDestroy(){this._stickyStyler?.destroy(),[this._rowOutlet?.viewContainer,this._headerRowOutlet?.viewContainer,this._footerRowOutlet?.viewContainer,this._cachedRenderRowsMap,this._customColumnDefs,this._customRowDefs,this._customHeaderRowDefs,this._customFooterRowDefs,this._columnDefsByName].forEach(A=>{A?.clear()}),this._headerRowDefs=[],this._footerRowDefs=[],this._defaultRowDef=null,this._onDestroy.next(),this._onDestroy.complete(),I8(this.dataSource)&&this.dataSource.disconnect(this)}renderRows(){this._renderRows=this._getAllRenderRows();let A=this._dataDiffer.diff(this._renderRows);if(!A){this._updateNoDataRow(),this.contentChanged.next();return}let i=this._rowOutlet.viewContainer;this._viewRepeater.applyChanges(A,i,(n,o,r)=>this._getEmbeddedViewArgs(n.item,r),n=>n.item.data,n=>{n.operation===SB.INSERTED&&n.context&&this._renderCellTemplateForItem(n.record.item.rowDef,n.context)}),this._updateRowIndexContext(),A.forEachIdentityChange(n=>{let o=i.get(n.currentIndex);o.context.$implicit=n.item.data}),this._updateNoDataRow(),this.contentChanged.next(),this.updateStickyColumnStyles()}addColumnDef(A){this._customColumnDefs.add(A)}removeColumnDef(A){this._customColumnDefs.delete(A)}addRowDef(A){this._customRowDefs.add(A)}removeRowDef(A){this._customRowDefs.delete(A)}addHeaderRowDef(A){this._customHeaderRowDefs.add(A),this._headerRowDefChanged=!0}removeHeaderRowDef(A){this._customHeaderRowDefs.delete(A),this._headerRowDefChanged=!0}addFooterRowDef(A){this._customFooterRowDefs.add(A),this._footerRowDefChanged=!0}removeFooterRowDef(A){this._customFooterRowDefs.delete(A),this._footerRowDefChanged=!0}setNoDataRow(A){this._customNoDataRow=A}updateStickyHeaderRowStyles(){let A=this._getRenderedRows(this._headerRowOutlet);if(this._isNativeHtmlTable){let n=OaA(this._headerRowOutlet,"thead");n&&(n.style.display=A.length?"":"none")}let i=this._headerRowDefs.map(n=>n.sticky);this._stickyStyler.clearStickyPositioning(A,["top"]),this._stickyStyler.stickRows(A,i,"top"),this._headerRowDefs.forEach(n=>n.resetStickyChanged())}updateStickyFooterRowStyles(){let A=this._getRenderedRows(this._footerRowOutlet);if(this._isNativeHtmlTable){let n=OaA(this._footerRowOutlet,"tfoot");n&&(n.style.display=A.length?"":"none")}let i=this._footerRowDefs.map(n=>n.sticky);this._stickyStyler.clearStickyPositioning(A,["bottom"]),this._stickyStyler.stickRows(A,i,"bottom"),this._stickyStyler.updateStickyFooterContainer(this._elementRef.nativeElement,i),this._footerRowDefs.forEach(n=>n.resetStickyChanged())}updateStickyColumnStyles(){let A=this._getRenderedRows(this._headerRowOutlet),i=this._getRenderedRows(this._rowOutlet),n=this._getRenderedRows(this._footerRowOutlet);(this._isNativeHtmlTable&&!this._fixedLayout||this._stickyColumnStylesNeedReset)&&(this._stickyStyler.clearStickyPositioning([...A,...i,...n],["left","right"]),this._stickyColumnStylesNeedReset=!1),A.forEach((o,r)=>{this._addStickyColumnStyles([o],this._headerRowDefs[r])}),this._rowDefs.forEach(o=>{let r=[];for(let s=0;s<i.length;s++)this._renderRows[s].rowDef===o&&r.push(i[s]);this._addStickyColumnStyles(r,o)}),n.forEach((o,r)=>{this._addStickyColumnStyles([o],this._footerRowDefs[r])}),Array.from(this._columnDefsByName.values()).forEach(o=>o.resetStickyChanged())}_outletAssigned(){!this._hasAllOutlets&&this._rowOutlet&&this._headerRowOutlet&&this._footerRowOutlet&&this._noDataRowOutlet&&(this._hasAllOutlets=!0,this._canRender()&&this._render())}_canRender(){return this._hasAllOutlets&&this._hasInitialized}_render(){this._cacheRowDefs(),this._cacheColumnDefs(),!this._headerRowDefs.length&&!this._footerRowDefs.length&&this._rowDefs.length;let i=this._renderUpdatedColumns()||this._headerRowDefChanged||this._footerRowDefChanged;this._stickyColumnStylesNeedReset=this._stickyColumnStylesNeedReset||i,this._forceRecalculateCellWidths=i,this._headerRowDefChanged&&(this._forceRenderHeaderRows(),this._headerRowDefChanged=!1),this._footerRowDefChanged&&(this._forceRenderFooterRows(),this._footerRowDefChanged=!1),this.dataSource&&this._rowDefs.length>0&&!this._renderChangeSubscription?this._observeRenderChanges():this._stickyColumnStylesNeedReset&&this.updateStickyColumnStyles(),this._checkStickyStates()}_getAllRenderRows(){let A=[],i=this._cachedRenderRowsMap;this._cachedRenderRowsMap=new Map;for(let n=0;n<this._data.length;n++){let o=this._data[n],r=this._getRenderRowsForData(o,n,i.get(o));this._cachedRenderRowsMap.has(o)||this._cachedRenderRowsMap.set(o,new WeakMap);for(let s=0;s<r.length;s++){let a=r[s],c=this._cachedRenderRowsMap.get(a.data);c.has(a.rowDef)?c.get(a.rowDef).push(a):c.set(a.rowDef,[a]),A.push(a)}}return A}_getRenderRowsForData(A,i,n){return this._getRowDefs(A,i).map(r=>{let s=n&&n.has(r)?n.get(r):[];if(s.length){let a=s.shift();return a.dataIndex=i,a}else return{data:A,rowDef:r,dataIndex:i}})}_cacheColumnDefs(){this._columnDefsByName.clear(),Py(this._getOwnDefs(this._contentColumnDefs),this._customColumnDefs).forEach(i=>{this._columnDefsByName.has(i.name),this._columnDefsByName.set(i.name,i)})}_cacheRowDefs(){this._headerRowDefs=Py(this._getOwnDefs(this._contentHeaderRowDefs),this._customHeaderRowDefs),this._footerRowDefs=Py(this._getOwnDefs(this._contentFooterRowDefs),this._customFooterRowDefs),this._rowDefs=Py(this._getOwnDefs(this._contentRowDefs),this._customRowDefs);let A=this._rowDefs.filter(i=>!i.when);!this.multiTemplateDataRows&&A.length>1,this._defaultRowDef=A[0]}_renderUpdatedColumns(){let A=(r,s)=>{let a=!!s.getColumnsDiff();return r||a},i=this._rowDefs.reduce(A,!1);i&&this._forceRenderDataRows();let n=this._headerRowDefs.reduce(A,!1);n&&this._forceRenderHeaderRows();let o=this._footerRowDefs.reduce(A,!1);return o&&this._forceRenderFooterRows(),i||n||o}_switchDataSource(A){this._data=[],I8(this.dataSource)&&this.dataSource.disconnect(this),this._renderChangeSubscription&&(this._renderChangeSubscription.unsubscribe(),this._renderChangeSubscription=null),A||(this._dataDiffer&&this._dataDiffer.diff([]),this._rowOutlet&&this._rowOutlet.viewContainer.clear()),this._dataSource=A}_observeRenderChanges(){if(!this.dataSource)return;let A;I8(this.dataSource)?A=this.dataSource.connect(this):I2(this.dataSource)?A=this.dataSource:Array.isArray(this.dataSource)&&(A=Me(this.dataSource)),this._renderChangeSubscription=A.pipe(St(this._onDestroy)).subscribe(i=>{this._data=i||[],this.renderRows()})}_forceRenderHeaderRows(){this._headerRowOutlet.viewContainer.length>0&&this._headerRowOutlet.viewContainer.clear(),this._headerRowDefs.forEach((A,i)=>this._renderRow(this._headerRowOutlet,A,i)),this.updateStickyHeaderRowStyles()}_forceRenderFooterRows(){this._footerRowOutlet.viewContainer.length>0&&this._footerRowOutlet.viewContainer.clear(),this._footerRowDefs.forEach((A,i)=>this._renderRow(this._footerRowOutlet,A,i)),this.updateStickyFooterRowStyles()}_addStickyColumnStyles(A,i){let n=Array.from(i?.columns||[]).map(s=>{let a=this._columnDefsByName.get(s);return a}),o=n.map(s=>s.sticky),r=n.map(s=>s.stickyEnd);this._stickyStyler.updateStickyColumns(A,o,r,!this._fixedLayout||this._forceRecalculateCellWidths)}_getRenderedRows(A){let i=[];for(let n=0;n<A.viewContainer.length;n++){let o=A.viewContainer.get(n);i.push(o.rootNodes[0])}return i}_getRowDefs(A,i){if(this._rowDefs.length==1)return[this._rowDefs[0]];let n=[];if(this.multiTemplateDataRows)n=this._rowDefs.filter(o=>!o.when||o.when(i,A));else{let o=this._rowDefs.find(r=>r.when&&r.when(i,A))||this._defaultRowDef;o&&n.push(o)}return n.length,n}_getEmbeddedViewArgs(A,i){let n=A.rowDef,o={$implicit:A.data};return{templateRef:n.template,context:o,index:i}}_renderRow(A,i,n,o={}){let r=A.viewContainer.createEmbeddedView(i.template,o,n);return this._renderCellTemplateForItem(i,o),r}_renderCellTemplateForItem(A,i){for(let n of this._getCellTemplates(A))td.mostRecentCellOutlet&&td.mostRecentCellOutlet._viewContainer.createEmbeddedView(n,i);this._changeDetectorRef.markForCheck()}_updateRowIndexContext(){let A=this._rowOutlet.viewContainer;for(let i=0,n=A.length;i<n;i++){let r=A.get(i).context;r.count=n,r.first=i===0,r.last=i===n-1,r.even=i%2===0,r.odd=!r.even,this.multiTemplateDataRows?(r.dataIndex=this._renderRows[i].dataIndex,r.renderIndex=i):r.index=this._renderRows[i].dataIndex}}_getCellTemplates(A){return!A||!A.columns?[]:Array.from(A.columns,i=>{let n=this._columnDefsByName.get(i);return A.extractCellTemplate(n)})}_forceRenderDataRows(){this._dataDiffer.diff([]),this._rowOutlet.viewContainer.clear(),this.renderRows()}_checkStickyStates(){let A=(i,n)=>i||n.hasStickyChanged();this._headerRowDefs.reduce(A,!1)&&this.updateStickyHeaderRowStyles(),this._footerRowDefs.reduce(A,!1)&&this.updateStickyFooterRowStyles(),Array.from(this._columnDefsByName.values()).reduce(A,!1)&&(this._stickyColumnStylesNeedReset=!0,this.updateStickyColumnStyles())}_setupStickyStyler(){let A=this._dir?this._dir.value:"ltr";this._stickyStyler=new hG(this._isNativeHtmlTable,this.stickyCssClass,A,this._coalescedStyleScheduler,this._platform.isBrowser,this.needsPositionStickyOnElement,this._stickyPositioningListener,this._injector),(this._dir?this._dir.change:Me()).pipe(St(this._onDestroy)).subscribe(i=>{this._stickyStyler.direction=i,this.updateStickyColumnStyles()})}_getOwnDefs(A){return A.filter(i=>!i._table||i._table===this)}_updateNoDataRow(){let A=this._customNoDataRow||this._noDataRow;if(!A)return;let i=this._rowOutlet.viewContainer.length===0;if(i===this._isShowingNoDataRow)return;let n=this._noDataRowOutlet.viewContainer;if(i){let o=n.createEmbeddedView(A.templateRef),r=o.rootNodes[0];o.rootNodes.length===1&&r?.nodeType===this._document.ELEMENT_NODE&&(r.setAttribute("role","row"),r.classList.add(A._contentClassName))}else n.clear();this._isShowingNoDataRow=i,this._changeDetectorRef.markForCheck()}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["cdk-table"],["table","cdk-table",""]],contentQueries:function(i,n,o){if(i&1&&(ci(o,VaA,5),ci(o,MQ,5),ci(o,$y,5),ci(o,If,5),ci(o,mG,5)),i&2){let r;XA(r=$A())&&(n._noDataRow=r.first),XA(r=$A())&&(n._contentColumnDefs=r),XA(r=$A())&&(n._contentRowDefs=r),XA(r=$A())&&(n._contentHeaderRowDefs=r),XA(r=$A())&&(n._contentFooterRowDefs=r)}},hostAttrs:[1,"cdk-table"],hostVars:2,hostBindings:function(i,n){i&2&&ue("cdk-table-fixed-layout",n.fixedLayout)},inputs:{trackBy:"trackBy",dataSource:"dataSource",multiTemplateDataRows:[2,"multiTemplateDataRows","multiTemplateDataRows",ie],fixedLayout:[2,"fixedLayout","fixedLayout",ie]},outputs:{contentChanged:"contentChanged"},exportAs:["cdkTable"],features:[ht([{provide:Yl,useExisting:t},{provide:Ru,useClass:RB},{provide:Vy,useClass:uG},{provide:Zy,useValue:null}])],ngContentSelectors:zTA,decls:5,vars:2,consts:[["role","rowgroup"],["headerRowOutlet",""],["rowOutlet",""],["noDataRowOutlet",""],["footerRowOutlet",""]],template:function(i,n){i&1&&(jt(HTA),Fe(0),Fe(1,1),_A(2,OTA,1,0)(3,PTA,7,0)(4,jTA,4,0)),i&2&&(G(2),GA(n._isServer?2:-1),G(),GA(n._isNativeHtmlTable?3:4))},dependencies:[yG,DG,bG,vG],styles:[".cdk-table-fixed-layout{table-layout:fixed}"],encapsulation:2})}return t})();function Py(t,e){return t.concat(Array.from(e))}function OaA(t,e){let A=e.toUpperCase(),i=t.viewContainer.element.nativeElement;for(;i;){let n=i.nodeType===1?i.nodeName:null;if(n===A)return i;if(n==="TABLE")break;i=i.parentNode}return null}var ZaA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[xu]})}return t})();var VTA=[[["caption"]],[["colgroup"],["col"]],"*"],ZTA=["caption","colgroup, col","*"];function WTA(t,e){t&1&&Fe(0,2)}function XTA(t,e){t&1&&(S(0,"thead",0),_r(1,1),F(),S(2,"tbody",2),_r(3,3)(4,4),F(),S(5,"tfoot",0),_r(6,5),F())}function $TA(t,e){t&1&&_r(0,1)(1,3)(2,4)(3,5)}var WaA=(()=>{class t extends MG{stickyCssClass="mat-mdc-table-sticky";needsPositionStickyOnElement=!1;static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-table"],["table","mat-table",""]],hostAttrs:[1,"mat-mdc-table","mdc-data-table__table"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mdc-table-fixed-layout",n.fixedLayout)},exportAs:["matTable"],features:[ht([{provide:MG,useExisting:t},{provide:Yl,useExisting:t},{provide:Vy,useClass:uG},{provide:Ru,useClass:RB},{provide:Zy,useValue:null}]),lt],ngContentSelectors:ZTA,decls:5,vars:2,consts:[["role","rowgroup"],["headerRowOutlet",""],["role","rowgroup",1,"mdc-data-table__content"],["rowOutlet",""],["noDataRowOutlet",""],["footerRowOutlet",""]],template:function(i,n){i&1&&(jt(VTA),Fe(0),Fe(1,1),_A(2,WTA,1,0)(3,XTA,7,0)(4,$TA,4,0)),i&2&&(G(2),GA(n._isServer?2:-1),G(),GA(n._isNativeHtmlTable?3:4))},dependencies:[yG,DG,bG,vG],styles:[".mat-mdc-table-sticky{position:sticky !important}mat-table{display:block}mat-header-row{min-height:56px}mat-row,mat-footer-row{min-height:48px}mat-row,mat-header-row,mat-footer-row{display:flex;border-width:0;border-bottom-width:1px;border-style:solid;align-items:center;box-sizing:border-box}mat-cell:first-of-type,mat-header-cell:first-of-type,mat-footer-cell:first-of-type{padding-left:24px}[dir=rtl] mat-cell:first-of-type:not(:only-of-type),[dir=rtl] mat-header-cell:first-of-type:not(:only-of-type),[dir=rtl] mat-footer-cell:first-of-type:not(:only-of-type){padding-left:0;padding-right:24px}mat-cell:last-of-type,mat-header-cell:last-of-type,mat-footer-cell:last-of-type{padding-right:24px}[dir=rtl] mat-cell:last-of-type:not(:only-of-type),[dir=rtl] mat-header-cell:last-of-type:not(:only-of-type),[dir=rtl] mat-footer-cell:last-of-type:not(:only-of-type){padding-right:0;padding-left:24px}mat-cell,mat-header-cell,mat-footer-cell{flex:1;display:flex;align-items:center;overflow:hidden;word-wrap:break-word;min-height:inherit}.mat-mdc-table{min-width:100%;border:0;border-spacing:0;table-layout:auto;white-space:normal;background-color:var(--mat-table-background-color, var(--mat-sys-surface))}.mdc-data-table__cell{box-sizing:border-box;overflow:hidden;text-align:left;text-overflow:ellipsis}[dir=rtl] .mdc-data-table__cell{text-align:right}.mdc-data-table__cell,.mdc-data-table__header-cell{padding:0 16px}.mat-mdc-header-row{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;height:var(--mat-table-header-container-height, 56px);color:var(--mat-table-header-headline-color, var(--mat-sys-on-surface, rgba(0, 0, 0, 0.87)));font-family:var(--mat-table-header-headline-font, var(--mat-sys-title-small-font, Roboto, sans-serif));line-height:var(--mat-table-header-headline-line-height, var(--mat-sys-title-small-line-height));font-size:var(--mat-table-header-headline-size, var(--mat-sys-title-small-size, 14px));font-weight:var(--mat-table-header-headline-weight, var(--mat-sys-title-small-weight, 500))}.mat-mdc-row{height:var(--mat-table-row-item-container-height, 52px);color:var(--mat-table-row-item-label-text-color, var(--mat-sys-on-surface, rgba(0, 0, 0, 0.87)))}.mat-mdc-row,.mdc-data-table__content{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;font-family:var(--mat-table-row-item-label-text-font, var(--mat-sys-body-medium-font, Roboto, sans-serif));line-height:var(--mat-table-row-item-label-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mat-table-row-item-label-text-size, var(--mat-sys-body-medium-size, 14px));font-weight:var(--mat-table-row-item-label-text-weight, var(--mat-sys-body-medium-weight))}.mat-mdc-footer-row{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;height:var(--mat-table-footer-container-height, 52px);color:var(--mat-table-row-item-label-text-color, var(--mat-sys-on-surface, rgba(0, 0, 0, 0.87)));font-family:var(--mat-table-footer-supporting-text-font, var(--mat-sys-body-medium-font, Roboto, sans-serif));line-height:var(--mat-table-footer-supporting-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mat-table-footer-supporting-text-size, var(--mat-sys-body-medium-size, 14px));font-weight:var(--mat-table-footer-supporting-text-weight, var(--mat-sys-body-medium-weight));letter-spacing:var(--mat-table-footer-supporting-text-tracking, var(--mat-sys-body-medium-tracking))}.mat-mdc-header-cell{border-bottom-color:var(--mat-table-row-item-outline-color, var(--mat-sys-outline, rgba(0, 0, 0, 0.12)));border-bottom-width:var(--mat-table-row-item-outline-width, 1px);border-bottom-style:solid;letter-spacing:var(--mat-table-header-headline-tracking, var(--mat-sys-title-small-tracking));font-weight:inherit;line-height:inherit;box-sizing:border-box;text-overflow:ellipsis;overflow:hidden;outline:none;text-align:left}[dir=rtl] .mat-mdc-header-cell{text-align:right}.mdc-data-table__row:last-child>.mat-mdc-header-cell{border-bottom:none}.mat-mdc-cell{border-bottom-color:var(--mat-table-row-item-outline-color, var(--mat-sys-outline, rgba(0, 0, 0, 0.12)));border-bottom-width:var(--mat-table-row-item-outline-width, 1px);border-bottom-style:solid;letter-spacing:var(--mat-table-row-item-label-text-tracking, var(--mat-sys-body-medium-tracking));line-height:inherit}.mdc-data-table__row:last-child>.mat-mdc-cell{border-bottom:none}.mat-mdc-footer-cell{letter-spacing:var(--mat-table-row-item-label-text-tracking, var(--mat-sys-body-medium-tracking))}mat-row.mat-mdc-row,mat-header-row.mat-mdc-header-row,mat-footer-row.mat-mdc-footer-row{border-bottom:none}.mat-mdc-table tbody,.mat-mdc-table tfoot,.mat-mdc-table thead,.mat-mdc-cell,.mat-mdc-footer-cell,.mat-mdc-header-row,.mat-mdc-row,.mat-mdc-footer-row,.mat-mdc-table .mat-mdc-header-cell{background:inherit}.mat-mdc-table mat-header-row.mat-mdc-header-row,.mat-mdc-table mat-row.mat-mdc-row,.mat-mdc-table mat-footer-row.mat-mdc-footer-cell{height:unset}mat-header-cell.mat-mdc-header-cell,mat-cell.mat-mdc-cell,mat-footer-cell.mat-mdc-footer-cell{align-self:stretch}"],encapsulation:2})}return t})(),XaA=(()=>{class t extends Wy{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matCellDef",""]],features:[ht([{provide:Wy,useExisting:t}]),lt]})}return t})(),$aA=(()=>{class t extends Xy{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matHeaderCellDef",""]],features:[ht([{provide:Xy,useExisting:t}]),lt]})}return t})();var AcA=(()=>{class t extends MQ{get name(){return this._name}set name(A){this._setNameInput(A)}_updateColumnCssClassName(){super._updateColumnCssClassName(),this._columnCssClassName.push(`mat-column-${this.cssClassFriendlyName}`)}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matColumnDef",""]],inputs:{name:[0,"matColumnDef","name"]},features:[ht([{provide:MQ,useExisting:t},{provide:"MAT_SORT_HEADER_COLUMN_DEF",useExisting:t}]),lt]})}return t})(),ecA=(()=>{class t extends jaA{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["mat-header-cell"],["th","mat-header-cell",""]],hostAttrs:["role","columnheader",1,"mat-mdc-header-cell","mdc-data-table__header-cell"],features:[lt]})}return t})();var tcA=(()=>{class t extends qaA{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["mat-cell"],["td","mat-cell",""]],hostAttrs:[1,"mat-mdc-cell","mdc-data-table__cell"],features:[lt]})}return t})();var icA=(()=>{class t extends If{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matHeaderRowDef",""]],inputs:{columns:[0,"matHeaderRowDef","columns"],sticky:[2,"matHeaderRowDefSticky","sticky",ie]},features:[ht([{provide:If,useExisting:t}]),lt]})}return t})();var ncA=(()=>{class t extends $y{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matRowDef",""]],inputs:{columns:[0,"matRowDefColumns","columns"],when:[0,"matRowDefWhen","when"]},features:[ht([{provide:$y,useExisting:t}]),lt]})}return t})(),ocA=(()=>{class t extends pG{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-header-row"],["tr","mat-header-row",""]],hostAttrs:["role","row",1,"mat-mdc-header-row","mdc-data-table__header-row"],exportAs:["matHeaderRow"],features:[ht([{provide:pG,useExisting:t}]),lt],decls:1,vars:0,consts:[["cdkCellOutlet",""]],template:function(i,n){i&1&&_r(0,0)},dependencies:[td],encapsulation:2})}return t})();var rcA=(()=>{class t extends wG{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-row"],["tr","mat-row",""]],hostAttrs:["role","row",1,"mat-mdc-row","mdc-data-table__row"],exportAs:["matRow"],features:[ht([{provide:wG,useExisting:t}]),lt],decls:1,vars:0,consts:[["cdkCellOutlet",""]],template:function(i,n){i&1&&_r(0,0)},dependencies:[td],encapsulation:2})}return t})();var scA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,ZaA,it]})}return t})(),AHA=9007199254740991,Cf=class extends g8{_data;_renderData=new Mi([]);_filter=new Mi("");_internalPageChanges=new OA;_renderChangesSubscription=null;filteredData;get data(){return this._data.value}set data(e){e=Array.isArray(e)?e:[],this._data.next(e),this._renderChangesSubscription||this._filterData(e)}get filter(){return this._filter.value}set filter(e){this._filter.next(e),this._renderChangesSubscription||this._filterData(this.data)}get sort(){return this._sort}set sort(e){this._sort=e,this._updateChangeSubscription()}_sort;get paginator(){return this._paginator}set paginator(e){this._paginator=e,this._updateChangeSubscription()}_paginator;sortingDataAccessor=(e,A)=>{let i=e[A];if(ak(i)){let n=Number(i);return n<AHA?n:i}return i};sortData=(e,A)=>{let i=A.active,n=A.direction;return!i||n==""?e:e.sort((o,r)=>{let s=this.sortingDataAccessor(o,i),a=this.sortingDataAccessor(r,i),c=typeof s,l=typeof a;c!==l&&(c==="number"&&(s+=""),l==="number"&&(a+=""));let I=0;return s!=null&&a!=null?s>a?I=1:s<a&&(I=-1):s!=null?I=1:a!=null&&(I=-1),I*(n=="asc"?1:-1)})};filterPredicate=(e,A)=>{let i=A.trim().toLowerCase();return Object.values(e).some(n=>`${n}`.toLowerCase().includes(i))};constructor(e=[]){super(),this._data=new Mi(e),this._updateChangeSubscription()}_updateChangeSubscription(){let e=this._sort?zn(this._sort.sortChange,this._sort.initialized):Me(null),A=this._paginator?zn(this._paginator.page,this._internalPageChanges,this._paginator.initialized):Me(null),i=this._data,n=Js([i,this._filter]).pipe(je(([s])=>this._filterData(s))),o=Js([n,e]).pipe(je(([s])=>this._orderData(s))),r=Js([o,A]).pipe(je(([s])=>this._pageData(s)));this._renderChangesSubscription?.unsubscribe(),this._renderChangesSubscription=r.subscribe(s=>this._renderData.next(s))}_filterData(e){return this.filteredData=this.filter==null||this.filter===""?e:e.filter(A=>this.filterPredicate(A,this.filter)),this.paginator&&this._updatePaginator(this.filteredData.length),this.filteredData}_orderData(e){return this.sort?this.sortData(e.slice(),this.sort):e}_pageData(e){if(!this.paginator)return e;let A=this.paginator.pageIndex*this.paginator.pageSize;return e.slice(A,A+this.paginator.pageSize)}_updatePaginator(e){Promise.resolve().then(()=>{let A=this.paginator;if(A&&(A.length=e,A.pageIndex>0)){let i=Math.ceil(A.length/A.pageSize)-1||0,n=Math.min(A.pageIndex,i);n!==A.pageIndex&&(A.pageIndex=n,this._internalPageChanges.next())}})}connect(){return this._renderChangesSubscription||this._updateChangeSubscription(),this._renderData}disconnect(){this._renderChangesSubscription?.unsubscribe(),this._renderChangesSubscription=null}};var acA=[{metricName:"tool_trajectory_avg_score",threshold:1},{metricName:"response_match_score",threshold:.7}];var us=[];for(let t=0;t<256;++t)us.push((t+256).toString(16).slice(1));function ccA(t,e=0){return(us[t[e+0]]+us[t[e+1]]+us[t[e+2]]+us[t[e+3]]+"-"+us[t[e+4]]+us[t[e+5]]+"-"+us[t[e+6]]+us[t[e+7]]+"-"+us[t[e+8]]+us[t[e+9]]+"-"+us[t[e+10]]+us[t[e+11]]+us[t[e+12]]+us[t[e+13]]+us[t[e+14]]+us[t[e+15]]).toLowerCase()}var kG,tHA=new Uint8Array(16);function SG(){if(!kG){if(typeof crypto>"u"||!crypto.getRandomValues)throw new Error("crypto.getRandomValues() not supported. See https://github.com/uuidjs/uuid#getrandomvalues-not-supported");kG=crypto.getRandomValues.bind(crypto)}return kG(tHA)}var iHA=typeof crypto<"u"&&crypto.randomUUID&&crypto.randomUUID.bind(crypto),RG={randomUUID:iHA};function nHA(t,e,A){if(RG.randomUUID&&!e&&!t)return RG.randomUUID();t=t||{};let i=t.random??t.rng?.()??SG();if(i.length<16)throw new Error("Random bytes length must be >= 16");if(i[6]=i[6]&15|64,i[8]=i[8]&63|128,e){if(A=A||0,A<0||A+16>e.length)throw new RangeError(`UUID byte range ${A}:${A+15} is out of buffer bounds`);for(let n=0;n<16;++n)e[A+n]=i[n];return e}return ccA(i)}var df=nHA;var Vc=class t{constructor(e){this.http=e}apiServerDomain=vs.getApiServerBaseUrl();getEvalSets(e){if(this.apiServerDomain!=null){let A=this.apiServerDomain+`/apps/${e}/eval_sets`;return this.http.get(A)}return new ct}createNewEvalSet(e,A){if(this.apiServerDomain!=null){let i=this.apiServerDomain+`/apps/${e}/eval_sets/${A}`;return this.http.post(i,{})}return new ct}listEvalCases(e,A){if(this.apiServerDomain!=null){let i=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/evals`;return this.http.get(i,{})}return new ct}addCurrentSession(e,A,i,n,o){let r=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/add_session`;return this.http.post(r,{evalId:i,sessionId:n,userId:o})}runEval(e,A,i,n){let o=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/run_eval`;return this.http.post(o,{evalIds:i,evalMetrics:n})}listEvalResults(e){if(this.apiServerDomain!=null){let A=this.apiServerDomain+`/apps/${e}/eval_results`;return this.http.get(A,{})}return new ct}getEvalResult(e,A){if(this.apiServerDomain!=null){let i=this.apiServerDomain+`/apps/${e}/eval_results/${A}`;return this.http.get(i,{})}return new ct}getEvalCase(e,A,i){if(this.apiServerDomain!=null){let n=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/evals/${i}`;return this.http.get(n,{})}return new ct}updateEvalCase(e,A,i,n){let o=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/evals/${i}`;return this.http.put(o,{evalId:i,conversation:n.conversation,sessionInput:n.sessionInput,creationTimestamp:n.creationTimestamp})}deleteEvalCase(e,A,i){let n=this.apiServerDomain+`/apps/${e}/eval_sets/${A}/evals/${i}`;return this.http.delete(n,{})}static \u0275fac=function(A){return new(A||t)(we(Ds))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var gcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["ng-component"]],hostAttrs:["cdk-text-field-style-loader",""],decls:0,vars:0,template:function(i,n){},styles:["textarea.cdk-textarea-autosize{resize:none}textarea.cdk-textarea-autosize-measuring{padding:2px 0 !important;box-sizing:content-box !important;height:auto !important;overflow:hidden !important}textarea.cdk-textarea-autosize-measuring-firefox{padding:2px 0 !important;box-sizing:content-box !important;height:0 !important}@keyframes cdk-text-field-autofill-start{/*!*/}@keyframes cdk-text-field-autofill-end{/*!*/}.cdk-text-field-autofill-monitored:-webkit-autofill{animation:cdk-text-field-autofill-start 0s 1ms}.cdk-text-field-autofill-monitored:not(:-webkit-autofill){animation:cdk-text-field-autofill-end 0s 1ms}"],encapsulation:2,changeDetection:0})}return t})(),lcA=bc({passive:!0}),IcA=(()=>{class t{_platform=f(Ii);_ngZone=f(Qe);_styleLoader=f(Rn);_monitoredElements=new Map;constructor(){}monitor(A){if(!this._platform.isBrowser)return sr;this._styleLoader.load(gcA);let i=ua(A),n=this._monitoredElements.get(i);if(n)return n.subject;let o=new OA,r="cdk-text-field-autofilled",s=a=>{a.animationName==="cdk-text-field-autofill-start"&&!i.classList.contains(r)?(i.classList.add(r),this._ngZone.run(()=>o.next({target:a.target,isAutofilled:!0}))):a.animationName==="cdk-text-field-autofill-end"&&i.classList.contains(r)&&(i.classList.remove(r),this._ngZone.run(()=>o.next({target:a.target,isAutofilled:!1})))};return this._ngZone.runOutsideAngular(()=>{i.addEventListener("animationstart",s,lcA),i.classList.add("cdk-text-field-autofill-monitored")}),this._monitoredElements.set(i,{subject:o,unlisten:()=>{i.removeEventListener("animationstart",s,lcA)}}),o}stopMonitoring(A){let i=ua(A),n=this._monitoredElements.get(i);n&&(n.unlisten(),n.subject.complete(),i.classList.remove("cdk-text-field-autofill-monitored"),i.classList.remove("cdk-text-field-autofilled"),this._monitoredElements.delete(i))}ngOnDestroy(){this._monitoredElements.forEach((A,i)=>this.stopMonitoring(i))}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})}return t})();var CcA=(()=>{class t{_elementRef=f(ee);_platform=f(Ii);_ngZone=f(Qe);_renderer=f(Wi);_resizeEvents=new OA;_previousValue;_initialHeight;_destroyed=new OA;_listenerCleanups;_minRows;_maxRows;_enabled=!0;_previousMinRows=-1;_textareaElement;get minRows(){return this._minRows}set minRows(A){this._minRows=zs(A),this._setMinHeight()}get maxRows(){return this._maxRows}set maxRows(A){this._maxRows=zs(A),this._setMaxHeight()}get enabled(){return this._enabled}set enabled(A){this._enabled!==A&&((this._enabled=A)?this.resizeToFitContent(!0):this.reset())}get placeholder(){return this._textareaElement.placeholder}set placeholder(A){this._cachedPlaceholderHeight=void 0,A?this._textareaElement.setAttribute("placeholder",A):this._textareaElement.removeAttribute("placeholder"),this._cacheTextareaPlaceholderHeight()}_cachedLineHeight;_cachedPlaceholderHeight;_document=f(at,{optional:!0});_hasFocus;_isViewInited=!1;constructor(){f(Rn).load(gcA),this._textareaElement=this._elementRef.nativeElement}_setMinHeight(){let A=this.minRows&&this._cachedLineHeight?`${this.minRows*this._cachedLineHeight}px`:null;A&&(this._textareaElement.style.minHeight=A)}_setMaxHeight(){let A=this.maxRows&&this._cachedLineHeight?`${this.maxRows*this._cachedLineHeight}px`:null;A&&(this._textareaElement.style.maxHeight=A)}ngAfterViewInit(){this._platform.isBrowser&&(this._initialHeight=this._textareaElement.style.height,this.resizeToFitContent(),this._ngZone.runOutsideAngular(()=>{this._listenerCleanups=[this._renderer.listen("window","resize",()=>this._resizeEvents.next()),this._renderer.listen(this._textareaElement,"focus",this._handleFocusEvent),this._renderer.listen(this._textareaElement,"blur",this._handleFocusEvent)],this._resizeEvents.pipe(yd(16)).subscribe(()=>{this._cachedLineHeight=this._cachedPlaceholderHeight=void 0,this.resizeToFitContent(!0)})}),this._isViewInited=!0,this.resizeToFitContent(!0))}ngOnDestroy(){this._listenerCleanups?.forEach(A=>A()),this._resizeEvents.complete(),this._destroyed.next(),this._destroyed.complete()}_cacheTextareaLineHeight(){if(this._cachedLineHeight)return;let A=this._textareaElement.cloneNode(!1),i=A.style;A.rows=1,i.position="absolute",i.visibility="hidden",i.border="none",i.padding="0",i.height="",i.minHeight="",i.maxHeight="",i.top=i.bottom=i.left=i.right="auto",i.overflow="hidden",this._textareaElement.parentNode.appendChild(A),this._cachedLineHeight=A.clientHeight,A.remove(),this._setMinHeight(),this._setMaxHeight()}_measureScrollHeight(){let A=this._textareaElement,i=A.style.marginBottom||"",n=this._platform.FIREFOX,o=n&&this._hasFocus,r=n?"cdk-textarea-autosize-measuring-firefox":"cdk-textarea-autosize-measuring";o&&(A.style.marginBottom=`${A.clientHeight}px`),A.classList.add(r);let s=A.scrollHeight-4;return A.classList.remove(r),o&&(A.style.marginBottom=i),s}_cacheTextareaPlaceholderHeight(){if(!this._isViewInited||this._cachedPlaceholderHeight!=null)return;if(!this.placeholder){this._cachedPlaceholderHeight=0;return}let A=this._textareaElement.value;this._textareaElement.value=this._textareaElement.placeholder,this._cachedPlaceholderHeight=this._measureScrollHeight(),this._textareaElement.value=A}_handleFocusEvent=A=>{this._hasFocus=A.type==="focus"};ngDoCheck(){this._platform.isBrowser&&this.resizeToFitContent()}resizeToFitContent(A=!1){if(!this._enabled||(this._cacheTextareaLineHeight(),this._cacheTextareaPlaceholderHeight(),!this._cachedLineHeight))return;let i=this._elementRef.nativeElement,n=i.value;if(!A&&this._minRows===this._previousMinRows&&n===this._previousValue)return;let o=this._measureScrollHeight(),r=Math.max(o,this._cachedPlaceholderHeight||0);i.style.height=`${r}px`,this._ngZone.runOutsideAngular(()=>{typeof requestAnimationFrame<"u"?requestAnimationFrame(()=>this._scrollToCaretPosition(i)):setTimeout(()=>this._scrollToCaretPosition(i))}),this._previousValue=n,this._previousMinRows=this._minRows}reset(){this._initialHeight!==void 0&&(this._textareaElement.style.height=this._initialHeight)}_noopInputHandler(){}_scrollToCaretPosition(A){let{selectionStart:i,selectionEnd:n}=A;!this._destroyed.isStopped&&this._hasFocus&&A.setSelectionRange(i,n)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["textarea","cdkTextareaAutosize",""]],hostAttrs:["rows","1",1,"cdk-textarea-autosize"],hostBindings:function(i,n){i&1&&hA("input",function(){return n._noopInputHandler()})},inputs:{minRows:[0,"cdkAutosizeMinRows","minRows"],maxRows:[0,"cdkAutosizeMaxRows","maxRows"],enabled:[2,"cdkTextareaAutosize","enabled",ie],placeholder:"placeholder"},exportAs:["cdkTextareaAutosize"]})}return t})(),dcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();var rHA=new dA("MAT_INPUT_VALUE_ACCESSOR"),sHA=["button","checkbox","file","hidden","image","radio","range","reset","submit"],aHA=new dA("MAT_INPUT_CONFIG"),iI=(()=>{class t{_elementRef=f(ee);_platform=f(Ii);ngControl=f(Ac,{optional:!0,self:!0});_autofillMonitor=f(IcA);_ngZone=f(Qe);_formField=f(Gu,{optional:!0});_renderer=f(Wi);_uid=f(sn).getId("mat-input-");_previousNativeValue;_inputValueAccessor;_signalBasedValueAccessor;_previousPlaceholder;_errorStateTracker;_config=f(aHA,{optional:!0});_cleanupIosKeyup;_cleanupWebkitWheel;_formFieldDescribedBy;_isServer;_isNativeSelect;_isTextarea;_isInFormField;focused=!1;stateChanges=new OA;controlType="mat-input";autofilled=!1;get disabled(){return this._disabled}set disabled(A){this._disabled=Xo(A),this.focused&&(this.focused=!1,this.stateChanges.next())}_disabled=!1;get id(){return this._id}set id(A){this._id=A||this._uid}_id;placeholder;name;get required(){return this._required??this.ngControl?.control?.hasValidator($a.required)??!1}set required(A){this._required=Xo(A)}_required;get type(){return this._type}set type(A){let i=this._type;this._type=A||"text",this._validateType(),!this._isTextarea&&ok().has(this._type)&&(this._elementRef.nativeElement.type=this._type),this._type!==i&&this._ensureWheelDefaultBehavior()}_type="text";get errorStateMatcher(){return this._errorStateTracker.matcher}set errorStateMatcher(A){this._errorStateTracker.matcher=A}userAriaDescribedBy;get value(){return this._signalBasedValueAccessor?this._signalBasedValueAccessor.value():this._inputValueAccessor.value}set value(A){A!==this.value&&(this._signalBasedValueAccessor?this._signalBasedValueAccessor.value.set(A):this._inputValueAccessor.value=A,this.stateChanges.next())}get readonly(){return this._readonly}set readonly(A){this._readonly=Xo(A)}_readonly=!1;disabledInteractive;get errorState(){return this._errorStateTracker.errorState}set errorState(A){this._errorStateTracker.errorState=A}_neverEmptyInputTypes=["date","datetime","datetime-local","month","time","week"].filter(A=>ok().has(A));constructor(){let A=f(su,{optional:!0}),i=f(GI,{optional:!0}),n=f(bB),o=f(rHA,{optional:!0,self:!0}),r=this._elementRef.nativeElement,s=r.nodeName.toLowerCase();o?p2(o.value)?this._signalBasedValueAccessor=o:this._inputValueAccessor=o:this._inputValueAccessor=r,this._previousNativeValue=this.value,this.id=this.id,this._platform.IOS&&this._ngZone.runOutsideAngular(()=>{this._cleanupIosKeyup=this._renderer.listen(r,"keyup",this._iOSKeyupListener)}),this._errorStateTracker=new $I(n,this.ngControl,i,A,this.stateChanges),this._isServer=!this._platform.isBrowser,this._isNativeSelect=s==="select",this._isTextarea=s==="textarea",this._isInFormField=!!this._formField,this.disabledInteractive=this._config?.disabledInteractive||!1,this._isNativeSelect&&(this.controlType=r.multiple?"mat-native-select-multiple":"mat-native-select"),this._signalBasedValueAccessor&&Nh(()=>{this._signalBasedValueAccessor.value(),this.stateChanges.next()})}ngAfterViewInit(){this._platform.isBrowser&&this._autofillMonitor.monitor(this._elementRef.nativeElement).subscribe(A=>{this.autofilled=A.isAutofilled,this.stateChanges.next()})}ngOnChanges(){this.stateChanges.next()}ngOnDestroy(){this.stateChanges.complete(),this._platform.isBrowser&&this._autofillMonitor.stopMonitoring(this._elementRef.nativeElement),this._cleanupIosKeyup?.(),this._cleanupWebkitWheel?.()}ngDoCheck(){this.ngControl&&(this.updateErrorState(),this.ngControl.disabled!==null&&this.ngControl.disabled!==this.disabled&&(this.disabled=this.ngControl.disabled,this.stateChanges.next())),this._dirtyCheckNativeValue(),this._dirtyCheckPlaceholder()}focus(A){this._elementRef.nativeElement.focus(A)}updateErrorState(){this._errorStateTracker.updateErrorState()}_focusChanged(A){if(A!==this.focused){if(!this._isNativeSelect&&A&&this.disabled&&this.disabledInteractive){let i=this._elementRef.nativeElement;i.type==="number"?(i.type="text",i.setSelectionRange(0,0),i.type="number"):i.setSelectionRange(0,0)}this.focused=A,this.stateChanges.next()}}_onInput(){}_dirtyCheckNativeValue(){let A=this._elementRef.nativeElement.value;this._previousNativeValue!==A&&(this._previousNativeValue=A,this.stateChanges.next())}_dirtyCheckPlaceholder(){let A=this._getPlaceholder();if(A!==this._previousPlaceholder){let i=this._elementRef.nativeElement;this._previousPlaceholder=A,A?i.setAttribute("placeholder",A):i.removeAttribute("placeholder")}}_getPlaceholder(){return this.placeholder||null}_validateType(){sHA.indexOf(this._type)>-1}_isNeverEmpty(){return this._neverEmptyInputTypes.indexOf(this._type)>-1}_isBadInput(){let A=this._elementRef.nativeElement.validity;return A&&A.badInput}get empty(){return!this._isNeverEmpty()&&!this._elementRef.nativeElement.value&&!this._isBadInput()&&!this.autofilled}get shouldLabelFloat(){if(this._isNativeSelect){let A=this._elementRef.nativeElement,i=A.options[0];return this.focused||A.multiple||!this.empty||!!(A.selectedIndex>-1&&i&&i.label)}else return this.focused&&!this.disabled||!this.empty}setDescribedByIds(A){let i=this._elementRef.nativeElement,n=i.getAttribute("aria-describedby"),o;if(n){let r=this._formFieldDescribedBy||A;o=A.concat(n.split(" ").filter(s=>s&&!r.includes(s)))}else o=A;this._formFieldDescribedBy=A,o.length?i.setAttribute("aria-describedby",o.join(" ")):i.removeAttribute("aria-describedby")}onContainerClick(){this.focused||this.focus()}_isInlineSelect(){let A=this._elementRef.nativeElement;return this._isNativeSelect&&(A.multiple||A.size>1)}_iOSKeyupListener=A=>{let i=A.target;!i.value&&i.selectionStart===0&&i.selectionEnd===0&&(i.setSelectionRange(1,1),i.setSelectionRange(0,0))};_webkitBlinkWheelListener=()=>{};_ensureWheelDefaultBehavior(){this._cleanupWebkitWheel?.(),this._type==="number"&&(this._platform.BLINK||this._platform.WEBKIT)&&(this._cleanupWebkitWheel=this._renderer.listen(this._elementRef.nativeElement,"wheel",this._webkitBlinkWheelListener))}_getReadonlyAttribute(){return this._isNativeSelect?null:this.readonly||this.disabled&&this.disabledInteractive?"true":null}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["input","matInput",""],["textarea","matInput",""],["select","matNativeControl",""],["input","matNativeControl",""],["textarea","matNativeControl",""]],hostAttrs:[1,"mat-mdc-input-element"],hostVars:21,hostBindings:function(i,n){i&1&&hA("focus",function(){return n._focusChanged(!0)})("blur",function(){return n._focusChanged(!1)})("input",function(){return n._onInput()}),i&2&&(Hs("id",n.id)("disabled",n.disabled&&!n.disabledInteractive)("required",n.required),Ne("name",n.name||null)("readonly",n._getReadonlyAttribute())("aria-disabled",n.disabled&&n.disabledInteractive?"true":null)("aria-invalid",n.empty&&n.required?null:n.errorState)("aria-required",n.required)("id",n.id),ue("mat-input-server",n._isServer)("mat-mdc-form-field-textarea-control",n._isInFormField&&n._isTextarea)("mat-mdc-form-field-input-control",n._isInFormField)("mat-mdc-input-disabled-interactive",n.disabledInteractive)("mdc-text-field__input",n._isInFormField)("mat-mdc-native-select-inline",n._isInlineSelect()))},inputs:{disabled:"disabled",id:"id",placeholder:"placeholder",name:"name",required:"required",type:"type",errorStateMatcher:"errorStateMatcher",userAriaDescribedBy:[0,"aria-describedby","userAriaDescribedBy"],value:"value",readonly:"readonly",disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie]},exportAs:["matInput"],features:[ht([{provide:_u,useExisting:t}]),ti]})}return t})(),e7=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,y0,y0,dcA,it]})}return t})();var Bf=class t{constructor(e,A,i){this.evalService=e;this.data=A;this.dialogRef=i}newCaseId="case"+df().slice(0,6);createNewEvalCase(){!this.newCaseId||this.newCaseId==""?alert("Cannot create eval set with empty id!"):this.evalService.addCurrentSession(this.data.appName,this.data.evalSetId,this.newCaseId,this.data.sessionId,this.data.userId).subscribe(e=>{this.dialogRef.close(!0)})}static \u0275fac=function(A){return new(A||t)(PA(Vc),PA(as),PA(Br))};static \u0275cmp=zA({type:t,selectors:[["app-add-eval-session-dialog"]],standalone:!1,decls:11,vars:1,consts:[["mat-dialog-title",""],[2,"padding-left","20px","padding-right","24px"],["matInput","",3,"ngModelChange","keydown.enter","ngModel"],["align","end"],["mat-button","","mat-dialog-close",""],["mat-button","","cdkFocusInitial","",3,"click"]],template:function(A,i){A&1&&(S(0,"h2",0),AA(1,"Add Current Session To Eval Set"),F(),S(2,"mat-dialog-content"),AA(3,` Please enter the eval case name
`),F(),S(4,"mat-form-field",1)(5,"input",2),da("ngModelChange",function(o){return Va(i.newCaseId,o)||(i.newCaseId=o),o}),hA("keydown.enter",function(){return i.createNewEvalCase()}),F()(),S(6,"mat-dialog-actions",3)(7,"button",4),AA(8,"Cancel"),F(),S(9,"button",5),hA("click",function(){return i.createNewEvalCase()}),AA(10,"Create"),F()()),A&2&&(G(5),Ca("ngModel",i.newCaseId))},dependencies:[vc,Ea,ec,Qg,iI,Dr,bs,ma,pa,hg],encapsulation:2})};var Ef=class t{constructor(e,A,i){this.evalService=e;this.data=A;this.dialogRef=i}newSetId="evalset"+df().slice(0,6);createNewEvalSet(){!this.newSetId||this.newSetId==""?alert("Cannot create eval set with empty id!"):this.evalService.createNewEvalSet(this.data.appName,this.newSetId).subscribe(e=>{this.dialogRef.close(!0)})}static \u0275fac=function(A){return new(A||t)(PA(Vc),PA(as),PA(Br))};static \u0275cmp=zA({type:t,selectors:[["app-new-eval-set-dialog-component"]],standalone:!1,decls:11,vars:1,consts:[["mat-dialog-title",""],[2,"padding-left","20px","padding-right","24px"],["matInput","",3,"ngModelChange","keydown.enter","ngModel"],["align","end"],["mat-button","","mat-dialog-close",""],["mat-button","","cdkFocusInitial","",3,"click"]],template:function(A,i){A&1&&(S(0,"h2",0),AA(1,"Create New Eval Set"),F(),S(2,"mat-dialog-content"),AA(3,` Please enter the eval set name
`),F(),S(4,"mat-form-field",1)(5,"input",2),da("ngModelChange",function(o){return Va(i.newSetId,o)||(i.newSetId=o),o}),hA("keydown.enter",function(){return i.createNewEvalSet()}),F()(),S(6,"mat-dialog-actions",3)(7,"button",4),AA(8,"Cancel"),F(),S(9,"button",5),hA("click",function(){return i.createNewEvalSet()}),AA(10,"Create"),F()()),A&2&&(G(5),Ca("ngModel",i.newSetId))},dependencies:[vc,Ea,ec,Qg,iI,Dr,bs,ma,pa,hg],encapsulation:2})};var cHA=["knob"],lHA=["valueIndicatorContainer"];function gHA(t,e){if(t&1&&(S(0,"div",2,1)(2,"div",5)(3,"span",6),AA(4),F()()()),t&2){let A=O();G(4),Gt(A.valueIndicatorText)}}var IHA=["trackActive"],CHA=["*"];function dHA(t,e){if(t&1&&JA(0,"div"),t&2){let A=e.$implicit,i=e.$index,n=O(3);fo(A===0?"mdc-slider__tick-mark--active":"mdc-slider__tick-mark--inactive"),uo("transform",n._calcTickMarkTransform(i))}}function BHA(t,e){if(t&1&&Dn(0,dHA,1,4,"div",8,Xd),t&2){let A=O(2);yn(A._tickMarks)}}function EHA(t,e){if(t&1&&(S(0,"div",6,1),_A(2,BHA,2,0),F()),t&2){let A=O();G(2),GA(A._cachedWidth?2:-1)}}function QHA(t,e){if(t&1&&JA(0,"mat-slider-visual-thumb",7),t&2){let A=O();yA("discrete",A.discrete)("thumbPosition",1)("valueIndicatorText",A.startValueIndicatorText)}}var ui=function(t){return t[t.START=1]="START",t[t.END=2]="END",t}(ui||{}),kQ=function(t){return t[t.ACTIVE=0]="ACTIVE",t[t.INACTIVE=1]="INACTIVE",t}(kQ||{}),xG=new dA("_MatSlider"),BcA=new dA("_MatSliderThumb"),hHA=new dA("_MatSliderRangeThumb"),EcA=new dA("_MatSliderVisualThumb");var uHA=(()=>{class t{_cdr=f(It);_ngZone=f(Qe);_slider=f(xG);_renderer=f(Wi);_listenerCleanups;discrete;thumbPosition;valueIndicatorText;_ripple;_knob;_valueIndicatorContainer;_sliderInput;_sliderInputEl;_hoverRippleRef;_focusRippleRef;_activeRippleRef;_isHovered=!1;_isActive=!1;_isValueIndicatorVisible=!1;_hostElement=f(ee).nativeElement;_platform=f(Ii);constructor(){}ngAfterViewInit(){let A=this._slider._getInput(this.thumbPosition);A&&(this._ripple.radius=24,this._sliderInput=A,this._sliderInputEl=this._sliderInput._hostElement,this._ngZone.runOutsideAngular(()=>{let i=this._sliderInputEl,n=this._renderer;this._listenerCleanups=[n.listen(i,"pointermove",this._onPointerMove),n.listen(i,"pointerdown",this._onDragStart),n.listen(i,"pointerup",this._onDragEnd),n.listen(i,"pointerleave",this._onMouseLeave),n.listen(i,"focus",this._onFocus),n.listen(i,"blur",this._onBlur)]}))}ngOnDestroy(){this._listenerCleanups?.forEach(A=>A())}_onPointerMove=A=>{if(this._sliderInput._isFocused)return;let i=this._hostElement.getBoundingClientRect(),n=this._slider._isCursorOnSliderThumb(A,i);this._isHovered=n,n?this._showHoverRipple():this._hideRipple(this._hoverRippleRef)};_onMouseLeave=()=>{this._isHovered=!1,this._hideRipple(this._hoverRippleRef)};_onFocus=()=>{this._hideRipple(this._hoverRippleRef),this._showFocusRipple(),this._hostElement.classList.add("mdc-slider__thumb--focused")};_onBlur=()=>{this._isActive||this._hideRipple(this._focusRippleRef),this._isHovered&&this._showHoverRipple(),this._hostElement.classList.remove("mdc-slider__thumb--focused")};_onDragStart=A=>{A.button===0&&(this._isActive=!0,this._showActiveRipple())};_onDragEnd=()=>{this._isActive=!1,this._hideRipple(this._activeRippleRef),this._sliderInput._isFocused||this._hideRipple(this._focusRippleRef),this._platform.SAFARI&&this._showHoverRipple()};_showHoverRipple(){this._isShowingRipple(this._hoverRippleRef)||(this._hoverRippleRef=this._showRipple({enterDuration:0,exitDuration:0}),this._hoverRippleRef?.element.classList.add("mat-mdc-slider-hover-ripple"))}_showFocusRipple(){this._isShowingRipple(this._focusRippleRef)||(this._focusRippleRef=this._showRipple({enterDuration:0,exitDuration:0},!0),this._focusRippleRef?.element.classList.add("mat-mdc-slider-focus-ripple"))}_showActiveRipple(){this._isShowingRipple(this._activeRippleRef)||(this._activeRippleRef=this._showRipple({enterDuration:225,exitDuration:400}),this._activeRippleRef?.element.classList.add("mat-mdc-slider-active-ripple"))}_isShowingRipple(A){return A?.state===Os.FADING_IN||A?.state===Os.VISIBLE}_showRipple(A,i){if(!this._slider.disabled&&(this._showValueIndicator(),this._slider._isRange&&this._slider._getThumb(this.thumbPosition===ui.START?ui.END:ui.START)._showValueIndicator(),!(this._slider._globalRippleOptions?.disabled&&!i)))return this._ripple.launch({animation:this._slider._noopAnimations?{enterDuration:0,exitDuration:0}:A,centered:!0,persistent:!0})}_hideRipple(A){if(A?.fadeOut(),this._isShowingAnyRipple())return;this._slider._isRange||this._hideValueIndicator();let i=this._getSibling();i._isShowingAnyRipple()||(this._hideValueIndicator(),i._hideValueIndicator())}_showValueIndicator(){this._hostElement.classList.add("mdc-slider__thumb--with-indicator")}_hideValueIndicator(){this._hostElement.classList.remove("mdc-slider__thumb--with-indicator")}_getSibling(){return this._slider._getThumb(this.thumbPosition===ui.START?ui.END:ui.START)}_getValueIndicatorContainer(){return this._valueIndicatorContainer?.nativeElement}_getKnob(){return this._knob.nativeElement}_isShowingAnyRipple(){return this._isShowingRipple(this._hoverRippleRef)||this._isShowingRipple(this._focusRippleRef)||this._isShowingRipple(this._activeRippleRef)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-slider-visual-thumb"]],viewQuery:function(i,n){if(i&1&&(Te(rs,5),Te(cHA,5),Te(lHA,5)),i&2){let o;XA(o=$A())&&(n._ripple=o.first),XA(o=$A())&&(n._knob=o.first),XA(o=$A())&&(n._valueIndicatorContainer=o.first)}},hostAttrs:[1,"mdc-slider__thumb","mat-mdc-slider-visual-thumb"],inputs:{discrete:"discrete",thumbPosition:"thumbPosition",valueIndicatorText:"valueIndicatorText"},features:[ht([{provide:EcA,useExisting:t}])],decls:4,vars:2,consts:[["knob",""],["valueIndicatorContainer",""],[1,"mdc-slider__value-indicator-container"],[1,"mdc-slider__thumb-knob"],["matRipple","",1,"mat-focus-indicator",3,"matRippleDisabled"],[1,"mdc-slider__value-indicator"],[1,"mdc-slider__value-indicator-text"]],template:function(i,n){i&1&&(_A(0,gHA,5,1,"div",2),JA(1,"div",3,0)(3,"div",4)),i&2&&(GA(n.discrete?0:-1),G(3),yA("matRippleDisabled",!0))},dependencies:[rs],styles:[".mat-mdc-slider-visual-thumb .mat-ripple{height:100%;width:100%}.mat-mdc-slider .mdc-slider__tick-marks{justify-content:start}.mat-mdc-slider .mdc-slider__tick-marks .mdc-slider__tick-mark--active,.mat-mdc-slider .mdc-slider__tick-marks .mdc-slider__tick-mark--inactive{position:absolute;left:2px}"],encapsulation:2,changeDetection:0})}return t})(),QcA=(()=>{class t{_ngZone=f(Qe);_cdr=f(It);_elementRef=f(ee);_dir=f(mo,{optional:!0});_globalRippleOptions=f(G2,{optional:!0});_trackActive;_thumbs;_input;_inputs;get disabled(){return this._disabled}set disabled(A){this._disabled=A;let i=this._getInput(ui.END),n=this._getInput(ui.START);i&&(i.disabled=this._disabled),n&&(n.disabled=this._disabled)}_disabled=!1;get discrete(){return this._discrete}set discrete(A){this._discrete=A,this._updateValueIndicatorUIs()}_discrete=!1;showTickMarks=!1;get min(){return this._min}set min(A){let i=isNaN(A)?this._min:A;this._min!==i&&this._updateMin(i)}_min=0;color;disableRipple=!1;_updateMin(A){let i=this._min;this._min=A,this._isRange?this._updateMinRange({old:i,new:A}):this._updateMinNonRange(A),this._onMinMaxOrStepChange()}_updateMinRange(A){let i=this._getInput(ui.END),n=this._getInput(ui.START),o=i.value,r=n.value;n.min=A.new,i.min=Math.max(A.new,n.value),n.max=Math.min(i.max,i.value),n._updateWidthInactive(),i._updateWidthInactive(),A.new<A.old?this._onTranslateXChangeBySideEffect(i,n):this._onTranslateXChangeBySideEffect(n,i),o!==i.value&&this._onValueChange(i),r!==n.value&&this._onValueChange(n)}_updateMinNonRange(A){let i=this._getInput(ui.END);if(i){let n=i.value;i.min=A,i._updateThumbUIByValue(),this._updateTrackUI(i),n!==i.value&&this._onValueChange(i)}}get max(){return this._max}set max(A){let i=isNaN(A)?this._max:A;this._max!==i&&this._updateMax(i)}_max=100;_updateMax(A){let i=this._max;this._max=A,this._isRange?this._updateMaxRange({old:i,new:A}):this._updateMaxNonRange(A),this._onMinMaxOrStepChange()}_updateMaxRange(A){let i=this._getInput(ui.END),n=this._getInput(ui.START),o=i.value,r=n.value;i.max=A.new,n.max=Math.min(A.new,i.value),i.min=n.value,i._updateWidthInactive(),n._updateWidthInactive(),A.new>A.old?this._onTranslateXChangeBySideEffect(n,i):this._onTranslateXChangeBySideEffect(i,n),o!==i.value&&this._onValueChange(i),r!==n.value&&this._onValueChange(n)}_updateMaxNonRange(A){let i=this._getInput(ui.END);if(i){let n=i.value;i.max=A,i._updateThumbUIByValue(),this._updateTrackUI(i),n!==i.value&&this._onValueChange(i)}}get step(){return this._step}set step(A){let i=isNaN(A)?this._step:A;this._step!==i&&this._updateStep(i)}_step=1;_updateStep(A){this._step=A,this._isRange?this._updateStepRange():this._updateStepNonRange(),this._onMinMaxOrStepChange()}_updateStepRange(){let A=this._getInput(ui.END),i=this._getInput(ui.START),n=A.value,o=i.value,r=i.value;A.min=this._min,i.max=this._max,A.step=this._step,i.step=this._step,this._platform.SAFARI&&(A.value=A.value,i.value=i.value),A.min=Math.max(this._min,i.value),i.max=Math.min(this._max,A.value),i._updateWidthInactive(),A._updateWidthInactive(),A.value<r?this._onTranslateXChangeBySideEffect(i,A):this._onTranslateXChangeBySideEffect(A,i),n!==A.value&&this._onValueChange(A),o!==i.value&&this._onValueChange(i)}_updateStepNonRange(){let A=this._getInput(ui.END);if(A){let i=A.value;A.step=this._step,this._platform.SAFARI&&(A.value=A.value),A._updateThumbUIByValue(),i!==A.value&&this._onValueChange(A)}}displayWith=A=>`${A}`;_tickMarks;_noopAnimations;_dirChangeSubscription;_resizeObserver;_cachedWidth;_cachedLeft;_rippleRadius=24;startValueIndicatorText="";endValueIndicatorText="";_endThumbTransform;_startThumbTransform;_isRange=!1;_isRtl=!1;_hasViewInitialized=!1;_tickMarkTrackWidth=0;_hasAnimation=!1;_resizeTimer=null;_platform=f(Ii);constructor(){f(Rn).load(lr);let A=f(Si,{optional:!0});this._noopAnimations=A==="NoopAnimations",this._dir&&(this._dirChangeSubscription=this._dir.change.subscribe(()=>this._onDirChange()),this._isRtl=this._dir.value==="rtl")}_knobRadius=8;_inputPadding;ngAfterViewInit(){this._platform.isBrowser&&this._updateDimensions();let A=this._getInput(ui.END),i=this._getInput(ui.START);this._isRange=!!A&&!!i,this._cdr.detectChanges();let n=this._getThumb(ui.END);this._rippleRadius=n._ripple.radius,this._inputPadding=this._rippleRadius-this._knobRadius,this._isRange?this._initUIRange(A,i):this._initUINonRange(A),this._updateTrackUI(A),this._updateTickMarkUI(),this._updateTickMarkTrackUI(),this._observeHostResize(),this._cdr.detectChanges()}_initUINonRange(A){A.initProps(),A.initUI(),this._updateValueIndicatorUI(A),this._hasViewInitialized=!0,A._updateThumbUIByValue()}_initUIRange(A,i){A.initProps(),A.initUI(),i.initProps(),i.initUI(),A._updateMinMax(),i._updateMinMax(),A._updateStaticStyles(),i._updateStaticStyles(),this._updateValueIndicatorUIs(),this._hasViewInitialized=!0,A._updateThumbUIByValue(),i._updateThumbUIByValue()}ngOnDestroy(){this._dirChangeSubscription.unsubscribe(),this._resizeObserver?.disconnect(),this._resizeObserver=null}_onDirChange(){this._isRtl=this._dir?.value==="rtl",this._isRange?this._onDirChangeRange():this._onDirChangeNonRange(),this._updateTickMarkUI()}_onDirChangeRange(){let A=this._getInput(ui.END),i=this._getInput(ui.START);A._setIsLeftThumb(),i._setIsLeftThumb(),A.translateX=A._calcTranslateXByValue(),i.translateX=i._calcTranslateXByValue(),A._updateStaticStyles(),i._updateStaticStyles(),A._updateWidthInactive(),i._updateWidthInactive(),A._updateThumbUIByValue(),i._updateThumbUIByValue()}_onDirChangeNonRange(){this._getInput(ui.END)._updateThumbUIByValue()}_observeHostResize(){typeof ResizeObserver>"u"||!ResizeObserver||this._ngZone.runOutsideAngular(()=>{this._resizeObserver=new ResizeObserver(()=>{this._isActive()||(this._resizeTimer&&clearTimeout(this._resizeTimer),this._onResize())}),this._resizeObserver.observe(this._elementRef.nativeElement)})}_isActive(){return this._getThumb(ui.START)._isActive||this._getThumb(ui.END)._isActive}_getValue(A=ui.END){let i=this._getInput(A);return i?i.value:this.min}_skipUpdate(){return!!(this._getInput(ui.START)?._skipUIUpdate||this._getInput(ui.END)?._skipUIUpdate)}_updateDimensions(){this._cachedWidth=this._elementRef.nativeElement.offsetWidth,this._cachedLeft=this._elementRef.nativeElement.getBoundingClientRect().left}_setTrackActiveStyles(A){let i=this._trackActive.nativeElement.style;i.left=A.left,i.right=A.right,i.transformOrigin=A.transformOrigin,i.transform=A.transform}_calcTickMarkTransform(A){let i=A*(this._tickMarkTrackWidth/(this._tickMarks.length-1));return`translateX(${this._isRtl?this._cachedWidth-6-i:i}px`}_onTranslateXChange(A){this._hasViewInitialized&&(this._updateThumbUI(A),this._updateTrackUI(A),this._updateOverlappingThumbUI(A))}_onTranslateXChangeBySideEffect(A,i){this._hasViewInitialized&&(A._updateThumbUIByValue(),i._updateThumbUIByValue())}_onValueChange(A){this._hasViewInitialized&&(this._updateValueIndicatorUI(A),this._updateTickMarkUI(),this._cdr.detectChanges())}_onMinMaxOrStepChange(){this._hasViewInitialized&&(this._updateTickMarkUI(),this._updateTickMarkTrackUI(),this._cdr.markForCheck())}_onResize(){if(this._hasViewInitialized){if(this._updateDimensions(),this._isRange){let A=this._getInput(ui.END),i=this._getInput(ui.START);A._updateThumbUIByValue(),i._updateThumbUIByValue(),A._updateStaticStyles(),i._updateStaticStyles(),A._updateMinMax(),i._updateMinMax(),A._updateWidthInactive(),i._updateWidthInactive()}else{let A=this._getInput(ui.END);A&&A._updateThumbUIByValue()}this._updateTickMarkUI(),this._updateTickMarkTrackUI(),this._cdr.detectChanges()}}_thumbsOverlap=!1;_areThumbsOverlapping(){let A=this._getInput(ui.START),i=this._getInput(ui.END);return!A||!i?!1:i.translateX-A.translateX<20}_updateOverlappingThumbClassNames(A){let i=A.getSibling(),n=this._getThumb(A.thumbPosition);this._getThumb(i.thumbPosition)._hostElement.classList.remove("mdc-slider__thumb--top"),n._hostElement.classList.toggle("mdc-slider__thumb--top",this._thumbsOverlap)}_updateOverlappingThumbUI(A){!this._isRange||this._skipUpdate()||this._thumbsOverlap!==this._areThumbsOverlapping()&&(this._thumbsOverlap=!this._thumbsOverlap,this._updateOverlappingThumbClassNames(A))}_updateThumbUI(A){if(this._skipUpdate())return;let i=this._getThumb(A.thumbPosition===ui.END?ui.END:ui.START);i._hostElement.style.transform=`translateX(${A.translateX}px)`}_updateValueIndicatorUI(A){if(this._skipUpdate())return;let i=this.displayWith(A.value);if(this._hasViewInitialized?A._valuetext.set(i):A._hostElement.setAttribute("aria-valuetext",i),this.discrete){A.thumbPosition===ui.START?this.startValueIndicatorText=i:this.endValueIndicatorText=i;let n=this._getThumb(A.thumbPosition);i.length<3?n._hostElement.classList.add("mdc-slider__thumb--short-value"):n._hostElement.classList.remove("mdc-slider__thumb--short-value")}}_updateValueIndicatorUIs(){let A=this._getInput(ui.END),i=this._getInput(ui.START);A&&this._updateValueIndicatorUI(A),i&&this._updateValueIndicatorUI(i)}_updateTickMarkTrackUI(){if(!this.showTickMarks||this._skipUpdate())return;let A=this._step&&this._step>0?this._step:1,n=(Math.floor(this.max/A)*A-this.min)/(this.max-this.min);this._tickMarkTrackWidth=(this._cachedWidth-6)*n}_updateTrackUI(A){this._skipUpdate()||(this._isRange?this._updateTrackUIRange(A):this._updateTrackUINonRange(A))}_updateTrackUIRange(A){let i=A.getSibling();if(!i||!this._cachedWidth)return;let n=Math.abs(i.translateX-A.translateX)/this._cachedWidth;A._isLeftThumb&&this._cachedWidth?this._setTrackActiveStyles({left:"auto",right:`${this._cachedWidth-i.translateX}px`,transformOrigin:"right",transform:`scaleX(${n})`}):this._setTrackActiveStyles({left:`${i.translateX}px`,right:"auto",transformOrigin:"left",transform:`scaleX(${n})`})}_updateTrackUINonRange(A){this._isRtl?this._setTrackActiveStyles({left:"auto",right:"0px",transformOrigin:"right",transform:`scaleX(${1-A.fillPercentage})`}):this._setTrackActiveStyles({left:"0px",right:"auto",transformOrigin:"left",transform:`scaleX(${A.fillPercentage})`})}_updateTickMarkUI(){if(!this.showTickMarks||this.step===void 0||this.min===void 0||this.max===void 0)return;let A=this.step>0?this.step:1;this._isRange?this._updateTickMarkUIRange(A):this._updateTickMarkUINonRange(A)}_updateTickMarkUINonRange(A){let i=this._getValue(),n=Math.max(Math.round((i-this.min)/A),0)+1,o=Math.max(Math.round((this.max-i)/A),0)-1;this._isRtl?n++:o++,this._tickMarks=Array(n).fill(kQ.ACTIVE).concat(Array(o).fill(kQ.INACTIVE))}_updateTickMarkUIRange(A){let i=this._getValue(),n=this._getValue(ui.START),o=Math.max(Math.round((n-this.min)/A),0),r=Math.max(Math.round((i-n)/A)+1,0),s=Math.max(Math.round((this.max-i)/A),0);this._tickMarks=Array(o).fill(kQ.INACTIVE).concat(Array(r).fill(kQ.ACTIVE),Array(s).fill(kQ.INACTIVE))}_getInput(A){if(A===ui.END&&this._input)return this._input;if(this._inputs?.length)return A===ui.START?this._inputs.first:this._inputs.last}_getThumb(A){return A===ui.END?this._thumbs?.last:this._thumbs?.first}_setTransition(A){this._hasAnimation=!this._platform.IOS&&A&&!this._noopAnimations,this._elementRef.nativeElement.classList.toggle("mat-mdc-slider-with-animation",this._hasAnimation)}_isCursorOnSliderThumb(A,i){let n=i.width/2,o=i.x+n,r=i.y+n,s=A.clientX-o,a=A.clientY-r;return Math.pow(s,2)+Math.pow(a,2)<Math.pow(n,2)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-slider"]],contentQueries:function(i,n,o){if(i&1&&(ci(o,BcA,5),ci(o,hHA,4)),i&2){let r;XA(r=$A())&&(n._input=r.first),XA(r=$A())&&(n._inputs=r)}},viewQuery:function(i,n){if(i&1&&(Te(IHA,5),Te(EcA,5)),i&2){let o;XA(o=$A())&&(n._trackActive=o.first),XA(o=$A())&&(n._thumbs=o)}},hostAttrs:[1,"mat-mdc-slider","mdc-slider"],hostVars:12,hostBindings:function(i,n){i&2&&(fo("mat-"+(n.color||"primary")),ue("mdc-slider--range",n._isRange)("mdc-slider--disabled",n.disabled)("mdc-slider--discrete",n.discrete)("mdc-slider--tick-marks",n.showTickMarks)("_mat-animation-noopable",n._noopAnimations))},inputs:{disabled:[2,"disabled","disabled",ie],discrete:[2,"discrete","discrete",ie],showTickMarks:[2,"showTickMarks","showTickMarks",ie],min:[2,"min","min",zi],color:"color",disableRipple:[2,"disableRipple","disableRipple",ie],max:[2,"max","max",zi],step:[2,"step","step",zi],displayWith:"displayWith"},exportAs:["matSlider"],features:[ht([{provide:xG,useExisting:t}])],ngContentSelectors:CHA,decls:9,vars:5,consts:[["trackActive",""],["tickMarkContainer",""],[1,"mdc-slider__track"],[1,"mdc-slider__track--inactive"],[1,"mdc-slider__track--active"],[1,"mdc-slider__track--active_fill"],[1,"mdc-slider__tick-marks"],[3,"discrete","thumbPosition","valueIndicatorText"],[3,"class","transform"]],template:function(i,n){i&1&&(jt(),Fe(0),S(1,"div",2),JA(2,"div",3),S(3,"div",4),JA(4,"div",5,0),F(),_A(6,EHA,3,1,"div",6),F(),_A(7,QHA,1,3,"mat-slider-visual-thumb",7),JA(8,"mat-slider-visual-thumb",7)),i&2&&(G(6),GA(n.showTickMarks?6:-1),G(),GA(n._isRange?7:-1),G(),yA("discrete",n.discrete)("thumbPosition",2)("valueIndicatorText",n.endValueIndicatorText))},dependencies:[uHA],styles:['.mdc-slider__track{position:absolute;top:50%;transform:translateY(-50%);width:100%;pointer-events:none;height:var(--mdc-slider-inactive-track-height, 4px)}.mdc-slider__track--active,.mdc-slider__track--inactive{display:flex;height:100%;position:absolute;width:100%}.mdc-slider__track--active{overflow:hidden;border-radius:var(--mdc-slider-active-track-shape, var(--mat-sys-corner-full));height:var(--mdc-slider-active-track-height, 4px);top:calc((var(--mdc-slider-inactive-track-height, 4px) - var(--mdc-slider-active-track-height, 4px))/2)}.mdc-slider__track--active_fill{border-top-style:solid;box-sizing:border-box;height:100%;width:100%;position:relative;transform-origin:left;transition:transform 80ms ease;border-color:var(--mdc-slider-active-track-color, var(--mat-sys-primary));border-top-width:var(--mdc-slider-active-track-height, 4px)}.mdc-slider--disabled .mdc-slider__track--active_fill{border-color:var(--mdc-slider-disabled-active-track-color, var(--mat-sys-on-surface))}[dir=rtl] .mdc-slider__track--active_fill{-webkit-transform-origin:right;transform-origin:right}.mdc-slider__track--inactive{left:0;top:0;opacity:.24;background-color:var(--mdc-slider-inactive-track-color, var(--mat-sys-surface-variant));height:var(--mdc-slider-inactive-track-height, 4px);border-radius:var(--mdc-slider-inactive-track-shape, var(--mat-sys-corner-full))}.mdc-slider--disabled .mdc-slider__track--inactive{background-color:var(--mdc-slider-disabled-inactive-track-color, var(--mat-sys-on-surface));opacity:.24}.mdc-slider__track--inactive::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:1px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}@media(forced-colors: active){.mdc-slider__track--inactive::before{border-color:CanvasText}}.mdc-slider__value-indicator-container{bottom:44px;left:50%;pointer-events:none;position:absolute;transform:translateX(-50%);transform:var(--mat-slider-value-indicator-container-transform, translateX(-50%) rotate(-45deg))}.mdc-slider__thumb--with-indicator .mdc-slider__value-indicator-container{pointer-events:auto}.mdc-slider__value-indicator{display:flex;align-items:center;border-radius:4px;height:32px;padding:0 12px;transform:scale(0);transform-origin:bottom;opacity:1;transition:transform 100ms cubic-bezier(0.4, 0, 1, 1);word-break:normal;background-color:var(--mdc-slider-label-container-color, var(--mat-sys-primary));color:var(--mdc-slider-label-label-text-color, var(--mat-sys-on-primary));width:var(--mat-slider-value-indicator-width, 28px);height:var(--mat-slider-value-indicator-height, 28px);padding:var(--mat-slider-value-indicator-padding, 0);opacity:var(--mat-slider-value-indicator-opacity, 1);border-radius:var(--mat-slider-value-indicator-border-radius, 50% 50% 50% 0)}.mdc-slider__thumb--with-indicator .mdc-slider__value-indicator{transition:transform 100ms cubic-bezier(0, 0, 0.2, 1);transform:scale(1)}.mdc-slider__value-indicator::before{border-left:6px solid rgba(0,0,0,0);border-right:6px solid rgba(0,0,0,0);border-top:6px solid;bottom:-5px;content:"";height:0;left:50%;position:absolute;transform:translateX(-50%);width:0;display:var(--mat-slider-value-indicator-caret-display, none);border-top-color:var(--mdc-slider-label-container-color, var(--mat-sys-primary))}.mdc-slider__value-indicator::after{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;border:1px solid rgba(0,0,0,0);border-radius:inherit;content:"";pointer-events:none}@media(forced-colors: active){.mdc-slider__value-indicator::after{border-color:CanvasText}}.mdc-slider__value-indicator-text{text-align:center;width:var(--mat-slider-value-indicator-width, 28px);transform:var(--mat-slider-value-indicator-text-transform, rotate(45deg));font-family:var(--mdc-slider-label-label-text-font, var(--mat-sys-label-medium-font));font-size:var(--mdc-slider-label-label-text-size, var(--mat-sys-label-medium-size));font-weight:var(--mdc-slider-label-label-text-weight, var(--mat-sys-label-medium-weight));line-height:var(--mdc-slider-label-label-text-line-height, var(--mat-sys-label-medium-line-height));letter-spacing:var(--mdc-slider-label-label-text-tracking, var(--mat-sys-label-medium-tracking))}.mdc-slider__thumb{-webkit-user-select:none;user-select:none;display:flex;left:-24px;outline:none;position:absolute;height:48px;width:48px;pointer-events:none}.mdc-slider--discrete .mdc-slider__thumb{transition:transform 80ms ease}.mdc-slider--disabled .mdc-slider__thumb{pointer-events:none}.mdc-slider__thumb--top{z-index:1}.mdc-slider__thumb-knob{position:absolute;box-sizing:border-box;left:50%;top:50%;transform:translate(-50%, -50%);border-style:solid;width:var(--mdc-slider-handle-width, 20px);height:var(--mdc-slider-handle-height, 20px);border-width:calc(var(--mdc-slider-handle-height, 20px)/2) calc(var(--mdc-slider-handle-width, 20px)/2);box-shadow:var(--mdc-slider-handle-elevation, var(--mat-sys-level1));background-color:var(--mdc-slider-handle-color, var(--mat-sys-primary));border-color:var(--mdc-slider-handle-color, var(--mat-sys-primary));border-radius:var(--mdc-slider-handle-shape, var(--mat-sys-corner-full))}.mdc-slider__thumb:hover .mdc-slider__thumb-knob{background-color:var(--mdc-slider-hover-handle-color, var(--mat-sys-primary));border-color:var(--mdc-slider-hover-handle-color, var(--mat-sys-primary))}.mdc-slider__thumb--focused .mdc-slider__thumb-knob{background-color:var(--mdc-slider-focus-handle-color, var(--mat-sys-primary));border-color:var(--mdc-slider-focus-handle-color, var(--mat-sys-primary))}.mdc-slider--disabled .mdc-slider__thumb-knob{background-color:var(--mdc-slider-disabled-handle-color, var(--mat-sys-on-surface));border-color:var(--mdc-slider-disabled-handle-color, var(--mat-sys-on-surface))}.mdc-slider__thumb--top .mdc-slider__thumb-knob,.mdc-slider__thumb--top.mdc-slider__thumb:hover .mdc-slider__thumb-knob,.mdc-slider__thumb--top.mdc-slider__thumb--focused .mdc-slider__thumb-knob{border:solid 1px #fff;box-sizing:content-box;border-color:var(--mdc-slider-with-overlap-handle-outline-color, var(--mat-sys-on-primary));border-width:var(--mdc-slider-with-overlap-handle-outline-width, 1px)}.mdc-slider__tick-marks{align-items:center;box-sizing:border-box;display:flex;height:100%;justify-content:space-between;padding:0 1px;position:absolute;width:100%}.mdc-slider__tick-mark--active,.mdc-slider__tick-mark--inactive{width:var(--mdc-slider-with-tick-marks-container-size, 2px);height:var(--mdc-slider-with-tick-marks-container-size, 2px);border-radius:var(--mdc-slider-with-tick-marks-container-shape, var(--mat-sys-corner-full))}.mdc-slider__tick-mark--inactive{opacity:var(--mdc-slider-with-tick-marks-inactive-container-opacity, 0.38);background-color:var(--mdc-slider-with-tick-marks-inactive-container-color, var(--mat-sys-on-surface-variant))}.mdc-slider--disabled .mdc-slider__tick-mark--inactive{opacity:var(--mdc-slider-with-tick-marks-inactive-container-opacity, 0.38);background-color:var(--mdc-slider-with-tick-marks-disabled-container-color, var(--mat-sys-on-surface))}.mdc-slider__tick-mark--active{opacity:var(--mdc-slider-with-tick-marks-active-container-opacity, 0.38);background-color:var(--mdc-slider-with-tick-marks-active-container-color, var(--mat-sys-on-primary))}.mdc-slider__input{cursor:pointer;left:2px;margin:0;height:44px;opacity:0;position:absolute;top:2px;width:44px;box-sizing:content-box}.mdc-slider__input.mat-mdc-slider-input-no-pointer-events{pointer-events:none}.mdc-slider__input.mat-slider__right-input{left:auto;right:0}.mat-mdc-slider{display:inline-block;box-sizing:border-box;outline:none;vertical-align:middle;cursor:pointer;height:48px;margin:0 8px;position:relative;touch-action:pan-y;width:auto;min-width:112px;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-slider.mdc-slider--disabled{cursor:auto;opacity:.38}.mat-mdc-slider .mdc-slider__thumb,.mat-mdc-slider .mdc-slider__track--active_fill{transition-duration:0ms}.mat-mdc-slider.mat-mdc-slider-with-animation .mdc-slider__thumb,.mat-mdc-slider.mat-mdc-slider-with-animation .mdc-slider__track--active_fill{transition-duration:80ms}.mat-mdc-slider.mdc-slider--discrete .mdc-slider__thumb,.mat-mdc-slider.mdc-slider--discrete .mdc-slider__track--active_fill{transition-duration:0ms}.mat-mdc-slider.mat-mdc-slider-with-animation .mdc-slider__thumb,.mat-mdc-slider.mat-mdc-slider-with-animation .mdc-slider__track--active_fill{transition-duration:80ms}.mat-mdc-slider .mat-ripple .mat-ripple-element{background-color:var(--mat-slider-ripple-color, var(--mat-sys-primary))}.mat-mdc-slider .mat-ripple .mat-mdc-slider-hover-ripple{background-color:var(--mat-slider-hover-state-layer-color, color-mix(in srgb, var(--mat-sys-primary) 5%, transparent))}.mat-mdc-slider .mat-ripple .mat-mdc-slider-focus-ripple,.mat-mdc-slider .mat-ripple .mat-mdc-slider-active-ripple{background-color:var(--mat-slider-focus-state-layer-color, color-mix(in srgb, var(--mat-sys-primary) 20%, transparent))}.mat-mdc-slider._mat-animation-noopable.mdc-slider--discrete .mdc-slider__thumb,.mat-mdc-slider._mat-animation-noopable.mdc-slider--discrete .mdc-slider__track--active_fill,.mat-mdc-slider._mat-animation-noopable .mdc-slider__value-indicator{transition:none}.mat-mdc-slider .mat-focus-indicator::before{border-radius:50%}.mdc-slider__thumb--focused .mat-focus-indicator::before{content:""}'],encapsulation:2,changeDetection:0})}return t})();var fHA={provide:yc,useExisting:Ir(()=>LG),multi:!0};var LG=(()=>{class t{_ngZone=f(Qe);_elementRef=f(ee);_cdr=f(It);_slider=f(xG);_platform=f(Ii);_listenerCleanups;get value(){return zi(this._hostElement.value,0)}set value(A){A=isNaN(A)?0:A;let i=A+"";if(!this._hasSetInitialValue){this._initialValue=i;return}this._isActive||this._setValue(i)}_setValue(A){this._hostElement.value=A,this._updateThumbUIByValue(),this._slider._onValueChange(this),this._cdr.detectChanges(),this._slider._cdr.markForCheck()}valueChange=new WA;dragStart=new WA;dragEnd=new WA;get translateX(){return this._slider.min>=this._slider.max?(this._translateX=this._tickMarkOffset,this._translateX):(this._translateX===void 0&&(this._translateX=this._calcTranslateXByValue()),this._translateX)}set translateX(A){this._translateX=A}_translateX;thumbPosition=ui.END;get min(){return zi(this._hostElement.min,0)}set min(A){this._hostElement.min=A+"",this._cdr.detectChanges()}get max(){return zi(this._hostElement.max,0)}set max(A){this._hostElement.max=A+"",this._cdr.detectChanges()}get step(){return zi(this._hostElement.step,0)}set step(A){this._hostElement.step=A+"",this._cdr.detectChanges()}get disabled(){return ie(this._hostElement.disabled)}set disabled(A){this._hostElement.disabled=A,this._cdr.detectChanges(),this._slider.disabled!==this.disabled&&(this._slider.disabled=this.disabled)}get percentage(){return this._slider.min>=this._slider.max?this._slider._isRtl?1:0:(this.value-this._slider.min)/(this._slider.max-this._slider.min)}get fillPercentage(){return this._slider._cachedWidth?this._translateX===0?0:this.translateX/this._slider._cachedWidth:this._slider._isRtl?1:0}_hostElement=this._elementRef.nativeElement;_valuetext=Jo("");_knobRadius=8;_tickMarkOffset=3;_isActive=!1;_isFocused=!1;_setIsFocused(A){this._isFocused=A}_hasSetInitialValue=!1;_initialValue;_formControl;_destroyed=new OA;_skipUIUpdate=!1;_onChangeFn;_onTouchedFn=()=>{};_isControlInitialized=!1;constructor(){let A=f(Wi);this._ngZone.runOutsideAngular(()=>{this._listenerCleanups=[A.listen(this._hostElement,"pointerdown",this._onPointerDown.bind(this)),A.listen(this._hostElement,"pointermove",this._onPointerMove.bind(this)),A.listen(this._hostElement,"pointerup",this._onPointerUp.bind(this))]})}ngOnDestroy(){this._listenerCleanups.forEach(A=>A()),this._destroyed.next(),this._destroyed.complete(),this.dragStart.complete(),this.dragEnd.complete()}initProps(){this._updateWidthInactive(),this.disabled!==this._slider.disabled&&(this._slider.disabled=!0),this.step=this._slider.step,this.min=this._slider.min,this.max=this._slider.max,this._initValue()}initUI(){this._updateThumbUIByValue()}_initValue(){this._hasSetInitialValue=!0,this._initialValue===void 0?this.value=this._getDefaultValue():(this._hostElement.value=this._initialValue,this._updateThumbUIByValue(),this._slider._onValueChange(this),this._cdr.detectChanges())}_getDefaultValue(){return this.min}_onBlur(){this._setIsFocused(!1),this._onTouchedFn()}_onFocus(){this._slider._setTransition(!1),this._slider._updateTrackUI(this),this._setIsFocused(!0)}_onChange(){this.valueChange.emit(this.value),this._isActive&&this._updateThumbUIByValue({withAnimation:!0})}_onInput(){this._onChangeFn?.(this.value),(this._slider.step||!this._isActive)&&this._updateThumbUIByValue({withAnimation:!0}),this._slider._onValueChange(this)}_onNgControlValueChange(){(!this._isActive||!this._isFocused)&&(this._slider._onValueChange(this),this._updateThumbUIByValue()),this._slider.disabled=this._formControl.disabled}_onPointerDown(A){if(!(this.disabled||A.button!==0)){if(this._platform.IOS){let i=this._slider._isCursorOnSliderThumb(A,this._slider._getThumb(this.thumbPosition)._hostElement.getBoundingClientRect());this._isActive=i,this._updateWidthActive(),this._slider._updateDimensions();return}this._isActive=!0,this._setIsFocused(!0),this._updateWidthActive(),this._slider._updateDimensions(),this._slider.step||this._updateThumbUIByPointerEvent(A,{withAnimation:!0}),this.disabled||(this._handleValueCorrection(A),this.dragStart.emit({source:this,parent:this._slider,value:this.value}))}}_handleValueCorrection(A){this._skipUIUpdate=!0,setTimeout(()=>{this._skipUIUpdate=!1,this._fixValue(A)},0)}_fixValue(A){let i=A.clientX-this._slider._cachedLeft,n=this._slider._cachedWidth,o=this._slider.step===0?1:this._slider.step,r=Math.floor((this._slider.max-this._slider.min)/o),s=this._slider._isRtl?1-i/n:i/n,c=Math.round(s*r)/r*(this._slider.max-this._slider.min)+this._slider.min,l=Math.round(c/o)*o,I=this.value;if(l===I){this._slider._onValueChange(this),this._slider.step>0?this._updateThumbUIByValue():this._updateThumbUIByPointerEvent(A,{withAnimation:this._slider._hasAnimation});return}this.value=l,this.valueChange.emit(this.value),this._onChangeFn?.(this.value),this._slider._onValueChange(this),this._slider.step>0?this._updateThumbUIByValue():this._updateThumbUIByPointerEvent(A,{withAnimation:this._slider._hasAnimation})}_onPointerMove(A){!this._slider.step&&this._isActive&&this._updateThumbUIByPointerEvent(A)}_onPointerUp(){this._isActive&&(this._isActive=!1,this._platform.SAFARI&&this._setIsFocused(!1),this.dragEnd.emit({source:this,parent:this._slider,value:this.value}),setTimeout(()=>this._updateWidthInactive(),this._platform.IOS?10:0))}_clamp(A){let i=this._tickMarkOffset,n=this._slider._cachedWidth-this._tickMarkOffset;return Math.max(Math.min(A,n),i)}_calcTranslateXByValue(){return this._slider._isRtl?(1-this.percentage)*(this._slider._cachedWidth-this._tickMarkOffset*2)+this._tickMarkOffset:this.percentage*(this._slider._cachedWidth-this._tickMarkOffset*2)+this._tickMarkOffset}_calcTranslateXByPointerEvent(A){return A.clientX-this._slider._cachedLeft}_updateWidthActive(){}_updateWidthInactive(){this._hostElement.style.padding=`0 ${this._slider._inputPadding}px`,this._hostElement.style.width=`calc(100% + ${this._slider._inputPadding-this._tickMarkOffset*2}px)`,this._hostElement.style.left=`-${this._slider._rippleRadius-this._tickMarkOffset}px`}_updateThumbUIByValue(A){this.translateX=this._clamp(this._calcTranslateXByValue()),this._updateThumbUI(A)}_updateThumbUIByPointerEvent(A,i){this.translateX=this._clamp(this._calcTranslateXByPointerEvent(A)),this._updateThumbUI(i)}_updateThumbUI(A){this._slider._setTransition(!!A?.withAnimation),this._slider._onTranslateXChange(this)}writeValue(A){(this._isControlInitialized||A!==null)&&(this.value=A)}registerOnChange(A){this._onChangeFn=A,this._isControlInitialized=!0}registerOnTouched(A){this._onTouchedFn=A}setDisabledState(A){this.disabled=A}focus(){this._hostElement.focus()}blur(){this._hostElement.blur()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["input","matSliderThumb",""]],hostAttrs:["type","range",1,"mdc-slider__input"],hostVars:1,hostBindings:function(i,n){i&1&&hA("change",function(){return n._onChange()})("input",function(){return n._onInput()})("blur",function(){return n._onBlur()})("focus",function(){return n._onFocus()}),i&2&&Ne("aria-valuetext",n._valuetext())},inputs:{value:[2,"value","value",zi]},outputs:{valueChange:"valueChange",dragStart:"dragStart",dragEnd:"dragEnd"},exportAs:["matSliderThumb"],features:[ht([fHA,{provide:BcA,useExisting:t}])]})}return t})();var hcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,Ps]})}return t})();var Qf=class t{constructor(e,A,i){this.dialogRef=e;this.fb=A;this.data=i;this.evalMetrics=this.data.evalMetrics,this.evalForm=this.fb.group({tool_trajectory_avg_score_threshold:[this.getEvalMetricThresholdFromData("tool_trajectory_avg_score"),[$a.required,$a.min(0),$a.max(1)]],response_match_score_threshold:[this.getEvalMetricThresholdFromData("response_match_score"),[$a.required,$a.min(0),$a.max(1)]]})}evalForm;evalMetrics=[];getEvalMetricThresholdFromData(e){return this.evalMetrics.find(A=>A.metricName===e)?.threshold??0}onStart(){if(this.evalForm.valid){let{tool_trajectory_avg_score_threshold:e,response_match_score_threshold:A}=this.evalForm.value;for(let i of this.evalMetrics)i.metricName==="tool_trajectory_avg_score"?i.threshold=e:i.metricName==="response_match_score"&&(i.threshold=A);this.dialogRef.close(this.evalMetrics)}}onCancel(){this.dialogRef.close(null)}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(Rz),PA(as))};static \u0275cmp=zA({type:t,selectors:[["app-run-eval-config-dialog"]],standalone:!1,decls:26,vars:3,consts:[[1,"dialog-container"],["mat-dialog-title","",1,"dialog-title"],[1,"eval-form",3,"formGroup"],[1,"metric-row"],[1,"metric-name"],[1,"flex-1","pl-4"],["min","0","max","1","step","0.1","thumbLabel","",1,"threshold-slider"],["matSliderThumb","","formControlName","tool_trajectory_avg_score_threshold"],[1,"threshold-value"],["matSliderThumb","","formControlName","response_match_score_threshold"],["align","end",1,"dialog-actions"],["mat-button","",1,"cancel-button",3,"click"],["mat-button","",1,"save-button",3,"click"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"h2",1),AA(2,"EVALUATION METRIC"),F(),S(3,"mat-dialog-content")(4,"form",2)(5,"div",3)(6,"div",4),AA(7,"Tool trajectory avg score: "),F(),S(8,"div",5)(9,"mat-slider",6),JA(10,"input",7),F(),S(11,"span",8),AA(12),F()()(),S(13,"div",3)(14,"div",4),AA(15,"Response match score: "),F(),S(16,"div",5)(17,"mat-slider",6),JA(18,"input",9),F(),S(19,"span",8),AA(20),F()()()()(),S(21,"mat-dialog-actions",10)(22,"button",11),hA("click",function(){return i.onCancel()}),AA(23,"Cancel"),F(),S(24,"button",12),hA("click",function(){return i.onStart()}),AA(25,"Start"),F()()()),A&2&&(G(4),yA("formGroup",i.evalForm),G(8),Et(" ",i.evalForm.controls.tool_trajectory_avg_score_threshold.value," "),G(8),Et(" ",i.evalForm.controls.response_match_score_threshold.value," "))},dependencies:[kz,vc,Ea,pz,Dr,bs,ma,pa,QcA,LG,GI,bM],styles:[".dialog-container[_ngcontent-%COMP%]{border-radius:12px;padding:18px;width:500px;box-shadow:0 8px 16px #0006}.threshold-slider[_ngcontent-%COMP%]{--mdc-slider-active-track-color: #4285f4;--mdc-slider-inactive-track-color: #616161;--mdc-slider-handle-color: #4285f4;--mdc-slider-ripple-color: #4285f4;width:100px}.metric-row[_ngcontent-%COMP%]{display:flex;flex-direction:row;align-items:center}.metric-name[_ngcontent-%COMP%]{width:250px}.threshold-value[_ngcontent-%COMP%]{margin-left:20px}.mdc-slider__thumb--with-indicator[_ngcontent-%COMP%]{background-color:var(--mdc-slider-handle-color, black);border:none!important;box-shadow:none!important}"]})};var Wg=class t{constructor(e){this.http=e}apiServerDomain=vs.getApiServerBaseUrl();createSession(e,A){if(this.apiServerDomain!=null){let i=this.apiServerDomain+`/apps/${A}/users/${e}/sessions`;return this.http.post(i,null)}return new ct}listSessions(e,A){if(this.apiServerDomain!=null){let i=this.apiServerDomain+`/apps/${A}/users/${e}/sessions`;return this.http.get(i)}return new ct}deleteSession(e,A,i){let n=this.apiServerDomain+`/apps/${A}/users/${e}/sessions/${i}`;return this.http.delete(n)}getSession(e,A,i){let n=this.apiServerDomain+`/apps/${A}/users/${e}/sessions/${i}`;return this.http.get(n)}importSession(e,A,i){if(this.apiServerDomain!=null){let n=this.apiServerDomain+`/apps/${A}/users/${e}/sessions`;return this.http.post(n,{appName:A,userId:e,events:i})}return new ct}static \u0275fac=function(A){return new(A||t)(we(Ds))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var pHA=["determinateSpinner"];function wHA(t,e){if(t&1&&(ar(),S(0,"svg",11),JA(1,"circle",12),F()),t&2){let A=O();Ne("viewBox",A._viewBox()),G(),uo("stroke-dasharray",A._strokeCircumference(),"px")("stroke-dashoffset",A._strokeCircumference()/2,"px")("stroke-width",A._circleStrokeWidth(),"%"),Ne("r",A._circleRadius())}}var DHA=new dA("mat-progress-spinner-default-options",{providedIn:"root",factory:yHA});function yHA(){return{diameter:ucA}}var ucA=100,vHA=10,fcA=(()=>{class t{_elementRef=f(ee);_noopAnimations;get color(){return this._color||this._defaultColor}set color(A){this._color=A}_color;_defaultColor="primary";_determinateCircle;constructor(){let A=f(Si,{optional:!0}),i=f(DHA);this._noopAnimations=A==="NoopAnimations"&&!!i&&!i._forceAnimations,this.mode=this._elementRef.nativeElement.nodeName.toLowerCase()==="mat-spinner"?"indeterminate":"determinate",i&&(i.color&&(this.color=this._defaultColor=i.color),i.diameter&&(this.diameter=i.diameter),i.strokeWidth&&(this.strokeWidth=i.strokeWidth))}mode;get value(){return this.mode==="determinate"?this._value:0}set value(A){this._value=Math.max(0,Math.min(100,A||0))}_value=0;get diameter(){return this._diameter}set diameter(A){this._diameter=A||0}_diameter=ucA;get strokeWidth(){return this._strokeWidth??this.diameter/10}set strokeWidth(A){this._strokeWidth=A||0}_strokeWidth;_circleRadius(){return(this.diameter-vHA)/2}_viewBox(){let A=this._circleRadius()*2+this.strokeWidth;return`0 0 ${A} ${A}`}_strokeCircumference(){return 2*Math.PI*this._circleRadius()}_strokeDashOffset(){return this.mode==="determinate"?this._strokeCircumference()*(100-this._value)/100:null}_circleStrokeWidth(){return this.strokeWidth/this.diameter*100}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-progress-spinner"],["mat-spinner"]],viewQuery:function(i,n){if(i&1&&Te(pHA,5),i&2){let o;XA(o=$A())&&(n._determinateCircle=o.first)}},hostAttrs:["role","progressbar","tabindex","-1",1,"mat-mdc-progress-spinner","mdc-circular-progress"],hostVars:18,hostBindings:function(i,n){i&2&&(Ne("aria-valuemin",0)("aria-valuemax",100)("aria-valuenow",n.mode==="determinate"?n.value:null)("mode",n.mode),fo("mat-"+n.color),uo("width",n.diameter,"px")("height",n.diameter,"px")("--mdc-circular-progress-size",n.diameter+"px")("--mdc-circular-progress-active-indicator-width",n.diameter+"px"),ue("_mat-animation-noopable",n._noopAnimations)("mdc-circular-progress--indeterminate",n.mode==="indeterminate"))},inputs:{color:"color",mode:"mode",value:[2,"value","value",zi],diameter:[2,"diameter","diameter",zi],strokeWidth:[2,"strokeWidth","strokeWidth",zi]},exportAs:["matProgressSpinner"],decls:14,vars:11,consts:[["circle",""],["determinateSpinner",""],["aria-hidden","true",1,"mdc-circular-progress__determinate-container"],["xmlns","http://www.w3.org/2000/svg","focusable","false",1,"mdc-circular-progress__determinate-circle-graphic"],["cx","50%","cy","50%",1,"mdc-circular-progress__determinate-circle"],["aria-hidden","true",1,"mdc-circular-progress__indeterminate-container"],[1,"mdc-circular-progress__spinner-layer"],[1,"mdc-circular-progress__circle-clipper","mdc-circular-progress__circle-left"],[3,"ngTemplateOutlet"],[1,"mdc-circular-progress__gap-patch"],[1,"mdc-circular-progress__circle-clipper","mdc-circular-progress__circle-right"],["xmlns","http://www.w3.org/2000/svg","focusable","false",1,"mdc-circular-progress__indeterminate-circle-graphic"],["cx","50%","cy","50%"]],template:function(i,n){if(i&1&&(_A(0,wHA,2,8,"ng-template",null,0,Fh),S(2,"div",2,1),ar(),S(4,"svg",3),JA(5,"circle",4),F()(),SI(),S(6,"div",5)(7,"div",6)(8,"div",7),_r(9,8),F(),S(10,"div",9),_r(11,8),F(),S(12,"div",10),_r(13,8),F()()()),i&2){let o=cr(1);G(4),Ne("viewBox",n._viewBox()),G(),uo("stroke-dasharray",n._strokeCircumference(),"px")("stroke-dashoffset",n._strokeDashOffset(),"px")("stroke-width",n._circleStrokeWidth(),"%"),Ne("r",n._circleRadius()),G(4),yA("ngTemplateOutlet",o),G(2),yA("ngTemplateOutlet",o),G(2),yA("ngTemplateOutlet",o)}},dependencies:[Yh],styles:[".mat-mdc-progress-spinner{display:block;overflow:hidden;line-height:0;position:relative;direction:ltr;transition:opacity 250ms cubic-bezier(0.4, 0, 0.6, 1)}.mat-mdc-progress-spinner circle{stroke-width:var(--mdc-circular-progress-active-indicator-width, 4px)}.mat-mdc-progress-spinner._mat-animation-noopable,.mat-mdc-progress-spinner._mat-animation-noopable .mdc-circular-progress__determinate-circle{transition:none !important}.mat-mdc-progress-spinner._mat-animation-noopable .mdc-circular-progress__indeterminate-circle-graphic,.mat-mdc-progress-spinner._mat-animation-noopable .mdc-circular-progress__spinner-layer,.mat-mdc-progress-spinner._mat-animation-noopable .mdc-circular-progress__indeterminate-container{animation:none !important}.mat-mdc-progress-spinner._mat-animation-noopable .mdc-circular-progress__indeterminate-container circle{stroke-dasharray:0 !important}@media(forced-colors: active){.mat-mdc-progress-spinner .mdc-circular-progress__indeterminate-circle-graphic,.mat-mdc-progress-spinner .mdc-circular-progress__determinate-circle{stroke:currentColor;stroke:CanvasText}}.mdc-circular-progress__determinate-container,.mdc-circular-progress__indeterminate-circle-graphic,.mdc-circular-progress__indeterminate-container,.mdc-circular-progress__spinner-layer{position:absolute;width:100%;height:100%}.mdc-circular-progress__determinate-container{transform:rotate(-90deg)}.mdc-circular-progress--indeterminate .mdc-circular-progress__determinate-container{opacity:0}.mdc-circular-progress__indeterminate-container{font-size:0;letter-spacing:0;white-space:nowrap;opacity:0}.mdc-circular-progress--indeterminate .mdc-circular-progress__indeterminate-container{opacity:1;animation:mdc-circular-progress-container-rotate 1568.2352941176ms linear infinite}.mdc-circular-progress__determinate-circle-graphic,.mdc-circular-progress__indeterminate-circle-graphic{fill:rgba(0,0,0,0)}.mat-mdc-progress-spinner .mdc-circular-progress__determinate-circle,.mat-mdc-progress-spinner .mdc-circular-progress__indeterminate-circle-graphic{stroke:var(--mdc-circular-progress-active-indicator-color, var(--mat-sys-primary))}@media(forced-colors: active){.mat-mdc-progress-spinner .mdc-circular-progress__determinate-circle,.mat-mdc-progress-spinner .mdc-circular-progress__indeterminate-circle-graphic{stroke:CanvasText}}.mdc-circular-progress__determinate-circle{transition:stroke-dashoffset 500ms cubic-bezier(0, 0, 0.2, 1)}.mdc-circular-progress__gap-patch{position:absolute;top:0;left:47.5%;box-sizing:border-box;width:5%;height:100%;overflow:hidden}.mdc-circular-progress__gap-patch .mdc-circular-progress__indeterminate-circle-graphic{left:-900%;width:2000%;transform:rotate(180deg)}.mdc-circular-progress__circle-clipper .mdc-circular-progress__indeterminate-circle-graphic{width:200%}.mdc-circular-progress__circle-right .mdc-circular-progress__indeterminate-circle-graphic{left:-100%}.mdc-circular-progress--indeterminate .mdc-circular-progress__circle-left .mdc-circular-progress__indeterminate-circle-graphic{animation:mdc-circular-progress-left-spin 1333ms cubic-bezier(0.4, 0, 0.2, 1) infinite both}.mdc-circular-progress--indeterminate .mdc-circular-progress__circle-right .mdc-circular-progress__indeterminate-circle-graphic{animation:mdc-circular-progress-right-spin 1333ms cubic-bezier(0.4, 0, 0.2, 1) infinite both}.mdc-circular-progress__circle-clipper{display:inline-flex;position:relative;width:50%;height:100%;overflow:hidden}.mdc-circular-progress--indeterminate .mdc-circular-progress__spinner-layer{animation:mdc-circular-progress-spinner-layer-rotate 5332ms cubic-bezier(0.4, 0, 0.2, 1) infinite both}@keyframes mdc-circular-progress-container-rotate{to{transform:rotate(360deg)}}@keyframes mdc-circular-progress-spinner-layer-rotate{12.5%{transform:rotate(135deg)}25%{transform:rotate(270deg)}37.5%{transform:rotate(405deg)}50%{transform:rotate(540deg)}62.5%{transform:rotate(675deg)}75%{transform:rotate(810deg)}87.5%{transform:rotate(945deg)}100%{transform:rotate(1080deg)}}@keyframes mdc-circular-progress-left-spin{from{transform:rotate(265deg)}50%{transform:rotate(130deg)}to{transform:rotate(265deg)}}@keyframes mdc-circular-progress-right-spin{from{transform:rotate(-265deg)}50%{transform:rotate(-130deg)}to{transform:rotate(-265deg)}}"],encapsulation:2,changeDetection:0})}return t})();var mcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it]})}return t})();function MHA(t,e){if(t&1){let A=be();S(0,"div",1)(1,"div"),AA(2,"All eval sets"),F(),S(3,"mat-icon",2),hA("click",function(){RA(A);let n=O();return xA(n.openNewEvalSetDialog())}),AA(4,"add"),F()()}}function kHA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",3)(2,"div",4),AA(3," Create New Evaluation Set "),F(),S(4,"div",5),AA(5," An evaluation set is a curated collection of evaluation cases, where each case includes input-output examples for assessing agent performance. "),F(),S(6,"div",6),hA("click",function(){RA(A);let n=O();return xA(n.openNewEvalSetDialog())}),AA(7," Create Evaluation Set "),F()()()}}function SHA(t,e){if(t&1){let A=be();S(0,"div",8),hA("click",function(){let n=RA(A).$implicit,o=O(2);return xA(o.selectEvalSet(n))}),S(1,"div",9)(2,"span",10),AA(3,"folder"),F(),S(4,"div",11),AA(5),F()(),S(6,"div")(7,"mat-icon",12),AA(8,"chevron_right"),F()()()}if(t&2){let A=e.$implicit;G(5),Gt(A)}}function RHA(t,e){if(t&1&&(S(0,"div"),Dn(1,SHA,9,1,"div",7,to),F()),t&2){let A=O();G(),yn(A.evalsets)}}function xHA(t,e){if(t&1){let A=be();S(0,"th",29)(1,"mat-checkbox",30),hA("change",function(n){RA(A);let o=O(4);return xA(n?o.toggleAllRows():null)}),F()()}if(t&2){let A=O(4);G(),yA("checked",A.selection.hasValue()&&A.isAllSelected())("indeterminate",A.selection.hasValue()&&!A.isAllSelected())}}function LHA(t,e){if(t&1){let A=be();S(0,"td",31)(1,"mat-checkbox",32),hA("click",function(n){return RA(A),xA(n.stopPropagation())})("change",function(n){let o=RA(A).$implicit,r=O(4);return xA(n?r.selection.toggle(o):null)}),F()()}if(t&2){let A=e.$implicit,i=O(4);G(),yA("checked",i.selection.isSelected(A))}}function FHA(t,e){t&1&&(S(0,"th",29),AA(1," Case ID "),F())}function NHA(t,e){if(t&1){let A=be();S(0,"td",33),hA("click",function(){let n=RA(A).$implicit,o=O(4);return xA(o.getEvalCase(n))}),AA(1),F()}if(t&2){let A,i=e.$implicit,n=O(4);ue("selected-eval-case",i===((A=n.selectedEvalCase())==null?null:A.evalId)),G(),Et(" ",i," ")}}function _HA(t,e){t&1&&(S(0,"th",29),AA(1," Result "),F())}function GHA(t,e){if(t&1){let A=be();S(0,"button",35),hA("click",function(){RA(A);let n=O().$implicit,o=O(4);return xA(o.getSession(n))}),S(1,"span",36),AA(2),F(),S(3,"div",37),AA(4),F()()}if(t&2){let A=O().$implicit,i=O(4);yA("ngClass",i.getEvalResultForCase(A)==1?"result-btn pass":"result-btn fail"),G(2),Et(" ",i.getEvalResultForCase(A)==1?"check":"close"," "),G(2),Et("",i.getEvalResultForCase(A)==1?"Pass":"Fail"," ")}}function UHA(t,e){if(t&1&&(S(0,"td",31),_A(1,GHA,5,3,"button",34),F()),t&2){let A=e.$implicit,i=O(4);G(),GA(i.getEvalResultForCase(A)?1:-1)}}function KHA(t,e){t&1&&JA(0,"tr",38)}function YHA(t,e){t&1&&JA(0,"tr",39)}function JHA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",16)(2,"button",17),hA("click",function(){RA(A);let n=O(3);return xA(n.openEvalConfigDialog())}),AA(3,"Run Evaluation"),F(),S(4,"mat-icon",18),hA("click",function(){RA(A);let n=O(3);return xA(n.toggleEvalHistoryButton())}),AA(5,"history"),F()(),S(6,"div",19)(7,"table",20),y2(8,21),_A(9,xHA,2,2,"th",22)(10,LHA,2,1,"td",23),v2(),y2(11,24),_A(12,FHA,2,0,"th",22)(13,NHA,2,3,"td",25),v2(),y2(14,26),_A(15,_HA,2,0,"th",22)(16,UHA,2,1,"td",23),v2(),_A(17,KHA,1,0,"tr",27)(18,YHA,1,0,"tr",28),F()()()}if(t&2){let A=O(3);G(7),yA("dataSource",A.dataSource),G(10),yA("matHeaderRowDef",A.displayedColumns),G(),yA("matRowDefColumns",A.displayedColumns)}}function THA(t,e){if(t&1&&(S(0,"div")(1,"span",50),AA(2,"|"),F(),S(3,"span",51),AA(4),F()()),t&2){let A=O().$implicit,i=O(4);G(4),Et("",i.getFailCountForCurrentResult(A.evaluationResults.evaluationResults)," Failed")}}function HHA(t,e){if(t&1&&(S(0,"span",52),AA(1),F()),t&2){let A=e.$implicit;G(),Ob(" ",A.metricName,": ",A.threshold," ")}}function zHA(t,e){if(t&1&&(S(0,"div",46),Dn(1,HHA,2,2,"span",52,to),F()),t&2){let A=O().$implicit,i=O(4);G(),yn(i.getEvalMetrics(A))}}function OHA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",53)(2,"span"),AA(3),F(),S(4,"button",54),hA("click",function(){let n=RA(A).$implicit,o=O(6);return xA(o.getHistorySession(n))}),S(5,"span",36),AA(6),F(),S(7,"div",37),AA(8),F()()()()}if(t&2){let A=e.$implicit;G(3),Et(" ",A.evalId," "),G(),yA("ngClass",A.finalEvalStatus==1?"result-btn pass":"result-btn fail"),G(2),Et(" ",A.finalEvalStatus==1?"check":"close"," "),G(2),Et("",A.finalEvalStatus==1?"PASS":"FAIL"," ")}}function PHA(t,e){if(t&1&&(S(0,"div",49),Dn(1,OHA,9,4,"div",null,to),F()),t&2){let A=O().$implicit,i=O(4);G(),yn(i.generateHistoryEvaluationDatasource(A.timestamp))}}function jHA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",40)(2,"div",41)(3,"div",42)(4,"div",43),AA(5),F(),S(6,"div",44)(7,"span",45),AA(8),F(),_A(9,THA,5,1,"div"),F(),_A(10,zHA,3,0,"div",46),F(),S(11,"div",47)(12,"mat-icon",48),hA("click",function(){let n=RA(A).$implicit,o=O(4);return xA(o.toggleHistoryStatusCard(n.timestamp))}),AA(13),F()()(),_A(14,PHA,3,0,"div",49),F()()}if(t&2){let A=e.$implicit,i=O(4);G(5),Gt(i.formatTimestamp(A.timestamp)),G(3),Et("",i.getPassCountForCurrentResult(A.evaluationResults.evaluationResults)," Passed"),G(),GA(i.getFailCountForCurrentResult(A.evaluationResults.evaluationResults)>0?9:-1),G(),GA(i.getEvalMetrics(A)?10:-1),G(3),Gt(i.getEvaluationStatusCardActionButtonIcon(A.timestamp)),G(),GA(i.isEvaluationStatusCardToggled(A.timestamp)?14:-1)}}function qHA(t,e){if(t&1&&(S(0,"div"),Dn(1,jHA,15,6,"div",null,to),F()),t&2){let A=O(3);G(),yn(A.getEvalHistoryOfCurrentSetSorted())}}function VHA(t,e){if(t&1&&(S(0,"div"),_A(1,JHA,19,3,"div")(2,qHA,3,0,"div"),F()),t&2){let A=O(2);G(),GA(A.showEvalHistory()?-1:1),G(),GA(A.showEvalHistory()?2:-1)}}function ZHA(t,e){if(t&1){let A=be();S(0,"button",55),hA("click",function(){RA(A);let n=O(2);return xA(n.openNewEvalCaseDialog())}),S(1,"div",56)(2,"mat-icon"),AA(3,"add"),F(),S(4,"div",57),AA(5),F()()()}if(t&2){let A=O(2);G(5),Et(" Add current session to ",A.selectedEvalSet," ")}}function WHA(t,e){t&1&&(S(0,"div"),JA(1,"mat-spinner",58),F()),t&2&&(G(),yA("diameter",28)("strokeWidth",3))}function XHA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",9)(2,"mat-icon",13),hA("click",function(){RA(A);let n=O();return xA(n.clearSelectedEvalSet())}),AA(3,"chevron_left"),F(),S(4,"div",14),hA("click",function(){RA(A);let n=O();return xA(n.clearSelectedEvalSet())}),AA(5),F()(),_A(6,VHA,3,2,"div")(7,ZHA,6,1,"button",15)(8,WHA,2,2,"div"),F()}if(t&2){let A=O();G(5),Et(" ",A.selectedEvalSet," "),G(),GA(A.evalCases.length>0&&!A.evalRunning()?6:-1),G(),GA(!A.evalRunning()&&!A.showEvalHistory()?7:-1),G(),GA(A.evalRunning()?8:-1)}}var id=class t{constructor(e,A){this.evalService=e;this.sessionService=A;this.evalCasesSubject.subscribe(i=>{!this.selectedEvalCase()&&this.deletedEvalCaseIndex>=0&&i.length>0?(this.selectNewEvalCase(i),this.deletedEvalCaseIndex=-1):i.length===0&&this.shouldReturnToSession.emit(!0)})}checkboxes;appName="";userId="";sessionId="";sessionSelected=new WA;shouldShowTab=new WA;evalNotInstalledMsg=new WA;evalCaseSelected=new WA;evalSetIdSelected=new WA;shouldReturnToSession=new WA;evalCasesSubject=new Mi([]);changeDetectorRef=f(It);flagService=f(KB);displayedColumns=["select","evalId","finalEvalStatus"];evalsets=[];selectedEvalSet="";evalCases=[];selectedEvalCase=Jo(null);deletedEvalCaseIndex=-1;dataSource=new Cf(this.evalCases);selection=new K2(!0,[]);showEvalHistory=Jo(!1);evalRunning=Jo(!1);evalMetrics=acA;currentEvalResultBySet=new Map;dialog=f(qs);appEvaluationResults={};ngOnChanges(e){e.appName&&(this.selectedEvalSet="",this.evalCases=[],this.getEvalSet(),this.getEvaluationResult())}ngOnInit(){}selectNewEvalCase(e){let A=this.deletedEvalCaseIndex;this.deletedEvalCaseIndex===e.length&&(A=0),this.getEvalCase(e[A])}getEvalSet(){this.appName!=""&&this.evalService.getEvalSets(this.appName).pipe(mr(e=>e.status===404&&e.statusText==="Not Found"?(this.shouldShowTab.emit(!1),Me(null)):Me([]))).subscribe(e=>{e!==null&&(this.shouldShowTab.emit(!0),this.evalsets=e)})}openNewEvalSetDialog(){this.dialog.open(Ef,{width:"600px",data:{appName:this.appName}}).afterClosed().subscribe(A=>{A&&this.getEvalSet()})}openNewEvalCaseDialog(){this.dialog.open(Bf,{width:"600px",data:{appName:this.appName,userId:this.userId,sessionId:this.sessionId,evalSetId:this.selectedEvalSet}}).afterClosed().subscribe(A=>{A&&this.listEvalCases()})}listEvalCases(){this.evalCases=[],this.evalService.listEvalCases(this.appName,this.selectedEvalSet).subscribe(e=>{this.evalCases=e,this.dataSource=new Cf(this.evalCases),this.evalCasesSubject.next(this.evalCases),this.changeDetectorRef.detectChanges()})}runEval(){if(this.evalRunning.set(!0),this.selection.selected.length==0){alert("No case selected!"),this.evalRunning.set(!1);return}this.evalService.runEval(this.appName,this.selectedEvalSet,this.selection.selected,this.evalMetrics).pipe(mr(e=>(e.error?.detail?.includes("not installed")&&this.evalNotInstalledMsg.emit(e.error.detail),Me([])))).subscribe(e=>{this.evalRunning.set(!1),this.currentEvalResultBySet.set(this.selectedEvalSet,e),this.getEvaluationResult()})}selectEvalSet(e){this.selectedEvalSet=e,this.listEvalCases()}clearSelectedEvalSet(){if(this.showEvalHistory()){this.toggleEvalHistoryButton();return}this.selectedEvalSet=""}isAllSelected(){let e=this.selection.selected.length,A=this.dataSource.data.length;return e===A}toggleAllRows(){if(this.isAllSelected()){this.selection.clear();return}this.selection.select(...this.dataSource.data)}getEvalResultForCase(e){let A=this.currentEvalResultBySet.get(this.selectedEvalSet)?.filter(i=>i.evalId==e);if(!(!A||A.length==0))return A[0].finalEvalStatus}formatToolUses(e){let A=[];for(let i of e)A.push({name:i.name,args:i.args});return A}addEvalCaseResultToEvents(e,A){let i=A.evalMetricResultPerInvocation,n=-1;if(i)for(let o=0;o<e.events.length;o++){let r=e.events[o];if(r.author==="user")n++;else{let s=i[n],a=1,c="",l=1,I=1;for(let C of s.evalMetricResults)if(C.evalStatus===2){a=2,c=C.metricName,l=C.score,I=C.threshold;break}r.evalStatus=a,(o===e.events.length-1||e.events[o+1].author==="user")&&this.addEvalFieldsToBotEvent(r,s,c,l,I)}}return e}addEvalFieldsToBotEvent(e,A,i,n,o){e.failedMetric=i,e.evalScore=n,e.evalThreshold=o,e.failedMetric==="tool_trajectory_avg_score"?(e.actualInvocationToolUses=this.formatToolUses(A.actualInvocation.intermediateData.toolUses),e.expectedInvocationToolUses=this.formatToolUses(A.expectedInvocation.intermediateData.toolUses)):e.failedMetric==="response_match_score"&&(e.actualFinalResponse=A.actualInvocation.finalResponse.parts[0].text,e.expectedFinalResponse=A.expectedInvocation.finalResponse.parts[0]?.text)}fromApiResultToSession(e){return{id:e?.id??"",appName:e?.appName??"",userId:e?.userId??"",state:e?.state??[],events:e?.events??[]}}getSession(e){let A=this.currentEvalResultBySet.get(this.selectedEvalSet)?.filter(n=>n.evalId==e)[0],i=A.sessionId;this.sessionService.getSession(this.userId,this.appName,i).subscribe(n=>{this.addEvalCaseResultToEvents(n,A);let o=this.fromApiResultToSession(n);this.sessionSelected.emit(o)})}toggleEvalHistoryButton(){this.showEvalHistory.set(!this.showEvalHistory())}getEvalHistoryOfCurrentSet(){return this.appEvaluationResults[this.appName][this.selectedEvalSet]}getEvalHistoryOfCurrentSetSorted(){let e=this.getEvalHistoryOfCurrentSet();return Object.keys(e).sort((n,o)=>o.localeCompare(n)).map(n=>({timestamp:n,evaluationResults:e[n]}))}getPassCountForCurrentResult(e){return e.filter(A=>A.finalEvalStatus==1).length}getFailCountForCurrentResult(e){return e.filter(A=>A.finalEvalStatus==2).length}formatTimestamp(e){let A=Number(e);if(isNaN(A))return"Invalid timestamp provided";let i=new Date(A*1e3);if(isNaN(i.getTime()))return"Invalid date created from timestamp";let n={month:"short",day:"numeric",year:"numeric",hour:"numeric",minute:"2-digit",hour12:!0};return new Intl.DateTimeFormat("en-US",n).format(i)}getEvaluationStatusCardActionButtonIcon(e){return this.getEvalHistoryOfCurrentSet()[e].isToggled?"keyboard_arrow_up":"keyboard_arrow_down"}toggleHistoryStatusCard(e){this.getEvalHistoryOfCurrentSet()[e].isToggled=!this.getEvalHistoryOfCurrentSet()[e].isToggled}isEvaluationStatusCardToggled(e){return this.getEvalHistoryOfCurrentSet()[e].isToggled}generateHistoryEvaluationDatasource(e){return this.getEvalHistoryOfCurrentSet()[e].evaluationResults}getHistorySession(e){this.addEvalCaseResultToEvents(e.sessionDetails,e);let A=this.fromApiResultToSession(e.sessionDetails);this.sessionSelected.emit(A)}getEvalCase(e){this.evalService.getEvalCase(this.appName,this.selectedEvalSet,e).subscribe(A=>{this.selectedEvalCase.set(A),this.evalCaseSelected.emit(A),this.evalSetIdSelected.emit(this.selectedEvalSet)})}resetEvalCase(){this.selectedEvalCase.set(null)}resetEvalResults(){this.currentEvalResultBySet.clear()}deleteEvalCase(e){this.evalService.deleteEvalCase(this.appName,this.selectedEvalSet,e).subscribe(A=>{this.deletedEvalCaseIndex=this.evalCases.indexOf(e),this.selectedEvalCase.set(null),this.listEvalCases()})}getEvaluationResult(){this.evalService.listEvalResults(this.appName).pipe(mr(e=>e.status===404&&e.statusText==="Not Found"?(this.shouldShowTab.emit(!1),Me(null)):Me([]))).subscribe(e=>{for(let A of e)this.evalService.getEvalResult(this.appName,A).subscribe(i=>{this.appEvaluationResults[this.appName]||(this.appEvaluationResults[this.appName]={}),this.appEvaluationResults[this.appName][i.evalSetId]||(this.appEvaluationResults[this.appName][i.evalSetId]={});let n=i.creationTimestamp;this.appEvaluationResults[this.appName][i.evalSetId][n]||(this.appEvaluationResults[this.appName][i.evalSetId][n]={isToggled:!1,evaluationResults:[]});let o={isToggled:!1,evaluationResults:i.evalCaseResults.map(r=>({setId:r.id,evalId:r.evalId,finalEvalStatus:r.finalEvalStatus,evalMetricResults:r.evalMetricResults,evalMetricResultPerInvocation:r.evalMetricResultPerInvocation,sessionId:r.sessionId,sessionDetails:r.sessionDetails,overallEvalMetricResults:r.overallEvalMetricResults??[]}))};this.appEvaluationResults[this.appName][i.evalSetId][n]=o})})}openEvalConfigDialog(){if(this.selection.selected.length==0){alert("No case selected!");return}this.dialog.open(Qf,{maxWidth:"90vw",maxHeight:"90vh",data:{evalMetrics:this.evalMetrics}}).afterClosed().subscribe(A=>{A&&(this.evalMetrics=A,this.runEval())})}getEvalMetrics(e){if(!e||!e.evaluationResults||!e.evaluationResults.evaluationResults)return this.evalMetrics;let A=e.evaluationResults.evaluationResults;return A.length===0?this.evalMetrics:typeof A[0].overallEvalMetricResults>"u"||!A[0].overallEvalMetricResults||A[0].overallEvalMetricResults.length===0?this.evalMetrics:A[0].overallEvalMetricResults.map(n=>({metricName:n.metricName,threshold:n.threshold}))}static \u0275fac=function(A){return new(A||t)(PA(Vc),PA(Wg))};static \u0275cmp=zA({type:t,selectors:[["app-eval-tab"]],viewQuery:function(A,i){if(A&1&&Te(bQ,5),A&2){let n;XA(n=$A())&&(i.checkboxes=n)}},inputs:{appName:"appName",userId:"userId",sessionId:"sessionId"},outputs:{sessionSelected:"sessionSelected",shouldShowTab:"shouldShowTab",evalNotInstalledMsg:"evalNotInstalledMsg",evalCaseSelected:"evalCaseSelected",evalSetIdSelected:"evalSetIdSelected",shouldReturnToSession:"shouldReturnToSession"},standalone:!1,features:[ti],decls:5,vars:4,consts:[[1,"eval-container"],[1,"eval-set-actions"],["matTooltip","Create new evaluation set",2,"cursor","pointer",3,"click"],[1,"empty-eval-info"],[1,"info-title"],[1,"info-detail"],[1,"info-create",3,"click"],[1,"eval-set-row"],[1,"eval-set-row",3,"click"],[2,"display","flex"],[1,"material-symbols-outlined",2,"margin-right","10px","padding-top","16px"],[2,"font-family","Roboto","font-size","14px","padding","16px","padding-top","20px"],[2,"padding-top","20px","color","#9AA0A6"],[2,"color","white","cursor","pointer",3,"click"],[2,"color","#9AA0A6","padding-top","2px","cursor","pointer",3,"click"],[1,"save-session-btn"],[1,"evaluation-tab-header"],[1,"run-eval-btn",3,"click"],["matTooltip","View eval run history",1,"evaluation-history-icon",3,"click"],[1,"mat-table-container",2,"margin-top","16px"],["mat-table","",3,"dataSource"],["matColumnDef","select"],["mat-header-cell","",4,"matHeaderCellDef"],["mat-cell","",4,"matCellDef"],["matColumnDef","evalId"],["mat-cell","","class","eval-case-id",3,"selected-eval-case","click",4,"matCellDef"],["matColumnDef","finalEvalStatus"],["mat-header-row","",4,"matHeaderRowDef"],["mat-row","",4,"matRowDef","matRowDefColumns"],["mat-header-cell",""],[3,"change","checked","indeterminate"],["mat-cell",""],[3,"click","change","checked"],["mat-cell","",1,"eval-case-id",3,"click"],["matTooltip","View eval run result",3,"ngClass"],["matTooltip","View eval run result",3,"click","ngClass"],[1,"material-symbols-outlined"],[2,"padding-top","4px"],["mat-header-row",""],["mat-row",""],[1,"status-card"],[1,"status-card__overview"],[1,"status-card__info"],[1,"status-card__timestamp"],[1,"status-card__summary"],[1,"status-card__passed"],[1,"status-card__metrics"],[1,"status-card__action"],[3,"click"],[1,"status-card__history-cases"],[1,"status-card__separator"],[1,"status-card__failed"],[1,"status-card__metric"],[1,"status-card__history-case"],[3,"click","ngClass"],[1,"save-session-btn",3,"click"],[1,"save-session-btn-detail"],[1,"save-session-btn-text"],[1,"eval-spinner",3,"diameter","strokeWidth"]],template:function(A,i){A&1&&(S(0,"div",0),_A(1,MHA,5,0,"div",1)(2,kHA,8,0,"div")(3,RHA,3,0,"div")(4,XHA,9,4,"div"),F()),A&2&&(G(),GA(i.selectedEvalSet==""?1:-1),G(),GA(i.evalsets.length==0?2:-1),G(),GA(i.evalsets.length>0&&i.selectedEvalSet==""?3:-1),G(),GA(i.selectedEvalSet!=""?4:-1))},dependencies:[Xa,P2,bQ,WaA,$aA,icA,AcA,XaA,ncA,ecA,tcA,ocA,rcA,_B,fcA],styles:[".eval-container[_ngcontent-%COMP%]{margin-top:20px;padding-left:25px;padding-right:25px}.eval-case-id[_ngcontent-%COMP%]{cursor:pointer}.eval-set-actions[_ngcontent-%COMP%]{display:flex;justify-content:space-between;color:#9aa0a6;font-style:normal;font-weight:700;font-size:14px}.empty-eval-info[_ngcontent-%COMP%]{margin-top:12px;background-color:#202124;border-radius:8px;box-shadow:0 2px 6px 2px #00000026,0 1px 2px #0000004d}.info-title[_ngcontent-%COMP%]{color:#e8eaed;font-family:Roboto;font-size:14px;font-weight:500;padding-top:13px;padding-right:16px;padding-left:16px}.info-detail[_ngcontent-%COMP%]{color:#e8eaed;font-family:Roboto;font-size:14px;font-weight:400;padding-top:13px;padding-right:16px;padding-left:16px;letter-spacing:.2px}.info-create[_ngcontent-%COMP%]{color:var(--Blue-300, #8ab4f8);font-size:14px;font-style:normal;font-weight:500;padding-right:16px;padding-left:16px;margin-top:19px;padding-bottom:16px;cursor:pointer}.eval-set-row[_ngcontent-%COMP%]{display:flex;justify-content:space-between;cursor:pointer}.selected-eval-case[_ngcontent-%COMP%]{font-weight:900;color:#8ab4f8}.save-session-btn[_ngcontent-%COMP%]{width:100%;background:linear-gradient(0deg,#8ab4f83d 0% 100%),#202124;border:none;border-radius:4px;margin-top:12px;cursor:pointer}.save-session-btn-detail[_ngcontent-%COMP%]{display:flex;padding:8px 16px 8px 12px;justify-content:center}.save-session-btn-text[_ngcontent-%COMP%]{padding-top:2px;color:var(--Blue-100, #d2e3fc);font-family:Google Sans;font-size:14px;font-style:normal;font-weight:500;line-height:20px;letter-spacing:.25px}.run-eval-btn[_ngcontent-%COMP%]{border-radius:4px;border:1px solid var(--Grey-700, #5f6368);background-color:transparent;padding:8px 24px;margin-top:16px;color:#8ab4f8;cursor:pointer}.run-eval-btn[_ngcontent-%COMP%]:hover{background-color:#202124}.result-btn[_ngcontent-%COMP%]{display:flex;background-color:transparent;border-radius:4px;border:1px solid var(--Grey-700, #5f6368);margin-top:4px;cursor:pointer}.result-btn[_ngcontent-%COMP%]:hover{background-color:#202124}.result-btn.pass[_ngcontent-%COMP%]{color:#44c265}.result-btn.fail[_ngcontent-%COMP%]{color:#ff8983}.evaluation-tab-header[_ngcontent-%COMP%]{display:flex;justify-content:space-between;align-items:center;width:100%}.evaluation-history-icon[_ngcontent-%COMP%]{cursor:pointer;margin-top:4px}.status-card[_ngcontent-%COMP%]{display:flex;flex-direction:column;align-items:center;border-radius:8px;background-color:#2d2d2d;padding:12px 16px;margin-top:12px}.status-card__overview[_ngcontent-%COMP%]{display:flex;justify-content:space-between;align-items:center;width:100%}.status-card__info[_ngcontent-%COMP%]{display:flex;flex-direction:column}.status-card__timestamp[_ngcontent-%COMP%]{font-size:.9em;color:#e0e0e0;margin-bottom:5px}.status-card__summary[_ngcontent-%COMP%]{display:flex;align-items:center;font-size:.95em;font-weight:500}.status-card__metrics[_ngcontent-%COMP%]{display:flex;align-items:center;font-size:.75em;font-weight:300;margin-top:3px}.status-card__metric[_ngcontent-%COMP%]{width:180px;color:#bbb}.status-card__failed[_ngcontent-%COMP%]{color:#ff6b6b}.status-card__separator[_ngcontent-%COMP%]{color:#666;margin:0 8px}.status-card__passed[_ngcontent-%COMP%]{color:#63e6be}.status-card__action[_ngcontent-%COMP%]{display:flex;align-items:center}.status-card__action[_ngcontent-%COMP%]   mat-icon[_ngcontent-%COMP%]{color:#bdbdbd;cursor:pointer;transition:transform .2s ease-in-out}.status-card__action[_ngcontent-%COMP%]   mat-icon[_ngcontent-%COMP%]:hover{opacity:.8}.status-card__action[_ngcontent-%COMP%]   .status-card__icon[_ngcontent-%COMP%]{color:#bdbdbd;font-size:1.2em;cursor:pointer}.status-card__action[_ngcontent-%COMP%]   .status-card__icon[_ngcontent-%COMP%]:hover{opacity:.8}.status-card__history-cases[_ngcontent-%COMP%]{display:flex;flex-direction:column;margin-top:3px;justify-content:flex-start;width:100%}.status-card__history-case[_ngcontent-%COMP%]{display:flex;justify-content:space-between;align-items:center;width:100%;margin-top:15px}.eval-spinner[_ngcontent-%COMP%]{margin-top:12px}"],changeDetection:0})};function AzA(t,e){t&1&&JA(0,"div",6)}function ezA(t,e){if(t&1&&(S(0,"div",3)(1,"div",5),Dn(2,AzA,1,0,"div",6,Xd),F(),S(4,"span",7),AA(5),F(),S(6,"div",8),AA(7),S(8,"span",9),AA(9),F()(),S(10,"div",10)(11,"div",11),AA(12),F()()()),t&2){let A=e.$implicit,i=O();G(2),yn(i.getArray(A.level)),G(3),Et(" ",i.getSpanIcon(A.span.name)," "),G(),uo("width",400-A.level*20,"px"),G(),Et(" ",A.span.name," "),G(2),Et(" (",(i.toMs(A.span.end_time)-i.toMs(A.span.start_time)).toFixed(2),"ms) "),G(2),uo("left",i.getRelativeStart(A.span),"%")("width",i.getRelativeWidth(A.span),"%"),G(),Et(" ",(i.toMs(A.span.end_time)-i.toMs(A.span.start_time)).toFixed(2),"ms ")}}var hf=class t{constructor(e,A){this.dialogRef=e;this.data=A}tree=[];baseStartTimeMs=0;totalDurationMs=1;flatTree=[];traceLabelIconMap=new Map([["Invocation","start"],["agent_run","directions_run"],["tool","build"],["call_llm","chat"]]);ngOnInit(){this.tree=this.buildSpanTree(this.data.spans),this.flatTree=this.flattenTree(this.tree);let e=this.getGlobalTimes(this.data.spans);this.baseStartTimeMs=e.start,this.totalDurationMs=e.duration}buildSpanTree(e){let A=e.map(o=>rA({},o)),i=new Map,n=[];return A.forEach(o=>i.set(o.span_id,o)),A.forEach(o=>{if(o.parent_span_id&&i.has(o.parent_span_id)){let r=i.get(o.parent_span_id);r.children=r.children||[],r.children.push(o)}else n.push(o)}),n}getGlobalTimes(e){let A=Math.min(...e.map(n=>this.toMs(n.start_time))),i=Math.max(...e.map(n=>this.toMs(n.end_time)));return{start:A,duration:i-A}}toMs(e){return e/1e6}getRelativeStart(e){return(this.toMs(e.start_time)-this.baseStartTimeMs)/this.totalDurationMs*100}getRelativeWidth(e){return(this.toMs(e.end_time)-this.toMs(e.start_time))/this.totalDurationMs*100}flattenTree(e,A=0){return e.flatMap(n=>[{span:n,level:A},...n.children?this.flattenTree(n.children,A+1):[]])}getSpanIcon(e){for(let[A,i]of this.traceLabelIconMap.entries())if(e.startsWith(A))return i;return"start"}getArray(e){return Array.from({length:e})}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(as))};static \u0275cmp=zA({type:t,selectors:[["app-trace-chart"]],standalone:!1,decls:9,vars:1,consts:[["mat-dialog-title",""],[2,"margin-top","8px"],[1,"trace-container"],[1,"trace-row"],["mat-button","","mat-dialog-close",""],[1,"trace-indent"],[1,"indent-connector"],[1,"material-symbols-outlined",2,"margin-right","8px"],[1,"trace-label"],[1,"trace-duration"],[1,"trace-bar-container"],[1,"trace-bar"]],template:function(A,i){A&1&&(S(0,"h2",0),AA(1),F(),S(2,"mat-dialog-content",1)(3,"div",2),Dn(4,ezA,13,10,"div",3,to),F()(),S(6,"mat-dialog-actions")(7,"button",4),AA(8,"Close"),F()()),A&2&&(G(),Et("Invocation ",i.data.invocId,""),G(3),yn(i.flatTree))},dependencies:[Dr,bs,ma,pa,hg],styles:[".trace-container[_ngcontent-%COMP%]{width:100%;white-space:nowrap;font-size:12px}.trace-label[_ngcontent-%COMP%]{width:400px;color:#e3e3e3;text-overflow:ellipsis;font-family:Google Sans Mono,monospace;font-size:14px;font-style:normal;font-weight:500;line-height:20px;letter-spacing:0px}.trace-bar-container[_ngcontent-%COMP%]{width:50vw;position:relative;height:16px}.trace-bar[_ngcontent-%COMP%]{position:absolute;height:18px;background-color:#2f4d65;border-radius:4px;padding-left:4px;overflow:hidden;font-size:11px;line-height:16px;color:#8dabbf;font-family:Google Sans}.trace-duration[_ngcontent-%COMP%]{color:#888;font-weight:400;margin-left:4px}.trace-row[_ngcontent-%COMP%]{display:flex;align-items:stretch;position:relative;height:32px}.trace-indent[_ngcontent-%COMP%]{display:flex;flex-shrink:0;height:100%}.indent-connector[_ngcontent-%COMP%]{width:20px;position:relative;height:100%}.vertical-line[_ngcontent-%COMP%]{position:absolute;top:0;bottom:0;left:9px;width:1px;background-color:#ccc}.horizontal-line[_ngcontent-%COMP%]{position:absolute;top:50%;left:9px;width:10px;height:1px;background-color:#ccc}"]})};var pcA=(()=>{class t{get vertical(){return this._vertical}set vertical(A){this._vertical=Xo(A)}_vertical=!1;get inset(){return this._inset}set inset(A){this._inset=Xo(A)}_inset=!1;static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-divider"]],hostAttrs:["role","separator",1,"mat-divider"],hostVars:7,hostBindings:function(i,n){i&2&&(Ne("aria-orientation",n.vertical?"vertical":"horizontal"),ue("mat-divider-vertical",n.vertical)("mat-divider-horizontal",!n.vertical)("mat-divider-inset",n.inset))},inputs:{vertical:"vertical",inset:"inset"},decls:0,vars:0,template:function(i,n){},styles:[".mat-divider{display:block;margin:0;border-top-style:solid;border-top-color:var(--mat-divider-color, var(--mat-sys-outline));border-top-width:var(--mat-divider-width, 1px)}.mat-divider.mat-divider-vertical{border-top:0;border-right-style:solid;border-right-color:var(--mat-divider-color, var(--mat-sys-outline));border-right-width:var(--mat-divider-width, 1px)}.mat-divider.mat-divider-inset{margin-left:80px}[dir=rtl] .mat-divider.mat-divider-inset{margin-left:auto;margin-right:80px}"],encapsulation:2,changeDetection:0})}return t})(),wcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,it]})}return t})();var izA=["*"],nzA='.mdc-list{margin:0;padding:8px 0;list-style-type:none}.mdc-list:focus{outline:none}.mdc-list-item{display:flex;position:relative;justify-content:flex-start;overflow:hidden;padding:0;align-items:stretch;cursor:pointer;padding-left:16px;padding-right:16px;background-color:var(--mdc-list-list-item-container-color, transparent);border-radius:var(--mdc-list-list-item-container-shape, var(--mat-sys-corner-none))}.mdc-list-item.mdc-list-item--selected{background-color:var(--mdc-list-list-item-selected-container-color)}.mdc-list-item:focus{outline:0}.mdc-list-item.mdc-list-item--disabled{cursor:auto}.mdc-list-item.mdc-list-item--with-one-line{height:var(--mdc-list-list-item-one-line-container-height, 48px)}.mdc-list-item.mdc-list-item--with-one-line .mdc-list-item__start{align-self:center;margin-top:0}.mdc-list-item.mdc-list-item--with-one-line .mdc-list-item__end{align-self:center;margin-top:0}.mdc-list-item.mdc-list-item--with-two-lines{height:var(--mdc-list-list-item-two-line-container-height, 64px)}.mdc-list-item.mdc-list-item--with-two-lines .mdc-list-item__start{align-self:flex-start;margin-top:16px}.mdc-list-item.mdc-list-item--with-two-lines .mdc-list-item__end{align-self:center;margin-top:0}.mdc-list-item.mdc-list-item--with-three-lines{height:var(--mdc-list-list-item-three-line-container-height, 88px)}.mdc-list-item.mdc-list-item--with-three-lines .mdc-list-item__start{align-self:flex-start;margin-top:16px}.mdc-list-item.mdc-list-item--with-three-lines .mdc-list-item__end{align-self:flex-start;margin-top:16px}.mdc-list-item.mdc-list-item--selected::before,.mdc-list-item.mdc-list-item--selected:focus::before,.mdc-list-item:not(.mdc-list-item--selected):focus::before{position:absolute;box-sizing:border-box;width:100%;height:100%;top:0;left:0;content:"";pointer-events:none}a.mdc-list-item{color:inherit;text-decoration:none}.mdc-list-item__start{fill:currentColor;flex-shrink:0;pointer-events:none}.mdc-list-item--with-leading-icon .mdc-list-item__start{color:var(--mdc-list-list-item-leading-icon-color, var(--mat-sys-on-surface-variant));width:var(--mdc-list-list-item-leading-icon-size, 24px);height:var(--mdc-list-list-item-leading-icon-size, 24px);margin-left:16px;margin-right:32px}[dir=rtl] .mdc-list-item--with-leading-icon .mdc-list-item__start{margin-left:32px;margin-right:16px}.mdc-list-item--with-leading-icon:hover .mdc-list-item__start{color:var(--mdc-list-list-item-hover-leading-icon-color)}.mdc-list-item--with-leading-avatar .mdc-list-item__start{width:var(--mdc-list-list-item-leading-avatar-size, 40px);height:var(--mdc-list-list-item-leading-avatar-size, 40px);margin-left:16px;margin-right:16px;border-radius:50%}.mdc-list-item--with-leading-avatar .mdc-list-item__start,[dir=rtl] .mdc-list-item--with-leading-avatar .mdc-list-item__start{margin-left:16px;margin-right:16px;border-radius:50%}.mdc-list-item__end{flex-shrink:0;pointer-events:none}.mdc-list-item--with-trailing-meta .mdc-list-item__end{font-family:var(--mdc-list-list-item-trailing-supporting-text-font, var(--mat-sys-label-small-font));line-height:var(--mdc-list-list-item-trailing-supporting-text-line-height, var(--mat-sys-label-small-line-height));font-size:var(--mdc-list-list-item-trailing-supporting-text-size, var(--mat-sys-label-small-size));font-weight:var(--mdc-list-list-item-trailing-supporting-text-weight, var(--mat-sys-label-small-weight));letter-spacing:var(--mdc-list-list-item-trailing-supporting-text-tracking, var(--mat-sys-label-small-tracking))}.mdc-list-item--with-trailing-icon .mdc-list-item__end{color:var(--mdc-list-list-item-trailing-icon-color, var(--mat-sys-on-surface-variant));width:var(--mdc-list-list-item-trailing-icon-size, 24px);height:var(--mdc-list-list-item-trailing-icon-size, 24px)}.mdc-list-item--with-trailing-icon:hover .mdc-list-item__end{color:var(--mdc-list-list-item-hover-trailing-icon-color)}.mdc-list-item.mdc-list-item--with-trailing-meta .mdc-list-item__end{color:var(--mdc-list-list-item-trailing-supporting-text-color, var(--mat-sys-on-surface-variant))}.mdc-list-item--selected.mdc-list-item--with-trailing-icon .mdc-list-item__end{color:var(--mdc-list-list-item-selected-trailing-icon-color, var(--mat-sys-primary))}.mdc-list-item__content{text-overflow:ellipsis;white-space:nowrap;overflow:hidden;align-self:center;flex:1;pointer-events:none}.mdc-list-item--with-two-lines .mdc-list-item__content,.mdc-list-item--with-three-lines .mdc-list-item__content{align-self:stretch}.mdc-list-item__primary-text{text-overflow:ellipsis;white-space:nowrap;overflow:hidden;color:var(--mdc-list-list-item-label-text-color, var(--mat-sys-on-surface));font-family:var(--mdc-list-list-item-label-text-font, var(--mat-sys-body-large-font));line-height:var(--mdc-list-list-item-label-text-line-height, var(--mat-sys-body-large-line-height));font-size:var(--mdc-list-list-item-label-text-size, var(--mat-sys-body-large-size));font-weight:var(--mdc-list-list-item-label-text-weight, var(--mat-sys-body-large-weight));letter-spacing:var(--mdc-list-list-item-label-text-tracking, var(--mat-sys-body-large-tracking))}.mdc-list-item:hover .mdc-list-item__primary-text{color:var(--mdc-list-list-item-hover-label-text-color, var(--mat-sys-on-surface))}.mdc-list-item:focus .mdc-list-item__primary-text{color:var(--mdc-list-list-item-focus-label-text-color, var(--mat-sys-on-surface))}.mdc-list-item--with-two-lines .mdc-list-item__primary-text,.mdc-list-item--with-three-lines .mdc-list-item__primary-text{display:block;margin-top:0;line-height:normal;margin-bottom:-20px}.mdc-list-item--with-two-lines .mdc-list-item__primary-text::before,.mdc-list-item--with-three-lines .mdc-list-item__primary-text::before{display:inline-block;width:0;height:28px;content:"";vertical-align:0}.mdc-list-item--with-two-lines .mdc-list-item__primary-text::after,.mdc-list-item--with-three-lines .mdc-list-item__primary-text::after{display:inline-block;width:0;height:20px;content:"";vertical-align:-20px}.mdc-list-item__secondary-text{text-overflow:ellipsis;white-space:nowrap;overflow:hidden;display:block;margin-top:0;color:var(--mdc-list-list-item-supporting-text-color, var(--mat-sys-on-surface-variant));font-family:var(--mdc-list-list-item-supporting-text-font, var(--mat-sys-body-medium-font));line-height:var(--mdc-list-list-item-supporting-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mdc-list-list-item-supporting-text-size, var(--mat-sys-body-medium-size));font-weight:var(--mdc-list-list-item-supporting-text-weight, var(--mat-sys-body-medium-weight));letter-spacing:var(--mdc-list-list-item-supporting-text-tracking, var(--mat-sys-body-medium-tracking))}.mdc-list-item__secondary-text::before{display:inline-block;width:0;height:20px;content:"";vertical-align:0}.mdc-list-item--with-three-lines .mdc-list-item__secondary-text{white-space:normal;line-height:20px}.mdc-list-item--with-overline .mdc-list-item__secondary-text{white-space:nowrap;line-height:auto}.mdc-list-item--with-leading-radio.mdc-list-item,.mdc-list-item--with-leading-checkbox.mdc-list-item,.mdc-list-item--with-leading-icon.mdc-list-item,.mdc-list-item--with-leading-avatar.mdc-list-item{padding-left:0;padding-right:16px}[dir=rtl] .mdc-list-item--with-leading-radio.mdc-list-item,[dir=rtl] .mdc-list-item--with-leading-checkbox.mdc-list-item,[dir=rtl] .mdc-list-item--with-leading-icon.mdc-list-item,[dir=rtl] .mdc-list-item--with-leading-avatar.mdc-list-item{padding-left:16px;padding-right:0}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines .mdc-list-item__primary-text,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines .mdc-list-item__primary-text,.mdc-list-item--with-leading-icon.mdc-list-item--with-two-lines .mdc-list-item__primary-text,.mdc-list-item--with-leading-avatar.mdc-list-item--with-two-lines .mdc-list-item__primary-text{display:block;margin-top:0;line-height:normal;margin-bottom:-20px}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines .mdc-list-item__primary-text::before,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines .mdc-list-item__primary-text::before,.mdc-list-item--with-leading-icon.mdc-list-item--with-two-lines .mdc-list-item__primary-text::before,.mdc-list-item--with-leading-avatar.mdc-list-item--with-two-lines .mdc-list-item__primary-text::before{display:inline-block;width:0;height:32px;content:"";vertical-align:0}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines .mdc-list-item__primary-text::after,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines .mdc-list-item__primary-text::after,.mdc-list-item--with-leading-icon.mdc-list-item--with-two-lines .mdc-list-item__primary-text::after,.mdc-list-item--with-leading-avatar.mdc-list-item--with-two-lines .mdc-list-item__primary-text::after{display:inline-block;width:0;height:20px;content:"";vertical-align:-20px}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end,.mdc-list-item--with-leading-icon.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end,.mdc-list-item--with-leading-avatar.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end{display:block;margin-top:0;line-height:normal}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end::before,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end::before,.mdc-list-item--with-leading-icon.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end::before,.mdc-list-item--with-leading-avatar.mdc-list-item--with-two-lines.mdc-list-item--with-trailing-meta .mdc-list-item__end::before{display:inline-block;width:0;height:32px;content:"";vertical-align:0}.mdc-list-item--with-trailing-icon.mdc-list-item,[dir=rtl] .mdc-list-item--with-trailing-icon.mdc-list-item{padding-left:0;padding-right:0}.mdc-list-item--with-trailing-icon .mdc-list-item__end{margin-left:16px;margin-right:16px}.mdc-list-item--with-trailing-meta.mdc-list-item{padding-left:16px;padding-right:0}[dir=rtl] .mdc-list-item--with-trailing-meta.mdc-list-item{padding-left:0;padding-right:16px}.mdc-list-item--with-trailing-meta .mdc-list-item__end{-webkit-user-select:none;user-select:none;margin-left:28px;margin-right:16px}[dir=rtl] .mdc-list-item--with-trailing-meta .mdc-list-item__end{margin-left:16px;margin-right:28px}.mdc-list-item--with-trailing-meta.mdc-list-item--with-three-lines .mdc-list-item__end,.mdc-list-item--with-trailing-meta.mdc-list-item--with-two-lines .mdc-list-item__end{display:block;line-height:normal;align-self:flex-start;margin-top:0}.mdc-list-item--with-trailing-meta.mdc-list-item--with-three-lines .mdc-list-item__end::before,.mdc-list-item--with-trailing-meta.mdc-list-item--with-two-lines .mdc-list-item__end::before{display:inline-block;width:0;height:28px;content:"";vertical-align:0}.mdc-list-item--with-leading-radio .mdc-list-item__start,.mdc-list-item--with-leading-checkbox .mdc-list-item__start{margin-left:8px;margin-right:24px}[dir=rtl] .mdc-list-item--with-leading-radio .mdc-list-item__start,[dir=rtl] .mdc-list-item--with-leading-checkbox .mdc-list-item__start{margin-left:24px;margin-right:8px}.mdc-list-item--with-leading-radio.mdc-list-item--with-two-lines .mdc-list-item__start,.mdc-list-item--with-leading-checkbox.mdc-list-item--with-two-lines .mdc-list-item__start{align-self:flex-start;margin-top:8px}.mdc-list-item--with-trailing-radio.mdc-list-item,.mdc-list-item--with-trailing-checkbox.mdc-list-item{padding-left:16px;padding-right:0}[dir=rtl] .mdc-list-item--with-trailing-radio.mdc-list-item,[dir=rtl] .mdc-list-item--with-trailing-checkbox.mdc-list-item{padding-left:0;padding-right:16px}.mdc-list-item--with-trailing-radio.mdc-list-item--with-leading-icon,.mdc-list-item--with-trailing-radio.mdc-list-item--with-leading-avatar,.mdc-list-item--with-trailing-checkbox.mdc-list-item--with-leading-icon,.mdc-list-item--with-trailing-checkbox.mdc-list-item--with-leading-avatar{padding-left:0}[dir=rtl] .mdc-list-item--with-trailing-radio.mdc-list-item--with-leading-icon,[dir=rtl] .mdc-list-item--with-trailing-radio.mdc-list-item--with-leading-avatar,[dir=rtl] .mdc-list-item--with-trailing-checkbox.mdc-list-item--with-leading-icon,[dir=rtl] .mdc-list-item--with-trailing-checkbox.mdc-list-item--with-leading-avatar{padding-right:0}.mdc-list-item--with-trailing-radio .mdc-list-item__end,.mdc-list-item--with-trailing-checkbox .mdc-list-item__end{margin-left:24px;margin-right:8px}[dir=rtl] .mdc-list-item--with-trailing-radio .mdc-list-item__end,[dir=rtl] .mdc-list-item--with-trailing-checkbox .mdc-list-item__end{margin-left:8px;margin-right:24px}.mdc-list-item--with-trailing-radio.mdc-list-item--with-three-lines .mdc-list-item__end,.mdc-list-item--with-trailing-checkbox.mdc-list-item--with-three-lines .mdc-list-item__end{align-self:flex-start;margin-top:8px}.mdc-list-group__subheader{margin:.75rem 16px}.mdc-list-item--disabled .mdc-list-item__start,.mdc-list-item--disabled .mdc-list-item__content,.mdc-list-item--disabled .mdc-list-item__end{opacity:1}.mdc-list-item--disabled .mdc-list-item__primary-text,.mdc-list-item--disabled .mdc-list-item__secondary-text{opacity:var(--mdc-list-list-item-disabled-label-text-opacity, 0.3)}.mdc-list-item--disabled.mdc-list-item--with-leading-icon .mdc-list-item__start{color:var(--mdc-list-list-item-disabled-leading-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-list-list-item-disabled-leading-icon-opacity, 0.38)}.mdc-list-item--disabled.mdc-list-item--with-trailing-icon .mdc-list-item__end{color:var(--mdc-list-list-item-disabled-trailing-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-list-list-item-disabled-trailing-icon-opacity, 0.38)}.mat-mdc-list-item.mat-mdc-list-item-both-leading-and-trailing,[dir=rtl] .mat-mdc-list-item.mat-mdc-list-item-both-leading-and-trailing{padding-left:0;padding-right:0}.mdc-list-item.mdc-list-item--disabled .mdc-list-item__primary-text{color:var(--mdc-list-list-item-disabled-label-text-color, var(--mat-sys-on-surface))}.mdc-list-item:hover::before{background-color:var(--mdc-list-list-item-hover-state-layer-color, var(--mat-sys-on-surface));opacity:var(--mdc-list-list-item-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mdc-list-item.mdc-list-item--disabled::before{background-color:var(--mdc-list-list-item-disabled-state-layer-color, var(--mat-sys-on-surface));opacity:var(--mdc-list-list-item-disabled-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mdc-list-item:focus::before{background-color:var(--mdc-list-list-item-focus-state-layer-color, var(--mat-sys-on-surface));opacity:var(--mdc-list-list-item-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}.mdc-list-item--disabled .mdc-radio,.mdc-list-item--disabled .mdc-checkbox{opacity:var(--mdc-list-list-item-disabled-label-text-opacity, 0.3)}.mdc-list-item--with-leading-avatar .mat-mdc-list-item-avatar{border-radius:var(--mdc-list-list-item-leading-avatar-shape, var(--mat-sys-corner-full));background-color:var(--mdc-list-list-item-leading-avatar-color, var(--mat-sys-primary-container))}.mat-mdc-list-item-icon{font-size:var(--mdc-list-list-item-leading-icon-size, 24px)}@media(forced-colors: active){a.mdc-list-item--activated::after{content:"";position:absolute;top:50%;right:16px;transform:translateY(-50%);width:10px;height:0;border-bottom:solid 10px;border-radius:10px}a.mdc-list-item--activated [dir=rtl]::after{right:auto;left:16px}}.mat-mdc-list-base{display:block}.mat-mdc-list-base .mdc-list-item__start,.mat-mdc-list-base .mdc-list-item__end,.mat-mdc-list-base .mdc-list-item__content{pointer-events:auto}.mat-mdc-list-item,.mat-mdc-list-option{width:100%;box-sizing:border-box;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-list-item:not(.mat-mdc-list-item-interactive),.mat-mdc-list-option:not(.mat-mdc-list-item-interactive){cursor:default}.mat-mdc-list-item .mat-divider-inset,.mat-mdc-list-option .mat-divider-inset{position:absolute;left:0;right:0;bottom:0}.mat-mdc-list-item .mat-mdc-list-item-avatar~.mat-divider-inset,.mat-mdc-list-option .mat-mdc-list-item-avatar~.mat-divider-inset{margin-left:72px}[dir=rtl] .mat-mdc-list-item .mat-mdc-list-item-avatar~.mat-divider-inset,[dir=rtl] .mat-mdc-list-option .mat-mdc-list-item-avatar~.mat-divider-inset{margin-right:72px}.mat-mdc-list-item-interactive::before{top:0;left:0;right:0;bottom:0;position:absolute;content:"";opacity:0;pointer-events:none;border-radius:inherit}.mat-mdc-list-item>.mat-focus-indicator{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-mdc-list-item:focus>.mat-focus-indicator::before{content:""}.mat-mdc-list-item.mdc-list-item--with-three-lines .mat-mdc-list-item-line.mdc-list-item__secondary-text{white-space:nowrap;line-height:normal}.mat-mdc-list-item.mdc-list-item--with-three-lines .mat-mdc-list-item-unscoped-content.mdc-list-item__secondary-text{display:-webkit-box;-webkit-box-orient:vertical;-webkit-line-clamp:2}mat-action-list button{background:none;color:inherit;border:none;font:inherit;outline:inherit;-webkit-tap-highlight-color:rgba(0,0,0,0);text-align:start}mat-action-list button::-moz-focus-inner{border:0}.mdc-list-item--with-leading-icon .mdc-list-item__start{margin-inline-start:var(--mat-list-list-item-leading-icon-start-space, 16px);margin-inline-end:var(--mat-list-list-item-leading-icon-end-space, 16px)}.mat-mdc-nav-list .mat-mdc-list-item{border-radius:var(--mat-list-active-indicator-shape, var(--mat-sys-corner-full));--mat-focus-indicator-border-radius:var(--mat-list-active-indicator-shape, var(--mat-sys-corner-full))}.mat-mdc-nav-list .mat-mdc-list-item.mdc-list-item--activated{background-color:var(--mat-list-active-indicator-color, var(--mat-sys-secondary-container))}',ozA=["unscopedContent"],rzA=["text"],szA=[[["","matListItemAvatar",""],["","matListItemIcon",""]],[["","matListItemTitle",""]],[["","matListItemLine",""]],"*",[["","matListItemMeta",""]],[["mat-divider"]]],azA=["[matListItemAvatar],[matListItemIcon]","[matListItemTitle]","[matListItemLine]","*","[matListItemMeta]","mat-divider"];var czA=new dA("ListOption"),lzA=(()=>{class t{_elementRef=f(ee);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matListItemTitle",""]],hostAttrs:[1,"mat-mdc-list-item-title","mdc-list-item__primary-text"]})}return t})(),gzA=(()=>{class t{_elementRef=f(ee);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matListItemLine",""]],hostAttrs:[1,"mat-mdc-list-item-line","mdc-list-item__secondary-text"]})}return t})(),IzA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matListItemMeta",""]],hostAttrs:[1,"mat-mdc-list-item-meta","mdc-list-item__end"]})}return t})(),DcA=(()=>{class t{_listOption=f(czA,{optional:!0});constructor(){}_isAlignedAtStart(){return!this._listOption||this._listOption?._getTogglePosition()==="after"}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,hostVars:4,hostBindings:function(i,n){i&2&&ue("mdc-list-item__start",n._isAlignedAtStart())("mdc-list-item__end",!n._isAlignedAtStart())}})}return t})(),CzA=(()=>{class t extends DcA{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matListItemAvatar",""]],hostAttrs:[1,"mat-mdc-list-item-avatar"],features:[lt]})}return t})(),dzA=(()=>{class t extends DcA{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matListItemIcon",""]],hostAttrs:[1,"mat-mdc-list-item-icon"],features:[lt]})}return t})(),BzA=new dA("MAT_LIST_CONFIG"),NG=(()=>{class t{_isNonInteractive=!0;get disableRipple(){return this._disableRipple}set disableRipple(A){this._disableRipple=Xo(A)}_disableRipple=!1;get disabled(){return this._disabled}set disabled(A){this._disabled=Xo(A)}_disabled=!1;_defaultOptions=f(BzA,{optional:!0});static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,hostVars:1,hostBindings:function(i,n){i&2&&Ne("aria-disabled",n.disabled)},inputs:{disableRipple:"disableRipple",disabled:"disabled"}})}return t})(),EzA=(()=>{class t{_elementRef=f(ee);_ngZone=f(Qe);_listBase=f(NG,{optional:!0});_platform=f(Ii);_hostElement;_isButtonElement;_noopAnimations;_avatars;_icons;set lines(A){this._explicitLines=zs(A,null),this._updateItemLines(!1)}_explicitLines=null;get disableRipple(){return this.disabled||this._disableRipple||this._noopAnimations||!!this._listBase?.disableRipple}set disableRipple(A){this._disableRipple=Xo(A)}_disableRipple=!1;get disabled(){return this._disabled||!!this._listBase?.disabled}set disabled(A){this._disabled=Xo(A)}_disabled=!1;_subscriptions=new Kt;_rippleRenderer=null;_hasUnscopedTextContent=!1;rippleConfig;get rippleDisabled(){return this.disableRipple||!!this.rippleConfig.disabled}constructor(){f(Rn).load(lr);let A=f(G2,{optional:!0}),i=f(Si,{optional:!0});this.rippleConfig=A||{},this._hostElement=this._elementRef.nativeElement,this._isButtonElement=this._hostElement.nodeName.toLowerCase()==="button",this._noopAnimations=i==="NoopAnimations",this._listBase&&!this._listBase._isNonInteractive&&this._initInteractiveListItem(),this._isButtonElement&&!this._hostElement.hasAttribute("type")&&this._hostElement.setAttribute("type","button")}ngAfterViewInit(){this._monitorProjectedLinesAndTitle(),this._updateItemLines(!0)}ngOnDestroy(){this._subscriptions.unsubscribe(),this._rippleRenderer!==null&&this._rippleRenderer._removeTriggerEvents()}_hasIconOrAvatar(){return!!(this._avatars.length||this._icons.length)}_initInteractiveListItem(){this._hostElement.classList.add("mat-mdc-list-item-interactive"),this._rippleRenderer=new vB(this,this._ngZone,this._hostElement,this._platform,f(Rt)),this._rippleRenderer.setupTriggerEvents(this._hostElement)}_monitorProjectedLinesAndTitle(){this._ngZone.runOutsideAngular(()=>{this._subscriptions.add(zn(this._lines.changes,this._titles.changes).subscribe(()=>this._updateItemLines(!1)))})}_updateItemLines(A){if(!this._lines||!this._titles||!this._unscopedContent)return;A&&this._checkDomForUnscopedTextContent();let i=this._explicitLines??this._inferLinesFromContent(),n=this._unscopedContent.nativeElement;if(this._hostElement.classList.toggle("mat-mdc-list-item-single-line",i<=1),this._hostElement.classList.toggle("mdc-list-item--with-one-line",i<=1),this._hostElement.classList.toggle("mdc-list-item--with-two-lines",i===2),this._hostElement.classList.toggle("mdc-list-item--with-three-lines",i===3),this._hasUnscopedTextContent){let o=this._titles.length===0&&i===1;n.classList.toggle("mdc-list-item__primary-text",o),n.classList.toggle("mdc-list-item__secondary-text",!o)}else n.classList.remove("mdc-list-item__primary-text"),n.classList.remove("mdc-list-item__secondary-text")}_inferLinesFromContent(){let A=this._titles.length+this._lines.length;return this._hasUnscopedTextContent&&(A+=1),A}_checkDomForUnscopedTextContent(){this._hasUnscopedTextContent=Array.from(this._unscopedContent.nativeElement.childNodes).filter(A=>A.nodeType!==A.COMMENT_NODE).some(A=>!!(A.textContent&&A.textContent.trim()))}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,contentQueries:function(i,n,o){if(i&1&&(ci(o,CzA,4),ci(o,dzA,4)),i&2){let r;XA(r=$A())&&(n._avatars=r),XA(r=$A())&&(n._icons=r)}},hostVars:4,hostBindings:function(i,n){i&2&&(Ne("aria-disabled",n.disabled)("disabled",n._isButtonElement&&n.disabled||null),ue("mdc-list-item--disabled",n.disabled))},inputs:{lines:"lines",disableRipple:"disableRipple",disabled:"disabled"}})}return t})();var ycA=(()=>{class t extends NG{static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-list"]],hostAttrs:[1,"mat-mdc-list","mat-mdc-list-base","mdc-list"],exportAs:["matList"],features:[ht([{provide:NG,useExisting:t}]),lt],ngContentSelectors:izA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},styles:[nzA],encapsulation:2,changeDetection:0})}return t})(),vcA=(()=>{class t extends EzA{_lines;_titles;_meta;_unscopedContent;_itemText;get activated(){return this._activated}set activated(A){this._activated=Xo(A)}_activated=!1;_getAriaCurrent(){return this._hostElement.nodeName==="A"&&this._activated?"page":null}_hasBothLeadingAndTrailing(){return this._meta.length!==0&&(this._avatars.length!==0||this._icons.length!==0)}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-list-item"],["a","mat-list-item",""],["button","mat-list-item",""]],contentQueries:function(i,n,o){if(i&1&&(ci(o,gzA,5),ci(o,lzA,5),ci(o,IzA,5)),i&2){let r;XA(r=$A())&&(n._lines=r),XA(r=$A())&&(n._titles=r),XA(r=$A())&&(n._meta=r)}},viewQuery:function(i,n){if(i&1&&(Te(ozA,5),Te(rzA,5)),i&2){let o;XA(o=$A())&&(n._unscopedContent=o.first),XA(o=$A())&&(n._itemText=o.first)}},hostAttrs:[1,"mat-mdc-list-item","mdc-list-item"],hostVars:13,hostBindings:function(i,n){i&2&&(Ne("aria-current",n._getAriaCurrent()),ue("mdc-list-item--activated",n.activated)("mdc-list-item--with-leading-avatar",n._avatars.length!==0)("mdc-list-item--with-leading-icon",n._icons.length!==0)("mdc-list-item--with-trailing-meta",n._meta.length!==0)("mat-mdc-list-item-both-leading-and-trailing",n._hasBothLeadingAndTrailing())("_mat-animation-noopable",n._noopAnimations))},inputs:{activated:"activated"},exportAs:["matListItem"],features:[lt],ngContentSelectors:azA,decls:10,vars:0,consts:[["unscopedContent",""],[1,"mdc-list-item__content"],[1,"mat-mdc-list-item-unscoped-content",3,"cdkObserveContent"],[1,"mat-focus-indicator"]],template:function(i,n){if(i&1){let o=be();jt(szA),Fe(0),S(1,"span",1),Fe(2,1),Fe(3,2),S(4,"span",2,0),hA("cdkObserveContent",function(){return RA(o),xA(n._updateItemLines(!0))}),Fe(6,3),F()(),Fe(7,4),Fe(8,5),JA(9,"div",3)}},dependencies:[V6],encapsulation:2,changeDetection:0})}return t})();var bcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[DB,it,Ps,Dk,wcA]})}return t})();var hzA=["button"],uzA=["*"];function fzA(t,e){if(t&1&&(S(0,"div",2),JA(1,"mat-pseudo-checkbox",6),F()),t&2){let A=O();G(),yA("disabled",A.disabled)}}var McA=new dA("MAT_BUTTON_TOGGLE_DEFAULT_OPTIONS",{providedIn:"root",factory:mzA});function mzA(){return{hideSingleSelectionIndicator:!1,hideMultipleSelectionIndicator:!1,disabledInteractive:!1}}var kcA=new dA("MatButtonToggleGroup"),pzA={provide:yc,useExisting:Ir(()=>_G),multi:!0},i7=class{source;value;constructor(e,A){this.source=e,this.value=A}},_G=(()=>{class t{_changeDetector=f(It);_dir=f(mo,{optional:!0});_multiple=!1;_disabled=!1;_disabledInteractive=!1;_selectionModel;_rawValue;_controlValueAccessorChangeFn=()=>{};_onTouched=()=>{};_buttonToggles;appearance;get name(){return this._name}set name(A){this._name=A,this._markButtonsForCheck()}_name=f(sn).getId("mat-button-toggle-group-");vertical;get value(){let A=this._selectionModel?this._selectionModel.selected:[];return this.multiple?A.map(i=>i.value):A[0]?A[0].value:void 0}set value(A){this._setSelectionByValue(A),this.valueChange.emit(this.value)}valueChange=new WA;get selected(){let A=this._selectionModel?this._selectionModel.selected:[];return this.multiple?A:A[0]||null}get multiple(){return this._multiple}set multiple(A){this._multiple=A,this._markButtonsForCheck()}get disabled(){return this._disabled}set disabled(A){this._disabled=A,this._markButtonsForCheck()}get disabledInteractive(){return this._disabledInteractive}set disabledInteractive(A){this._disabledInteractive=A,this._markButtonsForCheck()}get dir(){return this._dir&&this._dir.value==="rtl"?"rtl":"ltr"}change=new WA;get hideSingleSelectionIndicator(){return this._hideSingleSelectionIndicator}set hideSingleSelectionIndicator(A){this._hideSingleSelectionIndicator=A,this._markButtonsForCheck()}_hideSingleSelectionIndicator;get hideMultipleSelectionIndicator(){return this._hideMultipleSelectionIndicator}set hideMultipleSelectionIndicator(A){this._hideMultipleSelectionIndicator=A,this._markButtonsForCheck()}_hideMultipleSelectionIndicator;constructor(){let A=f(McA,{optional:!0});this.appearance=A&&A.appearance?A.appearance:"standard",this.hideSingleSelectionIndicator=A?.hideSingleSelectionIndicator??!1,this.hideMultipleSelectionIndicator=A?.hideMultipleSelectionIndicator??!1}ngOnInit(){this._selectionModel=new K2(this.multiple,void 0,!1)}ngAfterContentInit(){this._selectionModel.select(...this._buttonToggles.filter(A=>A.checked)),this.multiple||this._initializeTabIndex()}writeValue(A){this.value=A,this._changeDetector.markForCheck()}registerOnChange(A){this._controlValueAccessorChangeFn=A}registerOnTouched(A){this._onTouched=A}setDisabledState(A){this.disabled=A}_keydown(A){if(this.multiple||this.disabled)return;let n=A.target.id,o=this._buttonToggles.toArray().findIndex(s=>s.buttonId===n),r=null;switch(A.keyCode){case 32:case 13:r=this._buttonToggles.get(o)||null;break;case 38:r=this._getNextButton(o,-1);break;case 37:r=this._getNextButton(o,this.dir==="ltr"?-1:1);break;case 40:r=this._getNextButton(o,1);break;case 39:r=this._getNextButton(o,this.dir==="ltr"?1:-1);break;default:return}r&&(A.preventDefault(),r._onButtonClick(),r.focus())}_emitChangeEvent(A){let i=new i7(A,this.value);this._rawValue=i.value,this._controlValueAccessorChangeFn(i.value),this.change.emit(i)}_syncButtonToggle(A,i,n=!1,o=!1){!this.multiple&&this.selected&&!A.checked&&(this.selected.checked=!1),this._selectionModel?i?this._selectionModel.select(A):this._selectionModel.deselect(A):o=!0,o?Promise.resolve().then(()=>this._updateModelValue(A,n)):this._updateModelValue(A,n)}_isSelected(A){return this._selectionModel&&this._selectionModel.isSelected(A)}_isPrechecked(A){return typeof this._rawValue>"u"?!1:this.multiple&&Array.isArray(this._rawValue)?this._rawValue.some(i=>A.value!=null&&i===A.value):A.value===this._rawValue}_initializeTabIndex(){if(this._buttonToggles.forEach(A=>{A.tabIndex=-1}),this.selected)this.selected.tabIndex=0;else for(let A=0;A<this._buttonToggles.length;A++){let i=this._buttonToggles.get(A);if(!i.disabled){i.tabIndex=0;break}}this._markButtonsForCheck()}_getNextButton(A,i){let n=this._buttonToggles;for(let o=1;o<=n.length;o++){let r=(A+i*o+n.length)%n.length,s=n.get(r);if(s&&!s.disabled)return s}return null}_setSelectionByValue(A){if(this._rawValue=A,!this._buttonToggles)return;let i=this._buttonToggles.toArray();if(this.multiple&&A?(Array.isArray(A),this._clearSelection(),A.forEach(n=>this._selectValue(n,i))):(this._clearSelection(),this._selectValue(A,i)),!this.multiple&&i.every(n=>n.tabIndex===-1)){for(let n of i)if(!n.disabled){n.tabIndex=0;break}}}_clearSelection(){this._selectionModel.clear(),this._buttonToggles.forEach(A=>{A.checked=!1,this.multiple||(A.tabIndex=-1)})}_selectValue(A,i){for(let n of i)if(n.value===A){n.checked=!0,this._selectionModel.select(n),this.multiple||(n.tabIndex=0);break}}_updateModelValue(A,i){i&&this._emitChangeEvent(A),this.valueChange.emit(this.value)}_markButtonsForCheck(){this._buttonToggles?.forEach(A=>A._markForCheck())}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["mat-button-toggle-group"]],contentQueries:function(i,n,o){if(i&1&&ci(o,n7,5),i&2){let r;XA(r=$A())&&(n._buttonToggles=r)}},hostAttrs:[1,"mat-button-toggle-group"],hostVars:6,hostBindings:function(i,n){i&1&&hA("keydown",function(r){return n._keydown(r)}),i&2&&(Ne("role",n.multiple?"group":"radiogroup")("aria-disabled",n.disabled),ue("mat-button-toggle-vertical",n.vertical)("mat-button-toggle-group-appearance-standard",n.appearance==="standard"))},inputs:{appearance:"appearance",name:"name",vertical:[2,"vertical","vertical",ie],value:"value",multiple:[2,"multiple","multiple",ie],disabled:[2,"disabled","disabled",ie],disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie],hideSingleSelectionIndicator:[2,"hideSingleSelectionIndicator","hideSingleSelectionIndicator",ie],hideMultipleSelectionIndicator:[2,"hideMultipleSelectionIndicator","hideMultipleSelectionIndicator",ie]},outputs:{valueChange:"valueChange",change:"change"},exportAs:["matButtonToggleGroup"],features:[ht([pzA,{provide:kcA,useExisting:t}])]})}return t})(),n7=(()=>{class t{_changeDetectorRef=f(It);_elementRef=f(ee);_focusMonitor=f(dr);_idGenerator=f(sn);_animationMode=f(Si,{optional:!0});_checked=!1;ariaLabel;ariaLabelledby=null;_buttonElement;buttonToggleGroup;get buttonId(){return`${this.id}-button`}id;name;value;get tabIndex(){return this._tabIndex}set tabIndex(A){A!==this._tabIndex&&(this._tabIndex=A,this._markForCheck())}_tabIndex;disableRipple;get appearance(){return this.buttonToggleGroup?this.buttonToggleGroup.appearance:this._appearance}set appearance(A){this._appearance=A}_appearance;get checked(){return this.buttonToggleGroup?this.buttonToggleGroup._isSelected(this):this._checked}set checked(A){A!==this._checked&&(this._checked=A,this.buttonToggleGroup&&this.buttonToggleGroup._syncButtonToggle(this,this._checked),this._changeDetectorRef.markForCheck())}get disabled(){return this._disabled||this.buttonToggleGroup&&this.buttonToggleGroup.disabled}set disabled(A){this._disabled=A}_disabled=!1;get disabledInteractive(){return this._disabledInteractive||this.buttonToggleGroup!==null&&this.buttonToggleGroup.disabledInteractive}set disabledInteractive(A){this._disabledInteractive=A}_disabledInteractive;change=new WA;constructor(){f(Rn).load(lr);let A=f(kcA,{optional:!0}),i=f(new wr("tabindex"),{optional:!0})||"",n=f(McA,{optional:!0});this._tabIndex=parseInt(i)||0,this.buttonToggleGroup=A,this.appearance=n&&n.appearance?n.appearance:"standard",this.disabledInteractive=n?.disabledInteractive??!1}ngOnInit(){let A=this.buttonToggleGroup;this.id=this.id||this._idGenerator.getId("mat-button-toggle-"),A&&(A._isPrechecked(this)?this.checked=!0:A._isSelected(this)!==this._checked&&A._syncButtonToggle(this,this._checked))}ngAfterViewInit(){this._animationMode!=="NoopAnimations"&&this._elementRef.nativeElement.classList.add("mat-button-toggle-animations-enabled"),this._focusMonitor.monitor(this._elementRef,!0)}ngOnDestroy(){let A=this.buttonToggleGroup;this._focusMonitor.stopMonitoring(this._elementRef),A&&A._isSelected(this)&&A._syncButtonToggle(this,!1,!1,!0)}focus(A){this._buttonElement.nativeElement.focus(A)}_onButtonClick(){if(this.disabled)return;let A=this.isSingleSelector()?!0:!this._checked;if(A!==this._checked&&(this._checked=A,this.buttonToggleGroup&&(this.buttonToggleGroup._syncButtonToggle(this,this._checked,!0),this.buttonToggleGroup._onTouched())),this.isSingleSelector()){let i=this.buttonToggleGroup._buttonToggles.find(n=>n.tabIndex===0);i&&(i.tabIndex=-1),this.tabIndex=0}this.change.emit(new i7(this,this.value))}_markForCheck(){this._changeDetectorRef.markForCheck()}_getButtonName(){return this.isSingleSelector()?this.buttonToggleGroup.name:this.name||null}isSingleSelector(){return this.buttonToggleGroup&&!this.buttonToggleGroup.multiple}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-button-toggle"]],viewQuery:function(i,n){if(i&1&&Te(hzA,5),i&2){let o;XA(o=$A())&&(n._buttonElement=o.first)}},hostAttrs:["role","presentation",1,"mat-button-toggle"],hostVars:14,hostBindings:function(i,n){i&1&&hA("focus",function(){return n.focus()}),i&2&&(Ne("aria-label",null)("aria-labelledby",null)("id",n.id)("name",null),ue("mat-button-toggle-standalone",!n.buttonToggleGroup)("mat-button-toggle-checked",n.checked)("mat-button-toggle-disabled",n.disabled)("mat-button-toggle-disabled-interactive",n.disabledInteractive)("mat-button-toggle-appearance-standard",n.appearance==="standard"))},inputs:{ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],id:"id",name:"name",value:"value",tabIndex:"tabIndex",disableRipple:[2,"disableRipple","disableRipple",ie],appearance:"appearance",checked:[2,"checked","checked",ie],disabled:[2,"disabled","disabled",ie],disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie]},outputs:{change:"change"},exportAs:["matButtonToggle"],ngContentSelectors:uzA,decls:7,vars:13,consts:[["button",""],["type","button",1,"mat-button-toggle-button","mat-focus-indicator",3,"click","id","disabled"],[1,"mat-button-toggle-checkbox-wrapper"],[1,"mat-button-toggle-label-content"],[1,"mat-button-toggle-focus-overlay"],["matRipple","",1,"mat-button-toggle-ripple",3,"matRippleTrigger","matRippleDisabled"],["state","checked","aria-hidden","true","appearance","minimal",3,"disabled"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"button",1,0),hA("click",function(){return RA(o),xA(n._onButtonClick())}),_A(2,fzA,2,1,"div",2),S(3,"span",3),Fe(4),F()(),JA(5,"span",4)(6,"span",5)}if(i&2){let o=cr(1);yA("id",n.buttonId)("disabled",n.disabled&&!n.disabledInteractive||null),Ne("role",n.isSingleSelector()?"radio":"button")("tabindex",n.disabled&&!n.disabledInteractive?-1:n.tabIndex)("aria-pressed",n.isSingleSelector()?null:n.checked)("aria-checked",n.isSingleSelector()?n.checked:null)("name",n._getButtonName())("aria-label",n.ariaLabel)("aria-labelledby",n.ariaLabelledby)("aria-disabled",n.disabled&&n.disabledInteractive?"true":null),G(2),GA(n.buttonToggleGroup&&(!n.buttonToggleGroup.multiple&&!n.buttonToggleGroup.hideSingleSelectionIndicator||n.buttonToggleGroup.multiple&&!n.buttonToggleGroup.hideMultipleSelectionIndicator)?2:-1),G(4),yA("matRippleTrigger",o)("matRippleDisabled",n.disableRipple||n.disabled)}},dependencies:[rs,wk],styles:[".mat-button-toggle-standalone,.mat-button-toggle-group{position:relative;display:inline-flex;flex-direction:row;white-space:nowrap;overflow:hidden;-webkit-tap-highlight-color:rgba(0,0,0,0);transform:translateZ(0);border-radius:var(--mat-legacy-button-toggle-shape)}.mat-button-toggle-standalone:not([class*=mat-elevation-z]),.mat-button-toggle-group:not([class*=mat-elevation-z]){box-shadow:0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12)}@media(forced-colors: active){.mat-button-toggle-standalone,.mat-button-toggle-group{outline:solid 1px}}.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,.mat-button-toggle-group-appearance-standard{border-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full));border:solid 1px var(--mat-standard-button-toggle-divider-color, var(--mat-sys-outline))}.mat-button-toggle-standalone.mat-button-toggle-appearance-standard .mat-pseudo-checkbox,.mat-button-toggle-group-appearance-standard .mat-pseudo-checkbox{--mat-minimal-pseudo-checkbox-selected-checkmark-color: var(--mat-standard-button-toggle-selected-state-text-color, var(--mat-sys-on-secondary-container))}.mat-button-toggle-standalone.mat-button-toggle-appearance-standard:not([class*=mat-elevation-z]),.mat-button-toggle-group-appearance-standard:not([class*=mat-elevation-z]){box-shadow:none}@media(forced-colors: active){.mat-button-toggle-standalone.mat-button-toggle-appearance-standard,.mat-button-toggle-group-appearance-standard{outline:0}}.mat-button-toggle-vertical{flex-direction:column}.mat-button-toggle-vertical .mat-button-toggle-label-content{display:block}.mat-button-toggle{white-space:nowrap;position:relative;color:var(--mat-legacy-button-toggle-text-color);font-family:var(--mat-legacy-button-toggle-label-text-font);font-size:var(--mat-legacy-button-toggle-label-text-size);line-height:var(--mat-legacy-button-toggle-label-text-line-height);font-weight:var(--mat-legacy-button-toggle-label-text-weight);letter-spacing:var(--mat-legacy-button-toggle-label-text-tracking);--mat-minimal-pseudo-checkbox-selected-checkmark-color: var(--mat-legacy-button-toggle-selected-state-text-color)}.mat-button-toggle.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:var(--mat-legacy-button-toggle-focus-state-layer-opacity)}.mat-button-toggle .mat-icon svg{vertical-align:top}.mat-button-toggle-checkbox-wrapper{display:inline-block;justify-content:flex-start;align-items:center;width:0;height:18px;line-height:18px;overflow:hidden;box-sizing:border-box;position:absolute;top:50%;left:16px;transform:translate3d(0, -50%, 0)}[dir=rtl] .mat-button-toggle-checkbox-wrapper{left:auto;right:16px}.mat-button-toggle-appearance-standard .mat-button-toggle-checkbox-wrapper{left:12px}[dir=rtl] .mat-button-toggle-appearance-standard .mat-button-toggle-checkbox-wrapper{left:auto;right:12px}.mat-button-toggle-checked .mat-button-toggle-checkbox-wrapper{width:18px}.mat-button-toggle-animations-enabled .mat-button-toggle-checkbox-wrapper{transition:width 150ms 45ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-button-toggle-vertical .mat-button-toggle-checkbox-wrapper{transition:none}.mat-button-toggle-checked{color:var(--mat-legacy-button-toggle-selected-state-text-color);background-color:var(--mat-legacy-button-toggle-selected-state-background-color)}.mat-button-toggle-disabled{pointer-events:none;color:var(--mat-legacy-button-toggle-disabled-state-text-color);background-color:var(--mat-legacy-button-toggle-disabled-state-background-color);--mat-minimal-pseudo-checkbox-disabled-selected-checkmark-color: var(--mat-legacy-button-toggle-disabled-state-text-color)}.mat-button-toggle-disabled.mat-button-toggle-checked{background-color:var(--mat-legacy-button-toggle-disabled-selected-state-background-color)}.mat-button-toggle-disabled-interactive{pointer-events:auto}.mat-button-toggle-appearance-standard{color:var(--mat-standard-button-toggle-text-color, var(--mat-sys-on-surface));background-color:var(--mat-standard-button-toggle-background-color, transparent);font-family:var(--mat-standard-button-toggle-label-text-font, var(--mat-sys-label-large-font));font-size:var(--mat-standard-button-toggle-label-text-size, var(--mat-sys-label-large-size));line-height:var(--mat-standard-button-toggle-label-text-line-height, var(--mat-sys-label-large-line-height));font-weight:var(--mat-standard-button-toggle-label-text-weight, var(--mat-sys-label-large-weight));letter-spacing:var(--mat-standard-button-toggle-label-text-tracking, var(--mat-sys-label-large-tracking))}.mat-button-toggle-group-appearance-standard .mat-button-toggle-appearance-standard+.mat-button-toggle-appearance-standard{border-left:solid 1px var(--mat-standard-button-toggle-divider-color, var(--mat-sys-outline))}[dir=rtl] .mat-button-toggle-group-appearance-standard .mat-button-toggle-appearance-standard+.mat-button-toggle-appearance-standard{border-left:none;border-right:solid 1px var(--mat-standard-button-toggle-divider-color, var(--mat-sys-outline))}.mat-button-toggle-group-appearance-standard.mat-button-toggle-vertical .mat-button-toggle-appearance-standard+.mat-button-toggle-appearance-standard{border-left:none;border-right:none;border-top:solid 1px var(--mat-standard-button-toggle-divider-color, var(--mat-sys-outline))}.mat-button-toggle-appearance-standard.mat-button-toggle-checked{color:var(--mat-standard-button-toggle-selected-state-text-color, var(--mat-sys-on-secondary-container));background-color:var(--mat-standard-button-toggle-selected-state-background-color, var(--mat-sys-secondary-container))}.mat-button-toggle-appearance-standard.mat-button-toggle-disabled{color:var(--mat-standard-button-toggle-disabled-state-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-standard-button-toggle-disabled-state-background-color, transparent)}.mat-button-toggle-appearance-standard.mat-button-toggle-disabled .mat-pseudo-checkbox{--mat-minimal-pseudo-checkbox-disabled-selected-checkmark-color: var(--mat-standard-button-toggle-disabled-selected-state-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-button-toggle-appearance-standard.mat-button-toggle-disabled.mat-button-toggle-checked{color:var(--mat-standard-button-toggle-disabled-selected-state-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent));background-color:var(--mat-standard-button-toggle-disabled-selected-state-background-color, color-mix(in srgb, var(--mat-sys-on-surface) 12%, transparent))}.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay{background-color:var(--mat-standard-button-toggle-state-layer-color, var(--mat-sys-on-surface))}.mat-button-toggle-appearance-standard:hover .mat-button-toggle-focus-overlay{opacity:var(--mat-standard-button-toggle-hover-state-layer-opacity, var(--mat-sys-hover-state-layer-opacity))}.mat-button-toggle-appearance-standard.cdk-keyboard-focused .mat-button-toggle-focus-overlay{opacity:var(--mat-standard-button-toggle-focus-state-layer-opacity, var(--mat-sys-focus-state-layer-opacity))}@media(hover: none){.mat-button-toggle-appearance-standard:hover .mat-button-toggle-focus-overlay{display:none}}.mat-button-toggle-label-content{-webkit-user-select:none;user-select:none;display:inline-block;padding:0 16px;line-height:var(--mat-legacy-button-toggle-height);position:relative}.mat-button-toggle-appearance-standard .mat-button-toggle-label-content{padding:0 12px;line-height:var(--mat-standard-button-toggle-height, 40px)}.mat-button-toggle-label-content>*{vertical-align:middle}.mat-button-toggle-focus-overlay{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:inherit;pointer-events:none;opacity:0;background-color:var(--mat-legacy-button-toggle-state-layer-color)}@media(forced-colors: active){.mat-button-toggle-checked .mat-button-toggle-focus-overlay{border-bottom:solid 500px;opacity:.5;height:0}.mat-button-toggle-checked:hover .mat-button-toggle-focus-overlay{opacity:.6}.mat-button-toggle-checked.mat-button-toggle-appearance-standard .mat-button-toggle-focus-overlay{border-bottom:solid 500px}}.mat-button-toggle .mat-button-toggle-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}.mat-button-toggle-button{border:0;background:none;color:inherit;padding:0;margin:0;font:inherit;outline:none;width:100%;cursor:pointer}.mat-button-toggle-animations-enabled .mat-button-toggle-button{transition:padding 150ms 45ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-button-toggle-vertical .mat-button-toggle-button{transition:none}.mat-button-toggle-disabled .mat-button-toggle-button{cursor:default}.mat-button-toggle-button::-moz-focus-inner{border:0}.mat-button-toggle-checked .mat-button-toggle-button:has(.mat-button-toggle-checkbox-wrapper){padding-left:30px}[dir=rtl] .mat-button-toggle-checked .mat-button-toggle-button:has(.mat-button-toggle-checkbox-wrapper){padding-left:0;padding-right:30px}.mat-button-toggle-standalone.mat-button-toggle-appearance-standard{--mat-focus-indicator-border-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full))}.mat-button-toggle-group-appearance-standard:not(.mat-button-toggle-vertical) .mat-button-toggle:last-of-type .mat-button-toggle-button::before{border-top-right-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full));border-bottom-right-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full))}.mat-button-toggle-group-appearance-standard:not(.mat-button-toggle-vertical) .mat-button-toggle:first-of-type .mat-button-toggle-button::before{border-top-left-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full));border-bottom-left-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full))}.mat-button-toggle-group-appearance-standard.mat-button-toggle-vertical .mat-button-toggle:last-of-type .mat-button-toggle-button::before{border-bottom-right-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full));border-bottom-left-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full))}.mat-button-toggle-group-appearance-standard.mat-button-toggle-vertical .mat-button-toggle:first-of-type .mat-button-toggle-button::before{border-top-right-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full));border-top-left-radius:var(--mat-standard-button-toggle-shape, var(--mat-sys-corner-full))}"],encapsulation:2,changeDetection:0})}return t})(),ScA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,Ps,n7,it]})}return t})();function DzA(t,e){t&1&&(S(0,"p"),AA(1,"Conversations"),F())}function yzA(t,e){t&1&&(S(0,"p"),AA(1,"Trace"),F())}function vzA(t,e){if(t&1){let A=be();S(0,"mat-button-toggle-group",5),da("ngModelChange",function(n){RA(A);let o=O(2);return Va(o.view,n)||(o.view=n),xA(n)}),S(1,"mat-button-toggle",6),AA(2,"Events"),F(),S(3,"mat-button-toggle",7),AA(4,"Trace"),F()()}if(t&2){let A=O(2);Ca("ngModel",A.view)}}function bzA(t,e){if(t&1){let A=be();S(0,"mat-list-item",8),hA("click",function(){let n=RA(A).$implicit,o=O(3);return xA(o.selectEvent(n.key))}),S(1,"span",9),AA(2),F(),S(3,"span",10),AA(4),F()()}if(t&2){let A=e.$implicit,i=e.$index;G(2),Gt(i),G(2),Gt(A.value.title)}}function MzA(t,e){if(t&1&&(S(0,"mat-list",4),Dn(1,bzA,5,2,"mat-list-item",null,to),Za(3,"keyvalue"),F()),t&2){let A=O(2);G(),yn(Lh(3,0,A.eventsMap,A.mapOrderPreservingSort))}}function kzA(t,e){if(t&1){let A=be();S(0,"mat-list-item",8),hA("click",function(){let n=RA(A).$implicit,o=O(3);return xA(o.openDialog(n.key))}),S(1,"span",9),AA(2),F(),S(3,"span"),AA(4),F()()}if(t&2){let A=e.$implicit,i=e.$index,n=O(3);G(2),Gt(i),G(2),Et("Invocation ",n.findInvocIdFromTraceId(A.key),"")}}function SzA(t,e){if(t&1&&(S(0,"mat-list",4),Dn(1,kzA,5,2,"mat-list-item",null,to),Za(3,"keyvalue"),F()),t&2){let A=O(2);G(),yn(Lh(3,0,A.invocTraces,A.mapOrderPreservingSort))}}function RzA(t,e){if(t&1&&(S(0,"div",1)(1,"div",2),_A(2,DzA,2,0,"p")(3,yzA,2,0,"p")(4,vzA,5,1,"mat-button-toggle-group",3),F(),_A(5,MzA,4,3,"mat-list",4)(6,SzA,4,3,"mat-list",4),F()),t&2){let A=O();G(2),GA(A.isTraceView()?-1:2),G(),GA(A.isTraceView()?3:-1),G(),GA(A.traceData?4:-1),G(),GA(A.isTraceView()?-1:5),G(),GA(A.isTraceView()?6:-1)}}function xzA(t,e){t&1&&(S(0,"div")(1,"p"),AA(2,"No conversations"),F()())}var nd=class t{constructor(e){this.dialog=e}eventsMap=new Map;selectedEvent=new WA;traceData=[];llmRequest=void 0;llmResponse=void 0;llmRequestKey="gcp.vertex.agent.llm_request";llmResponseKey="gcp.vertex.agent.llm_response";isDetailsPanelOpen=!1;view="events";invocTraces=new Map;ngOnChanges(e){"traceData"in e&&this.prcessTraceDataToInvocTrace()}showJson=Array(this.eventsMap.size).fill(!1);toggleJson(e){this.showJson[e]=!this.showJson[e]}selectEvent(e){this.selectedEvent.emit(e)}isTraceView(){return this.view=="trace"}mapOrderPreservingSort=(e,A)=>0;prcessTraceDataToInvocTrace(){!this.traceData||this.traceData.length==0||(this.invocTraces=this.traceData.reduce((e,A)=>{let i=A.trace_id,n=e.get(i);return n?(n.push(A),n.sort((o,r)=>o.start_time-r.start_time)):e.set(i,[A]),e},new Map))}findInvocIdFromTraceId(e){return this.invocTraces.get(e)?.find(i=>i.attributes!==void 0&&"gcp.vertex.agent.invocation_id"in i.attributes).attributes["gcp.vertex.agent.invocation_id"]}openDialog(e){let A=this.dialog.open(hf,{width:"auto",maxWidth:"90vw",data:{spans:this.invocTraces.get(e),invocId:this.findInvocIdFromTraceId(e)}})}static \u0275fac=function(A){return new(A||t)(PA(qs))};static \u0275cmp=zA({type:t,selectors:[["app-event-tab"]],inputs:{eventsMap:"eventsMap",traceData:"traceData"},outputs:{selectedEvent:"selectedEvent"},standalone:!1,features:[ti],decls:3,vars:2,consts:[[1,"events-wrapper"],[1,"events-container"],[1,"event-header"],["name","fontStyle","aria-label","Font Style",2,"scale","0.8",3,"ngModel"],[1,"event-list"],["name","fontStyle","aria-label","Font Style",2,"scale","0.8",3,"ngModelChange","ngModel"],["value","events"],["value","trace"],[3,"click"],[1,"event-index"],[1,"event-title"]],template:function(A,i){A&1&&(S(0,"div",0),_A(1,RzA,7,5,"div",1)(2,xzA,3,0,"div"),F()),A&2&&(G(),GA(i.eventsMap.size>0?1:-1),G(),GA(i.eventsMap.size==0?2:-1))},dependencies:[Ea,ec,ycA,vcA,_G,n7,Th],styles:[".events-wrapper[_ngcontent-%COMP%]{padding-left:25px;padding-right:25px;color:#9aa0a6;font-size:14px;font-weight:700}.event-index[_ngcontent-%COMP%]{color:#80868b;font-family:Roboto;font-size:14px;font-style:normal;font-weight:400;margin-right:10px}.event-title[_ngcontent-%COMP%]{font-family:Google Sans Mono,monospace}.spacer[_ngcontent-%COMP%]{flex:1 1 auto}.events-container[_ngcontent-%COMP%]{margin-top:20px}.event-container[_ngcontent-%COMP%]{display:flex;flex-direction:row;margin-top:20px}.function-event-button[_ngcontent-%COMP%]{margin-top:11px}.event-list[_ngcontent-%COMP%]{--mat-list-active-indicator-color: orange}.event-list[_ngcontent-%COMP%]{--mdc-list-list-item-container-color: #2b2b2f}.event-list[_ngcontent-%COMP%]{--mdc-list-list-item-label-text-size: 14px}.event-list[_ngcontent-%COMP%]{--mdc-list-list-item-label-text-weight: 400}.event-list[_ngcontent-%COMP%]{--mdc-list-list-item-one-line-container-height: 52px}[_nghost-%COMP%]     .mdc-list-item{border:1px solid #5f6368;cursor:pointer}[_nghost-%COMP%]     .mdc-list-item:hover{background-color:#1c1b1c}.event-header[_ngcontent-%COMP%]{display:flex;justify-content:space-between}"]})};function FzA(t,e){t&1&&(S(0,"h2",0),AA(1,"Events List"),F())}function NzA(t,e){t&1&&(S(0,"h2",0),AA(1,"Send Response To Pending Event"),F())}function _zA(t,e){t&1&&(S(0,"h2",4),AA(1,"Events List"),F())}function GzA(t,e){t&1&&(S(0,"h2",4),AA(1,"Send Response To Pending Event"),F())}function UzA(t,e){if(t&1){let A=be();S(0,"div")(1,"p"),AA(2,"Name"),F(),S(3,"p"),AA(4),F(),S(5,"p"),AA(6,"Args"),F(),S(7,"p"),AA(8),F(),S(9,"mat-form-field",5)(10,"mat-label"),AA(11,"Response"),F(),S(12,"textarea",6),da("ngModelChange",function(n){RA(A);let o=O();return Va(o.selectedEvent.response,n)||(o.selectedEvent.response=n),xA(n)}),F()()()}if(t&2){let A=O();G(4),Gt(A.selectedEvent.name),G(4),Gt(A.argsToJson(A.selectedEvent.args)),G(4),Ca("ngModel",A.selectedEvent.response)}}function KzA(t,e){if(t&1){let A=be();S(0,"button",7),hA("click",function(){RA(A);let n=O();return xA(n.sendResponse())}),AA(1),F()}if(t&2){let A=O();yA("disabled",A.sending),G(),Et(" ",A.sending?"Sending...":"Send"," ")}}var uf=class t{constructor(e,A,i){this.dialogRef=e;this.data=A;this.agentService=i;this.selectedEvent=A.event,this.appName=A.appName,this.userId=A.userId,this.sessionId=A.sessionId,this.functionCallEventId=A.functionCallEventId}selectedEvent=null;appName;userId;sessionId;functionCallEventId;sending=!1;response=[];argsToJson(e){return JSON.stringify(e)}sendResponse(){this.sending=!0;let e={appName:this.appName,userId:this.userId,sessionId:this.sessionId,newMessage:{role:"user",parts:[]}};this.selectedEvent.response&&(e.functionCallEventId=this.functionCallEventId,e.newMessage.parts.push({function_response:{id:this.selectedEvent.id,name:this.selectedEvent.name,response:{response:this.selectedEvent.response}}})),this.agentService.runSse(e).subscribe({next:A=>Ao(this,null,function*(){let i=JSON.parse(A);this.response.push(i)}),error:A=>console.error("SSE error:",A),complete:()=>{this.sending=!1,this.dialogRef.close({response:this.response,events:[this.selectedEvent]})}})}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(as),PA(H2))};static \u0275cmp=zA({type:t,selectors:[["app-pending-event-dialog"]],standalone:!1,decls:10,vars:6,consts:[["mat-dialog-title",""],["mat-dialog-title","","class","dialog-title",4,"ngIf"],["mat-button","",3,"disabled"],["mat-button","","mat-dialog-close",""],["mat-dialog-title","",1,"dialog-title"],["appearance","outline",1,"response-textarea"],["matInput","",3,"ngModelChange","ngModel"],["mat-button","",3,"click","disabled"]],template:function(A,i){A&1&&(_A(0,FzA,2,0,"h2",0)(1,NzA,2,0,"h2",0)(2,_zA,2,0,"h2",1)(3,GzA,2,0,"h2",1),S(4,"mat-dialog-content"),_A(5,UzA,13,3,"div"),F(),S(6,"mat-dialog-actions"),_A(7,KzA,2,2,"button",2),S(8,"button",3),AA(9,"Close"),F()()),A&2&&(GA(i.selectedEvent?-1:0),G(),GA(i.selectedEvent?1:-1),G(),yA("ngIf",!i.selectedEvent),G(),yA("ngIf",i.selectedEvent),G(2),GA(i.selectedEvent?5:-1),G(2),GA(i.selectedEvent&&i.selectedEvent.response?7:-1))},dependencies:[Uh,vc,Ea,ec,Qg,Q8,iI,Dr,bs,ma,pa,hg],styles:[".response-textarea[_ngcontent-%COMP%]{min-width:500px;margin-top:15px}.dialog-title[_ngcontent-%COMP%]{font-weight:700;font-size:large}"]})};var SQ=class t{constructor(e,A){this.dialogRef=e;this.data=A}onConfirm(){this.dialogRef.close(!0)}onCancel(){this.dialogRef.close(!1)}static \u0275fac=function(A){return new(A||t)(PA(Br),PA(as))};static \u0275cmp=zA({type:t,selectors:[["app-delete-session-dialog"]],standalone:!1,decls:11,vars:4,consts:[[1,"confirm-delete-wrapper"],["mat-dialog-title",""],["align","end"],["mat-button","",3,"click"],["mat-button","","cdkFocusInitial","",3,"click"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"h2",1),AA(2),F(),S(3,"mat-dialog-content")(4,"p"),AA(5),F()(),S(6,"mat-dialog-actions",2)(7,"button",3),hA("click",function(){return i.onCancel()}),AA(8),F(),S(9,"button",4),hA("click",function(){return i.onConfirm()}),AA(10),F()()()),A&2&&(G(2),Gt(i.data.title),G(3),Gt(i.data.message),G(3),Gt(i.data.cancelButtonText),G(2),Gt(i.data.confirmButtonText))},dependencies:[Dr,bs,ma,pa],encapsulation:2})};function YzA(t,e){if(t&1){let A=be();S(0,"div",3),hA("click",function(){let n=RA(A).$implicit,o=O();return xA(o.getSession(n.id))}),S(1,"div",4)(2,"div",5),AA(3),F(),S(4,"div",6),AA(5),F()()()}if(t&2){let A=e.$implicit,i=O();yA("ngClass",A.id===i.sessionId?"session-item current":"session-item"),G(3),Et(" ",A.id," "),G(2),Et(" ",i.getDate(A)," ")}}var od=class t{constructor(e,A){this.sessionService=e;this.dialog=A;this.refreshSessionsSubject.pipe(jn(()=>this.sessionService.listSessions(this.userId,this.appName))).subscribe(i=>{i=i.sort((n,o)=>Number(o.lastUpdateTime)-Number(n.lastUpdateTime)),this.sessionList=i})}userId="";appName="";sessionId="";sessionSelected=new WA;sessionReloaded=new WA;sessionList=[];refreshSessionsSubject=new OA;ngOnInit(){setTimeout(()=>{this.refreshSessionsSubject.next()},500)}getSession(e){this.sessionService.getSession(this.userId,this.appName,e).subscribe(A=>{let i=this.fromApiResultToSession(A);this.sessionSelected.emit(i)})}getDate(e){let A=e.lastUpdateTime;return new Date(A*1e3).toLocaleString()}fromApiResultToSession(e){return{id:e?.id??"",appName:e?.appName??"",userId:e?.userId??"",state:e?.state??[],events:e?.events??[]}}reloadSession(e){this.sessionService.getSession(this.userId,this.appName,e).subscribe(A=>{let i=this.fromApiResultToSession(A);this.sessionReloaded.emit(i)})}refreshSession(e){if(this.refreshSessionsSubject.next(),!(this.sessionList.length<=1)){let A=this.sessionList.findIndex(i=>i.id==e);return A==this.sessionList.length-1&&(A=-1),this.sessionList[A+1]}}static \u0275fac=function(A){return new(A||t)(PA(Wg),PA(qs))};static \u0275cmp=zA({type:t,selectors:[["app-session-tab"]],inputs:{userId:"userId",appName:"appName",sessionId:"sessionId"},outputs:{sessionSelected:"sessionSelected",sessionReloaded:"sessionReloaded"},standalone:!1,decls:4,vars:0,consts:[[1,"session-wrapper"],[1,"session-tab-container",2,"margin-top","16px"],[3,"ngClass"],[3,"click","ngClass"],[1,"session-info"],[1,"session-id"],[1,"session-date"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"div",1),Dn(2,YzA,6,3,"div",2,to),F()()),A&2&&(G(2),yn(i.sessionList))},dependencies:[Xa],styles:[".session-wrapper[_ngcontent-%COMP%]{padding-left:25px;padding-right:25px;color:#9aa0a6;font-size:14px;font-weight:700}.session-item[_ngcontent-%COMP%]{display:flex;justify-content:space-between;border:none;background-color:#303030;border-radius:8px;margin-bottom:4px;cursor:pointer}.session-item[_ngcontent-%COMP%]:hover{background-color:#141414}.session-item.current[_ngcontent-%COMP%]{background-color:#004a77}.session-id[_ngcontent-%COMP%]{color:#e8eaed;font-family:Google Sans Mono,monospace;font-size:14px;font-style:normal;font-weight:500;line-height:20px;letter-spacing:.25px}.session-date[_ngcontent-%COMP%]{color:#9aa0a6;font-family:Roboto;font-size:12px;font-style:normal;font-weight:400;line-height:16px;letter-spacing:.3px}.session-info[_ngcontent-%COMP%]{padding:11px}"]})};var RQ=class t{constructor(e){this.http=e}apiServerDomain=vs.getApiServerBaseUrl();getLatestArtifact(e,A,i,n){let o=this.apiServerDomain+`/apps/${A}/users/${e}/sessions/${i}/artifacts/${n}`;return this.http.get(o)}getArtifactVersion(e,A,i,n,o){let r=this.apiServerDomain+`/apps/${A}/users/${e}/sessions/${i}/artifacts/${n}/versions/${o}`;return this.http.get(r)}static \u0275fac=function(A){return new(A||t)(we(Ds))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var HzA={url:"",deserializer:t=>JSON.parse(t.data),serializer:t=>JSON.stringify(t)},zzA="WebSocketSubject.error must be called with an object with an error code, and an optional reason: { code: number, reason: string }",ff=class t extends md{constructor(e,A){if(super(),this._socket=null,e instanceof ct)this.destination=A,this.source=e;else{let i=this._config=Object.assign({},HzA);if(this._output=new OA,typeof e=="string")i.url=e;else for(let n in e)e.hasOwnProperty(n)&&(i[n]=e[n]);if(!i.WebSocketCtor&&WebSocket)i.WebSocketCtor=WebSocket;else if(!i.WebSocketCtor)throw new Error("no WebSocket constructor can be found");this.destination=new Al}}lift(e){let A=new t(this._config,this.destination);return A.operator=e,A.source=this,A}_resetState(){this._socket=null,this.source||(this.destination=new Al),this._output=new OA}multiplex(e,A,i){let n=this;return new ct(o=>{try{n.next(e())}catch(s){o.error(s)}let r=n.subscribe({next:s=>{try{i(s)&&o.next(s)}catch(a){o.error(a)}},error:s=>o.error(s),complete:()=>o.complete()});return()=>{try{n.next(A())}catch(s){o.error(s)}r.unsubscribe()}})}_connectSocket(){let{WebSocketCtor:e,protocol:A,url:i,binaryType:n}=this._config,o=this._output,r=null;try{r=A?new e(i,A):new e(i),this._socket=r,n&&(this._socket.binaryType=n)}catch(a){o.error(a);return}let s=new Kt(()=>{this._socket=null,r&&r.readyState===1&&r.close()});r.onopen=a=>{let{_socket:c}=this;if(!c){r.close(),this._resetState();return}let{openObserver:l}=this._config;l&&l.next(a);let I=this.destination;this.destination=o0.create(C=>{if(r.readyState===1)try{let{serializer:d}=this._config;r.send(d(C))}catch(d){this.destination.error(d)}},C=>{let{closingObserver:d}=this._config;d&&d.next(void 0),C&&C.code?r.close(C.code,C.reason):o.error(new TypeError(zzA)),this._resetState()},()=>{let{closingObserver:C}=this._config;C&&C.next(void 0),r.close(),this._resetState()}),I&&I instanceof Al&&s.add(I.subscribe(this.destination))},r.onerror=a=>{this._resetState(),o.error(a)},r.onclose=a=>{r===this._socket&&this._resetState();let{closeObserver:c}=this._config;c&&c.next(a),a.wasClean?o.complete():o.error(a)},r.onmessage=a=>{try{let{deserializer:c}=this._config;o.next(c(a))}catch(c){o.error(c)}}}_subscribe(e){let{source:A}=this;return A?A.subscribe(e):(this._socket||this._connectSocket(),this._output.subscribe(e),e.add(()=>{let{_socket:i}=this;this._output.observers.length===0&&(i&&(i.readyState===1||i.readyState===0)&&i.close(),this._resetState())}),e)}unsubscribe(){let{_socket:e}=this;e&&(e.readyState===1||e.readyState===0)&&e.close(),this._resetState(),super.unsubscribe()}};var Xg=class t{socket$;messages$=new Mi("");audioContext=new AudioContext({sampleRate:22e3});audioBuffer=[];audioIntervalId=null;lastAudioTime=0;closeReasonSubject=new OA;constructor(){}connect(e){this.socket$=new ff({url:e,serializer:A=>JSON.stringify(A),deserializer:A=>A.data,closeObserver:{next:A=>{this.emitWsCloseReason(A.reason)}}}),this.socket$.subscribe(A=>{this.handleIncomingAudio(A),this.messages$.next(A)},A=>{console.error("WebSocket error:",A)}),this.audioIntervalId=setInterval(()=>this.processBufferedAudio(),250)}sendMessage(e){if(e.blob.data=this.arrayBufferToBase64(e.blob.data.buffer),!this.socket$||this.socket$.closed){console.error("WebSocket is not open.");return}this.socket$.next(e)}closeConnection(){clearInterval(this.audioIntervalId),this.audioIntervalId=null,this.socket$&&this.socket$.complete()}getMessages(){return this.messages$.asObservable()}arrayBufferToBase64(e){let A="",i=new Uint8Array(e),n=i.byteLength;for(let o=0;o<n;o++)A+=String.fromCharCode(i[o]);return btoa(A)}handleIncomingAudio(e){let A=JSON.parse(e);if(A.content&&A.content.parts&&A.content.parts[0].inlineData){let i=this.base64ToUint8Array(A.content.parts[0].inlineData.data);this.audioBuffer.push(i)}}processBufferedAudio(){if(this.audioBuffer.length===0)return;let e=this.audioBuffer.reduce((n,o)=>n+o.length,0),A=new Uint8Array(e),i=0;for(let n of this.audioBuffer)A.set(n,i),i+=n.length;this.playPCM(A),this.audioBuffer=[]}base64ToUint8Array(e){let A=atob(this.urlSafeBase64ToBase64(e)),i=A.length,n=new Uint8Array(i);for(let o=0;o<i;o++)n[o]=A.charCodeAt(o);return n}playPCM(e){let A=new Float32Array(e.length/2);for(let s=0;s<A.length;s++){let a=e[s*2]|e[s*2+1]<<8;a>=32768&&(a-=65536),A[s]=a/32768}let i=this.audioContext.createBuffer(1,A.length,22e3);i.copyToChannel(A,0);let n=this.audioContext.createBufferSource();n.buffer=i,n.connect(this.audioContext.destination);let o=this.audioContext.currentTime,r=Math.max(this.lastAudioTime,o);n.start(r),this.lastAudioTime=r+i.duration}urlSafeBase64ToBase64(e){let A=e.replace(/_/g,"/").replace(/-/g,"+");for(;A.length%4!==0;)A+="=";return A}emitWsCloseReason(e){this.closeReasonSubject.next(e)}onCloseReason(){return this.closeReasonSubject.asObservable()}static \u0275fac=function(A){return new(A||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var xQ=class t{constructor(e){this.wsService=e}mediaRecorder;stream;audioContext;source;processor;audioBuffer=[];audioIntervalId=null;startRecording(){return Ao(this,null,function*(){try{this.stream=yield navigator.mediaDevices.getUserMedia({audio:!0}),this.audioContext=new AudioContext,yield this.audioContext.audioWorklet.addModule("./assets/audio-processor.js"),this.source=this.audioContext.createMediaStreamSource(this.stream);let e=new AudioWorkletNode(this.audioContext,"audio-processor");e.port.onmessage=A=>{let i=A.data,n=this.float32ToPCM(i);this.audioBuffer.push(n)},this.source.connect(e),e.connect(this.audioContext.destination),this.audioIntervalId=setInterval(()=>this.sendBufferedAudio(),250)}catch(e){console.error("Error accessing microphone:",e)}})}sendBufferedAudio(){if(this.audioBuffer.length===0)return;let e=this.audioBuffer.reduce((o,r)=>o+r.length,0),A=new Uint8Array(e),i=0;for(let o of this.audioBuffer)A.set(o,i),i+=o.length;let n={blob:{mime_type:"audio/pcm",data:A}};this.wsService.sendMessage(n),this.audioBuffer=[]}stopRecording(){this.processor&&this.processor.disconnect(),this.source&&this.source.disconnect(),this.audioContext&&this.audioContext.close(),this.stream&&this.stream.getTracks().forEach(e=>e.stop()),this.audioIntervalId&&(clearInterval(this.audioIntervalId),this.audioIntervalId=null)}float32ToPCM(e){let A=new ArrayBuffer(e.length*2),i=new DataView(A);for(let n=0;n<e.length;n++){let o=Math.max(-1,Math.min(1,e[n]));o=o<0?o*32768:o*32767,i.setInt16(n*2,o,!0)}return new Uint8Array(A)}static \u0275fac=function(A){return new(A||t)(we(Xg))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var LQ=class t{constructor(e,A){this.wsService=e;this.renderer=A.createRenderer(null,null)}mediaRecorder;stream;renderer;videoElement;videoBuffer=[];videoIntervalId=null;createVideoElement(e){this.clearVideoElement(e),this.videoElement=this.renderer.createElement("video"),this.renderer.setAttribute(this.videoElement,"width","400"),this.renderer.setAttribute(this.videoElement,"height","300"),this.renderer.setAttribute(this.videoElement,"autoplay","true"),this.renderer.setAttribute(this.videoElement,"muted","true"),this.renderer.appendChild(e.nativeElement,this.videoElement)}startRecording(e){return Ao(this,null,function*(){this.createVideoElement(e);try{this.stream=yield navigator.mediaDevices.getUserMedia({video:!0}),this.videoElement.srcObject=this.stream,this.mediaRecorder=new MediaRecorder(this.stream,{mimeType:"video/webm"}),this.mediaRecorder.start(1e3),this.videoIntervalId=setInterval(()=>this.captureAndSendFrame(),1e3)}catch(A){console.error("Error accessing camera/microphone:",A)}})}captureAndSendFrame(){return Ao(this,null,function*(){try{let e=yield this.captureFrame(),i={blob:{mime_type:"image/jpeg",data:yield this.blobToUint8Array(e)}};this.wsService.sendMessage(i)}catch(e){console.error("Error capturing frame:",e)}})}blobToUint8Array(e){return Ao(this,null,function*(){let A=yield e.arrayBuffer();return new Uint8Array(A)})}captureFrame(){return Ao(this,null,function*(){return new Promise((e,A)=>{try{let i=document.createElement("canvas");i.width=this.videoElement.videoWidth,i.height=this.videoElement.videoHeight;let n=i.getContext("2d");if(!n){A(new Error("Canvas context not supported"));return}n.drawImage(this.videoElement,0,0,i.width,i.height),i.toBlob(o=>{o?e(o):A(new Error("Failed to create image blob"))},"image/png")}catch(i){A(i)}})})}sendBufferedVideo(){if(this.videoBuffer.length===0)return;let e=this.videoBuffer.reduce((o,r)=>o+r.length,0),A=new Uint8Array(e),i=0;for(let o of this.videoBuffer)A.set(o,i),i+=o.length;let n={blob:{mime_type:"image/jpeg",data:A}};this.wsService.sendMessage(n),this.videoBuffer=[]}stopRecording(e){this.mediaRecorder&&this.mediaRecorder.stop(),this.stream&&this.stream.getTracks().forEach(A=>A.stop()),clearInterval(this.videoIntervalId),this.clearVideoElement(e)}clearVideoElement(e){let A=e.nativeElement.querySelector("video");A&&this.renderer.removeChild(e.nativeElement,A)}static \u0275fac=function(A){return new(A||t)(we(Xg),we(ws))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var nI=class t{constructor(e){this.http=e}apiServerDomain=vs.getApiServerBaseUrl();getEventTrace(e){let A=this.apiServerDomain+`/debug/trace/${e}`;return this.http.get(A)}getTrace(e){let A=this.apiServerDomain+`/debug/trace/session/${e}`;return this.http.get(A)}getEvent(e,A,i,n){let o=this.apiServerDomain+`/apps/${A}/users/${e}/sessions/${i}/events/${n}/graph`;return this.http.get(o)}static \u0275fac=function(A){return new(A||t)(we(Ds))};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var $g=class t{selectedTraceRowSource=new Mi(void 0);selectedTraceRow$=this.selectedTraceRowSource.asObservable();eventDataSource=new Mi(void 0);eventData$=this.eventDataSource.asObservable();hoveredMessageIndiciesSource=new Mi([]);hoveredMessageIndicies$=this.hoveredMessageIndiciesSource.asObservable();messagesSource=new Mi([]);messages$=this.messagesSource.asObservable();selectedRow(e){this.selectedTraceRowSource.next(e)}setEventData(e){this.eventDataSource.next(e)}setMessages(e){this.messagesSource.next(e)}setHoveredMessages(e,A){if(!e){this.hoveredMessageIndiciesSource.next([]);return}let i=e.attributes,n=i&&i["gcp.vertex.agent.event_id"],o=0,r=[];for(let s of this.messagesSource.value){if(s.role=="user"){o++;continue}if(this.eventDataSource.value?.get(s.eventId).invocationId!=A){o++;continue}if(n)if(i["gcp.vertex.agent.event_id"]==s.eventId){r.push(o),o++;continue}else{o++;continue}else{r.push(o),o++;continue}}this.hoveredMessageIndiciesSource.next(r)}resetTraceService(){this.eventDataSource.next(void 0),this.messagesSource.next([]),this.hoveredMessageIndiciesSource.next([])}static \u0275fac=function(A){return new(A||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac,providedIn:"root"})};var jzA=["*"];var qzA=new dA("MAT_CARD_CONFIG"),xcA=(()=>{class t{appearance;constructor(){let A=f(qzA,{optional:!0});this.appearance=A?.appearance||"raised"}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-card"]],hostAttrs:[1,"mat-mdc-card","mdc-card"],hostVars:4,hostBindings:function(i,n){i&2&&ue("mat-mdc-card-outlined",n.appearance==="outlined")("mdc-card--outlined",n.appearance==="outlined")},inputs:{appearance:"appearance"},exportAs:["matCard"],ngContentSelectors:jzA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},styles:['.mat-mdc-card{display:flex;flex-direction:column;box-sizing:border-box;position:relative;border-style:solid;border-width:0;background-color:var(--mdc-elevated-card-container-color, var(--mat-sys-surface-container-low));border-color:var(--mdc-elevated-card-container-color, var(--mat-sys-surface-container-low));border-radius:var(--mdc-elevated-card-container-shape, var(--mat-sys-corner-medium));box-shadow:var(--mdc-elevated-card-container-elevation, var(--mat-sys-level1))}.mat-mdc-card::after{position:absolute;top:0;left:0;width:100%;height:100%;border:solid 1px rgba(0,0,0,0);content:"";display:block;pointer-events:none;box-sizing:border-box;border-radius:var(--mdc-elevated-card-container-shape, var(--mat-sys-corner-medium))}.mat-mdc-card-outlined{background-color:var(--mdc-outlined-card-container-color, var(--mat-sys-surface));border-radius:var(--mdc-outlined-card-container-shape, var(--mat-sys-corner-medium));border-width:var(--mdc-outlined-card-outline-width, 1px);border-color:var(--mdc-outlined-card-outline-color, var(--mat-sys-outline-variant));box-shadow:var(--mdc-outlined-card-container-elevation, var(--mat-sys-level0))}.mat-mdc-card-outlined::after{border:none}.mdc-card__media{position:relative;box-sizing:border-box;background-repeat:no-repeat;background-position:center;background-size:cover}.mdc-card__media::before{display:block;content:""}.mdc-card__media:first-child{border-top-left-radius:inherit;border-top-right-radius:inherit}.mdc-card__media:last-child{border-bottom-left-radius:inherit;border-bottom-right-radius:inherit}.mat-mdc-card-actions{display:flex;flex-direction:row;align-items:center;box-sizing:border-box;min-height:52px;padding:8px}.mat-mdc-card-title{font-family:var(--mat-card-title-text-font, var(--mat-sys-title-large-font));line-height:var(--mat-card-title-text-line-height, var(--mat-sys-title-large-line-height));font-size:var(--mat-card-title-text-size, var(--mat-sys-title-large-size));letter-spacing:var(--mat-card-title-text-tracking, var(--mat-sys-title-large-tracking));font-weight:var(--mat-card-title-text-weight, var(--mat-sys-title-large-weight))}.mat-mdc-card-subtitle{color:var(--mat-card-subtitle-text-color, var(--mat-sys-on-surface));font-family:var(--mat-card-subtitle-text-font, var(--mat-sys-title-medium-font));line-height:var(--mat-card-subtitle-text-line-height, var(--mat-sys-title-medium-line-height));font-size:var(--mat-card-subtitle-text-size, var(--mat-sys-title-medium-size));letter-spacing:var(--mat-card-subtitle-text-tracking, var(--mat-sys-title-medium-tracking));font-weight:var(--mat-card-subtitle-text-weight, var(--mat-sys-title-medium-weight))}.mat-mdc-card-title,.mat-mdc-card-subtitle{display:block;margin:0}.mat-mdc-card-avatar~.mat-mdc-card-header-text .mat-mdc-card-title,.mat-mdc-card-avatar~.mat-mdc-card-header-text .mat-mdc-card-subtitle{padding:16px 16px 0}.mat-mdc-card-header{display:flex;padding:16px 16px 0}.mat-mdc-card-content{display:block;padding:0 16px}.mat-mdc-card-content:first-child{padding-top:16px}.mat-mdc-card-content:last-child{padding-bottom:16px}.mat-mdc-card-title-group{display:flex;justify-content:space-between;width:100%}.mat-mdc-card-avatar{height:40px;width:40px;border-radius:50%;flex-shrink:0;margin-bottom:16px;object-fit:cover}.mat-mdc-card-avatar~.mat-mdc-card-header-text .mat-mdc-card-subtitle,.mat-mdc-card-avatar~.mat-mdc-card-header-text .mat-mdc-card-title{line-height:normal}.mat-mdc-card-sm-image{width:80px;height:80px}.mat-mdc-card-md-image{width:112px;height:112px}.mat-mdc-card-lg-image{width:152px;height:152px}.mat-mdc-card-xl-image{width:240px;height:240px}.mat-mdc-card-subtitle~.mat-mdc-card-title,.mat-mdc-card-title~.mat-mdc-card-subtitle,.mat-mdc-card-header .mat-mdc-card-header-text .mat-mdc-card-title,.mat-mdc-card-header .mat-mdc-card-header-text .mat-mdc-card-subtitle,.mat-mdc-card-title-group .mat-mdc-card-title,.mat-mdc-card-title-group .mat-mdc-card-subtitle{padding-top:0}.mat-mdc-card-content>:last-child:not(.mat-mdc-card-footer){margin-bottom:0}.mat-mdc-card-actions-align-end{justify-content:flex-end}'],encapsulation:2,changeDetection:0})}return t})();var LcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,it]})}return t})();var ZzA=t=>["segment",t],WzA=(t,e)=>({"segment-main":!0,expandable:t,expanded:e});function XzA(t,e){t&1&&JA(0,"div",9)}function $zA(t,e){if(t&1&&(S(0,"span",10),AA(1),F()),t&2){let A=O().$implicit;G(),Gt(A.description)}}function AOA(t,e){if(t&1&&(S(0,"section",11),JA(1,"ngx-json-viewer",12),F()),t&2){let A=O().$implicit,i=O();G(),yA("json",A.value)("expanded",i.expanded)("depth",i.depth)("_currentDepth",i._currentDepth+1)}}function eOA(t,e){if(t&1){let A=be();S(0,"section",2)(1,"section",3),hA("click",function(){let n=RA(A).$implicit,o=O();return xA(o.toggle(n))}),_A(2,XzA,1,0,"div",4),S(3,"span",5),AA(4),F(),S(5,"span",6),AA(6,": "),F(),_A(7,$zA,2,1,"span",7),F(),_A(8,AOA,2,4,"section",8),F()}if(t&2){let A=e.$implicit,i=O();yA("ngClass",qr(6,ZzA,"segment-type-"+A.type)),G(),yA("ngClass",b2(8,WzA,i.isExpandable(A),A.expanded)),G(),yA("ngIf",i.isExpandable(A)),G(2),Gt(A.key),G(3),yA("ngIf",!A.expanded||!i.isExpandable(A)),G(),yA("ngIf",A.expanded&&i.isExpandable(A))}}var FQ=(()=>{class t{constructor(){this.expanded=!0,this.depth=-1,this._currentDepth=0,this.segments=[]}ngOnChanges(){this.segments=[],this.json=this.decycle(this.json),typeof this.json=="object"?Object.keys(this.json).forEach(A=>{this.segments.push(this.parseKeyValue(A,this.json[A]))}):this.segments.push(this.parseKeyValue(`(${typeof this.json})`,this.json))}isExpandable(A){return A.type==="object"||A.type==="array"}toggle(A){this.isExpandable(A)&&(A.expanded=!A.expanded)}parseKeyValue(A,i){let n={key:A,value:i,type:void 0,description:""+i,expanded:this.isExpanded()};switch(typeof n.value){case"number":{n.type="number";break}case"boolean":{n.type="boolean";break}case"function":{n.type="function";break}case"string":{n.type="string",n.description='"'+n.value+'"';break}case"undefined":{n.type="undefined",n.description="undefined";break}case"object":{n.value===null?(n.type="null",n.description="null"):Array.isArray(n.value)?(n.type="array",n.description="Array["+n.value.length+"] "+JSON.stringify(n.value)):n.value instanceof Date?n.type="date":(n.type="object",n.description="Object "+JSON.stringify(n.value));break}}return n}isExpanded(){return this.expanded&&!(this.depth>-1&&this._currentDepth>=this.depth)}decycle(A){let i=new WeakMap;return function n(o,r){let s,a;return typeof o=="object"&&o!==null&&!(o instanceof Boolean)&&!(o instanceof Date)&&!(o instanceof Number)&&!(o instanceof RegExp)&&!(o instanceof String)?(s=i.get(o),s!==void 0?{$ref:s}:(i.set(o,r),Array.isArray(o)?(a=[],o.forEach(function(c,l){a[l]=n(c,r+"["+l+"]")})):(a={},Object.keys(o).forEach(function(c){a[c]=n(o[c],r+"["+JSON.stringify(c)+"]")})),a)):o}(A,"$")}}return t.\u0275fac=function(A){return new(A||t)},t.\u0275cmp=zA({type:t,selectors:[["ngx-json-viewer"]],inputs:{json:"json",expanded:"expanded",depth:"depth",_currentDepth:"_currentDepth"},standalone:!1,features:[ti],decls:2,vars:1,consts:[[1,"ngx-json-viewer"],[3,"ngClass",4,"ngFor","ngForOf"],[3,"ngClass"],[3,"click","ngClass"],["class","toggler",4,"ngIf"],[1,"segment-key"],[1,"segment-separator"],["class","segment-value",4,"ngIf"],["class","children",4,"ngIf"],[1,"toggler"],[1,"segment-value"],[1,"children"],[3,"json","expanded","depth","_currentDepth"]],template:function(A,i){A&1&&(S(0,"section",0),_A(1,eOA,9,11,"section",1),F()),A&2&&(G(),yA("ngForOf",i.segments))},dependencies:[Xa,Hp,Uh,t],styles:['@charset "UTF-8";.ngx-json-viewer[_ngcontent-%COMP%]{font-family:var(--ngx-json-font-family, monospace);font-size:var(--ngx-json-font-size, 1em);width:100%;height:100%;overflow:hidden;position:relative}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]{padding:2px;margin:1px 1px 1px 12px}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]{word-wrap:break-word}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]   .toggler[_ngcontent-%COMP%]{position:absolute;margin-left:-14px;margin-top:3px;font-size:.8em;line-height:1.2em;vertical-align:middle;color:var(--ngx-json-toggler, #787878)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]   .toggler[_ngcontent-%COMP%]:after{display:inline-block;content:"\\25ba";transition:transform .1s ease-in}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]   .segment-key[_ngcontent-%COMP%]{color:var(--ngx-json-key, #4E187C)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]   .segment-separator[_ngcontent-%COMP%]{color:var(--ngx-json-separator, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .segment-main[_ngcontent-%COMP%]   .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-value, #000)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment[_ngcontent-%COMP%]   .children[_ngcontent-%COMP%]{margin-left:12px}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-string[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-string, #FF6B6B)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-number[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-number, #009688)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-boolean[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-boolean, #B938A4)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-date[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-date, #05668D)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-array[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-array, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-object[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-object, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-function[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-function, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-null[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-null, #fff)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-undefined[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{color:var(--ngx-json-undefined, #fff)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-null[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{background-color:var(--ngx-json-null-bg, red)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-undefined[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-key[_ngcontent-%COMP%]{color:var(--ngx-json-undefined-key, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-undefined[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%] > .segment-value[_ngcontent-%COMP%]{background-color:var(--ngx-json-undefined-key, #999)}.ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-object[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%], .ngx-json-viewer[_ngcontent-%COMP%]   .segment-type-array[_ngcontent-%COMP%] > .segment-main[_ngcontent-%COMP%]{white-space:nowrap}.ngx-json-viewer[_ngcontent-%COMP%]   .expanded[_ngcontent-%COMP%] > .toggler[_ngcontent-%COMP%]:after{transform:rotate(90deg)}.ngx-json-viewer[_ngcontent-%COMP%]   .expandable[_ngcontent-%COMP%], .ngx-json-viewer[_ngcontent-%COMP%]   .expandable[_ngcontent-%COMP%] > .toggler[_ngcontent-%COMP%]{cursor:pointer}']}),t})(),FcA=(()=>{class t{}return t.\u0275fac=function(A){return new(A||t)},t.\u0275mod=Ce({type:t}),t.\u0275inj=Ie({imports:[f0]}),t})();var NcA=["*"],tOA=["content"],iOA=[[["mat-drawer"]],[["mat-drawer-content"]],"*"],nOA=["mat-drawer","mat-drawer-content","*"];function oOA(t,e){if(t&1){let A=be();S(0,"div",1),hA("click",function(){RA(A);let n=O();return xA(n._onBackdropClicked())}),F()}if(t&2){let A=O();ue("mat-drawer-shown",A._isShowingBackdrop())}}function rOA(t,e){t&1&&(S(0,"mat-drawer-content"),Fe(1,2),F())}var sOA=new dA("MAT_DRAWER_DEFAULT_AUTOSIZE",{providedIn:"root",factory:aOA}),_cA=new dA("MAT_DRAWER_CONTAINER");function aOA(){return!1}var YG=(()=>{class t extends D0{_platform=f(Ii);_changeDetectorRef=f(It);_container=f(TG);constructor(){let A=f(ee),i=f(Y2),n=f(Qe);super(A,i,n)}ngAfterContentInit(){this._container._contentMarginChanges.subscribe(()=>{this._changeDetectorRef.markForCheck()})}_shouldBeHidden(){if(this._platform.isBrowser)return!1;let{start:A,end:i}=this._container;return A!=null&&A.mode!=="over"&&A.opened||i!=null&&i.mode!=="over"&&i.opened}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-drawer-content"]],hostAttrs:[1,"mat-drawer-content"],hostVars:6,hostBindings:function(i,n){i&2&&(uo("margin-left",n._container._contentMargins.left,"px")("margin-right",n._container._contentMargins.right,"px"),ue("mat-drawer-content-hidden",n._shouldBeHidden()))},features:[ht([{provide:D0,useExisting:t}]),lt],ngContentSelectors:NcA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},encapsulation:2,changeDetection:0})}return t})(),JG=(()=>{class t{_elementRef=f(ee);_focusTrapFactory=f(n8);_focusMonitor=f(dr);_platform=f(Ii);_ngZone=f(Qe);_renderer=f(Wi);_interactivityChecker=f(Mu);_doc=f(at,{optional:!0});_container=f(_cA,{optional:!0});_focusTrap=null;_elementFocusedBeforeDrawerWasOpened=null;_eventCleanups;_isAttached;_anchor;get position(){return this._position}set position(A){A=A==="end"?"end":"start",A!==this._position&&(this._isAttached&&this._updatePositionInParent(A),this._position=A,this.onPositionChanged.emit())}_position="start";get mode(){return this._mode}set mode(A){this._mode=A,this._updateFocusTrapState(),this._modeChanged.next()}_mode="over";get disableClose(){return this._disableClose}set disableClose(A){this._disableClose=Xo(A)}_disableClose=!1;get autoFocus(){let A=this._autoFocus;return A??(this.mode==="side"?"dialog":"first-tabbable")}set autoFocus(A){(A==="true"||A==="false"||A==null)&&(A=Xo(A)),this._autoFocus=A}_autoFocus;get opened(){return this._opened}set opened(A){this.toggle(Xo(A))}_opened=!1;_openedVia;_animationStarted=new OA;_animationEnd=new OA;openedChange=new WA(!0);_openedStream=this.openedChange.pipe(kt(A=>A),je(()=>{}));openedStart=this._animationStarted.pipe(kt(()=>this.opened),vd(void 0));_closedStream=this.openedChange.pipe(kt(A=>!A),je(()=>{}));closedStart=this._animationStarted.pipe(kt(()=>!this.opened),vd(void 0));_destroyed=new OA;onPositionChanged=new WA;_content;_modeChanged=new OA;_injector=f(Rt);_changeDetectorRef=f(It);constructor(){this.openedChange.pipe(St(this._destroyed)).subscribe(A=>{A?(this._doc&&(this._elementFocusedBeforeDrawerWasOpened=this._doc.activeElement),this._takeFocus()):this._isFocusWithinDrawer()&&this._restoreFocus(this._openedVia||"program")}),this._ngZone.runOutsideAngular(()=>{let A=this._elementRef.nativeElement;oh(A,"keydown").pipe(kt(i=>i.keyCode===27&&!this.disableClose&&!ir(i)),St(this._destroyed)).subscribe(i=>this._ngZone.run(()=>{this.close(),i.stopPropagation(),i.preventDefault()})),this._eventCleanups=[this._renderer.listen(A,"transitionrun",this._handleTransitionEvent),this._renderer.listen(A,"transitionend",this._handleTransitionEvent),this._renderer.listen(A,"transitioncancel",this._handleTransitionEvent)]}),this._animationEnd.subscribe(()=>{this.openedChange.emit(this._opened)})}_forceFocus(A,i){this._interactivityChecker.isFocusable(A)||(A.tabIndex=-1,this._ngZone.runOutsideAngular(()=>{let n=()=>{o(),r(),A.removeAttribute("tabindex")},o=this._renderer.listen(A,"blur",n),r=this._renderer.listen(A,"mousedown",n)})),A.focus(i)}_focusByCssSelector(A,i){let n=this._elementRef.nativeElement.querySelector(A);n&&this._forceFocus(n,i)}_takeFocus(){if(!this._focusTrap)return;let A=this._elementRef.nativeElement;switch(this.autoFocus){case!1:case"dialog":return;case!0:case"first-tabbable":To(()=>{!this._focusTrap.focusInitialElement()&&typeof A.focus=="function"&&A.focus()},{injector:this._injector});break;case"first-heading":this._focusByCssSelector('h1, h2, h3, h4, h5, h6, [role="heading"]');break;default:this._focusByCssSelector(this.autoFocus);break}}_restoreFocus(A){this.autoFocus!=="dialog"&&(this._elementFocusedBeforeDrawerWasOpened?this._focusMonitor.focusVia(this._elementFocusedBeforeDrawerWasOpened,A):this._elementRef.nativeElement.blur(),this._elementFocusedBeforeDrawerWasOpened=null)}_isFocusWithinDrawer(){let A=this._doc.activeElement;return!!A&&this._elementRef.nativeElement.contains(A)}ngAfterViewInit(){this._isAttached=!0,this._position==="end"&&this._updatePositionInParent("end"),this._platform.isBrowser&&(this._focusTrap=this._focusTrapFactory.create(this._elementRef.nativeElement),this._updateFocusTrapState())}ngOnDestroy(){this._eventCleanups.forEach(A=>A()),this._focusTrap?.destroy(),this._anchor?.remove(),this._anchor=null,this._animationStarted.complete(),this._animationEnd.complete(),this._modeChanged.complete(),this._destroyed.next(),this._destroyed.complete()}open(A){return this.toggle(!0,A)}close(){return this.toggle(!1)}_closeViaBackdropClick(){return this._setOpen(!1,!0,"mouse")}toggle(A=!this.opened,i){A&&i&&(this._openedVia=i);let n=this._setOpen(A,!A&&this._isFocusWithinDrawer(),this._openedVia||"program");return A||(this._openedVia=null),n}_setOpen(A,i,n){return A===this._opened?Promise.resolve(A?"open":"close"):(this._opened=A,this._container?._transitionsEnabled?this._setIsAnimating(!0):setTimeout(()=>{this._animationStarted.next(),this._animationEnd.next()}),this._elementRef.nativeElement.classList.toggle("mat-drawer-opened",A),!A&&i&&this._restoreFocus(n),this._changeDetectorRef.markForCheck(),this._updateFocusTrapState(),new Promise(o=>{this.openedChange.pipe(On(1)).subscribe(r=>o(r?"open":"close"))}))}_setIsAnimating(A){this._elementRef.nativeElement.classList.toggle("mat-drawer-animating",A)}_getWidth(){return this._elementRef.nativeElement.offsetWidth||0}_updateFocusTrapState(){this._focusTrap&&(this._focusTrap.enabled=!!this._container?.hasBackdrop&&this.opened)}_updatePositionInParent(A){if(!this._platform.isBrowser)return;let i=this._elementRef.nativeElement,n=i.parentNode;A==="end"?(this._anchor||(this._anchor=this._doc.createComment("mat-drawer-anchor"),n.insertBefore(this._anchor,i)),n.appendChild(i)):this._anchor&&this._anchor.parentNode.insertBefore(i,this._anchor)}_handleTransitionEvent=A=>{let i=this._elementRef.nativeElement;A.target===i&&this._ngZone.run(()=>{A.type==="transitionrun"?this._animationStarted.next(A):(A.type==="transitionend"&&this._setIsAnimating(!1),this._animationEnd.next(A))})};static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-drawer"]],viewQuery:function(i,n){if(i&1&&Te(tOA,5),i&2){let o;XA(o=$A())&&(n._content=o.first)}},hostAttrs:["tabIndex","-1",1,"mat-drawer"],hostVars:11,hostBindings:function(i,n){i&2&&(Ne("align",null),uo("visibility",!n._container&&!n.opened?"hidden":null),ue("mat-drawer-end",n.position==="end")("mat-drawer-over",n.mode==="over")("mat-drawer-push",n.mode==="push")("mat-drawer-side",n.mode==="side"))},inputs:{position:"position",mode:"mode",disableClose:"disableClose",autoFocus:"autoFocus",opened:"opened"},outputs:{openedChange:"openedChange",_openedStream:"opened",openedStart:"openedStart",_closedStream:"closed",closedStart:"closedStart",onPositionChanged:"positionChanged"},exportAs:["matDrawer"],ngContentSelectors:NcA,decls:3,vars:0,consts:[["content",""],["cdkScrollable","",1,"mat-drawer-inner-container"]],template:function(i,n){i&1&&(jt(),S(0,"div",1,0),Fe(2),F())},dependencies:[D0],encapsulation:2,changeDetection:0})}return t})(),TG=(()=>{class t{_dir=f(mo,{optional:!0});_element=f(ee);_ngZone=f(Qe);_changeDetectorRef=f(It);_animationMode=f(Si,{optional:!0});_transitionsEnabled=!1;_allDrawers;_drawers=new ca;_content;_userContent;get start(){return this._start}get end(){return this._end}get autosize(){return this._autosize}set autosize(A){this._autosize=Xo(A)}_autosize=f(sOA);get hasBackdrop(){return this._drawerHasBackdrop(this._start)||this._drawerHasBackdrop(this._end)}set hasBackdrop(A){this._backdropOverride=A==null?null:Xo(A)}_backdropOverride;backdropClick=new WA;_start;_end;_left;_right;_destroyed=new OA;_doCheckSubject=new OA;_contentMargins={left:null,right:null};_contentMarginChanges=new OA;get scrollable(){return this._userContent||this._content}_injector=f(Rt);constructor(){let A=f(Ii),i=f(Mc);this._dir?.change.pipe(St(this._destroyed)).subscribe(()=>{this._validateDrawers(),this.updateContentMargins()}),i.change().pipe(St(this._destroyed)).subscribe(()=>this.updateContentMargins()),this._animationMode!=="NoopAnimations"&&A.isBrowser&&this._ngZone.runOutsideAngular(()=>{setTimeout(()=>{this._element.nativeElement.classList.add("mat-drawer-transition"),this._transitionsEnabled=!0},200)})}ngAfterContentInit(){this._allDrawers.changes.pipe(Pn(this._allDrawers),St(this._destroyed)).subscribe(A=>{this._drawers.reset(A.filter(i=>!i._container||i._container===this)),this._drawers.notifyOnChanges()}),this._drawers.changes.pipe(Pn(null)).subscribe(()=>{this._validateDrawers(),this._drawers.forEach(A=>{this._watchDrawerToggle(A),this._watchDrawerPosition(A),this._watchDrawerMode(A)}),(!this._drawers.length||this._isDrawerOpen(this._start)||this._isDrawerOpen(this._end))&&this.updateContentMargins(),this._changeDetectorRef.markForCheck()}),this._ngZone.runOutsideAngular(()=>{this._doCheckSubject.pipe(el(10),St(this._destroyed)).subscribe(()=>this.updateContentMargins())})}ngOnDestroy(){this._contentMarginChanges.complete(),this._doCheckSubject.complete(),this._drawers.destroy(),this._destroyed.next(),this._destroyed.complete()}open(){this._drawers.forEach(A=>A.open())}close(){this._drawers.forEach(A=>A.close())}updateContentMargins(){let A=0,i=0;if(this._left&&this._left.opened){if(this._left.mode=="side")A+=this._left._getWidth();else if(this._left.mode=="push"){let n=this._left._getWidth();A+=n,i-=n}}if(this._right&&this._right.opened){if(this._right.mode=="side")i+=this._right._getWidth();else if(this._right.mode=="push"){let n=this._right._getWidth();i+=n,A-=n}}A=A||null,i=i||null,(A!==this._contentMargins.left||i!==this._contentMargins.right)&&(this._contentMargins={left:A,right:i},this._ngZone.run(()=>this._contentMarginChanges.next(this._contentMargins)))}ngDoCheck(){this._autosize&&this._isPushed()&&this._ngZone.runOutsideAngular(()=>this._doCheckSubject.next())}_watchDrawerToggle(A){A._animationStarted.pipe(St(this._drawers.changes)).subscribe(()=>{this.updateContentMargins(),this._changeDetectorRef.markForCheck()}),A.mode!=="side"&&A.openedChange.pipe(St(this._drawers.changes)).subscribe(()=>this._setContainerClass(A.opened))}_watchDrawerPosition(A){A.onPositionChanged.pipe(St(this._drawers.changes)).subscribe(()=>{To({read:()=>this._validateDrawers()},{injector:this._injector})})}_watchDrawerMode(A){A._modeChanged.pipe(St(zn(this._drawers.changes,this._destroyed))).subscribe(()=>{this.updateContentMargins(),this._changeDetectorRef.markForCheck()})}_setContainerClass(A){let i=this._element.nativeElement.classList,n="mat-drawer-container-has-open";A?i.add(n):i.remove(n)}_validateDrawers(){this._start=this._end=null,this._drawers.forEach(A=>{A.position=="end"?(this._end!=null,this._end=A):(this._start!=null,this._start=A)}),this._right=this._left=null,this._dir&&this._dir.value==="rtl"?(this._left=this._end,this._right=this._start):(this._left=this._start,this._right=this._end)}_isPushed(){return this._isDrawerOpen(this._start)&&this._start.mode!="over"||this._isDrawerOpen(this._end)&&this._end.mode!="over"}_onBackdropClicked(){this.backdropClick.emit(),this._closeModalDrawersViaBackdrop()}_closeModalDrawersViaBackdrop(){[this._start,this._end].filter(A=>A&&!A.disableClose&&this._drawerHasBackdrop(A)).forEach(A=>A._closeViaBackdropClick())}_isShowingBackdrop(){return this._isDrawerOpen(this._start)&&this._drawerHasBackdrop(this._start)||this._isDrawerOpen(this._end)&&this._drawerHasBackdrop(this._end)}_isDrawerOpen(A){return A!=null&&A.opened}_drawerHasBackdrop(A){return this._backdropOverride==null?!!A&&A.mode!=="side":this._backdropOverride}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-drawer-container"]],contentQueries:function(i,n,o){if(i&1&&(ci(o,YG,5),ci(o,JG,5)),i&2){let r;XA(r=$A())&&(n._content=r.first),XA(r=$A())&&(n._allDrawers=r)}},viewQuery:function(i,n){if(i&1&&Te(YG,5),i&2){let o;XA(o=$A())&&(n._userContent=o.first)}},hostAttrs:[1,"mat-drawer-container"],hostVars:2,hostBindings:function(i,n){i&2&&ue("mat-drawer-container-explicit-backdrop",n._backdropOverride)},inputs:{autosize:"autosize",hasBackdrop:"hasBackdrop"},outputs:{backdropClick:"backdropClick"},exportAs:["matDrawerContainer"],features:[ht([{provide:_cA,useExisting:t}])],ngContentSelectors:nOA,decls:4,vars:2,consts:[[1,"mat-drawer-backdrop",3,"mat-drawer-shown"],[1,"mat-drawer-backdrop",3,"click"]],template:function(i,n){i&1&&(jt(iOA),_A(0,oOA,1,2,"div",0),Fe(1),Fe(2,1),_A(3,rOA,2,0,"mat-drawer-content")),i&2&&(GA(n.hasBackdrop?0:-1),G(3),GA(n._content?-1:3))},dependencies:[YG],styles:[".mat-drawer-container{position:relative;z-index:1;color:var(--mat-sidenav-content-text-color, var(--mat-sys-on-background));background-color:var(--mat-sidenav-content-background-color, var(--mat-sys-background));box-sizing:border-box;display:block;overflow:hidden}.mat-drawer-container[fullscreen]{top:0;left:0;right:0;bottom:0;position:absolute}.mat-drawer-container[fullscreen].mat-drawer-container-has-open{overflow:hidden}.mat-drawer-container.mat-drawer-container-explicit-backdrop .mat-drawer-side{z-index:3}.mat-drawer-container.ng-animate-disabled .mat-drawer-backdrop,.mat-drawer-container.ng-animate-disabled .mat-drawer-content,.ng-animate-disabled .mat-drawer-container .mat-drawer-backdrop,.ng-animate-disabled .mat-drawer-container .mat-drawer-content{transition:none}.mat-drawer-backdrop{top:0;left:0;right:0;bottom:0;position:absolute;display:block;z-index:3;visibility:hidden}.mat-drawer-backdrop.mat-drawer-shown{visibility:visible;background-color:var(--mat-sidenav-scrim-color, color-mix(in srgb, var(--mat-sys-neutral-variant20) 40%, transparent))}.mat-drawer-transition .mat-drawer-backdrop{transition-duration:400ms;transition-timing-function:cubic-bezier(0.25, 0.8, 0.25, 1);transition-property:background-color,visibility}@media(forced-colors: active){.mat-drawer-backdrop{opacity:.5}}.mat-drawer-content{position:relative;z-index:1;display:block;height:100%;overflow:auto}.mat-drawer-content.mat-drawer-content-hidden{opacity:0}.mat-drawer-transition .mat-drawer-content{transition-duration:400ms;transition-timing-function:cubic-bezier(0.25, 0.8, 0.25, 1);transition-property:transform,margin-left,margin-right}.mat-drawer{position:relative;z-index:4;color:var(--mat-sidenav-container-text-color, var(--mat-sys-on-surface-variant));box-shadow:var(--mat-sidenav-container-elevation-shadow, none);background-color:var(--mat-sidenav-container-background-color, var(--mat-sys-surface));border-top-right-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-bottom-right-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));width:var(--mat-sidenav-container-width, 360px);display:block;position:absolute;top:0;bottom:0;z-index:3;outline:0;box-sizing:border-box;overflow-y:auto;transform:translate3d(-100%, 0, 0)}@media(forced-colors: active){.mat-drawer,[dir=rtl] .mat-drawer.mat-drawer-end{border-right:solid 1px currentColor}}@media(forced-colors: active){[dir=rtl] .mat-drawer,.mat-drawer.mat-drawer-end{border-left:solid 1px currentColor;border-right:none}}.mat-drawer.mat-drawer-side{z-index:2}.mat-drawer.mat-drawer-end{right:0;transform:translate3d(100%, 0, 0);border-top-left-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-bottom-left-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-top-right-radius:0;border-bottom-right-radius:0}[dir=rtl] .mat-drawer{border-top-left-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-bottom-left-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-top-right-radius:0;border-bottom-right-radius:0;transform:translate3d(100%, 0, 0)}[dir=rtl] .mat-drawer.mat-drawer-end{border-top-right-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-bottom-right-radius:var(--mat-sidenav-container-shape, var(--mat-sys-corner-large));border-top-left-radius:0;border-bottom-left-radius:0;left:0;right:auto;transform:translate3d(-100%, 0, 0)}.mat-drawer-transition .mat-drawer{transition:transform 400ms cubic-bezier(0.25, 0.8, 0.25, 1)}.mat-drawer:not(.mat-drawer-opened):not(.mat-drawer-animating){visibility:hidden;box-shadow:none}.mat-drawer:not(.mat-drawer-opened):not(.mat-drawer-animating) .mat-drawer-inner-container{display:none}.mat-drawer.mat-drawer-opened.mat-drawer-opened{transform:none}.mat-drawer-side{box-shadow:none;border-right-color:var(--mat-sidenav-container-divider-color, transparent);border-right-width:1px;border-right-style:solid}.mat-drawer-side.mat-drawer-end{border-left-color:var(--mat-sidenav-container-divider-color, transparent);border-left-width:1px;border-left-style:solid;border-right:none}[dir=rtl] .mat-drawer-side{border-left-color:var(--mat-sidenav-container-divider-color, transparent);border-left-width:1px;border-left-style:solid;border-right:none}[dir=rtl] .mat-drawer-side.mat-drawer-end{border-right-color:var(--mat-sidenav-container-divider-color, transparent);border-right-width:1px;border-right-style:solid;border-left:none}.mat-drawer-inner-container{width:100%;height:100%;overflow:auto}.mat-sidenav-fixed{position:fixed}"],encapsulation:2,changeDetection:0})}return t})();var GcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,Cl,Cl,it]})}return t})();var PG=["*"];function lOA(t,e){t&1&&Fe(0)}var gOA=["tabListContainer"],IOA=["tabList"],COA=["tabListInner"],dOA=["nextPaginator"],BOA=["previousPaginator"],EOA=t=>({animationDuration:t}),QOA=(t,e)=>({value:t,params:e});function hOA(t,e){}var uOA=["tabBodyWrapper"],fOA=["tabHeader"];function mOA(t,e){}function pOA(t,e){if(t&1&&_A(0,mOA,0,0,"ng-template",12),t&2){let A=O().$implicit;yA("cdkPortalOutlet",A.templateLabel)}}function wOA(t,e){if(t&1&&AA(0),t&2){let A=O().$implicit;Gt(A.textLabel)}}function DOA(t,e){if(t&1){let A=be();S(0,"div",7,2),hA("click",function(){let n=RA(A),o=n.$implicit,r=n.$index,s=O(),a=cr(1);return xA(s._handleClick(o,a,r))})("cdkFocusChange",function(n){let o=RA(A).$index,r=O();return xA(r._tabFocusChanged(n,o))}),JA(2,"span",8)(3,"div",9),S(4,"span",10)(5,"span",11),_A(6,pOA,1,1,null,12)(7,wOA,1,1),F()()()}if(t&2){let A=e.$implicit,i=e.$index,n=cr(1),o=O();fo(A.labelClass),ue("mdc-tab--active",o.selectedIndex===i),yA("id",o._getTabLabelId(i))("disabled",A.disabled)("fitInkBarToContent",o.fitInkBarToContent),Ne("tabIndex",o._getTabIndex(i))("aria-posinset",i+1)("aria-setsize",o._tabs.length)("aria-controls",o._getTabContentId(i))("aria-selected",o.selectedIndex===i)("aria-label",A.ariaLabel||null)("aria-labelledby",!A.ariaLabel&&A.ariaLabelledby?A.ariaLabelledby:null),G(3),yA("matRippleTrigger",n)("matRippleDisabled",A.disabled||o.disableRipple),G(3),GA(A.templateLabel?6:7)}}function yOA(t,e){t&1&&Fe(0)}function vOA(t,e){if(t&1){let A=be();S(0,"mat-tab-body",13),hA("_onCentered",function(){RA(A);let n=O();return xA(n._removeTabBodyWrapperHeight())})("_onCentering",function(n){RA(A);let o=O();return xA(o._setTabBodyWrapperHeight(n))}),F()}if(t&2){let A=e.$implicit,i=e.$index,n=O();fo(A.bodyClass),ue("mat-mdc-tab-body-active",n.selectedIndex===i),yA("id",n._getTabContentId(i))("content",A.content)("position",A.position)("origin",A.origin)("animationDuration",n.animationDuration)("preserveContent",n.preserveContent),Ne("tabindex",n.contentTabIndex!=null&&n.selectedIndex===i?n.contentTabIndex:null)("aria-labelledby",n._getTabLabelId(i))("aria-hidden",n.selectedIndex!==i)}}var bOA=new dA("MatTabContent"),MOA=(()=>{class t{template=f(wn);constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matTabContent",""]],features:[ht([{provide:bOA,useExisting:t}])]})}return t})(),kOA=new dA("MatTabLabel"),YcA=new dA("MAT_TAB"),jG=(()=>{class t extends sP{_closestTab=f(YcA,{optional:!0});static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","mat-tab-label",""],["","matTabLabel",""]],features:[ht([{provide:kOA,useExisting:t}]),lt]})}return t})(),JcA=new dA("MAT_TAB_GROUP"),mf=(()=>{class t{_viewContainerRef=f(Nn);_closestTabGroup=f(JcA,{optional:!0});disabled=!1;get templateLabel(){return this._templateLabel}set templateLabel(A){this._setTemplateLabelInput(A)}_templateLabel;_explicitContent=void 0;_implicitContent;textLabel="";ariaLabel;ariaLabelledby;labelClass;bodyClass;_contentPortal=null;get content(){return this._contentPortal}_stateChanges=new OA;position=null;origin=null;isActive=!1;constructor(){f(Rn).load(lr)}ngOnChanges(A){(A.hasOwnProperty("textLabel")||A.hasOwnProperty("disabled"))&&this._stateChanges.next()}ngOnDestroy(){this._stateChanges.complete()}ngOnInit(){this._contentPortal=new ys(this._explicitContent||this._implicitContent,this._viewContainerRef)}_setTemplateLabelInput(A){A&&A._closestTab===this&&(this._templateLabel=A)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-tab"]],contentQueries:function(i,n,o){if(i&1&&(ci(o,jG,5),ci(o,MOA,7,wn)),i&2){let r;XA(r=$A())&&(n.templateLabel=r.first),XA(r=$A())&&(n._explicitContent=r.first)}},viewQuery:function(i,n){if(i&1&&Te(wn,7),i&2){let o;XA(o=$A())&&(n._implicitContent=o.first)}},hostAttrs:["hidden",""],inputs:{disabled:[2,"disabled","disabled",ie],textLabel:[0,"label","textLabel"],ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],labelClass:"labelClass",bodyClass:"bodyClass"},exportAs:["matTab"],features:[ht([{provide:YcA,useExisting:t}]),ti],ngContentSelectors:PG,decls:1,vars:0,template:function(i,n){i&1&&(jt(),_A(0,lOA,1,0,"ng-template"))},encapsulation:2})}return t})(),HG="mdc-tab-indicator--active",UcA="mdc-tab-indicator--no-transition",zG=class{_items;_currentItem;constructor(e){this._items=e}hide(){this._items.forEach(e=>e.deactivateInkBar()),this._currentItem=void 0}alignToElement(e){let A=this._items.find(n=>n.elementRef.nativeElement===e),i=this._currentItem;if(A!==i&&(i?.deactivateInkBar(),A)){let n=i?.elementRef.nativeElement.getBoundingClientRect?.();A.activateInkBar(n),this._currentItem=A}}},SOA=(()=>{class t{_elementRef=f(ee);_inkBarElement;_inkBarContentElement;_fitToContent=!1;get fitInkBarToContent(){return this._fitToContent}set fitInkBarToContent(A){this._fitToContent!==A&&(this._fitToContent=A,this._inkBarElement&&this._appendInkBarElement())}activateInkBar(A){let i=this._elementRef.nativeElement;if(!A||!i.getBoundingClientRect||!this._inkBarContentElement){i.classList.add(HG);return}let n=i.getBoundingClientRect(),o=A.width/n.width,r=A.left-n.left;i.classList.add(UcA),this._inkBarContentElement.style.setProperty("transform",`translateX(${r}px) scaleX(${o})`),i.getBoundingClientRect(),i.classList.remove(UcA),i.classList.add(HG),this._inkBarContentElement.style.setProperty("transform","")}deactivateInkBar(){this._elementRef.nativeElement.classList.remove(HG)}ngOnInit(){this._createInkBarElement()}ngOnDestroy(){this._inkBarElement?.remove(),this._inkBarElement=this._inkBarContentElement=null}_createInkBarElement(){let A=this._elementRef.nativeElement.ownerDocument||document,i=this._inkBarElement=A.createElement("span"),n=this._inkBarContentElement=A.createElement("span");i.className="mdc-tab-indicator",n.className="mdc-tab-indicator__content mdc-tab-indicator__content--underline",i.appendChild(this._inkBarContentElement),this._appendInkBarElement()}_appendInkBarElement(){this._inkBarElement;let A=this._fitToContent?this._elementRef.nativeElement.querySelector(".mdc-tab__content"):this._elementRef.nativeElement;A.appendChild(this._inkBarElement)}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,inputs:{fitInkBarToContent:[2,"fitInkBarToContent","fitInkBarToContent",ie]}})}return t})();var TcA=(()=>{class t extends SOA{elementRef=f(ee);disabled=!1;focus(){this.elementRef.nativeElement.focus()}getOffsetLeft(){return this.elementRef.nativeElement.offsetLeft}getOffsetWidth(){return this.elementRef.nativeElement.offsetWidth}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275dir=jA({type:t,selectors:[["","matTabLabelWrapper",""]],hostVars:3,hostBindings:function(i,n){i&2&&(Ne("aria-disabled",!!n.disabled),ue("mat-mdc-tab-disabled",n.disabled))},inputs:{disabled:[2,"disabled","disabled",ie]},features:[lt]})}return t})(),KcA={passive:!0},ROA=650,xOA=100,LOA=(()=>{class t{_elementRef=f(ee);_changeDetectorRef=f(It);_viewportRuler=f(Mc);_dir=f(mo,{optional:!0});_ngZone=f(Qe);_platform=f(Ii);_sharedResizeObserver=f(B8);_injector=f(Rt);_renderer=f(Wi);_animationMode=f(Si,{optional:!0});_eventCleanups;_scrollDistance=0;_selectedIndexChanged=!1;_destroyed=new OA;_showPaginationControls=!1;_disableScrollAfter=!0;_disableScrollBefore=!0;_tabLabelCount;_scrollDistanceChanged;_keyManager;_currentTextContent;_stopScrolling=new OA;disablePagination=!1;get selectedIndex(){return this._selectedIndex}set selectedIndex(A){let i=isNaN(A)?0:A;this._selectedIndex!=i&&(this._selectedIndexChanged=!0,this._selectedIndex=i,this._keyManager&&this._keyManager.updateActiveItem(i))}_selectedIndex=0;selectFocusedIndex=new WA;indexFocused=new WA;constructor(){this._eventCleanups=this._ngZone.runOutsideAngular(()=>[this._renderer.listen(this._elementRef.nativeElement,"mouseleave",()=>this._stopInterval())])}ngAfterViewInit(){this._eventCleanups.push(sk(this._renderer,this._previousPaginator.nativeElement,"touchstart",()=>this._handlePaginatorPress("before"),KcA),sk(this._renderer,this._nextPaginator.nativeElement,"touchstart",()=>this._handlePaginatorPress("after"),KcA))}ngAfterContentInit(){let A=this._dir?this._dir.change:Me("ltr"),i=this._sharedResizeObserver.observe(this._elementRef.nativeElement).pipe(el(32),St(this._destroyed)),n=this._viewportRuler.change(150).pipe(St(this._destroyed)),o=()=>{this.updatePagination(),this._alignInkBarToSelectedTab()};this._keyManager=new qI(this._items).withHorizontalOrientation(this._getLayoutDirection()).withHomeAndEnd().withWrap().skipPredicate(()=>!1),this._keyManager.updateActiveItem(this._selectedIndex),To(o,{injector:this._injector}),zn(A,n,i,this._items.changes,this._itemsResized()).pipe(St(this._destroyed)).subscribe(()=>{this._ngZone.run(()=>{Promise.resolve().then(()=>{this._scrollDistance=Math.max(0,Math.min(this._getMaxScrollDistance(),this._scrollDistance)),o()})}),this._keyManager.withHorizontalOrientation(this._getLayoutDirection())}),this._keyManager.change.subscribe(r=>{this.indexFocused.emit(r),this._setTabFocus(r)})}_itemsResized(){return typeof ResizeObserver!="function"?sr:this._items.changes.pipe(Pn(this._items),jn(A=>new ct(i=>this._ngZone.runOutsideAngular(()=>{let n=new ResizeObserver(o=>i.next(o));return A.forEach(o=>n.observe(o.elementRef.nativeElement)),()=>{n.disconnect()}}))),CI(1),kt(A=>A.some(i=>i.contentRect.width>0&&i.contentRect.height>0)))}ngAfterContentChecked(){this._tabLabelCount!=this._items.length&&(this.updatePagination(),this._tabLabelCount=this._items.length,this._changeDetectorRef.markForCheck()),this._selectedIndexChanged&&(this._scrollToLabel(this._selectedIndex),this._checkScrollingControls(),this._alignInkBarToSelectedTab(),this._selectedIndexChanged=!1,this._changeDetectorRef.markForCheck()),this._scrollDistanceChanged&&(this._updateTabScrollPosition(),this._scrollDistanceChanged=!1,this._changeDetectorRef.markForCheck())}ngOnDestroy(){this._eventCleanups.forEach(A=>A()),this._keyManager?.destroy(),this._destroyed.next(),this._destroyed.complete(),this._stopScrolling.complete()}_handleKeydown(A){if(!ir(A))switch(A.keyCode){case 13:case 32:if(this.focusIndex!==this.selectedIndex){let i=this._items.get(this.focusIndex);i&&!i.disabled&&(this.selectFocusedIndex.emit(this.focusIndex),this._itemSelected(A))}break;default:this._keyManager.onKeydown(A)}}_onContentChanges(){let A=this._elementRef.nativeElement.textContent;A!==this._currentTextContent&&(this._currentTextContent=A||"",this._ngZone.run(()=>{this.updatePagination(),this._alignInkBarToSelectedTab(),this._changeDetectorRef.markForCheck()}))}updatePagination(){this._checkPaginationEnabled(),this._checkScrollingControls(),this._updateTabScrollPosition()}get focusIndex(){return this._keyManager?this._keyManager.activeItemIndex:0}set focusIndex(A){!this._isValidIndex(A)||this.focusIndex===A||!this._keyManager||this._keyManager.setActiveItem(A)}_isValidIndex(A){return this._items?!!this._items.toArray()[A]:!0}_setTabFocus(A){if(this._showPaginationControls&&this._scrollToLabel(A),this._items&&this._items.length){this._items.toArray()[A].focus();let i=this._tabListContainer.nativeElement;this._getLayoutDirection()=="ltr"?i.scrollLeft=0:i.scrollLeft=i.scrollWidth-i.offsetWidth}}_getLayoutDirection(){return this._dir&&this._dir.value==="rtl"?"rtl":"ltr"}_updateTabScrollPosition(){if(this.disablePagination)return;let A=this.scrollDistance,i=this._getLayoutDirection()==="ltr"?-A:A;this._tabList.nativeElement.style.transform=`translateX(${Math.round(i)}px)`,(this._platform.TRIDENT||this._platform.EDGE)&&(this._tabListContainer.nativeElement.scrollLeft=0)}get scrollDistance(){return this._scrollDistance}set scrollDistance(A){this._scrollTo(A)}_scrollHeader(A){let i=this._tabListContainer.nativeElement.offsetWidth,n=(A=="before"?-1:1)*i/3;return this._scrollTo(this._scrollDistance+n)}_handlePaginatorClick(A){this._stopInterval(),this._scrollHeader(A)}_scrollToLabel(A){if(this.disablePagination)return;let i=this._items?this._items.toArray()[A]:null;if(!i)return;let n=this._tabListContainer.nativeElement.offsetWidth,{offsetLeft:o,offsetWidth:r}=i.elementRef.nativeElement,s,a;this._getLayoutDirection()=="ltr"?(s=o,a=s+r):(a=this._tabListInner.nativeElement.offsetWidth-o,s=a-r);let c=this.scrollDistance,l=this.scrollDistance+n;s<c?this.scrollDistance-=c-s:a>l&&(this.scrollDistance+=Math.min(a-l,s-c))}_checkPaginationEnabled(){if(this.disablePagination)this._showPaginationControls=!1;else{let A=this._tabListInner.nativeElement.scrollWidth,i=this._elementRef.nativeElement.offsetWidth,n=A-i>=5;n||(this.scrollDistance=0),n!==this._showPaginationControls&&(this._showPaginationControls=n,this._changeDetectorRef.markForCheck())}}_checkScrollingControls(){this.disablePagination?this._disableScrollAfter=this._disableScrollBefore=!0:(this._disableScrollBefore=this.scrollDistance==0,this._disableScrollAfter=this.scrollDistance==this._getMaxScrollDistance(),this._changeDetectorRef.markForCheck())}_getMaxScrollDistance(){let A=this._tabListInner.nativeElement.scrollWidth,i=this._tabListContainer.nativeElement.offsetWidth;return A-i||0}_alignInkBarToSelectedTab(){let A=this._items&&this._items.length?this._items.toArray()[this.selectedIndex]:null,i=A?A.elementRef.nativeElement:null;i?this._inkBar.alignToElement(i):this._inkBar.hide()}_stopInterval(){this._stopScrolling.next()}_handlePaginatorPress(A,i){i&&i.button!=null&&i.button!==0||(this._stopInterval(),II(ROA,xOA).pipe(St(zn(this._stopScrolling,this._destroyed))).subscribe(()=>{let{maxScrollDistance:n,distance:o}=this._scrollHeader(A);(o===0||o>=n)&&this._stopInterval()}))}_scrollTo(A){if(this.disablePagination)return{maxScrollDistance:0,distance:0};let i=this._getMaxScrollDistance();return this._scrollDistance=Math.max(0,Math.min(i,A)),this._scrollDistanceChanged=!0,this._checkScrollingControls(),{maxScrollDistance:i,distance:this._scrollDistance}}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,inputs:{disablePagination:[2,"disablePagination","disablePagination",ie],selectedIndex:[2,"selectedIndex","selectedIndex",zi]},outputs:{selectFocusedIndex:"selectFocusedIndex",indexFocused:"indexFocused"}})}return t})(),FOA=(()=>{class t extends LOA{_items;_tabListContainer;_tabList;_tabListInner;_nextPaginator;_previousPaginator;_inkBar;ariaLabel;ariaLabelledby;disableRipple=!1;ngAfterContentInit(){this._inkBar=new zG(this._items),super.ngAfterContentInit()}_itemSelected(A){A.preventDefault()}static \u0275fac=(()=>{let A;return function(n){return(A||(A=Hi(t)))(n||t)}})();static \u0275cmp=zA({type:t,selectors:[["mat-tab-header"]],contentQueries:function(i,n,o){if(i&1&&ci(o,TcA,4),i&2){let r;XA(r=$A())&&(n._items=r)}},viewQuery:function(i,n){if(i&1&&(Te(gOA,7),Te(IOA,7),Te(COA,7),Te(dOA,5),Te(BOA,5)),i&2){let o;XA(o=$A())&&(n._tabListContainer=o.first),XA(o=$A())&&(n._tabList=o.first),XA(o=$A())&&(n._tabListInner=o.first),XA(o=$A())&&(n._nextPaginator=o.first),XA(o=$A())&&(n._previousPaginator=o.first)}},hostAttrs:[1,"mat-mdc-tab-header"],hostVars:4,hostBindings:function(i,n){i&2&&ue("mat-mdc-tab-header-pagination-controls-enabled",n._showPaginationControls)("mat-mdc-tab-header-rtl",n._getLayoutDirection()=="rtl")},inputs:{ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],disableRipple:[2,"disableRipple","disableRipple",ie]},features:[lt],ngContentSelectors:PG,decls:13,vars:10,consts:[["previousPaginator",""],["tabListContainer",""],["tabList",""],["tabListInner",""],["nextPaginator",""],["mat-ripple","",1,"mat-mdc-tab-header-pagination","mat-mdc-tab-header-pagination-before",3,"click","mousedown","touchend","matRippleDisabled"],[1,"mat-mdc-tab-header-pagination-chevron"],[1,"mat-mdc-tab-label-container",3,"keydown"],["role","tablist",1,"mat-mdc-tab-list",3,"cdkObserveContent"],[1,"mat-mdc-tab-labels"],["mat-ripple","",1,"mat-mdc-tab-header-pagination","mat-mdc-tab-header-pagination-after",3,"mousedown","click","touchend","matRippleDisabled"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"div",5,0),hA("click",function(){return RA(o),xA(n._handlePaginatorClick("before"))})("mousedown",function(s){return RA(o),xA(n._handlePaginatorPress("before",s))})("touchend",function(){return RA(o),xA(n._stopInterval())}),JA(2,"div",6),F(),S(3,"div",7,1),hA("keydown",function(s){return RA(o),xA(n._handleKeydown(s))}),S(5,"div",8,2),hA("cdkObserveContent",function(){return RA(o),xA(n._onContentChanges())}),S(7,"div",9,3),Fe(9),F()()(),S(10,"div",10,4),hA("mousedown",function(s){return RA(o),xA(n._handlePaginatorPress("after",s))})("click",function(){return RA(o),xA(n._handlePaginatorClick("after"))})("touchend",function(){return RA(o),xA(n._stopInterval())}),JA(12,"div",6),F()}i&2&&(ue("mat-mdc-tab-header-pagination-disabled",n._disableScrollBefore),yA("matRippleDisabled",n._disableScrollBefore||n.disableRipple),G(3),ue("_mat-animation-noopable",n._animationMode==="NoopAnimations"),G(2),Ne("aria-label",n.ariaLabel||null)("aria-labelledby",n.ariaLabelledby||null),G(5),ue("mat-mdc-tab-header-pagination-disabled",n._disableScrollAfter),yA("matRippleDisabled",n._disableScrollAfter||n.disableRipple))},dependencies:[rs,V6],styles:[".mat-mdc-tab-header{display:flex;overflow:hidden;position:relative;flex-shrink:0}.mdc-tab-indicator .mdc-tab-indicator__content{transition-duration:var(--mat-tab-animation-duration, 250ms)}.mat-mdc-tab-header-pagination{-webkit-user-select:none;user-select:none;position:relative;display:none;justify-content:center;align-items:center;min-width:32px;cursor:pointer;z-index:2;-webkit-tap-highlight-color:rgba(0,0,0,0);touch-action:none;box-sizing:content-box;outline:0}.mat-mdc-tab-header-pagination::-moz-focus-inner{border:0}.mat-mdc-tab-header-pagination .mat-ripple-element{opacity:.12;background-color:var(--mat-tab-header-inactive-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-tab-header-pagination-controls-enabled .mat-mdc-tab-header-pagination{display:flex}.mat-mdc-tab-header-pagination-before,.mat-mdc-tab-header-rtl .mat-mdc-tab-header-pagination-after{padding-left:4px}.mat-mdc-tab-header-pagination-before .mat-mdc-tab-header-pagination-chevron,.mat-mdc-tab-header-rtl .mat-mdc-tab-header-pagination-after .mat-mdc-tab-header-pagination-chevron{transform:rotate(-135deg)}.mat-mdc-tab-header-rtl .mat-mdc-tab-header-pagination-before,.mat-mdc-tab-header-pagination-after{padding-right:4px}.mat-mdc-tab-header-rtl .mat-mdc-tab-header-pagination-before .mat-mdc-tab-header-pagination-chevron,.mat-mdc-tab-header-pagination-after .mat-mdc-tab-header-pagination-chevron{transform:rotate(45deg)}.mat-mdc-tab-header-pagination-chevron{border-style:solid;border-width:2px 2px 0 0;height:8px;width:8px;border-color:var(--mat-tab-header-pagination-icon-color, var(--mat-sys-on-surface))}.mat-mdc-tab-header-pagination-disabled{box-shadow:none;cursor:default;pointer-events:none}.mat-mdc-tab-header-pagination-disabled .mat-mdc-tab-header-pagination-chevron{opacity:.4}.mat-mdc-tab-list{flex-grow:1;position:relative;transition:transform 500ms cubic-bezier(0.35, 0, 0.25, 1)}._mat-animation-noopable .mat-mdc-tab-list{transition:none}.mat-mdc-tab-label-container{display:flex;flex-grow:1;overflow:hidden;z-index:1;border-bottom-style:solid;border-bottom-width:var(--mat-tab-header-divider-height, 1px);border-bottom-color:var(--mat-tab-header-divider-color, var(--mat-sys-surface-variant))}.mat-mdc-tab-group-inverted-header .mat-mdc-tab-label-container{border-bottom:none;border-top-style:solid;border-top-width:var(--mat-tab-header-divider-height, 1px);border-top-color:var(--mat-tab-header-divider-color, var(--mat-sys-surface-variant))}.mat-mdc-tab-labels{display:flex;flex:1 0 auto}[mat-align-tabs=center]>.mat-mdc-tab-header .mat-mdc-tab-labels{justify-content:center}[mat-align-tabs=end]>.mat-mdc-tab-header .mat-mdc-tab-labels{justify-content:flex-end}.cdk-drop-list .mat-mdc-tab-labels,.mat-mdc-tab-labels.cdk-drop-list{min-height:var(--mdc-secondary-navigation-tab-container-height, 48px)}.mat-mdc-tab::before{margin:5px}@media(forced-colors: active){.mat-mdc-tab[aria-disabled=true]{color:GrayText}}"],encapsulation:2})}return t})(),NOA=new dA("MAT_TABS_CONFIG"),_OA={translateTab:sc("translateTab",[js("center, void, left-origin-center, right-origin-center",po({transform:"none",visibility:"visible"})),js("left",po({transform:"translate3d(-100%, 0, 0)",minHeight:"1px",visibility:"hidden"})),js("right",po({transform:"translate3d(100%, 0, 0)",minHeight:"1px",visibility:"hidden"})),Vr("* => left, * => right, left => center, right => center",ss("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")),Vr("void => left-origin-center",[po({transform:"translate3d(-100%, 0, 0)",visibility:"hidden"}),ss("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")]),Vr("void => right-origin-center",[po({transform:"translate3d(100%, 0, 0)",visibility:"hidden"}),ss("{{animationDuration}} cubic-bezier(0.35, 0, 0.25, 1)")])])},GOA=(()=>{class t extends fa{_host=f(HcA);_centeringSub=Kt.EMPTY;_leavingSub=Kt.EMPTY;constructor(){super()}ngOnInit(){super.ngOnInit(),this._centeringSub=this._host._beforeCentering.pipe(Pn(this._host._isCenterPosition(this._host._position))).subscribe(A=>{this._host._content&&A&&!this.hasAttached()&&this.attach(this._host._content)}),this._leavingSub=this._host._afterLeavingCenter.subscribe(()=>{this._host.preserveContent||this.detach()})}ngOnDestroy(){super.ngOnDestroy(),this._centeringSub.unsubscribe(),this._leavingSub.unsubscribe()}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","matTabBodyHost",""]],features:[lt]})}return t})(),HcA=(()=>{class t{_elementRef=f(ee);_dir=f(mo,{optional:!0});_positionIndex;_dirChangeSubscription=Kt.EMPTY;_position;_translateTabComplete=new OA;_onCentering=new WA;_beforeCentering=new WA;_afterLeavingCenter=new WA;_onCentered=new WA(!0);_portalHost;_content;origin;animationDuration="500ms";preserveContent=!1;set position(A){this._positionIndex=A,this._computePositionAnimationState()}constructor(){if(this._dir){let A=f(It);this._dirChangeSubscription=this._dir.change.subscribe(i=>{this._computePositionAnimationState(i),A.markForCheck()})}this._translateTabComplete.subscribe(A=>{this._isCenterPosition(A.toState)&&this._isCenterPosition(this._position)&&this._onCentered.emit(),this._isCenterPosition(A.fromState)&&!this._isCenterPosition(this._position)&&this._afterLeavingCenter.emit()})}ngOnInit(){this._position=="center"&&this.origin!=null&&(this._position=this._computePositionFromOrigin(this.origin))}ngOnDestroy(){this._dirChangeSubscription.unsubscribe(),this._translateTabComplete.complete()}_onTranslateTabStarted(A){let i=this._isCenterPosition(A.toState);this._beforeCentering.emit(i),i&&this._onCentering.emit(this._elementRef.nativeElement.clientHeight)}_getLayoutDirection(){return this._dir&&this._dir.value==="rtl"?"rtl":"ltr"}_isCenterPosition(A){return A=="center"||A=="left-origin-center"||A=="right-origin-center"}_computePositionAnimationState(A=this._getLayoutDirection()){this._positionIndex<0?this._position=A=="ltr"?"left":"right":this._positionIndex>0?this._position=A=="ltr"?"right":"left":this._position="center"}_computePositionFromOrigin(A){let i=this._getLayoutDirection();return i=="ltr"&&A<=0||i=="rtl"&&A>0?"left-origin-center":"right-origin-center"}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-tab-body"]],viewQuery:function(i,n){if(i&1&&Te(fa,5),i&2){let o;XA(o=$A())&&(n._portalHost=o.first)}},hostAttrs:[1,"mat-mdc-tab-body"],inputs:{_content:[0,"content","_content"],origin:"origin",animationDuration:"animationDuration",preserveContent:"preserveContent",position:"position"},outputs:{_onCentering:"_onCentering",_beforeCentering:"_beforeCentering",_afterLeavingCenter:"_afterLeavingCenter",_onCentered:"_onCentered"},decls:3,vars:6,consts:[["content",""],["cdkScrollable","",1,"mat-mdc-tab-body-content"],["matTabBodyHost",""]],template:function(i,n){if(i&1){let o=be();S(0,"div",1,0),hA("@translateTab.start",function(s){return RA(o),xA(n._onTranslateTabStarted(s))})("@translateTab.done",function(s){return RA(o),xA(n._translateTabComplete.next(s))}),_A(2,hOA,0,0,"ng-template",2),F()}i&2&&yA("@translateTab",b2(3,QOA,n._position,qr(1,EOA,n.animationDuration)))},dependencies:[GOA,D0],styles:['.mat-mdc-tab-body{top:0;left:0;right:0;bottom:0;position:absolute;display:block;overflow:hidden;outline:0;flex-basis:100%}.mat-mdc-tab-body.mat-mdc-tab-body-active{position:relative;overflow-x:hidden;overflow-y:auto;z-index:1;flex-grow:1}.mat-mdc-tab-group.mat-mdc-tab-group-dynamic-height .mat-mdc-tab-body.mat-mdc-tab-body-active{overflow-y:hidden}.mat-mdc-tab-body-content{height:100%;overflow:auto}.mat-mdc-tab-group-dynamic-height .mat-mdc-tab-body-content{overflow:hidden}.mat-mdc-tab-body-content[style*="visibility: hidden"]{display:none}'],encapsulation:2,data:{animation:[_OA.translateTab]}})}return t})(),UOA=!0,o7=(()=>{class t{_elementRef=f(ee);_changeDetectorRef=f(It);_animationMode=f(Si,{optional:!0});_allTabs;_tabBodyWrapper;_tabHeader;_tabs=new ca;_indexToSelect=0;_lastFocusedTabIndex=null;_tabBodyWrapperHeight=0;_tabsSubscription=Kt.EMPTY;_tabLabelSubscription=Kt.EMPTY;color;get fitInkBarToContent(){return this._fitInkBarToContent}set fitInkBarToContent(A){this._fitInkBarToContent=A,this._changeDetectorRef.markForCheck()}_fitInkBarToContent=!1;stretchTabs=!0;alignTabs=null;dynamicHeight=!1;get selectedIndex(){return this._selectedIndex}set selectedIndex(A){this._indexToSelect=isNaN(A)?null:A}_selectedIndex=null;headerPosition="above";get animationDuration(){return this._animationDuration}set animationDuration(A){let i=A+"";this._animationDuration=/^\d+$/.test(i)?A+"ms":i}_animationDuration;get contentTabIndex(){return this._contentTabIndex}set contentTabIndex(A){this._contentTabIndex=isNaN(A)?null:A}_contentTabIndex;disablePagination=!1;disableRipple=!1;preserveContent=!1;get backgroundColor(){return this._backgroundColor}set backgroundColor(A){if(!UOA)throw new Error("mat-tab-group background color must be set through the Sass theming API");let i=this._elementRef.nativeElement.classList;i.remove("mat-tabs-with-background",`mat-background-${this.backgroundColor}`),A&&i.add("mat-tabs-with-background",`mat-background-${A}`),this._backgroundColor=A}_backgroundColor;ariaLabel;ariaLabelledby;selectedIndexChange=new WA;focusChange=new WA;animationDone=new WA;selectedTabChange=new WA(!0);_groupId;_isServer=!f(Ii).isBrowser;constructor(){let A=f(NOA,{optional:!0});this._groupId=f(sn).getId("mat-tab-group-"),this.animationDuration=A&&A.animationDuration?A.animationDuration:"500ms",this.disablePagination=A&&A.disablePagination!=null?A.disablePagination:!1,this.dynamicHeight=A&&A.dynamicHeight!=null?A.dynamicHeight:!1,A?.contentTabIndex!=null&&(this.contentTabIndex=A.contentTabIndex),this.preserveContent=!!A?.preserveContent,this.fitInkBarToContent=A&&A.fitInkBarToContent!=null?A.fitInkBarToContent:!1,this.stretchTabs=A&&A.stretchTabs!=null?A.stretchTabs:!0,this.alignTabs=A&&A.alignTabs!=null?A.alignTabs:null}ngAfterContentChecked(){let A=this._indexToSelect=this._clampTabIndex(this._indexToSelect);if(this._selectedIndex!=A){let i=this._selectedIndex==null;if(!i){this.selectedTabChange.emit(this._createChangeEvent(A));let n=this._tabBodyWrapper.nativeElement;n.style.minHeight=n.clientHeight+"px"}Promise.resolve().then(()=>{this._tabs.forEach((n,o)=>n.isActive=o===A),i||(this.selectedIndexChange.emit(A),this._tabBodyWrapper.nativeElement.style.minHeight="")})}this._tabs.forEach((i,n)=>{i.position=n-A,this._selectedIndex!=null&&i.position==0&&!i.origin&&(i.origin=A-this._selectedIndex)}),this._selectedIndex!==A&&(this._selectedIndex=A,this._lastFocusedTabIndex=null,this._changeDetectorRef.markForCheck())}ngAfterContentInit(){this._subscribeToAllTabChanges(),this._subscribeToTabLabels(),this._tabsSubscription=this._tabs.changes.subscribe(()=>{let A=this._clampTabIndex(this._indexToSelect);if(A===this._selectedIndex){let i=this._tabs.toArray(),n;for(let o=0;o<i.length;o++)if(i[o].isActive){this._indexToSelect=this._selectedIndex=o,this._lastFocusedTabIndex=null,n=i[o];break}!n&&i[A]&&Promise.resolve().then(()=>{i[A].isActive=!0,this.selectedTabChange.emit(this._createChangeEvent(A))})}this._changeDetectorRef.markForCheck()})}_subscribeToAllTabChanges(){this._allTabs.changes.pipe(Pn(this._allTabs)).subscribe(A=>{this._tabs.reset(A.filter(i=>i._closestTabGroup===this||!i._closestTabGroup)),this._tabs.notifyOnChanges()})}ngOnDestroy(){this._tabs.destroy(),this._tabsSubscription.unsubscribe(),this._tabLabelSubscription.unsubscribe()}realignInkBar(){this._tabHeader&&this._tabHeader._alignInkBarToSelectedTab()}updatePagination(){this._tabHeader&&this._tabHeader.updatePagination()}focusTab(A){let i=this._tabHeader;i&&(i.focusIndex=A)}_focusChanged(A){this._lastFocusedTabIndex=A,this.focusChange.emit(this._createChangeEvent(A))}_createChangeEvent(A){let i=new OG;return i.index=A,this._tabs&&this._tabs.length&&(i.tab=this._tabs.toArray()[A]),i}_subscribeToTabLabels(){this._tabLabelSubscription&&this._tabLabelSubscription.unsubscribe(),this._tabLabelSubscription=zn(...this._tabs.map(A=>A._stateChanges)).subscribe(()=>this._changeDetectorRef.markForCheck())}_clampTabIndex(A){return Math.min(this._tabs.length-1,Math.max(A||0,0))}_getTabLabelId(A){return`${this._groupId}-label-${A}`}_getTabContentId(A){return`${this._groupId}-content-${A}`}_setTabBodyWrapperHeight(A){if(!this.dynamicHeight||!this._tabBodyWrapperHeight)return;let i=this._tabBodyWrapper.nativeElement;i.style.height=this._tabBodyWrapperHeight+"px",this._tabBodyWrapper.nativeElement.offsetHeight&&(i.style.height=A+"px")}_removeTabBodyWrapperHeight(){let A=this._tabBodyWrapper.nativeElement;this._tabBodyWrapperHeight=A.clientHeight,A.style.height="",this.animationDone.emit()}_handleClick(A,i,n){i.focusIndex=n,A.disabled||(this.selectedIndex=n)}_getTabIndex(A){let i=this._lastFocusedTabIndex??this.selectedIndex;return A===i?0:-1}_tabFocusChanged(A,i){A&&A!=="mouse"&&A!=="touch"&&(this._tabHeader.focusIndex=i)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-tab-group"]],contentQueries:function(i,n,o){if(i&1&&ci(o,mf,5),i&2){let r;XA(r=$A())&&(n._allTabs=r)}},viewQuery:function(i,n){if(i&1&&(Te(uOA,5),Te(fOA,5)),i&2){let o;XA(o=$A())&&(n._tabBodyWrapper=o.first),XA(o=$A())&&(n._tabHeader=o.first)}},hostAttrs:[1,"mat-mdc-tab-group"],hostVars:11,hostBindings:function(i,n){i&2&&(Ne("mat-align-tabs",n.alignTabs),fo("mat-"+(n.color||"primary")),uo("--mat-tab-animation-duration",n.animationDuration),ue("mat-mdc-tab-group-dynamic-height",n.dynamicHeight)("mat-mdc-tab-group-inverted-header",n.headerPosition==="below")("mat-mdc-tab-group-stretch-tabs",n.stretchTabs))},inputs:{color:"color",fitInkBarToContent:[2,"fitInkBarToContent","fitInkBarToContent",ie],stretchTabs:[2,"mat-stretch-tabs","stretchTabs",ie],alignTabs:[0,"mat-align-tabs","alignTabs"],dynamicHeight:[2,"dynamicHeight","dynamicHeight",ie],selectedIndex:[2,"selectedIndex","selectedIndex",zi],headerPosition:"headerPosition",animationDuration:"animationDuration",contentTabIndex:[2,"contentTabIndex","contentTabIndex",zi],disablePagination:[2,"disablePagination","disablePagination",ie],disableRipple:[2,"disableRipple","disableRipple",ie],preserveContent:[2,"preserveContent","preserveContent",ie],backgroundColor:"backgroundColor",ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"]},outputs:{selectedIndexChange:"selectedIndexChange",focusChange:"focusChange",animationDone:"animationDone",selectedTabChange:"selectedTabChange"},exportAs:["matTabGroup"],features:[ht([{provide:JcA,useExisting:t}])],ngContentSelectors:PG,decls:9,vars:8,consts:[["tabHeader",""],["tabBodyWrapper",""],["tabNode",""],[3,"indexFocused","selectFocusedIndex","selectedIndex","disableRipple","disablePagination","aria-label","aria-labelledby"],["role","tab","matTabLabelWrapper","","cdkMonitorElementFocus","",1,"mdc-tab","mat-mdc-tab","mat-focus-indicator",3,"id","mdc-tab--active","class","disabled","fitInkBarToContent"],[1,"mat-mdc-tab-body-wrapper"],["role","tabpanel",3,"id","mat-mdc-tab-body-active","class","content","position","origin","animationDuration","preserveContent"],["role","tab","matTabLabelWrapper","","cdkMonitorElementFocus","",1,"mdc-tab","mat-mdc-tab","mat-focus-indicator",3,"click","cdkFocusChange","id","disabled","fitInkBarToContent"],[1,"mdc-tab__ripple"],["mat-ripple","",1,"mat-mdc-tab-ripple",3,"matRippleTrigger","matRippleDisabled"],[1,"mdc-tab__content"],[1,"mdc-tab__text-label"],[3,"cdkPortalOutlet"],["role","tabpanel",3,"_onCentered","_onCentering","id","content","position","origin","animationDuration","preserveContent"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"mat-tab-header",3,0),hA("indexFocused",function(s){return RA(o),xA(n._focusChanged(s))})("selectFocusedIndex",function(s){return RA(o),xA(n.selectedIndex=s)}),Dn(2,DOA,8,17,"div",4,to),F(),_A(4,yOA,1,0),S(5,"div",5,1),Dn(7,vOA,1,13,"mat-tab-body",6,to),F()}i&2&&(yA("selectedIndex",n.selectedIndex||0)("disableRipple",n.disableRipple)("disablePagination",n.disablePagination)("aria-label",n.ariaLabel)("aria-labelledby",n.ariaLabelledby),G(2),yn(n._tabs),G(2),GA(n._isServer?4:-1),G(),ue("_mat-animation-noopable",n._animationMode==="NoopAnimations"),G(2),yn(n._tabs))},dependencies:[FOA,TcA,PO,rs,fa,HcA],styles:['.mdc-tab{min-width:90px;padding:0 24px;display:flex;flex:1 0 auto;justify-content:center;box-sizing:border-box;border:none;outline:none;text-align:center;white-space:nowrap;cursor:pointer;z-index:1}.mdc-tab__content{display:flex;align-items:center;justify-content:center;height:inherit;pointer-events:none}.mdc-tab__text-label{transition:150ms color linear;display:inline-block;line-height:1;z-index:2}.mdc-tab--active .mdc-tab__text-label{transition-delay:100ms}._mat-animation-noopable .mdc-tab__text-label{transition:none}.mdc-tab-indicator{display:flex;position:absolute;top:0;left:0;justify-content:center;width:100%;height:100%;pointer-events:none;z-index:1}.mdc-tab-indicator__content{transition:var(--mat-tab-animation-duration, 250ms) transform cubic-bezier(0.4, 0, 0.2, 1);transform-origin:left;opacity:0}.mdc-tab-indicator__content--underline{align-self:flex-end;box-sizing:border-box;width:100%;border-top-style:solid}.mdc-tab-indicator--active .mdc-tab-indicator__content{opacity:1}._mat-animation-noopable .mdc-tab-indicator__content,.mdc-tab-indicator--no-transition .mdc-tab-indicator__content{transition:none}.mat-mdc-tab-ripple.mat-mdc-tab-ripple{position:absolute;top:0;left:0;bottom:0;right:0;pointer-events:none}.mat-mdc-tab{-webkit-tap-highlight-color:rgba(0,0,0,0);-webkit-font-smoothing:antialiased;-moz-osx-font-smoothing:grayscale;text-decoration:none;background:none;height:var(--mdc-secondary-navigation-tab-container-height, 48px);font-family:var(--mat-tab-header-label-text-font, var(--mat-sys-title-small-font));font-size:var(--mat-tab-header-label-text-size, var(--mat-sys-title-small-size));letter-spacing:var(--mat-tab-header-label-text-tracking, var(--mat-sys-title-small-tracking));line-height:var(--mat-tab-header-label-text-line-height, var(--mat-sys-title-small-line-height));font-weight:var(--mat-tab-header-label-text-weight, var(--mat-sys-title-small-weight))}.mat-mdc-tab.mdc-tab{flex-grow:0}.mat-mdc-tab .mdc-tab-indicator__content--underline{border-color:var(--mdc-tab-indicator-active-indicator-color, var(--mat-sys-primary));border-top-width:var(--mdc-tab-indicator-active-indicator-height, 2px);border-radius:var(--mdc-tab-indicator-active-indicator-shape, 0)}.mat-mdc-tab:hover .mdc-tab__text-label{color:var(--mat-tab-header-inactive-hover-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-tab:focus .mdc-tab__text-label{color:var(--mat-tab-header-inactive-focus-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-tab.mdc-tab--active .mdc-tab__text-label{color:var(--mat-tab-header-active-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-tab.mdc-tab--active .mdc-tab__ripple::before,.mat-mdc-tab.mdc-tab--active .mat-ripple-element{background-color:var(--mat-tab-header-active-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-tab.mdc-tab--active:hover .mdc-tab__text-label{color:var(--mat-tab-header-active-hover-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-tab.mdc-tab--active:hover .mdc-tab-indicator__content--underline{border-color:var(--mat-tab-header-active-hover-indicator-color, var(--mat-sys-primary))}.mat-mdc-tab.mdc-tab--active:focus .mdc-tab__text-label{color:var(--mat-tab-header-active-focus-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-tab.mdc-tab--active:focus .mdc-tab-indicator__content--underline{border-color:var(--mat-tab-header-active-focus-indicator-color, var(--mat-sys-primary))}.mat-mdc-tab.mat-mdc-tab-disabled{opacity:.4;pointer-events:none}.mat-mdc-tab.mat-mdc-tab-disabled .mdc-tab__content{pointer-events:none}.mat-mdc-tab.mat-mdc-tab-disabled .mdc-tab__ripple::before,.mat-mdc-tab.mat-mdc-tab-disabled .mat-ripple-element{background-color:var(--mat-tab-header-disabled-ripple-color)}.mat-mdc-tab .mdc-tab__ripple::before{content:"";display:block;position:absolute;top:0;left:0;right:0;bottom:0;opacity:0;pointer-events:none;background-color:var(--mat-tab-header-inactive-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-tab .mdc-tab__text-label{color:var(--mat-tab-header-inactive-label-text-color, var(--mat-sys-on-surface));display:inline-flex;align-items:center}.mat-mdc-tab .mdc-tab__content{position:relative;pointer-events:auto}.mat-mdc-tab:hover .mdc-tab__ripple::before{opacity:.04}.mat-mdc-tab.cdk-program-focused .mdc-tab__ripple::before,.mat-mdc-tab.cdk-keyboard-focused .mdc-tab__ripple::before{opacity:.12}.mat-mdc-tab .mat-ripple-element{opacity:.12;background-color:var(--mat-tab-header-inactive-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-tab-group.mat-mdc-tab-group-stretch-tabs>.mat-mdc-tab-header .mat-mdc-tab{flex-grow:1}.mat-mdc-tab-group{display:flex;flex-direction:column;max-width:100%}.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination{background-color:var(--mat-tab-header-with-background-background-color)}.mat-mdc-tab-group.mat-tabs-with-background.mat-primary>.mat-mdc-tab-header .mat-mdc-tab .mdc-tab__text-label{color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background.mat-primary>.mat-mdc-tab-header .mdc-tab-indicator__content--underline{border-color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background:not(.mat-primary)>.mat-mdc-tab-header .mat-mdc-tab:not(.mdc-tab--active) .mdc-tab__text-label{color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background:not(.mat-primary)>.mat-mdc-tab-header .mat-mdc-tab:not(.mdc-tab--active) .mdc-tab-indicator__content--underline{border-color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header .mat-mdc-tab-header-pagination-chevron,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header .mat-focus-indicator::before,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination .mat-mdc-tab-header-pagination-chevron,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination .mat-focus-indicator::before{border-color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header .mat-ripple-element,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header .mdc-tab__ripple::before,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination .mat-ripple-element,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination .mdc-tab__ripple::before{background-color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header .mat-mdc-tab-header-pagination-chevron,.mat-mdc-tab-group.mat-tabs-with-background>.mat-mdc-tab-header-pagination .mat-mdc-tab-header-pagination-chevron{color:var(--mat-tab-header-with-background-foreground-color)}.mat-mdc-tab-group.mat-mdc-tab-group-inverted-header{flex-direction:column-reverse}.mat-mdc-tab-group.mat-mdc-tab-group-inverted-header .mdc-tab-indicator__content--underline{align-self:flex-start}.mat-mdc-tab-body-wrapper{position:relative;overflow:hidden;display:flex;transition:height 500ms cubic-bezier(0.35, 0, 0.25, 1)}.mat-mdc-tab-body-wrapper._mat-animation-noopable{transition:none !important;animation:none !important}'],encapsulation:2})}return t})(),OG=class{index;tab};var zcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,it]})}return t})();function KOA(t,e){t&1&&JA(0,"div",2)}var YOA=new dA("MAT_PROGRESS_BAR_DEFAULT_OPTIONS");var jcA=(()=>{class t{_elementRef=f(ee);_ngZone=f(Qe);_changeDetectorRef=f(It);_renderer=f(Wi);_cleanupTransitionEnd;_animationMode=f(Si,{optional:!0});constructor(){let A=f(YOA,{optional:!0});this._isNoopAnimation=this._animationMode==="NoopAnimations",A&&(A.color&&(this.color=this._defaultColor=A.color),this.mode=A.mode||this.mode)}_isNoopAnimation=!1;get color(){return this._color||this._defaultColor}set color(A){this._color=A}_color;_defaultColor="primary";get value(){return this._value}set value(A){this._value=PcA(A||0),this._changeDetectorRef.markForCheck()}_value=0;get bufferValue(){return this._bufferValue||0}set bufferValue(A){this._bufferValue=PcA(A||0),this._changeDetectorRef.markForCheck()}_bufferValue=0;animationEnd=new WA;get mode(){return this._mode}set mode(A){this._mode=A,this._changeDetectorRef.markForCheck()}_mode="determinate";ngAfterViewInit(){this._ngZone.runOutsideAngular(()=>{this._cleanupTransitionEnd=this._renderer.listen(this._elementRef.nativeElement,"transitionend",this._transitionendHandler)})}ngOnDestroy(){this._cleanupTransitionEnd?.()}_getPrimaryBarTransform(){return`scaleX(${this._isIndeterminate()?1:this.value/100})`}_getBufferBarFlexBasis(){return`${this.mode==="buffer"?this.bufferValue:100}%`}_isIndeterminate(){return this.mode==="indeterminate"||this.mode==="query"}_transitionendHandler=A=>{this.animationEnd.observers.length===0||!A.target||!A.target.classList.contains("mdc-linear-progress__primary-bar")||(this.mode==="determinate"||this.mode==="buffer")&&this._ngZone.run(()=>this.animationEnd.next({value:this.value}))};static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-progress-bar"]],hostAttrs:["role","progressbar","aria-valuemin","0","aria-valuemax","100","tabindex","-1",1,"mat-mdc-progress-bar","mdc-linear-progress"],hostVars:10,hostBindings:function(i,n){i&2&&(Ne("aria-valuenow",n._isIndeterminate()?null:n.value)("mode",n.mode),fo("mat-"+n.color),ue("_mat-animation-noopable",n._isNoopAnimation)("mdc-linear-progress--animation-ready",!n._isNoopAnimation)("mdc-linear-progress--indeterminate",n._isIndeterminate()))},inputs:{color:"color",value:[2,"value","value",zi],bufferValue:[2,"bufferValue","bufferValue",zi],mode:"mode"},outputs:{animationEnd:"animationEnd"},exportAs:["matProgressBar"],decls:7,vars:5,consts:[["aria-hidden","true",1,"mdc-linear-progress__buffer"],[1,"mdc-linear-progress__buffer-bar"],[1,"mdc-linear-progress__buffer-dots"],["aria-hidden","true",1,"mdc-linear-progress__bar","mdc-linear-progress__primary-bar"],[1,"mdc-linear-progress__bar-inner"],["aria-hidden","true",1,"mdc-linear-progress__bar","mdc-linear-progress__secondary-bar"]],template:function(i,n){i&1&&(S(0,"div",0),JA(1,"div",1),_A(2,KOA,1,0,"div",2),F(),S(3,"div",3),JA(4,"span",4),F(),S(5,"div",5),JA(6,"span",4),F()),i&2&&(G(),uo("flex-basis",n._getBufferBarFlexBasis()),G(),GA(n.mode==="buffer"?2:-1),G(),uo("transform",n._getPrimaryBarTransform()))},styles:[`.mat-mdc-progress-bar{display:block;text-align:start}.mat-mdc-progress-bar[mode=query]{transform:scaleX(-1)}.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__buffer-dots,.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__primary-bar,.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__secondary-bar,.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__bar-inner.mdc-linear-progress__bar-inner{animation:none}.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__primary-bar,.mat-mdc-progress-bar._mat-animation-noopable .mdc-linear-progress__buffer-bar{transition:transform 1ms}.mdc-linear-progress{position:relative;width:100%;transform:translateZ(0);outline:1px solid rgba(0,0,0,0);overflow-x:hidden;transition:opacity 250ms 0ms cubic-bezier(0.4, 0, 0.6, 1);height:max(var(--mdc-linear-progress-track-height, 4px),var(--mdc-linear-progress-active-indicator-height, 4px))}@media(forced-colors: active){.mdc-linear-progress{outline-color:CanvasText}}.mdc-linear-progress__bar{position:absolute;top:0;bottom:0;margin:auto 0;width:100%;animation:none;transform-origin:top left;transition:transform 250ms 0ms cubic-bezier(0.4, 0, 0.6, 1);height:var(--mdc-linear-progress-active-indicator-height, 4px)}.mdc-linear-progress--indeterminate .mdc-linear-progress__bar{transition:none}[dir=rtl] .mdc-linear-progress__bar{right:0;transform-origin:center right}.mdc-linear-progress__bar-inner{display:inline-block;position:absolute;width:100%;animation:none;border-top-style:solid;border-color:var(--mdc-linear-progress-active-indicator-color, var(--mat-sys-primary));border-top-width:var(--mdc-linear-progress-active-indicator-height, 4px)}.mdc-linear-progress__buffer{display:flex;position:absolute;top:0;bottom:0;margin:auto 0;width:100%;overflow:hidden;height:var(--mdc-linear-progress-track-height, 4px);border-radius:var(--mdc-linear-progress-track-shape, var(--mat-sys-corner-none))}.mdc-linear-progress__buffer-dots{-webkit-mask-image:url("data:image/svg+xml,%3Csvg version='1.1' xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink' x='0px' y='0px' enable-background='new 0 0 5 2' xml:space='preserve' viewBox='0 0 5 2' preserveAspectRatio='xMinYMin slice'%3E%3Ccircle cx='1' cy='1' r='1'/%3E%3C/svg%3E");mask-image:url("data:image/svg+xml,%3Csvg version='1.1' xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink' x='0px' y='0px' enable-background='new 0 0 5 2' xml:space='preserve' viewBox='0 0 5 2' preserveAspectRatio='xMinYMin slice'%3E%3Ccircle cx='1' cy='1' r='1'/%3E%3C/svg%3E");background-repeat:repeat-x;flex:auto;transform:rotate(180deg);animation:mdc-linear-progress-buffering 250ms infinite linear;background-color:var(--mdc-linear-progress-track-color, var(--mat-sys-surface-variant))}@media(forced-colors: active){.mdc-linear-progress__buffer-dots{background-color:ButtonBorder}}[dir=rtl] .mdc-linear-progress__buffer-dots{animation:mdc-linear-progress-buffering-reverse 250ms infinite linear;transform:rotate(0)}.mdc-linear-progress__buffer-bar{flex:0 1 100%;transition:flex-basis 250ms 0ms cubic-bezier(0.4, 0, 0.6, 1);background-color:var(--mdc-linear-progress-track-color, var(--mat-sys-surface-variant))}.mdc-linear-progress__primary-bar{transform:scaleX(0)}.mdc-linear-progress--indeterminate .mdc-linear-progress__primary-bar{left:-145.166611%}.mdc-linear-progress--indeterminate.mdc-linear-progress--animation-ready .mdc-linear-progress__primary-bar{animation:mdc-linear-progress-primary-indeterminate-translate 2s infinite linear}.mdc-linear-progress--indeterminate.mdc-linear-progress--animation-ready .mdc-linear-progress__primary-bar>.mdc-linear-progress__bar-inner{animation:mdc-linear-progress-primary-indeterminate-scale 2s infinite linear}[dir=rtl] .mdc-linear-progress.mdc-linear-progress--animation-ready .mdc-linear-progress__primary-bar{animation-name:mdc-linear-progress-primary-indeterminate-translate-reverse}[dir=rtl] .mdc-linear-progress.mdc-linear-progress--indeterminate .mdc-linear-progress__primary-bar{right:-145.166611%;left:auto}.mdc-linear-progress__secondary-bar{display:none}.mdc-linear-progress--indeterminate .mdc-linear-progress__secondary-bar{left:-54.888891%;display:block}.mdc-linear-progress--indeterminate.mdc-linear-progress--animation-ready .mdc-linear-progress__secondary-bar{animation:mdc-linear-progress-secondary-indeterminate-translate 2s infinite linear}.mdc-linear-progress--indeterminate.mdc-linear-progress--animation-ready .mdc-linear-progress__secondary-bar>.mdc-linear-progress__bar-inner{animation:mdc-linear-progress-secondary-indeterminate-scale 2s infinite linear}[dir=rtl] .mdc-linear-progress.mdc-linear-progress--animation-ready .mdc-linear-progress__secondary-bar{animation-name:mdc-linear-progress-secondary-indeterminate-translate-reverse}[dir=rtl] .mdc-linear-progress.mdc-linear-progress--indeterminate .mdc-linear-progress__secondary-bar{right:-54.888891%;left:auto}@keyframes mdc-linear-progress-buffering{from{transform:rotate(180deg) translateX(calc(var(--mdc-linear-progress-track-height, 4px) * -2.5))}}@keyframes mdc-linear-progress-primary-indeterminate-translate{0%{transform:translateX(0)}20%{animation-timing-function:cubic-bezier(0.5, 0, 0.701732, 0.495819);transform:translateX(0)}59.15%{animation-timing-function:cubic-bezier(0.302435, 0.381352, 0.55, 0.956352);transform:translateX(83.67142%)}100%{transform:translateX(200.611057%)}}@keyframes mdc-linear-progress-primary-indeterminate-scale{0%{transform:scaleX(0.08)}36.65%{animation-timing-function:cubic-bezier(0.334731, 0.12482, 0.785844, 1);transform:scaleX(0.08)}69.15%{animation-timing-function:cubic-bezier(0.06, 0.11, 0.6, 1);transform:scaleX(0.661479)}100%{transform:scaleX(0.08)}}@keyframes mdc-linear-progress-secondary-indeterminate-translate{0%{animation-timing-function:cubic-bezier(0.15, 0, 0.515058, 0.409685);transform:translateX(0)}25%{animation-timing-function:cubic-bezier(0.31033, 0.284058, 0.8, 0.733712);transform:translateX(37.651913%)}48.35%{animation-timing-function:cubic-bezier(0.4, 0.627035, 0.6, 0.902026);transform:translateX(84.386165%)}100%{transform:translateX(160.277782%)}}@keyframes mdc-linear-progress-secondary-indeterminate-scale{0%{animation-timing-function:cubic-bezier(0.205028, 0.057051, 0.57661, 0.453971);transform:scaleX(0.08)}19.15%{animation-timing-function:cubic-bezier(0.152313, 0.196432, 0.648374, 1.004315);transform:scaleX(0.457104)}44.15%{animation-timing-function:cubic-bezier(0.257759, -0.003163, 0.211762, 1.38179);transform:scaleX(0.72796)}100%{transform:scaleX(0.08)}}@keyframes mdc-linear-progress-primary-indeterminate-translate-reverse{0%{transform:translateX(0)}20%{animation-timing-function:cubic-bezier(0.5, 0, 0.701732, 0.495819);transform:translateX(0)}59.15%{animation-timing-function:cubic-bezier(0.302435, 0.381352, 0.55, 0.956352);transform:translateX(-83.67142%)}100%{transform:translateX(-200.611057%)}}@keyframes mdc-linear-progress-secondary-indeterminate-translate-reverse{0%{animation-timing-function:cubic-bezier(0.15, 0, 0.515058, 0.409685);transform:translateX(0)}25%{animation-timing-function:cubic-bezier(0.31033, 0.284058, 0.8, 0.733712);transform:translateX(-37.651913%)}48.35%{animation-timing-function:cubic-bezier(0.4, 0.627035, 0.6, 0.902026);transform:translateX(-84.386165%)}100%{transform:translateX(-160.277782%)}}@keyframes mdc-linear-progress-buffering-reverse{from{transform:translateX(-10px)}}`],encapsulation:2,changeDetection:0})}return t})();function PcA(t,e=0,A=100){return Math.max(e,Math.min(A,t))}var qcA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it]})}return t})();function WG(){return{async:!1,breaks:!1,extensions:null,gfm:!0,hooks:null,pedantic:!1,renderer:null,silent:!1,tokenizer:null,walkTokens:null}}var sd=WG();function AlA(t){sd=t}var Df={exec:()=>null};function Zn(t,e=""){let A=typeof t=="string"?t:t.source,i={replace:(n,o)=>{let r=typeof o=="string"?o:o.source;return r=r.replace(ea.caret,"$1"),A=A.replace(n,r),i},getRegex:()=>new RegExp(A,e)};return i}var ea={codeRemoveIndent:/^(?: {1,4}| {0,3}\t)/gm,outputLinkReplace:/\\([\[\]])/g,indentCodeCompensation:/^(\s+)(?:```)/,beginningSpace:/^\s+/,endingHash:/#$/,startingSpaceChar:/^ /,endingSpaceChar:/ $/,nonSpaceChar:/[^ ]/,newLineCharGlobal:/\n/g,tabCharGlobal:/\t/g,multipleSpaceGlobal:/\s+/g,blankLine:/^[ \t]*$/,doubleBlankLine:/\n[ \t]*\n[ \t]*$/,blockquoteStart:/^ {0,3}>/,blockquoteSetextReplace:/\n {0,3}((?:=+|-+) *)(?=\n|$)/g,blockquoteSetextReplace2:/^ {0,3}>[ \t]?/gm,listReplaceTabs:/^\t+/,listReplaceNesting:/^ {1,4}(?=( {4})*[^ ])/g,listIsTask:/^\[[ xX]\] /,listReplaceTask:/^\[[ xX]\] +/,anyLine:/\n.*\n/,hrefBrackets:/^<(.*)>$/,tableDelimiter:/[:|]/,tableAlignChars:/^\||\| *$/g,tableRowBlankLine:/\n[ \t]*$/,tableAlignRight:/^ *-+: *$/,tableAlignCenter:/^ *:-+: *$/,tableAlignLeft:/^ *:-+ *$/,startATag:/^<a /i,endATag:/^<\/a>/i,startPreScriptTag:/^<(pre|code|kbd|script)(\s|>)/i,endPreScriptTag:/^<\/(pre|code|kbd|script)(\s|>)/i,startAngleBracket:/^</,endAngleBracket:/>$/,pedanticHrefTitle:/^([^'"]*[^\s])\s+(['"])(.*)\2/,unicodeAlphaNumeric:/[\p{L}\p{N}]/u,escapeTest:/[&<>"']/,escapeReplace:/[&<>"']/g,escapeTestNoEncode:/[<>"']|&(?!(#\d{1,7}|#[Xx][a-fA-F0-9]{1,6}|\w+);)/,escapeReplaceNoEncode:/[<>"']|&(?!(#\d{1,7}|#[Xx][a-fA-F0-9]{1,6}|\w+);)/g,unescapeTest:/&(#(?:\d+)|(?:#x[0-9A-Fa-f]+)|(?:\w+));?/ig,caret:/(^|[^\[])\^/g,percentDecode:/%25/g,findPipe:/\|/g,splitPipe:/ \|/,slashPipe:/\\\|/g,carriageReturn:/\r\n|\r/g,spaceLine:/^ +$/gm,notSpaceStart:/^\S*/,endingNewline:/\n$/,listItemRegex:t=>new RegExp(`^( {0,3}${t})((?:[	 ][^\\n]*)?(?:\\n|$))`),nextBulletRegex:t=>new RegExp(`^ {0,${Math.min(3,t-1)}}(?:[*+-]|\\d{1,9}[.)])((?:[ 	][^\\n]*)?(?:\\n|$))`),hrRegex:t=>new RegExp(`^ {0,${Math.min(3,t-1)}}((?:- *){3,}|(?:_ *){3,}|(?:\\* *){3,})(?:\\n+|$)`),fencesBeginRegex:t=>new RegExp(`^ {0,${Math.min(3,t-1)}}(?:\`\`\`|~~~)`),headingBeginRegex:t=>new RegExp(`^ {0,${Math.min(3,t-1)}}#`),htmlBeginRegex:t=>new RegExp(`^ {0,${Math.min(3,t-1)}}<(?:[a-z].*>|!--)`,"i")},TOA=/^(?:[ \t]*(?:\n|$))+/,HOA=/^((?: {4}| {0,3}\t)[^\n]+(?:\n(?:[ \t]*(?:\n|$))*)?)+/,zOA=/^ {0,3}(`{3,}(?=[^`\n]*(?:\n|$))|~{3,})([^\n]*)(?:\n|$)(?:|([\s\S]*?)(?:\n|$))(?: {0,3}\1[~`]* *(?=\n|$)|$)/,yf=/^ {0,3}((?:-[\t ]*){3,}|(?:_[ \t]*){3,}|(?:\*[ \t]*){3,})(?:\n+|$)/,OOA=/^ {0,3}(#{1,6})(?=\s|$)(.*)(?:\n+|$)/,XG=/(?:[*+-]|\d{1,9}[.)])/,elA=/^(?!bull |blockCode|fences|blockquote|heading|html|table)((?:.|\n(?!\s*?\n|bull |blockCode|fences|blockquote|heading|html|table))+?)\n {0,3}(=+|-+) *(?:\n+|$)/,tlA=Zn(elA).replace(/bull/g,XG).replace(/blockCode/g,/(?: {4}| {0,3}\t)/).replace(/fences/g,/ {0,3}(?:`{3,}|~{3,})/).replace(/blockquote/g,/ {0,3}>/).replace(/heading/g,/ {0,3}#{1,6}/).replace(/html/g,/ {0,3}<[^\n>]+>\n/).replace(/\|table/g,"").getRegex(),POA=Zn(elA).replace(/bull/g,XG).replace(/blockCode/g,/(?: {4}| {0,3}\t)/).replace(/fences/g,/ {0,3}(?:`{3,}|~{3,})/).replace(/blockquote/g,/ {0,3}>/).replace(/heading/g,/ {0,3}#{1,6}/).replace(/html/g,/ {0,3}<[^\n>]+>\n/).replace(/table/g,/ {0,3}\|?(?:[:\- ]*\|)+[\:\- ]*\n/).getRegex(),$G=/^([^\n]+(?:\n(?!hr|heading|lheading|blockquote|fences|list|html|table| +\n)[^\n]+)*)/,jOA=/^[^\n]+/,AU=/(?!\s*\])(?:\\.|[^\[\]\\])+/,qOA=Zn(/^ {0,3}\[(label)\]: *(?:\n[ \t]*)?([^<\s][^\s]*|<.*?>)(?:(?: +(?:\n[ \t]*)?| *\n[ \t]*)(title))? *(?:\n+|$)/).replace("label",AU).replace("title",/(?:"(?:\\"?|[^"\\])*"|'[^'\n]*(?:\n[^'\n]+)*\n?'|\([^()]*\))/).getRegex(),VOA=Zn(/^( {0,3}bull)([ \t][^\n]+?)?(?:\n|$)/).replace(/bull/g,XG).getRegex(),l7="address|article|aside|base|basefont|blockquote|body|caption|center|col|colgroup|dd|details|dialog|dir|div|dl|dt|fieldset|figcaption|figure|footer|form|frame|frameset|h[1-6]|head|header|hr|html|iframe|legend|li|link|main|menu|menuitem|meta|nav|noframes|ol|optgroup|option|p|param|search|section|summary|table|tbody|td|tfoot|th|thead|title|tr|track|ul",eU=/<!--(?:-?>|[\s\S]*?(?:-->|$))/,ZOA=Zn("^ {0,3}(?:<(script|pre|style|textarea)[\\s>][\\s\\S]*?(?:</\\1>[^\\n]*\\n+|$)|comment[^\\n]*(\\n+|$)|<\\?[\\s\\S]*?(?:\\?>\\n*|$)|<![A-Z][\\s\\S]*?(?:>\\n*|$)|<!\\[CDATA\\[[\\s\\S]*?(?:\\]\\]>\\n*|$)|</?(tag)(?: +|\\n|/?>)[\\s\\S]*?(?:(?:\\n[ 	]*)+\\n|$)|<(?!script|pre|style|textarea)([a-z][\\w-]*)(?:attribute)*? */?>(?=[ \\t]*(?:\\n|$))[\\s\\S]*?(?:(?:\\n[ 	]*)+\\n|$)|</(?!script|pre|style|textarea)[a-z][\\w-]*\\s*>(?=[ \\t]*(?:\\n|$))[\\s\\S]*?(?:(?:\\n[ 	]*)+\\n|$))","i").replace("comment",eU).replace("tag",l7).replace("attribute",/ +[a-zA-Z:_][\w.:-]*(?: *= *"[^"\n]*"| *= *'[^'\n]*'| *= *[^\s"'=<>`]+)?/).getRegex(),ilA=Zn($G).replace("hr",yf).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("|lheading","").replace("|table","").replace("blockquote"," {0,3}>").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",l7).getRegex(),WOA=Zn(/^( {0,3}> ?(paragraph|[^\n]*)(?:\n|$))+/).replace("paragraph",ilA).getRegex(),tU={blockquote:WOA,code:HOA,def:qOA,fences:zOA,heading:OOA,hr:yf,html:ZOA,lheading:tlA,list:VOA,newline:TOA,paragraph:ilA,table:Df,text:jOA},VcA=Zn("^ *([^\\n ].*)\\n {0,3}((?:\\| *)?:?-+:? *(?:\\| *:?-+:? *)*(?:\\| *)?)(?:\\n((?:(?! *\\n|hr|heading|blockquote|code|fences|list|html).*(?:\\n|$))*)\\n*|$)").replace("hr",yf).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("blockquote"," {0,3}>").replace("code","(?: {4}| {0,3}	)[^\\n]").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",l7).getRegex(),XOA=Ye(rA({},tU),{lheading:POA,table:VcA,paragraph:Zn($G).replace("hr",yf).replace("heading"," {0,3}#{1,6}(?:\\s|$)").replace("|lheading","").replace("table",VcA).replace("blockquote"," {0,3}>").replace("fences"," {0,3}(?:`{3,}(?=[^`\\n]*\\n)|~{3,})[^\\n]*\\n").replace("list"," {0,3}(?:[*+-]|1[.)]) ").replace("html","</?(?:tag)(?: +|\\n|/?>)|<(?:script|pre|style|textarea|!--)").replace("tag",l7).getRegex()}),$OA=Ye(rA({},tU),{html:Zn(`^ *(?:comment *(?:\\n|\\s*$)|<(tag)[\\s\\S]+?</\\1> *(?:\\n{2,}|\\s*$)|<tag(?:"[^"]*"|'[^']*'|\\s[^'"/>\\s]*)*?/?> *(?:\\n{2,}|\\s*$))`).replace("comment",eU).replace(/tag/g,"(?!(?:a|em|strong|small|s|cite|q|dfn|abbr|data|time|code|var|samp|kbd|sub|sup|i|b|u|mark|ruby|rt|rp|bdi|bdo|span|br|wbr|ins|del|img)\\b)\\w+(?!:|[^\\w\\s@]*@)\\b").getRegex(),def:/^ *\[([^\]]+)\]: *<?([^\s>]+)>?(?: +(["(][^\n]+[")]))? *(?:\n+|$)/,heading:/^(#{1,6})(.*)(?:\n+|$)/,fences:Df,lheading:/^(.+?)\n {0,3}(=+|-+) *(?:\n+|$)/,paragraph:Zn($G).replace("hr",yf).replace("heading",` *#{1,6} *[^
]`).replace("lheading",tlA).replace("|table","").replace("blockquote"," {0,3}>").replace("|fences","").replace("|list","").replace("|html","").replace("|tag","").getRegex()}),APA=/^\\([!"#$%&'()*+,\-./:;<=>?@\[\]\\^_`{|}~])/,ePA=/^(`+)([^`]|[^`][\s\S]*?[^`])\1(?!`)/,nlA=/^( {2,}|\\)\n(?!\s*$)/,tPA=/^(`+|[^`])(?:(?= {2,}\n)|[\s\S]*?(?:(?=[\\<!\[`*_]|\b_|$)|[^ ](?= {2,}\n)))/,g7=/[\p{P}\p{S}]/u,iU=/[\s\p{P}\p{S}]/u,olA=/[^\s\p{P}\p{S}]/u,iPA=Zn(/^((?![*_])punctSpace)/,"u").replace(/punctSpace/g,iU).getRegex(),rlA=/(?!~)[\p{P}\p{S}]/u,nPA=/(?!~)[\s\p{P}\p{S}]/u,oPA=/(?:[^\s\p{P}\p{S}]|~)/u,rPA=/\[[^[\]]*?\]\((?:\\.|[^\\\(\)]|\((?:\\.|[^\\\(\)])*\))*\)|`[^`]*?`|<[^<>]*?>/g,slA=/^(?:\*+(?:((?!\*)punct)|[^\s*]))|^_+(?:((?!_)punct)|([^\s_]))/,sPA=Zn(slA,"u").replace(/punct/g,g7).getRegex(),aPA=Zn(slA,"u").replace(/punct/g,rlA).getRegex(),alA="^[^_*]*?__[^_*]*?\\*[^_*]*?(?=__)|[^*]+(?=[^*])|(?!\\*)punct(\\*+)(?=[\\s]|$)|notPunctSpace(\\*+)(?!\\*)(?=punctSpace|$)|(?!\\*)punctSpace(\\*+)(?=notPunctSpace)|[\\s](\\*+)(?!\\*)(?=punct)|(?!\\*)punct(\\*+)(?!\\*)(?=punct)|notPunctSpace(\\*+)(?=notPunctSpace)",cPA=Zn(alA,"gu").replace(/notPunctSpace/g,olA).replace(/punctSpace/g,iU).replace(/punct/g,g7).getRegex(),lPA=Zn(alA,"gu").replace(/notPunctSpace/g,oPA).replace(/punctSpace/g,nPA).replace(/punct/g,rlA).getRegex(),gPA=Zn("^[^_*]*?\\*\\*[^_*]*?_[^_*]*?(?=\\*\\*)|[^_]+(?=[^_])|(?!_)punct(_+)(?=[\\s]|$)|notPunctSpace(_+)(?!_)(?=punctSpace|$)|(?!_)punctSpace(_+)(?=notPunctSpace)|[\\s](_+)(?!_)(?=punct)|(?!_)punct(_+)(?!_)(?=punct)","gu").replace(/notPunctSpace/g,olA).replace(/punctSpace/g,iU).replace(/punct/g,g7).getRegex(),IPA=Zn(/\\(punct)/,"gu").replace(/punct/g,g7).getRegex(),CPA=Zn(/^<(scheme:[^\s\x00-\x1f<>]*|email)>/).replace("scheme",/[a-zA-Z][a-zA-Z0-9+.-]{1,31}/).replace("email",/[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+(@)[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)+(?![-_])/).getRegex(),dPA=Zn(eU).replace("(?:-->|$)","-->").getRegex(),BPA=Zn("^comment|^</[a-zA-Z][\\w:-]*\\s*>|^<[a-zA-Z][\\w-]*(?:attribute)*?\\s*/?>|^<\\?[\\s\\S]*?\\?>|^<![a-zA-Z]+\\s[\\s\\S]*?>|^<!\\[CDATA\\[[\\s\\S]*?\\]\\]>").replace("comment",dPA).replace("attribute",/\s+[a-zA-Z:_][\w.:-]*(?:\s*=\s*"[^"]*"|\s*=\s*'[^']*'|\s*=\s*[^\s"'=<>`]+)?/).getRegex(),a7=/(?:\[(?:\\.|[^\[\]\\])*\]|\\.|`[^`]*`|[^\[\]\\`])*?/,EPA=Zn(/^!?\[(label)\]\(\s*(href)(?:(?:[ \t]*(?:\n[ \t]*)?)(title))?\s*\)/).replace("label",a7).replace("href",/<(?:\\.|[^\n<>\\])+>|[^ \t\n\x00-\x1f]*/).replace("title",/"(?:\\"?|[^"\\])*"|'(?:\\'?|[^'\\])*'|\((?:\\\)?|[^)\\])*\)/).getRegex(),clA=Zn(/^!?\[(label)\]\[(ref)\]/).replace("label",a7).replace("ref",AU).getRegex(),llA=Zn(/^!?\[(ref)\](?:\[\])?/).replace("ref",AU).getRegex(),QPA=Zn("reflink|nolink(?!\\()","g").replace("reflink",clA).replace("nolink",llA).getRegex(),nU={_backpedal:Df,anyPunctuation:IPA,autolink:CPA,blockSkip:rPA,br:nlA,code:ePA,del:Df,emStrongLDelim:sPA,emStrongRDelimAst:cPA,emStrongRDelimUnd:gPA,escape:APA,link:EPA,nolink:llA,punctuation:iPA,reflink:clA,reflinkSearch:QPA,tag:BPA,text:tPA,url:Df},hPA=Ye(rA({},nU),{link:Zn(/^!?\[(label)\]\((.*?)\)/).replace("label",a7).getRegex(),reflink:Zn(/^!?\[(label)\]\s*\[([^\]]*)\]/).replace("label",a7).getRegex()}),qG=Ye(rA({},nU),{emStrongRDelimAst:lPA,emStrongLDelim:aPA,url:Zn(/^((?:ftp|https?):\/\/|www\.)(?:[a-zA-Z0-9\-]+\.?)+[^\s<]*|^email/,"i").replace("email",/[A-Za-z0-9._+-]+(@)[a-zA-Z0-9-_]+(?:\.[a-zA-Z0-9-_]*[a-zA-Z0-9])+(?![-_])/).getRegex(),_backpedal:/(?:[^?!.,:;*_'"~()&]+|\([^)]*\)|&(?![a-zA-Z0-9]+;$)|[?!.,:;*_'"~)]+(?!$))+/,del:/^(~~?)(?=[^\s~])((?:\\.|[^\\])*?(?:\\.|[^\s~\\]))\1(?=[^~]|$)/,text:/^([`~]+|[^`~])(?:(?= {2,}\n)|(?=[a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-]+@)|[\s\S]*?(?:(?=[\\<!\[`*~_]|\b_|https?:\/\/|ftp:\/\/|www\.|$)|[^ ](?= {2,}\n)|[^a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-](?=[a-zA-Z0-9.!#$%&'*+\/=?_`{\|}~-]+@)))/}),uPA=Ye(rA({},qG),{br:Zn(nlA).replace("{2,}","*").getRegex(),text:Zn(qG.text).replace("\\b_","\\b_| {2,}\\n").replace(/\{2,\}/g,"*").getRegex()}),r7={normal:tU,gfm:XOA,pedantic:$OA},pf={normal:nU,gfm:qG,breaks:uPA,pedantic:hPA},fPA={"&":"&amp;","<":"&lt;",">":"&gt;",'"':"&quot;","'":"&#39;"},ZcA=t=>fPA[t];function A0(t,e){if(e){if(ea.escapeTest.test(t))return t.replace(ea.escapeReplace,ZcA)}else if(ea.escapeTestNoEncode.test(t))return t.replace(ea.escapeReplaceNoEncode,ZcA);return t}function WcA(t){try{t=encodeURI(t).replace(ea.percentDecode,"%")}catch{return null}return t}function XcA(t,e){let A=t.replace(ea.findPipe,(o,r,s)=>{let a=!1,c=r;for(;--c>=0&&s[c]==="\\";)a=!a;return a?"|":" |"}),i=A.split(ea.splitPipe),n=0;if(i[0].trim()||i.shift(),i.length>0&&!i.at(-1)?.trim()&&i.pop(),e)if(i.length>e)i.splice(e);else for(;i.length<e;)i.push("");for(;n<i.length;n++)i[n]=i[n].trim().replace(ea.slashPipe,"|");return i}function wf(t,e,A){let i=t.length;if(i===0)return"";let n=0;for(;n<i;){let o=t.charAt(i-n-1);if(o===e&&!A)n++;else if(o!==e&&A)n++;else break}return t.slice(0,i-n)}function mPA(t,e){if(t.indexOf(e[1])===-1)return-1;let A=0;for(let i=0;i<t.length;i++)if(t[i]==="\\")i++;else if(t[i]===e[0])A++;else if(t[i]===e[1]&&(A--,A<0))return i;return A>0?-2:-1}function $cA(t,e,A,i,n){let o=e.href,r=e.title||null,s=t[1].replace(n.other.outputLinkReplace,"$1");i.state.inLink=!0;let a={type:t[0].charAt(0)==="!"?"image":"link",raw:A,href:o,title:r,text:s,tokens:i.inlineTokens(s)};return i.state.inLink=!1,a}function pPA(t,e,A){let i=t.match(A.other.indentCodeCompensation);if(i===null)return e;let n=i[1];return e.split(`
`).map(o=>{let r=o.match(A.other.beginningSpace);if(r===null)return o;let[s]=r;return s.length>=n.length?o.slice(n.length):o}).join(`
`)}var c7=class{options;rules;lexer;constructor(t){this.options=t||sd}space(t){let e=this.rules.block.newline.exec(t);if(e&&e[0].length>0)return{type:"space",raw:e[0]}}code(t){let e=this.rules.block.code.exec(t);if(e){let A=e[0].replace(this.rules.other.codeRemoveIndent,"");return{type:"code",raw:e[0],codeBlockStyle:"indented",text:this.options.pedantic?A:wf(A,`
`)}}}fences(t){let e=this.rules.block.fences.exec(t);if(e){let A=e[0],i=pPA(A,e[3]||"",this.rules);return{type:"code",raw:A,lang:e[2]?e[2].trim().replace(this.rules.inline.anyPunctuation,"$1"):e[2],text:i}}}heading(t){let e=this.rules.block.heading.exec(t);if(e){let A=e[2].trim();if(this.rules.other.endingHash.test(A)){let i=wf(A,"#");(this.options.pedantic||!i||this.rules.other.endingSpaceChar.test(i))&&(A=i.trim())}return{type:"heading",raw:e[0],depth:e[1].length,text:A,tokens:this.lexer.inline(A)}}}hr(t){let e=this.rules.block.hr.exec(t);if(e)return{type:"hr",raw:wf(e[0],`
`)}}blockquote(t){let e=this.rules.block.blockquote.exec(t);if(e){let A=wf(e[0],`
`).split(`
`),i="",n="",o=[];for(;A.length>0;){let r=!1,s=[],a;for(a=0;a<A.length;a++)if(this.rules.other.blockquoteStart.test(A[a]))s.push(A[a]),r=!0;else if(!r)s.push(A[a]);else break;A=A.slice(a);let c=s.join(`
`),l=c.replace(this.rules.other.blockquoteSetextReplace,`
    $1`).replace(this.rules.other.blockquoteSetextReplace2,"");i=i?`${i}
${c}`:c,n=n?`${n}
${l}`:l;let I=this.lexer.state.top;if(this.lexer.state.top=!0,this.lexer.blockTokens(l,o,!0),this.lexer.state.top=I,A.length===0)break;let C=o.at(-1);if(C?.type==="code")break;if(C?.type==="blockquote"){let d=C,B=d.raw+`
`+A.join(`
`),E=this.blockquote(B);o[o.length-1]=E,i=i.substring(0,i.length-d.raw.length)+E.raw,n=n.substring(0,n.length-d.text.length)+E.text;break}else if(C?.type==="list"){let d=C,B=d.raw+`
`+A.join(`
`),E=this.list(B);o[o.length-1]=E,i=i.substring(0,i.length-C.raw.length)+E.raw,n=n.substring(0,n.length-d.raw.length)+E.raw,A=B.substring(o.at(-1).raw.length).split(`
`);continue}}return{type:"blockquote",raw:i,tokens:o,text:n}}}list(t){let e=this.rules.block.list.exec(t);if(e){let A=e[1].trim(),i=A.length>1,n={type:"list",raw:"",ordered:i,start:i?+A.slice(0,-1):"",loose:!1,items:[]};A=i?`\\d{1,9}\\${A.slice(-1)}`:`\\${A}`,this.options.pedantic&&(A=i?A:"[*+-]");let o=this.rules.other.listItemRegex(A),r=!1;for(;t;){let a=!1,c="",l="";if(!(e=o.exec(t))||this.rules.block.hr.test(t))break;c=e[0],t=t.substring(c.length);let I=e[2].split(`
`,1)[0].replace(this.rules.other.listReplaceTabs,u=>" ".repeat(3*u.length)),C=t.split(`
`,1)[0],d=!I.trim(),B=0;if(this.options.pedantic?(B=2,l=I.trimStart()):d?B=e[1].length+1:(B=e[2].search(this.rules.other.nonSpaceChar),B=B>4?1:B,l=I.slice(B),B+=e[1].length),d&&this.rules.other.blankLine.test(C)&&(c+=C+`
`,t=t.substring(C.length+1),a=!0),!a){let u=this.rules.other.nextBulletRegex(B),D=this.rules.other.hrRegex(B),L=this.rules.other.fencesBeginRegex(B),R=this.rules.other.headingBeginRegex(B),w=this.rules.other.htmlBeginRegex(B);for(;t;){let _=t.split(`
`,1)[0],K;if(C=_,this.options.pedantic?(C=C.replace(this.rules.other.listReplaceNesting,"  "),K=C):K=C.replace(this.rules.other.tabCharGlobal,"    "),L.test(C)||R.test(C)||w.test(C)||u.test(C)||D.test(C))break;if(K.search(this.rules.other.nonSpaceChar)>=B||!C.trim())l+=`
`+K.slice(B);else{if(d||I.replace(this.rules.other.tabCharGlobal,"    ").search(this.rules.other.nonSpaceChar)>=4||L.test(I)||R.test(I)||D.test(I))break;l+=`
`+C}!d&&!C.trim()&&(d=!0),c+=_+`
`,t=t.substring(_.length+1),I=K.slice(B)}}n.loose||(r?n.loose=!0:this.rules.other.doubleBlankLine.test(c)&&(r=!0));let E=null,h;this.options.gfm&&(E=this.rules.other.listIsTask.exec(l),E&&(h=E[0]!=="[ ] ",l=l.replace(this.rules.other.listReplaceTask,""))),n.items.push({type:"list_item",raw:c,task:!!E,checked:h,loose:!1,text:l,tokens:[]}),n.raw+=c}let s=n.items.at(-1);if(s)s.raw=s.raw.trimEnd(),s.text=s.text.trimEnd();else return;n.raw=n.raw.trimEnd();for(let a=0;a<n.items.length;a++)if(this.lexer.state.top=!1,n.items[a].tokens=this.lexer.blockTokens(n.items[a].text,[]),!n.loose){let c=n.items[a].tokens.filter(I=>I.type==="space"),l=c.length>0&&c.some(I=>this.rules.other.anyLine.test(I.raw));n.loose=l}if(n.loose)for(let a=0;a<n.items.length;a++)n.items[a].loose=!0;return n}}html(t){let e=this.rules.block.html.exec(t);if(e)return{type:"html",block:!0,raw:e[0],pre:e[1]==="pre"||e[1]==="script"||e[1]==="style",text:e[0]}}def(t){let e=this.rules.block.def.exec(t);if(e){let A=e[1].toLowerCase().replace(this.rules.other.multipleSpaceGlobal," "),i=e[2]?e[2].replace(this.rules.other.hrefBrackets,"$1").replace(this.rules.inline.anyPunctuation,"$1"):"",n=e[3]?e[3].substring(1,e[3].length-1).replace(this.rules.inline.anyPunctuation,"$1"):e[3];return{type:"def",tag:A,raw:e[0],href:i,title:n}}}table(t){let e=this.rules.block.table.exec(t);if(!e||!this.rules.other.tableDelimiter.test(e[2]))return;let A=XcA(e[1]),i=e[2].replace(this.rules.other.tableAlignChars,"").split("|"),n=e[3]?.trim()?e[3].replace(this.rules.other.tableRowBlankLine,"").split(`
`):[],o={type:"table",raw:e[0],header:[],align:[],rows:[]};if(A.length===i.length){for(let r of i)this.rules.other.tableAlignRight.test(r)?o.align.push("right"):this.rules.other.tableAlignCenter.test(r)?o.align.push("center"):this.rules.other.tableAlignLeft.test(r)?o.align.push("left"):o.align.push(null);for(let r=0;r<A.length;r++)o.header.push({text:A[r],tokens:this.lexer.inline(A[r]),header:!0,align:o.align[r]});for(let r of n)o.rows.push(XcA(r,o.header.length).map((s,a)=>({text:s,tokens:this.lexer.inline(s),header:!1,align:o.align[a]})));return o}}lheading(t){let e=this.rules.block.lheading.exec(t);if(e)return{type:"heading",raw:e[0],depth:e[2].charAt(0)==="="?1:2,text:e[1],tokens:this.lexer.inline(e[1])}}paragraph(t){let e=this.rules.block.paragraph.exec(t);if(e){let A=e[1].charAt(e[1].length-1)===`
`?e[1].slice(0,-1):e[1];return{type:"paragraph",raw:e[0],text:A,tokens:this.lexer.inline(A)}}}text(t){let e=this.rules.block.text.exec(t);if(e)return{type:"text",raw:e[0],text:e[0],tokens:this.lexer.inline(e[0])}}escape(t){let e=this.rules.inline.escape.exec(t);if(e)return{type:"escape",raw:e[0],text:e[1]}}tag(t){let e=this.rules.inline.tag.exec(t);if(e)return!this.lexer.state.inLink&&this.rules.other.startATag.test(e[0])?this.lexer.state.inLink=!0:this.lexer.state.inLink&&this.rules.other.endATag.test(e[0])&&(this.lexer.state.inLink=!1),!this.lexer.state.inRawBlock&&this.rules.other.startPreScriptTag.test(e[0])?this.lexer.state.inRawBlock=!0:this.lexer.state.inRawBlock&&this.rules.other.endPreScriptTag.test(e[0])&&(this.lexer.state.inRawBlock=!1),{type:"html",raw:e[0],inLink:this.lexer.state.inLink,inRawBlock:this.lexer.state.inRawBlock,block:!1,text:e[0]}}link(t){let e=this.rules.inline.link.exec(t);if(e){let A=e[2].trim();if(!this.options.pedantic&&this.rules.other.startAngleBracket.test(A)){if(!this.rules.other.endAngleBracket.test(A))return;let o=wf(A.slice(0,-1),"\\");if((A.length-o.length)%2===0)return}else{let o=mPA(e[2],"()");if(o===-2)return;if(o>-1){let s=(e[0].indexOf("!")===0?5:4)+e[1].length+o;e[2]=e[2].substring(0,o),e[0]=e[0].substring(0,s).trim(),e[3]=""}}let i=e[2],n="";if(this.options.pedantic){let o=this.rules.other.pedanticHrefTitle.exec(i);o&&(i=o[1],n=o[3])}else n=e[3]?e[3].slice(1,-1):"";return i=i.trim(),this.rules.other.startAngleBracket.test(i)&&(this.options.pedantic&&!this.rules.other.endAngleBracket.test(A)?i=i.slice(1):i=i.slice(1,-1)),$cA(e,{href:i&&i.replace(this.rules.inline.anyPunctuation,"$1"),title:n&&n.replace(this.rules.inline.anyPunctuation,"$1")},e[0],this.lexer,this.rules)}}reflink(t,e){let A;if((A=this.rules.inline.reflink.exec(t))||(A=this.rules.inline.nolink.exec(t))){let i=(A[2]||A[1]).replace(this.rules.other.multipleSpaceGlobal," "),n=e[i.toLowerCase()];if(!n){let o=A[0].charAt(0);return{type:"text",raw:o,text:o}}return $cA(A,n,A[0],this.lexer,this.rules)}}emStrong(t,e,A=""){let i=this.rules.inline.emStrongLDelim.exec(t);if(!i||i[3]&&A.match(this.rules.other.unicodeAlphaNumeric))return;if(!(i[1]||i[2]||"")||!A||this.rules.inline.punctuation.exec(A)){let o=[...i[0]].length-1,r,s,a=o,c=0,l=i[0][0]==="*"?this.rules.inline.emStrongRDelimAst:this.rules.inline.emStrongRDelimUnd;for(l.lastIndex=0,e=e.slice(-1*t.length+o);(i=l.exec(e))!=null;){if(r=i[1]||i[2]||i[3]||i[4]||i[5]||i[6],!r)continue;if(s=[...r].length,i[3]||i[4]){a+=s;continue}else if((i[5]||i[6])&&o%3&&!((o+s)%3)){c+=s;continue}if(a-=s,a>0)continue;s=Math.min(s,s+a+c);let I=[...i[0]][0].length,C=t.slice(0,o+i.index+I+s);if(Math.min(o,s)%2){let B=C.slice(1,-1);return{type:"em",raw:C,text:B,tokens:this.lexer.inlineTokens(B)}}let d=C.slice(2,-2);return{type:"strong",raw:C,text:d,tokens:this.lexer.inlineTokens(d)}}}}codespan(t){let e=this.rules.inline.code.exec(t);if(e){let A=e[2].replace(this.rules.other.newLineCharGlobal," "),i=this.rules.other.nonSpaceChar.test(A),n=this.rules.other.startingSpaceChar.test(A)&&this.rules.other.endingSpaceChar.test(A);return i&&n&&(A=A.substring(1,A.length-1)),{type:"codespan",raw:e[0],text:A}}}br(t){let e=this.rules.inline.br.exec(t);if(e)return{type:"br",raw:e[0]}}del(t){let e=this.rules.inline.del.exec(t);if(e)return{type:"del",raw:e[0],text:e[2],tokens:this.lexer.inlineTokens(e[2])}}autolink(t){let e=this.rules.inline.autolink.exec(t);if(e){let A,i;return e[2]==="@"?(A=e[1],i="mailto:"+A):(A=e[1],i=A),{type:"link",raw:e[0],text:A,href:i,tokens:[{type:"text",raw:A,text:A}]}}}url(t){let e;if(e=this.rules.inline.url.exec(t)){let A,i;if(e[2]==="@")A=e[0],i="mailto:"+A;else{let n;do n=e[0],e[0]=this.rules.inline._backpedal.exec(e[0])?.[0]??"";while(n!==e[0]);A=e[0],e[1]==="www."?i="http://"+e[0]:i=e[0]}return{type:"link",raw:e[0],text:A,href:i,tokens:[{type:"text",raw:A,text:A}]}}}inlineText(t){let e=this.rules.inline.text.exec(t);if(e){let A=this.lexer.state.inRawBlock;return{type:"text",raw:e[0],text:e[0],escaped:A}}}},s2=class VG{tokens;options;state;tokenizer;inlineQueue;constructor(e){this.tokens=[],this.tokens.links=Object.create(null),this.options=e||sd,this.options.tokenizer=this.options.tokenizer||new c7,this.tokenizer=this.options.tokenizer,this.tokenizer.options=this.options,this.tokenizer.lexer=this,this.inlineQueue=[],this.state={inLink:!1,inRawBlock:!1,top:!0};let A={other:ea,block:r7.normal,inline:pf.normal};this.options.pedantic?(A.block=r7.pedantic,A.inline=pf.pedantic):this.options.gfm&&(A.block=r7.gfm,this.options.breaks?A.inline=pf.breaks:A.inline=pf.gfm),this.tokenizer.rules=A}static get rules(){return{block:r7,inline:pf}}static lex(e,A){return new VG(A).lex(e)}static lexInline(e,A){return new VG(A).inlineTokens(e)}lex(e){e=e.replace(ea.carriageReturn,`
`),this.blockTokens(e,this.tokens);for(let A=0;A<this.inlineQueue.length;A++){let i=this.inlineQueue[A];this.inlineTokens(i.src,i.tokens)}return this.inlineQueue=[],this.tokens}blockTokens(e,A=[],i=!1){for(this.options.pedantic&&(e=e.replace(ea.tabCharGlobal,"    ").replace(ea.spaceLine,""));e;){let n;if(this.options.extensions?.block?.some(r=>(n=r.call({lexer:this},e,A))?(e=e.substring(n.raw.length),A.push(n),!0):!1))continue;if(n=this.tokenizer.space(e)){e=e.substring(n.raw.length);let r=A.at(-1);n.raw.length===1&&r!==void 0?r.raw+=`
`:A.push(n);continue}if(n=this.tokenizer.code(e)){e=e.substring(n.raw.length);let r=A.at(-1);r?.type==="paragraph"||r?.type==="text"?(r.raw+=`
`+n.raw,r.text+=`
`+n.text,this.inlineQueue.at(-1).src=r.text):A.push(n);continue}if(n=this.tokenizer.fences(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.heading(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.hr(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.blockquote(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.list(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.html(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.def(e)){e=e.substring(n.raw.length);let r=A.at(-1);r?.type==="paragraph"||r?.type==="text"?(r.raw+=`
`+n.raw,r.text+=`
`+n.raw,this.inlineQueue.at(-1).src=r.text):this.tokens.links[n.tag]||(this.tokens.links[n.tag]={href:n.href,title:n.title});continue}if(n=this.tokenizer.table(e)){e=e.substring(n.raw.length),A.push(n);continue}if(n=this.tokenizer.lheading(e)){e=e.substring(n.raw.length),A.push(n);continue}let o=e;if(this.options.extensions?.startBlock){let r=1/0,s=e.slice(1),a;this.options.extensions.startBlock.forEach(c=>{a=c.call({lexer:this},s),typeof a=="number"&&a>=0&&(r=Math.min(r,a))}),r<1/0&&r>=0&&(o=e.substring(0,r+1))}if(this.state.top&&(n=this.tokenizer.paragraph(o))){let r=A.at(-1);i&&r?.type==="paragraph"?(r.raw+=`
`+n.raw,r.text+=`
`+n.text,this.inlineQueue.pop(),this.inlineQueue.at(-1).src=r.text):A.push(n),i=o.length!==e.length,e=e.substring(n.raw.length);continue}if(n=this.tokenizer.text(e)){e=e.substring(n.raw.length);let r=A.at(-1);r?.type==="text"?(r.raw+=`
`+n.raw,r.text+=`
`+n.text,this.inlineQueue.pop(),this.inlineQueue.at(-1).src=r.text):A.push(n);continue}if(e){let r="Infinite loop on byte: "+e.charCodeAt(0);if(this.options.silent){console.error(r);break}else throw new Error(r)}}return this.state.top=!0,A}inline(e,A=[]){return this.inlineQueue.push({src:e,tokens:A}),A}inlineTokens(e,A=[]){let i=e,n=null;if(this.tokens.links){let s=Object.keys(this.tokens.links);if(s.length>0)for(;(n=this.tokenizer.rules.inline.reflinkSearch.exec(i))!=null;)s.includes(n[0].slice(n[0].lastIndexOf("[")+1,-1))&&(i=i.slice(0,n.index)+"["+"a".repeat(n[0].length-2)+"]"+i.slice(this.tokenizer.rules.inline.reflinkSearch.lastIndex))}for(;(n=this.tokenizer.rules.inline.anyPunctuation.exec(i))!=null;)i=i.slice(0,n.index)+"++"+i.slice(this.tokenizer.rules.inline.anyPunctuation.lastIndex);for(;(n=this.tokenizer.rules.inline.blockSkip.exec(i))!=null;)i=i.slice(0,n.index)+"["+"a".repeat(n[0].length-2)+"]"+i.slice(this.tokenizer.rules.inline.blockSkip.lastIndex);let o=!1,r="";for(;e;){o||(r=""),o=!1;let s;if(this.options.extensions?.inline?.some(c=>(s=c.call({lexer:this},e,A))?(e=e.substring(s.raw.length),A.push(s),!0):!1))continue;if(s=this.tokenizer.escape(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.tag(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.link(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.reflink(e,this.tokens.links)){e=e.substring(s.raw.length);let c=A.at(-1);s.type==="text"&&c?.type==="text"?(c.raw+=s.raw,c.text+=s.text):A.push(s);continue}if(s=this.tokenizer.emStrong(e,i,r)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.codespan(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.br(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.del(e)){e=e.substring(s.raw.length),A.push(s);continue}if(s=this.tokenizer.autolink(e)){e=e.substring(s.raw.length),A.push(s);continue}if(!this.state.inLink&&(s=this.tokenizer.url(e))){e=e.substring(s.raw.length),A.push(s);continue}let a=e;if(this.options.extensions?.startInline){let c=1/0,l=e.slice(1),I;this.options.extensions.startInline.forEach(C=>{I=C.call({lexer:this},l),typeof I=="number"&&I>=0&&(c=Math.min(c,I))}),c<1/0&&c>=0&&(a=e.substring(0,c+1))}if(s=this.tokenizer.inlineText(a)){e=e.substring(s.raw.length),s.raw.slice(-1)!=="_"&&(r=s.raw.slice(-1)),o=!0;let c=A.at(-1);c?.type==="text"?(c.raw+=s.raw,c.text+=s.text):A.push(s);continue}if(e){let c="Infinite loop on byte: "+e.charCodeAt(0);if(this.options.silent){console.error(c);break}else throw new Error(c)}}return A}},oI=class{options;parser;constructor(t){this.options=t||sd}space(t){return""}code({text:t,lang:e,escaped:A}){let i=(e||"").match(ea.notSpaceStart)?.[0],n=t.replace(ea.endingNewline,"")+`
`;return i?'<pre><code class="language-'+A0(i)+'">'+(A?n:A0(n,!0))+`</code></pre>
`:"<pre><code>"+(A?n:A0(n,!0))+`</code></pre>
`}blockquote({tokens:t}){return`<blockquote>
${this.parser.parse(t)}</blockquote>
`}html({text:t}){return t}heading({tokens:t,depth:e}){return`<h${e}>${this.parser.parseInline(t)}</h${e}>
`}hr(t){return`<hr>
`}list(t){let e=t.ordered,A=t.start,i="";for(let r=0;r<t.items.length;r++){let s=t.items[r];i+=this.listitem(s)}let n=e?"ol":"ul",o=e&&A!==1?' start="'+A+'"':"";return"<"+n+o+`>
`+i+"</"+n+`>
`}listitem(t){let e="";if(t.task){let A=this.checkbox({checked:!!t.checked});t.loose?t.tokens[0]?.type==="paragraph"?(t.tokens[0].text=A+" "+t.tokens[0].text,t.tokens[0].tokens&&t.tokens[0].tokens.length>0&&t.tokens[0].tokens[0].type==="text"&&(t.tokens[0].tokens[0].text=A+" "+A0(t.tokens[0].tokens[0].text),t.tokens[0].tokens[0].escaped=!0)):t.tokens.unshift({type:"text",raw:A+" ",text:A+" ",escaped:!0}):e+=A+" "}return e+=this.parser.parse(t.tokens,!!t.loose),`<li>${e}</li>
`}checkbox({checked:t}){return"<input "+(t?'checked="" ':"")+'disabled="" type="checkbox">'}paragraph({tokens:t}){return`<p>${this.parser.parseInline(t)}</p>
`}table(t){let e="",A="";for(let n=0;n<t.header.length;n++)A+=this.tablecell(t.header[n]);e+=this.tablerow({text:A});let i="";for(let n=0;n<t.rows.length;n++){let o=t.rows[n];A="";for(let r=0;r<o.length;r++)A+=this.tablecell(o[r]);i+=this.tablerow({text:A})}return i&&(i=`<tbody>${i}</tbody>`),`<table>
<thead>
`+e+`</thead>
`+i+`</table>
`}tablerow({text:t}){return`<tr>
${t}</tr>
`}tablecell(t){let e=this.parser.parseInline(t.tokens),A=t.header?"th":"td";return(t.align?`<${A} align="${t.align}">`:`<${A}>`)+e+`</${A}>
`}strong({tokens:t}){return`<strong>${this.parser.parseInline(t)}</strong>`}em({tokens:t}){return`<em>${this.parser.parseInline(t)}</em>`}codespan({text:t}){return`<code>${A0(t,!0)}</code>`}br(t){return"<br>"}del({tokens:t}){return`<del>${this.parser.parseInline(t)}</del>`}link({href:t,title:e,tokens:A}){let i=this.parser.parseInline(A),n=WcA(t);if(n===null)return i;t=n;let o='<a href="'+t+'"';return e&&(o+=' title="'+A0(e)+'"'),o+=">"+i+"</a>",o}image({href:t,title:e,text:A,tokens:i}){i&&(A=this.parser.parseInline(i,this.parser.textRenderer));let n=WcA(t);if(n===null)return A0(A);t=n;let o=`<img src="${t}" alt="${A}"`;return e&&(o+=` title="${A0(e)}"`),o+=">",o}text(t){return"tokens"in t&&t.tokens?this.parser.parseInline(t.tokens):"escaped"in t&&t.escaped?t.text:A0(t.text)}},oU=class{strong({text:t}){return t}em({text:t}){return t}codespan({text:t}){return t}del({text:t}){return t}html({text:t}){return t}text({text:t}){return t}link({text:t}){return""+t}image({text:t}){return""+t}br(){return""}},a2=class ZG{options;renderer;textRenderer;constructor(e){this.options=e||sd,this.options.renderer=this.options.renderer||new oI,this.renderer=this.options.renderer,this.renderer.options=this.options,this.renderer.parser=this,this.textRenderer=new oU}static parse(e,A){return new ZG(A).parse(e)}static parseInline(e,A){return new ZG(A).parseInline(e)}parse(e,A=!0){let i="";for(let n=0;n<e.length;n++){let o=e[n];if(this.options.extensions?.renderers?.[o.type]){let s=o,a=this.options.extensions.renderers[s.type].call({parser:this},s);if(a!==!1||!["space","hr","heading","code","table","blockquote","list","html","paragraph","text"].includes(s.type)){i+=a||"";continue}}let r=o;switch(r.type){case"space":{i+=this.renderer.space(r);continue}case"hr":{i+=this.renderer.hr(r);continue}case"heading":{i+=this.renderer.heading(r);continue}case"code":{i+=this.renderer.code(r);continue}case"table":{i+=this.renderer.table(r);continue}case"blockquote":{i+=this.renderer.blockquote(r);continue}case"list":{i+=this.renderer.list(r);continue}case"html":{i+=this.renderer.html(r);continue}case"paragraph":{i+=this.renderer.paragraph(r);continue}case"text":{let s=r,a=this.renderer.text(s);for(;n+1<e.length&&e[n+1].type==="text";)s=e[++n],a+=`
`+this.renderer.text(s);A?i+=this.renderer.paragraph({type:"paragraph",raw:a,text:a,tokens:[{type:"text",raw:a,text:a,escaped:!0}]}):i+=a;continue}default:{let s='Token with "'+r.type+'" type was not found.';if(this.options.silent)return console.error(s),"";throw new Error(s)}}}return i}parseInline(e,A=this.renderer){let i="";for(let n=0;n<e.length;n++){let o=e[n];if(this.options.extensions?.renderers?.[o.type]){let s=this.options.extensions.renderers[o.type].call({parser:this},o);if(s!==!1||!["escape","html","link","image","strong","em","codespan","br","del","text"].includes(o.type)){i+=s||"";continue}}let r=o;switch(r.type){case"escape":{i+=A.text(r);break}case"html":{i+=A.html(r);break}case"link":{i+=A.link(r);break}case"image":{i+=A.image(r);break}case"strong":{i+=A.strong(r);break}case"em":{i+=A.em(r);break}case"codespan":{i+=A.codespan(r);break}case"br":{i+=A.br(r);break}case"del":{i+=A.del(r);break}case"text":{i+=A.text(r);break}default:{let s='Token with "'+r.type+'" type was not found.';if(this.options.silent)return console.error(s),"";throw new Error(s)}}}return i}},s7=class{options;block;constructor(e){this.options=e||sd}static passThroughHooks=new Set(["preprocess","postprocess","processAllTokens"]);preprocess(e){return e}postprocess(e){return e}processAllTokens(e){return e}provideLexer(){return this.block?s2.lex:s2.lexInline}provideParser(){return this.block?a2.parse:a2.parseInline}},wPA=class{defaults=WG();options=this.setOptions;parse=this.parseMarkdown(!0);parseInline=this.parseMarkdown(!1);Parser=a2;Renderer=oI;TextRenderer=oU;Lexer=s2;Tokenizer=c7;Hooks=s7;constructor(...t){this.use(...t)}walkTokens(t,e){let A=[];for(let i of t)switch(A=A.concat(e.call(this,i)),i.type){case"table":{let n=i;for(let o of n.header)A=A.concat(this.walkTokens(o.tokens,e));for(let o of n.rows)for(let r of o)A=A.concat(this.walkTokens(r.tokens,e));break}case"list":{let n=i;A=A.concat(this.walkTokens(n.items,e));break}default:{let n=i;this.defaults.extensions?.childTokens?.[n.type]?this.defaults.extensions.childTokens[n.type].forEach(o=>{let r=n[o].flat(1/0);A=A.concat(this.walkTokens(r,e))}):n.tokens&&(A=A.concat(this.walkTokens(n.tokens,e)))}}return A}use(...t){let e=this.defaults.extensions||{renderers:{},childTokens:{}};return t.forEach(A=>{let i=rA({},A);if(i.async=this.defaults.async||i.async||!1,A.extensions&&(A.extensions.forEach(n=>{if(!n.name)throw new Error("extension name required");if("renderer"in n){let o=e.renderers[n.name];o?e.renderers[n.name]=function(...r){let s=n.renderer.apply(this,r);return s===!1&&(s=o.apply(this,r)),s}:e.renderers[n.name]=n.renderer}if("tokenizer"in n){if(!n.level||n.level!=="block"&&n.level!=="inline")throw new Error("extension level must be 'block' or 'inline'");let o=e[n.level];o?o.unshift(n.tokenizer):e[n.level]=[n.tokenizer],n.start&&(n.level==="block"?e.startBlock?e.startBlock.push(n.start):e.startBlock=[n.start]:n.level==="inline"&&(e.startInline?e.startInline.push(n.start):e.startInline=[n.start]))}"childTokens"in n&&n.childTokens&&(e.childTokens[n.name]=n.childTokens)}),i.extensions=e),A.renderer){let n=this.defaults.renderer||new oI(this.defaults);for(let o in A.renderer){if(!(o in n))throw new Error(`renderer '${o}' does not exist`);if(["options","parser"].includes(o))continue;let r=o,s=A.renderer[r],a=n[r];n[r]=(...c)=>{let l=s.apply(n,c);return l===!1&&(l=a.apply(n,c)),l||""}}i.renderer=n}if(A.tokenizer){let n=this.defaults.tokenizer||new c7(this.defaults);for(let o in A.tokenizer){if(!(o in n))throw new Error(`tokenizer '${o}' does not exist`);if(["options","rules","lexer"].includes(o))continue;let r=o,s=A.tokenizer[r],a=n[r];n[r]=(...c)=>{let l=s.apply(n,c);return l===!1&&(l=a.apply(n,c)),l}}i.tokenizer=n}if(A.hooks){let n=this.defaults.hooks||new s7;for(let o in A.hooks){if(!(o in n))throw new Error(`hook '${o}' does not exist`);if(["options","block"].includes(o))continue;let r=o,s=A.hooks[r],a=n[r];s7.passThroughHooks.has(o)?n[r]=c=>{if(this.defaults.async)return Promise.resolve(s.call(n,c)).then(I=>a.call(n,I));let l=s.call(n,c);return a.call(n,l)}:n[r]=(...c)=>{let l=s.apply(n,c);return l===!1&&(l=a.apply(n,c)),l}}i.hooks=n}if(A.walkTokens){let n=this.defaults.walkTokens,o=A.walkTokens;i.walkTokens=function(r){let s=[];return s.push(o.call(this,r)),n&&(s=s.concat(n.call(this,r))),s}}this.defaults=rA(rA({},this.defaults),i)}),this}setOptions(t){return this.defaults=rA(rA({},this.defaults),t),this}lexer(t,e){return s2.lex(t,e??this.defaults)}parser(t,e){return a2.parse(t,e??this.defaults)}parseMarkdown(t){return(A,i)=>{let n=rA({},i),o=rA(rA({},this.defaults),n),r=this.onError(!!o.silent,!!o.async);if(this.defaults.async===!0&&n.async===!1)return r(new Error("marked(): The async option was set to true by an extension. Remove async: false from the parse options object to return a Promise."));if(typeof A>"u"||A===null)return r(new Error("marked(): input parameter is undefined or null"));if(typeof A!="string")return r(new Error("marked(): input parameter is of type "+Object.prototype.toString.call(A)+", string expected"));o.hooks&&(o.hooks.options=o,o.hooks.block=t);let s=o.hooks?o.hooks.provideLexer():t?s2.lex:s2.lexInline,a=o.hooks?o.hooks.provideParser():t?a2.parse:a2.parseInline;if(o.async)return Promise.resolve(o.hooks?o.hooks.preprocess(A):A).then(c=>s(c,o)).then(c=>o.hooks?o.hooks.processAllTokens(c):c).then(c=>o.walkTokens?Promise.all(this.walkTokens(c,o.walkTokens)).then(()=>c):c).then(c=>a(c,o)).then(c=>o.hooks?o.hooks.postprocess(c):c).catch(r);try{o.hooks&&(A=o.hooks.preprocess(A));let c=s(A,o);o.hooks&&(c=o.hooks.processAllTokens(c)),o.walkTokens&&this.walkTokens(c,o.walkTokens);let l=a(c,o);return o.hooks&&(l=o.hooks.postprocess(l)),l}catch(c){return r(c)}}}onError(t,e){return A=>{if(A.message+=`
Please report this to https://github.com/markedjs/marked.`,t){let i="<p>An error occurred:</p><pre>"+A0(A.message+"",!0)+"</pre>";return e?Promise.resolve(i):i}if(e)return Promise.reject(A);throw A}}},rd=new wPA;function Mn(t,e){return rd.parse(t,e)}Mn.options=Mn.setOptions=function(t){return rd.setOptions(t),Mn.defaults=rd.defaults,AlA(Mn.defaults),Mn};Mn.getDefaults=WG;Mn.defaults=sd;Mn.use=function(...t){return rd.use(...t),Mn.defaults=rd.defaults,AlA(Mn.defaults),Mn};Mn.walkTokens=function(t,e){return rd.walkTokens(t,e)};Mn.parseInline=rd.parseInline;Mn.Parser=a2;Mn.parser=a2.parse;Mn.Renderer=oI;Mn.TextRenderer=oU;Mn.Lexer=s2;Mn.lexer=s2.lex;Mn.Tokenizer=c7;Mn.Hooks=s7;Mn.parse=Mn;var R7e=Mn.options,x7e=Mn.setOptions,L7e=Mn.use,F7e=Mn.walkTokens,N7e=Mn.parseInline;var _7e=a2.parse,G7e=s2.lex;var DPA=["*"],yPA="Copy",vPA="Copied",bPA=(()=>{class t{constructor(){this._buttonClick$=new OA,this.copied$=this._buttonClick$.pipe(jn(()=>zn(Me(!0),II(3e3).pipe(vd(!1)))),tl(),a0(1)),this.copiedText$=this.copied$.pipe(Pn(!1),je(A=>A?vPA:yPA))}onCopyToClipboardClick(){this._buttonClick$.next()}static{this.\u0275fac=function(i){return new(i||t)}}static{this.\u0275cmp=zA({type:t,selectors:[["markdown-clipboard"]],decls:4,vars:7,consts:[[1,"markdown-clipboard-button",3,"click"]],template:function(i,n){i&1&&(S(0,"button",0),Za(1,"async"),hA("click",function(){return n.onCopyToClipboardClick()}),AA(2),Za(3,"async"),F()),i&2&&(ue("copied",M2(1,3,n.copied$)),G(2),Gt(M2(3,5,n.copiedText$)))},dependencies:[Jh],encapsulation:2,changeDetection:0})}}return t})(),MPA=new dA("CLIPBOARD_OPTIONS");var rU=function(t){return t.CommandLine="command-line",t.LineHighlight="line-highlight",t.LineNumbers="line-numbers",t}(rU||{}),glA=new dA("MARKED_EXTENSIONS"),kPA=new dA("MARKED_OPTIONS"),SPA=new dA("MERMAID_OPTIONS"),RPA="[ngx-markdown] When using the `emoji` attribute you *have to* include Emoji-Toolkit files to `angular.json` or use imports. See README for more information",xPA="[ngx-markdown] When using the `katex` attribute you *have to* include KaTeX files to `angular.json` or use imports. See README for more information",LPA="[ngx-markdown] When using the `mermaid` attribute you *have to* include Mermaid files to `angular.json` or use imports. See README for more information",FPA="[ngx-markdown] When using the `clipboard` attribute you *have to* include Clipboard files to `angular.json` or use imports. See README for more information",NPA="[ngx-markdown] When using the `clipboard` attribute you *have to* provide the `viewContainerRef` parameter to `MarkdownService.render()` function",_PA="[ngx-markdown] When using the `src` attribute you *have to* pass the `HttpClient` as a parameter of the `forRoot` method. See README for more information",IlA=new dA("SECURITY_CONTEXT");var ClA=(()=>{class t{get options(){return this._options}set options(A){this._options=rA(rA({},this.DEFAULT_MARKED_OPTIONS),A)}get renderer(){return this.options.renderer}set renderer(A){this.options.renderer=A}constructor(A,i,n,o,r,s,a,c){this.clipboardOptions=A,this.extensions=i,this.mermaidOptions=o,this.platform=r,this.securityContext=s,this.http=a,this.sanitizer=c,this.DEFAULT_MARKED_OPTIONS={renderer:new oI},this.DEFAULT_KATEX_OPTIONS={delimiters:[{left:"$$",right:"$$",display:!0},{left:"$",right:"$",display:!1},{left:"\\(",right:"\\)",display:!1},{left:"\\begin{equation}",right:"\\end{equation}",display:!0},{left:"\\begin{align}",right:"\\end{align}",display:!0},{left:"\\begin{alignat}",right:"\\end{alignat}",display:!0},{left:"\\begin{gather}",right:"\\end{gather}",display:!0},{left:"\\begin{CD}",right:"\\end{CD}",display:!0},{left:"\\[",right:"\\]",display:!0}]},this.DEFAULT_MERMAID_OPTIONS={startOnLoad:!1},this.DEFAULT_CLIPBOARD_OPTIONS={buttonComponent:void 0},this.DEFAULT_PARSE_OPTIONS={decodeHtml:!1,inline:!1,emoji:!1,mermaid:!1,markedOptions:void 0,disableSanitizer:!1},this.DEFAULT_RENDER_OPTIONS={clipboard:!1,clipboardOptions:void 0,katex:!1,katexOptions:void 0,mermaid:!1,mermaidOptions:void 0},this._reload$=new OA,this.reload$=this._reload$.asObservable(),this.options=n}parse(A,i=this.DEFAULT_PARSE_OPTIONS){let{decodeHtml:n,inline:o,emoji:r,mermaid:s,disableSanitizer:a}=i,c=rA(rA({},this.options),i.markedOptions),l=c.renderer||this.renderer||new oI;this.extensions&&(this.renderer=this.extendsRendererForExtensions(l)),s&&(this.renderer=this.extendsRendererForMermaid(l));let I=this.trimIndentation(A),C=n?this.decodeHtml(I):I,d=r?this.parseEmoji(C):C,B=this.parseMarked(d,c,o);return(a?B:this.sanitizer.sanitize(this.securityContext,B))||""}render(A,i=this.DEFAULT_RENDER_OPTIONS,n){let{clipboard:o,clipboardOptions:r,katex:s,katexOptions:a,mermaid:c,mermaidOptions:l}=i;s&&this.renderKatex(A,rA(rA({},this.DEFAULT_KATEX_OPTIONS),a)),c&&this.renderMermaid(A,rA(rA(rA({},this.DEFAULT_MERMAID_OPTIONS),this.mermaidOptions),l)),o&&this.renderClipboard(A,n,rA(rA(rA({},this.DEFAULT_CLIPBOARD_OPTIONS),this.clipboardOptions),r)),this.highlight(A)}reload(){this._reload$.next()}getSource(A){if(!this.http)throw new Error(_PA);return this.http.get(A,{responseType:"text"}).pipe(je(i=>this.handleExtension(A,i)))}highlight(A){if(!sg(this.platform)||typeof Prism>"u"||typeof Prism.highlightAllUnder>"u")return;A||(A=document);let i=A.querySelectorAll('pre code:not([class*="language-"])');Array.prototype.forEach.call(i,n=>n.classList.add("language-none")),Prism.highlightAllUnder(A)}decodeHtml(A){if(!sg(this.platform))return A;let i=document.createElement("textarea");return i.innerHTML=A,i.value}extendsRendererForExtensions(A){let i=A;return i.\u0275NgxMarkdownRendererExtendedForExtensions===!0||(this.extensions?.length>0&&Mn.use(...this.extensions),i.\u0275NgxMarkdownRendererExtendedForExtensions=!0),A}extendsRendererForMermaid(A){let i=A;if(i.\u0275NgxMarkdownRendererExtendedForMermaid===!0)return A;let n=A.code;return A.code=o=>o.lang==="mermaid"?`<div class="mermaid">${o.text}</div>`:n(o),i.\u0275NgxMarkdownRendererExtendedForMermaid=!0,A}handleExtension(A,i){let n=A.lastIndexOf("://"),o=n>-1?A.substring(n+4):A,r=o.lastIndexOf("/"),s=r>-1?o.substring(r+1).split("?")[0]:"",a=s.lastIndexOf("."),c=a>-1?s.substring(a+1):"";return c&&c!=="md"?"```"+c+`
`+i+"\n```":i}parseMarked(A,i,n=!1){if(i.renderer){let o=rA({},i.renderer);delete o.\u0275NgxMarkdownRendererExtendedForExtensions,delete o.\u0275NgxMarkdownRendererExtendedForMermaid,delete i.renderer,Mn.use({renderer:o})}return n?Mn.parseInline(A,i):Mn.parse(A,i)}parseEmoji(A){if(!sg(this.platform))return A;if(typeof joypixels>"u"||typeof joypixels.shortnameToUnicode>"u")throw new Error(RPA);return joypixels.shortnameToUnicode(A)}renderKatex(A,i){if(sg(this.platform)){if(typeof katex>"u"||typeof renderMathInElement>"u")throw new Error(xPA);renderMathInElement(A,i)}}renderClipboard(A,i,n){if(!sg(this.platform))return;if(typeof ClipboardJS>"u")throw new Error(FPA);if(!i)throw new Error(NPA);let{buttonComponent:o,buttonTemplate:r}=n,s=A.querySelectorAll("pre");for(let a=0;a<s.length;a++){let c=s.item(a),l=document.createElement("div");l.style.position="relative",c.parentNode.insertBefore(l,c),l.appendChild(c);let I=document.createElement("div");I.classList.add("markdown-clipboard-toolbar"),I.style.position="absolute",I.style.top=".5em",I.style.right=".5em",I.style.zIndex="1",l.insertAdjacentElement("beforeend",I),l.onmouseenter=()=>I.classList.add("hover"),l.onmouseleave=()=>I.classList.remove("hover");let C;if(o){let B=i.createComponent(o);C=B.hostView,B.changeDetectorRef.markForCheck()}else if(r)C=i.createEmbeddedView(r);else{let B=i.createComponent(bPA);C=B.hostView,B.changeDetectorRef.markForCheck()}let d;C.rootNodes.forEach(B=>{I.appendChild(B),d=new ClipboardJS(B,{text:()=>c.innerText})}),C.onDestroy(()=>d.destroy())}}renderMermaid(A,i=this.DEFAULT_MERMAID_OPTIONS){if(!sg(this.platform))return;if(typeof mermaid>"u"||typeof mermaid.initialize>"u")throw new Error(LPA);let n=A.querySelectorAll(".mermaid");n.length!==0&&(mermaid.initialize(i),mermaid.run({nodes:n}))}trimIndentation(A){if(!A)return"";let i;return A.split(`
`).map(n=>{let o=i;return n.length>0&&(o=isNaN(o)?n.search(/\S|$/):Math.min(n.search(/\S|$/),o)),isNaN(i)&&(i=o),o?n.substring(o):n}).join(`
`)}static{this.\u0275fac=function(i){return new(i||t)(we(MPA,8),we(glA,8),we(kPA,8),we(SPA,8),we(og),we(IlA),we(Ds,8),we(cl))}}static{this.\u0275prov=NA({token:t,factory:t.\u0275fac})}}return t})(),dlA=(()=>{class t{get disableSanitizer(){return this._disableSanitizer}set disableSanitizer(A){this._disableSanitizer=this.coerceBooleanProperty(A)}get inline(){return this._inline}set inline(A){this._inline=this.coerceBooleanProperty(A)}get clipboard(){return this._clipboard}set clipboard(A){this._clipboard=this.coerceBooleanProperty(A)}get emoji(){return this._emoji}set emoji(A){this._emoji=this.coerceBooleanProperty(A)}get katex(){return this._katex}set katex(A){this._katex=this.coerceBooleanProperty(A)}get mermaid(){return this._mermaid}set mermaid(A){this._mermaid=this.coerceBooleanProperty(A)}get lineHighlight(){return this._lineHighlight}set lineHighlight(A){this._lineHighlight=this.coerceBooleanProperty(A)}get lineNumbers(){return this._lineNumbers}set lineNumbers(A){this._lineNumbers=this.coerceBooleanProperty(A)}get commandLine(){return this._commandLine}set commandLine(A){this._commandLine=this.coerceBooleanProperty(A)}constructor(A,i,n){this.element=A,this.markdownService=i,this.viewContainerRef=n,this.error=new WA,this.load=new WA,this.ready=new WA,this._clipboard=!1,this._commandLine=!1,this._disableSanitizer=!1,this._emoji=!1,this._inline=!1,this._katex=!1,this._lineHighlight=!1,this._lineNumbers=!1,this._mermaid=!1,this.destroyed$=new OA}ngOnChanges(){this.loadContent()}loadContent(){if(this.data!=null){this.handleData();return}if(this.src!=null){this.handleSrc();return}}ngAfterViewInit(){!this.data&&!this.src&&this.handleTransclusion(),this.markdownService.reload$.pipe(St(this.destroyed$)).subscribe(()=>this.loadContent())}ngOnDestroy(){this.destroyed$.next(),this.destroyed$.complete()}render(A,i=!1){return Ao(this,null,function*(){let n={decodeHtml:i,inline:this.inline,emoji:this.emoji,mermaid:this.mermaid,disableSanitizer:this.disableSanitizer},o={clipboard:this.clipboard,clipboardOptions:this.getClipboardOptions(),katex:this.katex,katexOptions:this.katexOptions,mermaid:this.mermaid,mermaidOptions:this.mermaidOptions},r=yield this.markdownService.parse(A,n);this.element.nativeElement.innerHTML=r,this.handlePlugins(),this.markdownService.render(this.element.nativeElement,o,this.viewContainerRef),this.ready.emit()})}coerceBooleanProperty(A){return A!=null&&`${String(A)}`!="false"}getClipboardOptions(){if(this.clipboardButtonComponent||this.clipboardButtonTemplate)return{buttonComponent:this.clipboardButtonComponent,buttonTemplate:this.clipboardButtonTemplate}}handleData(){this.render(this.data)}handleSrc(){this.markdownService.getSource(this.src).subscribe({next:A=>{this.render(A).then(()=>{this.load.emit(A)})},error:A=>this.error.emit(A)})}handleTransclusion(){this.render(this.element.nativeElement.innerHTML,!0)}handlePlugins(){this.commandLine&&(this.setPluginClass(this.element.nativeElement,rU.CommandLine),this.setPluginOptions(this.element.nativeElement,{dataFilterOutput:this.filterOutput,dataHost:this.host,dataPrompt:this.prompt,dataOutput:this.output,dataUser:this.user})),this.lineHighlight&&this.setPluginOptions(this.element.nativeElement,{dataLine:this.line,dataLineOffset:this.lineOffset}),this.lineNumbers&&(this.setPluginClass(this.element.nativeElement,rU.LineNumbers),this.setPluginOptions(this.element.nativeElement,{dataStart:this.start}))}setPluginClass(A,i){let n=A.querySelectorAll("pre");for(let o=0;o<n.length;o++){let r=i instanceof Array?i:[i];n.item(o).classList.add(...r)}}setPluginOptions(A,i){let n=A.querySelectorAll("pre");for(let o=0;o<n.length;o++)Object.keys(i).forEach(r=>{let s=i[r];if(s){let a=this.toLispCase(r);n.item(o).setAttribute(a,s.toString())}})}toLispCase(A){let i=A.match(/([A-Z])/g);if(!i)return A;let n=A.toString();for(let o=0,r=i.length;o<r;o++)n=n.replace(new RegExp(i[o]),"-"+i[o].toLowerCase());return n.slice(0,1)==="-"&&(n=n.slice(1)),n}static{this.\u0275fac=function(i){return new(i||t)(PA(ee),PA(ClA),PA(Nn))}}static{this.\u0275cmp=zA({type:t,selectors:[["markdown"],["","markdown",""]],inputs:{data:"data",src:"src",disableSanitizer:"disableSanitizer",inline:"inline",clipboard:"clipboard",clipboardButtonComponent:"clipboardButtonComponent",clipboardButtonTemplate:"clipboardButtonTemplate",emoji:"emoji",katex:"katex",katexOptions:"katexOptions",mermaid:"mermaid",mermaidOptions:"mermaidOptions",lineHighlight:"lineHighlight",line:"line",lineOffset:"lineOffset",lineNumbers:"lineNumbers",start:"start",commandLine:"commandLine",filterOutput:"filterOutput",host:"host",prompt:"prompt",output:"output",user:"user"},outputs:{error:"error",load:"load",ready:"ready"},features:[ti],ngContentSelectors:DPA,decls:1,vars:0,template:function(i,n){i&1&&(jt(),Fe(0))},encapsulation:2})}}return t})();function GPA(t){return[ClA,t?.loader??[],t?.clipboardOptions??[],t?.markedOptions??[],t?.mermaidOptions??[],KPA(t?.markedExtensions)??[],{provide:IlA,useValue:t?.sanitize??jr.HTML}]}function UPA(t){return t!=null&&t.provide!=null}function KPA(t){if(t)return t.reduce((e,A)=>{let i=UPA(A)?Ye(rA({},A),{multi:!0}):{provide:glA,useValue:A,multi:!0};return[...e,i]},[])}var BlA=(()=>{class t{static forRoot(A){return{ngModule:t,providers:[GPA(A)]}}static forChild(){return{ngModule:t}}static{this.\u0275fac=function(i){return new(i||t)}}static{this.\u0275mod=Ce({type:t})}static{this.\u0275inj=Ie({imports:[f0]})}}return t})();var JPA=["switch"],TPA=["*"];function HPA(t,e){t&1&&(S(0,"span",10),ar(),S(1,"svg",12),JA(2,"path",13),F(),S(3,"svg",14),JA(4,"path",15),F()())}var zPA=new dA("mat-slide-toggle-default-options",{providedIn:"root",factory:()=>({disableToggleValue:!1,hideIcon:!1,disabledInteractive:!1})}),OPA={provide:yc,useExisting:Ir(()=>C7),multi:!0},I7=class{source;checked;constructor(e,A){this.source=e,this.checked=A}},C7=(()=>{class t{_elementRef=f(ee);_focusMonitor=f(dr);_changeDetectorRef=f(It);defaults=f(zPA);_onChange=A=>{};_onTouched=()=>{};_validatorOnChange=()=>{};_uniqueId;_checked=!1;_createChangeEvent(A){return new I7(this,A)}_labelId;get buttonId(){return`${this.id||this._uniqueId}-button`}_switchElement;focus(){this._switchElement.nativeElement.focus()}_noopAnimations;_focused;name=null;id;labelPosition="after";ariaLabel=null;ariaLabelledby=null;ariaDescribedby;required;color;disabled=!1;disableRipple=!1;tabIndex=0;get checked(){return this._checked}set checked(A){this._checked=A,this._changeDetectorRef.markForCheck()}hideIcon;disabledInteractive;change=new WA;toggleChange=new WA;get inputId(){return`${this.id||this._uniqueId}-input`}constructor(){f(Rn).load(lr);let A=f(new wr("tabindex"),{optional:!0}),i=this.defaults,n=f(Si,{optional:!0});this.tabIndex=A==null?0:parseInt(A)||0,this.color=i.color||"accent",this._noopAnimations=n==="NoopAnimations",this.id=this._uniqueId=f(sn).getId("mat-mdc-slide-toggle-"),this.hideIcon=i.hideIcon??!1,this.disabledInteractive=i.disabledInteractive??!1,this._labelId=this._uniqueId+"-label"}ngAfterContentInit(){this._focusMonitor.monitor(this._elementRef,!0).subscribe(A=>{A==="keyboard"||A==="program"?(this._focused=!0,this._changeDetectorRef.markForCheck()):A||Promise.resolve().then(()=>{this._focused=!1,this._onTouched(),this._changeDetectorRef.markForCheck()})})}ngOnChanges(A){A.required&&this._validatorOnChange()}ngOnDestroy(){this._focusMonitor.stopMonitoring(this._elementRef)}writeValue(A){this.checked=!!A}registerOnChange(A){this._onChange=A}registerOnTouched(A){this._onTouched=A}validate(A){return this.required&&A.value!==!0?{required:!0}:null}registerOnValidatorChange(A){this._validatorOnChange=A}setDisabledState(A){this.disabled=A,this._changeDetectorRef.markForCheck()}toggle(){this.checked=!this.checked,this._onChange(this.checked)}_emitChangeEvent(){this._onChange(this.checked),this.change.emit(this._createChangeEvent(this.checked))}_handleClick(){this.disabled||(this.toggleChange.emit(),this.defaults.disableToggleValue||(this.checked=!this.checked,this._onChange(this.checked),this.change.emit(new I7(this,this.checked))))}_getAriaLabelledBy(){return this.ariaLabelledby?this.ariaLabelledby:this.ariaLabel?null:this._labelId}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-slide-toggle"]],viewQuery:function(i,n){if(i&1&&Te(JPA,5),i&2){let o;XA(o=$A())&&(n._switchElement=o.first)}},hostAttrs:[1,"mat-mdc-slide-toggle"],hostVars:13,hostBindings:function(i,n){i&2&&(Hs("id",n.id),Ne("tabindex",null)("aria-label",null)("name",null)("aria-labelledby",null),fo(n.color?"mat-"+n.color:""),ue("mat-mdc-slide-toggle-focused",n._focused)("mat-mdc-slide-toggle-checked",n.checked)("_mat-animation-noopable",n._noopAnimations))},inputs:{name:"name",id:"id",labelPosition:"labelPosition",ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],ariaDescribedby:[0,"aria-describedby","ariaDescribedby"],required:[2,"required","required",ie],color:"color",disabled:[2,"disabled","disabled",ie],disableRipple:[2,"disableRipple","disableRipple",ie],tabIndex:[2,"tabIndex","tabIndex",A=>A==null?0:zi(A)],checked:[2,"checked","checked",ie],hideIcon:[2,"hideIcon","hideIcon",ie],disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie]},outputs:{change:"change",toggleChange:"toggleChange"},exportAs:["matSlideToggle"],features:[ht([OPA,{provide:w0,useExisting:t,multi:!0}]),ti],ngContentSelectors:TPA,decls:13,vars:27,consts:[["switch",""],["mat-internal-form-field","",3,"labelPosition"],["role","switch","type","button",1,"mdc-switch",3,"click","tabIndex","disabled"],[1,"mdc-switch__track"],[1,"mdc-switch__handle-track"],[1,"mdc-switch__handle"],[1,"mdc-switch__shadow"],[1,"mdc-elevation-overlay"],[1,"mdc-switch__ripple"],["mat-ripple","",1,"mat-mdc-slide-toggle-ripple","mat-focus-indicator",3,"matRippleTrigger","matRippleDisabled","matRippleCentered"],[1,"mdc-switch__icons"],[1,"mdc-label",3,"click","for"],["viewBox","0 0 24 24","aria-hidden","true",1,"mdc-switch__icon","mdc-switch__icon--on"],["d","M19.69,5.23L8.96,15.96l-4.23-4.23L2.96,13.5l6,6L21.46,7L19.69,5.23z"],["viewBox","0 0 24 24","aria-hidden","true",1,"mdc-switch__icon","mdc-switch__icon--off"],["d","M20 13H4v-2h16v2z"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"div",1)(1,"button",2,0),hA("click",function(){return RA(o),xA(n._handleClick())}),JA(3,"span",3),S(4,"span",4)(5,"span",5)(6,"span",6),JA(7,"span",7),F(),S(8,"span",8),JA(9,"span",9),F(),_A(10,HPA,5,0,"span",10),F()()(),S(11,"label",11),hA("click",function(s){return RA(o),xA(s.stopPropagation())}),Fe(12),F()()}if(i&2){let o=cr(2);yA("labelPosition",n.labelPosition),G(),ue("mdc-switch--selected",n.checked)("mdc-switch--unselected",!n.checked)("mdc-switch--checked",n.checked)("mdc-switch--disabled",n.disabled)("mat-mdc-slide-toggle-disabled-interactive",n.disabledInteractive),yA("tabIndex",n.disabled&&!n.disabledInteractive?-1:n.tabIndex)("disabled",n.disabled&&!n.disabledInteractive),Ne("id",n.buttonId)("name",n.name)("aria-label",n.ariaLabel)("aria-labelledby",n._getAriaLabelledBy())("aria-describedby",n.ariaDescribedby)("aria-required",n.required||null)("aria-checked",n.checked)("aria-disabled",n.disabled&&n.disabledInteractive?"true":null),G(8),yA("matRippleTrigger",o)("matRippleDisabled",n.disableRipple||n.disabled)("matRippleCentered",!0),G(),GA(n.hideIcon?-1:10),G(),yA("for",n.buttonId),Ne("id",n._labelId)}},dependencies:[rs,MB],styles:['.mdc-switch{align-items:center;background:none;border:none;cursor:pointer;display:inline-flex;flex-shrink:0;margin:0;outline:none;overflow:visible;padding:0;position:relative;width:var(--mdc-switch-track-width, 52px)}.mdc-switch.mdc-switch--disabled{cursor:default;pointer-events:none}.mdc-switch.mat-mdc-slide-toggle-disabled-interactive{pointer-events:auto}.mdc-switch__track{overflow:hidden;position:relative;width:100%;height:var(--mdc-switch-track-height, 32px);border-radius:var(--mdc-switch-track-shape, var(--mat-sys-corner-full))}.mdc-switch--disabled.mdc-switch .mdc-switch__track{opacity:var(--mdc-switch-disabled-track-opacity, 0.12)}.mdc-switch__track::before,.mdc-switch__track::after{border:1px solid rgba(0,0,0,0);border-radius:inherit;box-sizing:border-box;content:"";height:100%;left:0;position:absolute;width:100%;border-width:var(--mat-switch-track-outline-width, 2px);border-color:var(--mat-switch-track-outline-color, var(--mat-sys-outline))}.mdc-switch--selected .mdc-switch__track::before,.mdc-switch--selected .mdc-switch__track::after{border-width:var(--mat-switch-selected-track-outline-width, 2px);border-color:var(--mat-switch-selected-track-outline-color, transparent)}.mdc-switch--disabled .mdc-switch__track::before,.mdc-switch--disabled .mdc-switch__track::after{border-width:var(--mat-switch-disabled-unselected-track-outline-width, 2px);border-color:var(--mat-switch-disabled-unselected-track-outline-color, var(--mat-sys-on-surface))}@media(forced-colors: active){.mdc-switch__track{border-color:currentColor}}.mdc-switch__track::before{transition:transform 75ms 0ms cubic-bezier(0, 0, 0.2, 1);transform:translateX(0);background:var(--mdc-switch-unselected-track-color, var(--mat-sys-surface-variant))}.mdc-switch--selected .mdc-switch__track::before{transition:transform 75ms 0ms cubic-bezier(0.4, 0, 0.6, 1);transform:translateX(100%)}[dir=rtl] .mdc-switch--selected .mdc-switch--selected .mdc-switch__track::before{transform:translateX(-100%)}.mdc-switch--selected .mdc-switch__track::before{opacity:var(--mat-switch-hidden-track-opacity, 0);transition:var(--mat-switch-hidden-track-transition, opacity 75ms)}.mdc-switch--unselected .mdc-switch__track::before{opacity:var(--mat-switch-visible-track-opacity, 1);transition:var(--mat-switch-visible-track-transition, opacity 75ms)}.mdc-switch:enabled:hover:not(:focus):not(:active) .mdc-switch__track::before{background:var(--mdc-switch-unselected-hover-track-color, var(--mat-sys-surface-variant))}.mdc-switch:enabled:focus:not(:active) .mdc-switch__track::before{background:var(--mdc-switch-unselected-focus-track-color, var(--mat-sys-surface-variant))}.mdc-switch:enabled:active .mdc-switch__track::before{background:var(--mdc-switch-unselected-pressed-track-color, var(--mat-sys-surface-variant))}.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:hover:not(:focus):not(:active) .mdc-switch__track::before,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:focus:not(:active) .mdc-switch__track::before,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:active .mdc-switch__track::before,.mdc-switch.mdc-switch--disabled .mdc-switch__track::before{background:var(--mdc-switch-disabled-unselected-track-color, var(--mat-sys-surface-variant))}.mdc-switch__track::after{transform:translateX(-100%);background:var(--mdc-switch-selected-track-color, var(--mat-sys-primary))}[dir=rtl] .mdc-switch__track::after{transform:translateX(100%)}.mdc-switch--selected .mdc-switch__track::after{transform:translateX(0)}.mdc-switch--selected .mdc-switch__track::after{opacity:var(--mat-switch-visible-track-opacity, 1);transition:var(--mat-switch-visible-track-transition, opacity 75ms)}.mdc-switch--unselected .mdc-switch__track::after{opacity:var(--mat-switch-hidden-track-opacity, 0);transition:var(--mat-switch-hidden-track-transition, opacity 75ms)}.mdc-switch:enabled:hover:not(:focus):not(:active) .mdc-switch__track::after{background:var(--mdc-switch-selected-hover-track-color, var(--mat-sys-primary))}.mdc-switch:enabled:focus:not(:active) .mdc-switch__track::after{background:var(--mdc-switch-selected-focus-track-color, var(--mat-sys-primary))}.mdc-switch:enabled:active .mdc-switch__track::after{background:var(--mdc-switch-selected-pressed-track-color, var(--mat-sys-primary))}.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:hover:not(:focus):not(:active) .mdc-switch__track::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:focus:not(:active) .mdc-switch__track::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:active .mdc-switch__track::after,.mdc-switch.mdc-switch--disabled .mdc-switch__track::after{background:var(--mdc-switch-disabled-selected-track-color, var(--mat-sys-on-surface))}.mdc-switch__handle-track{height:100%;pointer-events:none;position:absolute;top:0;transition:transform 75ms 0ms cubic-bezier(0.4, 0, 0.2, 1);left:0;right:auto;transform:translateX(0);width:calc(100% - var(--mdc-switch-handle-width))}[dir=rtl] .mdc-switch__handle-track{left:auto;right:0}.mdc-switch--selected .mdc-switch__handle-track{transform:translateX(100%)}[dir=rtl] .mdc-switch--selected .mdc-switch__handle-track{transform:translateX(-100%)}.mdc-switch__handle{display:flex;pointer-events:auto;position:absolute;top:50%;transform:translateY(-50%);left:0;right:auto;transition:width 75ms cubic-bezier(0.4, 0, 0.2, 1),height 75ms cubic-bezier(0.4, 0, 0.2, 1),margin 75ms cubic-bezier(0.4, 0, 0.2, 1);width:var(--mdc-switch-handle-width);height:var(--mdc-switch-handle-height);border-radius:var(--mdc-switch-handle-shape, var(--mat-sys-corner-full))}[dir=rtl] .mdc-switch__handle{left:auto;right:0}.mat-mdc-slide-toggle .mdc-switch--unselected .mdc-switch__handle{width:var(--mat-switch-unselected-handle-size, 16px);height:var(--mat-switch-unselected-handle-size, 16px);margin:var(--mat-switch-unselected-handle-horizontal-margin, 0 8px)}.mat-mdc-slide-toggle .mdc-switch--unselected .mdc-switch__handle:has(.mdc-switch__icons){margin:var(--mat-switch-unselected-with-icon-handle-horizontal-margin, 0 4px)}.mat-mdc-slide-toggle .mdc-switch--selected .mdc-switch__handle{width:var(--mat-switch-selected-handle-size, 24px);height:var(--mat-switch-selected-handle-size, 24px);margin:var(--mat-switch-selected-handle-horizontal-margin, 0 24px)}.mat-mdc-slide-toggle .mdc-switch--selected .mdc-switch__handle:has(.mdc-switch__icons){margin:var(--mat-switch-selected-with-icon-handle-horizontal-margin, 0 24px)}.mat-mdc-slide-toggle .mdc-switch__handle:has(.mdc-switch__icons){width:var(--mat-switch-with-icon-handle-size, 24px);height:var(--mat-switch-with-icon-handle-size, 24px)}.mat-mdc-slide-toggle .mdc-switch:active:not(.mdc-switch--disabled) .mdc-switch__handle{width:var(--mat-switch-pressed-handle-size, 28px);height:var(--mat-switch-pressed-handle-size, 28px)}.mat-mdc-slide-toggle .mdc-switch--selected:active:not(.mdc-switch--disabled) .mdc-switch__handle{margin:var(--mat-switch-selected-pressed-handle-horizontal-margin, 0 22px)}.mat-mdc-slide-toggle .mdc-switch--unselected:active:not(.mdc-switch--disabled) .mdc-switch__handle{margin:var(--mat-switch-unselected-pressed-handle-horizontal-margin, 0 2px)}.mdc-switch--disabled.mdc-switch--selected .mdc-switch__handle::after{opacity:var(--mat-switch-disabled-selected-handle-opacity, 1)}.mdc-switch--disabled.mdc-switch--unselected .mdc-switch__handle::after{opacity:var(--mat-switch-disabled-unselected-handle-opacity, 0.38)}.mdc-switch__handle::before,.mdc-switch__handle::after{border:1px solid rgba(0,0,0,0);border-radius:inherit;box-sizing:border-box;content:"";width:100%;height:100%;left:0;position:absolute;top:0;transition:background-color 75ms 0ms cubic-bezier(0.4, 0, 0.2, 1),border-color 75ms 0ms cubic-bezier(0.4, 0, 0.2, 1);z-index:-1}@media(forced-colors: active){.mdc-switch__handle::before,.mdc-switch__handle::after{border-color:currentColor}}.mdc-switch--selected:enabled .mdc-switch__handle::after{background:var(--mdc-switch-selected-handle-color, var(--mat-sys-on-primary))}.mdc-switch--selected:enabled:hover:not(:focus):not(:active) .mdc-switch__handle::after{background:var(--mdc-switch-selected-hover-handle-color, var(--mat-sys-primary-container))}.mdc-switch--selected:enabled:focus:not(:active) .mdc-switch__handle::after{background:var(--mdc-switch-selected-focus-handle-color, var(--mat-sys-primary-container))}.mdc-switch--selected:enabled:active .mdc-switch__handle::after{background:var(--mdc-switch-selected-pressed-handle-color, var(--mat-sys-primary-container))}.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled.mdc-switch--selected:hover:not(:focus):not(:active) .mdc-switch__handle::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled.mdc-switch--selected:focus:not(:active) .mdc-switch__handle::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled.mdc-switch--selected:active .mdc-switch__handle::after,.mdc-switch--selected.mdc-switch--disabled .mdc-switch__handle::after{background:var(--mdc-switch-disabled-selected-handle-color, var(--mat-sys-surface))}.mdc-switch--unselected:enabled .mdc-switch__handle::after{background:var(--mdc-switch-unselected-handle-color, var(--mat-sys-outline))}.mdc-switch--unselected:enabled:hover:not(:focus):not(:active) .mdc-switch__handle::after{background:var(--mdc-switch-unselected-hover-handle-color, var(--mat-sys-on-surface-variant))}.mdc-switch--unselected:enabled:focus:not(:active) .mdc-switch__handle::after{background:var(--mdc-switch-unselected-focus-handle-color, var(--mat-sys-on-surface-variant))}.mdc-switch--unselected:enabled:active .mdc-switch__handle::after{background:var(--mdc-switch-unselected-pressed-handle-color, var(--mat-sys-on-surface-variant))}.mdc-switch--unselected.mdc-switch--disabled .mdc-switch__handle::after{background:var(--mdc-switch-disabled-unselected-handle-color, var(--mat-sys-on-surface))}.mdc-switch__handle::before{background:var(--mdc-switch-handle-surface-color)}.mdc-switch__shadow{border-radius:inherit;bottom:0;left:0;position:absolute;right:0;top:0}.mdc-switch:enabled .mdc-switch__shadow{box-shadow:var(--mdc-switch-handle-elevation-shadow)}.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:hover:not(:focus):not(:active) .mdc-switch__shadow,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:focus:not(:active) .mdc-switch__shadow,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:active .mdc-switch__shadow,.mdc-switch.mdc-switch--disabled .mdc-switch__shadow{box-shadow:var(--mdc-switch-disabled-handle-elevation-shadow)}.mdc-switch__ripple{left:50%;position:absolute;top:50%;transform:translate(-50%, -50%);z-index:-1;width:var(--mdc-switch-state-layer-size, 40px);height:var(--mdc-switch-state-layer-size, 40px)}.mdc-switch__ripple::after{content:"";opacity:0}.mdc-switch--disabled .mdc-switch__ripple::after{display:none}.mat-mdc-slide-toggle-disabled-interactive .mdc-switch__ripple::after{display:block}.mdc-switch:hover .mdc-switch__ripple::after{opacity:.04;transition:75ms opacity cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-slide-toggle.mat-mdc-slide-toggle-focused .mdc-switch .mdc-switch__ripple::after{opacity:.12}.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:enabled:focus .mdc-switch__ripple::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:enabled:active .mdc-switch__ripple::after,.mat-mdc-slide-toggle-disabled-interactive.mdc-switch--disabled:enabled:hover:not(:focus) .mdc-switch__ripple::after,.mdc-switch--unselected:enabled:hover:not(:focus) .mdc-switch__ripple::after{background:var(--mdc-switch-unselected-hover-state-layer-color, var(--mat-sys-on-surface))}.mdc-switch--unselected:enabled:focus .mdc-switch__ripple::after{background:var(--mdc-switch-unselected-focus-state-layer-color, var(--mat-sys-on-surface))}.mdc-switch--unselected:enabled:active .mdc-switch__ripple::after{background:var(--mdc-switch-unselected-pressed-state-layer-color, var(--mat-sys-on-surface));opacity:var(--mdc-switch-unselected-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity));transition:opacity 75ms linear}.mdc-switch--selected:enabled:hover:not(:focus) .mdc-switch__ripple::after{background:var(--mdc-switch-selected-hover-state-layer-color, var(--mat-sys-primary))}.mdc-switch--selected:enabled:focus .mdc-switch__ripple::after{background:var(--mdc-switch-selected-focus-state-layer-color, var(--mat-sys-primary))}.mdc-switch--selected:enabled:active .mdc-switch__ripple::after{background:var(--mdc-switch-selected-pressed-state-layer-color, var(--mat-sys-primary));opacity:var(--mdc-switch-selected-pressed-state-layer-opacity, var(--mat-sys-pressed-state-layer-opacity));transition:opacity 75ms linear}.mdc-switch__icons{position:relative;height:100%;width:100%;z-index:1}.mdc-switch--disabled.mdc-switch--unselected .mdc-switch__icons{opacity:var(--mdc-switch-disabled-unselected-icon-opacity, 0.38)}.mdc-switch--disabled.mdc-switch--selected .mdc-switch__icons{opacity:var(--mdc-switch-disabled-selected-icon-opacity, 0.38)}.mdc-switch__icon{bottom:0;left:0;margin:auto;position:absolute;right:0;top:0;opacity:0;transition:opacity 30ms 0ms cubic-bezier(0.4, 0, 1, 1)}.mdc-switch--unselected .mdc-switch__icon{width:var(--mdc-switch-unselected-icon-size, 16px);height:var(--mdc-switch-unselected-icon-size, 16px);fill:var(--mdc-switch-unselected-icon-color, var(--mat-sys-surface-variant))}.mdc-switch--unselected.mdc-switch--disabled .mdc-switch__icon{fill:var(--mdc-switch-disabled-unselected-icon-color, var(--mat-sys-surface-variant))}.mdc-switch--selected .mdc-switch__icon{width:var(--mdc-switch-selected-icon-size, 16px);height:var(--mdc-switch-selected-icon-size, 16px);fill:var(--mdc-switch-selected-icon-color, var(--mat-sys-on-primary-container))}.mdc-switch--selected.mdc-switch--disabled .mdc-switch__icon{fill:var(--mdc-switch-disabled-selected-icon-color, var(--mat-sys-on-surface))}.mdc-switch--selected .mdc-switch__icon--on,.mdc-switch--unselected .mdc-switch__icon--off{opacity:1;transition:opacity 45ms 30ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-slide-toggle{-webkit-user-select:none;user-select:none;display:inline-block;-webkit-tap-highlight-color:rgba(0,0,0,0);outline:0}.mat-mdc-slide-toggle .mat-mdc-slide-toggle-ripple,.mat-mdc-slide-toggle .mdc-switch__ripple::after{top:0;left:0;right:0;bottom:0;position:absolute;border-radius:50%;pointer-events:none}.mat-mdc-slide-toggle .mat-mdc-slide-toggle-ripple:not(:empty),.mat-mdc-slide-toggle .mdc-switch__ripple::after:not(:empty){transform:translateZ(0)}.mat-mdc-slide-toggle.mat-mdc-slide-toggle-focused .mat-focus-indicator::before{content:""}.mat-mdc-slide-toggle .mat-internal-form-field{color:var(--mat-switch-label-text-color, var(--mat-sys-on-surface));font-family:var(--mat-switch-label-text-font, var(--mat-sys-body-medium-font));line-height:var(--mat-switch-label-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mat-switch-label-text-size, var(--mat-sys-body-medium-size));letter-spacing:var(--mat-switch-label-text-tracking, var(--mat-sys-body-medium-tracking));font-weight:var(--mat-switch-label-text-weight, var(--mat-sys-body-medium-weight))}.mat-mdc-slide-toggle .mat-ripple-element{opacity:.12}.mat-mdc-slide-toggle .mat-focus-indicator::before{border-radius:50%}.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__handle-track,.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__icon,.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__handle::before,.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__handle::after,.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__track::before,.mat-mdc-slide-toggle._mat-animation-noopable .mdc-switch__track::after{transition:none}.mat-mdc-slide-toggle .mdc-switch:enabled+.mdc-label{cursor:pointer}.mat-mdc-slide-toggle .mdc-switch--disabled+label{color:var(--mdc-switch-disabled-label-text-color)}'],encapsulation:2,changeDetection:0})}return t})();var ElA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[C7,it,it]})}return t})();var jPA=["mat-menu-item",""],qPA=[[["mat-icon"],["","matMenuItemIcon",""]],"*"],VPA=["mat-icon, [matMenuItemIcon]","*"];function ZPA(t,e){t&1&&(ar(),S(0,"svg",2),JA(1,"polygon",3),F())}var WPA=["*"];function XPA(t,e){if(t&1){let A=be();S(0,"div",0),hA("click",function(){RA(A);let n=O();return xA(n.closed.emit("click"))})("animationstart",function(n){RA(A);let o=O();return xA(o._onAnimationStart(n.animationName))})("animationend",function(n){RA(A);let o=O();return xA(o._onAnimationDone(n.animationName))})("animationcancel",function(n){RA(A);let o=O();return xA(o._onAnimationDone(n.animationName))}),S(1,"div",1),Fe(2),F()()}if(t&2){let A=O();fo(A._classList),ue("mat-menu-panel-animations-disabled",A._animationsDisabled)("mat-menu-panel-exit-animation",A._panelAnimationState==="void")("mat-menu-panel-animating",A._isAnimating),yA("id",A.panelId),Ne("aria-label",A.ariaLabel||null)("aria-labelledby",A.ariaLabelledby||null)("aria-describedby",A.ariaDescribedby||null)}}var aU=new dA("MAT_MENU_PANEL"),bf=(()=>{class t{_elementRef=f(ee);_document=f(at);_focusMonitor=f(dr);_parentMenu=f(aU,{optional:!0});_changeDetectorRef=f(It);role="menuitem";disabled=!1;disableRipple=!1;_hovered=new OA;_focused=new OA;_highlighted=!1;_triggersSubmenu=!1;constructor(){f(Rn).load(lr),this._parentMenu?.addItem?.(this)}focus(A,i){this._focusMonitor&&A?this._focusMonitor.focusVia(this._getHostElement(),A,i):this._getHostElement().focus(i),this._focused.next(this)}ngAfterViewInit(){this._focusMonitor&&this._focusMonitor.monitor(this._elementRef,!1)}ngOnDestroy(){this._focusMonitor&&this._focusMonitor.stopMonitoring(this._elementRef),this._parentMenu&&this._parentMenu.removeItem&&this._parentMenu.removeItem(this),this._hovered.complete(),this._focused.complete()}_getTabIndex(){return this.disabled?"-1":"0"}_getHostElement(){return this._elementRef.nativeElement}_checkDisabled(A){this.disabled&&(A.preventDefault(),A.stopPropagation())}_handleMouseEnter(){this._hovered.next(this)}getLabel(){let A=this._elementRef.nativeElement.cloneNode(!0),i=A.querySelectorAll("mat-icon, .material-icons");for(let n=0;n<i.length;n++)i[n].remove();return A.textContent?.trim()||""}_setHighlighted(A){this._highlighted=A,this._changeDetectorRef.markForCheck()}_setTriggersSubmenu(A){this._triggersSubmenu=A,this._changeDetectorRef.markForCheck()}_hasFocus(){return this._document&&this._document.activeElement===this._getHostElement()}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["","mat-menu-item",""]],hostAttrs:[1,"mat-mdc-menu-item","mat-focus-indicator"],hostVars:8,hostBindings:function(i,n){i&1&&hA("click",function(r){return n._checkDisabled(r)})("mouseenter",function(){return n._handleMouseEnter()}),i&2&&(Ne("role",n.role)("tabindex",n._getTabIndex())("aria-disabled",n.disabled)("disabled",n.disabled||null),ue("mat-mdc-menu-item-highlighted",n._highlighted)("mat-mdc-menu-item-submenu-trigger",n._triggersSubmenu))},inputs:{role:"role",disabled:[2,"disabled","disabled",ie],disableRipple:[2,"disableRipple","disableRipple",ie]},exportAs:["matMenuItem"],attrs:jPA,ngContentSelectors:VPA,decls:5,vars:3,consts:[[1,"mat-mdc-menu-item-text"],["matRipple","",1,"mat-mdc-menu-ripple",3,"matRippleDisabled","matRippleTrigger"],["viewBox","0 0 5 10","focusable","false","aria-hidden","true",1,"mat-mdc-menu-submenu-icon"],["points","0,0 5,5 0,10"]],template:function(i,n){i&1&&(jt(qPA),Fe(0),S(1,"span",0),Fe(2,1),F(),JA(3,"div",1),_A(4,ZPA,2,0,":svg:svg",2)),i&2&&(G(3),yA("matRippleDisabled",n.disableRipple||n.disabled)("matRippleTrigger",n._getHostElement()),G(),GA(n._triggersSubmenu?4:-1))},dependencies:[rs],encapsulation:2,changeDetection:0})}return t})();var $PA=new dA("MatMenuContent");var AjA=new dA("mat-menu-default-options",{providedIn:"root",factory:ejA});function ejA(){return{overlapTrigger:!1,xPosition:"after",yPosition:"below",backdropClass:"cdk-overlay-transparent-backdrop"}}var sU="_mat-menu-enter",d7="_mat-menu-exit",NQ=(()=>{class t{_elementRef=f(ee);_changeDetectorRef=f(It);_injector=f(Rt);_keyManager;_xPosition;_yPosition;_firstItemFocusRef;_exitFallbackTimeout;_animationsDisabled;_allItems;_directDescendantItems=new ca;_classList={};_panelAnimationState="void";_animationDone=new OA;_isAnimating=!1;parentMenu;direction;overlayPanelClass;backdropClass;ariaLabel;ariaLabelledby;ariaDescribedby;get xPosition(){return this._xPosition}set xPosition(A){this._xPosition=A,this.setPositionClasses()}get yPosition(){return this._yPosition}set yPosition(A){this._yPosition=A,this.setPositionClasses()}templateRef;items;lazyContent;overlapTrigger;hasBackdrop;set panelClass(A){let i=this._previousPanelClass,n=rA({},this._classList);i&&i.length&&i.split(" ").forEach(o=>{n[o]=!1}),this._previousPanelClass=A,A&&A.length&&(A.split(" ").forEach(o=>{n[o]=!0}),this._elementRef.nativeElement.className=""),this._classList=n}_previousPanelClass;get classList(){return this.panelClass}set classList(A){this.panelClass=A}closed=new WA;close=this.closed;panelId=f(sn).getId("mat-menu-panel-");constructor(){let A=f(AjA);this.overlayPanelClass=A.overlayPanelClass||"",this._xPosition=A.xPosition,this._yPosition=A.yPosition,this.backdropClass=A.backdropClass,this.overlapTrigger=A.overlapTrigger,this.hasBackdrop=A.hasBackdrop,this._animationsDisabled=f(Si,{optional:!0})==="NoopAnimations"}ngOnInit(){this.setPositionClasses()}ngAfterContentInit(){this._updateDirectDescendants(),this._keyManager=new qI(this._directDescendantItems).withWrap().withTypeAhead().withHomeAndEnd(),this._keyManager.tabOut.subscribe(()=>this.closed.emit("tab")),this._directDescendantItems.changes.pipe(Pn(this._directDescendantItems),jn(A=>zn(...A.map(i=>i._focused)))).subscribe(A=>this._keyManager.updateActiveItem(A)),this._directDescendantItems.changes.subscribe(A=>{let i=this._keyManager;if(this._panelAnimationState==="enter"&&i.activeItem?._hasFocus()){let n=A.toArray(),o=Math.max(0,Math.min(n.length-1,i.activeItemIndex||0));n[o]&&!n[o].disabled?i.setActiveItem(o):i.setNextItemActive()}})}ngOnDestroy(){this._keyManager?.destroy(),this._directDescendantItems.destroy(),this.closed.complete(),this._firstItemFocusRef?.destroy(),clearTimeout(this._exitFallbackTimeout)}_hovered(){return this._directDescendantItems.changes.pipe(Pn(this._directDescendantItems),jn(i=>zn(...i.map(n=>n._hovered))))}addItem(A){}removeItem(A){}_handleKeydown(A){let i=A.keyCode,n=this._keyManager;switch(i){case 27:ir(A)||(A.preventDefault(),this.closed.emit("keydown"));break;case 37:this.parentMenu&&this.direction==="ltr"&&this.closed.emit("keydown");break;case 39:this.parentMenu&&this.direction==="rtl"&&this.closed.emit("keydown");break;default:(i===38||i===40)&&n.setFocusOrigin("keyboard"),n.onKeydown(A);return}}focusFirstItem(A="program"){this._firstItemFocusRef?.destroy(),this._firstItemFocusRef=To(()=>{let i=this._resolvePanel();if(!i||!i.contains(document.activeElement)){let n=this._keyManager;n.setFocusOrigin(A).setFirstItemActive(),!n.activeItem&&i&&i.focus()}},{injector:this._injector})}resetActiveItem(){this._keyManager.setActiveItem(-1)}setElevation(A){}setPositionClasses(A=this.xPosition,i=this.yPosition){this._classList=Ye(rA({},this._classList),{"mat-menu-before":A==="before","mat-menu-after":A==="after","mat-menu-above":i==="above","mat-menu-below":i==="below"}),this._changeDetectorRef.markForCheck()}_onAnimationDone(A){let i=A===d7;(i||A===sU)&&(i&&(clearTimeout(this._exitFallbackTimeout),this._exitFallbackTimeout=void 0),this._animationDone.next(i?"void":"enter"),this._isAnimating=!1)}_onAnimationStart(A){(A===sU||A===d7)&&(this._isAnimating=!0)}_setIsOpen(A){if(this._panelAnimationState=A?"enter":"void",A){if(this._keyManager.activeItemIndex===0){let i=this._resolvePanel();i&&(i.scrollTop=0)}}else this._animationsDisabled||(this._exitFallbackTimeout=setTimeout(()=>this._onAnimationDone(d7),200));this._animationsDisabled&&setTimeout(()=>{this._onAnimationDone(A?sU:d7)}),this._changeDetectorRef.markForCheck()}_updateDirectDescendants(){this._allItems.changes.pipe(Pn(this._allItems)).subscribe(A=>{this._directDescendantItems.reset(A.filter(i=>i._parentMenu===this)),this._directDescendantItems.notifyOnChanges()})}_resolvePanel(){let A=null;return this._directDescendantItems.length&&(A=this._directDescendantItems.first._getHostElement().closest('[role="menu"]')),A}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-menu"]],contentQueries:function(i,n,o){if(i&1&&(ci(o,$PA,5),ci(o,bf,5),ci(o,bf,4)),i&2){let r;XA(r=$A())&&(n.lazyContent=r.first),XA(r=$A())&&(n._allItems=r),XA(r=$A())&&(n.items=r)}},viewQuery:function(i,n){if(i&1&&Te(wn,5),i&2){let o;XA(o=$A())&&(n.templateRef=o.first)}},hostVars:3,hostBindings:function(i,n){i&2&&Ne("aria-label",null)("aria-labelledby",null)("aria-describedby",null)},inputs:{backdropClass:"backdropClass",ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],ariaDescribedby:[0,"aria-describedby","ariaDescribedby"],xPosition:"xPosition",yPosition:"yPosition",overlapTrigger:[2,"overlapTrigger","overlapTrigger",ie],hasBackdrop:[2,"hasBackdrop","hasBackdrop",A=>A==null?null:ie(A)],panelClass:[0,"class","panelClass"],classList:"classList"},outputs:{closed:"closed",close:"close"},exportAs:["matMenu"],features:[ht([{provide:aU,useExisting:t}])],ngContentSelectors:WPA,decls:1,vars:0,consts:[["tabindex","-1","role","menu",1,"mat-mdc-menu-panel",3,"click","animationstart","animationend","animationcancel","id"],[1,"mat-mdc-menu-content"]],template:function(i,n){i&1&&(jt(),_A(0,XPA,3,12,"ng-template"))},styles:['mat-menu{display:none}.mat-mdc-menu-content{margin:0;padding:8px 0;outline:0}.mat-mdc-menu-content,.mat-mdc-menu-content .mat-mdc-menu-item .mat-mdc-menu-item-text{-moz-osx-font-smoothing:grayscale;-webkit-font-smoothing:antialiased;flex:1;white-space:normal;font-family:var(--mat-menu-item-label-text-font, var(--mat-sys-label-large-font));line-height:var(--mat-menu-item-label-text-line-height, var(--mat-sys-label-large-line-height));font-size:var(--mat-menu-item-label-text-size, var(--mat-sys-label-large-size));letter-spacing:var(--mat-menu-item-label-text-tracking, var(--mat-sys-label-large-tracking));font-weight:var(--mat-menu-item-label-text-weight, var(--mat-sys-label-large-weight))}@keyframes _mat-menu-enter{from{opacity:0;transform:scale(0.8)}to{opacity:1;transform:none}}@keyframes _mat-menu-exit{from{opacity:1}to{opacity:0}}.mat-mdc-menu-panel{min-width:112px;max-width:280px;overflow:auto;box-sizing:border-box;outline:0;animation:_mat-menu-enter 120ms cubic-bezier(0, 0, 0.2, 1);border-radius:var(--mat-menu-container-shape, var(--mat-sys-corner-extra-small));background-color:var(--mat-menu-container-color, var(--mat-sys-surface-container));box-shadow:var(--mat-menu-container-elevation-shadow, 0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12));will-change:transform,opacity}.mat-mdc-menu-panel.mat-menu-panel-exit-animation{animation:_mat-menu-exit 100ms 25ms linear forwards}.mat-mdc-menu-panel.mat-menu-panel-animations-disabled{animation:none}.mat-mdc-menu-panel.mat-menu-panel-animating{pointer-events:none}.mat-mdc-menu-panel.mat-menu-panel-animating:has(.mat-mdc-menu-content:empty){display:none}@media(forced-colors: active){.mat-mdc-menu-panel{outline:solid 1px}}.mat-mdc-menu-panel .mat-divider{color:var(--mat-menu-divider-color, var(--mat-sys-surface-variant));margin-bottom:var(--mat-menu-divider-bottom-spacing, 8px);margin-top:var(--mat-menu-divider-top-spacing, 8px)}.mat-mdc-menu-item{display:flex;position:relative;align-items:center;justify-content:flex-start;overflow:hidden;padding:0;cursor:pointer;width:100%;text-align:left;box-sizing:border-box;color:inherit;font-size:inherit;background:none;text-decoration:none;margin:0;min-height:48px;padding-left:var(--mat-menu-item-leading-spacing, 12px);padding-right:var(--mat-menu-item-trailing-spacing, 12px);-webkit-user-select:none;user-select:none;cursor:pointer;outline:none;border:none;-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-menu-item::-moz-focus-inner{border:0}[dir=rtl] .mat-mdc-menu-item{padding-left:var(--mat-menu-item-trailing-spacing, 12px);padding-right:var(--mat-menu-item-leading-spacing, 12px)}.mat-mdc-menu-item:has(.material-icons,mat-icon,[matButtonIcon]){padding-left:var(--mat-menu-item-with-icon-leading-spacing, 12px);padding-right:var(--mat-menu-item-with-icon-trailing-spacing, 12px)}[dir=rtl] .mat-mdc-menu-item:has(.material-icons,mat-icon,[matButtonIcon]){padding-left:var(--mat-menu-item-with-icon-trailing-spacing, 12px);padding-right:var(--mat-menu-item-with-icon-leading-spacing, 12px)}.mat-mdc-menu-item,.mat-mdc-menu-item:visited,.mat-mdc-menu-item:link{color:var(--mat-menu-item-label-text-color, var(--mat-sys-on-surface))}.mat-mdc-menu-item .mat-icon-no-color,.mat-mdc-menu-item .mat-mdc-menu-submenu-icon{color:var(--mat-menu-item-icon-color, var(--mat-sys-on-surface-variant))}.mat-mdc-menu-item[disabled]{cursor:default;opacity:.38}.mat-mdc-menu-item[disabled]::after{display:block;position:absolute;content:"";top:0;left:0;bottom:0;right:0}.mat-mdc-menu-item:focus{outline:0}.mat-mdc-menu-item .mat-icon{flex-shrink:0;margin-right:var(--mat-menu-item-spacing, 12px);height:var(--mat-menu-item-icon-size, 24px);width:var(--mat-menu-item-icon-size, 24px)}[dir=rtl] .mat-mdc-menu-item{text-align:right}[dir=rtl] .mat-mdc-menu-item .mat-icon{margin-right:0;margin-left:var(--mat-menu-item-spacing, 12px)}.mat-mdc-menu-item:not([disabled]):hover{background-color:var(--mat-menu-item-hover-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-hover-state-layer-opacity) * 100%), transparent))}.mat-mdc-menu-item:not([disabled]).cdk-program-focused,.mat-mdc-menu-item:not([disabled]).cdk-keyboard-focused,.mat-mdc-menu-item:not([disabled]).mat-mdc-menu-item-highlighted{background-color:var(--mat-menu-item-focus-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-focus-state-layer-opacity) * 100%), transparent))}@media(forced-colors: active){.mat-mdc-menu-item{margin-top:1px}}.mat-mdc-menu-submenu-icon{width:var(--mat-menu-item-icon-size, 24px);height:10px;fill:currentColor;padding-left:var(--mat-menu-item-spacing, 12px)}[dir=rtl] .mat-mdc-menu-submenu-icon{padding-right:var(--mat-menu-item-spacing, 12px);padding-left:0}[dir=rtl] .mat-mdc-menu-submenu-icon polygon{transform:scaleX(-1);transform-origin:center}@media(forced-colors: active){.mat-mdc-menu-submenu-icon{fill:CanvasText}}.mat-mdc-menu-item .mat-mdc-menu-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none}'],encapsulation:2,changeDetection:0})}return t})(),hlA=new dA("mat-menu-scroll-strategy",{providedIn:"root",factory:()=>{let t=f(nr);return()=>t.scrollStrategies.reposition()}});function tjA(t){return()=>t.scrollStrategies.reposition()}var ijA={provide:hlA,deps:[nr],useFactory:tjA},QlA=bc({passive:!0});var vf=new WeakMap,ulA=(()=>{class t{_overlay=f(nr);_element=f(ee);_viewContainerRef=f(Nn);_menuItemInstance=f(bf,{optional:!0,self:!0});_dir=f(mo,{optional:!0});_focusMonitor=f(dr);_ngZone=f(Qe);_scrollStrategy=f(hlA);_changeDetectorRef=f(It);_portal;_overlayRef=null;_menuOpen=!1;_closingActionsSubscription=Kt.EMPTY;_hoverSubscription=Kt.EMPTY;_menuCloseSubscription=Kt.EMPTY;_pendingRemoval;_parentMaterialMenu;_parentInnerPadding;_handleTouchStart=A=>{Su(A)||(this._openedBy="touch")};_openedBy=void 0;get _deprecatedMatMenuTriggerFor(){return this.menu}set _deprecatedMatMenuTriggerFor(A){this.menu=A}get menu(){return this._menu}set menu(A){A!==this._menu&&(this._menu=A,this._menuCloseSubscription.unsubscribe(),A&&(this._parentMaterialMenu,this._menuCloseSubscription=A.close.subscribe(i=>{this._destroyMenu(i),(i==="click"||i==="tab")&&this._parentMaterialMenu&&this._parentMaterialMenu.closed.emit(i)})),this._menuItemInstance?._setTriggersSubmenu(this.triggersSubmenu()))}_menu;menuData;restoreFocus=!0;menuOpened=new WA;onMenuOpen=this.menuOpened;menuClosed=new WA;onMenuClose=this.menuClosed;constructor(){let A=f(aU,{optional:!0});this._parentMaterialMenu=A instanceof NQ?A:void 0,this._element.nativeElement.addEventListener("touchstart",this._handleTouchStart,QlA)}ngAfterContentInit(){this._handleHover()}ngOnDestroy(){this.menu&&this._ownsMenu(this.menu)&&vf.delete(this.menu),this._element.nativeElement.removeEventListener("touchstart",this._handleTouchStart,QlA),this._pendingRemoval?.unsubscribe(),this._menuCloseSubscription.unsubscribe(),this._closingActionsSubscription.unsubscribe(),this._hoverSubscription.unsubscribe(),this._overlayRef&&(this._overlayRef.dispose(),this._overlayRef=null)}get menuOpen(){return this._menuOpen}get dir(){return this._dir&&this._dir.value==="rtl"?"rtl":"ltr"}triggersSubmenu(){return!!(this._menuItemInstance&&this._parentMaterialMenu&&this.menu)}toggleMenu(){return this._menuOpen?this.closeMenu():this.openMenu()}openMenu(){let A=this.menu;if(this._menuOpen||!A)return;this._pendingRemoval?.unsubscribe();let i=vf.get(A);vf.set(A,this),i&&i!==this&&i.closeMenu();let n=this._createOverlay(A),o=n.getConfig(),r=o.positionStrategy;this._setPosition(A,r),o.hasBackdrop=A.hasBackdrop==null?!this.triggersSubmenu():A.hasBackdrop,n.hasAttached()||(n.attach(this._getPortal(A)),A.lazyContent?.attach(this.menuData)),this._closingActionsSubscription=this._menuClosingActions().subscribe(()=>this.closeMenu()),A.parentMenu=this.triggersSubmenu()?this._parentMaterialMenu:void 0,A.direction=this.dir,A.focusFirstItem(this._openedBy||"program"),this._setIsMenuOpen(!0),A instanceof NQ&&(A._setIsOpen(!0),A._directDescendantItems.changes.pipe(St(A.close)).subscribe(()=>{r.withLockedPosition(!1).reapplyLastPosition(),r.withLockedPosition(!0)}))}closeMenu(){this.menu?.close.emit()}focus(A,i){this._focusMonitor&&A?this._focusMonitor.focusVia(this._element,A,i):this._element.nativeElement.focus(i)}updatePosition(){this._overlayRef?.updatePosition()}_destroyMenu(A){let i=this._overlayRef,n=this._menu;!i||!this.menuOpen||(this._closingActionsSubscription.unsubscribe(),this._pendingRemoval?.unsubscribe(),n instanceof NQ&&this._ownsMenu(n)?(this._pendingRemoval=n._animationDone.pipe(On(1)).subscribe(()=>{i.detach(),n.lazyContent?.detach()}),n._setIsOpen(!1)):(i.detach(),n?.lazyContent?.detach()),n&&this._ownsMenu(n)&&vf.delete(n),this.restoreFocus&&(A==="keydown"||!this._openedBy||!this.triggersSubmenu())&&this.focus(this._openedBy),this._openedBy=void 0,this._setIsMenuOpen(!1))}_setIsMenuOpen(A){A!==this._menuOpen&&(this._menuOpen=A,this._menuOpen?this.menuOpened.emit():this.menuClosed.emit(),this.triggersSubmenu()&&this._menuItemInstance._setHighlighted(A),this._changeDetectorRef.markForCheck())}_createOverlay(A){if(!this._overlayRef){let i=this._getOverlayConfig(A);this._subscribeToPositions(A,i.positionStrategy),this._overlayRef=this._overlay.create(i),this._overlayRef.keydownEvents().subscribe(n=>{this.menu instanceof NQ&&this.menu._handleKeydown(n)})}return this._overlayRef}_getOverlayConfig(A){return new Bg({positionStrategy:this._overlay.position().flexibleConnectedTo(this._element).withLockedPosition().withGrowAfterOpen().withTransformOriginOn(".mat-menu-panel, .mat-mdc-menu-panel"),backdropClass:A.backdropClass||"cdk-overlay-transparent-backdrop",panelClass:A.overlayPanelClass,scrollStrategy:this._scrollStrategy(),direction:this._dir||"ltr"})}_subscribeToPositions(A,i){A.setPositionClasses&&i.positionChanges.subscribe(n=>{this._ngZone.run(()=>{let o=n.connectionPair.overlayX==="start"?"after":"before",r=n.connectionPair.overlayY==="top"?"below":"above";A.setPositionClasses(o,r)})})}_setPosition(A,i){let[n,o]=A.xPosition==="before"?["end","start"]:["start","end"],[r,s]=A.yPosition==="above"?["bottom","top"]:["top","bottom"],[a,c]=[r,s],[l,I]=[n,o],C=0;if(this.triggersSubmenu()){if(I=n=A.xPosition==="before"?"start":"end",o=l=n==="end"?"start":"end",this._parentMaterialMenu){if(this._parentInnerPadding==null){let d=this._parentMaterialMenu.items.first;this._parentInnerPadding=d?d._getHostElement().offsetTop:0}C=r==="bottom"?this._parentInnerPadding:-this._parentInnerPadding}}else A.overlapTrigger||(a=r==="top"?"bottom":"top",c=s==="top"?"bottom":"top");i.withPositions([{originX:n,originY:a,overlayX:l,overlayY:r,offsetY:C},{originX:o,originY:a,overlayX:I,overlayY:r,offsetY:C},{originX:n,originY:c,overlayX:l,overlayY:s,offsetY:-C},{originX:o,originY:c,overlayX:I,overlayY:s,offsetY:-C}])}_menuClosingActions(){let A=this._overlayRef.backdropClick(),i=this._overlayRef.detachments(),n=this._parentMaterialMenu?this._parentMaterialMenu.closed:Me(),o=this._parentMaterialMenu?this._parentMaterialMenu._hovered().pipe(kt(r=>this._menuOpen&&r!==this._menuItemInstance)):Me();return zn(A,n,o,i)}_handleMousedown(A){ku(A)||(this._openedBy=A.button===0?"mouse":void 0,this.triggersSubmenu()&&A.preventDefault())}_handleKeydown(A){let i=A.keyCode;(i===13||i===32)&&(this._openedBy="keyboard"),this.triggersSubmenu()&&(i===39&&this.dir==="ltr"||i===37&&this.dir==="rtl")&&(this._openedBy="keyboard",this.openMenu())}_handleClick(A){this.triggersSubmenu()?(A.stopPropagation(),this.openMenu()):this.toggleMenu()}_handleHover(){this.triggersSubmenu()&&this._parentMaterialMenu&&(this._hoverSubscription=this._parentMaterialMenu._hovered().subscribe(A=>{A===this._menuItemInstance&&!A.disabled&&(this._openedBy="mouse",this.openMenu())}))}_getPortal(A){return(!this._portal||this._portal.templateRef!==A.templateRef)&&(this._portal=new ys(A.templateRef,this._viewContainerRef)),this._portal}_ownsMenu(A){return vf.get(A)===this}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["","mat-menu-trigger-for",""],["","matMenuTriggerFor",""]],hostAttrs:[1,"mat-mdc-menu-trigger"],hostVars:3,hostBindings:function(i,n){i&1&&hA("click",function(r){return n._handleClick(r)})("mousedown",function(r){return n._handleMousedown(r)})("keydown",function(r){return n._handleKeydown(r)}),i&2&&Ne("aria-haspopup",n.menu?"menu":null)("aria-expanded",n.menuOpen)("aria-controls",n.menuOpen?n.menu.panelId:null)},inputs:{_deprecatedMatMenuTriggerFor:[0,"mat-menu-trigger-for","_deprecatedMatMenuTriggerFor"],menu:[0,"matMenuTriggerFor","menu"],menuData:[0,"matMenuTriggerData","menuData"],restoreFocus:[0,"matMenuTriggerRestoreFocus","restoreFocus"]},outputs:{menuOpened:"menuOpened",onMenuOpen:"onMenuOpen",menuClosed:"menuClosed",onMenuClose:"onMenuClose"},exportAs:["matMenuTrigger"]})}return t})(),flA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[ijA],imports:[Ps,it,El,Cl,it]})}return t})(),mlA={transformMenu:sc("transformMenu",[js("void",po({opacity:0,transform:"scale(0.8)"})),Vr("void => enter",ss("120ms cubic-bezier(0, 0, 0.2, 1)",po({opacity:1,transform:"scale(1)"}))),Vr("* => void",ss("100ms 25ms linear",po({opacity:0})))]),fadeInItems:sc("fadeInItems",[js("showing",po({opacity:1})),Vr("void => *",[po({opacity:0}),ss("400ms 100ms cubic-bezier(0.55, 0, 0.55, 0.2)")])])},zve=mlA.fadeInItems,Ove=mlA.transformMenu;var Mf=class t{sessionState={};constructor(){}static \u0275fac=function(A){return new(A||t)};static \u0275cmp=zA({type:t,selectors:[["app-state-tab"]],inputs:{sessionState:"sessionState"},standalone:!1,decls:3,vars:1,consts:[[1,"state-wrapper"],[3,"json"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"div"),JA(2,"ngx-json-viewer",1),F()()),A&2&&(G(2),yA("json",i.sessionState))},dependencies:[FQ],styles:[".state-wrapper[_ngcontent-%COMP%]{padding-left:25px;padding-right:25px;margin-top:16px}"]})};var kf=class t{constructor(e,A){this.el=e;this.renderer=A;this.sideDrawerMaxWidth=window.innerWidth/2}sideDrawerMinWidth=310;sideDrawerMaxWidth;resizeHandle=null;resizingEvent={isResizing:!1,startingCursorX:0,startingWidth:0};ngAfterViewInit(){this.resizeHandle=document.getElementsByClassName("resize-handler")[0],this.renderer.listen(this.resizeHandle,"mousedown",e=>this.onResizeHandleMouseDown(e)),document.documentElement.style.setProperty("--side-drawer-width","570px"),this.renderer.setStyle(this.el.nativeElement,"width","var(--side-drawer-width)")}onResizeHandleMouseDown(e){this.resizingEvent={isResizing:!0,startingCursorX:e.clientX,startingWidth:this.sideDrawerWidth},e.preventDefault()}onMouseMove(e){if(!this.resizingEvent.isResizing)return;let A=e.clientX-this.resizingEvent.startingCursorX,i=this.resizingEvent.startingWidth+A;this.sideDrawerWidth=i,this.renderer.addClass(document.body,"resizing")}onMouseUp(){this.resizingEvent.isResizing=!1,this.renderer.removeClass(document.body,"resizing")}onResize(){this.sideDrawerMaxWidth=window.innerWidth/2,this.sideDrawerWidth=this.sideDrawerWidth}set sideDrawerWidth(e){let A=Math.min(Math.max(e,this.sideDrawerMinWidth),this.sideDrawerMaxWidth);document.body.style.setProperty("--side-drawer-width",`${A}px`)}get sideDrawerWidth(){let e=getComputedStyle(document.body).getPropertyValue("--side-drawer-width"),A=parseInt(e,10);return isNaN(A)?500:A}static \u0275fac=function(A){return new(A||t)(PA(ee),PA(Wi))};static \u0275dir=jA({type:t,selectors:[["","appResizableDrawer",""]],hostBindings:function(A,i){A&1&&hA("mousemove",function(o){return i.onMouseMove(o)},!1,Wd)("mouseup",function(){return i.onMouseUp()},!1,Wd)("resize",function(){return i.onResize()},!1,wp)},standalone:!1})};var Sf=class t{constructor(e,A){this.el=e;this.renderer=A;this.bottomMaxHeight=window.innerHeight}bottomMinHeight=310;bottomMaxHeight;resizeHandle=null;resizingEvent={isResizing:!1,startingCursorY:0,startingHeight:0};ngAfterViewInit(){this.resizeHandle=document.getElementsByClassName("bottom-resize-handler")[0],this.renderer.listen(this.resizeHandle,"mousedown",e=>this.onResizeHandleMouseDown(e)),document.documentElement.style.setProperty("--bottom-panel-height","310px"),this.renderer.setStyle(this.el.nativeElement,"height","var(--bottom-panel-height)")}onResizeHandleMouseDown(e){this.resizingEvent={isResizing:!0,startingCursorY:e.clientY,startingHeight:this.bottomPanelHeight},e.preventDefault()}onMouseMove(e){if(!this.resizingEvent.isResizing)return;let A=this.resizingEvent.startingCursorY-e.clientY,i=this.resizingEvent.startingHeight+A;this.bottomPanelHeight=i,this.renderer.addClass(document.body,"resizing")}onMouseUp(){this.resizingEvent.isResizing=!1,this.renderer.removeClass(document.body,"resizing")}onResize(){this.bottomMaxHeight=window.innerHeight/2,this.bottomPanelHeight=this.bottomPanelHeight}set bottomPanelHeight(e){let A=Math.min(Math.max(e,this.bottomMinHeight),this.bottomMaxHeight);document.body.style.setProperty("--bottom-panel-height",`${A}px`)}get bottomPanelHeight(){let e=getComputedStyle(document.body).getPropertyValue("--bottom-panel-height"),A=parseInt(e,10);return isNaN(A)?500:A}static \u0275fac=function(A){return new(A||t)(PA(ee),PA(Wi))};static \u0275dir=jA({type:t,selectors:[["","appResizableBottomPanel",""]],hostBindings:function(A,i){A&1&&hA("mousemove",function(o){return i.onMouseMove(o)},!1,Wd)("mouseup",function(){return i.onMouseUp()},!1,Wd)("resize",function(){return i.onResize()},!1,wp)},standalone:!1})};var plA=new dA("CdkAccordion");var wlA=(()=>{class t{accordion=f(plA,{optional:!0,skipSelf:!0});_changeDetectorRef=f(It);_expansionDispatcher=f(xB);_openCloseAllSubscription=Kt.EMPTY;closed=new WA;opened=new WA;destroyed=new WA;expandedChange=new WA;id=f(sn).getId("cdk-accordion-child-");get expanded(){return this._expanded}set expanded(A){if(this._expanded!==A){if(this._expanded=A,this.expandedChange.emit(A),A){this.opened.emit();let i=this.accordion?this.accordion.id:this.id;this._expansionDispatcher.notify(this.id,i)}else this.closed.emit();this._changeDetectorRef.markForCheck()}}_expanded=!1;disabled=!1;_removeUniqueSelectionListener=()=>{};constructor(){}ngOnInit(){this._removeUniqueSelectionListener=this._expansionDispatcher.listen((A,i)=>{this.accordion&&!this.accordion.multi&&this.accordion.id===i&&this.id!==A&&(this.expanded=!1)}),this.accordion&&(this._openCloseAllSubscription=this._subscribeToOpenCloseAllActions())}ngOnDestroy(){this.opened.complete(),this.closed.complete(),this.destroyed.emit(),this.destroyed.complete(),this._removeUniqueSelectionListener(),this._openCloseAllSubscription.unsubscribe()}toggle(){this.disabled||(this.expanded=!this.expanded)}close(){this.disabled||(this.expanded=!1)}open(){this.disabled||(this.expanded=!0)}_subscribeToOpenCloseAllActions(){return this.accordion._openCloseAllActions.subscribe(A=>{this.disabled||(this.expanded=A)})}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["cdk-accordion-item"],["","cdkAccordionItem",""]],inputs:{expanded:[2,"expanded","expanded",ie],disabled:[2,"disabled","disabled",ie]},outputs:{closed:"closed",opened:"opened",destroyed:"destroyed",expandedChange:"expandedChange"},exportAs:["cdkAccordionItem"],features:[ht([{provide:plA,useValue:void 0}])]})}return t})(),DlA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({})}return t})();var ajA=["body"],cjA=["bodyWrapper"],ljA=[[["mat-expansion-panel-header"]],"*",[["mat-action-row"]]],gjA=["mat-expansion-panel-header","*","mat-action-row"];function IjA(t,e){}var CjA=[[["mat-panel-title"]],[["mat-panel-description"]],"*"],djA=["mat-panel-title","mat-panel-description","*"];function BjA(t,e){t&1&&(S(0,"span",1),ar(),S(1,"svg",2),JA(2,"path",3),F()())}var ylA=new dA("MAT_ACCORDION"),vlA=new dA("MAT_EXPANSION_PANEL"),EjA=(()=>{class t{_template=f(wn);_expansionPanel=f(vlA,{optional:!0});constructor(){}static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["ng-template","matExpansionPanelContent",""]]})}return t})(),blA=new dA("MAT_EXPANSION_PANEL_DEFAULT_OPTIONS"),cU=(()=>{class t extends wlA{_viewContainerRef=f(Nn);_animationsDisabled=f(Si,{optional:!0})==="NoopAnimations";_document=f(at);_ngZone=f(Qe);_elementRef=f(ee);_renderer=f(Wi);_cleanupTransitionEnd;get hideToggle(){return this._hideToggle||this.accordion&&this.accordion.hideToggle}set hideToggle(A){this._hideToggle=A}_hideToggle=!1;get togglePosition(){return this._togglePosition||this.accordion&&this.accordion.togglePosition}set togglePosition(A){this._togglePosition=A}_togglePosition;afterExpand=new WA;afterCollapse=new WA;_inputChanges=new OA;accordion=f(ylA,{optional:!0,skipSelf:!0});_lazyContent;_body;_bodyWrapper;_portal;_headerId=f(sn).getId("mat-expansion-panel-header-");constructor(){super();let A=f(blA,{optional:!0});this._expansionDispatcher=f(xB),A&&(this.hideToggle=A.hideToggle)}_hasSpacing(){return this.accordion?this.expanded&&this.accordion.displayMode==="default":!1}_getExpandedState(){return this.expanded?"expanded":"collapsed"}toggle(){this.expanded=!this.expanded}close(){this.expanded=!1}open(){this.expanded=!0}ngAfterContentInit(){this._lazyContent&&this._lazyContent._expansionPanel===this&&this.opened.pipe(Pn(null),kt(()=>this.expanded&&!this._portal),On(1)).subscribe(()=>{this._portal=new ys(this._lazyContent._template,this._viewContainerRef)}),this._setupAnimationEvents()}ngOnChanges(A){this._inputChanges.next(A)}ngOnDestroy(){super.ngOnDestroy(),this._cleanupTransitionEnd?.(),this._inputChanges.complete()}_containsFocus(){if(this._body){let A=this._document.activeElement,i=this._body.nativeElement;return A===i||i.contains(A)}return!1}_transitionEndListener=({target:A,propertyName:i})=>{A===this._bodyWrapper?.nativeElement&&i==="grid-template-rows"&&this._ngZone.run(()=>{this.expanded?this.afterExpand.emit():this.afterCollapse.emit()})};_setupAnimationEvents(){this._ngZone.runOutsideAngular(()=>{this._animationsDisabled?(this.opened.subscribe(()=>this._ngZone.run(()=>this.afterExpand.emit())),this.closed.subscribe(()=>this._ngZone.run(()=>this.afterCollapse.emit()))):setTimeout(()=>{let A=this._elementRef.nativeElement;this._cleanupTransitionEnd=this._renderer.listen(A,"transitionend",this._transitionEndListener),A.classList.add("mat-expansion-panel-animations-enabled")},200)})}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-expansion-panel"]],contentQueries:function(i,n,o){if(i&1&&ci(o,EjA,5),i&2){let r;XA(r=$A())&&(n._lazyContent=r.first)}},viewQuery:function(i,n){if(i&1&&(Te(ajA,5),Te(cjA,5)),i&2){let o;XA(o=$A())&&(n._body=o.first),XA(o=$A())&&(n._bodyWrapper=o.first)}},hostAttrs:[1,"mat-expansion-panel"],hostVars:4,hostBindings:function(i,n){i&2&&ue("mat-expanded",n.expanded)("mat-expansion-panel-spacing",n._hasSpacing())},inputs:{hideToggle:[2,"hideToggle","hideToggle",ie],togglePosition:"togglePosition"},outputs:{afterExpand:"afterExpand",afterCollapse:"afterCollapse"},exportAs:["matExpansionPanel"],features:[ht([{provide:ylA,useValue:void 0},{provide:vlA,useExisting:t}]),lt,ti],ngContentSelectors:gjA,decls:9,vars:4,consts:[["bodyWrapper",""],["body",""],[1,"mat-expansion-panel-content-wrapper"],["role","region",1,"mat-expansion-panel-content",3,"id"],[1,"mat-expansion-panel-body"],[3,"cdkPortalOutlet"]],template:function(i,n){i&1&&(jt(ljA),Fe(0),S(1,"div",2,0)(3,"div",3,1)(5,"div",4),Fe(6,1),_A(7,IjA,0,0,"ng-template",5),F(),Fe(8,2),F()()),i&2&&(G(),Ne("inert",n.expanded?null:""),G(2),yA("id",n.id),Ne("aria-labelledby",n._headerId),G(4),yA("cdkPortalOutlet",n._portal))},dependencies:[fa],styles:[".mat-expansion-panel{box-sizing:content-box;display:block;margin:0;overflow:hidden;position:relative;background:var(--mat-expansion-container-background-color, var(--mat-sys-surface));color:var(--mat-expansion-container-text-color, var(--mat-sys-on-surface));border-radius:var(--mat-expansion-container-shape, 12px)}.mat-expansion-panel.mat-expansion-panel-animations-enabled{transition:margin 225ms cubic-bezier(0.4, 0, 0.2, 1),box-shadow 280ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-expansion-panel:not([class*=mat-elevation-z]){box-shadow:0px 3px 1px -2px rgba(0, 0, 0, 0.2), 0px 2px 2px 0px rgba(0, 0, 0, 0.14), 0px 1px 5px 0px rgba(0, 0, 0, 0.12)}.mat-accordion .mat-expansion-panel:not(.mat-expanded),.mat-accordion .mat-expansion-panel:not(.mat-expansion-panel-spacing){border-radius:0}.mat-accordion .mat-expansion-panel:first-of-type{border-top-right-radius:var(--mat-expansion-container-shape, 12px);border-top-left-radius:var(--mat-expansion-container-shape, 12px)}.mat-accordion .mat-expansion-panel:last-of-type{border-bottom-right-radius:var(--mat-expansion-container-shape, 12px);border-bottom-left-radius:var(--mat-expansion-container-shape, 12px)}@media(forced-colors: active){.mat-expansion-panel{outline:solid 1px}}.mat-expansion-panel-content-wrapper{display:grid;grid-template-rows:0fr;grid-template-columns:100%}.mat-expansion-panel-animations-enabled .mat-expansion-panel-content-wrapper{transition:grid-template-rows 225ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-expansion-panel.mat-expanded>.mat-expansion-panel-content-wrapper{grid-template-rows:1fr}@supports not (grid-template-rows: 0fr){.mat-expansion-panel-content-wrapper{height:0}.mat-expansion-panel.mat-expanded>.mat-expansion-panel-content-wrapper{height:auto}}.mat-expansion-panel-content{display:flex;flex-direction:column;overflow:visible;min-height:0;visibility:hidden;font-family:var(--mat-expansion-container-text-font, var(--mat-sys-body-large-font));font-size:var(--mat-expansion-container-text-size, var(--mat-sys-body-large-size));font-weight:var(--mat-expansion-container-text-weight, var(--mat-sys-body-large-weight));line-height:var(--mat-expansion-container-text-line-height, var(--mat-sys-body-large-line-height));letter-spacing:var(--mat-expansion-container-text-tracking, var(--mat-sys-body-large-tracking))}.mat-expansion-panel-animations-enabled .mat-expansion-panel-content{transition:visibility 190ms linear}.mat-expansion-panel.mat-expanded>.mat-expansion-panel-content-wrapper>.mat-expansion-panel-content{visibility:visible}.mat-expansion-panel-body{padding:0 24px 16px}.mat-expansion-panel-spacing{margin:16px 0}.mat-accordion>.mat-expansion-panel-spacing:first-child,.mat-accordion>*:first-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-top:0}.mat-accordion>.mat-expansion-panel-spacing:last-child,.mat-accordion>*:last-child:not(.mat-expansion-panel) .mat-expansion-panel-spacing{margin-bottom:0}.mat-action-row{border-top-style:solid;border-top-width:1px;display:flex;flex-direction:row;justify-content:flex-end;padding:16px 8px 16px 24px;border-top-color:var(--mat-expansion-actions-divider-color, var(--mat-sys-outline))}.mat-action-row .mat-button-base,.mat-action-row .mat-mdc-button-base{margin-left:8px}[dir=rtl] .mat-action-row .mat-button-base,[dir=rtl] .mat-action-row .mat-mdc-button-base{margin-left:0;margin-right:8px}"],encapsulation:2,changeDetection:0})}return t})();var MlA=(()=>{class t{panel=f(cU,{host:!0});_element=f(ee);_focusMonitor=f(dr);_changeDetectorRef=f(It);_parentChangeSubscription=Kt.EMPTY;constructor(){f(Rn).load(lr);let A=this.panel,i=f(blA,{optional:!0}),n=f(new wr("tabindex"),{optional:!0}),o=A.accordion?A.accordion._stateChanges.pipe(kt(r=>!!(r.hideToggle||r.togglePosition))):sr;this.tabIndex=parseInt(n||"")||0,this._parentChangeSubscription=zn(A.opened,A.closed,o,A._inputChanges.pipe(kt(r=>!!(r.hideToggle||r.disabled||r.togglePosition)))).subscribe(()=>this._changeDetectorRef.markForCheck()),A.closed.pipe(kt(()=>A._containsFocus())).subscribe(()=>this._focusMonitor.focusVia(this._element,"program")),i&&(this.expandedHeight=i.expandedHeight,this.collapsedHeight=i.collapsedHeight)}expandedHeight;collapsedHeight;tabIndex=0;get disabled(){return this.panel.disabled}_toggle(){this.disabled||this.panel.toggle()}_isExpanded(){return this.panel.expanded}_getExpandedState(){return this.panel._getExpandedState()}_getPanelId(){return this.panel.id}_getTogglePosition(){return this.panel.togglePosition}_showToggle(){return!this.panel.hideToggle&&!this.panel.disabled}_getHeaderHeight(){let A=this._isExpanded();return A&&this.expandedHeight?this.expandedHeight:!A&&this.collapsedHeight?this.collapsedHeight:null}_keydown(A){switch(A.keyCode){case 32:case 13:ir(A)||(A.preventDefault(),this._toggle());break;default:this.panel.accordion&&this.panel.accordion._handleHeaderKeydown(A);return}}focus(A,i){A?this._focusMonitor.focusVia(this._element,A,i):this._element.nativeElement.focus(i)}ngAfterViewInit(){this._focusMonitor.monitor(this._element).subscribe(A=>{A&&this.panel.accordion&&this.panel.accordion._handleHeaderFocus(this)})}ngOnDestroy(){this._parentChangeSubscription.unsubscribe(),this._focusMonitor.stopMonitoring(this._element)}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-expansion-panel-header"]],hostAttrs:["role","button",1,"mat-expansion-panel-header","mat-focus-indicator"],hostVars:13,hostBindings:function(i,n){i&1&&hA("click",function(){return n._toggle()})("keydown",function(r){return n._keydown(r)}),i&2&&(Ne("id",n.panel._headerId)("tabindex",n.disabled?-1:n.tabIndex)("aria-controls",n._getPanelId())("aria-expanded",n._isExpanded())("aria-disabled",n.panel.disabled),uo("height",n._getHeaderHeight()),ue("mat-expanded",n._isExpanded())("mat-expansion-toggle-indicator-after",n._getTogglePosition()==="after")("mat-expansion-toggle-indicator-before",n._getTogglePosition()==="before"))},inputs:{expandedHeight:"expandedHeight",collapsedHeight:"collapsedHeight",tabIndex:[2,"tabIndex","tabIndex",A=>A==null?0:zi(A)]},ngContentSelectors:djA,decls:5,vars:3,consts:[[1,"mat-content"],[1,"mat-expansion-indicator"],["xmlns","http://www.w3.org/2000/svg","viewBox","0 -960 960 960","aria-hidden","true","focusable","false"],["d","M480-345 240-585l56-56 184 184 184-184 56 56-240 240Z"]],template:function(i,n){i&1&&(jt(CjA),S(0,"span",0),Fe(1),Fe(2,1),Fe(3,2),F(),_A(4,BjA,3,0,"span",1)),i&2&&(ue("mat-content-hide-toggle",!n._showToggle()),G(4),GA(n._showToggle()?4:-1))},styles:['.mat-expansion-panel-header{display:flex;flex-direction:row;align-items:center;padding:0 24px;border-radius:inherit;height:var(--mat-expansion-header-collapsed-state-height, 48px);font-family:var(--mat-expansion-header-text-font, var(--mat-sys-title-medium-font));font-size:var(--mat-expansion-header-text-size, var(--mat-sys-title-medium-size));font-weight:var(--mat-expansion-header-text-weight, var(--mat-sys-title-medium-weight));line-height:var(--mat-expansion-header-text-line-height, var(--mat-sys-title-medium-line-height));letter-spacing:var(--mat-expansion-header-text-tracking, var(--mat-sys-title-medium-tracking))}.mat-expansion-panel-animations-enabled .mat-expansion-panel-header{transition:height 225ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-expansion-panel-header::before{border-radius:inherit}.mat-expansion-panel-header.mat-expanded{height:var(--mat-expansion-header-expanded-state-height, 64px)}.mat-expansion-panel-header[aria-disabled=true]{color:var(--mat-expansion-header-disabled-state-text-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-expansion-panel-header:not([aria-disabled=true]){cursor:pointer}.mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=true]):hover{background:var(--mat-expansion-header-hover-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-hover-state-layer-opacity) * 100%), transparent))}@media(hover: none){.mat-expansion-panel:not(.mat-expanded) .mat-expansion-panel-header:not([aria-disabled=true]):hover{background:var(--mat-expansion-container-background-color, var(--mat-sys-surface))}}.mat-expansion-panel .mat-expansion-panel-header:not([aria-disabled=true]).cdk-keyboard-focused,.mat-expansion-panel .mat-expansion-panel-header:not([aria-disabled=true]).cdk-program-focused{background:var(--mat-expansion-header-focus-state-layer-color, color-mix(in srgb, var(--mat-sys-on-surface) calc(var(--mat-sys-focus-state-layer-opacity) * 100%), transparent))}.mat-expansion-panel-header._mat-animation-noopable{transition:none}.mat-expansion-panel-header:focus,.mat-expansion-panel-header:hover{outline:none}.mat-expansion-panel-header.mat-expanded:focus,.mat-expansion-panel-header.mat-expanded:hover{background:inherit}.mat-expansion-panel-header.mat-expansion-toggle-indicator-before{flex-direction:row-reverse}.mat-expansion-panel-header.mat-expansion-toggle-indicator-before .mat-expansion-indicator{margin:0 16px 0 0}[dir=rtl] .mat-expansion-panel-header.mat-expansion-toggle-indicator-before .mat-expansion-indicator{margin:0 0 0 16px}.mat-content{display:flex;flex:1;flex-direction:row;overflow:hidden}.mat-content.mat-content-hide-toggle{margin-right:8px}[dir=rtl] .mat-content.mat-content-hide-toggle{margin-right:0;margin-left:8px}.mat-expansion-toggle-indicator-before .mat-content.mat-content-hide-toggle{margin-left:24px;margin-right:0}[dir=rtl] .mat-expansion-toggle-indicator-before .mat-content.mat-content-hide-toggle{margin-right:24px;margin-left:0}.mat-expansion-panel-header-title{color:var(--mat-expansion-header-text-color, var(--mat-sys-on-surface))}.mat-expansion-panel-header-title,.mat-expansion-panel-header-description{display:flex;flex-grow:1;flex-basis:0;margin-right:16px;align-items:center}[dir=rtl] .mat-expansion-panel-header-title,[dir=rtl] .mat-expansion-panel-header-description{margin-right:0;margin-left:16px}.mat-expansion-panel-header[aria-disabled=true] .mat-expansion-panel-header-title,.mat-expansion-panel-header[aria-disabled=true] .mat-expansion-panel-header-description{color:inherit}.mat-expansion-panel-header-description{flex-grow:2;color:var(--mat-expansion-header-description-color, var(--mat-sys-on-surface-variant))}.mat-expansion-panel-animations-enabled .mat-expansion-indicator{transition:transform 225ms cubic-bezier(0.4, 0, 0.2, 1)}.mat-expansion-panel-header.mat-expanded .mat-expansion-indicator{transform:rotate(180deg)}.mat-expansion-indicator::after{border-style:solid;border-width:0 2px 2px 0;content:"";display:inline-block;padding:3px;transform:rotate(45deg);vertical-align:middle;color:var(--mat-expansion-header-indicator-color, var(--mat-sys-on-surface-variant));display:var(--mat-expansion-legacy-header-indicator-display, none)}.mat-expansion-indicator svg{width:24px;height:24px;margin:0 -8px;vertical-align:middle;fill:var(--mat-expansion-header-indicator-color, var(--mat-sys-on-surface-variant));display:var(--mat-expansion-header-indicator-display, inline-block)}@media(forced-colors: active){.mat-expansion-panel-content{border-top:1px solid;border-top-left-radius:0;border-top-right-radius:0}}'],encapsulation:2,changeDetection:0})}return t})();var klA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275dir=jA({type:t,selectors:[["mat-panel-title"]],hostAttrs:[1,"mat-expansion-panel-header-title"]})}return t})();var SlA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,DlA,dg]})}return t})();var RlA=t=>({color:t});function hjA(t,e){t&1&&JA(0,"div",8)}function ujA(t,e){if(t&1&&(S(0,"span",14),AA(1),F()),t&2){let A=O().$implicit,i=O();uo("left",i.getRelativeStart(A.span)+5,"%"),G(),Et("",(i.toMs(A.span.end_time)-i.toMs(A.span.start_time)).toFixed(2),"ms")}}function fjA(t,e){if(t&1){let A=be();S(0,"div",5),hA("click",function(){let n=RA(A).$implicit,o=O();return xA(o.selectRow(n))})("mouseenter",function(){let n=RA(A).$implicit,o=O();return xA(o.onHover(n))})("mouseleave",function(){RA(A);let n=O();return xA(n.onHoverOut())}),S(1,"div",6)(2,"div",7),Dn(3,hjA,1,0,"div",8,Xd),F(),S(5,"span",9),AA(6),F(),S(7,"div",10),AA(8),F()(),S(9,"div",11)(10,"div",12),AA(11),F(),_A(12,ujA,2,3,"span",13),F()()}if(t&2){let A=e.$implicit,i=O();ue("selected",i.rowSelected(A)),G(3),yn(i.getArray(A.level)),G(2),yA("ngStyle",qr(14,RlA,i.isEventRow(A)?"#8AB4F8":"white")),G(),Et(" ",i.getSpanIcon(A.span.name)," "),G(),uo("width",400-A.level*20,"px"),yA("ngStyle",qr(16,RlA,i.isEventRow(A)?"#8AB4F8":"white")),G(),Et(" ",A.span.name," "),G(2),uo("left",i.getRelativeStart(A.span),"%")("width",i.getRelativeWidth(A.span),"%"),G(),Et(" ",(i.toMs(A.span.end_time)-i.toMs(A.span.start_time)).toFixed(2),"ms "),G(),GA(i.getRelativeWidth(A.span)<10?12:-1)}}var Rf=class t{constructor(e){this.traceService=e}spans=[];invocationId="";tree=[];eventData;baseStartTimeMs=0;totalDurationMs=1;flatTree=[];traceLabelIconMap=new Map([["Invocation","start"],["agent_run","directions_run"],["tool","build"],["call_llm","chat"]]);selectedRow=void 0;ngOnInit(){this.tree=this.buildSpanTree(this.spans),this.flatTree=this.flattenTree(this.tree);let e=this.getGlobalTimes(this.spans);this.baseStartTimeMs=e.start,this.totalDurationMs=e.duration,this.traceService.selectedTraceRow$.subscribe(A=>this.selectedRow=A),this.traceService.eventData$.subscribe(A=>this.eventData=A)}buildSpanTree(e){let A=e.map(o=>rA({},o)),i=new Map,n=[];return A.forEach(o=>i.set(o.span_id,o)),A.forEach(o=>{if(o.parent_span_id&&i.has(o.parent_span_id)){let r=i.get(o.parent_span_id);r.children=r.children||[],r.children.push(o)}else n.push(o)}),n}getGlobalTimes(e){let A=Math.min(...e.map(n=>this.toMs(n.start_time))),i=Math.max(...e.map(n=>this.toMs(n.end_time)));return{start:A,duration:i-A}}toMs(e){return e/1e6}getRelativeStart(e){return(this.toMs(e.start_time)-this.baseStartTimeMs)/this.totalDurationMs*100}getRelativeWidth(e){return(this.toMs(e.end_time)-this.toMs(e.start_time))/this.totalDurationMs*100}flattenTree(e,A=0){return e.flatMap(n=>[{span:n,level:A},...n.children?this.flattenTree(n.children,A+1):[]])}getSpanIcon(e){for(let[A,i]of this.traceLabelIconMap.entries())if(e.startsWith(A))return i;return"start"}getArray(e){return Array.from({length:e})}selectRow(e){if(this.selectedRow&&this.selectedRow.span_id==e.span.span_id){this.traceService.selectedRow(void 0),this.traceService.setHoveredMessages(void 0,this.invocationId);return}this.traceService.selectedRow(e.span),this.traceService.setHoveredMessages(e.span,this.invocationId)}rowSelected(e){return this.selectedRow==e.span}isEventRow(e){if(!e.span.attributes)return!1;let A=e?.span.attributes["gcp.vertex.agent.event_id"];return!!(A&&this.eventData&&this.eventData.has(A))}onHover(e){this.traceService.setHoveredMessages(e.span,this.invocationId)}onHoverOut(){this.traceService.setHoveredMessages(void 0,this.invocationId),this.selectedRow&&this.traceService.setHoveredMessages(this.selectedRow,this.invocationId)}static \u0275fac=function(A){return new(A||t)(PA($g))};static \u0275cmp=zA({type:t,selectors:[["app-trace-tree"]],inputs:{spans:"spans",invocationId:"invocationId"},standalone:!1,decls:8,vars:1,consts:[[2,"margin-top","15px"],[1,"invocation-id-container"],[1,"invocation-id"],[1,"trace-container"],[1,"trace-row",3,"selected"],[1,"trace-row",3,"click","mouseenter","mouseleave"],[1,"trace-row-left"],[1,"trace-indent"],[1,"indent-connector"],[1,"material-symbols-outlined",2,"margin-right","8px",3,"ngStyle"],[1,"trace-label",3,"ngStyle"],[1,"trace-bar-container"],[1,"trace-bar"],[2,"position","absolute","color","#8dabbf",3,"left"],[2,"position","absolute","color","#8dabbf"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"div",1),AA(2,"Invocation ID: "),S(3,"div",2),AA(4),F()(),S(5,"div",3),Dn(6,fjA,13,18,"div",4,to),F()()),A&2&&(G(4),Gt(i.invocationId),G(2),yn(i.flatTree))},dependencies:[Kh],styles:[".trace-container[_ngcontent-%COMP%]{width:100%;white-space:nowrap;font-size:12px}.trace-label[_ngcontent-%COMP%]{width:400px;color:#e3e3e3;font-family:Google Sans Mono,monospace;font-size:13px;font-style:normal;font-weight:500;line-height:20px;letter-spacing:0px;text-overflow:ellipsis;white-space:nowrap;overflow:hidden}.trace-bar-container[_ngcontent-%COMP%]{width:100%;position:relative;height:16px}.trace-bar[_ngcontent-%COMP%]{position:absolute;height:18px;background-color:#2f4d65;border-radius:4px;padding-left:4px;overflow:hidden;font-size:11px;line-height:16px;color:#8dabbf;font-family:Google Sans}.trace-duration[_ngcontent-%COMP%]{color:#888;font-weight:400;margin-left:4px}.trace-row[_ngcontent-%COMP%]{display:flex;align-items:stretch;position:relative;height:32px;align-items:center;cursor:pointer}.trace-row[_ngcontent-%COMP%]:hover, .trace-row.selected[_ngcontent-%COMP%]{background-color:#3b3d3c}.trace-indent[_ngcontent-%COMP%]{display:flex;flex-shrink:0;height:100%}.indent-connector[_ngcontent-%COMP%]{width:20px;position:relative;height:100%}.vertical-line[_ngcontent-%COMP%]{position:absolute;top:0;bottom:0;left:9px;width:1px;background-color:#ccc}.horizontal-line[_ngcontent-%COMP%]{position:absolute;top:50%;left:9px;width:10px;height:1px;background-color:#ccc}.trace-row-left[_ngcontent-%COMP%]{display:flex;width:50%}.invocation-id-container[_ngcontent-%COMP%]{color:#9aa0a6;font-size:14px;font-style:normal;font-weight:700;line-height:20px;letter-spacing:0px;margin-bottom:5px}.invocation-id[_ngcontent-%COMP%]{font-family:Google Sans Mono,monospace}"]})};function pjA(t,e){if(t&1&&(S(0,"div",3)(1,"mat-expansion-panel")(2,"mat-expansion-panel-header")(3,"mat-panel-title"),AA(4),F()(),JA(5,"app-trace-tree",4),F()()),t&2){let A=e.$implicit,i=O();G(4),Et(" ",i.invocToUserMsg.get(A.key)," "),G(),yA("spans",A.value)("invocationId",i.findInvocIdFromTraceId(A.key))}}var xf=class t{traceData=[];invocTraces=new Map;invocToUserMsg=new Map;constructor(){}ngOnInit(){}ngOnChanges(e){"traceData"in e&&this.rebuildTrace()}rebuildTrace(){this.invocTraces=this.traceData.reduce((e,A)=>{let i=A.trace_id,n=e.get(i);return n?(n.push(A),n.sort((o,r)=>o.start_time-r.start_time)):e.set(i,[A]),e},new Map);for(let[e,A]of this.invocTraces)this.invocToUserMsg.set(e,this.findUserMsgFromInvocGroup(A))}getArray(e){return Array.from({length:e})}findUserMsgFromInvocGroup(e){let A=e?.find(o=>o.attributes!==void 0&&"gcp.vertex.agent.invocation_id"in o.attributes);return JSON.parse(A.attributes["gcp.vertex.agent.llm_request"]).contents.filter(o=>o.role=="user").at(-1).parts[0]?.text??"[attachment]"}findInvocIdFromTraceId(e){return this.invocTraces.get(e)?.find(i=>i.attributes!==void 0&&"gcp.vertex.agent.invocation_id"in i.attributes).attributes["gcp.vertex.agent.invocation_id"]}mapOrderPreservingSort=(e,A)=>0;static \u0275fac=function(A){return new(A||t)};static \u0275cmp=zA({type:t,selectors:[["app-trace-tab"]],inputs:{traceData:"traceData"},standalone:!1,features:[ti],decls:7,vars:3,consts:[[2,"padding-left","25px","padding-right","25px"],["mat-dialog-title","",1,"trace-title"],[1,"trace-list-wrapper"],[1,"trace-item"],[3,"spans","invocationId"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"h2",1),AA(2,"Invocations"),F(),S(3,"div",2),Dn(4,pjA,6,3,"div",3,to),Za(6,"keyvalue"),F()()),A&2&&(G(4),yn(Lh(6,0,i.invocTraces,i.mapOrderPreservingSort)))},dependencies:[bs,cU,MlA,klA,Rf,Th],styles:[".trace-container[_ngcontent-%COMP%]{width:100%;white-space:nowrap;font-size:12px}.trace-title[_ngcontent-%COMP%]{color:#9aa0a6;font-size:14px;font-style:normal;font-weight:700;line-height:20px;letter-spacing:0px}.trace-label[_ngcontent-%COMP%]{width:400px;color:#e3e3e3;text-overflow:ellipsis;font-family:Google Sans Mono,monospace;font-size:14px;font-style:normal;font-weight:500;line-height:20px;letter-spacing:0px}.trace-bar-container[_ngcontent-%COMP%]{width:50vw;position:relative;height:16px}.trace-bar[_ngcontent-%COMP%]{position:absolute;height:18px;background-color:#2f4d65;border-radius:4px;padding-left:4px;overflow:hidden;font-size:11px;line-height:16px;color:#8dabbf;font-family:Google Sans}.trace-duration[_ngcontent-%COMP%]{color:#888;font-weight:400;margin-left:4px}.trace-row[_ngcontent-%COMP%]{display:flex;align-items:stretch;position:relative;height:32px}.trace-indent[_ngcontent-%COMP%]{display:flex;flex-shrink:0;height:100%}.indent-connector[_ngcontent-%COMP%]{width:20px;position:relative;height:100%}.vertical-line[_ngcontent-%COMP%]{position:absolute;top:0;bottom:0;left:9px;width:1px;background-color:#ccc}.horizontal-line[_ngcontent-%COMP%]{position:absolute;top:50%;left:9px;width:10px;height:1px;background-color:#ccc}.trace-item[_ngcontent-%COMP%]{margin-top:5px}.trace-item[_ngcontent-%COMP%]{--mat-expansion-container-background-color: #333537}.trace-item[_ngcontent-%COMP%]{--mat-expansion-header-focus-state-layer-color: red}.trace-item[_ngcontent-%COMP%]{--mat-expansion-header-description-color: #8e918f}.trace-item[_ngcontent-%COMP%]{--mat-expansion-header-text-size: 15}  .mat-expansion-panel-header.mat-expanded:focus{background-color:#444746!important}  .mat-expansion-panel-header.mat-expanded{background-color:#444746!important}  .mat-expansion-panel-header.mat-expanded:hover{background-color:#444746!important}  .mat-expansion-panel-header-title{text-overflow:ellipsis;white-space:nowrap;overflow:hidden}  .mat-expansion-panel-header-description{text-overflow:ellipsis;white-space:nowrap;overflow:hidden}"]})};function DjA(t,e){if(t&1){let A=be();S(0,"div",11),hA("click",function(){RA(A);let n=O();return xA(n.openViewImageDialog(n.rawSvgString))}),F()}if(t&2){let A=O();yA("innerHtml",A.renderedEventGraph,RI)}}var Lf=class t{constructor(e,A,i,n){this.dialog=e;this.traceService=A;this.eventService=i;this.sanitizer=n}userId="";sessionId="";appName="";panelClosed=new WA;renderedEventGraph;eventData;selectedRow=void 0;rawSvgString=null;llmRequest=void 0;llmResponse=void 0;llmRequestKey="gcp.vertex.agent.llm_request";llmResponseKey="gcp.vertex.agent.llm_response";ngOnInit(){this.traceService.selectedTraceRow$.subscribe(e=>{this.selectedRow=e;let A=this.getEventIdFromSpan();A&&(this.eventService.getEventTrace(A).subscribe(i=>{this.llmRequest=JSON.parse(i[this.llmRequestKey]),this.llmResponse=JSON.parse(i[this.llmResponseKey])}),this.getEventGraph(A))}),this.traceService.eventData$.subscribe(e=>this.eventData=e)}openViewImageDialog(e){let A=this.dialog.open(v0,{maxWidth:"90vw",maxHeight:"90vh",data:{imageData:e}})}getEventDetails(){if(this.eventData&&this.selectedRow)return this.eventData.get(this.getEventIdFromSpan())}getEventIdFromSpan(){if(this.selectedRow)return this.selectedRow.attributes["gcp.vertex.agent.event_id"]}getEventGraph(e){this.eventService.getEvent(this.userId,this.appName,this.sessionId,e).subscribe(A=>Ao(this,null,function*(){if(!A.dotSrc){this.renderedEventGraph=void 0;return}let i=A.dotSrc,o=(yield Yu()).renderString(i,{format:"svg",engine:"dot"});this.rawSvgString=o,this.renderedEventGraph=this.sanitizer.bypassSecurityTrustHtml(o)}))}closePanel(){this.panelClosed.emit(!0)}static \u0275fac=function(A){return new(A||t)(PA(qs),PA($g),PA(nI),PA(cl))};static \u0275cmp=zA({type:t,selectors:[["app-trace-event"]],inputs:{userId:"userId",sessionId:"sessionId",appName:"appName"},outputs:{panelClosed:"panelClosed"},standalone:!1,decls:17,vars:4,consts:[[1,"wrapper"],["mat-stretch-tabs","false","mat-align-tabs","start"],["label","Event"],[1,"json-viewer-container"],[3,"json"],["label","Request"],["label","Response"],["label","Graph"],[1,"event-graph-container"],[3,"innerHtml"],["mat-icon-button","",1,"tab-header-action",3,"click"],[3,"click","innerHtml"]],template:function(A,i){A&1&&(S(0,"div",0)(1,"mat-tab-group",1)(2,"mat-tab",2)(3,"div",3),JA(4,"ngx-json-viewer",4),F()(),S(5,"mat-tab",5)(6,"div",3),JA(7,"ngx-json-viewer",4),F()(),S(8,"mat-tab",6)(9,"div",3),JA(10,"ngx-json-viewer",4),F()(),S(11,"mat-tab",7)(12,"div",8),_A(13,DjA,1,1,"div",9),F()()(),S(14,"button",10),hA("click",function(){return i.closePanel()}),S(15,"mat-icon"),AA(16,"close"),F()()()),A&2&&(G(4),yA("json",i.getEventDetails()),G(3),yA("json",i.llmRequest),G(3),yA("json",i.llmResponse),G(3),GA(i.renderedEventGraph?13:-1))},dependencies:[P2,kB,FQ,mf,o7],styles:[".json-viewer-container[_ngcontent-%COMP%]{padding-top:8px;padding-left:12px;padding-right:12px;background-color:#1b1b1b}.event-graph-container[_ngcontent-%COMP%]{text-align:center;padding-top:20px}.event-graph-container[_ngcontent-%COMP%]     svg text{font-family:Google Sans Mono,monospace;font-size:11px}.wrapper[_ngcontent-%COMP%]{position:relative}.tab-header-action[_ngcontent-%COMP%]{position:absolute;top:0;right:0;height:48px;z-index:2;margin-right:10px}"]})};var vjA=["videoContainer"],bjA=["sideDrawer"],MjA=["autoScroll"],kjA=["messageTextarea"],SjA=["bottomPanel"],RjA=t=>({"edit-mode":t}),xjA=()=>[],LjA=(t,e)=>({"user-message":t,"bot-message":e}),FjA=(t,e)=>({"eval-pass":t,"eval-fail":e}),NjA=t=>({"eval-fail":t}),IU=t=>({"background-color":t}),_jA=(t,e)=>({"font-style":t,color:e}),LlA=t=>({"function-event-button-highlight":t}),CU=t=>({hidden:t});function GjA(t,e){if(t&1){let A=be();S(0,"span",29),hA("click",function(){RA(A);let n=O();return xA(n.toggleSidePanel())}),AA(1,"left_panel_open"),F()}}function UjA(t,e){if(t&1&&(S(0,"mat-option",18),AA(1),F()),t&2){let A=e.$implicit;yA("value",A),G(),Gt(A)}}function KjA(t,e){t&1&&Dn(0,UjA,2,2,"mat-option",18,to),t&2&&yn(e)}function YjA(t,e){if(t&1&&(S(0,"mat-option",18),AA(1),F()),t&2){let A=O();yA("value",A.selectedAppControl.value),G(),Gt(A.selectedAppControl.value)}}function JjA(t,e){t&1&&(S(0,"span",38),AA(1,"Trace"),F())}function TjA(t,e){t&1&&(S(0,"span",38),AA(1,"Events"),F())}function HjA(t,e){t&1&&(S(0,"span",38),AA(1,"State"),F())}function zjA(t,e){t&1&&(S(0,"span",38),AA(1,"Artifacts"),F())}function OjA(t,e){t&1&&(S(0,"span",38),AA(1,"Sessions"),F())}function PjA(t,e){t&1&&(S(0,"span",38),AA(1,"Eval"),F())}function jjA(t,e){if(t&1){let A=be();S(0,"mat-tab"),_A(1,PjA,2,0,"ng-template",32),S(2,"app-eval-tab",39),hA("shouldShowTab",function(n){RA(A);let o=O(2);return xA(o.handleShouldShowEvalTab(n))})("sessionSelected",function(n){RA(A);let o=O(2);return xA(o.updateWithSelectedSession(n))})("evalCaseSelected",function(n){RA(A);let o=O(2);return xA(o.updateWithSelectedEvalCase(n))})("evalSetIdSelected",function(n){RA(A);let o=O(2);return xA(o.updateSelectedEvalSetId(n))})("shouldReturnToSession",function(n){RA(A);let o=O(2);return xA(o.handleReturnToSession(n))})("evalNotInstalledMsg",function(n){RA(A);let o=O(2);return xA(o.handleEvalNotInstalled(n))}),F()()}if(t&2){let A=O(2);G(2),yA("appName",A.appName)("userId",A.userId)("sessionId",A.sessionId)}}function qjA(t,e){if(t&1){let A=be();S(0,"div",19)(1,"mat-tab-group",30),hA("selectedTabChange",function(n){RA(A);let o=O();return xA(o.handleTabChange(n))}),S(2,"mat-tab",31),_A(3,JjA,2,0,"ng-template",32),JA(4,"app-trace-tab",33),F(),S(5,"mat-tab",31),_A(6,TjA,2,0,"ng-template",32),S(7,"app-event-tab",34),hA("selectedEvent",function(n){RA(A);let o=O();return xA(o.selectEvent(n))}),F()(),S(8,"mat-tab"),_A(9,HjA,2,0,"ng-template",32),JA(10,"app-state-tab",35),F(),S(11,"mat-tab"),_A(12,zjA,2,0,"ng-template",32),JA(13,"app-artifact-tab",36),F(),S(14,"mat-tab"),_A(15,OjA,2,0,"ng-template",32),S(16,"app-session-tab",37),hA("sessionSelected",function(n){RA(A);let o=O();return xA(o.updateWithSelectedSession(n))})("sessionReloaded",function(n){RA(A);let o=O();return xA(o.updateSessionState(n))}),F()(),_A(17,jjA,3,3,"mat-tab"),F()()}if(t&2){let A=O();G(4),yA("traceData",A.traceData),G(3),yA("eventsMap",A.eventData)("traceData",A.traceData),G(3),yA("sessionState",A.currentSessionState),G(3),yA("artifacts",A.artifacts),G(3),yA("userId",A.userId)("appName",A.appName)("sessionId",A.sessionId),G(),GA(A.shouldShowEvalTab()?17:-1)}}function VjA(t,e){if(t&1){let A=be();S(0,"div",52),hA("click",function(){RA(A);let n=O(2);return xA(n.openViewImageDialog(n.rawSvgString))}),F()}if(t&2){let A=O(2);yA("innerHtml",A.renderedEventGraph,RI)}}function ZjA(t,e){if(t&1){let A=be();S(0,"div",20)(1,"div",40)(2,"div",41)(3,"mat-paginator",42),hA("page",function(n){RA(A);let o=O();return xA(o.handlePageEvent(n))}),F(),S(4,"button",43)(5,"mat-icon",44),hA("click",function(){RA(A);let n=O();return xA(n.closeSelectedEvent())}),AA(6,"close"),F()()()(),S(7,"div")(8,"mat-tab-group")(9,"mat-tab",45)(10,"div",46),_A(11,VjA,1,1,"div",47),F(),S(12,"div",48),JA(13,"ngx-json-viewer",49),F()(),S(14,"mat-tab",50)(15,"div",48),JA(16,"ngx-json-viewer",49),F()(),S(17,"mat-tab",51)(18,"div",48),JA(19,"ngx-json-viewer",49),F()()()()()}if(t&2){let A=O();G(3),yA("length",A.eventData.size)("pageSize",1)("pageIndex",A.selectedEventIndex),G(8),GA(A.renderedEventGraph?11:-1),G(2),yA("json",A.selectedEvent),G(3),yA("json",A.llmRequest),G(3),yA("json",A.llmResponse)}}function WjA(t,e){if(t&1){let A=be();S(0,"span",54),hA("click",function(){RA(A);let n=O(2);return xA(n.toggleSidePanel())}),AA(1,"left_panel_open"),F()}}function XjA(t,e){if(t&1){let A=be();S(0,"button",59),hA("click",function(){RA(A);let n=O(3);return xA(n.cancelEditEvalCase())}),AA(1,"Cancel"),F(),S(2,"button",60),hA("click",function(){RA(A);let n=O(3);return xA(n.saveEvalCase())}),AA(3," Save "),F()}if(t&2){let A=O(3);G(2),yA("disabled",!A.hasEvalCaseChanged()||A.isEvalCaseEditing())}}function $jA(t,e){if(t&1){let A=be();S(0,"span",61),hA("click",function(){RA(A);let n=O(3);return xA(n.editEvalCase())}),AA(1," edit "),F(),S(2,"span",62),hA("click",function(){RA(A);let n=O(3);return xA(n.deleteEvalCase())}),AA(3," delete "),F()}}function AqA(t,e){if(t&1&&(S(0,"div",55)(1,"div",56),AA(2,"Eval Case ID"),F(),S(3,"div",57),AA(4),F()(),S(5,"div",58),_A(6,XjA,4,1)(7,$jA,4,0),F()),t&2){let A=O(2);G(4),Gt(A.evalCase.evalId),G(2),GA(A.isEvalEditMode()?6:7)}}function eqA(t,e){if(t&1){let A=be();S(0,"span",71),hA("click",function(){RA(A);let n=O(3);return xA(n.importSession())}),AA(1," upload "),F()}}function tqA(t,e){if(t&1){let A=be();S(0,"div",55)(1,"div",56),AA(2,"Session ID"),F(),S(3,"div",57),AA(4),F()(),S(5,"div",58)(6,"div",63)(7,"mat-slide-toggle",64),hA("change",function(){RA(A);let n=O(2);return xA(n.toggleSse())}),AA(8," Token Streaming "),F()(),JA(9,"mat-divider",65),S(10,"div",66)(11,"div",67),hA("click",function(){RA(A);let n=O(2);return xA(n.onNewSessionClick())}),S(12,"mat-icon"),AA(13,"add"),F(),AA(14," New Session "),F(),S(15,"span",68),hA("click",function(){RA(A);let n=O(2);return xA(n.deleteSession(n.sessionId))}),AA(16," delete "),F(),S(17,"span",69),hA("click",function(){RA(A);let n=O(2);return xA(n.exportSession())}),AA(18," download "),F(),_A(19,eqA,2,0,"span",70),Za(20,"async"),F()()}if(t&2){let A=O(2);G(4),Gt(A.sessionId),G(3),yA("checked",A.enableSseIndicator()),G(2),yA("vertical",!0),G(10),GA(M2(20,4,A.importSessionEnabledObs)?19:-1)}}function iqA(t,e){if(t&1&&(S(0,"div",23),_A(1,WjA,2,0,"span",53)(2,AqA,8,2)(3,tqA,21,6),F()),t&2){let A=O();yA("ngClass",qr(3,RjA,A.isEvalEditMode())),G(),GA(A.showSidePanel?-1:1),G(),GA(A.evalCase?2:3)}}function nqA(t,e){t&1&&(S(0,"div",72)(1,"span"),AA(2,"Loading agents, please wait..."),F()())}function oqA(t,e){t&1&&(S(0,"span"),AA(1,"Welcome to ADK!"),JA(2,"br"),AA(3," Select an agent on the left to begin with."),F())}function rqA(t,e){if(t&1&&(AA(0," Error message: "),JA(1,"br"),S(2,"pre",74),AA(3),F()),t&2){let A=O(4);G(3),Gt(A.loadingError())}}function sqA(t,e){t&1&&(S(0,"pre",73),AA(1,"Warning: No agents found in current folder."),F())}function aqA(t,e){if(t&1&&(S(0,"div"),AA(1," Failed to load agents. To get started, run "),S(2,"pre"),AA(3,"adk web"),F(),AA(4," in the folder that contains the agents."),JA(5,"br"),_A(6,rqA,4,1)(7,sqA,2,0,"pre",73),F()),t&2){let A=O(3);G(6),GA(A.loadingError()?6:7)}}function cqA(t,e){if(t&1&&(S(0,"div",72),_A(1,oqA,4,0,"span"),Za(2,"async"),_A(3,aqA,8,1,"div"),F()),t&2){let A=O(2);G(),GA((M2(2,1,A.apps$)||cH(3,xjA)).length>0?1:3)}}function lqA(t,e){if(t&1&&_A(0,nqA,3,0,"div",72)(1,cqA,4,4,"div",72),t&2){let A=O();GA(A.isLoadingApps()?0:1)}}function gqA(t,e){if(t&1){let A=be();S(0,"button",75),hA("click",function(){RA(A);let n=O();return xA(n.openDialog())}),S(1,"mat-icon"),AA(2,"priority_high"),F()()}}function IqA(t,e){if(t&1){let A=be();S(0,"button",81),hA("click",function(){RA(A);let n=O().$index,o=O(2);return xA(o.clickEvent(n))}),S(1,"mat-icon",82),AA(2,"robot_2"),F()()}if(t&2){let A=O().$index,i=O(2);fo(i.customIconColorClass(A)),yA("matTooltip",i.getAgentNameFromEvent(A))}}function CqA(t,e){t&1&&JA(0,"mat-progress-bar",83)}function dqA(t,e){if(t&1&&JA(0,"img",88),t&2){let A=O().$implicit;yA("src",A.url,ja)}}function BqA(t,e){if(t&1&&(S(0,"mat-icon"),AA(1,"insert_drive_file"),F(),S(2,"a",89),AA(3),F()),t&2){let A=O().$implicit;G(2),yA("href",A.url,ja),G(),Gt(A.file.name)}}function EqA(t,e){if(t&1&&(S(0,"div",87),_A(1,dqA,1,1,"img",88)(2,BqA,4,2),F()),t&2){let A=e.$implicit;G(),GA(A.file.type.startsWith("image/")?1:-1),G(),GA(A.file.type.startsWith("image/")?-1:2)}}function QqA(t,e){if(t&1&&(S(0,"div",84),Dn(1,EqA,3,2,"div",87,to),F()),t&2){let A=O(2).$implicit;G(),yn(A.attachments)}}function hqA(t,e){t&1&&(S(0,"div",85),AA(1,"Thought"),F())}function uqA(t,e){if(t&1){let A=be();S(0,"div",90)(1,"textarea",92,3),da("ngModelChange",function(n){RA(A);let o=O(5);return Va(o.userEditEvalCaseMessage,n)||(o.userEditEvalCaseMessage=n),xA(n)}),hA("keydown",function(n){RA(A);let o=O(3).$implicit,r=O(2);return xA(r.handleKeydown(n,o))}),F(),S(3,"div",93)(4,"span",94),hA("click",function(){RA(A);let n=O(3).$implicit,o=O(2);return xA(o.cancelEditMessage(n))}),AA(5," close "),F(),S(6,"span",95),hA("click",function(){RA(A);let n=O(3).$implicit,o=O(2);return xA(o.saveEditMessage(n))}),AA(7," check "),F()()()}if(t&2){let A=O(5);G(),Ca("ngModel",A.userEditEvalCaseMessage)}}function fqA(t,e){if(t&1&&JA(0,"markdown",91),t&2){let A=O(3).$implicit;yA("data",A.text)("ngStyle",b2(2,_jA,A.thought?"italic":"normal",A.thought?"#9aa0a6":"white"))}}function mqA(t,e){if(t&1&&_A(0,uqA,8,1,"div",90)(1,fqA,1,5,"markdown",91),t&2){let A=O(2).$implicit;GA(A.isEditing?0:1)}}function pqA(t,e){if(t&1&&(S(0,"div"),JA(1,"div",96),F()),t&2){let A=O(2).$implicit,i=O(2);G(),yA("innerHTML",i.renderGooglerSearch(A.renderedContent),RI)}}function wqA(t,e){if(t&1&&(S(0,"code"),AA(1),F()),t&2){let A=O(2).$implicit;G(),Et(" ",A.executableCode.code," ")}}function DqA(t,e){if(t&1&&(S(0,"div")(1,"div"),AA(2),F(),S(3,"div"),AA(4),F()()),t&2){let A=O(2).$implicit;G(2),Et("Outcome: ",A.codeExecutionResult.outcome,""),G(2),Et("Output: ",A.codeExecutionResult.output,"")}}function yqA(t,e){if(t&1){let A=be();S(0,"div",97)(1,"img",98),hA("click",function(){RA(A);let n=O(4).$implicit,o=O(2);return xA(o.openViewImageDialog(n.inlineData.data))}),F()()}if(t&2){let A=O(4).$implicit;G(),yA("src",A.inlineData.data,ja)}}function vqA(t,e){if(t&1&&(S(0,"div"),JA(1,"app-audio-player",99),F()),t&2){let A=O(4).$implicit;G(),yA("base64data",A.inlineData.data)}}function bqA(t,e){if(t&1){let A=be();S(0,"div")(1,"div",100)(2,"mat-icon"),AA(3,"description"),F(),S(4,"button",101),hA("click",function(){RA(A);let n=O(4).$implicit,o=O(2);return xA(o.openBase64InNewTab(n.inlineData.data,n.inlineData.mimeType))}),AA(5),F()()()}if(t&2){let A=O(4).$implicit;G(5),Et(" ",A.inlineData.name," ")}}function MqA(t,e){if(t&1){let A=be();S(0,"div")(1,"button",101),hA("click",function(){RA(A);let n=O(4).$implicit,o=O(2);return xA(o.openBase64InNewTab(n.inlineData.data,n.inlineData.mimeType))}),AA(2),F()()}if(t&2){let A=O(4).$implicit;G(2),Et(" ",A.inlineData.name," ")}}function kqA(t,e){if(t&1&&(S(0,"div")(1,"div"),_A(2,yqA,2,1,"div",97)(3,vqA,2,1,"div")(4,bqA,6,1,"div")(5,MqA,3,1,"div"),F()()),t&2){let A,i=O(3).$implicit,n=O(2);G(2),GA((A=i.inlineData.mediaType)===n.MediaType.IMAGE?2:A===n.MediaType.AUDIO?3:A===n.MediaType.TEXT?4:5)}}function SqA(t,e){if(t&1){let A=be();S(0,"div")(1,"img",102),hA("click",function(){RA(A);let n=O(4).$implicit,o=O(2);return xA(o.openViewImageDialog(n.inlineData.data))}),F()()}if(t&2){let A=O(4).$implicit;G(),yA("src",A.inlineData.data,ja)}}function RqA(t,e){if(t&1&&(S(0,"div",87)(1,"mat-icon"),AA(2,"insert_drive_file"),F(),S(3,"a",89),AA(4),F()()),t&2){let A=O(4).$implicit;G(3),yA("href",A.inlineData.data,ja),G(),Gt(A.inlineData.displayName)}}function xqA(t,e){if(t&1&&(S(0,"div"),_A(1,SqA,2,1,"div")(2,RqA,5,2,"div",87),F()),t&2){let A=O(3).$implicit;G(),GA(A.inlineData.mimeType.startsWith("image/")?1:2)}}function LqA(t,e){if(t&1&&_A(0,kqA,6,1,"div")(1,xqA,3,1,"div"),t&2){let A=O(2).$implicit;GA(A.role==="bot"?0:1)}}function FqA(t,e){if(t&1&&(S(0,"div",105)(1,"div",106),AA(2,"Actual tool uses:"),F(),JA(3,"ngx-json-viewer",49),F(),S(4,"div",107)(5,"div",108),AA(6," Expected tool uses: "),F(),JA(7,"ngx-json-viewer",49),F()),t&2){let A=O(3).$implicit;G(3),yA("json",A.actualInvocationToolUses),G(4),yA("json",A.expectedInvocationToolUses)}}function NqA(t,e){if(t&1&&(S(0,"div",105)(1,"div",106),AA(2,"Actual response:"),F(),S(3,"div"),AA(4),F()(),S(5,"div",107)(6,"div",108),AA(7,"Expected response:"),F(),S(8,"div"),AA(9),F()()),t&2){let A=O(3).$implicit;G(4),Gt(A.actualFinalResponse),G(5),Gt(A.expectedFinalResponse)}}function _qA(t,e){if(t&1&&(S(0,"div",104)(1,"span",109),AA(2),F(),S(3,"span",110),AA(4),F()()),t&2){let A=O(3).$implicit;G(2),Et("Match score: ",A.evalScore,""),G(2),Et("Threshold: ",A.evalThreshold,"")}}function GqA(t,e){if(t&1&&(S(0,"div",86)(1,"div",103),_A(2,FqA,8,2)(3,NqA,10,2),F(),_A(4,_qA,5,2,"div",104),F()),t&2){let A=O(2).$implicit;G(2),GA(A.actualInvocationToolUses?2:A.actualFinalResponse?3:-1),G(2),GA(A.evalScore!==void 0&&A.evalThreshold!==void 0?4:-1)}}function UqA(t,e){if(t&1&&(S(0,"mat-card",78),_A(1,CqA,1,0,"mat-progress-bar",83)(2,QqA,3,0,"div",84),S(3,"div"),_A(4,hqA,2,0,"div",85),S(5,"div"),_A(6,mqA,2,1),F(),_A(7,pqA,2,1,"div"),F(),_A(8,wqA,2,1,"code")(9,DqA,5,2,"div")(10,LqA,2,1)(11,GqA,5,2,"div",86),F()),t&2){let A=O(),i=A.$implicit,n=A.$index,o=O(2);yA("ngClass",qr(11,NjA,i.evalStatus===2))("ngStyle",qr(13,IU,o.shouldMessageHighlighted(n)?"rgb(15, 82, 35)":"")),G(),GA(i.isLoading?1:-1),G(),GA(i.attachments?2:-1),G(2),GA(i.thought?4:-1),G(2),GA(i.text?6:-1),G(),GA(i.renderedContent?7:-1),G(),GA(i.executableCode?8:-1),G(),GA(i.codeExecutionResult?9:-1),G(),GA(i.inlineData?10:-1),G(),GA(i.failedMetric&&i.evalStatus===2?11:-1)}}function KqA(t,e){if(t&1){let A=be();S(0,"button",111),hA("click",function(){RA(A);let n=O().$index,o=O(2);return xA(o.clickEvent(n))}),S(1,"mat-icon"),AA(2,"bolt"),F(),AA(3),F()}if(t&2){let A=O(),i=A.$implicit,n=A.$index,o=O(2);yA("ngClass",qr(2,LlA,o.shouldMessageHighlighted(n))),G(3),Et(" ",i.functionCall.name," ")}}function YqA(t,e){if(t&1){let A=be();S(0,"button",111),hA("click",function(){RA(A);let n=O().$index,o=O(2);return xA(o.clickEvent(n))}),S(1,"mat-icon"),AA(2,"check"),F(),AA(3),F()}if(t&2){let A=O(),i=A.$implicit,n=A.$index,o=O(2);yA("ngClass",qr(2,LlA,o.shouldMessageHighlighted(n))),G(3),Et(" ",i.functionResponse.name," ")}}function JqA(t,e){if(t&1){let A=be();S(0,"div")(1,"span",112),hA("click",function(){RA(A);let n=O(2).$implicit,o=O(2);return xA(o.editEvalCaseMessage(n))}),AA(2," edit "),F(),S(3,"span",113),hA("click",function(){RA(A);let n=O(2),o=n.$implicit,r=n.$index,s=O(2);return xA(s.deleteEvalCaseMessage(o,r))}),AA(4," delete "),F()()}if(t&2){let A=O(4);G(),yA("ngClass",qr(2,CU,A.isEvalCaseEditing())),G(2),yA("ngClass",qr(4,CU,A.isEvalCaseEditing()))}}function TqA(t,e){if(t&1){let A=be();S(0,"div")(1,"span",114),hA("click",function(){RA(A);let n=O(2).$implicit,o=O(2);return xA(o.editFunctionArgs(n))}),AA(2," edit "),F()()}if(t&2){let A=O(4);G(),yA("ngClass",qr(1,CU,A.isEvalCaseEditing()))}}function HqA(t,e){if(t&1&&(_A(0,JqA,5,6,"div"),Za(1,"async"),_A(2,TqA,3,3,"div")),t&2){let A=O().$implicit,i=O(2);GA(A.text?0:M2(1,1,i.isEditFunctionArgsEnabledObs)&&A.functionCall?2:-1)}}function zqA(t,e){t&1&&(S(0,"button",43)(1,"mat-icon"),AA(2,"person"),F()())}function OqA(t,e){if(t&1&&(S(0,"div",76),_A(1,IqA,3,3,"button",77)(2,UqA,12,15,"mat-card",78)(3,KqA,4,4,"button",79)(4,YqA,4,4,"button",79),S(5,"div",76)(6,"span",80),AA(7),F(),S(8,"span"),AA(9),F()(),_A(10,HqA,3,3)(11,zqA,3,0,"button",43),F()),t&2){let A=e.$implicit,i=O(2);yA("ngClass",b2(10,LjA,A.role==="user",A.role==="bot")),G(),GA(A.role==="bot"?1:-1),G(),GA(!A.functionCall&&!A.functionResponse?2:-1),G(),GA(A.functionCall?3:-1),G(),GA(A.functionResponse?4:-1),G(),yA("ngClass",b2(13,FjA,A.evalStatus===1,A.evalStatus===2)),G(2),Gt(A.evalStatus===1?"check":A.evalStatus===2?"close":""),G(2),Gt(A.evalStatus===1?"Pass":A.evalStatus===2?"Fail":""),G(),GA(i.evalCase&&A.role==="bot"&&i.isEvalEditMode()?10:-1),G(),GA(A.role==="user"?11:-1)}}function PqA(t,e){if(t&1&&(S(0,"div",26,1),JA(2,"div",null,2),Dn(4,OqA,12,16,"div",76,to),F()),t&2){let A=O();G(4),yn(A.messages)}}function jqA(t,e){if(t&1){let A=be();S(0,"div",125),JA(1,"img",126),S(2,"button",127),hA("click",function(){RA(A);let n=O().$index,o=O(3);return xA(o.removeFile(n))}),S(3,"mat-icon",128),AA(4,"close"),F()()()}if(t&2){let A=O().$implicit;G(),yA("src",A.url,ja)}}function qqA(t,e){if(t&1){let A=be();S(0,"div",124)(1,"button",127),hA("click",function(){RA(A);let n=O().$index,o=O(3);return xA(o.removeFile(n))}),S(2,"mat-icon",128),AA(3,"close"),F()(),S(4,"div",129)(5,"mat-icon"),AA(6,"insert_drive_file"),F(),S(7,"span"),AA(8),F()()()}if(t&2){let A=O().$implicit;G(8),Gt(A.file.name)}}function VqA(t,e){if(t&1&&(S(0,"div"),_A(1,jqA,5,1,"div",125)(2,qqA,9,1,"div",124),F()),t&2){let A=e.$implicit;G(),GA(A.file.type.startsWith("image/")?1:A.file.type.startsWith("image/")?-1:2)}}function ZqA(t,e){if(t&1){let A=be();S(0,"div",124)(1,"button",127),hA("click",function(){RA(A);let n=O(3);return xA(n.removeStateUpdate())}),S(2,"mat-icon",128),AA(3,"close"),F()(),S(4,"div",129)(5,"span"),AA(6,"Updated session state"),F()()()}}function WqA(t,e){if(t&1&&(S(0,"div",117),Dn(1,VqA,3,1,"div",null,to),_A(3,ZqA,7,0,"div",124),F()),t&2){let A=O(2);G(),yn(A.selectedFiles),G(2),GA(A.updatedSessionState()?3:-1)}}function XqA(t,e){if(t&1){let A=be();S(0,"div",27)(1,"input",115,4),hA("change",function(n){RA(A);let o=O();return xA(o.onFileSelect(n))}),F(),S(3,"mat-form-field",116),_A(4,WqA,4,1,"div",117),S(5,"textarea",118),da("ngModelChange",function(n){RA(A);let o=O();return Va(o.userInput,n)||(o.userInput=n),xA(n)}),hA("keydown.enter",function(n){RA(A);let o=O();return xA(o.sendMessage(n))}),F(),S(6,"div",119)(7,"div")(8,"button",120),hA("click",function(){RA(A);let n=cr(2);return xA(n.click())}),S(9,"mat-icon"),AA(10,"attach_file"),F()(),S(11,"button",121)(12,"mat-icon"),AA(13,"more_vert"),F()(),S(14,"mat-menu",null,5)(16,"span",122),hA("click",function(){RA(A);let n=O();return xA(n.updateState())}),AA(17," Update state "),F()()(),S(18,"div")(19,"button",123),hA("click",function(){RA(A);let n=O();return xA(n.toggleAudioRecording())}),S(20,"mat-icon"),AA(21,"mic"),F()(),S(22,"button",123),hA("click",function(){RA(A);let n=O();return xA(n.toggleVideoRecording())}),S(23,"mat-icon"),AA(24,"videocam"),F()()()()()()}if(t&2){let A=cr(15),i=O();G(4),GA(i.selectedFiles.length&&i.appName!=""||i.updatedSessionState()?4:-1),G(),Ca("ngModel",i.userInput),G(6),yA("matMenuTriggerFor",A),G(8),yA("ngStyle",qr(7,IU,i.isAudioRecording?"rgb(234, 67, 53)":"rgb(51, 53, 55)"))("matTooltip",i.isAudioRecording?"Turn off microphone":"Use microphone"),G(3),yA("ngStyle",qr(9,IU,i.isVideoRecording?"rgb(234, 67, 53)":"rgb(51, 53, 55)"))("matTooltip",i.isVideoRecording?"Turn off camera":"Use camera")}}function $qA(t,e){if(t&1){let A=be();S(0,"div",28,6),JA(2,"div",130),S(3,"app-trace-event",131),hA("panelClosed",function(){RA(A);let n=O();return xA(n.closeTraceEventDetailPanel())}),F()()}if(t&2){let A=O();G(3),yA("userId",A.userId)("appName",A.appName)("sessionId",A.sessionId)}}var lU="root_agent";function AVA(t){for(t=t.replace(/-/g,"+").replace(/_/g,"/");t.length%4!==0;)t+="=";return t}var dU=class extends iC{nextPageLabel="Next Event";previousPageLabel="Previous Event";firstPageLabel="First Event";lastPageLabel="Last Event";getRangeLabel=(e,A,i)=>i===0?`Event 0 of ${i}`:(i=Math.max(i,0),`Event ${e*A+1} of ${i}`)},xlA="Restarting bidirectional streaming is not currently supported. Please refresh the page or start a new session.",Ff=class t{constructor(e,A,i,n,o,r,s,a,c,l,I,C,d,B,E){this.sanitizer=e;this.sessionService=A;this.artifactService=i;this.audioService=n;this.webSocketService=o;this.videoService=r;this.dialog=s;this.eventService=a;this.route=c;this.downloadService=l;this.evalService=I;this.traceService=C;this.location=d;this.renderer=B;this.document=E}videoContainer;sideDrawer;eventTabComponent;sessionTab;evalTab;scrollContainer;textarea;bottomPanelRef;_snackBar=f(UP);shouldShowEvalTab=Jo(!0);enableSseIndicator=Jo(!1);isChatMode=Jo(!0);isEvalCaseEditing=Jo(!1);hasEvalCaseChanged=Jo(!1);isEvalEditMode=Jo(!1);videoElement;currentMessage="";messages=[];lastTextChunk="";streamingTextMessage=null;latestThought="";artifacts=[];userInput="";userEditEvalCaseMessage="";userId="user";appName="";sessionId="";evalCase=null;updatedEvalCase=null;evalSetId="";isAudioRecording=!1;isVideoRecording=!1;longRunningEvents=[];functionCallEventId="";redirectUri=vs.getBaseUrlWithoutPath();showSidePanel=!0;useSse=!1;currentSessionState={};root_agent=lU;updatedSessionState=Jo(null);messagesSubject=new Mi([]);streamingTextMessageSubject=new Mi(null);scrollInterruptedSubject=new Mi(!0);isModelThinkingSubject=new Mi(!1);sessionHasUsedBidi=new Set;eventData=new Map;traceData=[];eventMessageIndexArray=[];renderedEventGraph;rawSvgString=null;selectedEvent=void 0;selectedEventIndex=void 0;llmRequest=void 0;llmResponse=void 0;llmRequestKey="gcp.vertex.agent.llm_request";llmResponseKey="gcp.vertex.agent.llm_response";getMediaTypeFromMimetype=M8;selectedFiles=[];previousMessageCount=0;openBase64InNewTab=cS;MediaType=Ou;router=f(Ig);activatedRoute=f(ha);selectedAppControl=new _I("",{nonNullable:!0});changeDetectorRef=f(It);agentService=f(H2);isLoadingApps=Jo(!1);loadingError=Jo("");apps$=Me([]).pipe(Qo(()=>{this.isLoadingApps.set(!0),this.selectedAppControl.disable()}),jn(()=>this.agentService.listApps().pipe(mr(e=>(this.loadingError.set(e.message),Me(void 0))))),On(1),Qo(e=>{this.isLoadingApps.set(!1),this.selectedAppControl.enable(),e?.length==1&&this.router.navigate([],{relativeTo:this.route,queryParams:{app:e[0]}})}),a0());featureFlagService=f(KB);importSessionEnabledObs=this.featureFlagService.isImportSessionEnabled();isEditFunctionArgsEnabledObs=this.featureFlagService.isEditFunctionArgsEnabled();isSessionUrlEnabledObs=this.featureFlagService.isSessionUrlEnabled();bottomPanelVisible=!1;hoveredEventMessageIndices=[];ngOnInit(){if(this.syncSelectedAppFromUrl(),this.updateSelectedAppUrl(),this.webSocketService.onCloseReason().subscribe(i=>{let n=`Please check server log for full details: 
`+i;this.openSnackBar(n,"OK")}),new URL(window.location.href).searchParams.has("code")){let i=window.location.href;window.opener?.postMessage({authResponseUrl:i},window.origin),window.close()}this.agentService.getApp().subscribe(i=>{this.appName=i}),Js([this.agentService.getLoadingState(),this.isModelThinkingSubject]).subscribe(([i,n])=>{let o=this.messages[this.messages.length-1];i?!o?.isLoading&&!this.streamingTextMessage&&(this.messages.push({role:"bot",isLoading:!0}),this.messagesSubject.next(this.messages)):o?.isLoading&&!n&&(this.messages.pop(),this.messagesSubject.next(this.messages),this.changeDetectorRef.detectChanges())}),Js([this.messagesSubject,this.scrollInterruptedSubject,this.streamingTextMessageSubject]).subscribe(([i,n,o])=>{n||setTimeout(()=>{this.scrollToBottom()},100)}),this.traceService.selectedTraceRow$.subscribe(i=>{let n=i?.attributes["gcp.vertex.agent.event_id"];n&&this.eventData.has(n)?this.bottomPanelVisible=!0:this.bottomPanelVisible=!1}),this.traceService.hoveredMessageIndicies$.subscribe(i=>this.hoveredEventMessageIndices=i)}ngAfterViewInit(){this.showSidePanel=!0,this.sideDrawer.open()}scrollToBottom(){setTimeout(()=>{this.scrollContainer.nativeElement.scrollTo({top:this.scrollContainer.nativeElement.scrollHeight,behavior:"smooth"})})}selectApp(e){e!=this.appName&&(this.agentService.setApp(e),this.isSessionUrlEnabledObs.subscribe(A=>{let i=this.activatedRoute.snapshot.queryParams.session;if(!A||!i){this.createSessionAndReset();return}i&&this.sessionService.getSession(this.userId,this.appName,i).pipe(On(1),mr(n=>(this.openSnackBar("Cannot find specified session. Creating a new one.","OK"),this.createSessionAndReset(),Me(null)))).subscribe(n=>{n&&this.updateWithSelectedSession(n)})}))}createSessionAndReset(){this.createSession(),this.eventData=new Map,this.eventMessageIndexArray=[],this.messages=[],this.artifacts=[],this.userInput="",this.longRunningEvents=[]}createSession(){this.sessionService.createSession(this.userId,this.appName).subscribe(e=>{this.currentSessionState=e.state,this.sessionId=e.id,this.sessionTab.refreshSession(),this.isSessionUrlEnabledObs.subscribe(A=>{A&&this.updateSelectedSessionUrl()})})}sendMessage(e){return Ao(this,null,function*(){if(this.messages.length===0&&(this.scrollContainer.nativeElement.addEventListener("wheel",()=>{this.scrollInterruptedSubject.next(!0)}),this.scrollContainer.nativeElement.addEventListener("touchmove",()=>{this.scrollInterruptedSubject.next(!0)})),this.scrollInterruptedSubject.next(!1),e.preventDefault(),!this.userInput.trim()&&this.selectedFiles.length<=0||e instanceof KeyboardEvent&&(e.isComposing||e.keyCode===229))return;if(this.userInput.trim()&&(this.messages.push({role:"user",text:this.userInput}),this.messagesSubject.next(this.messages)),this.selectedFiles.length>0){let n=this.selectedFiles.map(o=>({file:o.file,url:o.url}));this.messages.push({role:"user",attachments:n}),this.messagesSubject.next(this.messages)}let A={appName:this.appName,userId:this.userId,sessionId:this.sessionId,newMessage:{role:"user",parts:yield this.getUserMessageParts()},streaming:this.useSse,stateDelta:this.updatedSessionState()};this.selectedFiles=[];let i=this.eventMessageIndexArray.length-1;this.streamingTextMessage=null,this.agentService.runSse(A).subscribe({next:n=>Ao(this,null,function*(){if(n.startsWith('{"error"')){this.openSnackBar(n,"OK");return}let o=JSON.parse(n);if(o.error){this.openSnackBar(o.error,"OK");return}if(o.content)for(let r of o.content.parts)i+=1,this.processPart(o,r,i),this.traceService.setEventData(this.eventData);else o.errorMessage&&this.processErrorMessage(o,i);this.changeDetectorRef.detectChanges()}),error:n=>console.error("SSE error:",n),complete:()=>{this.streamingTextMessage=null,this.sessionTab.reloadSession(this.sessionId),this.eventService.getTrace(this.sessionId).pipe(mr(n=>n.status===404?Me(null):Me([]))).subscribe(n=>{this.traceData=n,this.changeDetectorRef.detectChanges()}),this.traceService.setMessages(this.messages)}}),this.userInput="",this.updatedSessionState.set(null),this.changeDetectorRef.detectChanges()})}processErrorMessage(e,A){this.storeEvents(e,e,A),this.insertMessageBeforeLoadingMessage({text:e.errorMessage,role:"bot"})}processPart(e,A,i){let n=e.groundingMetadata?.searchEntryPoint?.renderedContent;if(A.text){this.isModelThinkingSubject.next(!1);let o=A.text;if(A.thought){if(o!==this.latestThought){this.storeEvents(A,e,i);let r={role:"bot",text:this.processThoughtText(o),thought:!0,eventId:e.id};this.insertMessageBeforeLoadingMessage(r)}this.latestThought=o}else if(this.streamingTextMessage){if(n&&(this.streamingTextMessage.renderedContent=e.groundingMetadata.searchEntryPoint.renderedContent),o==this.streamingTextMessage.text){this.storeEvents(A,e,i),this.eventMessageIndexArray[i]=o,this.streamingTextMessage=null;return}this.streamingTextMessage.text+=o,this.streamingTextMessageSubject.next(this.streamingTextMessage)}else if(this.streamingTextMessage={role:"bot",text:this.processThoughtText(o),thought:!!A.thought,eventId:e.id},n&&(this.streamingTextMessage.renderedContent=e.groundingMetadata.searchEntryPoint.renderedContent),this.insertMessageBeforeLoadingMessage(this.streamingTextMessage),!this.useSse){this.storeEvents(A,e,i),this.eventMessageIndexArray[i]=o,this.streamingTextMessage=null;return}}else A.thought?this.isModelThinkingSubject.next(!0):(this.isModelThinkingSubject.next(!1),this.storeEvents(A,e,i),this.storeMessage(A,e,i,e.author==="user"?"user":"bot"))}getUserMessageParts(){return Ao(this,null,function*(){let e=[];if(this.userInput.trim()&&e.push({text:`${this.userInput}`}),this.selectedFiles.length>0)for(let A of this.selectedFiles)e.push({inlineData:{displayName:A.file.name,data:yield this.readFileAsBytes(A.file),mimeType:A.file.type}});return e})}readFileAsBytes(e){return new Promise((A,i)=>{let n=new FileReader;n.onload=o=>{let r=o.target.result.split(",")[1];A(r)},n.onerror=i,n.readAsDataURL(e)})}updateRedirectUri(e,A){try{let i=new URL(e);return i.searchParams.set("redirect_uri",A),i.toString()}catch(i){return console.warn("Failed to update redirect URI: ",i),e}}storeMessage(e,A,i,n,o,r){if(A?.author&&this.createAgentIconColorClass(A.author),A?.longRunningToolIds&&A.longRunningToolIds.length>0){this.getAsyncFunctionsFromParts(A.longRunningToolIds,A.content.parts);let a=this.longRunningEvents[0];if(a.args.authConfig&&a.args.authConfig.exchangedAuthCredential&&a.args.authConfig.exchangedAuthCredential.oauth2){let c=a.args.authConfig.exchangedAuthCredential.oauth2.authUri,l=this.updateRedirectUri(c,this.redirectUri);this.openOAuthPopup(l).then(I=>{this.functionCallEventId=A.id,this.sendOAuthResponse(a,I,this.redirectUri)}).catch(I=>{console.error("OAuth Error:",I)})}else this.functionCallEventId=A.id}if(A?.actions&&A.actions.artifactDelta)for(let a in A.actions.artifactDelta)A.actions.artifactDelta.hasOwnProperty(a)&&this.renderArtifact(a,A.actions.artifactDelta[a]);A?.evalStatus&&this.isChatMode.set(!1);let s={role:n,evalStatus:A?.evalStatus,failedMetric:A?.failedMetric,evalScore:A?.evalScore,evalThreshold:A?.evalThreshold,actualInvocationToolUses:A?.actualInvocationToolUses,expectedInvocationToolUses:A?.expectedInvocationToolUses,actualFinalResponse:A?.actualFinalResponse,expectedFinalResponse:A?.expectedFinalResponse,invocationIndex:o!==void 0?o:void 0,finalResponsePartIndex:r?.finalResponsePartIndex!==void 0?r.finalResponsePartIndex:void 0,toolUseIndex:r?.toolUseIndex!==void 0?r.toolUseIndex:void 0};if(e.inlineData){let a=this.formatBase64Data(e.inlineData.data,e.inlineData.mimeType);s.inlineData={displayName:e.inlineData.displayName,data:a,mimeType:e.inlineData.mimeType},this.eventMessageIndexArray[i]=e.inlineData}else if(e.text)s.text=e.text,s.thought=!!e.thought,A?.groundingMetadata&&A.groundingMetadata.searchEntryPoint&&A.groundingMetadata.searchEntryPoint.renderedContent&&(s.renderedContent=A.groundingMetadata.searchEntryPoint.renderedContent),s.eventId=A?.id,this.eventMessageIndexArray[i]=e.text;else if(e.functionCall)s.functionCall=e.functionCall,s.eventId=A?.id,this.eventMessageIndexArray[i]=e.functionCall;else if(e.functionResponse)s.functionResponse=e.functionResponse,s.eventId=A?.id,this.eventMessageIndexArray[i]=e.functionResponse;else if(e.executableCode)s.executableCode=e.executableCode,this.eventMessageIndexArray[i]=e.executableCode;else if(e.codeExecutionResult&&(s.codeExecutionResult=e.codeExecutionResult,this.eventMessageIndexArray[i]=e.codeExecutionResult,A.actions&&A.actions.artifact_delta))for(let a in A.actions.artifact_delta)A.actions.artifact_delta.hasOwnProperty(a)&&this.renderArtifact(a,A.actions.artifact_delta[a]);Object.keys(e).length>0&&this.insertMessageBeforeLoadingMessage(s)}insertMessageBeforeLoadingMessage(e){this.messages[this.messages.length-1]?.isLoading?this.messages.splice(this.messages.length-1,0,e):this.messages.push(e),this.messagesSubject.next(this.messages)}formatBase64Data(e,A){let i=AVA(e);return`data:${A};base64,${i}`}renderArtifact(e,A){let i={role:"bot",inlineData:{data:"",mimeType:"image/png"}};this.insertMessageBeforeLoadingMessage(i);let n=this.messages.length-2;this.artifactService.getArtifactVersion(this.userId,this.appName,this.sessionId,e,A).subscribe(o=>{let r=o.inlineData.mimeType,s=this.formatBase64Data(o.inlineData.data,r),a=M8(r),c={name:this.createDefaultArtifactName(r),data:s,mimeType:r,mediaType:a};this.messages[n]={role:"bot",inlineData:c},this.artifacts=[...this.artifacts,{id:e,data:s,mimeType:r,versionId:A,mediaType:M8(r)}]})}storeEvents(e,A,i){let n="";e.text?n+="text:"+e.text:e.functionCall?n+="functionCall:"+e.functionCall.name:e.functionResponse?n+="functionResponse:"+e.functionResponse.name:e.executableCode?n+="executableCode:"+e.executableCode.code.slice(0,10):e.codeExecutionResult?n+="codeExecutionResult:"+e.codeExecutionResult.outcome:e.errorMessage&&(n+="errorMessage:"+e.errorMessage),A.title=n,this.eventData.set(A.id,A),this.eventData=new Map(this.eventData)}sendOAuthResponse(e,A,i){this.longRunningEvents.pop();let n={appName:this.appName,userId:this.userId,sessionId:this.sessionId,newMessage:{role:"user",parts:[]}};var o=structuredClone(e.args.authConfig);o.exchangedAuthCredential.oauth2.authResponseUri=A,o.exchangedAuthCredential.oauth2.redirectUri=i,n.functionCallEventId=this.functionCallEventId,n.newMessage.parts.push({function_response:{id:e.id,name:e.name,response:o}});let r=[];this.agentService.runSse(n).subscribe({next:s=>Ao(this,null,function*(){let a=JSON.parse(s);r.push(a)}),error:s=>console.error("SSE error:",s),complete:()=>{this.processRunSseResponse(r)}})}processRunSseResponse(e){let A=this.eventMessageIndexArray.length-1;for(let i of e)if(i.content)for(let n of i.content.parts)A+=1,this.processPart(i,n,A)}openDialog(){this.dialog.open(uf,{width:"600px",data:{event:this.longRunningEvents[0],appName:this.appName,userId:this.userId,sessionId:this.sessionId,functionCallEventId:this.functionCallEventId}}).afterClosed().subscribe(A=>{A&&(this.removeFinishedLongRunningEvents(A.events),this.processRunSseResponse(A.response))})}removeFinishedLongRunningEvents(e){let A=new Set(e.map(i=>i.id));this.longRunningEvents=this.longRunningEvents.filter(i=>!A.has(i.id))}getAgentNameFromEvent(e){let A=this.messages[e].eventId;return this.eventData.get(A)?.author??lU}customIconColorClass(e){let A=this.getAgentNameFromEvent(e);return A!==lU?`custom-icon-color-${(0,gU.default)(A).replace("#","")}`:""}createAgentIconColorClass(e){let A=(0,gU.default)(e),i=`custom-icon-color-${A.replace("#","")}`;this.injectCustomIconColorStyle(i,A)}clickEvent(e){let A=this.messages[e].eventId;this.sideDrawer.open(),this.showSidePanel=!0,this.selectedEvent=this.eventData.get(A),this.selectedEventIndex=this.getIndexOfKeyInMap(A),this.eventService.getEventTrace(this.selectedEvent.id).subscribe(i=>{this.llmRequest=JSON.parse(i[this.llmRequestKey]),this.llmResponse=JSON.parse(i[this.llmResponseKey])}),this.eventService.getEvent(this.userId,this.appName,this.sessionId,this.selectedEvent.id).subscribe(i=>Ao(this,null,function*(){if(!i.dotSrc){this.renderedEventGraph=void 0;return}let n=i.dotSrc,r=(yield Yu()).renderString(n,{format:"svg",engine:"dot"});this.rawSvgString=r,this.renderedEventGraph=this.sanitizer.bypassSecurityTrustHtml(r)}))}userMessagesLength(e){return this.messages.slice(0,e).filter(A=>A.role=="user").length}ngOnDestroy(){this.webSocketService.closeConnection()}onAppSelection(e){this.isAudioRecording&&(this.stopAudioRecording(),this.isAudioRecording=!1),this.isVideoRecording&&(this.stopVideoRecording(),this.isVideoRecording=!1),this.evalTab?.resetEvalResults(),this.traceData=[],this.bottomPanelVisible=!1}toggleAudioRecording(){this.isAudioRecording?this.stopAudioRecording():this.startAudioRecording()}startAudioRecording(){if(this.sessionHasUsedBidi.has(this.sessionId)){this.openSnackBar(xlA,"OK");return}this.isAudioRecording=!0;let e=window.location.protocol==="https:"?"wss":"ws";this.webSocketService.connect(`${e}://${vs.getWSServerUrl()}/run_live?app_name=${this.appName}&user_id=${this.userId}&session_id=${this.sessionId}`),this.audioService.startRecording(),this.messages.push({role:"user",text:"Speaking..."}),this.messages.push({role:"bot",text:"Speaking..."}),this.messagesSubject.next(this.messages),this.sessionHasUsedBidi.add(this.sessionId)}stopAudioRecording(){this.audioService.stopRecording(),this.webSocketService.closeConnection(),this.isAudioRecording=!1}toggleVideoRecording(){this.isVideoRecording?this.stopVideoRecording():this.startVideoRecording()}startVideoRecording(){if(this.sessionHasUsedBidi.has(this.sessionId)){this.openSnackBar(xlA,"OK");return}this.isVideoRecording=!0;let e=window.location.protocol==="https:"?"wss":"ws";this.webSocketService.connect(`${e}://${vs.getWSServerUrl()}/run_live?app_name=${this.appName}&user_id=${this.userId}&session_id=${this.sessionId}`),this.videoService.startRecording(this.videoContainer),this.audioService.startRecording(),this.messages.push({role:"user",text:"Speaking..."}),this.messagesSubject.next(this.messages),this.sessionHasUsedBidi.add(this.sessionId)}stopVideoRecording(){this.audioService.stopRecording(),this.videoService.stopRecording(this.videoContainer),this.webSocketService.closeConnection(),this.isVideoRecording=!1}getAsyncFunctionsFromParts(e,A){for(let i of A)i.functionCall&&e.includes(i.functionCall.id)&&this.longRunningEvents.push(i.functionCall)}openOAuthPopup(e){return new Promise((A,i)=>{if(!window.open(e,"oauthPopup","width=600,height=700")){i("Popup blocked!");return}let o=r=>{if(r.origin!==window.location.origin)return;let{authResponseUrl:s}=r.data;s?(A(s),window.removeEventListener("message",o)):console.log("OAuth failed",r)};window.addEventListener("message",o)})}toggleSidePanel(){this.showSidePanel?this.sideDrawer.close():this.sideDrawer.open(),this.showSidePanel=!this.showSidePanel}handleTabChange(e){this.isChatMode()||(this.resetEditEvalCaseVars(),this.handleReturnToSession(!0))}handleShouldShowEvalTab(e){this.shouldShowEvalTab.set(e)}handleReturnToSession(e){this.sessionTab.getSession(this.sessionId),this.evalTab.resetEvalCase(),this.isChatMode.set(!0)}handleEvalNotInstalled(e){e&&this.openSnackBar(e,"OK")}resetEventsAndMessages(){this.eventData.clear(),this.eventMessageIndexArray=[],this.messages=[],this.messagesSubject.next(this.messages),this.artifacts=[]}updateWithSelectedSession(e){if(!e||!e.id||!e.events||!e.state)return;this.traceService.resetTraceService(),this.sessionId=e.id,this.currentSessionState=e.state,this.evalCase=null,this.isChatMode.set(!0),this.isSessionUrlEnabledObs.subscribe(i=>{i&&this.updateSelectedSessionUrl()}),this.resetEventsAndMessages();let A=0;e.events.forEach(i=>{i.content?.parts?.forEach(n=>{this.storeMessage(n,i,A,i.author==="user"?"user":"bot"),A+=1,i.author&&i.author!=="user"&&this.storeEvents(n,i,A)})}),this.eventService.getTrace(this.sessionId).subscribe(i=>{this.traceData=i,this.traceService.setEventData(this.eventData),this.traceService.setMessages(this.messages)}),this.bottomPanelVisible=!1}updateWithSelectedEvalCase(e){this.evalCase=e,this.isChatMode.set(!1),this.resetEventsAndMessages();let A=0,i=0;for(let n of e.conversation){if(n.userContent?.parts)for(let o of n.userContent.parts)this.storeMessage(o,null,A,"user"),A++;if(n.intermediateData?.toolUses){let o=0;for(let r of n.intermediateData.toolUses){let s={functionCall:{name:r.name,args:r.args}};this.storeMessage(s,null,A,"bot",i,{toolUseIndex:o}),A++,o++;let a={functionResponse:{name:r.name}};this.storeMessage(a,null,A,"bot"),A++}}if(n.finalResponse?.parts){let o=0;for(let r of n.finalResponse.parts)this.storeMessage(r,null,A,"bot",i,{finalResponsePartIndex:o}),A++,o++}i++}}updateSelectedEvalSetId(e){this.evalSetId=e}editEvalCaseMessage(e){this.isEvalCaseEditing.set(!0),this.userEditEvalCaseMessage=e.text,e.isEditing=!0,setTimeout(()=>{this.textarea?.nativeElement.focus();let A=this.textarea?.nativeElement.value.length;e.text.charAt(A-1)===`
`&&A--,this.textarea?.nativeElement.setSelectionRange(A,A)},0)}editFunctionArgs(e){this.isEvalCaseEditing.set(!0),this.dialog.open(vQ,{maxWidth:"90vw",maxHeight:"90vh",data:{dialogHeader:"Edit function arguments",functionName:e.functionCall.name,jsonContent:e.functionCall.args}}).afterClosed().subscribe(i=>{this.isEvalCaseEditing.set(!1),i&&(this.hasEvalCaseChanged.set(!0),e.functionCall.args=i,this.updatedEvalCase=structuredClone(this.evalCase),this.updatedEvalCase.conversation[e.invocationIndex].intermediateData.toolUses[e.toolUseIndex].args=i)})}saveEvalCase(){this.evalService.updateEvalCase(this.appName,this.evalSetId,this.updatedEvalCase.evalId,this.updatedEvalCase).subscribe(e=>{this.openSnackBar("Eval case updated","OK"),this.resetEditEvalCaseVars()})}cancelEditEvalCase(){this.resetEditEvalCaseVars(),this.updateWithSelectedEvalCase(this.evalCase)}resetEditEvalCaseVars(){this.hasEvalCaseChanged.set(!1),this.isEvalCaseEditing.set(!1),this.isEvalEditMode.set(!1),this.updatedEvalCase=null}cancelEditMessage(e){e.isEditing=!1,this.isEvalCaseEditing.set(!1)}saveEditMessage(e){this.hasEvalCaseChanged.set(!0),this.isEvalCaseEditing.set(!1),e.isEditing=!1,e.text=this.userEditEvalCaseMessage?this.userEditEvalCaseMessage:" ",this.updatedEvalCase=structuredClone(this.evalCase),this.updatedEvalCase.conversation[e.invocationIndex].finalResponse.parts[e.finalResponsePartIndex]={text:this.userEditEvalCaseMessage},this.userEditEvalCaseMessage=""}handleKeydown(e,A){e.key==="Enter"&&!e.shiftKey?(e.preventDefault(),this.saveEditMessage(A)):e.key==="Escape"&&this.cancelEditMessage(A)}deleteEvalCaseMessage(e,A){this.hasEvalCaseChanged.set(!0),this.messages.splice(A,1),this.messagesSubject.next(this.messages),this.updatedEvalCase=structuredClone(this.evalCase),this.updatedEvalCase.conversation[e.invocationIndex].finalResponse.parts.splice(e.finalResponsePartIndex,1)}editEvalCase(){this.isEvalEditMode.set(!0)}deleteEvalCase(){let e={title:"Confirm delete",message:`Are you sure you want to delete ${this.evalCase.evalId}?`,confirmButtonText:"Delete",cancelButtonText:"Cancel"};this.dialog.open(SQ,{width:"600px",data:e}).afterClosed().subscribe(i=>{i&&(this.evalTab.deleteEvalCase(this.evalCase.evalId),this.openSnackBar("Eval case deleted","OK"))})}updateSessionState(e){this.currentSessionState=e.state}onNewSessionClick(){this.createSession(),this.eventData.clear(),this.eventMessageIndexArray=[],this.messages=[],this.artifacts=[],this.traceData=[],this.bottomPanelVisible=!1,this.evalTab.showEvalHistory&&this.evalTab.toggleEvalHistoryButton()}onFileSelect(e){let A=e.target;if(A.files)for(let i=0;i<A.files.length;i++){let n=A.files[i],o=URL.createObjectURL(n);this.selectedFiles.push({file:n,url:o})}A.value=""}removeFile(e){URL.revokeObjectURL(this.selectedFiles[e].url),this.selectedFiles.splice(e,1)}toggleSse(){this.useSse=!this.useSse}selectEvent(e){this.selectedEvent=this.eventData.get(e),this.selectedEventIndex=this.getIndexOfKeyInMap(e),this.eventService.getEventTrace(this.selectedEvent.id).subscribe(A=>{this.llmRequest=JSON.parse(A[this.llmRequestKey]),this.llmResponse=JSON.parse(A[this.llmResponseKey])}),this.eventService.getEvent(this.userId,this.appName,this.sessionId,this.selectedEvent.id).subscribe(A=>Ao(this,null,function*(){if(!A.dotSrc){this.renderedEventGraph=void 0;return}let i=A.dotSrc,o=(yield Yu()).renderString(i,{format:"svg",engine:"dot"});this.rawSvgString=o,this.renderedEventGraph=this.sanitizer.bypassSecurityTrustHtml(o)}))}deleteSession(e){let A={title:"Confirm delete",message:`Are you sure you want to delete this session ${this.sessionId}?`,confirmButtonText:"Delete",cancelButtonText:"Cancel"};this.dialog.open(SQ,{width:"600px",data:A}).afterClosed().subscribe(n=>{n&&this.sessionService.deleteSession(this.userId,this.appName,e).subscribe(o=>{let r=this.sessionTab.refreshSession(e);r?this.sessionTab.getSession(r.id):window.location.reload()})})}syncSelectedAppFromUrl(){Js([this.router.events.pipe(kt(e=>e instanceof nc),je(()=>this.activatedRoute.snapshot.queryParams)),this.apps$]).subscribe(([e,A])=>{if(A&&A.length){let i=e.app;i&&A.includes(i)?this.selectedAppControl.setValue(i):i&&this.openSnackBar(`Agent '${i}' not found`,"OK")}})}updateSelectedAppUrl(){this.selectedAppControl.valueChanges.pipe(tl(),kt(Boolean)).subscribe(e=>{this.selectApp(e);let A=this.activatedRoute.snapshot.queryParams.app;e!==A&&this.router.navigate([],{queryParams:{app:e},queryParamsHandling:"merge"})})}updateSelectedSessionUrl(){let e=this.router.createUrlTree([],{queryParams:{session:this.sessionId},queryParamsHandling:"merge"}).toString();this.location.replaceState(e)}handlePageEvent(e){if(e.pageIndex>=0){let A=this.getKeyAtIndexInMap(e.pageIndex);A&&this.selectEvent(A)}}closeSelectedEvent(){this.selectedEvent=void 0,this.selectedEventIndex=void 0}getIndexOfKeyInMap(e){let A=0,i=(o,r)=>0,n=Array.from(this.eventData.keys()).sort(i);for(let o of n){if(o===e)return A;A++}}getKeyAtIndexInMap(e){let A=(n,o)=>0,i=Array.from(this.eventData.keys()).sort(A);if(e>=0&&e<i.length)return i[e]}openSnackBar(e,A){this._snackBar.open(e,A)}processThoughtText(e){return e.replace("/*PLANNING*/","").replace("/*ACTION*/","")}openLink(e){window.open(e,"_blank")}renderGooglerSearch(e){return this.sanitizer.bypassSecurityTrustHtml(e)}openViewImageDialog(e){let A=this.dialog.open(v0,{maxWidth:"90vw",maxHeight:"90vh",data:{imageData:e}})}createDefaultArtifactName(e){return!e||!e.includes("/")?"":e.replace("/",".")}exportSession(){this.sessionService.getSession(this.userId,this.appName,this.sessionId).subscribe(e=>{console.log(e),this.downloadService.downloadObjectAsJson(e,`session-${this.sessionId}.json`)})}updateState(){this.dialog.open(vQ,{maxWidth:"90vw",maxHeight:"90vh",data:{dialogHeader:"Update state",jsonContent:this.currentSessionState}}).afterClosed().subscribe(A=>{A&&this.updatedSessionState.set(A)})}removeStateUpdate(){this.updatedSessionState.set(null)}closeTraceEventDetailPanel(){this.bottomPanelVisible=!1,this.traceService.selectedRow(void 0),this.traceService.setHoveredMessages(void 0,"")}shouldMessageHighlighted(e){return this.hoveredEventMessageIndices.includes(e)}importSession(){let e=document.createElement("input");e.type="file",e.accept="application/json",e.onchange=()=>{if(!e.files||e.files.length===0)return;let A=e.files[0],i=new FileReader;i.onload=n=>{if(n.target?.result)try{let o=JSON.parse(n.target.result);if(!o.userId||!o.appName||!o.events){this.openSnackBar("Invalid session file format","OK");return}this.sessionService.importSession(o.userId,o.appName,o.events).subscribe(r=>{this.openSnackBar("Session imported","OK"),this.sessionTab.refreshSession()})}catch{this.openSnackBar("Error parsing session file","OK")}},i.readAsText(A)},e.click()}injectCustomIconColorStyle(e,A){if(this.document.getElementById(e))return;let i=this.renderer.createElement("style");this.renderer.setAttribute(i,"id",e),this.renderer.setAttribute(i,"type","text/css");let n=`
      .${e} {
        background-color: ${A} !important;
      }
    `;this.renderer.appendChild(i,this.renderer.createText(n)),this.renderer.appendChild(this.document.head,i)}static \u0275fac=function(A){return new(A||t)(PA(cl),PA(Wg),PA(RQ),PA(xQ),PA(Xg),PA(LQ),PA(qs),PA(nI),PA(ha),PA(O2),PA(Vc),PA($g),PA(Dc),PA(Wi),PA(at))};static \u0275cmp=zA({type:t,selectors:[["app-chat"]],viewQuery:function(A,i){if(A&1&&(Te(vjA,5,ee),Te(bjA,5),Te(nd,5),Te(od,5),Te(id,5),Te(MjA,5),Te(kjA,5),Te(SjA,5)),A&2){let n;XA(n=$A())&&(i.videoContainer=n.first),XA(n=$A())&&(i.sideDrawer=n.first),XA(n=$A())&&(i.eventTabComponent=n.first),XA(n=$A())&&(i.sessionTab=n.first),XA(n=$A())&&(i.evalTab=n.first),XA(n=$A())&&(i.scrollContainer=n.first),XA(n=$A())&&(i.textarea=n.first),XA(n=$A())&&(i.bottomPanelRef=n.first)}},standalone:!1,features:[ht([{provide:iC,useClass:dU}])],decls:28,vars:15,consts:[["sideDrawer",""],["autoScroll",""],["videoContainer",""],["messageTextarea",""],["fileInput",""],["moreMenu","matMenu"],["bottomPanel",""],["autosize","",1,"drawer-container"],["matTooltip","Open panel",1,"material-symbols-outlined",2,"position","absolute","width","24px","height","24px","color","#c4c7c5","cursor","pointer","margin-left","20px","margin-top","20px","z-index","9999"],["mode","side","appResizableDrawer","",1,"side-drawer"],[2,"margin-top","20px","margin-left","20px","display","flex"],[2,"width","100%"],[1,"drawer-header"],[1,"drawer-logo"],["src","assets/ADK-512-color.svg","width","32px","height","32px"],["matTooltip","Collapse panel",1,"material-symbols-outlined",2,"color","#c4c7c5","cursor","pointer","margin-right","15px",3,"click"],[1,"app-select-container"],[1,"app-select",3,"selectionChange","placeholder","formControl"],[1,"app-name-option",3,"value"],[1,"tabs-container"],[1,"details-panel-container"],[1,"resize-handler"],[1,"chat-container"],[1,"chat-toolbar",3,"ngClass"],[1,"chat-card"],["mat-fab","","color","primary",1,"fab-button"],[1,"chat-messages"],[1,"chat-input"],["appResizableBottomPanel","",1,"trace-detail-container"],["matTooltip","Open panel",1,"material-symbols-outlined",2,"position","absolute","width","24px","height","24px","color","#c4c7c5","cursor","pointer","margin-left","20px","margin-top","20px","z-index","9999",3,"click"],[3,"selectedTabChange"],[1,"tabs-header"],["mat-tab-label",""],[3,"traceData"],[3,"selectedEvent","eventsMap","traceData"],[3,"sessionState"],[3,"artifacts"],[3,"sessionSelected","sessionReloaded","userId","appName","sessionId"],[1,"tab-label"],[3,"shouldShowTab","sessionSelected","evalCaseSelected","evalSetIdSelected","shouldReturnToSession","evalNotInstalledMsg","appName","userId","sessionId"],[1,"details-content"],[2,"display","flex","justify-content","flex-end","margin-top","10px"],["aria-label","Select event",1,"event-paginator",3,"page","length","pageSize","pageIndex"],["mat-mini-fab",""],[3,"click"],["label","Event"],[1,"event-graph-container"],[3,"innerHtml"],[1,"json-viewer-container"],[3,"json"],["label","Request"],["label","Response"],[3,"click","innerHtml"],["matTooltip","Open panel",1,"material-symbols-outlined",2,"width","24px","height","24px","color","#c4c7c5","cursor","pointer","margin-left","20px","margin-top","-2px","z-index","9999"],["matTooltip","Open panel",1,"material-symbols-outlined",2,"width","24px","height","24px","color","#c4c7c5","cursor","pointer","margin-left","20px","margin-top","-2px","z-index","9999",3,"click"],[2,"display","flex"],[1,"toolbar-session-text"],[1,"toolbar-session-id"],[1,"toolbar-actions"],["mat-button","",2,"height","30px",3,"click"],["mat-flat-button","",2,"height","30px",3,"click","disabled"],["matTooltip","Edit current eval case",1,"material-symbols-outlined","toolbar-icon",3,"click"],["matTooltip","Delete current eval case",1,"material-symbols-outlined","toolbar-icon",3,"click"],[1,"toolbar-sse-toggle"],[1,"example-margin",3,"change","checked"],[2,"margin-left","8px","margin-right","8px","height","22px",3,"vertical"],[2,"display","flex","align-items","center"],[1,"toolbar-new-sesison",3,"click"],["matTooltip","Delete current session",1,"material-symbols-outlined","toolbar-icon",3,"click"],["matTooltip","Export current session",1,"material-symbols-outlined","toolbar-icon",3,"click"],["matTooltip","Import session",1,"material-symbols-outlined","toolbar-icon"],["matTooltip","Import session",1,"material-symbols-outlined","toolbar-icon",3,"click"],[1,"empty-state-container"],[1,"warning"],[1,"error"],["mat-fab","","color","primary",1,"fab-button",3,"click"],[3,"ngClass"],["mat-mini-fab","",3,"matTooltip","class"],[1,"message-card",3,"ngClass","ngStyle"],["mat-stroked-button","",1,"function-event-button",3,"ngClass"],[1,"material-symbols-outlined"],["mat-mini-fab","",3,"click","matTooltip"],["fontSet","material-symbols-outlined"],["mode","buffer",1,"loading-bar"],[1,"attachments"],[1,"thought-chip"],[1,"eval-compare-container"],[1,"attachment"],["alt","attachment",1,"image-preview-chat",3,"src"],["download","",3,"href"],[1,"edit-message-container"],[1,"message-text",3,"data","ngStyle"],["rows","4","cols","80",1,"message-textarea",3,"ngModelChange","keydown","ngModel"],[1,"edit-message-buttons-container"],["matTooltip","Cancel editing",1,"material-symbols-outlined",2,"width","24px","height","24px","color","#c4c7c5","cursor","pointer","margin-right","16px",3,"click"],["matTooltip","Save eval case message",1,"material-symbols-outlined",2,"width","24px","height","24px","color","rgb(97, 151, 202)","cursor","pointer","margin-right","16px",3,"click"],[3,"innerHTML"],[1,"generated-image-container"],["alt","image",1,"generated-image",3,"click","src"],[3,"base64data"],[1,"html-artifact-container"],[1,"link-style-button",3,"click"],["alt","image",1,"image-preview-chat",3,"click","src"],[1,"actual-expected-compare-container"],[1,"score-threshold-container"],[1,"actual-result"],[1,"eval-response-header","header-actual"],[1,"expected-result"],[1,"eval-response-header","header-expected"],[1,"header-actual"],[1,"header-expected"],["mat-stroked-button","",1,"function-event-button",3,"click","ngClass"],["matTooltip","Edit eval case message",1,"material-symbols-outlined","eval-case-edit-button",3,"click","ngClass"],["matTooltip","Delete eval case message",1,"material-symbols-outlined","eval-case-edit-button",3,"click","ngClass"],["matTooltip","Edit function arguments",1,"material-symbols-outlined","eval-case-edit-button",3,"click","ngClass"],["type","file","multiple","","hidden","",3,"change"],["appearance","outline",1,"input-field"],[1,"file-preview"],["matInput","","cdkTextareaAutosize","","cdkAutosizeMinRows","1","cdkAutosizeMaxRows","10","placeholder","Type a Message...",1,"chat-input-box",2,"caret-color","white",3,"ngModelChange","keydown.enter","ngModel"],[1,"chat-input-actions"],["mat-icon-button","","matTooltip","Upload local file",1,"function-event-button",3,"click"],["mat-icon-button","","matTooltip","More options",1,"function-event-button",3,"matMenuTriggerFor"],["mat-menu-item","",3,"click"],["mat-icon-button","","matSuffix","",3,"click","ngStyle","matTooltip"],[1,"file-container"],[1,"image-container"],["alt","preview",1,"image-preview",3,"src"],["mat-icon-button","",1,"delete-button",3,"click"],["color","warn"],[1,"file-info"],[1,"bottom-resize-handler"],[3,"panelClosed","userId","appName","sessionId"]],template:function(A,i){if(A&1){let n=be();S(0,"mat-drawer-container",7),_A(1,GjA,2,0,"span",8),S(2,"mat-drawer",9,0)(4,"div",10)(5,"div",11)(6,"div",12)(7,"div",13),JA(8,"img",14),AA(9," Agent Development Kit "),F(),S(10,"span",15),hA("click",function(){return RA(n),xA(i.toggleSidePanel())}),AA(11,"left_panel_close"),F()()()(),S(12,"div",16)(13,"mat-select",17),hA("selectionChange",function(r){return RA(n),xA(i.onAppSelection(r))}),_A(14,KjA,2,0),Za(15,"async"),_A(16,YjA,2,2,"mat-option",18),F()(),_A(17,qjA,18,9,"div",19)(18,ZjA,20,7,"div",20),JA(19,"div",21),F(),S(20,"div",22),_A(21,iqA,4,5,"div",23),S(22,"mat-card",24),_A(23,lqA,2,1)(24,gqA,3,0,"button",25)(25,PqA,6,0,"div",26)(26,XqA,25,11,"div",27),F(),_A(27,$qA,4,3,"div",28),F()()}if(A&2){let n;G(),GA(!i.showSidePanel&&i.appName===""?1:-1),G(12),yA("placeholder",i.isLoadingApps()?"Loading...":"Select an agent")("formControl",i.selectedAppControl),G(),GA((n=M2(15,13,i.apps$))?14:-1,n),G(2),GA(i.selectedAppControl.value&&i.isLoadingApps()?16:-1),G(),GA(i.appName!=""&&i.showSidePanel?17:-1),G(),GA(i.selectedEvent&&i.showSidePanel?18:-1),G(3),GA(i.appName!=""?21:-1),G(2),GA(i.selectedAppControl.value?-1:23),G(),GA(i.longRunningEvents.length>0?24:-1),G(),GA(i.appName!=""?25:-1),G(),GA(i.appName!=""&&i.isChatMode()?26:-1),G(),GA(i.bottomPanelVisible?27:-1)}},dependencies:[Xa,Kh,vc,Ea,ec,xcA,P2,Qg,bP,iI,CcA,Dr,kB,rP,oP,Ok,pcA,FQ,JG,TG,jG,mf,o7,NB,U2,_B,jcA,dlA,C7,vM,NQ,bf,ulA,nd,od,id,zu,Mf,nC,kf,Sf,xf,Lf,Jh],styles:[".expand-side-drawer[_ngcontent-%COMP%]{position:relative;top:4%;left:1%}.drawer-container[_ngcontent-%COMP%]{height:100%;background-color:#131314}.generated-image-container[_ngcontent-%COMP%]{max-width:400px}.generated-image[_ngcontent-%COMP%]{max-width:100%;min-width:40px;border-radius:8px}.chat-container[_ngcontent-%COMP%]{width:100%;height:100%;max-width:100%;margin:auto;display:flex;flex-direction:column;flex:1}.event-container[_ngcontent-%COMP%]{color:#fff}.html-artifact-container[_ngcontent-%COMP%], .drawer-header[_ngcontent-%COMP%]{width:100%;display:flex;justify-content:flex-start;align-items:center}.drawer-header[_ngcontent-%COMP%]   .mat-icon[_ngcontent-%COMP%]{width:36px;height:36px;color:#bdc1c6;cursor:pointer;display:flex;align-items:center;justify-content:center}.chat-card[_ngcontent-%COMP%]{display:flex;flex-direction:column;overflow:hidden;flex:1;min-height:12%;box-shadow:none;background-color:#131314}.loading-bar[_ngcontent-%COMP%]{width:100px;margin:15px}.chat-messages[_ngcontent-%COMP%]{flex-grow:1;overflow-y:auto;padding:20px;margin-top:16px}.message-card[_ngcontent-%COMP%]{padding:5px 20px;margin:5px;border-radius:20px;max-width:80%;font-size:14px;font-weight:400;position:relative;display:inline-block}.function-event-button[_ngcontent-%COMP%]{background-color:#fff;margin:5px 5px 10px}.function-event-button-highlight[_ngcontent-%COMP%]{background-color:#0f5223;border-color:#0f5223!important;color:#fff!important}.user-message[_ngcontent-%COMP%]{display:flex;justify-content:flex-end;align-items:center}.user-message[_ngcontent-%COMP%]   .message-card[_ngcontent-%COMP%]{background-color:#004a77;align-self:flex-end;color:#fff;box-shadow:none}.bot-message[_ngcontent-%COMP%]{display:flex;align-items:center}.bot-message[_ngcontent-%COMP%]   .message-card[_ngcontent-%COMP%]{background-color:#303030;align-self:flex-start;color:#fff;box-shadow:none}.bot-message[_ngcontent-%COMP%]:focus-within   .message-card[_ngcontent-%COMP%]{background-color:#131314;border:1px solid #8ab4f8}.message-textarea[_ngcontent-%COMP%]{background-color:#303030;max-width:100%;border:none;font-family:Google Sans,Helvetica Neue,sans-serif}.message-textarea[_ngcontent-%COMP%]:focus{background-color:#131314;outline:none}.edit-message-buttons-container[_ngcontent-%COMP%]{display:flex;justify-content:flex-end}.message-card[_ngcontent-%COMP%]   .eval-compare-container[_ngcontent-%COMP%]{visibility:hidden;position:absolute;left:10px;z-index:10;background-color:#484848;overflow:hidden;border-radius:20px;padding:5px 20px;margin-bottom:10px;font-size:16px}.message-card[_ngcontent-%COMP%]   .eval-compare-container[_ngcontent-%COMP%]   .actual-result[_ngcontent-%COMP%]{border-right:2px solid #8a8686;padding-right:8px;min-width:350px;max-width:350px}.message-card[_ngcontent-%COMP%]   .eval-compare-container[_ngcontent-%COMP%]   .expected-result[_ngcontent-%COMP%]{padding-left:12px;min-width:350px;max-width:350px}.message-card[_ngcontent-%COMP%]:hover   .eval-compare-container[_ngcontent-%COMP%]{visibility:visible}.actual-expected-compare-container[_ngcontent-%COMP%]{display:flex}.score-threshold-container[_ngcontent-%COMP%]{display:flex;justify-content:center;gap:10px;align-items:center;margin-top:15px;font-size:14px;font-weight:600}.eval-response-header[_ngcontent-%COMP%]{padding-bottom:5px;border-bottom:2px solid #8a8686;font-style:italic;font-weight:700}.header-expected[_ngcontent-%COMP%]{color:#44c265}.header-actual[_ngcontent-%COMP%]{color:#ff8983}.eval-case-edit-button[_ngcontent-%COMP%]{cursor:pointer;margin-left:4px;margin-right:4px}.eval-pass[_ngcontent-%COMP%]{display:flex;color:#44c265}.eval-fail[_ngcontent-%COMP%]{display:flex;color:#ff8983}.navigation-button-sidepanel[_ngcontent-%COMP%]{margin-left:auto;margin-right:20px}.chat-input[_ngcontent-%COMP%]{display:flex;padding:10px;width:60%;margin:0 auto}.hidden[_ngcontent-%COMP%]{visibility:hidden}.input-field[_ngcontent-%COMP%]{flex-grow:1}.input-field[_ngcontent-%COMP%]   textarea[_ngcontent-%COMP%]{color:#fff;border:none;padding:10px;box-sizing:content-box}.input-field[_ngcontent-%COMP%]   textarea[_ngcontent-%COMP%]::placeholder{color:#8e918f}.input-field[_ngcontent-%COMP%]   button[_ngcontent-%COMP%]{color:#fff;background-color:#333537}.chat-input-actions[_ngcontent-%COMP%]{width:106%;margin-top:10px;display:flex;justify-content:space-between;align-items:center;max-width:100%}.chat-input-actions[_ngcontent-%COMP%]   button[_ngcontent-%COMP%]{margin-left:10px;margin-right:10px}.fab-button[_ngcontent-%COMP%]{position:fixed;bottom:200px;right:100px;z-index:1000}.sidepanel-toggle[_ngcontent-%COMP%]{position:relative;top:100px;z-index:1000}.side-drawer[_ngcontent-%COMP%]{background-color:#1b1b1b;color:#fff;border-radius:0}.tabs-container[_ngcontent-%COMP%]{width:100%;margin-top:20px}.tab-label[_ngcontent-%COMP%]{font-size:14px}.file-preview[_ngcontent-%COMP%]{display:flex;flex-wrap:wrap;gap:5px;margin-top:2px;margin-bottom:8px}.file-item[_ngcontent-%COMP%]{display:flex;align-items:center;gap:5px;background:#eee;padding:5px;border-radius:4px}.image-preview[_ngcontent-%COMP%]{width:40px;height:40px;object-fit:cover;border-radius:4px}.image-preview-chat[_ngcontent-%COMP%]{max-width:90%;max-height:70vh;width:auto;height:auto;border-radius:8px;cursor:pointer;transition:transform .2s ease-in-out}button[_ngcontent-%COMP%]{margin-left:20px;margin-right:20px}.app-select[_ngcontent-%COMP%]{width:100%}.empty-state-container[_ngcontent-%COMP%]{color:#eee;height:100%;display:flex;flex-direction:column;justify-content:center;align-items:center;font-family:Google Sans,sans-serif;font-weight:400;letter-spacing:normal;line-height:24px;font-size:18px}.empty-state-container[_ngcontent-%COMP%]   pre.warning[_ngcontent-%COMP%]{color:#ffc185}.empty-state-container[_ngcontent-%COMP%]   pre.error[_ngcontent-%COMP%]{color:#ff4545}[_nghost-%COMP%]     .mat-mdc-unelevated-button:not(:disabled){color:#202124;background-color:#8ab4f8}[_nghost-%COMP%]     .message-text p{white-space:pre-line;word-break:break-word;overflow-wrap:break-word}[_nghost-%COMP%]     .mdc-linear-progress__buffer-dots{background:#fff}[_nghost-%COMP%]     .mat-mdc-select-arrow-wrapper{margin-left:4px}[_nghost-%COMP%]     .mat-mdc-text-field-wrapper{border:1px solid #8e918f}[_nghost-%COMP%]     .input-field .mat-mdc-text-field-wrapper{border:1px solid #8e918f;border-radius:16px}[_nghost-%COMP%]     .mdc-notched-outline__leading, [_nghost-%COMP%]     .mdc-notched-outline__notch, [_nghost-%COMP%]     .mdc-notched-outline__trailing{border:none}[_nghost-%COMP%]     .mat-mdc-form-field-icon-suffix{padding:0 10px 0 40px}[_nghost-%COMP%]     .segment-key{color:#d3d3d3!important}[_nghost-%COMP%]     .mat-mdc-mini-fab{background-color:#fff}[_nghost-%COMP%]     .mat-mdc-mini-fab mat-icon{color:#000}.mat-mdc-select-placeholder[_ngcontent-%COMP%]{margin-left:20px}.resize-handler[_ngcontent-%COMP%]{background:#5f6368;width:4px;border-radius:4px;position:absolute;display:block;height:20%;top:40%;right:0;z-index:9999;cursor:ew-resize}.bottom-resize-handler[_ngcontent-%COMP%]{background:#5f6368;height:5px;border-radius:4px;position:absolute;display:block;width:20%;left:40%;top:0;right:0;z-index:9999;cursor:ns-resize}.trace-detail-container[_ngcontent-%COMP%]{position:relative;background-color:#1b1b1b}.trace-detail-container[_ngcontent-%COMP%]   app-trace-event[_ngcontent-%COMP%]{padding-top:8px}.new-session-button[_ngcontent-%COMP%]{margin-top:0;margin-left:50px;width:130px;height:28px;font-size:14px}.app-select-container[_ngcontent-%COMP%]{width:35%;margin-top:12px;background-color:#212123;margin-left:10px;height:30px;display:flex;justify-content:space-between;padding-left:20px;padding-right:20px;border-radius:10px;padding-top:5px}.app-select-container[_ngcontent-%COMP%]{--mat-select-placeholder-text-color: #8ab4f8}.app-select-container[_ngcontent-%COMP%]{--mat-select-enabled-trigger-text-color: #8ab4f8}.app-select-container[_ngcontent-%COMP%]{--mat-select-enabled-arrow-color: #8ab4f8}.json-viewer-container[_ngcontent-%COMP%]{margin:10px}.event-paginator[_ngcontent-%COMP%]{margin-top:-8px;margin-right:auto;background-color:inherit;display:flex;justify-content:center}[_nghost-%COMP%]     .mat-mdc-paginator-page-size{display:none!important}.details-panel-container[_ngcontent-%COMP%]{position:absolute;width:100%;height:98%;left:0;right:0;bottom:0;background:#242424;display:inline-block;justify-content:center;align-items:center;z-index:10}.details-content[_ngcontent-%COMP%]{color:#fff;font-size:14px}.adk-checkbox[_ngcontent-%COMP%]{position:fixed;bottom:0;left:0;right:0;margin-bottom:20px;margin-left:20px}.drawer-header[_ngcontent-%COMP%]{display:flex;justify-content:space-between}.drawer-header[_ngcontent-%COMP%]{--mdc-filled-button-container-color: #89b4f8}.drawer-header[_ngcontent-%COMP%]{--mdc-filled-button-label-text-color: black}.chat-toolbar[_ngcontent-%COMP%]{position:sticky;top:0;height:48px;background:#1b1b1b;display:flex;align-items:center;z-index:10}.chat-toolbar.edit-mode[_ngcontent-%COMP%]{background:#44c2651a}.attachment[_ngcontent-%COMP%]{display:flex;align-items:center}.toolbar-actions[_ngcontent-%COMP%]{margin-left:auto;display:flex;align-items:center}.toolbar-session-text[_ngcontent-%COMP%]{color:#fdfdfd;font-family:Roboto;font-size:12px;font-style:normal;font-weight:500;line-height:12px;letter-spacing:.8px;text-transform:uppercase;margin-left:20px;padding-top:4px}.toolbar-session-id[_ngcontent-%COMP%]{color:#9aa0a6;font-family:Google Sans Mono,monospace;font-size:14px;font-style:normal;font-weight:400;line-height:20px;letter-spacing:.25px;margin-left:5px}.toolbar-icon[_ngcontent-%COMP%]{width:24px;height:24px;color:#c4c7c5;cursor:pointer;margin-right:16px}.toolbar-new-sesison[_ngcontent-%COMP%]{font-size:14px;margin-right:16px;color:#9aa0a6;cursor:pointer;display:flex;align-items:center}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mat-switch-label-text-size: 14px}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mat-switch-label-text-color: #9aa0a6}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-track-color: #8ab4f9}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-focus-track-color: #8ab4f9}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-hover-track-color: #8ab4f9}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-handle-color: #1b73e8}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-focus-handle-color: #1b73e8}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-selected-hover-handle-color: #1b73e8}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-track-height: 24px}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mdc-switch-track-width: 46px}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mat-switch-track-outline-color: #1b73e8}.toolbar-sse-toggle[_ngcontent-%COMP%]{--mat-switch-with-icon-handle-size: 20px}.image-container[_ngcontent-%COMP%]{position:relative;display:inline-block;border-radius:12px;overflow:hidden}.image-preview[_ngcontent-%COMP%]{display:block;width:100%;height:auto;border-radius:12px;width:80px;height:80px}.delete-button[_ngcontent-%COMP%]{position:absolute;top:1px;right:1px;background-color:#000000b3;border:none;border-radius:50%;padding:8px;cursor:pointer;color:#fff;display:flex;align-items:center;justify-content:center;margin-right:0;scale:.7}.delete-button[_ngcontent-%COMP%]   mat-icon[_ngcontent-%COMP%]{font-size:20px}.file-container[_ngcontent-%COMP%]{position:relative;display:flex;flex-direction:column;gap:8px;height:80px;background-color:#1e1e1e;border-radius:12px}.file-info[_ngcontent-%COMP%]{margin-right:60px;padding-top:20px;padding-left:16px}.thought-chip[_ngcontent-%COMP%]{border-radius:5px;background-color:#8ab4f8;width:80px;text-align:center;margin-top:5px}.event-graph-container[_ngcontent-%COMP%]{margin-top:16px;margin-bottom:16px;display:flex;justify-content:center;max-height:33%;cursor:pointer}.event-graph-container[_ngcontent-%COMP%]     svg{width:100%;height:100%;display:block;object-fit:contain}.event-graph-container[_ngcontent-%COMP%]     svg text{font-family:Google Sans Mono,monospace;font-size:11px}[_nghost-%COMP%]     pre{white-space:pre-wrap;word-break:break-word;overflow-x:auto;max-width:100%}.link-style-button[_ngcontent-%COMP%]{background:none;border:none;padding:0;font:inherit;color:#007bff!important;text-decoration:underline;cursor:pointer;outline:none;font-size:14px}.drawer-logo[_ngcontent-%COMP%]{margin-left:9px;display:flex;align-items:center;font-size:16px;font-style:normal;font-weight:500;line-height:24px;letter-spacing:.1px}.drawer-logo[_ngcontent-%COMP%]   img[_ngcontent-%COMP%]{margin-right:9px}  .mat-drawer-content{display:flex!important}  .mat-drawer{border-right:1px solid #444746!important}.app-name-option[_ngcontent-%COMP%], .app-select[_ngcontent-%COMP%]{color:#9aa0a6;font-family:Google Sans Mono,monospace;font-style:normal;font-weight:400;padding-left:unset}"],changeDetection:0})};var _Q=class t{title="agent_framework_web";userId="";appName="";sessionId="";constructor(){}static \u0275fac=function(A){return new(A||t)};static \u0275cmp=zA({type:t,selectors:[["app-root"]],standalone:!1,decls:1,vars:0,template:function(A,i){A&1&&JA(0,"app-chat")},dependencies:[Ff],encapsulation:2})};var tVA=[{path:"",component:_Q}],B7=class t{static \u0275fac=function(A){return new(A||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[O6.forRoot(tVA),O6]})};function FlA(t){return new Ae(3e3,!1)}function iVA(){return new Ae(3100,!1)}function nVA(){return new Ae(3101,!1)}function oVA(t){return new Ae(3001,!1)}function rVA(t){return new Ae(3003,!1)}function sVA(t){return new Ae(3004,!1)}function _lA(t,e){return new Ae(3005,!1)}function GlA(){return new Ae(3006,!1)}function UlA(){return new Ae(3007,!1)}function KlA(t,e){return new Ae(3008,!1)}function YlA(t){return new Ae(3002,!1)}function JlA(t,e,A,i,n){return new Ae(3010,!1)}function TlA(){return new Ae(3011,!1)}function HlA(){return new Ae(3012,!1)}function zlA(){return new Ae(3200,!1)}function OlA(){return new Ae(3202,!1)}function PlA(){return new Ae(3013,!1)}function jlA(t){return new Ae(3014,!1)}function qlA(t){return new Ae(3015,!1)}function VlA(t){return new Ae(3016,!1)}function ZlA(t,e){return new Ae(3404,!1)}function aVA(t){return new Ae(3502,!1)}function WlA(t){return new Ae(3503,!1)}function XlA(){return new Ae(3300,!1)}function $lA(t){return new Ae(3504,!1)}function AgA(t){return new Ae(3301,!1)}function egA(t,e){return new Ae(3302,!1)}function tgA(t){return new Ae(3303,!1)}function igA(t,e){return new Ae(3400,!1)}function ngA(t){return new Ae(3401,!1)}function ogA(t){return new Ae(3402,!1)}function rgA(t,e){return new Ae(3505,!1)}function c2(t){switch(t.length){case 0:return new Eg;case 1:return t[0];default:return new tC(t)}}function hU(t,e,A=new Map,i=new Map){let n=[],o=[],r=-1,s=null;if(e.forEach(a=>{let c=a.get("offset"),l=c==r,I=l&&s||new Map;a.forEach((C,d)=>{let B=d,E=C;if(d!=="offset")switch(B=t.normalizePropertyName(B,n),E){case FB:E=A.get(d);break;case kc:E=i.get(d);break;default:E=t.normalizeStyleValue(d,B,E,n);break}I.set(B,E)}),l||o.push(I),s=I,r=c}),n.length)throw aVA(n);return o}function E7(t,e,A,i){switch(e){case"start":t.onStart(()=>i(A&&BU(A,"start",t)));break;case"done":t.onDone(()=>i(A&&BU(A,"done",t)));break;case"destroy":t.onDestroy(()=>i(A&&BU(A,"destroy",t)));break}}function BU(t,e,A){let i=A.totalTime,n=!!A.disabled,o=Q7(t.element,t.triggerName,t.fromState,t.toState,e||t.phaseName,i??t.totalTime,n),r=t._data;return r!=null&&(o._data=r),o}function Q7(t,e,A,i,n="",o=0,r){return{element:t,triggerName:e,fromState:A,toState:i,phaseName:n,totalTime:o,disabled:!!r}}function Ya(t,e,A){let i=t.get(e);return i||t.set(e,i=A),i}function uU(t){let e=t.indexOf(":"),A=t.substring(1,e),i=t.slice(e+1);return[A,i]}var cVA=typeof document>"u"?null:document.documentElement;function h7(t){let e=t.parentNode||t.host||null;return e===cVA?null:e}function lVA(t){return t.substring(1,6)=="ebkit"}var ad=null,NlA=!1;function sgA(t){ad||(ad=gVA()||{},NlA=ad.style?"WebkitAppearance"in ad.style:!1);let e=!0;return ad.style&&!lVA(t)&&(e=t in ad.style,!e&&NlA&&(e="Webkit"+t.charAt(0).toUpperCase()+t.slice(1)in ad.style)),e}function gVA(){return typeof document<"u"?document.body:null}function fU(t,e){for(;e;){if(e===t)return!0;e=h7(e)}return!1}function mU(t,e,A){if(A)return Array.from(t.querySelectorAll(e));let i=t.querySelector(e);return i?[i]:[]}var IVA=1e3,pU="{{",CVA="}}",wU="ng-enter",u7="ng-leave",Nf="ng-trigger",_f=".ng-trigger",DU="ng-animating",f7=".ng-animating";function e0(t){if(typeof t=="number")return t;let e=t.match(/^(-?[\.\d]+)(m?s)/);return!e||e.length<2?0:EU(parseFloat(e[1]),e[2])}function EU(t,e){switch(e){case"s":return t*IVA;default:return t}}function Gf(t,e,A){return t.hasOwnProperty("duration")?t:dVA(t,e,A)}function dVA(t,e,A){let i=/^(-?[\.\d]+)(m?s)(?:\s+(-?[\.\d]+)(m?s))?(?:\s+([-a-z]+(?:\(.+?\))?))?$/i,n,o=0,r="";if(typeof t=="string"){let s=t.match(i);if(s===null)return e.push(FlA(t)),{duration:0,delay:0,easing:""};n=EU(parseFloat(s[1]),s[2]);let a=s[3];a!=null&&(o=EU(parseFloat(a),s[4]));let c=s[5];c&&(r=c)}else n=t;if(!A){let s=!1,a=e.length;n<0&&(e.push(iVA()),s=!0),o<0&&(e.push(nVA()),s=!0),s&&e.splice(a,0,FlA(t))}return{duration:n,delay:o,easing:r}}function agA(t){return t.length?t[0]instanceof Map?t:t.map(e=>new Map(Object.entries(e))):[]}function Jl(t,e,A){e.forEach((i,n)=>{let o=m7(n);A&&!A.has(n)&&A.set(n,t.style[o]),t.style[o]=i})}function rI(t,e){e.forEach((A,i)=>{let n=m7(i);t.style[n]=""})}function GQ(t){return Array.isArray(t)?t.length==1?t[0]:hP(t):t}function cgA(t,e,A){let i=e.params||{},n=yU(t);n.length&&n.forEach(o=>{i.hasOwnProperty(o)||A.push(oVA(o))})}var QU=new RegExp(`${pU}\\s*(.+?)\\s*${CVA}`,"g");function yU(t){let e=[];if(typeof t=="string"){let A;for(;A=QU.exec(t);)e.push(A[1]);QU.lastIndex=0}return e}function UQ(t,e,A){let i=`${t}`,n=i.replace(QU,(o,r)=>{let s=e[r];return s==null&&(A.push(rVA(r)),s=""),s.toString()});return n==i?t:n}var BVA=/-+([a-z0-9])/g;function m7(t){return t.replace(BVA,(...e)=>e[1].toUpperCase())}function lgA(t,e){return t===0||e===0}function ggA(t,e,A){if(A.size&&e.length){let i=e[0],n=[];if(A.forEach((o,r)=>{i.has(r)||n.push(r),i.set(r,o)}),n.length)for(let o=1;o<e.length;o++){let r=e[o];n.forEach(s=>r.set(s,p7(t,s)))}}return e}function Ja(t,e,A){switch(e.type){case Ci.Trigger:return t.visitTrigger(e,A);case Ci.State:return t.visitState(e,A);case Ci.Transition:return t.visitTransition(e,A);case Ci.Sequence:return t.visitSequence(e,A);case Ci.Group:return t.visitGroup(e,A);case Ci.Animate:return t.visitAnimate(e,A);case Ci.Keyframes:return t.visitKeyframes(e,A);case Ci.Style:return t.visitStyle(e,A);case Ci.Reference:return t.visitReference(e,A);case Ci.AnimateChild:return t.visitAnimateChild(e,A);case Ci.AnimateRef:return t.visitAnimateRef(e,A);case Ci.Query:return t.visitQuery(e,A);case Ci.Stagger:return t.visitStagger(e,A);default:throw sVA(e.type)}}function p7(t,e){return window.getComputedStyle(t)[e]}var TU=(()=>{class t{validateStyleProperty(A){return sgA(A)}containsElement(A,i){return fU(A,i)}getParentElement(A){return h7(A)}query(A,i,n){return mU(A,i,n)}computeStyle(A,i,n){return n||""}animate(A,i,n,o,r,s=[],a){return new Eg(n,o)}static \u0275fac=function(i){return new(i||t)};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})(),ld=class{static NOOP=new TU},gd=class{};var EVA=new Set(["width","height","minWidth","minHeight","maxWidth","maxHeight","left","top","bottom","right","fontSize","outlineWidth","outlineOffset","paddingTop","paddingLeft","paddingBottom","paddingRight","marginTop","marginLeft","marginBottom","marginRight","borderRadius","borderWidth","borderTopWidth","borderLeftWidth","borderRightWidth","borderBottomWidth","textIndent","perspective"]),b7=class extends gd{normalizePropertyName(e,A){return m7(e)}normalizeStyleValue(e,A,i,n){let o="",r=i.toString().trim();if(EVA.has(A)&&i!==0&&i!=="0")if(typeof i=="number")o="px";else{let s=i.match(/^[+-]?[\d\.]+([a-z]*)$/);s&&s[1].length==0&&n.push(_lA(e,i))}return r+o}};var M7="*";function QVA(t,e){let A=[];return typeof t=="string"?t.split(/\s*,\s*/).forEach(i=>hVA(i,A,e)):A.push(t),A}function hVA(t,e,A){if(t[0]==":"){let a=uVA(t,A);if(typeof a=="function"){e.push(a);return}t=a}let i=t.match(/^(\*|[-\w]+)\s*(<?[=-]>)\s*(\*|[-\w]+)$/);if(i==null||i.length<4)return A.push(qlA(t)),e;let n=i[1],o=i[2],r=i[3];e.push(IgA(n,r));let s=n==M7&&r==M7;o[0]=="<"&&!s&&e.push(IgA(r,n))}function uVA(t,e){switch(t){case":enter":return"void => *";case":leave":return"* => void";case":increment":return(A,i)=>parseFloat(i)>parseFloat(A);case":decrement":return(A,i)=>parseFloat(i)<parseFloat(A);default:return e.push(VlA(t)),"* => *"}}var w7=new Set(["true","1"]),D7=new Set(["false","0"]);function IgA(t,e){let A=w7.has(t)||D7.has(t),i=w7.has(e)||D7.has(e);return(n,o)=>{let r=t==M7||t==n,s=e==M7||e==o;return!r&&A&&typeof n=="boolean"&&(r=n?w7.has(t):D7.has(t)),!s&&i&&typeof o=="boolean"&&(s=o?w7.has(e):D7.has(e)),r&&s}}var pgA=":self",fVA=new RegExp(`s*${pgA}s*,?`,"g");function wgA(t,e,A,i){return new RU(t).build(e,A,i)}var CgA="",RU=class{_driver;constructor(e){this._driver=e}build(e,A,i){let n=new xU(A);return this._resetContextStyleTimingState(n),Ja(this,GQ(e),n)}_resetContextStyleTimingState(e){e.currentQuerySelector=CgA,e.collectedStyles=new Map,e.collectedStyles.set(CgA,new Map),e.currentTime=0}visitTrigger(e,A){let i=A.queryCount=0,n=A.depCount=0,o=[],r=[];return e.name.charAt(0)=="@"&&A.errors.push(GlA()),e.definitions.forEach(s=>{if(this._resetContextStyleTimingState(A),s.type==Ci.State){let a=s,c=a.name;c.toString().split(/\s*,\s*/).forEach(l=>{a.name=l,o.push(this.visitState(a,A))}),a.name=c}else if(s.type==Ci.Transition){let a=this.visitTransition(s,A);i+=a.queryCount,n+=a.depCount,r.push(a)}else A.errors.push(UlA())}),{type:Ci.Trigger,name:e.name,states:o,transitions:r,queryCount:i,depCount:n,options:null}}visitState(e,A){let i=this.visitStyle(e.styles,A),n=e.options&&e.options.params||null;if(i.containsDynamicStyles){let o=new Set,r=n||{};i.styles.forEach(s=>{s instanceof Map&&s.forEach(a=>{yU(a).forEach(c=>{r.hasOwnProperty(c)||o.add(c)})})}),o.size&&A.errors.push(KlA(e.name,[...o.values()]))}return{type:Ci.State,name:e.name,style:i,options:n?{params:n}:null}}visitTransition(e,A){A.queryCount=0,A.depCount=0;let i=Ja(this,GQ(e.animation),A),n=QVA(e.expr,A.errors);return{type:Ci.Transition,matchers:n,animation:i,queryCount:A.queryCount,depCount:A.depCount,options:cd(e.options)}}visitSequence(e,A){return{type:Ci.Sequence,steps:e.steps.map(i=>Ja(this,i,A)),options:cd(e.options)}}visitGroup(e,A){let i=A.currentTime,n=0,o=e.steps.map(r=>{A.currentTime=i;let s=Ja(this,r,A);return n=Math.max(n,A.currentTime),s});return A.currentTime=n,{type:Ci.Group,steps:o,options:cd(e.options)}}visitAnimate(e,A){let i=DVA(e.timings,A.errors);A.currentAnimateTimings=i;let n,o=e.styles?e.styles:po({});if(o.type==Ci.Keyframes)n=this.visitKeyframes(o,A);else{let r=e.styles,s=!1;if(!r){s=!0;let c={};i.easing&&(c.easing=i.easing),r=po(c)}A.currentTime+=i.duration+i.delay;let a=this.visitStyle(r,A);a.isEmptyStep=s,n=a}return A.currentAnimateTimings=null,{type:Ci.Animate,timings:i,style:n,options:null}}visitStyle(e,A){let i=this._makeStyleAst(e,A);return this._validateStyleAst(i,A),i}_makeStyleAst(e,A){let i=[],n=Array.isArray(e.styles)?e.styles:[e.styles];for(let s of n)typeof s=="string"?s===kc?i.push(s):A.errors.push(YlA(s)):i.push(new Map(Object.entries(s)));let o=!1,r=null;return i.forEach(s=>{if(s instanceof Map&&(s.has("easing")&&(r=s.get("easing"),s.delete("easing")),!o)){for(let a of s.values())if(a.toString().indexOf(pU)>=0){o=!0;break}}}),{type:Ci.Style,styles:i,easing:r,offset:e.offset,containsDynamicStyles:o,options:null}}_validateStyleAst(e,A){let i=A.currentAnimateTimings,n=A.currentTime,o=A.currentTime;i&&o>0&&(o-=i.duration+i.delay),e.styles.forEach(r=>{typeof r!="string"&&r.forEach((s,a)=>{let c=A.collectedStyles.get(A.currentQuerySelector),l=c.get(a),I=!0;l&&(o!=n&&o>=l.startTime&&n<=l.endTime&&(A.errors.push(JlA(a,l.startTime,l.endTime,o,n)),I=!1),o=l.startTime),I&&c.set(a,{startTime:o,endTime:n}),A.options&&cgA(s,A.options,A.errors)})})}visitKeyframes(e,A){let i={type:Ci.Keyframes,styles:[],options:null};if(!A.currentAnimateTimings)return A.errors.push(TlA()),i;let n=1,o=0,r=[],s=!1,a=!1,c=0,l=e.steps.map(u=>{let D=this._makeStyleAst(u,A),L=D.offset!=null?D.offset:wVA(D.styles),R=0;return L!=null&&(o++,R=D.offset=L),a=a||R<0||R>1,s=s||R<c,c=R,r.push(R),D});a&&A.errors.push(HlA()),s&&A.errors.push(zlA());let I=e.steps.length,C=0;o>0&&o<I?A.errors.push(OlA()):o==0&&(C=n/(I-1));let d=I-1,B=A.currentTime,E=A.currentAnimateTimings,h=E.duration;return l.forEach((u,D)=>{let L=C>0?D==d?1:C*D:r[D],R=L*h;A.currentTime=B+E.delay+R,E.duration=R,this._validateStyleAst(u,A),u.offset=L,i.styles.push(u)}),i}visitReference(e,A){return{type:Ci.Reference,animation:Ja(this,GQ(e.animation),A),options:cd(e.options)}}visitAnimateChild(e,A){return A.depCount++,{type:Ci.AnimateChild,options:cd(e.options)}}visitAnimateRef(e,A){return{type:Ci.AnimateRef,animation:this.visitReference(e.animation,A),options:cd(e.options)}}visitQuery(e,A){let i=A.currentQuerySelector,n=e.options||{};A.queryCount++,A.currentQuery=e;let[o,r]=mVA(e.selector);A.currentQuerySelector=i.length?i+" "+o:o,Ya(A.collectedStyles,A.currentQuerySelector,new Map);let s=Ja(this,GQ(e.animation),A);return A.currentQuery=null,A.currentQuerySelector=i,{type:Ci.Query,selector:o,limit:n.limit||0,optional:!!n.optional,includeSelf:r,animation:s,originalSelector:e.selector,options:cd(e.options)}}visitStagger(e,A){A.currentQuery||A.errors.push(PlA());let i=e.timings==="full"?{duration:0,delay:0,easing:"full"}:Gf(e.timings,A.errors,!0);return{type:Ci.Stagger,animation:Ja(this,GQ(e.animation),A),timings:i,options:null}}};function mVA(t){let e=!!t.split(/\s*,\s*/).find(A=>A==pgA);return e&&(t=t.replace(fVA,"")),t=t.replace(/@\*/g,_f).replace(/@\w+/g,A=>_f+"-"+A.slice(1)).replace(/:animating/g,f7),[t,e]}function pVA(t){return t?rA({},t):null}var xU=class{errors;queryCount=0;depCount=0;currentTransition=null;currentQuery=null;currentQuerySelector=null;currentAnimateTimings=null;currentTime=0;collectedStyles=new Map;options=null;unsupportedCSSPropertiesFound=new Set;constructor(e){this.errors=e}};function wVA(t){if(typeof t=="string")return null;let e=null;if(Array.isArray(t))t.forEach(A=>{if(A instanceof Map&&A.has("offset")){let i=A;e=parseFloat(i.get("offset")),i.delete("offset")}});else if(t instanceof Map&&t.has("offset")){let A=t;e=parseFloat(A.get("offset")),A.delete("offset")}return e}function DVA(t,e){if(t.hasOwnProperty("duration"))return t;if(typeof t=="number"){let o=Gf(t,e).duration;return vU(o,0,"")}let A=t;if(A.split(/\s+/).some(o=>o.charAt(0)=="{"&&o.charAt(1)=="{")){let o=vU(0,0,"");return o.dynamic=!0,o.strValue=A,o}let n=Gf(A,e);return vU(n.duration,n.delay,n.easing)}function cd(t){return t?(t=rA({},t),t.params&&(t.params=pVA(t.params))):t={},t}function vU(t,e,A){return{duration:t,delay:e,easing:A}}function HU(t,e,A,i,n,o,r=null,s=!1){return{type:1,element:t,keyframes:e,preStyleProps:A,postStyleProps:i,duration:n,delay:o,totalTime:n+o,easing:r,subTimeline:s}}var Kf=class{_map=new Map;get(e){return this._map.get(e)||[]}append(e,A){let i=this._map.get(e);i||this._map.set(e,i=[]),i.push(...A)}has(e){return this._map.has(e)}clear(){this._map.clear()}},yVA=1,vVA=":enter",bVA=new RegExp(vVA,"g"),MVA=":leave",kVA=new RegExp(MVA,"g");function DgA(t,e,A,i,n,o=new Map,r=new Map,s,a,c=[]){return new LU().buildKeyframes(t,e,A,i,n,o,r,s,a,c)}var LU=class{buildKeyframes(e,A,i,n,o,r,s,a,c,l=[]){c=c||new Kf;let I=new FU(e,A,c,n,o,l,[]);I.options=a;let C=a.delay?e0(a.delay):0;I.currentTimeline.delayNextStep(C),I.currentTimeline.setStyles([r],null,I.errors,a),Ja(this,i,I);let d=I.timelines.filter(B=>B.containsAnimation());if(d.length&&s.size){let B;for(let E=d.length-1;E>=0;E--){let h=d[E];if(h.element===A){B=h;break}}B&&!B.allowOnlyTimelineStyles()&&B.setStyles([s],null,I.errors,a)}return d.length?d.map(B=>B.buildKeyframes()):[HU(A,[],[],[],0,C,"",!1)]}visitTrigger(e,A){}visitState(e,A){}visitTransition(e,A){}visitAnimateChild(e,A){let i=A.subInstructions.get(A.element);if(i){let n=A.createSubContext(e.options),o=A.currentTimeline.currentTime,r=this._visitSubInstructions(i,n,n.options);o!=r&&A.transformIntoNewTimeline(r)}A.previousNode=e}visitAnimateRef(e,A){let i=A.createSubContext(e.options);i.transformIntoNewTimeline(),this._applyAnimationRefDelays([e.options,e.animation.options],A,i),this.visitReference(e.animation,i),A.transformIntoNewTimeline(i.currentTimeline.currentTime),A.previousNode=e}_applyAnimationRefDelays(e,A,i){for(let n of e){let o=n?.delay;if(o){let r=typeof o=="number"?o:e0(UQ(o,n?.params??{},A.errors));i.delayNextStep(r)}}}_visitSubInstructions(e,A,i){let o=A.currentTimeline.currentTime,r=i.duration!=null?e0(i.duration):null,s=i.delay!=null?e0(i.delay):null;return r!==0&&e.forEach(a=>{let c=A.appendInstructionToTimeline(a,r,s);o=Math.max(o,c.duration+c.delay)}),o}visitReference(e,A){A.updateOptions(e.options,!0),Ja(this,e.animation,A),A.previousNode=e}visitSequence(e,A){let i=A.subContextCount,n=A,o=e.options;if(o&&(o.params||o.delay)&&(n=A.createSubContext(o),n.transformIntoNewTimeline(),o.delay!=null)){n.previousNode.type==Ci.Style&&(n.currentTimeline.snapshotCurrentStyles(),n.previousNode=k7);let r=e0(o.delay);n.delayNextStep(r)}e.steps.length&&(e.steps.forEach(r=>Ja(this,r,n)),n.currentTimeline.applyStylesToKeyframe(),n.subContextCount>i&&n.transformIntoNewTimeline()),A.previousNode=e}visitGroup(e,A){let i=[],n=A.currentTimeline.currentTime,o=e.options&&e.options.delay?e0(e.options.delay):0;e.steps.forEach(r=>{let s=A.createSubContext(e.options);o&&s.delayNextStep(o),Ja(this,r,s),n=Math.max(n,s.currentTimeline.currentTime),i.push(s.currentTimeline)}),i.forEach(r=>A.currentTimeline.mergeTimelineCollectedStyles(r)),A.transformIntoNewTimeline(n),A.previousNode=e}_visitTiming(e,A){if(e.dynamic){let i=e.strValue,n=A.params?UQ(i,A.params,A.errors):i;return Gf(n,A.errors)}else return{duration:e.duration,delay:e.delay,easing:e.easing}}visitAnimate(e,A){let i=A.currentAnimateTimings=this._visitTiming(e.timings,A),n=A.currentTimeline;i.delay&&(A.incrementTime(i.delay),n.snapshotCurrentStyles());let o=e.style;o.type==Ci.Keyframes?this.visitKeyframes(o,A):(A.incrementTime(i.duration),this.visitStyle(o,A),n.applyStylesToKeyframe()),A.currentAnimateTimings=null,A.previousNode=e}visitStyle(e,A){let i=A.currentTimeline,n=A.currentAnimateTimings;!n&&i.hasCurrentStyleProperties()&&i.forwardFrame();let o=n&&n.easing||e.easing;e.isEmptyStep?i.applyEmptyStep(o):i.setStyles(e.styles,o,A.errors,A.options),A.previousNode=e}visitKeyframes(e,A){let i=A.currentAnimateTimings,n=A.currentTimeline.duration,o=i.duration,s=A.createSubContext().currentTimeline;s.easing=i.easing,e.styles.forEach(a=>{let c=a.offset||0;s.forwardTime(c*o),s.setStyles(a.styles,a.easing,A.errors,A.options),s.applyStylesToKeyframe()}),A.currentTimeline.mergeTimelineCollectedStyles(s),A.transformIntoNewTimeline(n+o),A.previousNode=e}visitQuery(e,A){let i=A.currentTimeline.currentTime,n=e.options||{},o=n.delay?e0(n.delay):0;o&&(A.previousNode.type===Ci.Style||i==0&&A.currentTimeline.hasCurrentStyleProperties())&&(A.currentTimeline.snapshotCurrentStyles(),A.previousNode=k7);let r=i,s=A.invokeQuery(e.selector,e.originalSelector,e.limit,e.includeSelf,!!n.optional,A.errors);A.currentQueryTotal=s.length;let a=null;s.forEach((c,l)=>{A.currentQueryIndex=l;let I=A.createSubContext(e.options,c);o&&I.delayNextStep(o),c===A.element&&(a=I.currentTimeline),Ja(this,e.animation,I),I.currentTimeline.applyStylesToKeyframe();let C=I.currentTimeline.currentTime;r=Math.max(r,C)}),A.currentQueryIndex=0,A.currentQueryTotal=0,A.transformIntoNewTimeline(r),a&&(A.currentTimeline.mergeTimelineCollectedStyles(a),A.currentTimeline.snapshotCurrentStyles()),A.previousNode=e}visitStagger(e,A){let i=A.parentContext,n=A.currentTimeline,o=e.timings,r=Math.abs(o.duration),s=r*(A.currentQueryTotal-1),a=r*A.currentQueryIndex;switch(o.duration<0?"reverse":o.easing){case"reverse":a=s-a;break;case"full":a=i.currentStaggerTime;break}let l=A.currentTimeline;a&&l.delayNextStep(a);let I=l.currentTime;Ja(this,e.animation,A),A.previousNode=e,i.currentStaggerTime=n.currentTime-I+(n.startTime-i.currentTimeline.startTime)}},k7={},FU=class t{_driver;element;subInstructions;_enterClassName;_leaveClassName;errors;timelines;parentContext=null;currentTimeline;currentAnimateTimings=null;previousNode=k7;subContextCount=0;options={};currentQueryIndex=0;currentQueryTotal=0;currentStaggerTime=0;constructor(e,A,i,n,o,r,s,a){this._driver=e,this.element=A,this.subInstructions=i,this._enterClassName=n,this._leaveClassName=o,this.errors=r,this.timelines=s,this.currentTimeline=a||new S7(this._driver,A,0),s.push(this.currentTimeline)}get params(){return this.options.params}updateOptions(e,A){if(!e)return;let i=e,n=this.options;i.duration!=null&&(n.duration=e0(i.duration)),i.delay!=null&&(n.delay=e0(i.delay));let o=i.params;if(o){let r=n.params;r||(r=this.options.params={}),Object.keys(o).forEach(s=>{(!A||!r.hasOwnProperty(s))&&(r[s]=UQ(o[s],r,this.errors))})}}_copyOptions(){let e={};if(this.options){let A=this.options.params;if(A){let i=e.params={};Object.keys(A).forEach(n=>{i[n]=A[n]})}}return e}createSubContext(e=null,A,i){let n=A||this.element,o=new t(this._driver,n,this.subInstructions,this._enterClassName,this._leaveClassName,this.errors,this.timelines,this.currentTimeline.fork(n,i||0));return o.previousNode=this.previousNode,o.currentAnimateTimings=this.currentAnimateTimings,o.options=this._copyOptions(),o.updateOptions(e),o.currentQueryIndex=this.currentQueryIndex,o.currentQueryTotal=this.currentQueryTotal,o.parentContext=this,this.subContextCount++,o}transformIntoNewTimeline(e){return this.previousNode=k7,this.currentTimeline=this.currentTimeline.fork(this.element,e),this.timelines.push(this.currentTimeline),this.currentTimeline}appendInstructionToTimeline(e,A,i){let n={duration:A??e.duration,delay:this.currentTimeline.currentTime+(i??0)+e.delay,easing:""},o=new NU(this._driver,e.element,e.keyframes,e.preStyleProps,e.postStyleProps,n,e.stretchStartingKeyframe);return this.timelines.push(o),n}incrementTime(e){this.currentTimeline.forwardTime(this.currentTimeline.duration+e)}delayNextStep(e){e>0&&this.currentTimeline.delayNextStep(e)}invokeQuery(e,A,i,n,o,r){let s=[];if(n&&s.push(this.element),e.length>0){e=e.replace(bVA,"."+this._enterClassName),e=e.replace(kVA,"."+this._leaveClassName);let a=i!=1,c=this._driver.query(this.element,e,a);i!==0&&(c=i<0?c.slice(c.length+i,c.length):c.slice(0,i)),s.push(...c)}return!o&&s.length==0&&r.push(jlA(A)),s}},S7=class t{_driver;element;startTime;_elementTimelineStylesLookup;duration=0;easing=null;_previousKeyframe=new Map;_currentKeyframe=new Map;_keyframes=new Map;_styleSummary=new Map;_localTimelineStyles=new Map;_globalTimelineStyles;_pendingStyles=new Map;_backFill=new Map;_currentEmptyStepKeyframe=null;constructor(e,A,i,n){this._driver=e,this.element=A,this.startTime=i,this._elementTimelineStylesLookup=n,this._elementTimelineStylesLookup||(this._elementTimelineStylesLookup=new Map),this._globalTimelineStyles=this._elementTimelineStylesLookup.get(A),this._globalTimelineStyles||(this._globalTimelineStyles=this._localTimelineStyles,this._elementTimelineStylesLookup.set(A,this._localTimelineStyles)),this._loadKeyframe()}containsAnimation(){switch(this._keyframes.size){case 0:return!1;case 1:return this.hasCurrentStyleProperties();default:return!0}}hasCurrentStyleProperties(){return this._currentKeyframe.size>0}get currentTime(){return this.startTime+this.duration}delayNextStep(e){let A=this._keyframes.size===1&&this._pendingStyles.size;this.duration||A?(this.forwardTime(this.currentTime+e),A&&this.snapshotCurrentStyles()):this.startTime+=e}fork(e,A){return this.applyStylesToKeyframe(),new t(this._driver,e,A||this.currentTime,this._elementTimelineStylesLookup)}_loadKeyframe(){this._currentKeyframe&&(this._previousKeyframe=this._currentKeyframe),this._currentKeyframe=this._keyframes.get(this.duration),this._currentKeyframe||(this._currentKeyframe=new Map,this._keyframes.set(this.duration,this._currentKeyframe))}forwardFrame(){this.duration+=yVA,this._loadKeyframe()}forwardTime(e){this.applyStylesToKeyframe(),this.duration=e,this._loadKeyframe()}_updateStyle(e,A){this._localTimelineStyles.set(e,A),this._globalTimelineStyles.set(e,A),this._styleSummary.set(e,{time:this.currentTime,value:A})}allowOnlyTimelineStyles(){return this._currentEmptyStepKeyframe!==this._currentKeyframe}applyEmptyStep(e){e&&this._previousKeyframe.set("easing",e);for(let[A,i]of this._globalTimelineStyles)this._backFill.set(A,i||kc),this._currentKeyframe.set(A,kc);this._currentEmptyStepKeyframe=this._currentKeyframe}setStyles(e,A,i,n){A&&this._previousKeyframe.set("easing",A);let o=n&&n.params||{},r=SVA(e,this._globalTimelineStyles);for(let[s,a]of r){let c=UQ(a,o,i);this._pendingStyles.set(s,c),this._localTimelineStyles.has(s)||this._backFill.set(s,this._globalTimelineStyles.get(s)??kc),this._updateStyle(s,c)}}applyStylesToKeyframe(){this._pendingStyles.size!=0&&(this._pendingStyles.forEach((e,A)=>{this._currentKeyframe.set(A,e)}),this._pendingStyles.clear(),this._localTimelineStyles.forEach((e,A)=>{this._currentKeyframe.has(A)||this._currentKeyframe.set(A,e)}))}snapshotCurrentStyles(){for(let[e,A]of this._localTimelineStyles)this._pendingStyles.set(e,A),this._updateStyle(e,A)}getFinalKeyframe(){return this._keyframes.get(this.duration)}get properties(){let e=[];for(let A in this._currentKeyframe)e.push(A);return e}mergeTimelineCollectedStyles(e){e._styleSummary.forEach((A,i)=>{let n=this._styleSummary.get(i);(!n||A.time>n.time)&&this._updateStyle(i,A.value)})}buildKeyframes(){this.applyStylesToKeyframe();let e=new Set,A=new Set,i=this._keyframes.size===1&&this.duration===0,n=[];this._keyframes.forEach((s,a)=>{let c=new Map([...this._backFill,...s]);c.forEach((l,I)=>{l===FB?e.add(I):l===kc&&A.add(I)}),i||c.set("offset",a/this.duration),n.push(c)});let o=[...e.values()],r=[...A.values()];if(i){let s=n[0],a=new Map(s);s.set("offset",0),a.set("offset",1),n=[s,a]}return HU(this.element,n,o,r,this.duration,this.startTime,this.easing,!1)}},NU=class extends S7{keyframes;preStyleProps;postStyleProps;_stretchStartingKeyframe;timings;constructor(e,A,i,n,o,r,s=!1){super(e,A,r.delay),this.keyframes=i,this.preStyleProps=n,this.postStyleProps=o,this._stretchStartingKeyframe=s,this.timings={duration:r.duration,delay:r.delay,easing:r.easing}}containsAnimation(){return this.keyframes.length>1}buildKeyframes(){let e=this.keyframes,{delay:A,duration:i,easing:n}=this.timings;if(this._stretchStartingKeyframe&&A){let o=[],r=i+A,s=A/r,a=new Map(e[0]);a.set("offset",0),o.push(a);let c=new Map(e[0]);c.set("offset",dgA(s)),o.push(c);let l=e.length-1;for(let I=1;I<=l;I++){let C=new Map(e[I]),d=C.get("offset"),B=A+d*i;C.set("offset",dgA(B/r)),o.push(C)}i=r,A=0,n="",e=o}return HU(this.element,e,this.preStyleProps,this.postStyleProps,i,A,n,!0)}};function dgA(t,e=3){let A=Math.pow(10,e-1);return Math.round(t*A)/A}function SVA(t,e){let A=new Map,i;return t.forEach(n=>{if(n==="*"){i??=e.keys();for(let o of i)A.set(o,kc)}else for(let[o,r]of n)A.set(o,r)}),A}function BgA(t,e,A,i,n,o,r,s,a,c,l,I,C){return{type:0,element:t,triggerName:e,isRemovalTransition:n,fromState:A,fromStyles:o,toState:i,toStyles:r,timelines:s,queriedElements:a,preStyleProps:c,postStyleProps:l,totalTime:I,errors:C}}var bU={},R7=class{_triggerName;ast;_stateStyles;constructor(e,A,i){this._triggerName=e,this.ast=A,this._stateStyles=i}match(e,A,i,n){return RVA(this.ast.matchers,e,A,i,n)}buildStyles(e,A,i){let n=this._stateStyles.get("*");return e!==void 0&&(n=this._stateStyles.get(e?.toString())||n),n?n.buildStyles(A,i):new Map}build(e,A,i,n,o,r,s,a,c,l){let I=[],C=this.ast.options&&this.ast.options.params||bU,d=s&&s.params||bU,B=this.buildStyles(i,d,I),E=a&&a.params||bU,h=this.buildStyles(n,E,I),u=new Set,D=new Map,L=new Map,R=n==="void",w={params:ygA(E,C),delay:this.ast.options?.delay},_=l?[]:DgA(e,A,this.ast.animation,o,r,B,h,w,c,I),K=0;return _.forEach(z=>{K=Math.max(z.duration+z.delay,K)}),I.length?BgA(A,this._triggerName,i,n,R,B,h,[],[],D,L,K,I):(_.forEach(z=>{let U=z.element,H=Ya(D,U,new Set);z.preStyleProps.forEach(j=>H.add(j));let q=Ya(L,U,new Set);z.postStyleProps.forEach(j=>q.add(j)),U!==A&&u.add(U)}),BgA(A,this._triggerName,i,n,R,B,h,_,[...u.values()],D,L,K))}};function RVA(t,e,A,i,n){return t.some(o=>o(e,A,i,n))}function ygA(t,e){let A=rA({},e);return Object.entries(t).forEach(([i,n])=>{n!=null&&(A[i]=n)}),A}var _U=class{styles;defaultParams;normalizer;constructor(e,A,i){this.styles=e,this.defaultParams=A,this.normalizer=i}buildStyles(e,A){let i=new Map,n=ygA(e,this.defaultParams);return this.styles.styles.forEach(o=>{typeof o!="string"&&o.forEach((r,s)=>{r&&(r=UQ(r,n,A));let a=this.normalizer.normalizePropertyName(s,A);r=this.normalizer.normalizeStyleValue(s,a,r,A),i.set(s,r)})}),i}};function xVA(t,e,A){return new GU(t,e,A)}var GU=class{name;ast;_normalizer;transitionFactories=[];fallbackTransition;states=new Map;constructor(e,A,i){this.name=e,this.ast=A,this._normalizer=i,A.states.forEach(n=>{let o=n.options&&n.options.params||{};this.states.set(n.name,new _U(n.style,o,i))}),EgA(this.states,"true","1"),EgA(this.states,"false","0"),A.transitions.forEach(n=>{this.transitionFactories.push(new R7(e,n,this.states))}),this.fallbackTransition=LVA(e,this.states)}get containsQueries(){return this.ast.queryCount>0}matchTransition(e,A,i,n){return this.transitionFactories.find(r=>r.match(e,A,i,n))||null}matchStyles(e,A,i){return this.fallbackTransition.buildStyles(e,A,i)}};function LVA(t,e,A){let i=[(r,s)=>!0],n={type:Ci.Sequence,steps:[],options:null},o={type:Ci.Transition,animation:n,matchers:i,options:null,queryCount:0,depCount:0};return new R7(t,o,e)}function EgA(t,e,A){t.has(e)?t.has(A)||t.set(A,t.get(e)):t.has(A)&&t.set(e,t.get(A))}var FVA=new Kf,UU=class{bodyNode;_driver;_normalizer;_animations=new Map;_playersById=new Map;players=[];constructor(e,A,i){this.bodyNode=e,this._driver=A,this._normalizer=i}register(e,A){let i=[],n=[],o=wgA(this._driver,A,i,n);if(i.length)throw WlA(i);this._animations.set(e,o)}_buildPlayer(e,A,i){let n=e.element,o=hU(this._normalizer,e.keyframes,A,i);return this._driver.animate(n,o,e.duration,e.delay,e.easing,[],!0)}create(e,A,i={}){let n=[],o=this._animations.get(e),r,s=new Map;if(o?(r=DgA(this._driver,A,o,wU,u7,new Map,new Map,i,FVA,n),r.forEach(l=>{let I=Ya(s,l.element,new Map);l.postStyleProps.forEach(C=>I.set(C,null))})):(n.push(XlA()),r=[]),n.length)throw $lA(n);s.forEach((l,I)=>{l.forEach((C,d)=>{l.set(d,this._driver.computeStyle(I,d,kc))})});let a=r.map(l=>{let I=s.get(l.element);return this._buildPlayer(l,new Map,I)}),c=c2(a);return this._playersById.set(e,c),c.onDestroy(()=>this.destroy(e)),this.players.push(c),c}destroy(e){let A=this._getPlayer(e);A.destroy(),this._playersById.delete(e);let i=this.players.indexOf(A);i>=0&&this.players.splice(i,1)}_getPlayer(e){let A=this._playersById.get(e);if(!A)throw AgA(e);return A}listen(e,A,i,n){let o=Q7(A,"","","");return E7(this._getPlayer(e),i,o,n),()=>{}}command(e,A,i,n){if(i=="register"){this.register(e,n[0]);return}if(i=="create"){let r=n[0]||{};this.create(e,A,r);return}let o=this._getPlayer(e);switch(i){case"play":o.play();break;case"pause":o.pause();break;case"reset":o.reset();break;case"restart":o.restart();break;case"finish":o.finish();break;case"init":o.init();break;case"setPosition":o.setPosition(parseFloat(n[0]));break;case"destroy":this.destroy(e);break}}},QgA="ng-animate-queued",NVA=".ng-animate-queued",MU="ng-animate-disabled",_VA=".ng-animate-disabled",GVA="ng-star-inserted",UVA=".ng-star-inserted",KVA=[],vgA={namespaceId:"",setForRemoval:!1,setForMove:!1,hasAnimation:!1,removedBeforeQueried:!1},YVA={namespaceId:"",setForMove:!1,setForRemoval:!1,hasAnimation:!1,removedBeforeQueried:!0},Tl="__ng_removed",Yf=class{namespaceId;value;options;get params(){return this.options.params}constructor(e,A=""){this.namespaceId=A;let i=e&&e.hasOwnProperty("value"),n=i?e.value:e;if(this.value=TVA(n),i){let o=e,{value:r}=o,s=J7(o,["value"]);this.options=s}else this.options={};this.options.params||(this.options.params={})}absorbOptions(e){let A=e.params;if(A){let i=this.options.params;Object.keys(A).forEach(n=>{i[n]==null&&(i[n]=A[n])})}}},Uf="void",kU=new Yf(Uf),KU=class{id;hostElement;_engine;players=[];_triggers=new Map;_queue=[];_elementListeners=new Map;_hostClassName;constructor(e,A,i){this.id=e,this.hostElement=A,this._engine=i,this._hostClassName="ng-tns-"+e,Zc(A,this._hostClassName)}listen(e,A,i,n){if(!this._triggers.has(A))throw egA(i,A);if(i==null||i.length==0)throw tgA(A);if(!HVA(i))throw igA(i,A);let o=Ya(this._elementListeners,e,[]),r={name:A,phase:i,callback:n};o.push(r);let s=Ya(this._engine.statesByElement,e,new Map);return s.has(A)||(Zc(e,Nf),Zc(e,Nf+"-"+A),s.set(A,kU)),()=>{this._engine.afterFlush(()=>{let a=o.indexOf(r);a>=0&&o.splice(a,1),this._triggers.has(A)||s.delete(A)})}}register(e,A){return this._triggers.has(e)?!1:(this._triggers.set(e,A),!0)}_getTrigger(e){let A=this._triggers.get(e);if(!A)throw ngA(e);return A}trigger(e,A,i,n=!0){let o=this._getTrigger(A),r=new Jf(this.id,A,e),s=this._engine.statesByElement.get(e);s||(Zc(e,Nf),Zc(e,Nf+"-"+A),this._engine.statesByElement.set(e,s=new Map));let a=s.get(A),c=new Yf(i,this.id);if(!(i&&i.hasOwnProperty("value"))&&a&&c.absorbOptions(a.options),s.set(A,c),a||(a=kU),!(c.value===Uf)&&a.value===c.value){if(!PVA(a.params,c.params)){let E=[],h=o.matchStyles(a.value,a.params,E),u=o.matchStyles(c.value,c.params,E);E.length?this._engine.reportError(E):this._engine.afterFlush(()=>{rI(e,h),Jl(e,u)})}return}let C=Ya(this._engine.playersByElement,e,[]);C.forEach(E=>{E.namespaceId==this.id&&E.triggerName==A&&E.queued&&E.destroy()});let d=o.matchTransition(a.value,c.value,e,c.params),B=!1;if(!d){if(!n)return;d=o.fallbackTransition,B=!0}return this._engine.totalQueuedPlayers++,this._queue.push({element:e,triggerName:A,transition:d,fromState:a,toState:c,player:r,isFallbackTransition:B}),B||(Zc(e,QgA),r.onStart(()=>{KQ(e,QgA)})),r.onDone(()=>{let E=this.players.indexOf(r);E>=0&&this.players.splice(E,1);let h=this._engine.playersByElement.get(e);if(h){let u=h.indexOf(r);u>=0&&h.splice(u,1)}}),this.players.push(r),C.push(r),r}deregister(e){this._triggers.delete(e),this._engine.statesByElement.forEach(A=>A.delete(e)),this._elementListeners.forEach((A,i)=>{this._elementListeners.set(i,A.filter(n=>n.name!=e))})}clearElementCache(e){this._engine.statesByElement.delete(e),this._elementListeners.delete(e);let A=this._engine.playersByElement.get(e);A&&(A.forEach(i=>i.destroy()),this._engine.playersByElement.delete(e))}_signalRemovalForInnerTriggers(e,A){let i=this._engine.driver.query(e,_f,!0);i.forEach(n=>{if(n[Tl])return;let o=this._engine.fetchNamespacesByElement(n);o.size?o.forEach(r=>r.triggerLeaveAnimation(n,A,!1,!0)):this.clearElementCache(n)}),this._engine.afterFlushAnimationsDone(()=>i.forEach(n=>this.clearElementCache(n)))}triggerLeaveAnimation(e,A,i,n){let o=this._engine.statesByElement.get(e),r=new Map;if(o){let s=[];if(o.forEach((a,c)=>{if(r.set(c,a.value),this._triggers.has(c)){let l=this.trigger(e,c,Uf,n);l&&s.push(l)}}),s.length)return this._engine.markElementAsRemoved(this.id,e,!0,A,r),i&&c2(s).onDone(()=>this._engine.processLeaveNode(e)),!0}return!1}prepareLeaveAnimationListeners(e){let A=this._elementListeners.get(e),i=this._engine.statesByElement.get(e);if(A&&i){let n=new Set;A.forEach(o=>{let r=o.name;if(n.has(r))return;n.add(r);let a=this._triggers.get(r).fallbackTransition,c=i.get(r)||kU,l=new Yf(Uf),I=new Jf(this.id,r,e);this._engine.totalQueuedPlayers++,this._queue.push({element:e,triggerName:r,transition:a,fromState:c,toState:l,player:I,isFallbackTransition:!0})})}}removeNode(e,A){let i=this._engine;if(e.childElementCount&&this._signalRemovalForInnerTriggers(e,A),this.triggerLeaveAnimation(e,A,!0))return;let n=!1;if(i.totalAnimations){let o=i.players.length?i.playersByQueriedElement.get(e):[];if(o&&o.length)n=!0;else{let r=e;for(;r=r.parentNode;)if(i.statesByElement.get(r)){n=!0;break}}}if(this.prepareLeaveAnimationListeners(e),n)i.markElementAsRemoved(this.id,e,!1,A);else{let o=e[Tl];(!o||o===vgA)&&(i.afterFlush(()=>this.clearElementCache(e)),i.destroyInnerAnimations(e),i._onRemovalComplete(e,A))}}insertNode(e,A){Zc(e,this._hostClassName)}drainQueuedTransitions(e){let A=[];return this._queue.forEach(i=>{let n=i.player;if(n.destroyed)return;let o=i.element,r=this._elementListeners.get(o);r&&r.forEach(s=>{if(s.name==i.triggerName){let a=Q7(o,i.triggerName,i.fromState.value,i.toState.value);a._data=e,E7(i.player,s.phase,a,s.callback)}}),n.markedForDestroy?this._engine.afterFlush(()=>{n.destroy()}):A.push(i)}),this._queue=[],A.sort((i,n)=>{let o=i.transition.ast.depCount,r=n.transition.ast.depCount;return o==0||r==0?o-r:this._engine.driver.containsElement(i.element,n.element)?1:-1})}destroy(e){this.players.forEach(A=>A.destroy()),this._signalRemovalForInnerTriggers(this.hostElement,e)}},YU=class{bodyNode;driver;_normalizer;players=[];newHostElements=new Map;playersByElement=new Map;playersByQueriedElement=new Map;statesByElement=new Map;disabledNodes=new Set;totalAnimations=0;totalQueuedPlayers=0;_namespaceLookup={};_namespaceList=[];_flushFns=[];_whenQuietFns=[];namespacesByHostElement=new Map;collectedEnterElements=[];collectedLeaveElements=[];onRemovalComplete=(e,A)=>{};_onRemovalComplete(e,A){this.onRemovalComplete(e,A)}constructor(e,A,i){this.bodyNode=e,this.driver=A,this._normalizer=i}get queuedPlayers(){let e=[];return this._namespaceList.forEach(A=>{A.players.forEach(i=>{i.queued&&e.push(i)})}),e}createNamespace(e,A){let i=new KU(e,A,this);return this.bodyNode&&this.driver.containsElement(this.bodyNode,A)?this._balanceNamespaceList(i,A):(this.newHostElements.set(A,i),this.collectEnterElement(A)),this._namespaceLookup[e]=i}_balanceNamespaceList(e,A){let i=this._namespaceList,n=this.namespacesByHostElement;if(i.length-1>=0){let r=!1,s=this.driver.getParentElement(A);for(;s;){let a=n.get(s);if(a){let c=i.indexOf(a);i.splice(c+1,0,e),r=!0;break}s=this.driver.getParentElement(s)}r||i.unshift(e)}else i.push(e);return n.set(A,e),e}register(e,A){let i=this._namespaceLookup[e];return i||(i=this.createNamespace(e,A)),i}registerTrigger(e,A,i){let n=this._namespaceLookup[e];n&&n.register(A,i)&&this.totalAnimations++}destroy(e,A){e&&(this.afterFlush(()=>{}),this.afterFlushAnimationsDone(()=>{let i=this._fetchNamespace(e);this.namespacesByHostElement.delete(i.hostElement);let n=this._namespaceList.indexOf(i);n>=0&&this._namespaceList.splice(n,1),i.destroy(A),delete this._namespaceLookup[e]}))}_fetchNamespace(e){return this._namespaceLookup[e]}fetchNamespacesByElement(e){let A=new Set,i=this.statesByElement.get(e);if(i){for(let n of i.values())if(n.namespaceId){let o=this._fetchNamespace(n.namespaceId);o&&A.add(o)}}return A}trigger(e,A,i,n){if(y7(A)){let o=this._fetchNamespace(e);if(o)return o.trigger(A,i,n),!0}return!1}insertNode(e,A,i,n){if(!y7(A))return;let o=A[Tl];if(o&&o.setForRemoval){o.setForRemoval=!1,o.setForMove=!0;let r=this.collectedLeaveElements.indexOf(A);r>=0&&this.collectedLeaveElements.splice(r,1)}if(e){let r=this._fetchNamespace(e);r&&r.insertNode(A,i)}n&&this.collectEnterElement(A)}collectEnterElement(e){this.collectedEnterElements.push(e)}markElementAsDisabled(e,A){A?this.disabledNodes.has(e)||(this.disabledNodes.add(e),Zc(e,MU)):this.disabledNodes.has(e)&&(this.disabledNodes.delete(e),KQ(e,MU))}removeNode(e,A,i){if(y7(A)){let n=e?this._fetchNamespace(e):null;n?n.removeNode(A,i):this.markElementAsRemoved(e,A,!1,i);let o=this.namespacesByHostElement.get(A);o&&o.id!==e&&o.removeNode(A,i)}else this._onRemovalComplete(A,i)}markElementAsRemoved(e,A,i,n,o){this.collectedLeaveElements.push(A),A[Tl]={namespaceId:e,setForRemoval:n,hasAnimation:i,removedBeforeQueried:!1,previousTriggersValues:o}}listen(e,A,i,n,o){return y7(A)?this._fetchNamespace(e).listen(A,i,n,o):()=>{}}_buildInstruction(e,A,i,n,o){return e.transition.build(this.driver,e.element,e.fromState.value,e.toState.value,i,n,e.fromState.options,e.toState.options,A,o)}destroyInnerAnimations(e){let A=this.driver.query(e,_f,!0);A.forEach(i=>this.destroyActiveAnimationsForElement(i)),this.playersByQueriedElement.size!=0&&(A=this.driver.query(e,f7,!0),A.forEach(i=>this.finishActiveQueriedAnimationOnElement(i)))}destroyActiveAnimationsForElement(e){let A=this.playersByElement.get(e);A&&A.forEach(i=>{i.queued?i.markedForDestroy=!0:i.destroy()})}finishActiveQueriedAnimationOnElement(e){let A=this.playersByQueriedElement.get(e);A&&A.forEach(i=>i.finish())}whenRenderingDone(){return new Promise(e=>{if(this.players.length)return c2(this.players).onDone(()=>e());e()})}processLeaveNode(e){let A=e[Tl];if(A&&A.setForRemoval){if(e[Tl]=vgA,A.namespaceId){this.destroyInnerAnimations(e);let i=this._fetchNamespace(A.namespaceId);i&&i.clearElementCache(e)}this._onRemovalComplete(e,A.setForRemoval)}e.classList?.contains(MU)&&this.markElementAsDisabled(e,!1),this.driver.query(e,_VA,!0).forEach(i=>{this.markElementAsDisabled(i,!1)})}flush(e=-1){let A=[];if(this.newHostElements.size&&(this.newHostElements.forEach((i,n)=>this._balanceNamespaceList(i,n)),this.newHostElements.clear()),this.totalAnimations&&this.collectedEnterElements.length)for(let i=0;i<this.collectedEnterElements.length;i++){let n=this.collectedEnterElements[i];Zc(n,GVA)}if(this._namespaceList.length&&(this.totalQueuedPlayers||this.collectedLeaveElements.length)){let i=[];try{A=this._flushAnimations(i,e)}finally{for(let n=0;n<i.length;n++)i[n]()}}else for(let i=0;i<this.collectedLeaveElements.length;i++){let n=this.collectedLeaveElements[i];this.processLeaveNode(n)}if(this.totalQueuedPlayers=0,this.collectedEnterElements.length=0,this.collectedLeaveElements.length=0,this._flushFns.forEach(i=>i()),this._flushFns=[],this._whenQuietFns.length){let i=this._whenQuietFns;this._whenQuietFns=[],A.length?c2(A).onDone(()=>{i.forEach(n=>n())}):i.forEach(n=>n())}}reportError(e){throw ogA(e)}_flushAnimations(e,A){let i=new Kf,n=[],o=new Map,r=[],s=new Map,a=new Map,c=new Map,l=new Set;this.disabledNodes.forEach(lA=>{l.add(lA);let vA=this.driver.query(lA,NVA,!0);for(let tA=0;tA<vA.length;tA++)l.add(vA[tA])});let I=this.bodyNode,C=Array.from(this.statesByElement.keys()),d=fgA(C,this.collectedEnterElements),B=new Map,E=0;d.forEach((lA,vA)=>{let tA=wU+E++;B.set(vA,tA),lA.forEach(cA=>Zc(cA,tA))});let h=[],u=new Set,D=new Set;for(let lA=0;lA<this.collectedLeaveElements.length;lA++){let vA=this.collectedLeaveElements[lA],tA=vA[Tl];tA&&tA.setForRemoval&&(h.push(vA),u.add(vA),tA.hasAnimation?this.driver.query(vA,UVA,!0).forEach(cA=>u.add(cA)):D.add(vA))}let L=new Map,R=fgA(C,Array.from(u));R.forEach((lA,vA)=>{let tA=u7+E++;L.set(vA,tA),lA.forEach(cA=>Zc(cA,tA))}),e.push(()=>{d.forEach((lA,vA)=>{let tA=B.get(vA);lA.forEach(cA=>KQ(cA,tA))}),R.forEach((lA,vA)=>{let tA=L.get(vA);lA.forEach(cA=>KQ(cA,tA))}),h.forEach(lA=>{this.processLeaveNode(lA)})});let w=[],_=[];for(let lA=this._namespaceList.length-1;lA>=0;lA--)this._namespaceList[lA].drainQueuedTransitions(A).forEach(tA=>{let cA=tA.player,pA=tA.element;if(w.push(cA),this.collectedEnterElements.length){let He=pA[Tl];if(He&&He.setForMove){if(He.previousTriggersValues&&He.previousTriggersValues.has(tA.triggerName)){let uA=He.previousTriggersValues.get(tA.triggerName),eA=this.statesByElement.get(tA.element);if(eA&&eA.has(tA.triggerName)){let UA=eA.get(tA.triggerName);UA.value=uA,eA.set(tA.triggerName,UA)}}cA.destroy();return}}let VA=!I||!this.driver.containsElement(I,pA),oe=L.get(pA),KA=B.get(pA),CA=this._buildInstruction(tA,i,KA,oe,VA);if(CA.errors&&CA.errors.length){_.push(CA);return}if(VA){cA.onStart(()=>rI(pA,CA.fromStyles)),cA.onDestroy(()=>Jl(pA,CA.toStyles)),n.push(cA);return}if(tA.isFallbackTransition){cA.onStart(()=>rI(pA,CA.fromStyles)),cA.onDestroy(()=>Jl(pA,CA.toStyles)),n.push(cA);return}let TA=[];CA.timelines.forEach(He=>{He.stretchStartingKeyframe=!0,this.disabledNodes.has(He.element)||TA.push(He)}),CA.timelines=TA,i.append(pA,CA.timelines);let Ze={instruction:CA,player:cA,element:pA};r.push(Ze),CA.queriedElements.forEach(He=>Ya(s,He,[]).push(cA)),CA.preStyleProps.forEach((He,uA)=>{if(He.size){let eA=a.get(uA);eA||a.set(uA,eA=new Set),He.forEach((UA,aA)=>eA.add(aA))}}),CA.postStyleProps.forEach((He,uA)=>{let eA=c.get(uA);eA||c.set(uA,eA=new Set),He.forEach((UA,aA)=>eA.add(aA))})});if(_.length){let lA=[];_.forEach(vA=>{lA.push(rgA(vA.triggerName,vA.errors))}),w.forEach(vA=>vA.destroy()),this.reportError(lA)}let K=new Map,z=new Map;r.forEach(lA=>{let vA=lA.element;i.has(vA)&&(z.set(vA,vA),this._beforeAnimationBuild(lA.player.namespaceId,lA.instruction,K))}),n.forEach(lA=>{let vA=lA.element;this._getPreviousPlayers(vA,!1,lA.namespaceId,lA.triggerName,null).forEach(cA=>{Ya(K,vA,[]).push(cA),cA.destroy()})});let U=h.filter(lA=>mgA(lA,a,c)),H=new Map;ugA(H,this.driver,D,c,kc).forEach(lA=>{mgA(lA,a,c)&&U.push(lA)});let j=new Map;d.forEach((lA,vA)=>{ugA(j,this.driver,new Set(lA),a,FB)}),U.forEach(lA=>{let vA=H.get(lA),tA=j.get(lA);H.set(lA,new Map([...vA?.entries()??[],...tA?.entries()??[]]))});let gA=[],QA=[],BA={};r.forEach(lA=>{let{element:vA,player:tA,instruction:cA}=lA;if(i.has(vA)){if(l.has(vA)){tA.onDestroy(()=>Jl(vA,cA.toStyles)),tA.disabled=!0,tA.overrideTotalTime(cA.totalTime),n.push(tA);return}let pA=BA;if(z.size>1){let oe=vA,KA=[];for(;oe=oe.parentNode;){let CA=z.get(oe);if(CA){pA=CA;break}KA.push(oe)}KA.forEach(CA=>z.set(CA,pA))}let VA=this._buildAnimation(tA.namespaceId,cA,K,o,j,H);if(tA.setRealPlayer(VA),pA===BA)gA.push(tA);else{let oe=this.playersByElement.get(pA);oe&&oe.length&&(tA.parentPlayer=c2(oe)),n.push(tA)}}else rI(vA,cA.fromStyles),tA.onDestroy(()=>Jl(vA,cA.toStyles)),QA.push(tA),l.has(vA)&&n.push(tA)}),QA.forEach(lA=>{let vA=o.get(lA.element);if(vA&&vA.length){let tA=c2(vA);lA.setRealPlayer(tA)}}),n.forEach(lA=>{lA.parentPlayer?lA.syncPlayerEvents(lA.parentPlayer):lA.destroy()});for(let lA=0;lA<h.length;lA++){let vA=h[lA],tA=vA[Tl];if(KQ(vA,u7),tA&&tA.hasAnimation)continue;let cA=[];if(s.size){let VA=s.get(vA);VA&&VA.length&&cA.push(...VA);let oe=this.driver.query(vA,f7,!0);for(let KA=0;KA<oe.length;KA++){let CA=s.get(oe[KA]);CA&&CA.length&&cA.push(...CA)}}let pA=cA.filter(VA=>!VA.destroyed);pA.length?zVA(this,vA,pA):this.processLeaveNode(vA)}return h.length=0,gA.forEach(lA=>{this.players.push(lA),lA.onDone(()=>{lA.destroy();let vA=this.players.indexOf(lA);this.players.splice(vA,1)}),lA.play()}),gA}afterFlush(e){this._flushFns.push(e)}afterFlushAnimationsDone(e){this._whenQuietFns.push(e)}_getPreviousPlayers(e,A,i,n,o){let r=[];if(A){let s=this.playersByQueriedElement.get(e);s&&(r=s)}else{let s=this.playersByElement.get(e);if(s){let a=!o||o==Uf;s.forEach(c=>{c.queued||!a&&c.triggerName!=n||r.push(c)})}}return(i||n)&&(r=r.filter(s=>!(i&&i!=s.namespaceId||n&&n!=s.triggerName))),r}_beforeAnimationBuild(e,A,i){let n=A.triggerName,o=A.element,r=A.isRemovalTransition?void 0:e,s=A.isRemovalTransition?void 0:n;for(let a of A.timelines){let c=a.element,l=c!==o,I=Ya(i,c,[]);this._getPreviousPlayers(c,l,r,s,A.toState).forEach(d=>{let B=d.getRealPlayer();B.beforeDestroy&&B.beforeDestroy(),d.destroy(),I.push(d)})}rI(o,A.fromStyles)}_buildAnimation(e,A,i,n,o,r){let s=A.triggerName,a=A.element,c=[],l=new Set,I=new Set,C=A.timelines.map(B=>{let E=B.element;l.add(E);let h=E[Tl];if(h&&h.removedBeforeQueried)return new Eg(B.duration,B.delay);let u=E!==a,D=OVA((i.get(E)||KVA).map(K=>K.getRealPlayer())).filter(K=>{let z=K;return z.element?z.element===E:!1}),L=o.get(E),R=r.get(E),w=hU(this._normalizer,B.keyframes,L,R),_=this._buildPlayer(B,w,D);if(B.subTimeline&&n&&I.add(E),u){let K=new Jf(e,s,E);K.setRealPlayer(_),c.push(K)}return _});c.forEach(B=>{Ya(this.playersByQueriedElement,B.element,[]).push(B),B.onDone(()=>JVA(this.playersByQueriedElement,B.element,B))}),l.forEach(B=>Zc(B,DU));let d=c2(C);return d.onDestroy(()=>{l.forEach(B=>KQ(B,DU)),Jl(a,A.toStyles)}),I.forEach(B=>{Ya(n,B,[]).push(d)}),d}_buildPlayer(e,A,i){return A.length>0?this.driver.animate(e.element,A,e.duration,e.delay,e.easing,i):new Eg(e.duration,e.delay)}},Jf=class{namespaceId;triggerName;element;_player=new Eg;_containsRealPlayer=!1;_queuedCallbacks=new Map;destroyed=!1;parentPlayer=null;markedForDestroy=!1;disabled=!1;queued=!0;totalTime=0;constructor(e,A,i){this.namespaceId=e,this.triggerName=A,this.element=i}setRealPlayer(e){this._containsRealPlayer||(this._player=e,this._queuedCallbacks.forEach((A,i)=>{A.forEach(n=>E7(e,i,void 0,n))}),this._queuedCallbacks.clear(),this._containsRealPlayer=!0,this.overrideTotalTime(e.totalTime),this.queued=!1)}getRealPlayer(){return this._player}overrideTotalTime(e){this.totalTime=e}syncPlayerEvents(e){let A=this._player;A.triggerCallback&&e.onStart(()=>A.triggerCallback("start")),e.onDone(()=>this.finish()),e.onDestroy(()=>this.destroy())}_queueEvent(e,A){Ya(this._queuedCallbacks,e,[]).push(A)}onDone(e){this.queued&&this._queueEvent("done",e),this._player.onDone(e)}onStart(e){this.queued&&this._queueEvent("start",e),this._player.onStart(e)}onDestroy(e){this.queued&&this._queueEvent("destroy",e),this._player.onDestroy(e)}init(){this._player.init()}hasStarted(){return this.queued?!1:this._player.hasStarted()}play(){!this.queued&&this._player.play()}pause(){!this.queued&&this._player.pause()}restart(){!this.queued&&this._player.restart()}finish(){this._player.finish()}destroy(){this.destroyed=!0,this._player.destroy()}reset(){!this.queued&&this._player.reset()}setPosition(e){this.queued||this._player.setPosition(e)}getPosition(){return this.queued?0:this._player.getPosition()}triggerCallback(e){let A=this._player;A.triggerCallback&&A.triggerCallback(e)}};function JVA(t,e,A){let i=t.get(e);if(i){if(i.length){let n=i.indexOf(A);i.splice(n,1)}i.length==0&&t.delete(e)}return i}function TVA(t){return t??null}function y7(t){return t&&t.nodeType===1}function HVA(t){return t=="start"||t=="done"}function hgA(t,e){let A=t.style.display;return t.style.display=e??"none",A}function ugA(t,e,A,i,n){let o=[];A.forEach(a=>o.push(hgA(a)));let r=[];i.forEach((a,c)=>{let l=new Map;a.forEach(I=>{let C=e.computeStyle(c,I,n);l.set(I,C),(!C||C.length==0)&&(c[Tl]=YVA,r.push(c))}),t.set(c,l)});let s=0;return A.forEach(a=>hgA(a,o[s++])),r}function fgA(t,e){let A=new Map;if(t.forEach(s=>A.set(s,[])),e.length==0)return A;let i=1,n=new Set(e),o=new Map;function r(s){if(!s)return i;let a=o.get(s);if(a)return a;let c=s.parentNode;return A.has(c)?a=c:n.has(c)?a=i:a=r(c),o.set(s,a),a}return e.forEach(s=>{let a=r(s);a!==i&&A.get(a).push(s)}),A}function Zc(t,e){t.classList?.add(e)}function KQ(t,e){t.classList?.remove(e)}function zVA(t,e,A){c2(A).onDone(()=>t.processLeaveNode(e))}function OVA(t){let e=[];return bgA(t,e),e}function bgA(t,e){for(let A=0;A<t.length;A++){let i=t[A];i instanceof tC?bgA(i.players,e):e.push(i)}}function PVA(t,e){let A=Object.keys(t),i=Object.keys(e);if(A.length!=i.length)return!1;for(let n=0;n<A.length;n++){let o=A[n];if(!e.hasOwnProperty(o)||t[o]!==e[o])return!1}return!0}function mgA(t,e,A){let i=A.get(t);if(!i)return!1;let n=e.get(t);return n?i.forEach(o=>n.add(o)):e.set(t,i),A.delete(t),!0}var YQ=class{_driver;_normalizer;_transitionEngine;_timelineEngine;_triggerCache={};onRemovalComplete=(e,A)=>{};constructor(e,A,i){this._driver=A,this._normalizer=i,this._transitionEngine=new YU(e.body,A,i),this._timelineEngine=new UU(e.body,A,i),this._transitionEngine.onRemovalComplete=(n,o)=>this.onRemovalComplete(n,o)}registerTrigger(e,A,i,n,o){let r=e+"-"+n,s=this._triggerCache[r];if(!s){let a=[],c=[],l=wgA(this._driver,o,a,c);if(a.length)throw ZlA(n,a);s=xVA(n,l,this._normalizer),this._triggerCache[r]=s}this._transitionEngine.registerTrigger(A,n,s)}register(e,A){this._transitionEngine.register(e,A)}destroy(e,A){this._transitionEngine.destroy(e,A)}onInsert(e,A,i,n){this._transitionEngine.insertNode(e,A,i,n)}onRemove(e,A,i){this._transitionEngine.removeNode(e,A,i)}disableAnimations(e,A){this._transitionEngine.markElementAsDisabled(e,A)}process(e,A,i,n){if(i.charAt(0)=="@"){let[o,r]=uU(i),s=n;this._timelineEngine.command(o,A,r,s)}else this._transitionEngine.trigger(e,A,i,n)}listen(e,A,i,n,o){if(i.charAt(0)=="@"){let[r,s]=uU(i);return this._timelineEngine.listen(r,A,s,o)}return this._transitionEngine.listen(e,A,i,n,o)}flush(e=-1){this._transitionEngine.flush(e)}get players(){return[...this._transitionEngine.players,...this._timelineEngine.players]}whenRenderingDone(){return this._transitionEngine.whenRenderingDone()}afterFlushAnimationsDone(e){this._transitionEngine.afterFlushAnimationsDone(e)}};function jVA(t,e){let A=null,i=null;return Array.isArray(e)&&e.length?(A=SU(e[0]),e.length>1&&(i=SU(e[e.length-1]))):e instanceof Map&&(A=SU(e)),A||i?new qVA(t,A,i):null}var qVA=(()=>{class t{_element;_startStyles;_endStyles;static initialStylesByElement=new WeakMap;_state=0;_initialStyles;constructor(A,i,n){this._element=A,this._startStyles=i,this._endStyles=n;let o=t.initialStylesByElement.get(A);o||t.initialStylesByElement.set(A,o=new Map),this._initialStyles=o}start(){this._state<1&&(this._startStyles&&Jl(this._element,this._startStyles,this._initialStyles),this._state=1)}finish(){this.start(),this._state<2&&(Jl(this._element,this._initialStyles),this._endStyles&&(Jl(this._element,this._endStyles),this._endStyles=null),this._state=1)}destroy(){this.finish(),this._state<3&&(t.initialStylesByElement.delete(this._element),this._startStyles&&(rI(this._element,this._startStyles),this._endStyles=null),this._endStyles&&(rI(this._element,this._endStyles),this._endStyles=null),Jl(this._element,this._initialStyles),this._state=3)}}return t})();function SU(t){let e=null;return t.forEach((A,i)=>{VVA(i)&&(e=e||new Map,e.set(i,A))}),e}function VVA(t){return t==="display"||t==="position"}var x7=class{element;keyframes;options;_specialStyles;_onDoneFns=[];_onStartFns=[];_onDestroyFns=[];_duration;_delay;_initialized=!1;_finished=!1;_started=!1;_destroyed=!1;_finalKeyframe;_originalOnDoneFns=[];_originalOnStartFns=[];domPlayer;time=0;parentPlayer=null;currentSnapshot=new Map;constructor(e,A,i,n){this.element=e,this.keyframes=A,this.options=i,this._specialStyles=n,this._duration=i.duration,this._delay=i.delay||0,this.time=this._duration+this._delay}_onFinish(){this._finished||(this._finished=!0,this._onDoneFns.forEach(e=>e()),this._onDoneFns=[])}init(){this._buildPlayer(),this._preparePlayerBeforeStart()}_buildPlayer(){if(this._initialized)return;this._initialized=!0;let e=this.keyframes;this.domPlayer=this._triggerWebAnimation(this.element,e,this.options),this._finalKeyframe=e.length?e[e.length-1]:new Map;let A=()=>this._onFinish();this.domPlayer.addEventListener("finish",A),this.onDestroy(()=>{this.domPlayer.removeEventListener("finish",A)})}_preparePlayerBeforeStart(){this._delay?this._resetDomPlayerState():this.domPlayer.pause()}_convertKeyframesToObject(e){let A=[];return e.forEach(i=>{A.push(Object.fromEntries(i))}),A}_triggerWebAnimation(e,A,i){return e.animate(this._convertKeyframesToObject(A),i)}onStart(e){this._originalOnStartFns.push(e),this._onStartFns.push(e)}onDone(e){this._originalOnDoneFns.push(e),this._onDoneFns.push(e)}onDestroy(e){this._onDestroyFns.push(e)}play(){this._buildPlayer(),this.hasStarted()||(this._onStartFns.forEach(e=>e()),this._onStartFns=[],this._started=!0,this._specialStyles&&this._specialStyles.start()),this.domPlayer.play()}pause(){this.init(),this.domPlayer.pause()}finish(){this.init(),this._specialStyles&&this._specialStyles.finish(),this._onFinish(),this.domPlayer.finish()}reset(){this._resetDomPlayerState(),this._destroyed=!1,this._finished=!1,this._started=!1,this._onStartFns=this._originalOnStartFns,this._onDoneFns=this._originalOnDoneFns}_resetDomPlayerState(){this.domPlayer&&this.domPlayer.cancel()}restart(){this.reset(),this.play()}hasStarted(){return this._started}destroy(){this._destroyed||(this._destroyed=!0,this._resetDomPlayerState(),this._onFinish(),this._specialStyles&&this._specialStyles.destroy(),this._onDestroyFns.forEach(e=>e()),this._onDestroyFns=[])}setPosition(e){this.domPlayer===void 0&&this.init(),this.domPlayer.currentTime=e*this.time}getPosition(){return+(this.domPlayer.currentTime??0)/this.time}get totalTime(){return this._delay+this._duration}beforeDestroy(){let e=new Map;this.hasStarted()&&this._finalKeyframe.forEach((i,n)=>{n!=="offset"&&e.set(n,this._finished?i:p7(this.element,n))}),this.currentSnapshot=e}triggerCallback(e){let A=e==="start"?this._onStartFns:this._onDoneFns;A.forEach(i=>i()),A.length=0}},L7=class{validateStyleProperty(e){return!0}validateAnimatableStyleProperty(e){return!0}containsElement(e,A){return fU(e,A)}getParentElement(e){return h7(e)}query(e,A,i){return mU(e,A,i)}computeStyle(e,A,i){return p7(e,A)}animate(e,A,i,n,o,r=[]){let s=n==0?"both":"forwards",a={duration:i,delay:n,fill:s};o&&(a.easing=o);let c=new Map,l=r.filter(d=>d instanceof x7);lgA(i,n)&&l.forEach(d=>{d.currentSnapshot.forEach((B,E)=>c.set(E,B))});let I=agA(A).map(d=>new Map(d));I=ggA(e,I,c);let C=jVA(e,I);return new x7(e,I,a,C)}};var v7="@",MgA="@.disabled",F7=class{namespaceId;delegate;engine;_onDestroy;\u0275type=0;constructor(e,A,i,n){this.namespaceId=e,this.delegate=A,this.engine=i,this._onDestroy=n}get data(){return this.delegate.data}destroyNode(e){this.delegate.destroyNode?.(e)}destroy(){this.engine.destroy(this.namespaceId,this.delegate),this.engine.afterFlushAnimationsDone(()=>{queueMicrotask(()=>{this.delegate.destroy()})}),this._onDestroy?.()}createElement(e,A){return this.delegate.createElement(e,A)}createComment(e){return this.delegate.createComment(e)}createText(e){return this.delegate.createText(e)}appendChild(e,A){this.delegate.appendChild(e,A),this.engine.onInsert(this.namespaceId,A,e,!1)}insertBefore(e,A,i,n=!0){this.delegate.insertBefore(e,A,i),this.engine.onInsert(this.namespaceId,A,e,n)}removeChild(e,A,i){this.parentNode(A)&&this.engine.onRemove(this.namespaceId,A,this.delegate)}selectRootElement(e,A){return this.delegate.selectRootElement(e,A)}parentNode(e){return this.delegate.parentNode(e)}nextSibling(e){return this.delegate.nextSibling(e)}setAttribute(e,A,i,n){this.delegate.setAttribute(e,A,i,n)}removeAttribute(e,A,i){this.delegate.removeAttribute(e,A,i)}addClass(e,A){this.delegate.addClass(e,A)}removeClass(e,A){this.delegate.removeClass(e,A)}setStyle(e,A,i,n){this.delegate.setStyle(e,A,i,n)}removeStyle(e,A,i){this.delegate.removeStyle(e,A,i)}setProperty(e,A,i){A.charAt(0)==v7&&A==MgA?this.disableAnimations(e,!!i):this.delegate.setProperty(e,A,i)}setValue(e,A){this.delegate.setValue(e,A)}listen(e,A,i,n){return this.delegate.listen(e,A,i,n)}disableAnimations(e,A){this.engine.disableAnimations(e,A)}},JU=class extends F7{factory;constructor(e,A,i,n,o){super(A,i,n,o),this.factory=e,this.namespaceId=A}setProperty(e,A,i){A.charAt(0)==v7?A.charAt(1)=="."&&A==MgA?(i=i===void 0?!0:!!i,this.disableAnimations(e,i)):this.engine.process(this.namespaceId,e,A.slice(1),i):this.delegate.setProperty(e,A,i)}listen(e,A,i,n){if(A.charAt(0)==v7){let o=ZVA(e),r=A.slice(1),s="";return r.charAt(0)!=v7&&([r,s]=WVA(r)),this.engine.listen(this.namespaceId,o,r,s,a=>{let c=a._data||-1;this.factory.scheduleListenerCallback(c,i,a)})}return this.delegate.listen(e,A,i,n)}};function ZVA(t){switch(t){case"body":return document.body;case"document":return document;case"window":return window;default:return t}}function WVA(t){let e=t.indexOf("."),A=t.substring(0,e),i=t.slice(e+1);return[A,i]}var N7=class{delegate;engine;_zone;_currentId=0;_microtaskId=1;_animationCallbacksBuffer=[];_rendererCache=new Map;_cdRecurDepth=0;constructor(e,A,i){this.delegate=e,this.engine=A,this._zone=i,A.onRemovalComplete=(n,o)=>{o?.removeChild(null,n)}}createRenderer(e,A){let i="",n=this.delegate.createRenderer(e,A);if(!e||!A?.data?.animation){let c=this._rendererCache,l=c.get(n);if(!l){let I=()=>c.delete(n);l=new F7(i,n,this.engine,I),c.set(n,l)}return l}let o=A.id,r=A.id+"-"+this._currentId;this._currentId++,this.engine.register(r,e);let s=c=>{Array.isArray(c)?c.forEach(s):this.engine.registerTrigger(o,r,e,c.name,c)};return A.data.animation.forEach(s),new JU(this,r,n,this.engine)}begin(){this._cdRecurDepth++,this.delegate.begin&&this.delegate.begin()}_scheduleCountTask(){queueMicrotask(()=>{this._microtaskId++})}scheduleListenerCallback(e,A,i){if(e>=0&&e<this._microtaskId){this._zone.run(()=>A(i));return}let n=this._animationCallbacksBuffer;n.length==0&&queueMicrotask(()=>{this._zone.run(()=>{n.forEach(o=>{let[r,s]=o;r(s)}),this._animationCallbacksBuffer=[]})}),n.push([A,i])}end(){this._cdRecurDepth--,this._cdRecurDepth==0&&this._zone.runOutsideAngular(()=>{this._scheduleCountTask(),this.engine.flush(this._microtaskId)}),this.delegate.end&&this.delegate.end()}whenRenderingDone(){return this.engine.whenRenderingDone()}componentReplaced(e){this.engine.flush(),this.delegate.componentReplaced?.(e)}};var $VA=(()=>{class t extends YQ{constructor(A,i,n){super(A,i,n)}ngOnDestroy(){this.flush()}static \u0275fac=function(i){return new(i||t)(we(at),we(ld),we(gd))};static \u0275prov=NA({token:t,factory:t.\u0275fac})}return t})();function AZA(){return new b7}function eZA(t,e,A){return new N7(t,e,A)}var SgA=[{provide:gd,useFactory:AZA},{provide:YQ,useClass:$VA},{provide:ws,useFactory:eZA,deps:[jh,YQ,Qe]}],tZA=[{provide:ld,useClass:TU},{provide:Si,useValue:"NoopAnimations"},...SgA],kgA=[{provide:ld,useFactory:()=>new L7},{provide:Si,useFactory:()=>"BrowserAnimations"},...SgA],_7=(()=>{class t{static withConfig(A){return{ngModule:t,providers:A.disableAnimations?tZA:kgA}}static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:kgA,imports:[Vh]})}return t})();var iZA=new dA("mat-chips-default-options",{providedIn:"root",factory:()=>({separatorKeyCodes:[13]})});var RgA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({providers:[bB,{provide:iZA,useValue:{separatorKeyCodes:[13]}}],imports:[it,Ps,it]})}return t})();var nZA=["input"],oZA=["formField"],rZA=["*"],zU=class{source;value;constructor(e,A){this.source=e,this.value=A}};var sZA=new dA("MatRadioGroup"),aZA=new dA("mat-radio-default-options",{providedIn:"root",factory:cZA});function cZA(){return{color:"accent",disabledInteractive:!1}}var lZA=(()=>{class t{_elementRef=f(ee);_changeDetector=f(It);_focusMonitor=f(dr);_radioDispatcher=f(xB);_defaultOptions=f(aZA,{optional:!0});_ngZone=f(Qe);_renderer=f(Wi);_uniqueId=f(sn).getId("mat-radio-");_cleanupClick;id=this._uniqueId;name;ariaLabel;ariaLabelledby;ariaDescribedby;disableRipple=!1;tabIndex=0;get checked(){return this._checked}set checked(A){this._checked!==A&&(this._checked=A,A&&this.radioGroup&&this.radioGroup.value!==this.value?this.radioGroup.selected=this:!A&&this.radioGroup&&this.radioGroup.value===this.value&&(this.radioGroup.selected=null),A&&this._radioDispatcher.notify(this.id,this.name),this._changeDetector.markForCheck())}get value(){return this._value}set value(A){this._value!==A&&(this._value=A,this.radioGroup!==null&&(this.checked||(this.checked=this.radioGroup.value===A),this.checked&&(this.radioGroup.selected=this)))}get labelPosition(){return this._labelPosition||this.radioGroup&&this.radioGroup.labelPosition||"after"}set labelPosition(A){this._labelPosition=A}_labelPosition;get disabled(){return this._disabled||this.radioGroup!==null&&this.radioGroup.disabled}set disabled(A){this._setDisabled(A)}get required(){return this._required||this.radioGroup&&this.radioGroup.required}set required(A){this._required=A}get color(){return this._color||this.radioGroup&&this.radioGroup.color||this._defaultOptions&&this._defaultOptions.color||"accent"}set color(A){this._color=A}_color;get disabledInteractive(){return this._disabledInteractive||this.radioGroup!==null&&this.radioGroup.disabledInteractive}set disabledInteractive(A){this._disabledInteractive=A}_disabledInteractive;change=new WA;radioGroup;get inputId(){return`${this.id||this._uniqueId}-input`}_checked=!1;_disabled;_required;_value=null;_removeUniqueSelectionListener=()=>{};_previousTabIndex;_inputElement;_rippleTrigger;_noopAnimations;_injector=f(Rt);constructor(){f(Rn).load(lr);let A=f(sZA,{optional:!0}),i=f(Si,{optional:!0}),n=f(new wr("tabindex"),{optional:!0});this.radioGroup=A,this._noopAnimations=i==="NoopAnimations",this._disabledInteractive=this._defaultOptions?.disabledInteractive??!1,n&&(this.tabIndex=zi(n,0))}focus(A,i){i?this._focusMonitor.focusVia(this._inputElement,i,A):this._inputElement.nativeElement.focus(A)}_markForCheck(){this._changeDetector.markForCheck()}ngOnInit(){this.radioGroup&&(this.checked=this.radioGroup.value===this._value,this.checked&&(this.radioGroup.selected=this),this.name=this.radioGroup.name),this._removeUniqueSelectionListener=this._radioDispatcher.listen((A,i)=>{A!==this.id&&i===this.name&&(this.checked=!1)})}ngDoCheck(){this._updateTabIndex()}ngAfterViewInit(){this._updateTabIndex(),this._focusMonitor.monitor(this._elementRef,!0).subscribe(A=>{!A&&this.radioGroup&&this.radioGroup._touch()}),this._ngZone.runOutsideAngular(()=>{this._cleanupClick=this._renderer.listen(this._inputElement.nativeElement,"click",this._onInputClick)})}ngOnDestroy(){this._cleanupClick?.(),this._focusMonitor.stopMonitoring(this._elementRef),this._removeUniqueSelectionListener()}_emitChangeEvent(){this.change.emit(new zU(this,this._value))}_isRippleDisabled(){return this.disableRipple||this.disabled}_onInputInteraction(A){if(A.stopPropagation(),!this.checked&&!this.disabled){let i=this.radioGroup&&this.value!==this.radioGroup.value;this.checked=!0,this._emitChangeEvent(),this.radioGroup&&(this.radioGroup._controlValueAccessorChangeFn(this.value),i&&this.radioGroup._emitChangeEvent())}}_onTouchTargetClick(A){this._onInputInteraction(A),(!this.disabled||this.disabledInteractive)&&this._inputElement?.nativeElement.focus()}_setDisabled(A){this._disabled!==A&&(this._disabled=A,this._changeDetector.markForCheck())}_onInputClick=A=>{this.disabled&&this.disabledInteractive&&A.preventDefault()};_updateTabIndex(){let A=this.radioGroup,i;if(!A||!A.selected||this.disabled?i=this.tabIndex:i=A.selected===this?this.tabIndex:-1,i!==this._previousTabIndex){let n=this._inputElement?.nativeElement;n&&(n.setAttribute("tabindex",i+""),this._previousTabIndex=i,To(()=>{queueMicrotask(()=>{A&&A.selected&&A.selected!==this&&document.activeElement===n&&(A.selected?._inputElement.nativeElement.focus(),document.activeElement===n&&this._inputElement.nativeElement.blur())})},{injector:this._injector}))}}static \u0275fac=function(i){return new(i||t)};static \u0275cmp=zA({type:t,selectors:[["mat-radio-button"]],viewQuery:function(i,n){if(i&1&&(Te(nZA,5),Te(oZA,7,ee)),i&2){let o;XA(o=$A())&&(n._inputElement=o.first),XA(o=$A())&&(n._rippleTrigger=o.first)}},hostAttrs:[1,"mat-mdc-radio-button"],hostVars:19,hostBindings:function(i,n){i&1&&hA("focus",function(){return n._inputElement.nativeElement.focus()}),i&2&&(Ne("id",n.id)("tabindex",null)("aria-label",null)("aria-labelledby",null)("aria-describedby",null),ue("mat-primary",n.color==="primary")("mat-accent",n.color==="accent")("mat-warn",n.color==="warn")("mat-mdc-radio-checked",n.checked)("mat-mdc-radio-disabled",n.disabled)("mat-mdc-radio-disabled-interactive",n.disabledInteractive)("_mat-animation-noopable",n._noopAnimations))},inputs:{id:"id",name:"name",ariaLabel:[0,"aria-label","ariaLabel"],ariaLabelledby:[0,"aria-labelledby","ariaLabelledby"],ariaDescribedby:[0,"aria-describedby","ariaDescribedby"],disableRipple:[2,"disableRipple","disableRipple",ie],tabIndex:[2,"tabIndex","tabIndex",A=>A==null?0:zi(A)],checked:[2,"checked","checked",ie],value:"value",labelPosition:"labelPosition",disabled:[2,"disabled","disabled",ie],required:[2,"required","required",ie],color:"color",disabledInteractive:[2,"disabledInteractive","disabledInteractive",ie]},outputs:{change:"change"},exportAs:["matRadioButton"],ngContentSelectors:rZA,decls:13,vars:17,consts:[["formField",""],["input",""],["mat-internal-form-field","",3,"labelPosition"],[1,"mdc-radio"],[1,"mat-mdc-radio-touch-target",3,"click"],["type","radio",1,"mdc-radio__native-control",3,"change","id","checked","disabled","required"],[1,"mdc-radio__background"],[1,"mdc-radio__outer-circle"],[1,"mdc-radio__inner-circle"],["mat-ripple","",1,"mat-radio-ripple","mat-focus-indicator",3,"matRippleTrigger","matRippleDisabled","matRippleCentered"],[1,"mat-ripple-element","mat-radio-persistent-ripple"],[1,"mdc-label",3,"for"]],template:function(i,n){if(i&1){let o=be();jt(),S(0,"div",2,0)(2,"div",3)(3,"div",4),hA("click",function(s){return RA(o),xA(n._onTouchTargetClick(s))}),F(),S(4,"input",5,1),hA("change",function(s){return RA(o),xA(n._onInputInteraction(s))}),F(),S(6,"div",6),JA(7,"div",7)(8,"div",8),F(),S(9,"div",9),JA(10,"div",10),F()(),S(11,"label",11),Fe(12),F()()}i&2&&(yA("labelPosition",n.labelPosition),G(2),ue("mdc-radio--disabled",n.disabled),G(2),yA("id",n.inputId)("checked",n.checked)("disabled",n.disabled&&!n.disabledInteractive)("required",n.required),Ne("name",n.name)("value",n.value)("aria-label",n.ariaLabel)("aria-labelledby",n.ariaLabelledby)("aria-describedby",n.ariaDescribedby)("aria-disabled",n.disabled&&n.disabledInteractive?"true":null),G(5),yA("matRippleTrigger",n._rippleTrigger.nativeElement)("matRippleDisabled",n._isRippleDisabled())("matRippleCentered",!0),G(2),yA("for",n.inputId))},dependencies:[rs,MB],styles:['.mat-mdc-radio-button{-webkit-tap-highlight-color:rgba(0,0,0,0)}.mat-mdc-radio-button .mdc-radio{display:inline-block;position:relative;flex:0 0 auto;box-sizing:content-box;width:20px;height:20px;cursor:pointer;will-change:opacity,transform,border-color,color;padding:calc((var(--mdc-radio-state-layer-size, 40px) - 20px)/2)}.mat-mdc-radio-button .mdc-radio:hover .mdc-radio__native-control:not([disabled]):not(:focus)~.mdc-radio__background::before{opacity:.04;transform:scale(1)}.mat-mdc-radio-button .mdc-radio:hover .mdc-radio__native-control:not([disabled])~.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-unselected-hover-icon-color, var(--mat-sys-on-surface))}.mat-mdc-radio-button .mdc-radio:hover .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button .mdc-radio:hover .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__inner-circle{border-color:var(--mdc-radio-selected-hover-icon-color, var(--mat-sys-primary))}.mat-mdc-radio-button .mdc-radio:active .mdc-radio__native-control:enabled:not(:checked)+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-unselected-pressed-icon-color, var(--mat-sys-on-surface))}.mat-mdc-radio-button .mdc-radio:active .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button .mdc-radio:active .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__inner-circle{border-color:var(--mdc-radio-selected-pressed-icon-color, var(--mat-sys-primary))}.mat-mdc-radio-button .mdc-radio__background{display:inline-block;position:relative;box-sizing:border-box;width:20px;height:20px}.mat-mdc-radio-button .mdc-radio__background::before{position:absolute;transform:scale(0, 0);border-radius:50%;opacity:0;pointer-events:none;content:"";transition:opacity 90ms cubic-bezier(0.4, 0, 0.6, 1),transform 90ms cubic-bezier(0.4, 0, 0.6, 1);width:var(--mdc-radio-state-layer-size, 40px);height:var(--mdc-radio-state-layer-size, 40px);top:calc(-1*(var(--mdc-radio-state-layer-size, 40px) - 20px)/2);left:calc(-1*(var(--mdc-radio-state-layer-size, 40px) - 20px)/2)}.mat-mdc-radio-button .mdc-radio__outer-circle{position:absolute;top:0;left:0;box-sizing:border-box;width:100%;height:100%;border-width:2px;border-style:solid;border-radius:50%;transition:border-color 90ms cubic-bezier(0.4, 0, 0.6, 1)}.mat-mdc-radio-button .mdc-radio__inner-circle{position:absolute;top:0;left:0;box-sizing:border-box;width:100%;height:100%;transform:scale(0, 0);border-width:10px;border-style:solid;border-radius:50%;transition:transform 90ms cubic-bezier(0.4, 0, 0.6, 1),border-color 90ms cubic-bezier(0.4, 0, 0.6, 1)}.mat-mdc-radio-button .mdc-radio__native-control{position:absolute;margin:0;padding:0;opacity:0;top:0;right:0;left:0;cursor:inherit;z-index:1;width:var(--mdc-radio-state-layer-size, 40px);height:var(--mdc-radio-state-layer-size, 40px)}.mat-mdc-radio-button .mdc-radio__native-control:checked+.mdc-radio__background,.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background{transition:opacity 90ms cubic-bezier(0, 0, 0.2, 1),transform 90ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-radio-button .mdc-radio__native-control:checked+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background .mdc-radio__outer-circle{transition:border-color 90ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-radio-button .mdc-radio__native-control:checked+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background .mdc-radio__inner-circle{transition:transform 90ms cubic-bezier(0, 0, 0.2, 1),border-color 90ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-radio-button .mdc-radio__native-control:focus+.mdc-radio__background::before{transform:scale(1);opacity:.12;transition:opacity 90ms cubic-bezier(0, 0, 0.2, 1),transform 90ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-radio-button .mdc-radio__native-control:disabled:not(:checked)+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-disabled-unselected-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-radio-disabled-unselected-icon-opacity, 0.38)}.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background{cursor:default}.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button .mdc-radio__native-control:disabled+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-disabled-selected-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-radio-disabled-selected-icon-opacity, 0.38)}.mat-mdc-radio-button .mdc-radio__native-control:enabled:not(:checked)+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-unselected-icon-color, var(--mat-sys-on-surface-variant))}.mat-mdc-radio-button .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button .mdc-radio__native-control:enabled:checked+.mdc-radio__background .mdc-radio__inner-circle{border-color:var(--mdc-radio-selected-icon-color, var(--mat-sys-primary))}.mat-mdc-radio-button .mdc-radio__native-control:enabled:focus:checked+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button .mdc-radio__native-control:enabled:focus:checked+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-selected-focus-icon-color, var(--mat-sys-primary))}.mat-mdc-radio-button .mdc-radio__native-control:checked+.mdc-radio__background .mdc-radio__inner-circle{transform:scale(0.5);transition:transform 90ms cubic-bezier(0, 0, 0.2, 1),border-color 90ms cubic-bezier(0, 0, 0.2, 1)}.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled{pointer-events:auto}.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__native-control:not(:checked)+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-disabled-unselected-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-radio-disabled-unselected-icon-opacity, 0.38)}.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled:hover .mdc-radio__native-control:checked+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled:hover .mdc-radio__native-control:checked+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__native-control:checked:focus+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__native-control:checked:focus+.mdc-radio__background .mdc-radio__outer-circle,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__native-control+.mdc-radio__background .mdc-radio__inner-circle,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__native-control+.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-disabled-selected-icon-color, var(--mat-sys-on-surface));opacity:var(--mdc-radio-disabled-selected-icon-opacity, 0.38)}.mat-mdc-radio-button._mat-animation-noopable .mdc-radio__background::before,.mat-mdc-radio-button._mat-animation-noopable .mdc-radio__outer-circle,.mat-mdc-radio-button._mat-animation-noopable .mdc-radio__inner-circle{transition:none !important}.mat-mdc-radio-button .mdc-radio__background::before{background-color:var(--mat-radio-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-radio-button.mat-mdc-radio-checked .mat-ripple-element,.mat-mdc-radio-button.mat-mdc-radio-checked .mdc-radio__background::before{background-color:var(--mat-radio-checked-ripple-color, var(--mat-sys-primary))}.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mat-ripple-element,.mat-mdc-radio-button.mat-mdc-radio-disabled-interactive .mdc-radio--disabled .mdc-radio__background::before{background-color:var(--mat-radio-ripple-color, var(--mat-sys-on-surface))}.mat-mdc-radio-button .mat-internal-form-field{color:var(--mat-radio-label-text-color, var(--mat-sys-on-surface));font-family:var(--mat-radio-label-text-font, var(--mat-sys-body-medium-font));line-height:var(--mat-radio-label-text-line-height, var(--mat-sys-body-medium-line-height));font-size:var(--mat-radio-label-text-size, var(--mat-sys-body-medium-size));letter-spacing:var(--mat-radio-label-text-tracking, var(--mat-sys-body-medium-tracking));font-weight:var(--mat-radio-label-text-weight, var(--mat-sys-body-medium-weight))}.mat-mdc-radio-button .mdc-radio--disabled+label{color:var(--mat-radio-disabled-label-color, color-mix(in srgb, var(--mat-sys-on-surface) 38%, transparent))}.mat-mdc-radio-button .mat-radio-ripple{top:0;left:0;right:0;bottom:0;position:absolute;pointer-events:none;border-radius:50%}.mat-mdc-radio-button .mat-radio-ripple .mat-ripple-element{opacity:.14}.mat-mdc-radio-button .mat-radio-ripple::before{border-radius:50%}.mat-mdc-radio-button .mdc-radio .mdc-radio__native-control:focus:enabled:not(:checked)~.mdc-radio__background .mdc-radio__outer-circle{border-color:var(--mdc-radio-unselected-focus-icon-color, var(--mat-sys-on-surface))}.mat-mdc-radio-button.cdk-focused .mat-focus-indicator::before{content:""}.mat-mdc-radio-disabled{cursor:default;pointer-events:none}.mat-mdc-radio-disabled.mat-mdc-radio-disabled-interactive{pointer-events:auto}.mat-mdc-radio-touch-target{position:absolute;top:50%;left:50%;height:48px;width:48px;transform:translate(-50%, -50%);display:var(--mat-radio-touch-target-display, block)}[dir=rtl] .mat-mdc-radio-touch-target{left:auto;right:50%;transform:translate(50%, -50%)}'],encapsulation:2,changeDetection:0})}return t})(),xgA=(()=>{class t{static \u0275fac=function(i){return new(i||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[it,Ps,lZA,it]})}return t})();var G7=class t{static \u0275fac=function(A){return new(A||t)};static \u0275mod=Ce({type:t});static \u0275inj=Ie({imports:[f0,B6,LcA,pq,y0,e7,AC,_P,bcA,FcA,GcA,zcA,xgA,u8,hcA,HaA,scA,SlA,f8,qcA,_7,mcA,BlA.forRoot(),ElA,xz,RgA,Bq,ScA,flA]})};var Tf=class t{static \u0275fac=function(A){return new(A||t)};static \u0275mod=Ce({type:t,bootstrap:[_Q]});static \u0275inj=Ie({providers:[Wg,H2,Xg,xQ,LQ,nI,Vc,RQ,O2,$g],imports:[G7,Vh,B6,IM,B7,e7,y0,AC,_7]})};fetch("./assets/config/runtime-config.json").then(t=>t.json()).then(t=>{window.runtimeConfig=t,Wp().bootstrapModule(Tf).catch(e=>console.error(e))});Wp().bootstrapModule(Tf).catch(t=>console.error(t));
